import numpy as np
from scipy.interpolate import interp1d



def SRD(IFO, f): 
    import numpy as np
    from scipy.interpolate import interp1d

    #wswfunction [S, fstop] = SRD(IFO,f)

    # % SRD - Design sensitivity curve for the specified GW detector.
    # %
    # %  [S, fstop] = SRD(DET,f)
    # %
    # %  DET    String designating desired detector. Must be one of 'aLIGO',
    # %         'aLIGO-*' (see below), aVirgo, aVirgo-*' (see below), 'ETB', 'ETC', 
    # %         'ETD', 'GEO', 'GEO-HF', 'KAGRA*' (see below), 'LIGO', 'LIGO-LV',
    # %         'LHO', 'LLO', 'LISA', 'LISA-LIGHT1', 'LISA-LIGHT2', 'LISA-WD', 'TAMA',
    # %         'Virgo', 'Virgo-LV', 'LIGO3-*' (see below), 'LIGO-A', 'LIGO-Voyager',
    # %         'LIGO-CosmicExplorer', or 'white* (see below)'.  
    # %         Not case sensitive.  
    # %  f      Vector of positive frequencies (Hz) at which design sensitivity 
    # %         is desired. 
    # %
    # %  S      One-sided power spectral density (1/Hz) for requested detector
    # %         and frequencies.  S has the same size as f.
    # %  fstop  Minimum and maximum frequency (Hz) at which fit is "guaranteed"
    # %         to be accurate.  A warning message is printed if min(f)<fstop(1)
    # %         or max(f)>fstop(2). 
    # %
    # % For full details of the source of each noise spectrum, see the code.
    # %
    # % Several different types of aLIGO spectra are available:
    # %
    # %  aLIGO        Defaults to aLIGO-ZDHP.
    # %  aLIGO-BBH    Optimized for 30-30 solar mass black hole binary inspirals.
    # %               "This is slightly more difficult from a control system 
    # %               point of view." 
    # %  aLIGO-BNS    Optimized for 1.4 solar mass neutron star binary. This is
    # %               slightly more difficult from a control system point of
    # %               view. Gives ~200 Mpc reach for a single interferometer, SNR
    # %               8, averaged over directions and polarizations.    
    # %  aLIGOEARLY-50    Possible early configuration, c. 2015, with excess
    # %                   low-frequency noise and ~50 Mpc BNS range.
    # %  aLIGOEARLY-100   Possible early configuration, c. 2016-17, with excess
    # %                   low-frequency noise and ~100 Mpc BNS range.
    # %  aLIGOEARLY-150   Possible early configuration, c. 2017-18, with ~150 Mpc
    # %                   BNS range. 
    # %  aLIGO-HF     High Freq: an example of a narrowband tuning at 1kHz.
    # %               Unlike all the other tunings, it needs a recycling mirror
    # %               with a higher reflectivity.   
    # %  aLIGO-NOSRM	No SRM: a possible early configuration, where no signal
    # %               recycling mirror is used.  
    # %  aLIGO-ZDHP	Zero Det, High Power: zero-detuning of the signal
    # %               recycling mirror, with high laser power. "This gives most
    # %               of the sensitivity with the fewest technical difficulties." 
    # %               This spectrum is only available to a maximum frequency of
    # %               5000 Hz.
    # %  aLIGO-ZDLP   Zero Det, Low Power: a possible early configuration with
    # %               zero-detuning of the signal recycling mirror, with low
    # %               laser power. 
    # %  aLIGO-Hooper Analytic approximation from Hooper et al., 1108.3186 / Phys
    # %               Rev D.86.024012 (SPIIR paper). WARNING: This does not match
    # %               any of the official aLIGO spectra.
    # %
    # % The available aVirgo spectra are:
    # %
    # %  aVIRGO       Defaults to aVIRGO-ZDHP.
    # %  aVIRGO-BNS   Optimized for 1.4 solar mass neutron star binary (as of May 2010). 
    # %  aVIRGO-NOSRM	No SRM: a possible early configuration, where no signal
    # %               recycling mirror is used.  
    # %  aVIRGO-ZDHP  Optimized wide-band sensitivity. 
    # %  aVIRGO-ZDLP  Zero Det, Low Power: a possible early configuration with
    # %               zero-detuning of the signal recycling mirror, with low
    # %               laser power. 
    # %  
    # % The available KAGRA spectra are:
    # %
    # %  KAGRA_Early  Anticipated sensitivity c.2019-20.
    # %  KAGRA_Mid    Anticipated sensitivity c.2020-21.
    # %  KAGRA_Late   Anticipated sensitivity c.2021-22.
    # %  KAGRA        Design sensitivity c.2022+.
    # %
    # % The available LIGO3 spectra are:
    # %
    # %  LIGO3-default Standard 3G: 100kg masses, 10dB squeezing, 2.5x thermal noise reduction
    # %  LIGO3-mid     mid-freq  3G: 100kg masses, 10dB squeezing, 5x thermal noise reduction
    # %  LIGO3-high    high-freq  3G: 100kg masses, 15dB squeezing, 2.5x thermal noise reduction
    # %  LIGO3         Defaults to LIGO3-default
    # %
    # % The 'LIGO-A' (A+), 'LIGO-Voyager', and 'LIGO-CosmicExplorer' options yield the 
    # % strawman sensitivity curves described in T1400316-v5. The implementation here
    # % is based on interpolation of sampled points extracted from Figure 3 of
    # % T1400316-v5 and should be accurate to a few percent in amplitude. These
    # % spectra are available from 10Hz to 2000Hz only.
    # %
    # % The choice IFO = 'whiteFS' where FS is a number returns a constant S =
    # % 2/FS Hz^{-1}, corresponding to unit-variance discrete white noise.  The
    # % choice 'white' uses FS='1'.  This may be handy for testing codes which
    # % use noise weighting.
    # %
    # % For information on where each spectrum is taken from, please read the
    # % relevant portion of the code.
    # %
    # % $Id: SRD.m 5600 2018-04-06 09:03:21Z patrick.sutton@LIGO.ORG $ 

    #% ---- Argument checks.
    ####narginchk(2,2);
    
    if type(IFO)!=str:
        raise TypeError('Detector argument must be a string.')

    if type(f).__module__!=np.__name__:
        raise TypeError('Input frequencies must be a numpy array')

    if np.size(f)!=len(f):
        raise TypeError('Input frequencies must be a vector.')

    if min(f)<=0:
        raise ValueError('Input frequencies must be positive.')

    #% ---- Recognize detector and compute noise spectrum.

    #if upper(IFO)==
            #  case 'ADVLIGO'
            # % ---- Deprecated.  The current aLIGO curves do not include a 
            # %      spectrum that looks like this one.  Select one of the 
            # %      aLIGO-* curves instead.
            # fstop = [0,Inf];
            # f0 = 215;
            # x = f/f0;
            # S = (1e-49)*( x.^(-4.14) -5*x.^(-2) + 111*(1-x.^2+x.^4/2)./(1+x.^2/2) );
    if IFO.upper()=='ALIGO':
        [S, fstop] = aligospectrum('aLIGO-ZDHP',f)
                # % ---- Analytic approximation to ALIGO-ZDHP spectrum data, by P Ajith,
                # %      https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=63161 
                # %      The fit is accurate to withn 2% everywhere above 20 Hz band except
                # %      that it does not include lines.  
                # % x = f/245.4;
                # % S = 1e-48 * (0.0152*x.^(-4) + 0.2935*x.^(9/4) + 2.7951*x.^(3/2) - 6.5080*x.^(3/4) + 17.7622); 
                # % fstop = [20 1e4];
                # % ---- Analytic approximation to the ALIGO-ZDHP spectra, taken
                # %      from Del Pozzo by email, 2012.06.03.  Based on 1108.1317,
                # %      but this paper contains an error.  Corrected formula below
                # %      is accurate to within 10% in amplitude everywhere in the
                # %      range [11,3539] Hz.
                # % S = 1.35e-50 * f .* ( 60000*(f/10).^(-30) + 5*(f/50).^(-6) ...
                # %     + 1.07*(f/100).^(-3.25) + 3.7*(f/200).^(-1.25) ...
                # %     + 0.9*(f/300).^(-0.08) + 0.85*(f/1000).^(0.8) ...
                # %     + 0.53*(f/2000).^(3) );
                # % fstop = [11,3539];
    elif IFO.upper()=='AVIRGO':
        [S, fstop] = avirgospectrum('aVIRGO-ZDHP',f)
                # % case 'ET'
                # %     % ---- Deprecated pre-design version of the ET spectrum, by Leone Bosi. 
                # %     fstop = [1,Inf]; 
                # %     f0 = 200;
                # %     x = f/f0;
                # %     %
                # %     p1 = -4.05049;
                # %     p2 = -0.687263;
                # %     a0 = 185.623;
                # %     b0 = 232.565;
                # %     b1 = 31.1777;
                # %     b2 = -64.7178;
                # %     b3 = 52.242;
                # %     b4 = -42.1643;
                # %     b5 = 10.1708;
                # %     b6 = 11.5317;
                # %     c1 = 13.5793;
                # %     c2 = -36.4586;
                # %     c3 = 18.5625;
                # %     c4 = 27.4297;
                # %     S0 = 1.449936*10^-52;
                # %     %
                # %     S  = S0*(x.^p1 + a0.*x.^p2 + b0 * ...
                # %         ( 1. + b1*x + b2*x.^2 + b3*x.^3 + b4*x.^4 + b5*x.^5 + b6*x.^6 ) ...
                # %         ./ (1. + c1*x + c2*x.^2 + c3*x.^3 + c4*x.^4) );
    elif IFO.upper()=='ETB':
        [S, fstop] = etb(f)
                # % ---- This following analytic approximation is good to within 10%
                # %      across the [1,10000] Hz band.
                # % [S, fstop] = ETB_fit();
                # % S = S.^2;
                # % ---- Read sampled spectrum supplied by B. S. Sathyaprakash.

    elif IFO.upper()=='ETC':
        [S, fstop] = ETC_fit(f)
        S = S**2
    elif IFO.upper()=='ETD':
        [S, fstop] = etd(f)
                #% ---- Read sampled spectrum supplied by B. S. Sathyaprakash.
    elif IFO.upper()=='GEO':
        fstop = [40,float('inf')]
        f0 = 150
        x = f/f0
        S = (1e-46)*( (3.4*x)**(-30) + 34*x**(-1) + 20*(1-x**2+x**4/2)/(1+x**2/2) )
                #% ---- From L. P. Grishchuk, V. M. Lipunov, K. A. Postnov, M. E.
                #%      Prokhorov, and B. S. Sathyaprakash, Physics-Uspekhi 71, 3
                #%      (2000). 
    elif IFO.upper()=='GEO-HF':
        fstop = [40,float('inf')];
        S = 7.18e-46*(1+(f/1059)**2) + 4.90e-41/f**2 + 8.91e-43/f + 1.6e-17/f**16;
                #% ---- From LIGO-P0900199, analytical fit with sub percent precision
                #%      across the 50-8000 Hz band (GEO-HF logbook p.60)
                #%      https://intranet.aei.uni-hannover.de/geo600/geohflogbook.nsf/7e8722dffa24dea0c1256de900406c84/4837a612ac990060c12575ce004e70fd?OpenDocument
                #%      fstop not well defined, there is not design for the control
                #%      noise which currently dominates the spectrum below 100Hz

    elif IFO.upper() in ['KAGRA','KAGRA_EARLY','KAGRA_MID','KAGRA_LATE']:
        [S, fstop] = kagraspectrum(IFO,f)
    elif IFO.upper() in ['LIGO3', 'LIGO3-DEFAULT', 'LIGO3-MID', 'LIGO3-HIGH']:
        [S, fstop] = ligo3spectrumold(IFO,f)
    elif IFO.upper() in ['LIGO-A', 'LIGO-VOYAGER', 'LIGO-COSMICEXPLORER']:
        [S, fstop] = ligo3spectrum(IFO,f)
    elif IFO.upper() in ['LIGO', 'LHO', 'LLO']:
        fstop = [40,float('inf')]
        f0 = 150
        x = f/f0
        S = 9*(1e-46)*( (4.49*x)**(-56) + 0.16*x**(-4.52) + 0.52 + 0.32*x**2 )
                #% ---- From L. P. Grishchuk, V. M. Lipunov, K. A. Postnov, M. E.
                #%      Prokhorov, and B. S. Sathyaprakash, Physics-Uspekhi 71, 3
                #%      (2000). 
    elif IFO.upper()=='LIGO-LV':
        fstop = [40,float('inf')]
        seismicReference = 3.3e-21 * (30/f)**14
        thermalReference = 3.8e-23 * (100/f)**2
        shotReference = 1.13e-23 * np.sqrt(1 + (f / 90)**2)
        S = shotReference**2 + thermalReference**2 + seismicReference**2
                #%----- Design used by Chatterji in generating simulated data for 
                #%      the LIGO-Virgo project Ia,b.  This design does not include
                #%      the lines that were added to the simulated data.  fstop is 
                #%      guessed.
    elif IFO.upper()=='LISA':
                #% % ---- From L. P. Grishchuk, V. M. Lipunov, K. A. Postnov, M. E.
                #% %      Prokhorov, and B. S. Sathyaprakash, Physics-Uspekhi 71, 3
                #% %      (2000).  
                #% fstop = [1e-5,Inf];
                #% f0 = 1e-3;
                #% x = f/f0;
                #% S = 420*(1e-46)*( (x/5.62).^(-14/3) + 1e3 + x.^2 );
                #% ---- Sky-averaged instrumental noise for LISA, from 
                #%      1001.3099v2.  Code by Ioannis Kamaretsos.
        L = 5*10**9        #%-- arm length of LISA [m]
        Sp = 4*10**-22      #%-- noise scale constant [m^2 Hz^-1]
        Sa = 9*10**-30      #%-- noise scale constant [m^2 Hz^-1]
        c = 299792458      #%-- speed of light [m/s]
        fo = c/(2*pi*L)    #%-- LISA arm transfer frequency [Hz]
            #% ---- Instrumental noise.
        S = (L**-2)*((1+0.5*(f**2/fo**2))*Sp + (1+10**-8*(f**-2))*(4*Sa*((2*pi)**-4*f**-4)))
        fstop = [1e-5,1e0]  #%-- made up

    elif IFO.upper()=='LISA-LIGHT1':
                #% ---- CLISA1_P005_LPF configuration from
                #%      https://lisa-light.aei.mpg.de/lisa-light/pub/DetectorConfigurations/NewConfigurations/NewLISAConfig_v1.pdf 
                #%      This does not include the WD-WD confusion background.
        c = 299792458     #%-- speed of light [m/s]
                #% ---- Acceleration noise [m^2/Hz].
        Sacc = (28.99e-30*(1 + 1.8e-4/f)**2)*(2*pi*f)**-4
                #% ---- Shot noise [m^2/Hz].
        Ssn = 2.25e-21
                #% ---- "Other measurement noise" [m^2/Hz].
        Somn = 6.4e-23
                #% ---- Arm length [m].
        L = 1e9
            #% ---- "Very approximate" sensitivity.
        S = 20/3 * ( 1 + (2*L*f/(0.41*c))**2 ) * (Sacc + Ssn + Somn) / L**2        
        fstop = [1e-5,1e0]  #%-- made up
    elif IFO.upper()=='LISA-LIGHT2':
                #% ---- 10LISA1_P2_DRS configuration from
                #%      https://lisa-light.aei.mpg.de/lisa-light/pub/DetectorConfigurations/NewConfigurations/NewLISAConfig_v1.pdf 
                #%      This does not include the WD-WD confusion background.
        c = 299792458     #%-- speed of light [m/s]
                #% ---- Acceleration noise [m^2/Hz].
        Sacc = 21.30e-30/(2*pi*f)**4
                #% ---- Shot noise [m^2/Hz].
        Ssn = 1.4e-24
                #% ---- "Other measurement noise" [m^2/Hz].
        Somn = 6.4e-23
                #% ---- Arm length [m].
        L = 1e9
                #% ---- "Very approximate" sensitivity.
        S = 20/3 * ( 1 + (2*L*f/(0.41*c))**2 ) * (Sacc + Ssn + Somn) / L**2        
                #% S = 20/3 * ( 1 + (2*L*f/(0.41*c)).^2 ) .* (4*Sacc + Ssn + Somn) / L^2;        
        fstop = [1e-5,1e0]  #%-- made up
    # elif IFO.upper()=='LISA-WD':
    #                 #% ---- Sky-averaged instrumental noise for LISA, including WD-WD
    #                 #%      confusion background.  From 1001.3099v2.  Code by Ioannis
    #                 #%      Kamaretsos. 
    #     L = 5*10**9        #%-- arm length of LISA [m]
    #     Sp = 4*10**-22     #%-- noise scale constant [m^2 Hz^-1]
    #     Sa = 9*10**-30     #%-- noise scale constant [m^2 Hz^3]
    #     c = 299792458      #%-- speed of light [m/s]
    #     fo = c/(2*pi*L)    #%-- LISA arm transfer frequency [Hz]
    #             #% ---- Instrumental noise.
    #     S = (L**-2)*((1+0.5*(f**2/fo**2))*Sp + (1+10**-8*(f**-2))*(4*Sa*((2*pi)**-4*f**-4)));
    #             #% ---- WD-WD confusion noise.  Note typo in f<1e-3 term has been
    #             #%      corrected: f^(-2/3) -> f^(-2.3).
    #     S_conf = zeros(size(f));
    #     ind = find(f <= 1e-3);
    #     S_conf(ind) = (10^(-44.62))*f(ind)**(-2.3);

    #     ind = find(f > 1e-3 & f <= 10^(-2.7));
    #     S_conf(ind) = (10^(-50.92))*f(ind)**(-4.4);

    #     ind = find(f > 10^(-2.7) & f <= 10^(-2.4));
    #     S_conf(ind) = (10^(-62.8))*f(ind).^(-8.8);

    #     ind = find(f > 10^(-2.4) & f <= 1e-2);
    #     S_conf(ind) = (10^(-89.68))*f(ind).^(-20);

    #     S = S + S_conf;
    #         fstop = [1e-5,1e0];  %-- made up
    elif IFO.upper()=='TAMA':
            #% ---- From L. P. Grishchuk, V. M. Lipunov, K. A. Postnov, M. E.
            #%      Prokhorov, and B. S. Sathyaprakash, Physics-Uspekhi 71, 3
            #%      (2000).  
        fstop = [75,float('inf')]
        f0 = 400
        x = f/f0
        S = 75*(1e-46)*( x**(-5) + 13*x**(-1) + 9 + 9*x**2 )
    elif IFO.upper()=='VIRGO':
            #% ---- From L. P. Grishchuk, V. M. Lipunov, K. A. Postnov, M. E.
            #%      Prokhorov, and B. S. Sathyaprakash, Physics-Uspekhi 71, 3
            #%      (2000).  
        fstop = [20,float('inf')];
        f0 = 500
        x = f/f0
        S = 3.24*(1e-46)*( (6.23*x)**(-5) + 2*x**(-1) + 1 + x**2 )
    elif IFO.upper()=='VIRGO-LV':
            #%----- Design used in generating simulated data for the LIGO-Virgo 
            #%      project Ia,b.  This design does not include the lines that
            #%      were added to the simulated data.  Note that f0 is reset
            #%      from 500 in 'VIRGO' to 300 in 'VIRGO-LV'. 
        fstop = [20,float('inf')]
        f0 = 500*0.6
        x = f/f0
        S = 3.24*(1e-46)*( (6.23*x)**(-5) + 2*x**(-1) + 2 + 1.2*x**2 )
    elif IFO.upper()=='WHITE':
        fstop = [f[0],f[-1]]
        S = 2 * np.ones(len(f))
    #     otherwise
    #         % ---- Check for one of the various aLIGO or aVirgo configurations,
    #         %      or white noise. 
    #         if (length(IFO)==12) && (strcmpi(IFO,'ALIGO-HOOPER'))
    #             % ---- Analytic approximation from Hooper et al., 1108.3186 / Phys
    #             %      Rev D.86.024012 (SPIIR paper). WARNING: This does not match
    #             %      any of the official aLIGO spectra.
    #             x = f/215;
    #             S = 1e-49 * (x.^(-4.14) - 5*x.^(-2) + 111*(1 - x.^2 + 0.5*x.^4)./(1+0.5*x.^2)); 
    #             fstop = [9 Inf];        
    #         elseif (length(IFO)>6) && (strcmpi(IFO(1:6),'ALIGO-'))
    #             [S, fstop] = aligospectrum(IFO,f);
    #         elseif (length(IFO)>11) && (strcmpi(IFO(1:11),'ALIGOEARLY-'))
    #             [S, fstop] = aligoearlyspectrum(IFO,f);
    #         elseif (length(IFO)>7) && (strcmpi(IFO(1:7),'AVIRGO-'))
    #             [S, fstop] = avirgospectrum(IFO,f);
    #         elseif (length(IFO)>=5) && (strcmpi(IFO(1:5),'WHITE'))
    #             fs = str2num(IFO(6:end));
    #             fstop = [f(1),f(end)];
    #             S = 2 * ones(size(f)) / fs;
    #         else
    #             error(['Detector ' IFO ' not recognized.']);
    #         end


    #% ---- Display warning message if asking for spectrum at frequencies below
    #%      or above the recommended cut-offs.
    if min(f)<fstop[0]:
        pass
        #print('Warning: Requested frequency range extends below minimum frequency of '+str(fstop[0])+' Hz for SRD fit for '+IFO+' detector.')

    if max(f)>fstop[1]:
        pass
        #print('Warning: Requested frequency range extends above maximum frequency of '+str(fstop[1])+' Hz for SRD fit for '+IFO+' detector.')



    #% ---- Done.
    return S, fstop

'''

# % -------------------------------------------------------------------------
# %                              ET-B SRD
# % -------------------------------------------------------------------------
# function[sqrtS, fstop] = ETB_fit(f)
# % ETB_fit - simple fit to the "ET-B" sensitivity curve of arXiv:0810.0604v2 
# % 
# % [sqrtS, fstop] = ETB_fit(f)
# %
# %   f       Vector of sampled frequencies (Hz).
# %
# %   S       Vector of the same shape as f, containing amplitude spectral 
# %           densities of ET_B (1/sqrt(Hz)).
# %   fstop   Minimum frequency (Hz) at which fit is "guaranteed" to be
# %           accurate.
# %
# % ETB_fit provides a simple fit to the "ET-B" design noise spectrum curve
# % described in arXiv:0810.0604v2.  For data see 
# %   https://workarea.et-gw.eu/et/WG3-Topology 
# % The following analytic approximation is good to within 10% across the
# % [1,10000] Hz band.  Compared to the sampled spectrum, the analytic
# % approximation Sanal has the properties 
# %        min((Sanal/S).^0.5)  = 0.90
# %        mean((Sanal/S).^0.5) = 0.95
# %        min((Sanal/S).^0.5)  = 1.03
# %
# % Created by Thomas Dent (thomas.dent@astro.cf.ac.uk) on 14/05/09

xt = f/100.
fit = 2.39*10**(-27)*xt**(-15.64) +0.349*xt**(-2.145) +1.76*xt**(-0.12) +0.409*xt**(1.10)
sqrtS = 10**(-25)*fit
fstop = [1,float('inf')]  #%-- chosen arbitrarily -- looks okay.


# % -------------------------------------------------------------------------
# %                              ET-C SRD
# % -------------------------------------------------------------------------

def ETC(f):   #function [ET_C, fstop] = ETC_fit(f)
# % ETC - SRD noise curve for the ET detector, version "C".
# %
# % This function produces for a given frequency vector the corresponding
# % amplitude spectral density of the ET-C sensitivity curve.  It is a fit to
# % the ET-C spectrum found in S. Hild et al., CQG 27 015003 (2010), with fit
# % written by S. Hild. The fit is only valid for frequencies between 1 Hz
# % and 10kHz and the the accuracy is to within 1 percent. Please also note
# % that the fit is not given by a single function, but is split into several
# % individual frequency bands.
# %
# % usage:
# %
# %   [S, fstop] = ETC_fit(f)
# %
# %   f       Vector of sampled frequencies (Hz).
# %
# %   S       Vector of the same shape as f, containing amplitude spectral 
# %           densities of ET_C (1/sqrt(Hz)).
# %   fstop   Minimum and maximum frequency (Hz) at which fit is "guaranteed"
# %           to be accurate. 
# %
# % A detailed description of ET-C can be found in:
# % S. Hild et al.: 'A xylophone configuration for a third-generation 
# % gravitational wave detector', CQG, 2010, 27, 015003.
# %
# % Stefan Hild, 4/1/2010 (stefan.hild@ligo.org)

#% ---- Check range of f.
fstop = [1,1e4];
# % if min(f)<1
# %     error('Minimum frequency 1 Hz.');
# % end
# % if max(f)>1e4
# %     warning('Maximum frequency 1e4 Hz.');
# % end

# % ---- Force f to be a row vector.
####isColumnVector = 0;
####if size(f,1)>size(f,2)
####    isColumnVector = 1;
####    f = f.';
####end

#% ---- Append extra element 1 Hz above highest f value.
f = list(f)
f.append(f[-1]+1)
f=np.array(f)

ET_C = f;    #% initializing vector for ET_C

# % -------------------------------------------------------------------------
# %                  Piecewise computation of ET_C SRD 
# % -------------------------------------------------------------------------

# % ---- Below here the functions for the individual frequency bands are given.

f1 = 4600        #% This is the frequency range from 4.6 to 10 kHz.
[y1-1, x1-1] = np.min(np.abs(f-f1))
ET_C(x1:len(f)) = 6.405e-28.*f(x1:length(f)).^0.9969-5.092e-27;

f2 = 1900;        % This is the frequency range from 1.9 to 4.6 kHz.
[y2, x2] = min(abs(f-f2));
ET_C(x2:x1-1) = 7.458e-28.*f(x2:x1-1).^0.9794-1.793e-26;

f3 = 210;
[y3, x3] = min(abs(f-f3));
ET_C(x3:x2-1) = -7.646e-55.*f(x3:x2-1).^9 +  1.54e-50.*f(x3:x2-1).^8 -1.331e-46.*f(x3:x2-1).^7 + 6.458e-43.*f(x3:x2-1).^6 -1.936e-39.*f(x3:x2-1).^5 +  3.719e-36.*f(x3:x2-1).^4 -4.606e-33.*f(x3:x2-1).^3 + 3.614e-30.*f(x3:x2-1).^2 -1.095e-27.*f(x3:x2-1) + 4.397e-25;

f4 = 110;
[y4, x4] = min(abs(f-f4));
ET_C(x4:x3-1)=-3.317e-32.*f(x4:x3-1).^3+2.027e-29.*f(x4:x3-1).^2+-4.354e-27.*f(x4:x3-1)+ 6.601e-25;

f5 = 80;
[y5, x5] = min(abs(f-f5));
ET_C(x5:x4-1) =   -2.935e-46.*f(x5:x4-1).^9 +  6.757e-43.*f(x5:x4-1).^8 -6.794e-40.*f(x5:x4-1).^7 + 3.916e-37.*f(x5:x4-1).^6 -1.426e-34.*f(x5:x4-1).^5 +   3.414e-32.*f(x5:x4-1).^4-5.399e-30.*f(x5:x4-1).^3 + 5.52e-28.*f(x5:x4-1).^2 -3.433e-26.*f(x5:x4-1) + 1.389e-24;

f6 = 51;
[y6, x6] = min(abs(f-f6));
ET_C(x6:x5-1) =  -2.998e-30.*f(x6:x5-1).^3 +7.391e-28.*f(x6:x5-1).^2 -6.303e-26.*f(x6:x5-1) +2.279e-24;

f7 = 38;
[y7, x7] = min(abs(f-f7));
ET_C(x7:x6-1) =  1.27e-29.*f(x7:x6-1).^3 -1.118e-27.*f(x7:x6-1).^2 +2.716e-27.*f(x7:x6-1) +1.674e-24;

f8 = 29.8;
[y8, x8] = min(abs(f-f8));
ET_C(x8:x7-1) =  3.151e-28.*f(x8:x7-1).^3  -3.676e-26.*f(x8:x7-1).^2 + 1.4e-24.*f(x8:x7-1) -1.655e-23;

f9 = 22;
[y9, x9] = min(abs(f-f9));
ET_C(x9:x8-1) =  -3.173e-28.*f(x9:x8-1).^3  +   3.028e-26.*f(x9:x8-1).^2 -9.138e-25.*f(x9:x8-1) + 9.6e-24;

f10 = 18.5;
[y10, x10] = min(abs(f-f10));
ET_C(x10:x9-1) =  -5.873e-29.*f(x10:x9-1).^4+ 5.781e-27.*f(x10:x9-1).^3   -2.06e-25.*f(x10:x9-1).^2 +  3.134e-24.*f(x10:x9-1)  -1.627e-23;

f11 = 9;
[y11, x11] = min(abs(f-f11));
ET_C(x11:x10-1) =  5.104e-28.*f(x11:x10-1).^3   -2.817e-26.*f(x11:x10-1).^2 +  4.761e-25.*f(x11:x10-1)  -1.466e-24;

f12 = 7.5;
[y12, x12] = min(abs(f-f12));
ET_C(x12:x11-1) =  -1.368e-26.*f(x12:x11-1).^3  + 3.581e-25.*f(x12:x11-1).^2  -3.031e-24.*f(x12:x11-1) +  9.151e-24;

f13 = 5.7;
[y13, x13] = min(abs(f-f13));
ET_C(x13:x12-1) =   -6.333e-26.*f(x13:x12-1).^3  + 1.509e-24.*f(x13:x12-1).^2 -1.193e-23.*f(x13:x12-1) +  3.210e-23;

f14 = 4.0;
[y14, x14] = min(abs(f-f14));
ET_C(x14:x13-1) =   1.362e-21.*f(x14:x13-1).^(-3.952)  -3.004e-27;

f15 = 3.0;
[y15, x15] = min(abs(f-f15));
ET_C(x15:x14-1) =  1.135e-21.*f(x15:x14-1).^( -3.806)   -1.16e-25;

f16 = 2.4;
[y16, x16] = min(abs(f-f16));
ET_C(x16:x15-1) =  8.52e-22.*f(x16:x15-1).^(-3.53)   -4.264e-25;

f17 = 2;
[y17, x17] = min(abs(f-f17));
ET_C(x17:x16-1) =   6.899e-22.*f(x17:x16-1).^(-3.276)   -9.181e-25;

f18 = 1.65;
[y18, x18] = min(abs(f-f18));
ET_C(x18:x17-1) =   7.056e-22.*f(x18:x17-1).^(-3.521)  + 9.042e-24;

f19 = 1.4;
[y19, x19] = min(abs(f-f19));
ET_C(x19:x18-1) = 2.717e-21.*f(x19:x18-1).^4 -1.842e-20.*f(x19:x18-1).^3  +  4.709e-20.*f(x19:x18-1).^2 -5.402e-20.*f(x19:x18-1) +2.3667e-20;

f20 = 1.26;
[y20, x20] = min(abs(f-f20));
ET_C(x20:x19-1) =  -1.98e-20.*f(x20:x19-1).^3  +  8.486e-20.*f(x20:x19-1).^2  -1.219e-19.*f(x20:x19-1) + 5.8894e-20;
  
f21 = 1.12;
[y21, x21] = min(abs(f-f21));
ET_C(x21:x20-1) =  -1.042e-19.*f(x21:x20-1).^3  +  4.045e-19.*f(x21:x20-1).^2  -5.257e-19.*f(x21:x20-1) +  2.29051e-19;

f22 = 1;
[y22, x22] = min(abs(f-f22));
ET_C(x22:x21-1) =  4.741e-21.*f(x22:x21-1).^(-11.72)  + 1.906e-23;

ET_C = ET_C(1:length(ET_C)-1);

% -------------------------------------------------------------------------

% ---- Return spectrum in the format (row/column vector) matching input.
if isColumnVector
    ET_C = ET_C.';
end
 
return
'''

# % -------------------------------------------------------------------------
# %                              aLIGO spectra
# % -------------------------------------------------------------------------

def aligoearlyspectrum(IFO,f):
    # % ALIGOEARLYSPECTRUM - Possible noise curves for early aLIGO.
    # %
    # % Arguments are the same as for the parent function.
    # %
    # % The aLIGO spectra are taken from LIGO Document T1200307_v3, 
    # % https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=93068

    # % ---- Recognize detector and compute noise spectrum.
    if IFO.upper=='ALIGOEARLY-50':
        col = 1
    elif IFO.upper=='ALIGOEARLY-100':
        col = 2
    elif IFO.upper=='ALIGOEARLY-150':
        col = 3
    else:
        raise TypeError('Detector '+IFO+' not recognized.')


    #% ---- Extract pre-sampled noise spectrum from data.  First column is
    #%      frequency, other columns map to configurations as above.


    data = np.array([[9.0, 2.0315133e-21, 1.7564389e-21, 1.7360779e-21], [9.0203998, 2.4115253e-21, 2.1869794e-21, 2.1708083e-21], [9.0408458, 3.09785e-21, 2.9280383e-21, 2.9160887e-21], [9.0613382, 4.6064717e-21, 4.4950573e-21, 4.4873528e-21], [9.081877, 1.0007255e-20, 9.9569194e-21, 9.9534749e-21], [9.1024624, 4.1375496e-20, 4.136346e-20, 4.1362638e-20], [9.1230944, 6.7000822e-21, 6.6260274e-21, 6.6209432e-21], [9.1437733, 3.7140344e-21, 3.5799561e-21, 3.5706222e-21], [9.1644989, 2.6361497e-21, 2.4454318e-21, 2.4318707e-21], [9.1852716, 2.097615e-21, 1.8546095e-21, 1.8368537e-21], [9.2060913, 1.7837581e-21, 1.4933317e-21, 1.4714229e-21], [9.2269583, 1.5832162e-21, 1.2503601e-21, 1.2243496e-21], [9.2478725, 1.4467633e-21, 1.0763148e-21, 1.0462626e-21], [9.2688341, 1.3494487e-21, 9.4592165e-22, 9.1189619e-22], [9.2898433, 1.2774003e-21, 8.4490099e-22, 8.0697861e-22], [9.3109, 1.222372e-21, 7.6457475e-22, 7.2283892e-22], [9.3320045, 1.1792069e-21, 6.9936437e-22, 6.5390499e-22], [9.3531568, 1.1445471e-21, 6.4552008e-22, 5.9643274e-22], [9.3743571, 1.1161329e-21, 6.0042814e-22, 5.4781329e-22], [9.3956054, 1.0924006e-21, 5.6221082e-22, 5.0617302e-22], [9.4169019, 1.0722422e-21, 5.2948451e-22, 4.7013159e-22], [9.4382467, 1.0548558e-21, 5.0120728e-22, 4.3864963e-22], [9.4596398, 1.0396507e-21, 4.7657977e-22, 4.1092956e-22], [9.4810815, 1.0261843e-21, 4.5497866e-22, 3.8634918e-22], [9.5025717, 1.0141201e-21, 4.3591091e-22, 3.6441594e-22], [9.5241107, 1.0031991e-21, 4.1898162e-22, 3.4473484e-22], [9.5456984, 9.9321909e-22, 4.0387096e-22, 3.2698543e-22], [9.5673351, 9.840209e-22, 3.9031731e-22, 3.1090499e-22], [9.5890209, 9.7547772e-22, 3.7810474e-22, 2.9627613e-22], [9.6107557, 9.6748765e-22, 3.6705367e-22, 2.8291738e-22], [9.6325399, 9.5996809e-22, 3.5701355e-22, 2.7067609e-22], [9.6543734, 9.5285155e-22, 3.4785743e-22, 2.594229e-22], [9.6762565, 9.4608248e-22, 3.3947757e-22, 2.4904746e-22], [9.6981891, 9.3961485e-22, 3.3178204e-22, 2.3945509e-22], [9.7201714, 9.3341028e-22, 3.2469202e-22, 2.3056411e-22], [9.7422036, 9.2743655e-22, 3.1813961e-22, 2.2230363e-22], [9.7642857, 9.2166648e-22, 3.1206603e-22, 2.1461184e-22], [9.7864179, 9.1607701e-22, 3.0642024e-22, 2.0743464e-22], [9.8086002, 9.1064848e-22, 3.011577e-22, 2.0072442e-22], [9.8308328, 9.0536404e-22, 2.9623948e-22, 1.9443917e-22], [9.8531158, 9.0020922e-22, 2.9163138e-22, 1.8854165e-22], [9.8754493, 8.951715e-22, 2.8730329e-22, 1.8299875e-22], [9.8978334, 8.9024001e-22, 2.8322861e-22, 1.7778092e-22], [9.9202683, 8.854053e-22, 2.7938378e-22, 1.7286171e-22], [9.942754, 8.8065911e-22, 2.7574789e-22, 1.6821737e-22], [9.9652907, 8.7599418e-22, 2.7230231e-22, 1.6382652e-22], [9.9878784, 8.7140414e-22, 2.6903044e-22, 1.5966988e-22], [10.010517, 8.6688336e-22, 2.6591743e-22, 1.5573003e-22], [10.033208, 8.6242686e-22, 2.6294998e-22, 1.5199117e-22], [10.055949, 8.5803021e-22, 2.6011614e-22, 1.4843892e-22], [10.078743, 8.5368948e-22, 2.5740515e-22, 1.450602e-22], [10.101588, 8.4940114e-22, 2.5480731e-22, 1.4184304e-22], [10.124484, 8.4516204e-22, 2.5231385e-22, 1.387765e-22], [10.147433, 8.4096935e-22, 2.4991684e-22, 1.3585058e-22], [10.170433, 8.3682052e-22, 2.476091e-22, 1.330561e-22], [10.193486, 8.3271326e-22, 2.453841e-22, 1.303846e-22], [10.216591, 8.286455e-22, 2.4323591e-22, 1.2782832e-22], [10.239749, 8.2461536e-22, 2.4115912e-22, 1.2538011e-22], [10.262958, 8.2062112e-22, 2.3914882e-22, 1.2303333e-22], [10.286221, 8.1666124e-22, 2.372005e-22, 1.2078187e-22], [10.309536, 8.127343e-22, 2.3531004e-22, 1.1862008e-22], [10.332904, 8.0883901e-22, 2.3347367e-22, 1.1654271e-22], [10.356325, 8.0497418e-22, 2.3168795e-22, 1.145449e-22], [10.379799, 8.0113874e-22, 2.2994969e-22, 1.1262213e-22], [10.403327, 7.9733167e-22, 2.2825598e-22, 1.1077018e-22], [10.426907, 7.9355207e-22, 2.2660413e-22, 1.0898516e-22], [10.450541, 7.8979908e-22, 2.2499167e-22, 1.072634e-22], [10.474229, 7.8607193e-22, 2.2341631e-22, 1.0560152e-22], [10.49797, 7.823699e-22, 2.2187595e-22, 1.0399634e-22], [10.521765, 7.786923e-22, 2.2036864e-22, 1.0244489e-22], [10.545615, 7.7503852e-22, 2.1889258e-22, 1.009444e-22], [10.569518, 7.7140798e-22, 2.1744609e-22, 9.9492265e-23], [10.593475, 7.6780014e-22, 2.1602762e-22, 9.8086061e-23], [10.617487, 7.642145e-22, 2.1463573e-22, 9.6723508e-23], [10.641553, 7.6065058e-22, 2.132691e-22, 9.5402468e-23], [10.665673, 7.5710794e-22, 2.1192647e-22, 9.4120931e-23], [10.689849, 7.5358617e-22, 2.1060668e-22, 9.2877013e-23], [10.714079, 7.5008489e-22, 2.0930866e-22, 9.1668939e-23], [10.738364, 7.4660373e-22, 2.0803141e-22, 9.0495044e-23], [10.762704, 7.4314234e-22, 2.0677397e-22, 8.9353759e-23], [10.787099, 7.397004e-22, 2.0553549e-22, 8.8243604e-23], [10.811549, 7.362776e-22, 2.0431513e-22, 8.7163188e-23], [10.836055, 7.3287366e-22, 2.0311213e-22, 8.6111195e-23], [10.860617, 7.294883e-22, 2.0192577e-22, 8.5086388e-23], [10.885234, 7.2612126e-22, 2.0075537e-22, 8.4087598e-23], [10.909907, 7.2277229e-22, 1.9960031e-22, 8.311372e-23], [10.934636, 7.1944115e-22, 1.9846e-22, 8.2163709e-23], [10.959421, 7.1612763e-22, 1.9733386e-22, 8.123658e-23], [10.984262, 7.1283151e-22, 1.9622139e-22, 8.0331397e-23], [11.009159, 7.0955258e-22, 1.9512209e-22, 7.9447278e-23], [11.034113, 7.0629066e-22, 1.9403549e-22, 7.8583388e-23], [11.059124, 7.0304554e-22, 1.9296116e-22, 7.7738933e-23], [11.084191, 6.9981706e-22, 1.9189868e-22, 7.6913161e-23], [11.109315, 6.9660504e-22, 1.9084767e-22, 7.6105362e-23], [11.134495, 6.9340931e-22, 1.8980776e-22, 7.5314858e-23], [11.159733, 6.9022972e-22, 1.887786e-22, 7.454101e-23], [11.185028, 6.8706612e-22, 1.8775986e-22, 7.3783208e-23], [11.210381, 6.8391835e-22, 1.8675124e-22, 7.3040875e-23], [11.235791, 6.8078627e-22, 1.8575244e-22, 7.2313459e-23], [11.261258, 6.7766975e-22, 1.8476317e-22, 7.1600438e-23], [11.286784, 6.7456864e-22, 1.8378318e-22, 7.0901315e-23], [11.312367, 6.7148283e-22, 1.8281222e-22, 7.0215617e-23], [11.338008, 6.6841219e-22, 1.8185004e-22, 6.9542893e-23], [11.363707, 6.6535659e-22, 1.8089643e-22, 6.8882715e-23], [11.389465, 6.6231592e-22, 1.7995116e-22, 6.8234672e-23], [11.41528, 6.5929006e-22, 1.7901403e-22, 6.7598374e-23], [11.441155, 6.562789e-22, 1.7808485e-22, 6.6973451e-23], [11.467088, 6.5328234e-22, 1.7716343e-22, 6.6359547e-23], [11.49308, 6.5030025e-22, 1.762496e-22, 6.5756323e-23], [11.51913, 6.4733255e-22, 1.7534318e-22, 6.5163455e-23], [11.54524, 6.4437914e-22, 1.7444401e-22, 6.4580633e-23], [11.571409, 6.414399e-22, 1.7355194e-22, 6.4007563e-23], [11.597637, 6.3851475e-22, 1.7266683e-22, 6.344396e-23], [11.623925, 6.3560359e-22, 1.7178853e-22, 6.2889556e-23], [11.650272, 6.3270633e-22, 1.709169e-22, 6.2344089e-23], [11.676679, 6.2982288e-22, 1.7005182e-22, 6.1807312e-23], [11.703146, 6.2695314e-22, 1.6919317e-22, 6.1278987e-23], [11.729673, 6.2409704e-22, 1.6834082e-22, 6.0758886e-23], [11.75626, 6.2125449e-22, 1.6749466e-22, 6.0246791e-23], [11.782907, 6.184254e-22, 1.6665459e-22, 5.974249e-23], [11.809615, 6.1560969e-22, 1.6582049e-22, 5.9245784e-23], [11.836383, 6.1280728e-22, 1.6499226e-22, 5.8756477e-23], [11.863212, 6.1001809e-22, 1.6416982e-22, 5.8274384e-23], [11.890102, 6.0724204e-22, 1.6335306e-22, 5.7799326e-23], [11.917052, 6.0447905e-22, 1.6254189e-22, 5.7331131e-23], [11.944064, 6.0172905e-22, 1.6173623e-22, 5.6869634e-23], [11.971137, 5.9899196e-22, 1.60936e-22, 5.6414674e-23], [11.998271, 5.9626771e-22, 1.6014111e-22, 5.5966098e-23], [12.025467, 5.9355623e-22, 1.5935148e-22, 5.5523759e-23], [12.052725, 5.9085744e-22, 1.5856704e-22, 5.5087513e-23], [12.080044, 5.8817127e-22, 1.5778771e-22, 5.4657223e-23], [12.107425, 5.8549765e-22, 1.5701343e-22, 5.4232755e-23], [12.134868, 5.8283651e-22, 1.5624413e-22, 5.3813982e-23], [12.162374, 5.8018779e-22, 1.5547973e-22, 5.3400779e-23], [12.189941, 5.7755141e-22, 1.5472019e-22, 5.2993026e-23], [12.217572, 5.7492731e-22, 1.5396542e-22, 5.2590608e-23], [12.245264, 5.7231542e-22, 1.5321538e-22, 5.2193413e-23], [12.27302, 5.6971568e-22, 1.5247e-22, 5.1801331e-23], [12.300839, 5.6712803e-22, 1.5172922e-22, 5.1414258e-23], [12.32872, 5.6455239e-22, 1.50993e-22, 5.1032093e-23], [12.356665, 5.6198871e-22, 1.5026127e-22, 5.0654736e-23], [12.384673, 5.5943691e-22, 1.4953399e-22, 5.0282092e-23], [12.412745, 5.5689695e-22, 1.4881111e-22, 4.9914067e-23], [12.44088, 5.5436876e-22, 1.4809256e-22, 4.9550573e-23], [12.469079, 5.5185227e-22, 1.4737832e-22, 4.9191522e-23], [12.497342, 5.4934743e-22, 1.4666832e-22, 4.8836829e-23], [12.525669, 5.4685417e-22, 1.4596252e-22, 4.8486411e-23], [12.55406, 5.4437244e-22, 1.4526088e-22, 4.8140188e-23], [12.582516, 5.4190218e-22, 1.4456335e-22, 4.7798083e-23], [12.611036, 5.3944333e-22, 1.438699e-22, 4.7460019e-23], [12.639621, 5.3699583e-22, 1.4318047e-22, 4.7125923e-23], [12.66827, 5.3455962e-22, 1.4249503e-22, 4.6795724e-23], [12.696985, 5.3213465e-22, 1.4181354e-22, 4.6469351e-23], [12.725764, 5.2972086e-22, 1.4113596e-22, 4.6146736e-23], [12.754609, 5.273182e-22, 1.4046225e-22, 4.5827813e-23], [12.783519, 5.249266e-22, 1.3979237e-22, 4.5512518e-23], [12.812495, 5.2254602e-22, 1.3912629e-22, 4.5200787e-23], [12.841536, 5.2017639e-22, 1.3846397e-22, 4.4892559e-23], [12.870644, 5.1781767e-22, 1.3780538e-22, 4.4587775e-23], [12.899817, 5.154698e-22, 1.3715047e-22, 4.4286375e-23], [12.929056, 5.1313272e-22, 1.3649923e-22, 4.3988302e-23], [12.958362, 5.1080639e-22, 1.3585161e-22, 4.3693502e-23], [12.987734, 5.0849075e-22, 1.3520758e-22, 4.3401918e-23], [13.017172, 5.0618574e-22, 1.3456711e-22, 4.31135e-23], [13.046677, 5.0389132e-22, 1.3393017e-22, 4.2828193e-23], [13.07625, 5.0160744e-22, 1.3329673e-22, 4.2545948e-23], [13.105889, 4.9933403e-22, 1.3266676e-22, 4.2266715e-23], [13.135595, 4.9707106e-22, 1.3204023e-22, 4.1990446e-23], [13.165369, 4.9481846e-22, 1.3141711e-22, 4.1717093e-23], [13.19521, 4.925762e-22, 1.3079737e-22, 4.144661e-23], [13.225119, 4.9034422e-22, 1.3018098e-22, 4.1178951e-23], [13.255096, 4.8812246e-22, 1.2956791e-22, 4.0914072e-23], [13.28514, 4.8591088e-22, 1.2895815e-22, 4.0651929e-23], [13.315253, 4.8370944e-22, 1.2835166e-22, 4.0392481e-23], [13.345434, 4.8151808e-22, 1.2774841e-22, 4.0135685e-23], [13.375683, 4.7933675e-22, 1.2714838e-22, 3.98815e-23], [13.406001, 4.7716541e-22, 1.2655155e-22, 3.9629887e-23], [13.436388, 4.75004e-22, 1.2595788e-22, 3.9380807e-23], [13.466843, 4.7285248e-22, 1.2536736e-22, 3.9134221e-23], [13.497368, 4.7071081e-22, 1.2477996e-22, 3.8890091e-23], [13.527961, 4.6857893e-22, 1.2419566e-22, 3.8648382e-23], [13.558624, 4.664568e-22, 1.2361443e-22, 3.8409056e-23], [13.589357, 4.6434437e-22, 1.2303625e-22, 3.8172079e-23], [13.620159, 4.6224159e-22, 1.224611e-22, 3.7937416e-23], [13.651031, 4.6014843e-22, 1.2188895e-22, 3.7705033e-23], [13.681973, 4.5806483e-22, 1.2131979e-22, 3.7474896e-23], [13.712985, 4.5599074e-22, 1.2075358e-22, 3.7246974e-23], [13.744068, 4.5392613e-22, 1.2019031e-22, 3.7021234e-23], [13.775221, 4.5187095e-22, 1.1962996e-22, 3.6797644e-23], [13.806444, 4.4982515e-22, 1.1907251e-22, 3.6576174e-23], [13.837739, 4.4778869e-22, 1.1851793e-22, 3.6356794e-23], [13.869104, 4.4576153e-22, 1.1796621e-22, 3.6139475e-23], [13.90054, 4.4374361e-22, 1.1741733e-22, 3.5924186e-23], [13.932048, 4.417349e-22, 1.1687125e-22, 3.5710899e-23], [13.963627, 4.3973536e-22, 1.1632798e-22, 3.5499588e-23], [13.995277, 4.3774494e-22, 1.1578748e-22, 3.5290223e-23], [14.027, 4.3576359e-22, 1.1524974e-22, 3.5082779e-23], [14.058794, 4.3379128e-22, 1.1471474e-22, 3.4877228e-23], [14.09066, 4.3182796e-22, 1.1418246e-22, 3.4673545e-23], [14.122599, 4.2987359e-22, 1.1365287e-22, 3.4471704e-23], [14.15461, 4.2792813e-22, 1.1312598e-22, 3.4271681e-23], [14.186693, 4.2599154e-22, 1.1260174e-22, 3.407345e-23], [14.218849, 4.2406377e-22, 1.1208016e-22, 3.3876988e-23], [14.251078, 4.2214478e-22, 1.115612e-22, 3.368227e-23], [14.28338, 4.2023453e-22, 1.1104486e-22, 3.3489275e-23], [14.315756, 4.1833299e-22, 1.1053111e-22, 3.3297978e-23], [14.348204, 4.1644011e-22, 1.1001994e-22, 3.3108357e-23], [14.380727, 4.1455584e-22, 1.0951134e-22, 3.2920392e-23], [14.413323, 4.1268016e-22, 1.0900528e-22, 3.2734058e-23], [14.445993, 4.1081301e-22, 1.0850174e-22, 3.2549337e-23], [14.478736, 4.0895437e-22, 1.0800072e-22, 3.2366206e-23], [14.511555, 4.0710419e-22, 1.075022e-22, 3.2184645e-23], [14.544447, 4.0526242e-22, 1.0700616e-22, 3.2004635e-23], [14.577414, 4.0342904e-22, 1.0651259e-22, 3.1826154e-23], [14.610456, 4.01604e-22, 1.0602146e-22, 3.1649185e-23], [14.643573, 3.9978726e-22, 1.0553277e-22, 3.1473707e-23], [14.676764, 3.9797879e-22, 1.050465e-22, 3.1299703e-23], [14.710031, 3.9617855e-22, 1.0456263e-22, 3.1127153e-23], [14.743374, 3.9438649e-22, 1.0408116e-22, 3.0956039e-23], [14.776792, 3.9260259e-22, 1.0360206e-22, 3.0786345e-23], [14.810286, 3.908268e-22, 1.0312532e-22, 3.0618051e-23], [14.843855, 3.8905908e-22, 1.0265092e-22, 3.0451142e-23], [14.877501, 3.872994e-22, 1.0217886e-22, 3.0285599e-23], [14.911223, 3.8554772e-22, 1.0170912e-22, 3.0121408e-23], [14.945021, 3.83804e-22, 1.0124168e-22, 2.995855e-23], [14.978896, 3.8206821e-22, 1.0077653e-22, 2.9797011e-23], [15.012848, 3.8034031e-22, 1.0031366e-22, 2.9636774e-23], [15.046877, 3.7862026e-22, 9.9853051e-23, 2.9477824e-23], [15.080983, 3.7690803e-22, 9.9394693e-23, 2.9320145e-23], [15.115166, 3.7520358e-22, 9.8938571e-23, 2.9163723e-23], [15.149427, 3.7350688e-22, 9.8484674e-23, 2.9008544e-23], [15.183765, 3.7181788e-22, 9.8032987e-23, 2.8854591e-23], [15.218181, 3.7013656e-22, 9.7583498e-23, 2.8701852e-23], [15.252676, 3.6846288e-22, 9.7136194e-23, 2.8550312e-23], [15.287248, 3.667968e-22, 9.6691062e-23, 2.8399957e-23], [15.321899, 3.6513828e-22, 9.624809e-23, 2.8250774e-23], [15.356628, 3.634873e-22, 9.5807266e-23, 2.810275e-23], [15.391436, 3.6184382e-22, 9.5368578e-23, 2.7955871e-23], [15.426323, 3.602078e-22, 9.4932012e-23, 2.7810125e-23], [15.461289, 3.5857921e-22, 9.4497557e-23, 2.7665498e-23], [15.496334, 3.5695801e-22, 9.4065202e-23, 2.7521978e-23], [15.531459, 3.5534418e-22, 9.3634934e-23, 2.7379554e-23], [15.566663, 3.5373767e-22, 9.3206741e-23, 2.7238213e-23], [15.601947, 3.5213846e-22, 9.2780613e-23, 2.7097942e-23], [15.637311, 3.5054651e-22, 9.2356536e-23, 2.6958731e-23], [15.672755, 3.4896178e-22, 9.1934501e-23, 2.6820568e-23], [15.70828, 3.4738425e-22, 9.1514495e-23, 2.6683442e-23], [15.743885, 3.4581387e-22, 9.1096508e-23, 2.654734e-23], [15.779571, 3.4425063e-22, 9.0680528e-23, 2.6412253e-23], [15.815337, 3.4269448e-22, 9.0266544e-23, 2.627817e-23], [15.851185, 3.4114539e-22, 8.9854545e-23, 2.614508e-23], [15.887114, 3.3960333e-22, 8.9444521e-23, 2.6012972e-23], [15.923125, 3.3806827e-22, 8.903646e-23, 2.5881836e-23], [15.959217, 3.3654017e-22, 8.8630352e-23, 2.5751662e-23], [15.99539, 3.3501901e-22, 8.8226186e-23, 2.5622441e-23], [16.031646, 3.3350475e-22, 8.7823952e-23, 2.5494161e-23], [16.067984, 3.3199736e-22, 8.742364e-23, 2.5366814e-23], [16.104405, 3.304968e-22, 8.7025238e-23, 2.524039e-23], [16.140908, 3.2900306e-22, 8.6628737e-23, 2.511488e-23], [16.177493, 3.2751609e-22, 8.6234127e-23, 2.4990274e-23], [16.214162, 3.2603587e-22, 8.5841397e-23, 2.4866563e-23], [16.250914, 3.2456236e-22, 8.5450537e-23, 2.4743738e-23], [16.287749, 3.2309553e-22, 8.5061537e-23, 2.4621791e-23], [16.324667, 3.2163536e-22, 8.4674389e-23, 2.4500713e-23], [16.36167, 3.201818e-22, 8.4289081e-23, 2.4380495e-23], [16.398756, 3.1873484e-22, 8.3905604e-23, 2.4261128e-23], [16.435926, 3.1729445e-22, 8.3523948e-23, 2.4142605e-23], [16.47318, 3.1586058e-22, 8.3144104e-23, 2.4024916e-23], [16.510519, 3.1443322e-22, 8.2766062e-23, 2.3908055e-23], [16.547942, 3.1301233e-22, 8.2389813e-23, 2.3792013e-23], [16.585451, 3.1159788e-22, 8.2015348e-23, 2.3676783e-23], [16.623044, 3.1018985e-22, 8.1642656e-23, 2.3562355e-23], [16.660723, 3.087882e-22, 8.127173e-23, 2.3448724e-23], [16.698486, 3.073929e-22, 8.0902559e-23, 2.3335881e-23], [16.736336, 3.0600393e-22, 8.0535135e-23, 2.3223819e-23], [16.774271, 3.0462126e-22, 8.0169448e-23, 2.311253e-23], [16.812293, 3.0324486e-22, 7.980549e-23, 2.3002008e-23], [16.8504, 3.0187469e-22, 7.9443252e-23, 2.2892245e-23], [16.888594, 3.0051074e-22, 7.9082724e-23, 2.2783235e-23], [16.926874, 2.9915297e-22, 7.8723899e-23, 2.267497e-23], [16.965242, 2.9780136e-22, 7.8366766e-23, 2.2567443e-23], [17.003696, 2.9645587e-22, 7.8011319e-23, 2.2460649e-23], [17.042237, 2.9511648e-22, 7.7657547e-23, 2.235458e-23], [17.080866, 2.9378316e-22, 7.7305443e-23, 2.2249229e-23], [17.119582, 2.9245588e-22, 7.6954998e-23, 2.214459e-23], [17.158386, 2.9113462e-22, 7.6606204e-23, 2.2040658e-23], [17.197278, 2.8981935e-22, 7.6259052e-23, 2.1937425e-23], [17.236258, 2.8851004e-22, 7.5913533e-23, 2.1834886e-23], [17.275326, 2.8720667e-22, 7.556964e-23, 2.1733034e-23], [17.314483, 2.859092e-22, 7.5227365e-23, 2.1631863e-23], [17.353729, 2.8461761e-22, 7.4886699e-23, 2.1531367e-23], [17.393064, 2.8333187e-22, 7.4547634e-23, 2.1431541e-23], [17.432488, 2.8205196e-22, 7.4210162e-23, 2.1332379e-23], [17.472001, 2.8077785e-22, 7.3874275e-23, 2.1233875e-23], [17.511604, 2.7950951e-22, 7.3539965e-23, 2.1136023e-23], [17.551296, 2.7824692e-22, 7.3207225e-23, 2.1038817e-23], [17.591079, 2.7699005e-22, 7.2876046e-23, 2.0942253e-23], [17.630952, 2.7573888e-22, 7.254642e-23, 2.0846325e-23], [17.670915, 2.7449338e-22, 7.2218341e-23, 2.0751027e-23], [17.710968, 2.7325352e-22, 7.1891799e-23, 2.0656355e-23], [17.751113, 2.7201927e-22, 7.1566789e-23, 2.0562302e-23], [17.791348, 2.7079062e-22, 7.1243301e-23, 2.0468865e-23], [17.831675, 2.6956754e-22, 7.0921328e-23, 2.0376037e-23], [17.872093, 2.6835e-22, 7.0600863e-23, 2.0283813e-23], [17.912603, 2.6713797e-22, 7.0281899e-23, 2.019219e-23], [17.953204, 2.6593143e-22, 6.9964428e-23, 2.0101161e-23], [17.993898, 2.6473037e-22, 6.9648442e-23, 2.0010722e-23], [18.034683, 2.6353474e-22, 6.9333935e-23, 1.9920868e-23], [18.075562, 2.6234453e-22, 6.9020899e-23, 1.9831595e-23], [18.116532, 2.6115971e-22, 6.8709327e-23, 1.9742897e-23], [18.157596, 2.5998025e-22, 6.8399211e-23, 1.9654771e-23], [18.198753, 2.5880615e-22, 6.8090545e-23, 1.9567211e-23], [18.240003, 2.5763736e-22, 6.7783321e-23, 1.9480213e-23], [18.281347, 2.5647386e-22, 6.7477532e-23, 1.9393773e-23], [18.322784, 2.5531564e-22, 6.7173172e-23, 1.9307886e-23], [18.364315, 2.5416266e-22, 6.6870233e-23, 1.9222547e-23], [18.405941, 2.530149e-22, 6.6568709e-23, 1.9137754e-23], [18.44766, 2.5187235e-22, 6.6268592e-23, 1.90535e-23], [18.489475, 2.5073497e-22, 6.5969876e-23, 1.8969783e-23], [18.531384, 2.4960274e-22, 6.5672553e-23, 1.8886598e-23], [18.573388, 2.4847564e-22, 6.5376618e-23, 1.880394e-23], [18.615487, 2.4735364e-22, 6.5082063e-23, 1.8721807e-23], [18.657682, 2.4623673e-22, 6.4788882e-23, 1.8640193e-23], [18.699972, 2.4512487e-22, 6.4497068e-23, 1.8559095e-23], [18.742358, 2.4401806e-22, 6.4206614e-23, 1.8478509e-23], [18.78484, 2.4291625e-22, 6.3917514e-23, 1.8398431e-23], [18.827419, 2.4181944e-22, 6.3629761e-23, 1.8318857e-23], [18.870094, 2.4072759e-22, 6.3343349e-23, 1.8239783e-23], [18.912866, 2.3964069e-22, 6.3058271e-23, 1.8161206e-23], [18.955734, 2.3855871e-22, 6.2774522e-23, 1.8083122e-23], [18.9987, 2.3748164e-22, 6.2492094e-23, 1.8005527e-23], [19.041764, 2.3640944e-22, 6.2210981e-23, 1.7928418e-23], [19.084925, 2.3534209e-22, 6.1931176e-23, 1.785179e-23], [19.128183, 2.3427958e-22, 6.1652675e-23, 1.7775642e-23], [19.17154, 2.3322189e-22, 6.1375469e-23, 1.7699968e-23], [19.214995, 2.3216898e-22, 6.1099554e-23, 1.7624765e-23], [19.258549, 2.3112084e-22, 6.0824923e-23, 1.7550031e-23], [19.302201, 2.3007745e-22, 6.0551569e-23, 1.7475761e-23], [19.345952, 2.2903879e-22, 6.0279487e-23, 1.7401953e-23], [19.389802, 2.2800483e-22, 6.0008671e-23, 1.7328603e-23], [19.433752, 2.2697555e-22, 5.9739114e-23, 1.7255707e-23], [19.477802, 2.2595093e-22, 5.9470811e-23, 1.7183262e-23], [19.521951, 2.2493095e-22, 5.9203755e-23, 1.7111266e-23], [19.5662, 2.239156e-22, 5.8937941e-23, 1.7039715e-23], [19.61055, 2.2290484e-22, 5.8673363e-23, 1.6968605e-23], [19.655, 2.2189866e-22, 5.8410015e-23, 1.6897934e-23], [19.699551, 2.2089703e-22, 5.814789e-23, 1.6827699e-23], [19.744203, 2.1989994e-22, 5.7886984e-23, 1.6757896e-23], [19.788956, 2.1890737e-22, 5.762729e-23, 1.6688522e-23], [19.83381, 2.1791929e-22, 5.7368803e-23, 1.6619575e-23], [19.878766, 2.1693569e-22, 5.7111516e-23, 1.6551052e-23], [19.923825, 2.1595654e-22, 5.6855425e-23, 1.6482949e-23], [19.968985, 2.1498183e-22, 5.6600524e-23, 1.6415263e-23], [20.014247, 2.1401153e-22, 5.6346807e-23, 1.6347992e-23], [20.059612, 2.1304562e-22, 5.6094267e-23, 1.6281133e-23], [20.10508, 2.1208409e-22, 5.5842901e-23, 1.6214683e-23], [20.150652, 2.1112692e-22, 5.5592702e-23, 1.614864e-23], [20.196326, 2.1017407e-22, 5.5343665e-23, 1.6082999e-23], [20.242104, 2.0922555e-22, 5.5095784e-23, 1.601776e-23], [20.287985, 2.0828132e-22, 5.4849054e-23, 1.5952918e-23], [20.333971, 2.0734136e-22, 5.4603469e-23, 1.5888471e-23], [20.380061, 2.0640566e-22, 5.4359024e-23, 1.5824417e-23], [20.426255, 2.054742e-22, 5.4115714e-23, 1.5760753e-23], [20.472554, 2.0454696e-22, 5.3873534e-23, 1.5697477e-23], [20.518958, 2.0362392e-22, 5.3632477e-23, 1.5634584e-23], [20.565467, 2.0270505e-22, 5.3392539e-23, 1.5572074e-23], [20.612082, 2.0179035e-22, 5.3153715e-23, 1.5509944e-23], [20.658802, 2.0087979e-22, 5.2915998e-23, 1.5448191e-23], [20.705628, 1.9997336e-22, 5.2679385e-23, 1.5386812e-23], [20.752561, 1.9907103e-22, 5.244387e-23, 1.5325805e-23], [20.799599, 1.9817278e-22, 5.2209448e-23, 1.5265168e-23], [20.846745, 1.9727861e-22, 5.1976113e-23, 1.5204897e-23], [20.893997, 1.9638848e-22, 5.1743861e-23, 1.5144992e-23], [20.941356, 1.9550238e-22, 5.1512686e-23, 1.5085449e-23], [20.988823, 1.946203e-22, 5.1282584e-23, 1.5026266e-23], [21.036397, 1.9374221e-22, 5.1053549e-23, 1.496744e-23], [21.084079, 1.928681e-22, 5.0825577e-23, 1.490897e-23], [21.131869, 1.9199795e-22, 5.0598662e-23, 1.4850853e-23], [21.179767, 1.9113174e-22, 5.03728e-23, 1.4793086e-23], [21.227774, 1.9026945e-22, 5.0147986e-23, 1.4735668e-23], [21.27589, 1.8941106e-22, 4.9924214e-23, 1.4678596e-23], [21.324115, 1.8855657e-22, 4.9701481e-23, 1.4621868e-23], [21.372449, 1.8770594e-22, 4.947978e-23, 1.4565481e-23], [21.420893, 1.8685917e-22, 4.9259108e-23, 1.4509434e-23], [21.469446, 1.8601623e-22, 4.903946e-23, 1.4453725e-23], [21.51811, 1.8517711e-22, 4.882083e-23, 1.439835e-23], [21.566884, 1.8434179e-22, 4.8603214e-23, 1.4343309e-23], [21.615768, 1.8351025e-22, 4.8386608e-23, 1.4288598e-23], [21.664764, 1.8268248e-22, 4.8171006e-23, 1.4234217e-23], [21.71387, 1.8185846e-22, 4.7956404e-23, 1.4180162e-23], [21.763087, 1.8103817e-22, 4.7742798e-23, 1.4126432e-23], [21.812417, 1.8022159e-22, 4.7530182e-23, 1.4073024e-23], [21.861858, 1.7940872e-22, 4.7318553e-23, 1.4019937e-23], [21.911411, 1.7859952e-22, 4.7107905e-23, 1.3967169e-23], [21.961076, 1.7779399e-22, 4.6898234e-23, 1.3914717e-23], [22.010854, 1.7699211e-22, 4.6689536e-23, 1.386258e-23], [22.060745, 1.7619386e-22, 4.6481806e-23, 1.3810756e-23], [22.110749, 1.7539923e-22, 4.6275039e-23, 1.3759242e-23], [22.160866, 1.7460819e-22, 4.6069231e-23, 1.3708037e-23], [22.211097, 1.7382074e-22, 4.5864379e-23, 1.3657139e-23], [22.261441, 1.7303686e-22, 4.5660476e-23, 1.3606545e-23], [22.3119, 1.7225652e-22, 4.5457519e-23, 1.3556255e-23], [22.362473, 1.7147972e-22, 4.5255504e-23, 1.3506266e-23], [22.413161, 1.7070644e-22, 4.5054426e-23, 1.3456577e-23], [22.463963, 1.6993666e-22, 4.4854281e-23, 1.3407184e-23], [22.514881, 1.6917037e-22, 4.4655065e-23, 1.3358088e-23], [22.565914, 1.6840755e-22, 4.4456773e-23, 1.3309285e-23], [22.617063, 1.6764819e-22, 4.42594e-23, 1.3260775e-23], [22.668328, 1.6689226e-22, 4.4062944e-23, 1.3212554e-23], [22.719709, 1.6613976e-22, 4.3867399e-23, 1.3164622e-23], [22.771207, 1.6539067e-22, 4.3672762e-23, 1.3116977e-23], [22.822821, 1.6464497e-22, 4.3479028e-23, 1.3069617e-23], [22.874552, 1.6390265e-22, 4.3286193e-23, 1.302254e-23], [22.926401, 1.6316369e-22, 4.3094253e-23, 1.2975745e-23], [22.978367, 1.6242808e-22, 4.2903204e-23, 1.2929229e-23], [23.03045, 1.616958e-22, 4.2713041e-23, 1.2882992e-23], [23.082652, 1.6096684e-22, 4.2523762e-23, 1.2837031e-23], [23.134972, 1.6024118e-22, 4.2335361e-23, 1.2791345e-23], [23.187411, 1.5951881e-22, 4.2147834e-23, 1.2745932e-23], [23.239969, 1.5879971e-22, 4.1961178e-23, 1.270079e-23], [23.292645, 1.5808387e-22, 4.1775389e-23, 1.2655918e-23], [23.345442, 1.5737127e-22, 4.1590462e-23, 1.2611315e-23], [23.398357, 1.566619e-22, 4.1406395e-23, 1.2566978e-23], [23.451393, 1.5595574e-22, 4.1223181e-23, 1.2522906e-23], [23.504549, 1.5525279e-22, 4.1040819e-23, 1.2479098e-23], [23.557825, 1.5455302e-22, 4.0859304e-23, 1.2435551e-23], [23.611223, 1.5385642e-22, 4.0678632e-23, 1.2392265e-23], [23.664741, 1.5316297e-22, 4.0498799e-23, 1.2349238e-23], [23.71838, 1.5247267e-22, 4.0319802e-23, 1.2306468e-23], [23.772142, 1.5178549e-22, 4.0141636e-23, 1.2263953e-23], [23.826025, 1.5110143e-22, 3.9964298e-23, 1.2221693e-23], [23.88003, 1.5042047e-22, 3.9787784e-23, 1.2179686e-23], [23.934157, 1.4974259e-22, 3.961209e-23, 1.213793e-23], [23.988407, 1.4906779e-22, 3.9437213e-23, 1.2096423e-23], [24.042781, 1.4839604e-22, 3.9263149e-23, 1.2055165e-23], [24.097277, 1.4772734e-22, 3.9089893e-23, 1.2014153e-23], [24.151897, 1.4706166e-22, 3.8917443e-23, 1.1973387e-23], [24.206641, 1.4639901e-22, 3.8745795e-23, 1.1932865e-23], [24.261508, 1.4573935e-22, 3.8574945e-23, 1.1892585e-23], [24.316501, 1.4508269e-22, 3.840489e-23, 1.1852546e-23], [24.371617, 1.44429e-22, 3.8235625e-23, 1.1812747e-23], [24.426859, 1.4377828e-22, 3.8067148e-23, 1.1773185e-23], [24.482226, 1.431305e-22, 3.7899454e-23, 1.1733861e-23], [24.537719, 1.4248566e-22, 3.7732541e-23, 1.1694772e-23], [24.593337, 1.4184374e-22, 3.7566404e-23, 1.1655918e-23], [24.649081, 1.4120473e-22, 3.740104e-23, 1.1617296e-23], [24.704952, 1.4056862e-22, 3.7236446e-23, 1.1578905e-23], [24.760949, 1.3993538e-22, 3.7072617e-23, 1.1540744e-23], [24.817074, 1.3930502e-22, 3.6909552e-23, 1.1502812e-23], [24.873325, 1.3867752e-22, 3.6747245e-23, 1.1465107e-23], [24.929704, 1.3805286e-22, 3.6585694e-23, 1.1427628e-23], [24.986211, 1.3743103e-22, 3.6424896e-23, 1.1390374e-23], [25.042846, 1.3681202e-22, 3.6264846e-23, 1.1353344e-23], [25.099609, 1.3619582e-22, 3.6105541e-23, 1.1316535e-23], [25.156501, 1.3558241e-22, 3.5946979e-23, 1.1279947e-23], [25.213522, 1.3497178e-22, 3.5789156e-23, 1.1243579e-23], [25.270672, 1.3436392e-22, 3.5632067e-23, 1.1207429e-23], [25.327951, 1.3375881e-22, 3.5475711e-23, 1.1171495e-23], [25.385361, 1.3315645e-22, 3.5320084e-23, 1.1135778e-23], [25.4429, 1.3255682e-22, 3.5165182e-23, 1.1100275e-23], [25.50057, 1.3195991e-22, 3.5011002e-23, 1.1064985e-23], [25.558371, 1.313657e-22, 3.4857541e-23, 1.1029907e-23], [25.616303, 1.3077418e-22, 3.4704796e-23, 1.0995041e-23], [25.674366, 1.3018535e-22, 3.4552763e-23, 1.0960383e-23], [25.73256, 1.2959919e-22, 3.4401439e-23, 1.0925934e-23], [25.790887, 1.2901569e-22, 3.4250822e-23, 1.0891692e-23], [25.849346, 1.2843483e-22, 3.4100907e-23, 1.0857656e-23], [25.907937, 1.278566e-22, 3.3951692e-23, 1.0823825e-23], [25.966661, 1.27281e-22, 3.3803173e-23, 1.0790198e-23], [26.025518, 1.2670801e-22, 3.3655348e-23, 1.0756773e-23], [26.084509, 1.2613761e-22, 3.3508213e-23, 1.0723549e-23], [26.143633, 1.255698e-22, 3.3361765e-23, 1.0690525e-23], [26.202891, 1.2500457e-22, 3.3216001e-23, 1.0657701e-23], [26.262284, 1.244419e-22, 3.3070918e-23, 1.0625074e-23], [26.321811, 1.2388178e-22, 3.2926512e-23, 1.0592643e-23], [26.381473, 1.233242e-22, 3.2782782e-23, 1.0560409e-23], [26.441271, 1.2276915e-22, 3.2639724e-23, 1.0528368e-23], [26.501204, 1.2221662e-22, 3.2497334e-23, 1.0496521e-23], [26.561272, 1.216666e-22, 3.235561e-23, 1.0464867e-23], [26.621477, 1.2111906e-22, 3.2214548e-23, 1.0433403e-23], [26.681819, 1.2057402e-22, 3.2074147e-23, 1.040213e-23], [26.742297, 1.2003144e-22, 3.1934402e-23, 1.0371045e-23], [26.802912, 1.1949133e-22, 3.1795312e-23, 1.0340148e-23], [26.863665, 1.1895366e-22, 3.1656872e-23, 1.0309438e-23], [26.924555, 1.1841844e-22, 3.151908e-23, 1.0278913e-23], [26.985584, 1.1788564e-22, 3.1381933e-23, 1.0248574e-23], [27.04675, 1.1735526e-22, 3.1245429e-23, 1.0218417e-23], [27.108056, 1.1682728e-22, 3.1109564e-23, 1.0188444e-23], [27.1695, 1.1630171e-22, 3.0974335e-23, 1.0158651e-23], [27.231084, 1.1577851e-22, 3.083974e-23, 1.012904e-23], [27.292807, 1.1525769e-22, 3.0705776e-23, 1.0099607e-23], [27.35467, 1.1473923e-22, 3.0572439e-23, 1.0070353e-23], [27.416673, 1.1422313e-22, 3.0439728e-23, 1.0041277e-23], [27.478817, 1.1370936e-22, 3.0307639e-23, 1.0012376e-23], [27.541102, 1.1319793e-22, 3.017617e-23, 9.9836511e-24], [27.603527, 1.1268882e-22, 3.0045318e-23, 9.9551004e-24], [27.666095, 1.1218202e-22, 2.9915079e-23, 9.926723e-24], [27.728804, 1.1167752e-22, 2.9785452e-23, 9.8985179e-24], [27.791655, 1.1117531e-22, 2.9656434e-23, 9.8704842e-24], [27.854649, 1.1067537e-22, 2.9528021e-23, 9.8426208e-24], [27.917786, 1.1017771e-22, 2.9400211e-23, 9.8149268e-24], [27.981065, 1.096823e-22, 2.9273002e-23, 9.7874012e-24], [28.044488, 1.0918915e-22, 2.914639e-23, 9.760043e-24], [28.108055, 1.0869823e-22, 2.9020374e-23, 9.7328513e-24], [28.171766, 1.0820954e-22, 2.889495e-23, 9.7058252e-24], [28.235622, 1.0772307e-22, 2.8770116e-23, 9.6789636e-24], [28.299622, 1.0723881e-22, 2.8645869e-23, 9.6522656e-24], [28.363767, 1.0675674e-22, 2.8522206e-23, 9.6257303e-24], [28.428057, 1.0627687e-22, 2.8399126e-23, 9.5993568e-24], [28.492494, 1.0579917e-22, 2.8276625e-23, 9.5731441e-24], [28.557076, 1.0532364e-22, 2.81547e-23, 9.5470914e-24], [28.621805, 1.0485027e-22, 2.803335e-23, 9.5211976e-24], [28.68668, 1.0437905e-22, 2.7912571e-23, 9.4954618e-24], [28.751703, 1.0390997e-22, 2.7792361e-23, 9.4698833e-24], [28.816873, 1.0344302e-22, 2.7672718e-23, 9.4444609e-24], [28.88219, 1.0297819e-22, 2.7553639e-23, 9.419194e-24], [28.947656, 1.0251547e-22, 2.7435122e-23, 9.3940815e-24], [29.01327, 1.0205485e-22, 2.7317163e-23, 9.3691225e-24], [29.079033, 1.0159633e-22, 2.7199761e-23, 9.3443163e-24], [29.144944, 1.0113988e-22, 2.7082914e-23, 9.3196618e-24], [29.211006, 1.0068551e-22, 2.6966617e-23, 9.2951583e-24], [29.277216, 1.0023321e-22, 2.685087e-23, 9.2708049e-24], [29.343577, 9.9782954e-23, 2.673567e-23, 9.2466006e-24], [29.410089, 9.9334747e-23, 2.6621014e-23, 9.2225446e-24], [29.476751, 9.8888576e-23, 2.6506899e-23, 9.1986362e-24], [29.543564, 9.8444432e-23, 2.6393325e-23, 9.1748743e-24], [29.610529, 9.8002306e-23, 2.6280287e-23, 9.1512582e-24], [29.677646, 9.7562189e-23, 2.6167784e-23, 9.1277871e-24], [29.744914, 9.7124072e-23, 2.6055813e-23, 9.10446e-24], [29.812335, 9.6687945e-23, 2.5944372e-23, 9.0812762e-24], [29.879909, 9.6253799e-23, 2.5833459e-23, 9.0582348e-24], [29.947636, 9.5821627e-23, 2.5723071e-23, 9.035335e-24], [30.015517, 9.5391418e-23, 2.5613207e-23, 9.0125759e-24], [30.083552, 9.4963165e-23, 2.5503862e-23, 8.9899568e-24], [30.15174, 9.4536858e-23, 2.5395036e-23, 8.9674769e-24], [30.220084, 9.4112488e-23, 2.5286726e-23, 8.9451353e-24], [30.288582, 9.3690048e-23, 2.517893e-23, 8.9229312e-24], [30.357235, 9.3269527e-23, 2.5071646e-23, 8.9008638e-24], [30.426044, 9.2850918e-23, 2.496487e-23, 8.8789324e-24], [30.495009, 9.2434212e-23, 2.4858602e-23, 8.8571361e-24], [30.56413, 9.20194e-23, 2.4752838e-23, 8.8354741e-24], [30.633408, 9.1606474e-23, 2.4647577e-23, 8.8139457e-24], [30.702844, 9.1195425e-23, 2.4542816e-23, 8.7925501e-24], [30.772436, 9.0786246e-23, 2.4438553e-23, 8.7712865e-24], [30.842186, 9.0378927e-23, 2.4334786e-23, 8.7501541e-24], [30.912094, 8.997346e-23, 2.4231512e-23, 8.7291521e-24], [30.982161, 8.9569837e-23, 2.4128731e-23, 8.7082799e-24], [31.052387, 8.9168049e-23, 2.4026438e-23, 8.6875366e-24], [31.122771, 8.8768089e-23, 2.3924633e-23, 8.6669215e-24], [31.193315, 8.8369948e-23, 2.3823312e-23, 8.6464338e-24], [31.26402, 8.7973618e-23, 2.3722475e-23, 8.6260728e-24], [31.334884, 8.7579091e-23, 2.3622118e-23, 8.6058377e-24], [31.405909, 8.7186358e-23, 2.352224e-23, 8.5857278e-24], [31.477095, 8.6795412e-23, 2.3422839e-23, 8.5657424e-24], [31.548442, 8.6406244e-23, 2.3323912e-23, 8.5458807e-24], [31.619951, 8.6018847e-23, 2.3225457e-23, 8.5261419e-24], [31.691623, 8.5633213e-23, 2.3127473e-23, 8.5065255e-24], [31.763456, 8.5249334e-23, 2.3029957e-23, 8.4870305e-24], [31.835453, 8.4867201e-23, 2.2932907e-23, 8.4676564e-24], [31.907612, 8.4486807e-23, 2.2836321e-23, 8.4484024e-24], [31.979936, 8.4108145e-23, 2.2740197e-23, 8.4292678e-24], [32.052423, 8.3731206e-23, 2.2644533e-23, 8.4102519e-24], [32.125074, 8.3355982e-23, 2.2549327e-23, 8.3913539e-24], [32.19789, 8.2982467e-23, 2.2454577e-23, 8.3725733e-24], [32.270871, 8.2610652e-23, 2.2360281e-23, 8.3539092e-24], [32.344018, 8.224053e-23, 2.2266438e-23, 8.3353609e-24], [32.41733, 8.1872092e-23, 2.2173044e-23, 8.3169279e-24], [32.490809, 8.1505332e-23, 2.2080098e-23, 8.2986094e-24], [32.564454, 8.1140243e-23, 2.1987598e-23, 8.2804047e-24], [32.638266, 8.0776815e-23, 2.1895543e-23, 8.2623132e-24], [32.712245, 8.0415043e-23, 2.1803929e-23, 8.2443341e-24], [32.786392, 8.0054918e-23, 2.1712756e-23, 8.2264669e-24], [32.860707, 7.9696433e-23, 2.1622021e-23, 8.2087107e-24], [32.935191, 7.9339582e-23, 2.1531722e-23, 8.191065e-24], [33.009843, 7.8984355e-23, 2.1441858e-23, 8.1735291e-24], [33.084665, 7.8630747e-23, 2.1352427e-23, 8.1561024e-24], [33.159656, 7.827875e-23, 2.1263426e-23, 8.1387841e-24], [33.234817, 7.7928357e-23, 2.1174854e-23, 8.1215737e-24], [33.310148, 7.757956e-23, 2.1086709e-23, 8.1044704e-24], [33.385651, 7.7232353e-23, 2.0998989e-23, 8.0874736e-24], [33.461324, 7.6886728e-23, 2.0911692e-23, 8.0705828e-24], [33.537169, 7.6542678e-23, 2.0824817e-23, 8.0537972e-24], [33.613186, 7.6200196e-23, 2.0738361e-23, 8.0371162e-24], [33.689375, 7.5859275e-23, 2.0652322e-23, 8.0205392e-24], [33.765737, 7.5519908e-23, 2.05667e-23, 8.0040655e-24], [33.842272, 7.5182088e-23, 2.0481491e-23, 7.9876946e-24], [33.91898, 7.4845808e-23, 2.0396695e-23, 7.9714257e-24], [33.995862, 7.4511062e-23, 2.0312309e-23, 7.9552584e-24], [34.072919, 7.4177842e-23, 2.0228332e-23, 7.9391919e-24], [34.15015, 7.3846141e-23, 2.0144761e-23, 7.9232256e-24], [34.227556, 7.3515954e-23, 2.0061596e-23, 7.907359e-24], [34.305138, 7.3187272e-23, 1.9978834e-23, 7.8915914e-24], [34.382895, 7.2860089e-23, 1.9896473e-23, 7.8759222e-24], [34.460829, 7.2534399e-23, 1.9814513e-23, 7.8603509e-24], [34.53894, 7.2210195e-23, 1.973295e-23, 7.8448768e-24], [34.617227, 7.1887469e-23, 1.9651784e-23, 7.8294993e-24], [34.695692, 7.1566217e-23, 1.9571012e-23, 7.8142178e-24], [34.774335, 7.124643e-23, 1.9490633e-23, 7.7990318e-24], [34.853156, 7.0928102e-23, 1.9410645e-23, 7.7839406e-24], [34.932155, 7.0611227e-23, 1.9331047e-23, 7.7689438e-24], [35.011334, 7.0295799e-23, 1.9251836e-23, 7.7540406e-24], [35.090692, 6.998181e-23, 1.9173012e-23, 7.7392305e-24], [35.17023, 6.9669255e-23, 1.9094571e-23, 7.724513e-24], [35.249949, 6.9358127e-23, 1.9016514e-23, 7.7098875e-24], [35.329848, 6.9048419e-23, 1.8938838e-23, 7.6953533e-24], [35.409928, 6.8740126e-23, 1.8861541e-23, 7.6809101e-24], [35.49019, 6.843324e-23, 1.8784622e-23, 7.666557e-24], [35.570633, 6.8127756e-23, 1.8708078e-23, 7.6522938e-24], [35.651259, 6.7823667e-23, 1.863191e-23, 7.6381196e-24], [35.732068, 6.7520967e-23, 1.8556114e-23, 7.6240341e-24], [35.81306, 6.721965e-23, 1.8480689e-23, 7.6100366e-24], [35.894235, 6.691971e-23, 1.8405634e-23, 7.5961267e-24], [35.975595, 6.6621141e-23, 1.8330947e-23, 7.5823037e-24], [36.057139, 6.6323935e-23, 1.8256627e-23, 7.5685671e-24], [36.138867, 6.6028088e-23, 1.8182671e-23, 7.5549165e-24], [36.220781, 6.5733593e-23, 1.8109079e-23, 7.5413511e-24], [36.302881, 6.5440444e-23, 1.8035848e-23, 7.5278706e-24], [36.385167, 6.5148636e-23, 1.7962977e-23, 7.5144744e-24], [36.467639, 6.4858161e-23, 1.7890465e-23, 7.5011619e-24], [36.550298, 6.4569015e-23, 1.7818309e-23, 7.4879327e-24], [36.633144, 6.428119e-23, 1.774651e-23, 7.4747861e-24], [36.716179, 6.3994683e-23, 1.7675064e-23, 7.4617218e-24], [36.799401, 6.3709485e-23, 1.760397e-23, 7.4487391e-24], [36.882812, 6.3425592e-23, 1.7533227e-23, 7.4358376e-24], [36.966413, 6.3142998e-23, 1.7462834e-23, 7.4230167e-24], [37.050202, 6.2861697e-23, 1.7392788e-23, 7.410276e-24], [37.134182, 6.2581683e-23, 1.7323088e-23, 7.3976149e-24], [37.218352, 6.230295e-23, 1.7253734e-23, 7.385033e-24], [37.302713, 6.2025493e-23, 1.7184723e-23, 7.3725296e-24], [37.387265, 6.1749307e-23, 1.7116053e-23, 7.3601044e-24], [37.472008, 6.1474384e-23, 1.7047724e-23, 7.3477569e-24], [37.556944, 6.120072e-23, 1.6979733e-23, 7.3354864e-24], [37.642072, 6.0928309e-23, 1.691208e-23, 7.3232927e-24], [37.727393, 6.0657146e-23, 1.6844763e-23, 7.3111751e-24], [37.812908, 6.0387224e-23, 1.6777781e-23, 7.2991332e-24], [37.898616, 6.0118539e-23, 1.6711131e-23, 7.2871665e-24], [37.984519, 5.9851084e-23, 1.6644813e-23, 7.2752745e-24], [38.070616, 5.9584855e-23, 1.6578826e-23, 7.2634567e-24], [38.156909, 5.9319846e-23, 1.6513166e-23, 7.2517128e-24], [38.243397, 5.9056051e-23, 1.6447835e-23, 7.2400421e-24], [38.330081, 5.8793466e-23, 1.6382829e-23, 7.2284443e-24], [38.416962, 5.8532084e-23, 1.6318147e-23, 7.2169188e-24], [38.504039, 5.82719e-23, 1.6253789e-23, 7.2054652e-24], [38.591314, 5.8012909e-23, 1.6189752e-23, 7.1940831e-24], [38.678787, 5.7755105e-23, 1.6126036e-23, 7.182772e-24], [38.766458, 5.7498484e-23, 1.6062639e-23, 7.1715314e-24], [38.854328, 5.724304e-23, 1.5999559e-23, 7.1603609e-24], [38.942397, 5.6988767e-23, 1.5936795e-23, 7.14926e-24], [39.030665, 5.6735662e-23, 1.5874346e-23, 7.1382282e-24], [39.119134, 5.6483717e-23, 1.5812211e-23, 7.1272652e-24], [39.207803, 5.6232929e-23, 1.5750387e-23, 7.1163705e-24], [39.296673, 5.5983291e-23, 1.5688875e-23, 7.1055437e-24], [39.385745, 5.57348e-23, 1.5627671e-23, 7.0947842e-24], [39.475018, 5.5487449e-23, 1.5566776e-23, 7.0840917e-24], [39.564494, 5.5241234e-23, 1.5506188e-23, 7.0734658e-24], [39.654172, 5.499615e-23, 1.5445905e-23, 7.062906e-24], [39.744054, 5.4752191e-23, 1.5385926e-23, 7.0524118e-24], [39.83414, 5.4509353e-23, 1.5326249e-23, 7.0419829e-24], [39.92443, 5.4267631e-23, 1.5266875e-23, 7.0316188e-24], [40.014924, 5.4027019e-23, 1.52078e-23, 7.0213192e-24], [40.105624, 5.3787513e-23, 1.5149024e-23, 7.0110835e-24], [40.196529, 5.3549107e-23, 1.5090546e-23, 7.0009114e-24], [40.28764, 5.3311798e-23, 1.5032364e-23, 6.9908025e-24], [40.378958, 5.3075579e-23, 1.4974478e-23, 6.9807563e-24], [40.470483, 5.2840447e-23, 1.4916885e-23, 6.9707725e-24], [40.562215, 5.2606396e-23, 1.4859584e-23, 6.9608506e-24], [40.654155, 5.2373422e-23, 1.4802575e-23, 6.9509902e-24], [40.746303, 5.2141519e-23, 1.4745855e-23, 6.9411909e-24], [40.838661, 5.1910683e-23, 1.4689425e-23, 6.9314524e-24], [40.931227, 5.1680909e-23, 1.4633282e-23, 6.9217742e-24], [41.024004, 5.1452193e-23, 1.4577425e-23, 6.912156e-24], [41.116991, 5.122453e-23, 1.4521853e-23, 6.9025972e-24], [41.210188, 5.0997915e-23, 1.4466565e-23, 6.8930977e-24], [41.303597, 5.0772343e-23, 1.4411559e-23, 6.8836569e-24], [41.397218, 5.054781e-23, 1.4356835e-23, 6.8742744e-24], [41.49105, 5.0324312e-23, 1.430239e-23, 6.86495e-24], [41.585096, 5.0101843e-23, 1.4248225e-23, 6.8556832e-24], [41.679354, 4.9880399e-23, 1.4194337e-23, 6.8464736e-24], [41.773827, 4.9659975e-23, 1.4140726e-23, 6.8373208e-24], [41.868513, 4.9440568e-23, 1.408739e-23, 6.8282246e-24], [41.963414, 4.9222171e-23, 1.4034329e-23, 6.8191844e-24], [42.05853, 4.9004782e-23, 1.398154e-23, 6.8102e-24], [42.153862, 4.8788396e-23, 1.3929023e-23, 6.801271e-24], [42.24941, 4.8573007e-23, 1.3876776e-23, 6.792397e-24], [42.345174, 4.8358612e-23, 1.3824799e-23, 6.7835776e-24], [42.441155, 4.8145207e-23, 1.3773091e-23, 6.7748125e-24], [42.537354, 4.7932786e-23, 1.3721649e-23, 6.7661013e-24], [42.633771, 4.7721345e-23, 1.3670473e-23, 6.7574436e-24], [42.730407, 4.7510881e-23, 1.3619562e-23, 6.7488392e-24], [42.827262, 4.7301389e-23, 1.3568915e-23, 6.7402877e-24], [42.924336, 4.7092864e-23, 1.351853e-23, 6.7317886e-24], [43.02163, 4.6885302e-23, 1.3468407e-23, 6.7233417e-24], [43.119145, 4.6678699e-23, 1.3418543e-23, 6.7149466e-24], [43.21688, 4.6473051e-23, 1.3368939e-23, 6.706603e-24], [43.314838, 4.6268353e-23, 1.3319593e-23, 6.6983105e-24], [43.413017, 4.6064602e-23, 1.3270504e-23, 6.6900688e-24], [43.511419, 4.5861792e-23, 1.3221671e-23, 6.6818775e-24], [43.610044, 4.5659921e-23, 1.3173092e-23, 6.6737363e-24], [43.708892, 4.5458983e-23, 1.3124767e-23, 6.665645e-24], [43.807965, 4.5258975e-23, 1.3076694e-23, 6.657603e-24], [43.907262, 4.5059892e-23, 1.3028873e-23, 6.6496102e-24], [44.006784, 4.4861731e-23, 1.2981302e-23, 6.6416662e-24], [44.106531, 4.4664487e-23, 1.293398e-23, 6.6337706e-24], [44.206505, 4.4468157e-23, 1.2886906e-23, 6.6259232e-24], [44.306706, 4.4272735e-23, 1.284008e-23, 6.6181236e-24], [44.407133, 4.4078219e-23, 1.2793499e-23, 6.6103715e-24], [44.507788, 4.3884604e-23, 1.2747163e-23, 6.6026665e-24], [44.608672, 4.3691887e-23, 1.2701071e-23, 6.5950085e-24], [44.709784, 4.3500062e-23, 1.2655222e-23, 6.5873969e-24], [44.811125, 4.3309127e-23, 1.2609615e-23, 6.5798317e-24], [44.912696, 4.3119077e-23, 1.2564248e-23, 6.5723123e-24], [45.014497, 4.2929909e-23, 1.2519121e-23, 6.5648386e-24], [45.116528, 4.2741619e-23, 1.2474232e-23, 6.5574102e-24], [45.218792, 4.2554202e-23, 1.2429582e-23, 6.5500268e-24], [45.321286, 4.2367655e-23, 1.2385167e-23, 6.5426881e-24], [45.424014, 4.2181974e-23, 1.2340989e-23, 6.5353939e-24], [45.526974, 4.1997155e-23, 1.2297044e-23, 6.5281437e-24], [45.630167, 4.1813195e-23, 1.2253333e-23, 6.5209374e-24], [45.733594, 4.1630089e-23, 1.2209855e-23, 6.5137746e-24], [45.837256, 4.1447834e-23, 1.2166608e-23, 6.506655e-24], [45.941153, 4.1266427e-23, 1.2123592e-23, 6.4995784e-24], [46.045285, 4.1085862e-23, 1.2080805e-23, 6.4925444e-24], [46.149653, 4.0906137e-23, 1.2038246e-23, 6.4855528e-24], [46.254258, 4.0727249e-23, 1.1995915e-23, 6.4786033e-24], [46.3591, 4.0549192e-23, 1.195381e-23, 6.4716956e-24], [46.46418, 4.0371964e-23, 1.1911931e-23, 6.4648293e-24], [46.569498, 4.0195562e-23, 1.1870276e-23, 6.4580044e-24], [46.675054, 4.001998e-23, 1.1828845e-23, 6.4512204e-24], [46.78085, 3.9845216e-23, 1.1787636e-23, 6.4444771e-24], [46.886885, 3.9671267e-23, 1.1746649e-23, 6.4377742e-24], [46.993161, 3.9498128e-23, 1.1705882e-23, 6.4311114e-24], [47.099678, 3.9325795e-23, 1.1665335e-23, 6.4244885e-24], [47.206436, 3.9154267e-23, 1.1625006e-23, 6.4179052e-24], [47.313436, 3.8983538e-23, 1.1584896e-23, 6.4113613e-24], [47.420679, 3.8813605e-23, 1.1545002e-23, 6.4048564e-24], [47.528165, 3.8644466e-23, 1.1505324e-23, 6.3983903e-24], [47.635894, 3.8476116e-23, 1.146586e-23, 6.3919628e-24], [47.743868, 3.8308552e-23, 1.1426611e-23, 6.3855736e-24], [47.852086, 3.814177e-23, 1.1387575e-23, 6.3792224e-24], [47.96055, 3.7975767e-23, 1.134875e-23, 6.3729089e-24], [48.069259, 3.781054e-23, 1.1310137e-23, 6.366633e-24], [48.178215, 3.7646085e-23, 1.1271734e-23, 6.3603944e-24], [48.287418, 3.74824e-23, 1.123354e-23, 6.3541928e-24], [48.396868, 3.7319479e-23, 1.1195554e-23, 6.3480279e-24], [48.506567, 3.7157321e-23, 1.1157776e-23, 6.3418996e-24], [48.616514, 3.6995921e-23, 1.1120205e-23, 6.3358075e-24], [48.72671, 3.6835277e-23, 1.1082839e-23, 6.3297515e-24], [48.837156, 3.6675385e-23, 1.1045677e-23, 6.3237312e-24], [48.947853, 3.6516242e-23, 1.100872e-23, 6.3177465e-24], [49.0588, 3.6357844e-23, 1.0971965e-23, 6.3117971e-24], [49.169999, 3.6200189e-23, 1.0935412e-23, 6.3058828e-24], [49.28145, 3.6043273e-23, 1.089906e-23, 6.3000033e-24], [49.393153, 3.5887092e-23, 1.0862909e-23, 6.2941584e-24], [49.50511, 3.5731645e-23, 1.0826957e-23, 6.2883479e-24], [49.61732, 3.5576926e-23, 1.0791203e-23, 6.2825715e-24], [49.729785, 3.5422934e-23, 1.0755647e-23, 6.276829e-24], [49.842505, 3.5269665e-23, 1.0720287e-23, 6.2711202e-24], [49.95548, 3.5117115e-23, 1.0685124e-23, 6.2654449e-24], [50.068711, 3.4965283e-23, 1.0650155e-23, 6.2598028e-24], [50.182199, 3.4814164e-23, 1.061538e-23, 6.2541936e-24], [50.295944, 3.4663755e-23, 1.0580798e-23, 6.2486173e-24], [50.409947, 3.4514054e-23, 1.0546409e-23, 6.2430735e-24], [50.524209, 3.4365057e-23, 1.0512211e-23, 6.2375621e-24], [50.638729, 3.4216762e-23, 1.0478204e-23, 6.2320827e-24], [50.753509, 3.4069164e-23, 1.0444387e-23, 6.2266353e-24], [50.868549, 3.3922262e-23, 1.0410758e-23, 6.2212196e-24], [50.98385, 3.3776051e-23, 1.0377318e-23, 6.2158354e-24], [51.099412, 3.363053e-23, 1.0344065e-23, 6.2104824e-24], [51.215237, 3.3485695e-23, 1.0310998e-23, 6.2051605e-24], [51.331323, 3.3341543e-23, 1.0278117e-23, 6.1998695e-24], [51.447673, 3.3198071e-23, 1.024542e-23, 6.1946091e-24], [51.564287, 3.3055276e-23, 1.0212907e-23, 6.1893791e-24], [51.681164, 3.2913155e-23, 1.0180578e-23, 6.1841794e-24], [51.798307, 3.2771705e-23, 1.014843e-23, 6.1790097e-24], [51.915716, 3.2630924e-23, 1.0116465e-23, 6.1738699e-24], [52.03339, 3.2490808e-23, 1.0084679e-23, 6.1687596e-24], [52.151331, 3.2351355e-23, 1.0053074e-23, 6.1636789e-24], [52.26954, 3.2212562e-23, 1.0021647e-23, 6.1586273e-24], [52.388016, 3.2074425e-23, 9.9903986e-24, 6.1536048e-24], [52.506761, 3.1936942e-23, 9.9593275e-24, 6.1486112e-24], [52.625775, 3.180011e-23, 9.928433e-24, 6.1436462e-24], [52.745059, 3.1663926e-23, 9.8977142e-24, 6.1387097e-24], [52.864613, 3.1528388e-23, 9.8671703e-24, 6.1338015e-24], [52.984439, 3.1393493e-23, 9.8368007e-24, 6.1289213e-24], [53.104535, 3.1259237e-23, 9.8066043e-24, 6.1240691e-24], [53.224904, 3.1125619e-23, 9.7765806e-24, 6.1192445e-24], [53.345546, 3.0992634e-23, 9.7467286e-24, 6.1144475e-24], [53.466462, 3.0860282e-23, 9.7170476e-24, 6.1096779e-24], [53.587651, 3.0728558e-23, 9.6875368e-24, 6.1049354e-24], [53.709115, 3.059746e-23, 9.6581955e-24, 6.1002199e-24], [53.830855, 3.0466986e-23, 9.6290228e-24, 6.0955312e-24], [53.95287, 3.0337132e-23, 9.6000181e-24, 6.0908692e-24], [54.075162, 3.0207896e-23, 9.5711804e-24, 6.0862336e-24], [54.197731, 3.0079276e-23, 9.5425091e-24, 6.0816243e-24], [54.320578, 2.9951268e-23, 9.5140033e-24, 6.0770411e-24], [54.443703, 2.9823871e-23, 9.4856624e-24, 6.0724838e-24], [54.567108, 2.969708e-23, 9.4574855e-24, 6.0679523e-24], [54.690792, 2.9570895e-23, 9.429472e-24, 6.0634464e-24], [54.814757, 2.9445311e-23, 9.401621e-24, 6.0589659e-24], [54.939002, 2.9320327e-23, 9.3739317e-24, 6.0545106e-24], [55.063529, 2.919594e-23, 9.3464035e-24, 6.0500805e-24], [55.188339, 2.9072148e-23, 9.3190357e-24, 6.0456752e-24], [55.313431, 2.8948947e-23, 9.2918273e-24, 6.0412948e-24], [55.438807, 2.8826336e-23, 9.2647778e-24, 6.0369389e-24], [55.564467, 2.8704311e-23, 9.2378863e-24, 6.0326075e-24], [55.690412, 2.858287e-23, 9.2111522e-24, 6.0283003e-24], [55.816642, 2.8462011e-23, 9.1845746e-24, 6.0240173e-24], [55.943158, 2.8341732e-23, 9.1581529e-24, 6.0197582e-24], [56.069962, 2.8222029e-23, 9.1318864e-24, 6.0155229e-24], [56.197052, 2.81029e-23, 9.1057743e-24, 6.0113113e-24], [56.324431, 2.7984342e-23, 9.0798158e-24, 6.0071232e-24], [56.452098, 2.7866355e-23, 9.0540103e-24, 6.0029584e-24], [56.580055, 2.7748933e-23, 9.028357e-24, 5.9988168e-24], [56.708302, 2.7632077e-23, 9.0028553e-24, 5.9946982e-24], [56.83684, 2.7515782e-23, 8.9775044e-24, 5.9906026e-24], [56.965668, 2.7400046e-23, 8.9523035e-24, 5.9865296e-24], [57.094789, 2.7284868e-23, 8.927252e-24, 5.9824793e-24], [57.224203, 2.7170245e-23, 8.9023493e-24, 5.9784514e-24], [57.35391, 2.7056174e-23, 8.8775944e-24, 5.9744458e-24], [57.483911, 2.6942652e-23, 8.8529869e-24, 5.9704624e-24], [57.614206, 2.6829679e-23, 8.8285259e-24, 5.966501e-24], [57.744797, 2.6717251e-23, 8.8042108e-24, 5.9625614e-24], [57.875684, 2.6605365e-23, 8.7800408e-24, 5.9586436e-24], [58.006867, 2.649402e-23, 8.7560154e-24, 5.9547474e-24], [58.138348, 2.6383214e-23, 8.7321337e-24, 5.9508726e-24], [58.270127, 2.6272943e-23, 8.7083951e-24, 5.9470191e-24], [58.402205, 2.6163207e-23, 8.6847989e-24, 5.9431868e-24], [58.534582, 2.6054001e-23, 8.6613444e-24, 5.9393756e-24], [58.667259, 2.5945325e-23, 8.638031e-24, 5.9355853e-24], [58.800237, 2.5837175e-23, 8.614858e-24, 5.9318157e-24], [58.933516, 2.572955e-23, 8.5918246e-24, 5.9280668e-24], [59.067097, 2.5622448e-23, 8.5689302e-24, 5.9243384e-24], [59.200981, 2.5515866e-23, 8.5461742e-24, 5.9206304e-24], [59.335169, 2.5409801e-23, 8.5235558e-24, 5.9169427e-24], [59.46966, 2.5304252e-23, 8.5010745e-24, 5.913275e-24], [59.604457, 2.5199217e-23, 8.4787294e-24, 5.9096274e-24], [59.739559, 2.5094693e-23, 8.4565201e-24, 5.9059996e-24], [59.874967, 2.4990678e-23, 8.4344457e-24, 5.9023916e-24], [60.010683, 2.4887171e-23, 8.4125057e-24, 5.8988032e-24], [60.146705, 2.4784167e-23, 8.3906994e-24, 5.8952343e-24], [60.283037, 2.4681667e-23, 8.3690261e-24, 5.8916848e-24], [60.419677, 2.4579667e-23, 8.3474852e-24, 5.8881545e-24], [60.556627, 2.4478166e-23, 8.326076e-24, 5.8846433e-24], [60.693887, 2.437716e-23, 8.3047978e-24, 5.8811512e-24], [60.831458, 2.4276649e-23, 8.2836502e-24, 5.877678e-24], [60.969342, 2.417663e-23, 8.2626323e-24, 5.8742235e-24], [61.107537, 2.4077101e-23, 8.2417435e-24, 5.8707877e-24], [61.246046, 2.3978059e-23, 8.2209833e-24, 5.8673705e-24], [61.384869, 2.3879504e-23, 8.2003509e-24, 5.8639717e-24], [61.524007, 2.3781432e-23, 8.1798458e-24, 5.8605912e-24], [61.66346, 2.3683841e-23, 8.1594672e-24, 5.8572289e-24], [61.803229, 2.3586731e-23, 8.1392146e-24, 5.8538847e-24], [61.943315, 2.3490098e-23, 8.1190874e-24, 5.8505584e-24], [62.083718, 2.339394e-23, 8.0990849e-24, 5.8472501e-24], [62.22444, 2.3298256e-23, 8.0792064e-24, 5.8439595e-24], [62.365481, 2.3203044e-23, 8.0594515e-24, 5.8406865e-24], [62.506841, 2.3108301e-23, 8.0398193e-24, 5.8374311e-24], [62.648522, 2.3014026e-23, 8.0203094e-24, 5.8341932e-24], [62.790524, 2.2920216e-23, 8.0009211e-24, 5.8309725e-24], [62.932847, 2.282687e-23, 7.9816538e-24, 5.8277691e-24], [63.075494, 2.2733985e-23, 7.9625068e-24, 5.8245828e-24], [63.218463, 2.264156e-23, 7.9434797e-24, 5.8214136e-24], [63.361757, 2.2549593e-23, 7.9245717e-24, 5.8182612e-24], [63.505376, 2.2458082e-23, 7.9057822e-24, 5.8151257e-24], [63.64932, 2.2367024e-23, 7.8871107e-24, 5.8120068e-24], [63.79359, 2.2276419e-23, 7.8685566e-24, 5.8089046e-24], [63.938187, 2.2186263e-23, 7.8501192e-24, 5.8058189e-24], [64.083113, 2.2096556e-23, 7.8317979e-24, 5.8027496e-24], [64.228366, 2.2007294e-23, 7.8135922e-24, 5.7996967e-24], [64.373949, 2.1918478e-23, 7.7955015e-24, 5.7966599e-24], [64.519862, 2.1830103e-23, 7.7775251e-24, 5.7936393e-24], [64.666105, 2.174217e-23, 7.7596625e-24, 5.7906347e-24], [64.81268, 2.1654675e-23, 7.7419132e-24, 5.787646e-24], [64.959588, 2.1567617e-23, 7.7242764e-24, 5.7846732e-24], [65.106828, 2.1480994e-23, 7.7067516e-24, 5.7817161e-24], [65.254402, 2.1394804e-23, 7.6893383e-24, 5.7787747e-24], [65.40231, 2.1309046e-23, 7.6720359e-24, 5.7758489e-24], [65.550554, 2.1223717e-23, 7.6548437e-24, 5.7729385e-24], [65.699134, 2.1138816e-23, 7.6377613e-24, 5.7700435e-24], [65.84805, 2.1054341e-23, 7.620788e-24, 5.7671638e-24], [65.997304, 2.097029e-23, 7.6039232e-24, 5.7642993e-24], [66.146897, 2.0886662e-23, 7.5871665e-24, 5.7614499e-24], [66.296828, 2.0803454e-23, 7.5705172e-24, 5.7586156e-24], [66.447099, 2.0720666e-23, 7.5539747e-24, 5.7557962e-24], [66.597711, 2.0638294e-23, 7.5375386e-24, 5.7529916e-24], [66.748665, 2.0556338e-23, 7.5212082e-24, 5.7502019e-24], [66.89996, 2.0474795e-23, 7.5049829e-24, 5.7474268e-24], [67.051599, 2.0393664e-23, 7.4888624e-24, 5.7446663e-24], [67.203581, 2.0312943e-23, 7.4728458e-24, 5.7419204e-24], [67.355907, 2.0232631e-23, 7.4569328e-24, 5.7391889e-24], [67.508579, 2.0152726e-23, 7.4411228e-24, 5.7364717e-24], [67.661597, 2.0073225e-23, 7.4254152e-24, 5.7337689e-24], [67.814962, 1.9994128e-23, 7.4098094e-24, 5.7310802e-24], [67.968674, 1.9915432e-23, 7.394305e-24, 5.7284057e-24], [68.122735, 1.9837137e-23, 7.3789014e-24, 5.7257452e-24], [68.277145, 1.9759239e-23, 7.363598e-24, 5.7230987e-24], [68.431905, 1.9681739e-23, 7.3483943e-24, 5.720466e-24], [68.587015, 1.9604633e-23, 7.3332898e-24, 5.7178472e-24], [68.742478, 1.952792e-23, 7.3182839e-24, 5.7152421e-24], [68.898292, 1.9451599e-23, 7.3033761e-24, 5.7126507e-24], [69.05446, 1.9375669e-23, 7.2885659e-24, 5.7100728e-24], [69.210982, 1.9300126e-23, 7.2738527e-24, 5.7075085e-24], [69.367859, 1.922497e-23, 7.2592361e-24, 5.7049576e-24], [69.525091, 1.91502e-23, 7.2447154e-24, 5.70242e-24], [69.68268, 1.9075813e-23, 7.2302902e-24, 5.6998958e-24], [69.840626, 1.9001807e-23, 7.2159599e-24, 5.6973848e-24], [69.998929, 1.8928183e-23, 7.2017241e-24, 5.6948869e-24], [70.157592, 1.8854937e-23, 7.1875821e-24, 5.6924022e-24], [70.316614, 1.8782068e-23, 7.1735336e-24, 5.6899304e-24], [70.475997, 1.8709574e-23, 7.1595779e-24, 5.6874716e-24], [70.635741, 1.8637455e-23, 7.1457146e-24, 5.6850256e-24], [70.795847, 1.8565708e-23, 7.1319432e-24, 5.6825925e-24], [70.956316, 1.8494332e-23, 7.1182631e-24, 5.6801721e-24], [71.117149, 1.8423326e-23, 7.1046739e-24, 5.6777643e-24], [71.278346, 1.8352687e-23, 7.091175e-24, 5.6753692e-24], [71.439908, 1.8282414e-23, 7.077766e-24, 5.6729866e-24], [71.601837, 1.8212507e-23, 7.0644463e-24, 5.6706165e-24], [71.764133, 1.8142962e-23, 7.0512155e-24, 5.6682588e-24], [71.926797, 1.8073779e-23, 7.038073e-24, 5.6659135e-24], [72.089829, 1.8004957e-23, 7.0250184e-24, 5.6635804e-24], [72.253231, 1.7936493e-23, 7.0120511e-24, 5.6612596e-24], [72.417004, 1.7868386e-23, 6.9991707e-24, 5.6589509e-24], [72.581147, 1.7800636e-23, 6.9863768e-24, 5.6566544e-24], [72.745663, 1.7733239e-23, 6.9736687e-24, 5.6543698e-24], [72.910551, 1.7666195e-23, 6.961046e-24, 5.6520973e-24], [73.075813, 1.7599503e-23, 6.9485083e-24, 5.6498367e-24], [73.24145, 1.753316e-23, 6.9360551e-24, 5.6475879e-24], [73.407463, 1.7467166e-23, 6.9236859e-24, 5.6453509e-24], [73.573851, 1.7401518e-23, 6.9114001e-24, 5.6431257e-24], [73.740617, 1.7336217e-23, 6.8991974e-24, 5.6409122e-24], [73.907761, 1.7271259e-23, 6.8870773e-24, 5.6387103e-24], [74.075283, 1.7206644e-23, 6.8750392e-24, 5.6365199e-24], [74.243185, 1.714237e-23, 6.8630828e-24, 5.6343411e-24], [74.411468, 1.7078436e-23, 6.8512076e-24, 5.6321737e-24], [74.580133, 1.701484e-23, 6.8394131e-24, 5.6300178e-24], [74.749179, 1.6951581e-23, 6.8276988e-24, 5.6278732e-24], [74.918609, 1.6888658e-23, 6.8160643e-24, 5.6257399e-24], [75.088423, 1.6826069e-23, 6.8045091e-24, 5.6236178e-24], [75.258621, 1.6763813e-23, 6.7930327e-24, 5.621507e-24], [75.429206, 1.6701888e-23, 6.7816348e-24, 5.6194072e-24], [75.600177, 1.6640293e-23, 6.7703149e-24, 5.6173186e-24], [75.771536, 1.6579027e-23, 6.7590724e-24, 5.615241e-24], [75.943283, 1.6518088e-23, 6.7479071e-24, 5.6131744e-24], [76.115419, 1.6457476e-23, 6.7368183e-24, 5.6111187e-24], [76.287946, 1.6397187e-23, 6.7258057e-24, 5.6090739e-24], [76.460863, 1.6337222e-23, 6.7148689e-24, 5.6070399e-24], [76.634173, 1.6277579e-23, 6.7040073e-24, 5.6050167e-24], [76.807875, 1.6218256e-23, 6.6932206e-24, 5.6030043e-24], [76.981971, 1.6159253e-23, 6.6825084e-24, 5.6010025e-24], [77.156462, 1.6100567e-23, 6.6718701e-24, 5.5990114e-24], [77.331348, 1.6042198e-23, 6.6613053e-24, 5.5970309e-24], [77.506631, 1.5984145e-23, 6.6508137e-24, 5.5950609e-24], [77.682311, 1.5926405e-23, 6.6403948e-24, 5.5931014e-24], [77.858389, 1.5868978e-23, 6.6300482e-24, 5.5911524e-24], [78.034866, 1.5811862e-23, 6.6197733e-24, 5.5892138e-24], [78.211743, 1.5755056e-23, 6.60957e-24, 5.5872855e-24], [78.389021, 1.5698559e-23, 6.5994376e-24, 5.5853676e-24], [78.566701, 1.5642369e-23, 6.5893758e-24, 5.5834599e-24], [78.744784, 1.5586486e-23, 6.5793842e-24, 5.5815625e-24], [78.923271, 1.5530907e-23, 6.5694623e-24, 5.5796752e-24], [79.102161, 1.5475632e-23, 6.5596097e-24, 5.5777981e-24], [79.281458, 1.542066e-23, 6.5498261e-24, 5.5759312e-24], [79.461161, 1.5365988e-23, 6.540111e-24, 5.5740742e-24], [79.641271, 1.5311617e-23, 6.5304641e-24, 5.5722273e-24], [79.821789, 1.5257544e-23, 6.5208848e-24, 5.5703904e-24], [80.002717, 1.5203768e-23, 6.5113728e-24, 5.5685634e-24], [80.184054, 1.5150289e-23, 6.5019278e-24, 5.5667463e-24], [80.365803, 1.5097104e-23, 6.4925492e-24, 5.564939e-24], [80.547964, 1.5044214e-23, 6.4832368e-24, 5.5631416e-24], [80.730537, 1.4991615e-23, 6.47399e-24, 5.561354e-24], [80.913525, 1.4939308e-23, 6.4648086e-24, 5.5595761e-24], [81.096927, 1.4887291e-23, 6.4556921e-24, 5.5578079e-24], [81.280745, 1.4835562e-23, 6.4466402e-24, 5.5560493e-24], [81.464979, 1.4784122e-23, 6.4376524e-24, 5.5543004e-24], [81.649631, 1.4732967e-23, 6.4287284e-24, 5.5525611e-24], [81.834702, 1.4682098e-23, 6.4198677e-24, 5.5508313e-24], [82.020192, 1.4631513e-23, 6.4110701e-24, 5.5491111e-24], [82.206103, 1.4581211e-23, 6.4023351e-24, 5.5474003e-24], [82.392434, 1.453119e-23, 6.3936623e-24, 5.545699e-24], [82.579189, 1.448145e-23, 6.3850514e-24, 5.544007e-24], [82.766366, 1.4431989e-23, 6.376502e-24, 5.5423245e-24], [82.953968, 1.4382806e-23, 6.3680137e-24, 5.5406513e-24], [83.141995, 1.43339e-23, 6.3595861e-24, 5.5389874e-24], [83.330449, 1.428527e-23, 6.3512189e-24, 5.5373328e-24], [83.519329, 1.4236914e-23, 6.3429118e-24, 5.5356874e-24], [83.708638, 1.4188832e-23, 6.3346643e-24, 5.5340512e-24], [83.898375, 1.4141023e-23, 6.3264761e-24, 5.5324242e-24], [84.088543, 1.4093484e-23, 6.3183468e-24, 5.5308063e-24], [84.279142, 1.4046216e-23, 6.3102761e-24, 5.5291976e-24], [84.470173, 1.3999216e-23, 6.3022636e-24, 5.5275979e-24], [84.661636, 1.3952485e-23, 6.2943089e-24, 5.5260073e-24], [84.853534, 1.390602e-23, 6.2864118e-24, 5.5244256e-24], [85.045867, 1.385982e-23, 6.2785718e-24, 5.522853e-24], [85.238636, 1.3813885e-23, 6.2707886e-24, 5.5212893e-24], [85.431841, 1.3768213e-23, 6.2630618e-24, 5.5197346e-24], [85.625485, 1.3722804e-23, 6.2553912e-24, 5.5181887e-24], [85.819567, 1.3677655e-23, 6.2477763e-24, 5.5166517e-24], [86.01409, 1.3632767e-23, 6.2402169e-24, 5.5151236e-24], [86.209053, 1.3588138e-23, 6.2327125e-24, 5.5136042e-24], [86.404458, 1.3543766e-23, 6.2252629e-24, 5.5120937e-24], [86.600306, 1.3499651e-23, 6.2178677e-24, 5.5105918e-24], [86.796598, 1.3455791e-23, 6.2105265e-24, 5.5090988e-24], [86.993335, 1.3412187e-23, 6.2032391e-24, 5.5076144e-24], [87.190518, 1.3368835e-23, 6.1960051e-24, 5.5061386e-24], [87.388148, 1.3325736e-23, 6.1888242e-24, 5.5046716e-24], [87.586226, 1.3282889e-23, 6.181696e-24, 5.5032131e-24], [87.784753, 1.3240291e-23, 6.1746203e-24, 5.5017633e-24], [87.983729, 1.3197943e-23, 6.1675966e-24, 5.500322e-24], [88.183157, 1.3155843e-23, 6.1606248e-24, 5.4988892e-24], [88.383037, 1.311399e-23, 6.1537044e-24, 5.497465e-24], [88.58337, 1.3072383e-23, 6.1468351e-24, 5.4960492e-24], [88.784157, 1.3031022e-23, 6.1400167e-24, 5.4946419e-24], [88.985399, 1.2989904e-23, 6.1332488e-24, 5.4932431e-24], [89.187097, 1.2949029e-23, 6.126531e-24, 5.4918527e-24], [89.389252, 1.2908395e-23, 6.1198632e-24, 5.4904707e-24], [89.591866, 1.2868003e-23, 6.1132449e-24, 5.489097e-24], [89.794939, 1.282785e-23, 6.1066759e-24, 5.4877317e-24], [89.998472, 1.2787937e-23, 6.1001559e-24, 5.4863748e-24], [90.202466, 1.2748261e-23, 6.0936846e-24, 5.4850261e-24], [90.406923, 1.2708821e-23, 6.0872615e-24, 5.4836857e-24], [90.611843, 1.2669618e-23, 6.0808866e-24, 5.4823536e-24], [90.817228, 1.2630649e-23, 6.0745594e-24, 5.4810297e-24], [91.023079, 1.2591914e-23, 6.0682797e-24, 5.4797141e-24], [91.229395, 1.2553411e-23, 6.0620471e-24, 5.4784066e-24], [91.43618, 1.251514e-23, 6.0558614e-24, 5.4771073e-24], [91.643433, 1.24771e-23, 6.0497223e-24, 5.4758162e-24], [91.851156, 1.243929e-23, 6.0436295e-24, 5.4745332e-24], [92.05935, 1.2401708e-23, 6.0375827e-24, 5.4732583e-24], [92.268016, 1.2364354e-23, 6.0315817e-24, 5.4719916e-24], [92.477155, 1.2327227e-23, 6.025626e-24, 5.4707329e-24], [92.686767, 1.2290325e-23, 6.0197155e-24, 5.4694822e-24], [92.896855, 1.2253648e-23, 6.0138499e-24, 5.4682396e-24], [93.107419, 1.2217195e-23, 6.0080289e-24, 5.4670051e-24], [93.318461, 1.2180965e-23, 6.0022521e-24, 5.4657785e-24], [93.52998, 1.2144957e-23, 5.9965195e-24, 5.4645599e-24], [93.741979, 1.2109169e-23, 5.9908306e-24, 5.4633493e-24], [93.954459, 1.2073602e-23, 5.9851851e-24, 5.4621466e-24], [94.16742, 1.2038253e-23, 5.9795829e-24, 5.4609519e-24], [94.380864, 1.2003123e-23, 5.9740237e-24, 5.4597651e-24], [94.594792, 1.1968209e-23, 5.9685071e-24, 5.4585862e-24], [94.809205, 1.1933511e-23, 5.963033e-24, 5.4574152e-24], [95.024104, 1.1899029e-23, 5.957601e-24, 5.456252e-24], [95.239489, 1.1864761e-23, 5.9522109e-24, 5.4550967e-24], [95.455363, 1.1830706e-23, 5.9468624e-24, 5.4539493e-24], [95.671727, 1.1796864e-23, 5.9415553e-24, 5.4528096e-24], [95.88858, 1.1763232e-23, 5.9362893e-24, 5.4516778e-24], [96.105926, 1.1729812e-23, 5.9310642e-24, 5.4505538e-24], [96.323764, 1.16966e-23, 5.9258797e-24, 5.4494375e-24], [96.542095, 1.1663598e-23, 5.9207355e-24, 5.448329e-24], [96.760922, 1.1630803e-23, 5.9156315e-24, 5.4472282e-24], [96.980244, 1.1598214e-23, 5.9105672e-24, 5.4461352e-24], [97.200064, 1.1565832e-23, 5.9055426e-24, 5.4450499e-24], [97.420382, 1.1533654e-23, 5.9005574e-24, 5.4439723e-24], [97.641199, 1.1501681e-23, 5.8956113e-24, 5.4429024e-24], [97.862517, 1.146991e-23, 5.890704e-24, 5.4418402e-24], [98.084336, 1.1438342e-23, 5.8858354e-24, 5.4407857e-24], [98.306659, 1.1406975e-23, 5.8810051e-24, 5.4397388e-24], [98.529485, 1.1375808e-23, 5.8762131e-24, 5.4386996e-24], [98.752816, 1.1344841e-23, 5.8714589e-24, 5.437668e-24], [98.976653, 1.1314073e-23, 5.8667424e-24, 5.436644e-24], [99.200998, 1.1283502e-23, 5.8620634e-24, 5.4356276e-24], [99.425852, 1.1253128e-23, 5.8574216e-24, 5.4346188e-24], [99.651215, 1.1222949e-23, 5.8528168e-24, 5.4336177e-24], [99.877088, 1.1192966e-23, 5.8482487e-24, 5.432624e-24], [100.10347, 1.1163177e-23, 5.8437172e-24, 5.431638e-24], [100.33037, 1.1133581e-23, 5.8392219e-24, 5.4306595e-24], [100.55779, 1.1104178e-23, 5.8347628e-24, 5.4296886e-24], [100.78571, 1.1074966e-23, 5.8303395e-24, 5.4287251e-24], [101.01416, 1.1045945e-23, 5.8259519e-24, 5.4277693e-24], [101.24312, 1.1017113e-23, 5.8215997e-24, 5.4268209e-24], [101.47261, 1.098847e-23, 5.8172827e-24, 5.42588e-24], [101.70261, 1.0960016e-23, 5.8130007e-24, 5.4249467e-24], [101.93313, 1.0931748e-23, 5.8087534e-24, 5.4240208e-24], [102.16418, 1.0903667e-23, 5.8045407e-24, 5.4231024e-24], [102.39575, 1.0875771e-23, 5.8003624e-24, 5.4221915e-24], [102.62784, 1.084806e-23, 5.7962182e-24, 5.421288e-24], [102.86046, 1.0820533e-23, 5.7921079e-24, 5.420392e-24], [103.09361, 1.0793188e-23, 5.7880313e-24, 5.4195034e-24], [103.32729, 1.0766026e-23, 5.7839883e-24, 5.4186223e-24], [103.56149, 1.0739045e-23, 5.7799786e-24, 5.4177486e-24], [103.79623, 1.0712244e-23, 5.776002e-24, 5.4168823e-24], [104.0315, 1.0685623e-23, 5.7720583e-24, 5.4160235e-24], [104.2673, 1.065918e-23, 5.7681473e-24, 5.4151721e-24], [104.50364, 1.0632915e-23, 5.7642688e-24, 5.414328e-24], [104.74051, 1.0606828e-23, 5.7604226e-24, 5.4134914e-24], [104.97792, 1.0580916e-23, 5.7566086e-24, 5.4126621e-24], [105.21587, 1.055518e-23, 5.7528265e-24, 5.4118403e-24], [105.45435, 1.0529618e-23, 5.7490761e-24, 5.4110258e-24], [105.69338, 1.0504231e-23, 5.7453572e-24, 5.4102187e-24], [105.93295, 1.0479016e-23, 5.7416697e-24, 5.409419e-24], [106.17306, 1.0453973e-23, 5.7380133e-24, 5.4086266e-24], [106.41372, 1.0429101e-23, 5.734388e-24, 5.4078416e-24], [106.65492, 1.04044e-23, 5.7307934e-24, 5.4070639e-24], [106.89667, 1.0379869e-23, 5.7272294e-24, 5.4062936e-24], [107.13897, 1.0355506e-23, 5.7236958e-24, 5.4055306e-24], [107.38181, 1.0331312e-23, 5.7201925e-24, 5.404775e-24], [107.62521, 1.0307285e-23, 5.7167192e-24, 5.4040267e-24], [107.86916, 1.0283424e-23, 5.7132759e-24, 5.4032857e-24], [108.11366, 1.0259729e-23, 5.7098622e-24, 5.4025521e-24], [108.35871, 1.0236199e-23, 5.706478e-24, 5.4018258e-24], [108.60432, 1.0212833e-23, 5.7031232e-24, 5.4011068e-24], [108.85049, 1.018963e-23, 5.6997976e-24, 5.4003951e-24], [109.09722, 1.0166589e-23, 5.696501e-24, 5.3996907e-24], [109.3445, 1.0143711e-23, 5.6932332e-24, 5.3989937e-24], [109.59235, 1.0120993e-23, 5.6899941e-24, 5.398304e-24], [109.84075, 1.0098435e-23, 5.6867834e-24, 5.3976215e-24], [110.08972, 1.0076037e-23, 5.6836012e-24, 5.3969464e-24], [110.33926, 1.0053798e-23, 5.680447e-24, 5.3962786e-24], [110.58936, 1.0031716e-23, 5.6773209e-24, 5.395618e-24], [110.84002, 1.0009791e-23, 5.6742226e-24, 5.3949648e-24], [111.09126, 9.988023e-24, 5.671152e-24, 5.3943189e-24], [111.34306, 9.9664102e-24, 5.6681089e-24, 5.3936803e-24], [111.59544, 9.9449521e-24, 5.6650932e-24, 5.3930489e-24], [111.84839, 9.9236481e-24, 5.6621046e-24, 5.3924249e-24], [112.10191, 9.9024973e-24, 5.6591431e-24, 5.3918081e-24], [112.356, 9.881499e-24, 5.6562085e-24, 5.3911986e-24], [112.61067, 9.8606525e-24, 5.6533006e-24, 5.3905965e-24], [112.86592, 9.8399569e-24, 5.6504193e-24, 5.3900016e-24], [113.12175, 9.8194116e-24, 5.6475644e-24, 5.389414e-24], [113.37815, 9.7990158e-24, 5.6447358e-24, 5.3888337e-24], [113.63514, 9.7787687e-24, 5.6419334e-24, 5.3882607e-24], [113.89271, 9.7586697e-24, 5.6391569e-24, 5.387695e-24], [114.15087, 9.7387179e-24, 5.6364062e-24, 5.3871365e-24], [114.40961, 9.7189127e-24, 5.6336812e-24, 5.3865854e-24], [114.66893, 9.6992533e-24, 5.6309818e-24, 5.3860415e-24], [114.92885, 9.6797389e-24, 5.6283078e-24, 5.385505e-24], [115.18935, 9.6603689e-24, 5.625659e-24, 5.3849757e-24], [115.45044, 9.6411426e-24, 5.6230353e-24, 5.3844538e-24], [115.71213, 9.6220591e-24, 5.6204366e-24, 5.3839391e-24], [115.9744, 9.6031178e-24, 5.6178628e-24, 5.3834317e-24], [116.23728, 9.5843179e-24, 5.6153136e-24, 5.3829317e-24], [116.50075, 9.5656588e-24, 5.612789e-24, 5.3824389e-24], [116.76481, 9.5471398e-24, 5.6102889e-24, 5.3819535e-24], [117.02948, 9.52876e-24, 5.607813e-24, 5.3814753e-24], [117.29474, 9.5105188e-24, 5.6053613e-24, 5.3810045e-24], [117.5606, 9.4924155e-24, 5.6029336e-24, 5.3805409e-24], [117.82707, 9.4744494e-24, 5.6005298e-24, 5.3800847e-24], [118.09414, 9.4566198e-24, 5.5981498e-24, 5.3796358e-24], [118.36182, 9.438926e-24, 5.5957935e-24, 5.3791942e-24], [118.63011, 9.4213673e-24, 5.5934606e-24, 5.37876e-24], [118.899, 9.4039429e-24, 5.5911511e-24, 5.3783331e-24], [119.1685, 9.3866523e-24, 5.5888649e-24, 5.3779135e-24], [119.43861, 9.3694947e-24, 5.5866019e-24, 5.3775012e-24], [119.70934, 9.3524694e-24, 5.5843618e-24, 5.3770963e-24], [119.98068, 9.3355757e-24, 5.5821447e-24, 5.3766987e-24], [120.25263, 9.318813e-24, 5.5799503e-24, 5.3763085e-24], [120.5252, 9.3021805e-24, 5.5777785e-24, 5.3759256e-24], [120.79839, 9.2856776e-24, 5.5756293e-24, 5.3755501e-24], [121.07219, 9.2693037e-24, 5.5735026e-24, 5.3751819e-24], [121.34662, 9.253058e-24, 5.5713981e-24, 5.3748211e-24], [121.62167, 9.2369399e-24, 5.5693158e-24, 5.3744677e-24], [121.89734, 9.2209486e-24, 5.5672555e-24, 5.3741216e-24], [122.17364, 9.2050836e-24, 5.5652173e-24, 5.3737829e-24], [122.45057, 9.1893442e-24, 5.5632008e-24, 5.3734516e-24], [122.72812, 9.1737297e-24, 5.5612061e-24, 5.3731277e-24], [123.0063, 9.1582394e-24, 5.559233e-24, 5.3728112e-24], [123.28511, 9.1428727e-24, 5.5572815e-24, 5.3725021e-24], [123.56455, 9.127629e-24, 5.5553513e-24, 5.3722004e-24], [123.84463, 9.1125076e-24, 5.5534425e-24, 5.3719062e-24], [124.12534, 9.0975077e-24, 5.5515548e-24, 5.3716193e-24], [124.40669, 9.0826289e-24, 5.5496882e-24, 5.3713399e-24], [124.68868, 9.0678705e-24, 5.5478426e-24, 5.3710679e-24], [124.9713, 9.0532317e-24, 5.5460178e-24, 5.3708033e-24], [125.25457, 9.038712e-24, 5.5442139e-24, 5.3705462e-24], [125.53847, 9.0243107e-24, 5.5424306e-24, 5.3702965e-24], [125.82302, 9.0100273e-24, 5.5406679e-24, 5.3700543e-24], [126.10822, 8.9958609e-24, 5.5389256e-24, 5.3698196e-24], [126.39406, 8.9818112e-24, 5.5372037e-24, 5.3695923e-24], [126.68055, 8.9678773e-24, 5.5355021e-24, 5.3693725e-24], [126.96769, 8.9540587e-24, 5.5338207e-24, 5.3691602e-24], [127.25548, 8.9403548e-24, 5.5321594e-24, 5.3689555e-24], [127.54393, 8.9267649e-24, 5.530518e-24, 5.3687582e-24], [127.83302, 8.9132884e-24, 5.5288966e-24, 5.3685684e-24], [128.12277, 8.8999247e-24, 5.5272949e-24, 5.3683862e-24], [128.41318, 8.8866733e-24, 5.5257129e-24, 5.3682114e-24], [128.70425, 8.8735334e-24, 5.5241506e-24, 5.3680443e-24], [128.99598, 8.8605045e-24, 5.5226078e-24, 5.3678846e-24], [129.28836, 8.8475859e-24, 5.5210844e-24, 5.3677326e-24], [129.58142, 8.8347772e-24, 5.5195804e-24, 5.3675881e-24], [129.87513, 8.8220776e-24, 5.5180956e-24, 5.3674511e-24], [130.16951, 8.8094866e-24, 5.51663e-24, 5.3673218e-24], [130.46456, 8.7970035e-24, 5.5151835e-24, 5.3672e-24], [130.76028, 8.7846279e-24, 5.513756e-24, 5.3670859e-24], [131.05666, 8.772359e-24, 5.5123474e-24, 5.3669794e-24], [131.35372, 8.7601964e-24, 5.5109577e-24, 5.3668804e-24], [131.65145, 8.7481394e-24, 5.5095867e-24, 5.3667892e-24], [131.94986, 8.7361874e-24, 5.5082343e-24, 5.3667055e-24], [132.24894, 8.7243399e-24, 5.5069006e-24, 5.3666295e-24], [132.5487, 8.7125963e-24, 5.5055854e-24, 5.3665612e-24], [132.84915, 8.700956e-24, 5.5042886e-24, 5.3665006e-24], [133.15027, 8.6894185e-24, 5.5030101e-24, 5.3664476e-24], [133.45207, 8.6779831e-24, 5.5017499e-24, 5.3664023e-24], [133.75456, 8.6666493e-24, 5.5005079e-24, 5.3663648e-24], [134.05773, 8.6554166e-24, 5.499284e-24, 5.3663349e-24], [134.36159, 8.6442844e-24, 5.4980782e-24, 5.3663128e-24], [134.66614, 8.6332521e-24, 5.4968904e-24, 5.3662984e-24], [134.97138, 8.6223191e-24, 5.4957204e-24, 5.3662918e-24], [135.27732, 8.611485e-24, 5.4945683e-24, 5.3662929e-24], [135.58394, 8.6007491e-24, 5.4934339e-24, 5.3663018e-24], [135.89126, 8.5901109e-24, 5.4923172e-24, 5.3663185e-24], [136.19928, 8.5795699e-24, 5.4912181e-24, 5.366343e-24], [136.508, 8.5691255e-24, 5.4901366e-24, 5.3663753e-24], [136.81741, 8.5587772e-24, 5.4890725e-24, 5.3664154e-24], [137.12753, 8.5485245e-24, 5.4880258e-24, 5.3664633e-24], [137.43835, 8.5383667e-24, 5.4869965e-24, 5.3665191e-24], [137.74987, 8.5283034e-24, 5.4859844e-24, 5.3665828e-24], [138.0621, 8.5183341e-24, 5.4849895e-24, 5.3666543e-24], [138.37504, 8.5084582e-24, 5.4840118e-24, 5.3667337e-24], [138.68868, 8.4986751e-24, 5.4830511e-24, 5.366821e-24], [139.00304, 8.4889845e-24, 5.4821075e-24, 5.3669162e-24], [139.31811, 8.4793857e-24, 5.4811808e-24, 5.3670194e-24], [139.6339, 8.4698782e-24, 5.480271e-24, 5.3671304e-24], [139.9504, 8.4604615e-24, 5.479378e-24, 5.3672495e-24], [140.26761, 8.4511351e-24, 5.4785018e-24, 5.3673764e-24], [140.58555, 8.4418985e-24, 5.4776423e-24, 5.3675114e-24], [140.90421, 8.4327512e-24, 5.4767994e-24, 5.3676544e-24], [141.22359, 8.4236927e-24, 5.4759731e-24, 5.3678053e-24], [141.54369, 8.4147224e-24, 5.4751634e-24, 5.3679643e-24], [141.86452, 8.4058399e-24, 5.4743701e-24, 5.3681313e-24], [142.18608, 8.3970447e-24, 5.4735932e-24, 5.3683064e-24], [142.50836, 8.3883363e-24, 5.4728327e-24, 5.3684895e-24], [142.83138, 8.3797142e-24, 5.4720885e-24, 5.3686808e-24], [143.15513, 8.3711779e-24, 5.4713605e-24, 5.3688801e-24], [143.47961, 8.3627269e-24, 5.4706487e-24, 5.3690875e-24], [143.80482, 8.3543607e-24, 5.4699531e-24, 5.369303e-24], [144.13078, 8.3460789e-24, 5.4692735e-24, 5.3695267e-24], [144.45747, 8.3378809e-24, 5.46861e-24, 5.3697585e-24], [144.78491, 8.3297663e-24, 5.4679625e-24, 5.3699985e-24], [145.11308, 8.3217347e-24, 5.4673309e-24, 5.3702466e-24], [145.442, 8.3137855e-24, 5.4667152e-24, 5.370503e-24], [145.77167, 8.3059183e-24, 5.4661154e-24, 5.3707676e-24], [146.10208, 8.2981326e-24, 5.4655313e-24, 5.3710404e-24], [146.43324, 8.2904279e-24, 5.464963e-24, 5.3713215e-24], [146.76515, 8.2828039e-24, 5.4644104e-24, 5.3716108e-24], [147.09782, 8.2752599e-24, 5.4638734e-24, 5.3719084e-24], [147.43123, 8.2677957e-24, 5.463352e-24, 5.3722143e-24], [147.76541, 8.2604106e-24, 5.4628462e-24, 5.3725285e-24], [148.10034, 8.2531044e-24, 5.4623558e-24, 5.3728511e-24], [148.43603, 8.2458764e-24, 5.461881e-24, 5.373182e-24], [148.77248, 8.2387264e-24, 5.4614216e-24, 5.3735212e-24], [149.1097, 8.2316537e-24, 5.4609775e-24, 5.3738689e-24], [149.44768, 8.2246581e-24, 5.4605488e-24, 5.3742249e-24], [149.78642, 8.217739e-24, 5.4601354e-24, 5.3745894e-24], [150.12593, 8.2108961e-24, 5.4597372e-24, 5.3749622e-24], [150.46621, 8.2041288e-24, 5.4593543e-24, 5.3753436e-24], [150.80727, 8.1974368e-24, 5.4589865e-24, 5.3757334e-24], [151.14909, 8.1908196e-24, 5.4586339e-24, 5.3761316e-24], [151.4917, 8.1842768e-24, 5.4582964e-24, 5.3765384e-24], [151.83507, 8.177808e-24, 5.4579739e-24, 5.3769537e-24], [152.17923, 8.1714128e-24, 5.4576664e-24, 5.3773776e-24], [152.52417, 8.1650907e-24, 5.457374e-24, 5.37781e-24], [152.86988, 8.1588414e-24, 5.4570965e-24, 5.378251e-24], [153.21638, 8.1526643e-24, 5.4568339e-24, 5.3787005e-24], [153.56367, 8.1465592e-24, 5.4565861e-24, 5.3791587e-24], [153.91175, 8.1405256e-24, 5.4563533e-24, 5.3796255e-24], [154.26061, 8.1345631e-24, 5.4561352e-24, 5.380101e-24], [154.61026, 8.1286712e-24, 5.4559319e-24, 5.3805851e-24], [154.96071, 8.1228497e-24, 5.4557433e-24, 5.381078e-24], [155.31195, 8.1170981e-24, 5.4555695e-24, 5.3815795e-24], [155.66399, 8.1114159e-24, 5.4554103e-24, 5.3820898e-24], [156.01682, 8.1058029e-24, 5.4552658e-24, 5.3826088e-24], [156.37046, 8.1002586e-24, 5.4551359e-24, 5.3831366e-24], [156.72489, 8.0947826e-24, 5.4550206e-24, 5.3836731e-24], [157.08013, 8.0893746e-24, 5.4549198e-24, 5.3842185e-24], [157.43618, 8.0840342e-24, 5.4548336e-24, 5.3847727e-24], [157.79303, 8.0787609e-24, 5.4547619e-24, 5.3853357e-24], [158.15069, 8.0735544e-24, 5.4547046e-24, 5.3859076e-24], [158.50916, 8.0684144e-24, 5.4546618e-24, 5.3864884e-24], [158.86844, 8.0633405e-24, 5.4546334e-24, 5.3870782e-24], [159.22854, 8.0583322e-24, 5.4546194e-24, 5.3876768e-24], [159.58946, 8.0533893e-24, 5.4546197e-24, 5.3882844e-24], [159.95119, 8.0485113e-24, 5.4546344e-24, 5.3889009e-24], [160.31374, 8.0436979e-24, 5.4546634e-24, 5.3895265e-24], [160.67712, 8.0389488e-24, 5.4547067e-24, 5.390161e-24], [161.04131, 8.0342636e-24, 5.4547643e-24, 5.3908046e-24], [161.40634, 8.0296419e-24, 5.454836e-24, 5.3914573e-24], [161.77219, 8.0250833e-24, 5.4549221e-24, 5.392119e-24], [162.13887, 8.0205876e-24, 5.4550223e-24, 5.3927898e-24], [162.50638, 8.0161543e-24, 5.4551366e-24, 5.3934698e-24], [162.87472, 8.0117832e-24, 5.4552652e-24, 5.3941588e-24], [163.2439, 8.0074739e-24, 5.4554078e-24, 5.3948571e-24], [163.61392, 8.003226e-24, 5.4555646e-24, 5.3955645e-24], [163.98477, 7.9990393e-24, 5.4557354e-24, 5.3962812e-24], [164.35647, 7.9949133e-24, 5.4559203e-24, 5.397007e-24], [164.729, 7.9908477e-24, 5.4561193e-24, 5.3977422e-24], [165.10239, 7.9868423e-24, 5.4563323e-24, 5.3984866e-24], [165.47661, 7.9828966e-24, 5.4565593e-24, 5.3992403e-24], [165.85169, 7.9790104e-24, 5.4568003e-24, 5.4000033e-24], [166.22762, 7.9751834e-24, 5.4570553e-24, 5.4007756e-24], [166.6044, 7.9714151e-24, 5.4573243e-24, 5.4015574e-24], [166.98203, 7.9677053e-24, 5.4576072e-24, 5.4023485e-24], [167.36052, 7.9640537e-24, 5.457904e-24, 5.403149e-24], [167.73986, 7.96046e-24, 5.4582147e-24, 5.403959e-24], [168.12007, 7.9569238e-24, 5.4585393e-24, 5.4047784e-24], [168.50114, 7.9534448e-24, 5.4588779e-24, 5.4056073e-24], [168.88307, 7.9500228e-24, 5.4592302e-24, 5.4064457e-24], [169.26587, 7.9466574e-24, 5.4595965e-24, 5.4072937e-24], [169.64953, 7.9433484e-24, 5.4599765e-24, 5.4081512e-24], [170.03407, 7.9400953e-24, 5.4603704e-24, 5.4090183e-24], [170.41948, 7.936898e-24, 5.4607782e-24, 5.409895e-24], [170.80576, 7.9337561e-24, 5.4611997e-24, 5.4107813e-24], [171.19291, 7.9306693e-24, 5.461635e-24, 5.4116773e-24], [171.58095, 7.9276374e-24, 5.4620841e-24, 5.4125829e-24], [171.96986, 7.92466e-24, 5.462547e-24, 5.4134983e-24], [172.35965, 7.9217369e-24, 5.4630236e-24, 5.4144234e-24], [172.75033, 7.9188678e-24, 5.463514e-24, 5.4153582e-24], [173.14189, 7.9160523e-24, 5.4640181e-24, 5.4163028e-24], [173.53435, 7.9132903e-24, 5.4645359e-24, 5.4172572e-24], [173.92769, 7.9105815e-24, 5.4650675e-24, 5.4182215e-24], [174.32192, 7.9079255e-24, 5.4656128e-24, 5.4191956e-24], [174.71704, 7.905322e-24, 5.4661717e-24, 5.4201796e-24], [175.11306, 7.9027709e-24, 5.4667444e-24, 5.4211734e-24], [175.50998, 7.9002719e-24, 5.4673308e-24, 5.4221772e-24], [175.9078, 7.8978246e-24, 5.4679309e-24, 5.423191e-24], [176.30652, 7.8954289e-24, 5.4685446e-24, 5.4242148e-24], [176.70615, 7.8930844e-24, 5.469172e-24, 5.4252485e-24], [177.10668, 7.8907909e-24, 5.4698131e-24, 5.4262923e-24], [177.50811, 7.8885482e-24, 5.4704678e-24, 5.4273462e-24], [177.91046, 7.8863559e-24, 5.4711362e-24, 5.4284101e-24], [178.31372, 7.8842139e-24, 5.4718183e-24, 5.4294842e-24], [178.7179, 7.8821218e-24, 5.472514e-24, 5.4305684e-24], [179.12299, 7.8800794e-24, 5.4732233e-24, 5.4316628e-24], [179.52899, 7.8780866e-24, 5.4739463e-24, 5.4327673e-24], [179.93592, 7.8761429e-24, 5.4746829e-24, 5.4338821e-24], [180.34377, 7.8742483e-24, 5.4754331e-24, 5.4350072e-24], [180.75255, 7.8724024e-24, 5.476197e-24, 5.4361425e-24], [181.16225, 7.870605e-24, 5.4769746e-24, 5.4372881e-24], [181.57288, 7.8688558e-24, 5.4777657e-24, 5.4384441e-24], [181.98444, 7.8671547e-24, 5.4785705e-24, 5.4396104e-24], [182.39693, 7.8655015e-24, 5.4793889e-24, 5.4407871e-24], [182.81036, 7.8638957e-24, 5.4802209e-24, 5.4419742e-24], [183.22473, 7.8623374e-24, 5.4810666e-24, 5.4431718e-24], [183.64003, 7.8608261e-24, 5.4819259e-24, 5.4443798e-24], [184.05628, 7.8593617e-24, 5.4827988e-24, 5.4455984e-24], [184.47347, 7.857944e-24, 5.4836854e-24, 5.4468274e-24], [184.89161, 7.8565727e-24, 5.4845856e-24, 5.4480671e-24], [185.31069, 7.8552477e-24, 5.4854994e-24, 5.4493173e-24], [185.73072, 7.8539686e-24, 5.4864269e-24, 5.4505781e-24], [186.15171, 7.8527354e-24, 5.487368e-24, 5.4518496e-24], [186.57365, 7.8515477e-24, 5.4883227e-24, 5.4531317e-24], [186.99654, 7.8504054e-24, 5.4892911e-24, 5.4544246e-24], [187.4204, 7.8493083e-24, 5.4902731e-24, 5.4557282e-24], [187.84521, 7.8482561e-24, 5.4912688e-24, 5.4570425e-24], [188.27099, 7.8472487e-24, 5.4922782e-24, 5.4583677e-24], [188.69774, 7.8462858e-24, 5.4933012e-24, 5.4597036e-24], [189.12545, 7.8453672e-24, 5.4943379e-24, 5.4610505e-24], [189.55413, 7.8444928e-24, 5.4953882e-24, 5.4624082e-24], [189.98378, 7.8436624e-24, 5.4964522e-24, 5.4637768e-24], [190.4144, 7.8428756e-24, 5.4975299e-24, 5.4651563e-24], [190.846, 7.8421325e-24, 5.4986213e-24, 5.4665468e-24], [191.27858, 7.8414327e-24, 5.4997264e-24, 5.4679484e-24], [191.71214, 7.840776e-24, 5.5008452e-24, 5.4693609e-24], [192.14669, 7.8401624e-24, 5.5019777e-24, 5.4707845e-24], [192.58222, 7.8395916e-24, 5.503124e-24, 5.4722193e-24], [193.01873, 7.8390633e-24, 5.5042839e-24, 5.4736651e-24], [193.45624, 7.8385775e-24, 5.5054576e-24, 5.4751221e-24], [193.89473, 7.838134e-24, 5.506645e-24, 5.4765903e-24], [194.33422, 7.8377325e-24, 5.5078462e-24, 5.4780697e-24], [194.77471, 7.8373729e-24, 5.5090611e-24, 5.4795603e-24], [195.21619, 7.837055e-24, 5.5102898e-24, 5.4810622e-24], [195.65868, 7.8367786e-24, 5.5115323e-24, 5.4825755e-24], [196.10217, 7.8365437e-24, 5.5127886e-24, 5.4841001e-24], [196.54666, 7.8363499e-24, 5.5140587e-24, 5.485636e-24], [196.99216, 7.8361972e-24, 5.5153426e-24, 5.4871834e-24], [197.43867, 7.8360853e-24, 5.5166403e-24, 5.4887422e-24], [197.8862, 7.8360141e-24, 5.5179518e-24, 5.4903125e-24], [198.33473, 7.8359835e-24, 5.5192772e-24, 5.4918943e-24], [198.78429, 7.8359932e-24, 5.5206164e-24, 5.4934876e-24], [199.23486, 7.8360431e-24, 5.5219695e-24, 5.4950925e-24], [199.68646, 7.8361332e-24, 5.5233365e-24, 5.496709e-24], [200.13907, 7.8362631e-24, 5.5247174e-24, 5.4983372e-24], [200.59272, 7.8364327e-24, 5.5261122e-24, 5.499977e-24], [201.04739, 7.8366419e-24, 5.5275209e-24, 5.5016285e-24], [201.50309, 7.8368906e-24, 5.5289435e-24, 5.5032917e-24], [201.95983, 7.8371786e-24, 5.5303801e-24, 5.5049667e-24], [202.4176, 7.8375057e-24, 5.5318306e-24, 5.5066536e-24], [202.87641, 7.8378718e-24, 5.5332951e-24, 5.5083522e-24], [203.33626, 7.8382767e-24, 5.5347736e-24, 5.5100627e-24], [203.79715, 7.8387204e-24, 5.5362662e-24, 5.5117851e-24], [204.25908, 7.8392026e-24, 5.5377727e-24, 5.5135195e-24], [204.72207, 7.8397232e-24, 5.5392933e-24, 5.5152658e-24], [205.1861, 7.8402822e-24, 5.5408279e-24, 5.5170242e-24], [205.65118, 7.8408792e-24, 5.5423766e-24, 5.5187945e-24], [206.11732, 7.8415143e-24, 5.5439393e-24, 5.520577e-24], [206.58451, 7.8421873e-24, 5.5455162e-24, 5.5223715e-24], [207.05277, 7.842898e-24, 5.5471072e-24, 5.5241782e-24], [207.52208, 7.8436464e-24, 5.5487123e-24, 5.5259971e-24], [207.99246, 7.8444322e-24, 5.5503316e-24, 5.5278282e-24], [208.46391, 7.8452554e-24, 5.5519651e-24, 5.5296715e-24], [208.93642, 7.8461158e-24, 5.5536127e-24, 5.5315271e-24], [209.41, 7.8470133e-24, 5.5552746e-24, 5.533395e-24], [209.88466, 7.8479478e-24, 5.5569506e-24, 5.5352753e-24], [210.3604, 7.8489192e-24, 5.558641e-24, 5.537168e-24], [210.83721, 7.8499273e-24, 5.5603455e-24, 5.5390731e-24], [211.3151, 7.8509721e-24, 5.5620644e-24, 5.5409906e-24], [211.79408, 7.8520534e-24, 5.5637976e-24, 5.5429206e-24], [212.27414, 7.853171e-24, 5.5655451e-24, 5.5448632e-24], [212.75529, 7.8543249e-24, 5.5673069e-24, 5.5468184e-24], [213.23753, 7.855515e-24, 5.5690831e-24, 5.5487861e-24], [213.72086, 7.8567412e-24, 5.5708737e-24, 5.5507665e-24], [214.20529, 7.8580033e-24, 5.5726786e-24, 5.5527596e-24], [214.69082, 7.8593013e-24, 5.5744981e-24, 5.5547653e-24], [215.17745, 7.8606349e-24, 5.5763319e-24, 5.5567838e-24], [215.66518, 7.8620042e-24, 5.5781802e-24, 5.5588152e-24], [216.15401, 7.863409e-24, 5.5800431e-24, 5.5608593e-24], [216.64396, 7.8648493e-24, 5.5819204e-24, 5.5629163e-24], [217.13501, 7.8663248e-24, 5.5838122e-24, 5.5649862e-24], [217.62718, 7.8678356e-24, 5.5857187e-24, 5.567069e-24], [218.12046, 7.8693815e-24, 5.5876397e-24, 5.5691648e-24], [218.61486, 7.8709624e-24, 5.5895753e-24, 5.5712736e-24], [219.11039, 7.8725782e-24, 5.5915255e-24, 5.5733954e-24], [219.60703, 7.8742289e-24, 5.5934904e-24, 5.5755303e-24], [220.1048, 7.8759143e-24, 5.59547e-24, 5.5776784e-24], [220.6037, 7.8776344e-24, 5.5974642e-24, 5.5798396e-24], [221.10373, 7.879389e-24, 5.5994732e-24, 5.582014e-24], [221.60489, 7.8811781e-24, 5.6014969e-24, 5.5842017e-24], [222.10719, 7.8830015e-24, 5.6035354e-24, 5.5864026e-24], [222.61063, 7.8848593e-24, 5.6055887e-24, 5.5886168e-24], [223.11521, 7.8867512e-24, 5.6076569e-24, 5.5908444e-24], [223.62093, 7.8886773e-24, 5.6097398e-24, 5.5930854e-24], [224.1278, 7.8906374e-24, 5.6118377e-24, 5.5953398e-24], [224.63582, 7.8926315e-24, 5.6139504e-24, 5.5976077e-24], [225.14499, 7.8946594e-24, 5.6160781e-24, 5.599889e-24], [225.65531, 7.8967212e-24, 5.6182207e-24, 5.602184e-24], [226.16679, 7.8988166e-24, 5.6203783e-24, 5.6044925e-24], [226.67943, 7.9009458e-24, 5.6225509e-24, 5.6068146e-24], [227.19323, 7.9031084e-24, 5.6247386e-24, 5.6091504e-24], [227.7082, 7.9053046e-24, 5.6269413e-24, 5.6114999e-24], [228.22433, 7.9075342e-24, 5.6291591e-24, 5.6138631e-24], [228.74164, 7.9097972e-24, 5.631392e-24, 5.6162402e-24], [229.26011, 7.9120934e-24, 5.6336401e-24, 5.618631e-24], [229.77976, 7.9144229e-24, 5.6359033e-24, 5.6210357e-24], [230.30059, 7.9167855e-24, 5.6381817e-24, 5.6234544e-24], [230.8226, 7.9191811e-24, 5.6404754e-24, 5.6258869e-24], [231.3458, 7.9216098e-24, 5.6427843e-24, 5.6283335e-24], [231.87017, 7.9240715e-24, 5.6451086e-24, 5.630794e-24], [232.39574, 7.926566e-24, 5.6474481e-24, 5.6332687e-24], [232.9225, 7.9290933e-24, 5.649803e-24, 5.6357574e-24], [233.45045, 7.9316534e-24, 5.6521732e-24, 5.6382603e-24], [233.9796, 7.9342462e-24, 5.6545589e-24, 5.6407774e-24], [234.50995, 7.9368716e-24, 5.65696e-24, 5.6433087e-24], [235.0415, 7.9395297e-24, 5.6593766e-24, 5.6458543e-24], [235.57425, 7.9422202e-24, 5.6618086e-24, 5.6484142e-24], [236.10822, 7.9449433e-24, 5.6642562e-24, 5.6509884e-24], [236.64339, 7.9476987e-24, 5.6667194e-24, 5.653577e-24], [237.17978, 7.9504865e-24, 5.6691981e-24, 5.6561801e-24], [237.71738, 7.9533066e-24, 5.6716925e-24, 5.6587977e-24], [238.2562, 7.956159e-24, 5.6742025e-24, 5.6614298e-24], [238.79624, 7.9590436e-24, 5.6767282e-24, 5.6640764e-24], [239.33751, 7.9619604e-24, 5.6792697e-24, 5.6667376e-24], [239.88, 7.9649093e-24, 5.6818268e-24, 5.6694135e-24], [240.42372, 7.9678902e-24, 5.6843998e-24, 5.6721041e-24], [240.96868, 7.9709032e-24, 5.6869886e-24, 5.6748094e-24], [241.51487, 7.9739481e-24, 5.6895932e-24, 5.6775295e-24], [242.0623, 7.977025e-24, 5.6922137e-24, 5.6802644e-24], [242.61097, 7.9801338e-24, 5.6948502e-24, 5.6830142e-24], [243.16088, 7.9832744e-24, 5.6975026e-24, 5.6857788e-24], [243.71204, 7.9864468e-24, 5.7001709e-24, 5.6885585e-24], [244.26445, 7.989651e-24, 5.7028553e-24, 5.6913531e-24], [244.81811, 7.992887e-24, 5.7055557e-24, 5.6941627e-24], [245.37302, 7.9961546e-24, 5.7082723e-24, 5.6969874e-24], [245.92919, 7.9994539e-24, 5.7110049e-24, 5.6998273e-24], [246.48663, 8.0027848e-24, 5.7137537e-24, 5.7026823e-24], [247.04533, 8.0061473e-24, 5.7165187e-24, 5.7055525e-24], [247.60529, 8.0095414e-24, 5.7192999e-24, 5.7084379e-24], [248.16652, 8.012967e-24, 5.7220973e-24, 5.7113387e-24], [248.72903, 8.016424e-24, 5.7249111e-24, 5.7142548e-24], [249.29281, 8.0199126e-24, 5.7277412e-24, 5.7171863e-24], [249.85787, 8.0234325e-24, 5.7305877e-24, 5.7201332e-24], [250.4242, 8.0269839e-24, 5.7334505e-24, 5.7230956e-24], [250.99183, 8.0305666e-24, 5.7363298e-24, 5.7260735e-24], [251.56074, 8.0341807e-24, 5.7392256e-24, 5.7290669e-24], [252.13093, 8.0378261e-24, 5.7421379e-24, 5.732076e-24], [252.70243, 8.0415028e-24, 5.7450668e-24, 5.7351007e-24], [253.27521, 8.0452107e-24, 5.7480122e-24, 5.7381412e-24], [253.8493, 8.0489499e-24, 5.7509742e-24, 5.7411973e-24], [254.42468, 8.0527204e-24, 5.753953e-24, 5.7442693e-24], [255.00137, 8.056522e-24, 5.7569484e-24, 5.7473571e-24], [255.57937, 8.0603548e-24, 5.7599605e-24, 5.7504607e-24], [256.15868, 8.0642188e-24, 5.7629894e-24, 5.7535803e-24], [256.7393, 8.0681139e-24, 5.7660352e-24, 5.7567159e-24], [257.32123, 8.0720402e-24, 5.7690978e-24, 5.7598675e-24], [257.90449, 8.0759975e-24, 5.7721772e-24, 5.7630351e-24], [258.48907, 8.079986e-24, 5.7752737e-24, 5.7662189e-24], [259.07497, 8.0840055e-24, 5.778387e-24, 5.7694188e-24], [259.6622, 8.0880561e-24, 5.7815174e-24, 5.7726349e-24], [260.25076, 8.0921377e-24, 5.7846648e-24, 5.7758673e-24], [260.84066, 8.0962504e-24, 5.7878293e-24, 5.7791159e-24], [261.43189, 8.1003941e-24, 5.791011e-24, 5.7823809e-24], [262.02446, 8.1045688e-24, 5.7942098e-24, 5.7856623e-24], [262.61838, 8.1087745e-24, 5.7974258e-24, 5.7889601e-24], [263.21364, 8.1130112e-24, 5.800659e-24, 5.7922744e-24], [263.81025, 8.1172789e-24, 5.8039095e-24, 5.7956053e-24], [264.40822, 8.1215775e-24, 5.8071774e-24, 5.7989527e-24], [265.00754, 8.1259072e-24, 5.8104626e-24, 5.8023167e-24], [265.60821, 8.1302678e-24, 5.8137652e-24, 5.8056974e-24], [266.21025, 8.1346593e-24, 5.8170853e-24, 5.8090948e-24], [266.81366, 8.1390818e-24, 5.8204228e-24, 5.812509e-24], [267.41843, 8.1435352e-24, 5.8237779e-24, 5.81594e-24], [268.02457, 8.1480196e-24, 5.8271506e-24, 5.8193879e-24], [268.63209, 8.152535e-24, 5.8305408e-24, 5.8228527e-24], [269.24098, 8.1570812e-24, 5.8339488e-24, 5.8263344e-24], [269.85125, 8.1616585e-24, 5.8373744e-24, 5.8298331e-24], [270.46291, 8.1662666e-24, 5.8408177e-24, 5.8333489e-24], [271.07595, 8.1709057e-24, 5.8442789e-24, 5.8368819e-24], [271.69039, 8.1755757e-24, 5.8477578e-24, 5.8404319e-24], [272.30621, 8.1802767e-24, 5.8512547e-24, 5.8439992e-24], [272.92343, 8.1850086e-24, 5.8547694e-24, 5.8475837e-24], [273.54205, 8.1897715e-24, 5.8583021e-24, 5.8511855e-24], [274.16208, 8.1945653e-24, 5.8618528e-24, 5.8548046e-24], [274.7835, 8.1993901e-24, 5.8654215e-24, 5.8584412e-24], [275.40634, 8.2042458e-24, 5.8690083e-24, 5.8620952e-24], [276.03059, 8.2091325e-24, 5.8726132e-24, 5.8657667e-24], [276.65625, 8.2140502e-24, 5.8762363e-24, 5.8694557e-24], [277.28333, 8.2189989e-24, 5.8798776e-24, 5.8731624e-24], [277.91183, 8.2239786e-24, 5.8835372e-24, 5.8768867e-24], [278.54176, 8.2289892e-24, 5.8872151e-24, 5.8806286e-24], [279.17312, 8.2340309e-24, 5.8909113e-24, 5.8843884e-24], [279.8059, 8.2391036e-24, 5.8946259e-24, 5.8881659e-24], [280.44012, 8.2442074e-24, 5.8983589e-24, 5.8919612e-24], [281.07578, 8.2493422e-24, 5.9021104e-24, 5.8957745e-24], [281.71288, 8.254508e-24, 5.9058805e-24, 5.8996057e-24], [282.35142, 8.2597049e-24, 5.9096691e-24, 5.9034549e-24], [282.99141, 8.2649329e-24, 5.9134763e-24, 5.9073222e-24], [283.63285, 8.270192e-24, 5.9173022e-24, 5.9112075e-24], [284.27575, 8.2754823e-24, 5.9211468e-24, 5.915111e-24], [284.9201, 8.2808036e-24, 5.9250101e-24, 5.9190327e-24], [285.56591, 8.2861561e-24, 5.9288923e-24, 5.9229727e-24], [286.21319, 8.2915398e-24, 5.9327933e-24, 5.9269309e-24], [286.86193, 8.2969546e-24, 5.9367132e-24, 5.9309075e-24], [287.51214, 8.3024007e-24, 5.940652e-24, 5.9349025e-24], [288.16383, 8.3078779e-24, 5.9446098e-24, 5.938916e-24], [288.817, 8.3133865e-24, 5.9485866e-24, 5.9429479e-24], [289.47164, 8.3189262e-24, 5.9525826e-24, 5.9469985e-24], [290.12777, 8.3244973e-24, 5.9565976e-24, 5.9510676e-24], [290.78539, 8.3300996e-24, 5.9606318e-24, 5.9551554e-24], [291.44449, 8.3357333e-24, 5.9646853e-24, 5.9592619e-24], [292.1051, 8.3413984e-24, 5.968758e-24, 5.9633872e-24], [292.76719, 8.3470948e-24, 5.97285e-24, 5.9675313e-24], [293.43079, 8.3528226e-24, 5.9769614e-24, 5.9716943e-24], [294.0959, 8.3585818e-24, 5.9810921e-24, 5.9758762e-24], [294.76251, 8.3643724e-24, 5.9852424e-24, 5.9800771e-24], [295.43063, 8.3701946e-24, 5.9894121e-24, 5.984297e-24], [296.10026, 8.3760482e-24, 5.9936015e-24, 5.988536e-24], [296.77142, 8.3819334e-24, 5.9978104e-24, 5.9927941e-24], [297.44409, 8.3878501e-24, 6.0020389e-24, 5.9970715e-24], [298.11829, 8.3937983e-24, 6.0062872e-24, 6.001368e-24], [298.79402, 8.3997782e-24, 6.0105552e-24, 6.0056839e-24], [299.47128, 8.4057898e-24, 6.0148431e-24, 6.0100191e-24], [300.15007, 8.411833e-24, 6.0191507e-24, 6.0143737e-24], [300.83041, 8.4179079e-24, 6.0234783e-24, 6.0187477e-24], [301.51228, 8.4240145e-24, 6.0278258e-24, 6.0231413e-24], [302.1957, 8.4301529e-24, 6.0321934e-24, 6.0275544e-24], [302.88067, 8.4363231e-24, 6.0365809e-24, 6.0319872e-24], [303.5672, 8.4425251e-24, 6.0409886e-24, 6.0364396e-24], [304.25528, 8.4487589e-24, 6.0454164e-24, 6.0409117e-24], [304.94491, 8.4550247e-24, 6.0498644e-24, 6.0454037e-24], [305.63612, 8.4613223e-24, 6.0543327e-24, 6.0499154e-24], [306.32888, 8.467652e-24, 6.0588212e-24, 6.0544471e-24], [307.02322, 8.4740136e-24, 6.0633301e-24, 6.0589987e-24], [307.71913, 8.4804072e-24, 6.0678594e-24, 6.0635702e-24], [308.41662, 8.486833e-24, 6.0724092e-24, 6.0681619e-24], [309.11569, 8.4932908e-24, 6.0769795e-24, 6.0727736e-24], [309.81635, 8.4997808e-24, 6.0815703e-24, 6.0774055e-24], [310.51859, 8.5063029e-24, 6.0861817e-24, 6.0820576e-24], [311.22243, 8.5128573e-24, 6.0908138e-24, 6.08673e-24], [311.92786, 8.5194439e-24, 6.0954666e-24, 6.0914227e-24], [312.63489, 8.5260628e-24, 6.1001402e-24, 6.0961358e-24], [313.34352, 8.532714e-24, 6.1048345e-24, 6.1008693e-24], [314.05376, 8.5393977e-24, 6.1095497e-24, 6.1056234e-24], [314.76561, 8.5461137e-24, 6.1142859e-24, 6.1103979e-24], [315.47907, 8.5528622e-24, 6.119043e-24, 6.1151931e-24], [316.19415, 8.5596432e-24, 6.1238212e-24, 6.120009e-24], [316.91085, 8.5664567e-24, 6.1286204e-24, 6.1248455e-24], [317.62917, 8.5733028e-24, 6.1334407e-24, 6.1297028e-24], [318.34912, 8.5801815e-24, 6.1382822e-24, 6.134581e-24], [319.07071, 8.5870929e-24, 6.143145e-24, 6.13948e-24], [319.79393, 8.5940371e-24, 6.148029e-24, 6.1444e-24], [320.51878, 8.6010139e-24, 6.1529344e-24, 6.149341e-24], [321.24529, 8.6080236e-24, 6.1578612e-24, 6.154303e-24], [321.97343, 8.6150661e-24, 6.1628094e-24, 6.1592862e-24], [322.70323, 8.6221416e-24, 6.1677791e-24, 6.1642905e-24], [323.43469, 8.6292499e-24, 6.1727704e-24, 6.169316e-24], [324.1678, 8.6363913e-24, 6.1777833e-24, 6.1743628e-24], [324.90257, 8.6435657e-24, 6.1828179e-24, 6.179431e-24], [325.63901, 8.6507731e-24, 6.1878741e-24, 6.1845205e-24], [326.37712, 8.6580137e-24, 6.1929522e-24, 6.1896316e-24], [327.1169, 8.6652875e-24, 6.198052e-24, 6.1947641e-24], [327.85835, 8.6725945e-24, 6.2031738e-24, 6.1999182e-24], [328.60149, 8.6799348e-24, 6.2083174e-24, 6.2050939e-24], [329.34632, 8.6873084e-24, 6.2134831e-24, 6.2102913e-24], [330.09283, 8.6947154e-24, 6.2186708e-24, 6.2155104e-24], [330.84103, 8.7021558e-24, 6.2238806e-24, 6.2207513e-24], [331.59093, 8.7096297e-24, 6.2291126e-24, 6.2260141e-24], [332.34253, 8.7171372e-24, 6.2343668e-24, 6.2312988e-24], [333.09583, 8.7246782e-24, 6.2396432e-24, 6.2366055e-24], [333.85084, 8.7322528e-24, 6.244942e-24, 6.2419342e-24], [334.60756, 8.7398612e-24, 6.2502631e-24, 6.247285e-24], [335.366, 8.7475033e-24, 6.2556067e-24, 6.252658e-24], [336.12615, 8.7551792e-24, 6.2609728e-24, 6.2580531e-24], [336.88803, 8.7628889e-24, 6.2663614e-24, 6.2634706e-24], [337.65163, 8.7706325e-24, 6.2717726e-24, 6.2689103e-24], [338.41697, 8.7784101e-24, 6.2772064e-24, 6.2743724e-24], [339.18404, 8.7862217e-24, 6.282663e-24, 6.279857e-24], [339.95285, 8.7940674e-24, 6.2881424e-24, 6.2853641e-24], [340.7234, 8.8019472e-24, 6.2936446e-24, 6.2908937e-24], [341.4957, 8.8098612e-24, 6.2991696e-24, 6.296446e-24], [342.26975, 8.8178095e-24, 6.3047176e-24, 6.302021e-24], [343.04555, 8.825792e-24, 6.3102886e-24, 6.3076186e-24], [343.82312, 8.8338089e-24, 6.3158827e-24, 6.3132391e-24], [344.60244, 8.8418602e-24, 6.3214999e-24, 6.3188825e-24], [345.38353, 8.849946e-24, 6.3271402e-24, 6.3245487e-24], [346.16639, 8.8580663e-24, 6.3328038e-24, 6.330238e-24], [346.95103, 8.8662212e-24, 6.3384907e-24, 6.3359503e-24], [347.73744, 8.8744107e-24, 6.3442009e-24, 6.3416856e-24], [348.52564, 8.882635e-24, 6.3499345e-24, 6.3474442e-24], [349.31562, 8.890894e-24, 6.3556916e-24, 6.3532259e-24], [350.1074, 8.8991879e-24, 6.3614722e-24, 6.359031e-24], [350.90097, 8.9075167e-24, 6.3672763e-24, 6.3648594e-24], [351.69633, 8.9158804e-24, 6.3731041e-24, 6.3707112e-24], [352.4935, 8.9242791e-24, 6.3789556e-24, 6.3765864e-24], [353.29248, 8.9327129e-24, 6.3848309e-24, 6.3824852e-24], [354.09327, 8.9411818e-24, 6.3907299e-24, 6.3884075e-24], [354.89587, 8.949686e-24, 6.3966528e-24, 6.3943535e-24], [355.70029, 8.9582254e-24, 6.4025997e-24, 6.4003232e-24], [356.50654, 8.9668002e-24, 6.4085706e-24, 6.4063167e-24], [357.31461, 8.9754103e-24, 6.4145655e-24, 6.412334e-24], [358.12452, 8.9840559e-24, 6.4205845e-24, 6.4183752e-24], [358.93626, 8.992737e-24, 6.4266276e-24, 6.4244404e-24], [359.74984, 9.0014537e-24, 6.432695e-24, 6.4305295e-24], [360.56526, 9.0102061e-24, 6.4387867e-24, 6.4366428e-24], [361.38254, 9.0189942e-24, 6.4449027e-24, 6.4427802e-24], [362.20166, 9.0278181e-24, 6.4510432e-24, 6.4489417e-24], [363.02265, 9.0366778e-24, 6.4572081e-24, 6.4551276e-24], [363.84549, 9.0455734e-24, 6.4633975e-24, 6.4613377e-24], [364.6702, 9.0545051e-24, 6.4696115e-24, 6.4675723e-24], [365.49677, 9.0634727e-24, 6.4758502e-24, 6.4738313e-24], [366.32522, 9.0724765e-24, 6.4821136e-24, 6.4801148e-24], [367.15555, 9.0815165e-24, 6.4884017e-24, 6.4864229e-24], [367.98776, 9.0905928e-24, 6.4947147e-24, 6.4927556e-24], [368.82186, 9.0997054e-24, 6.5010526e-24, 6.4991131e-24], [369.65785, 9.1088543e-24, 6.5074154e-24, 6.5054952e-24], [370.49573, 9.1180398e-24, 6.5138032e-24, 6.5119023e-24], [371.33551, 9.1272618e-24, 6.5202161e-24, 6.5183342e-24], [372.1772, 9.1365203e-24, 6.5266542e-24, 6.524791e-24], [373.02079, 9.1458156e-24, 6.5331174e-24, 6.5312729e-24], [373.8663, 9.1551476e-24, 6.5396059e-24, 6.5377798e-24], [374.71372, 9.1645164e-24, 6.5461198e-24, 6.5443119e-24], [375.56306, 9.1739221e-24, 6.552659e-24, 6.5508692e-24], [376.41433, 9.1833648e-24, 6.5592236e-24, 6.5574518e-24], [377.26753, 9.1928445e-24, 6.5658138e-24, 6.5640597e-24], [378.12266, 9.2023613e-24, 6.5724295e-24, 6.570693e-24], [378.97973, 9.2119153e-24, 6.5790709e-24, 6.5773518e-24], [379.83874, 9.2215066e-24, 6.585738e-24, 6.5840361e-24], [380.6997, 9.2311351e-24, 6.5924308e-24, 6.590746e-24], [381.56261, 9.2408011e-24, 6.5991494e-24, 6.5974815e-24], [382.42747, 9.2505045e-24, 6.605894e-24, 6.6042428e-24], [383.2943, 9.2602455e-24, 6.6126645e-24, 6.6110298e-24], [384.16309, 9.2700241e-24, 6.619461e-24, 6.6178427e-24], [385.03385, 9.2798404e-24, 6.6262835e-24, 6.6246815e-24], [385.90659, 9.2896945e-24, 6.6331322e-24, 6.6315463e-24], [386.7813, 9.2995864e-24, 6.6400071e-24, 6.6384371e-24], [387.658, 9.3095162e-24, 6.6469083e-24, 6.6453541e-24], [388.53668, 9.3194841e-24, 6.6538358e-24, 6.6522972e-24], [389.41735, 9.32949e-24, 6.6607897e-24, 6.6592666e-24], [390.30002, 9.339534e-24, 6.6677701e-24, 6.6662622e-24], [391.1847, 9.3496163e-24, 6.674777e-24, 6.6732843e-24], [392.07137, 9.3597369e-24, 6.6818104e-24, 6.6803328e-24], [392.96006, 9.3698959e-24, 6.6888705e-24, 6.6874078e-24], [393.85076, 9.3800933e-24, 6.6959574e-24, 6.6945093e-24], [394.74348, 9.3903293e-24, 6.703071e-24, 6.7016375e-24], [395.63822, 9.4006039e-24, 6.7102115e-24, 6.7087924e-24], [396.53499, 9.4109172e-24, 6.7173788e-24, 6.7159741e-24], [397.43379, 9.4212693e-24, 6.7245732e-24, 6.7231826e-24], [398.33464, 9.4316602e-24, 6.7317946e-24, 6.730418e-24], [399.23752, 9.4420901e-24, 6.7390431e-24, 6.7376804e-24], [400.14245, 9.452559e-24, 6.7463188e-24, 6.7449699e-24], [401.04943, 9.463067e-24, 6.7536217e-24, 6.7522864e-24], [401.95846, 9.4736141e-24, 6.760952e-24, 6.7596301e-24], [402.86956, 9.4842005e-24, 6.7683096e-24, 6.7670011e-24], [403.78272, 9.4948263e-24, 6.7756947e-24, 6.7743993e-24], [404.69795, 9.5054915e-24, 6.7831072e-24, 6.781825e-24], [405.61526, 9.5161962e-24, 6.7905474e-24, 6.7892781e-24], [406.53465, 9.5269404e-24, 6.7980152e-24, 6.7967587e-24], [407.45611, 9.5377244e-24, 6.8055107e-24, 6.8042669e-24], [408.37967, 9.5485481e-24, 6.8130339e-24, 6.8118028e-24], [409.30532, 9.5594116e-24, 6.8205851e-24, 6.8193663e-24], [410.23307, 9.5703151e-24, 6.8281641e-24, 6.8269577e-24], [411.16293, 9.5812586e-24, 6.8357711e-24, 6.834577e-24], [412.09489, 9.5922422e-24, 6.8434062e-24, 6.8422241e-24], [413.02896, 9.6032659e-24, 6.8510694e-24, 6.8498993e-24], [413.96515, 9.6143299e-24, 6.8587608e-24, 6.8576026e-24], [414.90346, 9.6254343e-24, 6.8664805e-24, 6.865334e-24], [415.8439, 9.636579e-24, 6.8742285e-24, 6.8730936e-24], [416.78647, 9.6477644e-24, 6.8820049e-24, 6.8808815e-24], [417.73117, 9.6589903e-24, 6.8898098e-24, 6.8886978e-24], [418.67802, 9.6702569e-24, 6.8976433e-24, 6.8965425e-24], [419.62701, 9.6815643e-24, 6.9055053e-24, 6.9044158e-24], [420.57816, 9.6929126e-24, 6.9133961e-24, 6.9123176e-24], [421.53146, 9.7043019e-24, 6.9213156e-24, 6.920248e-24], [422.48692, 9.7157322e-24, 6.929264e-24, 6.9282073e-24], [423.44455, 9.7272037e-24, 6.9372413e-24, 6.9361953e-24], [424.40435, 9.7387164e-24, 6.9452476e-24, 6.9442122e-24], [425.36632, 9.7502704e-24, 6.953283e-24, 6.9522581e-24], [426.33047, 9.7618659e-24, 6.9613475e-24, 6.960333e-24], [427.29681, 9.7735029e-24, 6.9694412e-24, 6.9684371e-24], [428.26534, 9.7851815e-24, 6.9775642e-24, 6.9765703e-24], [429.23607, 9.7969018e-24, 6.9857167e-24, 6.9847329e-24], [430.20899, 9.8086639e-24, 6.9938986e-24, 6.9929248e-24], [431.18412, 9.820468e-24, 7.0021101e-24, 7.0011462e-24], [432.16147, 9.832314e-24, 7.0103512e-24, 7.0093971e-24], [433.14102, 9.8442022e-24, 7.018622e-24, 7.0176776e-24], [434.1228, 9.8561326e-24, 7.0269227e-24, 7.0259879e-24], [435.1068, 9.8681053e-24, 7.0352532e-24, 7.034328e-24], [436.09303, 9.8801205e-24, 7.0436138e-24, 7.042698e-24], [437.0815, 9.8921782e-24, 7.0520045e-24, 7.0510981e-24], [438.07221, 9.9042785e-24, 7.0604254e-24, 7.0595282e-24], [439.06516, 9.9164217e-24, 7.0688766e-24, 7.0679886e-24], [440.06037, 9.9286077e-24, 7.0773583e-24, 7.0764793e-24], [441.05783, 9.9408367e-24, 7.0858705e-24, 7.0850005e-24], [442.05755, 9.9531089e-24, 7.0944134e-24, 7.0935522e-24], [443.05953, 9.9654244e-24, 7.1029871e-24, 7.1021347e-24], [444.06379, 9.9777833e-24, 7.1115917e-24, 7.1107481e-24], [445.07033, 9.9901857e-24, 7.1202274e-24, 7.1193924e-24], [446.07914, 1.0002632e-23, 7.1288944e-24, 7.128068e-24], [447.09024, 1.0015122e-23, 7.1375929e-24, 7.1367748e-24], [448.10364, 1.0027656e-23, 7.146323e-24, 7.1455133e-24], [449.11933, 1.0040235e-23, 7.1550849e-24, 7.1542835e-24], [450.13732, 1.0052858e-23, 7.163879e-24, 7.1630858e-24], [451.15762, 1.0065525e-23, 7.1727054e-24, 7.1719204e-24], [452.18024, 1.0078238e-23, 7.1815646e-24, 7.1807875e-24], [453.20517, 1.0090996e-23, 7.1904568e-24, 7.1896877e-24], [454.23242, 1.01038e-23, 7.1993825e-24, 7.1986213e-24], [455.26201, 1.011665e-23, 7.2083421e-24, 7.2075887e-24], [456.29392, 1.0129546e-23, 7.2173364e-24, 7.2165907e-24], [457.32818, 1.0142489e-23, 7.2263659e-24, 7.2256279e-24], [458.36478, 1.015548e-23, 7.2354317e-24, 7.2347012e-24], [459.40373, 1.016852e-23, 7.2445346e-24, 7.2438116e-24], [460.44503, 1.0181609e-23, 7.2536761e-24, 7.2529606e-24], [461.4887, 1.0194749e-23, 7.2628579e-24, 7.2621498e-24], [462.53473, 1.0207942e-23, 7.2720823e-24, 7.2713813e-24], [463.58313, 1.0221188e-23, 7.281352e-24, 7.2806583e-24], [464.63391, 1.0234493e-23, 7.290671e-24, 7.2899844e-24], [465.68707, 1.0247858e-23, 7.3000446e-24, 7.2993651e-24], [466.74261, 1.026129e-23, 7.3094802e-24, 7.3088076e-24], [467.80055, 1.0274796e-23, 7.3189884e-24, 7.3183228e-24], [468.86089, 1.0288387e-23, 7.328585e-24, 7.3279263e-24], [469.92363, 1.0302081e-23, 7.3382941e-24, 7.3376422e-24], [470.98878, 1.0315905e-23, 7.3481543e-24, 7.3475091e-24], [472.05634, 1.0329906e-23, 7.3582302e-24, 7.3575917e-24], [473.12633, 1.0344165e-23, 7.3686376e-24, 7.3680058e-24], [474.19874, 1.0358842e-23, 7.379601e-24, 7.3789758e-24], [475.27358, 1.0374283e-23, 7.3916033e-24, 7.3909847e-24], [476.35085, 1.0391347e-23, 7.4058505e-24, 7.4052387e-24], [477.43057, 1.0412718e-23, 7.4261062e-24, 7.4255016e-24], [478.51274, 1.0450754e-23, 7.4696566e-24, 7.4690609e-24], [479.59735, 1.0648976e-23, 7.7351589e-24, 7.7345888e-24], [480.68443, 1.1939038e-23, 9.4247775e-24, 9.4243138e-24], [481.77397, 1.0524153e-23, 7.5429907e-24, 7.5424166e-24], [482.86598, 1.0488183e-23, 7.4828695e-24, 7.482296e-24], [483.96047, 1.0490462e-23, 7.4761559e-24, 7.4755871e-24], [485.05743, 1.0500055e-23, 7.4796619e-24, 7.4790985e-24], [486.15688, 1.0512019e-23, 7.4864609e-24, 7.4859031e-24], [487.25883, 1.0524994e-23, 7.4946482e-24, 7.494096e-24], [488.36327, 1.0538488e-23, 7.5035305e-24, 7.5029839e-24], [489.47022, 1.0552286e-23, 7.5128068e-24, 7.5122658e-24], [490.57967, 1.0566283e-23, 7.5223294e-24, 7.521794e-24], [491.69164, 1.0580422e-23, 7.5320192e-24, 7.5314893e-24], [492.80613, 1.0594672e-23, 7.5418307e-24, 7.5413062e-24], [493.92315, 1.0609011e-23, 7.551736e-24, 7.551217e-24], [495.04269, 1.0623429e-23, 7.5617178e-24, 7.5612041e-24], [496.16478, 1.0637916e-23, 7.5717643e-24, 7.5712559e-24], [497.28941, 1.0652468e-23, 7.5818677e-24, 7.5813646e-24], [498.41659, 1.0667079e-23, 7.5920226e-24, 7.5915246e-24], [499.54632, 1.0681749e-23, 7.6022249e-24, 7.6017321e-24], [500.67861, 1.0696473e-23, 7.6124719e-24, 7.6119842e-24], [501.81347, 1.0711252e-23, 7.6227615e-24, 7.6222788e-24], [502.9509, 1.0726085e-23, 7.633092e-24, 7.6326144e-24], [504.09091, 1.0740969e-23, 7.6434624e-24, 7.6429897e-24], [505.23351, 1.0755905e-23, 7.6538716e-24, 7.6534039e-24], [506.37869, 1.0770892e-23, 7.6643191e-24, 7.6638562e-24], [507.52647, 1.078593e-23, 7.6748042e-24, 7.6743461e-24], [508.67686, 1.0801019e-23, 7.6853265e-24, 7.6848732e-24], [509.82984, 1.0816158e-23, 7.6958858e-24, 7.6954371e-24], [510.98545, 1.0831348e-23, 7.7064817e-24, 7.7060377e-24], [512.14367, 1.0846588e-23, 7.7171141e-24, 7.7166747e-24], [513.30452, 1.0861878e-23, 7.7277828e-24, 7.7273479e-24], [514.46799, 1.0877218e-23, 7.7384877e-24, 7.7380573e-24], [515.63411, 1.0892608e-23, 7.7492287e-24, 7.7488028e-24], [516.80287, 1.0908048e-23, 7.7600058e-24, 7.7595843e-24], [517.97428, 1.0923539e-23, 7.7708189e-24, 7.7704019e-24], [519.14834, 1.093908e-23, 7.781668e-24, 7.7812553e-24], [520.32506, 1.0954671e-23, 7.7925532e-24, 7.7921448e-24], [521.50446, 1.0970313e-23, 7.8034744e-24, 7.8030702e-24], [522.68652, 1.0986004e-23, 7.8144316e-24, 7.8140316e-24], [523.87127, 1.1001747e-23, 7.8254248e-24, 7.825029e-24], [525.05869, 1.101754e-23, 7.8364542e-24, 7.8360625e-24], [526.24882, 1.1033383e-23, 7.8475197e-24, 7.8471321e-24], [527.44163, 1.1049277e-23, 7.8586213e-24, 7.8582377e-24], [528.63716, 1.1065222e-23, 7.8697592e-24, 7.8693796e-24], [529.83539, 1.1081218e-23, 7.8809333e-24, 7.8805577e-24], [531.03634, 1.1097265e-23, 7.8921437e-24, 7.891772e-24], [532.24001, 1.1113362e-23, 7.9033906e-24, 7.9030227e-24], [533.44641, 1.1129511e-23, 7.9146738e-24, 7.9143098e-24], [534.65554, 1.1145711e-23, 7.9259936e-24, 7.9256334e-24], [535.86741, 1.1161963e-23, 7.9373499e-24, 7.9369934e-24], [537.08203, 1.1178266e-23, 7.9487428e-24, 7.9483901e-24], [538.29941, 1.119462e-23, 7.9601725e-24, 7.9598234e-24], [539.51954, 1.1211026e-23, 7.9716388e-24, 7.9712935e-24], [540.74244, 1.1227484e-23, 7.9831421e-24, 7.9828003e-24], [541.96811, 1.1243994e-23, 7.9946822e-24, 7.994344e-24], [543.19656, 1.1260556e-23, 8.0062593e-24, 8.0059246e-24], [544.42779, 1.1277169e-23, 8.0178734e-24, 8.0175422e-24], [545.66181, 1.1293835e-23, 8.0295247e-24, 8.029197e-24], [546.89863, 1.1310553e-23, 8.0412131e-24, 8.0408888e-24], [548.13826, 1.1327324e-23, 8.0529388e-24, 8.0526179e-24], [549.38069, 1.1344147e-23, 8.0647019e-24, 8.0643843e-24], [550.62594, 1.1361022e-23, 8.0765024e-24, 8.0761881e-24], [551.87402, 1.137795e-23, 8.0883403e-24, 8.0880294e-24], [553.12492, 1.1394932e-23, 8.1002158e-24, 8.0999082e-24], [554.37866, 1.1411965e-23, 8.112129e-24, 8.1118246e-24], [555.63523, 1.1429052e-23, 8.1240799e-24, 8.1237786e-24], [556.89466, 1.1446193e-23, 8.1360686e-24, 8.1357705e-24], [558.15694, 1.1463386e-23, 8.1480952e-24, 8.1478002e-24], [559.42209, 1.1480633e-23, 8.1601597e-24, 8.1598678e-24], [560.6901, 1.1497933e-23, 8.1722623e-24, 8.1719735e-24], [561.96098, 1.1515287e-23, 8.184403e-24, 8.1841172e-24], [563.23475, 1.1532694e-23, 8.1965819e-24, 8.1962991e-24], [564.5114, 1.1550155e-23, 8.208799e-24, 8.2085192e-24], [565.79095, 1.156767e-23, 8.2210545e-24, 8.2207776e-24], [567.07339, 1.1585239e-23, 8.2333485e-24, 8.2330745e-24], [568.35874, 1.1602863e-23, 8.245681e-24, 8.2454098e-24], [569.64701, 1.162054e-23, 8.258052e-24, 8.2577837e-24], [570.9382, 1.1638272e-23, 8.2704618e-24, 8.2701963e-24], [572.23231, 1.1656059e-23, 8.2829103e-24, 8.2826476e-24], [573.52936, 1.16739e-23, 8.2953976e-24, 8.2951377e-24], [574.82935, 1.1691795e-23, 8.3079239e-24, 8.3076667e-24], [576.13228, 1.1709746e-23, 8.3204892e-24, 8.3202347e-24], [577.43817, 1.1727751e-23, 8.3330935e-24, 8.3328417e-24], [578.74701, 1.1745812e-23, 8.3457371e-24, 8.3454879e-24], [580.05883, 1.1763928e-23, 8.3584199e-24, 8.3581733e-24], [581.37361, 1.1782099e-23, 8.371142e-24, 8.370898e-24], [582.69138, 1.1800325e-23, 8.3839035e-24, 8.3836621e-24], [584.01213, 1.1818607e-23, 8.3967046e-24, 8.3964657e-24], [585.33588, 1.1836944e-23, 8.4095452e-24, 8.4093089e-24], [586.66263, 1.1855338e-23, 8.4224255e-24, 8.4221916e-24], [587.99238, 1.1873787e-23, 8.4353455e-24, 8.4351142e-24], [589.32515, 1.1892292e-23, 8.4483054e-24, 8.4480765e-24], [590.66094, 1.1910853e-23, 8.4613052e-24, 8.4610787e-24], [591.99976, 1.1929471e-23, 8.474345e-24, 8.4741209e-24], [593.34161, 1.1948144e-23, 8.4874249e-24, 8.4872031e-24], [594.68651, 1.1966875e-23, 8.5005449e-24, 8.5003255e-24], [596.03445, 1.1985661e-23, 8.5137053e-24, 8.5134881e-24], [597.38545, 1.2004505e-23, 8.5269059e-24, 8.5266911e-24], [598.73951, 1.2023405e-23, 8.540147e-24, 8.5399344e-24], [600.09664, 1.2042363e-23, 8.5534286e-24, 8.5532183e-24], [601.45684, 1.2061377e-23, 8.5667508e-24, 8.5665427e-24], [602.82013, 1.2080448e-23, 8.5801137e-24, 8.5799078e-24], [604.18651, 1.2099577e-23, 8.5935173e-24, 8.5933136e-24], [605.55598, 1.2118763e-23, 8.6069618e-24, 8.6067602e-24], [606.92856, 1.2138007e-23, 8.6204472e-24, 8.6202478e-24], [608.30425, 1.2157309e-23, 8.6339737e-24, 8.6337764e-24], [609.68306, 1.2176668e-23, 8.6475412e-24, 8.647346e-24], [611.065, 1.2196085e-23, 8.66115e-24, 8.6609568e-24], [612.45006, 1.221556e-23, 8.6748001e-24, 8.6746089e-24], [613.83827, 1.2235093e-23, 8.6884915e-24, 8.6883024e-24], [615.22962, 1.2254684e-23, 8.7022244e-24, 8.7020373e-24], [616.62413, 1.2274334e-23, 8.7159988e-24, 8.7158137e-24], [618.0218, 1.2294043e-23, 8.7298148e-24, 8.7296317e-24], [619.42263, 1.231381e-23, 8.7436726e-24, 8.7434914e-24], [620.82664, 1.2333635e-23, 8.7575722e-24, 8.7573929e-24], [622.23384, 1.235352e-23, 8.7715137e-24, 8.7713363e-24], [623.64422, 1.2373464e-23, 8.7854971e-24, 8.7853216e-24], [625.0578, 1.2393466e-23, 8.7995227e-24, 8.799349e-24], [626.47458, 1.2413528e-23, 8.8135903e-24, 8.8134185e-24], [627.89458, 1.243365e-23, 8.8277003e-24, 8.8275302e-24], [629.31779, 1.2453831e-23, 8.8418525e-24, 8.8416843e-24], [630.74423, 1.2474071e-23, 8.8560472e-24, 8.8558808e-24], [632.1739, 1.2494371e-23, 8.8702844e-24, 8.8701197e-24], [633.60681, 1.2514731e-23, 8.8845642e-24, 8.8844013e-24], [635.04297, 1.2535151e-23, 8.8988867e-24, 8.8987255e-24], [636.48239, 1.2555631e-23, 8.913252e-24, 8.9130925e-24], [637.92507, 1.2576172e-23, 8.9276601e-24, 8.9275023e-24], [639.37102, 1.2596773e-23, 8.9421112e-24, 8.9419551e-24], [640.82024, 1.2617434e-23, 8.9566053e-24, 8.9564509e-24], [642.27276, 1.2638156e-23, 8.9711426e-24, 8.9709898e-24], [643.72856, 1.2658938e-23, 8.9857231e-24, 8.9855719e-24], [645.18766, 1.2679782e-23, 9.0003469e-24, 9.0001973e-24], [646.65007, 1.2700686e-23, 9.0150141e-24, 9.0148662e-24], [648.1158, 1.2721652e-23, 9.0297249e-24, 9.0295784e-24], [649.58485, 1.2742679e-23, 9.0444792e-24, 9.0443343e-24], [651.05722, 1.2763767e-23, 9.0592772e-24, 9.0591338e-24], [652.53294, 1.2784917e-23, 9.0741189e-24, 9.0739771e-24], [654.012, 1.2806129e-23, 9.0890045e-24, 9.0888643e-24], [655.49441, 1.2827402e-23, 9.1039341e-24, 9.1037953e-24], [656.98018, 1.2848737e-23, 9.1189077e-24, 9.1187704e-24], [658.46932, 1.2870134e-23, 9.1339255e-24, 9.1337896e-24], [659.96184, 1.2891593e-23, 9.1489875e-24, 9.1488531e-24], [661.45774, 1.2913115e-23, 9.1640938e-24, 9.1639608e-24], [662.95703, 1.2934699e-23, 9.1792445e-24, 9.1791129e-24], [664.45971, 1.2956345e-23, 9.1944397e-24, 9.1943095e-24], [665.96581, 1.2978055e-23, 9.2096795e-24, 9.2095507e-24], [667.47531, 1.2999827e-23, 9.224964e-24, 9.2248366e-24], [668.98824, 1.3021662e-23, 9.2402933e-24, 9.2401672e-24], [670.5046, 1.304356e-23, 9.2556674e-24, 9.2555427e-24], [672.02439, 1.3065521e-23, 9.2710865e-24, 9.2709632e-24], [673.54763, 1.3087546e-23, 9.2865507e-24, 9.2864287e-24], [675.07433, 1.3109634e-23, 9.30206e-24, 9.3019393e-24], [676.60448, 1.3131785e-23, 9.3176146e-24, 9.3174952e-24], [678.1381, 1.3154001e-23, 9.3332145e-24, 9.3330964e-24], [679.6752, 1.317628e-23, 9.3488599e-24, 9.348743e-24], [681.21578, 1.3198623e-23, 9.3645508e-24, 9.3644351e-24], [682.75985, 1.322103e-23, 9.3802873e-24, 9.3801728e-24], [684.30743, 1.3243502e-23, 9.3960695e-24, 9.3959563e-24], [685.85851, 1.3266038e-23, 9.4118976e-24, 9.4117855e-24], [687.41311, 1.3288639e-23, 9.4277715e-24, 9.4276607e-24], [688.97123, 1.3311304e-23, 9.4436914e-24, 9.4435818e-24], [690.53288, 1.3334034e-23, 9.4596575e-24, 9.459549e-24], [692.09807, 1.3356829e-23, 9.4756697e-24, 9.4755624e-24], [693.66681, 1.3379689e-23, 9.4917283e-24, 9.4916221e-24], [695.23911, 1.3402614e-23, 9.5078332e-24, 9.5077281e-24], [696.81497, 1.3425604e-23, 9.5239846e-24, 9.5238807e-24], [698.3944, 1.344866e-23, 9.5401825e-24, 9.5400797e-24], [699.97741, 1.3471782e-23, 9.5564272e-24, 9.5563255e-24], [701.56401, 1.3494969e-23, 9.5727186e-24, 9.572618e-24], [703.1542, 1.3518223e-23, 9.5890568e-24, 9.5889573e-24], [704.748, 1.3541542e-23, 9.6054421e-24, 9.6053436e-24], [706.34542, 1.3564927e-23, 9.6218744e-24, 9.621777e-24], [707.94645, 1.3588379e-23, 9.6383538e-24, 9.6382574e-24], [709.55111, 1.3611897e-23, 9.6548805e-24, 9.6547852e-24], [711.15941, 1.3635482e-23, 9.6714545e-24, 9.6713602e-24], [712.77136, 1.3659133e-23, 9.688076e-24, 9.6879828e-24], [714.38695, 1.3682852e-23, 9.7047451e-24, 9.7046528e-24], [716.00622, 1.3706637e-23, 9.7214618e-24, 9.7213705e-24], [717.62915, 1.3730489e-23, 9.7382262e-24, 9.7381359e-24], [719.25576, 1.3754409e-23, 9.7550385e-24, 9.7549491e-24], [720.88605, 1.3778396e-23, 9.7718987e-24, 9.7718103e-24], [722.52004, 1.3802451e-23, 9.7888069e-24, 9.7887195e-24], [724.15774, 1.3826573e-23, 9.8057633e-24, 9.8056768e-24], [725.79915, 1.3850763e-23, 9.822768e-24, 9.8226824e-24], [727.44428, 1.3875021e-23, 9.8398209e-24, 9.8397362e-24], [729.09313, 1.3899347e-23, 9.8569223e-24, 9.8568386e-24], [730.74573, 1.3923742e-23, 9.8740722e-24, 9.8739894e-24], [732.40207, 1.3948204e-23, 9.8912708e-24, 9.8911888e-24], [734.06216, 1.3972736e-23, 9.9085181e-24, 9.908437e-24], [735.72602, 1.3997336e-23, 9.9258142e-24, 9.925734e-24], [737.39365, 1.4022005e-23, 9.9431593e-24, 9.94308e-24], [739.06506, 1.4046742e-23, 9.9605534e-24, 9.9604749e-24], [740.74025, 1.4071549e-23, 9.9779967e-24, 9.977919e-24], [742.41925, 1.4096425e-23, 9.9954891e-24, 9.9954123e-24], [744.10205, 1.4121371e-23, 1.0013031e-23, 1.0012955e-23], [745.78866, 1.4146386e-23, 1.0030622e-23, 1.0030547e-23], [747.4791, 1.417147e-23, 1.0048263e-23, 1.0048189e-23], [749.17337, 1.4196625e-23, 1.0065953e-23, 1.006588e-23], [750.87148, 1.4221849e-23, 1.0083694e-23, 1.0083621e-23], [752.57344, 1.4247144e-23, 1.0101484e-23, 1.0101412e-23], [754.27925, 1.4272509e-23, 1.0119324e-23, 1.0119253e-23], [755.98893, 1.4297944e-23, 1.0137214e-23, 1.0137143e-23], [757.70249, 1.432345e-23, 1.0155154e-23, 1.0155084e-23], [759.41993, 1.4349026e-23, 1.0173144e-23, 1.0173076e-23], [761.14127, 1.4374674e-23, 1.0191185e-23, 1.0191117e-23], [762.8665, 1.4400392e-23, 1.0209277e-23, 1.0209209e-23], [764.59565, 1.4426181e-23, 1.0227418e-23, 1.0227352e-23], [766.32872, 1.4452042e-23, 1.0245611e-23, 1.0245545e-23], [768.06571, 1.4477974e-23, 1.0263854e-23, 1.0263789e-23], [769.80664, 1.4503978e-23, 1.0282149e-23, 1.0282084e-23], [771.55152, 1.4530053e-23, 1.0300494e-23, 1.030043e-23], [773.30035, 1.4556201e-23, 1.031889e-23, 1.0318827e-23], [775.05315, 1.458242e-23, 1.0337338e-23, 1.0337275e-23], [776.80992, 1.4608711e-23, 1.0355836e-23, 1.0355774e-23], [778.57067, 1.4635075e-23, 1.0374387e-23, 1.0374325e-23], [780.33541, 1.4661511e-23, 1.0392988e-23, 1.0392928e-23], [782.10415, 1.468802e-23, 1.0411642e-23, 1.0411582e-23], [783.8769, 1.4714602e-23, 1.0430347e-23, 1.0430287e-23], [785.65367, 1.4741256e-23, 1.0449104e-23, 1.0449045e-23], [787.43447, 1.4767984e-23, 1.0467912e-23, 1.0467854e-23], [789.2193, 1.4794785e-23, 1.0486773e-23, 1.0486716e-23], [791.00818, 1.4821659e-23, 1.0505686e-23, 1.050563e-23], [792.80112, 1.4848607e-23, 1.0524652e-23, 1.0524595e-23], [794.59812, 1.4875628e-23, 1.0543669e-23, 1.0543614e-23], [796.39919, 1.4902723e-23, 1.0562739e-23, 1.0562684e-23], [798.20434, 1.4929893e-23, 1.0581862e-23, 1.0581808e-23], [800.01358, 1.4957136e-23, 1.0601037e-23, 1.0600984e-23], [801.82693, 1.4984454e-23, 1.0620266e-23, 1.0620212e-23], [803.64439, 1.5011846e-23, 1.0639547e-23, 1.0639494e-23], [805.46596, 1.5039312e-23, 1.0658881e-23, 1.0658829e-23], [807.29167, 1.5066854e-23, 1.0678268e-23, 1.0678217e-23], [809.12151, 1.509447e-23, 1.0697708e-23, 1.0697658e-23], [810.9555, 1.5122162e-23, 1.0717202e-23, 1.0717152e-23], [812.79365, 1.5149928e-23, 1.073675e-23, 1.07367e-23], [814.63596, 1.517777e-23, 1.075635e-23, 1.0756301e-23], [816.48245, 1.5205688e-23, 1.0776005e-23, 1.0775956e-23], [818.33312, 1.5233681e-23, 1.0795713e-23, 1.0795665e-23], [820.18799, 1.526175e-23, 1.0815475e-23, 1.0815428e-23], [822.04707, 1.5289895e-23, 1.0835291e-23, 1.0835244e-23], [823.91036, 1.5318116e-23, 1.0855162e-23, 1.0855115e-23], [825.77787, 1.5346413e-23, 1.0875086e-23, 1.087504e-23], [827.64961, 1.5374787e-23, 1.0895065e-23, 1.0895019e-23], [829.5256, 1.5403238e-23, 1.0915098e-23, 1.0915053e-23], [831.40584, 1.5431765e-23, 1.0935186e-23, 1.0935141e-23], [833.29034, 1.546037e-23, 1.0955329e-23, 1.0955284e-23], [835.17911, 1.5489051e-23, 1.0975526e-23, 1.0975482e-23], [837.07216, 1.551781e-23, 1.0995778e-23, 1.0995735e-23], [838.96951, 1.5546646e-23, 1.1016085e-23, 1.1016043e-23], [840.87115, 1.5575559e-23, 1.1036448e-23, 1.1036405e-23], [842.77711, 1.5604551e-23, 1.1056865e-23, 1.1056823e-23], [844.68738, 1.563362e-23, 1.1077338e-23, 1.1077297e-23], [846.60199, 1.5662768e-23, 1.1097866e-23, 1.1097825e-23], [848.52093, 1.5691993e-23, 1.111845e-23, 1.111841e-23], [850.44423, 1.5721297e-23, 1.113909e-23, 1.113905e-23], [852.37188, 1.575068e-23, 1.1159785e-23, 1.1159746e-23], [854.30391, 1.5780141e-23, 1.1180537e-23, 1.1180497e-23], [856.24031, 1.5809681e-23, 1.1201344e-23, 1.1201305e-23], [858.1811, 1.5839301e-23, 1.1222208e-23, 1.1222169e-23], [860.12629, 1.5868999e-23, 1.1243127e-23, 1.1243089e-23], [862.07589, 1.5898777e-23, 1.1264103e-23, 1.1264066e-23], [864.02991, 1.5928634e-23, 1.1285136e-23, 1.1285099e-23], [865.98836, 1.5958571e-23, 1.1306225e-23, 1.1306189e-23], [867.95124, 1.5988588e-23, 1.1327371e-23, 1.1327335e-23], [869.91858, 1.6018685e-23, 1.1348574e-23, 1.1348538e-23], [871.89038, 1.6048862e-23, 1.1369834e-23, 1.1369799e-23], [873.86664, 1.607912e-23, 1.1391151e-23, 1.1391116e-23], [875.84739, 1.6109458e-23, 1.1412525e-23, 1.141249e-23], [877.83262, 1.6139876e-23, 1.1433956e-23, 1.1433922e-23], [879.82235, 1.6170376e-23, 1.1455445e-23, 1.1455411e-23], [881.8166, 1.6200956e-23, 1.1476992e-23, 1.1476958e-23], [883.81536, 1.6231618e-23, 1.1498596e-23, 1.1498562e-23], [885.81866, 1.6262361e-23, 1.1520258e-23, 1.1520225e-23], [887.82649, 1.6293186e-23, 1.1541977e-23, 1.1541945e-23], [889.83888, 1.6324093e-23, 1.1563755e-23, 1.1563723e-23], [891.85583, 1.6355081e-23, 1.1585591e-23, 1.1585559e-23], [893.87734, 1.6386151e-23, 1.1607485e-23, 1.1607454e-23], [895.90345, 1.6417304e-23, 1.1629438e-23, 1.1629407e-23], [897.93414, 1.6448539e-23, 1.1651449e-23, 1.1651418e-23], [899.96944, 1.6479856e-23, 1.1673519e-23, 1.1673488e-23], [902.00935, 1.6511257e-23, 1.1695647e-23, 1.1695617e-23], [904.05388, 1.654274e-23, 1.1717835e-23, 1.1717805e-23], [906.10305, 1.6574306e-23, 1.1740081e-23, 1.1740051e-23], [908.15686, 1.6605956e-23, 1.1762386e-23, 1.1762357e-23], [910.21533, 1.6637688e-23, 1.1784751e-23, 1.1784722e-23], [912.27846, 1.6669505e-23, 1.1807175e-23, 1.1807147e-23], [914.34627, 1.6701405e-23, 1.1829659e-23, 1.1829631e-23], [916.41877, 1.673339e-23, 1.1852202e-23, 1.1852174e-23], [918.49597, 1.6765458e-23, 1.1874805e-23, 1.1874778e-23], [920.57787, 1.6797611e-23, 1.1897468e-23, 1.1897441e-23], [922.66449, 1.6829848e-23, 1.1920191e-23, 1.1920164e-23], [924.75584, 1.686217e-23, 1.1942975e-23, 1.1942948e-23], [926.85193, 1.6894577e-23, 1.1965818e-23, 1.1965792e-23], [928.95278, 1.6927069e-23, 1.1988722e-23, 1.1988696e-23], [931.05838, 1.6959647e-23, 1.2011687e-23, 1.2011661e-23], [933.16876, 1.699231e-23, 1.2034713e-23, 1.2034687e-23], [935.28392, 1.7025059e-23, 1.20578e-23, 1.2057774e-23], [937.40388, 1.7057893e-23, 1.2080948e-23, 1.2080923e-23], [939.52864, 1.7090815e-23, 1.2104158e-23, 1.2104133e-23], [941.65821, 1.7123823e-23, 1.212743e-23, 1.2127406e-23], [943.79262, 1.7156919e-23, 1.2150765e-23, 1.2150741e-23], [945.93186, 1.7190103e-23, 1.2174165e-23, 1.2174141e-23], [948.07595, 1.7223377e-23, 1.2197631e-23, 1.2197607e-23], [950.2249, 1.7256745e-23, 1.2221168e-23, 1.2221145e-23], [952.37872, 1.7290215e-23, 1.2244788e-23, 1.2244765e-23], [954.53743, 1.732381e-23, 1.2268523e-23, 1.22685e-23], [956.70102, 1.735763e-23, 1.2292516e-23, 1.2292494e-23], [958.86952, 1.7392621e-23, 1.2318099e-23, 1.2318077e-23], [961.04294, 1.7506943e-23, 1.2455377e-23, 1.2455355e-23], [963.22128, 1.7459609e-23, 1.2364856e-23, 1.2364834e-23], [965.40456, 1.7492862e-23, 1.23878e-23, 1.2387778e-23], [967.59279, 1.7526766e-23, 1.2411602e-23, 1.2411581e-23], [969.78598, 1.7560852e-23, 1.2435599e-23, 1.2435578e-23], [971.98413, 1.7595055e-23, 1.24597e-23, 1.2459679e-23], [974.18728, 1.7629359e-23, 1.2483879e-23, 1.2483859e-23], [976.39541, 1.7663757e-23, 1.250813e-23, 1.2508109e-23], [978.60855, 1.7698247e-23, 1.2532448e-23, 1.2532427e-23], [980.82671, 1.7732828e-23, 1.2556831e-23, 1.2556811e-23], [983.04989, 1.7767499e-23, 1.258128e-23, 1.258126e-23], [985.27812, 1.7802261e-23, 1.2605793e-23, 1.2605773e-23], [987.51139, 1.7837114e-23, 1.2630371e-23, 1.2630351e-23], [989.74973, 1.7872057e-23, 1.2655013e-23, 1.2654994e-23], [991.99314, 1.790709e-23, 1.267972e-23, 1.2679701e-23], [994.24163, 1.7942214e-23, 1.2704491e-23, 1.2704472e-23], [996.49523, 1.7977429e-23, 1.2729326e-23, 1.2729307e-23], [998.75393, 1.8012734e-23, 1.2754226e-23, 1.2754208e-23], [1001.0177, 1.8048131e-23, 1.2779191e-23, 1.2779173e-23], [1003.2867, 1.8083619e-23, 1.280422e-23, 1.2804202e-23], [1005.5608, 1.8119198e-23, 1.2829315e-23, 1.2829297e-23], [1007.84, 1.815487e-23, 1.2854474e-23, 1.2854457e-23], [1010.1245, 1.8190632e-23, 1.2879699e-23, 1.2879682e-23], [1012.414, 1.8226487e-23, 1.2904989e-23, 1.2904972e-23], [1014.7088, 1.8262434e-23, 1.2930345e-23, 1.2930328e-23], [1017.0088, 1.8298474e-23, 1.2955766e-23, 1.2955749e-23], [1019.314, 1.8334606e-23, 1.2981252e-23, 1.2981236e-23], [1021.6244, 1.8370831e-23, 1.3006805e-23, 1.3006788e-23], [1023.9401, 1.8407149e-23, 1.3032423e-23, 1.3032407e-23], [1026.261, 1.8443559e-23, 1.3058108e-23, 1.3058091e-23], [1028.5872, 1.8480064e-23, 1.3083858e-23, 1.3083842e-23], [1030.9186, 1.8516661e-23, 1.3109675e-23, 1.310966e-23], [1033.2553, 1.8553353e-23, 1.3135559e-23, 1.3135543e-23], [1035.5973, 1.8590138e-23, 1.3161509e-23, 1.3161494e-23], [1037.9447, 1.8627018e-23, 1.3187526e-23, 1.3187511e-23], [1040.2973, 1.8663992e-23, 1.321361e-23, 1.3213595e-23], [1042.6553, 1.870106e-23, 1.3239761e-23, 1.3239746e-23], [1045.0186, 1.8738223e-23, 1.3265979e-23, 1.3265964e-23], [1047.3873, 1.8775481e-23, 1.3292264e-23, 1.3292249e-23], [1049.7614, 1.8812834e-23, 1.3318617e-23, 1.3318602e-23], [1052.1408, 1.8850282e-23, 1.3345037e-23, 1.3345023e-23], [1054.5256, 1.8887826e-23, 1.3371525e-23, 1.3371511e-23], [1056.9159, 1.8925465e-23, 1.3398081e-23, 1.3398067e-23], [1059.3115, 1.8963201e-23, 1.3424705e-23, 1.3424692e-23], [1061.7126, 1.9001032e-23, 1.3451397e-23, 1.3451384e-23], [1064.1191, 1.903896e-23, 1.3478158e-23, 1.3478144e-23], [1066.5311, 1.9076984e-23, 1.3504987e-23, 1.3504974e-23], [1068.9486, 1.9115105e-23, 1.3531884e-23, 1.3531871e-23], [1071.3715, 1.9153323e-23, 1.3558851e-23, 1.3558838e-23], [1073.7999, 1.9191638e-23, 1.3585886e-23, 1.3585873e-23], [1076.2338, 1.923005e-23, 1.361299e-23, 1.3612977e-23], [1078.6733, 1.926856e-23, 1.3640163e-23, 1.364015e-23], [1081.1182, 1.9307167e-23, 1.3667406e-23, 1.3667393e-23], [1083.5687, 1.9345873e-23, 1.3694718e-23, 1.3694705e-23], [1086.0248, 1.9384676e-23, 1.3722099e-23, 1.3722087e-23], [1088.4864, 1.9423578e-23, 1.374955e-23, 1.3749538e-23], [1090.9536, 1.9462579e-23, 1.3777072e-23, 1.377706e-23], [1093.4265, 1.9501678e-23, 1.3804663e-23, 1.3804651e-23], [1095.9049, 1.9540876e-23, 1.3832324e-23, 1.3832312e-23], [1098.3889, 1.9580173e-23, 1.3860056e-23, 1.3860044e-23], [1100.8785, 1.961957e-23, 1.3887858e-23, 1.3887846e-23], [1103.3738, 1.9659066e-23, 1.3915731e-23, 1.3915719e-23], [1105.8748, 1.9698662e-23, 1.3943674e-23, 1.3943663e-23], [1108.3814, 1.9738358e-23, 1.3971688e-23, 1.3971677e-23], [1110.8937, 1.9778154e-23, 1.3999774e-23, 1.3999763e-23], [1113.4117, 1.9818051e-23, 1.4027931e-23, 1.402792e-23], [1115.9354, 1.9858048e-23, 1.4056158e-23, 1.4056148e-23], [1118.4649, 1.9898146e-23, 1.4084458e-23, 1.4084447e-23], [1121.0, 1.9938345e-23, 1.4112829e-23, 1.4112819e-23], [1123.5409, 1.9978645e-23, 1.4141272e-23, 1.4141262e-23], [1126.0876, 2.0019047e-23, 1.4169787e-23, 1.4169777e-23], [1128.64, 2.005955e-23, 1.4198374e-23, 1.4198364e-23], [1131.1983, 2.0100156e-23, 1.4227033e-23, 1.4227023e-23], [1133.7623, 2.0140863e-23, 1.4255764e-23, 1.4255755e-23], [1136.3321, 2.0181673e-23, 1.4284569e-23, 1.4284559e-23], [1138.9078, 2.0222586e-23, 1.4313445e-23, 1.4313436e-23], [1141.4893, 2.0263601e-23, 1.4342395e-23, 1.4342386e-23], [1144.0766, 2.0304719e-23, 1.4371418e-23, 1.4371408e-23], [1146.6698, 2.034594e-23, 1.4400514e-23, 1.4400504e-23], [1149.2689, 2.0387265e-23, 1.4429683e-23, 1.4429673e-23], [1151.8739, 2.0428693e-23, 1.4458925e-23, 1.4458916e-23], [1154.4848, 2.0470225e-23, 1.4488241e-23, 1.4488232e-23], [1157.1016, 2.0511862e-23, 1.4517631e-23, 1.4517623e-23], [1159.7243, 2.0553602e-23, 1.4547095e-23, 1.4547086e-23], [1162.353, 2.0595447e-23, 1.4576633e-23, 1.4576625e-23], [1164.9877, 2.0637397e-23, 1.4606246e-23, 1.4606237e-23], [1167.6283, 2.0679451e-23, 1.4635932e-23, 1.4635924e-23], [1170.2749, 2.0721611e-23, 1.4665693e-23, 1.4665685e-23], [1172.9275, 2.0763877e-23, 1.4695529e-23, 1.4695521e-23], [1175.5861, 2.0806247e-23, 1.472544e-23, 1.4725431e-23], [1178.2507, 2.0848724e-23, 1.4755425e-23, 1.4755417e-23], [1180.9214, 2.0891307e-23, 1.4785486e-23, 1.4785478e-23], [1183.5981, 2.0933996e-23, 1.4815623e-23, 1.4815615e-23], [1186.2809, 2.0976791e-23, 1.4845834e-23, 1.4845826e-23], [1188.9698, 2.1019694e-23, 1.4876122e-23, 1.4876114e-23], [1191.6648, 2.1062703e-23, 1.4906485e-23, 1.4906477e-23], [1194.3658, 2.1105819e-23, 1.4936924e-23, 1.4936916e-23], [1197.073, 2.1149043e-23, 1.4967439e-23, 1.4967431e-23], [1199.7864, 2.1192374e-23, 1.4998031e-23, 1.4998023e-23], [1202.5059, 2.1235813e-23, 1.5028698e-23, 1.5028691e-23], [1205.2315, 2.1279361e-23, 1.5059443e-23, 1.5059436e-23], [1207.9634, 2.1323016e-23, 1.5090264e-23, 1.5090257e-23], [1210.7014, 2.1366781e-23, 1.5121163e-23, 1.5121155e-23], [1213.4456, 2.1410654e-23, 1.5152138e-23, 1.5152131e-23], [1216.1961, 2.1454636e-23, 1.518319e-23, 1.5183183e-23], [1218.9527, 2.1498727e-23, 1.521432e-23, 1.5214314e-23], [1221.7157, 2.1542927e-23, 1.5245528e-23, 1.5245521e-23], [1224.4849, 2.1587238e-23, 1.5276813e-23, 1.5276807e-23], [1227.2603, 2.1631658e-23, 1.5308177e-23, 1.530817e-23], [1230.0421, 2.1676189e-23, 1.5339618e-23, 1.5339611e-23], [1232.8302, 2.1720829e-23, 1.5371137e-23, 1.5371131e-23], [1235.6246, 2.1765581e-23, 1.5402735e-23, 1.5402729e-23], [1238.4253, 2.1810443e-23, 1.5434412e-23, 1.5434405e-23], [1241.2323, 2.1855417e-23, 1.5466167e-23, 1.546616e-23], [1244.0458, 2.1900501e-23, 1.5498001e-23, 1.5497995e-23], [1246.8656, 2.1945698e-23, 1.5529914e-23, 1.5529908e-23], [1249.6918, 2.1991006e-23, 1.5561906e-23, 1.55619e-23], [1252.5244, 2.2036426e-23, 1.5593978e-23, 1.5593972e-23], [1255.3634, 2.2081958e-23, 1.5626129e-23, 1.5626123e-23], [1258.2089, 2.2127603e-23, 1.565836e-23, 1.5658354e-23], [1261.0608, 2.2173361e-23, 1.5690671e-23, 1.5690665e-23], [1263.9192, 2.2219232e-23, 1.5723062e-23, 1.5723056e-23], [1266.784, 2.2265216e-23, 1.5755533e-23, 1.5755528e-23], [1269.6554, 2.2311313e-23, 1.5788085e-23, 1.5788079e-23], [1272.5332, 2.2357524e-23, 1.5820717e-23, 1.5820712e-23], [1275.4176, 2.2403849e-23, 1.585343e-23, 1.5853425e-23], [1278.3085, 2.2450289e-23, 1.5886224e-23, 1.5886218e-23], [1281.206, 2.2496843e-23, 1.5919099e-23, 1.5919093e-23], [1284.11, 2.2543511e-23, 1.5952055e-23, 1.595205e-23], [1287.0206, 2.2590294e-23, 1.5985093e-23, 1.5985087e-23], [1289.9379, 2.2637193e-23, 1.6018212e-23, 1.6018207e-23], [1292.8617, 2.2684207e-23, 1.6051413e-23, 1.6051408e-23], [1295.7922, 2.2731336e-23, 1.6084696e-23, 1.6084691e-23], [1298.7292, 2.2778582e-23, 1.6118061e-23, 1.6118056e-23], [1301.673, 2.2825943e-23, 1.6151508e-23, 1.6151503e-23], [1304.6234, 2.2873422e-23, 1.6185038e-23, 1.6185033e-23], [1307.5806, 2.2921016e-23, 1.621865e-23, 1.6218645e-23], [1310.5444, 2.2968728e-23, 1.6252345e-23, 1.625234e-23], [1313.5149, 2.3016557e-23, 1.6286124e-23, 1.6286119e-23], [1316.4922, 2.3064503e-23, 1.6319985e-23, 1.631998e-23], [1319.4762, 2.3112566e-23, 1.635393e-23, 1.6353925e-23], [1322.467, 2.3160748e-23, 1.6387958e-23, 1.6387953e-23], [1325.4645, 2.3209048e-23, 1.642207e-23, 1.6422065e-23], [1328.4689, 2.3257466e-23, 1.6456266e-23, 1.6456261e-23], [1331.4801, 2.3306003e-23, 1.6490545e-23, 1.6490541e-23], [1334.4981, 2.3354659e-23, 1.6524909e-23, 1.6524905e-23], [1337.5229, 2.3403434e-23, 1.6559358e-23, 1.6559354e-23], [1340.5546, 2.3452328e-23, 1.6593891e-23, 1.6593887e-23], [1343.5931, 2.3501342e-23, 1.6628509e-23, 1.6628505e-23], [1346.6386, 2.3550476e-23, 1.6663212e-23, 1.6663207e-23], [1349.6909, 2.359973e-23, 1.6698e-23, 1.6697995e-23], [1352.7502, 2.3649105e-23, 1.6732873e-23, 1.6732869e-23], [1355.8164, 2.36986e-23, 1.6767831e-23, 1.6767827e-23], [1358.8896, 2.3748216e-23, 1.6802876e-23, 1.6802872e-23], [1361.9697, 2.3797953e-23, 1.6838006e-23, 1.6838002e-23], [1365.0568, 2.3847812e-23, 1.6873222e-23, 1.6873218e-23], [1368.1509, 2.3897793e-23, 1.6908525e-23, 1.6908521e-23], [1371.252, 2.3947895e-23, 1.6943914e-23, 1.694391e-23], [1374.3601, 2.399812e-23, 1.6979389e-23, 1.6979385e-23], [1377.4753, 2.4048467e-23, 1.7014951e-23, 1.7014947e-23], [1380.5976, 2.4098937e-23, 1.70506e-23, 1.7050597e-23], [1383.7269, 2.4149531e-23, 1.7086337e-23, 1.7086333e-23], [1386.8633, 2.4200247e-23, 1.712216e-23, 1.7122157e-23], [1390.0068, 2.4251087e-23, 1.7158072e-23, 1.7158068e-23], [1393.1575, 2.4302051e-23, 1.719407e-23, 1.7194067e-23], [1396.3153, 2.4353138e-23, 1.7230157e-23, 1.7230154e-23], [1399.4802, 2.4404351e-23, 1.7266332e-23, 1.7266329e-23], [1402.6523, 2.4455687e-23, 1.7302595e-23, 1.7302592e-23], [1405.8316, 2.4507149e-23, 1.7338947e-23, 1.7338943e-23], [1409.0182, 2.4558736e-23, 1.7375387e-23, 1.7375384e-23], [1412.2119, 2.4610449e-23, 1.7411916e-23, 1.7411913e-23], [1415.4129, 2.4662287e-23, 1.7448535e-23, 1.7448531e-23], [1418.6211, 2.4714251e-23, 1.7485242e-23, 1.7485239e-23], [1421.8366, 2.4766342e-23, 1.752204e-23, 1.7522037e-23], [1425.0594, 2.4818559e-23, 1.7558927e-23, 1.7558924e-23], [1428.2895, 2.4870905e-23, 1.7595906e-23, 1.7595903e-23], [1431.527, 2.4923379e-23, 1.7632978e-23, 1.7632975e-23], [1434.7717, 2.4975989e-23, 1.7670151e-23, 1.7670148e-23], [1438.0238, 2.5028784e-23, 1.7707496e-23, 1.7707493e-23], [1441.2833, 2.5140746e-23, 1.7828283e-23, 1.782828e-23], [1444.5502, 2.5134603e-23, 1.7782239e-23, 1.7782236e-23], [1447.8245, 2.5187656e-23, 1.7819677e-23, 1.7819674e-23], [1451.1062, 2.5240883e-23, 1.7857271e-23, 1.7857268e-23], [1454.3953, 2.5294246e-23, 1.7894967e-23, 1.7894964e-23], [1457.6919, 2.5347741e-23, 1.7932757e-23, 1.7932754e-23], [1460.996, 2.5401366e-23, 1.7970639e-23, 1.7970637e-23], [1464.3075, 2.5455121e-23, 1.8008615e-23, 1.8008612e-23], [1467.6266, 2.5509007e-23, 1.8046682e-23, 1.8046679e-23], [1470.9532, 2.5563023e-23, 1.8084842e-23, 1.808484e-23], [1474.2873, 2.5617169e-23, 1.8123095e-23, 1.8123092e-23], [1477.629, 2.5671447e-23, 1.8161441e-23, 1.8161438e-23], [1480.9783, 2.5725856e-23, 1.8199879e-23, 1.8199876e-23], [1484.3351, 2.5780396e-23, 1.823841e-23, 1.8238408e-23], [1487.6996, 2.5835068e-23, 1.8277035e-23, 1.8277033e-23], [1491.0717, 2.5889872e-23, 1.8315754e-23, 1.8315751e-23], [1494.4514, 2.5944808e-23, 1.8354566e-23, 1.8354563e-23], [1497.8388, 2.5999877e-23, 1.8393472e-23, 1.8393469e-23], [1501.2338, 2.6055079e-23, 1.8432472e-23, 1.8432469e-23], [1504.6366, 2.6110414e-23, 1.8471566e-23, 1.8471563e-23], [1508.0471, 2.6165883e-23, 1.8510755e-23, 1.8510752e-23], [1511.4653, 2.6221485e-23, 1.8550038e-23, 1.8550036e-23], [1514.8912, 2.6277221e-23, 1.8589417e-23, 1.8589414e-23], [1518.325, 2.6333092e-23, 1.862889e-23, 1.8628888e-23], [1521.7665, 2.6389097e-23, 1.8668459e-23, 1.8668457e-23], [1525.2158, 2.6445237e-23, 1.8708124e-23, 1.8708121e-23], [1528.6729, 2.6501513e-23, 1.8747884e-23, 1.8747881e-23], [1532.1378, 2.6557923e-23, 1.878774e-23, 1.8787737e-23], [1535.6106, 2.661447e-23, 1.8827692e-23, 1.882769e-23], [1539.0913, 2.6671153e-23, 1.886774e-23, 1.8867738e-23], [1542.5799, 2.6727972e-23, 1.8907885e-23, 1.8907883e-23], [1546.0764, 2.6784927e-23, 1.8948127e-23, 1.8948125e-23], [1549.5808, 2.684202e-23, 1.8988466e-23, 1.8988464e-23], [1553.0931, 2.689925e-23, 1.9028902e-23, 1.9028899e-23], [1556.6134, 2.6956617e-23, 1.9069435e-23, 1.9069433e-23], [1560.1417, 2.7014122e-23, 1.9110066e-23, 1.9110064e-23], [1563.678, 2.7071766e-23, 1.9150794e-23, 1.9150792e-23], [1567.2223, 2.7129547e-23, 1.9191621e-23, 1.9191619e-23], [1570.7746, 2.7187468e-23, 1.9232546e-23, 1.9232544e-23], [1574.335, 2.7245527e-23, 1.9273569e-23, 1.9273567e-23], [1577.9035, 2.7303726e-23, 1.9314691e-23, 1.9314689e-23], [1581.48, 2.7362065e-23, 1.9355912e-23, 1.935591e-23], [1585.0647, 2.7420543e-23, 1.9397232e-23, 1.939723e-23], [1588.6575, 2.7479162e-23, 1.9438651e-23, 1.9438649e-23], [1592.2584, 2.7537921e-23, 1.9480169e-23, 1.9480167e-23], [1595.8675, 2.7596821e-23, 1.9521788e-23, 1.9521786e-23], [1599.4847, 2.7655862e-23, 1.9563506e-23, 1.9563504e-23], [1603.1102, 2.7715045e-23, 1.9605324e-23, 1.9605322e-23], [1606.7439, 2.7774369e-23, 1.9647243e-23, 1.9647241e-23], [1610.3858, 2.7833836e-23, 1.9689262e-23, 1.968926e-23], [1614.036, 2.7893445e-23, 1.9731382e-23, 1.973138e-23], [1617.6944, 2.7953196e-23, 1.9773603e-23, 1.9773601e-23], [1621.3611, 2.801309e-23, 1.9815925e-23, 1.9815924e-23], [1625.0362, 2.8073128e-23, 1.9858349e-23, 1.9858347e-23], [1628.7196, 2.813331e-23, 1.9900874e-23, 1.9900873e-23], [1632.4113, 2.8193635e-23, 1.9943502e-23, 1.99435e-23], [1636.1114, 2.8254104e-23, 1.9986231e-23, 1.9986229e-23], [1639.8199, 2.8314718e-23, 2.0029063e-23, 2.0029061e-23], [1643.5368, 2.8375477e-23, 2.0071997e-23, 2.0071995e-23], [1647.2621, 2.8436381e-23, 2.0115034e-23, 2.0115032e-23], [1650.9958, 2.8497431e-23, 2.0158173e-23, 2.0158172e-23], [1654.738, 2.8558626e-23, 2.0201417e-23, 2.0201415e-23], [1658.4887, 2.8619968e-23, 2.0244763e-23, 2.0244762e-23], [1662.2479, 2.8681455e-23, 2.0288213e-23, 2.0288212e-23], [1666.0157, 2.874309e-23, 2.0331767e-23, 2.0331766e-23], [1669.7919, 2.8804872e-23, 2.0375425e-23, 2.0375424e-23], [1673.5768, 2.8866801e-23, 2.0419188e-23, 2.0419186e-23], [1677.3702, 2.8928877e-23, 2.0463055e-23, 2.0463053e-23], [1681.1722, 2.8991102e-23, 2.0507027e-23, 2.0507025e-23], [1684.9828, 2.9053475e-23, 2.0551103e-23, 2.0551102e-23], [1688.802, 2.9115997e-23, 2.0595285e-23, 2.0595284e-23], [1692.6299, 2.9178668e-23, 2.0639573e-23, 2.0639571e-23], [1696.4665, 2.9241488e-23, 2.0683966e-23, 2.0683964e-23], [1700.3118, 2.9304457e-23, 2.0728465e-23, 2.0728463e-23], [1704.1658, 2.9367577e-23, 2.077307e-23, 2.0773069e-23], [1708.0286, 2.9430847e-23, 2.0817781e-23, 2.081778e-23], [1711.9001, 2.9494267e-23, 2.08626e-23, 2.0862598e-23], [1715.7803, 2.9557839e-23, 2.0907525e-23, 2.0907523e-23], [1719.6694, 2.9621562e-23, 2.0952557e-23, 2.0952555e-23], [1723.5673, 2.9685436e-23, 2.0997696e-23, 2.0997695e-23], [1727.474, 2.9749462e-23, 2.1042943e-23, 2.1042941e-23], [1731.3895, 2.9813641e-23, 2.1088297e-23, 2.1088296e-23], [1735.314, 2.9877972e-23, 2.113376e-23, 2.1133759e-23], [1739.2473, 2.9942456e-23, 2.1179331e-23, 2.1179329e-23], [1743.1896, 3.0007093e-23, 2.122501e-23, 2.1225009e-23], [1747.1408, 3.0071884e-23, 2.1270798e-23, 2.1270797e-23], [1751.1009, 3.0136828e-23, 2.1316695e-23, 2.1316694e-23], [1755.07, 3.0201927e-23, 2.1362701e-23, 2.13627e-23], [1759.0482, 3.0267181e-23, 2.1408816e-23, 2.1408815e-23], [1763.0353, 3.0332589e-23, 2.1455041e-23, 2.145504e-23], [1767.0315, 3.0398153e-23, 2.1501376e-23, 2.1501375e-23], [1771.0367, 3.0463872e-23, 2.1547822e-23, 2.154782e-23], [1775.051, 3.0529748e-23, 2.1594377e-23, 2.1594376e-23], [1779.0744, 3.0595779e-23, 2.1641043e-23, 2.1641042e-23], [1783.1069, 3.0661967e-23, 2.168782e-23, 2.1687819e-23], [1787.1486, 3.0728312e-23, 2.1734708e-23, 2.1734707e-23], [1791.1994, 3.0794815e-23, 2.1781707e-23, 2.1781706e-23], [1795.2594, 3.0861475e-23, 2.1828818e-23, 2.1828817e-23], [1799.3287, 3.0928293e-23, 2.1876041e-23, 2.187604e-23], [1803.4071, 3.0995269e-23, 2.1923375e-23, 2.1923374e-23], [1807.4948, 3.1062404e-23, 2.1970822e-23, 2.1970821e-23], [1811.5917, 3.1129698e-23, 2.2018382e-23, 2.2018381e-23], [1815.698, 3.1197151e-23, 2.2066054e-23, 2.2066053e-23], [1819.8135, 3.1264764e-23, 2.211384e-23, 2.2113839e-23], [1823.9384, 3.1332537e-23, 2.2161738e-23, 2.2161737e-23], [1828.0726, 3.1400471e-23, 2.220975e-23, 2.2209749e-23], [1832.2162, 3.1468565e-23, 2.2257876e-23, 2.2257875e-23], [1836.3692, 3.1536821e-23, 2.2306116e-23, 2.2306115e-23], [1840.5316, 3.1605237e-23, 2.235447e-23, 2.2354469e-23], [1844.7034, 3.1673816e-23, 2.2402939e-23, 2.2402938e-23], [1848.8847, 3.1742557e-23, 2.2451522e-23, 2.2451521e-23], [1853.0754, 3.181146e-23, 2.2500221e-23, 2.250022e-23], [1857.2757, 3.1880527e-23, 2.2549034e-23, 2.2549034e-23], [1861.4855, 3.1949756e-23, 2.2597964e-23, 2.2597963e-23], [1865.7048, 3.2019149e-23, 2.2647009e-23, 2.2647008e-23], [1869.9337, 3.2088706e-23, 2.269617e-23, 2.2696169e-23], [1874.1722, 3.2158427e-23, 2.2745447e-23, 2.2745446e-23], [1878.4202, 3.2228313e-23, 2.2794841e-23, 2.279484e-23], [1882.678, 3.2298364e-23, 2.2844351e-23, 2.284435e-23], [1886.9453, 3.2368581e-23, 2.2893979e-23, 2.2893978e-23], [1891.2223, 3.2438963e-23, 2.2943724e-23, 2.2943723e-23], [1895.5091, 3.2509511e-23, 2.2993586e-23, 2.2993585e-23], [1899.8055, 3.2580226e-23, 2.3043567e-23, 2.3043566e-23], [1904.1117, 3.2651107e-23, 2.3093665e-23, 2.3093665e-23], [1908.4276, 3.2722157e-23, 2.3143883e-23, 2.3143882e-23], [1912.7534, 3.2793374e-23, 2.319422e-23, 2.3194219e-23], [1917.0889, 3.2864767e-23, 2.3244686e-23, 2.3244686e-23], [1921.4343, 3.2954526e-23, 2.3320994e-23, 2.3320993e-23], [1925.7895, 3.3008042e-23, 2.3345953e-23, 2.3345952e-23], [1930.1545, 3.3079922e-23, 2.3396751e-23, 2.3396751e-23], [1934.5295, 3.315198e-23, 2.3447681e-23, 2.344768e-23], [1938.9144, 3.3224209e-23, 2.3498732e-23, 2.3498731e-23], [1943.3092, 3.3296608e-23, 2.3549904e-23, 2.3549903e-23], [1947.714, 3.3369178e-23, 2.3601197e-23, 2.3601196e-23], [1952.1288, 3.3441918e-23, 2.365261e-23, 2.365261e-23], [1956.5536, 3.351483e-23, 2.3704145e-23, 2.3704145e-23], [1960.9884, 3.3587914e-23, 2.3755802e-23, 2.3755801e-23], [1965.4332, 3.366117e-23, 2.380758e-23, 2.3807579e-23], [1969.8882, 3.3734598e-23, 2.385948e-23, 2.385948e-23], [1974.3532, 3.3808199e-23, 2.3911503e-23, 2.3911502e-23], [1978.8284, 3.3881973e-23, 2.3963648e-23, 2.3963647e-23], [1983.3137, 3.395592e-23, 2.4015916e-23, 2.4015915e-23], [1987.8091, 3.4030042e-23, 2.4068306e-23, 2.4068306e-23], [1992.3148, 3.4104338e-23, 2.4120821e-23, 2.412082e-23], [1996.8307, 3.4178808e-23, 2.4173458e-23, 2.4173458e-23], [2001.3568, 3.4253453e-23, 2.422622e-23, 2.4226219e-23], [2005.8931, 3.4328274e-23, 2.4279106e-23, 2.4279105e-23], [2010.4398, 3.4403271e-23, 2.4332116e-23, 2.4332115e-23], [2014.9967, 3.4478444e-23, 2.4385251e-23, 2.438525e-23], [2019.564, 3.4553793e-23, 2.443851e-23, 2.443851e-23], [2024.1416, 3.462932e-23, 2.4491895e-23, 2.4491895e-23], [2028.7296, 3.4705024e-23, 2.4545406e-23, 2.4545405e-23], [2033.328, 3.4780905e-23, 2.4599042e-23, 2.4599041e-23], [2037.9369, 3.4856965e-23, 2.4652804e-23, 2.4652804e-23], [2042.5562, 3.4933204e-23, 2.4706693e-23, 2.4706692e-23], [2047.1859, 3.5009621e-23, 2.4760708e-23, 2.4760708e-23], [2051.8261, 3.5086217e-23, 2.481485e-23, 2.481485e-23], [2056.4769, 3.5162994e-23, 2.4869119e-23, 2.4869119e-23], [2061.1382, 3.523995e-23, 2.4923516e-23, 2.4923516e-23], [2065.8101, 3.5317087e-23, 2.497804e-23, 2.497804e-23], [2070.4925, 3.5394405e-23, 2.5032693e-23, 2.5032692e-23], [2075.1856, 3.5471904e-23, 2.5087473e-23, 2.5087473e-23], [2079.8893, 3.5549585e-23, 2.5142383e-23, 2.5142382e-23], [2084.6037, 3.5627447e-23, 2.5197421e-23, 2.519742e-23], [2089.3287, 3.5705493e-23, 2.5252588e-23, 2.5252587e-23], [2094.0645, 3.5783721e-23, 2.5307884e-23, 2.5307884e-23], [2098.811, 3.5862133e-23, 2.5363311e-23, 2.536331e-23], [2103.5682, 3.5940728e-23, 2.5418867e-23, 2.5418866e-23], [2108.3363, 3.6019507e-23, 2.5474553e-23, 2.5474553e-23], [2113.1151, 3.6098471e-23, 2.553037e-23, 2.553037e-23], [2117.9048, 3.617762e-23, 2.5586318e-23, 2.5586318e-23], [2122.7053, 3.6256954e-23, 2.5642397e-23, 2.5642397e-23], [2127.5168, 3.6336474e-23, 2.5698608e-23, 2.5698607e-23], [2132.3391, 3.641618e-23, 2.575495e-23, 2.5754949e-23], [2137.1723, 3.6496073e-23, 2.5811424e-23, 2.5811423e-23], [2142.0165, 3.6576152e-23, 2.586803e-23, 2.586803e-23], [2146.8717, 3.6656419e-23, 2.5924769e-23, 2.5924769e-23], [2151.7379, 3.6736874e-23, 2.5981641e-23, 2.598164e-23], [2156.6151, 3.6817517e-23, 2.6038646e-23, 2.6038645e-23], [2161.5034, 3.6898349e-23, 2.6095784e-23, 2.6095784e-23], [2166.4028, 3.6979369e-23, 2.6153056e-23, 2.6153056e-23], [2171.3132, 3.7060579e-23, 2.6210462e-23, 2.6210462e-23], [2176.2348, 3.7141979e-23, 2.6268003e-23, 2.6268002e-23], [2181.1676, 3.7223569e-23, 2.6325678e-23, 2.6325677e-23], [2186.1115, 3.730535e-23, 2.6383487e-23, 2.6383487e-23], [2191.0667, 3.7387322e-23, 2.6441433e-23, 2.6441432e-23], [2196.033, 3.7469485e-23, 2.6499513e-23, 2.6499513e-23], [2201.0106, 3.7551841e-23, 2.655773e-23, 2.6557729e-23], [2205.9996, 3.7634389e-23, 2.6616082e-23, 2.6616082e-23], [2210.9998, 3.7717129e-23, 2.6674571e-23, 2.6674571e-23], [2216.0113, 3.7800063e-23, 2.6733197e-23, 2.6733197e-23], [2221.0342, 3.7883191e-23, 2.679196e-23, 2.6791959e-23], [2226.0685, 3.7966512e-23, 2.685086e-23, 2.6850859e-23], [2231.1142, 3.8050029e-23, 2.6909897e-23, 2.6909897e-23], [2236.1714, 3.813374e-23, 2.6969073e-23, 2.6969072e-23], [2241.24, 3.8217646e-23, 2.7028386e-23, 2.7028386e-23], [2246.3201, 3.8301748e-23, 2.7087839e-23, 2.7087838e-23], [2251.4117, 3.8386047e-23, 2.714743e-23, 2.7147429e-23], [2256.5148, 3.8470542e-23, 2.720716e-23, 2.720716e-23], [2261.6295, 3.8555234e-23, 2.726703e-23, 2.7267029e-23], [2266.7558, 3.8640124e-23, 2.7327039e-23, 2.7327039e-23], [2271.8938, 3.8725212e-23, 2.7387188e-23, 2.7387188e-23], [2277.0433, 3.8810498e-23, 2.7447478e-23, 2.7447478e-23], [2282.2046, 3.8895983e-23, 2.7507909e-23, 2.7507908e-23], [2287.3775, 3.8981668e-23, 2.756848e-23, 2.756848e-23], [2292.5622, 3.9067552e-23, 2.7629193e-23, 2.7629193e-23], [2297.7586, 3.9153636e-23, 2.7690047e-23, 2.7690047e-23], [2302.9668, 3.9239921e-23, 2.7751044e-23, 2.7751043e-23], [2308.1868, 3.9326407e-23, 2.7812182e-23, 2.7812182e-23], [2313.4187, 3.9413094e-23, 2.7873463e-23, 2.7873463e-23], [2318.6624, 3.9499984e-23, 2.7934887e-23, 2.7934887e-23], [2323.9179, 3.9587075e-23, 2.7996454e-23, 2.7996454e-23], [2329.1854, 3.967437e-23, 2.8058165e-23, 2.8058164e-23], [2334.4649, 3.9761868e-23, 2.8120019e-23, 2.8120019e-23], [2339.7563, 3.9849569e-23, 2.8182018e-23, 2.8182017e-23], [2345.0597, 3.9937475e-23, 2.8244161e-23, 2.824416e-23], [2350.3751, 4.0025586e-23, 2.8306448e-23, 2.8306448e-23], [2355.7025, 4.0113901e-23, 2.8368881e-23, 2.8368881e-23], [2361.0421, 4.0202422e-23, 2.8431459e-23, 2.8431459e-23], [2366.3937, 4.0291149e-23, 2.8494183e-23, 2.8494183e-23], [2371.7575, 4.0380083e-23, 2.8557053e-23, 2.8557053e-23], [2377.1334, 4.0469223e-23, 2.8620069e-23, 2.8620069e-23], [2382.5215, 4.0558571e-23, 2.8683233e-23, 2.8683232e-23], [2387.9219, 4.0648127e-23, 2.8746543e-23, 2.8746543e-23], [2393.3344, 4.0737892e-23, 2.8810001e-23, 2.8810001e-23], [2398.7593, 4.0827871e-23, 2.8873616e-23, 2.8873616e-23], [2404.1964, 4.0918061e-23, 2.893738e-23, 2.893738e-23], [2409.6459, 4.1008438e-23, 2.9001261e-23, 2.9001261e-23], [2415.1077, 4.1099039e-23, 2.906531e-23, 2.9065309e-23], [2420.5819, 4.1189851e-23, 2.9129508e-23, 2.9129508e-23], [2426.0685, 4.1280874e-23, 2.9193856e-23, 2.9193856e-23], [2431.5675, 4.1372109e-23, 2.9258354e-23, 2.9258354e-23], [2437.079, 4.1463556e-23, 2.9323002e-23, 2.9323002e-23], [2442.603, 4.1555215e-23, 2.93878e-23, 2.93878e-23], [2448.1395, 4.1647087e-23, 2.9452749e-23, 2.9452749e-23], [2453.6885, 4.1739173e-23, 2.9517849e-23, 2.9517848e-23], [2459.2502, 4.1831472e-23, 2.95831e-23, 2.9583099e-23], [2464.8244, 4.1923986e-23, 2.9648502e-23, 2.9648502e-23], [2470.4113, 4.2016715e-23, 2.9714057e-23, 2.9714057e-23], [2476.0108, 4.2109659e-23, 2.9779764e-23, 2.9779763e-23], [2481.6231, 4.2202819e-23, 2.9845623e-23, 2.9845623e-23], [2487.248, 4.2296195e-23, 2.9911636e-23, 2.9911636e-23], [2492.8857, 4.2389788e-23, 2.9977802e-23, 2.9977801e-23], [2498.5362, 4.2483598e-23, 3.0044121e-23, 3.0044121e-23], [2504.1995, 4.2577626e-23, 3.0110595e-23, 3.0110594e-23], [2509.8756, 4.2671872e-23, 3.0177223e-23, 3.0177222e-23], [2515.5646, 4.2766337e-23, 3.0244005e-23, 3.0244005e-23], [2521.2665, 4.286102e-23, 3.0310943e-23, 3.0310942e-23], [2526.9813, 4.2955924e-23, 3.0378035e-23, 3.0378035e-23], [2532.7091, 4.3051047e-23, 3.0445284e-23, 3.0445284e-23], [2538.4499, 4.3146392e-23, 3.0512689e-23, 3.0512689e-23], [2544.2036, 4.3241957e-23, 3.058025e-23, 3.058025e-23], [2549.9704, 4.3337744e-23, 3.0647968e-23, 3.0647967e-23], [2555.7503, 4.3433753e-23, 3.0715843e-23, 3.0715842e-23], [2561.5433, 4.3529984e-23, 3.0783875e-23, 3.0783875e-23], [2567.3494, 4.3626439e-23, 3.0852065e-23, 3.0852065e-23], [2573.1686, 4.3723117e-23, 3.0920413e-23, 3.0920413e-23], [2579.0011, 4.382002e-23, 3.098892e-23, 3.098892e-23], [2584.8468, 4.3917147e-23, 3.1057586e-23, 3.1057586e-23], [2590.7057, 4.4014498e-23, 3.1126411e-23, 3.112641e-23], [2596.5779, 4.4112076e-23, 3.1195395e-23, 3.1195395e-23], [2602.4634, 4.420988e-23, 3.1264539e-23, 3.1264539e-23], [2608.3623, 4.430791e-23, 3.1333844e-23, 3.1333844e-23], [2614.2745, 4.4406167e-23, 3.1403309e-23, 3.1403309e-23], [2620.2001, 4.4504652e-23, 3.1472935e-23, 3.1472935e-23], [2626.1392, 4.4603364e-23, 3.1542722e-23, 3.1542722e-23], [2632.0917, 4.4702306e-23, 3.1612671e-23, 3.1612671e-23], [2638.0577, 4.4801476e-23, 3.1682783e-23, 3.1682782e-23], [2644.0373, 4.4900876e-23, 3.1753056e-23, 3.1753056e-23], [2650.0304, 4.5000507e-23, 3.1823492e-23, 3.1823492e-23], [2656.037, 4.5100368e-23, 3.1894092e-23, 3.1894092e-23], [2662.0573, 4.520046e-23, 3.1964855e-23, 3.1964855e-23], [2668.0913, 4.5300783e-23, 3.2035782e-23, 3.2035781e-23], [2674.1389, 4.5401339e-23, 3.2106873e-23, 3.2106872e-23], [2680.2002, 4.5502127e-23, 3.2178128e-23, 3.2178128e-23], [2686.2753, 4.5603149e-23, 3.2249549e-23, 3.2249548e-23], [2692.3641, 4.5704404e-23, 3.2321134e-23, 3.2321134e-23], [2698.4667, 4.5805894e-23, 3.2392886e-23, 3.2392886e-23], [2704.5832, 4.5907618e-23, 3.2464803e-23, 3.2464803e-23], [2710.7135, 4.6009578e-23, 3.2536887e-23, 3.2536887e-23], [2716.8577, 4.6111773e-23, 3.2609138e-23, 3.2609138e-23], [2723.0159, 4.6214205e-23, 3.2681556e-23, 3.2681556e-23], [2729.188, 4.6316874e-23, 3.2754141e-23, 3.2754141e-23], [2735.3741, 4.641978e-23, 3.2826895e-23, 3.2826895e-23], [2741.5742, 4.6522924e-23, 3.2899816e-23, 3.2899816e-23], [2747.7884, 4.6626306e-23, 3.2972906e-23, 3.2972906e-23], [2754.0166, 4.6729927e-23, 3.3046166e-23, 3.3046166e-23], [2760.259, 4.6833788e-23, 3.3119594e-23, 3.3119594e-23], [2766.5155, 4.6937889e-23, 3.3193193e-23, 3.3193193e-23], [2772.7862, 4.704223e-23, 3.3266961e-23, 3.3266961e-23], [2779.0712, 4.7146813e-23, 3.33409e-23, 3.33409e-23], [2785.3703, 4.7251637e-23, 3.341501e-23, 3.341501e-23], [2791.6838, 4.7356703e-23, 3.3489291e-23, 3.3489291e-23], [2798.0115, 4.7462012e-23, 3.3563744e-23, 3.3563744e-23], [2804.3536, 4.7567564e-23, 3.3638369e-23, 3.3638369e-23], [2810.7101, 4.767336e-23, 3.3713167e-23, 3.3713167e-23], [2817.081, 4.7779401e-23, 3.3788137e-23, 3.3788137e-23], [2823.4663, 4.7885686e-23, 3.386328e-23, 3.386328e-23], [2829.8661, 4.7992217e-23, 3.3938597e-23, 3.3938597e-23], [2836.2804, 4.8098993e-23, 3.4014088e-23, 3.4014088e-23], [2842.7092, 4.8206016e-23, 3.4089754e-23, 3.4089754e-23], [2849.1526, 4.8313287e-23, 3.4165594e-23, 3.4165594e-23], [2855.6106, 4.8420804e-23, 3.4241609e-23, 3.4241609e-23], [2862.0833, 4.852857e-23, 3.43178e-23, 3.43178e-23], [2868.5706, 4.8636585e-23, 3.4394167e-23, 3.4394167e-23], [2875.0726, 4.874485e-23, 3.447071e-23, 3.447071e-23], [2881.5894, 4.885343e-23, 3.4547523e-23, 3.4547523e-23], [2888.1209, 4.8962128e-23, 3.4624327e-23, 3.4624327e-23], [2894.6673, 4.9071143e-23, 3.47014e-23, 3.47014e-23], [2901.2285, 4.9180409e-23, 3.4778652e-23, 3.4778652e-23], [2907.8045, 4.9289928e-23, 3.4856083e-23, 3.4856082e-23], [2914.3955, 4.9399699e-23, 3.4933692e-23, 3.4933692e-23], [2921.0014, 4.9509724e-23, 3.501148e-23, 3.501148e-23], [2927.6222, 4.9620002e-23, 3.5089447e-23, 3.5089447e-23], [2934.2581, 4.9730534e-23, 3.5167595e-23, 3.5167595e-23], [2940.909, 4.9841322e-23, 3.5245923e-23, 3.5245922e-23], [2947.575, 4.9952365e-23, 3.5324431e-23, 3.5324431e-23], [2954.2561, 5.0063664e-23, 3.5403121e-23, 3.540312e-23], [2960.9524, 5.0175219e-23, 3.5481992e-23, 3.5481991e-23], [2967.6638, 5.0287032e-23, 3.5561044e-23, 3.5561044e-23], [2974.3904, 5.0399103e-23, 3.564028e-23, 3.5640279e-23], [2981.1323, 5.0511431e-23, 3.5719697e-23, 3.5719697e-23], [2987.8895, 5.0624019e-23, 3.5799298e-23, 3.5799298e-23], [2994.662, 5.0736866e-23, 3.5879083e-23, 3.5879083e-23], [3001.4498, 5.0849973e-23, 3.5959051e-23, 3.5959051e-23], [3008.253, 5.0963341e-23, 3.6039204e-23, 3.6039204e-23], [3015.0716, 5.107697e-23, 3.6119541e-23, 3.6119541e-23], [3021.9057, 5.1190861e-23, 3.6200064e-23, 3.6200064e-23], [3028.7553, 5.1305014e-23, 3.6280772e-23, 3.6280772e-23], [3035.6204, 5.1419431e-23, 3.6361667e-23, 3.6361666e-23], [3042.5011, 5.153411e-23, 3.6442747e-23, 3.6442747e-23], [3049.3974, 5.1649054e-23, 3.6524014e-23, 3.6524014e-23], [3056.3093, 5.1764263e-23, 3.6605469e-23, 3.6605469e-23], [3063.2368, 5.1879737e-23, 3.6687111e-23, 3.6687111e-23], [3070.1801, 5.1995476e-23, 3.6768942e-23, 3.6768942e-23], [3077.1391, 5.2111483e-23, 3.685096e-23, 3.685096e-23], [3084.1139, 5.2227756e-23, 3.6933168e-23, 3.6933168e-23], [3091.1045, 5.2344297e-23, 3.7015565e-23, 3.7015565e-23], [3098.1109, 5.2461107e-23, 3.7098152e-23, 3.7098152e-23], [3105.1332, 5.2578185e-23, 3.7180929e-23, 3.7180929e-23], [3112.1714, 5.2695533e-23, 3.7263897e-23, 3.7263896e-23], [3119.2256, 5.281315e-23, 3.7347055e-23, 3.7347055e-23], [3126.2958, 5.2931039e-23, 3.7430405e-23, 3.7430405e-23], [3133.382, 5.3049199e-23, 3.7513947e-23, 3.7513947e-23], [3140.4843, 5.3167631e-23, 3.7597681e-23, 3.7597681e-23], [3147.6026, 5.3286335e-23, 3.7681608e-23, 3.7681608e-23], [3154.7371, 5.3405312e-23, 3.7765728e-23, 3.7765728e-23], [3161.8878, 5.3524564e-23, 3.7850042e-23, 3.7850042e-23], [3169.0546, 5.3644089e-23, 3.793455e-23, 3.793455e-23], [3176.2378, 5.3763889e-23, 3.8019252e-23, 3.8019252e-23], [3183.4372, 5.3883965e-23, 3.8104149e-23, 3.8104149e-23], [3190.6529, 5.4004318e-23, 3.8189241e-23, 3.8189241e-23], [3197.885, 5.4124947e-23, 3.827453e-23, 3.8274529e-23], [3205.1334, 5.4245853e-23, 3.8360014e-23, 3.8360014e-23], [3212.3983, 5.4367038e-23, 3.8445695e-23, 3.8445695e-23], [3219.6797, 5.4488501e-23, 3.8531573e-23, 3.8531573e-23], [3226.9775, 5.4610243e-23, 3.8617649e-23, 3.8617649e-23], [3234.2919, 5.4732265e-23, 3.8703922e-23, 3.8703922e-23], [3241.6229, 5.4854568e-23, 3.8790394e-23, 3.8790394e-23], [3248.9705, 5.4977152e-23, 3.8877065e-23, 3.8877065e-23], [3256.3348, 5.5100018e-23, 3.8963935e-23, 3.8963935e-23], [3263.7158, 5.5223166e-23, 3.9051005e-23, 3.9051005e-23], [3271.1134, 5.5346598e-23, 3.9138275e-23, 3.9138275e-23], [3278.5279, 5.5470313e-23, 3.9225746e-23, 3.9225746e-23], [3285.9591, 5.5594312e-23, 3.9313418e-23, 3.9313417e-23], [3293.4072, 5.5718596e-23, 3.9401291e-23, 3.9401291e-23], [3300.8722, 5.5843166e-23, 3.9489366e-23, 3.9489366e-23], [3308.3541, 5.5968022e-23, 3.9577644e-23, 3.9577644e-23], [3315.853, 5.6093165e-23, 3.9666124e-23, 3.9666124e-23], [3323.3688, 5.6218596e-23, 3.9754808e-23, 3.9754808e-23], [3330.9017, 5.6344314e-23, 3.9843696e-23, 3.9843696e-23], [3338.4517, 5.6470322e-23, 3.9932788e-23, 3.9932788e-23], [3346.0188, 5.6596619e-23, 4.0022085e-23, 4.0022085e-23], [3353.603, 5.6723206e-23, 4.0111587e-23, 4.0111587e-23], [3361.2044, 5.6850091e-23, 4.0201305e-23, 4.0201305e-23], [3368.8231, 5.6977253e-23, 4.0291208e-23, 4.0291208e-23], [3376.459, 5.7104714e-23, 4.0381328e-23, 4.0381328e-23], [3384.1122, 5.7232468e-23, 4.0471655e-23, 4.0471655e-23], [3391.7828, 5.7360515e-23, 4.0562189e-23, 4.0562189e-23], [3399.4708, 5.7488856e-23, 4.0652932e-23, 4.0652932e-23], [3407.1762, 5.7617492e-23, 4.0743883e-23, 4.0743883e-23], [3414.899, 5.7746424e-23, 4.0835042e-23, 4.0835042e-23], [3422.6394, 5.7875651e-23, 4.0926411e-23, 4.0926411e-23], [3430.3973, 5.8005175e-23, 4.101799e-23, 4.101799e-23], [3438.1728, 5.8134996e-23, 4.1109779e-23, 4.1109779e-23], [3445.9659, 5.8265115e-23, 4.1201779e-23, 4.1201779e-23], [3453.7767, 5.8395533e-23, 4.129399e-23, 4.129399e-23], [3461.6051, 5.8526251e-23, 4.1386413e-23, 4.1386413e-23], [3469.4514, 5.8657268e-23, 4.1479048e-23, 4.1479048e-23], [3477.3154, 5.8788586e-23, 4.1571896e-23, 4.1571896e-23], [3485.1972, 5.8920205e-23, 4.1664956e-23, 4.1664956e-23], [3493.0969, 5.9052126e-23, 4.1758231e-23, 4.1758231e-23], [3501.0145, 5.918435e-23, 4.1851719e-23, 4.1851719e-23], [3508.9501, 5.9316877e-23, 4.1945422e-23, 4.1945422e-23], [3516.9036, 5.9449709e-23, 4.203934e-23, 4.203934e-23], [3524.8752, 5.9582845e-23, 4.2133473e-23, 4.2133473e-23], [3532.8648, 5.9716286e-23, 4.2227823e-23, 4.2227823e-23], [3540.8726, 5.9850034e-23, 4.2322388e-23, 4.2322388e-23], [3548.8985, 5.9984088e-23, 4.2417171e-23, 4.2417171e-23], [3556.9425, 6.011845e-23, 4.2512172e-23, 4.2512171e-23], [3565.0049, 6.025312e-23, 4.260739e-23, 4.260739e-23], [3573.0855, 6.0388099e-23, 4.2702826e-23, 4.2702826e-23], [3581.1844, 6.0523387e-23, 4.2798482e-23, 4.2798482e-23], [3589.3016, 6.0658985e-23, 4.2894357e-23, 4.2894357e-23], [3597.4373, 6.0794895e-23, 4.2990451e-23, 4.2990451e-23], [3605.5914, 6.0931116e-23, 4.3086767e-23, 4.3086767e-23], [3613.764, 6.106765e-23, 4.3183303e-23, 4.3183303e-23], [3621.9551, 6.1204496e-23, 4.328006e-23, 4.328006e-23], [3630.1648, 6.1341656e-23, 4.3377039e-23, 4.3377039e-23], [3638.3931, 6.1479131e-23, 4.3474241e-23, 4.3474241e-23], [3646.64, 6.1616921e-23, 4.3571666e-23, 4.3571666e-23], [3654.9057, 6.1755026e-23, 4.3669313e-23, 4.3669313e-23], [3663.19, 6.1893448e-23, 4.3767185e-23, 4.3767185e-23], [3671.4932, 6.2032188e-23, 4.3865281e-23, 4.3865281e-23], [3679.8152, 6.2171245e-23, 4.3963602e-23, 4.3963602e-23], [3688.156, 6.2310621e-23, 4.4062149e-23, 4.4062148e-23], [3696.5157, 6.2450317e-23, 4.4160921e-23, 4.4160921e-23], [3704.8944, 6.2590332e-23, 4.4259919e-23, 4.4259919e-23], [3713.2921, 6.2730669e-23, 4.4359145e-23, 4.4359145e-23], [3721.7088, 6.2871327e-23, 4.4458598e-23, 4.4458598e-23], [3730.1446, 6.3012307e-23, 4.4558278e-23, 4.4558278e-23], [3738.5995, 6.315361e-23, 4.4658188e-23, 4.4658188e-23], [3747.0736, 6.3295237e-23, 4.4758326e-23, 4.4758326e-23], [3755.5668, 6.3437188e-23, 4.4858693e-23, 4.4858693e-23], [3764.0794, 6.3579465e-23, 4.4959291e-23, 4.4959291e-23], [3772.6112, 6.3722067e-23, 4.5060119e-23, 4.5060119e-23], [3781.1624, 6.3864996e-23, 4.5161178e-23, 4.5161178e-23], [3789.7329, 6.4008253e-23, 4.5262468e-23, 4.5262468e-23], [3798.3229, 6.4151837e-23, 4.5363991e-23, 4.5363991e-23], [3806.9323, 6.4295751e-23, 4.5465746e-23, 4.5465746e-23], [3815.5613, 6.4439994e-23, 4.5567734e-23, 4.5567734e-23], [3824.2098, 6.4584567e-23, 4.5669956e-23, 4.5669956e-23], [3832.8779, 6.4729471e-23, 4.5772412e-23, 4.5772412e-23], [3841.5657, 6.4874712e-23, 4.5875108e-23, 4.5875108e-23], [3850.2732, 6.5020276e-23, 4.5978028e-23, 4.5978028e-23], [3859.0004, 6.5166178e-23, 4.6081189e-23, 4.6081189e-23], [3867.7473, 6.5312414e-23, 4.6184587e-23, 4.6184587e-23], [3876.5141, 6.5458985e-23, 4.6288221e-23, 4.6288221e-23], [3885.3008, 6.5605891e-23, 4.6392093e-23, 4.6392093e-23], [3894.1074, 6.5753134e-23, 4.6496202e-23, 4.6496202e-23], [3902.934, 6.5900713e-23, 4.660055e-23, 4.660055e-23], [3911.7805, 6.6048631e-23, 4.6705137e-23, 4.6705137e-23], [3920.6471, 6.6196887e-23, 4.6809963e-23, 4.6809963e-23], [3929.5339, 6.6345482e-23, 4.6915029e-23, 4.6915029e-23], [3938.4407, 6.6494418e-23, 4.7020335e-23, 4.7020335e-23], [3947.3677, 6.6643694e-23, 4.7125883e-23, 4.7125883e-23], [3956.315, 6.6793312e-23, 4.7231672e-23, 4.7231672e-23], [3965.2826, 6.6943272e-23, 4.7337703e-23, 4.7337703e-23], [3974.2705, 6.7093575e-23, 4.7443977e-23, 4.7443977e-23], [3983.2787, 6.7244223e-23, 4.7550495e-23, 4.7550495e-23], [3992.3074, 6.7395215e-23, 4.7657256e-23, 4.7657256e-23], [4001.3565, 6.7546552e-23, 4.7764261e-23, 4.7764261e-23], [4010.4262, 6.7698236e-23, 4.7871511e-23, 4.7871511e-23], [4019.5164, 6.7850266e-23, 4.7979006e-23, 4.7979006e-23], [4028.6272, 6.8002645e-23, 4.8086748e-23, 4.8086748e-23], [4037.7587, 6.8155372e-23, 4.8194736e-23, 4.8194736e-23], [4046.9108, 6.8308449e-23, 4.8302971e-23, 4.8302971e-23], [4056.0837, 6.8461875e-23, 4.8411454e-23, 4.8411453e-23], [4065.2774, 6.8615653e-23, 4.8520184e-23, 4.8520184e-23], [4074.492, 6.8769782e-23, 4.8629164e-23, 4.8629164e-23], [4083.7274, 6.8924264e-23, 4.8738393e-23, 4.8738393e-23], [4092.9837, 6.9079099e-23, 4.8847871e-23, 4.8847871e-23], [4102.2611, 6.9234288e-23, 4.89576e-23, 4.89576e-23], [4111.5594, 6.9389832e-23, 4.906758e-23, 4.906758e-23], [4120.8789, 6.9545732e-23, 4.9177812e-23, 4.9177812e-23], [4130.2194, 6.9701988e-23, 4.9288296e-23, 4.9288296e-23], [4139.5812, 6.9858602e-23, 4.9399032e-23, 4.9399032e-23], [4148.9641, 7.0015573e-23, 4.9510022e-23, 4.9510022e-23], [4158.3684, 7.0172904e-23, 4.9621265e-23, 4.9621265e-23], [4167.7939, 7.0330594e-23, 4.9732763e-23, 4.9732763e-23], [4177.2408, 7.0488645e-23, 4.9844516e-23, 4.9844515e-23], [4186.7091, 7.0647057e-23, 4.9956524e-23, 4.9956524e-23], [4196.1989, 7.0805831e-23, 5.0068788e-23, 5.0068788e-23], [4205.7102, 7.0964968e-23, 5.0181309e-23, 5.0181309e-23], [4215.243, 7.1124469e-23, 5.0294087e-23, 5.0294087e-23], [4224.7975, 7.1284335e-23, 5.0407123e-23, 5.0407123e-23], [4234.3736, 7.1444566e-23, 5.0520418e-23, 5.0520418e-23], [4243.9714, 7.1605163e-23, 5.0633971e-23, 5.0633971e-23], [4253.591, 7.1766127e-23, 5.0747784e-23, 5.0747784e-23], [4263.2324, 7.1927459e-23, 5.0861857e-23, 5.0861857e-23], [4272.8956, 7.208916e-23, 5.0976191e-23, 5.0976191e-23], [4282.5807, 7.225123e-23, 5.1090786e-23, 5.1090786e-23], [4292.2878, 7.2413671e-23, 5.1205643e-23, 5.1205643e-23], [4302.0169, 7.2576482e-23, 5.1320763e-23, 5.1320763e-23], [4311.768, 7.2739666e-23, 5.1436146e-23, 5.1436146e-23], [4321.5413, 7.2903225e-23, 5.1551794e-23, 5.1551794e-23], [4331.3367, 7.3067154e-23, 5.1667703e-23, 5.1667703e-23], [4341.1543, 7.3231458e-23, 5.1783878e-23, 5.1783878e-23], [4350.9941, 7.3396139e-23, 5.1900319e-23, 5.1900319e-23], [4360.8563, 7.3561195e-23, 5.2017026e-23, 5.2017026e-23], [4370.7408, 7.3726629e-23, 5.2134e-23, 5.2134e-23], [4380.6477, 7.3892441e-23, 5.225124e-23, 5.225124e-23], [4390.5771, 7.4058631e-23, 5.2368749e-23, 5.2368749e-23], [4400.5289, 7.4225201e-23, 5.2486526e-23, 5.2486526e-23], [4410.5034, 7.4392151e-23, 5.2604573e-23, 5.2604573e-23], [4420.5004, 7.4559483e-23, 5.2722889e-23, 5.2722889e-23], [4430.5201, 7.4727197e-23, 5.2841475e-23, 5.2841475e-23], [4440.5625, 7.4895294e-23, 5.2960332e-23, 5.2960332e-23], [4450.6277, 7.5063775e-23, 5.307946e-23, 5.307946e-23], [4460.7157, 7.5232641e-23, 5.3198861e-23, 5.3198861e-23], [4470.8265, 7.5401892e-23, 5.3318534e-23, 5.3318534e-23], [4480.9603, 7.557153e-23, 5.3438481e-23, 5.3438481e-23], [4491.117, 7.5741555e-23, 5.3558702e-23, 5.3558702e-23], [4501.2968, 7.5911969e-23, 5.3679197e-23, 5.3679197e-23], [4511.4996, 7.6082772e-23, 5.3799967e-23, 5.3799967e-23], [4521.7256, 7.6253964e-23, 5.3921013e-23, 5.3921013e-23], [4531.9747, 7.6425548e-23, 5.4042336e-23, 5.4042336e-23], [4542.2471, 7.6597523e-23, 5.4163936e-23, 5.4163936e-23], [4552.5428, 7.6769891e-23, 5.4285813e-23, 5.4285813e-23], [4562.8618, 7.6942652e-23, 5.4407968e-23, 5.4407968e-23], [4573.2041, 7.7115808e-23, 5.4530403e-23, 5.4530403e-23], [4583.57, 7.7289359e-23, 5.4653117e-23, 5.4653117e-23], [4593.9593, 7.7463307e-23, 5.4776111e-23, 5.4776111e-23], [4604.3722, 7.7637651e-23, 5.4899386e-23, 5.4899386e-23], [4614.8086, 7.7812394e-23, 5.5022942e-23, 5.5022942e-23], [4625.2688, 7.7987535e-23, 5.5146781e-23, 5.5146781e-23], [4635.7526, 7.8163076e-23, 5.5270902e-23, 5.5270902e-23], [4646.2602, 7.8339018e-23, 5.5395306e-23, 5.5395306e-23], [4656.7916, 7.8515361e-23, 5.5519995e-23, 5.5519995e-23], [4667.3469, 7.8692107e-23, 5.5644968e-23, 5.5644968e-23], [4677.9261, 7.8869256e-23, 5.5770226e-23, 5.5770226e-23], [4688.5293, 7.9046809e-23, 5.589577e-23, 5.589577e-23], [4699.1565, 7.9224768e-23, 5.6021601e-23, 5.6021601e-23], [4709.8078, 7.9403133e-23, 5.6147719e-23, 5.6147719e-23], [4720.4833, 7.9581904e-23, 5.6274125e-23, 5.6274125e-23], [4731.1829, 7.9761084e-23, 5.6400819e-23, 5.6400819e-23], [4741.9069, 7.9940673e-23, 5.6527802e-23, 5.6527802e-23], [4752.6551, 8.0120671e-23, 5.6655076e-23, 5.6655076e-23], [4763.4276, 8.030108e-23, 5.6782639e-23, 5.6782639e-23], [4774.2246, 8.04819e-23, 5.6910494e-23, 5.6910494e-23], [4785.0461, 8.0663134e-23, 5.703864e-23, 5.703864e-23], [4795.8921, 8.084478e-23, 5.7167079e-23, 5.7167079e-23], [4806.7627, 8.1026842e-23, 5.7295811e-23, 5.7295811e-23], [4817.6579, 8.1209318e-23, 5.7424836e-23, 5.7424836e-23], [4828.5778, 8.1392211e-23, 5.7554156e-23, 5.7554156e-23], [4839.5225, 8.1575521e-23, 5.7683771e-23, 5.7683771e-23], [4850.4919, 8.1759249e-23, 5.7813682e-23, 5.7813682e-23], [4861.4863, 8.1943396e-23, 5.7943889e-23, 5.7943889e-23], [4872.5055, 8.2127963e-23, 5.8074393e-23, 5.8074393e-23], [4883.5498, 8.2312951e-23, 5.8205194e-23, 5.8205194e-23], [4894.619, 8.2498362e-23, 5.8336294e-23, 5.8336294e-23], [4905.7134, 8.2684195e-23, 5.8467694e-23, 5.8467694e-23], [4916.8329, 8.2870452e-23, 5.8599393e-23, 5.8599392e-23], [4927.9776, 8.3057133e-23, 5.8731392e-23, 5.8731392e-23], [4939.1476, 8.3244241e-23, 5.8863692e-23, 5.8863692e-23], [4950.3429, 8.3431775e-23, 5.8996294e-23, 5.8996294e-23], [4961.5635, 8.3619737e-23, 5.9129199e-23, 5.9129199e-23], [4972.8096, 8.3808127e-23, 5.9262406e-23, 5.9262406e-23], [4984.0812, 8.3996947e-23, 5.9395918e-23, 5.9395918e-23], [4995.3784, 8.4186198e-23, 5.9529734e-23, 5.9529734e-23], [5006.7011, 8.437588e-23, 5.9663855e-23, 5.9663855e-23], [5018.0495, 8.4565995e-23, 5.9798282e-23, 5.9798282e-23], [5029.4236, 8.4756543e-23, 5.9933015e-23, 5.9933015e-23], [5040.8236, 8.4947526e-23, 6.0068056e-23, 6.0068056e-23], [5052.2493, 8.5138944e-23, 6.0203404e-23, 6.0203404e-23], [5063.701, 8.5330799e-23, 6.0339062e-23, 6.0339062e-23], [5075.1786, 8.5523091e-23, 6.0475028e-23, 6.0475028e-23], [5086.6822, 8.5715821e-23, 6.0611305e-23, 6.0611305e-23], [5098.2119, 8.5908991e-23, 6.0747892e-23, 6.0747892e-23], [5109.7677, 8.6102601e-23, 6.0884791e-23, 6.0884791e-23], [5121.3497, 8.6296653e-23, 6.1022001e-23, 6.1022001e-23], [5132.958, 8.6491146e-23, 6.1159525e-23, 6.1159525e-23], [5144.5926, 8.6686084e-23, 6.1297362e-23, 6.1297362e-23], [5156.2535, 8.6881465e-23, 6.1435514e-23, 6.1435514e-23], [5167.9409, 8.7077292e-23, 6.157398e-23, 6.157398e-23], [5179.6548, 8.7273566e-23, 6.1712762e-23, 6.1712762e-23], [5191.3953, 8.7470286e-23, 6.185186e-23, 6.185186e-23], [5203.1623, 8.7667455e-23, 6.1991275e-23, 6.1991275e-23], [5214.956, 8.7865074e-23, 6.2131008e-23, 6.2131008e-23], [5226.7765, 8.8063143e-23, 6.227106e-23, 6.227106e-23], [5238.6237, 8.8261663e-23, 6.2411431e-23, 6.2411431e-23], [5250.4978, 8.8460636e-23, 6.2552122e-23, 6.2552122e-23], [5262.3988, 8.8660062e-23, 6.2693133e-23, 6.2693133e-23], [5274.3268, 8.8859942e-23, 6.2834466e-23, 6.2834466e-23], [5286.2818, 8.9060279e-23, 6.2976121e-23, 6.2976121e-23], [5298.2639, 8.9261071e-23, 6.3118098e-23, 6.3118098e-23], [5310.2732, 8.9462321e-23, 6.32604e-23, 6.32604e-23], [5322.3097, 8.966403e-23, 6.3403025e-23, 6.3403025e-23], [5334.3735, 8.9866199e-23, 6.3545976e-23, 6.3545976e-23], [5346.4646, 9.0068828e-23, 6.3689252e-23, 6.3689252e-23], [5358.5831, 9.0271919e-23, 6.3832855e-23, 6.3832855e-23], [5370.7291, 9.0475472e-23, 6.3976785e-23, 6.3976785e-23], [5382.9027, 9.0679489e-23, 6.4121043e-23, 6.4121043e-23], [5395.1038, 9.0883972e-23, 6.426563e-23, 6.426563e-23], [5407.3326, 9.108892e-23, 6.4410546e-23, 6.4410546e-23], [5419.5891, 9.1294335e-23, 6.4555792e-23, 6.4555792e-23], [5431.8733, 9.1500217e-23, 6.4701369e-23, 6.4701369e-23], [5444.1855, 9.1706569e-23, 6.4847278e-23, 6.4847278e-23], [5456.5255, 9.1913391e-23, 6.499352e-23, 6.499352e-23], [5468.8935, 9.2120685e-23, 6.5140094e-23, 6.5140094e-23], [5481.2895, 9.232845e-23, 6.5287003e-23, 6.5287003e-23], [5493.7136, 9.2536689e-23, 6.5434246e-23, 6.5434246e-23], [5506.1659, 9.2745402e-23, 6.5581824e-23, 6.5581824e-23], [5518.6465, 9.295459e-23, 6.5729739e-23, 6.5729739e-23], [5531.1553, 9.3164255e-23, 6.5877991e-23, 6.5877991e-23], [5543.6924, 9.3374398e-23, 6.602658e-23, 6.602658e-23], [5556.258, 9.3585019e-23, 6.6175508e-23, 6.6175508e-23], [5568.8521, 9.379612e-23, 6.6324775e-23, 6.6324775e-23], [5581.4747, 9.4007702e-23, 6.6474382e-23, 6.6474382e-23], [5594.1259, 9.4219765e-23, 6.662433e-23, 6.662433e-23], [5606.8058, 9.4432312e-23, 6.6774619e-23, 6.6774619e-23], [5619.5144, 9.4645343e-23, 6.6925251e-23, 6.6925251e-23], [5632.2518, 9.4858858e-23, 6.7076226e-23, 6.7076226e-23], [5645.0181, 9.5072861e-23, 6.7227544e-23, 6.7227544e-23], [5657.8134, 9.528735e-23, 6.7379207e-23, 6.7379207e-23], [5670.6376, 9.5502328e-23, 6.7531216e-23, 6.7531216e-23], [5683.4909, 9.5717795e-23, 6.7683571e-23, 6.7683571e-23], [5696.3734, 9.5933753e-23, 6.7836273e-23, 6.7836273e-23], [5709.285, 9.6150203e-23, 6.7989322e-23, 6.7989322e-23], [5722.226, 9.6367146e-23, 6.814272e-23, 6.814272e-23], [5735.1962, 9.6584583e-23, 6.8296467e-23, 6.8296467e-23], [5748.1959, 9.6802514e-23, 6.8450565e-23, 6.8450565e-23], [5761.225, 9.7020943e-23, 6.8605013e-23, 6.8605013e-23], [5774.2836, 9.7239868e-23, 6.8759813e-23, 6.8759813e-23], [5787.3719, 9.7459292e-23, 6.8914966e-23, 6.8914966e-23], [5800.4898, 9.7679215e-23, 6.9070471e-23, 6.9070471e-23], [5813.6374, 9.7899639e-23, 6.9226331e-23, 6.9226331e-23], [5826.8149, 9.8120565e-23, 6.9382546e-23, 6.9382546e-23], [5840.0222, 9.8341994e-23, 6.9539116e-23, 6.9539116e-23], [5853.2594, 9.8563927e-23, 6.9696043e-23, 6.9696043e-23], [5866.5267, 9.8786365e-23, 6.9853327e-23, 6.9853327e-23], [5879.824, 9.900931e-23, 7.0010969e-23, 7.0010969e-23], [5893.1515, 9.9232762e-23, 7.016897e-23, 7.016897e-23], [5906.5091, 9.9456723e-23, 7.0327331e-23, 7.0327331e-23], [5919.8971, 9.9681194e-23, 7.0486053e-23, 7.0486053e-23], [5933.3154, 9.9906175e-23, 7.0645135e-23, 7.0645135e-23], [5946.7641, 1.0013167e-22, 7.080458e-23, 7.080458e-23], [5960.2433, 1.0035768e-22, 7.0964388e-23, 7.0964388e-23], [5973.753, 1.005842e-22, 7.1124559e-23, 7.1124559e-23], [5987.2934, 1.0081124e-22, 7.1285095e-23, 7.1285095e-23], [6000.8645, 1.0103879e-22, 7.1445997e-23, 7.1445997e-23], [6014.4663, 1.0126686e-22, 7.1607264e-23, 7.1607264e-23], [6028.0989, 1.0149545e-22, 7.1768899e-23, 7.1768899e-23], [6041.7625, 1.0172456e-22, 7.1930902e-23, 7.1930902e-23], [6055.457, 1.019542e-22, 7.2093273e-23, 7.2093273e-23], [6069.1826, 1.0218435e-22, 7.2256014e-23, 7.2256014e-23], [6082.9393, 1.0241503e-22, 7.2419126e-23, 7.2419126e-23], [6096.7271, 1.0264623e-22, 7.2582609e-23, 7.2582609e-23], [6110.5462, 1.0287796e-22, 7.2746463e-23, 7.2746463e-23], [6124.3967, 1.0311022e-22, 7.2910691e-23, 7.2910691e-23], [6138.2785, 1.0334301e-22, 7.3075292e-23, 7.3075292e-23], [6152.1918, 1.0357632e-22, 7.3240268e-23, 7.3240268e-23], [6166.1366, 1.0381017e-22, 7.340562e-23, 7.340562e-23], [6180.113, 1.0404455e-22, 7.3571347e-23, 7.3571347e-23], [6194.1211, 1.0427946e-22, 7.3737452e-23, 7.3737452e-23], [6208.161, 1.0451491e-22, 7.3903935e-23, 7.3903935e-23], [6222.2327, 1.0475089e-22, 7.4070797e-23, 7.4070797e-23], [6236.3363, 1.0498741e-22, 7.4238038e-23, 7.4238038e-23], [6250.4718, 1.0522447e-22, 7.440566e-23, 7.440566e-23], [6264.6394, 1.0546207e-22, 7.4573664e-23, 7.4573664e-23], [6278.8391, 1.057002e-22, 7.4742049e-23, 7.4742049e-23], [6293.071, 1.0593888e-22, 7.4910818e-23, 7.4910818e-23], [6307.3352, 1.0617811e-22, 7.5079971e-23, 7.5079971e-23], [6321.6317, 1.0641787e-22, 7.5249509e-23, 7.5249509e-23], [6335.9605, 1.0665819e-22, 7.5419432e-23, 7.5419432e-23], [6350.3219, 1.0689904e-22, 7.5589742e-23, 7.5589742e-23], [6364.7158, 1.0714045e-22, 7.576044e-23, 7.576044e-23], [6379.1424, 1.0738241e-22, 7.5931525e-23, 7.5931525e-23], [6393.6016, 1.0762491e-22, 7.6103001e-23, 7.6103001e-23], [6408.0936, 1.0786797e-22, 7.6274866e-23, 7.6274866e-23], [6422.6185, 1.0811158e-22, 7.6447122e-23, 7.6447122e-23], [6437.1763, 1.0835575e-22, 7.661977e-23, 7.661977e-23], [6451.7671, 1.0860047e-22, 7.6792811e-23, 7.6792811e-23], [6466.3909, 1.0884575e-22, 7.6966246e-23, 7.6966246e-23], [6481.0479, 1.0909158e-22, 7.7140075e-23, 7.7140075e-23], [6495.7382, 1.0933798e-22, 7.7314299e-23, 7.7314299e-23], [6510.4617, 1.0958493e-22, 7.748892e-23, 7.748892e-23], [6525.2186, 1.0983245e-22, 7.7663938e-23, 7.7663938e-23], [6540.0089, 1.1008053e-22, 7.7839354e-23, 7.7839354e-23], [6554.8328, 1.1032918e-22, 7.8015169e-23, 7.8015169e-23], [6569.6903, 1.1057839e-22, 7.8191385e-23, 7.8191385e-23], [6584.5814, 1.1082816e-22, 7.8368e-23, 7.8368e-23], [6599.5063, 1.1107851e-22, 7.8545018e-23, 7.8545018e-23], [6614.465, 1.1132942e-22, 7.8722438e-23, 7.8722438e-23], [6629.4577, 1.1158091e-22, 7.8900262e-23, 7.8900262e-23], [6644.4843, 1.1183296e-22, 7.9078491e-23, 7.9078491e-23], [6659.545, 1.1208559e-22, 7.9257124e-23, 7.9257124e-23], [6674.6398, 1.123388e-22, 7.9436164e-23, 7.9436164e-23], [6689.7688, 1.1259258e-22, 7.9615611e-23, 7.9615611e-23], [6704.9321, 1.1284694e-22, 7.9795467e-23, 7.9795467e-23], [6720.1298, 1.1310187e-22, 7.9975731e-23, 7.9975731e-23], [6735.362, 1.1335739e-22, 8.0156405e-23, 8.0156405e-23], [6750.6286, 1.1361349e-22, 8.033749e-23, 8.033749e-23], [6765.9299, 1.1387017e-22, 8.0518988e-23, 8.0518988e-23], [6781.2659, 1.1412743e-22, 8.0700897e-23, 8.0700897e-23], [6796.6366, 1.1438528e-22, 8.0883221e-23, 8.0883221e-23], [6812.0421, 1.1464371e-22, 8.1065959e-23, 8.1065959e-23], [6827.4826, 1.1490274e-22, 8.1249112e-23, 8.1249112e-23], [6842.9581, 1.1516235e-22, 8.1432683e-23, 8.1432683e-23], [6858.4686, 1.1542255e-22, 8.161667e-23, 8.161667e-23], [6874.0143, 1.1568334e-22, 8.1801076e-23, 8.1801076e-23], [6889.5953, 1.1594473e-22, 8.1985901e-23, 8.1985901e-23], [6905.2115, 1.1620671e-22, 8.2171147e-23, 8.2171147e-23], [6920.8632, 1.1646928e-22, 8.2356814e-23, 8.2356814e-23], [6936.5503, 1.1673246e-22, 8.2542903e-23, 8.2542903e-23], [6952.273, 1.1699623e-22, 8.2729415e-23, 8.2729415e-23], [6968.0313, 1.172606e-22, 8.2916351e-23, 8.2916351e-23], [6983.8254, 1.1752557e-22, 8.3103712e-23, 8.3103712e-23], [6999.6552, 1.1779115e-22, 8.3291499e-23, 8.3291499e-23], [7015.5209, 1.1805733e-22, 8.3479713e-23, 8.3479713e-23], [7031.4226, 1.1832411e-22, 8.3668356e-23, 8.3668356e-23], [7047.3604, 1.185915e-22, 8.3857427e-23, 8.3857427e-23], [7063.3342, 1.188595e-22, 8.4046927e-23, 8.4046927e-23], [7079.3443, 1.1912811e-22, 8.4236859e-23, 8.4236859e-23], [7095.3906, 1.1939733e-22, 8.4427223e-23, 8.4427223e-23], [7111.4734, 1.1966716e-22, 8.4618019e-23, 8.4618019e-23], [7127.5925, 1.199376e-22, 8.4809249e-23, 8.4809249e-23], [7143.7482, 1.2020866e-22, 8.5000914e-23, 8.5000914e-23], [7159.9406, 1.2048033e-22, 8.5193014e-23, 8.5193014e-23], [7176.1696, 1.2075263e-22, 8.5385552e-23, 8.5385552e-23], [7192.4354, 1.2102554e-22, 8.5578527e-23, 8.5578527e-23], [7208.7381, 1.2129907e-22, 8.577194e-23, 8.577194e-23], [7225.0778, 1.2157322e-22, 8.5965793e-23, 8.5965793e-23], [7241.4544, 1.21848e-22, 8.6160087e-23, 8.6160087e-23], [7257.8682, 1.221234e-22, 8.6354823e-23, 8.6354823e-23], [7274.3192, 1.2239943e-22, 8.6550001e-23, 8.6550001e-23], [7290.8075, 1.2267608e-22, 8.6745623e-23, 8.6745623e-23], [7307.3332, 1.2295337e-22, 8.694169e-23, 8.694169e-23], [7323.8963, 1.2323128e-22, 8.7138202e-23, 8.7138202e-23], [7340.497, 1.2350983e-22, 8.7335162e-23, 8.7335162e-23], [7357.1353, 1.2378901e-22, 8.7532568e-23, 8.7532568e-23], [7373.8113, 1.2406882e-22, 8.7730424e-23, 8.7730424e-23], [7390.5251, 1.2434927e-22, 8.7928729e-23, 8.7928729e-23], [7407.2768, 1.2463036e-22, 8.8127485e-23, 8.8127485e-23], [7424.0664, 1.2491208e-22, 8.8326692e-23, 8.8326692e-23], [7440.8941, 1.2519445e-22, 8.8526353e-23, 8.8526353e-23], [7457.76, 1.2547745e-22, 8.8726467e-23, 8.8726467e-23], [7474.6641, 1.2576111e-22, 8.8927036e-23, 8.8927036e-23], [7491.6065, 1.260454e-22, 8.9128061e-23, 8.9128061e-23], [7508.5873, 1.2633034e-22, 8.9329543e-23, 8.9329543e-23], [7525.6066, 1.2661593e-22, 8.9531482e-23, 8.9531482e-23], [7542.6644, 1.2690217e-22, 8.9733881e-23, 8.9733881e-23], [7559.761, 1.2718906e-22, 8.993674e-23, 8.993674e-23], [7576.8963, 1.274766e-22, 9.0140059e-23, 9.0140059e-23], [7594.0704, 1.2776479e-22, 9.0343841e-23, 9.0343841e-23], [7611.2834, 1.2805364e-22, 9.0548086e-23, 9.0548086e-23], [7628.5355, 1.2834315e-22, 9.0752795e-23, 9.0752795e-23], [7645.8267, 1.2863331e-22, 9.0957969e-23, 9.0957969e-23], [7663.157, 1.2892414e-22, 9.116361e-23, 9.116361e-23], [7680.5267, 1.2921562e-22, 9.1369717e-23, 9.1369717e-23], [7697.9357, 1.2950776e-22, 9.1576294e-23, 9.1576294e-23], [7715.3842, 1.2980057e-22, 9.1783339e-23, 9.1783339e-23], [7732.8722, 1.3009405e-22, 9.1990855e-23, 9.1990855e-23], [7750.3999, 1.3038819e-22, 9.2198843e-23, 9.2198843e-23], [7767.9673, 1.30683e-22, 9.2407303e-23, 9.2407303e-23], [7785.5745, 1.3097848e-22, 9.2616237e-23, 9.2616237e-23], [7803.2216, 1.3127464e-22, 9.2825646e-23, 9.2825646e-23], [7820.9087, 1.3157146e-22, 9.3035531e-23, 9.3035531e-23], [7838.6359, 1.3186896e-22, 9.3245892e-23, 9.3245892e-23], [7856.4033, 1.3216713e-22, 9.3456732e-23, 9.3456732e-23], [7874.211, 1.3246599e-22, 9.366805e-23, 9.366805e-23], [7892.059, 1.3276552e-22, 9.3879849e-23, 9.3879849e-23], [7909.9475, 1.3306573e-22, 9.4092129e-23, 9.4092129e-23], [7927.8766, 1.3336663e-22, 9.4304891e-23, 9.4304891e-23], [7945.8462, 1.336682e-22, 9.4518137e-23, 9.4518137e-23], [7963.8566, 1.3397047e-22, 9.4731867e-23, 9.4731867e-23], [7981.9079, 1.3427342e-22, 9.4946083e-23, 9.4946083e-23], [8000.0, 1.3457706e-22, 9.5160785e-23, 9.5160785e-23]])


    f1 = data[:,0]
    S1 = data[:,col]**2
    fstop = [f1[0],f1[-1]]

    #% ---- Interpolate to desired frequencies.  Set noise to infinity outside
    #%      band of frequencies covered by f1.
    ##S = inf(size(f)); 
    ##k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop


# % -------------------------------------------------------------------------
# %                              aLIGO spectra
# % -------------------------------------------------------------------------

def aligospectrum(IFO,f):
# % ALIGOSPECTRUM - Design noise curve for different configurations of aLIGO.
# %
# % Arguments are the same as for the parent function.
# %
# % The aLIGO spectra are taken from LIGO Document T0900288-v3, 
# % https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=2974

    source = 'T0900288-v3'
# % ---- Recognize detector and compute noise spectrum.
    if IFO.upper()=='ALIGO-NOSRM':
        col = 1
    elif IFO.upper()=='ALIGO-ZDLP':
        col = 3
    elif IFO.upper()=='ALIGO-ZDHP':
        #% col = 4;
        #% ZDHP baseline has been updated in
        #% https://dcc.ligo.org/LIGO-T1800044 - v4 with slightly lower
        #% range.
        source = 'T1800044-v4'
        col = 1       
    elif IFO.upper()=='ALIGO-BNS':
        col = 5
    elif IFO.upper()=='ALIGO-BBH':
        col = 6
    elif IFO.upper()=='ALIGO-HF':
        col = 7
    else:
        raise ValueError('Detector '+IFO+' not recognized.')


# % ---- Extract pre-sampled noise spectrum from data.  First column is
# %      frequency, other columns map to configurations as above.
    if source=='T0900288-v3':
        data = np.array([[9.0, 1.74133e-21, 1.73571e-21, 1.73707e-21, 1.7391e-21, 1.73646e-21, 1.74313e-21], [9.02047, 2.17689e-21, 2.17244e-21, 2.17352e-21, 2.17512e-21, 2.17303e-21, 2.17831e-21], [9.04099, 2.92623e-21, 2.92296e-21, 2.92375e-21, 2.92493e-21, 2.92339e-21, 2.92728e-21], [9.06155, 4.51505e-21, 4.51295e-21, 4.51345e-21, 4.51422e-21, 4.51322e-21, 4.51572e-21], [9.08216, 1.01283e-20, 1.01274e-20, 1.01276e-20, 1.0128e-20, 1.01275e-20, 1.01286e-20], [9.10282, 3.79178e-20, 3.79175e-20, 3.79176e-20, 3.79177e-20, 3.79176e-20, 3.79178e-20], [9.12353, 6.50675e-21, 6.50533e-21, 6.50567e-21, 6.50619e-21, 6.50552e-21, 6.5072e-21], [9.14428, 3.53294e-21, 3.53035e-21, 3.53097e-21, 3.53191e-21, 3.53069e-21, 3.53375e-21], [9.16508, 2.41373e-21, 2.40997e-21, 2.41088e-21, 2.41224e-21, 2.41047e-21, 2.41491e-21], [9.18593, 1.82732e-21, 1.8224e-21, 1.82359e-21, 1.82537e-21, 1.82305e-21, 1.82886e-21], [9.20682, 1.46687e-21, 1.4608e-21, 1.46227e-21, 1.46447e-21, 1.4616e-21, 1.46876e-21], [9.22776, 1.2232e-21, 1.21597e-21, 1.21772e-21, 1.22034e-21, 1.21692e-21, 1.22544e-21], [9.24875, 1.0477e-21, 1.03933e-21, 1.04136e-21, 1.04439e-21, 1.04043e-21, 1.05028e-21], [9.26979, 9.15454e-22, 9.05952e-22, 9.08259e-22, 9.11697e-22, 9.072e-22, 9.18367e-22], [9.29087, 8.12362e-22, 8.01738e-22, 8.0432e-22, 8.08165e-22, 8.03134e-22, 8.15602e-22], [9.312, 7.2985e-22, 7.18117e-22, 7.20972e-22, 7.25218e-22, 7.19658e-22, 7.33409e-22], [9.33319, 6.62403e-22, 6.49573e-22, 6.527e-22, 6.57344e-22, 6.51259e-22, 6.66273e-22], [9.35441, 6.06313e-22, 5.924e-22, 5.95796e-22, 6.00833e-22, 5.9423e-22, 6.10485e-22], [9.37569, 5.58993e-22, 5.44013e-22, 5.47675e-22, 5.53101e-22, 5.45984e-22, 5.63459e-22], [9.39702, 5.18587e-22, 5.02555e-22, 5.06481e-22, 5.12289e-22, 5.04666e-22, 5.23337e-22], [9.41839, 4.83723e-22, 4.66657e-22, 4.70844e-22, 4.77028e-22, 4.68907e-22, 4.88748e-22], [9.43981, 4.53371e-22, 4.35288e-22, 4.39732e-22, 4.46287e-22, 4.37674e-22, 4.5866e-22], [9.46129, 4.26738e-22, 4.07657e-22, 4.12356e-22, 4.19275e-22, 4.10177e-22, 4.32282e-22], [9.48281, 4.03206e-22, 3.83146e-22, 3.88096e-22, 3.95372e-22, 3.85799e-22, 4.08994e-22], [9.50438, 3.82285e-22, 3.61266e-22, 3.66464e-22, 3.74091e-22, 3.64049e-22, 3.88307e-22], [9.52599, 3.63582e-22, 3.41625e-22, 3.47068e-22, 3.55037e-22, 3.44537e-22, 3.69828e-22], [9.54766, 3.46779e-22, 3.23905e-22, 3.29589e-22, 3.37894e-22, 3.26943e-22, 3.53238e-22], [9.56938, 3.31614e-22, 3.07845e-22, 3.13765e-22, 3.22399e-22, 3.11007e-22, 3.38277e-22], [9.59114, 3.17871e-22, 2.93229e-22, 2.99382e-22, 3.08336e-22, 2.96512e-22, 3.24727e-22], [9.61296, 3.0537e-22, 2.79876e-22, 2.86259e-22, 2.95525e-22, 2.83279e-22, 3.12408e-22], [9.63483, 2.93958e-22, 2.67636e-22, 2.74243e-22, 2.83813e-22, 2.71156e-22, 3.01169e-22], [9.65674, 2.83506e-22, 2.56379e-22, 2.63206e-22, 2.73073e-22, 2.60014e-22, 2.9088e-22], [9.67871, 2.73905e-22, 2.45996e-22, 2.5304e-22, 2.63195e-22, 2.49743e-22, 2.81432e-22], [9.70072, 2.65061e-22, 2.36393e-22, 2.43648e-22, 2.54083e-22, 2.40249e-22, 2.72732e-22], [9.72279, 2.56893e-22, 2.27488e-22, 2.34951e-22, 2.45657e-22, 2.31451e-22, 2.64698e-22], [9.7449, 2.49329e-22, 2.19212e-22, 2.26877e-22, 2.37846e-22, 2.23279e-22, 2.57259e-22], [9.76707, 2.42309e-22, 2.11502e-22, 2.19365e-22, 2.3059e-22, 2.15671e-22, 2.50356e-22], [9.78928, 2.35779e-22, 2.04305e-22, 2.12362e-22, 2.23833e-22, 2.08574e-22, 2.43934e-22], [9.81155, 2.29692e-22, 1.97574e-22, 2.05819e-22, 2.17529e-22, 2.01939e-22, 2.37946e-22], [9.83387, 2.24007e-22, 1.91266e-22, 1.99696e-22, 2.11636e-22, 1.95726e-22, 2.32352e-22], [9.85623, 2.18686e-22, 1.85345e-22, 1.93955e-22, 2.06117e-22, 1.89897e-22, 2.27114e-22], [9.87865, 2.13696e-22, 1.79779e-22, 1.88563e-22, 2.00939e-22, 1.84419e-22, 2.22201e-22], [9.90112, 2.0901e-22, 1.74536e-22, 1.83491e-22, 1.96073e-22, 1.79263e-22, 2.17583e-22], [9.92364, 2.046e-22, 1.69592e-22, 1.78712e-22, 1.91493e-22, 1.74402e-22, 2.13235e-22], [9.94621, 2.00444e-22, 1.64922e-22, 1.74204e-22, 1.87175e-22, 1.69814e-22, 2.09134e-22], [9.96884, 1.96521e-22, 1.60506e-22, 1.69944e-22, 1.83097e-22, 1.65476e-22, 2.0526e-22], [9.99151, 1.92812e-22, 1.56324e-22, 1.65913e-22, 1.79243e-22, 1.6137e-22, 2.01593e-22], [10.0142, 1.89299e-22, 1.52358e-22, 1.62095e-22, 1.75593e-22, 1.57478e-22, 1.98118e-22], [10.037, 1.85969e-22, 1.48594e-22, 1.58473e-22, 1.72132e-22, 1.53784e-22, 1.94818e-22], [10.0598, 1.82806e-22, 1.45016e-22, 1.55034e-22, 1.68846e-22, 1.50275e-22, 1.91681e-22], [10.0827, 1.79799e-22, 1.41612e-22, 1.51764e-22, 1.65724e-22, 1.46937e-22, 1.88694e-22], [10.1057, 1.76935e-22, 1.38369e-22, 1.48651e-22, 1.62751e-22, 1.43759e-22, 1.85846e-22], [10.1286, 1.74205e-22, 1.35278e-22, 1.45685e-22, 1.59919e-22, 1.40729e-22, 1.83127e-22], [10.1517, 1.71599e-22, 1.32328e-22, 1.42856e-22, 1.57218e-22, 1.37837e-22, 1.80527e-22], [10.1748, 1.69108e-22, 1.29509e-22, 1.40154e-22, 1.54637e-22, 1.35075e-22, 1.78039e-22], [10.1979, 1.66724e-22, 1.26814e-22, 1.37572e-22, 1.5217e-22, 1.32434e-22, 1.75653e-22], [10.2211, 1.6444e-22, 1.24234e-22, 1.35101e-22, 1.49809e-22, 1.29907e-22, 1.73364e-22], [10.2444, 1.6225e-22, 1.21763e-22, 1.32735e-22, 1.47547e-22, 1.27486e-22, 1.71165e-22], [10.2677, 1.60147e-22, 1.19394e-22, 1.30467e-22, 1.45377e-22, 1.25165e-22, 1.69049e-22], [10.291, 1.58125e-22, 1.17121e-22, 1.28291e-22, 1.43294e-22, 1.22937e-22, 1.67011e-22], [10.3144, 1.5618e-22, 1.14937e-22, 1.26201e-22, 1.41292e-22, 1.20797e-22, 1.65047e-22], [10.3379, 1.54306e-22, 1.12839e-22, 1.24193e-22, 1.39366e-22, 1.18741e-22, 1.63151e-22], [10.3614, 1.525e-22, 1.10821e-22, 1.22261e-22, 1.37512e-22, 1.16762e-22, 1.6132e-22], [10.385, 1.50756e-22, 1.08878e-22, 1.20401e-22, 1.35725e-22, 1.14857e-22, 1.59549e-22], [10.4086, 1.49072e-22, 1.07007e-22, 1.18609e-22, 1.34002e-22, 1.13021e-22, 1.57834e-22], [10.4323, 1.47443e-22, 1.05203e-22, 1.16882e-22, 1.32338e-22, 1.11251e-22, 1.56173e-22], [10.456, 1.45867e-22, 1.03462e-22, 1.15215e-22, 1.30731e-22, 1.09543e-22, 1.54563e-22], [10.4798, 1.44341e-22, 1.01783e-22, 1.13605e-22, 1.29177e-22, 1.07894e-22, 1.52999e-22], [10.5036, 1.42861e-22, 1.0016e-22, 1.12049e-22, 1.27673e-22, 1.063e-22, 1.51481e-22], [10.5275, 1.41425e-22, 9.85919e-23, 1.10544e-22, 1.26216e-22, 1.04759e-22, 1.50004e-22], [10.5515, 1.40031e-22, 9.7075e-23, 1.09088e-22, 1.24805e-22, 1.03267e-22, 1.48567e-22], [10.5755, 1.38676e-22, 9.56071e-23, 1.07679e-22, 1.23436e-22, 1.01823e-22, 1.47168e-22], [10.5995, 1.37359e-22, 9.41855e-23, 1.06312e-22, 1.22107e-22, 1.00424e-22, 1.45805e-22], [10.6236, 1.36077e-22, 9.28082e-23, 1.04988e-22, 1.20816e-22, 9.90673e-23, 1.44476e-22], [10.6478, 1.34829e-22, 9.14728e-23, 1.03702e-22, 1.19561e-22, 9.77512e-23, 1.43178e-22], [10.672, 1.33612e-22, 9.01774e-23, 1.02454e-22, 1.18341e-22, 9.64736e-23, 1.41911e-22], [10.6963, 1.32426e-22, 8.89201e-23, 1.01242e-22, 1.17154e-22, 9.52328e-23, 1.40673e-22], [10.7206, 1.31268e-22, 8.76992e-23, 1.00064e-22, 1.15997e-22, 9.4027e-23, 1.39463e-22], [10.745, 1.30137e-22, 8.65129e-23, 9.89177e-23, 1.1487e-22, 9.28545e-23, 1.38279e-22], [10.7694, 1.29033e-22, 8.53596e-23, 9.78025e-23, 1.13771e-22, 9.17138e-23, 1.37119e-22], [10.7939, 1.27954e-22, 8.4238e-23, 9.67166e-23, 1.127e-22, 9.06034e-23, 1.35984e-22], [10.8185, 1.26898e-22, 8.31465e-23, 9.56589e-23, 1.11653e-22, 8.95221e-23, 1.34871e-22], [10.8431, 1.25864e-22, 8.20839e-23, 9.46279e-23, 1.10631e-22, 8.84684e-23, 1.3378e-22], [10.8677, 1.24852e-22, 8.10489e-23, 9.36225e-23, 1.09633e-22, 8.74412e-23, 1.32709e-22], [10.8925, 1.23861e-22, 8.00403e-23, 9.26416e-23, 1.08657e-22, 8.64393e-23, 1.31659e-22], [10.9172, 1.2289e-22, 7.9057e-23, 9.16842e-23, 1.07702e-22, 8.54617e-23, 1.30627e-22], [10.9421, 1.21937e-22, 7.80981e-23, 9.07492e-23, 1.06767e-22, 8.45074e-23, 1.29613e-22], [10.967, 1.21003e-22, 7.71624e-23, 8.98357e-23, 1.05852e-22, 8.35753e-23, 1.28617e-22], [10.9919, 1.20085e-22, 7.6249e-23, 8.89428e-23, 1.04956e-22, 8.26646e-23, 1.27638e-22], [11.0169, 1.19185e-22, 7.53571e-23, 8.80696e-23, 1.04078e-22, 8.17743e-23, 1.26674e-22], [11.042, 1.183e-22, 7.44858e-23, 8.72155e-23, 1.03217e-22, 8.09038e-23, 1.25726e-22], [11.0671, 1.1743e-22, 7.36342e-23, 8.63795e-23, 1.02373e-22, 8.00521e-23, 1.24793e-22], [11.0923, 1.16576e-22, 7.28018e-23, 8.55611e-23, 1.01544e-22, 7.92187e-23, 1.23874e-22], [11.1175, 1.15735e-22, 7.19876e-23, 8.47596e-23, 1.00731e-22, 7.84027e-23, 1.22969e-22], [11.1428, 1.14908e-22, 7.11911e-23, 8.39742e-23, 9.99324e-23, 7.76036e-23, 1.22077e-22], [11.1681, 1.14094e-22, 7.04116e-23, 8.32044e-23, 9.9148e-23, 7.68206e-23, 1.21197e-22], [11.1935, 1.13293e-22, 6.96484e-23, 8.24496e-23, 9.83773e-23, 7.60533e-23, 1.2033e-22], [11.219, 1.12504e-22, 6.8901e-23, 8.17093e-23, 9.76197e-23, 7.5301e-23, 1.19475e-22], [11.2445, 1.11727e-22, 6.81689e-23, 8.0983e-23, 9.68749e-23, 7.45632e-23, 1.18632e-22], [11.2701, 1.10961e-22, 6.74515e-23, 8.02701e-23, 9.61422e-23, 7.38394e-23, 1.17799e-22], [11.2957, 1.10206e-22, 6.67483e-23, 7.95703e-23, 9.54214e-23, 7.31291e-23, 1.16977e-22], [11.3214, 1.09462e-22, 6.60587e-23, 7.88829e-23, 9.47119e-23, 7.24319e-23, 1.16166e-22], [11.3472, 1.08728e-22, 6.53825e-23, 7.82077e-23, 9.40135e-23, 7.17473e-23, 1.15365e-22], [11.373, 1.08004e-22, 6.47191e-23, 7.75443e-23, 9.33258e-23, 7.1075e-23, 1.14573e-22], [11.3988, 1.0729e-22, 6.40681e-23, 7.68922e-23, 9.26484e-23, 7.04144e-23, 1.13791e-22], [11.4248, 1.06585e-22, 6.34291e-23, 7.6251e-23, 9.1981e-23, 6.97653e-23, 1.13018e-22], [11.4507, 1.05889e-22, 6.28017e-23, 7.56205e-23, 9.13233e-23, 6.91272e-23, 1.12254e-22], [11.4768, 1.05202e-22, 6.21857e-23, 7.50003e-23, 9.0675e-23, 6.84998e-23, 1.11499e-22], [11.5029, 1.04523e-22, 6.15805e-23, 7.43901e-23, 9.00358e-23, 6.78829e-23, 1.10752e-22], [11.5291, 1.03853e-22, 6.0986e-23, 7.37895e-23, 8.94054e-23, 6.7276e-23, 1.10014e-22], [11.5553, 1.03191e-22, 6.04018e-23, 7.31983e-23, 8.87836e-23, 6.6679e-23, 1.09283e-22], [11.5816, 1.02536e-22, 5.98275e-23, 7.26163e-23, 8.81701e-23, 6.60914e-23, 1.0856e-22], [11.6079, 1.0189e-22, 5.9263e-23, 7.20431e-23, 8.75648e-23, 6.5513e-23, 1.07845e-22], [11.6343, 1.0125e-22, 5.87079e-23, 7.14785e-23, 8.69673e-23, 6.49436e-23, 1.07137e-22], [11.6608, 1.00618e-22, 5.8162e-23, 7.09223e-23, 8.63774e-23, 6.43829e-23, 1.06436e-22], [11.6873, 9.99932e-23, 5.76249e-23, 7.03742e-23, 8.5795e-23, 6.38307e-23, 1.05742e-22], [11.7139, 9.93749e-23, 5.70966e-23, 6.98339e-23, 8.52199e-23, 6.32867e-23, 1.05056e-22], [11.7405, 9.87634e-23, 5.65767e-23, 6.93014e-23, 8.46519e-23, 6.27508e-23, 1.04375e-22], [11.7672, 9.81584e-23, 5.60649e-23, 6.87764e-23, 8.40907e-23, 6.22226e-23, 1.03702e-22], [11.794, 9.75598e-23, 5.55612e-23, 6.82587e-23, 8.35363e-23, 6.17021e-23, 1.03035e-22], [11.8208, 9.69674e-23, 5.50653e-23, 6.77481e-23, 8.29884e-23, 6.11889e-23, 1.02374e-22], [11.8477, 9.63811e-23, 5.4577e-23, 6.72444e-23, 8.24468e-23, 6.0683e-23, 1.01719e-22], [11.8747, 9.58008e-23, 5.40961e-23, 6.67475e-23, 8.19116e-23, 6.01842e-23, 1.0107e-22], [11.9017, 9.52262e-23, 5.36225e-23, 6.62571e-23, 8.13824e-23, 5.96922e-23, 1.00427e-22], [11.9287, 9.46574e-23, 5.31559e-23, 6.57732e-23, 8.08592e-23, 5.92069e-23, 9.97901e-23], [11.9559, 9.40941e-23, 5.26961e-23, 6.52956e-23, 8.03417e-23, 5.87281e-23, 9.91586e-23], [11.9831, 9.35363e-23, 5.22431e-23, 6.48241e-23, 7.983e-23, 5.82558e-23, 9.85326e-23], [12.0103, 9.29838e-23, 5.17967e-23, 6.43586e-23, 7.93238e-23, 5.77897e-23, 9.7912e-23], [12.0376, 9.24365e-23, 5.13566e-23, 6.38989e-23, 7.88231e-23, 5.73296e-23, 9.72967e-23], [12.065, 9.18943e-23, 5.09229e-23, 6.3445e-23, 7.83277e-23, 5.68756e-23, 9.66867e-23], [12.0925, 9.13572e-23, 5.04952e-23, 6.29966e-23, 7.78374e-23, 5.64274e-23, 9.60818e-23], [12.12, 9.08249e-23, 5.00736e-23, 6.25537e-23, 7.73523e-23, 5.59849e-23, 9.54818e-23], [12.1475, 9.02975e-23, 4.96578e-23, 6.21162e-23, 7.68722e-23, 5.55479e-23, 9.48869e-23], [12.1752, 8.97749e-23, 4.92477e-23, 6.16839e-23, 7.6397e-23, 5.51164e-23, 9.42968e-23], [12.2029, 8.92568e-23, 4.88433e-23, 6.12567e-23, 7.59265e-23, 5.46903e-23, 9.37114e-23], [12.2306, 8.87434e-23, 4.84444e-23, 6.08346e-23, 7.54608e-23, 5.42694e-23, 9.31307e-23], [12.2584, 8.82344e-23, 4.80508e-23, 6.04173e-23, 7.49997e-23, 5.38536e-23, 9.25547e-23], [12.2863, 8.77298e-23, 4.76625e-23, 6.00049e-23, 7.45431e-23, 5.34429e-23, 9.19832e-23], [12.3143, 8.72296e-23, 4.72793e-23, 5.95972e-23, 7.40909e-23, 5.3037e-23, 9.14161e-23], [12.3423, 8.67336e-23, 4.69012e-23, 5.91941e-23, 7.36431e-23, 5.2636e-23, 9.08535e-23], [12.3704, 8.62417e-23, 4.65281e-23, 5.87956e-23, 7.31996e-23, 5.22397e-23, 9.02951e-23], [12.3985, 8.5754e-23, 4.61598e-23, 5.84015e-23, 7.27602e-23, 5.1848e-23, 8.9741e-23], [12.4267, 8.52703e-23, 4.57963e-23, 5.80117e-23, 7.2325e-23, 5.14609e-23, 8.91911e-23], [12.455, 8.47906e-23, 4.54374e-23, 5.76263e-23, 7.18938e-23, 5.10782e-23, 8.86454e-23], [12.4833, 8.43147e-23, 4.50832e-23, 5.7245e-23, 7.14666e-23, 5.06999e-23, 8.81037e-23], [12.5117, 8.38428e-23, 4.47334e-23, 5.68678e-23, 7.10433e-23, 5.03259e-23, 8.7566e-23], [12.5401, 8.33746e-23, 4.4388e-23, 5.64947e-23, 7.06239e-23, 4.99562e-23, 8.70322e-23], [12.5687, 8.29101e-23, 4.4047e-23, 5.61256e-23, 7.02082e-23, 4.95905e-23, 8.65024e-23], [12.5973, 8.24493e-23, 4.37102e-23, 5.57603e-23, 6.97962e-23, 4.92289e-23, 8.59764e-23], [12.6259, 8.19921e-23, 4.33776e-23, 5.53989e-23, 6.93879e-23, 4.88713e-23, 8.54542e-23], [12.6546, 8.15385e-23, 4.3049e-23, 5.50413e-23, 6.89832e-23, 4.85176e-23, 8.49357e-23], [12.6834, 8.10884e-23, 4.27245e-23, 5.46873e-23, 6.8582e-23, 4.81678e-23, 8.44209e-23], [12.7123, 8.06417e-23, 4.24039e-23, 5.4337e-23, 6.81843e-23, 4.78217e-23, 8.39097e-23], [12.7412, 8.01984e-23, 4.20873e-23, 5.39902e-23, 6.779e-23, 4.74794e-23, 8.34022e-23], [12.7702, 7.97585e-23, 4.17744e-23, 5.36469e-23, 6.73991e-23, 4.71407e-23, 8.28981e-23], [12.7992, 7.93219e-23, 4.14652e-23, 5.33071e-23, 6.70115e-23, 4.68055e-23, 8.23976e-23], [12.8283, 7.88886e-23, 4.11598e-23, 5.29707e-23, 6.66271e-23, 4.6474e-23, 8.19005e-23], [12.8575, 7.84585e-23, 4.08579e-23, 5.26376e-23, 6.6246e-23, 4.61459e-23, 8.14068e-23], [12.8867, 7.80315e-23, 4.05597e-23, 5.23078e-23, 6.5868e-23, 4.58212e-23, 8.09165e-23], [12.916, 7.76077e-23, 4.02649e-23, 5.19812e-23, 6.54932e-23, 4.54998e-23, 8.04295e-23], [12.9454, 7.7187e-23, 3.99735e-23, 5.16577e-23, 6.51214e-23, 4.51818e-23, 7.99458e-23], [12.9749, 7.67693e-23, 3.96855e-23, 5.13374e-23, 6.47527e-23, 4.4867e-23, 7.94654e-23], [13.0044, 7.63546e-23, 3.94009e-23, 5.10202e-23, 6.43869e-23, 4.45554e-23, 7.89881e-23], [13.034, 7.59429e-23, 3.91195e-23, 5.0706e-23, 6.40241e-23, 4.4247e-23, 7.8514e-23], [13.0636, 7.55341e-23, 3.88413e-23, 5.03948e-23, 6.36642e-23, 4.39416e-23, 7.80431e-23], [13.0933, 7.51283e-23, 3.85663e-23, 5.00865e-23, 6.33072e-23, 4.36393e-23, 7.75752e-23], [13.1231, 7.47252e-23, 3.82944e-23, 4.9781e-23, 6.2953e-23, 4.334e-23, 7.71105e-23], [13.153, 7.4325e-23, 3.80255e-23, 4.94785e-23, 6.26015e-23, 4.30436e-23, 7.66487e-23], [13.1829, 7.39276e-23, 3.77597e-23, 4.91787e-23, 6.22528e-23, 4.27502e-23, 7.61899e-23], [13.2129, 7.35329e-23, 3.74968e-23, 4.88816e-23, 6.19068e-23, 4.24596e-23, 7.57341e-23], [13.2429, 7.3141e-23, 3.72369e-23, 4.85873e-23, 6.15635e-23, 4.21719e-23, 7.52812e-23], [13.273, 7.27517e-23, 3.69798e-23, 4.82956e-23, 6.12228e-23, 4.18869e-23, 7.48312e-23], [13.3032, 7.23651e-23, 3.67255e-23, 4.80066e-23, 6.08847e-23, 4.16046e-23, 7.43841e-23], [13.3335, 7.19811e-23, 3.6474e-23, 4.77202e-23, 6.05491e-23, 4.1325e-23, 7.39398e-23], [13.3638, 7.15997e-23, 3.62253e-23, 4.74363e-23, 6.02161e-23, 4.10481e-23, 7.34983e-23], [13.3942, 7.12209e-23, 3.59793e-23, 4.7155e-23, 5.98856e-23, 4.07739e-23, 7.30596e-23], [13.4247, 7.08446e-23, 3.57359e-23, 4.68761e-23, 5.95576e-23, 4.05021e-23, 7.26237e-23], [13.4552, 7.04708e-23, 3.54951e-23, 4.65997e-23, 5.9232e-23, 4.0233e-23, 7.21904e-23], [13.4858, 7.00994e-23, 3.52569e-23, 4.63256e-23, 5.89088e-23, 3.99663e-23, 7.17599e-23], [13.5165, 6.97306e-23, 3.50212e-23, 4.6054e-23, 5.85879e-23, 3.97021e-23, 7.1332e-23], [13.5472, 6.93641e-23, 3.47881e-23, 4.57847e-23, 5.82694e-23, 3.94403e-23, 7.09067e-23], [13.578, 6.9e-23, 3.45574e-23, 4.55177e-23, 5.79533e-23, 3.91809e-23, 7.0484e-23], [13.6089, 6.86383e-23, 3.43291e-23, 4.5253e-23, 5.76394e-23, 3.89239e-23, 7.0064e-23], [13.6399, 6.82789e-23, 3.41032e-23, 4.49905e-23, 5.73277e-23, 3.86693e-23, 6.96465e-23], [13.6709, 6.79219e-23, 3.38797e-23, 4.47303e-23, 5.70183e-23, 3.84169e-23, 6.92315e-23], [13.702, 6.75671e-23, 3.36585e-23, 4.44722e-23, 5.67111e-23, 3.81668e-23, 6.8819e-23], [13.7332, 6.72146e-23, 3.34395e-23, 4.42163e-23, 5.64061e-23, 3.79189e-23, 6.8409e-23], [13.7644, 6.68644e-23, 3.32229e-23, 4.39625e-23, 5.61032e-23, 3.76732e-23, 6.80015e-23], [13.7957, 6.65163e-23, 3.30084e-23, 4.37108e-23, 5.58025e-23, 3.74297e-23, 6.75964e-23], [13.8271, 6.61705e-23, 3.27961e-23, 4.34612e-23, 5.55038e-23, 3.71883e-23, 6.71938e-23], [13.8585, 6.58268e-23, 3.2586e-23, 4.32136e-23, 5.52072e-23, 3.69491e-23, 6.67935e-23], [13.8901, 6.54853e-23, 3.2378e-23, 4.2968e-23, 5.49127e-23, 3.67119e-23, 6.63956e-23], [13.9217, 6.51459e-23, 3.21721e-23, 4.27245e-23, 5.46202e-23, 3.64768e-23, 6.6e-23], [13.9533, 6.48086e-23, 3.19683e-23, 4.24829e-23, 5.43297e-23, 3.62438e-23, 6.56068e-23], [13.9851, 6.44734e-23, 3.17665e-23, 4.22432e-23, 5.40412e-23, 3.60127e-23, 6.52159e-23], [14.0169, 6.41403e-23, 3.15667e-23, 4.20054e-23, 5.37547e-23, 3.57836e-23, 6.48273e-23], [14.0488, 6.38092e-23, 3.13689e-23, 4.17696e-23, 5.34701e-23, 3.55565e-23, 6.44409e-23], [14.0807, 6.34801e-23, 3.11731e-23, 4.15355e-23, 5.31874e-23, 3.53313e-23, 6.40568e-23], [14.1127, 6.31531e-23, 3.09791e-23, 4.13034e-23, 5.29066e-23, 3.5108e-23, 6.36749e-23], [14.1448, 6.2828e-23, 3.07871e-23, 4.1073e-23, 5.26277e-23, 3.48866e-23, 6.32952e-23], [14.177, 6.25049e-23, 3.0597e-23, 4.08445e-23, 5.23506e-23, 3.4667e-23, 6.29177e-23], [14.2093, 6.21837e-23, 3.04086e-23, 4.06177e-23, 5.20754e-23, 3.44493e-23, 6.25424e-23], [14.2416, 6.18645e-23, 3.02222e-23, 4.03927e-23, 5.1802e-23, 3.42334e-23, 6.21692e-23], [14.274, 6.15472e-23, 3.00375e-23, 4.01694e-23, 5.15304e-23, 3.40192e-23, 6.17982e-23], [14.3064, 6.12317e-23, 2.98546e-23, 3.99478e-23, 5.12606e-23, 3.38069e-23, 6.14292e-23], [14.339, 6.09182e-23, 2.96734e-23, 3.97278e-23, 5.09925e-23, 3.35962e-23, 6.10624e-23], [14.3716, 6.06065e-23, 2.94939e-23, 3.95096e-23, 5.07262e-23, 3.33873e-23, 6.06977e-23], [14.4043, 6.02966e-23, 2.93162e-23, 3.9293e-23, 5.04616e-23, 3.31801e-23, 6.0335e-23], [14.4371, 5.99886e-23, 2.91401e-23, 3.9078e-23, 5.01987e-23, 3.29746e-23, 5.99743e-23], [14.4699, 5.96824e-23, 2.89657e-23, 3.88647e-23, 4.99375e-23, 3.27707e-23, 5.96157e-23], [14.5028, 5.9378e-23, 2.8793e-23, 3.86529e-23, 4.96779e-23, 3.25685e-23, 5.92591e-23], [14.5358, 5.90753e-23, 2.86218e-23, 3.84427e-23, 4.942e-23, 3.23679e-23, 5.89045e-23], [14.5689, 5.87744e-23, 2.84523e-23, 3.8234e-23, 4.91638e-23, 3.21688e-23, 5.85519e-23], [14.602, 5.84753e-23, 2.82843e-23, 3.80269e-23, 4.89091e-23, 3.19714e-23, 5.82013e-23], [14.6352, 5.81779e-23, 2.81179e-23, 3.78213e-23, 4.86561e-23, 3.17755e-23, 5.78526e-23], [14.6685, 5.78822e-23, 2.7953e-23, 3.76172e-23, 4.84047e-23, 3.15812e-23, 5.75058e-23], [14.7019, 5.75882e-23, 2.77896e-23, 3.74146e-23, 4.81548e-23, 3.13884e-23, 5.7161e-23], [14.7353, 5.72959e-23, 2.76277e-23, 3.72134e-23, 4.79065e-23, 3.11971e-23, 5.6818e-23], [14.7688, 5.70052e-23, 2.74673e-23, 3.70137e-23, 4.76598e-23, 3.10073e-23, 5.6477e-23], [14.8024, 5.67163e-23, 2.73084e-23, 3.68155e-23, 4.74145e-23, 3.08189e-23, 5.61378e-23], [14.8361, 5.64289e-23, 2.71508e-23, 3.66186e-23, 4.71708e-23, 3.06321e-23, 5.58005e-23], [14.8698, 5.61432e-23, 2.69948e-23, 3.64231e-23, 4.69286e-23, 3.04466e-23, 5.5465e-23], [14.9036, 5.58591e-23, 2.68401e-23, 3.62291e-23, 4.66879e-23, 3.02626e-23, 5.51314e-23], [14.9375, 5.55766e-23, 2.66868e-23, 3.60363e-23, 4.64486e-23, 3.008e-23, 5.47996e-23], [14.9715, 5.52957e-23, 2.65348e-23, 3.5845e-23, 4.62108e-23, 2.98988e-23, 5.44696e-23], [15.0056, 5.50164e-23, 2.63843e-23, 3.5655e-23, 4.59745e-23, 2.9719e-23, 5.41414e-23], [15.0397, 5.47387e-23, 2.6235e-23, 3.54663e-23, 4.57395e-23, 2.95405e-23, 5.3815e-23], [15.0739, 5.44625e-23, 2.60871e-23, 3.52789e-23, 4.5506e-23, 2.93634e-23, 5.34904e-23], [15.1082, 5.41878e-23, 2.59405e-23, 3.50929e-23, 4.52739e-23, 2.91876e-23, 5.31675e-23], [15.1426, 5.39147e-23, 2.57951e-23, 3.49081e-23, 4.50432e-23, 2.90131e-23, 5.28463e-23], [15.177, 5.3643e-23, 2.56511e-23, 3.47245e-23, 4.48139e-23, 2.88399e-23, 5.25269e-23], [15.2115, 5.33729e-23, 2.55083e-23, 3.45423e-23, 4.4586e-23, 2.8668e-23, 5.22092e-23], [15.2461, 5.31043e-23, 2.53667e-23, 3.43613e-23, 4.43594e-23, 2.84974e-23, 5.18932e-23], [15.2808, 5.28371e-23, 2.52264e-23, 3.41815e-23, 4.41341e-23, 2.83281e-23, 5.15789e-23], [15.3156, 5.25715e-23, 2.50873e-23, 3.40029e-23, 4.39102e-23, 2.816e-23, 5.12663e-23], [15.3504, 5.23072e-23, 2.49493e-23, 3.38256e-23, 4.36876e-23, 2.79931e-23, 5.09553e-23], [15.3853, 5.20445e-23, 2.48126e-23, 3.36494e-23, 4.34663e-23, 2.78275e-23, 5.06461e-23], [15.4203, 5.17831e-23, 2.46771e-23, 3.34744e-23, 4.32463e-23, 2.76631e-23, 5.03384e-23], [15.4554, 5.15232e-23, 2.45426e-23, 3.33006e-23, 4.30276e-23, 2.74999e-23, 5.00324e-23], [15.4905, 5.12647e-23, 2.44094e-23, 3.31279e-23, 4.28102e-23, 2.73378e-23, 4.97281e-23], [15.5258, 5.10076e-23, 2.42773e-23, 3.29564e-23, 4.2594e-23, 2.7177e-23, 4.94253e-23], [15.5611, 5.07519e-23, 2.41463e-23, 3.2786e-23, 4.23791e-23, 2.70173e-23, 4.91241e-23], [15.5965, 5.04976e-23, 2.40164e-23, 3.26168e-23, 4.21655e-23, 2.68587e-23, 4.88246e-23], [15.632, 5.02446e-23, 2.38876e-23, 3.24487e-23, 4.19531e-23, 2.67013e-23, 4.85266e-23], [15.6675, 4.9993e-23, 2.37598e-23, 3.22816e-23, 4.17419e-23, 2.6545e-23, 4.82302e-23], [15.7032, 4.97428e-23, 2.36332e-23, 3.21157e-23, 4.15319e-23, 2.63899e-23, 4.79354e-23], [15.7389, 4.94939e-23, 2.35076e-23, 3.19508e-23, 4.13231e-23, 2.62358e-23, 4.76421e-23], [15.7747, 4.92463e-23, 2.3383e-23, 3.17871e-23, 4.11155e-23, 2.60829e-23, 4.73503e-23], [15.8105, 4.90001e-23, 2.32595e-23, 3.16243e-23, 4.09092e-23, 2.5931e-23, 4.70601e-23], [15.8465, 4.87552e-23, 2.3137e-23, 3.14627e-23, 4.07039e-23, 2.57802e-23, 4.67714e-23], [15.8826, 4.85116e-23, 2.30155e-23, 3.1302e-23, 4.04999e-23, 2.56305e-23, 4.64843e-23], [15.9187, 4.82693e-23, 2.28951e-23, 3.11424e-23, 4.0297e-23, 2.54818e-23, 4.61986e-23], [15.9549, 4.80282e-23, 2.27756e-23, 3.09838e-23, 4.00952e-23, 2.53341e-23, 4.59144e-23], [15.9912, 4.77885e-23, 2.26571e-23, 3.08263e-23, 3.98946e-23, 2.51875e-23, 4.56317e-23], [16.0276, 4.755e-23, 2.25395e-23, 3.06697e-23, 3.96952e-23, 2.5042e-23, 4.53505e-23], [16.064, 4.73128e-23, 2.2423e-23, 3.05142e-23, 3.94968e-23, 2.48974e-23, 4.50707e-23], [16.1005, 4.70768e-23, 2.23073e-23, 3.03596e-23, 3.92996e-23, 2.47539e-23, 4.47924e-23], [16.1372, 4.68421e-23, 2.21927e-23, 3.0206e-23, 3.91034e-23, 2.46113e-23, 4.45156e-23], [16.1739, 4.66086e-23, 2.20789e-23, 3.00533e-23, 3.89084e-23, 2.44697e-23, 4.42402e-23], [16.2107, 4.63763e-23, 2.19661e-23, 2.99017e-23, 3.87145e-23, 2.43291e-23, 4.39662e-23], [16.2475, 4.61452e-23, 2.18541e-23, 2.97509e-23, 3.85216e-23, 2.41895e-23, 4.36936e-23], [16.2845, 4.59154e-23, 2.17431e-23, 2.96011e-23, 3.83298e-23, 2.40509e-23, 4.34225e-23], [16.3215, 4.56868e-23, 2.1633e-23, 2.94523e-23, 3.81391e-23, 2.39132e-23, 4.31528e-23], [16.3587, 4.54593e-23, 2.15237e-23, 2.93044e-23, 3.79494e-23, 2.37764e-23, 4.28844e-23], [16.3959, 4.52331e-23, 2.14154e-23, 2.91574e-23, 3.77608e-23, 2.36406e-23, 4.26175e-23], [16.4332, 4.5008e-23, 2.13079e-23, 2.90113e-23, 3.75732e-23, 2.35057e-23, 4.23519e-23], [16.4705, 4.47841e-23, 2.12012e-23, 2.88661e-23, 3.73866e-23, 2.33717e-23, 4.20877e-23], [16.508, 4.45613e-23, 2.10954e-23, 2.87218e-23, 3.72011e-23, 2.32387e-23, 4.18248e-23], [16.5455, 4.43397e-23, 2.09904e-23, 2.85784e-23, 3.70166e-23, 2.31065e-23, 4.15633e-23], [16.5832, 4.41193e-23, 2.08863e-23, 2.84358e-23, 3.68331e-23, 2.29753e-23, 4.13032e-23], [16.6209, 4.39e-23, 2.0783e-23, 2.82942e-23, 3.66506e-23, 2.28449e-23, 4.10444e-23], [16.6587, 4.36818e-23, 2.06805e-23, 2.81534e-23, 3.64691e-23, 2.27154e-23, 4.07869e-23], [16.6966, 4.34648e-23, 2.05788e-23, 2.80134e-23, 3.62886e-23, 2.25868e-23, 4.05308e-23], [16.7346, 4.32489e-23, 2.04779e-23, 2.78743e-23, 3.6109e-23, 2.24591e-23, 4.0276e-23], [16.7726, 4.3034e-23, 2.03778e-23, 2.77361e-23, 3.59305e-23, 2.23322e-23, 4.00224e-23], [16.8108, 4.28203e-23, 2.02785e-23, 2.75987e-23, 3.57529e-23, 2.22061e-23, 3.97702e-23], [16.849, 4.26077e-23, 2.018e-23, 2.74621e-23, 3.55763e-23, 2.2081e-23, 3.95193e-23], [16.8874, 4.23962e-23, 2.00822e-23, 2.73263e-23, 3.54006e-23, 2.19566e-23, 3.92697e-23], [16.9258, 4.21858e-23, 1.99852e-23, 2.71914e-23, 3.52259e-23, 2.18331e-23, 3.90213e-23], [16.9643, 4.19764e-23, 1.98889e-23, 2.70573e-23, 3.50521e-23, 2.17104e-23, 3.87742e-23], [17.0028, 4.17681e-23, 1.97934e-23, 2.69239e-23, 3.48793e-23, 2.15885e-23, 3.85284e-23], [17.0415, 4.15609e-23, 1.96987e-23, 2.67914e-23, 3.47073e-23, 2.14675e-23, 3.82839e-23], [17.0803, 4.13548e-23, 1.96046e-23, 2.66597e-23, 3.45364e-23, 2.13472e-23, 3.80406e-23], [17.1191, 4.11496e-23, 1.95113e-23, 2.65287e-23, 3.43663e-23, 2.12278e-23, 3.77985e-23], [17.1581, 4.09456e-23, 1.94187e-23, 2.63985e-23, 3.41971e-23, 2.11091e-23, 3.75577e-23], [17.1971, 4.07425e-23, 1.93269e-23, 2.62691e-23, 3.40288e-23, 2.09913e-23, 3.73181e-23], [17.2362, 4.05406e-23, 1.92357e-23, 2.61405e-23, 3.38615e-23, 2.08742e-23, 3.70797e-23], [17.2754, 4.03396e-23, 1.91452e-23, 2.60126e-23, 3.3695e-23, 2.07579e-23, 3.68426e-23], [17.3147, 4.01396e-23, 1.90555e-23, 2.58855e-23, 3.35294e-23, 2.06423e-23, 3.66066e-23], [17.3541, 3.99407e-23, 1.89664e-23, 2.57591e-23, 3.33647e-23, 2.05275e-23, 3.63719e-23], [17.3936, 3.97428e-23, 1.8878e-23, 2.56334e-23, 3.32009e-23, 2.04135e-23, 3.61384e-23], [17.4331, 3.95458e-23, 1.87902e-23, 2.55085e-23, 3.30379e-23, 2.03002e-23, 3.5906e-23], [17.4728, 3.93499e-23, 1.87032e-23, 2.53844e-23, 3.28758e-23, 2.01877e-23, 3.56749e-23], [17.5125, 3.9155e-23, 1.86168e-23, 2.52609e-23, 3.27146e-23, 2.00759e-23, 3.54449e-23], [17.5524, 3.8961e-23, 1.85311e-23, 2.51382e-23, 3.25542e-23, 1.99649e-23, 3.52161e-23], [17.5923, 3.8768e-23, 1.8446e-23, 2.50162e-23, 3.23947e-23, 1.98546e-23, 3.49885e-23], [17.6323, 3.8576e-23, 1.83615e-23, 2.48949e-23, 3.2236e-23, 1.9745e-23, 3.4762e-23], [17.6724, 3.8385e-23, 1.82777e-23, 2.47743e-23, 3.20781e-23, 1.96361e-23, 3.45367e-23], [17.7126, 3.81949e-23, 1.81946e-23, 2.46544e-23, 3.19211e-23, 1.9528e-23, 3.43125e-23], [17.7529, 3.80058e-23, 1.81121e-23, 2.45352e-23, 3.17648e-23, 1.94205e-23, 3.40895e-23], [17.7933, 3.78176e-23, 1.80301e-23, 2.44167e-23, 3.16095e-23, 1.93138e-23, 3.38676e-23], [17.8337, 3.76304e-23, 1.79489e-23, 2.42989e-23, 3.14549e-23, 1.92077e-23, 3.36469e-23], [17.8743, 3.74441e-23, 1.78682e-23, 2.41817e-23, 3.13011e-23, 1.91024e-23, 3.34273e-23], [17.915, 3.72588e-23, 1.77881e-23, 2.40653e-23, 3.11482e-23, 1.89977e-23, 3.32088e-23], [17.9557, 3.70743e-23, 1.77086e-23, 2.39495e-23, 3.0996e-23, 1.88937e-23, 3.29914e-23], [17.9966, 3.68908e-23, 1.76298e-23, 2.38343e-23, 3.08446e-23, 1.87904e-23, 3.27751e-23], [18.0375, 3.67083e-23, 1.75515e-23, 2.37198e-23, 3.0694e-23, 1.86878e-23, 3.25599e-23], [18.0785, 3.65266e-23, 1.74738e-23, 2.3606e-23, 3.05442e-23, 1.85859e-23, 3.23459e-23], [18.1196, 3.63458e-23, 1.73967e-23, 2.34928e-23, 3.03952e-23, 1.84846e-23, 3.21329e-23], [18.1609, 3.6166e-23, 1.73202e-23, 2.33803e-23, 3.0247e-23, 1.8384e-23, 3.1921e-23], [18.2022, 3.5987e-23, 1.72442e-23, 2.32684e-23, 3.00995e-23, 1.8284e-23, 3.17102e-23], [18.2436, 3.5809e-23, 1.71689e-23, 2.31572e-23, 2.99528e-23, 1.81847e-23, 3.15004e-23], [18.2851, 3.56318e-23, 1.7094e-23, 2.30465e-23, 2.98069e-23, 1.8086e-23, 3.12918e-23], [18.3267, 3.54555e-23, 1.70198e-23, 2.29366e-23, 2.96617e-23, 1.7988e-23, 3.10842e-23], [18.3683, 3.52801e-23, 1.69461e-23, 2.28272e-23, 2.95173e-23, 1.78906e-23, 3.08776e-23], [18.4101, 3.51055e-23, 1.68729e-23, 2.27184e-23, 2.93736e-23, 1.77939e-23, 3.06722e-23], [18.452, 3.49318e-23, 1.68003e-23, 2.26103e-23, 2.92306e-23, 1.76977e-23, 3.04677e-23], [18.494, 3.4759e-23, 1.67282e-23, 2.25028e-23, 2.90884e-23, 1.76022e-23, 3.02644e-23], [18.536, 3.45871e-23, 1.66567e-23, 2.23959e-23, 2.89469e-23, 1.75074e-23, 3.0062e-23], [18.5782, 3.4416e-23, 1.65857e-23, 2.22896e-23, 2.88062e-23, 1.74131e-23, 2.98607e-23], [18.6204, 3.42457e-23, 1.65152e-23, 2.21838e-23, 2.86662e-23, 1.73195e-23, 2.96605e-23], [18.6628, 3.40763e-23, 1.64453e-23, 2.20787e-23, 2.85269e-23, 1.72265e-23, 2.94612e-23], [18.7053, 3.39077e-23, 1.63759e-23, 2.19742e-23, 2.83883e-23, 1.71341e-23, 2.9263e-23], [18.7478, 3.374e-23, 1.6307e-23, 2.18703e-23, 2.82504e-23, 1.70423e-23, 2.90658e-23], [18.7904, 3.35731e-23, 1.62386e-23, 2.17669e-23, 2.81133e-23, 1.6951e-23, 2.88696e-23], [18.8332, 3.3407e-23, 1.61707e-23, 2.16641e-23, 2.79768e-23, 1.68604e-23, 2.86745e-23], [18.876, 3.32418e-23, 1.61033e-23, 2.15619e-23, 2.78411e-23, 1.67704e-23, 2.84803e-23], [18.919, 3.30774e-23, 1.60364e-23, 2.14603e-23, 2.7706e-23, 1.6681e-23, 2.82871e-23], [18.962, 3.29137e-23, 1.597e-23, 2.13592e-23, 2.75716e-23, 1.65921e-23, 2.80949e-23], [19.0051, 3.27509e-23, 1.59041e-23, 2.12587e-23, 2.74379e-23, 1.65039e-23, 2.79037e-23], [19.0483, 3.25889e-23, 1.58387e-23, 2.11588e-23, 2.7305e-23, 1.64162e-23, 2.77135e-23], [19.0917, 3.24277e-23, 1.57738e-23, 2.10594e-23, 2.71726e-23, 1.6329e-23, 2.75243e-23], [19.1351, 3.22673e-23, 1.57093e-23, 2.09605e-23, 2.7041e-23, 1.62425e-23, 2.73361e-23], [19.1786, 3.21077e-23, 1.56454e-23, 2.08623e-23, 2.691e-23, 1.61565e-23, 2.71488e-23], [19.2222, 3.19489e-23, 1.55819e-23, 2.07645e-23, 2.67797e-23, 1.60711e-23, 2.69625e-23], [19.266, 3.17908e-23, 1.55188e-23, 2.06673e-23, 2.66501e-23, 1.59862e-23, 2.67772e-23], [19.3098, 3.16336e-23, 1.54563e-23, 2.05707e-23, 2.65211e-23, 1.59019e-23, 2.65928e-23], [19.3537, 3.14771e-23, 1.53942e-23, 2.04745e-23, 2.63928e-23, 1.58182e-23, 2.64094e-23], [19.3977, 3.13213e-23, 1.53325e-23, 2.03789e-23, 2.62651e-23, 1.5735e-23, 2.62269e-23], [19.4419, 3.11664e-23, 1.52714e-23, 2.02839e-23, 2.61381e-23, 1.56524e-23, 2.60454e-23], [19.4861, 3.10122e-23, 1.52106e-23, 2.01893e-23, 2.60117e-23, 1.55703e-23, 2.58648e-23], [19.5304, 3.08588e-23, 1.51503e-23, 2.00953e-23, 2.5886e-23, 1.54887e-23, 2.56852e-23], [19.5748, 3.07061e-23, 1.50905e-23, 2.00018e-23, 2.57609e-23, 1.54077e-23, 2.55065e-23], [19.6193, 3.05542e-23, 1.50311e-23, 1.99089e-23, 2.56365e-23, 1.53272e-23, 2.53288e-23], [19.664, 3.0403e-23, 1.49722e-23, 1.98164e-23, 2.55126e-23, 1.52472e-23, 2.51519e-23], [19.7087, 3.02526e-23, 1.49136e-23, 1.97244e-23, 2.53894e-23, 1.51678e-23, 2.4976e-23], [19.7535, 3.01029e-23, 1.48556e-23, 1.9633e-23, 2.52669e-23, 1.50889e-23, 2.4801e-23], [19.7985, 2.99539e-23, 1.47979e-23, 1.9542e-23, 2.51449e-23, 1.50105e-23, 2.4627e-23], [19.8435, 2.98057e-23, 1.47407e-23, 1.94516e-23, 2.50236e-23, 1.49326e-23, 2.44538e-23], [19.8886, 2.96582e-23, 1.46839e-23, 1.93616e-23, 2.49029e-23, 1.48553e-23, 2.42816e-23], [19.9339, 2.95114e-23, 1.46275e-23, 1.92721e-23, 2.47828e-23, 1.47785e-23, 2.41102e-23], [19.9792, 2.93654e-23, 1.45715e-23, 1.91832e-23, 2.46633e-23, 1.47022e-23, 2.39398e-23], [20.0246, 2.922e-23, 1.45159e-23, 1.90947e-23, 2.45444e-23, 1.46263e-23, 2.37703e-23], [20.0702, 2.90754e-23, 1.44608e-23, 1.90067e-23, 2.44261e-23, 1.4551e-23, 2.36016e-23], [20.1158, 2.89315e-23, 1.44061e-23, 1.89192e-23, 2.43084e-23, 1.44762e-23, 2.34339e-23], [20.1616, 2.87883e-23, 1.43517e-23, 1.88321e-23, 2.41913e-23, 1.44019e-23, 2.32671e-23], [20.2075, 2.86458e-23, 1.42978e-23, 1.87456e-23, 2.40747e-23, 1.43281e-23, 2.31011e-23], [20.2534, 2.8504e-23, 1.42443e-23, 1.86595e-23, 2.39588e-23, 1.42548e-23, 2.2936e-23], [20.2995, 2.83629e-23, 1.41911e-23, 1.85738e-23, 2.38435e-23, 1.4182e-23, 2.27718e-23], [20.3457, 2.82225e-23, 1.41384e-23, 1.84887e-23, 2.37287e-23, 1.41097e-23, 2.26085e-23], [20.3919, 2.80828e-23, 1.4086e-23, 1.8404e-23, 2.36145e-23, 1.40378e-23, 2.2446e-23], [20.4383, 2.79437e-23, 1.4034e-23, 1.83198e-23, 2.35009e-23, 1.39665e-23, 2.22845e-23], [20.4848, 2.78054e-23, 1.39825e-23, 1.8236e-23, 2.33879e-23, 1.38956e-23, 2.21237e-23], [20.5314, 2.76677e-23, 1.39313e-23, 1.81527e-23, 2.32754e-23, 1.38252e-23, 2.19639e-23], [20.5781, 2.75307e-23, 1.38804e-23, 1.80698e-23, 2.31635e-23, 1.37553e-23, 2.18049e-23], [20.6249, 2.73944e-23, 1.383e-23, 1.79874e-23, 2.30522e-23, 1.36858e-23, 2.16468e-23], [20.6718, 2.72587e-23, 1.37799e-23, 1.79054e-23, 2.29414e-23, 1.36169e-23, 2.14895e-23], [20.7188, 2.71237e-23, 1.37302e-23, 1.78239e-23, 2.28312e-23, 1.35484e-23, 2.13331e-23], [20.766, 2.69894e-23, 1.36809e-23, 1.77428e-23, 2.27216e-23, 1.34803e-23, 2.11775e-23], [20.8132, 2.68557e-23, 1.36319e-23, 1.76622e-23, 2.26124e-23, 1.34128e-23, 2.10228e-23], [20.8605, 2.67227e-23, 1.35833e-23, 1.7582e-23, 2.25039e-23, 1.33456e-23, 2.08689e-23], [20.908, 2.65903e-23, 1.35351e-23, 1.75023e-23, 2.23959e-23, 1.3279e-23, 2.07159e-23], [20.9556, 2.64586e-23, 1.34872e-23, 1.74229e-23, 2.22884e-23, 1.32128e-23, 2.05637e-23], [21.0032, 2.63275e-23, 1.34397e-23, 1.7344e-23, 2.21814e-23, 1.31471e-23, 2.04123e-23], [21.051, 2.61971e-23, 1.33925e-23, 1.72656e-23, 2.2075e-23, 1.30818e-23, 2.02618e-23], [21.0989, 2.60673e-23, 1.33457e-23, 1.71875e-23, 2.19692e-23, 1.30169e-23, 2.01121e-23], [21.1469, 2.59381e-23, 1.32992e-23, 1.71099e-23, 2.18638e-23, 1.29526e-23, 1.99632e-23], [21.195, 2.58096e-23, 1.32531e-23, 1.70327e-23, 2.1759e-23, 1.28886e-23, 1.98152e-23], [21.2432, 2.56817e-23, 1.32073e-23, 1.69559e-23, 2.16548e-23, 1.28251e-23, 1.9668e-23], [21.2915, 2.55544e-23, 1.31618e-23, 1.68795e-23, 2.1551e-23, 1.27621e-23, 1.95216e-23], [21.3399, 2.54278e-23, 1.31167e-23, 1.68036e-23, 2.14478e-23, 1.26994e-23, 1.9376e-23], [21.3885, 2.53018e-23, 1.30719e-23, 1.6728e-23, 2.1345e-23, 1.26373e-23, 1.92312e-23], [21.4371, 2.51764e-23, 1.30275e-23, 1.66529e-23, 2.12428e-23, 1.25755e-23, 1.90873e-23], [21.4859, 2.50516e-23, 1.29834e-23, 1.65782e-23, 2.11411e-23, 1.25142e-23, 1.89441e-23], [21.5347, 2.49274e-23, 1.29396e-23, 1.65038e-23, 2.104e-23, 1.24533e-23, 1.88018e-23], [21.5837, 2.48038e-23, 1.28962e-23, 1.64299e-23, 2.09393e-23, 1.23929e-23, 1.86603e-23], [21.6328, 2.46809e-23, 1.2853e-23, 1.63564e-23, 2.08391e-23, 1.23329e-23, 1.85196e-23], [21.682, 2.45585e-23, 1.28102e-23, 1.62833e-23, 2.07394e-23, 1.22733e-23, 1.83797e-23], [21.7313, 2.44367e-23, 1.27678e-23, 1.62105e-23, 2.06403e-23, 1.22141e-23, 1.82405e-23], [21.7808, 2.43156e-23, 1.27256e-23, 1.61382e-23, 2.05416e-23, 1.21553e-23, 1.81022e-23], [21.8303, 2.4195e-23, 1.26838e-23, 1.60662e-23, 2.04434e-23, 1.2097e-23, 1.79647e-23], [21.88, 2.4075e-23, 1.26422e-23, 1.59947e-23, 2.03457e-23, 1.20391e-23, 1.7828e-23], [21.9297, 2.39556e-23, 1.2601e-23, 1.59235e-23, 2.02485e-23, 1.19816e-23, 1.76921e-23], [21.9796, 2.38368e-23, 1.25601e-23, 1.58527e-23, 2.01518e-23, 1.19245e-23, 1.7557e-23], [22.0296, 2.37186e-23, 1.25195e-23, 1.57823e-23, 2.00556e-23, 1.18678e-23, 1.74226e-23], [22.0797, 2.3601e-23, 1.24792e-23, 1.57123e-23, 1.99598e-23, 1.18115e-23, 1.72891e-23], [22.1299, 2.34839e-23, 1.24392e-23, 1.56426e-23, 1.98646e-23, 1.17556e-23, 1.71563e-23], [22.1803, 2.33674e-23, 1.23995e-23, 1.55733e-23, 1.97698e-23, 1.17002e-23, 1.70243e-23], [22.2307, 2.32515e-23, 1.23601e-23, 1.55044e-23, 1.96755e-23, 1.16451e-23, 1.68931e-23], [22.2813, 2.31362e-23, 1.2321e-23, 1.54359e-23, 1.95816e-23, 1.15905e-23, 1.67627e-23], [22.332, 2.30214e-23, 1.22822e-23, 1.53677e-23, 1.94882e-23, 1.15362e-23, 1.66331e-23], [22.3828, 2.29072e-23, 1.22437e-23, 1.52999e-23, 1.93953e-23, 1.14823e-23, 1.65043e-23], [22.4337, 2.27935e-23, 1.22054e-23, 1.52325e-23, 1.93029e-23, 1.14289e-23, 1.63762e-23], [22.4847, 2.26804e-23, 1.21675e-23, 1.51655e-23, 1.92109e-23, 1.13758e-23, 1.62489e-23], [22.5358, 2.25679e-23, 1.21299e-23, 1.50988e-23, 1.91194e-23, 1.13231e-23, 1.61224e-23], [22.5871, 2.24559e-23, 1.20925e-23, 1.50324e-23, 1.90283e-23, 1.12708e-23, 1.59967e-23], [22.6385, 2.23444e-23, 1.20554e-23, 1.49664e-23, 1.89377e-23, 1.12189e-23, 1.58717e-23], [22.69, 2.22335e-23, 1.20186e-23, 1.49008e-23, 1.88476e-23, 1.11674e-23, 1.57475e-23], [22.7416, 2.21232e-23, 1.19821e-23, 1.48355e-23, 1.87579e-23, 1.11163e-23, 1.56241e-23], [22.7933, 2.20134e-23, 1.19459e-23, 1.47706e-23, 1.86686e-23, 1.10656e-23, 1.55014e-23], [22.8452, 2.19041e-23, 1.19099e-23, 1.4706e-23, 1.85798e-23, 1.10152e-23, 1.53796e-23], [22.8971, 2.17954e-23, 1.18742e-23, 1.46418e-23, 1.84915e-23, 1.09652e-23, 1.52584e-23], [22.9492, 2.16872e-23, 1.18388e-23, 1.45779e-23, 1.84035e-23, 1.09156e-23, 1.51381e-23], [23.0014, 2.15796e-23, 1.18036e-23, 1.45144e-23, 1.83161e-23, 1.08664e-23, 1.50185e-23], [23.0537, 2.14724e-23, 1.17688e-23, 1.44512e-23, 1.8229e-23, 1.08175e-23, 1.48997e-23], [23.1062, 2.13658e-23, 1.17341e-23, 1.43883e-23, 1.81424e-23, 1.0769e-23, 1.47817e-23], [23.1587, 2.12598e-23, 1.16998e-23, 1.43258e-23, 1.80562e-23, 1.07209e-23, 1.46644e-23], [23.2114, 2.11542e-23, 1.16657e-23, 1.42637e-23, 1.79705e-23, 1.06732e-23, 1.45479e-23], [23.2642, 2.10492e-23, 1.16319e-23, 1.42018e-23, 1.78852e-23, 1.06258e-23, 1.44322e-23], [23.3171, 2.09447e-23, 1.15983e-23, 1.41403e-23, 1.78003e-23, 1.05788e-23, 1.43172e-23], [23.3701, 2.08407e-23, 1.1565e-23, 1.40791e-23, 1.77159e-23, 1.05322e-23, 1.4203e-23], [23.4233, 2.07372e-23, 1.15319e-23, 1.40183e-23, 1.76318e-23, 1.04859e-23, 1.40895e-23], [23.4766, 2.06342e-23, 1.14991e-23, 1.39578e-23, 1.75482e-23, 1.044e-23, 1.39768e-23], [23.53, 2.05318e-23, 1.14666e-23, 1.38976e-23, 1.7465e-23, 1.03944e-23, 1.38649e-23], [23.5835, 2.04298e-23, 1.14343e-23, 1.38377e-23, 1.73822e-23, 1.03492e-23, 1.37538e-23], [23.6371, 2.03284e-23, 1.14022e-23, 1.37782e-23, 1.72999e-23, 1.03044e-23, 1.36434e-23], [23.6909, 2.02274e-23, 1.13704e-23, 1.3719e-23, 1.72179e-23, 1.02599e-23, 1.35337e-23], [23.7448, 2.0127e-23, 1.13388e-23, 1.36601e-23, 1.71364e-23, 1.02157e-23, 1.34249e-23], [23.7988, 2.0027e-23, 1.13075e-23, 1.36015e-23, 1.70553e-23, 1.01719e-23, 1.33168e-23], [23.8529, 1.99276e-23, 1.12764e-23, 1.35432e-23, 1.69745e-23, 1.01285e-23, 1.32094e-23], [23.9072, 1.98286e-23, 1.12456e-23, 1.34853e-23, 1.68942e-23, 1.00854e-23, 1.31028e-23], [23.9616, 1.97301e-23, 1.1215e-23, 1.34277e-23, 1.68143e-23, 1.00427e-23, 1.2997e-23], [24.0161, 1.96321e-23, 1.11846e-23, 1.33703e-23, 1.67348e-23, 1.00003e-23, 1.2892e-23], [24.0707, 1.95346e-23, 1.11545e-23, 1.33133e-23, 1.66557e-23, 9.95823e-24, 1.27877e-23], [24.1254, 1.94376e-23, 1.11246e-23, 1.32566e-23, 1.6577e-23, 9.91652e-24, 1.26842e-23], [24.1803, 1.93411e-23, 1.10949e-23, 1.32002e-23, 1.64987e-23, 9.87515e-24, 1.25815e-23], [24.2353, 1.9245e-23, 1.10655e-23, 1.31442e-23, 1.64207e-23, 9.83412e-24, 1.24795e-23], [24.2904, 1.91495e-23, 1.10363e-23, 1.30884e-23, 1.63432e-23, 9.79343e-24, 1.23783e-23], [24.3457, 1.90544e-23, 1.10073e-23, 1.30329e-23, 1.62661e-23, 9.75307e-24, 1.22778e-23], [24.4011, 1.89597e-23, 1.09786e-23, 1.29777e-23, 1.61893e-23, 9.71305e-24, 1.21782e-23], [24.4566, 1.88656e-23, 1.09501e-23, 1.29228e-23, 1.61129e-23, 9.67336e-24, 1.20793e-23], [24.5122, 1.87719e-23, 1.09217e-23, 1.28682e-23, 1.60369e-23, 9.634e-24, 1.19811e-23], [24.568, 1.86786e-23, 1.08937e-23, 1.28139e-23, 1.59613e-23, 9.59497e-24, 1.18838e-23], [24.6238, 1.85859e-23, 1.08658e-23, 1.276e-23, 1.58861e-23, 9.55627e-24, 1.17872e-23], [24.6798, 1.84936e-23, 1.08382e-23, 1.27063e-23, 1.58113e-23, 9.5179e-24, 1.16914e-23], [24.736, 1.84018e-23, 1.08107e-23, 1.26528e-23, 1.57368e-23, 9.47985e-24, 1.15964e-23], [24.7922, 1.83104e-23, 1.07835e-23, 1.25997e-23, 1.56627e-23, 9.44213e-24, 1.15022e-23], [24.8486, 1.82195e-23, 1.07565e-23, 1.25469e-23, 1.5589e-23, 9.40473e-24, 1.14087e-23], [24.9052, 1.8129e-23, 1.07297e-23, 1.24944e-23, 1.55156e-23, 9.36765e-24, 1.1316e-23], [24.9618, 1.8039e-23, 1.07031e-23, 1.24421e-23, 1.54427e-23, 9.33089e-24, 1.12241e-23], [25.0186, 1.79494e-23, 1.06768e-23, 1.23901e-23, 1.53701e-23, 9.29444e-24, 1.1133e-23], [25.0755, 1.78603e-23, 1.06506e-23, 1.23384e-23, 1.52978e-23, 9.25832e-24, 1.10427e-23], [25.1325, 1.77716e-23, 1.06247e-23, 1.2287e-23, 1.52259e-23, 9.22251e-24, 1.09531e-23], [25.1897, 1.76834e-23, 1.05989e-23, 1.22359e-23, 1.51544e-23, 9.18701e-24, 1.08644e-23], [25.247, 1.75956e-23, 1.05734e-23, 1.21851e-23, 1.50833e-23, 9.15182e-24, 1.07765e-23], [25.3044, 1.75083e-23, 1.0548e-23, 1.21345e-23, 1.50125e-23, 9.11695e-24, 1.06893e-23], [25.362, 1.74213e-23, 1.05229e-23, 1.20842e-23, 1.49421e-23, 9.08238e-24, 1.06029e-23], [25.4197, 1.73349e-23, 1.04979e-23, 1.20342e-23, 1.4872e-23, 9.04812e-24, 1.05174e-23], [25.4775, 1.72488e-23, 1.04732e-23, 1.19844e-23, 1.48023e-23, 9.01417e-24, 1.04326e-23], [25.5354, 1.71632e-23, 1.04486e-23, 1.1935e-23, 1.47329e-23, 8.98052e-24, 1.03487e-23], [25.5935, 1.70781e-23, 1.04243e-23, 1.18857e-23, 1.46639e-23, 8.94717e-24, 1.02655e-23], [25.6517, 1.69933e-23, 1.04001e-23, 1.18368e-23, 1.45952e-23, 8.91413e-24, 1.01832e-23], [25.7101, 1.6909e-23, 1.03761e-23, 1.17881e-23, 1.45269e-23, 8.88139e-24, 1.01017e-23], [25.7685, 1.68251e-23, 1.03523e-23, 1.17397e-23, 1.44589e-23, 8.84894e-24, 1.00209e-23], [25.8272, 1.67416e-23, 1.03288e-23, 1.16916e-23, 1.43913e-23, 8.81679e-24, 9.94105e-24], [25.8859, 1.66586e-23, 1.03054e-23, 1.16437e-23, 1.4324e-23, 8.78494e-24, 9.86198e-24], [25.9448, 1.6576e-23, 1.02821e-23, 1.15961e-23, 1.42571e-23, 8.75339e-24, 9.78373e-24], [26.0038, 1.64938e-23, 1.02591e-23, 1.15488e-23, 1.41905e-23, 8.72212e-24, 9.7063e-24], [26.0629, 1.6412e-23, 1.02363e-23, 1.15017e-23, 1.41242e-23, 8.69115e-24, 9.6297e-24], [26.1222, 1.63306e-23, 1.02136e-23, 1.14549e-23, 1.40583e-23, 8.66047e-24, 9.55394e-24], [26.1816, 1.62496e-23, 1.01911e-23, 1.14083e-23, 1.39927e-23, 8.63007e-24, 9.479e-24], [26.2412, 1.6169e-23, 1.01688e-23, 1.1362e-23, 1.39274e-23, 8.59997e-24, 9.40491e-24], [26.3009, 1.60889e-23, 1.01467e-23, 1.13159e-23, 1.38625e-23, 8.57015e-24, 9.33166e-24], [26.3607, 1.60092e-23, 1.01247e-23, 1.12701e-23, 1.37979e-23, 8.54061e-24, 9.25924e-24], [26.4207, 1.59298e-23, 1.0103e-23, 1.12246e-23, 1.37336e-23, 8.51135e-24, 9.18768e-24], [26.4808, 1.58509e-23, 1.00814e-23, 1.11793e-23, 1.36697e-23, 8.48238e-24, 9.11697e-24], [26.541, 1.57723e-23, 1.006e-23, 1.11342e-23, 1.36061e-23, 8.45369e-24, 9.04711e-24], [26.6014, 1.56942e-23, 1.00387e-23, 1.10894e-23, 1.35428e-23, 8.42527e-24, 8.97811e-24], [26.6619, 1.56165e-23, 1.00176e-23, 1.10449e-23, 1.34798e-23, 8.39714e-24, 8.90997e-24], [26.7225, 1.55391e-23, 9.99674e-24, 1.10005e-23, 1.34172e-23, 8.36928e-24, 8.84269e-24], [26.7833, 1.54622e-23, 9.97601e-24, 1.09565e-23, 1.33549e-23, 8.34169e-24, 8.77628e-24], [26.8442, 1.53856e-23, 9.95544e-24, 1.09127e-23, 1.32929e-23, 8.31437e-24, 8.71074e-24], [26.9053, 1.53095e-23, 9.93504e-24, 1.08691e-23, 1.32312e-23, 8.28733e-24, 8.64607e-24], [26.9665, 1.52337e-23, 9.91481e-24, 1.08258e-23, 1.31698e-23, 8.26056e-24, 8.58228e-24], [27.0278, 1.51583e-23, 9.89474e-24, 1.07827e-23, 1.31088e-23, 8.23405e-24, 8.51937e-24], [27.0893, 1.50833e-23, 9.87483e-24, 1.07398e-23, 1.3048e-23, 8.20781e-24, 8.45735e-24], [27.1509, 1.50087e-23, 9.85508e-24, 1.06972e-23, 1.29876e-23, 8.18184e-24, 8.39622e-24], [27.2127, 1.49345e-23, 9.8355e-24, 1.06548e-23, 1.29275e-23, 8.15613e-24, 8.33597e-24], [27.2746, 1.48607e-23, 9.81607e-24, 1.06127e-23, 1.28677e-23, 8.13068e-24, 8.27662e-24], [27.3366, 1.47872e-23, 9.79681e-24, 1.05707e-23, 1.28082e-23, 8.10549e-24, 8.21817e-24], [27.3988, 1.47141e-23, 9.7777e-24, 1.05291e-23, 1.2749e-23, 8.08057e-24, 8.16063e-24], [27.4611, 1.46414e-23, 9.75874e-24, 1.04876e-23, 1.26901e-23, 8.0559e-24, 8.10398e-24], [27.5236, 1.4569e-23, 9.73995e-24, 1.04464e-23, 1.26315e-23, 8.03149e-24, 8.04824e-24], [27.5862, 1.44971e-23, 9.7213e-24, 1.04054e-23, 1.25733e-23, 8.00733e-24, 7.99342e-24], [27.6489, 1.44255e-23, 9.70281e-24, 1.03647e-23, 1.25153e-23, 7.98343e-24, 7.93951e-24], [27.7118, 1.43543e-23, 9.68447e-24, 1.03241e-23, 1.24576e-23, 7.95978e-24, 7.88652e-24], [27.7748, 1.42834e-23, 9.66628e-24, 1.02838e-23, 1.24002e-23, 7.93638e-24, 7.83445e-24], [27.838, 1.42129e-23, 9.64824e-24, 1.02438e-23, 1.23431e-23, 7.91323e-24, 7.7833e-24], [27.9013, 1.41428e-23, 9.63035e-24, 1.02039e-23, 1.22863e-23, 7.89032e-24, 7.73308e-24], [27.9648, 1.40731e-23, 9.6126e-24, 1.01643e-23, 1.22298e-23, 7.86767e-24, 7.68378e-24], [28.0284, 1.40037e-23, 9.595e-24, 1.01249e-23, 1.21736e-23, 7.84526e-24, 7.63542e-24], [28.0921, 1.39346e-23, 9.57755e-24, 1.00857e-23, 1.21177e-23, 7.82309e-24, 7.588e-24], [28.156, 1.3866e-23, 9.56024e-24, 1.00467e-23, 1.20621e-23, 7.80116e-24, 7.54151e-24], [28.2201, 1.37977e-23, 9.54307e-24, 1.0008e-23, 1.20068e-23, 7.77948e-24, 7.49596e-24], [28.2843, 1.37297e-23, 9.52605e-24, 9.96948e-24, 1.19517e-23, 7.75803e-24, 7.45135e-24], [28.3486, 1.36621e-23, 9.50916e-24, 9.93117e-24, 1.1897e-23, 7.73682e-24, 7.40768e-24], [28.4131, 1.35949e-23, 9.49242e-24, 9.89308e-24, 1.18425e-23, 7.71585e-24, 7.36496e-24], [28.4777, 1.3528e-23, 9.47581e-24, 9.8552e-24, 1.17883e-23, 7.69512e-24, 7.32319e-24], [28.5425, 1.34614e-23, 9.45934e-24, 9.81754e-24, 1.17344e-23, 7.67461e-24, 7.28236e-24], [28.6074, 1.33952e-23, 9.44301e-24, 9.78009e-24, 1.16808e-23, 7.65434e-24, 7.24249e-24], [28.6725, 1.33294e-23, 9.42682e-24, 9.74285e-24, 1.16275e-23, 7.6343e-24, 7.20356e-24], [28.7377, 1.32639e-23, 9.41075e-24, 9.70583e-24, 1.15744e-23, 7.61448e-24, 7.16559e-24], [28.8031, 1.31987e-23, 9.39483e-24, 9.66901e-24, 1.15216e-23, 7.5949e-24, 7.12857e-24], [28.8686, 1.31339e-23, 9.37903e-24, 9.6324e-24, 1.14691e-23, 7.57554e-24, 7.0925e-24], [28.9342, 1.30694e-23, 9.36337e-24, 9.596e-24, 1.14169e-23, 7.5564e-24, 7.05738e-24], [29.0001, 1.30053e-23, 9.34783e-24, 9.55981e-24, 1.13649e-23, 7.53749e-24, 7.02322e-24], [29.066, 1.29415e-23, 9.33243e-24, 9.52382e-24, 1.13132e-23, 7.5188e-24, 6.99001e-24], [29.1321, 1.28781e-23, 9.31715e-24, 9.48804e-24, 1.12618e-23, 7.50033e-24, 6.95775e-24], [29.1984, 1.28149e-23, 9.30201e-24, 9.45246e-24, 1.12107e-23, 7.48207e-24, 6.92645e-24], [29.2648, 1.27522e-23, 9.28699e-24, 9.41709e-24, 1.11598e-23, 7.46404e-24, 6.89609e-24], [29.3314, 1.26897e-23, 9.27209e-24, 9.38191e-24, 1.11092e-23, 7.44622e-24, 6.86669e-24], [29.3981, 1.26276e-23, 9.25732e-24, 9.34694e-24, 1.10589e-23, 7.42861e-24, 6.83823e-24], [29.465, 1.25658e-23, 9.24267e-24, 9.31217e-24, 1.10088e-23, 7.41122e-24, 6.81072e-24], [29.532, 1.25044e-23, 9.22815e-24, 9.27759e-24, 1.0959e-23, 7.39403e-24, 6.78416e-24], [29.5992, 1.24432e-23, 9.21375e-24, 9.24322e-24, 1.09094e-23, 7.37706e-24, 6.75853e-24], [29.6665, 1.23824e-23, 9.19947e-24, 9.20904e-24, 1.08602e-23, 7.3603e-24, 6.73385e-24], [29.734, 1.2322e-23, 9.1853e-24, 9.17506e-24, 1.08111e-23, 7.34374e-24, 6.7101e-24], [29.8016, 1.22618e-23, 9.17126e-24, 9.14127e-24, 1.07624e-23, 7.32739e-24, 6.68729e-24], [29.8694, 1.2202e-23, 9.15734e-24, 9.10768e-24, 1.07139e-23, 7.31124e-24, 6.6654e-24], [29.9373, 1.21425e-23, 9.14353e-24, 9.07428e-24, 1.06656e-23, 7.29529e-24, 6.64444e-24], [30.0054, 1.20833e-23, 9.12984e-24, 9.04108e-24, 1.06176e-23, 7.27955e-24, 6.6244e-24], [30.0737, 1.20244e-23, 9.11627e-24, 9.00806e-24, 1.05699e-23, 7.264e-24, 6.60528e-24], [30.1421, 1.19659e-23, 9.10281e-24, 8.97524e-24, 1.05224e-23, 7.24865e-24, 6.58708e-24], [30.2106, 1.19077e-23, 9.08946e-24, 8.9426e-24, 1.04752e-23, 7.2335e-24, 6.56978e-24], [30.2793, 1.18498e-23, 9.07623e-24, 8.91016e-24, 1.04282e-23, 7.21854e-24, 6.55338e-24], [30.3482, 1.17922e-23, 9.0631e-24, 8.8779e-24, 1.03815e-23, 7.20378e-24, 6.53788e-24], [30.4172, 1.17349e-23, 9.05009e-24, 8.84583e-24, 1.03351e-23, 7.18921e-24, 6.52327e-24], [30.4864, 1.16779e-23, 9.03719e-24, 8.81394e-24, 1.02888e-23, 7.17482e-24, 6.50954e-24], [30.5558, 1.16213e-23, 9.0244e-24, 8.78225e-24, 1.02429e-23, 7.16063e-24, 6.4967e-24], [30.6253, 1.15649e-23, 9.01172e-24, 8.75073e-24, 1.01971e-23, 7.14663e-24, 6.48472e-24], [30.6949, 1.15089e-23, 8.99914e-24, 8.7194e-24, 1.01517e-23, 7.13281e-24, 6.47361e-24], [30.7647, 1.14531e-23, 8.98667e-24, 8.68825e-24, 1.01064e-23, 7.11918e-24, 6.46336e-24], [30.8347, 1.13977e-23, 8.97431e-24, 8.65728e-24, 1.00615e-23, 7.10573e-24, 6.45395e-24], [30.9049, 1.13426e-23, 8.96205e-24, 8.6265e-24, 1.00167e-23, 7.09246e-24, 6.44539e-24], [30.9752, 1.12877e-23, 8.9499e-24, 8.59589e-24, 9.9722e-24, 7.07937e-24, 6.43766e-24], [31.0456, 1.12332e-23, 8.93785e-24, 8.56546e-24, 9.92794e-24, 7.06646e-24, 6.43076e-24], [31.1162, 1.1179e-23, 8.9259e-24, 8.53521e-24, 9.88392e-24, 7.05373e-24, 6.42467e-24], [31.187, 1.11251e-23, 8.91406e-24, 8.50514e-24, 9.84013e-24, 7.04117e-24, 6.4194e-24], [31.2579, 1.10714e-23, 8.90231e-24, 8.47524e-24, 9.79659e-24, 7.02879e-24, 6.41492e-24], [31.329, 1.10181e-23, 8.89067e-24, 8.44552e-24, 9.75328e-24, 7.01658e-24, 6.41123e-24], [31.4003, 1.09651e-23, 8.87912e-24, 8.41598e-24, 9.7102e-24, 7.00454e-24, 6.40832e-24], [31.4717, 1.09124e-23, 8.86767e-24, 8.3866e-24, 9.66736e-24, 6.99267e-24, 6.40618e-24], [31.5433, 1.08599e-23, 8.85633e-24, 8.35741e-24, 9.62476e-24, 6.98098e-24, 6.4048e-24], [31.6151, 1.08078e-23, 8.84507e-24, 8.32838e-24, 9.58239e-24, 6.96944e-24, 6.40418e-24], [31.687, 1.07559e-23, 8.83392e-24, 8.29952e-24, 9.54024e-24, 6.95808e-24, 6.40429e-24], [31.759, 1.07043e-23, 8.82286e-24, 8.27084e-24, 9.49833e-24, 6.94688e-24, 6.40514e-24], [31.8313, 1.06531e-23, 8.81189e-24, 8.24233e-24, 9.45665e-24, 6.93584e-24, 6.40671e-24], [31.9037, 1.06021e-23, 8.80102e-24, 8.21398e-24, 9.4152e-24, 6.92497e-24, 6.40898e-24], [31.9762, 1.05514e-23, 8.79024e-24, 8.1858e-24, 9.37398e-24, 6.91425e-24, 6.41196e-24], [32.049, 1.0501e-23, 8.77956e-24, 8.15779e-24, 9.33298e-24, 6.90369e-24, 6.41562e-24], [32.1219, 1.04508e-23, 8.76896e-24, 8.12995e-24, 9.29221e-24, 6.89329e-24, 6.41997e-24], [32.1949, 1.0401e-23, 8.75846e-24, 8.10227e-24, 9.25166e-24, 6.88305e-24, 6.42498e-24], [32.2682, 1.03514e-23, 8.74804e-24, 8.07476e-24, 9.21134e-24, 6.87296e-24, 6.43064e-24], [32.3416, 1.03021e-23, 8.73772e-24, 8.04741e-24, 9.17124e-24, 6.86302e-24, 6.43695e-24], [32.4151, 1.02531e-23, 8.72748e-24, 8.02022e-24, 9.13137e-24, 6.85324e-24, 6.44389e-24], [32.4889, 1.02044e-23, 8.71734e-24, 7.9932e-24, 9.09171e-24, 6.8436e-24, 6.45146e-24], [32.5628, 1.0156e-23, 8.70728e-24, 7.96633e-24, 9.05228e-24, 6.83412e-24, 6.45964e-24], [32.6368, 1.01078e-23, 8.6973e-24, 7.93963e-24, 9.01306e-24, 6.82478e-24, 6.46841e-24], [32.7111, 1.00599e-23, 8.68741e-24, 7.91309e-24, 8.97406e-24, 6.81559e-24, 6.47778e-24], [32.7855, 1.00123e-23, 8.67761e-24, 7.88671e-24, 8.93528e-24, 6.80654e-24, 6.48772e-24], [32.86, 9.96495e-24, 8.66789e-24, 7.86048e-24, 8.89672e-24, 6.79764e-24, 6.49823e-24], [32.9348, 9.91788e-24, 8.65826e-24, 7.83442e-24, 8.85837e-24, 6.78887e-24, 6.5093e-24], [33.0097, 9.87108e-24, 8.64871e-24, 7.80851e-24, 8.82024e-24, 6.78025e-24, 6.52091e-24], [33.0848, 9.82456e-24, 8.63924e-24, 7.78276e-24, 8.78232e-24, 6.77177e-24, 6.53305e-24], [33.16, 9.7783e-24, 8.62985e-24, 7.75716e-24, 8.74461e-24, 6.76343e-24, 6.54572e-24], [33.2354, 9.73232e-24, 8.62054e-24, 7.73172e-24, 8.70712e-24, 6.75522e-24, 6.5589e-24], [33.311, 9.6866e-24, 8.61132e-24, 7.70643e-24, 8.66983e-24, 6.74715e-24, 6.57257e-24], [33.3868, 9.64115e-24, 8.60217e-24, 7.68129e-24, 8.63276e-24, 6.73921e-24, 6.58674e-24], [33.4628, 9.59596e-24, 8.5931e-24, 7.65631e-24, 8.5959e-24, 6.7314e-24, 6.60138e-24], [33.5389, 9.55104e-24, 8.58412e-24, 7.63147e-24, 8.55924e-24, 6.72373e-24, 6.61649e-24], [33.6152, 9.50639e-24, 8.5752e-24, 7.60679e-24, 8.5228e-24, 6.71619e-24, 6.63206e-24], [33.6916, 9.46199e-24, 8.56637e-24, 7.58226e-24, 8.48655e-24, 6.70877e-24, 6.64807e-24], [33.7682, 9.41786e-24, 8.55761e-24, 7.55788e-24, 8.45052e-24, 6.70148e-24, 6.66452e-24], [33.8451, 9.37399e-24, 8.54893e-24, 7.53365e-24, 8.41469e-24, 6.69432e-24, 6.6814e-24], [33.922, 9.33038e-24, 8.54032e-24, 7.50956e-24, 8.37906e-24, 6.68728e-24, 6.69869e-24], [33.9992, 9.28703e-24, 8.53179e-24, 7.48562e-24, 8.34364e-24, 6.68037e-24, 6.71638e-24], [34.0765, 9.24394e-24, 8.52333e-24, 7.46183e-24, 8.30842e-24, 6.67357e-24, 6.73447e-24], [34.154, 9.2011e-24, 8.51494e-24, 7.43818e-24, 8.27341e-24, 6.6669e-24, 6.75294e-24], [34.2317, 9.15852e-24, 8.50663e-24, 7.41468e-24, 8.23859e-24, 6.66035e-24, 6.77179e-24], [34.3096, 9.1162e-24, 8.49839e-24, 7.39132e-24, 8.20397e-24, 6.65392e-24, 6.791e-24], [34.3876, 9.07413e-24, 8.49022e-24, 7.36811e-24, 8.16955e-24, 6.6476e-24, 6.81057e-24], [34.4658, 9.03231e-24, 8.48212e-24, 7.34504e-24, 8.13533e-24, 6.6414e-24, 6.83049e-24], [34.5442, 8.99075e-24, 8.47409e-24, 7.32211e-24, 8.10131e-24, 6.63531e-24, 6.85074e-24], [34.6228, 8.94943e-24, 8.46613e-24, 7.29932e-24, 8.06748e-24, 6.62934e-24, 6.87131e-24], [34.7016, 8.90837e-24, 8.45824e-24, 7.27668e-24, 8.03385e-24, 6.62348e-24, 6.89221e-24], [34.7805, 8.86756e-24, 8.45042e-24, 7.25417e-24, 8.00042e-24, 6.61773e-24, 6.91341e-24], [34.8596, 8.827e-24, 8.44267e-24, 7.2318e-24, 7.96718e-24, 6.61209e-24, 6.93492e-24], [34.9389, 8.78668e-24, 8.43498e-24, 7.20957e-24, 7.93413e-24, 6.60655e-24, 6.95671e-24], [35.0184, 8.74661e-24, 8.42736e-24, 7.18748e-24, 7.90127e-24, 6.60113e-24, 6.97879e-24], [35.098, 8.70679e-24, 8.41981e-24, 7.16552e-24, 7.86861e-24, 6.59581e-24, 7.00114e-24], [35.1779, 8.66721e-24, 8.41232e-24, 7.1437e-24, 7.83614e-24, 6.59059e-24, 7.02376e-24], [35.2579, 8.62787e-24, 8.4049e-24, 7.12202e-24, 7.80386e-24, 6.58548e-24, 7.04664e-24], [35.3381, 8.58878e-24, 8.39754e-24, 7.10047e-24, 7.77176e-24, 6.58047e-24, 7.06977e-24], [35.4184, 8.54993e-24, 8.39024e-24, 7.07906e-24, 7.73986e-24, 6.57556e-24, 7.09314e-24], [35.499, 8.51133e-24, 8.38301e-24, 7.05778e-24, 7.70814e-24, 6.57075e-24, 7.11674e-24], [35.5798, 8.47296e-24, 8.37584e-24, 7.03663e-24, 7.67661e-24, 6.56604e-24, 7.14057e-24], [35.6607, 8.43483e-24, 8.36874e-24, 7.01562e-24, 7.64527e-24, 6.56143e-24, 7.16462e-24], [35.7418, 8.39695e-24, 8.36169e-24, 6.99473e-24, 7.61411e-24, 6.55691e-24, 7.18888e-24], [35.8231, 8.35929e-24, 8.35471e-24, 6.97398e-24, 7.58314e-24, 6.55249e-24, 7.21335e-24], [35.9046, 8.32188e-24, 8.34779e-24, 6.95336e-24, 7.55235e-24, 6.54817e-24, 7.23802e-24], [35.9862, 8.2847e-24, 8.34092e-24, 6.93287e-24, 7.52174e-24, 6.54393e-24, 7.26287e-24], [36.0681, 8.24776e-24, 8.33412e-24, 6.9125e-24, 7.49131e-24, 6.53979e-24, 7.28791e-24], [36.1501, 8.21106e-24, 8.32737e-24, 6.89227e-24, 7.46107e-24, 6.53574e-24, 7.31313e-24], [36.2324, 8.17458e-24, 8.32069e-24, 6.87216e-24, 7.43101e-24, 6.53178e-24, 7.33852e-24], [36.3148, 8.13834e-24, 8.31406e-24, 6.85218e-24, 7.40113e-24, 6.52791e-24, 7.36407e-24], [36.3974, 8.10233e-24, 8.30749e-24, 6.83232e-24, 7.37142e-24, 6.52412e-24, 7.38978e-24], [36.4802, 8.06655e-24, 8.30098e-24, 6.81259e-24, 7.3419e-24, 6.52042e-24, 7.41564e-24], [36.5631, 8.03101e-24, 8.29452e-24, 6.79299e-24, 7.31255e-24, 6.51681e-24, 7.44165e-24], [36.6463, 7.99569e-24, 8.28812e-24, 6.77351e-24, 7.28338e-24, 6.51328e-24, 7.4678e-24], [36.7297, 7.9606e-24, 8.28178e-24, 6.75415e-24, 7.25439e-24, 6.50983e-24, 7.49409e-24], [36.8132, 7.92574e-24, 8.27549e-24, 6.73491e-24, 7.22557e-24, 6.50647e-24, 7.5205e-24], [36.8969, 7.8911e-24, 8.26925e-24, 6.7158e-24, 7.19693e-24, 6.50319e-24, 7.54704e-24], [36.9809, 7.85669e-24, 8.26307e-24, 6.69681e-24, 7.16846e-24, 6.49998e-24, 7.57369e-24], [37.065, 7.82251e-24, 8.25694e-24, 6.67794e-24, 7.14016e-24, 6.49686e-24, 7.60046e-24], [37.1493, 7.78855e-24, 8.25087e-24, 6.65919e-24, 7.11204e-24, 6.49381e-24, 7.62733e-24], [37.2338, 7.75481e-24, 8.24485e-24, 6.64056e-24, 7.08408e-24, 6.49085e-24, 7.65431e-24], [37.3185, 7.7213e-24, 8.23888e-24, 6.62205e-24, 7.0563e-24, 6.48795e-24, 7.68139e-24], [37.4034, 7.68801e-24, 8.23296e-24, 6.60366e-24, 7.02869e-24, 6.48513e-24, 7.70855e-24], [37.4884, 7.65494e-24, 8.22709e-24, 6.58539e-24, 7.00125e-24, 6.48239e-24, 7.73581e-24], [37.5737, 7.62209e-24, 8.22128e-24, 6.56723e-24, 6.97398e-24, 6.47972e-24, 7.76315e-24], [37.6592, 7.58946e-24, 8.21551e-24, 6.54919e-24, 6.94688e-24, 6.47712e-24, 7.79057e-24], [37.7448, 7.55705e-24, 8.20979e-24, 6.53126e-24, 6.91994e-24, 6.47459e-24, 7.81806e-24], [37.8307, 7.52486e-24, 8.20413e-24, 6.51345e-24, 6.89317e-24, 6.47213e-24, 7.84562e-24], [37.9167, 7.49288e-24, 8.19851e-24, 6.49576e-24, 6.86657e-24, 6.46974e-24, 7.87325e-24], [38.003, 7.46112e-24, 8.19294e-24, 6.47818e-24, 6.84013e-24, 6.46742e-24, 7.90094e-24], [38.0894, 7.42958e-24, 8.18742e-24, 6.46071e-24, 6.81385e-24, 6.46516e-24, 7.92869e-24], [38.176, 7.39825e-24, 8.18195e-24, 6.44336e-24, 6.78774e-24, 6.46298e-24, 7.95649e-24], [38.2629, 7.36713e-24, 8.17653e-24, 6.42611e-24, 6.7618e-24, 6.46085e-24, 7.98435e-24], [38.3499, 7.33623e-24, 8.17115e-24, 6.40898e-24, 6.73601e-24, 6.45879e-24, 8.01225e-24], [38.4371, 7.30554e-24, 8.16582e-24, 6.39196e-24, 6.71039e-24, 6.4568e-24, 8.04019e-24], [38.5246, 7.27507e-24, 8.16053e-24, 6.37505e-24, 6.68493e-24, 6.45486e-24, 8.06817e-24], [38.6122, 7.2448e-24, 8.15529e-24, 6.35826e-24, 6.65963e-24, 6.45299e-24, 8.09619e-24], [38.7, 7.21474e-24, 8.1501e-24, 6.34156e-24, 6.63449e-24, 6.45118e-24, 8.12424e-24], [38.788, 7.1849e-24, 8.14495e-24, 6.32498e-24, 6.6095e-24, 6.44943e-24, 8.15232e-24], [38.8763, 7.15526e-24, 8.13984e-24, 6.30851e-24, 6.58468e-24, 6.44774e-24, 8.18042e-24], [38.9647, 7.12583e-24, 8.13478e-24, 6.29214e-24, 6.56001e-24, 6.44611e-24, 8.20855e-24], [39.0533, 7.0966e-24, 8.12976e-24, 6.27588e-24, 6.5355e-24, 6.44453e-24, 8.2367e-24], [39.1422, 7.06759e-24, 8.12479e-24, 6.25973e-24, 6.51115e-24, 6.44301e-24, 8.26486e-24], [39.2312, 7.03877e-24, 8.11986e-24, 6.24368e-24, 6.48695e-24, 6.44154e-24, 8.29304e-24], [39.3204, 7.01017e-24, 8.11497e-24, 6.22774e-24, 6.46291e-24, 6.44013e-24, 8.32123e-24], [39.4099, 6.98176e-24, 8.11012e-24, 6.2119e-24, 6.43902e-24, 6.43878e-24, 8.34942e-24], [39.4995, 6.95357e-24, 8.10532e-24, 6.19616e-24, 6.41529e-24, 6.43747e-24, 8.37763e-24], [39.5893, 6.92557e-24, 8.10056e-24, 6.18053e-24, 6.3917e-24, 6.43622e-24, 8.40583e-24], [39.6794, 6.89777e-24, 8.09583e-24, 6.165e-24, 6.36827e-24, 6.43502e-24, 8.43403e-24], [39.7696, 6.87018e-24, 8.09115e-24, 6.14958e-24, 6.345e-24, 6.43387e-24, 8.46224e-24], [39.8601, 6.84279e-24, 8.08651e-24, 6.13425e-24, 6.32187e-24, 6.43278e-24, 8.49043e-24], [39.9508, 6.81559e-24, 8.08191e-24, 6.11903e-24, 6.29889e-24, 6.43173e-24, 8.51862e-24], [40.0416, 6.7886e-24, 8.07734e-24, 6.1039e-24, 6.27606e-24, 6.43072e-24, 8.54681e-24], [40.1327, 6.7618e-24, 8.07282e-24, 6.08888e-24, 6.25338e-24, 6.42977e-24, 8.57497e-24], [40.224, 6.7352e-24, 8.06834e-24, 6.07395e-24, 6.23085e-24, 6.42886e-24, 8.60313e-24], [40.3155, 6.7088e-24, 8.06389e-24, 6.05913e-24, 6.20847e-24, 6.428e-24, 8.63127e-24], [40.4072, 6.68259e-24, 8.05948e-24, 6.0444e-24, 6.18623e-24, 6.42718e-24, 8.65939e-24], [40.4991, 6.65658e-24, 8.05511e-24, 6.02977e-24, 6.16414e-24, 6.42641e-24, 8.68749e-24], [40.5912, 6.63076e-24, 8.05078e-24, 6.01523e-24, 6.1422e-24, 6.42568e-24, 8.71557e-24], [40.6836, 6.60514e-24, 8.04648e-24, 6.0008e-24, 6.1204e-24, 6.42499e-24, 8.74362e-24], [40.7761, 6.57971e-24, 8.04222e-24, 5.98645e-24, 6.09874e-24, 6.42435e-24, 8.77165e-24], [40.8688, 6.55447e-24, 8.038e-24, 5.97221e-24, 6.07723e-24, 6.42375e-24, 8.79965e-24], [40.9618, 6.52942e-24, 8.03381e-24, 5.95806e-24, 6.05586e-24, 6.42319e-24, 8.82762e-24], [41.055, 6.50456e-24, 8.02966e-24, 5.944e-24, 6.03464e-24, 6.42266e-24, 8.85556e-24], [41.1483, 6.4799e-24, 8.02555e-24, 5.93004e-24, 6.01355e-24, 6.42218e-24, 8.88347e-24], [41.2419, 6.45542e-24, 8.02146e-24, 5.91617e-24, 5.99261e-24, 6.42174e-24, 8.91134e-24], [41.3358, 6.43113e-24, 8.01742e-24, 5.90239e-24, 5.9718e-24, 6.42133e-24, 8.93917e-24], [41.4298, 6.40703e-24, 8.01341e-24, 5.8887e-24, 5.95114e-24, 6.42097e-24, 8.96697e-24], [41.524, 6.38312e-24, 8.00943e-24, 5.87511e-24, 5.93062e-24, 6.42063e-24, 8.99473e-24], [41.6185, 6.35939e-24, 8.00549e-24, 5.8616e-24, 5.91023e-24, 6.42034e-24, 9.02245e-24], [41.7131, 6.33585e-24, 8.00158e-24, 5.84819e-24, 5.88998e-24, 6.42008e-24, 9.05012e-24], [41.808, 6.3125e-24, 7.9977e-24, 5.83487e-24, 5.86987e-24, 6.41985e-24, 9.07775e-24], [41.9031, 6.28932e-24, 7.99385e-24, 5.82163e-24, 5.8499e-24, 6.41966e-24, 9.10534e-24], [41.9984, 6.26634e-24, 7.99004e-24, 5.80849e-24, 5.83006e-24, 6.4195e-24, 9.13288e-24], [42.0939, 6.24353e-24, 7.98626e-24, 5.79543e-24, 5.81035e-24, 6.41938e-24, 9.16037e-24], [42.1897, 6.22091e-24, 7.98251e-24, 5.78246e-24, 5.79078e-24, 6.41929e-24, 9.18781e-24], [42.2856, 6.19847e-24, 7.9788e-24, 5.76958e-24, 5.77135e-24, 6.41922e-24, 9.2152e-24], [42.3818, 6.17621e-24, 7.97511e-24, 5.75679e-24, 5.75205e-24, 6.41919e-24, 9.24254e-24], [42.4782, 6.15413e-24, 7.97146e-24, 5.74408e-24, 5.73288e-24, 6.41919e-24, 9.26983e-24], [42.5748, 6.13222e-24, 7.96784e-24, 5.73146e-24, 5.71384e-24, 6.41922e-24, 9.29707e-24], [42.6717, 6.1105e-24, 7.96425e-24, 5.71892e-24, 5.69494e-24, 6.41928e-24, 9.32425e-24], [42.7687, 6.08896e-24, 7.96068e-24, 5.70646e-24, 5.67616e-24, 6.41937e-24, 9.35138e-24], [42.866, 6.06759e-24, 7.95715e-24, 5.6941e-24, 5.65752e-24, 6.41948e-24, 9.37844e-24], [42.9635, 6.0464e-24, 7.95365e-24, 5.68181e-24, 5.639e-24, 6.41962e-24, 9.40545e-24], [43.0613, 6.02538e-24, 7.95018e-24, 5.66961e-24, 5.62062e-24, 6.41979e-24, 9.43241e-24], [43.1592, 6.00454e-24, 7.94673e-24, 5.65749e-24, 5.60236e-24, 6.41999e-24, 9.4593e-24], [43.2574, 5.98388e-24, 7.94332e-24, 5.64545e-24, 5.58423e-24, 6.42021e-24, 9.48613e-24], [43.3558, 5.96339e-24, 7.93993e-24, 5.63349e-24, 5.56623e-24, 6.42046e-24, 9.5129e-24], [43.4544, 5.94307e-24, 7.93657e-24, 5.62162e-24, 5.54835e-24, 6.42073e-24, 9.53961e-24], [43.5532, 5.92292e-24, 7.93324e-24, 5.60983e-24, 5.5306e-24, 6.42102e-24, 9.56626e-24], [43.6523, 5.90294e-24, 7.92994e-24, 5.59811e-24, 5.51298e-24, 6.42134e-24, 9.59284e-24], [43.7516, 5.88314e-24, 7.92666e-24, 5.58648e-24, 5.49548e-24, 6.42168e-24, 9.61936e-24], [43.8511, 5.86351e-24, 7.92342e-24, 5.57492e-24, 5.4781e-24, 6.42204e-24, 9.64581e-24], [43.9508, 5.84404e-24, 7.9202e-24, 5.56345e-24, 5.46085e-24, 6.42243e-24, 9.6722e-24], [44.0508, 5.82474e-24, 7.917e-24, 5.55205e-24, 5.44372e-24, 6.42284e-24, 9.69852e-24], [44.151, 5.80562e-24, 7.91384e-24, 5.54073e-24, 5.42672e-24, 6.42326e-24, 9.72477e-24], [44.2514, 5.78666e-24, 7.91069e-24, 5.52948e-24, 5.40983e-24, 6.42371e-24, 9.75095e-24], [44.3521, 5.76786e-24, 7.90758e-24, 5.51832e-24, 5.39307e-24, 6.42418e-24, 9.77707e-24], [44.453, 5.74923e-24, 7.90449e-24, 5.50722e-24, 5.37642e-24, 6.42467e-24, 9.80311e-24], [44.5541, 5.73077e-24, 7.90143e-24, 5.49621e-24, 5.3599e-24, 6.42517e-24, 9.82909e-24], [44.6554, 5.71247e-24, 7.89839e-24, 5.48527e-24, 5.34349e-24, 6.4257e-24, 9.855e-24], [44.757, 5.69434e-24, 7.89537e-24, 5.47441e-24, 5.32721e-24, 6.42624e-24, 9.88083e-24], [44.8588, 5.67637e-24, 7.89238e-24, 5.46361e-24, 5.31104e-24, 6.4268e-24, 9.9066e-24], [44.9608, 5.65856e-24, 7.88942e-24, 5.4529e-24, 5.29499e-24, 6.42738e-24, 9.93229e-24], [45.0631, 5.64091e-24, 7.88648e-24, 5.44225e-24, 5.27905e-24, 6.42798e-24, 9.95791e-24], [45.1656, 5.62342e-24, 7.88356e-24, 5.43168e-24, 5.26324e-24, 6.42859e-24, 9.98345e-24], [45.2683, 5.6061e-24, 7.88067e-24, 5.42119e-24, 5.24753e-24, 6.42921e-24, 1.00089e-23], [45.3713, 5.58893e-24, 7.8778e-24, 5.41076e-24, 5.23195e-24, 6.42985e-24, 1.00343e-23], [45.4745, 5.57192e-24, 7.87496e-24, 5.40041e-24, 5.21647e-24, 6.43051e-24, 1.00597e-23], [45.5779, 5.55508e-24, 7.87214e-24, 5.39012e-24, 5.20112e-24, 6.43118e-24, 1.00849e-23], [45.6816, 5.53838e-24, 7.86934e-24, 5.37991e-24, 5.18587e-24, 6.43186e-24, 1.01101e-23], [45.7855, 5.52185e-24, 7.86656e-24, 5.36977e-24, 5.17074e-24, 6.43256e-24, 1.01352e-23], [45.8896, 5.50547e-24, 7.86381e-24, 5.35969e-24, 5.15572e-24, 6.43327e-24, 1.01602e-23], [45.994, 5.48925e-24, 7.86108e-24, 5.34969e-24, 5.14081e-24, 6.434e-24, 1.01851e-23], [46.0986, 5.47318e-24, 7.85837e-24, 5.33976e-24, 5.12601e-24, 6.43473e-24, 1.021e-23], [46.2035, 5.45726e-24, 7.85568e-24, 5.32989e-24, 5.11132e-24, 6.43548e-24, 1.02348e-23], [46.3086, 5.4415e-24, 7.85302e-24, 5.32009e-24, 5.09674e-24, 6.43624e-24, 1.02595e-23], [46.4139, 5.42589e-24, 7.85038e-24, 5.31036e-24, 5.08227e-24, 6.43701e-24, 1.02842e-23], [46.5195, 5.41044e-24, 7.84775e-24, 5.3007e-24, 5.06791e-24, 6.43779e-24, 1.03087e-23], [46.6253, 5.39513e-24, 7.84515e-24, 5.2911e-24, 5.05366e-24, 6.43859e-24, 1.03332e-23], [46.7313, 5.37998e-24, 7.84257e-24, 5.28157e-24, 5.03951e-24, 6.43939e-24, 1.03576e-23], [46.8376, 5.36497e-24, 7.84001e-24, 5.2721e-24, 5.02547e-24, 6.4402e-24, 1.03819e-23], [46.9442, 5.35012e-24, 7.83747e-24, 5.2627e-24, 5.01154e-24, 6.44102e-24, 1.04062e-23], [47.0509, 5.33541e-24, 7.83495e-24, 5.25337e-24, 4.99771e-24, 6.44185e-24, 1.04304e-23], [47.158, 5.32085e-24, 7.83246e-24, 5.24409e-24, 4.98399e-24, 6.44269e-24, 1.04545e-23], [47.2652, 5.30644e-24, 7.82998e-24, 5.23489e-24, 4.97038e-24, 6.44353e-24, 1.04785e-23], [47.3727, 5.29217e-24, 7.82752e-24, 5.22574e-24, 4.95686e-24, 6.44439e-24, 1.05024e-23], [47.4805, 5.27805e-24, 7.82508e-24, 5.21666e-24, 4.94345e-24, 6.44525e-24, 1.05263e-23], [47.5885, 5.26407e-24, 7.82266e-24, 5.20765e-24, 4.93014e-24, 6.44611e-24, 1.055e-23], [47.6967, 5.25024e-24, 7.82026e-24, 5.19869e-24, 4.91694e-24, 6.44699e-24, 1.05737e-23], [47.8052, 5.23655e-24, 7.81788e-24, 5.1898e-24, 4.90383e-24, 6.44787e-24, 1.05973e-23], [47.914, 5.22301e-24, 7.81552e-24, 5.18097e-24, 4.89083e-24, 6.44876e-24, 1.06209e-23], [48.0229, 5.20961e-24, 7.81317e-24, 5.17219e-24, 4.87793e-24, 6.44965e-24, 1.06443e-23], [48.1322, 5.19634e-24, 7.81085e-24, 5.16348e-24, 4.86513e-24, 6.45055e-24, 1.06677e-23], [48.2417, 5.18322e-24, 7.80854e-24, 5.15483e-24, 4.85242e-24, 6.45145e-24, 1.0691e-23], [48.3514, 5.17024e-24, 7.80625e-24, 5.14624e-24, 4.83982e-24, 6.45236e-24, 1.07142e-23], [48.4614, 5.1574e-24, 7.80398e-24, 5.13771e-24, 4.82731e-24, 6.45327e-24, 1.07373e-23], [48.5716, 5.1447e-24, 7.80172e-24, 5.12924e-24, 4.8149e-24, 6.45419e-24, 1.07604e-23], [48.6821, 5.13213e-24, 7.79949e-24, 5.12083e-24, 4.80259e-24, 6.45511e-24, 1.07833e-23], [48.7928, 5.1197e-24, 7.79727e-24, 5.11247e-24, 4.79038e-24, 6.45603e-24, 1.08062e-23], [48.9038, 5.10741e-24, 7.79507e-24, 5.10418e-24, 4.77826e-24, 6.45696e-24, 1.0829e-23], [49.015, 5.09526e-24, 7.79289e-24, 5.09594e-24, 4.76623e-24, 6.45789e-24, 1.08518e-23], [49.1265, 5.08324e-24, 7.79072e-24, 5.08775e-24, 4.75431e-24, 6.45882e-24, 1.08744e-23], [49.2383, 5.07135e-24, 7.78857e-24, 5.07963e-24, 4.74247e-24, 6.45975e-24, 1.0897e-23], [49.3502, 5.0596e-24, 7.78643e-24, 5.07156e-24, 4.73073e-24, 6.46069e-24, 1.09194e-23], [49.4625, 5.04798e-24, 7.78432e-24, 5.06355e-24, 4.71908e-24, 6.46163e-24, 1.09419e-23], [49.575, 5.03649e-24, 7.78222e-24, 5.05559e-24, 4.70753e-24, 6.46257e-24, 1.09642e-23], [49.6878, 5.02514e-24, 7.78013e-24, 5.04768e-24, 4.69606e-24, 6.46351e-24, 1.09864e-23], [49.8008, 5.01391e-24, 7.77806e-24, 5.03984e-24, 4.68469e-24, 6.46445e-24, 1.10086e-23], [49.9141, 5.00282e-24, 7.77601e-24, 5.03204e-24, 4.67341e-24, 6.46539e-24, 1.10306e-23], [50.0276, 4.99185e-24, 7.77397e-24, 5.0243e-24, 4.66222e-24, 6.46634e-24, 1.10526e-23], [50.1414, 4.98102e-24, 7.77195e-24, 5.01661e-24, 4.65112e-24, 6.46728e-24, 1.10745e-23], [50.2554, 4.97031e-24, 7.76995e-24, 5.00898e-24, 4.6401e-24, 6.46822e-24, 1.10964e-23], [50.3697, 4.95973e-24, 7.76796e-24, 5.0014e-24, 4.62918e-24, 6.46916e-24, 1.11181e-23], [50.4843, 4.94928e-24, 7.76598e-24, 4.99387e-24, 4.61835e-24, 6.4701e-24, 1.11398e-23], [50.5991, 4.93895e-24, 7.76402e-24, 4.9864e-24, 4.6076e-24, 6.47104e-24, 1.11614e-23], [50.7142, 4.92875e-24, 7.76207e-24, 4.97897e-24, 4.59694e-24, 6.47198e-24, 1.11829e-23], [50.8296, 4.91867e-24, 7.76014e-24, 4.9716e-24, 4.58636e-24, 6.47292e-24, 1.12043e-23], [50.9452, 4.90872e-24, 7.75823e-24, 4.96428e-24, 4.57588e-24, 6.47386e-24, 1.12257e-23], [51.0611, 4.89889e-24, 7.75632e-24, 4.957e-24, 4.56547e-24, 6.47479e-24, 1.12469e-23], [51.1772, 4.88918e-24, 7.75444e-24, 4.94978e-24, 4.55515e-24, 6.47572e-24, 1.12681e-23], [51.2936, 4.87959e-24, 7.75256e-24, 4.94261e-24, 4.54492e-24, 6.47665e-24, 1.12892e-23], [51.4103, 4.87013e-24, 7.7507e-24, 4.93549e-24, 4.53477e-24, 6.47758e-24, 1.13103e-23], [51.5272, 4.86079e-24, 7.74886e-24, 4.92842e-24, 4.52471e-24, 6.4785e-24, 1.13312e-23], [51.6444, 4.85156e-24, 7.74703e-24, 4.92139e-24, 4.51472e-24, 6.47943e-24, 1.13521e-23], [51.7619, 4.84246e-24, 7.74521e-24, 4.91442e-24, 4.50482e-24, 6.48034e-24, 1.13728e-23], [51.8796, 4.83347e-24, 7.7434e-24, 4.90749e-24, 4.495e-24, 6.48126e-24, 1.13935e-23], [51.9976, 4.8246e-24, 7.74161e-24, 4.90061e-24, 4.48526e-24, 6.48217e-24, 1.14142e-23], [52.1159, 4.81585e-24, 7.73983e-24, 4.89378e-24, 4.4756e-24, 6.48307e-24, 1.14347e-23], [52.2345, 4.80722e-24, 7.73807e-24, 4.887e-24, 4.46602e-24, 6.48398e-24, 1.14552e-23], [52.3533, 4.7987e-24, 7.73632e-24, 4.88026e-24, 4.45653e-24, 6.48488e-24, 1.14756e-23], [52.4724, 4.79029e-24, 7.73458e-24, 4.87357e-24, 4.44711e-24, 6.48577e-24, 1.14959e-23], [52.5917, 4.782e-24, 7.73285e-24, 4.86692e-24, 4.43777e-24, 6.48666e-24, 1.15161e-23], [52.7113, 4.77383e-24, 7.73114e-24, 4.86032e-24, 4.4285e-24, 6.48754e-24, 1.15362e-23], [52.8312, 4.76577e-24, 7.72943e-24, 4.85377e-24, 4.41932e-24, 6.48842e-24, 1.15563e-23], [52.9514, 4.75782e-24, 7.72775e-24, 4.84726e-24, 4.41021e-24, 6.48929e-24, 1.15763e-23], [53.0718, 4.74998e-24, 7.72607e-24, 4.8408e-24, 4.40118e-24, 6.49016e-24, 1.15962e-23], [53.1925, 4.74225e-24, 7.7244e-24, 4.83438e-24, 4.39222e-24, 6.49102e-24, 1.1616e-23], [53.3135, 4.73463e-24, 7.72275e-24, 4.828e-24, 4.38334e-24, 6.49188e-24, 1.16358e-23], [53.4348, 4.72713e-24, 7.72111e-24, 4.82167e-24, 4.37453e-24, 6.49273e-24, 1.16554e-23], [53.5563, 4.71973e-24, 7.71948e-24, 4.81538e-24, 4.3658e-24, 6.49357e-24, 1.1675e-23], [53.6782, 4.71244e-24, 7.71787e-24, 4.80914e-24, 4.35714e-24, 6.49441e-24, 1.16945e-23], [53.8003, 4.70525e-24, 7.71626e-24, 4.80293e-24, 4.34856e-24, 6.49524e-24, 1.1714e-23], [53.9226, 4.69818e-24, 7.71467e-24, 4.79677e-24, 4.34004e-24, 6.49606e-24, 1.17333e-23], [54.0453, 4.69121e-24, 7.71308e-24, 4.79066e-24, 4.3316e-24, 6.49688e-24, 1.17526e-23], [54.1682, 4.68435e-24, 7.71151e-24, 4.78458e-24, 4.32324e-24, 6.49768e-24, 1.17718e-23], [54.2914, 4.67759e-24, 7.70995e-24, 4.77855e-24, 4.31494e-24, 6.49849e-24, 1.17909e-23], [54.4149, 4.67093e-24, 7.7084e-24, 4.77255e-24, 4.30671e-24, 6.49928e-24, 1.181e-23], [54.5387, 4.66438e-24, 7.70686e-24, 4.7666e-24, 4.29856e-24, 6.50007e-24, 1.18289e-23], [54.6627, 4.65794e-24, 7.70534e-24, 4.76069e-24, 4.29047e-24, 6.50084e-24, 1.18478e-23], [54.7871, 4.65159e-24, 7.70382e-24, 4.75482e-24, 4.28245e-24, 6.50161e-24, 1.18666e-23], [54.9117, 4.64535e-24, 7.70231e-24, 4.74899e-24, 4.27451e-24, 6.50238e-24, 1.18854e-23], [55.0366, 4.63921e-24, 7.70082e-24, 4.7432e-24, 4.26663e-24, 6.50313e-24, 1.1904e-23], [55.1618, 4.63316e-24, 7.69933e-24, 4.73745e-24, 4.25882e-24, 6.50387e-24, 1.19226e-23], [55.2872, 4.62722e-24, 7.69786e-24, 4.73174e-24, 4.25107e-24, 6.50461e-24, 1.19411e-23], [55.413, 4.62138e-24, 7.69639e-24, 4.72606e-24, 4.24339e-24, 6.50534e-24, 1.19595e-23], [55.539, 4.61564e-24, 7.69494e-24, 4.72043e-24, 4.23578e-24, 6.50606e-24, 1.19779e-23], [55.6654, 4.60999e-24, 7.6935e-24, 4.71483e-24, 4.22824e-24, 6.50677e-24, 1.19962e-23], [55.792, 4.60444e-24, 7.69206e-24, 4.70927e-24, 4.22076e-24, 6.50747e-24, 1.20144e-23], [55.9189, 4.59899e-24, 7.69064e-24, 4.70375e-24, 4.21334e-24, 6.50816e-24, 1.20325e-23], [56.0461, 4.59363e-24, 7.68922e-24, 4.69827e-24, 4.20599e-24, 6.50884e-24, 1.20505e-23], [56.1736, 4.58837e-24, 7.68782e-24, 4.69282e-24, 4.1987e-24, 6.50951e-24, 1.20685e-23], [56.3013, 4.5832e-24, 7.68642e-24, 4.68742e-24, 4.19148e-24, 6.51017e-24, 1.20864e-23], [56.4294, 4.57813e-24, 7.68504e-24, 4.68204e-24, 4.18432e-24, 6.51082e-24, 1.21043e-23], [56.5577, 4.57315e-24, 7.68366e-24, 4.67671e-24, 4.17722e-24, 6.51146e-24, 1.2122e-23], [56.6864, 4.56827e-24, 7.68229e-24, 4.67141e-24, 4.17018e-24, 6.51209e-24, 1.21397e-23], [56.8153, 4.56348e-24, 7.68094e-24, 4.66614e-24, 4.16321e-24, 6.51271e-24, 1.21573e-23], [56.9445, 4.55877e-24, 7.67959e-24, 4.66091e-24, 4.1563e-24, 6.51332e-24, 1.21748e-23], [57.0741, 4.55416e-24, 7.67825e-24, 4.65572e-24, 4.14944e-24, 6.51392e-24, 1.21923e-23], [57.2039, 4.54964e-24, 7.67692e-24, 4.65056e-24, 4.14265e-24, 6.51451e-24, 1.22096e-23], [57.334, 4.54521e-24, 7.6756e-24, 4.64544e-24, 4.13592e-24, 6.51509e-24, 1.2227e-23], [57.4644, 4.54087e-24, 7.67428e-24, 4.64035e-24, 4.12924e-24, 6.51565e-24, 1.22442e-23], [57.5951, 4.53662e-24, 7.67298e-24, 4.63529e-24, 4.12263e-24, 6.51621e-24, 1.22614e-23], [57.7261, 4.53245e-24, 7.67168e-24, 4.63027e-24, 4.11607e-24, 6.51675e-24, 1.22784e-23], [57.8574, 4.52838e-24, 7.6704e-24, 4.62528e-24, 4.10957e-24, 6.51728e-24, 1.22955e-23], [57.989, 4.52438e-24, 7.66912e-24, 4.62033e-24, 4.10313e-24, 6.5178e-24, 1.23124e-23], [58.1209, 4.52048e-24, 7.66785e-24, 4.61541e-24, 4.09675e-24, 6.51831e-24, 1.23293e-23], [58.2531, 4.51666e-24, 7.66659e-24, 4.61052e-24, 4.09042e-24, 6.51881e-24, 1.23461e-23], [58.3856, 4.51293e-24, 7.66533e-24, 4.60567e-24, 4.08415e-24, 6.51929e-24, 1.23628e-23], [58.5184, 4.50928e-24, 7.66409e-24, 4.60084e-24, 4.07793e-24, 6.51977e-24, 1.23795e-23], [58.6515, 4.50571e-24, 7.66285e-24, 4.59605e-24, 4.07177e-24, 6.52023e-24, 1.2396e-23], [58.7849, 4.50223e-24, 7.66162e-24, 4.59129e-24, 4.06566e-24, 6.52068e-24, 1.24126e-23], [58.9187, 4.49883e-24, 7.6604e-24, 4.58656e-24, 4.05961e-24, 6.52111e-24, 1.2429e-23], [59.0527, 4.49551e-24, 7.65919e-24, 4.58187e-24, 4.05361e-24, 6.52153e-24, 1.24454e-23], [59.187, 4.49228e-24, 7.65798e-24, 4.5772e-24, 4.04766e-24, 6.52195e-24, 1.24617e-23], [59.3216, 4.48912e-24, 7.65679e-24, 4.57257e-24, 4.04177e-24, 6.52234e-24, 1.24779e-23], [59.4565, 4.48605e-24, 7.6556e-24, 4.56796e-24, 4.03593e-24, 6.52273e-24, 1.24941e-23], [59.5918, 4.48305e-24, 7.65441e-24, 4.56339e-24, 4.03014e-24, 6.5231e-24, 1.25101e-23], [59.7273, 4.48014e-24, 7.65324e-24, 4.55885e-24, 4.02441e-24, 6.52346e-24, 1.25262e-23], [59.8632, 4.4773e-24, 7.65207e-24, 4.55433e-24, 4.01872e-24, 6.52381e-24, 1.25421e-23], [59.9993, 4.47454e-24, 7.65091e-24, 4.54985e-24, 4.01309e-24, 6.52414e-24, 1.2558e-23], [60.1358, 4.47186e-24, 7.64976e-24, 4.5454e-24, 4.0075e-24, 6.52446e-24, 1.25738e-23], [60.2726, 4.46926e-24, 7.64862e-24, 4.54097e-24, 4.00197e-24, 6.52477e-24, 1.25895e-23], [60.4097, 4.46673e-24, 7.64748e-24, 4.53658e-24, 3.99648e-24, 6.52506e-24, 1.26052e-23], [60.5471, 4.46428e-24, 7.64635e-24, 4.53221e-24, 3.99105e-24, 6.52534e-24, 1.26208e-23], [60.6848, 4.46191e-24, 7.64522e-24, 4.52787e-24, 3.98566e-24, 6.5256e-24, 1.26364e-23], [60.8229, 4.45961e-24, 7.64411e-24, 4.52356e-24, 3.98032e-24, 6.52586e-24, 1.26518e-23], [60.9612, 4.45738e-24, 7.643e-24, 4.51928e-24, 3.97503e-24, 6.52609e-24, 1.26672e-23], [61.0999, 4.45523e-24, 7.6419e-24, 4.51502e-24, 3.96979e-24, 6.52632e-24, 1.26826e-23], [61.2388, 4.45315e-24, 7.6408e-24, 4.5108e-24, 3.9646e-24, 6.52653e-24, 1.26978e-23], [61.3781, 4.45114e-24, 7.63971e-24, 4.5066e-24, 3.95945e-24, 6.52672e-24, 1.2713e-23], [61.5177, 4.44921e-24, 7.63863e-24, 4.50243e-24, 3.95434e-24, 6.5269e-24, 1.27281e-23], [61.6577, 4.44735e-24, 7.63755e-24, 4.49828e-24, 3.94929e-24, 6.52707e-24, 1.27432e-23], [61.7979, 4.44556e-24, 7.63648e-24, 4.49416e-24, 3.94427e-24, 6.52722e-24, 1.27582e-23], [61.9385, 4.44384e-24, 7.63542e-24, 4.49007e-24, 3.93931e-24, 6.52736e-24, 1.27731e-23], [62.0794, 4.44219e-24, 7.63437e-24, 4.48601e-24, 3.93439e-24, 6.52748e-24, 1.2788e-23], [62.2206, 4.44061e-24, 7.63332e-24, 4.48197e-24, 3.92951e-24, 6.52759e-24, 1.28028e-23], [62.3621, 4.4391e-24, 7.63227e-24, 4.47796e-24, 3.92467e-24, 6.52768e-24, 1.28175e-23], [62.5039, 4.43766e-24, 7.63124e-24, 4.47397e-24, 3.91988e-24, 6.52776e-24, 1.28322e-23], [62.6461, 4.43629e-24, 7.63021e-24, 4.47001e-24, 3.91514e-24, 6.52783e-24, 1.28468e-23], [62.7886, 4.43498e-24, 7.62918e-24, 4.46607e-24, 3.91043e-24, 6.52788e-24, 1.28613e-23], [62.9314, 4.43375e-24, 7.62816e-24, 4.46216e-24, 3.90577e-24, 6.52791e-24, 1.28758e-23], [63.0746, 4.43258e-24, 7.62715e-24, 4.45828e-24, 3.90115e-24, 6.52793e-24, 1.28902e-23], [63.218, 4.43147e-24, 7.62615e-24, 4.45441e-24, 3.89657e-24, 6.52793e-24, 1.29045e-23], [63.3618, 4.43043e-24, 7.62515e-24, 4.45058e-24, 3.89203e-24, 6.52792e-24, 1.29188e-23], [63.5059, 4.42946e-24, 7.62415e-24, 4.44676e-24, 3.88754e-24, 6.52789e-24, 1.2933e-23], [63.6504, 4.42855e-24, 7.62316e-24, 4.44298e-24, 3.88308e-24, 6.52785e-24, 1.29472e-23], [63.7952, 4.42771e-24, 7.62218e-24, 4.43921e-24, 3.87866e-24, 6.52779e-24, 1.29613e-23], [63.9403, 4.42693e-24, 7.6212e-24, 4.43547e-24, 3.87429e-24, 6.52772e-24, 1.29753e-23], [64.0857, 4.42622e-24, 7.62023e-24, 4.43175e-24, 3.86995e-24, 6.52763e-24, 1.29892e-23], [64.2315, 4.42557e-24, 7.61927e-24, 4.42806e-24, 3.86565e-24, 6.52752e-24, 1.30031e-23], [64.3776, 4.42498e-24, 7.61831e-24, 4.42439e-24, 3.86139e-24, 6.5274e-24, 1.3017e-23], [64.524, 4.42445e-24, 7.61735e-24, 4.42074e-24, 3.85717e-24, 6.52726e-24, 1.30307e-23], [64.6708, 4.42399e-24, 7.61641e-24, 4.41712e-24, 3.85299e-24, 6.52711e-24, 1.30444e-23], [64.8179, 4.42358e-24, 7.61546e-24, 4.41352e-24, 3.84884e-24, 6.52694e-24, 1.30581e-23], [64.9653, 4.42324e-24, 7.61453e-24, 4.40994e-24, 3.84474e-24, 6.52675e-24, 1.30717e-23], [65.1131, 4.42296e-24, 7.61359e-24, 4.40638e-24, 3.84067e-24, 6.52655e-24, 1.30852e-23], [65.2612, 4.42274e-24, 7.61267e-24, 4.40285e-24, 3.83663e-24, 6.52633e-24, 1.30986e-23], [65.4096, 4.42257e-24, 7.61175e-24, 4.39933e-24, 3.83263e-24, 6.5261e-24, 1.3112e-23], [65.5584, 4.42247e-24, 7.61083e-24, 4.39584e-24, 3.82867e-24, 6.52585e-24, 1.31254e-23], [65.7075, 4.42243e-24, 7.60992e-24, 4.39237e-24, 3.82474e-24, 6.52558e-24, 1.31387e-23], [65.857, 4.42244e-24, 7.60901e-24, 4.38892e-24, 3.82085e-24, 6.5253e-24, 1.31519e-23], [66.0068, 4.42252e-24, 7.60811e-24, 4.3855e-24, 3.81699e-24, 6.52499e-24, 1.3165e-23], [66.1569, 4.42265e-24, 7.60722e-24, 4.38209e-24, 3.81317e-24, 6.52468e-24, 1.31781e-23], [66.3074, 4.42284e-24, 7.60633e-24, 4.37871e-24, 3.80938e-24, 6.52434e-24, 1.31911e-23], [66.4582, 4.42308e-24, 7.60544e-24, 4.37534e-24, 3.80563e-24, 6.52399e-24, 1.32041e-23], [66.6094, 4.42338e-24, 7.60456e-24, 4.372e-24, 3.8019e-24, 6.52363e-24, 1.3217e-23], [66.7609, 4.42374e-24, 7.60369e-24, 4.36867e-24, 3.79822e-24, 6.52324e-24, 1.32299e-23], [66.9127, 4.42415e-24, 7.60282e-24, 4.36537e-24, 3.79456e-24, 6.52284e-24, 1.32427e-23], [67.0649, 4.42462e-24, 7.60195e-24, 4.36209e-24, 3.79094e-24, 6.52242e-24, 1.32554e-23], [67.2175, 4.42515e-24, 7.60109e-24, 4.35883e-24, 3.78735e-24, 6.52198e-24, 1.32681e-23], [67.3703, 4.42572e-24, 7.60023e-24, 4.35558e-24, 3.78379e-24, 6.52153e-24, 1.32807e-23], [67.5236, 4.42636e-24, 7.59938e-24, 4.35236e-24, 3.78026e-24, 6.52106e-24, 1.32933e-23], [67.6772, 4.42704e-24, 7.59854e-24, 4.34915e-24, 3.77677e-24, 6.52058e-24, 1.33058e-23], [67.8311, 4.42778e-24, 7.5977e-24, 4.34597e-24, 3.7733e-24, 6.52007e-24, 1.33182e-23], [67.9854, 4.42858e-24, 7.59686e-24, 4.3428e-24, 3.76987e-24, 6.51955e-24, 1.33306e-23], [68.14, 4.42942e-24, 7.59603e-24, 4.33966e-24, 3.76646e-24, 6.51901e-24, 1.33429e-23], [68.295, 4.43032e-24, 7.5952e-24, 4.33653e-24, 3.76309e-24, 6.51845e-24, 1.33552e-23], [68.4504, 4.43127e-24, 7.59437e-24, 4.33342e-24, 3.75974e-24, 6.51788e-24, 1.33674e-23], [68.6061, 4.43227e-24, 7.59356e-24, 4.33033e-24, 3.75643e-24, 6.51729e-24, 1.33795e-23], [68.7621, 4.43333e-24, 7.59274e-24, 4.32725e-24, 3.75314e-24, 6.51668e-24, 1.33916e-23], [68.9185, 4.43443e-24, 7.59193e-24, 4.3242e-24, 3.74989e-24, 6.51605e-24, 1.34037e-23], [69.0753, 4.43559e-24, 7.59113e-24, 4.32116e-24, 3.74666e-24, 6.5154e-24, 1.34157e-23], [69.2324, 4.43679e-24, 7.59032e-24, 4.31814e-24, 3.74346e-24, 6.51474e-24, 1.34276e-23], [69.3899, 4.43805e-24, 7.58953e-24, 4.31514e-24, 3.74029e-24, 6.51406e-24, 1.34394e-23], [69.5477, 4.43935e-24, 7.58873e-24, 4.31216e-24, 3.73714e-24, 6.51336e-24, 1.34513e-23], [69.7059, 4.44071e-24, 7.58795e-24, 4.30919e-24, 3.73402e-24, 6.51265e-24, 1.3463e-23], [69.8644, 4.44211e-24, 7.58716e-24, 4.30625e-24, 3.73093e-24, 6.51191e-24, 1.34747e-23], [70.0234, 4.44357e-24, 7.58638e-24, 4.30331e-24, 3.72787e-24, 6.51116e-24, 1.34864e-23], [70.1826, 4.44507e-24, 7.58561e-24, 4.3004e-24, 3.72483e-24, 6.51039e-24, 1.3498e-23], [70.3423, 4.44662e-24, 7.58483e-24, 4.2975e-24, 3.72182e-24, 6.5096e-24, 1.35095e-23], [70.5023, 4.44821e-24, 7.58407e-24, 4.29462e-24, 3.71884e-24, 6.50879e-24, 1.3521e-23], [70.6626, 4.44986e-24, 7.5833e-24, 4.29176e-24, 3.71588e-24, 6.50797e-24, 1.35324e-23], [70.8233, 4.45155e-24, 7.58254e-24, 4.28891e-24, 3.71295e-24, 6.50712e-24, 1.35438e-23], [70.9844, 4.45329e-24, 7.58179e-24, 4.28608e-24, 3.71004e-24, 6.50626e-24, 1.35551e-23], [71.1459, 4.45507e-24, 7.58104e-24, 4.28327e-24, 3.70716e-24, 6.50538e-24, 1.35664e-23], [71.3077, 4.45691e-24, 7.58029e-24, 4.28047e-24, 3.7043e-24, 6.50448e-24, 1.35776e-23], [71.4699, 4.45878e-24, 7.57955e-24, 4.27768e-24, 3.70146e-24, 6.50356e-24, 1.35887e-23], [71.6325, 4.46071e-24, 7.57881e-24, 4.27492e-24, 3.69865e-24, 6.50263e-24, 1.35998e-23], [71.7954, 4.46268e-24, 7.57807e-24, 4.27217e-24, 3.69587e-24, 6.50167e-24, 1.36109e-23], [71.9587, 4.46469e-24, 7.57734e-24, 4.26943e-24, 3.6931e-24, 6.5007e-24, 1.36219e-23], [72.1224, 4.46675e-24, 7.57661e-24, 4.26671e-24, 3.69036e-24, 6.49971e-24, 1.36328e-23], [72.2864, 4.46886e-24, 7.57589e-24, 4.26401e-24, 3.68765e-24, 6.4987e-24, 1.36437e-23], [72.4509, 4.471e-24, 7.57517e-24, 4.26132e-24, 3.68495e-24, 6.49767e-24, 1.36546e-23], [72.6157, 4.4732e-24, 7.57445e-24, 4.25864e-24, 3.68228e-24, 6.49662e-24, 1.36654e-23], [72.7808, 4.47543e-24, 7.57374e-24, 4.25598e-24, 3.67963e-24, 6.49555e-24, 1.36761e-23], [72.9464, 4.47771e-24, 7.57303e-24, 4.25334e-24, 3.67701e-24, 6.49447e-24, 1.36868e-23], [73.1123, 4.48004e-24, 7.57232e-24, 4.25071e-24, 3.6744e-24, 6.49336e-24, 1.36974e-23], [73.2786, 4.4824e-24, 7.57162e-24, 4.24809e-24, 3.67182e-24, 6.49224e-24, 1.3708e-23], [73.4453, 4.48481e-24, 7.57092e-24, 4.24549e-24, 3.66926e-24, 6.4911e-24, 1.37185e-23], [73.6123, 4.48727e-24, 7.57023e-24, 4.24291e-24, 3.66672e-24, 6.48994e-24, 1.3729e-23], [73.7798, 4.48976e-24, 7.56954e-24, 4.24033e-24, 3.6642e-24, 6.48875e-24, 1.37394e-23], [73.9476, 4.4923e-24, 7.56885e-24, 4.23778e-24, 3.6617e-24, 6.48755e-24, 1.37498e-23], [74.1158, 4.49487e-24, 7.56816e-24, 4.23523e-24, 3.65922e-24, 6.48634e-24, 1.37601e-23], [74.2844, 4.49749e-24, 7.56748e-24, 4.2327e-24, 3.65676e-24, 6.4851e-24, 1.37704e-23], [74.4533, 4.50016e-24, 7.56681e-24, 4.23019e-24, 3.65432e-24, 6.48384e-24, 1.37806e-23], [74.6227, 4.50286e-24, 7.56613e-24, 4.22768e-24, 3.6519e-24, 6.48256e-24, 1.37908e-23], [74.7924, 4.5056e-24, 7.56546e-24, 4.22519e-24, 3.6495e-24, 6.48127e-24, 1.38009e-23], [74.9625, 4.50839e-24, 7.56479e-24, 4.22272e-24, 3.64712e-24, 6.47995e-24, 1.3811e-23], [75.133, 4.51121e-24, 7.56413e-24, 4.22026e-24, 3.64476e-24, 6.47861e-24, 1.3821e-23], [75.3039, 4.51408e-24, 7.56347e-24, 4.21781e-24, 3.64242e-24, 6.47726e-24, 1.3831e-23], [75.4752, 4.51698e-24, 7.56281e-24, 4.21537e-24, 3.64009e-24, 6.47589e-24, 1.38409e-23], [75.6469, 4.51993e-24, 7.56216e-24, 4.21295e-24, 3.63779e-24, 6.47449e-24, 1.38508e-23], [75.819, 4.52291e-24, 7.56151e-24, 4.21054e-24, 3.6355e-24, 6.47308e-24, 1.38606e-23], [75.9914, 4.52594e-24, 7.56086e-24, 4.20814e-24, 3.63323e-24, 6.47165e-24, 1.38704e-23], [76.1643, 4.529e-24, 7.56022e-24, 4.20576e-24, 3.63098e-24, 6.47019e-24, 1.38802e-23], [76.3375, 4.53211e-24, 7.55958e-24, 4.20339e-24, 3.62874e-24, 6.46872e-24, 1.38898e-23], [76.5111, 4.53525e-24, 7.55894e-24, 4.20103e-24, 3.62653e-24, 6.46723e-24, 1.38995e-23], [76.6852, 4.53843e-24, 7.55831e-24, 4.19868e-24, 3.62433e-24, 6.46572e-24, 1.39091e-23], [76.8596, 4.54165e-24, 7.55768e-24, 4.19635e-24, 3.62214e-24, 6.46418e-24, 1.39186e-23], [77.0344, 4.5449e-24, 7.55705e-24, 4.19402e-24, 3.61998e-24, 6.46263e-24, 1.39281e-23], [77.2096, 4.5482e-24, 7.55643e-24, 4.19171e-24, 3.61783e-24, 6.46106e-24, 1.39375e-23], [77.3852, 4.55153e-24, 7.55581e-24, 4.18942e-24, 3.61569e-24, 6.45947e-24, 1.39469e-23], [77.5613, 4.5549e-24, 7.55519e-24, 4.18713e-24, 3.61358e-24, 6.45786e-24, 1.39563e-23], [77.7377, 4.55831e-24, 7.55457e-24, 4.18486e-24, 3.61148e-24, 6.45623e-24, 1.39656e-23], [77.9145, 4.56176e-24, 7.55396e-24, 4.18259e-24, 3.60939e-24, 6.45458e-24, 1.39748e-23], [78.0917, 4.56524e-24, 7.55335e-24, 4.18034e-24, 3.60732e-24, 6.4529e-24, 1.39841e-23], [78.2694, 4.56876e-24, 7.55275e-24, 4.1781e-24, 3.60526e-24, 6.45121e-24, 1.39932e-23], [78.4474, 4.57232e-24, 7.55215e-24, 4.17588e-24, 3.60322e-24, 6.4495e-24, 1.40023e-23], [78.6258, 4.57591e-24, 7.55155e-24, 4.17366e-24, 3.6012e-24, 6.44777e-24, 1.40114e-23], [78.8047, 4.57954e-24, 7.55095e-24, 4.17145e-24, 3.59919e-24, 6.44602e-24, 1.40204e-23], [78.9839, 4.58321e-24, 7.55036e-24, 4.16926e-24, 3.59719e-24, 6.44424e-24, 1.40294e-23], [79.1636, 4.58691e-24, 7.54976e-24, 4.16708e-24, 3.59521e-24, 6.44245e-24, 1.40383e-23], [79.3436, 4.59065e-24, 7.54918e-24, 4.16491e-24, 3.59325e-24, 6.44064e-24, 1.40472e-23], [79.5241, 4.59443e-24, 7.54859e-24, 4.16275e-24, 3.59129e-24, 6.43881e-24, 1.40561e-23], [79.705, 4.59824e-24, 7.54801e-24, 4.1606e-24, 3.58935e-24, 6.43695e-24, 1.40649e-23], [79.8863, 4.60208e-24, 7.54743e-24, 4.15846e-24, 3.58743e-24, 6.43508e-24, 1.40736e-23], [80.068, 4.60596e-24, 7.54686e-24, 4.15633e-24, 3.58552e-24, 6.43318e-24, 1.40823e-23], [80.2501, 4.60988e-24, 7.54628e-24, 4.15421e-24, 3.58362e-24, 6.43127e-24, 1.4091e-23], [80.4326, 4.61383e-24, 7.54571e-24, 4.1521e-24, 3.58173e-24, 6.42933e-24, 1.40996e-23], [80.6156, 4.61782e-24, 7.54515e-24, 4.15e-24, 3.57986e-24, 6.42738e-24, 1.41082e-23], [80.7989, 4.62184e-24, 7.54458e-24, 4.14792e-24, 3.578e-24, 6.4254e-24, 1.41167e-23], [80.9827, 4.6259e-24, 7.54402e-24, 4.14584e-24, 3.57615e-24, 6.4234e-24, 1.41252e-23], [81.1669, 4.62999e-24, 7.54346e-24, 4.14377e-24, 3.57432e-24, 6.42138e-24, 1.41337e-23], [81.3516, 4.63412e-24, 7.54291e-24, 4.14172e-24, 3.5725e-24, 6.41934e-24, 1.41421e-23], [81.5366, 4.63828e-24, 7.54235e-24, 4.13967e-24, 3.57069e-24, 6.41728e-24, 1.41504e-23], [81.7221, 4.64247e-24, 7.5418e-24, 4.13764e-24, 3.56889e-24, 6.4152e-24, 1.41587e-23], [81.9079, 4.6467e-24, 7.54125e-24, 4.13561e-24, 3.5671e-24, 6.4131e-24, 1.4167e-23], [82.0942, 4.65096e-24, 7.54071e-24, 4.13359e-24, 3.56533e-24, 6.41098e-24, 1.41752e-23], [82.281, 4.65526e-24, 7.54017e-24, 4.13159e-24, 3.56356e-24, 6.40884e-24, 1.41834e-23], [82.4681, 4.65959e-24, 7.53963e-24, 4.12959e-24, 3.56181e-24, 6.40667e-24, 1.41916e-23], [82.6557, 4.66395e-24, 7.53909e-24, 4.1276e-24, 3.56007e-24, 6.40449e-24, 1.41997e-23], [82.8437, 4.66835e-24, 7.53856e-24, 4.12562e-24, 3.55834e-24, 6.40228e-24, 1.42077e-23], [83.0321, 4.67278e-24, 7.53803e-24, 4.12365e-24, 3.55662e-24, 6.40006e-24, 1.42157e-23], [83.221, 4.67724e-24, 7.5375e-24, 4.12169e-24, 3.55491e-24, 6.39781e-24, 1.42237e-23], [83.4103, 4.68174e-24, 7.53697e-24, 4.11974e-24, 3.55322e-24, 6.39554e-24, 1.42316e-23], [83.6, 4.68627e-24, 7.53645e-24, 4.1178e-24, 3.55153e-24, 6.39325e-24, 1.42395e-23], [83.7902, 4.69083e-24, 7.53593e-24, 4.11587e-24, 3.54985e-24, 6.39094e-24, 1.42474e-23], [83.9808, 4.69543e-24, 7.53541e-24, 4.11395e-24, 3.54818e-24, 6.38861e-24, 1.42552e-23], [84.1718, 4.70005e-24, 7.5349e-24, 4.11203e-24, 3.54653e-24, 6.38626e-24, 1.42629e-23], [84.3632, 4.70471e-24, 7.53438e-24, 4.11013e-24, 3.54488e-24, 6.38388e-24, 1.42707e-23], [84.5551, 4.70941e-24, 7.53387e-24, 4.10823e-24, 3.54324e-24, 6.38149e-24, 1.42783e-23], [84.7475, 4.71413e-24, 7.53337e-24, 4.10635e-24, 3.54162e-24, 6.37907e-24, 1.4286e-23], [84.9402, 4.71889e-24, 7.53286e-24, 4.10447e-24, 3.54e-24, 6.37663e-24, 1.42936e-23], [85.1334, 4.72368e-24, 7.53236e-24, 4.1026e-24, 3.53839e-24, 6.37417e-24, 1.43011e-23], [85.3271, 4.7285e-24, 7.53186e-24, 4.10073e-24, 3.53679e-24, 6.37169e-24, 1.43087e-23], [85.5211, 4.73336e-24, 7.53136e-24, 4.09888e-24, 3.5352e-24, 6.36919e-24, 1.43161e-23], [85.7157, 4.73825e-24, 7.53087e-24, 4.09704e-24, 3.53362e-24, 6.36666e-24, 1.43236e-23], [85.9106, 4.74316e-24, 7.53038e-24, 4.0952e-24, 3.53204e-24, 6.36412e-24, 1.4331e-23], [86.106, 4.74811e-24, 7.52989e-24, 4.09337e-24, 3.53048e-24, 6.36155e-24, 1.43383e-23], [86.3019, 4.7531e-24, 7.5294e-24, 4.09155e-24, 3.52892e-24, 6.35896e-24, 1.43457e-23], [86.4982, 4.75811e-24, 7.52892e-24, 4.08974e-24, 3.52738e-24, 6.35636e-24, 1.43529e-23], [86.6949, 4.76315e-24, 7.52844e-24, 4.08794e-24, 3.52584e-24, 6.35372e-24, 1.43602e-23], [86.8921, 4.76823e-24, 7.52796e-24, 4.08614e-24, 3.52431e-24, 6.35107e-24, 1.43674e-23], [87.0898, 4.77334e-24, 7.52748e-24, 4.08436e-24, 3.52278e-24, 6.3484e-24, 1.43745e-23], [87.2879, 4.77848e-24, 7.52701e-24, 4.08258e-24, 3.52127e-24, 6.3457e-24, 1.43817e-23], [87.4864, 4.78365e-24, 7.52654e-24, 4.08081e-24, 3.51976e-24, 6.34298e-24, 1.43888e-23], [87.6854, 4.78885e-24, 7.52607e-24, 4.07904e-24, 3.51826e-24, 6.34024e-24, 1.43958e-23], [87.8849, 4.79408e-24, 7.5256e-24, 4.07729e-24, 3.51677e-24, 6.33748e-24, 1.44028e-23], [88.0848, 4.79935e-24, 7.52514e-24, 4.07554e-24, 3.51528e-24, 6.3347e-24, 1.44098e-23], [88.2851, 4.80464e-24, 7.52468e-24, 4.0738e-24, 3.5138e-24, 6.3319e-24, 1.44167e-23], [88.4859, 4.80997e-24, 7.52422e-24, 4.07207e-24, 3.51233e-24, 6.32907e-24, 1.44236e-23], [88.6872, 4.81532e-24, 7.52376e-24, 4.07034e-24, 3.51087e-24, 6.32622e-24, 1.44304e-23], [88.8889, 4.82071e-24, 7.52331e-24, 4.06862e-24, 3.50941e-24, 6.32335e-24, 1.44373e-23], [89.0911, 4.82613e-24, 7.52286e-24, 4.06691e-24, 3.50796e-24, 6.32046e-24, 1.4444e-23], [89.2937, 4.83158e-24, 7.52241e-24, 4.06521e-24, 3.50652e-24, 6.31754e-24, 1.44508e-23], [89.4968, 4.83706e-24, 7.52196e-24, 4.06351e-24, 3.50508e-24, 6.31461e-24, 1.44575e-23], [89.7004, 4.84257e-24, 7.52152e-24, 4.06183e-24, 3.50365e-24, 6.31165e-24, 1.44641e-23], [89.9044, 4.84811e-24, 7.52107e-24, 4.06014e-24, 3.50223e-24, 6.30867e-24, 1.44708e-23], [90.1089, 4.85368e-24, 7.52063e-24, 4.05847e-24, 3.50081e-24, 6.30567e-24, 1.44774e-23], [90.3139, 4.85928e-24, 7.5202e-24, 4.0568e-24, 3.4994e-24, 6.30264e-24, 1.44839e-23], [90.5193, 4.86491e-24, 7.51976e-24, 4.05514e-24, 3.49799e-24, 6.2996e-24, 1.44904e-23], [90.7252, 4.87057e-24, 7.51933e-24, 4.05349e-24, 3.49659e-24, 6.29653e-24, 1.44969e-23], [90.9316, 4.87626e-24, 7.5189e-24, 4.05185e-24, 3.4952e-24, 6.29344e-24, 1.45034e-23], [91.1384, 4.88198e-24, 7.51847e-24, 4.05021e-24, 3.49381e-24, 6.29033e-24, 1.45098e-23], [91.3457, 4.88774e-24, 7.51805e-24, 4.04857e-24, 3.49243e-24, 6.28719e-24, 1.45161e-23], [91.5535, 4.89352e-24, 7.51763e-24, 4.04695e-24, 3.49105e-24, 6.28404e-24, 1.45225e-23], [91.7617, 4.89933e-24, 7.51721e-24, 4.04533e-24, 3.48968e-24, 6.28086e-24, 1.45288e-23], [91.9704, 4.90517e-24, 7.51679e-24, 4.04372e-24, 3.48831e-24, 6.27766e-24, 1.4535e-23], [92.1796, 4.91104e-24, 7.51637e-24, 4.04211e-24, 3.48695e-24, 6.27443e-24, 1.45412e-23], [92.3893, 4.91695e-24, 7.51596e-24, 4.04052e-24, 3.48559e-24, 6.27119e-24, 1.45474e-23], [92.5994, 4.92288e-24, 7.51555e-24, 4.03892e-24, 3.48424e-24, 6.26792e-24, 1.45536e-23], [92.8101, 4.92884e-24, 7.51514e-24, 4.03734e-24, 3.48289e-24, 6.26463e-24, 1.45597e-23], [93.0212, 4.93483e-24, 7.51474e-24, 4.03576e-24, 3.48155e-24, 6.26131e-24, 1.45658e-23], [93.2328, 4.94085e-24, 7.51433e-24, 4.03419e-24, 3.48022e-24, 6.25798e-24, 1.45718e-23], [93.4448, 4.9469e-24, 7.51393e-24, 4.03262e-24, 3.47888e-24, 6.25462e-24, 1.45779e-23], [93.6574, 4.95299e-24, 7.51353e-24, 4.03106e-24, 3.47756e-24, 6.25124e-24, 1.45838e-23], [93.8704, 4.9591e-24, 7.51314e-24, 4.02951e-24, 3.47623e-24, 6.24784e-24, 1.45898e-23], [94.0839, 4.96524e-24, 7.51274e-24, 4.02796e-24, 3.47491e-24, 6.24441e-24, 1.45957e-23], [94.2979, 4.97141e-24, 7.51235e-24, 4.02642e-24, 3.4736e-24, 6.24096e-24, 1.46016e-23], [94.5124, 4.9776e-24, 7.51196e-24, 4.02489e-24, 3.47229e-24, 6.23749e-24, 1.46074e-23], [94.7274, 4.98383e-24, 7.51157e-24, 4.02336e-24, 3.47098e-24, 6.234e-24, 1.46132e-23], [94.9428, 4.99009e-24, 7.51119e-24, 4.02184e-24, 3.46968e-24, 6.23049e-24, 1.4619e-23], [95.1588, 4.99638e-24, 7.51081e-24, 4.02032e-24, 3.46838e-24, 6.22695e-24, 1.46247e-23], [95.3752, 5.0027e-24, 7.51043e-24, 4.01881e-24, 3.46709e-24, 6.22339e-24, 1.46304e-23], [95.5922, 5.00904e-24, 7.51005e-24, 4.0173e-24, 3.4658e-24, 6.2198e-24, 1.46361e-23], [95.8096, 5.01542e-24, 7.50967e-24, 4.01581e-24, 3.46451e-24, 6.2162e-24, 1.46417e-23], [96.0275, 5.02183e-24, 7.5093e-24, 4.01431e-24, 3.46322e-24, 6.21257e-24, 1.46473e-23], [96.246, 5.02826e-24, 7.50893e-24, 4.01283e-24, 3.46194e-24, 6.20892e-24, 1.46529e-23], [96.4649, 5.03473e-24, 7.50856e-24, 4.01135e-24, 3.46067e-24, 6.20524e-24, 1.46584e-23], [96.6843, 5.04122e-24, 7.5082e-24, 4.00987e-24, 3.45939e-24, 6.20155e-24, 1.46639e-23], [96.9042, 5.04774e-24, 7.50783e-24, 4.0084e-24, 3.45812e-24, 6.19783e-24, 1.46694e-23], [97.1246, 5.0543e-24, 7.50747e-24, 4.00694e-24, 3.45686e-24, 6.19408e-24, 1.46748e-23], [97.3455, 5.06088e-24, 7.50711e-24, 4.00548e-24, 3.45559e-24, 6.19032e-24, 1.46802e-23], [97.567, 5.06749e-24, 7.50676e-24, 4.00403e-24, 3.45433e-24, 6.18653e-24, 1.46856e-23], [97.7889, 5.07413e-24, 7.5064e-24, 4.00258e-24, 3.45307e-24, 6.18272e-24, 1.46909e-23], [98.0113, 5.0808e-24, 7.50605e-24, 4.00114e-24, 3.45182e-24, 6.17889e-24, 1.46962e-23], [98.2342, 5.0875e-24, 7.5057e-24, 3.9997e-24, 3.45057e-24, 6.17503e-24, 1.47015e-23], [98.4577, 5.09423e-24, 7.50535e-24, 3.99827e-24, 3.44932e-24, 6.17115e-24, 1.47067e-23], [98.6816, 5.10098e-24, 7.50501e-24, 3.99685e-24, 3.44807e-24, 6.16725e-24, 1.47119e-23], [98.9061, 5.10777e-24, 7.50467e-24, 3.99543e-24, 3.44683e-24, 6.16332e-24, 1.47171e-23], [99.1311, 5.11459e-24, 7.50433e-24, 3.99402e-24, 3.44558e-24, 6.15937e-24, 1.47222e-23], [99.3565, 5.12143e-24, 7.50399e-24, 3.99261e-24, 3.44434e-24, 6.1554e-24, 1.47273e-23], [99.5825, 5.12831e-24, 7.50365e-24, 3.9912e-24, 3.44311e-24, 6.15141e-24, 1.47324e-23], [99.809, 5.13521e-24, 7.50332e-24, 3.98981e-24, 3.44187e-24, 6.14739e-24, 1.47374e-23], [100.036, 5.14214e-24, 7.50299e-24, 3.98841e-24, 3.44064e-24, 6.14335e-24, 1.47424e-23], [100.264, 5.1491e-24, 7.50266e-24, 3.98702e-24, 3.43941e-24, 6.13929e-24, 1.47474e-23], [100.492, 5.15609e-24, 7.50233e-24, 3.98564e-24, 3.43818e-24, 6.1352e-24, 1.47524e-23], [100.72, 5.16312e-24, 7.50201e-24, 3.98427e-24, 3.43696e-24, 6.13109e-24, 1.47573e-23], [100.949, 5.17016e-24, 7.50169e-24, 3.98289e-24, 3.43573e-24, 6.12696e-24, 1.47622e-23], [101.179, 5.17724e-24, 7.50137e-24, 3.98153e-24, 3.43451e-24, 6.1228e-24, 1.4767e-23], [101.409, 5.18435e-24, 7.50105e-24, 3.98016e-24, 3.43329e-24, 6.11863e-24, 1.47718e-23], [101.64, 5.19149e-24, 7.50074e-24, 3.97881e-24, 3.43207e-24, 6.11442e-24, 1.47766e-23], [101.871, 5.19866e-24, 7.50043e-24, 3.97746e-24, 3.43085e-24, 6.1102e-24, 1.47814e-23], [102.103, 5.20585e-24, 7.50012e-24, 3.97611e-24, 3.42964e-24, 6.10595e-24, 1.47861e-23], [102.335, 5.21308e-24, 7.49981e-24, 3.97477e-24, 3.42842e-24, 6.10168e-24, 1.47908e-23], [102.568, 5.22033e-24, 7.4995e-24, 3.97343e-24, 3.42721e-24, 6.09738e-24, 1.47955e-23], [102.801, 5.22761e-24, 7.4992e-24, 3.9721e-24, 3.426e-24, 6.09307e-24, 1.48001e-23], [103.035, 5.23493e-24, 7.4989e-24, 3.97077e-24, 3.42479e-24, 6.08873e-24, 1.48047e-23], [103.269, 5.24227e-24, 7.4986e-24, 3.96945e-24, 3.42358e-24, 6.08436e-24, 1.48093e-23], [103.504, 5.24964e-24, 7.49831e-24, 3.96813e-24, 3.42237e-24, 6.07998e-24, 1.48138e-23], [103.739, 5.25704e-24, 7.49801e-24, 3.96682e-24, 3.42117e-24, 6.07557e-24, 1.48183e-23], [103.975, 5.26447e-24, 7.49772e-24, 3.96551e-24, 3.41996e-24, 6.07113e-24, 1.48228e-23], [104.212, 5.27193e-24, 7.49743e-24, 3.9642e-24, 3.41876e-24, 6.06668e-24, 1.48273e-23], [104.449, 5.27942e-24, 7.49715e-24, 3.9629e-24, 3.41756e-24, 6.0622e-24, 1.48317e-23], [104.687, 5.28693e-24, 7.49686e-24, 3.96161e-24, 3.41636e-24, 6.05769e-24, 1.48361e-23], [104.925, 5.29448e-24, 7.49658e-24, 3.96032e-24, 3.41515e-24, 6.05317e-24, 1.48405e-23], [105.163, 5.30205e-24, 7.4963e-24, 3.95904e-24, 3.41396e-24, 6.04862e-24, 1.48448e-23], [105.403, 5.30966e-24, 7.49602e-24, 3.95776e-24, 3.41276e-24, 6.04404e-24, 1.48491e-23], [105.642, 5.31729e-24, 7.49575e-24, 3.95648e-24, 3.41156e-24, 6.03945e-24, 1.48534e-23], [105.883, 5.32496e-24, 7.49548e-24, 3.95521e-24, 3.41036e-24, 6.03483e-24, 1.48576e-23], [106.123, 5.33265e-24, 7.49521e-24, 3.95394e-24, 3.40916e-24, 6.03019e-24, 1.48618e-23], [106.365, 5.34037e-24, 7.49494e-24, 3.95268e-24, 3.40797e-24, 6.02552e-24, 1.4866e-23], [106.607, 5.34813e-24, 7.49467e-24, 3.95142e-24, 3.40677e-24, 6.02083e-24, 1.48702e-23], [106.849, 5.35591e-24, 7.49441e-24, 3.95017e-24, 3.40558e-24, 6.01612e-24, 1.48743e-23], [107.092, 5.36372e-24, 7.49415e-24, 3.94892e-24, 3.40438e-24, 6.01138e-24, 1.48784e-23], [107.336, 5.37156e-24, 7.49389e-24, 3.94767e-24, 3.40319e-24, 6.00662e-24, 1.48825e-23], [107.58, 5.37943e-24, 7.49364e-24, 3.94643e-24, 3.40199e-24, 6.00184e-24, 1.48865e-23], [107.825, 5.38733e-24, 7.49338e-24, 3.9452e-24, 3.4008e-24, 5.99704e-24, 1.48905e-23], [108.07, 5.39526e-24, 7.49313e-24, 3.94397e-24, 3.39961e-24, 5.99221e-24, 1.48945e-23], [108.316, 5.40322e-24, 7.49288e-24, 3.94274e-24, 3.39841e-24, 5.98736e-24, 1.48984e-23], [108.562, 5.4112e-24, 7.49264e-24, 3.94152e-24, 3.39722e-24, 5.98248e-24, 1.49024e-23], [108.809, 5.41922e-24, 7.49239e-24, 3.9403e-24, 3.39603e-24, 5.97758e-24, 1.49063e-23], [109.057, 5.42727e-24, 7.49215e-24, 3.93908e-24, 3.39483e-24, 5.97266e-24, 1.49101e-23], [109.305, 5.43535e-24, 7.49191e-24, 3.93787e-24, 3.39364e-24, 5.96772e-24, 1.4914e-23], [109.553, 5.44345e-24, 7.49168e-24, 3.93667e-24, 3.39245e-24, 5.96275e-24, 1.49178e-23], [109.802, 5.45159e-24, 7.49144e-24, 3.93546e-24, 3.39126e-24, 5.95776e-24, 1.49216e-23], [110.052, 5.45975e-24, 7.49121e-24, 3.93427e-24, 3.39006e-24, 5.95274e-24, 1.49253e-23], [110.302, 5.46795e-24, 7.49098e-24, 3.93307e-24, 3.38887e-24, 5.9477e-24, 1.49291e-23], [110.553, 5.47617e-24, 7.49076e-24, 3.93188e-24, 3.38768e-24, 5.94264e-24, 1.49328e-23], [110.805, 5.48443e-24, 7.49053e-24, 3.9307e-24, 3.38648e-24, 5.93756e-24, 1.49364e-23], [111.057, 5.49271e-24, 7.49031e-24, 3.92951e-24, 3.38529e-24, 5.93245e-24, 1.49401e-23], [111.309, 5.50103e-24, 7.49009e-24, 3.92834e-24, 3.38409e-24, 5.92732e-24, 1.49437e-23], [111.563, 5.50937e-24, 7.48987e-24, 3.92716e-24, 3.3829e-24, 5.92217e-24, 1.49473e-23], [111.816, 5.51775e-24, 7.48966e-24, 3.92599e-24, 3.3817e-24, 5.91699e-24, 1.49508e-23], [112.071, 5.52615e-24, 7.48945e-24, 3.92483e-24, 3.38051e-24, 5.91179e-24, 1.49544e-23], [112.326, 5.53459e-24, 7.48924e-24, 3.92367e-24, 3.37931e-24, 5.90656e-24, 1.49579e-23], [112.581, 5.54305e-24, 7.48903e-24, 3.92251e-24, 3.37812e-24, 5.90132e-24, 1.49614e-23], [112.837, 5.55154e-24, 7.48882e-24, 3.92136e-24, 3.37692e-24, 5.89605e-24, 1.49648e-23], [113.094, 5.56007e-24, 7.48862e-24, 3.92021e-24, 3.37572e-24, 5.89076e-24, 1.49682e-23], [113.351, 5.56862e-24, 7.48842e-24, 3.91906e-24, 3.37452e-24, 5.88544e-24, 1.49716e-23], [113.609, 5.57721e-24, 7.48822e-24, 3.91792e-24, 3.37332e-24, 5.8801e-24, 1.4975e-23], [113.867, 5.58582e-24, 7.48803e-24, 3.91678e-24, 3.37213e-24, 5.87474e-24, 1.49784e-23], [114.126, 5.59446e-24, 7.48784e-24, 3.91565e-24, 3.37092e-24, 5.86935e-24, 1.49817e-23], [114.386, 5.60314e-24, 7.48765e-24, 3.91452e-24, 3.36972e-24, 5.86395e-24, 1.4985e-23], [114.646, 5.61184e-24, 7.48746e-24, 3.91339e-24, 3.36852e-24, 5.85852e-24, 1.49882e-23], [114.907, 5.62058e-24, 7.48727e-24, 3.91227e-24, 3.36732e-24, 5.85306e-24, 1.49915e-23], [115.168, 5.62934e-24, 7.48709e-24, 3.91115e-24, 3.36611e-24, 5.84759e-24, 1.49947e-23], [115.43, 5.63814e-24, 7.48691e-24, 3.91003e-24, 3.36491e-24, 5.84209e-24, 1.49979e-23], [115.693, 5.64696e-24, 7.48673e-24, 3.90892e-24, 3.3637e-24, 5.83656e-24, 1.5001e-23], [115.956, 5.65582e-24, 7.48656e-24, 3.90781e-24, 3.3625e-24, 5.83102e-24, 1.50041e-23], [116.22, 5.66471e-24, 7.48639e-24, 3.90671e-24, 3.36129e-24, 5.82545e-24, 1.50072e-23], [116.484, 5.67362e-24, 7.48622e-24, 3.90561e-24, 3.36008e-24, 5.81986e-24, 1.50103e-23], [116.749, 5.68257e-24, 7.48605e-24, 3.90451e-24, 3.35887e-24, 5.81425e-24, 1.50134e-23], [117.014, 5.69155e-24, 7.48588e-24, 3.90342e-24, 3.35766e-24, 5.80861e-24, 1.50164e-23], [117.281, 5.70056e-24, 7.48572e-24, 3.90233e-24, 3.35645e-24, 5.80295e-24, 1.50194e-23], [117.547, 5.7096e-24, 7.48556e-24, 3.90125e-24, 3.35523e-24, 5.79727e-24, 1.50224e-23], [117.815, 5.71867e-24, 7.48541e-24, 3.90016e-24, 3.35402e-24, 5.79157e-24, 1.50253e-23], [118.083, 5.72777e-24, 7.48525e-24, 3.89909e-24, 3.3528e-24, 5.78584e-24, 1.50282e-23], [118.351, 5.7369e-24, 7.4851e-24, 3.89801e-24, 3.35158e-24, 5.78009e-24, 1.50311e-23], [118.621, 5.74606e-24, 7.48495e-24, 3.89694e-24, 3.35037e-24, 5.77432e-24, 1.5034e-23], [118.89, 5.75525e-24, 7.4848e-24, 3.89587e-24, 3.34915e-24, 5.76852e-24, 1.50369e-23], [119.161, 5.76448e-24, 7.48466e-24, 3.89481e-24, 3.34792e-24, 5.76271e-24, 1.50397e-23], [119.432, 5.77373e-24, 7.48452e-24, 3.89375e-24, 3.3467e-24, 5.75687e-24, 1.50425e-23], [119.703, 5.78301e-24, 7.48438e-24, 3.89269e-24, 3.34548e-24, 5.75101e-24, 1.50452e-23], [119.976, 5.79233e-24, 7.48424e-24, 3.89164e-24, 3.34425e-24, 5.74512e-24, 1.5048e-23], [120.249, 5.80168e-24, 7.48411e-24, 3.89059e-24, 3.34302e-24, 5.73922e-24, 1.50507e-23], [120.522, 5.81105e-24, 7.48398e-24, 3.88954e-24, 3.34179e-24, 5.73329e-24, 1.50534e-23], [120.796, 5.82046e-24, 7.48385e-24, 3.8885e-24, 3.34056e-24, 5.72734e-24, 1.50561e-23], [121.071, 5.8299e-24, 7.48372e-24, 3.88746e-24, 3.33933e-24, 5.72137e-24, 1.50587e-23], [121.346, 5.83937e-24, 7.4836e-24, 3.88642e-24, 3.3381e-24, 5.71537e-24, 1.50613e-23], [121.622, 5.84888e-24, 7.48348e-24, 3.88539e-24, 3.33686e-24, 5.70936e-24, 1.50639e-23], [121.899, 5.85841e-24, 7.48336e-24, 3.88436e-24, 3.33563e-24, 5.70332e-24, 1.50665e-23], [122.176, 5.86798e-24, 7.48325e-24, 3.88334e-24, 3.33439e-24, 5.69726e-24, 1.5069e-23], [122.454, 5.87757e-24, 7.48313e-24, 3.88231e-24, 3.33315e-24, 5.69118e-24, 1.50715e-23], [122.733, 5.8872e-24, 7.48302e-24, 3.88129e-24, 3.3319e-24, 5.68507e-24, 1.5074e-23], [123.012, 5.89686e-24, 7.48292e-24, 3.88028e-24, 3.33066e-24, 5.67895e-24, 1.50765e-23], [123.292, 5.90655e-24, 7.48281e-24, 3.87927e-24, 3.32941e-24, 5.6728e-24, 1.50789e-23], [123.572, 5.91627e-24, 7.48271e-24, 3.87826e-24, 3.32817e-24, 5.66663e-24, 1.50813e-23], [123.853, 5.92602e-24, 7.48261e-24, 3.87725e-24, 3.32692e-24, 5.66044e-24, 1.50837e-23], [124.135, 5.93581e-24, 7.48252e-24, 3.87625e-24, 3.32567e-24, 5.65423e-24, 1.50861e-23], [124.417, 5.94562e-24, 7.48242e-24, 3.87525e-24, 3.32441e-24, 5.648e-24, 1.50884e-23], [124.7, 5.95547e-24, 7.48233e-24, 3.87426e-24, 3.32316e-24, 5.64174e-24, 1.50908e-23], [124.984, 5.96535e-24, 7.48224e-24, 3.87327e-24, 3.3219e-24, 5.63547e-24, 1.50931e-23], [125.268, 5.97526e-24, 7.48216e-24, 3.87228e-24, 3.32064e-24, 5.62917e-24, 1.50953e-23], [125.553, 5.98521e-24, 7.48207e-24, 3.87129e-24, 3.31938e-24, 5.62286e-24, 1.50976e-23], [125.839, 5.99518e-24, 7.48199e-24, 3.87031e-24, 3.31812e-24, 5.61652e-24, 1.50998e-23], [126.125, 6.00519e-24, 7.48192e-24, 3.86933e-24, 3.31685e-24, 5.61016e-24, 1.5102e-23], [126.412, 6.01523e-24, 7.48184e-24, 3.86836e-24, 3.31558e-24, 5.60378e-24, 1.51042e-23], [126.699, 6.0253e-24, 7.48177e-24, 3.86739e-24, 3.31431e-24, 5.59738e-24, 1.51063e-23], [126.988, 6.0354e-24, 7.4817e-24, 3.86642e-24, 3.31304e-24, 5.59096e-24, 1.51084e-23], [127.276, 6.04554e-24, 7.48164e-24, 3.86545e-24, 3.31177e-24, 5.58452e-24, 1.51105e-23], [127.566, 6.0557e-24, 7.48157e-24, 3.86449e-24, 3.31049e-24, 5.57806e-24, 1.51126e-23], [127.856, 6.0659e-24, 7.48151e-24, 3.86353e-24, 3.30922e-24, 5.57157e-24, 1.51147e-23], [128.147, 6.07613e-24, 7.48146e-24, 3.86257e-24, 3.30794e-24, 5.56507e-24, 1.51167e-23], [128.438, 6.0864e-24, 7.4814e-24, 3.86162e-24, 3.30665e-24, 5.55855e-24, 1.51187e-23], [128.731, 6.09669e-24, 7.48135e-24, 3.86067e-24, 3.30537e-24, 5.55201e-24, 1.51207e-23], [129.023, 6.10702e-24, 7.4813e-24, 3.85973e-24, 3.30408e-24, 5.54544e-24, 1.51226e-23], [129.317, 6.11738e-24, 7.48125e-24, 3.85878e-24, 3.30279e-24, 5.53886e-24, 1.51246e-23], [129.611, 6.12778e-24, 7.48121e-24, 3.85784e-24, 3.3015e-24, 5.53226e-24, 1.51265e-23], [129.906, 6.1382e-24, 7.48117e-24, 3.85691e-24, 3.30021e-24, 5.52564e-24, 1.51283e-23], [130.201, 6.14866e-24, 7.48113e-24, 3.85597e-24, 3.29891e-24, 5.519e-24, 1.51302e-23], [130.497, 6.15915e-24, 7.4811e-24, 3.85504e-24, 3.29761e-24, 5.51234e-24, 1.5132e-23], [130.794, 6.16967e-24, 7.48107e-24, 3.85412e-24, 3.29631e-24, 5.50566e-24, 1.51339e-23], [131.092, 6.18023e-24, 7.48104e-24, 3.85319e-24, 3.29501e-24, 5.49896e-24, 1.51357e-23], [131.39, 6.19082e-24, 7.48101e-24, 3.85227e-24, 3.2937e-24, 5.49224e-24, 1.51374e-23], [131.689, 6.20144e-24, 7.48099e-24, 3.85135e-24, 3.2924e-24, 5.4855e-24, 1.51392e-23], [131.988, 6.2121e-24, 7.48097e-24, 3.85044e-24, 3.29109e-24, 5.47875e-24, 1.51409e-23], [132.288, 6.22279e-24, 7.48095e-24, 3.84953e-24, 3.28977e-24, 5.47197e-24, 1.51426e-23], [132.589, 6.23351e-24, 7.48094e-24, 3.84862e-24, 3.28846e-24, 5.46518e-24, 1.51443e-23], [132.891, 6.24426e-24, 7.48092e-24, 3.84772e-24, 3.28714e-24, 5.45837e-24, 1.51459e-23], [133.193, 6.25505e-24, 7.48092e-24, 3.84681e-24, 3.28582e-24, 5.45154e-24, 1.51475e-23], [133.496, 6.26587e-24, 7.48091e-24, 3.84591e-24, 3.2845e-24, 5.4447e-24, 1.51491e-23], [133.8, 6.27672e-24, 7.48091e-24, 3.84502e-24, 3.28317e-24, 5.43783e-24, 1.51507e-23], [134.104, 6.28761e-24, 7.48091e-24, 3.84413e-24, 3.28185e-24, 5.43095e-24, 1.51523e-23], [134.409, 6.29853e-24, 7.48091e-24, 3.84324e-24, 3.28051e-24, 5.42405e-24, 1.51538e-23], [134.715, 6.30948e-24, 7.48092e-24, 3.84235e-24, 3.27918e-24, 5.41713e-24, 1.51553e-23], [135.021, 6.32047e-24, 7.48093e-24, 3.84147e-24, 3.27785e-24, 5.4102e-24, 1.51568e-23], [135.328, 6.33149e-24, 7.48094e-24, 3.84059e-24, 3.27651e-24, 5.40324e-24, 1.51583e-23], [135.636, 6.34254e-24, 7.48096e-24, 3.83971e-24, 3.27517e-24, 5.39627e-24, 1.51597e-23], [135.945, 6.35363e-24, 7.48098e-24, 3.83884e-24, 3.27383e-24, 5.38929e-24, 1.51611e-23], [136.254, 6.36475e-24, 7.481e-24, 3.83796e-24, 3.27248e-24, 5.38229e-24, 1.51625e-23], [136.564, 6.3759e-24, 7.48102e-24, 3.8371e-24, 3.27113e-24, 5.37527e-24, 1.51639e-23], [136.875, 6.38709e-24, 7.48105e-24, 3.83623e-24, 3.26978e-24, 5.36823e-24, 1.51653e-23], [137.186, 6.39831e-24, 7.48108e-24, 3.83537e-24, 3.26843e-24, 5.36118e-24, 1.51666e-23], [137.498, 6.40957e-24, 7.48112e-24, 3.83451e-24, 3.26707e-24, 5.35411e-24, 1.51679e-23], [137.811, 6.42086e-24, 7.48115e-24, 3.83366e-24, 3.26571e-24, 5.34703e-24, 1.51692e-23], [138.124, 6.43218e-24, 7.48119e-24, 3.8328e-24, 3.26435e-24, 5.33993e-24, 1.51704e-23], [138.438, 6.44354e-24, 7.48124e-24, 3.83195e-24, 3.26298e-24, 5.33282e-24, 1.51717e-23], [138.753, 6.45493e-24, 7.48128e-24, 3.83111e-24, 3.26162e-24, 5.32569e-24, 1.51729e-23], [139.069, 6.46636e-24, 7.48133e-24, 3.83026e-24, 3.26025e-24, 5.31855e-24, 1.51741e-23], [139.385, 6.47782e-24, 7.48139e-24, 3.82942e-24, 3.25887e-24, 5.31139e-24, 1.51753e-23], [139.702, 6.48931e-24, 7.48144e-24, 3.82858e-24, 3.2575e-24, 5.30422e-24, 1.51764e-23], [140.02, 6.50084e-24, 7.4815e-24, 3.82775e-24, 3.25612e-24, 5.29703e-24, 1.51775e-23], [140.338, 6.5124e-24, 7.48157e-24, 3.82692e-24, 3.25474e-24, 5.28983e-24, 1.51786e-23], [140.658, 6.524e-24, 7.48163e-24, 3.82609e-24, 3.25335e-24, 5.28261e-24, 1.51797e-23], [140.978, 6.53563e-24, 7.4817e-24, 3.82526e-24, 3.25197e-24, 5.27538e-24, 1.51808e-23], [141.298, 6.5473e-24, 7.48177e-24, 3.82444e-24, 3.25058e-24, 5.26814e-24, 1.51818e-23], [141.62, 6.559e-24, 7.48185e-24, 3.82362e-24, 3.24919e-24, 5.26089e-24, 1.51828e-23], [141.942, 6.57074e-24, 7.48193e-24, 3.82281e-24, 3.24779e-24, 5.25362e-24, 1.51838e-23], [142.265, 6.58251e-24, 7.48201e-24, 3.82199e-24, 3.24639e-24, 5.24634e-24, 1.51848e-23], [142.588, 6.59431e-24, 7.48209e-24, 3.82118e-24, 3.24499e-24, 5.23904e-24, 1.51857e-23], [142.913, 6.60615e-24, 7.48218e-24, 3.82037e-24, 3.24359e-24, 5.23174e-24, 1.51867e-23], [143.238, 6.61803e-24, 7.48227e-24, 3.81957e-24, 3.24218e-24, 5.22442e-24, 1.51876e-23], [143.563, 6.62994e-24, 7.48237e-24, 3.81877e-24, 3.24077e-24, 5.21709e-24, 1.51884e-23], [143.89, 6.64188e-24, 7.48247e-24, 3.81797e-24, 3.23936e-24, 5.20975e-24, 1.51893e-23], [144.217, 6.65386e-24, 7.48257e-24, 3.81717e-24, 3.23795e-24, 5.2024e-24, 1.51901e-23], [144.545, 6.66588e-24, 7.48267e-24, 3.81638e-24, 3.23653e-24, 5.19503e-24, 1.51909e-23], [144.874, 6.67793e-24, 7.48278e-24, 3.81559e-24, 3.23511e-24, 5.18766e-24, 1.51917e-23], [145.204, 6.69001e-24, 7.48289e-24, 3.81481e-24, 3.23368e-24, 5.18028e-24, 1.51925e-23], [145.534, 6.70213e-24, 7.48301e-24, 3.81402e-24, 3.23226e-24, 5.17288e-24, 1.51933e-23], [145.865, 6.71429e-24, 7.48313e-24, 3.81324e-24, 3.23083e-24, 5.16548e-24, 1.5194e-23], [146.197, 6.72648e-24, 7.48325e-24, 3.81246e-24, 3.22939e-24, 5.15807e-24, 1.51947e-23], [146.529, 6.73871e-24, 7.48338e-24, 3.81169e-24, 3.22796e-24, 5.15064e-24, 1.51954e-23], [146.862, 6.75097e-24, 7.4835e-24, 3.81092e-24, 3.22652e-24, 5.14321e-24, 1.5196e-23], [147.197, 6.76327e-24, 7.48364e-24, 3.81015e-24, 3.22508e-24, 5.13577e-24, 1.51967e-23], [147.531, 6.77561e-24, 7.48377e-24, 3.80938e-24, 3.22363e-24, 5.12833e-24, 1.51973e-23], [147.867, 6.78798e-24, 7.48391e-24, 3.80862e-24, 3.22219e-24, 5.12087e-24, 1.51979e-23], [148.203, 6.80038e-24, 7.48405e-24, 3.80786e-24, 3.22074e-24, 5.11341e-24, 1.51984e-23], [148.54, 6.81283e-24, 7.4842e-24, 3.8071e-24, 3.21928e-24, 5.10594e-24, 1.5199e-23], [148.878, 6.8253e-24, 7.48435e-24, 3.80635e-24, 3.21783e-24, 5.09846e-24, 1.51995e-23], [149.217, 6.83782e-24, 7.4845e-24, 3.8056e-24, 3.21637e-24, 5.09098e-24, 1.52e-23], [149.556, 6.85037e-24, 7.48466e-24, 3.80485e-24, 3.2149e-24, 5.08349e-24, 1.52005e-23], [149.896, 6.86295e-24, 7.48482e-24, 3.8041e-24, 3.21344e-24, 5.07599e-24, 1.5201e-23], [150.237, 6.87558e-24, 7.48498e-24, 3.80336e-24, 3.21197e-24, 5.06849e-24, 1.52014e-23], [150.579, 6.88823e-24, 7.48515e-24, 3.80262e-24, 3.2105e-24, 5.06098e-24, 1.52018e-23], [150.922, 6.90093e-24, 7.48532e-24, 3.80189e-24, 3.20903e-24, 5.05347e-24, 1.52022e-23], [151.265, 6.91366e-24, 7.48549e-24, 3.80115e-24, 3.20755e-24, 5.04596e-24, 1.52026e-23], [151.609, 6.92643e-24, 7.48567e-24, 3.80042e-24, 3.20607e-24, 5.03844e-24, 1.52029e-23], [151.954, 6.93923e-24, 7.48585e-24, 3.7997e-24, 3.20459e-24, 5.03092e-24, 1.52033e-23], [152.299, 6.95207e-24, 7.48604e-24, 3.79897e-24, 3.2031e-24, 5.02339e-24, 1.52036e-23], [152.646, 6.96495e-24, 7.48622e-24, 3.79825e-24, 3.20161e-24, 5.01586e-24, 1.52039e-23], [152.993, 6.97787e-24, 7.48642e-24, 3.79753e-24, 3.20012e-24, 5.00833e-24, 1.52041e-23], [153.341, 6.99082e-24, 7.48661e-24, 3.79682e-24, 3.19862e-24, 5.0008e-24, 1.52044e-23], [153.69, 7.00381e-24, 7.48681e-24, 3.7961e-24, 3.19713e-24, 4.99327e-24, 1.52046e-23], [154.039, 7.01683e-24, 7.48701e-24, 3.79539e-24, 3.19562e-24, 4.98573e-24, 1.52048e-23], [154.39, 7.02989e-24, 7.48722e-24, 3.79469e-24, 3.19412e-24, 4.9782e-24, 1.5205e-23], [154.741, 7.04299e-24, 7.48743e-24, 3.79398e-24, 3.19261e-24, 4.97066e-24, 1.52051e-23], [155.093, 7.05613e-24, 7.48765e-24, 3.79328e-24, 3.1911e-24, 4.96312e-24, 1.52053e-23], [155.446, 7.0693e-24, 7.48786e-24, 3.79258e-24, 3.18959e-24, 4.95559e-24, 1.52054e-23], [155.799, 7.08251e-24, 7.48809e-24, 3.79189e-24, 3.18807e-24, 4.94806e-24, 1.52055e-23], [156.154, 7.09576e-24, 7.48831e-24, 3.7912e-24, 3.18655e-24, 4.94052e-24, 1.52055e-23], [156.509, 7.10905e-24, 7.48854e-24, 3.79051e-24, 3.18503e-24, 4.933e-24, 1.52056e-23], [156.865, 7.12237e-24, 7.48877e-24, 3.78982e-24, 3.18351e-24, 4.92547e-24, 1.52056e-23], [157.222, 7.13573e-24, 7.48901e-24, 3.78914e-24, 3.18198e-24, 4.91794e-24, 1.52056e-23], [157.579, 7.14913e-24, 7.48925e-24, 3.78846e-24, 3.18045e-24, 4.91042e-24, 1.52056e-23], [157.938, 7.16256e-24, 7.48949e-24, 3.78778e-24, 3.17891e-24, 4.90291e-24, 1.52056e-23], [158.297, 7.17604e-24, 7.48974e-24, 3.78711e-24, 3.17738e-24, 4.8954e-24, 1.52055e-23], [158.657, 7.18955e-24, 7.48999e-24, 3.78644e-24, 3.17584e-24, 4.88789e-24, 1.52054e-23], [159.018, 7.2031e-24, 7.49025e-24, 3.78577e-24, 3.17429e-24, 4.88039e-24, 1.52053e-23], [159.379, 7.21668e-24, 7.49051e-24, 3.78511e-24, 3.17275e-24, 4.8729e-24, 1.52052e-23], [159.742, 7.23031e-24, 7.49077e-24, 3.78444e-24, 3.1712e-24, 4.86541e-24, 1.52051e-23], [160.105, 7.24397e-24, 7.49104e-24, 3.78378e-24, 3.16965e-24, 4.85793e-24, 1.52049e-23], [160.47, 7.25767e-24, 7.49131e-24, 3.78313e-24, 3.16809e-24, 4.85046e-24, 1.52047e-23], [160.835, 7.27141e-24, 7.49159e-24, 3.78247e-24, 3.16653e-24, 4.843e-24, 1.52045e-23], [161.2, 7.28519e-24, 7.49187e-24, 3.78182e-24, 3.16497e-24, 4.83555e-24, 1.52043e-23], [161.567, 7.29901e-24, 7.49215e-24, 3.78118e-24, 3.16341e-24, 4.8281e-24, 1.5204e-23], [161.935, 7.31286e-24, 7.49244e-24, 3.78053e-24, 3.16184e-24, 4.82067e-24, 1.52037e-23], [162.303, 7.32676e-24, 7.49273e-24, 3.77989e-24, 3.16027e-24, 4.81325e-24, 1.52034e-23], [162.672, 7.34069e-24, 7.49302e-24, 3.77925e-24, 3.1587e-24, 4.80584e-24, 1.52031e-23], [163.042, 7.35466e-24, 7.49332e-24, 3.77862e-24, 3.15712e-24, 4.79844e-24, 1.52028e-23], [163.413, 7.36867e-24, 7.49362e-24, 3.77799e-24, 3.15555e-24, 4.79105e-24, 1.52024e-23], [163.785, 7.38272e-24, 7.49393e-24, 3.77736e-24, 3.15396e-24, 4.78368e-24, 1.5202e-23], [164.157, 7.3968e-24, 7.49424e-24, 3.77673e-24, 3.15238e-24, 4.77632e-24, 1.52016e-23], [164.53, 7.41093e-24, 7.49456e-24, 3.77611e-24, 3.15079e-24, 4.76898e-24, 1.52012e-23], [164.905, 7.42509e-24, 7.49488e-24, 3.77549e-24, 3.1492e-24, 4.76165e-24, 1.52007e-23], [165.28, 7.4393e-24, 7.4952e-24, 3.77487e-24, 3.14761e-24, 4.75434e-24, 1.52003e-23], [165.656, 7.45354e-24, 7.49553e-24, 3.77426e-24, 3.14601e-24, 4.74705e-24, 1.51998e-23], [166.033, 7.46782e-24, 7.49586e-24, 3.77364e-24, 3.14441e-24, 4.73977e-24, 1.51993e-23], [166.41, 7.48215e-24, 7.49619e-24, 3.77304e-24, 3.14281e-24, 4.73251e-24, 1.51987e-23], [166.789, 7.49651e-24, 7.49653e-24, 3.77243e-24, 3.14121e-24, 4.72527e-24, 1.51982e-23], [167.168, 7.51091e-24, 7.49688e-24, 3.77183e-24, 3.1396e-24, 4.71806e-24, 1.51976e-23], [167.548, 7.52535e-24, 7.49723e-24, 3.77123e-24, 3.13799e-24, 4.71086e-24, 1.5197e-23], [167.929, 7.53983e-24, 7.49758e-24, 3.77063e-24, 3.13638e-24, 4.70368e-24, 1.51964e-23], [168.311, 7.55435e-24, 7.49794e-24, 3.77004e-24, 3.13476e-24, 4.69652e-24, 1.51957e-23], [168.694, 7.56891e-24, 7.4983e-24, 3.76945e-24, 3.13314e-24, 4.68939e-24, 1.51951e-23], [169.078, 7.5835e-24, 7.49866e-24, 3.76886e-24, 3.13152e-24, 4.68228e-24, 1.51944e-23], [169.463, 7.59814e-24, 7.49903e-24, 3.76828e-24, 3.1299e-24, 4.6752e-24, 1.51937e-23], [169.848, 7.61282e-24, 7.4994e-24, 3.7677e-24, 3.12827e-24, 4.66814e-24, 1.5193e-23], [170.234, 7.62754e-24, 7.49978e-24, 3.76712e-24, 3.12664e-24, 4.66111e-24, 1.51922e-23], [170.621, 7.6423e-24, 7.50016e-24, 3.76655e-24, 3.12501e-24, 4.6541e-24, 1.51914e-23], [171.01, 7.6571e-24, 7.50055e-24, 3.76597e-24, 3.12337e-24, 4.64712e-24, 1.51906e-23], [171.399, 7.67194e-24, 7.50094e-24, 3.7654e-24, 3.12173e-24, 4.64017e-24, 1.51898e-23], [171.788, 7.68682e-24, 7.50134e-24, 3.76484e-24, 3.12009e-24, 4.63325e-24, 1.5189e-23], [172.179, 7.70174e-24, 7.50174e-24, 3.76428e-24, 3.11845e-24, 4.62636e-24, 1.51881e-23], [172.571, 7.7167e-24, 7.50214e-24, 3.76372e-24, 3.1168e-24, 4.6195e-24, 1.51873e-23], [172.963, 7.7317e-24, 7.50255e-24, 3.76316e-24, 3.11515e-24, 4.61267e-24, 1.51863e-23], [173.357, 7.74674e-24, 7.50296e-24, 3.76261e-24, 3.1135e-24, 4.60588e-24, 1.51854e-23], [173.751, 7.76183e-24, 7.50338e-24, 3.76206e-24, 3.11185e-24, 4.59912e-24, 1.51845e-23], [174.146, 7.77695e-24, 7.5038e-24, 3.76151e-24, 3.11019e-24, 4.59239e-24, 1.51835e-23], [174.542, 7.79211e-24, 7.50423e-24, 3.76097e-24, 3.10853e-24, 4.5857e-24, 1.51825e-23], [174.939, 7.80732e-24, 7.50466e-24, 3.76042e-24, 3.10687e-24, 4.57905e-24, 1.51815e-23], [175.337, 7.82257e-24, 7.50509e-24, 3.75989e-24, 3.1052e-24, 4.57244e-24, 1.51805e-23], [175.736, 7.83785e-24, 7.50553e-24, 3.75935e-24, 3.10353e-24, 4.56586e-24, 1.51794e-23], [176.136, 7.85318e-24, 7.50598e-24, 3.75882e-24, 3.10186e-24, 4.55932e-24, 1.51783e-23], [176.536, 7.86855e-24, 7.50642e-24, 3.75829e-24, 3.10019e-24, 4.55282e-24, 1.51772e-23], [176.938, 7.88396e-24, 7.50688e-24, 3.75777e-24, 3.09851e-24, 4.54637e-24, 1.51761e-23], [177.34, 7.89942e-24, 7.50734e-24, 3.75724e-24, 3.09684e-24, 4.53996e-24, 1.5175e-23], [177.744, 7.91491e-24, 7.5078e-24, 3.75672e-24, 3.09516e-24, 4.53359e-24, 1.51738e-23], [178.148, 7.93045e-24, 7.50826e-24, 3.75621e-24, 3.09347e-24, 4.52726e-24, 1.51726e-23], [178.553, 7.94603e-24, 7.50874e-24, 3.75569e-24, 3.09179e-24, 4.52099e-24, 1.51714e-23], [178.959, 7.96164e-24, 7.50921e-24, 3.75518e-24, 3.0901e-24, 4.51475e-24, 1.51702e-23], [179.367, 7.97731e-24, 7.50969e-24, 3.75468e-24, 3.08841e-24, 4.50857e-24, 1.51689e-23], [179.775, 7.99301e-24, 7.51018e-24, 3.75417e-24, 3.08672e-24, 4.50244e-24, 1.51677e-23], [180.183, 8.00875e-24, 7.51067e-24, 3.75367e-24, 3.08502e-24, 4.49635e-24, 1.51664e-23], [180.593, 8.02454e-24, 7.51116e-24, 3.75318e-24, 3.08332e-24, 4.49032e-24, 1.51651e-23], [181.004, 8.04037e-24, 7.51166e-24, 3.75268e-24, 3.08162e-24, 4.48434e-24, 1.51637e-23], [181.416, 8.05624e-24, 7.51217e-24, 3.75219e-24, 3.07992e-24, 4.47841e-24, 1.51624e-23], [181.828, 8.07216e-24, 7.51268e-24, 3.7517e-24, 3.07822e-24, 4.47253e-24, 1.5161e-23], [182.242, 8.08811e-24, 7.51319e-24, 3.75122e-24, 3.07651e-24, 4.46672e-24, 1.51596e-23], [182.656, 8.10411e-24, 7.51371e-24, 3.75074e-24, 3.0748e-24, 4.46096e-24, 1.51581e-23], [183.072, 8.12015e-24, 7.51423e-24, 3.75026e-24, 3.07309e-24, 4.45525e-24, 1.51567e-23], [183.488, 8.13624e-24, 7.51476e-24, 3.74978e-24, 3.07138e-24, 4.44961e-24, 1.51552e-23], [183.906, 8.15236e-24, 7.5153e-24, 3.74931e-24, 3.06966e-24, 4.44403e-24, 1.51537e-23], [184.324, 8.16853e-24, 7.51583e-24, 3.74884e-24, 3.06794e-24, 4.4385e-24, 1.51522e-23], [184.743, 8.18474e-24, 7.51638e-24, 3.74838e-24, 3.06622e-24, 4.43305e-24, 1.51507e-23], [185.163, 8.201e-24, 7.51693e-24, 3.74792e-24, 3.0645e-24, 4.42765e-24, 1.51491e-23], [185.585, 8.2173e-24, 7.51748e-24, 3.74746e-24, 3.06278e-24, 4.42232e-24, 1.51475e-23], [186.007, 8.23364e-24, 7.51804e-24, 3.747e-24, 3.06105e-24, 4.41706e-24, 1.51459e-23], [186.43, 8.25002e-24, 7.5186e-24, 3.74655e-24, 3.05932e-24, 4.41186e-24, 1.51443e-23], [186.854, 8.26645e-24, 7.51917e-24, 3.7461e-24, 3.05759e-24, 4.40674e-24, 1.51426e-23], [187.279, 8.28292e-24, 7.51974e-24, 3.74565e-24, 3.05586e-24, 4.40168e-24, 1.5141e-23], [187.705, 8.29944e-24, 7.52032e-24, 3.74521e-24, 3.05412e-24, 4.3967e-24, 1.51393e-23], [188.132, 8.316e-24, 7.5209e-24, 3.74477e-24, 3.05239e-24, 4.39179e-24, 1.51376e-23], [188.56, 8.3326e-24, 7.52149e-24, 3.74433e-24, 3.05065e-24, 4.38696e-24, 1.51358e-23], [188.989, 8.34924e-24, 7.52208e-24, 3.7439e-24, 3.04891e-24, 4.3822e-24, 1.5134e-23], [189.419, 8.36593e-24, 7.52268e-24, 3.74347e-24, 3.04716e-24, 4.37751e-24, 1.51323e-23], [189.849, 8.38267e-24, 7.52328e-24, 3.74304e-24, 3.04542e-24, 4.37291e-24, 1.51305e-23], [190.281, 8.39944e-24, 7.52389e-24, 3.74262e-24, 3.04367e-24, 4.36838e-24, 1.51286e-23], [190.714, 8.41627e-24, 7.52451e-24, 3.7422e-24, 3.04193e-24, 4.36394e-24, 1.51268e-23], [191.148, 8.43313e-24, 7.52512e-24, 3.74178e-24, 3.04018e-24, 4.35958e-24, 1.51249e-23], [191.583, 8.45004e-24, 7.52575e-24, 3.74137e-24, 3.03842e-24, 4.3553e-24, 1.5123e-23], [192.018, 8.46699e-24, 7.52638e-24, 3.74096e-24, 3.03667e-24, 4.35111e-24, 1.51211e-23], [192.455, 8.48399e-24, 7.52701e-24, 3.74055e-24, 3.03492e-24, 4.34701e-24, 1.51191e-23], [192.893, 8.50104e-24, 7.52765e-24, 3.74015e-24, 3.03316e-24, 4.34299e-24, 1.51172e-23], [193.332, 8.51812e-24, 7.5283e-24, 3.73975e-24, 3.0314e-24, 4.33906e-24, 1.51152e-23], [193.771, 8.53526e-24, 7.52895e-24, 3.73935e-24, 3.02964e-24, 4.33523e-24, 1.51132e-23], [194.212, 8.55243e-24, 7.5296e-24, 3.73896e-24, 3.02788e-24, 4.33148e-24, 1.51111e-23], [194.654, 8.56965e-24, 7.53026e-24, 3.73857e-24, 3.02612e-24, 4.32783e-24, 1.51091e-23], [195.097, 8.58692e-24, 7.53093e-24, 3.73818e-24, 3.02435e-24, 4.32428e-24, 1.5107e-23], [195.54, 8.60423e-24, 7.5316e-24, 3.7378e-24, 3.02259e-24, 4.32082e-24, 1.51049e-23], [195.985, 8.62159e-24, 7.53228e-24, 3.73742e-24, 3.02082e-24, 4.31747e-24, 1.51028e-23], [196.431, 8.63899e-24, 7.53296e-24, 3.73704e-24, 3.01905e-24, 4.31421e-24, 1.51006e-23], [196.878, 8.65644e-24, 7.53365e-24, 3.73667e-24, 3.01728e-24, 4.31105e-24, 1.50984e-23], [197.326, 8.67393e-24, 7.53434e-24, 3.7363e-24, 3.01551e-24, 4.308e-24, 1.50963e-23], [197.774, 8.69147e-24, 7.53504e-24, 3.73593e-24, 3.01374e-24, 4.30505e-24, 1.5094e-23], [198.224, 8.70905e-24, 7.53574e-24, 3.73557e-24, 3.01197e-24, 4.3022e-24, 1.50918e-23], [198.675, 8.72668e-24, 7.53645e-24, 3.73521e-24, 3.01019e-24, 4.29947e-24, 1.50895e-23], [199.127, 8.74435e-24, 7.53717e-24, 3.73485e-24, 3.00842e-24, 4.29684e-24, 1.50872e-23], [199.58, 8.76208e-24, 7.53789e-24, 3.7345e-24, 3.00664e-24, 4.29432e-24, 1.50849e-23], [200.034, 8.77984e-24, 7.53861e-24, 3.73415e-24, 3.00486e-24, 4.29192e-24, 1.50826e-23], [200.489, 8.79765e-24, 7.53935e-24, 3.73381e-24, 3.00308e-24, 4.28962e-24, 1.50802e-23], [200.945, 8.81551e-24, 7.54008e-24, 3.73346e-24, 3.0013e-24, 4.28745e-24, 1.50779e-23], [201.402, 8.83342e-24, 7.54083e-24, 3.73312e-24, 2.99952e-24, 4.28539e-24, 1.50755e-23], [201.86, 8.85137e-24, 7.54158e-24, 3.73279e-24, 2.99774e-24, 4.28345e-24, 1.5073e-23], [202.319, 8.86936e-24, 7.54233e-24, 3.73246e-24, 2.99596e-24, 4.28163e-24, 1.50706e-23], [202.779, 8.88741e-24, 7.54309e-24, 3.73213e-24, 2.99417e-24, 4.27993e-24, 1.50681e-23], [203.241, 8.9055e-24, 7.54386e-24, 3.7318e-24, 2.99239e-24, 4.27835e-24, 1.50656e-23], [203.703, 8.92364e-24, 7.54463e-24, 3.73148e-24, 2.9906e-24, 4.2769e-24, 1.50631e-23], [204.166, 8.94182e-24, 7.54541e-24, 3.73116e-24, 2.98882e-24, 4.27557e-24, 1.50605e-23], [204.631, 8.96005e-24, 7.54619e-24, 3.73085e-24, 2.98703e-24, 4.27437e-24, 1.5058e-23], [205.096, 8.97833e-24, 7.54698e-24, 3.73054e-24, 2.98525e-24, 4.2733e-24, 1.50554e-23], [205.563, 8.99665e-24, 7.54777e-24, 3.73023e-24, 2.98346e-24, 4.27236e-24, 1.50528e-23], [206.03, 9.01502e-24, 7.54857e-24, 3.72993e-24, 2.98167e-24, 4.27156e-24, 1.50501e-23], [206.499, 9.03344e-24, 7.54938e-24, 3.72962e-24, 2.97988e-24, 4.27088e-24, 1.50475e-23], [206.968, 9.05191e-24, 7.55019e-24, 3.72933e-24, 2.97809e-24, 4.27035e-24, 1.50448e-23], [207.439, 9.07042e-24, 7.55101e-24, 3.72903e-24, 2.9763e-24, 4.26995e-24, 1.50421e-23], [207.911, 9.08898e-24, 7.55183e-24, 3.72874e-24, 2.97451e-24, 4.26969e-24, 1.50393e-23], [208.384, 9.10759e-24, 7.55266e-24, 3.72846e-24, 2.97272e-24, 4.26957e-24, 1.50366e-23], [208.858, 9.12624e-24, 7.5535e-24, 3.72817e-24, 2.97093e-24, 4.26959e-24, 1.50338e-23], [209.333, 9.14495e-24, 7.55434e-24, 3.7279e-24, 2.96914e-24, 4.26975e-24, 1.5031e-23], [209.809, 9.1637e-24, 7.55519e-24, 3.72762e-24, 2.96735e-24, 4.27006e-24, 1.50281e-23], [210.286, 9.1825e-24, 7.55605e-24, 3.72735e-24, 2.96556e-24, 4.27051e-24, 1.50253e-23], [210.765, 9.20134e-24, 7.55691e-24, 3.72708e-24, 2.96377e-24, 4.27112e-24, 1.50224e-23], [211.244, 9.22024e-24, 7.55777e-24, 3.72682e-24, 2.96198e-24, 4.27187e-24, 1.50195e-23], [211.725, 9.23918e-24, 7.55864e-24, 3.72655e-24, 2.96019e-24, 4.27277e-24, 1.50166e-23], [212.206, 9.25817e-24, 7.55952e-24, 3.7263e-24, 2.9584e-24, 4.27383e-24, 1.50136e-23], [212.689, 9.27721e-24, 7.56041e-24, 3.72604e-24, 2.95661e-24, 4.27504e-24, 1.50106e-23], [213.173, 9.2963e-24, 7.5613e-24, 3.72579e-24, 2.95482e-24, 4.27641e-24, 1.50076e-23], [213.658, 9.31544e-24, 7.5622e-24, 3.72555e-24, 2.95303e-24, 4.27793e-24, 1.50046e-23], [214.143, 9.33462e-24, 7.5631e-24, 3.7253e-24, 2.95124e-24, 4.27961e-24, 1.50015e-23], [214.631, 9.35386e-24, 7.56401e-24, 3.72506e-24, 2.94946e-24, 4.28145e-24, 1.49985e-23], [215.119, 9.37314e-24, 7.56493e-24, 3.72483e-24, 2.94767e-24, 4.28346e-24, 1.49954e-23], [215.608, 9.39248e-24, 7.56585e-24, 3.7246e-24, 2.94588e-24, 4.28563e-24, 1.49922e-23], [216.098, 9.41186e-24, 7.56678e-24, 3.72437e-24, 2.9441e-24, 4.28796e-24, 1.49891e-23], [216.59, 9.43129e-24, 7.56771e-24, 3.72414e-24, 2.94231e-24, 4.29045e-24, 1.49859e-23], [217.083, 9.45077e-24, 7.56865e-24, 3.72392e-24, 2.94052e-24, 4.29312e-24, 1.49827e-23], [217.576, 9.4703e-24, 7.5696e-24, 3.72371e-24, 2.93874e-24, 4.29595e-24, 1.49795e-23], [218.071, 9.48988e-24, 7.57056e-24, 3.72349e-24, 2.93696e-24, 4.29896e-24, 1.49762e-23], [218.567, 9.5095e-24, 7.57152e-24, 3.72328e-24, 2.93517e-24, 4.30213e-24, 1.4973e-23], [219.064, 9.52918e-24, 7.57248e-24, 3.72308e-24, 2.93339e-24, 4.30548e-24, 1.49697e-23], [219.563, 9.54891e-24, 7.57346e-24, 3.72288e-24, 2.93161e-24, 4.30901e-24, 1.49663e-23], [220.062, 9.56869e-24, 7.57444e-24, 3.72268e-24, 2.92983e-24, 4.31271e-24, 1.4963e-23], [220.563, 9.58851e-24, 7.57542e-24, 3.72249e-24, 2.92806e-24, 4.31658e-24, 1.49596e-23], [221.064, 9.60839e-24, 7.57642e-24, 3.72229e-24, 2.92628e-24, 4.32064e-24, 1.49562e-23], [221.567, 9.62832e-24, 7.57742e-24, 3.72211e-24, 2.9245e-24, 4.32487e-24, 1.49528e-23], [222.071, 9.6483e-24, 7.57842e-24, 3.72193e-24, 2.92273e-24, 4.32929e-24, 1.49493e-23], [222.576, 9.66833e-24, 7.57944e-24, 3.72175e-24, 2.92096e-24, 4.33389e-24, 1.49459e-23], [223.083, 9.68841e-24, 7.58046e-24, 3.72157e-24, 2.91919e-24, 4.33867e-24, 1.49424e-23], [223.59, 9.70854e-24, 7.58148e-24, 3.7214e-24, 2.91742e-24, 4.34364e-24, 1.49388e-23], [224.099, 9.72872e-24, 7.58252e-24, 3.72123e-24, 2.91565e-24, 4.34879e-24, 1.49353e-23], [224.608, 9.74895e-24, 7.58356e-24, 3.72107e-24, 2.91388e-24, 4.35414e-24, 1.49317e-23], [225.119, 9.76923e-24, 7.5846e-24, 3.72091e-24, 2.91212e-24, 4.35967e-24, 1.49281e-23], [225.631, 9.78956e-24, 7.58566e-24, 3.72075e-24, 2.91036e-24, 4.36539e-24, 1.49244e-23], [226.144, 9.80994e-24, 7.58672e-24, 3.7206e-24, 2.9086e-24, 4.3713e-24, 1.49208e-23], [226.659, 9.83038e-24, 7.58778e-24, 3.72045e-24, 2.90684e-24, 4.3774e-24, 1.49171e-23], [227.174, 9.85087e-24, 7.58886e-24, 3.72031e-24, 2.90508e-24, 4.3837e-24, 1.49134e-23], [227.691, 9.8714e-24, 7.58994e-24, 3.72017e-24, 2.90333e-24, 4.39019e-24, 1.49097e-23], [228.209, 9.89199e-24, 7.59103e-24, 3.72003e-24, 2.90158e-24, 4.39688e-24, 1.49059e-23], [228.728, 9.91263e-24, 7.59212e-24, 3.7199e-24, 2.89983e-24, 4.40377e-24, 1.49021e-23], [229.248, 9.93333e-24, 7.59322e-24, 3.71977e-24, 2.89809e-24, 4.41085e-24, 1.48983e-23], [229.77, 9.95407e-24, 7.59433e-24, 3.71965e-24, 2.89634e-24, 4.41813e-24, 1.48945e-23], [230.292, 9.97487e-24, 7.59545e-24, 3.71953e-24, 2.8946e-24, 4.42561e-24, 1.48906e-23], [230.816, 9.99571e-24, 7.59657e-24, 3.71941e-24, 2.89286e-24, 4.43329e-24, 1.48867e-23], [231.341, 1.00166e-23, 7.5977e-24, 3.7193e-24, 2.89113e-24, 4.44118e-24, 1.48828e-23], [231.867, 1.00376e-23, 7.59884e-24, 3.71919e-24, 2.8894e-24, 4.44926e-24, 1.48788e-23], [232.395, 1.00586e-23, 7.59998e-24, 3.71909e-24, 2.88767e-24, 4.45755e-24, 1.48748e-23], [232.923, 1.00796e-23, 7.60113e-24, 3.71899e-24, 2.88594e-24, 4.46605e-24, 1.48708e-23], [233.453, 1.01007e-23, 7.60229e-24, 3.71889e-24, 2.88422e-24, 4.47475e-24, 1.48668e-23], [233.984, 1.01219e-23, 7.60346e-24, 3.7188e-24, 2.8825e-24, 4.48366e-24, 1.48628e-23], [234.516, 1.01431e-23, 7.60463e-24, 3.71871e-24, 2.88078e-24, 4.49277e-24, 1.48587e-23], [235.05, 1.01644e-23, 7.60581e-24, 3.71863e-24, 2.87907e-24, 4.50209e-24, 1.48546e-23], [235.585, 1.01857e-23, 7.607e-24, 3.71855e-24, 2.87736e-24, 4.51163e-24, 1.48504e-23], [236.12, 1.02071e-23, 7.60819e-24, 3.71847e-24, 2.87565e-24, 4.52137e-24, 1.48463e-23], [236.657, 1.02285e-23, 7.6094e-24, 3.7184e-24, 2.87395e-24, 4.53132e-24, 1.48421e-23], [237.196, 1.025e-23, 7.61061e-24, 3.71833e-24, 2.87225e-24, 4.54148e-24, 1.48379e-23], [237.735, 1.02715e-23, 7.61182e-24, 3.71827e-24, 2.87056e-24, 4.55186e-24, 1.48336e-23], [238.276, 1.02931e-23, 7.61305e-24, 3.71821e-24, 2.86887e-24, 4.56244e-24, 1.48294e-23], [238.818, 1.03147e-23, 7.61428e-24, 3.71816e-24, 2.86718e-24, 4.57324e-24, 1.48251e-23], [239.361, 1.03364e-23, 7.61552e-24, 3.7181e-24, 2.8655e-24, 4.58426e-24, 1.48207e-23], [239.906, 1.03582e-23, 7.61677e-24, 3.71806e-24, 2.86382e-24, 4.59548e-24, 1.48164e-23], [240.451, 1.038e-23, 7.61802e-24, 3.71802e-24, 2.86215e-24, 4.60693e-24, 1.4812e-23], [240.998, 1.04019e-23, 7.61928e-24, 3.71798e-24, 2.86048e-24, 4.61859e-24, 1.48076e-23], [241.546, 1.04238e-23, 7.62055e-24, 3.71794e-24, 2.85882e-24, 4.63046e-24, 1.48031e-23], [242.096, 1.04457e-23, 7.62183e-24, 3.71791e-24, 2.85716e-24, 4.64255e-24, 1.47987e-23], [242.646, 1.04677e-23, 7.62312e-24, 3.71789e-24, 2.8555e-24, 4.65486e-24, 1.47942e-23], [243.198, 1.04898e-23, 7.62441e-24, 3.71787e-24, 2.85385e-24, 4.66739e-24, 1.47897e-23], [243.752, 1.0512e-23, 7.62571e-24, 3.71785e-24, 2.85221e-24, 4.68014e-24, 1.47851e-23], [244.306, 1.05341e-23, 7.62702e-24, 3.71784e-24, 2.85057e-24, 4.6931e-24, 1.47805e-23], [244.862, 1.05564e-23, 7.62833e-24, 3.71783e-24, 2.84893e-24, 4.70628e-24, 1.47759e-23], [245.419, 1.05787e-23, 7.62966e-24, 3.71782e-24, 2.8473e-24, 4.71969e-24, 1.47713e-23], [245.977, 1.0601e-23, 7.63099e-24, 3.71782e-24, 2.84568e-24, 4.73331e-24, 1.47666e-23], [246.536, 1.06234e-23, 7.63233e-24, 3.71783e-24, 2.84406e-24, 4.74715e-24, 1.47619e-23], [247.097, 1.06459e-23, 7.63368e-24, 3.71784e-24, 2.84245e-24, 4.76121e-24, 1.47572e-23], [247.659, 1.06684e-23, 7.63503e-24, 3.71785e-24, 2.84085e-24, 4.7755e-24, 1.47525e-23], [248.222, 1.0691e-23, 7.6364e-24, 3.71787e-24, 2.83925e-24, 4.79e-24, 1.47477e-23], [248.787, 1.07136e-23, 7.63777e-24, 3.71789e-24, 2.83765e-24, 4.80473e-24, 1.47429e-23], [249.353, 1.07363e-23, 7.63915e-24, 3.71792e-24, 2.83606e-24, 4.81968e-24, 1.4738e-23], [249.92, 1.0759e-23, 7.64053e-24, 3.71795e-24, 2.83448e-24, 4.83485e-24, 1.47332e-23], [250.489, 1.07818e-23, 7.64193e-24, 3.71798e-24, 2.83291e-24, 4.85024e-24, 1.47283e-23], [251.058, 1.08046e-23, 7.64333e-24, 3.71802e-24, 2.83134e-24, 4.86585e-24, 1.47234e-23], [251.629, 1.08275e-23, 7.64474e-24, 3.71806e-24, 2.82978e-24, 4.88169e-24, 1.47184e-23], [252.202, 1.08505e-23, 7.64616e-24, 3.71811e-24, 2.82822e-24, 4.89775e-24, 1.47134e-23], [252.775, 1.08735e-23, 7.64759e-24, 3.71817e-24, 2.82667e-24, 4.91403e-24, 1.47084e-23], [253.35, 1.08966e-23, 7.64903e-24, 3.71822e-24, 2.82513e-24, 4.93053e-24, 1.47034e-23], [253.927, 1.09197e-23, 7.65047e-24, 3.71829e-24, 2.8236e-24, 4.94726e-24, 1.46983e-23], [254.504, 1.09429e-23, 7.65193e-24, 3.71835e-24, 2.82207e-24, 4.96421e-24, 1.46932e-23], [255.083, 1.09662e-23, 7.65339e-24, 3.71842e-24, 2.82055e-24, 4.98138e-24, 1.46881e-23], [255.663, 1.09895e-23, 7.65486e-24, 3.7185e-24, 2.81904e-24, 4.99878e-24, 1.46829e-23], [256.245, 1.10128e-23, 7.65634e-24, 3.71858e-24, 2.81753e-24, 5.0164e-24, 1.46777e-23], [256.828, 1.10362e-23, 7.65782e-24, 3.71866e-24, 2.81604e-24, 5.03424e-24, 1.46725e-23], [257.412, 1.10597e-23, 7.65932e-24, 3.71875e-24, 2.81455e-24, 5.05231e-24, 1.46673e-23], [257.997, 1.10832e-23, 7.66082e-24, 3.71884e-24, 2.81307e-24, 5.0706e-24, 1.4662e-23], [258.584, 1.11068e-23, 7.66233e-24, 3.71894e-24, 2.8116e-24, 5.08911e-24, 1.46567e-23], [259.172, 1.11305e-23, 7.66385e-24, 3.71905e-24, 2.81013e-24, 5.10785e-24, 1.46513e-23], [259.762, 1.11542e-23, 7.66538e-24, 3.71915e-24, 2.80868e-24, 5.12681e-24, 1.4646e-23], [260.353, 1.11779e-23, 7.66692e-24, 3.71926e-24, 2.80723e-24, 5.14599e-24, 1.46406e-23], [260.945, 1.12018e-23, 7.66847e-24, 3.71938e-24, 2.80579e-24, 5.16539e-24, 1.46351e-23], [261.538, 1.12256e-23, 7.67002e-24, 3.7195e-24, 2.80437e-24, 5.18502e-24, 1.46297e-23], [262.133, 1.12496e-23, 7.67159e-24, 3.71963e-24, 2.80295e-24, 5.20487e-24, 1.46242e-23], [262.73, 1.12736e-23, 7.67316e-24, 3.71976e-24, 2.80153e-24, 5.22494e-24, 1.46186e-23], [263.327, 1.12976e-23, 7.67474e-24, 3.71989e-24, 2.80013e-24, 5.24523e-24, 1.46131e-23], [263.926, 1.13217e-23, 7.67633e-24, 3.72003e-24, 2.79874e-24, 5.26575e-24, 1.46075e-23], [264.526, 1.13459e-23, 7.67793e-24, 3.72018e-24, 2.79736e-24, 5.28649e-24, 1.46019e-23], [265.128, 1.13701e-23, 7.67954e-24, 3.72033e-24, 2.79599e-24, 5.30745e-24, 1.45962e-23], [265.731, 1.13944e-23, 7.68115e-24, 3.72048e-24, 2.79462e-24, 5.32863e-24, 1.45906e-23], [266.336, 1.14188e-23, 7.68278e-24, 3.72064e-24, 2.79327e-24, 5.35003e-24, 1.45849e-23], [266.941, 1.14432e-23, 7.68441e-24, 3.72081e-24, 2.79193e-24, 5.37166e-24, 1.45791e-23], [267.549, 1.14677e-23, 7.68606e-24, 3.72097e-24, 2.7906e-24, 5.3935e-24, 1.45733e-23], [268.157, 1.14922e-23, 7.68771e-24, 3.72115e-24, 2.78928e-24, 5.41557e-24, 1.45675e-23], [268.767, 1.15168e-23, 7.68937e-24, 3.72133e-24, 2.78797e-24, 5.43785e-24, 1.45617e-23], [269.378, 1.15414e-23, 7.69104e-24, 3.72151e-24, 2.78667e-24, 5.46036e-24, 1.45558e-23], [269.991, 1.15661e-23, 7.69272e-24, 3.7217e-24, 2.78538e-24, 5.48309e-24, 1.45499e-23], [270.605, 1.15909e-23, 7.69441e-24, 3.72189e-24, 2.7841e-24, 5.50603e-24, 1.4544e-23], [271.221, 1.16157e-23, 7.69611e-24, 3.72209e-24, 2.78284e-24, 5.5292e-24, 1.4538e-23], [271.838, 1.16406e-23, 7.69782e-24, 3.72229e-24, 2.78158e-24, 5.55258e-24, 1.4532e-23], [272.456, 1.16655e-23, 7.69954e-24, 3.7225e-24, 2.78034e-24, 5.57618e-24, 1.4526e-23], [273.076, 1.16906e-23, 7.70126e-24, 3.72271e-24, 2.77911e-24, 5.60001e-24, 1.45199e-23], [273.697, 1.17156e-23, 7.703e-24, 3.72293e-24, 2.77789e-24, 5.62405e-24, 1.45138e-23], [274.319, 1.17407e-23, 7.70474e-24, 3.72315e-24, 2.77668e-24, 5.6483e-24, 1.45077e-23], [274.943, 1.17659e-23, 7.7065e-24, 3.72338e-24, 2.77549e-24, 5.67278e-24, 1.45015e-23], [275.569, 1.17912e-23, 7.70826e-24, 3.72361e-24, 2.77431e-24, 5.69747e-24, 1.44953e-23], [276.195, 1.18165e-23, 7.71003e-24, 3.72385e-24, 2.77314e-24, 5.72238e-24, 1.44891e-23], [276.824, 1.18419e-23, 7.71182e-24, 3.72409e-24, 2.77198e-24, 5.7475e-24, 1.44829e-23], [277.453, 1.18673e-23, 7.71361e-24, 3.72434e-24, 2.77084e-24, 5.77284e-24, 1.44766e-23], [278.084, 1.18928e-23, 7.71541e-24, 3.72459e-24, 2.76971e-24, 5.7984e-24, 1.44702e-23], [278.717, 1.19184e-23, 7.71722e-24, 3.72485e-24, 2.76859e-24, 5.82417e-24, 1.44639e-23], [279.351, 1.1944e-23, 7.71905e-24, 3.72511e-24, 2.76749e-24, 5.85016e-24, 1.44575e-23], [279.986, 1.19697e-23, 7.72088e-24, 3.72538e-24, 2.7664e-24, 5.87636e-24, 1.44511e-23], [280.623, 1.19954e-23, 7.72272e-24, 3.72565e-24, 2.76532e-24, 5.90278e-24, 1.44446e-23], [281.261, 1.20212e-23, 7.72457e-24, 3.72593e-24, 2.76426e-24, 5.92941e-24, 1.44381e-23], [281.901, 1.20471e-23, 7.72643e-24, 3.72621e-24, 2.76321e-24, 5.95625e-24, 1.44316e-23], [282.542, 1.2073e-23, 7.7283e-24, 3.7265e-24, 2.76218e-24, 5.98331e-24, 1.4425e-23], [283.185, 1.2099e-23, 7.73018e-24, 3.72679e-24, 2.76116e-24, 6.01058e-24, 1.44184e-23], [283.829, 1.21251e-23, 7.73207e-24, 3.72709e-24, 2.76016e-24, 6.03806e-24, 1.44118e-23], [284.475, 1.21512e-23, 7.73397e-24, 3.72739e-24, 2.75917e-24, 6.06576e-24, 1.44051e-23], [285.122, 1.21774e-23, 7.73588e-24, 3.7277e-24, 2.7582e-24, 6.09366e-24, 1.43984e-23], [285.77, 1.22036e-23, 7.7378e-24, 3.72802e-24, 2.75724e-24, 6.12178e-24, 1.43917e-23], [286.42, 1.22299e-23, 7.73973e-24, 3.72834e-24, 2.7563e-24, 6.15011e-24, 1.43849e-23], [287.072, 1.22563e-23, 7.74167e-24, 3.72866e-24, 2.75537e-24, 6.17865e-24, 1.43781e-23], [287.725, 1.22827e-23, 7.74362e-24, 3.72899e-24, 2.75446e-24, 6.20739e-24, 1.43713e-23], [288.379, 1.23092e-23, 7.74558e-24, 3.72933e-24, 2.75357e-24, 6.23635e-24, 1.43644e-23], [289.035, 1.23358e-23, 7.74755e-24, 3.72967e-24, 2.75269e-24, 6.26552e-24, 1.43575e-23], [289.693, 1.23624e-23, 7.74954e-24, 3.73001e-24, 2.75183e-24, 6.29489e-24, 1.43505e-23], [290.352, 1.23891e-23, 7.75153e-24, 3.73036e-24, 2.75099e-24, 6.32448e-24, 1.43436e-23], [291.012, 1.24159e-23, 7.75353e-24, 3.73072e-24, 2.75016e-24, 6.35427e-24, 1.43365e-23], [291.674, 1.24427e-23, 7.75554e-24, 3.73108e-24, 2.74935e-24, 6.38427e-24, 1.43295e-23], [292.337, 1.24696e-23, 7.75757e-24, 3.73145e-24, 2.74856e-24, 6.41448e-24, 1.43224e-23], [293.002, 1.24965e-23, 7.7596e-24, 3.73182e-24, 2.74778e-24, 6.44489e-24, 1.43153e-23], [293.669, 1.25235e-23, 7.76164e-24, 3.7322e-24, 2.74702e-24, 6.47551e-24, 1.43081e-23], [294.337, 1.25506e-23, 7.7637e-24, 3.73258e-24, 2.74628e-24, 6.50633e-24, 1.43009e-23], [295.006, 1.25778e-23, 7.76576e-24, 3.73297e-24, 2.74556e-24, 6.53737e-24, 1.42937e-23], [295.677, 1.2605e-23, 7.76784e-24, 3.73337e-24, 2.74486e-24, 6.5686e-24, 1.42864e-23], [296.35, 1.26322e-23, 7.76992e-24, 3.73377e-24, 2.74418e-24, 6.60004e-24, 1.42791e-23], [297.024, 1.26596e-23, 7.77202e-24, 3.73417e-24, 2.74351e-24, 6.63169e-24, 1.42718e-23], [297.699, 1.2687e-23, 7.77413e-24, 3.73458e-24, 2.74286e-24, 6.66354e-24, 1.42644e-23], [298.377, 1.27144e-23, 7.77625e-24, 3.735e-24, 2.74224e-24, 6.69559e-24, 1.4257e-23], [299.055, 1.2742e-23, 7.77838e-24, 3.73542e-24, 2.74163e-24, 6.72785e-24, 1.42496e-23], [299.735, 1.27696e-23, 7.78052e-24, 3.73585e-24, 2.74104e-24, 6.76031e-24, 1.42421e-23], [300.417, 1.27973e-23, 7.78267e-24, 3.73628e-24, 2.74047e-24, 6.79297e-24, 1.42346e-23], [301.101, 1.2825e-23, 7.78483e-24, 3.73672e-24, 2.73992e-24, 6.82583e-24, 1.4227e-23], [301.785, 1.28528e-23, 7.787e-24, 3.73717e-24, 2.7394e-24, 6.85889e-24, 1.42194e-23], [302.472, 1.28807e-23, 7.78919e-24, 3.73762e-24, 2.73889e-24, 6.89216e-24, 1.42118e-23], [303.16, 1.29086e-23, 7.79138e-24, 3.73807e-24, 2.7384e-24, 6.92563e-24, 1.42041e-23], [303.849, 1.29366e-23, 7.79359e-24, 3.73853e-24, 2.73794e-24, 6.95929e-24, 1.41964e-23], [304.541, 1.29647e-23, 7.7958e-24, 3.739e-24, 2.73749e-24, 6.99316e-24, 1.41886e-23], [305.233, 1.29928e-23, 7.79803e-24, 3.73947e-24, 2.73707e-24, 7.02723e-24, 1.41809e-23], [305.928, 1.3021e-23, 7.80027e-24, 3.73995e-24, 2.73667e-24, 7.06149e-24, 1.4173e-23], [306.623, 1.30493e-23, 7.80252e-24, 3.74044e-24, 2.73629e-24, 7.09596e-24, 1.41652e-23], [307.321, 1.30776e-23, 7.80479e-24, 3.74093e-24, 2.73593e-24, 7.13062e-24, 1.41573e-23], [308.02, 1.3106e-23, 7.80706e-24, 3.74142e-24, 2.73559e-24, 7.16548e-24, 1.41493e-23], [308.72, 1.31345e-23, 7.80934e-24, 3.74193e-24, 2.73528e-24, 7.20054e-24, 1.41414e-23], [309.423, 1.3163e-23, 7.81164e-24, 3.74243e-24, 2.73499e-24, 7.2358e-24, 1.41333e-23], [310.126, 1.31916e-23, 7.81395e-24, 3.74295e-24, 2.73472e-24, 7.27125e-24, 1.41253e-23], [310.832, 1.32203e-23, 7.81627e-24, 3.74347e-24, 2.73448e-24, 7.3069e-24, 1.41172e-23], [311.539, 1.32491e-23, 7.8186e-24, 3.74399e-24, 2.73426e-24, 7.34275e-24, 1.41091e-23], [312.248, 1.32779e-23, 7.82094e-24, 3.74452e-24, 2.73406e-24, 7.37879e-24, 1.41009e-23], [312.958, 1.33068e-23, 7.8233e-24, 3.74506e-24, 2.73389e-24, 7.41503e-24, 1.40927e-23], [313.67, 1.33357e-23, 7.82566e-24, 3.7456e-24, 2.73374e-24, 7.45146e-24, 1.40844e-23], [314.383, 1.33647e-23, 7.82804e-24, 3.74615e-24, 2.73362e-24, 7.48809e-24, 1.40761e-23], [315.098, 1.33938e-23, 7.83043e-24, 3.74671e-24, 2.73352e-24, 7.52491e-24, 1.40678e-23], [315.815, 1.3423e-23, 7.83283e-24, 3.74727e-24, 2.73344e-24, 7.56193e-24, 1.40594e-23], [316.533, 1.34522e-23, 7.83524e-24, 3.74784e-24, 2.73339e-24, 7.59914e-24, 1.4051e-23], [317.253, 1.34815e-23, 7.83767e-24, 3.74841e-24, 2.73337e-24, 7.63654e-24, 1.40426e-23], [317.975, 1.35109e-23, 7.84011e-24, 3.74899e-24, 2.73337e-24, 7.67414e-24, 1.40341e-23], [318.698, 1.35403e-23, 7.84256e-24, 3.74957e-24, 2.7334e-24, 7.71193e-24, 1.40256e-23], [319.423, 1.35698e-23, 7.84502e-24, 3.75016e-24, 2.73345e-24, 7.74992e-24, 1.4017e-23], [320.149, 1.35994e-23, 7.84749e-24, 3.75076e-24, 2.73353e-24, 7.78809e-24, 1.40084e-23], [320.878, 1.36291e-23, 7.84998e-24, 3.75136e-24, 2.73364e-24, 7.82646e-24, 1.39998e-23], [321.608, 1.36588e-23, 7.85247e-24, 3.75197e-24, 2.73377e-24, 7.86502e-24, 1.39911e-23], [322.339, 1.36886e-23, 7.85498e-24, 3.75259e-24, 2.73393e-24, 7.90377e-24, 1.39823e-23], [323.072, 1.37184e-23, 7.8575e-24, 3.75321e-24, 2.73412e-24, 7.94271e-24, 1.39736e-23], [323.807, 1.37484e-23, 7.86004e-24, 3.75384e-24, 2.73433e-24, 7.98184e-24, 1.39647e-23], [324.544, 1.37784e-23, 7.86259e-24, 3.75447e-24, 2.73457e-24, 8.02116e-24, 1.39559e-23], [325.282, 1.38084e-23, 7.86514e-24, 3.75512e-24, 2.73484e-24, 8.06068e-24, 1.3947e-23], [326.022, 1.38386e-23, 7.86772e-24, 3.75576e-24, 2.73514e-24, 8.10038e-24, 1.39381e-23], [326.763, 1.38688e-23, 7.8703e-24, 3.75642e-24, 2.73547e-24, 8.14027e-24, 1.39291e-23], [327.506, 1.38991e-23, 7.8729e-24, 3.75707e-24, 2.73583e-24, 8.18035e-24, 1.39201e-23], [328.251, 1.39295e-23, 7.87551e-24, 3.75774e-24, 2.73621e-24, 8.22062e-24, 1.3911e-23], [328.998, 1.39599e-23, 7.87813e-24, 3.75841e-24, 2.73663e-24, 8.26108e-24, 1.39019e-23], [329.746, 1.39904e-23, 7.88076e-24, 3.75909e-24, 2.73707e-24, 8.30173e-24, 1.38927e-23], [330.496, 1.4021e-23, 7.88341e-24, 3.75978e-24, 2.73754e-24, 8.34256e-24, 1.38835e-23], [331.248, 1.40516e-23, 7.88607e-24, 3.76047e-24, 2.73805e-24, 8.38358e-24, 1.38743e-23], [332.002, 1.40823e-23, 7.88874e-24, 3.76116e-24, 2.73858e-24, 8.4248e-24, 1.3865e-23], [332.757, 1.41131e-23, 7.89143e-24, 3.76187e-24, 2.73914e-24, 8.46619e-24, 1.38557e-23], [333.514, 1.4144e-23, 7.89413e-24, 3.76258e-24, 2.73974e-24, 8.50778e-24, 1.38464e-23], [334.272, 1.41749e-23, 7.89684e-24, 3.7633e-24, 2.74036e-24, 8.54955e-24, 1.3837e-23], [335.033, 1.4206e-23, 7.89956e-24, 3.76402e-24, 2.74102e-24, 8.59151e-24, 1.38275e-23], [335.795, 1.4237e-23, 7.9023e-24, 3.76475e-24, 2.74171e-24, 8.63365e-24, 1.3818e-23], [336.558, 1.42682e-23, 7.90505e-24, 3.76549e-24, 2.74243e-24, 8.67598e-24, 1.38085e-23], [337.324, 1.42994e-23, 7.90781e-24, 3.76623e-24, 2.74318e-24, 8.7185e-24, 1.37989e-23], [338.091, 1.43308e-23, 7.91059e-24, 3.76698e-24, 2.74397e-24, 8.7612e-24, 1.37893e-23], [338.86, 1.43621e-23, 7.91338e-24, 3.76774e-24, 2.74478e-24, 8.80409e-24, 1.37796e-23], [339.631, 1.43936e-23, 7.91618e-24, 3.7685e-24, 2.74563e-24, 8.84716e-24, 1.37699e-23], [340.403, 1.44251e-23, 7.919e-24, 3.76927e-24, 2.74652e-24, 8.89042e-24, 1.37601e-23], [341.178, 1.44567e-23, 7.92183e-24, 3.77005e-24, 2.74743e-24, 8.93386e-24, 1.37503e-23], [341.954, 1.44884e-23, 7.92467e-24, 3.77083e-24, 2.74838e-24, 8.97749e-24, 1.37405e-23], [342.732, 1.45202e-23, 7.92753e-24, 3.77162e-24, 2.74936e-24, 9.0213e-24, 1.37306e-23], [343.511, 1.4552e-23, 7.9304e-24, 3.77241e-24, 2.75038e-24, 9.0653e-24, 1.37207e-23], [344.293, 1.45839e-23, 7.93329e-24, 3.77322e-24, 2.75143e-24, 9.10948e-24, 1.37107e-23], [345.076, 1.46159e-23, 7.93618e-24, 3.77403e-24, 2.75252e-24, 9.15384e-24, 1.37007e-23], [345.861, 1.46479e-23, 7.93909e-24, 3.77484e-24, 2.75364e-24, 9.19838e-24, 1.36906e-23], [346.647, 1.46801e-23, 7.94202e-24, 3.77567e-24, 2.7548e-24, 9.24311e-24, 1.36805e-23], [347.436, 1.47123e-23, 7.94496e-24, 3.7765e-24, 2.75599e-24, 9.28803e-24, 1.36703e-23], [348.226, 1.47446e-23, 7.94791e-24, 3.77734e-24, 2.75721e-24, 9.33312e-24, 1.36601e-23], [349.018, 1.47769e-23, 7.95088e-24, 3.77818e-24, 2.75848e-24, 9.3784e-24, 1.36499e-23], [349.812, 1.48094e-23, 7.95386e-24, 3.77903e-24, 2.75978e-24, 9.42386e-24, 1.36396e-23], [350.608, 1.48419e-23, 7.95685e-24, 3.77989e-24, 2.76111e-24, 9.4695e-24, 1.36292e-23], [351.405, 1.48745e-23, 7.95986e-24, 3.78075e-24, 2.76248e-24, 9.51533e-24, 1.36189e-23], [352.204, 1.49071e-23, 7.96288e-24, 3.78163e-24, 2.76389e-24, 9.56133e-24, 1.36084e-23], [353.005, 1.49399e-23, 7.96592e-24, 3.7825e-24, 2.76534e-24, 9.60752e-24, 1.35979e-23], [353.808, 1.49727e-23, 7.96897e-24, 3.78339e-24, 2.76682e-24, 9.65389e-24, 1.35874e-23], [354.613, 1.50056e-23, 7.97204e-24, 3.78428e-24, 2.76834e-24, 9.70044e-24, 1.35768e-23], [355.42, 1.50386e-23, 7.97512e-24, 3.78518e-24, 2.7699e-24, 9.74717e-24, 1.35662e-23], [356.228, 1.50716e-23, 7.97821e-24, 3.78609e-24, 2.77149e-24, 9.79409e-24, 1.35555e-23], [357.038, 1.51048e-23, 7.98132e-24, 3.78701e-24, 2.77313e-24, 9.84118e-24, 1.35448e-23], [357.851, 1.5138e-23, 7.98444e-24, 3.78793e-24, 2.7748e-24, 9.88846e-24, 1.35341e-23], [358.664, 1.51713e-23, 7.98758e-24, 3.78886e-24, 2.77651e-24, 9.93591e-24, 1.35232e-23], [359.48, 1.52046e-23, 7.99073e-24, 3.78979e-24, 2.77826e-24, 9.98355e-24, 1.35124e-23], [360.298, 1.52381e-23, 7.9939e-24, 3.79073e-24, 2.78005e-24, 1.00314e-23, 1.35015e-23], [361.117, 1.52716e-23, 7.99708e-24, 3.79168e-24, 2.78188e-24, 1.00794e-23, 1.34905e-23], [361.939, 1.53052e-23, 8.00027e-24, 3.79264e-24, 2.78375e-24, 1.01275e-23, 1.34795e-23], [362.762, 1.53389e-23, 8.00348e-24, 3.79361e-24, 2.78566e-24, 1.01759e-23, 1.34684e-23], [363.587, 1.53727e-23, 8.00671e-24, 3.79458e-24, 2.78761e-24, 1.02244e-23, 1.34573e-23], [364.414, 1.54065e-23, 8.00995e-24, 3.79556e-24, 2.7896e-24, 1.02731e-23, 1.34462e-23], [365.243, 1.54404e-23, 8.0132e-24, 3.79654e-24, 2.79163e-24, 1.0322e-23, 1.3435e-23], [366.074, 1.54744e-23, 8.01647e-24, 3.79754e-24, 2.7937e-24, 1.03711e-23, 1.34237e-23], [366.907, 1.55085e-23, 8.01976e-24, 3.79854e-24, 2.79582e-24, 1.04204e-23, 1.34124e-23], [367.741, 1.55426e-23, 8.02306e-24, 3.79955e-24, 2.79797e-24, 1.04698e-23, 1.34011e-23], [368.578, 1.55769e-23, 8.02637e-24, 3.80056e-24, 2.80017e-24, 1.05194e-23, 1.33897e-23], [369.416, 1.56112e-23, 8.0297e-24, 3.80159e-24, 2.80241e-24, 1.05692e-23, 1.33783e-23], [370.256, 1.56456e-23, 8.03305e-24, 3.80262e-24, 2.80469e-24, 1.06192e-23, 1.33668e-23], [371.098, 1.56801e-23, 8.03641e-24, 3.80366e-24, 2.80701e-24, 1.06693e-23, 1.33552e-23], [371.942, 1.57147e-23, 8.03979e-24, 3.8047e-24, 2.80938e-24, 1.07196e-23, 1.33436e-23], [372.788, 1.57493e-23, 8.04318e-24, 3.80576e-24, 2.81179e-24, 1.07702e-23, 1.3332e-23], [373.636, 1.5784e-23, 8.04659e-24, 3.80682e-24, 2.81424e-24, 1.08208e-23, 1.33203e-23], [374.486, 1.58188e-23, 8.05001e-24, 3.80789e-24, 2.81674e-24, 1.08717e-23, 1.33085e-23], [375.338, 1.58537e-23, 8.05345e-24, 3.80896e-24, 2.81928e-24, 1.09227e-23, 1.32967e-23], [376.192, 1.58887e-23, 8.0569e-24, 3.81005e-24, 2.82186e-24, 1.0974e-23, 1.32849e-23], [377.047, 1.59237e-23, 8.06037e-24, 3.81114e-24, 2.82449e-24, 1.10254e-23, 1.3273e-23], [377.905, 1.59589e-23, 8.06385e-24, 3.81224e-24, 2.82716e-24, 1.10769e-23, 1.3261e-23], [378.765, 1.59941e-23, 8.06735e-24, 3.81335e-24, 2.82987e-24, 1.11287e-23, 1.3249e-23], [379.626, 1.60294e-23, 8.07087e-24, 3.81446e-24, 2.83263e-24, 1.11806e-23, 1.32369e-23], [380.49, 1.60648e-23, 8.0744e-24, 3.81558e-24, 2.83544e-24, 1.12327e-23, 1.32248e-23], [381.355, 1.61002e-23, 8.07795e-24, 3.81671e-24, 2.83829e-24, 1.1285e-23, 1.32127e-23], [382.223, 1.61358e-23, 8.08152e-24, 3.81785e-24, 2.84119e-24, 1.13375e-23, 1.32005e-23], [383.092, 1.61714e-23, 8.0851e-24, 3.819e-24, 2.84413e-24, 1.13901e-23, 1.31882e-23], [383.963, 1.62071e-23, 8.08869e-24, 3.82015e-24, 2.84712e-24, 1.14429e-23, 1.31759e-23], [384.837, 1.62429e-23, 8.09231e-24, 3.82132e-24, 2.85016e-24, 1.14959e-23, 1.31635e-23], [385.712, 1.62788e-23, 8.09594e-24, 3.82249e-24, 2.85324e-24, 1.15491e-23, 1.31511e-23], [386.589, 1.63148e-23, 8.09958e-24, 3.82366e-24, 2.85636e-24, 1.16024e-23, 1.31386e-23], [387.469, 1.63508e-23, 8.10324e-24, 3.82485e-24, 2.85954e-24, 1.1656e-23, 1.31261e-23], [388.35, 1.63869e-23, 8.10692e-24, 3.82604e-24, 2.86276e-24, 1.17097e-23, 1.31135e-23], [389.233, 1.64232e-23, 8.11062e-24, 3.82725e-24, 2.86603e-24, 1.17635e-23, 1.31008e-23], [390.119, 1.64595e-23, 8.11433e-24, 3.82846e-24, 2.86934e-24, 1.18176e-23, 1.30881e-23], [391.006, 1.64959e-23, 8.11805e-24, 3.82967e-24, 2.87271e-24, 1.18718e-23, 1.30754e-23], [391.895, 1.65323e-23, 8.1218e-24, 3.8309e-24, 2.87612e-24, 1.19262e-23, 1.30626e-23], [392.787, 1.65689e-23, 8.12556e-24, 3.83213e-24, 2.87958e-24, 1.19808e-23, 1.30498e-23], [393.68, 1.66055e-23, 8.12934e-24, 3.83338e-24, 2.88309e-24, 1.20356e-23, 1.30368e-23], [394.576, 1.66423e-23, 8.13313e-24, 3.83463e-24, 2.88665e-24, 1.20905e-23, 1.30239e-23], [395.473, 1.66791e-23, 8.13694e-24, 3.83589e-24, 2.89025e-24, 1.21456e-23, 1.30109e-23], [396.373, 1.6716e-23, 8.14077e-24, 3.83715e-24, 2.89391e-24, 1.22009e-23, 1.29978e-23], [397.274, 1.6753e-23, 8.14462e-24, 3.83843e-24, 2.89761e-24, 1.22564e-23, 1.29847e-23], [398.178, 1.679e-23, 8.14848e-24, 3.83971e-24, 2.90136e-24, 1.2312e-23, 1.29715e-23], [399.084, 1.68272e-23, 8.15236e-24, 3.841e-24, 2.90516e-24, 1.23679e-23, 1.29582e-23], [399.991, 1.68645e-23, 8.15626e-24, 3.84231e-24, 2.90902e-24, 1.24239e-23, 1.29449e-23], [400.901, 1.69018e-23, 8.16017e-24, 3.84361e-24, 2.91292e-24, 1.248e-23, 1.29316e-23], [401.813, 1.69392e-23, 8.1641e-24, 3.84493e-24, 2.91687e-24, 1.25364e-23, 1.29182e-23], [402.727, 1.69767e-23, 8.16805e-24, 3.84626e-24, 2.92087e-24, 1.25929e-23, 1.29047e-23], [403.643, 1.70143e-23, 8.17202e-24, 3.84759e-24, 2.92492e-24, 1.26496e-23, 1.28912e-23], [404.561, 1.7052e-23, 8.176e-24, 3.84893e-24, 2.92903e-24, 1.27065e-23, 1.28776e-23], [405.481, 1.70898e-23, 8.18e-24, 3.85029e-24, 2.93318e-24, 1.27635e-23, 1.2864e-23], [406.404, 1.71276e-23, 8.18402e-24, 3.85165e-24, 2.93739e-24, 1.28208e-23, 1.28503e-23], [407.328, 1.71656e-23, 8.18806e-24, 3.85301e-24, 2.94164e-24, 1.28782e-23, 1.28366e-23], [408.254, 1.72036e-23, 8.19211e-24, 3.85439e-24, 2.94595e-24, 1.29358e-23, 1.28228e-23], [409.183, 1.72417e-23, 8.19618e-24, 3.85578e-24, 2.95031e-24, 1.29935e-23, 1.28089e-23], [410.114, 1.728e-23, 8.20027e-24, 3.85717e-24, 2.95472e-24, 1.30515e-23, 1.2795e-23], [411.047, 1.73183e-23, 8.20438e-24, 3.85857e-24, 2.95919e-24, 1.31096e-23, 1.2781e-23], [411.982, 1.73567e-23, 8.20851e-24, 3.85999e-24, 2.9637e-24, 1.31679e-23, 1.2767e-23], [412.919, 1.73951e-23, 8.21265e-24, 3.86141e-24, 2.96827e-24, 1.32263e-23, 1.27529e-23], [413.858, 1.74337e-23, 8.21681e-24, 3.86284e-24, 2.97289e-24, 1.3285e-23, 1.27387e-23], [414.799, 1.74724e-23, 8.22099e-24, 3.86428e-24, 2.97756e-24, 1.33438e-23, 1.27245e-23], [415.743, 1.75111e-23, 8.22519e-24, 3.86572e-24, 2.98229e-24, 1.34028e-23, 1.27103e-23], [416.688, 1.755e-23, 8.22941e-24, 3.86718e-24, 2.98707e-24, 1.34619e-23, 1.26959e-23], [417.636, 1.75889e-23, 8.23364e-24, 3.86864e-24, 2.9919e-24, 1.35213e-23, 1.26815e-23], [418.586, 1.76279e-23, 8.23789e-24, 3.87012e-24, 2.99679e-24, 1.35808e-23, 1.26671e-23], [419.538, 1.76671e-23, 8.24217e-24, 3.8716e-24, 3.00172e-24, 1.36405e-23, 1.26526e-23], [420.492, 1.77063e-23, 8.24646e-24, 3.87309e-24, 3.00672e-24, 1.37004e-23, 1.2638e-23], [421.449, 1.77456e-23, 8.25077e-24, 3.8746e-24, 3.01176e-24, 1.37604e-23, 1.26234e-23], [422.407, 1.77849e-23, 8.25509e-24, 3.87611e-24, 3.01686e-24, 1.38207e-23, 1.26087e-23], [423.368, 1.78244e-23, 8.25944e-24, 3.87763e-24, 3.02202e-24, 1.38811e-23, 1.2594e-23], [424.331, 1.7864e-23, 8.26381e-24, 3.87916e-24, 3.02723e-24, 1.39416e-23, 1.25792e-23], [425.296, 1.79037e-23, 8.26819e-24, 3.88069e-24, 3.03249e-24, 1.40024e-23, 1.25643e-23], [426.264, 1.79434e-23, 8.27259e-24, 3.88224e-24, 3.03781e-24, 1.40633e-23, 1.25494e-23], [427.233, 1.79833e-23, 8.27702e-24, 3.8838e-24, 3.04318e-24, 1.41244e-23, 1.25344e-23], [428.205, 1.80232e-23, 8.28146e-24, 3.88537e-24, 3.04861e-24, 1.41857e-23, 1.25194e-23], [429.179, 1.80633e-23, 8.28592e-24, 3.88694e-24, 3.05409e-24, 1.42472e-23, 1.25042e-23], [430.155, 1.81034e-23, 8.2904e-24, 3.88853e-24, 3.05963e-24, 1.43088e-23, 1.24891e-23], [431.134, 1.81436e-23, 8.2949e-24, 3.89012e-24, 3.06523e-24, 1.43706e-23, 1.24738e-23], [432.114, 1.81839e-23, 8.29942e-24, 3.89173e-24, 3.07088e-24, 1.44326e-23, 1.24586e-23], [433.097, 1.82243e-23, 8.30396e-24, 3.89334e-24, 3.07658e-24, 1.44948e-23, 1.24432e-23], [434.082, 1.82648e-23, 8.30851e-24, 3.89497e-24, 3.08234e-24, 1.45571e-23, 1.24278e-23], [435.07, 1.83054e-23, 8.31309e-24, 3.8966e-24, 3.08816e-24, 1.46196e-23, 1.24123e-23], [436.059, 1.83461e-23, 8.31769e-24, 3.89824e-24, 3.09403e-24, 1.46823e-23, 1.23968e-23], [437.051, 1.83869e-23, 8.32231e-24, 3.8999e-24, 3.09997e-24, 1.47452e-23, 1.23812e-23], [438.045, 1.84278e-23, 8.32695e-24, 3.90156e-24, 3.10595e-24, 1.48083e-23, 1.23655e-23], [439.042, 1.84688e-23, 8.3316e-24, 3.90324e-24, 3.112e-24, 1.48715e-23, 1.23498e-23], [440.04, 1.85098e-23, 8.33628e-24, 3.90492e-24, 3.1181e-24, 1.49349e-23, 1.2334e-23], [441.041, 1.8551e-23, 8.34098e-24, 3.90662e-24, 3.12426e-24, 1.49985e-23, 1.23181e-23], [442.044, 1.85923e-23, 8.3457e-24, 3.90833e-24, 3.13047e-24, 1.50622e-23, 1.23022e-23], [443.05, 1.86336e-23, 8.35044e-24, 3.91004e-24, 3.13675e-24, 1.51261e-23, 1.22862e-23], [444.058, 1.86751e-23, 8.3552e-24, 3.91177e-24, 3.14308e-24, 1.51903e-23, 1.22702e-23], [445.068, 1.87166e-23, 8.35998e-24, 3.91351e-24, 3.14947e-24, 1.52545e-23, 1.22541e-23], [446.08, 1.87583e-23, 8.36478e-24, 3.91526e-24, 3.15592e-24, 1.5319e-23, 1.22379e-23], [447.095, 1.88e-23, 8.3696e-24, 3.91702e-24, 3.16243e-24, 1.53836e-23, 1.22216e-23], [448.111, 1.88419e-23, 8.37445e-24, 3.91879e-24, 3.169e-24, 1.54485e-23, 1.22053e-23], [449.131, 1.88838e-23, 8.37931e-24, 3.92058e-24, 3.17563e-24, 1.55135e-23, 1.2189e-23], [450.152, 1.89259e-23, 8.3842e-24, 3.92238e-24, 3.18232e-24, 1.55786e-23, 1.21725e-23], [451.176, 1.8968e-23, 8.38911e-24, 3.92419e-24, 3.18906e-24, 1.5644e-23, 1.2156e-23], [452.202, 1.90102e-23, 8.39404e-24, 3.92601e-24, 3.19588e-24, 1.57095e-23, 1.21395e-23], [453.231, 1.90526e-23, 8.399e-24, 3.92785e-24, 3.20275e-24, 1.57752e-23, 1.21229e-23], [454.262, 1.9095e-23, 8.40397e-24, 3.9297e-24, 3.20969e-24, 1.58411e-23, 1.21062e-23], [455.295, 1.91375e-23, 8.40897e-24, 3.93157e-24, 3.21669e-24, 1.59072e-23, 1.20894e-23], [456.331, 1.91802e-23, 8.414e-24, 3.93346e-24, 3.22375e-24, 1.59734e-23, 1.20726e-23], [457.369, 1.92229e-23, 8.41905e-24, 3.93536e-24, 3.23089e-24, 1.60398e-23, 1.20557e-23], [458.409, 1.92658e-23, 8.42412e-24, 3.93728e-24, 3.23809e-24, 1.61064e-23, 1.20388e-23], [459.452, 1.93087e-23, 8.42922e-24, 3.93923e-24, 3.24536e-24, 1.61732e-23, 1.20217e-23], [460.497, 1.93517e-23, 8.43435e-24, 3.94119e-24, 3.25271e-24, 1.62402e-23, 1.20047e-23], [461.544, 1.93949e-23, 8.43951e-24, 3.94319e-24, 3.26013e-24, 1.63073e-23, 1.19875e-23], [462.594, 1.94381e-23, 8.4447e-24, 3.94521e-24, 3.26764e-24, 1.63747e-23, 1.19703e-23], [463.646, 1.94815e-23, 8.44992e-24, 3.94727e-24, 3.27524e-24, 1.64422e-23, 1.19531e-23], [464.701, 1.95249e-23, 8.45518e-24, 3.94938e-24, 3.28293e-24, 1.65099e-23, 1.19358e-23], [465.758, 1.95685e-23, 8.46048e-24, 3.95154e-24, 3.29074e-24, 1.65777e-23, 1.19184e-23], [466.817, 1.96122e-23, 8.46583e-24, 3.95377e-24, 3.29867e-24, 1.66458e-23, 1.1901e-23], [467.879, 1.9656e-23, 8.47123e-24, 3.95608e-24, 3.30675e-24, 1.67141e-23, 1.18835e-23], [468.943, 1.96999e-23, 8.47671e-24, 3.95851e-24, 3.31502e-24, 1.67826e-23, 1.18661e-23], [470.01, 1.97439e-23, 8.48229e-24, 3.96111e-24, 3.32353e-24, 1.68513e-23, 1.18486e-23], [471.079, 1.97881e-23, 8.48799e-24, 3.96396e-24, 3.33237e-24, 1.69202e-23, 1.18311e-23], [472.15, 1.98325e-23, 8.49388e-24, 3.96716e-24, 3.34169e-24, 1.69894e-23, 1.18136e-23], [473.224, 1.98771e-23, 8.50007e-24, 3.97097e-24, 3.35175e-24, 1.70589e-23, 1.17963e-23], [474.301, 1.9922e-23, 8.50677e-24, 3.97583e-24, 3.3631e-24, 1.71288e-23, 1.17793e-23], [475.38, 1.99674e-23, 8.51443e-24, 3.98271e-24, 3.37687e-24, 1.71993e-23, 1.17629e-23], [476.461, 2.00138e-23, 8.52423e-24, 3.99412e-24, 3.39599e-24, 1.72711e-23, 1.17479e-23], [477.545, 2.00628e-23, 8.53996e-24, 4.01809e-24, 3.42982e-24, 1.73459e-23, 1.17372e-23], [478.631, 2.01223e-23, 8.58017e-24, 4.0935e-24, 3.52345e-24, 1.74329e-23, 1.17443e-23], [479.72, 2.0307e-23, 8.90924e-24, 4.73656e-24, 4.25826e-24, 1.7664e-23, 1.1965e-23], [480.811, 2.05924e-23, 9.45131e-24, 5.68421e-24, 5.29606e-24, 1.80094e-23, 1.23528e-23], [481.904, 2.02644e-23, 8.61397e-24, 4.13592e-24, 3.5895e-24, 1.76521e-23, 1.17021e-23], [483.0, 2.02887e-23, 8.57031e-24, 4.03453e-24, 3.47811e-24, 1.76988e-23, 1.16474e-23], [484.099, 2.03287e-23, 8.56328e-24, 4.00982e-24, 3.45544e-24, 1.77635e-23, 1.16195e-23], [485.2, 2.03719e-23, 8.56388e-24, 4.00126e-24, 3.45159e-24, 1.78322e-23, 1.15971e-23], [486.304, 2.04163e-23, 8.56702e-24, 3.99809e-24, 3.45408e-24, 1.79022e-23, 1.15766e-23], [487.41, 2.04613e-23, 8.57126e-24, 3.99724e-24, 3.45931e-24, 1.79729e-23, 1.15567e-23], [488.519, 2.05067e-23, 8.57606e-24, 3.99753e-24, 3.46592e-24, 1.80441e-23, 1.15371e-23], [489.63, 2.05522e-23, 8.58117e-24, 3.99846e-24, 3.47331e-24, 1.81156e-23, 1.15177e-23], [490.743, 2.05979e-23, 8.58648e-24, 3.99978e-24, 3.48121e-24, 1.81874e-23, 1.14983e-23], [491.86, 2.06438e-23, 8.59192e-24, 4.00134e-24, 3.48944e-24, 1.82594e-23, 1.1479e-23], [492.978, 2.06898e-23, 8.59746e-24, 4.00307e-24, 3.4979e-24, 1.83316e-23, 1.14596e-23], [494.1, 2.0736e-23, 8.60307e-24, 4.00492e-24, 3.50656e-24, 1.8404e-23, 1.14402e-23], [495.224, 2.07823e-23, 8.60874e-24, 4.00686e-24, 3.51536e-24, 1.84766e-23, 1.14208e-23], [496.35, 2.08286e-23, 8.61446e-24, 4.00886e-24, 3.52428e-24, 1.85494e-23, 1.14012e-23], [497.479, 2.08751e-23, 8.62022e-24, 4.01092e-24, 3.53332e-24, 1.86224e-23, 1.13817e-23], [498.611, 2.09218e-23, 8.62602e-24, 4.01302e-24, 3.54244e-24, 1.86955e-23, 1.1362e-23], [499.745, 2.09685e-23, 8.63186e-24, 4.01515e-24, 3.55166e-24, 1.87689e-23, 1.13423e-23], [500.881, 2.10153e-23, 8.63773e-24, 4.01732e-24, 3.56096e-24, 1.88425e-23, 1.13226e-23], [502.021, 2.10623e-23, 8.64363e-24, 4.01951e-24, 3.57033e-24, 1.89162e-23, 1.13027e-23], [503.163, 2.11093e-23, 8.64956e-24, 4.02172e-24, 3.57978e-24, 1.89901e-23, 1.12828e-23], [504.307, 2.11565e-23, 8.65552e-24, 4.02396e-24, 3.5893e-24, 1.90642e-23, 1.12628e-23], [505.454, 2.12038e-23, 8.66151e-24, 4.02622e-24, 3.59889e-24, 1.91385e-23, 1.12428e-23], [506.604, 2.12512e-23, 8.66752e-24, 4.02849e-24, 3.60855e-24, 1.9213e-23, 1.12227e-23], [507.756, 2.12987e-23, 8.67356e-24, 4.03078e-24, 3.61827e-24, 1.92876e-23, 1.12025e-23], [508.911, 2.13463e-23, 8.67963e-24, 4.03309e-24, 3.62806e-24, 1.93625e-23, 1.11822e-23], [510.069, 2.1394e-23, 8.68572e-24, 4.03542e-24, 3.63791e-24, 1.94375e-23, 1.11618e-23], [511.229, 2.14419e-23, 8.69184e-24, 4.03775e-24, 3.64783e-24, 1.95127e-23, 1.11414e-23], [512.392, 2.14898e-23, 8.69799e-24, 4.04011e-24, 3.65781e-24, 1.95881e-23, 1.11209e-23], [513.557, 2.15379e-23, 8.70416e-24, 4.04248e-24, 3.66785e-24, 1.96636e-23, 1.11003e-23], [514.725, 2.15861e-23, 8.71036e-24, 4.04486e-24, 3.67796e-24, 1.97394e-23, 1.10797e-23], [515.896, 2.16343e-23, 8.71658e-24, 4.04725e-24, 3.68812e-24, 1.98153e-23, 1.1059e-23], [517.069, 2.16827e-23, 8.72283e-24, 4.04966e-24, 3.69835e-24, 1.98914e-23, 1.10382e-23], [518.246, 2.17312e-23, 8.72911e-24, 4.05208e-24, 3.70864e-24, 1.99677e-23, 1.10173e-23], [519.424, 2.17799e-23, 8.73541e-24, 4.05452e-24, 3.71899e-24, 2.00442e-23, 1.09963e-23], [520.606, 2.18286e-23, 8.74173e-24, 4.05697e-24, 3.7294e-24, 2.01209e-23, 1.09753e-23], [521.79, 2.18774e-23, 8.74809e-24, 4.05943e-24, 3.73987e-24, 2.01977e-23, 1.09542e-23], [522.977, 2.19264e-23, 8.75446e-24, 4.0619e-24, 3.7504e-24, 2.02747e-23, 1.0933e-23], [524.166, 2.19754e-23, 8.76087e-24, 4.06439e-24, 3.76099e-24, 2.0352e-23, 1.09117e-23], [525.359, 2.20246e-23, 8.7673e-24, 4.06689e-24, 3.77165e-24, 2.04294e-23, 1.08904e-23], [526.554, 2.20739e-23, 8.77375e-24, 4.0694e-24, 3.78236e-24, 2.05069e-23, 1.08689e-23], [527.751, 2.21233e-23, 8.78023e-24, 4.07192e-24, 3.79313e-24, 2.05847e-23, 1.08474e-23], [528.952, 2.21729e-23, 8.78674e-24, 4.07446e-24, 3.80397e-24, 2.06627e-23, 1.08258e-23], [530.155, 2.22225e-23, 8.79327e-24, 4.07701e-24, 3.81486e-24, 2.07408e-23, 1.08042e-23], [531.361, 2.22722e-23, 8.79983e-24, 4.07957e-24, 3.82582e-24, 2.08191e-23, 1.07824e-23], [532.569, 2.23221e-23, 8.80642e-24, 4.08215e-24, 3.83683e-24, 2.08976e-23, 1.07606e-23], [533.781, 2.23721e-23, 8.81303e-24, 4.08474e-24, 3.8479e-24, 2.09763e-23, 1.07387e-23], [534.995, 2.24222e-23, 8.81967e-24, 4.08734e-24, 3.85904e-24, 2.10552e-23, 1.07167e-23], [536.212, 2.24724e-23, 8.82633e-24, 4.08995e-24, 3.87023e-24, 2.11343e-23, 1.06946e-23], [537.431, 2.25227e-23, 8.83302e-24, 4.09258e-24, 3.88148e-24, 2.12135e-23, 1.06725e-23], [538.654, 2.25732e-23, 8.83974e-24, 4.09521e-24, 3.8928e-24, 2.12929e-23, 1.06503e-23], [539.879, 2.26237e-23, 8.84648e-24, 4.09786e-24, 3.90417e-24, 2.13725e-23, 1.0628e-23], [541.107, 2.26744e-23, 8.85325e-24, 4.10053e-24, 3.9156e-24, 2.14523e-23, 1.06056e-23], [542.338, 2.27252e-23, 8.86005e-24, 4.1032e-24, 3.9271e-24, 2.15323e-23, 1.05831e-23], [543.571, 2.27761e-23, 8.86688e-24, 4.10589e-24, 3.93865e-24, 2.16125e-23, 1.05606e-23], [544.808, 2.28271e-23, 8.87373e-24, 4.1086e-24, 3.95026e-24, 2.16929e-23, 1.05379e-23], [546.047, 2.28783e-23, 8.8806e-24, 4.11131e-24, 3.96193e-24, 2.17734e-23, 1.05152e-23], [547.289, 2.29296e-23, 8.88751e-24, 4.11404e-24, 3.97366e-24, 2.18541e-23, 1.04924e-23], [548.534, 2.29809e-23, 8.89444e-24, 4.11678e-24, 3.98545e-24, 2.19351e-23, 1.04695e-23], [549.781, 2.30324e-23, 8.9014e-24, 4.11953e-24, 3.9973e-24, 2.20162e-23, 1.04466e-23], [551.032, 2.30841e-23, 8.90838e-24, 4.1223e-24, 4.00921e-24, 2.20975e-23, 1.04235e-23], [552.285, 2.31358e-23, 8.9154e-24, 4.12507e-24, 4.02117e-24, 2.2179e-23, 1.04004e-23], [553.542, 2.31877e-23, 8.92244e-24, 4.12787e-24, 4.0332e-24, 2.22606e-23, 1.03772e-23], [554.801, 2.32396e-23, 8.9295e-24, 4.13067e-24, 4.04529e-24, 2.23425e-23, 1.03539e-23], [556.063, 2.32917e-23, 8.9366e-24, 4.13349e-24, 4.05743e-24, 2.24245e-23, 1.03305e-23], [557.327, 2.33439e-23, 8.94372e-24, 4.13632e-24, 4.06964e-24, 2.25067e-23, 1.0307e-23], [558.595, 2.33963e-23, 8.95087e-24, 4.13916e-24, 4.0819e-24, 2.25892e-23, 1.02835e-23], [559.866, 2.34487e-23, 8.95805e-24, 4.14202e-24, 4.09423e-24, 2.26718e-23, 1.02599e-23], [561.139, 2.35013e-23, 8.96526e-24, 4.14489e-24, 4.10661e-24, 2.27546e-23, 1.02362e-23], [562.415, 2.3554e-23, 8.97249e-24, 4.14777e-24, 4.11905e-24, 2.28376e-23, 1.02124e-23], [563.695, 2.36068e-23, 8.97976e-24, 4.15067e-24, 4.13155e-24, 2.29207e-23, 1.01885e-23], [564.977, 2.36598e-23, 8.98705e-24, 4.15358e-24, 4.14411e-24, 2.30041e-23, 1.01645e-23], [566.262, 2.37128e-23, 8.99436e-24, 4.1565e-24, 4.15673e-24, 2.30876e-23, 1.01404e-23], [567.55, 2.3766e-23, 9.00171e-24, 4.15943e-24, 4.16941e-24, 2.31714e-23, 1.01163e-23], [568.841, 2.38193e-23, 9.00909e-24, 4.16238e-24, 4.18214e-24, 2.32553e-23, 1.00921e-23], [570.135, 2.38728e-23, 9.01649e-24, 4.16535e-24, 4.19494e-24, 2.33394e-23, 1.00678e-23], [571.432, 2.39263e-23, 9.02392e-24, 4.16832e-24, 4.20779e-24, 2.34237e-23, 1.00434e-23], [572.731, 2.398e-23, 9.03138e-24, 4.17131e-24, 4.22071e-24, 2.35082e-23, 1.00189e-23], [574.034, 2.40338e-23, 9.03887e-24, 4.17431e-24, 4.23368e-24, 2.35929e-23, 9.99433e-24], [575.34, 2.40877e-23, 9.04639e-24, 4.17733e-24, 4.24671e-24, 2.36778e-23, 9.96967e-24], [576.648, 2.41418e-23, 9.05394e-24, 4.18036e-24, 4.2598e-24, 2.37629e-23, 9.94493e-24], [577.96, 2.41959e-23, 9.06152e-24, 4.1834e-24, 4.27295e-24, 2.38481e-23, 9.92011e-24], [579.275, 2.42502e-23, 9.06912e-24, 4.18646e-24, 4.28615e-24, 2.39336e-23, 9.89519e-24], [580.592, 2.43046e-23, 9.07675e-24, 4.18953e-24, 4.29942e-24, 2.40192e-23, 9.87019e-24], [581.913, 2.43592e-23, 9.08442e-24, 4.19261e-24, 4.31274e-24, 2.41051e-23, 9.8451e-24], [583.236, 2.44139e-23, 9.09211e-24, 4.19571e-24, 4.32613e-24, 2.41911e-23, 9.81993e-24], [584.563, 2.44687e-23, 9.09983e-24, 4.19882e-24, 4.33957e-24, 2.42773e-23, 9.79467e-24], [585.893, 2.45236e-23, 9.10759e-24, 4.20195e-24, 4.35307e-24, 2.43637e-23, 9.76932e-24], [587.225, 2.45786e-23, 9.11537e-24, 4.20509e-24, 4.36663e-24, 2.44503e-23, 9.74388e-24], [588.561, 2.46338e-23, 9.12318e-24, 4.20824e-24, 4.38024e-24, 2.45371e-23, 9.71835e-24], [589.9, 2.46891e-23, 9.13102e-24, 4.21141e-24, 4.39392e-24, 2.46241e-23, 9.69274e-24], [591.241, 2.47445e-23, 9.13889e-24, 4.21459e-24, 4.40765e-24, 2.47113e-23, 9.66703e-24], [592.586, 2.48001e-23, 9.14679e-24, 4.21778e-24, 4.42145e-24, 2.47986e-23, 9.64124e-24], [593.934, 2.48558e-23, 9.15472e-24, 4.22099e-24, 4.4353e-24, 2.48862e-23, 9.61536e-24], [595.285, 2.49116e-23, 9.16268e-24, 4.22421e-24, 4.44921e-24, 2.4974e-23, 9.58939e-24], [596.639, 2.49675e-23, 9.17067e-24, 4.22745e-24, 4.46317e-24, 2.50619e-23, 9.56334e-24], [597.996, 2.50236e-23, 9.17869e-24, 4.2307e-24, 4.4772e-24, 2.51501e-23, 9.53719e-24], [599.356, 2.50798e-23, 9.18674e-24, 4.23397e-24, 4.49128e-24, 2.52384e-23, 9.51095e-24], [600.72, 2.51361e-23, 9.19482e-24, 4.23725e-24, 4.50542e-24, 2.53269e-23, 9.48463e-24], [602.086, 2.51926e-23, 9.20294e-24, 4.24054e-24, 4.51962e-24, 2.54157e-23, 9.45821e-24], [603.456, 2.52492e-23, 9.21108e-24, 4.24385e-24, 4.53388e-24, 2.55046e-23, 9.43171e-24], [604.828, 2.53059e-23, 9.21925e-24, 4.24717e-24, 4.5482e-24, 2.55937e-23, 9.40511e-24], [606.204, 2.53627e-23, 9.22746e-24, 4.25051e-24, 4.56257e-24, 2.5683e-23, 9.37843e-24], [607.583, 2.54197e-23, 9.23569e-24, 4.25386e-24, 4.577e-24, 2.57725e-23, 9.35165e-24], [608.965, 2.54768e-23, 9.24396e-24, 4.25723e-24, 4.5915e-24, 2.58622e-23, 9.32479e-24], [610.35, 2.55341e-23, 9.25225e-24, 4.26061e-24, 4.60604e-24, 2.59521e-23, 9.29783e-24], [611.738, 2.55914e-23, 9.26058e-24, 4.264e-24, 4.62065e-24, 2.60422e-23, 9.27078e-24], [613.13, 2.56489e-23, 9.26894e-24, 4.26741e-24, 4.63531e-24, 2.61325e-23, 9.24365e-24], [614.524, 2.57066e-23, 9.27733e-24, 4.27083e-24, 4.65004e-24, 2.6223e-23, 9.21642e-24], [615.922, 2.57643e-23, 9.28575e-24, 4.27427e-24, 4.66482e-24, 2.63137e-23, 9.1891e-24], [617.323, 2.58222e-23, 9.29421e-24, 4.27772e-24, 4.67965e-24, 2.64046e-23, 9.16169e-24], [618.727, 2.58803e-23, 9.30269e-24, 4.28119e-24, 4.69455e-24, 2.64957e-23, 9.13419e-24], [620.134, 2.59384e-23, 9.31121e-24, 4.28468e-24, 4.7095e-24, 2.65869e-23, 9.10659e-24], [621.545, 2.59967e-23, 9.31976e-24, 4.28817e-24, 4.72451e-24, 2.66784e-23, 9.07891e-24], [622.959, 2.60552e-23, 9.32834e-24, 4.29168e-24, 4.73958e-24, 2.67701e-23, 9.05113e-24], [624.376, 2.61137e-23, 9.33695e-24, 4.29521e-24, 4.75471e-24, 2.6862e-23, 9.02326e-24], [625.796, 2.61724e-23, 9.3456e-24, 4.29875e-24, 4.76989e-24, 2.6954e-23, 8.9953e-24], [627.219, 2.62313e-23, 9.35427e-24, 4.30231e-24, 4.78513e-24, 2.70463e-23, 8.96725e-24], [628.646, 2.62902e-23, 9.36298e-24, 4.30588e-24, 4.80043e-24, 2.71388e-23, 8.9391e-24], [630.076, 2.63493e-23, 9.37172e-24, 4.30947e-24, 4.81579e-24, 2.72314e-23, 8.91087e-24], [631.509, 2.64086e-23, 9.3805e-24, 4.31307e-24, 4.8312e-24, 2.73243e-23, 8.88254e-24], [632.945, 2.6468e-23, 9.3893e-24, 4.31669e-24, 4.84667e-24, 2.74174e-23, 8.85411e-24], [634.385, 2.65275e-23, 9.39814e-24, 4.32032e-24, 4.8622e-24, 2.75106e-23, 8.8256e-24], [635.828, 2.65871e-23, 9.40701e-24, 4.32397e-24, 4.87779e-24, 2.76041e-23, 8.79699e-24], [637.274, 2.66469e-23, 9.41592e-24, 4.32763e-24, 4.89343e-24, 2.76978e-23, 8.76829e-24], [638.724, 2.67069e-23, 9.42485e-24, 4.33131e-24, 4.90913e-24, 2.77916e-23, 8.7395e-24], [640.177, 2.67669e-23, 9.43382e-24, 4.335e-24, 4.92489e-24, 2.78857e-23, 8.71061e-24], [641.633, 2.68271e-23, 9.44283e-24, 4.33871e-24, 4.9407e-24, 2.798e-23, 8.68163e-24], [643.092, 2.68875e-23, 9.45186e-24, 4.34244e-24, 4.95658e-24, 2.80744e-23, 8.65256e-24], [644.555, 2.69479e-23, 9.46093e-24, 4.34618e-24, 4.97251e-24, 2.81691e-23, 8.62339e-24], [646.021, 2.70085e-23, 9.47003e-24, 4.34993e-24, 4.98849e-24, 2.8264e-23, 8.59413e-24], [647.49, 2.70693e-23, 9.47917e-24, 4.3537e-24, 5.00454e-24, 2.83591e-23, 8.56478e-24], [648.963, 2.71302e-23, 9.48834e-24, 4.35749e-24, 5.02064e-24, 2.84543e-23, 8.53533e-24], [650.439, 2.71912e-23, 9.49754e-24, 4.36129e-24, 5.0368e-24, 2.85498e-23, 8.50579e-24], [651.919, 2.72524e-23, 9.50678e-24, 4.36511e-24, 5.05301e-24, 2.86455e-23, 8.47615e-24], [653.402, 2.73137e-23, 9.51605e-24, 4.36894e-24, 5.06929e-24, 2.87414e-23, 8.44642e-24], [654.888, 2.73752e-23, 9.52536e-24, 4.37279e-24, 5.08562e-24, 2.88375e-23, 8.4166e-24], [656.377, 2.74368e-23, 9.53469e-24, 4.37666e-24, 5.102e-24, 2.89338e-23, 8.38668e-24], [657.87, 2.74985e-23, 9.54407e-24, 4.38054e-24, 5.11845e-24, 2.90303e-23, 8.35667e-24], [659.367, 2.75604e-23, 9.55347e-24, 4.38443e-24, 5.13495e-24, 2.9127e-23, 8.32657e-24], [660.866, 2.76224e-23, 9.56292e-24, 4.38835e-24, 5.15151e-24, 2.92239e-23, 8.29637e-24], [662.37, 2.76846e-23, 9.57239e-24, 4.39227e-24, 5.16812e-24, 2.9321e-23, 8.26607e-24], [663.876, 2.77469e-23, 9.5819e-24, 4.39622e-24, 5.18479e-24, 2.94183e-23, 8.23568e-24], [665.386, 2.78093e-23, 9.59145e-24, 4.40018e-24, 5.20152e-24, 2.95158e-23, 8.2052e-24], [666.9, 2.78719e-23, 9.60103e-24, 4.40416e-24, 5.21831e-24, 2.96136e-23, 8.17462e-24], [668.417, 2.79346e-23, 9.61064e-24, 4.40815e-24, 5.23515e-24, 2.97115e-23, 8.14395e-24], [669.937, 2.79975e-23, 9.62029e-24, 4.41216e-24, 5.25205e-24, 2.98096e-23, 8.11319e-24], [671.461, 2.80605e-23, 9.62997e-24, 4.41618e-24, 5.269e-24, 2.9908e-23, 8.08233e-24], [672.988, 2.81237e-23, 9.63969e-24, 4.42022e-24, 5.28602e-24, 3.00065e-23, 8.05137e-24], [674.519, 2.8187e-23, 9.64945e-24, 4.42428e-24, 5.30309e-24, 3.01053e-23, 8.02032e-24], [676.053, 2.82505e-23, 9.65923e-24, 4.42835e-24, 5.32021e-24, 3.02043e-23, 7.98918e-24], [677.591, 2.83141e-23, 9.66906e-24, 4.43245e-24, 5.33739e-24, 3.03035e-23, 7.95794e-24], [679.132, 2.83778e-23, 9.67892e-24, 4.43655e-24, 5.35463e-24, 3.04028e-23, 7.92661e-24], [680.677, 2.84417e-23, 9.68881e-24, 4.44068e-24, 5.37193e-24, 3.05024e-23, 7.89518e-24], [682.225, 2.85058e-23, 9.69875e-24, 4.44481e-24, 5.38928e-24, 3.06022e-23, 7.86365e-24], [683.777, 2.85699e-23, 9.70871e-24, 4.44897e-24, 5.40669e-24, 3.07022e-23, 7.83204e-24], [685.332, 2.86343e-23, 9.71872e-24, 4.45314e-24, 5.42416e-24, 3.08025e-23, 7.80033e-24], [686.891, 2.86988e-23, 9.72875e-24, 4.45733e-24, 5.44168e-24, 3.09029e-23, 7.76852e-24], [688.453, 2.87634e-23, 9.73883e-24, 4.46154e-24, 5.45926e-24, 3.10035e-23, 7.73662e-24], [690.019, 2.88282e-23, 9.74894e-24, 4.46576e-24, 5.4769e-24, 3.11044e-23, 7.70462e-24], [691.589, 2.88931e-23, 9.75909e-24, 4.47e-24, 5.49459e-24, 3.12055e-23, 7.67253e-24], [693.162, 2.89582e-23, 9.76927e-24, 4.47426e-24, 5.51234e-24, 3.13067e-23, 7.64035e-24], [694.739, 2.90234e-23, 9.77949e-24, 4.47853e-24, 5.53014e-24, 3.14082e-23, 7.60807e-24], [696.319, 2.90888e-23, 9.78975e-24, 4.48282e-24, 5.548e-24, 3.15099e-23, 7.5757e-24], [697.903, 2.91543e-23, 9.80004e-24, 4.48713e-24, 5.56592e-24, 3.16118e-23, 7.54323e-24], [699.49, 2.922e-23, 9.81037e-24, 4.49145e-24, 5.5839e-24, 3.17139e-23, 7.51067e-24], [701.081, 2.92858e-23, 9.82073e-24, 4.4958e-24, 5.60193e-24, 3.18163e-23, 7.47801e-24], [702.676, 2.93518e-23, 9.83114e-24, 4.50015e-24, 5.62002e-24, 3.19188e-23, 7.44526e-24], [704.274, 2.94179e-23, 9.84158e-24, 4.50453e-24, 5.63816e-24, 3.20216e-23, 7.41242e-24], [705.876, 2.94842e-23, 9.85206e-24, 4.50892e-24, 5.65636e-24, 3.21245e-23, 7.37948e-24], [707.481, 2.95506e-23, 9.86257e-24, 4.51333e-24, 5.67462e-24, 3.22277e-23, 7.34645e-24], [709.091, 2.96172e-23, 9.87312e-24, 4.51776e-24, 5.69293e-24, 3.23311e-23, 7.31333e-24], [710.704, 2.96839e-23, 9.88371e-24, 4.5222e-24, 5.7113e-24, 3.24347e-23, 7.28011e-24], [712.32, 2.97508e-23, 9.89434e-24, 4.52667e-24, 5.72973e-24, 3.25386e-23, 7.2468e-24], [713.94, 2.98178e-23, 9.905e-24, 4.53115e-24, 5.74821e-24, 3.26426e-23, 7.21339e-24], [715.564, 2.9885e-23, 9.9157e-24, 4.53564e-24, 5.76675e-24, 3.27469e-23, 7.1799e-24], [717.192, 2.99524e-23, 9.92644e-24, 4.54016e-24, 5.78534e-24, 3.28513e-23, 7.14631e-24], [718.823, 3.00199e-23, 9.93722e-24, 4.54469e-24, 5.80399e-24, 3.2956e-23, 7.11262e-24], [720.458, 3.00875e-23, 9.94804e-24, 4.54924e-24, 5.8227e-24, 3.30609e-23, 7.07885e-24], [722.097, 3.01553e-23, 9.95889e-24, 4.5538e-24, 5.84147e-24, 3.31661e-23, 7.04498e-24], [723.739, 3.02233e-23, 9.96978e-24, 4.55839e-24, 5.86029e-24, 3.32714e-23, 7.01102e-24], [725.386, 3.02914e-23, 9.98071e-24, 4.56299e-24, 5.87916e-24, 3.3377e-23, 6.97697e-24], [727.035, 3.03597e-23, 9.99168e-24, 4.56761e-24, 5.8981e-24, 3.34827e-23, 6.94282e-24], [728.689, 3.04281e-23, 1.00027e-23, 4.57225e-24, 5.91708e-24, 3.35887e-23, 6.90859e-24], [730.347, 3.04967e-23, 1.00137e-23, 4.5769e-24, 5.93613e-24, 3.36949e-23, 6.87426e-24], [732.008, 3.05655e-23, 1.00248e-23, 4.58158e-24, 5.95523e-24, 3.38014e-23, 6.83984e-24], [733.673, 3.06344e-23, 1.00359e-23, 4.58627e-24, 5.97439e-24, 3.3908e-23, 6.80534e-24], [735.342, 3.07034e-23, 1.00471e-23, 4.59098e-24, 5.9936e-24, 3.40149e-23, 6.77074e-24], [737.014, 3.07726e-23, 1.00583e-23, 4.59571e-24, 6.01287e-24, 3.4122e-23, 6.73605e-24], [738.691, 3.0842e-23, 1.00695e-23, 4.60045e-24, 6.0322e-24, 3.42293e-23, 6.70127e-24], [740.371, 3.09116e-23, 1.00808e-23, 4.60521e-24, 6.05158e-24, 3.43368e-23, 6.6664e-24], [742.055, 3.09813e-23, 1.00921e-23, 4.61e-24, 6.07102e-24, 3.44446e-23, 6.63145e-24], [743.743, 3.10511e-23, 1.01035e-23, 4.6148e-24, 6.09052e-24, 3.45525e-23, 6.5964e-24], [745.434, 3.11211e-23, 1.01149e-23, 4.61961e-24, 6.11007e-24, 3.46607e-23, 6.56127e-24], [747.13, 3.11913e-23, 1.01263e-23, 4.62445e-24, 6.12968e-24, 3.47691e-23, 6.52604e-24], [748.829, 3.12616e-23, 1.01378e-23, 4.62931e-24, 6.14934e-24, 3.48778e-23, 6.49073e-24], [750.533, 3.13321e-23, 1.01493e-23, 4.63418e-24, 6.16906e-24, 3.49866e-23, 6.45534e-24], [752.24, 3.14028e-23, 1.01609e-23, 4.63907e-24, 6.18884e-24, 3.50957e-23, 6.41985e-24], [753.951, 3.14736e-23, 1.01725e-23, 4.64398e-24, 6.20867e-24, 3.5205e-23, 6.38428e-24], [755.666, 3.15446e-23, 1.01841e-23, 4.64891e-24, 6.22856e-24, 3.53146e-23, 6.34863e-24], [757.385, 3.16157e-23, 1.01958e-23, 4.65386e-24, 6.24851e-24, 3.54243e-23, 6.31289e-24], [759.107, 3.1687e-23, 1.02075e-23, 4.65882e-24, 6.26851e-24, 3.55343e-23, 6.27706e-24], [760.834, 3.17585e-23, 1.02193e-23, 4.66381e-24, 6.28857e-24, 3.56445e-23, 6.24115e-24], [762.564, 3.18301e-23, 1.02311e-23, 4.66881e-24, 6.30868e-24, 3.57549e-23, 6.20516e-24], [764.299, 3.19019e-23, 1.02429e-23, 4.67383e-24, 6.32885e-24, 3.58656e-23, 6.16908e-24], [766.037, 3.19739e-23, 1.02548e-23, 4.67887e-24, 6.34908e-24, 3.59765e-23, 6.13292e-24], [767.78, 3.2046e-23, 1.02667e-23, 4.68393e-24, 6.36936e-24, 3.60876e-23, 6.09668e-24], [769.526, 3.21183e-23, 1.02787e-23, 4.68901e-24, 6.3897e-24, 3.61989e-23, 6.06036e-24], [771.277, 3.21908e-23, 1.02907e-23, 4.69411e-24, 6.41009e-24, 3.63105e-23, 6.02396e-24], [773.031, 3.22634e-23, 1.03027e-23, 4.69923e-24, 6.43055e-24, 3.64223e-23, 5.98747e-24], [774.789, 3.23362e-23, 1.03148e-23, 4.70436e-24, 6.45105e-24, 3.65343e-23, 5.95091e-24], [776.551, 3.24091e-23, 1.03269e-23, 4.70952e-24, 6.47162e-24, 3.66465e-23, 5.91427e-24], [778.318, 3.24823e-23, 1.03391e-23, 4.71469e-24, 6.49224e-24, 3.6759e-23, 5.87756e-24], [780.088, 3.25556e-23, 1.03513e-23, 4.71988e-24, 6.51292e-24, 3.68717e-23, 5.84076e-24], [781.862, 3.2629e-23, 1.03636e-23, 4.7251e-24, 6.53365e-24, 3.69847e-23, 5.80389e-24], [783.641, 3.27026e-23, 1.03759e-23, 4.73033e-24, 6.55444e-24, 3.70978e-23, 5.76695e-24], [785.423, 3.27764e-23, 1.03882e-23, 4.73558e-24, 6.57528e-24, 3.72112e-23, 5.72993e-24], [787.21, 3.28504e-23, 1.04006e-23, 4.74085e-24, 6.59619e-24, 3.73249e-23, 5.69284e-24], [789.0, 3.29245e-23, 1.0413e-23, 4.74614e-24, 6.61715e-24, 3.74387e-23, 5.65567e-24], [790.795, 3.29988e-23, 1.04255e-23, 4.75145e-24, 6.63816e-24, 3.75528e-23, 5.61844e-24], [792.594, 3.30733e-23, 1.0438e-23, 4.75678e-24, 6.65923e-24, 3.76671e-23, 5.58113e-24], [794.397, 3.31479e-23, 1.04505e-23, 4.76212e-24, 6.68036e-24, 3.77817e-23, 5.54376e-24], [796.203, 3.32227e-23, 1.04631e-23, 4.76749e-24, 6.70154e-24, 3.78965e-23, 5.50632e-24], [798.014, 3.32977e-23, 1.04757e-23, 4.77288e-24, 6.72279e-24, 3.80115e-23, 5.46881e-24], [799.83, 3.33729e-23, 1.04884e-23, 4.77829e-24, 6.74408e-24, 3.81267e-23, 5.43124e-24], [801.649, 3.34482e-23, 1.05011e-23, 4.78371e-24, 6.76544e-24, 3.82422e-23, 5.3936e-24], [803.472, 3.35237e-23, 1.05139e-23, 4.78916e-24, 6.78685e-24, 3.8358e-23, 5.3559e-24], [805.3, 3.35994e-23, 1.05267e-23, 4.79463e-24, 6.80831e-24, 3.84739e-23, 5.31814e-24], [807.132, 3.36752e-23, 1.05395e-23, 4.80011e-24, 6.82984e-24, 3.85901e-23, 5.28032e-24], [808.967, 3.37512e-23, 1.05524e-23, 4.80562e-24, 6.85142e-24, 3.87065e-23, 5.24244e-24], [810.807, 3.38274e-23, 1.05654e-23, 4.81114e-24, 6.87305e-24, 3.88232e-23, 5.2045e-24], [812.652, 3.39038e-23, 1.05783e-23, 4.81669e-24, 6.89475e-24, 3.89401e-23, 5.16651e-24], [814.5, 3.39803e-23, 1.05914e-23, 4.82226e-24, 6.91649e-24, 3.90572e-23, 5.12847e-24], [816.353, 3.4057e-23, 1.06044e-23, 4.82784e-24, 6.9383e-24, 3.91746e-23, 5.09037e-24], [818.21, 3.41339e-23, 1.06175e-23, 4.83345e-24, 6.96016e-24, 3.92922e-23, 5.05223e-24], [820.071, 3.4211e-23, 1.06307e-23, 4.83908e-24, 6.98208e-24, 3.94101e-23, 5.01404e-24], [821.936, 3.42882e-23, 1.06439e-23, 4.84473e-24, 7.00406e-24, 3.95282e-23, 4.9758e-24], [823.806, 3.43656e-23, 1.06571e-23, 4.85039e-24, 7.02609e-24, 3.96465e-23, 4.93751e-24], [825.679, 3.44432e-23, 1.06704e-23, 4.85608e-24, 7.04818e-24, 3.97651e-23, 4.89919e-24], [827.557, 3.4521e-23, 1.06837e-23, 4.86179e-24, 7.07033e-24, 3.98839e-23, 4.86082e-24], [829.44, 3.4599e-23, 1.06971e-23, 4.86752e-24, 7.09253e-24, 4.0003e-23, 4.82242e-24], [831.326, 3.46771e-23, 1.07105e-23, 4.87327e-24, 7.11479e-24, 4.01223e-23, 4.78399e-24], [833.217, 3.47554e-23, 1.0724e-23, 4.87904e-24, 7.13711e-24, 4.02418e-23, 4.74552e-24], [835.113, 3.48339e-23, 1.07375e-23, 4.88483e-24, 7.15948e-24, 4.03616e-23, 4.70702e-24], [837.012, 3.49125e-23, 1.07511e-23, 4.89064e-24, 7.18191e-24, 4.04816e-23, 4.66849e-24], [838.916, 3.49914e-23, 1.07647e-23, 4.89647e-24, 7.2044e-24, 4.06018e-23, 4.62994e-24], [840.824, 3.50704e-23, 1.07783e-23, 4.90232e-24, 7.22694e-24, 4.07224e-23, 4.59137e-24], [842.737, 3.51496e-23, 1.0792e-23, 4.9082e-24, 7.24954e-24, 4.08431e-23, 4.55278e-24], [844.653, 3.5229e-23, 1.08057e-23, 4.91409e-24, 7.2722e-24, 4.09641e-23, 4.51417e-24], [846.575, 3.53086e-23, 1.08195e-23, 4.92001e-24, 7.29492e-24, 4.10853e-23, 4.47555e-24], [848.5, 3.53883e-23, 1.08333e-23, 4.92595e-24, 7.31769e-24, 4.12068e-23, 4.43693e-24], [850.43, 3.54682e-23, 1.08472e-23, 4.9319e-24, 7.34052e-24, 4.13286e-23, 4.3983e-24], [852.365, 3.55484e-23, 1.08611e-23, 4.93788e-24, 7.36341e-24, 4.14505e-23, 4.35966e-24], [854.303, 3.56287e-23, 1.08751e-23, 4.94388e-24, 7.38635e-24, 4.15728e-23, 4.32103e-24], [856.247, 3.57091e-23, 1.08891e-23, 4.9499e-24, 7.40935e-24, 4.16952e-23, 4.28241e-24], [858.194, 3.57898e-23, 1.09032e-23, 4.95594e-24, 7.43241e-24, 4.18179e-23, 4.24379e-24], [860.146, 3.58706e-23, 1.09173e-23, 4.96201e-24, 7.45552e-24, 4.19409e-23, 4.20519e-24], [862.103, 3.59517e-23, 1.09314e-23, 4.96809e-24, 7.47869e-24, 4.20641e-23, 4.16661e-24], [864.064, 3.60329e-23, 1.09456e-23, 4.9742e-24, 7.50192e-24, 4.21876e-23, 4.12806e-24], [866.029, 3.61143e-23, 1.09598e-23, 4.98033e-24, 7.52521e-24, 4.23113e-23, 4.08953e-24], [867.999, 3.61959e-23, 1.09741e-23, 4.98647e-24, 7.54856e-24, 4.24353e-23, 4.05104e-24], [869.973, 3.62777e-23, 1.09885e-23, 4.99264e-24, 7.57196e-24, 4.25595e-23, 4.01258e-24], [871.952, 3.63596e-23, 1.10028e-23, 4.99884e-24, 7.59542e-24, 4.26839e-23, 3.97417e-24], [873.935, 3.64418e-23, 1.10173e-23, 5.00505e-24, 7.61893e-24, 4.28087e-23, 3.93581e-24], [875.923, 3.65241e-23, 1.10318e-23, 5.01129e-24, 7.64251e-24, 4.29336e-23, 3.89751e-24], [877.915, 3.66066e-23, 1.10463e-23, 5.01754e-24, 7.66614e-24, 4.30588e-23, 3.85927e-24], [879.912, 3.66893e-23, 1.10609e-23, 5.02382e-24, 7.68983e-24, 4.31843e-23, 3.8211e-24], [881.914, 3.67723e-23, 1.10755e-23, 5.03012e-24, 7.71358e-24, 4.331e-23, 3.78301e-24], [883.92, 3.68553e-23, 1.10901e-23, 5.03645e-24, 7.73738e-24, 4.3436e-23, 3.745e-24], [885.93, 3.69386e-23, 1.11049e-23, 5.04279e-24, 7.76124e-24, 4.35623e-23, 3.70708e-24], [887.945, 3.70221e-23, 1.11196e-23, 5.04916e-24, 7.78516e-24, 4.36887e-23, 3.66926e-24], [889.965, 3.71058e-23, 1.11344e-23, 5.05554e-24, 7.80914e-24, 4.38155e-23, 3.63155e-24], [891.989, 3.71896e-23, 1.11493e-23, 5.06196e-24, 7.83318e-24, 4.39425e-23, 3.59395e-24], [894.018, 3.72737e-23, 1.11642e-23, 5.06839e-24, 7.85727e-24, 4.40697e-23, 3.55648e-24], [896.052, 3.73579e-23, 1.11791e-23, 5.07484e-24, 7.88142e-24, 4.41973e-23, 3.51914e-24], [898.09, 3.74423e-23, 1.11941e-23, 5.08132e-24, 7.90563e-24, 4.4325e-23, 3.48194e-24], [900.133, 3.75269e-23, 1.12092e-23, 5.08782e-24, 7.9299e-24, 4.44531e-23, 3.4449e-24], [902.18, 3.76118e-23, 1.12243e-23, 5.09434e-24, 7.95422e-24, 4.45813e-23, 3.40802e-24], [904.232, 3.76968e-23, 1.12395e-23, 5.10089e-24, 7.97861e-24, 4.47099e-23, 3.37131e-24], [906.289, 3.7782e-23, 1.12547e-23, 5.10745e-24, 8.00305e-24, 4.48387e-23, 3.33479e-24], [908.35, 3.78674e-23, 1.12699e-23, 5.11404e-24, 8.02755e-24, 4.49678e-23, 3.29847e-24], [910.416, 3.7953e-23, 1.12852e-23, 5.12066e-24, 8.05211e-24, 4.50971e-23, 3.26236e-24], [912.487, 3.80387e-23, 1.13005e-23, 5.12729e-24, 8.07673e-24, 4.52267e-23, 3.22648e-24], [914.563, 3.81247e-23, 1.13159e-23, 5.13395e-24, 8.1014e-24, 4.53565e-23, 3.19083e-24], [916.643, 3.82109e-23, 1.13314e-23, 5.14063e-24, 8.12614e-24, 4.54866e-23, 3.15543e-24], [918.728, 3.82973e-23, 1.13469e-23, 5.14733e-24, 8.15093e-24, 4.5617e-23, 3.1203e-24], [920.818, 3.83839e-23, 1.13624e-23, 5.15406e-24, 8.17578e-24, 4.57476e-23, 3.08545e-24], [922.912, 3.84706e-23, 1.1378e-23, 5.16081e-24, 8.20069e-24, 4.58785e-23, 3.0509e-24], [925.011, 3.85576e-23, 1.13937e-23, 5.16758e-24, 8.22566e-24, 4.60097e-23, 3.01667e-24], [927.115, 3.86448e-23, 1.14094e-23, 5.17437e-24, 8.25068e-24, 4.61411e-23, 2.98276e-24], [929.224, 3.87321e-23, 1.14251e-23, 5.18119e-24, 8.27577e-24, 4.62728e-23, 2.94921e-24], [931.338, 3.88197e-23, 1.14409e-23, 5.18804e-24, 8.30091e-24, 4.64048e-23, 2.91603e-24], [933.456, 3.89075e-23, 1.14568e-23, 5.1949e-24, 8.32612e-24, 4.6537e-23, 2.88324e-24], [935.579, 3.89954e-23, 1.14727e-23, 5.20179e-24, 8.35138e-24, 4.66695e-23, 2.85087e-24], [937.707, 3.90836e-23, 1.14886e-23, 5.20871e-24, 8.37671e-24, 4.68023e-23, 2.81892e-24], [939.84, 3.9172e-23, 1.15046e-23, 5.21565e-24, 8.40209e-24, 4.69353e-23, 2.78743e-24], [941.978, 3.92606e-23, 1.15207e-23, 5.22262e-24, 8.42753e-24, 4.70686e-23, 2.75643e-24], [944.12, 3.93493e-23, 1.15368e-23, 5.22961e-24, 8.45304e-24, 4.72022e-23, 2.72593e-24], [946.268, 3.94383e-23, 1.15529e-23, 5.23664e-24, 8.47861e-24, 4.7336e-23, 2.69598e-24], [948.42, 3.95275e-23, 1.15691e-23, 5.2437e-24, 8.50425e-24, 4.74701e-23, 2.6666e-24], [950.578, 3.96169e-23, 1.15854e-23, 5.25081e-24, 8.52996e-24, 4.76045e-23, 2.63785e-24], [952.74, 3.97065e-23, 1.16018e-23, 5.258e-24, 8.55577e-24, 4.77392e-23, 2.60982e-24], [954.907, 3.97963e-23, 1.16183e-23, 5.26537e-24, 8.58172e-24, 4.78741e-23, 2.58274e-24], [957.079, 3.98864e-23, 1.16351e-23, 5.27337e-24, 8.60811e-24, 4.80094e-23, 2.55758e-24], [959.256, 3.99776e-23, 1.16548e-23, 5.28762e-24, 8.63837e-24, 4.81457e-23, 2.54612e-24], [961.438, 4.00727e-23, 1.16875e-23, 5.33051e-24, 8.68621e-24, 4.82853e-23, 2.59496e-24], [963.624, 4.01578e-23, 1.1685e-23, 5.29579e-24, 8.68658e-24, 4.84167e-23, 2.48298e-24], [965.816, 4.02484e-23, 1.17009e-23, 5.30138e-24, 8.71168e-24, 4.85528e-23, 2.45548e-24], [968.013, 4.03393e-23, 1.17174e-23, 5.30823e-24, 8.73759e-24, 4.86894e-23, 2.43143e-24], [970.215, 4.04305e-23, 1.17341e-23, 5.31534e-24, 8.76371e-24, 4.88262e-23, 2.40873e-24], [972.422, 4.0522e-23, 1.17508e-23, 5.32255e-24, 8.78994e-24, 4.89633e-23, 2.38708e-24], [974.634, 4.06136e-23, 1.17676e-23, 5.32982e-24, 8.81624e-24, 4.91007e-23, 2.36641e-24], [976.85, 4.07055e-23, 1.17845e-23, 5.33714e-24, 8.84261e-24, 4.92384e-23, 2.34673e-24], [979.072, 4.07976e-23, 1.18014e-23, 5.34448e-24, 8.86905e-24, 4.93763e-23, 2.32804e-24], [981.299, 4.08898e-23, 1.18184e-23, 5.35185e-24, 8.89556e-24, 4.95145e-23, 2.31038e-24], [983.531, 4.09823e-23, 1.18354e-23, 5.35925e-24, 8.92212e-24, 4.9653e-23, 2.29377e-24], [985.768, 4.1075e-23, 1.18525e-23, 5.36668e-24, 8.94875e-24, 4.97918e-23, 2.27825e-24], [988.011, 4.11679e-23, 1.18696e-23, 5.37413e-24, 8.97544e-24, 4.99309e-23, 2.26384e-24], [990.258, 4.12611e-23, 1.18868e-23, 5.38161e-24, 9.00219e-24, 5.00702e-23, 2.25058e-24], [992.51, 4.13544e-23, 1.19041e-23, 5.38912e-24, 9.029e-24, 5.02099e-23, 2.2385e-24], [994.768, 4.14479e-23, 1.19214e-23, 5.39665e-24, 9.05587e-24, 5.03498e-23, 2.22762e-24], [997.031, 4.15417e-23, 1.19388e-23, 5.4042e-24, 9.0828e-24, 5.049e-23, 2.21797e-24], [999.298, 4.16357e-23, 1.19562e-23, 5.41178e-24, 9.10979e-24, 5.06305e-23, 2.20958e-24], [1001.57, 4.17299e-23, 1.19737e-23, 5.41938e-24, 9.13684e-24, 5.07712e-23, 2.20247e-24], [1003.85, 4.18243e-23, 1.19912e-23, 5.42701e-24, 9.16396e-24, 5.09123e-23, 2.19667e-24], [1006.13, 4.19189e-23, 1.20088e-23, 5.43467e-24, 9.19113e-24, 5.10536e-23, 2.1922e-24], [1008.42, 4.20137e-23, 1.20264e-23, 5.44235e-24, 9.21837e-24, 5.11952e-23, 2.18907e-24], [1010.72, 4.21088e-23, 1.20441e-23, 5.45005e-24, 9.24567e-24, 5.13371e-23, 2.1873e-24], [1013.01, 4.22041e-23, 1.20618e-23, 5.45778e-24, 9.27303e-24, 5.14793e-23, 2.18691e-24], [1015.32, 4.22996e-23, 1.20796e-23, 5.46554e-24, 9.30045e-24, 5.16218e-23, 2.1879e-24], [1017.63, 4.23953e-23, 1.20975e-23, 5.47332e-24, 9.32793e-24, 5.17646e-23, 2.1903e-24], [1019.94, 4.24912e-23, 1.21154e-23, 5.48112e-24, 9.35547e-24, 5.19076e-23, 2.19409e-24], [1022.26, 4.25873e-23, 1.21333e-23, 5.48895e-24, 9.38307e-24, 5.2051e-23, 2.19929e-24], [1024.59, 4.26837e-23, 1.21514e-23, 5.49681e-24, 9.41074e-24, 5.21946e-23, 2.2059e-24], [1026.92, 4.27803e-23, 1.21694e-23, 5.50469e-24, 9.43846e-24, 5.23386e-23, 2.21391e-24], [1029.25, 4.28771e-23, 1.21876e-23, 5.5126e-24, 9.46625e-24, 5.24828e-23, 2.22332e-24], [1031.59, 4.29741e-23, 1.22057e-23, 5.52053e-24, 9.4941e-24, 5.26273e-23, 2.23413e-24], [1033.94, 4.30714e-23, 1.2224e-23, 5.52849e-24, 9.52201e-24, 5.27721e-23, 2.24632e-24], [1036.29, 4.31688e-23, 1.22423e-23, 5.53647e-24, 9.54999e-24, 5.29172e-23, 2.25989e-24], [1038.65, 4.32665e-23, 1.22606e-23, 5.54448e-24, 9.57802e-24, 5.30626e-23, 2.27481e-24], [1041.01, 4.33644e-23, 1.22791e-23, 5.55252e-24, 9.60612e-24, 5.32083e-23, 2.29108e-24], [1043.38, 4.34626e-23, 1.22975e-23, 5.56058e-24, 9.63428e-24, 5.33543e-23, 2.30867e-24], [1045.75, 4.35609e-23, 1.2316e-23, 5.56867e-24, 9.6625e-24, 5.35005e-23, 2.32757e-24], [1048.13, 4.36595e-23, 1.23346e-23, 5.57678e-24, 9.69079e-24, 5.36471e-23, 2.34776e-24], [1050.52, 4.37583e-23, 1.23533e-23, 5.58492e-24, 9.71914e-24, 5.3794e-23, 2.36921e-24], [1052.91, 4.38573e-23, 1.2372e-23, 5.59308e-24, 9.74755e-24, 5.39412e-23, 2.39189e-24], [1055.3, 4.39566e-23, 1.23907e-23, 5.60127e-24, 9.77602e-24, 5.40886e-23, 2.41579e-24], [1057.7, 4.40561e-23, 1.24095e-23, 5.60949e-24, 9.80455e-24, 5.42364e-23, 2.44089e-24], [1060.11, 4.41558e-23, 1.24284e-23, 5.61773e-24, 9.83315e-24, 5.43845e-23, 2.46714e-24], [1062.52, 4.42557e-23, 1.24473e-23, 5.626e-24, 9.86181e-24, 5.45329e-23, 2.49454e-24], [1064.94, 4.43559e-23, 1.24663e-23, 5.6343e-24, 9.89053e-24, 5.46815e-23, 2.52304e-24], [1067.36, 4.44563e-23, 1.24853e-23, 5.64262e-24, 9.91932e-24, 5.48305e-23, 2.55262e-24], [1069.79, 4.45569e-23, 1.25044e-23, 5.65097e-24, 9.94817e-24, 5.49798e-23, 2.58327e-24], [1072.22, 4.46578e-23, 1.25236e-23, 5.65934e-24, 9.97708e-24, 5.51293e-23, 2.61494e-24], [1074.66, 4.47589e-23, 1.25428e-23, 5.66775e-24, 1.00061e-23, 5.52792e-23, 2.64761e-24], [1077.1, 4.48602e-23, 1.25621e-23, 5.67617e-24, 1.00351e-23, 5.54294e-23, 2.68125e-24], [1079.55, 4.49617e-23, 1.25814e-23, 5.68463e-24, 1.00642e-23, 5.55799e-23, 2.71585e-24], [1082.01, 4.50635e-23, 1.26008e-23, 5.69311e-24, 1.00934e-23, 5.57307e-23, 2.75136e-24], [1084.47, 4.51655e-23, 1.26202e-23, 5.70162e-24, 1.01226e-23, 5.58818e-23, 2.78777e-24], [1086.93, 4.52678e-23, 1.26397e-23, 5.71015e-24, 1.01519e-23, 5.60332e-23, 2.82505e-24], [1089.41, 4.53702e-23, 1.26593e-23, 5.71871e-24, 1.01812e-23, 5.61849e-23, 2.86318e-24], [1091.89, 4.54729e-23, 1.26789e-23, 5.7273e-24, 1.02107e-23, 5.6337e-23, 2.90213e-24], [1094.37, 4.55759e-23, 1.26986e-23, 5.73592e-24, 1.02401e-23, 5.64893e-23, 2.94188e-24], [1096.86, 4.56791e-23, 1.27184e-23, 5.74456e-24, 1.02697e-23, 5.6642e-23, 2.9824e-24], [1099.35, 4.57825e-23, 1.27382e-23, 5.75323e-24, 1.02993e-23, 5.67949e-23, 3.02367e-24], [1101.85, 4.58861e-23, 1.2758e-23, 5.76193e-24, 1.0329e-23, 5.69482e-23, 3.06568e-24], [1104.36, 4.599e-23, 1.27779e-23, 5.77065e-24, 1.03587e-23, 5.71018e-23, 3.1084e-24], [1106.87, 4.60941e-23, 1.27979e-23, 5.7794e-24, 1.03885e-23, 5.72557e-23, 3.1518e-24], [1109.39, 4.61985e-23, 1.2818e-23, 5.78818e-24, 1.04184e-23, 5.74099e-23, 3.19588e-24], [1111.91, 4.63031e-23, 1.28381e-23, 5.79699e-24, 1.04483e-23, 5.75644e-23, 3.24061e-24], [1114.44, 4.64079e-23, 1.28582e-23, 5.80582e-24, 1.04783e-23, 5.77192e-23, 3.28597e-24], [1116.98, 4.6513e-23, 1.28784e-23, 5.81468e-24, 1.05084e-23, 5.78744e-23, 3.33195e-24], [1119.52, 4.66183e-23, 1.28987e-23, 5.82357e-24, 1.05385e-23, 5.80299e-23, 3.37852e-24], [1122.06, 4.67239e-23, 1.29191e-23, 5.83249e-24, 1.05687e-23, 5.81856e-23, 3.42568e-24], [1124.62, 4.68297e-23, 1.29395e-23, 5.84143e-24, 1.0599e-23, 5.83417e-23, 3.4734e-24], [1127.17, 4.69357e-23, 1.29599e-23, 5.8504e-24, 1.06293e-23, 5.84982e-23, 3.52168e-24], [1129.74, 4.7042e-23, 1.29805e-23, 5.8594e-24, 1.06597e-23, 5.86549e-23, 3.57049e-24], [1132.31, 4.71485e-23, 1.30011e-23, 5.86843e-24, 1.06901e-23, 5.8812e-23, 3.61983e-24], [1134.88, 4.72553e-23, 1.30217e-23, 5.87748e-24, 1.07207e-23, 5.89693e-23, 3.66967e-24], [1137.46, 4.73623e-23, 1.30424e-23, 5.88657e-24, 1.07513e-23, 5.9127e-23, 3.72001e-24], [1140.05, 4.74695e-23, 1.30632e-23, 5.89568e-24, 1.07819e-23, 5.92851e-23, 3.77084e-24], [1142.64, 4.7577e-23, 1.3084e-23, 5.90482e-24, 1.08126e-23, 5.94434e-23, 3.82214e-24], [1145.24, 4.76848e-23, 1.31049e-23, 5.91399e-24, 1.08434e-23, 5.96021e-23, 3.8739e-24], [1147.85, 4.77928e-23, 1.31259e-23, 5.92318e-24, 1.08743e-23, 5.97611e-23, 3.92611e-24], [1150.46, 4.7901e-23, 1.31469e-23, 5.93241e-24, 1.09052e-23, 5.99204e-23, 3.97876e-24], [1153.08, 4.80095e-23, 1.3168e-23, 5.94166e-24, 1.09362e-23, 6.008e-23, 4.03184e-24], [1155.7, 4.81182e-23, 1.31891e-23, 5.95094e-24, 1.09672e-23, 6.024e-23, 4.08533e-24], [1158.33, 4.82272e-23, 1.32103e-23, 5.96025e-24, 1.09983e-23, 6.04003e-23, 4.13924e-24], [1160.96, 4.83364e-23, 1.32316e-23, 5.96959e-24, 1.10295e-23, 6.05609e-23, 4.19355e-24], [1163.6, 4.84459e-23, 1.32529e-23, 5.97895e-24, 1.10608e-23, 6.07219e-23, 4.24826e-24], [1166.25, 4.85556e-23, 1.32743e-23, 5.98835e-24, 1.10921e-23, 6.08832e-23, 4.30335e-24], [1168.9, 4.86656e-23, 1.32958e-23, 5.99777e-24, 1.11234e-23, 6.10448e-23, 4.35881e-24], [1171.56, 4.87758e-23, 1.33173e-23, 6.00722e-24, 1.11549e-23, 6.12067e-23, 4.41465e-24], [1174.23, 4.88862e-23, 1.33389e-23, 6.0167e-24, 1.11864e-23, 6.1369e-23, 4.47085e-24], [1176.9, 4.8997e-23, 1.33605e-23, 6.02621e-24, 1.1218e-23, 6.15316e-23, 4.5274e-24], [1179.57, 4.9108e-23, 1.33822e-23, 6.03575e-24, 1.12496e-23, 6.16945e-23, 4.58431e-24], [1182.26, 4.92192e-23, 1.3404e-23, 6.04532e-24, 1.12814e-23, 6.18578e-23, 4.64155e-24], [1184.95, 4.93307e-23, 1.34259e-23, 6.05492e-24, 1.13132e-23, 6.20214e-23, 4.69913e-24], [1187.64, 4.94424e-23, 1.34478e-23, 6.06454e-24, 1.1345e-23, 6.21854e-23, 4.75705e-24], [1190.34, 4.95544e-23, 1.34697e-23, 6.0742e-24, 1.13769e-23, 6.23496e-23, 4.81528e-24], [1193.05, 4.96666e-23, 1.34918e-23, 6.08388e-24, 1.14089e-23, 6.25143e-23, 4.87384e-24], [1195.76, 4.97791e-23, 1.35139e-23, 6.0936e-24, 1.1441e-23, 6.26792e-23, 4.93271e-24], [1198.48, 4.98919e-23, 1.3536e-23, 6.10334e-24, 1.14731e-23, 6.28445e-23, 4.99189e-24], [1201.21, 5.00049e-23, 1.35583e-23, 6.11311e-24, 1.15053e-23, 6.30101e-23, 5.05137e-24], [1203.94, 5.01182e-23, 1.35805e-23, 6.12292e-24, 1.15376e-23, 6.31761e-23, 5.11116e-24], [1206.68, 5.02317e-23, 1.36029e-23, 6.13275e-24, 1.15699e-23, 6.33424e-23, 5.17123e-24], [1209.43, 5.03455e-23, 1.36253e-23, 6.14261e-24, 1.16023e-23, 6.35091e-23, 5.2316e-24], [1212.18, 5.04596e-23, 1.36478e-23, 6.1525e-24, 1.16348e-23, 6.36761e-23, 5.29226e-24], [1214.93, 5.05739e-23, 1.36704e-23, 6.16242e-24, 1.16673e-23, 6.38434e-23, 5.3532e-24], [1217.7, 5.06885e-23, 1.3693e-23, 6.17237e-24, 1.16999e-23, 6.40111e-23, 5.41442e-24], [1220.47, 5.08033e-23, 1.37157e-23, 6.18235e-24, 1.17326e-23, 6.41791e-23, 5.47591e-24], [1223.24, 5.09184e-23, 1.37384e-23, 6.19236e-24, 1.17653e-23, 6.43475e-23, 5.53768e-24], [1226.02, 5.10337e-23, 1.37612e-23, 6.20241e-24, 1.17981e-23, 6.45162e-23, 5.59972e-24], [1228.81, 5.11494e-23, 1.37841e-23, 6.21248e-24, 1.1831e-23, 6.46852e-23, 5.66202e-24], [1231.61, 5.12652e-23, 1.38071e-23, 6.22258e-24, 1.1864e-23, 6.48547e-23, 5.72458e-24], [1234.41, 5.13814e-23, 1.38301e-23, 6.23271e-24, 1.1897e-23, 6.50244e-23, 5.7874e-24], [1237.22, 5.14978e-23, 1.38532e-23, 6.24287e-24, 1.19301e-23, 6.51945e-23, 5.85048e-24], [1240.03, 5.16145e-23, 1.38763e-23, 6.25306e-24, 1.19633e-23, 6.5365e-23, 5.91381e-24], [1242.85, 5.17314e-23, 1.38995e-23, 6.26328e-24, 1.19965e-23, 6.55358e-23, 5.97739e-24], [1245.68, 5.18486e-23, 1.39228e-23, 6.27353e-24, 1.20298e-23, 6.5707e-23, 6.04122e-24], [1248.51, 5.19661e-23, 1.39461e-23, 6.28382e-24, 1.20632e-23, 6.58785e-23, 6.1053e-24], [1251.35, 5.20838e-23, 1.39696e-23, 6.29413e-24, 1.20966e-23, 6.60503e-23, 6.16962e-24], [1254.2, 5.22019e-23, 1.3993e-23, 6.30448e-24, 1.21301e-23, 6.62226e-23, 6.23418e-24], [1257.05, 5.23201e-23, 1.40166e-23, 6.31485e-24, 1.21637e-23, 6.63951e-23, 6.29898e-24], [1259.91, 5.24387e-23, 1.40402e-23, 6.32526e-24, 1.21974e-23, 6.65681e-23, 6.36402e-24], [1262.78, 5.25575e-23, 1.40639e-23, 6.33569e-24, 1.22311e-23, 6.67414e-23, 6.42929e-24], [1265.65, 5.26766e-23, 1.40877e-23, 6.34616e-24, 1.22649e-23, 6.6915e-23, 6.49479e-24], [1268.53, 5.2796e-23, 1.41115e-23, 6.35666e-24, 1.22988e-23, 6.7089e-23, 6.56052e-24], [1271.41, 5.29156e-23, 1.41354e-23, 6.36719e-24, 1.23327e-23, 6.72634e-23, 6.62649e-24], [1274.3, 5.30355e-23, 1.41593e-23, 6.37775e-24, 1.23667e-23, 6.74381e-23, 6.69268e-24], [1277.2, 5.31557e-23, 1.41834e-23, 6.38834e-24, 1.24008e-23, 6.76132e-23, 6.75909e-24], [1280.11, 5.32761e-23, 1.42075e-23, 6.39896e-24, 1.2435e-23, 6.77886e-23, 6.82573e-24], [1283.02, 5.33969e-23, 1.42316e-23, 6.40962e-24, 1.24692e-23, 6.79644e-23, 6.89259e-24], [1285.94, 5.35179e-23, 1.42559e-23, 6.4203e-24, 1.25035e-23, 6.81406e-23, 6.95967e-24], [1288.86, 5.36391e-23, 1.42802e-23, 6.43102e-24, 1.25379e-23, 6.83171e-23, 7.02697e-24], [1291.79, 5.37607e-23, 1.43045e-23, 6.44177e-24, 1.25723e-23, 6.8494e-23, 7.09449e-24], [1294.73, 5.38825e-23, 1.4329e-23, 6.45255e-24, 1.26068e-23, 6.86713e-23, 7.16222e-24], [1297.68, 5.40046e-23, 1.43535e-23, 6.46336e-24, 1.26414e-23, 6.88489e-23, 7.23017e-24], [1300.63, 5.4127e-23, 1.43781e-23, 6.4742e-24, 1.26761e-23, 6.90269e-23, 7.29833e-24], [1303.59, 5.42497e-23, 1.44027e-23, 6.48507e-24, 1.27108e-23, 6.92053e-23, 7.3667e-24], [1306.55, 5.43726e-23, 1.44275e-23, 6.49598e-24, 1.27456e-23, 6.9384e-23, 7.43528e-24], [1309.53, 5.44959e-23, 1.44523e-23, 6.50692e-24, 1.27805e-23, 6.95631e-23, 7.50407e-24], [1312.5, 5.46194e-23, 1.44771e-23, 6.51789e-24, 1.28154e-23, 6.97426e-23, 7.57307e-24], [1315.49, 5.47432e-23, 1.45021e-23, 6.52889e-24, 1.28505e-23, 6.99224e-23, 7.64228e-24], [1318.48, 5.48672e-23, 1.45271e-23, 6.53993e-24, 1.28856e-23, 7.01027e-23, 7.7117e-24], [1321.48, 5.49916e-23, 1.45521e-23, 6.55099e-24, 1.29208e-23, 7.02833e-23, 7.78132e-24], [1324.49, 5.51162e-23, 1.45773e-23, 6.56209e-24, 1.2956e-23, 7.04642e-23, 7.85114e-24], [1327.5, 5.52411e-23, 1.46025e-23, 6.57322e-24, 1.29913e-23, 7.06456e-23, 7.92117e-24], [1330.52, 5.53664e-23, 1.46278e-23, 6.58439e-24, 1.30267e-23, 7.08273e-23, 7.9914e-24], [1333.54, 5.54918e-23, 1.46532e-23, 6.59558e-24, 1.30622e-23, 7.10094e-23, 8.06183e-24], [1336.58, 5.56176e-23, 1.46786e-23, 6.60681e-24, 1.30977e-23, 7.11918e-23, 8.13246e-24], [1339.62, 5.57437e-23, 1.47041e-23, 6.61807e-24, 1.31334e-23, 7.13747e-23, 8.20329e-24], [1342.67, 5.587e-23, 1.47297e-23, 6.62937e-24, 1.31691e-23, 7.15579e-23, 8.27432e-24], [1345.72, 5.59967e-23, 1.47553e-23, 6.64069e-24, 1.32048e-23, 7.17415e-23, 8.34555e-24], [1348.78, 5.61236e-23, 1.4781e-23, 6.65205e-24, 1.32407e-23, 7.19255e-23, 8.41697e-24], [1351.85, 5.62508e-23, 1.48068e-23, 6.66344e-24, 1.32766e-23, 7.21099e-23, 8.48859e-24], [1354.92, 5.63783e-23, 1.48327e-23, 6.67487e-24, 1.33126e-23, 7.22947e-23, 8.56041e-24], [1358.0, 5.65061e-23, 1.48586e-23, 6.68633e-24, 1.33487e-23, 7.24798e-23, 8.63242e-24], [1361.09, 5.66342e-23, 1.48847e-23, 6.69782e-24, 1.33848e-23, 7.26653e-23, 8.70463e-24], [1364.19, 5.67626e-23, 1.49107e-23, 6.70934e-24, 1.3421e-23, 7.28512e-23, 8.77703e-24], [1367.29, 5.68912e-23, 1.49369e-23, 6.7209e-24, 1.34573e-23, 7.30375e-23, 8.84962e-24], [1370.4, 5.70202e-23, 1.49631e-23, 6.73249e-24, 1.34937e-23, 7.32242e-23, 8.92241e-24], [1373.52, 5.71495e-23, 1.49894e-23, 6.74411e-24, 1.35302e-23, 7.34113e-23, 8.99539e-24], [1376.64, 5.7279e-23, 1.50158e-23, 6.75577e-24, 1.35667e-23, 7.35988e-23, 9.06855e-24], [1379.77, 5.74089e-23, 1.50423e-23, 6.76746e-24, 1.36033e-23, 7.37866e-23, 9.14191e-24], [1382.91, 5.7539e-23, 1.50688e-23, 6.77918e-24, 1.364e-23, 7.39749e-23, 9.21546e-24], [1386.06, 5.76695e-23, 1.50954e-23, 6.79094e-24, 1.36767e-23, 7.41635e-23, 9.2892e-24], [1389.21, 5.78002e-23, 1.51221e-23, 6.80273e-24, 1.37136e-23, 7.43525e-23, 9.36313e-24], [1392.37, 5.79312e-23, 1.51488e-23, 6.81455e-24, 1.37505e-23, 7.45419e-23, 9.43725e-24], [1395.54, 5.80626e-23, 1.51756e-23, 6.82641e-24, 1.37875e-23, 7.47318e-23, 9.51156e-24], [1398.71, 5.81942e-23, 1.52025e-23, 6.83831e-24, 1.38245e-23, 7.4922e-23, 9.58605e-24], [1401.89, 5.83261e-23, 1.52295e-23, 6.85023e-24, 1.38617e-23, 7.51126e-23, 9.66073e-24], [1405.08, 5.84584e-23, 1.52566e-23, 6.86219e-24, 1.38989e-23, 7.53036e-23, 9.7356e-24], [1408.28, 5.85909e-23, 1.52837e-23, 6.87419e-24, 1.39362e-23, 7.5495e-23, 9.81066e-24], [1411.48, 5.87237e-23, 1.53109e-23, 6.88622e-24, 1.39736e-23, 7.56868e-23, 9.8859e-24], [1414.69, 5.88569e-23, 1.53382e-23, 6.89828e-24, 1.4011e-23, 7.5879e-23, 9.96132e-24], [1417.91, 5.89903e-23, 1.53655e-23, 6.91038e-24, 1.40486e-23, 7.60716e-23, 1.00369e-23], [1421.14, 5.9124e-23, 1.53929e-23, 6.92251e-24, 1.40862e-23, 7.62646e-23, 1.01127e-23], [1424.37, 5.92581e-23, 1.54204e-23, 6.93468e-24, 1.41239e-23, 7.6458e-23, 1.01887e-23], [1427.61, 5.93925e-23, 1.5448e-23, 6.94689e-24, 1.41616e-23, 7.66519e-23, 1.02649e-23], [1430.85, 5.95271e-23, 1.54757e-23, 6.95913e-24, 1.41995e-23, 7.68461e-23, 1.03412e-23], [1434.11, 5.96621e-23, 1.55034e-23, 6.97144e-24, 1.42374e-23, 7.70407e-23, 1.04178e-23], [1437.37, 5.97974e-23, 1.55313e-23, 6.9839e-24, 1.42755e-23, 7.72358e-23, 1.04946e-23], [1440.64, 5.99343e-23, 1.55644e-23, 7.00786e-24, 1.43193e-23, 7.74322e-23, 1.05792e-23], [1443.92, 6.00689e-23, 1.55872e-23, 7.00881e-24, 1.43519e-23, 7.76271e-23, 1.06487e-23], [1447.2, 6.0205e-23, 1.56151e-23, 7.02088e-24, 1.439e-23, 7.78233e-23, 1.07258e-23], [1450.49, 6.03415e-23, 1.56432e-23, 7.03328e-24, 1.44283e-23, 7.802e-23, 1.08032e-23], [1453.79, 6.04784e-23, 1.56714e-23, 7.04574e-24, 1.44667e-23, 7.8217e-23, 1.08808e-23], [1457.1, 6.06155e-23, 1.56997e-23, 7.05824e-24, 1.45051e-23, 7.84145e-23, 1.09586e-23], [1460.41, 6.07529e-23, 1.5728e-23, 7.07079e-24, 1.45437e-23, 7.86124e-23, 1.10366e-23], [1463.74, 6.08907e-23, 1.57565e-23, 7.08337e-24, 1.45824e-23, 7.88108e-23, 1.11148e-23], [1467.07, 6.10288e-23, 1.5785e-23, 7.09599e-24, 1.46211e-23, 7.90095e-23, 1.11931e-23], [1470.4, 6.11672e-23, 1.58135e-23, 7.10864e-24, 1.46599e-23, 7.92087e-23, 1.12717e-23], [1473.75, 6.13059e-23, 1.58422e-23, 7.12133e-24, 1.46988e-23, 7.94082e-23, 1.13504e-23], [1477.1, 6.14449e-23, 1.5871e-23, 7.13406e-24, 1.47378e-23, 7.96082e-23, 1.14293e-23], [1480.46, 6.15842e-23, 1.58998e-23, 7.14682e-24, 1.47768e-23, 7.98086e-23, 1.15084e-23], [1483.83, 6.17239e-23, 1.59287e-23, 7.15962e-24, 1.48159e-23, 8.00095e-23, 1.15876e-23], [1487.2, 6.18638e-23, 1.59576e-23, 7.17245e-24, 1.48552e-23, 8.02107e-23, 1.16671e-23], [1490.58, 6.20041e-23, 1.59867e-23, 7.18532e-24, 1.48945e-23, 8.04124e-23, 1.17467e-23], [1493.97, 6.21447e-23, 1.60158e-23, 7.19822e-24, 1.49338e-23, 8.06145e-23, 1.18265e-23], [1497.37, 6.22857e-23, 1.60451e-23, 7.21116e-24, 1.49733e-23, 8.0817e-23, 1.19064e-23], [1500.78, 6.24269e-23, 1.60744e-23, 7.22414e-24, 1.50129e-23, 8.10199e-23, 1.19866e-23], [1504.19, 6.25685e-23, 1.61037e-23, 7.23715e-24, 1.50525e-23, 8.12233e-23, 1.20669e-23], [1507.61, 6.27104e-23, 1.61332e-23, 7.2502e-24, 1.50922e-23, 8.14271e-23, 1.21475e-23], [1511.04, 6.28526e-23, 1.61627e-23, 7.26329e-24, 1.5132e-23, 8.16314e-23, 1.22281e-23], [1514.48, 6.29951e-23, 1.61923e-23, 7.27641e-24, 1.51719e-23, 8.1836e-23, 1.2309e-23], [1517.92, 6.3138e-23, 1.6222e-23, 7.28957e-24, 1.52118e-23, 8.20411e-23, 1.23901e-23], [1521.38, 6.32812e-23, 1.62518e-23, 7.30276e-24, 1.52519e-23, 8.22466e-23, 1.24713e-23], [1524.84, 6.34247e-23, 1.62817e-23, 7.31599e-24, 1.5292e-23, 8.24526e-23, 1.25527e-23], [1528.31, 6.35686e-23, 1.63116e-23, 7.32926e-24, 1.53322e-23, 8.2659e-23, 1.26343e-23], [1531.78, 6.37127e-23, 1.63416e-23, 7.34257e-24, 1.53725e-23, 8.28658e-23, 1.27161e-23], [1535.27, 6.38572e-23, 1.63717e-23, 7.35591e-24, 1.54129e-23, 8.3073e-23, 1.2798e-23], [1538.76, 6.40021e-23, 1.64019e-23, 7.36929e-24, 1.54534e-23, 8.32807e-23, 1.28801e-23], [1542.26, 6.41472e-23, 1.64322e-23, 7.3827e-24, 1.54939e-23, 8.34889e-23, 1.29624e-23], [1545.77, 6.42927e-23, 1.64625e-23, 7.39616e-24, 1.55345e-23, 8.36974e-23, 1.30449e-23], [1549.28, 6.44385e-23, 1.6493e-23, 7.40965e-24, 1.55753e-23, 8.39064e-23, 1.31275e-23], [1552.81, 6.45847e-23, 1.65235e-23, 7.42317e-24, 1.56161e-23, 8.41159e-23, 1.32104e-23], [1556.34, 6.47312e-23, 1.65541e-23, 7.43674e-24, 1.5657e-23, 8.43258e-23, 1.32934e-23], [1559.88, 6.4878e-23, 1.65848e-23, 7.45034e-24, 1.56979e-23, 8.45361e-23, 1.33766e-23], [1563.43, 6.50251e-23, 1.66155e-23, 7.46398e-24, 1.5739e-23, 8.47469e-23, 1.34599e-23], [1566.98, 6.51726e-23, 1.66464e-23, 7.47766e-24, 1.57802e-23, 8.49581e-23, 1.35435e-23], [1570.55, 6.53205e-23, 1.66773e-23, 7.49137e-24, 1.58214e-23, 8.51698e-23, 1.36272e-23], [1574.12, 6.54686e-23, 1.67083e-23, 7.50512e-24, 1.58627e-23, 8.53819e-23, 1.37111e-23], [1577.7, 6.56171e-23, 1.67394e-23, 7.51891e-24, 1.59041e-23, 8.55945e-23, 1.37952e-23], [1581.29, 6.57659e-23, 1.67706e-23, 7.53274e-24, 1.59456e-23, 8.58075e-23, 1.38794e-23], [1584.88, 6.59151e-23, 1.68018e-23, 7.54661e-24, 1.59872e-23, 8.60209e-23, 1.39639e-23], [1588.49, 6.60646e-23, 1.68332e-23, 7.56051e-24, 1.60289e-23, 8.62348e-23, 1.40485e-23], [1592.1, 6.62145e-23, 1.68646e-23, 7.57445e-24, 1.60706e-23, 8.64492e-23, 1.41333e-23], [1595.72, 6.63647e-23, 1.68961e-23, 7.58843e-24, 1.61125e-23, 8.6664e-23, 1.42182e-23], [1599.35, 6.65152e-23, 1.69277e-23, 7.60245e-24, 1.61544e-23, 8.68793e-23, 1.43034e-23], [1602.99, 6.66661e-23, 1.69594e-23, 7.6165e-24, 1.61964e-23, 8.7095e-23, 1.43887e-23], [1606.64, 6.68173e-23, 1.69912e-23, 7.6306e-24, 1.62385e-23, 8.73112e-23, 1.44742e-23], [1610.29, 6.69689e-23, 1.70231e-23, 7.64473e-24, 1.62807e-23, 8.75278e-23, 1.45599e-23], [1613.95, 6.71208e-23, 1.7055e-23, 7.6589e-24, 1.6323e-23, 8.77449e-23, 1.46457e-23], [1617.63, 6.72731e-23, 1.7087e-23, 7.67311e-24, 1.63654e-23, 8.79625e-23, 1.47317e-23], [1621.31, 6.74257e-23, 1.71191e-23, 7.68736e-24, 1.64078e-23, 8.81805e-23, 1.48179e-23], [1624.99, 6.75786e-23, 1.71513e-23, 7.70165e-24, 1.64504e-23, 8.8399e-23, 1.49043e-23], [1628.69, 6.77319e-23, 1.71836e-23, 7.71598e-24, 1.6493e-23, 8.86179e-23, 1.49909e-23], [1632.39, 6.78856e-23, 1.7216e-23, 7.73034e-24, 1.65357e-23, 8.88374e-23, 1.50776e-23], [1636.11, 6.80396e-23, 1.72484e-23, 7.74475e-24, 1.65786e-23, 8.90572e-23, 1.51645e-23], [1639.83, 6.81939e-23, 1.7281e-23, 7.75919e-24, 1.66215e-23, 8.92776e-23, 1.52516e-23], [1643.56, 6.83486e-23, 1.73136e-23, 7.77367e-24, 1.66645e-23, 8.94984e-23, 1.53389e-23], [1647.3, 6.85037e-23, 1.73463e-23, 7.7882e-24, 1.67075e-23, 8.97197e-23, 1.54264e-23], [1651.04, 6.86591e-23, 1.73791e-23, 7.80276e-24, 1.67507e-23, 8.99414e-23, 1.5514e-23], [1654.8, 6.88149e-23, 1.7412e-23, 7.81736e-24, 1.6794e-23, 9.01636e-23, 1.56018e-23], [1658.56, 6.8971e-23, 1.7445e-23, 7.832e-24, 1.68374e-23, 9.03863e-23, 1.56898e-23], [1662.34, 6.91275e-23, 1.74781e-23, 7.84668e-24, 1.68808e-23, 9.06095e-23, 1.57779e-23], [1666.12, 6.92843e-23, 1.75112e-23, 7.8614e-24, 1.69243e-23, 9.08331e-23, 1.58663e-23], [1669.91, 6.94415e-23, 1.75445e-23, 7.87616e-24, 1.6968e-23, 9.10572e-23, 1.59548e-23], [1673.7, 6.9599e-23, 1.75778e-23, 7.89096e-24, 1.70117e-23, 9.12818e-23, 1.60435e-23], [1677.51, 6.97569e-23, 1.76112e-23, 7.9058e-24, 1.70555e-23, 9.15068e-23, 1.61324e-23], [1681.33, 6.99152e-23, 1.76447e-23, 7.92067e-24, 1.70994e-23, 9.17324e-23, 1.62214e-23], [1685.15, 7.00738e-23, 1.76783e-23, 7.93559e-24, 1.71434e-23, 9.19584e-23, 1.63106e-23], [1688.98, 7.02328e-23, 1.7712e-23, 7.95055e-24, 1.71875e-23, 9.21849e-23, 1.64e-23], [1692.83, 7.03921e-23, 1.77458e-23, 7.96555e-24, 1.72317e-23, 9.24119e-23, 1.64896e-23], [1696.68, 7.05518e-23, 1.77797e-23, 7.98059e-24, 1.7276e-23, 9.26394e-23, 1.65794e-23], [1700.54, 7.07119e-23, 1.78136e-23, 7.99567e-24, 1.73203e-23, 9.28673e-23, 1.66693e-23], [1704.4, 7.08723e-23, 1.78476e-23, 8.01079e-24, 1.73648e-23, 9.30957e-23, 1.67595e-23], [1708.28, 7.10331e-23, 1.78818e-23, 8.02595e-24, 1.74094e-23, 9.33247e-23, 1.68498e-23], [1712.17, 7.11943e-23, 1.7916e-23, 8.04116e-24, 1.7454e-23, 9.35541e-23, 1.69402e-23], [1716.06, 7.13558e-23, 1.79503e-23, 8.0564e-24, 1.74988e-23, 9.3784e-23, 1.70309e-23], [1719.96, 7.15177e-23, 1.79847e-23, 8.07168e-24, 1.75436e-23, 9.40143e-23, 1.71217e-23], [1723.88, 7.168e-23, 1.80192e-23, 8.08701e-24, 1.75885e-23, 9.42452e-23, 1.72128e-23], [1727.8, 7.18427e-23, 1.80538e-23, 8.10237e-24, 1.76336e-23, 9.44766e-23, 1.73039e-23], [1731.73, 7.20057e-23, 1.80885e-23, 8.11778e-24, 1.76787e-23, 9.47085e-23, 1.73953e-23], [1735.67, 7.2169e-23, 1.81233e-23, 8.13323e-24, 1.77239e-23, 9.49408e-23, 1.74869e-23], [1739.61, 7.23328e-23, 1.81581e-23, 8.14872e-24, 1.77692e-23, 9.51737e-23, 1.75786e-23], [1743.57, 7.24969e-23, 1.81931e-23, 8.16425e-24, 1.78146e-23, 9.5407e-23, 1.76705e-23], [1747.54, 7.26614e-23, 1.82281e-23, 8.17982e-24, 1.78601e-23, 9.56409e-23, 1.77626e-23], [1751.51, 7.28263e-23, 1.82632e-23, 8.19543e-24, 1.79057e-23, 9.58752e-23, 1.78549e-23], [1755.5, 7.29915e-23, 1.82985e-23, 8.21109e-24, 1.79514e-23, 9.61101e-23, 1.79473e-23], [1759.49, 7.31572e-23, 1.83338e-23, 8.22678e-24, 1.79972e-23, 9.63454e-23, 1.804e-23], [1763.49, 7.33232e-23, 1.83692e-23, 8.24252e-24, 1.80431e-23, 9.65813e-23, 1.81328e-23], [1767.5, 7.34895e-23, 1.84047e-23, 8.2583e-24, 1.80891e-23, 9.68176e-23, 1.82258e-23], [1771.52, 7.36563e-23, 1.84403e-23, 8.27412e-24, 1.81352e-23, 9.70545e-23, 1.83189e-23], [1775.55, 7.38234e-23, 1.8476e-23, 8.28999e-24, 1.81813e-23, 9.72918e-23, 1.84123e-23], [1779.59, 7.3991e-23, 1.85118e-23, 8.30589e-24, 1.82276e-23, 9.75297e-23, 1.85058e-23], [1783.64, 7.41589e-23, 1.85477e-23, 8.32184e-24, 1.8274e-23, 9.77681e-23, 1.85996e-23], [1787.69, 7.43271e-23, 1.85837e-23, 8.33783e-24, 1.83204e-23, 9.8007e-23, 1.86934e-23], [1791.76, 7.44958e-23, 1.86197e-23, 8.35386e-24, 1.8367e-23, 9.82464e-23, 1.87875e-23], [1795.84, 7.46649e-23, 1.86559e-23, 8.36994e-24, 1.84137e-23, 9.84863e-23, 1.88818e-23], [1799.92, 7.48343e-23, 1.86921e-23, 8.38606e-24, 1.84604e-23, 9.87267e-23, 1.89762e-23], [1804.02, 7.50041e-23, 1.87285e-23, 8.40222e-24, 1.85073e-23, 9.89677e-23, 1.90708e-23], [1808.12, 7.51743e-23, 1.87649e-23, 8.41842e-24, 1.85543e-23, 9.92091e-23, 1.91656e-23], [1812.23, 7.53449e-23, 1.88015e-23, 8.43467e-24, 1.86013e-23, 9.94511e-23, 1.92606e-23], [1816.35, 7.55159e-23, 1.88381e-23, 8.45096e-24, 1.86485e-23, 9.96936e-23, 1.93558e-23], [1820.48, 7.56873e-23, 1.88748e-23, 8.46729e-24, 1.86957e-23, 9.99366e-23, 1.94511e-23], [1824.63, 7.5859e-23, 1.89117e-23, 8.48366e-24, 1.87431e-23, 1.0018e-22, 1.95467e-23], [1828.78, 7.60312e-23, 1.89486e-23, 8.50008e-24, 1.87906e-23, 1.00424e-22, 1.96424e-23], [1832.94, 7.62037e-23, 1.89856e-23, 8.51654e-24, 1.88381e-23, 1.00669e-22, 1.97383e-23], [1837.1, 7.63767e-23, 1.90227e-23, 8.53305e-24, 1.88858e-23, 1.00914e-22, 1.98344e-23], [1841.28, 7.655e-23, 1.90599e-23, 8.5496e-24, 1.89335e-23, 1.0116e-22, 1.99306e-23], [1845.47, 7.67237e-23, 1.90972e-23, 8.56619e-24, 1.89814e-23, 1.01406e-22, 2.00271e-23], [1849.67, 7.68978e-23, 1.91346e-23, 8.58282e-24, 1.90294e-23, 1.01652e-22, 2.01237e-23], [1853.88, 7.70724e-23, 1.91721e-23, 8.5995e-24, 1.90774e-23, 1.019e-22, 2.02205e-23], [1858.09, 7.72473e-23, 1.92097e-23, 8.61622e-24, 1.91256e-23, 1.02147e-22, 2.03175e-23], [1862.32, 7.74226e-23, 1.92474e-23, 8.63299e-24, 1.91738e-23, 1.02396e-22, 2.04147e-23], [1866.56, 7.75983e-23, 1.92852e-23, 8.6498e-24, 1.92222e-23, 1.02644e-22, 2.0512e-23], [1870.8, 7.77744e-23, 1.93231e-23, 8.66666e-24, 1.92707e-23, 1.02894e-22, 2.06096e-23], [1875.06, 7.79509e-23, 1.93611e-23, 8.68356e-24, 1.93193e-23, 1.03144e-22, 2.07073e-23], [1879.32, 7.81278e-23, 1.93992e-23, 8.7005e-24, 1.93679e-23, 1.03394e-22, 2.08052e-23], [1883.6, 7.83052e-23, 1.94374e-23, 8.71749e-24, 1.94167e-23, 1.03645e-22, 2.09033e-23], [1887.88, 7.84829e-23, 1.94757e-23, 8.73452e-24, 1.94656e-23, 1.03896e-22, 2.10016e-23], [1892.17, 7.8661e-23, 1.95141e-23, 8.7516e-24, 1.95146e-23, 1.04148e-22, 2.11e-23], [1896.48, 7.88395e-23, 1.95525e-23, 8.76872e-24, 1.95637e-23, 1.04401e-22, 2.11987e-23], [1900.79, 7.90185e-23, 1.95911e-23, 8.78588e-24, 1.96129e-23, 1.04654e-22, 2.12975e-23], [1905.12, 7.91978e-23, 1.96298e-23, 8.8031e-24, 1.96621e-23, 1.04908e-22, 2.13966e-23], [1909.45, 7.93776e-23, 1.96686e-23, 8.82035e-24, 1.97115e-23, 1.05162e-22, 2.14958e-23], [1913.79, 7.95577e-23, 1.97075e-23, 8.83766e-24, 1.97611e-23, 1.05417e-22, 2.15952e-23], [1918.14, 7.97383e-23, 1.97465e-23, 8.85507e-24, 1.98107e-23, 1.05672e-22, 2.16948e-23], [1922.51, 7.99194e-23, 1.9786e-23, 8.87332e-24, 1.98608e-23, 1.05928e-22, 2.17949e-23], [1926.88, 8.01007e-23, 1.98247e-23, 8.88985e-24, 1.99102e-23, 1.06184e-22, 2.18945e-23], [1931.26, 8.02825e-23, 1.9864e-23, 8.90731e-24, 1.99601e-23, 1.06441e-22, 2.19946e-23], [1935.66, 8.04647e-23, 1.99034e-23, 8.92483e-24, 2.00101e-23, 1.06699e-22, 2.20949e-23], [1940.06, 8.06473e-23, 1.99429e-23, 8.9424e-24, 2.00603e-23, 1.06957e-22, 2.21954e-23], [1944.47, 8.08304e-23, 1.99825e-23, 8.96002e-24, 2.01105e-23, 1.07216e-22, 2.22961e-23], [1948.89, 8.10139e-23, 2.00222e-23, 8.97768e-24, 2.01609e-23, 1.07475e-22, 2.2397e-23], [1953.33, 8.11977e-23, 2.00619e-23, 8.99539e-24, 2.02113e-23, 1.07734e-22, 2.24981e-23], [1957.77, 8.1382e-23, 2.01018e-23, 9.01315e-24, 2.02619e-23, 1.07995e-22, 2.25993e-23], [1962.22, 8.15668e-23, 2.01418e-23, 9.03095e-24, 2.03126e-23, 1.08256e-22, 2.27008e-23], [1966.69, 8.17519e-23, 2.01819e-23, 9.0488e-24, 2.03633e-23, 1.08517e-22, 2.28024e-23], [1971.16, 8.19375e-23, 2.02221e-23, 9.06669e-24, 2.04142e-23, 1.08779e-22, 2.29042e-23], [1975.64, 8.21235e-23, 2.02624e-23, 9.08463e-24, 2.04652e-23, 1.09042e-22, 2.30063e-23], [1980.14, 8.23099e-23, 2.03028e-23, 9.10262e-24, 2.05163e-23, 1.09305e-22, 2.31085e-23], [1984.64, 8.24967e-23, 2.03433e-23, 9.12065e-24, 2.05675e-23, 1.09568e-22, 2.32109e-23], [1989.16, 8.2684e-23, 2.0384e-23, 9.13873e-24, 2.06188e-23, 1.09833e-22, 2.33134e-23], [1993.68, 8.28717e-23, 2.04247e-23, 9.15686e-24, 2.06703e-23, 1.10097e-22, 2.34162e-23], [1998.21, 8.30598e-23, 2.04655e-23, 9.17503e-24, 2.07218e-23, 1.10363e-22, 2.35192e-23], [2002.76, 8.32483e-23, 2.05064e-23, 9.19325e-24, 2.07735e-23, 1.10629e-22, 2.36223e-23], [2007.32, 8.34373e-23, 2.05475e-23, 9.21152e-24, 2.08252e-23, 1.10895e-22, 2.37257e-23], [2011.88, 8.36267e-23, 2.05886e-23, 9.22983e-24, 2.08771e-23, 1.11162e-22, 2.38292e-23], [2016.46, 8.38165e-23, 2.06298e-23, 9.24819e-24, 2.09291e-23, 1.1143e-22, 2.39329e-23], [2021.04, 8.40068e-23, 2.06712e-23, 9.2666e-24, 2.09811e-23, 1.11698e-22, 2.40368e-23], [2025.64, 8.41975e-23, 2.07126e-23, 9.28506e-24, 2.10333e-23, 1.11967e-22, 2.41409e-23], [2030.25, 8.43886e-23, 2.07542e-23, 9.30356e-24, 2.10856e-23, 1.12237e-22, 2.42452e-23], [2034.87, 8.45802e-23, 2.07958e-23, 9.32211e-24, 2.11381e-23, 1.12507e-22, 2.43497e-23], [2039.49, 8.47722e-23, 2.08376e-23, 9.34071e-24, 2.11906e-23, 1.12777e-22, 2.44544e-23], [2044.13, 8.49646e-23, 2.08795e-23, 9.35936e-24, 2.12432e-23, 1.13049e-22, 2.45593e-23], [2048.78, 8.51575e-23, 2.09215e-23, 9.37805e-24, 2.1296e-23, 1.1332e-22, 2.46643e-23], [2053.44, 8.53508e-23, 2.09635e-23, 9.39679e-24, 2.13489e-23, 1.13593e-22, 2.47696e-23], [2058.11, 8.55446e-23, 2.10057e-23, 9.41558e-24, 2.14018e-23, 1.13866e-22, 2.4875e-23], [2062.8, 8.57388e-23, 2.1048e-23, 9.43442e-24, 2.14549e-23, 1.14139e-22, 2.49807e-23], [2067.49, 8.59334e-23, 2.10904e-23, 9.45331e-24, 2.15081e-23, 1.14413e-22, 2.50865e-23], [2072.19, 8.61285e-23, 2.1133e-23, 9.47224e-24, 2.15614e-23, 1.14688e-22, 2.51925e-23], [2076.9, 8.6324e-23, 2.11756e-23, 9.49123e-24, 2.16149e-23, 1.14963e-22, 2.52988e-23], [2081.63, 8.652e-23, 2.12183e-23, 9.51026e-24, 2.16684e-23, 1.15239e-22, 2.54052e-23], [2086.36, 8.67164e-23, 2.12612e-23, 9.52934e-24, 2.17221e-23, 1.15516e-22, 2.55118e-23], [2091.11, 8.69133e-23, 2.13041e-23, 9.54847e-24, 2.17759e-23, 1.15793e-22, 2.56186e-23], [2095.86, 8.71106e-23, 2.13472e-23, 9.56765e-24, 2.18297e-23, 1.16071e-22, 2.57256e-23], [2100.63, 8.73084e-23, 2.13903e-23, 9.58688e-24, 2.18837e-23, 1.16349e-22, 2.58328e-23], [2105.41, 8.75066e-23, 2.14336e-23, 9.60615e-24, 2.19379e-23, 1.16628e-22, 2.59402e-23], [2110.2, 8.77052e-23, 2.1477e-23, 9.62548e-24, 2.19921e-23, 1.16908e-22, 2.60478e-23], [2115.0, 8.79043e-23, 2.15205e-23, 9.64485e-24, 2.20464e-23, 1.17188e-22, 2.61556e-23], [2119.81, 8.81039e-23, 2.15641e-23, 9.66428e-24, 2.21009e-23, 1.17468e-22, 2.62635e-23], [2124.63, 8.83039e-23, 2.16078e-23, 9.68375e-24, 2.21555e-23, 1.1775e-22, 2.63717e-23], [2129.46, 8.85044e-23, 2.16516e-23, 9.70328e-24, 2.22102e-23, 1.18032e-22, 2.64801e-23], [2134.31, 8.87053e-23, 2.16955e-23, 9.72285e-24, 2.2265e-23, 1.18314e-22, 2.65887e-23], [2139.16, 8.89067e-23, 2.17396e-23, 9.74247e-24, 2.23199e-23, 1.18598e-22, 2.66974e-23], [2144.03, 8.91086e-23, 2.17838e-23, 9.76215e-24, 2.2375e-23, 1.18881e-22, 2.68064e-23], [2148.9, 8.93109e-23, 2.1828e-23, 9.78187e-24, 2.24301e-23, 1.19166e-22, 2.69156e-23], [2153.79, 8.95137e-23, 2.18724e-23, 9.80164e-24, 2.24854e-23, 1.19451e-22, 2.70249e-23], [2158.69, 8.97169e-23, 2.19169e-23, 9.82147e-24, 2.25408e-23, 1.19737e-22, 2.71345e-23], [2163.6, 8.99206e-23, 2.19615e-23, 9.84134e-24, 2.25964e-23, 1.20023e-22, 2.72443e-23], [2168.52, 9.01247e-23, 2.20062e-23, 9.86127e-24, 2.2652e-23, 1.2031e-22, 2.73542e-23], [2173.45, 9.03294e-23, 2.2051e-23, 9.88124e-24, 2.27077e-23, 1.20597e-22, 2.74644e-23], [2178.4, 9.05344e-23, 2.2096e-23, 9.90127e-24, 2.27636e-23, 1.20885e-22, 2.75747e-23], [2183.35, 9.074e-23, 2.2141e-23, 9.92134e-24, 2.28196e-23, 1.21174e-22, 2.76853e-23], [2188.32, 9.0946e-23, 2.21862e-23, 9.94147e-24, 2.28757e-23, 1.21464e-22, 2.77961e-23], [2193.3, 9.11525e-23, 2.22315e-23, 9.96165e-24, 2.2932e-23, 1.21754e-22, 2.7907e-23], [2198.29, 9.13595e-23, 2.22769e-23, 9.98188e-24, 2.29883e-23, 1.22044e-22, 2.80182e-23], [2203.29, 9.15669e-23, 2.23224e-23, 1.00022e-23, 2.30448e-23, 1.22336e-22, 2.81295e-23], [2208.3, 9.17748e-23, 2.2368e-23, 1.00225e-23, 2.31014e-23, 1.22628e-22, 2.82411e-23], [2213.32, 9.19832e-23, 2.24137e-23, 1.00429e-23, 2.31581e-23, 1.2292e-22, 2.83529e-23], [2218.35, 9.2192e-23, 2.24596e-23, 1.00633e-23, 2.3215e-23, 1.23213e-22, 2.84648e-23], [2223.4, 9.24014e-23, 2.25056e-23, 1.00838e-23, 2.32719e-23, 1.23507e-22, 2.8577e-23], [2228.46, 9.26112e-23, 2.25516e-23, 1.01043e-23, 2.3329e-23, 1.23802e-22, 2.86894e-23], [2233.53, 9.28214e-23, 2.25978e-23, 1.01249e-23, 2.33862e-23, 1.24097e-22, 2.8802e-23], [2238.61, 9.30322e-23, 2.26441e-23, 1.01456e-23, 2.34435e-23, 1.24393e-22, 2.89148e-23], [2243.7, 9.32434e-23, 2.26906e-23, 1.01663e-23, 2.3501e-23, 1.24689e-22, 2.90277e-23], [2248.8, 9.34551e-23, 2.27371e-23, 1.0187e-23, 2.35586e-23, 1.24986e-22, 2.91409e-23], [2253.92, 9.36673e-23, 2.27838e-23, 1.02078e-23, 2.36163e-23, 1.25284e-22, 2.92543e-23], [2259.04, 9.388e-23, 2.28305e-23, 1.02287e-23, 2.36741e-23, 1.25582e-22, 2.93679e-23], [2264.18, 9.40932e-23, 2.28774e-23, 1.02496e-23, 2.3732e-23, 1.25881e-22, 2.94817e-23], [2269.33, 9.43069e-23, 2.29245e-23, 1.02705e-23, 2.37901e-23, 1.26181e-22, 2.95957e-23], [2274.49, 9.4521e-23, 2.29716e-23, 1.02915e-23, 2.38483e-23, 1.26481e-22, 2.97099e-23], [2279.67, 9.47356e-23, 2.30188e-23, 1.03126e-23, 2.39066e-23, 1.26782e-22, 2.98244e-23], [2284.85, 9.49507e-23, 2.30662e-23, 1.03337e-23, 2.39651e-23, 1.27084e-22, 2.9939e-23], [2290.05, 9.51663e-23, 2.31137e-23, 1.03549e-23, 2.40236e-23, 1.27386e-22, 3.00538e-23], [2295.26, 9.53824e-23, 2.31613e-23, 1.03761e-23, 2.40823e-23, 1.27689e-22, 3.01689e-23], [2300.48, 9.5599e-23, 2.3209e-23, 1.03974e-23, 2.41412e-23, 1.27993e-22, 3.02841e-23], [2305.71, 9.58161e-23, 2.32568e-23, 1.04187e-23, 2.42001e-23, 1.28297e-22, 3.03996e-23], [2310.96, 9.60337e-23, 2.33048e-23, 1.04401e-23, 2.42592e-23, 1.28602e-22, 3.05153e-23], [2316.21, 9.62517e-23, 2.33529e-23, 1.04615e-23, 2.43184e-23, 1.28908e-22, 3.06311e-23], [2321.48, 9.64703e-23, 2.34011e-23, 1.0483e-23, 2.43777e-23, 1.29214e-22, 3.07472e-23], [2326.76, 9.66894e-23, 2.34494e-23, 1.05045e-23, 2.44372e-23, 1.29521e-22, 3.08635e-23], [2332.05, 9.69089e-23, 2.34978e-23, 1.05261e-23, 2.44967e-23, 1.29828e-22, 3.098e-23], [2337.36, 9.7129e-23, 2.35464e-23, 1.05478e-23, 2.45565e-23, 1.30137e-22, 3.10967e-23], [2342.67, 9.73495e-23, 2.3595e-23, 1.05695e-23, 2.46163e-23, 1.30446e-22, 3.12137e-23], [2348.0, 9.75706e-23, 2.36438e-23, 1.05912e-23, 2.46763e-23, 1.30755e-22, 3.13308e-23], [2353.34, 9.77922e-23, 2.36927e-23, 1.0613e-23, 2.47364e-23, 1.31066e-22, 3.14481e-23], [2358.7, 9.80142e-23, 2.37418e-23, 1.06349e-23, 2.47966e-23, 1.31377e-22, 3.15657e-23], [2364.06, 9.82368e-23, 2.37909e-23, 1.06568e-23, 2.48569e-23, 1.31688e-22, 3.16835e-23], [2369.44, 9.84599e-23, 2.38402e-23, 1.06788e-23, 2.49174e-23, 1.32001e-22, 3.18014e-23], [2374.83, 9.86835e-23, 2.38896e-23, 1.07008e-23, 2.4978e-23, 1.32314e-22, 3.19196e-23], [2380.23, 9.89076e-23, 2.39392e-23, 1.07229e-23, 2.50388e-23, 1.32628e-22, 3.2038e-23], [2385.64, 9.91322e-23, 2.39888e-23, 1.07451e-23, 2.50997e-23, 1.32942e-22, 3.21567e-23], [2391.07, 9.93573e-23, 2.40386e-23, 1.07673e-23, 2.51607e-23, 1.33257e-22, 3.22755e-23], [2396.51, 9.95829e-23, 2.40885e-23, 1.07895e-23, 2.52218e-23, 1.33573e-22, 3.23946e-23], [2401.96, 9.98576e-23, 2.43384e-23, 1.1251e-23, 2.5474e-23, 1.33926e-22, 3.26625e-23], [2407.42, 1.00036e-22, 2.41887e-23, 1.08342e-23, 2.53445e-23, 1.34207e-22, 3.26333e-23], [2412.9, 1.00263e-22, 2.42389e-23, 1.08566e-23, 2.5406e-23, 1.34524e-22, 3.2753e-23], [2418.39, 1.00491e-22, 2.42893e-23, 1.08791e-23, 2.54677e-23, 1.34843e-22, 3.28729e-23], [2423.89, 1.00719e-22, 2.43398e-23, 1.09016e-23, 2.55295e-23, 1.35162e-22, 3.2993e-23], [2429.4, 1.00948e-22, 2.43905e-23, 1.09242e-23, 2.55914e-23, 1.35482e-22, 3.31134e-23], [2434.93, 1.01177e-22, 2.44412e-23, 1.09468e-23, 2.56535e-23, 1.35803e-22, 3.32339e-23], [2440.47, 1.01407e-22, 2.44921e-23, 1.09695e-23, 2.57157e-23, 1.36124e-22, 3.33547e-23], [2446.02, 1.01637e-22, 2.45431e-23, 1.09923e-23, 2.5778e-23, 1.36446e-22, 3.34757e-23], [2451.58, 1.01868e-22, 2.45943e-23, 1.10151e-23, 2.58405e-23, 1.36769e-22, 3.35969e-23], [2457.16, 1.02099e-22, 2.46456e-23, 1.1038e-23, 2.59031e-23, 1.37093e-22, 3.37183e-23], [2462.75, 1.02331e-22, 2.4697e-23, 1.10609e-23, 2.59659e-23, 1.37417e-22, 3.384e-23], [2468.35, 1.02563e-22, 2.47485e-23, 1.10839e-23, 2.60287e-23, 1.37742e-22, 3.39619e-23], [2473.96, 1.02796e-22, 2.48001e-23, 1.11069e-23, 2.60918e-23, 1.38067e-22, 3.40839e-23], [2479.59, 1.0303e-22, 2.48519e-23, 1.113e-23, 2.61549e-23, 1.38394e-22, 3.42063e-23], [2485.23, 1.03264e-22, 2.49038e-23, 1.11532e-23, 2.62182e-23, 1.38721e-22, 3.43288e-23], [2490.88, 1.03498e-22, 2.49558e-23, 1.11764e-23, 2.62816e-23, 1.39049e-22, 3.44515e-23], [2496.55, 1.03733e-22, 2.5008e-23, 1.11997e-23, 2.63452e-23, 1.39377e-22, 3.45745e-23], [2502.23, 1.03969e-22, 2.50603e-23, 1.1223e-23, 2.64089e-23, 1.39706e-22, 3.46977e-23], [2507.92, 1.04205e-22, 2.51127e-23, 1.12464e-23, 2.64727e-23, 1.40036e-22, 3.48211e-23], [2513.62, 1.04442e-22, 2.51653e-23, 1.12698e-23, 2.65367e-23, 1.40367e-22, 3.49448e-23], [2519.34, 1.04679e-22, 2.52179e-23, 1.12933e-23, 2.66008e-23, 1.40698e-22, 3.50686e-23], [2525.07, 1.04917e-22, 2.52707e-23, 1.13169e-23, 2.6665e-23, 1.4103e-22, 3.51927e-23], [2530.81, 1.05155e-22, 2.53237e-23, 1.13405e-23, 2.67294e-23, 1.41363e-22, 3.5317e-23], [2536.57, 1.05394e-22, 2.53767e-23, 1.13641e-23, 2.67939e-23, 1.41697e-22, 3.54415e-23], [2542.34, 1.05633e-22, 2.54299e-23, 1.13879e-23, 2.68586e-23, 1.42031e-22, 3.55663e-23], [2548.12, 1.05873e-22, 2.54833e-23, 1.14117e-23, 2.69234e-23, 1.42366e-22, 3.56913e-23], [2553.92, 1.06113e-22, 2.55367e-23, 1.14355e-23, 2.69884e-23, 1.42702e-22, 3.58165e-23], [2559.73, 1.06354e-22, 2.55903e-23, 1.14594e-23, 2.70534e-23, 1.43039e-22, 3.59419e-23], [2565.55, 1.06596e-22, 2.5644e-23, 1.14834e-23, 2.71187e-23, 1.43376e-22, 3.60676e-23], [2571.39, 1.06838e-22, 2.56979e-23, 1.15074e-23, 2.7184e-23, 1.43714e-22, 3.61935e-23], [2577.23, 1.07081e-22, 2.57519e-23, 1.15315e-23, 2.72495e-23, 1.44053e-22, 3.63196e-23], [2583.1, 1.07324e-22, 2.5806e-23, 1.15557e-23, 2.73152e-23, 1.44392e-22, 3.64459e-23], [2588.97, 1.07568e-22, 2.58602e-23, 1.15799e-23, 2.7381e-23, 1.44732e-22, 3.65725e-23], [2594.86, 1.07812e-22, 2.59146e-23, 1.16041e-23, 2.74469e-23, 1.45073e-22, 3.66993e-23], [2600.76, 1.08057e-22, 2.59691e-23, 1.16284e-23, 2.7513e-23, 1.45415e-22, 3.68263e-23], [2606.68, 1.08302e-22, 2.60238e-23, 1.16528e-23, 2.75792e-23, 1.45758e-22, 3.69536e-23], [2612.61, 1.08548e-22, 2.60785e-23, 1.16773e-23, 2.76456e-23, 1.46101e-22, 3.70811e-23], [2618.55, 1.08795e-22, 2.61334e-23, 1.17018e-23, 2.77121e-23, 1.46445e-22, 3.72088e-23], [2624.51, 1.09042e-22, 2.61885e-23, 1.17263e-23, 2.77788e-23, 1.4679e-22, 3.73368e-23], [2630.48, 1.0929e-22, 2.62437e-23, 1.1751e-23, 2.78456e-23, 1.47135e-22, 3.74649e-23], [2636.46, 1.09538e-22, 2.6299e-23, 1.17757e-23, 2.79125e-23, 1.47481e-22, 3.75934e-23], [2642.46, 1.09787e-22, 2.63545e-23, 1.18004e-23, 2.79796e-23, 1.47828e-22, 3.7722e-23], [2648.47, 1.10036e-22, 2.641e-23, 1.18252e-23, 2.80468e-23, 1.48176e-22, 3.78509e-23], [2654.49, 1.10286e-22, 2.64658e-23, 1.18501e-23, 2.81142e-23, 1.48525e-22, 3.798e-23], [2660.53, 1.10536e-22, 2.65216e-23, 1.1875e-23, 2.81817e-23, 1.48874e-22, 3.81093e-23], [2666.58, 1.10787e-22, 2.65776e-23, 1.19e-23, 2.82494e-23, 1.49224e-22, 3.82389e-23], [2672.64, 1.11039e-22, 2.66338e-23, 1.1925e-23, 2.83172e-23, 1.49575e-22, 3.83687e-23], [2678.72, 1.11291e-22, 2.669e-23, 1.19501e-23, 2.83852e-23, 1.49927e-22, 3.84988e-23], [2684.82, 1.11544e-22, 2.67464e-23, 1.19753e-23, 2.84533e-23, 1.50279e-22, 3.86291e-23], [2690.92, 1.11797e-22, 2.6803e-23, 1.20006e-23, 2.85216e-23, 1.50633e-22, 3.87596e-23], [2697.04, 1.12051e-22, 2.68597e-23, 1.20259e-23, 2.859e-23, 1.50987e-22, 3.88904e-23], [2703.18, 1.12306e-22, 2.69165e-23, 1.20512e-23, 2.86586e-23, 1.51341e-22, 3.90214e-23], [2709.33, 1.12561e-22, 2.69734e-23, 1.20766e-23, 2.87273e-23, 1.51697e-22, 3.91526e-23], [2715.49, 1.12817e-22, 2.70305e-23, 1.21021e-23, 2.87961e-23, 1.52053e-22, 3.92841e-23], [2721.67, 1.13073e-22, 2.70878e-23, 1.21277e-23, 2.88651e-23, 1.5241e-22, 3.94158e-23], [2727.86, 1.1333e-22, 2.71451e-23, 1.21533e-23, 2.89343e-23, 1.52768e-22, 3.95477e-23], [2734.06, 1.13587e-22, 2.72027e-23, 1.21789e-23, 2.90036e-23, 1.53127e-22, 3.96799e-23], [2740.28, 1.13845e-22, 2.72603e-23, 1.22047e-23, 2.90731e-23, 1.53487e-22, 3.98123e-23], [2746.51, 1.14104e-22, 2.73181e-23, 1.22305e-23, 2.91427e-23, 1.53847e-22, 3.9945e-23], [2752.76, 1.14363e-22, 2.7376e-23, 1.22563e-23, 2.92124e-23, 1.54208e-22, 4.00779e-23], [2759.02, 1.14623e-22, 2.74341e-23, 1.22823e-23, 2.92823e-23, 1.5457e-22, 4.02111e-23], [2765.3, 1.14883e-22, 2.74923e-23, 1.23082e-23, 2.93524e-23, 1.54932e-22, 4.03445e-23], [2771.59, 1.15144e-22, 2.75507e-23, 1.23343e-23, 2.94226e-23, 1.55296e-22, 4.04781e-23], [2777.89, 1.15406e-22, 2.76092e-23, 1.23604e-23, 2.9493e-23, 1.5566e-22, 4.0612e-23], [2784.21, 1.15668e-22, 2.76678e-23, 1.23866e-23, 2.95635e-23, 1.56025e-22, 4.07461e-23], [2790.54, 1.1593e-22, 2.77266e-23, 1.24128e-23, 2.96342e-23, 1.56391e-22, 4.08805e-23], [2796.89, 1.16194e-22, 2.77855e-23, 1.24391e-23, 2.9705e-23, 1.56758e-22, 4.10151e-23], [2803.25, 1.16458e-22, 2.78446e-23, 1.24655e-23, 2.9776e-23, 1.57126e-22, 4.115e-23], [2809.63, 1.16722e-22, 2.79038e-23, 1.24919e-23, 2.98472e-23, 1.57494e-22, 4.12851e-23], [2816.02, 1.16987e-22, 2.79632e-23, 1.25184e-23, 2.99185e-23, 1.57863e-22, 4.14204e-23], [2822.42, 1.17253e-22, 2.80227e-23, 1.2545e-23, 2.99899e-23, 1.58233e-22, 4.1556e-23], [2828.84, 1.1752e-22, 2.80823e-23, 1.25716e-23, 3.00615e-23, 1.58604e-22, 4.16919e-23], [2835.28, 1.17787e-22, 2.81421e-23, 1.25983e-23, 3.01333e-23, 1.58975e-22, 4.1828e-23], [2841.73, 1.18054e-22, 2.8202e-23, 1.2625e-23, 3.02052e-23, 1.59348e-22, 4.19643e-23], [2848.19, 1.18322e-22, 2.82621e-23, 1.26519e-23, 3.02773e-23, 1.59721e-22, 4.21009e-23], [2854.67, 1.18591e-22, 2.83224e-23, 1.26787e-23, 3.03495e-23, 1.60095e-22, 4.22377e-23], [2861.16, 1.1886e-22, 2.83827e-23, 1.27057e-23, 3.04219e-23, 1.6047e-22, 4.23748e-23], [2867.67, 1.1913e-22, 2.84432e-23, 1.27327e-23, 3.04945e-23, 1.60846e-22, 4.25122e-23], [2874.19, 1.19401e-22, 2.85039e-23, 1.27598e-23, 3.05672e-23, 1.61222e-22, 4.26498e-23], [2880.73, 1.19672e-22, 2.85647e-23, 1.2787e-23, 3.06401e-23, 1.616e-22, 4.27876e-23], [2887.28, 1.19944e-22, 2.86257e-23, 1.28142e-23, 3.07131e-23, 1.61978e-22, 4.29257e-23], [2893.85, 1.20217e-22, 2.86868e-23, 1.28414e-23, 3.07863e-23, 1.62357e-22, 4.3064e-23], [2900.43, 1.2049e-22, 2.8748e-23, 1.28688e-23, 3.08596e-23, 1.62737e-22, 4.32026e-23], [2907.03, 1.20763e-22, 2.88094e-23, 1.28962e-23, 3.09331e-23, 1.63118e-22, 4.33415e-23], [2913.64, 1.21038e-22, 2.8871e-23, 1.29237e-23, 3.10068e-23, 1.63499e-22, 4.34806e-23], [2920.27, 1.21313e-22, 2.89327e-23, 1.29512e-23, 3.10806e-23, 1.63882e-22, 4.362e-23], [2926.91, 1.21588e-22, 2.89945e-23, 1.29788e-23, 3.11546e-23, 1.64265e-22, 4.37596e-23], [2933.57, 1.21865e-22, 2.90565e-23, 1.30065e-23, 3.12287e-23, 1.64649e-22, 4.38995e-23], [2940.24, 1.22141e-22, 2.91186e-23, 1.30343e-23, 3.1303e-23, 1.65034e-22, 4.40396e-23], [2946.93, 1.22419e-22, 2.91809e-23, 1.30621e-23, 3.13775e-23, 1.6542e-22, 4.418e-23], [2953.63, 1.22697e-22, 2.92434e-23, 1.30899e-23, 3.14522e-23, 1.65807e-22, 4.43206e-23], [2960.35, 1.22976e-22, 2.9306e-23, 1.31179e-23, 3.1527e-23, 1.66194e-22, 4.44615e-23], [2967.08, 1.23255e-22, 2.93687e-23, 1.31459e-23, 3.16019e-23, 1.66583e-22, 4.46027e-23], [2973.83, 1.23535e-22, 2.94316e-23, 1.3174e-23, 3.1677e-23, 1.66972e-22, 4.47441e-23], [2980.6, 1.23816e-22, 2.94946e-23, 1.32021e-23, 3.17523e-23, 1.67362e-22, 4.48858e-23], [2987.38, 1.24097e-22, 2.95578e-23, 1.32304e-23, 3.18278e-23, 1.67753e-22, 4.50278e-23], [2994.17, 1.24379e-22, 2.96212e-23, 1.32586e-23, 3.19034e-23, 1.68145e-22, 4.517e-23], [3000.98, 1.24661e-22, 2.96847e-23, 1.3287e-23, 3.19792e-23, 1.68538e-22, 4.53124e-23], [3007.81, 1.24945e-22, 2.97484e-23, 1.33154e-23, 3.20552e-23, 1.68931e-22, 4.54552e-23], [3014.65, 1.25229e-22, 2.98122e-23, 1.33439e-23, 3.21313e-23, 1.69326e-22, 4.55981e-23], [3021.51, 1.25513e-22, 2.98761e-23, 1.33725e-23, 3.22076e-23, 1.69721e-22, 4.57414e-23], [3028.38, 1.25798e-22, 2.99402e-23, 1.34011e-23, 3.2284e-23, 1.70117e-22, 4.58849e-23], [3035.27, 1.26084e-22, 3.00045e-23, 1.34298e-23, 3.23606e-23, 1.70514e-22, 4.60287e-23], [3042.17, 1.2637e-22, 3.00689e-23, 1.34586e-23, 3.24374e-23, 1.70912e-22, 4.61727e-23], [3049.09, 1.26657e-22, 3.01335e-23, 1.34874e-23, 3.25144e-23, 1.71311e-22, 4.63171e-23], [3056.03, 1.26945e-22, 3.01983e-23, 1.35163e-23, 3.25915e-23, 1.71711e-22, 4.64616e-23], [3062.98, 1.27234e-22, 3.02631e-23, 1.35453e-23, 3.26688e-23, 1.72112e-22, 4.66065e-23], [3069.95, 1.27523e-22, 3.03282e-23, 1.35743e-23, 3.27463e-23, 1.72513e-22, 4.67516e-23], [3076.93, 1.27812e-22, 3.03934e-23, 1.36035e-23, 3.28239e-23, 1.72916e-22, 4.6897e-23], [3083.93, 1.28103e-22, 3.04587e-23, 1.36327e-23, 3.29017e-23, 1.73319e-22, 4.70426e-23], [3090.94, 1.28394e-22, 3.05243e-23, 1.36619e-23, 3.29797e-23, 1.73723e-22, 4.71886e-23], [3097.97, 1.28685e-22, 3.05899e-23, 1.36912e-23, 3.30578e-23, 1.74128e-22, 4.73348e-23], [3105.02, 1.28978e-22, 3.06558e-23, 1.37206e-23, 3.31361e-23, 1.74534e-22, 4.74812e-23], [3112.08, 1.29271e-22, 3.07218e-23, 1.37501e-23, 3.32146e-23, 1.74941e-22, 4.76279e-23], [3119.16, 1.29565e-22, 3.07879e-23, 1.37796e-23, 3.32933e-23, 1.75349e-22, 4.7775e-23], [3126.25, 1.29859e-22, 3.08542e-23, 1.38093e-23, 3.33721e-23, 1.75758e-22, 4.79222e-23], [3133.37, 1.30154e-22, 3.09207e-23, 1.38389e-23, 3.34511e-23, 1.76168e-22, 4.80698e-23], [3140.49, 1.3045e-22, 3.09873e-23, 1.38687e-23, 3.35303e-23, 1.76578e-22, 4.82176e-23], [3147.64, 1.30746e-22, 3.10541e-23, 1.38985e-23, 3.36097e-23, 1.7699e-22, 4.83657e-23], [3154.8, 1.31043e-22, 3.1121e-23, 1.39284e-23, 3.36892e-23, 1.77402e-22, 4.85141e-23], [3161.97, 1.31341e-22, 3.11881e-23, 1.39584e-23, 3.37689e-23, 1.77815e-22, 4.86627e-23], [3169.16, 1.31639e-22, 3.12554e-23, 1.39884e-23, 3.38488e-23, 1.78229e-22, 4.88116e-23], [3176.37, 1.31938e-22, 3.13228e-23, 1.40185e-23, 3.39288e-23, 1.78645e-22, 4.89609e-23], [3183.6, 1.32238e-22, 3.13904e-23, 1.40487e-23, 3.40091e-23, 1.79061e-22, 4.91103e-23], [3190.84, 1.32539e-22, 3.14581e-23, 1.4079e-23, 3.40895e-23, 1.79478e-22, 4.92601e-23], [3198.1, 1.3284e-22, 3.1526e-23, 1.41093e-23, 3.417e-23, 1.79896e-22, 4.94101e-23], [3205.37, 1.33141e-22, 3.15941e-23, 1.41397e-23, 3.42508e-23, 1.80314e-22, 4.95604e-23], [3212.66, 1.33444e-22, 3.16623e-23, 1.41702e-23, 3.43317e-23, 1.80734e-22, 4.9711e-23], [3219.97, 1.33747e-22, 3.17307e-23, 1.42007e-23, 3.44128e-23, 1.81155e-22, 4.98619e-23], [3227.29, 1.34051e-22, 3.17993e-23, 1.42314e-23, 3.44941e-23, 1.81577e-22, 5.00131e-23], [3234.63, 1.34356e-22, 3.1868e-23, 1.42621e-23, 3.45756e-23, 1.81999e-22, 5.01645e-23], [3241.99, 1.34661e-22, 3.19369e-23, 1.42928e-23, 3.46573e-23, 1.82423e-22, 5.03162e-23], [3249.36, 1.34967e-22, 3.2006e-23, 1.43237e-23, 3.47391e-23, 1.82847e-22, 5.04682e-23], [3256.76, 1.35273e-22, 3.20752e-23, 1.43546e-23, 3.48211e-23, 1.83272e-22, 5.06205e-23], [3264.16, 1.35581e-22, 3.21446e-23, 1.43856e-23, 3.49033e-23, 1.83699e-22, 5.07731e-23], [3271.59, 1.35889e-22, 3.22141e-23, 1.44166e-23, 3.49857e-23, 1.84126e-22, 5.0926e-23], [3279.03, 1.36198e-22, 3.22838e-23, 1.44478e-23, 3.50682e-23, 1.84554e-22, 5.10791e-23], [3286.49, 1.36507e-22, 3.23537e-23, 1.4479e-23, 3.51509e-23, 1.84984e-22, 5.12325e-23], [3293.96, 1.36817e-22, 3.24238e-23, 1.45103e-23, 3.52339e-23, 1.85414e-22, 5.13862e-23], [3301.46, 1.37128e-22, 3.2494e-23, 1.45416e-23, 3.5317e-23, 1.85845e-22, 5.15402e-23], [3308.96, 1.3744e-22, 3.25644e-23, 1.45731e-23, 3.54002e-23, 1.86277e-22, 5.16945e-23], [3316.49, 1.37752e-22, 3.26349e-23, 1.46046e-23, 3.54837e-23, 1.8671e-22, 5.18491e-23], [3324.03, 1.38065e-22, 3.27056e-23, 1.46362e-23, 3.55673e-23, 1.87144e-22, 5.2004e-23], [3331.6, 1.38379e-22, 3.27765e-23, 1.46678e-23, 3.56512e-23, 1.87579e-22, 5.21592e-23], [3339.17, 1.38693e-22, 3.28476e-23, 1.46996e-23, 3.57352e-23, 1.88015e-22, 5.23146e-23], [3346.77, 1.39008e-22, 3.29188e-23, 1.47314e-23, 3.58194e-23, 1.88452e-22, 5.24704e-23], [3354.38, 1.39324e-22, 3.29902e-23, 1.47633e-23, 3.59038e-23, 1.8889e-22, 5.26264e-23], [3362.01, 1.39641e-22, 3.30618e-23, 1.47954e-23, 3.59884e-23, 1.89329e-22, 5.27828e-23], [3369.66, 1.39958e-22, 3.31335e-23, 1.48273e-23, 3.60731e-23, 1.89768e-22, 5.29394e-23], [3377.32, 1.40276e-22, 3.32054e-23, 1.48594e-23, 3.61581e-23, 1.90209e-22, 5.30963e-23], [3385.0, 1.40595e-22, 3.32775e-23, 1.48916e-23, 3.62432e-23, 1.90651e-22, 5.32535e-23], [3392.7, 1.40914e-22, 3.33497e-23, 1.49239e-23, 3.63285e-23, 1.91094e-22, 5.3411e-23], [3400.42, 1.41234e-22, 3.34221e-23, 1.49562e-23, 3.6414e-23, 1.91538e-22, 5.35688e-23], [3408.16, 1.41555e-22, 3.34947e-23, 1.49887e-23, 3.64997e-23, 1.91982e-22, 5.37269e-23], [3415.91, 1.41877e-22, 3.35675e-23, 1.50212e-23, 3.65856e-23, 1.92428e-22, 5.38853e-23], [3423.68, 1.42199e-22, 3.36404e-23, 1.50538e-23, 3.66717e-23, 1.92875e-22, 5.4044e-23], [3431.46, 1.42522e-22, 3.37135e-23, 1.50864e-23, 3.6758e-23, 1.93323e-22, 5.4203e-23], [3439.27, 1.42846e-22, 3.37868e-23, 1.51192e-23, 3.68444e-23, 1.93771e-22, 5.43623e-23], [3447.09, 1.43171e-22, 3.38603e-23, 1.5152e-23, 3.69311e-23, 1.94221e-22, 5.45219e-23], [3454.93, 1.43496e-22, 3.39339e-23, 1.51849e-23, 3.70179e-23, 1.94672e-22, 5.46818e-23], [3462.79, 1.43822e-22, 3.40077e-23, 1.52178e-23, 3.71049e-23, 1.95124e-22, 5.48421e-23], [3470.67, 1.44149e-22, 3.40817e-23, 1.52509e-23, 3.71921e-23, 1.95577e-22, 5.50026e-23], [3478.56, 1.44476e-22, 3.41559e-23, 1.5284e-23, 3.72796e-23, 1.9603e-22, 5.51634e-23], [3486.47, 1.44805e-22, 3.42302e-23, 1.53172e-23, 3.73672e-23, 1.96485e-22, 5.53245e-23], [3494.4, 1.45134e-22, 3.43047e-23, 1.53505e-23, 3.7455e-23, 1.96941e-22, 5.54859e-23], [3502.35, 1.45463e-22, 3.43794e-23, 1.53839e-23, 3.7543e-23, 1.97398e-22, 5.56477e-23], [3510.32, 1.45794e-22, 3.44543e-23, 1.54173e-23, 3.76312e-23, 1.97856e-22, 5.58097e-23], [3518.3, 1.46125e-22, 3.45293e-23, 1.54509e-23, 3.77195e-23, 1.98314e-22, 5.5972e-23], [3526.31, 1.46457e-22, 3.46045e-23, 1.54845e-23, 3.78081e-23, 1.98774e-22, 5.61347e-23], [3534.33, 1.4679e-22, 3.46799e-23, 1.55182e-23, 3.78969e-23, 1.99235e-22, 5.62976e-23], [3542.37, 1.47124e-22, 3.47555e-23, 1.55519e-23, 3.79859e-23, 1.99697e-22, 5.64609e-23], [3550.42, 1.47458e-22, 3.48313e-23, 1.55858e-23, 3.8075e-23, 2.0016e-22, 5.66245e-23], [3558.5, 1.47793e-22, 3.49072e-23, 1.56197e-23, 3.81644e-23, 2.00624e-22, 5.67884e-23], [3566.59, 1.48129e-22, 3.49833e-23, 1.56537e-23, 3.8254e-23, 2.01089e-22, 5.69526e-23], [3574.71, 1.48465e-22, 3.50596e-23, 1.56878e-23, 3.83437e-23, 2.01555e-22, 5.71171e-23], [3582.84, 1.48803e-22, 3.51361e-23, 1.5722e-23, 3.84337e-23, 2.02022e-22, 5.72819e-23], [3590.99, 1.49141e-22, 3.52128e-23, 1.57562e-23, 3.85238e-23, 2.02491e-22, 5.74471e-23], [3599.15, 1.4948e-22, 3.52896e-23, 1.57905e-23, 3.86142e-23, 2.0296e-22, 5.76125e-23], [3607.34, 1.49819e-22, 3.53666e-23, 1.58249e-23, 3.87047e-23, 2.0343e-22, 5.77783e-23], [3615.55, 1.5016e-22, 3.54438e-23, 1.58594e-23, 3.87955e-23, 2.03901e-22, 5.79444e-23], [3623.77, 1.50501e-22, 3.55212e-23, 1.5894e-23, 3.88865e-23, 2.04374e-22, 5.81108e-23], [3632.01, 1.50843e-22, 3.55988e-23, 1.59287e-23, 3.89776e-23, 2.04847e-22, 5.82775e-23], [3640.27, 1.51186e-22, 3.56765e-23, 1.59634e-23, 3.9069e-23, 2.05322e-22, 5.84446e-23], [3648.55, 1.51529e-22, 3.57545e-23, 1.59982e-23, 3.91605e-23, 2.05797e-22, 5.86119e-23], [3656.85, 1.51874e-22, 3.58326e-23, 1.60331e-23, 3.92523e-23, 2.06274e-22, 5.87796e-23], [3665.17, 1.52219e-22, 3.59109e-23, 1.60681e-23, 3.93443e-23, 2.06752e-22, 5.89476e-23], [3673.51, 1.52565e-22, 3.59894e-23, 1.61032e-23, 3.94364e-23, 2.0723e-22, 5.91159e-23], [3681.86, 1.52911e-22, 3.60681e-23, 1.61384e-23, 3.95288e-23, 2.0771e-22, 5.92846e-23], [3690.24, 1.53259e-22, 3.61469e-23, 1.61736e-23, 3.96214e-23, 2.08191e-22, 5.94535e-23], [3698.63, 1.53607e-22, 3.6226e-23, 1.62089e-23, 3.97142e-23, 2.08673e-22, 5.96228e-23], [3707.04, 1.53956e-22, 3.63052e-23, 1.62443e-23, 3.98072e-23, 2.09156e-22, 5.97925e-23], [3715.48, 1.54306e-22, 3.63846e-23, 1.62798e-23, 3.99004e-23, 2.0964e-22, 5.99624e-23], [3723.93, 1.54657e-22, 3.64643e-23, 1.63154e-23, 3.99938e-23, 2.10125e-22, 6.01327e-23], [3732.4, 1.55008e-22, 3.65441e-23, 1.6351e-23, 4.00874e-23, 2.10612e-22, 6.03033e-23], [3740.89, 1.5536e-22, 3.6624e-23, 1.63868e-23, 4.01812e-23, 2.11099e-22, 6.04742e-23], [3749.4, 1.55714e-22, 3.67042e-23, 1.64226e-23, 4.02752e-23, 2.11587e-22, 6.06454e-23], [3757.92, 1.56067e-22, 3.67846e-23, 1.64585e-23, 4.03695e-23, 2.12077e-22, 6.0817e-23], [3766.47, 1.56422e-22, 3.68651e-23, 1.64945e-23, 4.04639e-23, 2.12568e-22, 6.09889e-23], [3775.04, 1.56777e-22, 3.69459e-23, 1.65306e-23, 4.05586e-23, 2.13059e-22, 6.11612e-23], [3783.62, 1.57134e-22, 3.70268e-23, 1.65667e-23, 4.06534e-23, 2.13552e-22, 6.13337e-23], [3792.23, 1.57491e-22, 3.71079e-23, 1.6603e-23, 4.07485e-23, 2.14046e-22, 6.15066e-23], [3800.86, 1.57849e-22, 3.71893e-23, 1.66393e-23, 4.08438e-23, 2.14541e-22, 6.16799e-23], [3809.5, 1.58207e-22, 3.72708e-23, 1.66758e-23, 4.09393e-23, 2.15037e-22, 6.18535e-23], [3818.17, 1.58567e-22, 3.73525e-23, 1.67123e-23, 4.1035e-23, 2.15535e-22, 6.20274e-23], [3826.85, 1.58927e-22, 3.74344e-23, 1.67489e-23, 4.11309e-23, 2.16033e-22, 6.22016e-23], [3835.56, 1.59288e-22, 3.75165e-23, 1.67855e-23, 4.1227e-23, 2.16533e-22, 6.23762e-23], [3844.28, 1.5965e-22, 3.75988e-23, 1.68223e-23, 4.13234e-23, 2.17033e-22, 6.25511e-23], [3853.02, 1.60013e-22, 3.76812e-23, 1.68592e-23, 4.14199e-23, 2.17535e-22, 6.27264e-23], [3861.79, 1.60377e-22, 3.77639e-23, 1.68961e-23, 4.15167e-23, 2.18038e-22, 6.2902e-23], [3870.57, 1.60741e-22, 3.78468e-23, 1.69331e-23, 4.16137e-23, 2.18542e-22, 6.30779e-23], [3879.38, 1.61107e-22, 3.79298e-23, 1.69702e-23, 4.17109e-23, 2.19047e-22, 6.32542e-23], [3888.2, 1.61473e-22, 3.80131e-23, 1.70074e-23, 4.18083e-23, 2.19553e-22, 6.34308e-23], [3897.04, 1.6184e-22, 3.80965e-23, 1.70447e-23, 4.19059e-23, 2.20061e-22, 6.36077e-23], [3905.91, 1.62207e-22, 3.81802e-23, 1.70821e-23, 4.20038e-23, 2.20569e-22, 6.3785e-23], [3914.79, 1.62576e-22, 3.8264e-23, 1.71196e-23, 4.21018e-23, 2.21079e-22, 6.39627e-23], [3923.7, 1.62946e-22, 3.83481e-23, 1.71571e-23, 4.22001e-23, 2.2159e-22, 6.41407e-23], [3932.62, 1.63316e-22, 3.84323e-23, 1.71948e-23, 4.22986e-23, 2.22101e-22, 6.4319e-23], [3941.57, 1.63687e-22, 3.85167e-23, 1.72325e-23, 4.23973e-23, 2.22615e-22, 6.44977e-23], [3950.53, 1.64059e-22, 3.86014e-23, 1.72703e-23, 4.24963e-23, 2.23129e-22, 6.46767e-23], [3959.52, 1.64432e-22, 3.86862e-23, 1.73083e-23, 4.25954e-23, 2.23644e-22, 6.48561e-23], [3968.52, 1.64805e-22, 3.87712e-23, 1.73463e-23, 4.26948e-23, 2.24161e-22, 6.50358e-23], [3977.55, 1.6518e-22, 3.88565e-23, 1.73843e-23, 4.27944e-23, 2.24678e-22, 6.52159e-23], [3986.6, 1.65555e-22, 3.89419e-23, 1.74225e-23, 4.28943e-23, 2.25197e-22, 6.53963e-23], [3995.67, 1.65932e-22, 3.90275e-23, 1.74608e-23, 4.29943e-23, 2.25717e-22, 6.55771e-23], [4004.75, 1.66309e-22, 3.91134e-23, 1.74992e-23, 4.30946e-23, 2.26239e-22, 6.57582e-23], [4013.86, 1.66687e-22, 3.91994e-23, 1.75376e-23, 4.31951e-23, 2.26761e-22, 6.59397e-23], [4022.99, 1.67066e-22, 3.92857e-23, 1.75761e-23, 4.32958e-23, 2.27284e-22, 6.61216e-23], [4032.14, 1.67445e-22, 3.93721e-23, 1.76148e-23, 4.33967e-23, 2.27809e-22, 6.63037e-23], [4041.32, 1.67826e-22, 3.94587e-23, 1.76535e-23, 4.34979e-23, 2.28335e-22, 6.64863e-23], [4050.51, 1.68207e-22, 3.95456e-23, 1.76923e-23, 4.35993e-23, 2.28862e-22, 6.66692e-23], [4059.72, 1.68589e-22, 3.96327e-23, 1.77312e-23, 4.37009e-23, 2.2939e-22, 6.68525e-23], [4068.95, 1.68973e-22, 3.97199e-23, 1.77702e-23, 4.38027e-23, 2.2992e-22, 6.70361e-23], [4078.21, 1.69357e-22, 3.98074e-23, 1.78093e-23, 4.39048e-23, 2.3045e-22, 6.72201e-23], [4087.49, 1.69741e-22, 3.9895e-23, 1.78485e-23, 4.40071e-23, 2.30982e-22, 6.74044e-23], [4096.78, 1.70127e-22, 3.99829e-23, 1.78877e-23, 4.41096e-23, 2.31515e-22, 6.75891e-23], [4106.1, 1.70514e-22, 4.0071e-23, 1.79271e-23, 4.42123e-23, 2.32049e-22, 6.77742e-23], [4115.44, 1.70901e-22, 4.01593e-23, 1.79666e-23, 4.43153e-23, 2.32585e-22, 6.79596e-23], [4124.8, 1.7129e-22, 4.02478e-23, 1.80061e-23, 4.44185e-23, 2.33121e-22, 6.81454e-23], [4134.18, 1.71679e-22, 4.03365e-23, 1.80458e-23, 4.45219e-23, 2.33659e-22, 6.83315e-23], [4143.59, 1.72069e-22, 4.04254e-23, 1.80855e-23, 4.46256e-23, 2.34198e-22, 6.8518e-23], [4153.01, 1.7246e-22, 4.05145e-23, 1.81253e-23, 4.47295e-23, 2.34738e-22, 6.87049e-23], [4162.46, 1.72852e-22, 4.06038e-23, 1.81652e-23, 4.48336e-23, 2.3528e-22, 6.88922e-23], [4171.93, 1.73245e-22, 4.06934e-23, 1.82053e-23, 4.4938e-23, 2.35823e-22, 6.90798e-23], [4181.42, 1.73639e-22, 4.07831e-23, 1.82454e-23, 4.50426e-23, 2.36366e-22, 6.92678e-23], [4190.93, 1.74033e-22, 4.08731e-23, 1.82856e-23, 4.51474e-23, 2.36912e-22, 6.94561e-23], [4200.46, 1.74429e-22, 4.09633e-23, 1.83259e-23, 4.52524e-23, 2.37458e-22, 6.96449e-23], [4210.01, 1.74825e-22, 4.10536e-23, 1.83663e-23, 4.53577e-23, 2.38005e-22, 6.9834e-23], [4219.59, 1.75223e-22, 4.11442e-23, 1.84068e-23, 4.54633e-23, 2.38554e-22, 7.00234e-23], [4229.19, 1.75621e-22, 4.1235e-23, 1.84473e-23, 4.5569e-23, 2.39104e-22, 7.02133e-23], [4238.81, 1.7602e-22, 4.1326e-23, 1.8488e-23, 4.5675e-23, 2.39655e-22, 7.04035e-23], [4248.45, 1.7642e-22, 4.14173e-23, 1.85288e-23, 4.57812e-23, 2.40208e-22, 7.05941e-23], [4258.11, 1.76821e-22, 4.15087e-23, 1.85697e-23, 4.58877e-23, 2.40762e-22, 7.07851e-23], [4267.8, 1.77223e-22, 4.16004e-23, 1.86106e-23, 4.59944e-23, 2.41316e-22, 7.09764e-23], [4277.5, 1.77626e-22, 4.16922e-23, 1.86517e-23, 4.61014e-23, 2.41873e-22, 7.11681e-23], [4287.23, 1.78029e-22, 4.17843e-23, 1.86928e-23, 4.62085e-23, 2.4243e-22, 7.13602e-23], [4296.99, 1.78434e-22, 4.18766e-23, 1.87341e-23, 4.6316e-23, 2.42989e-22, 7.15527e-23], [4306.76, 1.7884e-22, 4.19692e-23, 1.87755e-23, 4.64236e-23, 2.43549e-22, 7.17456e-23], [4316.56, 1.79246e-22, 4.20619e-23, 1.88169e-23, 4.65315e-23, 2.4411e-22, 7.19388e-23], [4326.37, 1.79653e-22, 4.21548e-23, 1.88584e-23, 4.66396e-23, 2.44672e-22, 7.21325e-23], [4336.21, 1.80062e-22, 4.2248e-23, 1.89001e-23, 4.6748e-23, 2.45236e-22, 7.23265e-23], [4346.08, 1.80471e-22, 4.23414e-23, 1.89418e-23, 4.68566e-23, 2.45801e-22, 7.25209e-23], [4355.96, 1.80881e-22, 4.2435e-23, 1.89837e-23, 4.69655e-23, 2.46367e-22, 7.27157e-23], [4365.87, 1.81292e-22, 4.25288e-23, 1.90256e-23, 4.70746e-23, 2.46935e-22, 7.29108e-23], [4375.8, 1.81704e-22, 4.26229e-23, 1.90676e-23, 4.7184e-23, 2.47504e-22, 7.31064e-23], [4385.75, 1.82117e-22, 4.27171e-23, 1.91098e-23, 4.72936e-23, 2.48074e-22, 7.33023e-23], [4395.73, 1.82531e-22, 4.28116e-23, 1.9152e-23, 4.74034e-23, 2.48645e-22, 7.34987e-23], [4405.73, 1.82946e-22, 4.29063e-23, 1.91943e-23, 4.75135e-23, 2.49218e-22, 7.36954e-23], [4415.75, 1.83362e-22, 4.30013e-23, 1.92368e-23, 4.76238e-23, 2.49792e-22, 7.38925e-23], [4425.79, 1.83779e-22, 4.30964e-23, 1.92793e-23, 4.77344e-23, 2.50367e-22, 7.409e-23], [4435.86, 1.84196e-22, 4.31918e-23, 1.93219e-23, 4.78452e-23, 2.50944e-22, 7.42879e-23], [4445.95, 1.84615e-22, 4.32874e-23, 1.93646e-23, 4.79563e-23, 2.51521e-22, 7.44862e-23], [4456.06, 1.85034e-22, 4.33832e-23, 1.94075e-23, 4.80676e-23, 2.52101e-22, 7.46849e-23], [4466.2, 1.85455e-22, 4.34792e-23, 1.94504e-23, 4.81792e-23, 2.52681e-22, 7.4884e-23], [4476.36, 1.85877e-22, 4.35755e-23, 1.94934e-23, 4.8291e-23, 2.53263e-22, 7.50834e-23], [4486.54, 1.86299e-22, 4.3672e-23, 1.95365e-23, 4.8403e-23, 2.53846e-22, 7.52833e-23], [4496.74, 1.86722e-22, 4.37687e-23, 1.95798e-23, 4.85153e-23, 2.5443e-22, 7.54836e-23], [4506.97, 1.87147e-22, 4.38656e-23, 1.96231e-23, 4.86279e-23, 2.55016e-22, 7.56843e-23], [4517.22, 1.87572e-22, 4.39628e-23, 1.96665e-23, 4.87407e-23, 2.55603e-22, 7.58853e-23], [4527.5, 1.87998e-22, 4.40602e-23, 1.97101e-23, 4.88538e-23, 2.56191e-22, 7.60868e-23], [4537.8, 1.88426e-22, 4.41578e-23, 1.97537e-23, 4.89671e-23, 2.56781e-22, 7.62887e-23], [4548.12, 1.88854e-22, 4.42557e-23, 1.97974e-23, 4.90807e-23, 2.57372e-22, 7.6491e-23], [4558.46, 1.89283e-22, 4.43537e-23, 1.98413e-23, 4.91945e-23, 2.57964e-22, 7.66937e-23], [4568.83, 1.89713e-22, 4.4452e-23, 1.98852e-23, 4.93086e-23, 2.58557e-22, 7.68968e-23], [4579.22, 1.90145e-22, 4.45506e-23, 1.99293e-23, 4.94229e-23, 2.59152e-22, 7.71003e-23], [4589.64, 1.90577e-22, 4.46493e-23, 1.99734e-23, 4.95375e-23, 2.59749e-22, 7.73042e-23], [4600.08, 1.9101e-22, 4.47483e-23, 2.00177e-23, 4.96524e-23, 2.60346e-22, 7.75085e-23], [4610.54, 1.91444e-22, 4.48476e-23, 2.0062e-23, 4.97675e-23, 2.60945e-22, 7.77132e-23], [4621.03, 1.91879e-22, 4.4947e-23, 2.01065e-23, 4.98828e-23, 2.61546e-22, 7.79184e-23], [4631.54, 1.92315e-22, 4.50467e-23, 2.0151e-23, 4.99984e-23, 2.62147e-22, 7.81239e-23], [4642.07, 1.92753e-22, 4.51466e-23, 2.01957e-23, 5.01143e-23, 2.6275e-22, 7.83299e-23], [4652.63, 1.93191e-22, 4.52468e-23, 2.02405e-23, 5.02305e-23, 2.63355e-22, 7.85363e-23], [4663.22, 1.9363e-22, 4.53472e-23, 2.02853e-23, 5.03469e-23, 2.6396e-22, 7.8743e-23], [4673.82, 1.9407e-22, 4.54478e-23, 2.03303e-23, 5.04635e-23, 2.64567e-22, 7.89502e-23], [4684.45, 1.94511e-22, 4.55487e-23, 2.03754e-23, 5.05804e-23, 2.65176e-22, 7.91579e-23], [4695.11, 1.94953e-22, 4.56498e-23, 2.04206e-23, 5.06976e-23, 2.65786e-22, 7.93659e-23], [4705.79, 1.95396e-22, 4.57511e-23, 2.04659e-23, 5.0815e-23, 2.66397e-22, 7.95744e-23], [4716.49, 1.9584e-22, 4.58526e-23, 2.05113e-23, 5.09327e-23, 2.6701e-22, 7.97832e-23], [4727.22, 1.96285e-22, 4.59544e-23, 2.05568e-23, 5.10507e-23, 2.67624e-22, 7.99925e-23], [4737.97, 1.96732e-22, 4.60565e-23, 2.06024e-23, 5.11689e-23, 2.68239e-22, 8.02023e-23], [4748.75, 1.97179e-22, 4.61588e-23, 2.06481e-23, 5.12874e-23, 2.68856e-22, 8.04124e-23], [4759.55, 1.97627e-22, 4.62613e-23, 2.06939e-23, 5.14062e-23, 2.69474e-22, 8.0623e-23], [4770.38, 1.98076e-22, 4.6364e-23, 2.07399e-23, 5.15252e-23, 2.70093e-22, 8.0834e-23], [4781.23, 1.98526e-22, 4.6467e-23, 2.07859e-23, 5.16445e-23, 2.70714e-22, 8.10454e-23], [4792.1, 1.98978e-22, 4.65702e-23, 2.0832e-23, 5.1764e-23, 2.71336e-22, 8.12572e-23], [4803.0, 1.9943e-22, 4.66737e-23, 2.08783e-23, 5.18839e-23, 2.7196e-22, 8.14695e-23], [4813.93, 1.99883e-22, 4.67774e-23, 2.09247e-23, 5.2004e-23, 2.72585e-22, 8.16822e-23], [4824.88, 2.00337e-22, 4.68814e-23, 2.09711e-23, 5.21243e-23, 2.73212e-22, 8.18953e-23], [4835.85, 2.00793e-22, 4.69856e-23, 2.10177e-23, 5.2245e-23, 2.7384e-22, 8.21089e-23], [4846.85, 2.01249e-22, 4.709e-23, 2.10644e-23, 5.23659e-23, 2.74469e-22, 8.23229e-23], [4857.87, 2.01707e-22, 4.71947e-23, 2.11112e-23, 5.2487e-23, 2.751e-22, 8.25373e-23], [4868.92, 2.02165e-22, 4.72996e-23, 2.11581e-23, 5.26085e-23, 2.75732e-22, 8.27522e-23], [4880.0, 2.02625e-22, 4.74048e-23, 2.12051e-23, 5.27302e-23, 2.76365e-22, 8.29675e-23], [4891.1, 2.03085e-22, 4.75102e-23, 2.12522e-23, 5.28522e-23, 2.77e-22, 8.31832e-23], [4902.22, 2.03547e-22, 4.76159e-23, 2.12995e-23, 5.29744e-23, 2.77637e-22, 8.33994e-23], [4913.37, 2.04009e-22, 4.77218e-23, 2.13468e-23, 5.3097e-23, 2.78275e-22, 8.3616e-23], [4924.55, 2.04473e-22, 4.78279e-23, 2.13942e-23, 5.32198e-23, 2.78914e-22, 8.3833e-23], [4935.75, 2.04938e-22, 4.79343e-23, 2.14418e-23, 5.33428e-23, 2.79555e-22, 8.40505e-23], [4946.98, 2.05404e-22, 4.8041e-23, 2.14895e-23, 5.34662e-23, 2.80197e-22, 8.42684e-23], [4958.23, 2.05871e-22, 4.81478e-23, 2.15373e-23, 5.35898e-23, 2.80841e-22, 8.44868e-23], [4969.51, 2.06339e-22, 4.8255e-23, 2.15852e-23, 5.37137e-23, 2.81486e-22, 8.47056e-23], [4980.81, 2.06808e-22, 4.83624e-23, 2.16332e-23, 5.38379e-23, 2.82132e-22, 8.49249e-23], [4992.14, 2.07278e-22, 4.847e-23, 2.16813e-23, 5.39624e-23, 2.8278e-22, 8.51446e-23], [5003.5, 2.07749e-22, 4.85779e-23, 2.17295e-23, 5.40871e-23, 2.8343e-22, 8.53647e-23], [5014.88, 2.08221e-22, 4.8686e-23, 2.17779e-23, 5.42122e-23, 2.84081e-22, 8.55853e-23], [5026.28, 2.08694e-22, 4.87944e-23, 2.18263e-23, 5.43375e-23, 2.84733e-22, 8.58064e-23], [5037.72, 2.09169e-22, 4.89031e-23, 2.18749e-23, 5.4463e-23, 2.85387e-22, 8.60279e-23], [5049.17, 2.09644e-22, 4.9012e-23, 2.19236e-23, 5.45889e-23, 2.86042e-22, 8.62498e-23], [5060.66, 2.10121e-22, 4.91211e-23, 2.19724e-23, 5.47151e-23, 2.86699e-22, 8.64723e-23], [5072.17, 2.10598e-22, 4.92305e-23, 2.20213e-23, 5.48415e-23, 2.87358e-22, 8.66951e-23], [5083.71, 2.11077e-22, 4.93402e-23, 2.20703e-23, 5.49682e-23, 2.88017e-22, 8.69184e-23], [5095.27, 2.11557e-22, 4.94501e-23, 2.21194e-23, 5.50952e-23, 2.88679e-22, 8.71422e-23], [5106.86, 2.12038e-22, 4.95603e-23, 2.21687e-23, 5.52225e-23, 2.89341e-22, 8.73664e-23], [5118.48, 2.1252e-22, 4.96707e-23, 2.2218e-23, 5.535e-23, 2.90006e-22, 8.75911e-23], [5130.12, 2.13003e-22, 4.97814e-23, 2.22675e-23, 5.54779e-23, 2.90671e-22, 8.78163e-23], [5141.79, 2.13487e-22, 4.98923e-23, 2.23171e-23, 5.5606e-23, 2.91339e-22, 8.80419e-23], [5153.48, 2.13972e-22, 5.00035e-23, 2.23668e-23, 5.57345e-23, 2.92007e-22, 8.82679e-23], [5165.2, 2.14459e-22, 5.01149e-23, 2.24166e-23, 5.58632e-23, 2.92678e-22, 8.84945e-23], [5176.95, 2.14946e-22, 5.02266e-23, 2.24666e-23, 5.59922e-23, 2.93349e-22, 8.87214e-23], [5188.73, 2.15435e-22, 5.03386e-23, 2.25166e-23, 5.61215e-23, 2.94023e-22, 8.89489e-23], [5200.53, 2.15924e-22, 5.04508e-23, 2.25668e-23, 5.62511e-23, 2.94698e-22, 8.91768e-23], [5212.36, 2.16415e-22, 5.05633e-23, 2.26171e-23, 5.63809e-23, 2.95374e-22, 8.94052e-23], [5224.22, 2.16907e-22, 5.06761e-23, 2.26675e-23, 5.65111e-23, 2.96052e-22, 8.96341e-23], [5236.1, 2.174e-22, 5.07891e-23, 2.2718e-23, 5.66415e-23, 2.96731e-22, 8.98634e-23], [5248.01, 2.17894e-22, 5.09024e-23, 2.27686e-23, 5.67723e-23, 2.97412e-22, 9.00932e-23], [5259.95, 2.1839e-22, 5.10159e-23, 2.28194e-23, 5.69033e-23, 2.98094e-22, 9.03235e-23], [5271.91, 2.18886e-22, 5.11297e-23, 2.28703e-23, 5.70347e-23, 2.98778e-22, 9.05542e-23], [5283.9, 2.19384e-22, 5.12438e-23, 2.29213e-23, 5.71663e-23, 2.99464e-22, 9.07854e-23], [5295.92, 2.19882e-22, 5.13581e-23, 2.29724e-23, 5.72982e-23, 3.00151e-22, 9.10171e-23], [5307.97, 2.20382e-22, 5.14727e-23, 2.30236e-23, 5.74304e-23, 3.0084e-22, 9.12493e-23], [5320.04, 2.20883e-22, 5.15875e-23, 2.30749e-23, 5.7563e-23, 3.0153e-22, 9.14819e-23], [5332.14, 2.21385e-22, 5.17026e-23, 2.31264e-23, 5.76958e-23, 3.02222e-22, 9.1715e-23], [5344.27, 2.21888e-22, 5.1818e-23, 2.3178e-23, 5.78289e-23, 3.02915e-22, 9.19487e-23], [5356.42, 2.22393e-22, 5.19337e-23, 2.32297e-23, 5.79623e-23, 3.0361e-22, 9.21827e-23], [5368.61, 2.22898e-22, 5.20496e-23, 2.32815e-23, 5.8096e-23, 3.04306e-22, 9.24173e-23], [5380.82, 2.23405e-22, 5.21658e-23, 2.33335e-23, 5.823e-23, 3.05004e-22, 9.26523e-23], [5393.06, 2.23913e-22, 5.22823e-23, 2.33855e-23, 5.83643e-23, 3.05704e-22, 9.28879e-23], [5405.32, 2.24422e-22, 5.2399e-23, 2.34377e-23, 5.8499e-23, 3.06405e-22, 9.31239e-23], [5417.62, 2.24932e-22, 5.2516e-23, 2.349e-23, 5.86339e-23, 3.07108e-22, 9.33604e-23], [5429.94, 2.25443e-22, 5.26333e-23, 2.35425e-23, 5.87691e-23, 3.07812e-22, 9.35974e-23], [5442.29, 2.25956e-22, 5.27508e-23, 2.3595e-23, 5.89046e-23, 3.08518e-22, 9.38349e-23], [5454.67, 2.26469e-22, 5.28687e-23, 2.36477e-23, 5.90404e-23, 3.09225e-22, 9.40729e-23], [5467.08, 2.26984e-22, 5.29867e-23, 2.37005e-23, 5.91766e-23, 3.09934e-22, 9.43113e-23], [5479.51, 2.275e-22, 5.31051e-23, 2.37534e-23, 5.9313e-23, 3.10645e-22, 9.45503e-23], [5491.98, 2.28017e-22, 5.32238e-23, 2.38064e-23, 5.94497e-23, 3.11357e-22, 9.47897e-23], [5504.47, 2.28536e-22, 5.33427e-23, 2.38596e-23, 5.95868e-23, 3.12071e-22, 9.50297e-23], [5516.99, 2.29055e-22, 5.34619e-23, 2.39129e-23, 5.97241e-23, 3.12787e-22, 9.52701e-23], [5529.54, 2.29576e-22, 5.35813e-23, 2.39663e-23, 5.98618e-23, 3.13504e-22, 9.55111e-23], [5542.12, 2.30098e-22, 5.37011e-23, 2.40198e-23, 5.99998e-23, 3.14223e-22, 9.57525e-23], [5554.72, 2.30621e-22, 5.38211e-23, 2.40735e-23, 6.01381e-23, 3.14943e-22, 9.59944e-23], [5567.36, 2.31145e-22, 5.39414e-23, 2.41273e-23, 6.02767e-23, 3.15665e-22, 9.62369e-23], [5580.02, 2.31671e-22, 5.4062e-23, 2.41812e-23, 6.04156e-23, 3.16389e-22, 9.64798e-23], [5592.71, 2.32197e-22, 5.41828e-23, 2.42352e-23, 6.05548e-23, 3.17114e-22, 9.67233e-23], [5605.43, 2.32725e-22, 5.43039e-23, 2.42894e-23, 6.06943e-23, 3.17841e-22, 9.69672e-23], [5618.18, 2.33254e-22, 5.44254e-23, 2.43437e-23, 6.08342e-23, 3.1857e-22, 9.72117e-23], [5630.96, 2.33784e-22, 5.4547e-23, 2.43981e-23, 6.09743e-23, 3.193e-22, 9.74566e-23], [5643.77, 2.34316e-22, 5.4669e-23, 2.44526e-23, 6.11148e-23, 3.20032e-22, 9.77021e-23], [5656.61, 2.34848e-22, 5.47913e-23, 2.45073e-23, 6.12556e-23, 3.20765e-22, 9.79481e-23], [5669.47, 2.35382e-22, 5.49138e-23, 2.4562e-23, 6.13967e-23, 3.215e-22, 9.81946e-23], [5682.37, 2.35917e-22, 5.50366e-23, 2.46169e-23, 6.15381e-23, 3.22237e-22, 9.84416e-23], [5695.29, 2.36454e-22, 5.51598e-23, 2.4672e-23, 6.16799e-23, 3.22975e-22, 9.86891e-23], [5708.25, 2.36991e-22, 5.52831e-23, 2.47272e-23, 6.18219e-23, 3.23716e-22, 9.89371e-23], [5721.23, 2.3753e-22, 5.54068e-23, 2.47825e-23, 6.19643e-23, 3.24457e-22, 9.91857e-23], [5734.25, 2.3807e-22, 5.55308e-23, 2.48379e-23, 6.2107e-23, 3.25201e-22, 9.94347e-23], [5747.29, 2.38611e-22, 5.5655e-23, 2.48934e-23, 6.225e-23, 3.25946e-22, 9.96843e-23], [5760.36, 2.39153e-22, 5.57796e-23, 2.49491e-23, 6.23934e-23, 3.26693e-22, 9.99344e-23], [5773.46, 2.39697e-22, 5.59044e-23, 2.50049e-23, 6.2537e-23, 3.27441e-22, 1.00185e-22], [5786.6, 2.40242e-22, 5.60295e-23, 2.50608e-23, 6.2681e-23, 3.28192e-22, 1.00436e-22], [5799.76, 2.40788e-22, 5.61549e-23, 2.51169e-23, 6.28253e-23, 3.28944e-22, 1.00688e-22], [5812.95, 2.41336e-22, 5.62806e-23, 2.51731e-23, 6.29699e-23, 3.29697e-22, 1.0094e-22], [5826.17, 2.41884e-22, 5.64066e-23, 2.52294e-23, 6.31149e-23, 3.30452e-22, 1.01193e-22], [5839.42, 2.42434e-22, 5.65329e-23, 2.52859e-23, 6.32602e-23, 3.31209e-22, 1.01446e-22], [5852.71, 2.42985e-22, 5.66594e-23, 2.53425e-23, 6.34058e-23, 3.31968e-22, 1.017e-22], [5866.02, 2.43537e-22, 5.67863e-23, 2.53992e-23, 6.35517e-23, 3.32729e-22, 1.01954e-22], [5879.36, 2.44091e-22, 5.69135e-23, 2.5456e-23, 6.3698e-23, 3.33491e-22, 1.02209e-22], [5892.73, 2.44646e-22, 5.70409e-23, 2.5513e-23, 6.38446e-23, 3.34255e-22, 1.02464e-22], [5906.14, 2.45202e-22, 5.71686e-23, 2.55701e-23, 6.39915e-23, 3.3502e-22, 1.0272e-22], [5919.57, 2.4576e-22, 5.72967e-23, 2.56274e-23, 6.41388e-23, 3.35788e-22, 1.02977e-22], [5933.04, 2.46318e-22, 5.7425e-23, 2.56848e-23, 6.42864e-23, 3.36557e-22, 1.03234e-22], [5946.53, 2.46878e-22, 5.75536e-23, 2.57423e-23, 6.44343e-23, 3.37328e-22, 1.03491e-22], [5960.06, 2.47439e-22, 5.76826e-23, 2.57999e-23, 6.45825e-23, 3.381e-22, 1.03749e-22], [5973.61, 2.48002e-22, 5.78118e-23, 2.58577e-23, 6.47311e-23, 3.38874e-22, 1.04008e-22], [5987.2, 2.48566e-22, 5.79413e-23, 2.59156e-23, 6.48801e-23, 3.3965e-22, 1.04267e-22], [6000.82, 2.49131e-22, 5.80711e-23, 2.59736e-23, 6.50293e-23, 3.40428e-22, 1.04527e-22], [6014.47, 2.49697e-22, 5.82012e-23, 2.60318e-23, 6.51789e-23, 3.41208e-22, 1.04787e-22], [6028.15, 2.50265e-22, 5.83317e-23, 2.60901e-23, 6.53288e-23, 3.41989e-22, 1.05048e-22], [6041.86, 2.50834e-22, 5.84624e-23, 2.61486e-23, 6.54791e-23, 3.42772e-22, 1.05309e-22], [6055.6, 2.51404e-22, 5.85934e-23, 2.62071e-23, 6.56297e-23, 3.43557e-22, 1.05571e-22], [6069.38, 2.51975e-22, 5.87247e-23, 2.62659e-23, 6.57806e-23, 3.44344e-22, 1.05833e-22], [6083.18, 2.52548e-22, 5.88564e-23, 2.63247e-23, 6.59319e-23, 3.45132e-22, 1.06096e-22], [6097.02, 2.53122e-22, 5.89883e-23, 2.63837e-23, 6.60835e-23, 3.45922e-22, 1.0636e-22], [6110.89, 2.53698e-22, 5.91205e-23, 2.64428e-23, 6.62355e-23, 3.46714e-22, 1.06624e-22], [6124.79, 2.54275e-22, 5.92531e-23, 2.65021e-23, 6.63878e-23, 3.47508e-22, 1.06888e-22], [6138.72, 2.54853e-22, 5.93859e-23, 2.65615e-23, 6.65405e-23, 3.48304e-22, 1.07154e-22], [6152.68, 2.55432e-22, 5.95191e-23, 2.6621e-23, 6.66934e-23, 3.49101e-22, 1.07419e-22], [6166.67, 2.56013e-22, 5.96525e-23, 2.66807e-23, 6.68468e-23, 3.499e-22, 1.07686e-22], [6180.7, 2.56595e-22, 5.97863e-23, 2.67405e-23, 6.70005e-23, 3.50701e-22, 1.07952e-22], [6194.76, 2.57178e-22, 5.99204e-23, 2.68004e-23, 6.71545e-23, 3.51504e-22, 1.0822e-22], [6208.85, 2.57763e-22, 6.00548e-23, 2.68605e-23, 6.73089e-23, 3.52308e-22, 1.08488e-22], [6222.97, 2.58349e-22, 6.01895e-23, 2.69208e-23, 6.74636e-23, 3.53115e-22, 1.08756e-22], [6237.13, 2.58936e-22, 6.03245e-23, 2.69811e-23, 6.76187e-23, 3.53923e-22, 1.09025e-22], [6251.31, 2.59525e-22, 6.04598e-23, 2.70416e-23, 6.77741e-23, 3.54733e-22, 1.09295e-22], [6265.53, 2.60115e-22, 6.05954e-23, 2.71023e-23, 6.79298e-23, 3.55545e-22, 1.09565e-22], [6279.78, 2.60706e-22, 6.07314e-23, 2.7163e-23, 6.8086e-23, 3.56359e-22, 1.09836e-22], [6294.07, 2.61299e-22, 6.08676e-23, 2.7224e-23, 6.82424e-23, 3.57174e-22, 1.10108e-22], [6308.38, 2.61893e-22, 6.10042e-23, 2.7285e-23, 6.83993e-23, 3.57992e-22, 1.10379e-22], [6322.73, 2.62488e-22, 6.11411e-23, 2.73462e-23, 6.85564e-23, 3.58811e-22, 1.10652e-22], [6337.11, 2.63085e-22, 6.12783e-23, 2.74076e-23, 6.8714e-23, 3.59632e-22, 1.10925e-22], [6351.53, 2.63683e-22, 6.14158e-23, 2.74691e-23, 6.88718e-23, 3.60455e-22, 1.11199e-22], [6365.98, 2.64282e-22, 6.15536e-23, 2.75307e-23, 6.90301e-23, 3.6128e-22, 1.11473e-22], [6380.46, 2.64883e-22, 6.16918e-23, 2.75925e-23, 6.91887e-23, 3.62106e-22, 1.11748e-22], [6394.97, 2.65485e-22, 6.18302e-23, 2.76544e-23, 6.93476e-23, 3.62935e-22, 1.12023e-22], [6409.51, 2.66089e-22, 6.1969e-23, 2.77164e-23, 6.95069e-23, 3.63765e-22, 1.12299e-22], [6424.09, 2.66694e-22, 6.21081e-23, 2.77786e-23, 6.96666e-23, 3.64598e-22, 1.12576e-22], [6438.71, 2.673e-22, 6.22476e-23, 2.7841e-23, 6.98266e-23, 3.65432e-22, 1.12853e-22], [6453.35, 2.67908e-22, 6.23873e-23, 2.79035e-23, 6.9987e-23, 3.66268e-22, 1.13131e-22], [6468.03, 2.68517e-22, 6.25274e-23, 2.79661e-23, 7.01478e-23, 3.67106e-22, 1.13409e-22], [6482.74, 2.69127e-22, 6.26678e-23, 2.80289e-23, 7.03089e-23, 3.67946e-22, 1.13688e-22], [6497.49, 2.69739e-22, 6.28085e-23, 2.80918e-23, 7.04704e-23, 3.68788e-22, 1.13968e-22], [6512.27, 2.70352e-22, 6.29496e-23, 2.81548e-23, 7.06322e-23, 3.69631e-22, 1.14248e-22], [6527.08, 2.70967e-22, 6.30909e-23, 2.8218e-23, 7.07944e-23, 3.70477e-22, 1.14528e-22], [6541.92, 2.71583e-22, 6.32326e-23, 2.82814e-23, 7.0957e-23, 3.71324e-22, 1.1481e-22], [6556.8, 2.72201e-22, 6.33746e-23, 2.83449e-23, 7.11199e-23, 3.72174e-22, 1.15092e-22], [6571.72, 2.72819e-22, 6.3517e-23, 2.84086e-23, 7.12833e-23, 3.73025e-22, 1.15374e-22], [6586.67, 2.7344e-22, 6.36597e-23, 2.84723e-23, 7.14469e-23, 3.73878e-22, 1.15657e-22], [6601.65, 2.74061e-22, 6.38027e-23, 2.85363e-23, 7.1611e-23, 3.74733e-22, 1.15941e-22], [6616.66, 2.74684e-22, 6.3946e-23, 2.86004e-23, 7.17754e-23, 3.75591e-22, 1.16225e-22], [6631.71, 2.75309e-22, 6.40897e-23, 2.86646e-23, 7.19402e-23, 3.7645e-22, 1.1651e-22], [6646.8, 2.75935e-22, 6.42337e-23, 2.8729e-23, 7.21053e-23, 3.77311e-22, 1.16796e-22], [6661.92, 2.76562e-22, 6.4378e-23, 2.87935e-23, 7.22708e-23, 3.78174e-22, 1.17082e-22], [6677.07, 2.77191e-22, 6.45226e-23, 2.88582e-23, 7.24367e-23, 3.79038e-22, 1.17368e-22], [6692.26, 2.77821e-22, 6.46676e-23, 2.8923e-23, 7.2603e-23, 3.79905e-22, 1.17656e-22], [6707.48, 2.78453e-22, 6.4813e-23, 2.8988e-23, 7.27697e-23, 3.80774e-22, 1.17944e-22], [6722.74, 2.79086e-22, 6.49586e-23, 2.90531e-23, 7.29367e-23, 3.81645e-22, 1.18232e-22], [6738.03, 2.7972e-22, 6.51046e-23, 2.91184e-23, 7.31041e-23, 3.82518e-22, 1.18521e-22], [6753.35, 2.80356e-22, 6.52509e-23, 2.91838e-23, 7.32719e-23, 3.83392e-22, 1.18811e-22], [6768.71, 2.80993e-22, 6.53976e-23, 2.92494e-23, 7.344e-23, 3.84269e-22, 1.19102e-22], [6784.11, 2.81632e-22, 6.55446e-23, 2.93152e-23, 7.36086e-23, 3.85148e-22, 1.19393e-22], [6799.54, 2.82273e-22, 6.5692e-23, 2.9381e-23, 7.37775e-23, 3.86028e-22, 1.19684e-22], [6815.01, 2.82914e-22, 6.58396e-23, 2.94471e-23, 7.39468e-23, 3.86911e-22, 1.19976e-22], [6830.51, 2.83557e-22, 6.59877e-23, 2.95133e-23, 7.41165e-23, 3.87796e-22, 1.20269e-22], [6846.04, 2.84202e-22, 6.6136e-23, 2.95796e-23, 7.42865e-23, 3.88682e-22, 1.20563e-22], [6861.62, 2.84848e-22, 6.62847e-23, 2.96461e-23, 7.4457e-23, 3.89571e-22, 1.20857e-22], [6877.22, 2.85496e-22, 6.64338e-23, 2.97127e-23, 7.46278e-23, 3.90462e-22, 1.21152e-22], [6892.87, 2.86145e-22, 6.65832e-23, 2.97795e-23, 7.4799e-23, 3.91355e-22, 1.21447e-22], [6908.54, 2.86796e-22, 6.67329e-23, 2.98465e-23, 7.49706e-23, 3.92249e-22, 1.21743e-22], [6924.26, 2.87448e-22, 6.6883e-23, 2.99136e-23, 7.51426e-23, 3.93146e-22, 1.2204e-22], [6940.01, 2.88101e-22, 6.70334e-23, 2.99808e-23, 7.5315e-23, 3.94045e-22, 1.22337e-22], [6955.79, 2.88756e-22, 6.71842e-23, 3.00483e-23, 7.54877e-23, 3.94946e-22, 1.22635e-22], [6971.62, 2.89413e-22, 6.73353e-23, 3.01158e-23, 7.56609e-23, 3.95848e-22, 1.22933e-22], [6987.47, 2.90071e-22, 6.74868e-23, 3.01836e-23, 7.58344e-23, 3.96753e-22, 1.23233e-22], [7003.37, 2.9073e-22, 6.76386e-23, 3.02514e-23, 7.60084e-23, 3.9766e-22, 1.23532e-22], [7019.3, 2.91391e-22, 6.77907e-23, 3.03195e-23, 7.61827e-23, 3.98569e-22, 1.23833e-22], [7035.26, 2.92053e-22, 6.79432e-23, 3.03877e-23, 7.63574e-23, 3.9948e-22, 1.24134e-22], [7051.26, 2.92717e-22, 6.80961e-23, 3.0456e-23, 7.65325e-23, 4.00393e-22, 1.24436e-22], [7067.3, 2.93383e-22, 6.82493e-23, 3.05245e-23, 7.67081e-23, 4.01309e-22, 1.24738e-22], [7083.38, 2.9405e-22, 6.84029e-23, 3.05932e-23, 7.6884e-23, 4.02226e-22, 1.25041e-22], [7099.49, 2.94718e-22, 6.85568e-23, 3.0662e-23, 7.70603e-23, 4.03145e-22, 1.25345e-22], [7115.64, 2.95389e-22, 6.87111e-23, 3.0731e-23, 7.7237e-23, 4.04067e-22, 1.25649e-22], [7131.82, 2.9606e-22, 6.88657e-23, 3.08001e-23, 7.74141e-23, 4.0499e-22, 1.25954e-22], [7148.04, 2.96733e-22, 6.90207e-23, 3.08694e-23, 7.75916e-23, 4.05916e-22, 1.2626e-22], [7164.3, 2.97408e-22, 6.9176e-23, 3.09389e-23, 7.77695e-23, 4.06843e-22, 1.26566e-22], [7180.6, 2.98084e-22, 6.93317e-23, 3.10085e-23, 7.79478e-23, 4.07773e-22, 1.26873e-22], [7196.93, 2.98762e-22, 6.94878e-23, 3.10783e-23, 7.81265e-23, 4.08705e-22, 1.27181e-22], [7213.3, 2.99441e-22, 6.96442e-23, 3.11482e-23, 7.83056e-23, 4.09639e-22, 1.27489e-22], [7229.71, 3.00122e-22, 6.98009e-23, 3.12183e-23, 7.84851e-23, 4.10575e-22, 1.27798e-22], [7246.15, 3.00804e-22, 6.99581e-23, 3.12886e-23, 7.8665e-23, 4.11513e-22, 1.28107e-22], [7262.64, 3.01488e-22, 7.01156e-23, 3.1359e-23, 7.88453e-23, 4.12454e-22, 1.28418e-22], [7279.15, 3.02173e-22, 7.02734e-23, 3.14296e-23, 7.90261e-23, 4.13396e-22, 1.28729e-22], [7295.71, 3.0286e-22, 7.04316e-23, 3.15003e-23, 7.92072e-23, 4.14341e-22, 1.2904e-22], [7312.31, 3.03549e-22, 7.05902e-23, 3.15712e-23, 7.93888e-23, 4.15288e-22, 1.29353e-22], [7328.94, 3.04239e-22, 7.07492e-23, 3.16423e-23, 7.95707e-23, 4.16236e-22, 1.29665e-22], [7345.61, 3.04931e-22, 7.09085e-23, 3.17135e-23, 7.97531e-23, 4.17188e-22, 1.29979e-22], [7362.32, 3.05624e-22, 7.10682e-23, 3.17849e-23, 7.99359e-23, 4.18141e-22, 1.30293e-22], [7379.06, 3.06319e-22, 7.12282e-23, 3.18565e-23, 8.01191e-23, 4.19096e-22, 1.30608e-22], [7395.85, 3.07015e-22, 7.13886e-23, 3.19282e-23, 8.03027e-23, 4.20054e-22, 1.30924e-22], [7412.67, 3.07713e-22, 7.15494e-23, 3.20001e-23, 8.04867e-23, 4.21013e-22, 1.3124e-22], [7429.53, 3.08413e-22, 7.17105e-23, 3.20722e-23, 8.06711e-23, 4.21975e-22, 1.31557e-22], [7446.43, 3.09114e-22, 7.18721e-23, 3.21444e-23, 8.0856e-23, 4.22939e-22, 1.31875e-22], [7463.37, 3.09817e-22, 7.20339e-23, 3.22168e-23, 8.10413e-23, 4.23905e-22, 1.32193e-22], [7480.34, 3.10521e-22, 7.21962e-23, 3.22893e-23, 8.1227e-23, 4.24874e-22, 1.32512e-22], [7497.36, 3.11227e-22, 7.23588e-23, 3.23621e-23, 8.14131e-23, 4.25844e-22, 1.32832e-22], [7514.41, 3.11935e-22, 7.25218e-23, 3.2435e-23, 8.15996e-23, 4.26817e-22, 1.33152e-22], [7531.5, 3.12644e-22, 7.26852e-23, 3.2508e-23, 8.17865e-23, 4.27792e-22, 1.33473e-22], [7548.63, 3.13355e-22, 7.2849e-23, 3.25812e-23, 8.19739e-23, 4.28769e-22, 1.33795e-22], [7565.8, 3.14068e-22, 7.30131e-23, 3.26546e-23, 8.21617e-23, 4.29749e-22, 1.34118e-22], [7583.01, 3.14782e-22, 7.31776e-23, 3.27282e-23, 8.23499e-23, 4.30731e-22, 1.34441e-22], [7600.26, 3.15497e-22, 7.33425e-23, 3.28019e-23, 8.25386e-23, 4.31714e-22, 1.34765e-22], [7617.55, 3.16215e-22, 7.35078e-23, 3.28758e-23, 8.27276e-23, 4.32701e-22, 1.35089e-22], [7634.87, 3.16934e-22, 7.36734e-23, 3.29499e-23, 8.29171e-23, 4.33689e-22, 1.35415e-22], [7652.24, 3.17654e-22, 7.38395e-23, 3.30241e-23, 8.31071e-23, 4.34679e-22, 1.35741e-22], [7669.65, 3.18376e-22, 7.40059e-23, 3.30985e-23, 8.32974e-23, 4.35672e-22, 1.36067e-22], [7687.09, 3.191e-22, 7.41727e-23, 3.31731e-23, 8.34882e-23, 4.36667e-22, 1.36395e-22], [7704.58, 3.19826e-22, 7.43398e-23, 3.32479e-23, 8.36794e-23, 4.37665e-22, 1.36723e-22], [7722.1, 3.20553e-22, 7.45074e-23, 3.33228e-23, 8.38711e-23, 4.38664e-22, 1.37051e-22], [7739.66, 3.21282e-22, 7.46753e-23, 3.33979e-23, 8.40632e-23, 4.39666e-22, 1.37381e-22], [7757.27, 3.22012e-22, 7.48436e-23, 3.34732e-23, 8.42557e-23, 4.4067e-22, 1.37711e-22], [7774.91, 3.22744e-22, 7.50124e-23, 3.35486e-23, 8.44486e-23, 4.41677e-22, 1.38042e-22], [7792.6, 3.23478e-22, 7.51815e-23, 3.36242e-23, 8.4642e-23, 4.42685e-22, 1.38374e-22], [7810.32, 3.24214e-22, 7.53509e-23, 3.37e-23, 8.48358e-23, 4.43696e-22, 1.38706e-22], [7828.09, 3.24951e-22, 7.55208e-23, 3.3776e-23, 8.50301e-23, 4.44709e-22, 1.39039e-22], [7845.89, 3.2569e-22, 7.56911e-23, 3.38521e-23, 8.52248e-23, 4.45725e-22, 1.39373e-22], [7863.74, 3.2643e-22, 7.58617e-23, 3.39284e-23, 8.54199e-23, 4.46743e-22, 1.39708e-22], [7881.63, 3.27172e-22, 7.60328e-23, 3.40049e-23, 8.56155e-23, 4.47763e-22, 1.40043e-22], [7899.55, 3.27916e-22, 7.62042e-23, 3.40815e-23, 8.58115e-23, 4.48785e-22, 1.40379e-22], [7917.52, 3.28662e-22, 7.63761e-23, 3.41584e-23, 8.6008e-23, 4.4981e-22, 1.40715e-22], [7935.53, 3.29409e-22, 7.65483e-23, 3.42354e-23, 8.62049e-23, 4.50837e-22, 1.41053e-22], [7953.58, 3.30158e-22, 7.67209e-23, 3.43126e-23, 8.64023e-23, 4.51867e-22, 1.41391e-22], [7971.67, 3.30909e-22, 7.68939e-23, 3.439e-23, 8.66001e-23, 4.52898e-22, 1.4173e-22], [7989.8, 3.31661e-22, 7.70674e-23, 3.44675e-23, 8.67983e-23, 4.53933e-22, 1.42069e-22], [8007.98, 3.32415e-22, 7.72412e-23, 3.45452e-23, 8.6997e-23, 4.54969e-22, 1.4241e-22], [8026.19, 3.33171e-22, 7.74154e-23, 3.46231e-23, 8.71962e-23, 4.56008e-22, 1.42751e-22], [8044.45, 3.33928e-22, 7.759e-23, 3.47012e-23, 8.73958e-23, 4.57049e-22, 1.43093e-22], [8062.75, 3.34688e-22, 7.7765e-23, 3.47795e-23, 8.75958e-23, 4.58092e-22, 1.43435e-22], [8081.09, 3.35449e-22, 7.79404e-23, 3.48579e-23, 8.77963e-23, 4.59138e-22, 1.43779e-22], [8099.47, 3.36211e-22, 7.81162e-23, 3.49365e-23, 8.79973e-23, 4.60187e-22, 1.44123e-22], [8117.89, 3.36976e-22, 7.82925e-23, 3.50153e-23, 8.81987e-23, 4.61237e-22, 1.44467e-22], [8136.35, 3.37742e-22, 7.84691e-23, 3.50943e-23, 8.84005e-23, 4.6229e-22, 1.44813e-22], [8154.86, 3.3851e-22, 7.86461e-23, 3.51735e-23, 8.86029e-23, 4.63346e-22, 1.45159e-22], [8173.41, 3.39279e-22, 7.88235e-23, 3.52528e-23, 8.88056e-23, 4.64403e-22, 1.45506e-22], [8192.0, 3.40051e-22, 7.90014e-23, 3.53323e-23, 8.90089e-23, 4.65463e-22, 1.45854e-22]])     
    elif source=='T1800044-v4':
        data = np.array([[5.0, 1.9997e-20], [5.0115, 1.9462e-20], [5.0231, 1.8942e-20], [5.0347, 1.8437e-20], [5.0463, 1.7946e-20], [5.0579, 1.7469e-20], [5.0696, 1.7005e-20], [5.0813, 1.6555e-20], [5.093, 1.6117e-20], [5.1047, 1.5691e-20], [5.1165, 1.5277e-20], [5.1283, 1.4875e-20], [5.1401, 1.4484e-20], [5.152, 1.4104e-20], [5.1639, 1.3734e-20], [5.1758, 1.3375e-20], [5.1877, 1.3026e-20], [5.1997, 1.2686e-20], [5.2117, 1.2355e-20], [5.2237, 1.2034e-20], [5.2357, 1.1722e-20], [5.2478, 1.1418e-20], [5.2599, 1.1122e-20], [5.272, 1.0835e-20], [5.2842, 1.0555e-20], [5.2964, 1.0283e-20], [5.3086, 1.0018e-20], [5.3208, 9.7607e-21], [5.3331, 9.5103e-21], [5.3454, 9.2666e-21], [5.3577, 9.0295e-21], [5.3701, 8.7988e-21], [5.3825, 8.5744e-21], [5.3949, 8.3559e-21], [5.4073, 8.1434e-21], [5.4198, 7.9365e-21], [5.4323, 7.7352e-21], [5.4448, 7.5393e-21], [5.4574, 7.3487e-21], [5.4699, 7.1631e-21], [5.4826, 6.9824e-21], [5.4952, 6.8066e-21], [5.5079, 6.6354e-21], [5.5206, 6.4688e-21], [5.5333, 6.3066e-21], [5.5461, 6.1487e-21], [5.5589, 5.995e-21], [5.5717, 5.8453e-21], [5.5845, 5.6996e-21], [5.5974, 5.5577e-21], [5.6103, 5.4195e-21], [5.6232, 5.285e-21], [5.6362, 5.154e-21], [5.6492, 5.0264e-21], [5.6622, 4.9022e-21], [5.6753, 4.7812e-21], [5.6884, 4.6634e-21], [5.7015, 4.5486e-21], [5.7146, 4.4369e-21], [5.7278, 4.328e-21], [5.741, 4.222e-21], [5.7543, 4.1187e-21], [5.7675, 4.0181e-21], [5.7808, 3.9201e-21], [5.7942, 3.8246e-21], [5.8075, 3.7316e-21], [5.8209, 3.641e-21], [5.8343, 3.5528e-21], [5.8478, 3.4668e-21], [5.8613, 3.383e-21], [5.8748, 3.3014e-21], [5.8884, 3.2218e-21], [5.9019, 3.1443e-21], [5.9155, 3.0688e-21], [5.9292, 2.9953e-21], [5.9429, 2.9236e-21], [5.9566, 2.8537e-21], [5.9703, 2.7856e-21], [5.9841, 2.7193e-21], [5.9979, 2.6546e-21], [6.0117, 2.5916e-21], [6.0256, 2.5302e-21], [6.0394, 2.4703e-21], [6.0534, 2.412e-21], [6.0673, 2.3551e-21], [6.0813, 2.2997e-21], [6.0954, 2.2457e-21], [6.1094, 2.1931e-21], [6.1235, 2.1418e-21], [6.1376, 2.0917e-21], [6.1518, 2.043e-21], [6.166, 1.9955e-21], [6.1802, 1.9491e-21], [6.1944, 1.904e-21], [6.2087, 1.86e-21], [6.223, 1.817e-21], [6.2374, 1.7752e-21], [6.2518, 1.7344e-21], [6.2662, 1.6947e-21], [6.2806, 1.6559e-21], [6.2951, 1.6181e-21], [6.3096, 1.5812e-21], [6.3242, 1.5453e-21], [6.3388, 1.5103e-21], [6.3534, 1.4762e-21], [6.368, 1.4429e-21], [6.3827, 1.4104e-21], [6.3974, 1.3788e-21], [6.4122, 1.3479e-21], [6.427, 1.3178e-21], [6.4418, 1.2885e-21], [6.4566, 1.2599e-21], [6.4715, 1.232e-21], [6.4865, 1.2048e-21], [6.5014, 1.1783e-21], [6.5164, 1.1524e-21], [6.5314, 1.1272e-21], [6.5465, 1.1027e-21], [6.5616, 1.0787e-21], [6.5767, 1.0553e-21], [6.5919, 1.0326e-21], [6.6071, 1.0103e-21], [6.6223, 9.8868e-22], [6.6376, 9.6756e-22], [6.6529, 9.4697e-22], [6.6682, 9.2689e-22], [6.6836, 9.0732e-22], [6.699, 8.8823e-22], [6.7145, 8.6962e-22], [6.73, 8.5148e-22], [6.7455, 8.3378e-22], [6.761, 8.1654e-22], [6.7766, 7.9972e-22], [6.7923, 7.8332e-22], [6.8079, 7.6734e-22], [6.8236, 7.5175e-22], [6.8394, 7.3656e-22], [6.8551, 7.2174e-22], [6.8709, 7.073e-22], [6.8868, 6.9323e-22], [6.9027, 6.795e-22], [6.9186, 6.6612e-22], [6.9345, 6.5308e-22], [6.9505, 6.4037e-22], [6.9666, 6.2798e-22], [6.9826, 6.159e-22], [6.9987, 6.0412e-22], [7.0149, 5.9265e-22], [7.031, 5.8146e-22], [7.0473, 5.7056e-22], [7.0635, 5.5994e-22], [7.0798, 5.4958e-22], [7.0961, 5.3949e-22], [7.1125, 5.2966e-22], [7.1289, 5.2007e-22], [7.1453, 5.1073e-22], [7.1618, 5.0163e-22], [7.1783, 4.9277e-22], [7.1949, 4.8413e-22], [7.2115, 4.7571e-22], [7.2281, 4.6751e-22], [7.2448, 4.5952e-22], [7.2615, 4.5173e-22], [7.2782, 4.4415e-22], [7.295, 4.3676e-22], [7.3118, 4.2957e-22], [7.3287, 4.2256e-22], [7.3456, 4.1573e-22], [7.3625, 4.0908e-22], [7.3795, 4.0261e-22], [7.3965, 3.963e-22], [7.4136, 3.9016e-22], [7.4307, 3.8418e-22], [7.4478, 3.7836e-22], [7.465, 3.7269e-22], [7.4822, 3.6717e-22], [7.4994, 3.618e-22], [7.5167, 3.5656e-22], [7.5341, 3.5147e-22], [7.5514, 3.4651e-22], [7.5689, 3.4169e-22], [7.5863, 3.3699e-22], [7.6038, 3.3242e-22], [7.6213, 3.2797e-22], [7.6389, 3.2365e-22], [7.6565, 3.1943e-22], [7.6742, 3.1534e-22], [7.6919, 3.1135e-22], [7.7096, 3.0747e-22], [7.7274, 3.037e-22], [7.7452, 3.0004e-22], [7.7631, 2.9647e-22], [7.781, 2.93e-22], [7.7989, 2.8963e-22], [7.8169, 2.8635e-22], [7.8349, 2.8316e-22], [7.853, 2.8007e-22], [7.8711, 2.7706e-22], [7.8893, 2.7413e-22], [7.9074, 2.7129e-22], [7.9257, 2.6853e-22], [7.944, 2.6586e-22], [7.9623, 2.6326e-22], [7.9806, 2.6073e-22], [7.999, 2.5828e-22], [8.0175, 2.5591e-22], [8.036, 2.5361e-22], [8.0545, 2.5137e-22], [8.0731, 2.4921e-22], [8.0917, 2.4712e-22], [8.1104, 2.4509e-22], [8.1291, 2.4312e-22], [8.1478, 2.4123e-22], [8.1666, 2.3939e-22], [8.1854, 2.3762e-22], [8.2043, 2.3591e-22], [8.2232, 2.3426e-22], [8.2422, 2.3268e-22], [8.2612, 2.3115e-22], [8.2802, 2.2968e-22], [8.2993, 2.2827e-22], [8.3185, 2.2692e-22], [8.3377, 2.2563e-22], [8.3569, 2.244e-22], [8.3762, 2.2322e-22], [8.3955, 2.221e-22], [8.4148, 2.2104e-22], [8.4342, 2.2004e-22], [8.4537, 2.191e-22], [8.4732, 2.1822e-22], [8.4927, 2.174e-22], [8.5123, 2.1664e-22], [8.5319, 2.1594e-22], [8.5516, 2.1531e-22], [8.5713, 2.1474e-22], [8.5911, 2.1423e-22], [8.6109, 2.138e-22], [8.6308, 2.1343e-22], [8.6507, 2.1314e-22], [8.6706, 2.1292e-22], [8.6906, 2.1278e-22], [8.7106, 2.1272e-22], [8.7307, 2.1274e-22], [8.7509, 2.1285e-22], [8.771, 2.1305e-22], [8.7913, 2.1335e-22], [8.8115, 2.1375e-22], [8.8319, 2.1426e-22], [8.8522, 2.1489e-22], [8.8726, 2.1564e-22], [8.8931, 2.1651e-22], [8.9136, 2.1753e-22], [8.9342, 2.187e-22], [8.9548, 2.2003e-22], [8.9754, 2.2154e-22], [8.9961, 2.2323e-22], [9.0169, 2.2513e-22], [9.0377, 2.2726e-22], [9.0585, 2.2963e-22], [9.0794, 2.3227e-22], [9.1003, 2.3521e-22], [9.1213, 2.3848e-22], [9.1423, 2.4211e-22], [9.1634, 2.4615e-22], [9.1846, 2.5065e-22], [9.2057, 2.5566e-22], [9.227, 2.6125e-22], [9.2482, 2.675e-22], [9.2696, 2.7451e-22], [9.2909, 2.8238e-22], [9.3124, 2.9126e-22], [9.3338, 3.0131e-22], [9.3554, 3.1274e-22], [9.3769, 3.2581e-22], [9.3986, 3.4084e-22], [9.4202, 3.5826e-22], [9.442, 3.786e-22], [9.4637, 4.026e-22], [9.4856, 4.3124e-22], [9.5074, 4.659e-22], [9.5294, 5.0857e-22], [9.5513, 5.6218e-22], [9.5734, 6.3135e-22], [9.5954, 7.237e-22], [9.6176, 8.5276e-22], [9.6397, 1.0452e-21], [9.662, 1.3616e-21], [9.6842, 1.9762e-21], [9.7066, 3.6774e-21], [9.729, 3.0668e-20], [9.7514, 4.7297e-21], [9.7739, 2.1756e-21], [9.7964, 1.4056e-21], [9.819, 1.0349e-21], [9.8417, 8.1736e-22], [9.8643, 6.7459e-22], [9.8871, 5.7392e-22], [9.9099, 4.9927e-22], [9.9327, 4.4183e-22], [9.9557, 3.9636e-22], [9.9786, 3.5955e-22], [10.002, 3.2919e-22], [10.025, 3.0377e-22], [10.048, 2.8223e-22], [10.071, 2.6375e-22], [10.094, 2.4777e-22], [10.117, 2.3383e-22], [10.141, 2.2158e-22], [10.164, 2.1074e-22], [10.188, 2.011e-22], [10.211, 1.9248e-22], [10.235, 1.8473e-22], [10.258, 1.7773e-22], [10.282, 1.7139e-22], [10.306, 1.6562e-22], [10.329, 1.6035e-22], [10.353, 1.5552e-22], [10.377, 1.5108e-22], [10.401, 1.4699e-22], [10.425, 1.432e-22], [10.449, 1.3969e-22], [10.473, 1.3643e-22], [10.497, 1.3339e-22], [10.522, 1.3056e-22], [10.546, 1.279e-22], [10.57, 1.2541e-22], [10.594, 1.2306e-22], [10.619, 1.2085e-22], [10.643, 1.1877e-22], [10.668, 1.1679e-22], [10.693, 1.1492e-22], [10.717, 1.1315e-22], [10.742, 1.1146e-22], [10.767, 1.0986e-22], [10.791, 1.0832e-22], [10.816, 1.0686e-22], [10.841, 1.0545e-22], [10.866, 1.0411e-22], [10.891, 1.0282e-22], [10.916, 1.0158e-22], [10.942, 1.0039e-22], [10.967, 9.9242e-23], [10.992, 9.8134e-23], [11.018, 9.7064e-23], [11.043, 9.603e-23], [11.068, 9.5029e-23], [11.094, 9.406e-23], [11.12, 9.312e-23], [11.145, 9.2207e-23], [11.171, 9.132e-23], [11.197, 9.0458e-23], [11.222, 8.9619e-23], [11.248, 8.8802e-23], [11.274, 8.8005e-23], [11.3, 8.7228e-23], [11.326, 8.6469e-23], [11.352, 8.5728e-23], [11.379, 8.5004e-23], [11.405, 8.4296e-23], [11.431, 8.3602e-23], [11.458, 8.2923e-23], [11.484, 8.2258e-23], [11.51, 8.1606e-23], [11.537, 8.0966e-23], [11.564, 8.0339e-23], [11.59, 7.9722e-23], [11.617, 7.9117e-23], [11.644, 7.8523e-23], [11.671, 7.7938e-23], [11.697, 7.7364e-23], [11.724, 7.6798e-23], [11.752, 7.6242e-23], [11.779, 7.5694e-23], [11.806, 7.5155e-23], [11.833, 7.4624e-23], [11.86, 7.4101e-23], [11.888, 7.3585e-23], [11.915, 7.3076e-23], [11.943, 7.2575e-23], [11.97, 7.208e-23], [11.998, 7.1592e-23], [12.025, 7.1111e-23], [12.053, 7.0636e-23], [12.081, 7.0166e-23], [12.109, 6.9703e-23], [12.137, 6.9246e-23], [12.165, 6.8794e-23], [12.193, 6.8347e-23], [12.221, 6.7906e-23], [12.249, 6.747e-23], [12.277, 6.7039e-23], [12.306, 6.6612e-23], [12.334, 6.6191e-23], [12.362, 6.5774e-23], [12.391, 6.5362e-23], [12.419, 6.4955e-23], [12.448, 6.4551e-23], [12.477, 6.4152e-23], [12.506, 6.3757e-23], [12.534, 6.3367e-23], [12.563, 6.298e-23], [12.592, 6.2597e-23], [12.621, 6.2218e-23], [12.65, 6.1843e-23], [12.68, 6.1471e-23], [12.709, 6.1103e-23], [12.738, 6.0739e-23], [12.767, 6.0378e-23], [12.797, 6.002e-23], [12.826, 5.9666e-23], [12.856, 5.9315e-23], [12.886, 5.8968e-23], [12.915, 5.8623e-23], [12.945, 5.8282e-23], [12.975, 5.7944e-23], [13.005, 5.7609e-23], [13.035, 5.7276e-23], [13.065, 5.6947e-23], [13.095, 5.6621e-23], [13.125, 5.6297e-23], [13.156, 5.5976e-23], [13.186, 5.5658e-23], [13.216, 5.5343e-23], [13.247, 5.503e-23], [13.277, 5.472e-23], [13.308, 5.4413e-23], [13.339, 5.4108e-23], [13.369, 5.3806e-23], [13.4, 5.3506e-23], [13.431, 5.3208e-23], [13.462, 5.2913e-23], [13.493, 5.2621e-23], [13.524, 5.233e-23], [13.555, 5.2042e-23], [13.587, 5.1756e-23], [13.618, 5.1473e-23], [13.649, 5.1192e-23], [13.681, 5.0912e-23], [13.712, 5.0636e-23], [13.744, 5.0361e-23], [13.776, 5.0088e-23], [13.808, 4.9817e-23], [13.839, 4.9549e-23], [13.871, 4.9282e-23], [13.903, 4.9018e-23], [13.935, 4.8755e-23], [13.968, 4.8494e-23], [14.0, 4.8235e-23], [14.032, 4.7979e-23], [14.064, 4.7724e-23], [14.097, 4.7471e-23], [14.129, 4.7219e-23], [14.162, 4.697e-23], [14.195, 4.6722e-23], [14.227, 4.6476e-23], [14.26, 4.6232e-23], [14.293, 4.599e-23], [14.326, 4.5749e-23], [14.359, 4.551e-23], [14.392, 4.5273e-23], [14.425, 4.5037e-23], [14.459, 4.4803e-23], [14.492, 4.4571e-23], [14.525, 4.434e-23], [14.559, 4.411e-23], [14.592, 4.3883e-23], [14.626, 4.3657e-23], [14.66, 4.3432e-23], [14.694, 4.3209e-23], [14.727, 4.2987e-23], [14.761, 4.2767e-23], [14.795, 4.2549e-23], [14.83, 4.2332e-23], [14.864, 4.2116e-23], [14.898, 4.1902e-23], [14.932, 4.1689e-23], [14.967, 4.1477e-23], [15.001, 4.1267e-23], [15.036, 4.1058e-23], [15.071, 4.0851e-23], [15.105, 4.0645e-23], [15.14, 4.044e-23], [15.175, 4.0237e-23], [15.21, 4.0035e-23], [15.245, 3.9834e-23], [15.28, 3.9635e-23], [15.316, 3.9436e-23], [15.351, 3.9239e-23], [15.386, 3.9044e-23], [15.422, 3.8849e-23], [15.457, 3.8656e-23], [15.493, 3.8464e-23], [15.529, 3.8273e-23], [15.564, 3.8083e-23], [15.6, 3.7895e-23], [15.636, 3.7708e-23], [15.672, 3.7522e-23], [15.709, 3.7337e-23], [15.745, 3.7153e-23], [15.781, 3.697e-23], [15.817, 3.6788e-23], [15.854, 3.6608e-23], [15.89, 3.6429e-23], [15.927, 3.625e-23], [15.964, 3.6073e-23], [16.001, 3.5897e-23], [16.038, 3.5722e-23], [16.075, 3.5548e-23], [16.112, 3.5375e-23], [16.149, 3.5203e-23], [16.186, 3.5032e-23], [16.223, 3.4862e-23], [16.261, 3.4693e-23], [16.298, 3.4525e-23], [16.336, 3.4358e-23], [16.374, 3.4192e-23], [16.411, 3.4028e-23], [16.449, 3.3864e-23], [16.487, 3.3701e-23], [16.525, 3.3539e-23], [16.563, 3.3378e-23], [16.601, 3.3217e-23], [16.64, 3.3058e-23], [16.678, 3.29e-23], [16.716, 3.2743e-23], [16.755, 3.2586e-23], [16.794, 3.2431e-23], [16.832, 3.2276e-23], [16.871, 3.2122e-23], [16.91, 3.1969e-23], [16.949, 3.1818e-23], [16.988, 3.1666e-23], [17.027, 3.1516e-23], [17.067, 3.1367e-23], [17.106, 3.1218e-23], [17.145, 3.1071e-23], [17.185, 3.0924e-23], [17.225, 3.0778e-23], [17.264, 3.0633e-23], [17.304, 3.0488e-23], [17.344, 3.0345e-23], [17.384, 3.0202e-23], [17.424, 3.006e-23], [17.464, 2.9919e-23], [17.505, 2.9779e-23], [17.545, 2.964e-23], [17.585, 2.9501e-23], [17.626, 2.9363e-23], [17.667, 2.9226e-23], [17.707, 2.9089e-23], [17.748, 2.8954e-23], [17.789, 2.8819e-23], [17.83, 2.8685e-23], [17.871, 2.8551e-23], [17.912, 2.8419e-23], [17.954, 2.8287e-23], [17.995, 2.8156e-23], [18.037, 2.8025e-23], [18.078, 2.7896e-23], [18.12, 2.7767e-23], [18.162, 2.7638e-23], [18.204, 2.7511e-23], [18.246, 2.7384e-23], [18.288, 2.7258e-23], [18.33, 2.7132e-23], [18.372, 2.7007e-23], [18.414, 2.6883e-23], [18.457, 2.676e-23], [18.499, 2.6637e-23], [18.542, 2.6515e-23], [18.585, 2.6394e-23], [18.628, 2.6273e-23], [18.671, 2.6153e-23], [18.714, 2.6034e-23], [18.757, 2.5915e-23], [18.8, 2.5797e-23], [18.844, 2.5679e-23], [18.887, 2.5562e-23], [18.931, 2.5446e-23], [18.974, 2.5331e-23], [19.018, 2.5216e-23], [19.062, 2.5101e-23], [19.106, 2.4988e-23], [19.15, 2.4875e-23], [19.194, 2.4762e-23], [19.238, 2.465e-23], [19.283, 2.4539e-23], [19.327, 2.4428e-23], [19.372, 2.4318e-23], [19.416, 2.4209e-23], [19.461, 2.41e-23], [19.506, 2.3992e-23], [19.551, 2.3884e-23], [19.596, 2.3777e-23], [19.641, 2.367e-23], [19.686, 2.3564e-23], [19.732, 2.3459e-23], [19.777, 2.3354e-23], [19.823, 2.325e-23], [19.869, 2.3146e-23], [19.915, 2.3043e-23], [19.96, 2.294e-23], [20.006, 2.2838e-23], [20.053, 2.2737e-23], [20.099, 2.2636e-23], [20.145, 2.2535e-23], [20.192, 2.2435e-23], [20.238, 2.2336e-23], [20.285, 2.2237e-23], [20.332, 2.2139e-23], [20.379, 2.2041e-23], [20.426, 2.1943e-23], [20.473, 2.1847e-23], [20.52, 2.175e-23], [20.567, 2.1655e-23], [20.615, 2.1559e-23], [20.662, 2.1465e-23], [20.71, 2.1371e-23], [20.758, 2.1277e-23], [20.805, 2.1184e-23], [20.853, 2.1091e-23], [20.901, 2.0999e-23], [20.95, 2.0907e-23], [20.998, 2.0816e-23], [21.046, 2.0725e-23], [21.095, 2.0634e-23], [21.144, 2.0545e-23], [21.192, 2.0455e-23], [21.241, 2.0366e-23], [21.29, 2.0278e-23], [21.339, 2.019e-23], [21.389, 2.0103e-23], [21.438, 2.0015e-23], [21.487, 1.9929e-23], [21.537, 1.9843e-23], [21.586, 1.9757e-23], [21.636, 1.9672e-23], [21.686, 1.9587e-23], [21.736, 1.9503e-23], [21.786, 1.9419e-23], [21.837, 1.9336e-23], [21.887, 1.9253e-23], [21.937, 1.917e-23], [21.988, 1.9088e-23], [22.039, 1.9006e-23], [22.089, 1.8925e-23], [22.14, 1.8844e-23], [22.191, 1.8764e-23], [22.243, 1.8684e-23], [22.294, 1.8604e-23], [22.345, 1.8525e-23], [22.397, 1.8446e-23], [22.449, 1.8368e-23], [22.5, 1.829e-23], [22.552, 1.8213e-23], [22.604, 1.8136e-23], [22.656, 1.8059e-23], [22.709, 1.7983e-23], [22.761, 1.7907e-23], [22.813, 1.7831e-23], [22.866, 1.7756e-23], [22.919, 1.7681e-23], [22.972, 1.7607e-23], [23.025, 1.7533e-23], [23.078, 1.746e-23], [23.131, 1.7386e-23], [23.184, 1.7314e-23], [23.238, 1.7241e-23], [23.291, 1.7169e-23], [23.345, 1.7098e-23], [23.399, 1.7026e-23], [23.453, 1.6956e-23], [23.507, 1.6885e-23], [23.561, 1.6815e-23], [23.615, 1.6745e-23], [23.67, 1.6676e-23], [23.724, 1.6607e-23], [23.779, 1.6538e-23], [23.834, 1.647e-23], [23.889, 1.6402e-23], [23.944, 1.6334e-23], [23.999, 1.6267e-23], [24.055, 1.62e-23], [24.11, 1.6133e-23], [24.166, 1.6067e-23], [24.221, 1.6001e-23], [24.277, 1.5936e-23], [24.333, 1.5871e-23], [24.389, 1.5806e-23], [24.446, 1.5741e-23], [24.502, 1.5677e-23], [24.558, 1.5613e-23], [24.615, 1.555e-23], [24.672, 1.5487e-23], [24.729, 1.5424e-23], [24.786, 1.5362e-23], [24.843, 1.5299e-23], [24.9, 1.5238e-23], [24.958, 1.5176e-23], [25.015, 1.5115e-23], [25.073, 1.5054e-23], [25.131, 1.4993e-23], [25.189, 1.4933e-23], [25.247, 1.4873e-23], [25.305, 1.4814e-23], [25.363, 1.4754e-23], [25.422, 1.4695e-23], [25.48, 1.4637e-23], [25.539, 1.4578e-23], [25.598, 1.452e-23], [25.657, 1.4463e-23], [25.716, 1.4405e-23], [25.776, 1.4348e-23], [25.835, 1.4291e-23], [25.895, 1.4235e-23], [25.954, 1.4178e-23], [26.014, 1.4122e-23], [26.074, 1.4067e-23], [26.134, 1.4011e-23], [26.194, 1.3956e-23], [26.255, 1.3901e-23], [26.315, 1.3847e-23], [26.376, 1.3793e-23], [26.437, 1.3739e-23], [26.498, 1.3685e-23], [26.559, 1.3632e-23], [26.62, 1.3579e-23], [26.682, 1.3526e-23], [26.743, 1.3473e-23], [26.805, 1.3421e-23], [26.867, 1.3369e-23], [26.929, 1.3317e-23], [26.991, 1.3266e-23], [27.053, 1.3215e-23], [27.115, 1.3164e-23], [27.178, 1.3113e-23], [27.241, 1.3063e-23], [27.303, 1.3013e-23], [27.366, 1.2963e-23], [27.429, 1.2913e-23], [27.493, 1.2864e-23], [27.556, 1.2815e-23], [27.62, 1.2766e-23], [27.683, 1.2718e-23], [27.747, 1.2669e-23], [27.811, 1.2621e-23], [27.875, 1.2574e-23], [27.94, 1.2526e-23], [28.004, 1.2479e-23], [28.069, 1.2432e-23], [28.133, 1.2385e-23], [28.198, 1.2338e-23], [28.263, 1.2292e-23], [28.328, 1.2246e-23], [28.394, 1.22e-23], [28.459, 1.2155e-23], [28.525, 1.2109e-23], [28.591, 1.2064e-23], [28.656, 1.202e-23], [28.723, 1.1975e-23], [28.789, 1.1931e-23], [28.855, 1.1887e-23], [28.922, 1.1843e-23], [28.988, 1.1799e-23], [29.055, 1.1756e-23], [29.122, 1.1712e-23], [29.189, 1.1669e-23], [29.257, 1.1627e-23], [29.324, 1.1584e-23], [29.392, 1.1542e-23], [29.46, 1.15e-23], [29.528, 1.1458e-23], [29.596, 1.1416e-23], [29.664, 1.1375e-23], [29.732, 1.1334e-23], [29.801, 1.1293e-23], [29.87, 1.1252e-23], [29.938, 1.1212e-23], [30.007, 1.1171e-23], [30.077, 1.1131e-23], [30.146, 1.1091e-23], [30.216, 1.1052e-23], [30.285, 1.1012e-23], [30.355, 1.0973e-23], [30.425, 1.0934e-23], [30.495, 1.0895e-23], [30.566, 1.0857e-23], [30.636, 1.0818e-23], [30.707, 1.078e-23], [30.777, 1.0742e-23], [30.848, 1.0704e-23], [30.92, 1.0667e-23], [30.991, 1.0629e-23], [31.062, 1.0592e-23], [31.134, 1.0555e-23], [31.206, 1.0518e-23], [31.278, 1.0482e-23], [31.35, 1.0445e-23], [31.422, 1.0409e-23], [31.495, 1.0373e-23], [31.567, 1.0337e-23], [31.64, 1.0302e-23], [31.713, 1.0266e-23], [31.786, 1.0231e-23], [31.859, 1.0196e-23], [31.933, 1.0161e-23], [32.007, 1.0126e-23], [32.08, 1.0092e-23], [32.154, 1.0057e-23], [32.228, 1.0023e-23], [32.303, 9.9894e-24], [32.377, 9.9556e-24], [32.452, 9.9221e-24], [32.527, 9.8887e-24], [32.602, 9.8554e-24], [32.677, 9.8224e-24], [32.752, 9.7896e-24], [32.828, 9.7569e-24], [32.904, 9.7244e-24], [32.979, 9.692e-24], [33.055, 9.6599e-24], [33.132, 9.6279e-24], [33.208, 9.5961e-24], [33.285, 9.5644e-24], [33.361, 9.533e-24], [33.438, 9.5017e-24], [33.515, 9.4705e-24], [33.593, 9.4396e-24], [33.67, 9.4088e-24], [33.748, 9.3782e-24], [33.826, 9.3477e-24], [33.904, 9.3174e-24], [33.982, 9.2873e-24], [34.06, 9.2573e-24], [34.139, 9.2275e-24], [34.218, 9.1979e-24], [34.296, 9.1684e-24], [34.376, 9.139e-24], [34.455, 9.1099e-24], [34.534, 9.0809e-24], [34.614, 9.052e-24], [34.694, 9.0233e-24], [34.774, 8.9948e-24], [34.854, 8.9664e-24], [34.934, 8.9382e-24], [35.015, 8.9101e-24], [35.096, 8.8822e-24], [35.176, 8.8544e-24], [35.258, 8.8268e-24], [35.339, 8.7993e-24], [35.42, 8.772e-24], [35.502, 8.7448e-24], [35.584, 8.7178e-24], [35.666, 8.6909e-24], [35.748, 8.6642e-24], [35.831, 8.6376e-24], [35.913, 8.6112e-24], [35.996, 8.5849e-24], [36.079, 8.5587e-24], [36.162, 8.5327e-24], [36.246, 8.5069e-24], [36.329, 8.4811e-24], [36.413, 8.4556e-24], [36.497, 8.4301e-24], [36.581, 8.4048e-24], [36.666, 8.3796e-24], [36.75, 8.3546e-24], [36.835, 8.3297e-24], [36.92, 8.3049e-24], [37.005, 8.2803e-24], [37.09, 8.2558e-24], [37.176, 8.2315e-24], [37.262, 8.2073e-24], [37.347, 8.1832e-24], [37.434, 8.1592e-24], [37.52, 8.1354e-24], [37.606, 8.1117e-24], [37.693, 8.0881e-24], [37.78, 8.0647e-24], [37.867, 8.0414e-24], [37.955, 8.0182e-24], [38.042, 7.9952e-24], [38.13, 7.9722e-24], [38.218, 7.9494e-24], [38.306, 7.9268e-24], [38.394, 7.9042e-24], [38.483, 7.8818e-24], [38.571, 7.8595e-24], [38.66, 7.8373e-24], [38.75, 7.8152e-24], [38.839, 7.7933e-24], [38.928, 7.7715e-24], [39.018, 7.7498e-24], [39.108, 7.7282e-24], [39.198, 7.7067e-24], [39.289, 7.6854e-24], [39.379, 7.6641e-24], [39.47, 7.643e-24], [39.561, 7.622e-24], [39.652, 7.6012e-24], [39.744, 7.5804e-24], [39.836, 7.5597e-24], [39.927, 7.5392e-24], [40.019, 7.5188e-24], [40.112, 7.4985e-24], [40.204, 7.4783e-24], [40.297, 7.4582e-24], [40.39, 7.4382e-24], [40.483, 7.4183e-24], [40.576, 7.3986e-24], [40.67, 7.3789e-24], [40.764, 7.3594e-24], [40.858, 7.3399e-24], [40.952, 7.3206e-24], [41.046, 7.3014e-24], [41.141, 7.2822e-24], [41.236, 7.2632e-24], [41.331, 7.2443e-24], [41.426, 7.2255e-24], [41.522, 7.2068e-24], [41.618, 7.1882e-24], [41.714, 7.1697e-24], [41.81, 7.1513e-24], [41.906, 7.133e-24], [42.003, 7.1148e-24], [42.1, 7.0967e-24], [42.197, 7.0787e-24], [42.294, 7.0608e-24], [42.392, 7.043e-24], [42.489, 7.0253e-24], [42.587, 7.0077e-24], [42.686, 6.9902e-24], [42.784, 6.9728e-24], [42.883, 6.9555e-24], [42.981, 6.9383e-24], [43.081, 6.9212e-24], [43.18, 6.9041e-24], [43.28, 6.8872e-24], [43.379, 6.8703e-24], [43.479, 6.8536e-24], [43.58, 6.8369e-24], [43.68, 6.8203e-24], [43.781, 6.8039e-24], [43.882, 6.7875e-24], [43.983, 6.7712e-24], [44.084, 6.755e-24], [44.186, 6.7388e-24], [44.288, 6.7228e-24], [44.39, 6.7069e-24], [44.492, 6.691e-24], [44.595, 6.6752e-24], [44.698, 6.6595e-24], [44.801, 6.6439e-24], [44.904, 6.6284e-24], [45.008, 6.613e-24], [45.112, 6.5977e-24], [45.216, 6.5824e-24], [45.32, 6.5672e-24], [45.424, 6.5521e-24], [45.529, 6.5371e-24], [45.634, 6.5222e-24], [45.739, 6.5073e-24], [45.845, 6.4925e-24], [45.951, 6.4779e-24], [46.057, 6.4632e-24], [46.163, 6.4487e-24], [46.269, 6.4343e-24], [46.376, 6.4199e-24], [46.483, 6.4056e-24], [46.59, 6.3914e-24], [46.697, 6.3772e-24], [46.805, 6.3632e-24], [46.913, 6.3492e-24], [47.021, 6.3353e-24], [47.13, 6.3214e-24], [47.238, 6.3077e-24], [47.347, 6.294e-24], [47.457, 6.2804e-24], [47.566, 6.2668e-24], [47.676, 6.2534e-24], [47.786, 6.24e-24], [47.896, 6.2267e-24], [48.006, 6.2134e-24], [48.117, 6.2002e-24], [48.228, 6.1871e-24], [48.339, 6.1741e-24], [48.451, 6.1611e-24], [48.562, 6.1482e-24], [48.674, 6.1354e-24], [48.787, 6.1227e-24], [48.899, 6.11e-24], [49.012, 6.0974e-24], [49.125, 6.0848e-24], [49.238, 6.0723e-24], [49.352, 6.0599e-24], [49.465, 6.0476e-24], [49.579, 6.0353e-24], [49.694, 6.023e-24], [49.808, 6.0109e-24], [49.923, 5.9988e-24], [50.038, 5.9868e-24], [50.154, 5.9748e-24], [50.269, 5.9629e-24], [50.385, 5.9511e-24], [50.502, 5.9393e-24], [50.618, 5.9276e-24], [50.735, 5.916e-24], [50.852, 5.9044e-24], [50.969, 5.8929e-24], [51.087, 5.8814e-24], [51.204, 5.87e-24], [51.322, 5.8587e-24], [51.441, 5.8474e-24], [51.559, 5.8362e-24], [51.678, 5.825e-24], [51.797, 5.8139e-24], [51.917, 5.8029e-24], [52.037, 5.7919e-24], [52.157, 5.7809e-24], [52.277, 5.7701e-24], [52.397, 5.7593e-24], [52.518, 5.7485e-24], [52.639, 5.7378e-24], [52.761, 5.7272e-24], [52.882, 5.7166e-24], [53.004, 5.706e-24], [53.127, 5.6956e-24], [53.249, 5.6851e-24], [53.372, 5.6748e-24], [53.495, 5.6645e-24], [53.618, 5.6542e-24], [53.742, 5.644e-24], [53.866, 5.6338e-24], [53.99, 5.6237e-24], [54.115, 5.6137e-24], [54.239, 5.6037e-24], [54.365, 5.5938e-24], [54.49, 5.5839e-24], [54.616, 5.574e-24], [54.741, 5.5642e-24], [54.868, 5.5545e-24], [54.994, 5.5448e-24], [55.121, 5.5352e-24], [55.248, 5.5256e-24], [55.376, 5.516e-24], [55.503, 5.5065e-24], [55.631, 5.4971e-24], [55.76, 5.4877e-24], [55.888, 5.4783e-24], [56.017, 5.469e-24], [56.146, 5.4598e-24], [56.276, 5.4506e-24], [56.405, 5.4414e-24], [56.535, 5.4323e-24], [56.666, 5.4232e-24], [56.797, 5.4142e-24], [56.928, 5.4052e-24], [57.059, 5.3963e-24], [57.19, 5.3874e-24], [57.322, 5.3786e-24], [57.454, 5.3698e-24], [57.587, 5.361e-24], [57.72, 5.3523e-24], [57.853, 5.3437e-24], [57.986, 5.3351e-24], [58.12, 5.3265e-24], [58.254, 5.3179e-24], [58.388, 5.3095e-24], [58.523, 5.301e-24], [58.658, 5.2926e-24], [58.793, 5.2842e-24], [58.929, 5.2759e-24], [59.065, 5.2676e-24], [59.201, 5.2594e-24], [59.337, 5.2512e-24], [59.474, 5.243e-24], [59.611, 5.2349e-24], [59.749, 5.2268e-24], [59.887, 5.2188e-24], [60.025, 5.2108e-24], [60.163, 5.2028e-24], [60.302, 5.1949e-24], [60.441, 5.187e-24], [60.58, 5.1792e-24], [60.72, 5.1714e-24], [60.86, 5.1636e-24], [61.0, 5.1559e-24], [61.141, 5.1482e-24], [61.282, 5.1405e-24], [61.423, 5.1329e-24], [61.565, 5.1253e-24], [61.707, 5.1178e-24], [61.849, 5.1103e-24], [61.992, 5.1028e-24], [62.135, 5.0953e-24], [62.278, 5.0879e-24], [62.422, 5.0806e-24], [62.566, 5.0732e-24], [62.71, 5.0659e-24], [62.855, 5.0587e-24], [62.999, 5.0515e-24], [63.145, 5.0443e-24], [63.29, 5.0371e-24], [63.436, 5.03e-24], [63.583, 5.0229e-24], [63.729, 5.0158e-24], [63.876, 5.0088e-24], [64.023, 5.0018e-24], [64.171, 4.9949e-24], [64.319, 4.9879e-24], [64.467, 4.981e-24], [64.616, 4.9742e-24], [64.765, 4.9673e-24], [64.914, 4.9605e-24], [65.064, 4.9538e-24], [65.214, 4.947e-24], [65.365, 4.9403e-24], [65.515, 4.9337e-24], [65.666, 4.927e-24], [65.818, 4.9204e-24], [65.97, 4.9138e-24], [66.122, 4.9073e-24], [66.274, 4.9007e-24], [66.427, 4.8943e-24], [66.58, 4.8878e-24], [66.734, 4.8814e-24], [66.888, 4.875e-24], [67.042, 4.8686e-24], [67.196, 4.8622e-24], [67.351, 4.8559e-24], [67.507, 4.8496e-24], [67.662, 4.8434e-24], [67.818, 4.8371e-24], [67.975, 4.8309e-24], [68.132, 4.8248e-24], [68.289, 4.8186e-24], [68.446, 4.8125e-24], [68.604, 4.8064e-24], [68.762, 4.8003e-24], [68.921, 4.7943e-24], [69.08, 4.7883e-24], [69.239, 4.7823e-24], [69.399, 4.7763e-24], [69.559, 4.7704e-24], [69.719, 4.7645e-24], [69.88, 4.7586e-24], [70.041, 4.7527e-24], [70.202, 4.7469e-24], [70.364, 4.7411e-24], [70.527, 4.7353e-24], [70.689, 4.7296e-24], [70.852, 4.7238e-24], [71.016, 4.7181e-24], [71.179, 4.7124e-24], [71.344, 4.7068e-24], [71.508, 4.7012e-24], [71.673, 4.6955e-24], [71.838, 4.69e-24], [72.004, 4.6844e-24], [72.17, 4.6789e-24], [72.336, 4.6734e-24], [72.503, 4.6679e-24], [72.67, 4.6624e-24], [72.838, 4.657e-24], [73.006, 4.6515e-24], [73.174, 4.6461e-24], [73.343, 4.6408e-24], [73.512, 4.6354e-24], [73.682, 4.6301e-24], [73.852, 4.6248e-24], [74.022, 4.6195e-24], [74.193, 4.6142e-24], [74.364, 4.609e-24], [74.535, 4.6038e-24], [74.707, 4.5986e-24], [74.879, 4.5934e-24], [75.052, 4.5882e-24], [75.225, 4.5831e-24], [75.399, 4.578e-24], [75.572, 4.5729e-24], [75.747, 4.5678e-24], [75.921, 4.5628e-24], [76.096, 4.5577e-24], [76.272, 4.5527e-24], [76.448, 4.5477e-24], [76.624, 4.5427e-24], [76.801, 4.5378e-24], [76.978, 4.5329e-24], [77.155, 4.528e-24], [77.333, 4.5231e-24], [77.512, 4.5182e-24], [77.69, 4.5133e-24], [77.87, 4.5085e-24], [78.049, 4.5037e-24], [78.229, 4.4989e-24], [78.409, 4.4941e-24], [78.59, 4.4894e-24], [78.772, 4.4846e-24], [78.953, 4.4799e-24], [79.135, 4.4752e-24], [79.318, 4.4705e-24], [79.501, 4.4659e-24], [79.684, 4.4612e-24], [79.868, 4.4566e-24], [80.052, 4.452e-24], [80.236, 4.4474e-24], [80.421, 4.4428e-24], [80.607, 4.4382e-24], [80.793, 4.4337e-24], [80.979, 4.4292e-24], [81.166, 4.4247e-24], [81.353, 4.4202e-24], [81.541, 4.4157e-24], [81.729, 4.4113e-24], [81.917, 4.4068e-24], [82.106, 4.4024e-24], [82.295, 4.398e-24], [82.485, 4.3936e-24], [82.675, 4.3893e-24], [82.866, 4.3849e-24], [83.057, 4.3806e-24], [83.249, 4.3762e-24], [83.441, 4.3719e-24], [83.633, 4.3676e-24], [83.826, 4.3634e-24], [84.019, 4.3591e-24], [84.213, 4.3549e-24], [84.407, 4.3506e-24], [84.602, 4.3464e-24], [84.797, 4.3422e-24], [84.992, 4.3381e-24], [85.188, 4.3339e-24], [85.385, 4.3297e-24], [85.582, 4.3256e-24], [85.779, 4.3215e-24], [85.977, 4.3174e-24], [86.175, 4.3133e-24], [86.374, 4.3092e-24], [86.573, 4.3052e-24], [86.773, 4.3011e-24], [86.973, 4.2971e-24], [87.173, 4.2931e-24], [87.374, 4.2891e-24], [87.576, 4.2851e-24], [87.778, 4.2811e-24], [87.98, 4.2771e-24], [88.183, 4.2732e-24], [88.386, 4.2693e-24], [88.59, 4.2654e-24], [88.795, 4.2614e-24], [88.999, 4.2576e-24], [89.205, 4.2537e-24], [89.41, 4.2498e-24], [89.616, 4.246e-24], [89.823, 4.2421e-24], [90.03, 4.2383e-24], [90.238, 4.2345e-24], [90.446, 4.2307e-24], [90.655, 4.2269e-24], [90.864, 4.2232e-24], [91.073, 4.2194e-24], [91.283, 4.2157e-24], [91.494, 4.2119e-24], [91.705, 4.2082e-24], [91.916, 4.2045e-24], [92.128, 4.2008e-24], [92.34, 4.1971e-24], [92.553, 4.1935e-24], [92.767, 4.1898e-24], [92.981, 4.1862e-24], [93.195, 4.1825e-24], [93.41, 4.1789e-24], [93.625, 4.1753e-24], [93.841, 4.1717e-24], [94.058, 4.1681e-24], [94.275, 4.1646e-24], [94.492, 4.161e-24], [94.71, 4.1574e-24], [94.928, 4.1539e-24], [95.147, 4.1504e-24], [95.367, 4.1469e-24], [95.587, 4.1434e-24], [95.807, 4.1399e-24], [96.028, 4.1364e-24], [96.249, 4.1329e-24], [96.471, 4.1295e-24], [96.694, 4.1261e-24], [96.917, 4.1226e-24], [97.14, 4.1192e-24], [97.364, 4.1158e-24], [97.589, 4.1124e-24], [97.814, 4.109e-24], [98.039, 4.1056e-24], [98.266, 4.1023e-24], [98.492, 4.0989e-24], [98.719, 4.0956e-24], [98.947, 4.0922e-24], [99.175, 4.0889e-24], [99.404, 4.0856e-24], [99.633, 4.0823e-24], [99.863, 4.079e-24], [100.09, 4.0757e-24], [100.32, 4.0725e-24], [100.56, 4.0692e-24], [100.79, 4.066e-24], [101.02, 4.0627e-24], [101.25, 4.0595e-24], [101.49, 4.0563e-24], [101.72, 4.0531e-24], [101.95, 4.0499e-24], [102.19, 4.0467e-24], [102.43, 4.0435e-24], [102.66, 4.0403e-24], [102.9, 4.0372e-24], [103.14, 4.034e-24], [103.37, 4.0309e-24], [103.61, 4.0278e-24], [103.85, 4.0247e-24], [104.09, 4.0216e-24], [104.33, 4.0185e-24], [104.57, 4.0154e-24], [104.81, 4.0123e-24], [105.05, 4.0092e-24], [105.3, 4.0062e-24], [105.54, 4.0031e-24], [105.78, 4.0001e-24], [106.03, 3.997e-24], [106.27, 3.994e-24], [106.52, 3.991e-24], [106.76, 3.988e-24], [107.01, 3.985e-24], [107.25, 3.982e-24], [107.5, 3.979e-24], [107.75, 3.9761e-24], [108.0, 3.9731e-24], [108.25, 3.9702e-24], [108.5, 3.9672e-24], [108.75, 3.9643e-24], [109.0, 3.9614e-24], [109.25, 3.9585e-24], [109.5, 3.9556e-24], [109.75, 3.9527e-24], [110.01, 3.9498e-24], [110.26, 3.9469e-24], [110.51, 3.944e-24], [110.77, 3.9412e-24], [111.02, 3.9383e-24], [111.28, 3.9355e-24], [111.54, 3.9327e-24], [111.79, 3.9298e-24], [112.05, 3.927e-24], [112.31, 3.9242e-24], [112.57, 3.9214e-24], [112.83, 3.9186e-24], [113.09, 3.9158e-24], [113.35, 3.913e-24], [113.61, 3.9103e-24], [113.87, 3.9075e-24], [114.14, 3.9048e-24], [114.4, 3.902e-24], [114.66, 3.8993e-24], [114.93, 3.8966e-24], [115.19, 3.8939e-24], [115.46, 3.8911e-24], [115.72, 3.8884e-24], [115.99, 3.8857e-24], [116.26, 3.8831e-24], [116.53, 3.8804e-24], [116.8, 3.8777e-24], [117.06, 3.8751e-24], [117.33, 3.8724e-24], [117.61, 3.8698e-24], [117.88, 3.8671e-24], [118.15, 3.8645e-24], [118.42, 3.8619e-24], [118.69, 3.8593e-24], [118.97, 3.8566e-24], [119.24, 3.854e-24], [119.52, 3.8515e-24], [119.79, 3.8489e-24], [120.07, 3.8463e-24], [120.35, 3.8437e-24], [120.62, 3.8412e-24], [120.9, 3.8386e-24], [121.18, 3.8361e-24], [121.46, 3.8335e-24], [121.74, 3.831e-24], [122.02, 3.8285e-24], [122.3, 3.826e-24], [122.58, 3.8235e-24], [122.87, 3.821e-24], [123.15, 3.8185e-24], [123.43, 3.816e-24], [123.72, 3.8135e-24], [124.0, 3.811e-24], [124.29, 3.8086e-24], [124.58, 3.8061e-24], [124.86, 3.8037e-24], [125.15, 3.8012e-24], [125.44, 3.7988e-24], [125.73, 3.7964e-24], [126.02, 3.7939e-24], [126.31, 3.7915e-24], [126.6, 3.7891e-24], [126.89, 3.7867e-24], [127.19, 3.7843e-24], [127.48, 3.7819e-24], [127.77, 3.7796e-24], [128.07, 3.7772e-24], [128.36, 3.7748e-24], [128.66, 3.7725e-24], [128.96, 3.7701e-24], [129.25, 3.7678e-24], [129.55, 3.7654e-24], [129.85, 3.7631e-24], [130.15, 3.7608e-24], [130.45, 3.7585e-24], [130.75, 3.7562e-24], [131.05, 3.7539e-24], [131.35, 3.7516e-24], [131.66, 3.7493e-24], [131.96, 3.747e-24], [132.26, 3.7447e-24], [132.57, 3.7425e-24], [132.88, 3.7402e-24], [133.18, 3.738e-24], [133.49, 3.7357e-24], [133.8, 3.7335e-24], [134.11, 3.7312e-24], [134.41, 3.729e-24], [134.72, 3.7268e-24], [135.04, 3.7246e-24], [135.35, 3.7224e-24], [135.66, 3.7202e-24], [135.97, 3.718e-24], [136.29, 3.7158e-24], [136.6, 3.7136e-24], [136.91, 3.7114e-24], [137.23, 3.7093e-24], [137.55, 3.7071e-24], [137.86, 3.705e-24], [138.18, 3.7028e-24], [138.5, 3.7007e-24], [138.82, 3.6985e-24], [139.14, 3.6964e-24], [139.46, 3.6943e-24], [139.78, 3.6922e-24], [140.1, 3.6901e-24], [140.43, 3.688e-24], [140.75, 3.6859e-24], [141.08, 3.6838e-24], [141.4, 3.6817e-24], [141.73, 3.6796e-24], [142.05, 3.6776e-24], [142.38, 3.6755e-24], [142.71, 3.6735e-24], [143.04, 3.6714e-24], [143.37, 3.6694e-24], [143.7, 3.6673e-24], [144.03, 3.6653e-24], [144.36, 3.6633e-24], [144.7, 3.6613e-24], [145.03, 3.6592e-24], [145.36, 3.6572e-24], [145.7, 3.6552e-24], [146.04, 3.6533e-24], [146.37, 3.6513e-24], [146.71, 3.6493e-24], [147.05, 3.6473e-24], [147.39, 3.6454e-24], [147.73, 3.6434e-24], [148.07, 3.6414e-24], [148.41, 3.6395e-24], [148.75, 3.6376e-24], [149.09, 3.6356e-24], [149.44, 3.6337e-24], [149.78, 3.6318e-24], [150.13, 3.6299e-24], [150.47, 3.6279e-24], [150.82, 3.626e-24], [151.17, 3.6241e-24], [151.52, 3.6222e-24], [151.87, 3.6204e-24], [152.22, 3.6185e-24], [152.57, 3.6166e-24], [152.92, 3.6147e-24], [153.27, 3.6129e-24], [153.63, 3.611e-24], [153.98, 3.6092e-24], [154.34, 3.6073e-24], [154.69, 3.6055e-24], [155.05, 3.6037e-24], [155.41, 3.6019e-24], [155.76, 3.6e-24], [156.12, 3.5982e-24], [156.48, 3.5964e-24], [156.84, 3.5946e-24], [157.21, 3.5928e-24], [157.57, 3.5911e-24], [157.93, 3.5893e-24], [158.3, 3.5875e-24], [158.66, 3.5857e-24], [159.03, 3.584e-24], [159.39, 3.5822e-24], [159.76, 3.5805e-24], [160.13, 3.5787e-24], [160.5, 3.577e-24], [160.87, 3.5753e-24], [161.24, 3.5735e-24], [161.61, 3.5718e-24], [161.98, 3.5701e-24], [162.36, 3.5684e-24], [162.73, 3.5667e-24], [163.11, 3.565e-24], [163.48, 3.5633e-24], [163.86, 3.5616e-24], [164.24, 3.56e-24], [164.62, 3.5583e-24], [165.0, 3.5566e-24], [165.38, 3.555e-24], [165.76, 3.5533e-24], [166.14, 3.5517e-24], [166.52, 3.55e-24], [166.91, 3.5484e-24], [167.29, 3.5468e-24], [167.68, 3.5452e-24], [168.07, 3.5436e-24], [168.45, 3.5419e-24], [168.84, 3.5403e-24], [169.23, 3.5388e-24], [169.62, 3.5372e-24], [170.01, 3.5356e-24], [170.4, 3.534e-24], [170.8, 3.5324e-24], [171.19, 3.5309e-24], [171.59, 3.5293e-24], [171.98, 3.5278e-24], [172.38, 3.5262e-24], [172.78, 3.5247e-24], [173.17, 3.5231e-24], [173.57, 3.5216e-24], [173.97, 3.5201e-24], [174.37, 3.5186e-24], [174.78, 3.5171e-24], [175.18, 3.5156e-24], [175.58, 3.5141e-24], [175.99, 3.5126e-24], [176.39, 3.5111e-24], [176.8, 3.5096e-24], [177.21, 3.5081e-24], [177.62, 3.5067e-24], [178.03, 3.5052e-24], [178.44, 3.5038e-24], [178.85, 3.5023e-24], [179.26, 3.5009e-24], [179.68, 3.4994e-24], [180.09, 3.498e-24], [180.5, 3.4966e-24], [180.92, 3.4952e-24], [181.34, 3.4938e-24], [181.76, 3.4924e-24], [182.18, 3.491e-24], [182.6, 3.4896e-24], [183.02, 3.4882e-24], [183.44, 3.4868e-24], [183.86, 3.4854e-24], [184.29, 3.4841e-24], [184.71, 3.4827e-24], [185.14, 3.4813e-24], [185.56, 3.48e-24], [185.99, 3.4787e-24], [186.42, 3.4773e-24], [186.85, 3.476e-24], [187.28, 3.4747e-24], [187.71, 3.4733e-24], [188.15, 3.472e-24], [188.58, 3.4707e-24], [189.01, 3.4694e-24], [189.45, 3.4681e-24], [189.89, 3.4669e-24], [190.33, 3.4656e-24], [190.76, 3.4643e-24], [191.2, 3.463e-24], [191.65, 3.4618e-24], [192.09, 3.4605e-24], [192.53, 3.4593e-24], [192.97, 3.458e-24], [193.42, 3.4568e-24], [193.87, 3.4556e-24], [194.31, 3.4544e-24], [194.76, 3.4531e-24], [195.21, 3.4519e-24], [195.66, 3.4507e-24], [196.11, 3.4495e-24], [196.56, 3.4483e-24], [197.02, 3.4472e-24], [197.47, 3.446e-24], [197.93, 3.4448e-24], [198.38, 3.4436e-24], [198.84, 3.4425e-24], [199.3, 3.4413e-24], [199.76, 3.4402e-24], [200.22, 3.439e-24], [200.68, 3.4379e-24], [201.14, 3.4368e-24], [201.61, 3.4357e-24], [202.07, 3.4346e-24], [202.54, 3.4335e-24], [203.0, 3.4324e-24], [203.47, 3.4313e-24], [203.94, 3.4302e-24], [204.41, 3.4291e-24], [204.88, 3.428e-24], [205.36, 3.427e-24], [205.83, 3.4259e-24], [206.3, 3.4248e-24], [206.78, 3.4238e-24], [207.26, 3.4228e-24], [207.73, 3.4217e-24], [208.21, 3.4207e-24], [208.69, 3.4197e-24], [209.18, 3.4187e-24], [209.66, 3.4177e-24], [210.14, 3.4167e-24], [210.63, 3.4157e-24], [211.11, 3.4147e-24], [211.6, 3.4137e-24], [212.09, 3.4127e-24], [212.58, 3.4118e-24], [213.07, 3.4108e-24], [213.56, 3.4098e-24], [214.05, 3.4089e-24], [214.54, 3.408e-24], [215.04, 3.407e-24], [215.53, 3.4061e-24], [216.03, 3.4052e-24], [216.53, 3.4043e-24], [217.03, 3.4034e-24], [217.53, 3.4025e-24], [218.03, 3.4016e-24], [218.53, 3.4007e-24], [219.04, 3.3998e-24], [219.54, 3.3989e-24], [220.05, 3.3981e-24], [220.56, 3.3972e-24], [221.06, 3.3964e-24], [221.57, 3.3955e-24], [222.08, 3.3947e-24], [222.6, 3.3938e-24], [223.11, 3.393e-24], [223.62, 3.3922e-24], [224.14, 3.3914e-24], [224.66, 3.3906e-24], [225.18, 3.3898e-24], [225.69, 3.389e-24], [226.22, 3.3882e-24], [226.74, 3.3875e-24], [227.26, 3.3867e-24], [227.78, 3.3859e-24], [228.31, 3.3852e-24], [228.84, 3.3844e-24], [229.36, 3.3837e-24], [229.89, 3.383e-24], [230.42, 3.3822e-24], [230.95, 3.3815e-24], [231.49, 3.3808e-24], [232.02, 3.3801e-24], [232.56, 3.3794e-24], [233.09, 3.3787e-24], [233.63, 3.378e-24], [234.17, 3.3774e-24], [234.71, 3.3767e-24], [235.25, 3.376e-24], [235.79, 3.3754e-24], [236.34, 3.3747e-24], [236.88, 3.3741e-24], [237.43, 3.3735e-24], [237.97, 3.3729e-24], [238.52, 3.3722e-24], [239.07, 3.3716e-24], [239.62, 3.371e-24], [240.18, 3.3704e-24], [240.73, 3.3699e-24], [241.29, 3.3693e-24], [241.84, 3.3687e-24], [242.4, 3.3681e-24], [242.96, 3.3676e-24], [243.52, 3.367e-24], [244.08, 3.3665e-24], [244.64, 3.366e-24], [245.21, 3.3654e-24], [245.77, 3.3649e-24], [246.34, 3.3644e-24], [246.91, 3.3639e-24], [247.48, 3.3634e-24], [248.05, 3.3629e-24], [248.62, 3.3624e-24], [249.19, 3.362e-24], [249.77, 3.3615e-24], [250.34, 3.3611e-24], [250.92, 3.3606e-24], [251.5, 3.3602e-24], [252.08, 3.3597e-24], [252.66, 3.3593e-24], [253.24, 3.3589e-24], [253.83, 3.3585e-24], [254.41, 3.3581e-24], [255.0, 3.3577e-24], [255.59, 3.3573e-24], [256.18, 3.3569e-24], [256.77, 3.3565e-24], [257.36, 3.3562e-24], [257.95, 3.3558e-24], [258.55, 3.3555e-24], [259.14, 3.3551e-24], [259.74, 3.3548e-24], [260.34, 3.3545e-24], [260.94, 3.3542e-24], [261.54, 3.3539e-24], [262.15, 3.3536e-24], [262.75, 3.3533e-24], [263.36, 3.353e-24], [263.96, 3.3527e-24], [264.57, 3.3525e-24], [265.18, 3.3522e-24], [265.79, 3.352e-24], [266.41, 3.3517e-24], [267.02, 3.3515e-24], [267.64, 3.3513e-24], [268.25, 3.351e-24], [268.87, 3.3508e-24], [269.49, 3.3506e-24], [270.11, 3.3504e-24], [270.74, 3.3503e-24], [271.36, 3.3501e-24], [271.99, 3.3499e-24], [272.61, 3.3498e-24], [273.24, 3.3496e-24], [273.87, 3.3495e-24], [274.5, 3.3493e-24], [275.14, 3.3492e-24], [275.77, 3.3491e-24], [276.41, 3.349e-24], [277.05, 3.3489e-24], [277.68, 3.3488e-24], [278.32, 3.3487e-24], [278.97, 3.3486e-24], [279.61, 3.3486e-24], [280.25, 3.3485e-24], [280.9, 3.3485e-24], [281.55, 3.3484e-24], [282.2, 3.3484e-24], [282.85, 3.3484e-24], [283.5, 3.3484e-24], [284.15, 3.3484e-24], [284.81, 3.3484e-24], [285.47, 3.3484e-24], [286.13, 3.3484e-24], [286.78, 3.3485e-24], [287.45, 3.3485e-24], [288.11, 3.3485e-24], [288.77, 3.3486e-24], [289.44, 3.3487e-24], [290.11, 3.3488e-24], [290.78, 3.3488e-24], [291.45, 3.3489e-24], [292.12, 3.349e-24], [292.79, 3.3492e-24], [293.47, 3.3493e-24], [294.14, 3.3494e-24], [294.82, 3.3495e-24], [295.5, 3.3497e-24], [296.18, 3.3499e-24], [296.87, 3.35e-24], [297.55, 3.3502e-24], [298.24, 3.3504e-24], [298.92, 3.3506e-24], [299.61, 3.3508e-24], [300.31, 3.351e-24], [301.0, 3.3512e-24], [301.69, 3.3515e-24], [302.39, 3.3517e-24], [303.08, 3.352e-24], [303.78, 3.3522e-24], [304.48, 3.3525e-24], [305.19, 3.3528e-24], [305.89, 3.3531e-24], [306.6, 3.3534e-24], [307.3, 3.3537e-24], [308.01, 3.354e-24], [308.72, 3.3543e-24], [309.43, 3.3547e-24], [310.15, 3.355e-24], [310.86, 3.3554e-24], [311.58, 3.3557e-24], [312.3, 3.3561e-24], [313.02, 3.3565e-24], [313.74, 3.3569e-24], [314.46, 3.3573e-24], [315.19, 3.3577e-24], [315.91, 3.3581e-24], [316.64, 3.3586e-24], [317.37, 3.359e-24], [318.11, 3.3595e-24], [318.84, 3.36e-24], [319.57, 3.3604e-24], [320.31, 3.3609e-24], [321.05, 3.3614e-24], [321.79, 3.3619e-24], [322.53, 3.3624e-24], [323.28, 3.363e-24], [324.02, 3.3635e-24], [324.77, 3.3641e-24], [325.52, 3.3646e-24], [326.27, 3.3652e-24], [327.02, 3.3658e-24], [327.77, 3.3663e-24], [328.53, 3.3669e-24], [329.29, 3.3676e-24], [330.05, 3.3682e-24], [330.81, 3.3688e-24], [331.57, 3.3694e-24], [332.34, 3.3701e-24], [333.1, 3.3708e-24], [333.87, 3.3714e-24], [334.64, 3.3721e-24], [335.41, 3.3728e-24], [336.19, 3.3735e-24], [336.96, 3.3742e-24], [337.74, 3.3749e-24], [338.52, 3.3757e-24], [339.3, 3.3764e-24], [340.08, 3.3772e-24], [340.86, 3.378e-24], [341.65, 3.3787e-24], [342.44, 3.3795e-24], [343.23, 3.3803e-24], [344.02, 3.3811e-24], [344.81, 3.382e-24], [345.61, 3.3828e-24], [346.4, 3.3836e-24], [347.2, 3.3845e-24], [348.0, 3.3854e-24], [348.81, 3.3862e-24], [349.61, 3.3871e-24], [350.42, 3.388e-24], [351.23, 3.3889e-24], [352.04, 3.3899e-24], [352.85, 3.3908e-24], [353.66, 3.3917e-24], [354.48, 3.3927e-24], [355.29, 3.3937e-24], [356.11, 3.3946e-24], [356.93, 3.3956e-24], [357.76, 3.3966e-24], [358.58, 3.3976e-24], [359.41, 3.3987e-24], [360.24, 3.3997e-24], [361.07, 3.4007e-24], [361.9, 3.4018e-24], [362.74, 3.4029e-24], [363.57, 3.4039e-24], [364.41, 3.405e-24], [365.25, 3.4061e-24], [366.09, 3.4073e-24], [366.94, 3.4084e-24], [367.78, 3.4095e-24], [368.63, 3.4107e-24], [369.48, 3.4118e-24], [370.33, 3.413e-24], [371.19, 3.4142e-24], [372.04, 3.4154e-24], [372.9, 3.4166e-24], [373.76, 3.4178e-24], [374.62, 3.4191e-24], [375.49, 3.4203e-24], [376.35, 3.4216e-24], [377.22, 3.4228e-24], [378.09, 3.4241e-24], [378.96, 3.4254e-24], [379.84, 3.4267e-24], [380.71, 3.428e-24], [381.59, 3.4294e-24], [382.47, 3.4307e-24], [383.35, 3.4321e-24], [384.24, 3.4334e-24], [385.12, 3.4348e-24], [386.01, 3.4362e-24], [386.9, 3.4376e-24], [387.79, 3.439e-24], [388.69, 3.4405e-24], [389.58, 3.4419e-24], [390.48, 3.4434e-24], [391.38, 3.4448e-24], [392.28, 3.4463e-24], [393.19, 3.4478e-24], [394.1, 3.4493e-24], [395.0, 3.4508e-24], [395.92, 3.4524e-24], [396.83, 3.4539e-24], [397.74, 3.4555e-24], [398.66, 3.457e-24], [399.58, 3.4586e-24], [400.5, 3.4602e-24], [401.43, 3.4618e-24], [402.35, 3.4635e-24], [403.28, 3.4651e-24], [404.21, 3.4667e-24], [405.14, 3.4684e-24], [406.08, 3.4701e-24], [407.01, 3.4718e-24], [407.95, 3.4735e-24], [408.89, 3.4752e-24], [409.83, 3.4769e-24], [410.78, 3.4786e-24], [411.73, 3.4804e-24], [412.68, 3.4822e-24], [413.63, 3.4839e-24], [414.58, 3.4857e-24], [415.54, 3.4875e-24], [416.5, 3.4894e-24], [417.46, 3.4912e-24], [418.42, 3.493e-24], [419.38, 3.4949e-24], [420.35, 3.4968e-24], [421.32, 3.4987e-24], [422.29, 3.5006e-24], [423.27, 3.5025e-24], [424.24, 3.5044e-24], [425.22, 3.5063e-24], [426.2, 3.5083e-24], [427.18, 3.5103e-24], [428.17, 3.5123e-24], [429.16, 3.5143e-24], [430.15, 3.5163e-24], [431.14, 3.5183e-24], [432.13, 3.5203e-24], [433.13, 3.5224e-24], [434.13, 3.5245e-24], [435.13, 3.5265e-24], [436.13, 3.5286e-24], [437.14, 3.5307e-24], [438.14, 3.5329e-24], [439.16, 3.535e-24], [440.17, 3.5372e-24], [441.18, 3.5393e-24], [442.2, 3.5415e-24], [443.22, 3.5437e-24], [444.24, 3.5459e-24], [445.27, 3.5481e-24], [446.29, 3.5504e-24], [447.32, 3.5526e-24], [448.35, 3.5549e-24], [449.39, 3.5572e-24], [450.42, 3.5595e-24], [451.46, 3.5618e-24], [452.5, 3.5641e-24], [453.55, 3.5665e-24], [454.59, 3.5688e-24], [455.64, 3.5712e-24], [456.69, 3.5736e-24], [457.75, 3.576e-24], [458.8, 3.5784e-24], [459.86, 3.5809e-24], [460.92, 3.5833e-24], [461.98, 3.5858e-24], [463.05, 3.5882e-24], [464.12, 3.5907e-24], [465.19, 3.5933e-24], [466.26, 3.5958e-24], [467.33, 3.5983e-24], [468.41, 3.6009e-24], [469.49, 3.6035e-24], [470.57, 3.6061e-24], [471.66, 3.6087e-24], [472.75, 3.6113e-24], [473.84, 3.6139e-24], [474.93, 3.6166e-24], [476.02, 3.6193e-24], [477.12, 3.622e-24], [478.22, 3.6247e-24], [479.33, 3.6274e-24], [480.43, 3.6302e-24], [481.54, 3.633e-24], [482.65, 3.6358e-24], [483.76, 3.6386e-24], [484.88, 3.6415e-24], [486.0, 3.6445e-24], [487.12, 3.6475e-24], [488.24, 3.6506e-24], [489.37, 3.6538e-24], [490.49, 3.6574e-24], [491.63, 3.6615e-24], [492.76, 3.6668e-24], [493.9, 3.6763e-24], [495.03, 3.7073e-24], [496.18, 4.6937e-24], [497.32, 3.773e-24], [498.47, 3.6946e-24], [499.62, 3.6862e-24], [500.77, 3.6856e-24], [501.92, 3.6871e-24], [503.08, 3.6893e-24], [504.24, 3.6919e-24], [505.4, 3.6946e-24], [506.57, 3.6975e-24], [507.74, 3.7005e-24], [508.91, 3.7035e-24], [510.08, 3.7066e-24], [511.26, 3.7097e-24], [512.44, 3.7128e-24], [513.62, 3.7159e-24], [514.8, 3.7191e-24], [515.99, 3.7223e-24], [517.18, 3.7255e-24], [518.37, 3.7288e-24], [519.57, 3.732e-24], [520.77, 3.7353e-24], [521.97, 3.7386e-24], [523.17, 3.7419e-24], [524.38, 3.7452e-24], [525.59, 3.7486e-24], [526.8, 3.7519e-24], [528.01, 3.7553e-24], [529.23, 3.7587e-24], [530.45, 3.7622e-24], [531.67, 3.7656e-24], [532.9, 3.7691e-24], [534.13, 3.7725e-24], [535.36, 3.776e-24], [536.6, 3.7795e-24], [537.83, 3.7831e-24], [539.07, 3.7866e-24], [540.32, 3.7902e-24], [541.56, 3.7938e-24], [542.81, 3.7974e-24], [544.06, 3.801e-24], [545.32, 3.8046e-24], [546.57, 3.8083e-24], [547.84, 3.8119e-24], [549.1, 3.8156e-24], [550.36, 3.8193e-24], [551.63, 3.823e-24], [552.91, 3.8268e-24], [554.18, 3.8305e-24], [555.46, 3.8343e-24], [556.74, 3.8381e-24], [558.02, 3.8419e-24], [559.31, 3.8458e-24], [560.6, 3.8496e-24], [561.89, 3.8535e-24], [563.19, 3.8574e-24], [564.49, 3.8613e-24], [565.79, 3.8652e-24], [567.09, 3.8691e-24], [568.4, 3.8731e-24], [569.71, 3.8771e-24], [571.03, 3.8811e-24], [572.34, 3.8851e-24], [573.66, 3.8891e-24], [574.99, 3.8932e-24], [576.31, 3.8972e-24], [577.64, 3.9013e-24], [578.97, 3.9054e-24], [580.31, 3.9096e-24], [581.65, 3.9137e-24], [582.99, 3.9179e-24], [584.33, 3.9221e-24], [585.68, 3.9263e-24], [587.03, 3.9305e-24], [588.38, 3.9347e-24], [589.74, 3.939e-24], [591.1, 3.9433e-24], [592.46, 3.9476e-24], [593.83, 3.9519e-24], [595.2, 3.9562e-24], [596.57, 3.9606e-24], [597.95, 3.9649e-24], [599.33, 3.9693e-24], [600.71, 3.9737e-24], [602.09, 3.9782e-24], [603.48, 3.9826e-24], [604.87, 3.9871e-24], [606.27, 3.9916e-24], [607.67, 3.9961e-24], [609.07, 4.0006e-24], [610.47, 4.0052e-24], [611.88, 4.0097e-24], [613.29, 4.0143e-24], [614.7, 4.0189e-24], [616.12, 4.0236e-24], [617.54, 4.0282e-24], [618.97, 4.0329e-24], [620.39, 4.0376e-24], [621.83, 4.0423e-24], [623.26, 4.047e-24], [624.7, 4.0517e-24], [626.14, 4.0565e-24], [627.58, 4.0613e-24], [629.03, 4.0661e-24], [630.48, 4.0709e-24], [631.93, 4.0758e-24], [633.39, 4.0806e-24], [634.85, 4.0855e-24], [636.31, 4.0904e-24], [637.78, 4.0953e-24], [639.25, 4.1003e-24], [640.73, 4.1053e-24], [642.2, 4.1103e-24], [643.68, 4.1153e-24], [645.17, 4.1203e-24], [646.66, 4.1253e-24], [648.15, 4.1304e-24], [649.64, 4.1355e-24], [651.14, 4.1406e-24], [652.64, 4.1458e-24], [654.15, 4.1509e-24], [655.66, 4.1561e-24], [657.17, 4.1613e-24], [658.68, 4.1665e-24], [660.2, 4.1717e-24], [661.72, 4.177e-24], [663.25, 4.1823e-24], [664.78, 4.1876e-24], [666.31, 4.1929e-24], [667.85, 4.1982e-24], [669.39, 4.2036e-24], [670.93, 4.209e-24], [672.48, 4.2144e-24], [674.03, 4.2198e-24], [675.59, 4.2253e-24], [677.14, 4.2307e-24], [678.7, 4.2362e-24], [680.27, 4.2417e-24], [681.84, 4.2473e-24], [683.41, 4.2528e-24], [684.99, 4.2584e-24], [686.57, 4.264e-24], [688.15, 4.2696e-24], [689.74, 4.2753e-24], [691.33, 4.2809e-24], [692.92, 4.2866e-24], [694.52, 4.2923e-24], [696.12, 4.2981e-24], [697.73, 4.3038e-24], [699.33, 4.3096e-24], [700.95, 4.3154e-24], [702.56, 4.3212e-24], [704.18, 4.3271e-24], [705.81, 4.3329e-24], [707.44, 4.3388e-24], [709.07, 4.3447e-24], [710.7, 4.3506e-24], [712.34, 4.3566e-24], [713.98, 4.3626e-24], [715.63, 4.3686e-24], [717.28, 4.3746e-24], [718.93, 4.3806e-24], [720.59, 4.3867e-24], [722.25, 4.3928e-24], [723.92, 4.3989e-24], [725.59, 4.405e-24], [727.26, 4.4112e-24], [728.94, 4.4173e-24], [730.62, 4.4235e-24], [732.3, 4.4298e-24], [733.99, 4.436e-24], [735.69, 4.4423e-24], [737.38, 4.4486e-24], [739.08, 4.4549e-24], [740.79, 4.4612e-24], [742.5, 4.4676e-24], [744.21, 4.474e-24], [745.92, 4.4804e-24], [747.64, 4.4868e-24], [749.37, 4.4933e-24], [751.1, 4.4998e-24], [752.83, 4.5063e-24], [754.56, 4.5128e-24], [756.3, 4.5193e-24], [758.05, 4.5259e-24], [759.8, 4.5325e-24], [761.55, 4.5391e-24], [763.3, 4.5458e-24], [765.06, 4.5524e-24], [766.83, 4.5591e-24], [768.6, 4.5658e-24], [770.37, 4.5726e-24], [772.15, 4.5793e-24], [773.93, 4.5861e-24], [775.71, 4.5929e-24], [777.5, 4.5997e-24], [779.29, 4.6066e-24], [781.09, 4.6135e-24], [782.89, 4.6204e-24], [784.7, 4.6273e-24], [786.51, 4.6343e-24], [788.32, 4.6413e-24], [790.14, 4.6483e-24], [791.96, 4.6553e-24], [793.79, 4.6623e-24], [795.62, 4.6694e-24], [797.45, 4.6765e-24], [799.29, 4.6836e-24], [801.13, 4.6908e-24], [802.98, 4.698e-24], [804.83, 4.7052e-24], [806.69, 4.7124e-24], [808.55, 4.7196e-24], [810.41, 4.7269e-24], [812.28, 4.7342e-24], [814.16, 4.7415e-24], [816.03, 4.7489e-24], [817.91, 4.7562e-24], [819.8, 4.7636e-24], [821.69, 4.7711e-24], [823.59, 4.7785e-24], [825.49, 4.786e-24], [827.39, 4.7935e-24], [829.3, 4.801e-24], [831.21, 4.8086e-24], [833.13, 4.8161e-24], [835.05, 4.8237e-24], [836.97, 4.8314e-24], [838.9, 4.839e-24], [840.84, 4.8467e-24], [842.78, 4.8544e-24], [844.72, 4.8621e-24], [846.67, 4.8699e-24], [848.62, 4.8776e-24], [850.58, 4.8855e-24], [852.54, 4.8933e-24], [854.5, 4.9011e-24], [856.47, 4.909e-24], [858.45, 4.9169e-24], [860.43, 4.9249e-24], [862.41, 4.9328e-24], [864.4, 4.9408e-24], [866.4, 4.9488e-24], [868.39, 4.9569e-24], [870.4, 4.9649e-24], [872.4, 4.973e-24], [874.41, 4.9812e-24], [876.43, 4.9893e-24], [878.45, 4.9975e-24], [880.48, 5.0057e-24], [882.51, 5.0139e-24], [884.54, 5.0221e-24], [886.58, 5.0304e-24], [888.63, 5.0387e-24], [890.68, 5.0471e-24], [892.73, 5.0554e-24], [894.79, 5.0638e-24], [896.85, 5.0722e-24], [898.92, 5.0807e-24], [900.99, 5.0891e-24], [903.07, 5.0976e-24], [905.15, 5.1061e-24], [907.24, 5.1147e-24], [909.33, 5.1233e-24], [911.43, 5.1319e-24], [913.53, 5.1405e-24], [915.64, 5.1492e-24], [917.75, 5.1578e-24], [919.87, 5.1665e-24], [921.99, 5.1753e-24], [924.11, 5.1841e-24], [926.25, 5.1929e-24], [928.38, 5.2017e-24], [930.52, 5.2105e-24], [932.67, 5.2194e-24], [934.82, 5.2283e-24], [936.97, 5.2372e-24], [939.13, 5.2462e-24], [941.3, 5.2552e-24], [943.47, 5.2642e-24], [945.65, 5.2733e-24], [947.83, 5.2823e-24], [950.01, 5.2914e-24], [952.2, 5.3006e-24], [954.4, 5.3097e-24], [956.6, 5.3189e-24], [958.81, 5.3281e-24], [961.02, 5.3374e-24], [963.23, 5.3467e-24], [965.45, 5.356e-24], [967.68, 5.3653e-24], [969.91, 5.3747e-24], [972.15, 5.3841e-24], [974.39, 5.3935e-24], [976.64, 5.4029e-24], [978.89, 5.4124e-24], [981.15, 5.4219e-24], [983.41, 5.4315e-24], [985.68, 5.4411e-24], [987.95, 5.4508e-24], [990.23, 5.4612e-24], [992.51, 5.5229e-24], [994.8, 5.4813e-24], [997.1, 5.4896e-24], [999.39, 5.4991e-24], [1001.7, 5.5088e-24], [1004.0, 5.5186e-24], [1006.3, 5.5284e-24], [1008.6, 5.5382e-24], [1011.0, 5.5481e-24], [1013.3, 5.558e-24], [1015.6, 5.568e-24], [1018.0, 5.5779e-24], [1020.3, 5.5879e-24], [1022.7, 5.598e-24], [1025.0, 5.608e-24], [1027.4, 5.6181e-24], [1029.8, 5.6283e-24], [1032.1, 5.6384e-24], [1034.5, 5.6486e-24], [1036.9, 5.6588e-24], [1039.3, 5.6691e-24], [1041.7, 5.6793e-24], [1044.1, 5.6896e-24], [1046.5, 5.7e-24], [1048.9, 5.7103e-24], [1051.3, 5.7207e-24], [1053.8, 5.7312e-24], [1056.2, 5.7416e-24], [1058.6, 5.7521e-24], [1061.1, 5.7626e-24], [1063.5, 5.7732e-24], [1066.0, 5.7838e-24], [1068.4, 5.7944e-24], [1070.9, 5.805e-24], [1073.4, 5.8157e-24], [1075.8, 5.8264e-24], [1078.3, 5.8372e-24], [1080.8, 5.8479e-24], [1083.3, 5.8587e-24], [1085.8, 5.8696e-24], [1088.3, 5.8805e-24], [1090.8, 5.8913e-24], [1093.3, 5.9023e-24], [1095.8, 5.9132e-24], [1098.4, 5.9242e-24], [1100.9, 5.9353e-24], [1103.4, 5.9463e-24], [1106.0, 5.9574e-24], [1108.5, 5.9685e-24], [1111.1, 5.9797e-24], [1113.7, 5.9909e-24], [1116.2, 6.0021e-24], [1118.8, 6.0133e-24], [1121.4, 6.0246e-24], [1124.0, 6.0359e-24], [1126.6, 6.0473e-24], [1129.2, 6.0586e-24], [1131.8, 6.0701e-24], [1134.4, 6.0815e-24], [1137.0, 6.093e-24], [1139.6, 6.1045e-24], [1142.2, 6.116e-24], [1144.9, 6.1276e-24], [1147.5, 6.1392e-24], [1150.2, 6.1509e-24], [1152.8, 6.1625e-24], [1155.5, 6.1742e-24], [1158.1, 6.186e-24], [1160.8, 6.1978e-24], [1163.5, 6.2096e-24], [1166.2, 6.2214e-24], [1168.9, 6.2333e-24], [1171.5, 6.2452e-24], [1174.2, 6.2571e-24], [1177.0, 6.2691e-24], [1179.7, 6.2811e-24], [1182.4, 6.2932e-24], [1185.1, 6.3052e-24], [1187.9, 6.3173e-24], [1190.6, 6.3295e-24], [1193.3, 6.3417e-24], [1196.1, 6.3539e-24], [1198.8, 6.3661e-24], [1201.6, 6.3784e-24], [1204.4, 6.3907e-24], [1207.2, 6.4031e-24], [1209.9, 6.4155e-24], [1212.7, 6.4279e-24], [1215.5, 6.4403e-24], [1218.3, 6.4528e-24], [1221.1, 6.4653e-24], [1224.0, 6.4779e-24], [1226.8, 6.4905e-24], [1229.6, 6.5031e-24], [1232.4, 6.5158e-24], [1235.3, 6.5285e-24], [1238.1, 6.5412e-24], [1241.0, 6.554e-24], [1243.9, 6.5668e-24], [1246.7, 6.5796e-24], [1249.6, 6.5925e-24], [1252.5, 6.6054e-24], [1255.4, 6.6184e-24], [1258.3, 6.6313e-24], [1261.2, 6.6444e-24], [1264.1, 6.6574e-24], [1267.0, 6.6705e-24], [1269.9, 6.6836e-24], [1272.8, 6.6968e-24], [1275.8, 6.71e-24], [1278.7, 6.7232e-24], [1281.7, 6.7365e-24], [1284.6, 6.7498e-24], [1287.6, 6.7631e-24], [1290.5, 6.7765e-24], [1293.5, 6.7899e-24], [1296.5, 6.8034e-24], [1299.5, 6.8168e-24], [1302.5, 6.8304e-24], [1305.5, 6.8439e-24], [1308.5, 6.8575e-24], [1311.5, 6.8712e-24], [1314.5, 6.8848e-24], [1317.6, 6.8985e-24], [1320.6, 6.9123e-24], [1323.7, 6.9261e-24], [1326.7, 6.9399e-24], [1329.8, 6.9537e-24], [1332.8, 6.9676e-24], [1335.9, 6.9816e-24], [1339.0, 6.9955e-24], [1342.1, 7.0096e-24], [1345.2, 7.0236e-24], [1348.3, 7.0377e-24], [1351.4, 7.0518e-24], [1354.5, 7.066e-24], [1357.6, 7.0802e-24], [1360.8, 7.0944e-24], [1363.9, 7.1087e-24], [1367.0, 7.123e-24], [1370.2, 7.1373e-24], [1373.4, 7.1517e-24], [1376.5, 7.1661e-24], [1379.7, 7.1806e-24], [1382.9, 7.1951e-24], [1386.1, 7.2096e-24], [1389.3, 7.2242e-24], [1392.5, 7.2388e-24], [1395.7, 7.2535e-24], [1398.9, 7.2682e-24], [1402.1, 7.2829e-24], [1405.4, 7.2977e-24], [1408.6, 7.3125e-24], [1411.8, 7.3274e-24], [1415.1, 7.3423e-24], [1418.4, 7.3572e-24], [1421.6, 7.3722e-24], [1424.9, 7.3872e-24], [1428.2, 7.4022e-24], [1431.5, 7.4173e-24], [1434.8, 7.4325e-24], [1438.1, 7.4476e-24], [1441.4, 7.4629e-24], [1444.7, 7.4781e-24], [1448.1, 7.4934e-24], [1451.4, 7.5087e-24], [1454.8, 7.5241e-24], [1458.1, 7.5395e-24], [1461.5, 7.555e-24], [1464.8, 7.5705e-24], [1468.2, 7.586e-24], [1471.6, 7.6016e-24], [1475.0, 7.6172e-24], [1478.4, 7.6329e-24], [1481.8, 7.6486e-24], [1485.2, 7.6644e-24], [1488.7, 7.6837e-24], [1492.1, 7.6961e-24], [1495.5, 7.7118e-24], [1499.0, 7.7277e-24], [1502.4, 7.7436e-24], [1505.9, 7.7596e-24], [1509.4, 7.7756e-24], [1512.9, 7.7917e-24], [1516.3, 7.8078e-24], [1519.8, 7.824e-24], [1523.3, 7.8402e-24], [1526.9, 7.8564e-24], [1530.4, 7.8727e-24], [1533.9, 7.889e-24], [1537.4, 7.9054e-24], [1541.0, 7.9218e-24], [1544.5, 7.9382e-24], [1548.1, 7.9547e-24], [1551.7, 7.9713e-24], [1555.3, 7.9879e-24], [1558.8, 8.0045e-24], [1562.4, 8.0211e-24], [1566.0, 8.0378e-24], [1569.6, 8.0546e-24], [1573.3, 8.0714e-24], [1576.9, 8.0882e-24], [1580.5, 8.1051e-24], [1584.2, 8.122e-24], [1587.8, 8.139e-24], [1591.5, 8.156e-24], [1595.2, 8.1731e-24], [1598.8, 8.1902e-24], [1602.5, 8.2073e-24], [1606.2, 8.2245e-24], [1609.9, 8.2418e-24], [1613.6, 8.259e-24], [1617.4, 8.2763e-24], [1621.1, 8.2937e-24], [1624.8, 8.3111e-24], [1628.6, 8.3286e-24], [1632.3, 8.3461e-24], [1636.1, 8.3636e-24], [1639.9, 8.3812e-24], [1643.6, 8.3989e-24], [1647.4, 8.4165e-24], [1651.2, 8.4343e-24], [1655.0, 8.452e-24], [1658.9, 8.4698e-24], [1662.7, 8.4877e-24], [1666.5, 8.5056e-24], [1670.4, 8.5236e-24], [1674.2, 8.5416e-24], [1678.1, 8.5596e-24], [1681.9, 8.5777e-24], [1685.8, 8.5958e-24], [1689.7, 8.614e-24], [1693.6, 8.6322e-24], [1697.5, 8.6505e-24], [1701.4, 8.6688e-24], [1705.4, 8.6872e-24], [1709.3, 8.7056e-24], [1713.2, 8.7241e-24], [1717.2, 8.7426e-24], [1721.1, 8.7611e-24], [1725.1, 8.7797e-24], [1729.1, 8.7984e-24], [1733.1, 8.8171e-24], [1737.1, 8.8358e-24], [1741.1, 8.8546e-24], [1745.1, 8.8735e-24], [1749.1, 8.8923e-24], [1753.1, 8.9113e-24], [1757.2, 8.9303e-24], [1761.2, 8.9493e-24], [1765.3, 8.9684e-24], [1769.4, 8.9875e-24], [1773.5, 9.0067e-24], [1777.5, 9.0259e-24], [1781.6, 9.0452e-24], [1785.8, 9.0645e-24], [1789.9, 9.0838e-24], [1794.0, 9.1033e-24], [1798.1, 9.1227e-24], [1802.3, 9.1422e-24], [1806.4, 9.1618e-24], [1810.6, 9.1814e-24], [1814.8, 9.2011e-24], [1819.0, 9.2208e-24], [1823.2, 9.2405e-24], [1827.4, 9.2603e-24], [1831.6, 9.2802e-24], [1835.8, 9.3001e-24], [1840.0, 9.3201e-24], [1844.3, 9.3401e-24], [1848.5, 9.3601e-24], [1852.8, 9.3803e-24], [1857.1, 9.4004e-24], [1861.3, 9.4206e-24], [1865.6, 9.4409e-24], [1869.9, 9.4612e-24], [1874.3, 9.4816e-24], [1878.6, 9.502e-24], [1882.9, 9.5224e-24], [1887.2, 9.5429e-24], [1891.6, 9.5635e-24], [1896.0, 9.5841e-24], [1900.3, 9.6048e-24], [1904.7, 9.6255e-24], [1909.1, 9.6463e-24], [1913.5, 9.6671e-24], [1917.9, 9.688e-24], [1922.3, 9.7089e-24], [1926.8, 9.7299e-24], [1931.2, 9.7509e-24], [1935.7, 9.772e-24], [1940.1, 9.7931e-24], [1944.6, 9.8143e-24], [1949.1, 9.8356e-24], [1953.6, 9.8569e-24], [1958.1, 9.8782e-24], [1962.6, 9.8996e-24], [1967.1, 9.9211e-24], [1971.7, 9.9426e-24], [1976.2, 9.9641e-24], [1980.8, 9.9857e-24], [1985.3, 1.0011e-23], [1989.9, 1.0029e-23], [1994.5, 1.0051e-23], [1999.1, 1.0073e-23], [2003.7, 1.0095e-23], [2008.3, 1.0117e-23], [2013.0, 1.0139e-23], [2017.6, 1.0161e-23], [2022.3, 1.0183e-23], [2026.9, 1.0205e-23], [2031.6, 1.0227e-23], [2036.3, 1.0249e-23], [2041.0, 1.0272e-23], [2045.7, 1.0294e-23], [2050.4, 1.0316e-23], [2055.1, 1.0339e-23], [2059.9, 1.0361e-23], [2064.6, 1.0384e-23], [2069.4, 1.0407e-23], [2074.2, 1.0429e-23], [2078.9, 1.0452e-23], [2083.7, 1.0475e-23], [2088.5, 1.0498e-23], [2093.4, 1.0521e-23], [2098.2, 1.0544e-23], [2103.0, 1.0567e-23], [2107.9, 1.059e-23], [2112.7, 1.0613e-23], [2117.6, 1.0636e-23], [2122.5, 1.0659e-23], [2127.4, 1.0683e-23], [2132.3, 1.0706e-23], [2137.2, 1.073e-23], [2142.1, 1.0753e-23], [2147.1, 1.0777e-23], [2152.0, 1.08e-23], [2157.0, 1.0824e-23], [2162.0, 1.0848e-23], [2167.0, 1.0871e-23], [2171.9, 1.0895e-23], [2177.0, 1.0919e-23], [2182.0, 1.0943e-23], [2187.0, 1.0967e-23], [2192.1, 1.0991e-23], [2197.1, 1.1015e-23], [2202.2, 1.104e-23], [2207.3, 1.1064e-23], [2212.3, 1.1088e-23], [2217.4, 1.1113e-23], [2222.6, 1.1137e-23], [2227.7, 1.1161e-23], [2232.8, 1.1186e-23], [2238.0, 1.1211e-23], [2243.1, 1.1235e-23], [2248.3, 1.126e-23], [2253.5, 1.1285e-23], [2258.7, 1.131e-23], [2263.9, 1.1335e-23], [2269.1, 1.136e-23], [2274.3, 1.1385e-23], [2279.6, 1.141e-23], [2284.8, 1.1435e-23], [2290.1, 1.146e-23], [2295.4, 1.1485e-23], [2300.7, 1.1511e-23], [2306.0, 1.1536e-23], [2311.3, 1.1562e-23], [2316.6, 1.1587e-23], [2322.0, 1.1613e-23], [2327.3, 1.1638e-23], [2332.7, 1.1664e-23], [2338.1, 1.169e-23], [2343.5, 1.1716e-23], [2348.9, 1.1742e-23], [2354.3, 1.1768e-23], [2359.7, 1.1794e-23], [2365.2, 1.182e-23], [2370.6, 1.1846e-23], [2376.1, 1.1872e-23], [2381.6, 1.1898e-23], [2387.1, 1.1925e-23], [2392.6, 1.1951e-23], [2398.1, 1.1978e-23], [2403.6, 1.2004e-23], [2409.2, 1.2031e-23], [2414.7, 1.2057e-23], [2420.3, 1.2084e-23], [2425.9, 1.2111e-23], [2431.5, 1.2138e-23], [2437.1, 1.2165e-23], [2442.7, 1.2192e-23], [2448.3, 1.2219e-23], [2454.0, 1.2246e-23], [2459.6, 1.2273e-23], [2465.3, 1.23e-23], [2471.0, 1.2328e-23], [2476.7, 1.2355e-23], [2482.4, 1.2384e-23], [2488.1, 1.241e-23], [2493.8, 1.2437e-23], [2499.6, 1.2465e-23], [2505.4, 1.2493e-23], [2511.1, 1.2521e-23], [2516.9, 1.2548e-23], [2522.7, 1.2576e-23], [2528.5, 1.2604e-23], [2534.4, 1.2632e-23], [2540.2, 1.2661e-23], [2546.1, 1.2689e-23], [2552.0, 1.2717e-23], [2557.8, 1.2745e-23], [2563.7, 1.2774e-23], [2569.6, 1.2802e-23], [2575.6, 1.2831e-23], [2581.5, 1.2859e-23], [2587.5, 1.2888e-23], [2593.4, 1.2917e-23], [2599.4, 1.2945e-23], [2605.4, 1.2974e-23], [2611.4, 1.3003e-23], [2617.4, 1.3032e-23], [2623.5, 1.3061e-23], [2629.5, 1.309e-23], [2635.6, 1.312e-23], [2641.7, 1.3149e-23], [2647.8, 1.3178e-23], [2653.9, 1.3208e-23], [2660.0, 1.3237e-23], [2666.1, 1.3267e-23], [2672.3, 1.3296e-23], [2678.4, 1.3326e-23], [2684.6, 1.3356e-23], [2690.8, 1.3386e-23], [2697.0, 1.3416e-23], [2703.2, 1.3446e-23], [2709.5, 1.3476e-23], [2715.7, 1.3506e-23], [2722.0, 1.3536e-23], [2728.2, 1.3566e-23], [2734.5, 1.3597e-23], [2740.8, 1.3627e-23], [2747.2, 1.3658e-23], [2753.5, 1.3688e-23], [2759.8, 1.3719e-23], [2766.2, 1.375e-23], [2772.6, 1.3781e-23], [2779.0, 1.3811e-23], [2785.4, 1.3842e-23], [2791.8, 1.3873e-23], [2798.2, 1.3904e-23], [2804.7, 1.3936e-23], [2811.2, 1.3967e-23], [2817.6, 1.3998e-23], [2824.1, 1.403e-23], [2830.7, 1.4061e-23], [2837.2, 1.4093e-23], [2843.7, 1.4124e-23], [2850.3, 1.4156e-23], [2856.9, 1.4188e-23], [2863.4, 1.422e-23], [2870.1, 1.4251e-23], [2876.7, 1.4283e-23], [2883.3, 1.4316e-23], [2890.0, 1.4348e-23], [2896.6, 1.438e-23], [2903.3, 1.4412e-23], [2910.0, 1.4445e-23], [2916.7, 1.4477e-23], [2923.4, 1.451e-23], [2930.2, 1.4542e-23], [2936.9, 1.4575e-23], [2943.7, 1.4608e-23], [2950.5, 1.4641e-23], [2957.3, 1.4673e-23], [2964.1, 1.4706e-23], [2970.9, 1.474e-23], [2977.8, 1.4773e-23], [2984.7, 1.4806e-23], [2991.5, 1.4839e-23], [2998.4, 1.4873e-23], [3005.4, 1.4906e-23], [3012.3, 1.494e-23], [3019.2, 1.4973e-23], [3026.2, 1.5007e-23], [3033.2, 1.5041e-23], [3040.2, 1.5075e-23], [3047.2, 1.5109e-23], [3054.2, 1.5143e-23], [3061.3, 1.5177e-23], [3068.3, 1.5211e-23], [3075.4, 1.5245e-23], [3082.5, 1.528e-23], [3089.6, 1.5314e-23], [3096.7, 1.5349e-23], [3103.9, 1.5383e-23], [3111.0, 1.5418e-23], [3118.2, 1.5453e-23], [3125.4, 1.5488e-23], [3132.6, 1.5523e-23], [3139.8, 1.5558e-23], [3147.0, 1.5593e-23], [3154.3, 1.5628e-23], [3161.6, 1.5663e-23], [3168.9, 1.5699e-23], [3176.2, 1.5734e-23], [3183.5, 1.577e-23], [3190.8, 1.5805e-23], [3198.2, 1.5841e-23], [3205.6, 1.5877e-23], [3213.0, 1.5913e-23], [3220.4, 1.5949e-23], [3227.8, 1.5985e-23], [3235.2, 1.6021e-23], [3242.7, 1.6057e-23], [3250.2, 1.6093e-23], [3257.7, 1.613e-23], [3265.2, 1.6166e-23], [3272.7, 1.6203e-23], [3280.3, 1.6239e-23], [3287.8, 1.6276e-23], [3295.4, 1.6313e-23], [3303.0, 1.635e-23], [3310.6, 1.6387e-23], [3318.3, 1.6424e-23], [3325.9, 1.6461e-23], [3333.6, 1.6498e-23], [3341.3, 1.6536e-23], [3349.0, 1.6573e-23], [3356.7, 1.6611e-23], [3364.4, 1.6648e-23], [3372.2, 1.6686e-23], [3380.0, 1.6724e-23], [3387.8, 1.6762e-23], [3395.6, 1.6799e-23], [3403.4, 1.6838e-23], [3411.3, 1.6876e-23], [3419.1, 1.6914e-23], [3427.0, 1.6952e-23], [3434.9, 1.6991e-23], [3442.8, 1.7029e-23], [3450.8, 1.7068e-23], [3458.7, 1.7106e-23], [3466.7, 1.7145e-23], [3474.7, 1.7225e-23], [3482.7, 1.7223e-23], [3490.7, 1.7262e-23], [3498.8, 1.7301e-23], [3506.9, 1.7341e-23], [3514.9, 1.738e-23], [3523.1, 1.7419e-23], [3531.2, 1.7459e-23], [3539.3, 1.7499e-23], [3547.5, 1.7538e-23], [3555.7, 1.7578e-23], [3563.9, 1.7618e-23], [3572.1, 1.7658e-23], [3580.3, 1.7698e-23], [3588.6, 1.7738e-23], [3596.8, 1.7778e-23], [3605.1, 1.7819e-23], [3613.5, 1.7859e-23], [3621.8, 1.79e-23], [3630.1, 1.7941e-23], [3638.5, 1.7981e-23], [3646.9, 1.8022e-23], [3655.3, 1.8063e-23], [3663.7, 1.8104e-23], [3672.2, 1.8145e-23], [3680.7, 1.8187e-23], [3689.1, 1.8228e-23], [3697.7, 1.8269e-23], [3706.2, 1.8311e-23], [3714.7, 1.8353e-23], [3723.3, 1.8394e-23], [3731.9, 1.8436e-23], [3740.5, 1.8478e-23], [3749.1, 1.852e-23], [3757.8, 1.8562e-23], [3766.4, 1.8604e-23], [3775.1, 1.8647e-23], [3783.8, 1.8689e-23], [3792.5, 1.8732e-23], [3801.3, 1.8774e-23], [3810.0, 1.8817e-23], [3818.8, 1.886e-23], [3827.6, 1.8903e-23], [3836.5, 1.8946e-23], [3845.3, 1.8989e-23], [3854.2, 1.9032e-23], [3863.1, 1.9076e-23], [3872.0, 1.9119e-23], [3880.9, 1.9163e-23], [3889.9, 1.9206e-23], [3898.8, 1.925e-23], [3907.8, 1.9294e-23], [3916.8, 1.9338e-23], [3925.9, 1.9382e-23], [3934.9, 1.9426e-23], [3944.0, 1.947e-23], [3953.1, 1.9515e-23], [3962.2, 1.9559e-23], [3971.3, 1.9604e-23], [3980.5, 1.9648e-23], [3989.7, 1.9693e-23], [3998.9, 1.9738e-23], [4008.1, 1.9783e-23], [4017.3, 1.9828e-23], [4026.6, 1.9873e-23], [4035.9, 1.9919e-23], [4045.2, 1.9964e-23], [4054.5, 2.001e-23], [4063.9, 2.0055e-23], [4073.2, 2.0101e-23], [4082.6, 2.0147e-23], [4092.1, 2.0193e-23], [4101.5, 2.0239e-23], [4110.9, 2.0285e-23], [4120.4, 2.0332e-23], [4129.9, 2.0378e-23], [4139.5, 2.0424e-23], [4149.0, 2.0471e-23], [4158.6, 2.0518e-23], [4168.2, 2.0565e-23], [4177.8, 2.0612e-23], [4187.4, 2.0659e-23], [4197.1, 2.0706e-23], [4206.7, 2.0753e-23], [4216.4, 2.0801e-23], [4226.2, 2.0848e-23], [4235.9, 2.0896e-23], [4245.7, 2.0943e-23], [4255.5, 2.0991e-23], [4265.3, 2.1039e-23], [4275.1, 2.1087e-23], [4285.0, 2.1136e-23], [4294.8, 2.1184e-23], [4304.8, 2.1232e-23], [4314.7, 2.1281e-23], [4324.6, 2.1329e-23], [4334.6, 2.1378e-23], [4344.6, 2.1427e-23], [4354.6, 2.1476e-23], [4364.7, 2.1525e-23], [4374.7, 2.1574e-23], [4384.8, 2.1624e-23], [4394.9, 2.1673e-23], [4405.1, 2.1723e-23], [4415.2, 2.1772e-23], [4425.4, 2.1822e-23], [4435.6, 2.1872e-23], [4445.8, 2.1922e-23], [4456.1, 2.1972e-23], [4466.4, 2.2023e-23], [4476.7, 2.2073e-23], [4487.0, 2.2124e-23], [4497.3, 2.2174e-23], [4507.7, 2.2225e-23], [4518.1, 2.2276e-23], [4528.5, 2.2327e-23], [4539.0, 2.2378e-23], [4549.4, 2.2429e-23], [4559.9, 2.248e-23], [4570.4, 2.2532e-23], [4581.0, 2.2584e-23], [4591.5, 2.2635e-23], [4602.1, 2.2687e-23], [4612.7, 2.2739e-23], [4623.4, 2.2791e-23], [4634.0, 2.2843e-23], [4644.7, 2.2896e-23], [4655.4, 2.2948e-23], [4666.2, 2.3001e-23], [4676.9, 2.3053e-23], [4687.7, 2.3106e-23], [4698.5, 2.3159e-23], [4709.4, 2.3212e-23], [4720.2, 2.3265e-23], [4731.1, 2.3319e-23], [4742.0, 2.3372e-23], [4752.9, 2.3426e-23], [4763.9, 2.3479e-23], [4774.9, 2.3533e-23], [4785.9, 2.3587e-23], [4796.9, 2.3641e-23], [4808.0, 2.3695e-23], [4819.1, 2.375e-23], [4830.2, 2.3804e-23], [4841.3, 2.3859e-23], [4852.5, 2.3914e-23], [4863.7, 2.3968e-23], [4874.9, 2.4023e-23], [4886.1, 2.4079e-23], [4897.4, 2.4134e-23], [4908.7, 2.4189e-23], [4920.0, 2.4245e-23], [4931.4, 2.43e-23], [4942.7, 2.4356e-23], [4954.1, 2.4412e-23], [4965.6, 2.4468e-23], [4977.0, 2.4524e-23], [4988.5, 2.458e-23], [5000.0, 2.4637e-23]])
    else:
        raise ValueError('Value '+source+' of internal variable source not recognised')

    f1 = data[:,0]
    S1 = data[:,col]**2
    fstop = [f1[0],f1[-1]]

    # % ---- Interpolate to desired frequencies.  Set noise to infinity outside
    # %      band of frequencies covered by f1.
    # S = inf(size(f)); 
    # k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop


# % -------------------------------------------------------------------------
# %                              Advanced Virgo spectra
# % -------------------------------------------------------------------------

def avirgospectrum(IFO,f):
# % AVIRGOSPECTRUM - Design noise curves for advanced Virgo.
# %
# % Arguments are the same as for the parent function.
# %
# % The spectra are taken from the ligovirgo CVS, 
# % ligovirgo/dac/policy/CoordinatedFollowUpProgram/data/AdV_evolution2.fig
# % These are the official design curves as of June 2012.
# % The AVIRGO-BNS spectrum matches that from
# % https://wwwcascina.virgo.infn.it/advirgo/, corresponding to proposal
# % VIR-0337A-10 and used for ER1 simulated data generation.  

# % The following analytic fit is copied from Manzotti & Dietz,
# % https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=p1200014
# % I have not checked its accuracy, and it is not used in this function.
# %   S.^0.5 = 1.259e-24 * ( ...
# %       + 0.07*exp(-0.142-1.437*x+0.407*x.^2) ...
# %       + 3.10*exp(-0.466-1.043*x-0.548*x.^2) ...
# %       + 0.40*exp(-0.304+2.896*x-0.293*x.^2) ...
# %       + 0.09*exp( 1.466+3.722*x-0.984*x.^2) ...
# %   );
# % where 
# %   x := log(f/300Hz)

# % ---- Recognize detector and compute noise spectrum.
    if IFO.upper()=='AVIRGO-NOSRM':
        col = 1
    elif IFO.upper()=='AVIRGO-ZDLP':
        col = 2
    elif IFO.upper()=='AVIRGO-BNS':
        col = 3
    elif IFO.upper()=='AVIRGO-ZDHP':
        col = 4
    else:
        raise ValueError('Detector '+IFO+' not recognized.')

# % ---- Extract pre-sampled noise spectrum from data.  First column is
# %      frequency, other columns map to configurations as above.
    data = np.array([[10.0, 1.10777e-21, 1.09837e-21, 1.10758e-21, 1.10062e-21], [10.0231, 1.09746e-21, 1.08806e-21, 1.09727e-21, 1.09031e-21], [10.0462, 1.08725e-21, 1.07785e-21, 1.08706e-21, 1.0801e-21], [10.0693, 1.07714e-21, 1.06773e-21, 1.07694e-21, 1.06998e-21], [10.0926, 1.06712e-21, 1.05772e-21, 1.06693e-21, 1.05997e-21], [10.1158, 1.05721e-21, 1.0478e-21, 1.05701e-21, 1.05005e-21], [10.1392, 1.04739e-21, 1.03798e-21, 1.04719e-21, 1.04024e-21], [10.1625, 1.03767e-21, 1.02826e-21, 1.03747e-21, 1.03051e-21], [10.186, 1.02804e-21, 1.01863e-21, 1.02784e-21, 1.02088e-21], [10.2095, 1.0185e-21, 1.0091e-21, 1.0183e-21, 1.01135e-21], [10.233, 1.00906e-21, 9.99654e-22, 1.00886e-21, 1.00191e-21], [10.2566, 9.99711e-22, 9.90303e-22, 9.99508e-22, 9.92554e-22], [10.2803, 9.90451e-22, 9.81042e-22, 9.90246e-22, 9.83294e-22], [10.304, 9.81281e-22, 9.71871e-22, 9.81074e-22, 9.74123e-22], [10.3277, 9.72199e-22, 9.62789e-22, 9.7199e-22, 9.6504e-22], [10.3515, 9.63204e-22, 9.53794e-22, 9.62994e-22, 9.56046e-22], [10.3754, 9.54297e-22, 9.44886e-22, 9.54085e-22, 9.47138e-22], [10.3993, 9.45475e-22, 9.36064e-22, 9.45262e-22, 9.38316e-22], [10.4233, 9.36739e-22, 9.27327e-22, 9.36523e-22, 9.29579e-22], [10.4474, 9.28086e-22, 9.18674e-22, 9.27869e-22, 9.20926e-22], [10.4714, 9.19517e-22, 9.10104e-22, 9.19298e-22, 9.12356e-22], [10.4956, 9.11029e-22, 9.01617e-22, 9.10809e-22, 9.03869e-22], [10.5198, 9.02624e-22, 8.93211e-22, 9.02402e-22, 8.95463e-22], [10.5441, 8.94299e-22, 8.84886e-22, 8.94075e-22, 8.87138e-22], [10.5684, 8.86053e-22, 8.76641e-22, 8.85828e-22, 8.78892e-22], [10.5927, 8.77887e-22, 8.68474e-22, 8.7766e-22, 8.70726e-22], [10.6172, 8.69799e-22, 8.60386e-22, 8.6957e-22, 8.62638e-22], [10.6417, 8.61788e-22, 8.52375e-22, 8.61558e-22, 8.54627e-22], [10.6662, 8.53854e-22, 8.44441e-22, 8.53622e-22, 8.46693e-22], [10.6908, 8.45996e-22, 8.36583e-22, 8.45762e-22, 8.38835e-22], [10.7154, 8.38213e-22, 8.288e-22, 8.37976e-22, 8.31051e-22], [10.7401, 8.30504e-22, 8.21091e-22, 8.30266e-22, 8.23342e-22], [10.7649, 8.22868e-22, 8.13455e-22, 8.22628e-22, 8.15707e-22], [10.7897, 8.15305e-22, 8.05893e-22, 8.15064e-22, 8.08144e-22], [10.8146, 8.07815e-22, 7.98402e-22, 8.07572e-22, 8.00654e-22], [10.8396, 8.00395e-22, 7.90983e-22, 8.0015e-22, 7.93235e-22], [10.8646, 7.93047e-22, 7.83635e-22, 7.928e-22, 7.85886e-22], [10.8896, 7.85768e-22, 7.76356e-22, 7.85519e-22, 7.78607e-22], [10.9147, 7.78558e-22, 7.69147e-22, 7.78308e-22, 7.71398e-22], [10.9399, 7.71417e-22, 7.62006e-22, 7.71165e-22, 7.64257e-22], [10.9651, 7.64344e-22, 7.54933e-22, 7.6409e-22, 7.57184e-22], [10.9904, 7.57338e-22, 7.47928e-22, 7.57082e-22, 7.50178e-22], [11.0157, 7.50398e-22, 7.40988e-22, 7.50141e-22, 7.43239e-22], [11.0412, 7.43524e-22, 7.34115e-22, 7.43265e-22, 7.36366e-22], [11.0666, 7.36716e-22, 7.27307e-22, 7.36455e-22, 7.29558e-22], [11.0921, 7.29972e-22, 7.20564e-22, 7.29709e-22, 7.22814e-22], [11.1177, 7.23292e-22, 7.13884e-22, 7.23027e-22, 7.16134e-22], [11.1433, 7.16675e-22, 7.07268e-22, 7.16408e-22, 7.09518e-22], [11.169, 7.10121e-22, 7.00715e-22, 7.09852e-22, 7.02964e-22], [11.1948, 7.03629e-22, 6.94223e-22, 7.03359e-22, 6.96473e-22], [11.2206, 6.97198e-22, 6.87793e-22, 6.96926e-22, 6.90043e-22], [11.2465, 6.90828e-22, 6.81424e-22, 6.90554e-22, 6.83673e-22], [11.2724, 6.84519e-22, 6.75116e-22, 6.84243e-22, 6.77364e-22], [11.2984, 6.78269e-22, 6.68867e-22, 6.77991e-22, 6.71115e-22], [11.3245, 6.72078e-22, 6.62677e-22, 6.71799e-22, 6.64925e-22], [11.3506, 6.65946e-22, 6.56545e-22, 6.65665e-22, 6.58793e-22], [11.3768, 6.59872e-22, 6.50472e-22, 6.59588e-22, 6.5272e-22], [11.403, 6.53855e-22, 6.44456e-22, 6.5357e-22, 6.46703e-22], [11.4293, 6.47894e-22, 6.38497e-22, 6.47608e-22, 6.40744e-22], [11.4556, 6.41991e-22, 6.32594e-22, 6.41702e-22, 6.34841e-22], [11.4821, 6.36142e-22, 6.26747e-22, 6.35852e-22, 6.28994e-22], [11.5085, 6.3035e-22, 6.20955e-22, 6.30057e-22, 6.23202e-22], [11.5351, 6.24611e-22, 6.15218e-22, 6.24317e-22, 6.17464e-22], [11.5617, 6.18927e-22, 6.09535e-22, 6.18631e-22, 6.11781e-22], [11.5883, 6.13297e-22, 6.03906e-22, 6.12999e-22, 6.06151e-22], [11.6151, 6.07719e-22, 5.9833e-22, 6.07419e-22, 6.00575e-22], [11.6418, 6.02195e-22, 5.92806e-22, 6.01893e-22, 5.95051e-22], [11.6687, 5.96722e-22, 5.87335e-22, 5.96418e-22, 5.89579e-22], [11.6956, 5.91301e-22, 5.81915e-22, 5.90995e-22, 5.84159e-22], [11.7226, 5.85931e-22, 5.76547e-22, 5.85623e-22, 5.7879e-22], [11.7496, 5.80612e-22, 5.71229e-22, 5.80302e-22, 5.73472e-22], [11.7767, 5.75342e-22, 5.65961e-22, 5.75031e-22, 5.68204e-22], [11.8039, 5.70123e-22, 5.60743e-22, 5.69809e-22, 5.62985e-22], [11.8311, 5.64952e-22, 5.55574e-22, 5.64637e-22, 5.57816e-22], [11.8584, 5.59831e-22, 5.50454e-22, 5.59513e-22, 5.52696e-22], [11.8857, 5.54758e-22, 5.45382e-22, 5.54438e-22, 5.47623e-22], [11.9131, 5.49732e-22, 5.40358e-22, 5.49411e-22, 5.42599e-22], [11.9406, 5.44754e-22, 5.35382e-22, 5.44431e-22, 5.37622e-22], [11.9681, 5.39823e-22, 5.30452e-22, 5.39497e-22, 5.32692e-22], [11.9957, 5.34938e-22, 5.25569e-22, 5.34611e-22, 5.27809e-22], [12.0234, 5.30099e-22, 5.20732e-22, 5.2977e-22, 5.22971e-22], [12.0511, 5.25306e-22, 5.15941e-22, 5.24975e-22, 5.18179e-22], [12.0789, 5.20558e-22, 5.11195e-22, 5.20225e-22, 5.13433e-22], [12.1068, 5.15855e-22, 5.06493e-22, 5.1552e-22, 5.08731e-22], [12.1347, 5.11196e-22, 5.01837e-22, 5.10859e-22, 5.04073e-22], [12.1627, 5.06581e-22, 4.97224e-22, 5.06242e-22, 4.9946e-22], [12.1907, 5.02009e-22, 4.92654e-22, 5.01668e-22, 4.9489e-22], [12.2188, 4.97481e-22, 4.88128e-22, 4.97138e-22, 4.90363e-22], [12.247, 4.92995e-22, 4.83644e-22, 4.9265e-22, 4.85879e-22], [12.2752, 4.88552e-22, 4.79203e-22, 4.88205e-22, 4.81437e-22], [12.3035, 4.84151e-22, 4.74804e-22, 4.83802e-22, 4.77038e-22], [12.3319, 4.79791e-22, 4.70446e-22, 4.7944e-22, 4.72679e-22], [12.3603, 4.75472e-22, 4.6613e-22, 4.75119e-22, 4.68362e-22], [12.3888, 4.71194e-22, 4.61854e-22, 4.70839e-22, 4.64086e-22], [12.4174, 4.66957e-22, 4.57619e-22, 4.666e-22, 4.5985e-22], [12.4461, 4.62759e-22, 4.53424e-22, 4.624e-22, 4.55655e-22], [12.4748, 4.58602e-22, 4.49269e-22, 4.58241e-22, 4.51499e-22], [12.5035, 4.54483e-22, 4.45153e-22, 4.5412e-22, 4.47382e-22], [12.5324, 4.50404e-22, 4.41076e-22, 4.50038e-22, 4.43305e-22], [12.5613, 4.46363e-22, 4.37038e-22, 4.45996e-22, 4.39266e-22], [12.5902, 4.4236e-22, 4.33038e-22, 4.41991e-22, 4.35265e-22], [12.6193, 4.38395e-22, 4.29076e-22, 4.38024e-22, 4.31302e-22], [12.6484, 4.34468e-22, 4.25151e-22, 4.34095e-22, 4.27377e-22], [12.6775, 4.30578e-22, 4.21264e-22, 4.30203e-22, 4.23489e-22], [12.7068, 4.26725e-22, 4.17414e-22, 4.26348e-22, 4.19638e-22], [12.7361, 4.22909e-22, 4.136e-22, 4.22529e-22, 4.15824e-22], [12.7654, 4.19128e-22, 4.09823e-22, 4.18747e-22, 4.12046e-22], [12.7949, 4.15384e-22, 4.06081e-22, 4.15e-22, 4.08303e-22], [12.8244, 4.11675e-22, 4.02376e-22, 4.1129e-22, 4.04597e-22], [12.8539, 4.08002e-22, 3.98706e-22, 4.07614e-22, 4.00926e-22], [12.8836, 4.04363e-22, 3.9507e-22, 4.03974e-22, 3.9729e-22], [12.9133, 4.0076e-22, 3.9147e-22, 4.00368e-22, 3.93688e-22], [12.9431, 3.9719e-22, 3.87904e-22, 3.96796e-22, 3.90121e-22], [12.9729, 3.93655e-22, 3.84372e-22, 3.93259e-22, 3.86589e-22], [13.0028, 3.90154e-22, 3.80874e-22, 3.89755e-22, 3.8309e-22], [13.0328, 3.86686e-22, 3.77409e-22, 3.86285e-22, 3.79624e-22], [13.0629, 3.83251e-22, 3.73978e-22, 3.82848e-22, 3.76192e-22], [13.093, 3.79849e-22, 3.7058e-22, 3.79445e-22, 3.72793e-22], [13.1232, 3.7648e-22, 3.67215e-22, 3.76073e-22, 3.69427e-22], [13.1534, 3.73143e-22, 3.63882e-22, 3.72735e-22, 3.66093e-22], [13.1838, 3.69839e-22, 3.60581e-22, 3.69428e-22, 3.62791e-22], [13.2142, 3.66566e-22, 3.57312e-22, 3.66153e-22, 3.59521e-22], [13.2447, 3.63325e-22, 3.54075e-22, 3.6291e-22, 3.56283e-22], [13.2752, 3.60115e-22, 3.50869e-22, 3.59698e-22, 3.53076e-22], [13.3058, 3.56937e-22, 3.47695e-22, 3.56517e-22, 3.49901e-22], [13.3365, 3.53789e-22, 3.44551e-22, 3.53367e-22, 3.46756e-22], [13.3672, 3.50672e-22, 3.41438e-22, 3.50248e-22, 3.43642e-22], [13.3981, 3.47585e-22, 3.38356e-22, 3.47159e-22, 3.40559e-22], [13.429, 3.44528e-22, 3.35304e-22, 3.44101e-22, 3.37505e-22], [13.4599, 3.41502e-22, 3.32282e-22, 3.41072e-22, 3.34482e-22], [13.491, 3.38505e-22, 3.2929e-22, 3.38073e-22, 3.31488e-22], [13.5221, 3.35537e-22, 3.26327e-22, 3.35103e-22, 3.28525e-22], [13.5533, 3.32599e-22, 3.23394e-22, 3.32163e-22, 3.2559e-22], [13.5845, 3.2969e-22, 3.20489e-22, 3.29252e-22, 3.22685e-22], [13.6158, 3.2681e-22, 3.17614e-22, 3.26369e-22, 3.19808e-22], [13.6472, 3.23958e-22, 3.14768e-22, 3.23516e-22, 3.1696e-22], [13.6787, 3.21135e-22, 3.1195e-22, 3.20691e-22, 3.14141e-22], [13.7103, 3.18341e-22, 3.09161e-22, 3.17894e-22, 3.11351e-22], [13.7419, 3.15574e-22, 3.064e-22, 3.15125e-22, 3.08588e-22], [13.7736, 3.12836e-22, 3.03667e-22, 3.12385e-22, 3.05854e-22], [13.8053, 3.10125e-22, 3.00962e-22, 3.09672e-22, 3.03147e-22], [13.8372, 3.07442e-22, 2.98285e-22, 3.06987e-22, 3.00469e-22], [13.8691, 3.04786e-22, 2.95636e-22, 3.04329e-22, 2.97818e-22], [13.9011, 3.02158e-22, 2.93014e-22, 3.01699e-22, 2.95194e-22], [13.9331, 2.99557e-22, 2.90419e-22, 2.99096e-22, 2.92598e-22], [13.9652, 2.96984e-22, 2.87852e-22, 2.9652e-22, 2.90029e-22], [13.9974, 2.94437e-22, 2.85312e-22, 2.93971e-22, 2.87487e-22], [14.0297, 2.91917e-22, 2.82798e-22, 2.91449e-22, 2.84972e-22], [14.0621, 2.89424e-22, 2.80312e-22, 2.88954e-22, 2.82484e-22], [14.0945, 2.86957e-22, 2.77853e-22, 2.86485e-22, 2.80023e-22], [14.127, 2.84518e-22, 2.75421e-22, 2.84043e-22, 2.77589e-22], [14.1596, 2.82104e-22, 2.73015e-22, 2.81628e-22, 2.75181e-22], [14.1922, 2.79718e-22, 2.70636e-22, 2.79239e-22, 2.728e-22], [14.225, 2.77358e-22, 2.68284e-22, 2.76877e-22, 2.70446e-22], [14.2578, 2.75024e-22, 2.65959e-22, 2.74541e-22, 2.68119e-22], [14.2906, 2.72717e-22, 2.6366e-22, 2.72232e-22, 2.65818e-22], [14.3236, 2.70436e-22, 2.61388e-22, 2.6995e-22, 2.63543e-22], [14.3566, 2.68182e-22, 2.59143e-22, 2.67693e-22, 2.61296e-22], [14.3897, 2.65954e-22, 2.56925e-22, 2.65464e-22, 2.59075e-22], [14.4229, 2.63753e-22, 2.54733e-22, 2.63261e-22, 2.56881e-22], [14.4562, 2.61579e-22, 2.52569e-22, 2.61084e-22, 2.54714e-22], [14.4895, 2.59432e-22, 2.50432e-22, 2.58935e-22, 2.52574e-22], [14.5229, 2.57311e-22, 2.48322e-22, 2.56813e-22, 2.50462e-22], [14.5564, 2.55218e-22, 2.46239e-22, 2.54717e-22, 2.48376e-22], [14.59, 2.53152e-22, 2.44185e-22, 2.52649e-22, 2.46319e-22], [14.6236, 2.51113e-22, 2.42158e-22, 2.50609e-22, 2.44289e-22], [14.6573, 2.49103e-22, 2.40159e-22, 2.48596e-22, 2.42287e-22], [14.6911, 2.4712e-22, 2.38189e-22, 2.46612e-22, 2.40314e-22], [14.725, 2.45166e-22, 2.36248e-22, 2.44656e-22, 2.3837e-22], [14.759, 2.43241e-22, 2.34336e-22, 2.42729e-22, 2.36454e-22], [14.793, 2.41345e-22, 2.32455e-22, 2.40831e-22, 2.34569e-22], [14.8271, 2.39479e-22, 2.30603e-22, 2.38963e-22, 2.32714e-22], [14.8613, 2.37643e-22, 2.28783e-22, 2.37126e-22, 2.30889e-22], [14.8956, 2.35838e-22, 2.26994e-22, 2.3532e-22, 2.29096e-22], [14.9299, 2.34065e-22, 2.25237e-22, 2.33545e-22, 2.27336e-22], [14.9644, 2.32325e-22, 2.23514e-22, 2.31803e-22, 2.25608e-22], [14.9989, 2.30618e-22, 2.21825e-22, 2.30094e-22, 2.23914e-22], [15.0335, 2.28945e-22, 2.20171e-22, 2.28419e-22, 2.22255e-22], [15.0681, 2.27307e-22, 2.18554e-22, 2.26781e-22, 2.20633e-22], [15.1029, 2.25707e-22, 2.16974e-22, 2.25178e-22, 2.19048e-22], [15.1377, 2.24144e-22, 2.15433e-22, 2.23615e-22, 2.17501e-22], [15.1726, 2.22621e-22, 2.13933e-22, 2.2209e-22, 2.15996e-22], [15.2076, 2.21139e-22, 2.12476e-22, 2.20608e-22, 2.14532e-22], [15.2427, 2.19701e-22, 2.11063e-22, 2.19168e-22, 2.13113e-22], [15.2778, 2.18308e-22, 2.09697e-22, 2.17774e-22, 2.1174e-22], [15.3131, 2.16964e-22, 2.08381e-22, 2.16429e-22, 2.10417e-22], [15.3484, 2.15669e-22, 2.07117e-22, 2.15134e-22, 2.09145e-22], [15.3838, 2.14429e-22, 2.05909e-22, 2.13893e-22, 2.07929e-22], [15.4192, 2.13246e-22, 2.0476e-22, 2.12709e-22, 2.06772e-22], [15.4548, 2.12124e-22, 2.03675e-22, 2.11587e-22, 2.05677e-22], [15.4904, 2.11069e-22, 2.02657e-22, 2.10531e-22, 2.0465e-22], [15.5261, 2.10084e-22, 2.01714e-22, 2.09546e-22, 2.03696e-22], [15.5619, 2.09175e-22, 2.00849e-22, 2.08638e-22, 2.02821e-22], [15.5978, 2.08351e-22, 2.00072e-22, 2.07813e-22, 2.02031e-22], [15.6338, 2.07617e-22, 1.99389e-22, 2.0708e-22, 2.01336e-22], [15.6699, 2.06983e-22, 1.98809e-22, 2.06447e-22, 2.00743e-22], [15.706, 2.0646e-22, 1.98344e-22, 2.05925e-22, 2.00263e-22], [15.7422, 2.06058e-22, 1.98004e-22, 2.05524e-22, 1.99908e-22], [15.7785, 2.05791e-22, 1.97806e-22, 2.05259e-22, 1.99692e-22], [15.8149, 2.05676e-22, 1.97764e-22, 2.05146e-22, 1.99632e-22], [15.8514, 2.05731e-22, 1.97898e-22, 2.05204e-22, 1.99747e-22], [15.8879, 2.05978e-22, 1.98231e-22, 2.05454e-22, 2.00059e-22], [15.9246, 2.06443e-22, 1.9879e-22, 2.05922e-22, 2.00594e-22], [15.9613, 2.07157e-22, 1.99606e-22, 2.06641e-22, 2.01385e-22], [15.9981, 2.08158e-22, 2.00717e-22, 2.07646e-22, 2.02469e-22], [16.035, 2.09489e-22, 2.0217e-22, 2.08983e-22, 2.03892e-22], [16.072, 2.11206e-22, 2.04019e-22, 2.10706e-22, 2.05709e-22], [16.109, 2.13374e-22, 2.06332e-22, 2.12882e-22, 2.07987e-22], [16.1462, 2.16076e-22, 2.09192e-22, 2.15592e-22, 2.10809e-22], [16.1834, 2.19412e-22, 2.12703e-22, 2.18938e-22, 2.14277e-22], [16.2207, 2.23512e-22, 2.16994e-22, 2.23049e-22, 2.18522e-22], [16.2581, 2.28538e-22, 2.22229e-22, 2.28087e-22, 2.23706e-22], [16.2956, 2.34698e-22, 2.28619e-22, 2.34261e-22, 2.30041e-22], [16.3332, 2.42267e-22, 2.3644e-22, 2.41845e-22, 2.37801e-22], [16.3709, 2.51608e-22, 2.46056e-22, 2.51204e-22, 2.47352e-22], [16.4086, 2.63213e-22, 2.57963e-22, 2.62829e-22, 2.59187e-22], [16.4464, 2.77768e-22, 2.72846e-22, 2.77406e-22, 2.73992e-22], [16.4844, 2.96247e-22, 2.91682e-22, 2.95909e-22, 2.92744e-22], [16.5224, 3.20091e-22, 3.15912e-22, 3.1978e-22, 3.16883e-22], [16.5605, 3.51517e-22, 3.47753e-22, 3.51235e-22, 3.48626e-22], [16.5987, 3.94122e-22, 3.908e-22, 3.93871e-22, 3.91569e-22], [16.6369, 4.54151e-22, 4.51299e-22, 4.53935e-22, 4.51959e-22], [16.6753, 5.43501e-22, 5.41143e-22, 5.43321e-22, 5.41688e-22], [16.7138, 6.87979e-22, 6.86136e-22, 6.87838e-22, 6.86562e-22], [16.7523, 9.56104e-22, 9.54792e-22, 9.56003e-22, 9.55095e-22], [16.7909, 1.60994e-21, 1.60917e-21, 1.60988e-21, 1.60934e-21], [16.8297, 5.42605e-21, 5.42582e-21, 5.42603e-21, 5.42588e-21], [16.8685, 3.82716e-21, 3.82684e-21, 3.82713e-21, 3.82691e-21], [16.9074, 1.41349e-21, 1.41264e-21, 1.41343e-21, 1.41284e-21], [16.9464, 8.69075e-22, 8.677e-22, 8.68967e-22, 8.68016e-22], [16.9854, 6.3052e-22, 6.28641e-22, 6.30371e-22, 6.29074e-22], [17.0246, 4.97637e-22, 4.95278e-22, 4.97449e-22, 4.95821e-22], [17.0639, 4.13567e-22, 4.10752e-22, 4.13342e-22, 4.114e-22], [17.1032, 3.5598e-22, 3.52738e-22, 3.55719e-22, 3.53484e-22], [17.1427, 3.14319e-22, 3.10678e-22, 3.14026e-22, 3.11517e-22], [17.1822, 2.82948e-22, 2.78938e-22, 2.82624e-22, 2.79862e-22], [17.2218, 2.58585e-22, 2.54233e-22, 2.58231e-22, 2.55237e-22], [17.2615, 2.3919e-22, 2.34524e-22, 2.38809e-22, 2.35601e-22], [17.3013, 2.23431e-22, 2.18479e-22, 2.23026e-22, 2.19623e-22], [17.3412, 2.10405e-22, 2.05189e-22, 2.09976e-22, 2.06394e-22], [17.3812, 1.99474e-22, 1.94019e-22, 1.99024e-22, 1.9528e-22], [17.4213, 1.9018e-22, 1.84507e-22, 1.8971e-22, 1.85819e-22], [17.4615, 1.82186e-22, 1.76314e-22, 1.81698e-22, 1.77673e-22], [17.5017, 1.75236e-22, 1.69184e-22, 1.74731e-22, 1.70585e-22], [17.5421, 1.69137e-22, 1.62921e-22, 1.68616e-22, 1.6436e-22], [17.5825, 1.63737e-22, 1.57371e-22, 1.63201e-22, 1.58846e-22], [17.6231, 1.58917e-22, 1.52416e-22, 1.58368e-22, 1.53922e-22], [17.6637, 1.54584e-22, 1.47959e-22, 1.54023e-22, 1.49494e-22], [17.7045, 1.50661e-22, 1.43924e-22, 1.50088e-22, 1.45485e-22], [17.7453, 1.47088e-22, 1.40247e-22, 1.46503e-22, 1.41833e-22], [17.7862, 1.43813e-22, 1.36879e-22, 1.43218e-22, 1.38487e-22], [17.8272, 1.40795e-22, 1.33776e-22, 1.4019e-22, 1.35404e-22], [17.8683, 1.38001e-22, 1.30904e-22, 1.37387e-22, 1.32549e-22], [17.9095, 1.35401e-22, 1.28233e-22, 1.34778e-22, 1.29895e-22], [17.9508, 1.32971e-22, 1.25738e-22, 1.3234e-22, 1.27415e-22], [17.9922, 1.30691e-22, 1.23399e-22, 1.30052e-22, 1.25089e-22], [18.0337, 1.28544e-22, 1.21197e-22, 1.27897e-22, 1.229e-22], [18.0753, 1.26514e-22, 1.19118e-22, 1.25861e-22, 1.20832e-22], [18.117, 1.2459e-22, 1.17148e-22, 1.23929e-22, 1.18873e-22], [18.1588, 1.2276e-22, 1.15276e-22, 1.22092e-22, 1.1701e-22], [18.2006, 1.21014e-22, 1.13492e-22, 1.20341e-22, 1.15235e-22], [18.2426, 1.19345e-22, 1.11788e-22, 1.18665e-22, 1.13539e-22], [18.2847, 1.17745e-22, 1.10157e-22, 1.1706e-22, 1.11914e-22], [18.3268, 1.16208e-22, 1.08591e-22, 1.15517e-22, 1.10355e-22], [18.3691, 1.14729e-22, 1.07085e-22, 1.14033e-22, 1.08855e-22], [18.4115, 1.13302e-22, 1.05634e-22, 1.126e-22, 1.07409e-22], [18.4539, 1.11924e-22, 1.04233e-22, 1.11217e-22, 1.06013e-22], [18.4965, 1.1059e-22, 1.02879e-22, 1.09878e-22, 1.04663e-22], [18.5391, 1.09297e-22, 1.01568e-22, 1.0858e-22, 1.03355e-22], [18.5819, 1.08042e-22, 1.00297e-22, 1.07321e-22, 1.02087e-22], [18.6247, 1.06823e-22, 9.90622e-23, 1.06097e-22, 1.00856e-22], [18.6677, 1.05636e-22, 9.78621e-23, 1.04905e-22, 9.9658e-23], [18.7107, 1.0448e-22, 9.66941e-23, 1.03745e-22, 9.84921e-23], [18.7539, 1.03353e-22, 9.55561e-23, 1.02614e-22, 9.7356e-23], [18.7971, 1.02253e-22, 9.44463e-23, 1.01509e-22, 9.62476e-23], [18.8405, 1.01178e-22, 9.3363e-23, 1.0043e-22, 9.51655e-23], [18.8839, 1.00127e-22, 9.23046e-23, 9.93752e-23, 9.41081e-23], [18.9275, 9.9099e-23, 9.12698e-23, 9.83427e-23, 9.30739e-23], [18.9711, 9.8092e-23, 9.02572e-23, 9.73317e-23, 9.20617e-23], [19.0149, 9.71051e-23, 8.92657e-23, 9.63409e-23, 9.10705e-23], [19.0587, 9.61375e-23, 8.82943e-23, 9.53694e-23, 9.0099e-23], [19.1027, 9.51881e-23, 8.73419e-23, 9.44162e-23, 8.91464e-23], [19.1467, 9.42561e-23, 8.64077e-23, 9.34805e-23, 8.82117e-23], [19.1909, 9.33408e-23, 8.54908e-23, 9.25615e-23, 8.72942e-23], [19.2351, 9.24413e-23, 8.45904e-23, 9.16584e-23, 8.63931e-23], [19.2795, 9.15571e-23, 8.37059e-23, 9.07705e-23, 8.55077e-23], [19.3239, 9.06875e-23, 8.28367e-23, 8.98973e-23, 8.46373e-23], [19.3685, 8.98319e-23, 8.1982e-23, 8.90382e-23, 8.37813e-23], [19.4132, 8.89898e-23, 8.11413e-23, 8.81926e-23, 8.29393e-23], [19.4579, 8.81607e-23, 8.03142e-23, 8.736e-23, 8.21106e-23], [19.5028, 8.73441e-23, 7.95001e-23, 8.654e-23, 8.12948e-23], [19.5478, 8.65396e-23, 7.86986e-23, 8.57321e-23, 8.04914e-23], [19.5928, 8.57468e-23, 7.79092e-23, 8.49359e-23, 7.97001e-23], [19.638, 8.49653e-23, 7.71315e-23, 8.41511e-23, 7.89203e-23], [19.6833, 8.41948e-23, 7.63652e-23, 8.33772e-23, 7.81519e-23], [19.7287, 8.34349e-23, 7.561e-23, 8.2614e-23, 7.73943e-23], [19.7742, 8.26853e-23, 7.48654e-23, 8.18611e-23, 7.66473e-23], [19.8198, 8.19457e-23, 7.41313e-23, 8.11183e-23, 7.59106e-23], [19.8655, 8.12159e-23, 7.34072e-23, 8.03852e-23, 7.51839e-23], [19.9113, 8.04955e-23, 7.26929e-23, 7.96616e-23, 7.44669e-23], [19.9572, 7.97843e-23, 7.19882e-23, 7.89472e-23, 7.37594e-23], [20.0032, 7.90821e-23, 7.12928e-23, 7.82418e-23, 7.30611e-23], [20.0494, 7.83887e-23, 7.06065e-23, 7.75453e-23, 7.23717e-23], [20.0956, 7.77038e-23, 6.99291e-23, 7.68573e-23, 7.16911e-23], [20.1419, 7.70273e-23, 6.92602e-23, 7.61776e-23, 7.10191e-23], [20.1884, 7.6359e-23, 6.85999e-23, 7.55062e-23, 7.03554e-23], [20.2349, 7.56986e-23, 6.79478e-23, 7.48427e-23, 6.96999e-23], [20.2816, 7.5046e-23, 6.73037e-23, 7.4187e-23, 6.90524e-23], [20.3284, 7.44011e-23, 6.66676e-23, 7.3539e-23, 6.84127e-23], [20.3753, 7.37636e-23, 6.60392e-23, 7.28985e-23, 6.77807e-23], [20.4222, 7.31335e-23, 6.54184e-23, 7.22654e-23, 6.71561e-23], [20.4693, 7.25106e-23, 6.4805e-23, 7.16394e-23, 6.65389e-23], [20.5165, 7.18947e-23, 6.41988e-23, 7.10205e-23, 6.5929e-23], [20.5638, 7.12857e-23, 6.35999e-23, 7.04085e-23, 6.5326e-23], [20.6113, 7.06834e-23, 6.30079e-23, 6.98033e-23, 6.473e-23], [20.6588, 7.00879e-23, 6.24228e-23, 6.92047e-23, 6.41408e-23], [20.7064, 6.94989e-23, 6.18445e-23, 6.86127e-23, 6.35583e-23], [20.7542, 6.89163e-23, 6.12728e-23, 6.80272e-23, 6.29824e-23], [20.802, 6.834e-23, 6.07077e-23, 6.7448e-23, 6.24129e-23], [20.85, 6.77699e-23, 6.01489e-23, 6.6875e-23, 6.18498e-23], [20.8981, 6.72059e-23, 5.95965e-23, 6.63081e-23, 6.12929e-23], [20.9463, 6.6648e-23, 5.90503e-23, 6.57472e-23, 6.07421e-23], [20.9946, 6.60959e-23, 5.85101e-23, 6.51922e-23, 6.01973e-23], [21.043, 6.55497e-23, 5.7976e-23, 6.46431e-23, 5.96586e-23], [21.0915, 6.50091e-23, 5.74478e-23, 6.40997e-23, 5.91256e-23], [21.1402, 6.44743e-23, 5.69255e-23, 6.35619e-23, 5.85984e-23], [21.1889, 6.39449e-23, 5.64089e-23, 6.30298e-23, 5.80769e-23], [21.2378, 6.34211e-23, 5.58979e-23, 6.2503e-23, 5.7561e-23], [21.2868, 6.29026e-23, 5.53925e-23, 6.19817e-23, 5.70506e-23], [21.3358, 6.23895e-23, 5.48926e-23, 6.14657e-23, 5.65456e-23], [21.385, 6.18815e-23, 5.43981e-23, 6.0955e-23, 5.60459e-23], [21.4344, 6.13788e-23, 5.3909e-23, 6.04494e-23, 5.55516e-23], [21.4838, 6.08811e-23, 5.34251e-23, 5.9949e-23, 5.50624e-23], [21.5333, 6.03885e-23, 5.29465e-23, 5.94535e-23, 5.45783e-23], [21.583, 5.99008e-23, 5.24729e-23, 5.89631e-23, 5.40994e-23], [21.6327, 5.9418e-23, 5.20044e-23, 5.84775e-23, 5.36254e-23], [21.6826, 5.89401e-23, 5.15409e-23, 5.79967e-23, 5.31563e-23], [21.7326, 5.84668e-23, 5.10824e-23, 5.75208e-23, 5.26921e-23], [21.7828, 5.79983e-23, 5.06287e-23, 5.70495e-23, 5.22327e-23], [21.833, 5.75345e-23, 5.01798e-23, 5.65829e-23, 5.1778e-23], [21.8833, 5.70752e-23, 4.97356e-23, 5.61208e-23, 5.13281e-23], [21.9338, 5.66204e-23, 4.92962e-23, 5.56633e-23, 5.08827e-23], [21.9844, 5.61701e-23, 4.88613e-23, 5.52103e-23, 5.04419e-23], [22.0351, 5.57242e-23, 4.84311e-23, 5.47617e-23, 5.00056e-23], [22.0859, 5.52827e-23, 4.80053e-23, 5.43174e-23, 4.95737e-23], [22.1368, 5.48454e-23, 4.7584e-23, 5.38775e-23, 4.91463e-23], [22.1879, 5.44124e-23, 4.71672e-23, 5.34418e-23, 4.87231e-23], [22.239, 5.39836e-23, 4.67546e-23, 5.30103e-23, 4.83043e-23], [22.2903, 5.3559e-23, 4.63464e-23, 5.2583e-23, 4.78897e-23], [22.3417, 5.31384e-23, 4.59425e-23, 5.21598e-23, 4.74793e-23], [22.3932, 5.2722e-23, 4.55427e-23, 5.17406e-23, 4.70731e-23], [22.4449, 5.23095e-23, 4.51471e-23, 5.13255e-23, 4.66709e-23], [22.4966, 5.19009e-23, 4.47556e-23, 5.09143e-23, 4.62728e-23], [22.5485, 5.14963e-23, 4.43682e-23, 5.0507e-23, 4.58787e-23], [22.6005, 5.10956e-23, 4.39848e-23, 5.01037e-23, 4.54885e-23], [22.6526, 5.06986e-23, 4.36054e-23, 4.97041e-23, 4.51022e-23], [22.7049, 5.03055e-23, 4.32299e-23, 4.93083e-23, 4.47198e-23], [22.7572, 4.99161e-23, 4.28584e-23, 4.89163e-23, 4.43413e-23], [22.8097, 4.95304e-23, 4.24906e-23, 4.8528e-23, 4.39665e-23], [22.8623, 4.91483e-23, 4.21267e-23, 4.81434e-23, 4.35954e-23], [22.915, 4.87699e-23, 4.17665e-23, 4.77623e-23, 4.32281e-23], [22.9679, 4.8395e-23, 4.14101e-23, 4.73849e-23, 4.28644e-23], [23.0208, 4.80236e-23, 4.10573e-23, 4.7011e-23, 4.25043e-23], [23.0739, 4.76558e-23, 4.07082e-23, 4.66406e-23, 4.21478e-23], [23.1271, 4.72914e-23, 4.03627e-23, 4.62736e-23, 4.17948e-23], [23.1804, 4.69304e-23, 4.00208e-23, 4.59101e-23, 4.14454e-23], [23.2339, 4.65729e-23, 3.96824e-23, 4.555e-23, 4.10994e-23], [23.2875, 4.62186e-23, 3.93475e-23, 4.51932e-23, 4.07568e-23], [23.3412, 4.58677e-23, 3.9016e-23, 4.48398e-23, 4.04176e-23], [23.395, 4.552e-23, 3.8688e-23, 4.44896e-23, 4.00818e-23], [23.4489, 4.51756e-23, 3.83634e-23, 4.41427e-23, 3.97493e-23], [23.503, 4.48344e-23, 3.80421e-23, 4.3799e-23, 3.94201e-23], [23.5572, 4.44964e-23, 3.77242e-23, 4.34585e-23, 3.90941e-23], [23.6115, 4.41615e-23, 3.74095e-23, 4.31211e-23, 3.87713e-23], [23.666, 4.38297e-23, 3.70981e-23, 4.27868e-23, 3.84518e-23], [23.7206, 4.3501e-23, 3.67899e-23, 4.24556e-23, 3.81353e-23], [23.7753, 4.31753e-23, 3.64849e-23, 4.21275e-23, 3.7822e-23], [23.8301, 4.28526e-23, 3.6183e-23, 4.18024e-23, 3.75118e-23], [23.885, 4.2533e-23, 3.58843e-23, 4.14803e-23, 3.72046e-23], [23.9401, 4.22162e-23, 3.55887e-23, 4.11611e-23, 3.69005e-23], [23.9953, 4.19024e-23, 3.52962e-23, 4.08449e-23, 3.65994e-23], [24.0507, 4.15914e-23, 3.50066e-23, 4.05315e-23, 3.63012e-23], [24.1061, 4.12834e-23, 3.47201e-23, 4.0221e-23, 3.60059e-23], [24.1617, 4.09781e-23, 3.44366e-23, 3.99134e-23, 3.57136e-23], [24.2174, 4.06757e-23, 3.4156e-23, 3.96086e-23, 3.54241e-23], [24.2733, 4.0376e-23, 3.38783e-23, 3.93065e-23, 3.51374e-23], [24.3292, 4.0079e-23, 3.36035e-23, 3.90072e-23, 3.48536e-23], [24.3854, 3.97848e-23, 3.33316e-23, 3.87106e-23, 3.45726e-23], [24.4416, 3.94933e-23, 3.30625e-23, 3.84167e-23, 3.42943e-23], [24.4979, 3.92044e-23, 3.27962e-23, 3.81255e-23, 3.40188e-23], [24.5544, 3.89181e-23, 3.25327e-23, 3.78369e-23, 3.37459e-23], [24.6111, 3.86345e-23, 3.22719e-23, 3.7551e-23, 3.34758e-23], [24.6678, 3.83534e-23, 3.20139e-23, 3.72676e-23, 3.32083e-23], [24.7247, 3.80749e-23, 3.17586e-23, 3.69868e-23, 3.29434e-23], [24.7817, 3.77989e-23, 3.15059e-23, 3.67085e-23, 3.26811e-23], [24.8389, 3.75254e-23, 3.12559e-23, 3.64327e-23, 3.24214e-23], [24.8961, 3.72544e-23, 3.10085e-23, 3.61595e-23, 3.21642e-23], [24.9536, 3.69859e-23, 3.07637e-23, 3.58887e-23, 3.19096e-23], [25.0111, 3.67197e-23, 3.05215e-23, 3.56203e-23, 3.16575e-23], [25.0688, 3.6456e-23, 3.02818e-23, 3.53543e-23, 3.14078e-23], [25.1266, 3.61947e-23, 3.00447e-23, 3.50908e-23, 3.11606e-23], [25.1845, 3.59357e-23, 2.981e-23, 3.48296e-23, 3.09158e-23], [25.2426, 3.56791e-23, 2.95779e-23, 3.45707e-23, 3.06734e-23], [25.3008, 3.54247e-23, 2.93482e-23, 3.43142e-23, 3.04334e-23], [25.3592, 3.51727e-23, 2.91209e-23, 3.40599e-23, 3.01958e-23], [25.4176, 3.49229e-23, 2.8896e-23, 3.3808e-23, 2.99605e-23], [25.4762, 3.46754e-23, 2.86735e-23, 3.35583e-23, 2.97275e-23], [25.535, 3.443e-23, 2.84534e-23, 3.33108e-23, 2.94967e-23], [25.5939, 3.41869e-23, 2.82356e-23, 3.30655e-23, 2.92683e-23], [25.6529, 3.3946e-23, 2.80202e-23, 3.28224e-23, 2.90421e-23], [25.7121, 3.37072e-23, 2.7807e-23, 3.25815e-23, 2.88181e-23], [25.7713, 3.34705e-23, 2.75961e-23, 3.23428e-23, 2.85963e-23], [25.8308, 3.3236e-23, 2.73875e-23, 3.21061e-23, 2.83767e-23], [25.8903, 3.30035e-23, 2.71811e-23, 3.18716e-23, 2.81593e-23], [25.95, 3.27732e-23, 2.69769e-23, 3.16391e-23, 2.7944e-23], [26.0099, 3.25448e-23, 2.67749e-23, 3.14087e-23, 2.77308e-23], [26.0699, 3.23186e-23, 2.65751e-23, 3.11804e-23, 2.75197e-23], [26.13, 3.20943e-23, 2.63774e-23, 3.09541e-23, 2.73107e-23], [26.1902, 3.1872e-23, 2.61819e-23, 3.07297e-23, 2.71037e-23], [26.2506, 3.16517e-23, 2.59884e-23, 3.05074e-23, 2.68988e-23], [26.3112, 3.14334e-23, 2.57971e-23, 3.02871e-23, 2.66959e-23], [26.3718, 3.12169e-23, 2.56078e-23, 3.00686e-23, 2.6495e-23], [26.4326, 3.10024e-23, 2.54206e-23, 2.98521e-23, 2.6296e-23], [26.4936, 3.07898e-23, 2.52354e-23, 2.96376e-23, 2.60991e-23], [26.5547, 3.05791e-23, 2.50523e-23, 2.94249e-23, 2.59041e-23], [26.6159, 3.03703e-23, 2.48711e-23, 2.92141e-23, 2.57109e-23], [26.6773, 3.01633e-23, 2.46919e-23, 2.90051e-23, 2.55198e-23], [26.7388, 2.99581e-23, 2.45147e-23, 2.8798e-23, 2.53304e-23], [26.8005, 2.97547e-23, 2.43394e-23, 2.85927e-23, 2.5143e-23], [26.8623, 2.95531e-23, 2.41661e-23, 2.83892e-23, 2.49574e-23], [26.9242, 2.93533e-23, 2.39946e-23, 2.81875e-23, 2.47736e-23], [26.9863, 2.91553e-23, 2.3825e-23, 2.79876e-23, 2.45917e-23], [27.0486, 2.89589e-23, 2.36573e-23, 2.77894e-23, 2.44116e-23], [27.1109, 2.87644e-23, 2.34915e-23, 2.75929e-23, 2.42332e-23], [27.1734, 2.85715e-23, 2.33275e-23, 2.73982e-23, 2.40566e-23], [27.2361, 2.83803e-23, 2.31653e-23, 2.72052e-23, 2.38817e-23], [27.2989, 2.81908e-23, 2.30049e-23, 2.70138e-23, 2.37086e-23], [27.3619, 2.80029e-23, 2.28463e-23, 2.68242e-23, 2.35372e-23], [27.425, 2.78167e-23, 2.26895e-23, 2.66362e-23, 2.33675e-23], [27.4882, 2.76321e-23, 2.25344e-23, 2.64498e-23, 2.31995e-23], [27.5516, 2.74492e-23, 2.2381e-23, 2.6265e-23, 2.30331e-23], [27.6151, 2.72678e-23, 2.22294e-23, 2.60819e-23, 2.28684e-23], [27.6788, 2.7088e-23, 2.20795e-23, 2.59004e-23, 2.27053e-23], [27.7426, 2.69098e-23, 2.19312e-23, 2.57204e-23, 2.25439e-23], [27.8066, 2.67331e-23, 2.17846e-23, 2.5542e-23, 2.2384e-23], [27.8707, 2.6558e-23, 2.16397e-23, 2.53652e-23, 2.22257e-23], [27.935, 2.63843e-23, 2.14965e-23, 2.51898e-23, 2.2069e-23], [27.9994, 2.62122e-23, 2.13548e-23, 2.5016e-23, 2.19139e-23], [28.064, 2.60416e-23, 2.12148e-23, 2.48438e-23, 2.17603e-23], [28.1287, 2.58725e-23, 2.10763e-23, 2.4673e-23, 2.16082e-23], [28.1936, 2.57048e-23, 2.09394e-23, 2.45037e-23, 2.14577e-23], [28.2586, 2.55386e-23, 2.08041e-23, 2.43358e-23, 2.13086e-23], [28.3237, 2.53738e-23, 2.06704e-23, 2.41694e-23, 2.1161e-23], [28.3891, 2.52105e-23, 2.05381e-23, 2.40044e-23, 2.10149e-23], [28.4545, 2.50485e-23, 2.04074e-23, 2.38409e-23, 2.08703e-23], [28.5201, 2.4888e-23, 2.02782e-23, 2.36788e-23, 2.07271e-23], [28.5859, 2.47288e-23, 2.01505e-23, 2.3518e-23, 2.05853e-23], [28.6518, 2.4571e-23, 2.00243e-23, 2.33587e-23, 2.0445e-23], [28.7179, 2.44146e-23, 1.98995e-23, 2.32007e-23, 2.0306e-23], [28.7841, 2.42595e-23, 1.97761e-23, 2.30441e-23, 2.01685e-23], [28.8505, 2.41057e-23, 1.96542e-23, 2.28888e-23, 2.00323e-23], [28.917, 2.39533e-23, 1.95338e-23, 2.27349e-23, 1.98975e-23], [28.9837, 2.38022e-23, 1.94147e-23, 2.25823e-23, 1.9764e-23], [29.0506, 2.36524e-23, 1.9297e-23, 2.2431e-23, 1.96318e-23], [29.1175, 2.35038e-23, 1.91807e-23, 2.2281e-23, 1.9501e-23], [29.1847, 2.33565e-23, 1.90658e-23, 2.21323e-23, 1.93715e-23], [29.252, 2.32105e-23, 1.89522e-23, 2.19848e-23, 1.92433e-23], [29.3194, 2.30658e-23, 1.88399e-23, 2.18387e-23, 1.91164e-23], [29.3871, 2.29222e-23, 1.8729e-23, 2.16937e-23, 1.89907e-23], [29.4548, 2.27799e-23, 1.86194e-23, 2.155e-23, 1.88663e-23], [29.5227, 2.26389e-23, 1.85111e-23, 2.14076e-23, 1.87432e-23], [29.5908, 2.2499e-23, 1.84041e-23, 2.12663e-23, 1.86213e-23], [29.6591, 2.23603e-23, 1.82983e-23, 2.11263e-23, 1.85006e-23], [29.7275, 2.22228e-23, 1.81938e-23, 2.09875e-23, 1.83812e-23], [29.796, 2.20864e-23, 1.80906e-23, 2.08498e-23, 1.82629e-23], [29.8647, 2.19512e-23, 1.79886e-23, 2.07133e-23, 1.81458e-23], [29.9336, 2.18172e-23, 1.78878e-23, 2.0578e-23, 1.803e-23], [30.0026, 2.16843e-23, 1.77882e-23, 2.04438e-23, 1.79152e-23], [30.0718, 2.15525e-23, 1.76899e-23, 2.03108e-23, 1.78017e-23], [30.1411, 2.14218e-23, 1.75927e-23, 2.01789e-23, 1.76892e-23], [30.2106, 2.12923e-23, 1.74967e-23, 2.00481e-23, 1.7578e-23], [30.2803, 2.11638e-23, 1.74018e-23, 1.99184e-23, 1.74678e-23], [30.3501, 2.10365e-23, 1.73081e-23, 1.97898e-23, 1.73587e-23], [30.4201, 2.09102e-23, 1.72156e-23, 1.96623e-23, 1.72508e-23], [30.4903, 2.07849e-23, 1.71242e-23, 1.95359e-23, 1.71439e-23], [30.5606, 2.06608e-23, 1.70338e-23, 1.94106e-23, 1.70381e-23], [30.6311, 2.05376e-23, 1.69446e-23, 1.92863e-23, 1.69334e-23], [30.7017, 2.04155e-23, 1.68565e-23, 1.91631e-23, 1.68298e-23], [30.7725, 2.02945e-23, 1.67695e-23, 1.90409e-23, 1.67272e-23], [30.8435, 2.01744e-23, 1.66835e-23, 1.89198e-23, 1.66256e-23], [30.9146, 2.00554e-23, 1.65986e-23, 1.87996e-23, 1.65251e-23], [30.9859, 1.99373e-23, 1.65148e-23, 1.86805e-23, 1.64255e-23], [31.0573, 1.98203e-23, 1.64319e-23, 1.85624e-23, 1.6327e-23], [31.1289, 1.97042e-23, 1.63501e-23, 1.84453e-23, 1.62295e-23], [31.2007, 1.95891e-23, 1.62694e-23, 1.83292e-23, 1.6133e-23], [31.2727, 1.9475e-23, 1.61896e-23, 1.8214e-23, 1.60375e-23], [31.3448, 1.93618e-23, 1.61108e-23, 1.80999e-23, 1.59429e-23], [31.4171, 1.92495e-23, 1.6033e-23, 1.79866e-23, 1.58493e-23], [31.4895, 1.91382e-23, 1.59562e-23, 1.78744e-23, 1.57566e-23], [31.5621, 1.90279e-23, 1.58803e-23, 1.77631e-23, 1.56649e-23], [31.6349, 1.89184e-23, 1.58054e-23, 1.76527e-23, 1.55741e-23], [31.7079, 1.88099e-23, 1.57314e-23, 1.75432e-23, 1.54842e-23], [31.781, 1.87022e-23, 1.56583e-23, 1.74347e-23, 1.53953e-23], [31.8543, 1.85955e-23, 1.55862e-23, 1.7327e-23, 1.53072e-23], [31.9277, 1.84896e-23, 1.5515e-23, 1.72203e-23, 1.522e-23], [32.0014, 1.83846e-23, 1.54447e-23, 1.71145e-23, 1.51338e-23], [32.0752, 1.82805e-23, 1.53753e-23, 1.70095e-23, 1.50484e-23], [32.1491, 1.81773e-23, 1.53067e-23, 1.69054e-23, 1.49639e-23], [32.2233, 1.80749e-23, 1.5239e-23, 1.68022e-23, 1.48802e-23], [32.2976, 1.79733e-23, 1.51722e-23, 1.66999e-23, 1.47974e-23], [32.372, 1.78726e-23, 1.51062e-23, 1.65984e-23, 1.47154e-23], [32.4467, 1.77728e-23, 1.50411e-23, 1.64978e-23, 1.46343e-23], [32.5215, 1.76737e-23, 1.49768e-23, 1.6398e-23, 1.4554e-23], [32.5965, 1.75755e-23, 1.49134e-23, 1.6299e-23, 1.44745e-23], [32.6717, 1.74781e-23, 1.48507e-23, 1.62009e-23, 1.43958e-23], [32.747, 1.73815e-23, 1.47888e-23, 1.61035e-23, 1.43179e-23], [32.8225, 1.72856e-23, 1.47278e-23, 1.6007e-23, 1.42409e-23], [32.8982, 1.71906e-23, 1.46675e-23, 1.59113e-23, 1.41646e-23], [32.9741, 1.70964e-23, 1.4608e-23, 1.58164e-23, 1.4089e-23], [33.0501, 1.70029e-23, 1.45493e-23, 1.57223e-23, 1.40143e-23], [33.1263, 1.69102e-23, 1.44913e-23, 1.5629e-23, 1.39403e-23], [33.2027, 1.68182e-23, 1.44341e-23, 1.55364e-23, 1.38671e-23], [33.2793, 1.6727e-23, 1.43776e-23, 1.54446e-23, 1.37946e-23], [33.356, 1.66366e-23, 1.43218e-23, 1.53536e-23, 1.37228e-23], [33.433, 1.65469e-23, 1.42668e-23, 1.52633e-23, 1.36518e-23], [33.51, 1.64579e-23, 1.42125e-23, 1.51738e-23, 1.35815e-23], [33.5873, 1.63697e-23, 1.41589e-23, 1.5085e-23, 1.3512e-23], [33.6648, 1.62821e-23, 1.4106e-23, 1.4997e-23, 1.34431e-23], [33.7424, 1.61953e-23, 1.40537e-23, 1.49097e-23, 1.33749e-23], [33.8202, 1.61092e-23, 1.40022e-23, 1.48231e-23, 1.33075e-23], [33.8982, 1.60238e-23, 1.39513e-23, 1.47372e-23, 1.32407e-23], [33.9764, 1.59391e-23, 1.39011e-23, 1.4652e-23, 1.31746e-23], [34.0547, 1.58551e-23, 1.38516e-23, 1.45676e-23, 1.31092e-23], [34.1333, 1.57718e-23, 1.38026e-23, 1.44838e-23, 1.30444e-23], [34.212, 1.56891e-23, 1.37544e-23, 1.44008e-23, 1.29803e-23], [34.2909, 1.56072e-23, 1.37068e-23, 1.43184e-23, 1.29168e-23], [34.3699, 1.55258e-23, 1.36597e-23, 1.42367e-23, 1.2854e-23], [34.4492, 1.54452e-23, 1.36134e-23, 1.41557e-23, 1.27919e-23], [34.5286, 1.53652e-23, 1.35676e-23, 1.40753e-23, 1.27304e-23], [34.6083, 1.52858e-23, 1.35224e-23, 1.39956e-23, 1.26695e-23], [34.6881, 1.52071e-23, 1.34778e-23, 1.39166e-23, 1.26092e-23], [34.7681, 1.51291e-23, 1.34338e-23, 1.38383e-23, 1.25495e-23], [34.8482, 1.50516e-23, 1.33904e-23, 1.37605e-23, 1.24904e-23], [34.9286, 1.49748e-23, 1.33475e-23, 1.36834e-23, 1.2432e-23], [35.0091, 1.48986e-23, 1.33053e-23, 1.3607e-23, 1.23741e-23], [35.0899, 1.48231e-23, 1.32635e-23, 1.35312e-23, 1.23168e-23], [35.1708, 1.47481e-23, 1.32223e-23, 1.3456e-23, 1.22601e-23], [35.2519, 1.46737e-23, 1.31817e-23, 1.33814e-23, 1.2204e-23], [35.3332, 1.46e-23, 1.31416e-23, 1.33075e-23, 1.21485e-23], [35.4147, 1.45268e-23, 1.3102e-23, 1.32341e-23, 1.20935e-23], [35.4963, 1.44542e-23, 1.3063e-23, 1.31614e-23, 1.20391e-23], [35.5782, 1.43822e-23, 1.30245e-23, 1.30892e-23, 1.19852e-23], [35.6602, 1.43108e-23, 1.29864e-23, 1.30177e-23, 1.19319e-23], [35.7425, 1.424e-23, 1.29489e-23, 1.29467e-23, 1.18791e-23], [35.8249, 1.41697e-23, 1.29119e-23, 1.28764e-23, 1.18268e-23], [35.9075, 1.41e-23, 1.28754e-23, 1.28066e-23, 1.17751e-23], [35.9903, 1.40308e-23, 1.28393e-23, 1.27374e-23, 1.17239e-23], [36.0733, 1.39622e-23, 1.28037e-23, 1.26687e-23, 1.16732e-23], [36.1565, 1.38942e-23, 1.27686e-23, 1.26006e-23, 1.1623e-23], [36.2398, 1.38267e-23, 1.2734e-23, 1.25331e-23, 1.15733e-23], [36.3234, 1.37597e-23, 1.26998e-23, 1.24661e-23, 1.15242e-23], [36.4072, 1.36933e-23, 1.26661e-23, 1.23997e-23, 1.14755e-23], [36.4911, 1.36274e-23, 1.26328e-23, 1.23339e-23, 1.14273e-23], [36.5753, 1.35621e-23, 1.25999e-23, 1.22685e-23, 1.13796e-23], [36.6596, 1.34972e-23, 1.25675e-23, 1.22037e-23, 1.13324e-23], [36.7442, 1.34329e-23, 1.25355e-23, 1.21395e-23, 1.12857e-23], [36.8289, 1.33691e-23, 1.2504e-23, 1.20758e-23, 1.12394e-23], [36.9138, 1.33058e-23, 1.24728e-23, 1.20126e-23, 1.11936e-23], [36.999, 1.3243e-23, 1.24421e-23, 1.19499e-23, 1.11482e-23], [37.0843, 1.31807e-23, 1.24117e-23, 1.18877e-23, 1.11033e-23], [37.1698, 1.31189e-23, 1.23818e-23, 1.18261e-23, 1.10589e-23], [37.2555, 1.30576e-23, 1.23523e-23, 1.17649e-23, 1.10149e-23], [37.3414, 1.29968e-23, 1.23231e-23, 1.17043e-23, 1.09713e-23], [37.4275, 1.29365e-23, 1.22943e-23, 1.16441e-23, 1.09282e-23], [37.5138, 1.28766e-23, 1.2266e-23, 1.15845e-23, 1.08855e-23], [37.6003, 1.28172e-23, 1.22379e-23, 1.15253e-23, 1.08432e-23], [37.687, 1.27583e-23, 1.22103e-23, 1.14667e-23, 1.08014e-23], [37.7739, 1.26999e-23, 1.2183e-23, 1.14085e-23, 1.076e-23], [37.8611, 1.26419e-23, 1.21561e-23, 1.13508e-23, 1.07189e-23], [37.9484, 1.25844e-23, 1.21295e-23, 1.12935e-23, 1.06783e-23], [38.0359, 1.25274e-23, 1.21033e-23, 1.12368e-23, 1.06381e-23], [38.1236, 1.24707e-23, 1.20774e-23, 1.11805e-23, 1.05983e-23], [38.2115, 1.24146e-23, 1.20518e-23, 1.11246e-23, 1.05589e-23], [38.2996, 1.23589e-23, 1.20266e-23, 1.10693e-23, 1.05198e-23], [38.3879, 1.23036e-23, 1.20017e-23, 1.10144e-23, 1.04812e-23], [38.4765, 1.22488e-23, 1.19771e-23, 1.09599e-23, 1.04429e-23], [38.5652, 1.21944e-23, 1.19529e-23, 1.09059e-23, 1.0405e-23], [38.6541, 1.21404e-23, 1.19289e-23, 1.08523e-23, 1.03675e-23], [38.7433, 1.20869e-23, 1.19053e-23, 1.07992e-23, 1.03304e-23], [38.8326, 1.20338e-23, 1.1882e-23, 1.07465e-23, 1.02936e-23], [38.9221, 1.19811e-23, 1.1859e-23, 1.06943e-23, 1.02572e-23], [39.0119, 1.19288e-23, 1.18363e-23, 1.06424e-23, 1.02211e-23], [39.1019, 1.18769e-23, 1.18138e-23, 1.0591e-23, 1.01854e-23], [39.192, 1.18255e-23, 1.17917e-23, 1.05401e-23, 1.015e-23], [39.2824, 1.17745e-23, 1.17698e-23, 1.04895e-23, 1.0115e-23], [39.373, 1.17238e-23, 1.17483e-23, 1.04394e-23, 1.00803e-23], [39.4638, 1.16736e-23, 1.1727e-23, 1.03897e-23, 1.0046e-23], [39.5548, 1.16237e-23, 1.17059e-23, 1.03404e-23, 1.0012e-23], [39.646, 1.15743e-23, 1.16852e-23, 1.02915e-23, 9.97827e-24], [39.7374, 1.15252e-23, 1.16647e-23, 1.0243e-23, 9.94491e-24], [39.8291, 1.14765e-23, 1.16444e-23, 1.01949e-23, 9.91188e-24], [39.9209, 1.14282e-23, 1.16245e-23, 1.01472e-23, 9.87916e-24], [40.013, 1.13803e-23, 1.16047e-23, 1.00999e-23, 9.84676e-24], [40.1052, 1.13328e-23, 1.15853e-23, 1.0053e-23, 9.81467e-24], [40.1977, 1.12856e-23, 1.1566e-23, 1.00065e-23, 9.78288e-24], [40.2904, 1.12389e-23, 1.1547e-23, 9.96032e-24, 9.7514e-24], [40.3833, 1.11924e-23, 1.15283e-23, 9.91457e-24, 9.72022e-24], [40.4764, 1.11464e-23, 1.15098e-23, 9.8692e-24, 9.68934e-24], [40.5698, 1.11007e-23, 1.14915e-23, 9.8242e-24, 9.65875e-24], [40.6633, 1.10554e-23, 1.14734e-23, 9.77958e-24, 9.62845e-24], [40.7571, 1.10104e-23, 1.14556e-23, 9.73532e-24, 9.59844e-24], [40.8511, 1.09658e-23, 1.1438e-23, 9.69144e-24, 9.56872e-24], [40.9453, 1.09215e-23, 1.14206e-23, 9.64792e-24, 9.53927e-24], [41.0397, 1.08776e-23, 1.14034e-23, 9.60476e-24, 9.5101e-24], [41.1344, 1.0834e-23, 1.13864e-23, 9.56196e-24, 9.48121e-24], [41.2292, 1.07908e-23, 1.13697e-23, 9.51952e-24, 9.45259e-24], [41.3243, 1.07479e-23, 1.13531e-23, 9.47742e-24, 9.42423e-24], [41.4196, 1.07054e-23, 1.13368e-23, 9.43569e-24, 9.39615e-24], [41.5151, 1.06631e-23, 1.13206e-23, 9.39429e-24, 9.36832e-24], [41.6108, 1.06213e-23, 1.13047e-23, 9.35325e-24, 9.34075e-24], [41.7068, 1.05797e-23, 1.12889e-23, 9.31254e-24, 9.31344e-24], [41.803, 1.05385e-23, 1.12734e-23, 9.27218e-24, 9.28639e-24], [41.8994, 1.04976e-23, 1.1258e-23, 9.23215e-24, 9.25958e-24], [41.996, 1.0457e-23, 1.12428e-23, 9.19246e-24, 9.23302e-24], [42.0928, 1.04167e-23, 1.12278e-23, 9.15309e-24, 9.20671e-24], [42.1899, 1.03768e-23, 1.12129e-23, 9.11406e-24, 9.18063e-24], [42.2872, 1.03372e-23, 1.11983e-23, 9.07536e-24, 9.1548e-24], [42.3847, 1.02979e-23, 1.11838e-23, 9.03698e-24, 9.1292e-24], [42.4824, 1.02588e-23, 1.11695e-23, 8.99892e-24, 9.10384e-24], [42.5804, 1.02201e-23, 1.11553e-23, 8.96117e-24, 9.07871e-24], [42.6786, 1.01818e-23, 1.11414e-23, 8.92375e-24, 9.05381e-24], [42.777, 1.01437e-23, 1.11276e-23, 8.88664e-24, 9.02913e-24], [42.8757, 1.01059e-23, 1.11139e-23, 8.84984e-24, 9.00468e-24], [42.9745, 1.00684e-23, 1.11004e-23, 8.81335e-24, 8.98044e-24], [43.0736, 1.00312e-23, 1.10871e-23, 8.77717e-24, 8.95643e-24], [43.173, 9.99426e-24, 1.10739e-23, 8.74129e-24, 8.93263e-24], [43.2725, 9.95765e-24, 1.10609e-23, 8.70572e-24, 8.90905e-24], [43.3723, 9.92132e-24, 1.1048e-23, 8.67044e-24, 8.88567e-24], [43.4723, 9.88528e-24, 1.10353e-23, 8.63546e-24, 8.86251e-24], [43.5726, 9.84952e-24, 1.10227e-23, 8.60078e-24, 8.83955e-24], [43.673, 9.81404e-24, 1.10103e-23, 8.56639e-24, 8.81679e-24], [43.7737, 9.77884e-24, 1.0998e-23, 8.53229e-24, 8.79423e-24], [43.8747, 9.74393e-24, 1.09858e-23, 8.49847e-24, 8.77188e-24], [43.9759, 9.70928e-24, 1.09738e-23, 8.46495e-24, 8.74972e-24], [44.0773, 9.67491e-24, 1.09619e-23, 8.43171e-24, 8.72775e-24], [44.1789, 9.64081e-24, 1.09501e-23, 8.39875e-24, 8.70598e-24], [44.2808, 9.60698e-24, 1.09385e-23, 8.36606e-24, 8.6844e-24], [44.3829, 9.57342e-24, 1.0927e-23, 8.33366e-24, 8.66301e-24], [44.4853, 9.54012e-24, 1.09156e-23, 8.30153e-24, 8.6418e-24], [44.5878, 9.50709e-24, 1.09044e-23, 8.26968e-24, 8.62077e-24], [44.6907, 9.47431e-24, 1.08933e-23, 8.23809e-24, 8.59993e-24], [44.7937, 9.44179e-24, 1.08823e-23, 8.20677e-24, 8.57927e-24], [44.897, 9.40954e-24, 1.08714e-23, 8.17572e-24, 8.55878e-24], [45.0005, 9.37753e-24, 1.08606e-23, 8.14494e-24, 8.53847e-24], [45.1043, 9.34578e-24, 1.085e-23, 8.11442e-24, 8.51833e-24], [45.2083, 9.31428e-24, 1.08395e-23, 8.08416e-24, 8.49837e-24], [45.3126, 9.28303e-24, 1.0829e-23, 8.05415e-24, 8.47857e-24], [45.4171, 9.25203e-24, 1.08187e-23, 8.02441e-24, 8.45894e-24], [45.5218, 9.22127e-24, 1.08085e-23, 7.99491e-24, 8.43948e-24], [45.6268, 9.19076e-24, 1.07984e-23, 7.96567e-24, 8.42018e-24], [45.732, 9.16048e-24, 1.07885e-23, 7.93668e-24, 8.40105e-24], [45.8374, 9.13045e-24, 1.07786e-23, 7.90794e-24, 8.38207e-24], [45.9431, 9.10066e-24, 1.07688e-23, 7.87945e-24, 8.36325e-24], [46.0491, 9.0711e-24, 1.07591e-23, 7.8512e-24, 8.34459e-24], [46.1553, 9.04177e-24, 1.07495e-23, 7.82319e-24, 8.32608e-24], [46.2617, 9.01268e-24, 1.07401e-23, 7.79543e-24, 8.30773e-24], [46.3684, 8.98382e-24, 1.07307e-23, 7.7679e-24, 8.28952e-24], [46.4753, 8.95519e-24, 1.07214e-23, 7.74061e-24, 8.27147e-24], [46.5825, 8.92679e-24, 1.07122e-23, 7.71356e-24, 8.25356e-24], [46.6899, 8.89861e-24, 1.07031e-23, 7.68674e-24, 8.2358e-24], [46.7976, 8.87065e-24, 1.06941e-23, 7.66015e-24, 8.21819e-24], [46.9055, 8.84292e-24, 1.06852e-23, 7.63379e-24, 8.20071e-24], [47.0137, 8.81541e-24, 1.06763e-23, 7.60766e-24, 8.18338e-24], [47.1221, 8.78812e-24, 1.06676e-23, 7.58175e-24, 8.16619e-24], [47.2307, 8.76104e-24, 1.06589e-23, 7.55607e-24, 8.14914e-24], [47.3397, 8.73419e-24, 1.06504e-23, 7.53062e-24, 8.13222e-24], [47.4488, 8.70754e-24, 1.06419e-23, 7.50538e-24, 8.11544e-24], [47.5582, 8.68111e-24, 1.06335e-23, 7.48036e-24, 8.09879e-24], [47.6679, 8.65489e-24, 1.06251e-23, 7.45556e-24, 8.08227e-24], [47.7778, 8.62888e-24, 1.06169e-23, 7.43098e-24, 8.06588e-24], [47.888, 8.60307e-24, 1.06087e-23, 7.40661e-24, 8.04963e-24], [47.9984, 8.57747e-24, 1.06006e-23, 7.38245e-24, 8.0335e-24], [48.1091, 8.55208e-24, 1.05926e-23, 7.3585e-24, 8.01749e-24], [48.2201, 8.52689e-24, 1.05847e-23, 7.33477e-24, 8.00161e-24], [48.3313, 8.5019e-24, 1.05768e-23, 7.31124e-24, 7.98586e-24], [48.4427, 8.47712e-24, 1.0569e-23, 7.28791e-24, 7.97022e-24], [48.5544, 8.45253e-24, 1.05613e-23, 7.26479e-24, 7.95471e-24], [48.6664, 8.42814e-24, 1.05536e-23, 7.24188e-24, 7.93931e-24], [48.7786, 8.40394e-24, 1.05461e-23, 7.21916e-24, 7.92404e-24], [48.8911, 8.37994e-24, 1.05385e-23, 7.19664e-24, 7.90888e-24], [49.0038, 8.35613e-24, 1.05311e-23, 7.17432e-24, 7.89383e-24], [49.1168, 8.33251e-24, 1.05237e-23, 7.1522e-24, 7.8789e-24], [49.2301, 8.30909e-24, 1.05164e-23, 7.13027e-24, 7.86409e-24], [49.3436, 8.28585e-24, 1.05092e-23, 7.10853e-24, 7.84938e-24], [49.4574, 8.2628e-24, 1.0502e-23, 7.08699e-24, 7.83478e-24], [49.5715, 8.23993e-24, 1.04949e-23, 7.06564e-24, 7.8203e-24], [49.6858, 8.21725e-24, 1.04878e-23, 7.04447e-24, 7.80592e-24], [49.8004, 8.19475e-24, 1.04808e-23, 7.02349e-24, 7.79164e-24], [49.9152, 8.17244e-24, 1.04739e-23, 7.0027e-24, 7.77748e-24], [50.0303, 8.1503e-24, 1.0467e-23, 6.98209e-24, 7.76341e-24], [50.1457, 8.12834e-24, 1.04602e-23, 6.96166e-24, 7.74945e-24], [50.2613, 8.10657e-24, 1.04534e-23, 6.94142e-24, 7.7356e-24], [50.3772, 8.08496e-24, 1.04467e-23, 6.92135e-24, 7.72184e-24], [50.4934, 8.06354e-24, 1.04401e-23, 6.90146e-24, 7.70818e-24], [50.6098, 8.04228e-24, 1.04335e-23, 6.88175e-24, 7.69462e-24], [50.7265, 8.0212e-24, 1.04269e-23, 6.86221e-24, 7.68116e-24], [50.8435, 8.0003e-24, 1.04204e-23, 6.84285e-24, 7.6678e-24], [50.9607, 7.97956e-24, 1.0414e-23, 6.82366e-24, 7.65453e-24], [51.0783, 7.95899e-24, 1.04076e-23, 6.80464e-24, 7.64136e-24], [51.196, 7.93858e-24, 1.04013e-23, 6.78579e-24, 7.62827e-24], [51.3141, 7.91835e-24, 1.0395e-23, 6.76711e-24, 7.61529e-24], [51.4324, 7.89828e-24, 1.03888e-23, 6.74859e-24, 7.60239e-24], [51.551, 7.87837e-24, 1.03826e-23, 6.73024e-24, 7.58958e-24], [51.6699, 7.85863e-24, 1.03765e-23, 6.71206e-24, 7.57687e-24], [51.7891, 7.83904e-24, 1.03704e-23, 6.69404e-24, 7.56424e-24], [51.9085, 7.81962e-24, 1.03644e-23, 6.67617e-24, 7.5517e-24], [52.0282, 7.80036e-24, 1.03584e-23, 6.65847e-24, 7.53924e-24], [52.1482, 7.78126e-24, 1.03524e-23, 6.64093e-24, 7.52687e-24], [52.2684, 7.76231e-24, 1.03466e-23, 6.62355e-24, 7.51459e-24], [52.389, 7.74352e-24, 1.03407e-23, 6.60632e-24, 7.50239e-24], [52.5098, 7.72488e-24, 1.03349e-23, 6.58924e-24, 7.49027e-24], [52.6309, 7.7064e-24, 1.03291e-23, 6.57232e-24, 7.47824e-24], [52.7522, 7.68807e-24, 1.03234e-23, 6.55555e-24, 7.46629e-24], [52.8739, 7.66989e-24, 1.03177e-23, 6.53894e-24, 7.45441e-24], [52.9958, 7.65186e-24, 1.03121e-23, 6.52247e-24, 7.44262e-24], [53.118, 7.63398e-24, 1.03065e-23, 6.50615e-24, 7.4309e-24], [53.2405, 7.61624e-24, 1.03009e-23, 6.48998e-24, 7.41927e-24], [53.3633, 7.59866e-24, 1.02954e-23, 6.47395e-24, 7.40771e-24], [53.4863, 7.58122e-24, 1.029e-23, 6.45807e-24, 7.39622e-24], [53.6097, 7.56393e-24, 1.02845e-23, 6.44233e-24, 7.38481e-24], [53.7333, 7.54677e-24, 1.02791e-23, 6.42673e-24, 7.37348e-24], [53.8572, 7.52977e-24, 1.02738e-23, 6.41128e-24, 7.36222e-24], [53.9814, 7.5129e-24, 1.02684e-23, 6.39597e-24, 7.35103e-24], [54.1059, 7.49618e-24, 1.02632e-23, 6.38079e-24, 7.33992e-24], [54.2306, 7.47959e-24, 1.02579e-23, 6.36575e-24, 7.32888e-24], [54.3557, 7.46314e-24, 1.02527e-23, 6.35085e-24, 7.3179e-24], [54.481, 7.44683e-24, 1.02475e-23, 6.33608e-24, 7.307e-24], [54.6067, 7.43066e-24, 1.02424e-23, 6.32145e-24, 7.29617e-24], [54.7326, 7.41463e-24, 1.02373e-23, 6.30695e-24, 7.2854e-24], [54.8588, 7.39872e-24, 1.02322e-23, 6.29258e-24, 7.27471e-24], [54.9853, 7.38296e-24, 1.02272e-23, 6.27834e-24, 7.26408e-24], [55.1121, 7.36732e-24, 1.02222e-23, 6.26423e-24, 7.25352e-24], [55.2392, 7.35182e-24, 1.02172e-23, 6.25025e-24, 7.24302e-24], [55.3666, 7.33645e-24, 1.02122e-23, 6.2364e-24, 7.23259e-24], [55.4943, 7.3212e-24, 1.02073e-23, 6.22267e-24, 7.22222e-24], [55.6222, 7.30609e-24, 1.02025e-23, 6.20907e-24, 7.21192e-24], [55.7505, 7.29111e-24, 1.01976e-23, 6.19559e-24, 7.20168e-24], [55.8791, 7.27625e-24, 1.01928e-23, 6.18223e-24, 7.1915e-24], [56.0079, 7.26152e-24, 1.0188e-23, 6.169e-24, 7.18138e-24], [56.1371, 7.24691e-24, 1.01833e-23, 6.15589e-24, 7.17133e-24], [56.2665, 7.23243e-24, 1.01785e-23, 6.14289e-24, 7.16134e-24], [56.3963, 7.21807e-24, 1.01738e-23, 6.13002e-24, 7.1514e-24], [56.5263, 7.20384e-24, 1.01692e-23, 6.11726e-24, 7.14153e-24], [56.6567, 7.18972e-24, 1.01645e-23, 6.10462e-24, 7.13172e-24], [56.7873, 7.17573e-24, 1.01599e-23, 6.09209e-24, 7.12196e-24], [56.9183, 7.16186e-24, 1.01554e-23, 6.07968e-24, 7.11226e-24], [57.0495, 7.14811e-24, 1.01508e-23, 6.06738e-24, 7.10262e-24], [57.1811, 7.13447e-24, 1.01463e-23, 6.0552e-24, 7.09303e-24], [57.3129, 7.12096e-24, 1.01418e-23, 6.04312e-24, 7.0835e-24], [57.4451, 7.10756e-24, 1.01373e-23, 6.03116e-24, 7.07403e-24], [57.5776, 7.09427e-24, 1.01328e-23, 6.0193e-24, 7.06461e-24], [57.7104, 7.08111e-24, 1.01284e-23, 6.00755e-24, 7.05525e-24], [57.8434, 7.06805e-24, 1.0124e-23, 5.99592e-24, 7.04594e-24], [57.9768, 7.05511e-24, 1.01197e-23, 5.98438e-24, 7.03668e-24], [58.1105, 7.04228e-24, 1.01153e-23, 5.97295e-24, 7.02748e-24], [58.2445, 7.02957e-24, 1.0111e-23, 5.96163e-24, 7.01832e-24], [58.3788, 7.01696e-24, 1.01067e-23, 5.95041e-24, 7.00922e-24], [58.5135, 7.00447e-24, 1.01024e-23, 5.93929e-24, 7.00018e-24], [58.6484, 6.99209e-24, 1.00982e-23, 5.92828e-24, 6.99118e-24], [58.7836, 6.97981e-24, 1.00939e-23, 5.91736e-24, 6.98223e-24], [58.9192, 6.96764e-24, 1.00897e-23, 5.90655e-24, 6.97333e-24], [59.0551, 6.95558e-24, 1.00856e-23, 5.89583e-24, 6.96448e-24], [59.1912, 6.94363e-24, 1.00814e-23, 5.88521e-24, 6.95568e-24], [59.3277, 6.93178e-24, 1.00773e-23, 5.87469e-24, 6.94693e-24], [59.4645, 6.92004e-24, 1.00732e-23, 5.86426e-24, 6.93823e-24], [59.6017, 6.9084e-24, 1.00691e-23, 5.85393e-24, 6.92957e-24], [59.7391, 6.89687e-24, 1.0065e-23, 5.8437e-24, 6.92097e-24], [59.8769, 6.88544e-24, 1.0061e-23, 5.83355e-24, 6.9124e-24], [60.0149, 6.87411e-24, 1.00569e-23, 5.8235e-24, 6.90389e-24], [60.1533, 6.86288e-24, 1.00529e-23, 5.81354e-24, 6.89542e-24], [60.2921, 6.85175e-24, 1.00489e-23, 5.80368e-24, 6.88699e-24], [60.4311, 6.84073e-24, 1.0045e-23, 5.7939e-24, 6.87861e-24], [60.5704, 6.8298e-24, 1.0041e-23, 5.78421e-24, 6.87028e-24], [60.7101, 6.81897e-24, 1.00371e-23, 5.77461e-24, 6.86199e-24], [60.8501, 6.80824e-24, 1.00332e-23, 5.76509e-24, 6.85374e-24], [60.9904, 6.79761e-24, 1.00293e-23, 5.75567e-24, 6.84553e-24], [61.1311, 6.78707e-24, 1.00255e-23, 5.74633e-24, 6.83737e-24], [61.2721, 6.77663e-24, 1.00216e-23, 5.73707e-24, 6.82925e-24], [61.4133, 6.76629e-24, 1.00178e-23, 5.7279e-24, 6.82118e-24], [61.555, 6.75604e-24, 1.0014e-23, 5.71881e-24, 6.81314e-24], [61.6969, 6.74588e-24, 1.00102e-23, 5.7098e-24, 6.80515e-24], [61.8392, 6.73582e-24, 1.00064e-23, 5.70087e-24, 6.79719e-24], [61.9818, 6.72585e-24, 1.00027e-23, 5.69203e-24, 6.78928e-24], [62.1247, 6.71597e-24, 9.99892e-24, 5.68327e-24, 6.78141e-24], [62.268, 6.70619e-24, 9.99521e-24, 5.67458e-24, 6.77357e-24], [62.4116, 6.69649e-24, 9.99151e-24, 5.66597e-24, 6.76578e-24], [62.5555, 6.68689e-24, 9.98783e-24, 5.65744e-24, 6.75803e-24], [62.6997, 6.67738e-24, 9.98416e-24, 5.64899e-24, 6.75031e-24], [62.8443, 6.66795e-24, 9.98051e-24, 5.64062e-24, 6.74263e-24], [62.9892, 6.65861e-24, 9.97688e-24, 5.63232e-24, 6.735e-24], [63.1345, 6.64937e-24, 9.97327e-24, 5.62409e-24, 6.72739e-24], [63.2801, 6.64021e-24, 9.96967e-24, 5.61594e-24, 6.71983e-24], [63.426, 6.63113e-24, 9.96609e-24, 5.60786e-24, 6.7123e-24], [63.5723, 6.62214e-24, 9.96252e-24, 5.59985e-24, 6.70481e-24], [63.7189, 6.61324e-24, 9.95898e-24, 5.59192e-24, 6.69736e-24], [63.8658, 6.60442e-24, 9.95544e-24, 5.58406e-24, 6.68994e-24], [64.0131, 6.59569e-24, 9.95193e-24, 5.57626e-24, 6.68256e-24], [64.1607, 6.58704e-24, 9.94842e-24, 5.56854e-24, 6.67521e-24], [64.3087, 6.57848e-24, 9.94494e-24, 5.56088e-24, 6.6679e-24], [64.457, 6.56999e-24, 9.94147e-24, 5.5533e-24, 6.66063e-24], [64.6056, 6.56159e-24, 9.93801e-24, 5.54578e-24, 6.65339e-24], [64.7546, 6.55328e-24, 9.93457e-24, 5.53833e-24, 6.64618e-24], [64.9039, 6.54504e-24, 9.93115e-24, 5.53094e-24, 6.63901e-24], [65.0536, 6.53688e-24, 9.92774e-24, 5.52362e-24, 6.63187e-24], [65.2036, 6.52881e-24, 9.92434e-24, 5.51636e-24, 6.62476e-24], [65.3539, 6.52081e-24, 9.92096e-24, 5.50917e-24, 6.61769e-24], [65.5046, 6.51289e-24, 9.91759e-24, 5.50204e-24, 6.61065e-24], [65.6557, 6.50505e-24, 9.91424e-24, 5.49498e-24, 6.60364e-24], [65.8071, 6.49729e-24, 9.9109e-24, 5.48797e-24, 6.59666e-24], [65.9589, 6.48961e-24, 9.90757e-24, 5.48103e-24, 6.58972e-24], [66.111, 6.482e-24, 9.90426e-24, 5.47415e-24, 6.58281e-24], [66.2634, 6.47448e-24, 9.90096e-24, 5.46733e-24, 6.57593e-24], [66.4162, 6.46702e-24, 9.89768e-24, 5.46057e-24, 6.56908e-24], [66.5694, 6.45965e-24, 9.89441e-24, 5.45387e-24, 6.56226e-24], [66.7229, 6.45234e-24, 9.89115e-24, 5.44722e-24, 6.55547e-24], [66.8767, 6.44512e-24, 9.88791e-24, 5.44064e-24, 6.54871e-24], [67.031, 6.43796e-24, 9.88468e-24, 5.43411e-24, 6.54199e-24], [67.1855, 6.43089e-24, 9.88146e-24, 5.42763e-24, 6.53529e-24], [67.3405, 6.42388e-24, 9.87825e-24, 5.42122e-24, 6.52862e-24], [67.4958, 6.41695e-24, 9.87506e-24, 5.41486e-24, 6.52199e-24], [67.6514, 6.41009e-24, 9.87188e-24, 5.40855e-24, 6.51538e-24], [67.8074, 6.4033e-24, 9.86871e-24, 5.4023e-24, 6.5088e-24], [67.9638, 6.39658e-24, 9.86556e-24, 5.3961e-24, 6.50225e-24], [68.1205, 6.38994e-24, 9.86242e-24, 5.38995e-24, 6.49572e-24], [68.2776, 6.38336e-24, 9.85929e-24, 5.38386e-24, 6.48923e-24], [68.435, 6.37686e-24, 9.85617e-24, 5.37781e-24, 6.48277e-24], [68.5928, 6.37042e-24, 9.85306e-24, 5.37182e-24, 6.47633e-24], [68.751, 6.36406e-24, 9.84997e-24, 5.36588e-24, 6.46992e-24], [68.9096, 6.35776e-24, 9.84689e-24, 5.35999e-24, 6.46353e-24], [69.0685, 6.35153e-24, 9.84382e-24, 5.35415e-24, 6.45718e-24], [69.2277, 6.34537e-24, 9.84076e-24, 5.34836e-24, 6.45085e-24], [69.3874, 6.33928e-24, 9.83771e-24, 5.34262e-24, 6.44455e-24], [69.5474, 6.33325e-24, 9.83468e-24, 5.33692e-24, 6.43827e-24], [69.7078, 6.3273e-24, 9.83165e-24, 5.33127e-24, 6.43202e-24], [69.8685, 6.3214e-24, 9.82864e-24, 5.32567e-24, 6.4258e-24], [70.0296, 6.31558e-24, 9.82564e-24, 5.32012e-24, 6.4196e-24], [70.1911, 6.30982e-24, 9.82265e-24, 5.31461e-24, 6.41343e-24], [70.353, 6.30413e-24, 9.81967e-24, 5.30914e-24, 6.40728e-24], [70.5152, 6.2985e-24, 9.8167e-24, 5.30372e-24, 6.40116e-24], [70.6778, 6.29293e-24, 9.81375e-24, 5.29835e-24, 6.39507e-24], [70.8408, 6.28743e-24, 9.8108e-24, 5.29302e-24, 6.389e-24], [71.0042, 6.28199e-24, 9.80786e-24, 5.28773e-24, 6.38295e-24], [71.1679, 6.27662e-24, 9.80494e-24, 5.28249e-24, 6.37693e-24], [71.332, 6.27131e-24, 9.80202e-24, 5.27728e-24, 6.37093e-24], [71.4965, 6.26606e-24, 9.79912e-24, 5.27212e-24, 6.36496e-24], [71.6614, 6.26088e-24, 9.79623e-24, 5.267e-24, 6.35901e-24], [71.8266, 6.25576e-24, 9.79334e-24, 5.26193e-24, 6.35309e-24], [71.9923, 6.2507e-24, 9.79047e-24, 5.25689e-24, 6.34719e-24], [72.1583, 6.2457e-24, 9.78761e-24, 5.25189e-24, 6.34131e-24], [72.3247, 6.24076e-24, 9.78476e-24, 5.24693e-24, 6.33546e-24], [72.4914, 6.23588e-24, 9.78191e-24, 5.24201e-24, 6.32963e-24], [72.6586, 6.23106e-24, 9.77908e-24, 5.23713e-24, 6.32382e-24], [72.8262, 6.22631e-24, 9.77626e-24, 5.23229e-24, 6.31803e-24], [72.9941, 6.22161e-24, 9.77345e-24, 5.22748e-24, 6.31227e-24], [73.1624, 6.21697e-24, 9.77064e-24, 5.22271e-24, 6.30653e-24], [73.3311, 6.21239e-24, 9.76785e-24, 5.21798e-24, 6.30081e-24], [73.5002, 6.20787e-24, 9.76507e-24, 5.21329e-24, 6.29512e-24], [73.6697, 6.20341e-24, 9.7623e-24, 5.20863e-24, 6.28944e-24], [73.8396, 6.199e-24, 9.75953e-24, 5.204e-24, 6.28379e-24], [74.0099, 6.19466e-24, 9.75678e-24, 5.19942e-24, 6.27816e-24], [74.1806, 6.19037e-24, 9.75403e-24, 5.19486e-24, 6.27255e-24], [74.3516, 6.18614e-24, 9.7513e-24, 5.19034e-24, 6.26697e-24], [74.5231, 6.18196e-24, 9.74857e-24, 5.18585e-24, 6.2614e-24], [74.6949, 6.17784e-24, 9.74585e-24, 5.1814e-24, 6.25586e-24], [74.8672, 6.17378e-24, 9.74315e-24, 5.17698e-24, 6.25033e-24], [75.0398, 6.16978e-24, 9.74045e-24, 5.17259e-24, 6.24483e-24], [75.2129, 6.16583e-24, 9.73776e-24, 5.16824e-24, 6.23935e-24], [75.3863, 6.16193e-24, 9.73508e-24, 5.16392e-24, 6.23388e-24], [75.5601, 6.15809e-24, 9.73241e-24, 5.15962e-24, 6.22844e-24], [75.7344, 6.15431e-24, 9.72974e-24, 5.15536e-24, 6.22302e-24], [75.909, 6.15058e-24, 9.72709e-24, 5.15113e-24, 6.21762e-24], [76.0841, 6.1469e-24, 9.72444e-24, 5.14693e-24, 6.21224e-24], [76.2595, 6.14328e-24, 9.72181e-24, 5.14276e-24, 6.20688e-24], [76.4354, 6.13971e-24, 9.71918e-24, 5.13862e-24, 6.20153e-24], [76.6116, 6.1362e-24, 9.71656e-24, 5.13451e-24, 6.19621e-24], [76.7883, 6.13274e-24, 9.71395e-24, 5.13042e-24, 6.19091e-24], [76.9654, 6.12933e-24, 9.71135e-24, 5.12637e-24, 6.18562e-24], [77.1429, 6.12597e-24, 9.70876e-24, 5.12234e-24, 6.18036e-24], [77.3208, 6.12267e-24, 9.70617e-24, 5.11834e-24, 6.17511e-24], [77.4991, 6.11942e-24, 9.7036e-24, 5.11437e-24, 6.16989e-24], [77.6778, 6.11622e-24, 9.70103e-24, 5.11042e-24, 6.16468e-24], [77.8569, 6.11308e-24, 9.69847e-24, 5.1065e-24, 6.15949e-24], [78.0364, 6.10998e-24, 9.69592e-24, 5.10261e-24, 6.15432e-24], [78.2164, 6.10694e-24, 9.69338e-24, 5.09874e-24, 6.14917e-24], [78.3968, 6.10395e-24, 9.69084e-24, 5.0949e-24, 6.14403e-24], [78.5775, 6.101e-24, 9.68832e-24, 5.09108e-24, 6.13892e-24], [78.7587, 6.09811e-24, 9.6858e-24, 5.08729e-24, 6.13382e-24], [78.9404, 6.09527e-24, 9.68329e-24, 5.08352e-24, 6.12874e-24], [79.1224, 6.09248e-24, 9.68079e-24, 5.07978e-24, 6.12367e-24], [79.3049, 6.08974e-24, 9.67829e-24, 5.07606e-24, 6.11863e-24], [79.4877, 6.08705e-24, 9.67581e-24, 5.07237e-24, 6.1136e-24], [79.671, 6.08441e-24, 9.67333e-24, 5.06869e-24, 6.10859e-24], [79.8548, 6.08182e-24, 9.67086e-24, 5.06504e-24, 6.1036e-24], [80.0389, 6.07928e-24, 9.6684e-24, 5.06142e-24, 6.09863e-24], [80.2235, 6.07678e-24, 9.66594e-24, 5.05781e-24, 6.09367e-24], [80.4085, 6.07434e-24, 9.6635e-24, 5.05423e-24, 6.08873e-24], [80.5939, 6.07194e-24, 9.66106e-24, 5.05067e-24, 6.0838e-24], [80.7797, 6.06959e-24, 9.65863e-24, 5.04713e-24, 6.07889e-24], [80.966, 6.06729e-24, 9.6562e-24, 5.04361e-24, 6.074e-24], [81.1527, 6.06504e-24, 9.65379e-24, 5.04011e-24, 6.06913e-24], [81.3399, 6.06284e-24, 9.65138e-24, 5.03663e-24, 6.06427e-24], [81.5274, 6.06068e-24, 9.64898e-24, 5.03318e-24, 6.05943e-24], [81.7154, 6.05857e-24, 9.64659e-24, 5.02974e-24, 6.05461e-24], [81.9039, 6.05651e-24, 9.6442e-24, 5.02632e-24, 6.0498e-24], [82.0927, 6.05449e-24, 9.64182e-24, 5.02292e-24, 6.045e-24], [82.282, 6.05252e-24, 9.63945e-24, 5.01954e-24, 6.04023e-24], [82.4718, 6.0506e-24, 9.63709e-24, 5.01618e-24, 6.03547e-24], [82.662, 6.04873e-24, 9.63473e-24, 5.01284e-24, 6.03072e-24], [82.8526, 6.0469e-24, 9.63238e-24, 5.00951e-24, 6.02599e-24], [83.0437, 6.04511e-24, 9.63004e-24, 5.00621e-24, 6.02128e-24], [83.2352, 6.04338e-24, 9.62771e-24, 5.00292e-24, 6.01658e-24], [83.4271, 6.04168e-24, 9.62538e-24, 4.99965e-24, 6.0119e-24], [83.6195, 6.04004e-24, 9.62306e-24, 4.99639e-24, 6.00723e-24], [83.8123, 6.03844e-24, 9.62075e-24, 4.99316e-24, 6.00258e-24], [84.0056, 6.03688e-24, 9.61845e-24, 4.98994e-24, 5.99794e-24], [84.1993, 6.03537e-24, 9.61615e-24, 4.98673e-24, 5.99332e-24], [84.3935, 6.03391e-24, 9.61386e-24, 4.98354e-24, 5.98871e-24], [84.5881, 6.03249e-24, 9.61157e-24, 4.98037e-24, 5.98412e-24], [84.7831, 6.03111e-24, 9.6093e-24, 4.97721e-24, 5.97954e-24], [84.9786, 6.02978e-24, 9.60703e-24, 4.97407e-24, 5.97498e-24], [85.1746, 6.02849e-24, 9.60476e-24, 4.97095e-24, 5.97043e-24], [85.371, 6.02725e-24, 9.60251e-24, 4.96783e-24, 5.9659e-24], [85.5679, 6.02605e-24, 9.60026e-24, 4.96474e-24, 5.96138e-24], [85.7652, 6.0249e-24, 9.59802e-24, 4.96166e-24, 5.95687e-24], [85.963, 6.02379e-24, 9.59578e-24, 4.95859e-24, 5.95238e-24], [86.1612, 6.02272e-24, 9.59355e-24, 4.95553e-24, 5.9479e-24], [86.3599, 6.0217e-24, 9.59133e-24, 4.95249e-24, 5.94344e-24], [86.559, 6.02072e-24, 9.58912e-24, 4.94947e-24, 5.93899e-24], [86.7586, 6.01978e-24, 9.58691e-24, 4.94645e-24, 5.93456e-24], [86.9587, 6.01888e-24, 9.58471e-24, 4.94345e-24, 5.93014e-24], [87.1592, 6.01803e-24, 9.58252e-24, 4.94046e-24, 5.92573e-24], [87.3602, 6.01722e-24, 9.58033e-24, 4.93749e-24, 5.92134e-24], [87.5617, 6.01646e-24, 9.57815e-24, 4.93453e-24, 5.91696e-24], [87.7636, 6.01574e-24, 9.57597e-24, 4.93157e-24, 5.9126e-24], [87.966, 6.01506e-24, 9.57381e-24, 4.92864e-24, 5.90824e-24], [88.1688, 6.01442e-24, 9.57165e-24, 4.92571e-24, 5.90391e-24], [88.3722, 6.01382e-24, 9.56949e-24, 4.92279e-24, 5.89958e-24], [88.5759, 6.01327e-24, 9.56734e-24, 4.91989e-24, 5.89527e-24], [88.7802, 6.01276e-24, 9.5652e-24, 4.917e-24, 5.89097e-24], [88.9849, 6.01229e-24, 9.56307e-24, 4.91411e-24, 5.88668e-24], [89.1901, 6.01186e-24, 9.56094e-24, 4.91124e-24, 5.88241e-24], [89.3958, 6.01147e-24, 9.55882e-24, 4.90838e-24, 5.87815e-24], [89.6019, 6.01113e-24, 9.5567e-24, 4.90553e-24, 5.87391e-24], [89.8086, 6.01083e-24, 9.55459e-24, 4.90269e-24, 5.86967e-24], [90.0157, 6.01056e-24, 9.55249e-24, 4.89986e-24, 5.86545e-24], [90.2232, 6.01034e-24, 9.5504e-24, 4.89704e-24, 5.86124e-24], [90.4313, 6.01016e-24, 9.54831e-24, 4.89423e-24, 5.85705e-24], [90.6398, 6.01003e-24, 9.54622e-24, 4.89142e-24, 5.85287e-24], [90.8489, 6.00993e-24, 9.54415e-24, 4.88863e-24, 5.8487e-24], [91.0583, 6.00987e-24, 9.54208e-24, 4.88585e-24, 5.84454e-24], [91.2683, 6.00986e-24, 9.54001e-24, 4.88307e-24, 5.84039e-24], [91.4788, 6.00988e-24, 9.53795e-24, 4.88031e-24, 5.83626e-24], [91.6897, 6.00995e-24, 9.5359e-24, 4.87755e-24, 5.83214e-24], [91.9012, 6.01005e-24, 9.53386e-24, 4.8748e-24, 5.82803e-24], [92.1131, 6.0102e-24, 9.53182e-24, 4.87206e-24, 5.82393e-24], [92.3255, 6.01039e-24, 9.52979e-24, 4.86932e-24, 5.81985e-24], [92.5384, 6.01061e-24, 9.52776e-24, 4.8666e-24, 5.81578e-24], [92.7518, 6.01088e-24, 9.52574e-24, 4.86388e-24, 5.81172e-24], [92.9657, 6.01119e-24, 9.52372e-24, 4.86117e-24, 5.80767e-24], [93.1801, 6.01154e-24, 9.52172e-24, 4.85847e-24, 5.80363e-24], [93.395, 6.01192e-24, 9.51971e-24, 4.85577e-24, 5.79961e-24], [93.6103, 6.01235e-24, 9.51772e-24, 4.85308e-24, 5.79559e-24], [93.8262, 6.01282e-24, 9.51573e-24, 4.8504e-24, 5.79159e-24], [94.0426, 6.01332e-24, 9.51374e-24, 4.84772e-24, 5.7876e-24], [94.2594, 6.01387e-24, 9.51177e-24, 4.84505e-24, 5.78363e-24], [94.4768, 6.01446e-24, 9.5098e-24, 4.84239e-24, 5.77966e-24], [94.6946, 6.01508e-24, 9.50783e-24, 4.83973e-24, 5.7757e-24], [94.913, 6.01575e-24, 9.50587e-24, 4.83708e-24, 5.77176e-24], [95.1319, 6.01645e-24, 9.50392e-24, 4.83443e-24, 5.76783e-24], [95.3513, 6.01719e-24, 9.50197e-24, 4.83179e-24, 5.76391e-24], [95.5711, 6.01798e-24, 9.50003e-24, 4.82916e-24, 5.76e-24], [95.7915, 6.0188e-24, 9.49809e-24, 4.82653e-24, 5.7561e-24], [96.0124, 6.01966e-24, 9.49616e-24, 4.82391e-24, 5.75221e-24], [96.2338, 6.02056e-24, 9.49424e-24, 4.82129e-24, 5.74833e-24], [96.4557, 6.0215e-24, 9.49232e-24, 4.81867e-24, 5.74447e-24], [96.6782, 6.02248e-24, 9.49041e-24, 4.81606e-24, 5.74061e-24], [96.9011, 6.02349e-24, 9.4885e-24, 4.81346e-24, 5.73677e-24], [97.1246, 6.02455e-24, 9.4866e-24, 4.81086e-24, 5.73294e-24], [97.3485, 6.02564e-24, 9.48471e-24, 4.80826e-24, 5.72912e-24], [97.573, 6.02678e-24, 9.48282e-24, 4.80567e-24, 5.72531e-24], [97.798, 6.02795e-24, 9.48094e-24, 4.80309e-24, 5.72151e-24], [98.0236, 6.02916e-24, 9.47906e-24, 4.8005e-24, 5.71772e-24], [98.2496, 6.03041e-24, 9.47719e-24, 4.79792e-24, 5.71394e-24], [98.4762, 6.03169e-24, 9.47533e-24, 4.79535e-24, 5.71017e-24], [98.7032, 6.03302e-24, 9.47347e-24, 4.79277e-24, 5.70641e-24], [98.9309, 6.03438e-24, 9.47161e-24, 4.79021e-24, 5.70266e-24], [99.159, 6.03579e-24, 9.46976e-24, 4.78764e-24, 5.69893e-24], [99.3877, 6.03723e-24, 9.46792e-24, 4.78508e-24, 5.6952e-24], [99.6168, 6.03871e-24, 9.46609e-24, 4.78252e-24, 5.69148e-24], [99.8466, 6.04022e-24, 9.46426e-24, 4.77996e-24, 5.68778e-24], [100.077, 6.04178e-24, 9.46243e-24, 4.77741e-24, 5.68408e-24], [100.308, 6.04337e-24, 9.46061e-24, 4.77486e-24, 5.6804e-24], [100.539, 6.045e-24, 9.4588e-24, 4.77231e-24, 5.67672e-24], [100.771, 6.04667e-24, 9.45699e-24, 4.76977e-24, 5.67306e-24], [101.003, 6.04838e-24, 9.45519e-24, 4.76722e-24, 5.6694e-24], [101.236, 6.05013e-24, 9.45339e-24, 4.76468e-24, 5.66576e-24], [101.469, 6.05191e-24, 9.4516e-24, 4.76214e-24, 5.66212e-24], [101.703, 6.05373e-24, 9.44982e-24, 4.7596e-24, 5.6585e-24], [101.938, 6.05559e-24, 9.44804e-24, 4.75707e-24, 5.65488e-24], [102.173, 6.05749e-24, 9.44626e-24, 4.75454e-24, 5.65128e-24], [102.409, 6.05942e-24, 9.44449e-24, 4.752e-24, 5.64768e-24], [102.645, 6.06139e-24, 9.44273e-24, 4.74947e-24, 5.6441e-24], [102.882, 6.0634e-24, 9.44097e-24, 4.74694e-24, 5.64052e-24], [103.119, 6.06545e-24, 9.43922e-24, 4.74442e-24, 5.63695e-24], [103.357, 6.06754e-24, 9.43748e-24, 4.74189e-24, 5.6334e-24], [103.595, 6.06966e-24, 9.43574e-24, 4.73936e-24, 5.62985e-24], [103.834, 6.07182e-24, 9.434e-24, 4.73684e-24, 5.62631e-24], [104.073, 6.07402e-24, 9.43227e-24, 4.73432e-24, 5.62278e-24], [104.313, 6.07626e-24, 9.43055e-24, 4.73179e-24, 5.61927e-24], [104.554, 6.07853e-24, 9.42883e-24, 4.72927e-24, 5.61576e-24], [104.795, 6.08085e-24, 9.42712e-24, 4.72675e-24, 5.61226e-24], [105.037, 6.08319e-24, 9.42541e-24, 4.72423e-24, 5.60877e-24], [105.279, 6.08558e-24, 9.42371e-24, 4.72171e-24, 5.60529e-24], [105.522, 6.08801e-24, 9.42201e-24, 4.71919e-24, 5.60182e-24], [105.765, 6.09047e-24, 9.42032e-24, 4.71667e-24, 5.59835e-24], [106.009, 6.09297e-24, 9.41863e-24, 4.71415e-24, 5.5949e-24], [106.253, 6.0955e-24, 9.41695e-24, 4.71163e-24, 5.59146e-24], [106.498, 6.09808e-24, 9.41528e-24, 4.70911e-24, 5.58802e-24], [106.744, 6.10069e-24, 9.41361e-24, 4.70659e-24, 5.5846e-24], [106.99, 6.10334e-24, 9.41195e-24, 4.70407e-24, 5.58118e-24], [107.237, 6.10602e-24, 9.41029e-24, 4.70155e-24, 5.57777e-24], [107.484, 6.10875e-24, 9.40864e-24, 4.69902e-24, 5.57438e-24], [107.732, 6.11151e-24, 9.40699e-24, 4.6965e-24, 5.57099e-24], [107.98, 6.1143e-24, 9.40535e-24, 4.69398e-24, 5.56761e-24], [108.229, 6.11714e-24, 9.40371e-24, 4.69146e-24, 5.56424e-24], [108.479, 6.12001e-24, 9.40208e-24, 4.68893e-24, 5.56087e-24], [108.729, 6.12292e-24, 9.40045e-24, 4.68641e-24, 5.55752e-24], [108.98, 6.12587e-24, 9.39883e-24, 4.68388e-24, 5.55418e-24], [109.231, 6.12886e-24, 9.39722e-24, 4.68135e-24, 5.55084e-24], [109.483, 6.13188e-24, 9.39561e-24, 4.67882e-24, 5.54751e-24], [109.735, 6.13494e-24, 9.394e-24, 4.67629e-24, 5.54419e-24], [109.988, 6.13803e-24, 9.3924e-24, 4.67376e-24, 5.54088e-24], [110.242, 6.14117e-24, 9.39081e-24, 4.67123e-24, 5.53758e-24], [110.496, 6.14434e-24, 9.38922e-24, 4.6687e-24, 5.53429e-24], [110.751, 6.14755e-24, 9.38764e-24, 4.66616e-24, 5.53101e-24], [111.007, 6.15079e-24, 9.38606e-24, 4.66362e-24, 5.52773e-24], [111.262, 6.15408e-24, 9.38449e-24, 4.66108e-24, 5.52447e-24], [111.519, 6.1574e-24, 9.38292e-24, 4.65854e-24, 5.52121e-24], [111.776, 6.16075e-24, 9.38136e-24, 4.656e-24, 5.51796e-24], [112.034, 6.16415e-24, 9.3798e-24, 4.65346e-24, 5.51472e-24], [112.292, 6.16758e-24, 9.37825e-24, 4.65091e-24, 5.51148e-24], [112.551, 6.17105e-24, 9.3767e-24, 4.64836e-24, 5.50826e-24], [112.811, 6.17455e-24, 9.37516e-24, 4.64581e-24, 5.50504e-24], [113.071, 6.1781e-24, 9.37362e-24, 4.64325e-24, 5.50184e-24], [113.332, 6.18168e-24, 9.37209e-24, 4.6407e-24, 5.49864e-24], [113.593, 6.1853e-24, 9.37057e-24, 4.63814e-24, 5.49545e-24], [113.855, 6.18895e-24, 9.36905e-24, 4.63558e-24, 5.49226e-24], [114.118, 6.19265e-24, 9.36753e-24, 4.63302e-24, 5.48909e-24], [114.381, 6.19638e-24, 9.36602e-24, 4.63045e-24, 5.48592e-24], [114.644, 6.20014e-24, 9.36452e-24, 4.62788e-24, 5.48276e-24], [114.909, 6.20395e-24, 9.36302e-24, 4.62531e-24, 5.47961e-24], [115.174, 6.20779e-24, 9.36153e-24, 4.62274e-24, 5.47647e-24], [115.439, 6.21167e-24, 9.36004e-24, 4.62016e-24, 5.47334e-24], [115.706, 6.21559e-24, 9.35855e-24, 4.61758e-24, 5.47021e-24], [115.972, 6.21954e-24, 9.35708e-24, 4.615e-24, 5.46709e-24], [116.24, 6.22353e-24, 9.3556e-24, 4.61241e-24, 5.46398e-24], [116.508, 6.22756e-24, 9.35413e-24, 4.60982e-24, 5.46088e-24], [116.777, 6.23162e-24, 9.35267e-24, 4.60723e-24, 5.45779e-24], [117.046, 6.23573e-24, 9.35121e-24, 4.60463e-24, 5.4547e-24], [117.316, 6.23987e-24, 9.34976e-24, 4.60203e-24, 5.45162e-24], [117.586, 6.24405e-24, 9.34831e-24, 4.59943e-24, 5.44855e-24], [117.857, 6.24826e-24, 9.34687e-24, 4.59682e-24, 5.44549e-24], [118.129, 6.25251e-24, 9.34544e-24, 4.59421e-24, 5.44244e-24], [118.402, 6.2568e-24, 9.344e-24, 4.5916e-24, 5.43939e-24], [118.675, 6.26113e-24, 9.34258e-24, 4.58898e-24, 5.43635e-24], [118.948, 6.2655e-24, 9.34116e-24, 4.58636e-24, 5.43332e-24], [119.223, 6.2699e-24, 9.33974e-24, 4.58374e-24, 5.4303e-24], [119.498, 6.27434e-24, 9.33833e-24, 4.58111e-24, 5.42728e-24], [119.773, 6.27882e-24, 9.33692e-24, 4.57848e-24, 5.42427e-24], [120.049, 6.28333e-24, 9.33552e-24, 4.57584e-24, 5.42127e-24], [120.326, 6.28788e-24, 9.33412e-24, 4.5732e-24, 5.41828e-24], [120.604, 6.29247e-24, 9.33273e-24, 4.57056e-24, 5.41529e-24], [120.882, 6.2971e-24, 9.33135e-24, 4.56791e-24, 5.41232e-24], [121.161, 6.30177e-24, 9.32997e-24, 4.56526e-24, 5.40935e-24], [121.44, 6.30647e-24, 9.32859e-24, 4.5626e-24, 5.40638e-24], [121.72, 6.31121e-24, 9.32722e-24, 4.55994e-24, 5.40343e-24], [122.001, 6.31599e-24, 9.32586e-24, 4.55727e-24, 5.40048e-24], [122.282, 6.3208e-24, 9.3245e-24, 4.55461e-24, 5.39754e-24], [122.564, 6.32565e-24, 9.32314e-24, 4.55193e-24, 5.39461e-24], [122.847, 6.33055e-24, 9.32179e-24, 4.54925e-24, 5.39168e-24], [123.13, 6.33547e-24, 9.32045e-24, 4.54657e-24, 5.38877e-24], [123.414, 6.34044e-24, 9.31911e-24, 4.54388e-24, 5.38585e-24], [123.698, 6.34544e-24, 9.31777e-24, 4.54119e-24, 5.38295e-24], [123.984, 6.35049e-24, 9.31644e-24, 4.53849e-24, 5.38006e-24], [124.27, 6.35557e-24, 9.31512e-24, 4.53579e-24, 5.37717e-24], [124.556, 6.36068e-24, 9.3138e-24, 4.53308e-24, 5.37429e-24], [124.843, 6.36584e-24, 9.31249e-24, 4.53037e-24, 5.37141e-24], [125.131, 6.37103e-24, 9.31118e-24, 4.52766e-24, 5.36855e-24], [125.42, 6.37626e-24, 9.30987e-24, 4.52494e-24, 5.36569e-24], [125.709, 6.38153e-24, 9.30857e-24, 4.52221e-24, 5.36283e-24], [125.999, 6.38684e-24, 9.30728e-24, 4.51948e-24, 5.35999e-24], [126.289, 6.39218e-24, 9.30599e-24, 4.51674e-24, 5.35715e-24], [126.581, 6.39756e-24, 9.30471e-24, 4.514e-24, 5.35432e-24], [126.873, 6.40299e-24, 9.30343e-24, 4.51126e-24, 5.3515e-24], [127.165, 6.40844e-24, 9.30215e-24, 4.5085e-24, 5.34868e-24], [127.458, 6.41394e-24, 9.30089e-24, 4.50575e-24, 5.34587e-24], [127.752, 6.41948e-24, 9.29962e-24, 4.50299e-24, 5.34307e-24], [128.047, 6.42505e-24, 9.29836e-24, 4.50022e-24, 5.34028e-24], [128.342, 6.43066e-24, 9.29711e-24, 4.49745e-24, 5.33749e-24], [128.638, 6.43631e-24, 9.29586e-24, 4.49467e-24, 5.33471e-24], [128.935, 6.442e-24, 9.29462e-24, 4.49188e-24, 5.33193e-24], [129.232, 6.44772e-24, 9.29338e-24, 4.4891e-24, 5.32917e-24], [129.53, 6.45348e-24, 9.29215e-24, 4.4863e-24, 5.32641e-24], [129.829, 6.45929e-24, 9.29092e-24, 4.4835e-24, 5.32365e-24], [130.128, 6.46513e-24, 9.28969e-24, 4.4807e-24, 5.32091e-24], [130.428, 6.47101e-24, 9.28848e-24, 4.47788e-24, 5.31817e-24], [130.729, 6.47692e-24, 9.28726e-24, 4.47507e-24, 5.31544e-24], [131.031, 6.48288e-24, 9.28605e-24, 4.47225e-24, 5.31271e-24], [131.333, 6.48887e-24, 9.28485e-24, 4.46942e-24, 5.30999e-24], [131.636, 6.4949e-24, 9.28365e-24, 4.46658e-24, 5.30728e-24], [131.939, 6.50097e-24, 9.28246e-24, 4.46374e-24, 5.30458e-24], [132.243, 6.50708e-24, 9.28127e-24, 4.4609e-24, 5.30188e-24], [132.548, 6.51323e-24, 9.28009e-24, 4.45805e-24, 5.29919e-24], [132.854, 6.51942e-24, 9.27891e-24, 4.45519e-24, 5.2965e-24], [133.16, 6.52564e-24, 9.27774e-24, 4.45233e-24, 5.29382e-24], [133.467, 6.53191e-24, 9.27657e-24, 4.44946e-24, 5.29115e-24], [133.775, 6.53821e-24, 9.27541e-24, 4.44658e-24, 5.28849e-24], [134.084, 6.54455e-24, 9.27425e-24, 4.4437e-24, 5.28583e-24], [134.393, 6.55093e-24, 9.2731e-24, 4.44081e-24, 5.28318e-24], [134.703, 6.55735e-24, 9.27195e-24, 4.43792e-24, 5.28054e-24], [135.013, 6.5638e-24, 9.27081e-24, 4.43502e-24, 5.2779e-24], [135.325, 6.5703e-24, 9.26967e-24, 4.43212e-24, 5.27527e-24], [135.637, 6.57684e-24, 9.26854e-24, 4.4292e-24, 5.27265e-24], [135.95, 6.58341e-24, 9.26741e-24, 4.42629e-24, 5.27003e-24], [136.263, 6.59002e-24, 9.26629e-24, 4.42336e-24, 5.26742e-24], [136.577, 6.59667e-24, 9.26517e-24, 4.42043e-24, 5.26481e-24], [136.892, 6.60336e-24, 9.26406e-24, 4.41749e-24, 5.26222e-24], [137.208, 6.61009e-24, 9.26295e-24, 4.41455e-24, 5.25962e-24], [137.524, 6.61686e-24, 9.26184e-24, 4.4116e-24, 5.25704e-24], [137.841, 6.62367e-24, 9.26075e-24, 4.40865e-24, 5.25446e-24], [138.159, 6.63052e-24, 9.25965e-24, 4.40568e-24, 5.25189e-24], [138.478, 6.6374e-24, 9.25857e-24, 4.40272e-24, 5.24933e-24], [138.797, 6.64433e-24, 9.25749e-24, 4.39974e-24, 5.24677e-24], [139.117, 6.65129e-24, 9.25641e-24, 4.39676e-24, 5.24421e-24], [139.438, 6.6583e-24, 9.25534e-24, 4.39377e-24, 5.24167e-24], [139.76, 6.66534e-24, 9.25427e-24, 4.39078e-24, 5.23913e-24], [140.082, 6.67242e-24, 9.25321e-24, 4.38778e-24, 5.2366e-24], [140.405, 6.67954e-24, 9.25215e-24, 4.38477e-24, 5.23407e-24], [140.729, 6.68671e-24, 9.2511e-24, 4.38175e-24, 5.23155e-24], [141.053, 6.69391e-24, 9.25005e-24, 4.37873e-24, 5.22904e-24], [141.379, 6.70115e-24, 9.24901e-24, 4.37571e-24, 5.22653e-24], [141.705, 6.70843e-24, 9.24797e-24, 4.37267e-24, 5.22403e-24], [142.031, 6.71575e-24, 9.24694e-24, 4.36963e-24, 5.22154e-24], [142.359, 6.72311e-24, 9.24591e-24, 4.36658e-24, 5.21905e-24], [142.687, 6.7305e-24, 9.24489e-24, 4.36353e-24, 5.21657e-24], [143.016, 6.73794e-24, 9.24387e-24, 4.36047e-24, 5.21409e-24], [143.346, 6.74542e-24, 9.24286e-24, 4.3574e-24, 5.21162e-24], [143.677, 6.75294e-24, 9.24185e-24, 4.35433e-24, 5.20916e-24], [144.008, 6.7605e-24, 9.24085e-24, 4.35125e-24, 5.2067e-24], [144.34, 6.76809e-24, 9.23985e-24, 4.34816e-24, 5.20425e-24], [144.673, 6.77573e-24, 9.23886e-24, 4.34506e-24, 5.20181e-24], [145.006, 6.78341e-24, 9.23788e-24, 4.34196e-24, 5.19937e-24], [145.341, 6.79113e-24, 9.23689e-24, 4.33885e-24, 5.19694e-24], [145.676, 6.79888e-24, 9.23592e-24, 4.33574e-24, 5.19452e-24], [146.012, 6.80668e-24, 9.23495e-24, 4.33262e-24, 5.1921e-24], [146.349, 6.81452e-24, 9.23398e-24, 4.32949e-24, 5.18969e-24], [146.686, 6.8224e-24, 9.23302e-24, 4.32635e-24, 5.18728e-24], [147.024, 6.83031e-24, 9.23206e-24, 4.32321e-24, 5.18488e-24], [147.363, 6.83827e-24, 9.23111e-24, 4.32006e-24, 5.18248e-24], [147.703, 6.84627e-24, 9.23016e-24, 4.31691e-24, 5.1801e-24], [148.044, 6.85431e-24, 9.22922e-24, 4.31374e-24, 5.17771e-24], [148.385, 6.86239e-24, 9.22829e-24, 4.31057e-24, 5.17534e-24], [148.727, 6.87051e-24, 9.22736e-24, 4.3074e-24, 5.17297e-24], [149.07, 6.87867e-24, 9.22643e-24, 4.30421e-24, 5.17061e-24], [149.414, 6.88687e-24, 9.22551e-24, 4.30102e-24, 5.16825e-24], [149.759, 6.89511e-24, 9.22459e-24, 4.29782e-24, 5.1659e-24], [150.104, 6.90339e-24, 9.22368e-24, 4.29462e-24, 5.16355e-24], [150.45, 6.91171e-24, 9.22278e-24, 4.29141e-24, 5.16121e-24], [150.797, 6.92007e-24, 9.22188e-24, 4.28819e-24, 5.15888e-24], [151.145, 6.92848e-24, 9.22098e-24, 4.28496e-24, 5.15655e-24], [151.493, 6.93692e-24, 9.22009e-24, 4.28173e-24, 5.15423e-24], [151.843, 6.94541e-24, 9.21921e-24, 4.27849e-24, 5.15192e-24], [152.193, 6.95393e-24, 9.21833e-24, 4.27524e-24, 5.14961e-24], [152.544, 6.9625e-24, 9.21745e-24, 4.27199e-24, 5.14731e-24], [152.896, 6.97111e-24, 9.21658e-24, 4.26873e-24, 5.14501e-24], [153.248, 6.97975e-24, 9.21572e-24, 4.26546e-24, 5.14272e-24], [153.601, 6.98844e-24, 9.21486e-24, 4.26219e-24, 5.14043e-24], [153.956, 6.99717e-24, 9.214e-24, 4.2589e-24, 5.13815e-24], [154.311, 7.00595e-24, 9.21315e-24, 4.25562e-24, 5.13588e-24], [154.667, 7.01476e-24, 9.21231e-24, 4.25232e-24, 5.13361e-24], [155.023, 7.02361e-24, 9.21147e-24, 4.24902e-24, 5.13135e-24], [155.381, 7.03251e-24, 9.21063e-24, 4.24571e-24, 5.1291e-24], [155.739, 7.04144e-24, 9.20981e-24, 4.24239e-24, 5.12685e-24], [156.098, 7.05042e-24, 9.20898e-24, 4.23907e-24, 5.1246e-24], [156.458, 7.05944e-24, 9.20816e-24, 4.23574e-24, 5.12237e-24], [156.819, 7.0685e-24, 9.20735e-24, 4.2324e-24, 5.12014e-24], [157.181, 7.0776e-24, 9.20654e-24, 4.22905e-24, 5.11791e-24], [157.543, 7.08675e-24, 9.20574e-24, 4.2257e-24, 5.11569e-24], [157.906, 7.09593e-24, 9.20494e-24, 4.22234e-24, 5.11348e-24], [158.27, 7.10516e-24, 9.20415e-24, 4.21898e-24, 5.11127e-24], [158.635, 7.11442e-24, 9.20336e-24, 4.21561e-24, 5.10907e-24], [159.001, 7.12373e-24, 9.20258e-24, 4.21223e-24, 5.10687e-24], [159.368, 7.13309e-24, 9.2018e-24, 4.20884e-24, 5.10468e-24], [159.735, 7.14248e-24, 9.20103e-24, 4.20545e-24, 5.10249e-24], [160.104, 7.15191e-24, 9.20026e-24, 4.20205e-24, 5.10032e-24], [160.473, 7.16139e-24, 9.1995e-24, 4.19864e-24, 5.09814e-24], [160.843, 7.17091e-24, 9.19874e-24, 4.19523e-24, 5.09598e-24], [161.214, 7.18047e-24, 9.19799e-24, 4.19181e-24, 5.09381e-24], [161.586, 7.19007e-24, 9.19724e-24, 4.18838e-24, 5.09166e-24], [161.958, 7.19972e-24, 9.1965e-24, 4.18495e-24, 5.08951e-24], [162.332, 7.20941e-24, 9.19577e-24, 4.1815e-24, 5.08736e-24], [162.706, 7.21914e-24, 9.19504e-24, 4.17806e-24, 5.08523e-24], [163.081, 7.22891e-24, 9.19431e-24, 4.1746e-24, 5.08309e-24], [163.457, 7.23872e-24, 9.19359e-24, 4.17114e-24, 5.08097e-24], [163.834, 7.24858e-24, 9.19288e-24, 4.16767e-24, 5.07885e-24], [164.212, 7.25848e-24, 9.19217e-24, 4.1642e-24, 5.07673e-24], [164.591, 7.26842e-24, 9.19146e-24, 4.16072e-24, 5.07462e-24], [164.97, 7.2784e-24, 9.19076e-24, 4.15723e-24, 5.07252e-24], [165.351, 7.28843e-24, 9.19007e-24, 4.15373e-24, 5.07042e-24], [165.732, 7.29849e-24, 9.18938e-24, 4.15023e-24, 5.06833e-24], [166.114, 7.30861e-24, 9.1887e-24, 4.14672e-24, 5.06624e-24], [166.497, 7.31876e-24, 9.18802e-24, 4.14321e-24, 5.06416e-24], [166.881, 7.32896e-24, 9.18735e-24, 4.13969e-24, 5.06208e-24], [167.266, 7.3392e-24, 9.18668e-24, 4.13616e-24, 5.06001e-24], [167.652, 7.34948e-24, 9.18602e-24, 4.13263e-24, 5.05795e-24], [168.038, 7.3598e-24, 9.18536e-24, 4.12909e-24, 5.05589e-24], [168.426, 7.37017e-24, 9.18471e-24, 4.12554e-24, 5.05384e-24], [168.814, 7.38058e-24, 9.18406e-24, 4.12199e-24, 5.05179e-24], [169.204, 7.39104e-24, 9.18342e-24, 4.11843e-24, 5.04975e-24], [169.594, 7.40153e-24, 9.18278e-24, 4.11486e-24, 5.04772e-24], [169.985, 7.41207e-24, 9.18215e-24, 4.11129e-24, 5.04569e-24], [170.377, 7.42266e-24, 9.18153e-24, 4.10771e-24, 5.04366e-24], [170.77, 7.43328e-24, 9.18091e-24, 4.10413e-24, 5.04164e-24], [171.163, 7.44395e-24, 9.1803e-24, 4.10053e-24, 5.03963e-24], [171.558, 7.45467e-24, 9.17969e-24, 4.09694e-24, 5.03762e-24], [171.954, 7.46542e-24, 9.17908e-24, 4.09333e-24, 5.03562e-24], [172.35, 7.47622e-24, 9.17848e-24, 4.08972e-24, 5.03362e-24], [172.748, 7.48707e-24, 9.17789e-24, 4.08611e-24, 5.03163e-24], [173.146, 7.49795e-24, 9.1773e-24, 4.08249e-24, 5.02965e-24], [173.545, 7.50888e-24, 9.17672e-24, 4.07886e-24, 5.02767e-24], [173.946, 7.51986e-24, 9.17614e-24, 4.07523e-24, 5.0257e-24], [174.347, 7.53088e-24, 9.17557e-24, 4.07159e-24, 5.02373e-24], [174.749, 7.54194e-24, 9.17501e-24, 4.06794e-24, 5.02176e-24], [175.152, 7.55304e-24, 9.17445e-24, 4.06429e-24, 5.01981e-24], [175.556, 7.56419e-24, 9.17389e-24, 4.06064e-24, 5.01786e-24], [175.96, 7.57539e-24, 9.17334e-24, 4.05697e-24, 5.01591e-24], [176.366, 7.58662e-24, 9.1728e-24, 4.05331e-24, 5.01397e-24], [176.773, 7.59791e-24, 9.17226e-24, 4.04963e-24, 5.01203e-24], [177.181, 7.60923e-24, 9.17173e-24, 4.04596e-24, 5.0101e-24], [177.589, 7.6206e-24, 9.1712e-24, 4.04227e-24, 5.00818e-24], [177.999, 7.63201e-24, 9.17067e-24, 4.03858e-24, 5.00626e-24], [178.409, 7.64347e-24, 9.17016e-24, 4.03489e-24, 5.00435e-24], [178.821, 7.65498e-24, 9.16965e-24, 4.03119e-24, 5.00244e-24], [179.233, 7.66652e-24, 9.16914e-24, 4.02748e-24, 5.00054e-24], [179.646, 7.67811e-24, 9.16864e-24, 4.02377e-24, 4.99864e-24], [180.061, 7.68975e-24, 9.16814e-24, 4.02006e-24, 4.99675e-24], [180.476, 7.70143e-24, 9.16765e-24, 4.01634e-24, 4.99487e-24], [180.892, 7.71316e-24, 9.16717e-24, 4.01261e-24, 4.99299e-24], [181.309, 7.72493e-24, 9.16669e-24, 4.00888e-24, 4.99112e-24], [181.727, 7.73674e-24, 9.16622e-24, 4.00514e-24, 4.98925e-24], [182.146, 7.7486e-24, 9.16575e-24, 4.00141e-24, 4.98738e-24], [182.566, 7.7605e-24, 9.16529e-24, 3.99766e-24, 4.98553e-24], [182.987, 7.77245e-24, 9.16483e-24, 3.99391e-24, 4.98367e-24], [183.409, 7.78445e-24, 9.16438e-24, 3.99016e-24, 4.98183e-24], [183.832, 7.79648e-24, 9.16394e-24, 3.9864e-24, 4.97999e-24], [184.256, 7.80857e-24, 9.1635e-24, 3.98264e-24, 4.97815e-24], [184.681, 7.8207e-24, 9.16306e-24, 3.97887e-24, 4.97632e-24], [185.107, 7.83287e-24, 9.16264e-24, 3.9751e-24, 4.97449e-24], [185.534, 7.84509e-24, 9.16221e-24, 3.97133e-24, 4.97268e-24], [185.962, 7.85736e-24, 9.1618e-24, 3.96755e-24, 4.97086e-24], [186.39, 7.86967e-24, 9.16138e-24, 3.96376e-24, 4.96905e-24], [186.82, 7.88202e-24, 9.16098e-24, 3.95998e-24, 4.96725e-24], [187.251, 7.89442e-24, 9.16058e-24, 3.95619e-24, 4.96545e-24], [187.683, 7.90687e-24, 9.16018e-24, 3.95239e-24, 4.96366e-24], [188.116, 7.91936e-24, 9.1598e-24, 3.94859e-24, 4.96187e-24], [188.549, 7.9319e-24, 9.15941e-24, 3.94479e-24, 4.96009e-24], [188.984, 7.94448e-24, 9.15903e-24, 3.94099e-24, 4.95832e-24], [189.42, 7.95711e-24, 9.15866e-24, 3.93718e-24, 4.95655e-24], [189.857, 7.96979e-24, 9.1583e-24, 3.93337e-24, 4.95478e-24], [190.295, 7.98251e-24, 9.15794e-24, 3.92955e-24, 4.95302e-24], [190.733, 7.99528e-24, 9.15758e-24, 3.92574e-24, 4.95127e-24], [191.173, 8.00809e-24, 9.15723e-24, 3.92192e-24, 4.94952e-24], [191.614, 8.02095e-24, 9.15689e-24, 3.91809e-24, 4.94778e-24], [192.056, 8.03385e-24, 9.15655e-24, 3.91427e-24, 4.94604e-24], [192.499, 8.04681e-24, 9.15622e-24, 3.91044e-24, 4.94431e-24], [192.943, 8.0598e-24, 9.15589e-24, 3.90661e-24, 4.94258e-24], [193.388, 8.07285e-24, 9.15557e-24, 3.90277e-24, 4.94086e-24], [193.834, 8.08594e-24, 9.15526e-24, 3.89894e-24, 4.93914e-24], [194.281, 8.09907e-24, 9.15495e-24, 3.8951e-24, 4.93743e-24], [194.729, 8.11226e-24, 9.15465e-24, 3.89126e-24, 4.93573e-24], [195.178, 8.12549e-24, 9.15435e-24, 3.88742e-24, 4.93403e-24], [195.628, 8.13876e-24, 9.15406e-24, 3.88358e-24, 4.93233e-24], [196.079, 8.15209e-24, 9.15378e-24, 3.87973e-24, 4.93064e-24], [196.531, 8.16546e-24, 9.1535e-24, 3.87589e-24, 4.92896e-24], [196.984, 8.17887e-24, 9.15322e-24, 3.87204e-24, 4.92728e-24], [197.438, 8.19234e-24, 9.15296e-24, 3.86819e-24, 4.92561e-24], [197.894, 8.20585e-24, 9.15269e-24, 3.86434e-24, 4.92394e-24], [198.35, 8.21941e-24, 9.15244e-24, 3.86049e-24, 4.92228e-24], [198.808, 8.23301e-24, 9.15219e-24, 3.85664e-24, 4.92062e-24], [199.266, 8.24666e-24, 9.15194e-24, 3.85279e-24, 4.91897e-24], [199.725, 8.26036e-24, 9.15171e-24, 3.84894e-24, 4.91732e-24], [200.186, 8.27411e-24, 9.15147e-24, 3.84508e-24, 4.91568e-24], [200.648, 8.2879e-24, 9.15125e-24, 3.84123e-24, 4.91405e-24], [201.11, 8.30175e-24, 9.15103e-24, 3.83738e-24, 4.91242e-24], [201.574, 8.31563e-24, 9.15081e-24, 3.83352e-24, 4.91079e-24], [202.039, 8.32957e-24, 9.15061e-24, 3.82967e-24, 4.90917e-24], [202.505, 8.34355e-24, 9.1504e-24, 3.82582e-24, 4.90756e-24], [202.972, 8.35759e-24, 9.15021e-24, 3.82197e-24, 4.90595e-24], [203.44, 8.37167e-24, 9.15002e-24, 3.81811e-24, 4.90435e-24], [203.909, 8.38579e-24, 9.14983e-24, 3.81426e-24, 4.90275e-24], [204.379, 8.39997e-24, 9.14966e-24, 3.81041e-24, 4.90116e-24], [204.851, 8.41419e-24, 9.14949e-24, 3.80657e-24, 4.89957e-24], [205.323, 8.42846e-24, 9.14932e-24, 3.80272e-24, 4.89799e-24], [205.796, 8.44278e-24, 9.14916e-24, 3.79888e-24, 4.89642e-24], [206.271, 8.45715e-24, 9.14901e-24, 3.79503e-24, 4.89485e-24], [206.747, 8.47156e-24, 9.14886e-24, 3.79119e-24, 4.89328e-24], [207.223, 8.48603e-24, 9.14872e-24, 3.78735e-24, 4.89172e-24], [207.701, 8.50054e-24, 9.14858e-24, 3.78352e-24, 4.89017e-24], [208.18, 8.5151e-24, 9.14845e-24, 3.77968e-24, 4.88862e-24], [208.66, 8.52971e-24, 9.14833e-24, 3.77585e-24, 4.88707e-24], [209.141, 8.54437e-24, 9.14821e-24, 3.77202e-24, 4.88554e-24], [209.624, 8.55907e-24, 9.1481e-24, 3.7682e-24, 4.884e-24], [210.107, 8.57383e-24, 9.148e-24, 3.76437e-24, 4.88247e-24], [210.592, 8.58863e-24, 9.1479e-24, 3.76056e-24, 4.88095e-24], [211.077, 8.60348e-24, 9.14781e-24, 3.75674e-24, 4.87944e-24], [211.564, 8.61839e-24, 9.14772e-24, 3.75293e-24, 4.87792e-24], [212.052, 8.63334e-24, 9.14765e-24, 3.74912e-24, 4.87642e-24], [212.541, 8.64834e-24, 9.14757e-24, 3.74532e-24, 4.87492e-24], [213.031, 8.66338e-24, 9.14751e-24, 3.74153e-24, 4.87342e-24], [213.522, 8.67848e-24, 9.14745e-24, 3.73773e-24, 4.87193e-24], [214.015, 8.69363e-24, 9.14739e-24, 3.73395e-24, 4.87045e-24], [214.508, 8.70883e-24, 9.14734e-24, 3.73016e-24, 4.86897e-24], [215.003, 8.72407e-24, 9.1473e-24, 3.72639e-24, 4.86749e-24], [215.499, 8.73937e-24, 9.14727e-24, 3.72262e-24, 4.86603e-24], [215.996, 8.75471e-24, 9.14724e-24, 3.71885e-24, 4.86456e-24], [216.494, 8.77011e-24, 9.14722e-24, 3.71509e-24, 4.86311e-24], [216.993, 8.78555e-24, 9.1472e-24, 3.71134e-24, 4.86165e-24], [217.493, 8.80104e-24, 9.14719e-24, 3.7076e-24, 4.86021e-24], [217.995, 8.81659e-24, 9.14719e-24, 3.70386e-24, 4.85876e-24], [218.498, 8.83218e-24, 9.14719e-24, 3.70013e-24, 4.85733e-24], [219.001, 8.84782e-24, 9.1472e-24, 3.6964e-24, 4.8559e-24], [219.506, 8.86352e-24, 9.14722e-24, 3.69269e-24, 4.85447e-24], [220.013, 8.87926e-24, 9.14724e-24, 3.68898e-24, 4.85305e-24], [220.52, 8.89506e-24, 9.14727e-24, 3.68528e-24, 4.85164e-24], [221.028, 8.9109e-24, 9.14731e-24, 3.68159e-24, 4.85023e-24], [221.538, 8.92679e-24, 9.14735e-24, 3.67791e-24, 4.84882e-24], [222.049, 8.94274e-24, 9.1474e-24, 3.67424e-24, 4.84743e-24], [222.561, 8.95874e-24, 9.14746e-24, 3.67058e-24, 4.84603e-24], [223.074, 8.97478e-24, 9.14752e-24, 3.66692e-24, 4.84464e-24], [223.589, 8.99088e-24, 9.14759e-24, 3.66328e-24, 4.84326e-24], [224.104, 9.00703e-24, 9.14766e-24, 3.65965e-24, 4.84189e-24], [224.621, 9.02322e-24, 9.14775e-24, 3.65603e-24, 4.84051e-24], [225.139, 9.03947e-24, 9.14784e-24, 3.65242e-24, 4.83915e-24], [225.658, 9.05577e-24, 9.14793e-24, 3.64882e-24, 4.83779e-24], [226.179, 9.07212e-24, 9.14803e-24, 3.64523e-24, 4.83643e-24], [226.7, 9.08853e-24, 9.14814e-24, 3.64166e-24, 4.83508e-24], [227.223, 9.10498e-24, 9.14826e-24, 3.63809e-24, 4.83374e-24], [227.747, 9.12149e-24, 9.14838e-24, 3.63454e-24, 4.8324e-24], [228.272, 9.13804e-24, 9.14851e-24, 3.631e-24, 4.83106e-24], [228.798, 9.15465e-24, 9.14865e-24, 3.62748e-24, 4.82974e-24], [229.326, 9.17131e-24, 9.14879e-24, 3.62397e-24, 4.82841e-24], [229.855, 9.18802e-24, 9.14894e-24, 3.62047e-24, 4.8271e-24], [230.385, 9.20478e-24, 9.1491e-24, 3.61699e-24, 4.82578e-24], [230.916, 9.2216e-24, 9.14926e-24, 3.61353e-24, 4.82448e-24], [231.449, 9.23846e-24, 9.14943e-24, 3.61007e-24, 4.82318e-24], [231.982, 9.25538e-24, 9.14961e-24, 3.60664e-24, 4.82188e-24], [232.517, 9.27235e-24, 9.14979e-24, 3.60322e-24, 4.82059e-24], [233.054, 9.28937e-24, 9.14998e-24, 3.59981e-24, 4.8193e-24], [233.591, 9.30645e-24, 9.15018e-24, 3.59642e-24, 4.81802e-24], [234.13, 9.32357e-24, 9.15038e-24, 3.59305e-24, 4.81675e-24], [234.67, 9.34075e-24, 9.15059e-24, 3.5897e-24, 4.81548e-24], [235.211, 9.35798e-24, 9.15081e-24, 3.58636e-24, 4.81422e-24], [235.753, 9.37526e-24, 9.15104e-24, 3.58305e-24, 4.81296e-24], [236.297, 9.3926e-24, 9.15127e-24, 3.57975e-24, 4.81171e-24], [236.842, 9.40999e-24, 9.15151e-24, 3.57647e-24, 4.81046e-24], [237.388, 9.42743e-24, 9.15176e-24, 3.57321e-24, 4.80922e-24], [237.935, 9.44493e-24, 9.15201e-24, 3.56997e-24, 4.80798e-24], [238.484, 9.46247e-24, 9.15227e-24, 3.56675e-24, 4.80675e-24], [239.034, 9.48007e-24, 9.15254e-24, 3.56355e-24, 4.80553e-24], [239.585, 9.49773e-24, 9.15282e-24, 3.56037e-24, 4.80431e-24], [240.138, 9.51543e-24, 9.1531e-24, 3.55721e-24, 4.8031e-24], [240.691, 9.53319e-24, 9.15339e-24, 3.55408e-24, 4.80189e-24], [241.246, 9.551e-24, 9.15368e-24, 3.55097e-24, 4.80068e-24], [241.803, 9.56887e-24, 9.15399e-24, 3.54788e-24, 4.79949e-24], [242.36, 9.58679e-24, 9.1543e-24, 3.54481e-24, 4.79829e-24], [242.919, 9.60476e-24, 9.15462e-24, 3.54177e-24, 4.79711e-24], [243.479, 9.62279e-24, 9.15494e-24, 3.53875e-24, 4.79593e-24], [244.041, 9.64087e-24, 9.15528e-24, 3.53576e-24, 4.79475e-24], [244.604, 9.659e-24, 9.15562e-24, 3.53279e-24, 4.79358e-24], [245.168, 9.67719e-24, 9.15596e-24, 3.52985e-24, 4.79242e-24], [245.733, 9.69543e-24, 9.15632e-24, 3.52693e-24, 4.79126e-24], [246.3, 9.71373e-24, 9.15668e-24, 3.52404e-24, 4.7901e-24], [246.868, 9.73208e-24, 9.15705e-24, 3.52118e-24, 4.78895e-24], [247.437, 9.75049e-24, 9.15743e-24, 3.51835e-24, 4.78781e-24], [248.007, 9.76894e-24, 9.15781e-24, 3.51554e-24, 4.78667e-24], [248.579, 9.78746e-24, 9.1582e-24, 3.51277e-24, 4.78554e-24], [249.153, 9.80603e-24, 9.1586e-24, 3.51002e-24, 4.78442e-24], [249.727, 9.82465e-24, 9.15901e-24, 3.5073e-24, 4.7833e-24], [250.303, 9.84333e-24, 9.15942e-24, 3.50461e-24, 4.78218e-24], [250.88, 9.86206e-24, 9.15984e-24, 3.50196e-24, 4.78107e-24], [251.459, 9.88084e-24, 9.16027e-24, 3.49933e-24, 4.77997e-24], [252.039, 9.89969e-24, 9.16071e-24, 3.49674e-24, 4.77887e-24], [252.62, 9.91858e-24, 9.16116e-24, 3.49418e-24, 4.77778e-24], [253.202, 9.93754e-24, 9.16161e-24, 3.49165e-24, 4.77669e-24], [253.786, 9.95654e-24, 9.16207e-24, 3.48916e-24, 4.77561e-24], [254.372, 9.97561e-24, 9.16253e-24, 3.4867e-24, 4.77453e-24], [254.958, 9.99472e-24, 9.16301e-24, 3.48428e-24, 4.77346e-24], [255.546, 1.00139e-23, 9.16349e-24, 3.48189e-24, 4.7724e-24], [256.135, 1.00331e-23, 9.16398e-24, 3.47954e-24, 4.77134e-24], [256.726, 1.00524e-23, 9.16448e-24, 3.47722e-24, 4.77028e-24], [257.318, 1.00718e-23, 9.16499e-24, 3.47494e-24, 4.76924e-24], [257.911, 1.00911e-23, 9.1655e-24, 3.4727e-24, 4.76819e-24], [258.506, 1.01106e-23, 9.16602e-24, 3.4705e-24, 4.76716e-24], [259.102, 1.01301e-23, 9.16655e-24, 3.46834e-24, 4.76612e-24], [259.7, 1.01497e-23, 9.16709e-24, 3.46622e-24, 4.7651e-24], [260.299, 1.01693e-23, 9.16763e-24, 3.46413e-24, 4.76408e-24], [260.899, 1.0189e-23, 9.16819e-24, 3.46209e-24, 4.76306e-24], [261.5, 1.02087e-23, 9.16875e-24, 3.46009e-24, 4.76206e-24], [262.103, 1.02285e-23, 9.16932e-24, 3.45813e-24, 4.76105e-24], [262.708, 1.02483e-23, 9.16989e-24, 3.45622e-24, 4.76006e-24], [263.314, 1.02682e-23, 9.17048e-24, 3.45435e-24, 4.75906e-24], [263.921, 1.02882e-23, 9.17107e-24, 3.45252e-24, 4.75808e-24], [264.53, 1.03082e-23, 9.17167e-24, 3.45074e-24, 4.7571e-24], [265.14, 1.03283e-23, 9.17228e-24, 3.449e-24, 4.75612e-24], [265.751, 1.03484e-23, 9.1729e-24, 3.44731e-24, 4.75515e-24], [266.364, 1.03686e-23, 9.17352e-24, 3.44567e-24, 4.75419e-24], [266.978, 1.03889e-23, 9.17416e-24, 3.44408e-24, 4.75323e-24], [267.594, 1.04092e-23, 9.1748e-24, 3.44253e-24, 4.75228e-24], [268.211, 1.04295e-23, 9.17545e-24, 3.44103e-24, 4.75133e-24], [268.829, 1.04499e-23, 9.1761e-24, 3.43958e-24, 4.75039e-24], [269.449, 1.04704e-23, 9.17677e-24, 3.43819e-24, 4.74946e-24], [270.071, 1.0491e-23, 9.17744e-24, 3.43684e-24, 4.74853e-24], [270.693, 1.05115e-23, 9.17813e-24, 3.43555e-24, 4.74761e-24], [271.318, 1.05322e-23, 9.17882e-24, 3.43431e-24, 4.74669e-24], [271.943, 1.05529e-23, 9.17952e-24, 3.43312e-24, 4.74578e-24], [272.57, 1.05737e-23, 9.18022e-24, 3.43198e-24, 4.74487e-24], [273.199, 1.05945e-23, 9.18094e-24, 3.4309e-24, 4.74397e-24], [273.829, 1.06154e-23, 9.18166e-24, 3.42988e-24, 4.74307e-24], [274.46, 1.06363e-23, 9.1824e-24, 3.42891e-24, 4.74218e-24], [275.093, 1.06573e-23, 9.18314e-24, 3.428e-24, 4.7413e-24], [275.728, 1.06784e-23, 9.18389e-24, 3.42715e-24, 4.74042e-24], [276.363, 1.06995e-23, 9.18465e-24, 3.42636e-24, 4.73955e-24], [277.001, 1.07207e-23, 9.18541e-24, 3.42563e-24, 4.73869e-24], [277.639, 1.07419e-23, 9.18619e-24, 3.42495e-24, 4.73783e-24], [278.28, 1.07632e-23, 9.18697e-24, 3.42434e-24, 4.73697e-24], [278.921, 1.07845e-23, 9.18776e-24, 3.42379e-24, 4.73612e-24], [279.565, 1.0806e-23, 9.18857e-24, 3.4233e-24, 4.73528e-24], [280.209, 1.08274e-23, 9.18937e-24, 3.42288e-24, 4.73445e-24], [280.855, 1.0849e-23, 9.19019e-24, 3.42251e-24, 4.73361e-24], [281.503, 1.08705e-23, 9.19102e-24, 3.42222e-24, 4.73279e-24], [282.152, 1.08922e-23, 9.19186e-24, 3.42199e-24, 4.73197e-24], [282.803, 1.09139e-23, 9.1927e-24, 3.42183e-24, 4.73116e-24], [283.455, 1.09357e-23, 9.19355e-24, 3.42173e-24, 4.73035e-24], [284.109, 1.09575e-23, 9.19442e-24, 3.4217e-24, 4.72955e-24], [284.764, 1.09794e-23, 9.19529e-24, 3.42174e-24, 4.72875e-24], [285.421, 1.10013e-23, 9.19617e-24, 3.42185e-24, 4.72796e-24], [286.079, 1.10234e-23, 9.19706e-24, 3.42203e-24, 4.72718e-24], [286.738, 1.10454e-23, 9.19795e-24, 3.42229e-24, 4.7264e-24], [287.4, 1.10676e-23, 9.19886e-24, 3.42261e-24, 4.72563e-24], [288.062, 1.10898e-23, 9.19978e-24, 3.42301e-24, 4.72486e-24], [288.727, 1.1112e-23, 9.2007e-24, 3.42348e-24, 4.7241e-24], [289.392, 1.11343e-23, 9.20163e-24, 3.42403e-24, 4.72335e-24], [290.06, 1.11567e-23, 9.20258e-24, 3.42465e-24, 4.7226e-24], [290.729, 1.11791e-23, 9.20353e-24, 3.42535e-24, 4.72186e-24], [291.399, 1.12016e-23, 9.20449e-24, 3.42613e-24, 4.72112e-24], [292.071, 1.12242e-23, 9.20546e-24, 3.42699e-24, 4.72039e-24], [292.745, 1.12468e-23, 9.20644e-24, 3.42792e-24, 4.71967e-24], [293.42, 1.12695e-23, 9.20743e-24, 3.42893e-24, 4.71895e-24], [294.096, 1.12922e-23, 9.20843e-24, 3.43003e-24, 4.71824e-24], [294.774, 1.1315e-23, 9.20943e-24, 3.43121e-24, 4.71753e-24], [295.454, 1.13379e-23, 9.21045e-24, 3.43246e-24, 4.71683e-24], [296.136, 1.13608e-23, 9.21148e-24, 3.43381e-24, 4.71614e-24], [296.818, 1.13838e-23, 9.21251e-24, 3.43523e-24, 4.71545e-24], [297.503, 1.14069e-23, 9.21356e-24, 3.43674e-24, 4.71477e-24], [298.189, 1.143e-23, 9.21461e-24, 3.43834e-24, 4.71409e-24], [298.877, 1.14532e-23, 9.21567e-24, 3.44002e-24, 4.71342e-24], [299.566, 1.14764e-23, 9.21675e-24, 3.4418e-24, 4.71276e-24], [300.257, 1.14997e-23, 9.21783e-24, 3.44366e-24, 4.7121e-24], [300.949, 1.15231e-23, 9.21892e-24, 3.44561e-24, 4.71145e-24], [301.643, 1.15465e-23, 9.22002e-24, 3.44765e-24, 4.7108e-24], [302.339, 1.157e-23, 9.22113e-24, 3.44978e-24, 4.71016e-24], [303.036, 1.15936e-23, 9.22225e-24, 3.452e-24, 4.70953e-24], [303.735, 1.16172e-23, 9.22339e-24, 3.45432e-24, 4.7089e-24], [304.435, 1.16409e-23, 9.22453e-24, 3.45672e-24, 4.70828e-24], [305.137, 1.16646e-23, 9.22568e-24, 3.45923e-24, 4.70767e-24], [305.841, 1.16884e-23, 9.22684e-24, 3.46183e-24, 4.70706e-24], [306.546, 1.17123e-23, 9.228e-24, 3.46452e-24, 4.70646e-24], [307.253, 1.17362e-23, 9.22918e-24, 3.46732e-24, 4.70586e-24], [307.961, 1.17602e-23, 9.23037e-24, 3.47021e-24, 4.70527e-24], [308.671, 1.17843e-23, 9.23157e-24, 3.47319e-24, 4.70469e-24], [309.383, 1.18084e-23, 9.23278e-24, 3.47628e-24, 4.70411e-24], [310.097, 1.18326e-23, 9.234e-24, 3.47947e-24, 4.70354e-24], [310.812, 1.18569e-23, 9.23523e-24, 3.48276e-24, 4.70297e-24], [311.529, 1.18812e-23, 9.23647e-24, 3.48616e-24, 4.70241e-24], [312.247, 1.19056e-23, 9.23772e-24, 3.48965e-24, 4.70186e-24], [312.967, 1.193e-23, 9.23898e-24, 3.49325e-24, 4.70132e-24], [313.689, 1.19545e-23, 9.24025e-24, 3.49696e-24, 4.70078e-24], [314.412, 1.19791e-23, 9.24153e-24, 3.50077e-24, 4.70024e-24], [315.137, 1.20038e-23, 9.24281e-24, 3.50469e-24, 4.69971e-24], [315.864, 1.20285e-23, 9.24411e-24, 3.50871e-24, 4.69919e-24], [316.592, 1.20533e-23, 9.24542e-24, 3.51284e-24, 4.69868e-24], [317.322, 1.20781e-23, 9.24674e-24, 3.51708e-24, 4.69817e-24], [318.054, 1.2103e-23, 9.24808e-24, 3.52144e-24, 4.69767e-24], [318.787, 1.2128e-23, 9.24942e-24, 3.5259e-24, 4.69717e-24], [319.523, 1.2153e-23, 9.25077e-24, 3.53047e-24, 4.69668e-24], [320.259, 1.21781e-23, 9.25213e-24, 3.53516e-24, 4.6962e-24], [320.998, 1.22033e-23, 9.2535e-24, 3.53995e-24, 4.69573e-24], [321.738, 1.22285e-23, 9.25488e-24, 3.54487e-24, 4.69526e-24], [322.48, 1.22538e-23, 9.25628e-24, 3.54989e-24, 4.69479e-24], [323.224, 1.22792e-23, 9.25768e-24, 3.55503e-24, 4.69434e-24], [323.969, 1.23046e-23, 9.2591e-24, 3.56029e-24, 4.69388e-24], [324.716, 1.23301e-23, 9.26052e-24, 3.56567e-24, 4.69344e-24], [325.465, 1.23557e-23, 9.26196e-24, 3.57116e-24, 4.693e-24], [326.215, 1.23814e-23, 9.2634e-24, 3.57677e-24, 4.69257e-24], [326.968, 1.24071e-23, 9.26486e-24, 3.5825e-24, 4.69215e-24], [327.722, 1.24328e-23, 9.26633e-24, 3.58835e-24, 4.69173e-24], [328.477, 1.24587e-23, 9.26781e-24, 3.59432e-24, 4.69132e-24], [329.235, 1.24846e-23, 9.2693e-24, 3.60042e-24, 4.69091e-24], [329.994, 1.25105e-23, 9.2708e-24, 3.60663e-24, 4.69051e-24], [330.755, 1.25366e-23, 9.27231e-24, 3.61297e-24, 4.69012e-24], [331.518, 1.25627e-23, 9.27383e-24, 3.61943e-24, 4.68974e-24], [332.282, 1.25889e-23, 9.27537e-24, 3.62602e-24, 4.68936e-24], [333.049, 1.26151e-23, 9.27691e-24, 3.63273e-24, 4.68899e-24], [333.817, 1.26414e-23, 9.27847e-24, 3.63957e-24, 4.68862e-24], [334.586, 1.26678e-23, 9.28003e-24, 3.64653e-24, 4.68826e-24], [335.358, 1.26943e-23, 9.28161e-24, 3.65362e-24, 4.68791e-24], [336.131, 1.27208e-23, 9.2832e-24, 3.66084e-24, 4.68756e-24], [336.906, 1.27474e-23, 9.2848e-24, 3.66819e-24, 4.68723e-24], [337.683, 1.2774e-23, 9.28641e-24, 3.67567e-24, 4.68689e-24], [338.462, 1.28008e-23, 9.28803e-24, 3.68328e-24, 4.68657e-24], [339.242, 1.28276e-23, 9.28967e-24, 3.69102e-24, 4.68625e-24], [340.025, 1.28544e-23, 9.29131e-24, 3.69889e-24, 4.68594e-24], [340.809, 1.28814e-23, 9.29297e-24, 3.70689e-24, 4.68563e-24], [341.595, 1.29084e-23, 9.29464e-24, 3.71502e-24, 4.68533e-24], [342.382, 1.29354e-23, 9.29632e-24, 3.72329e-24, 4.68504e-24], [343.172, 1.29626e-23, 9.29801e-24, 3.73169e-24, 4.68476e-24], [343.963, 1.29898e-23, 9.29971e-24, 3.74023e-24, 4.68448e-24], [344.757, 1.30171e-23, 9.30143e-24, 3.7489e-24, 4.68421e-24], [345.552, 1.30444e-23, 9.30315e-24, 3.75771e-24, 4.68394e-24], [346.348, 1.30719e-23, 9.30489e-24, 3.76666e-24, 4.68369e-24], [347.147, 1.30994e-23, 9.30664e-24, 3.77574e-24, 4.68343e-24], [347.948, 1.31269e-23, 9.3084e-24, 3.78495e-24, 4.68319e-24], [348.75, 1.31546e-23, 9.31018e-24, 3.79431e-24, 4.68295e-24], [349.554, 1.31823e-23, 9.31196e-24, 3.80381e-24, 4.68272e-24], [350.36, 1.321e-23, 9.31376e-24, 3.81344e-24, 4.6825e-24], [351.168, 1.32379e-23, 9.31557e-24, 3.82322e-24, 4.68228e-24], [351.978, 1.32658e-23, 9.31739e-24, 3.83313e-24, 4.68207e-24], [352.79, 1.32938e-23, 9.31922e-24, 3.84319e-24, 4.68187e-24], [353.603, 1.33219e-23, 9.32107e-24, 3.85339e-24, 4.68168e-24], [354.419, 1.335e-23, 9.32292e-24, 3.86373e-24, 4.68149e-24], [355.236, 1.33782e-23, 9.32479e-24, 3.87421e-24, 4.68131e-24], [356.055, 1.34065e-23, 9.32667e-24, 3.88483e-24, 4.68113e-24], [356.876, 1.34348e-23, 9.32857e-24, 3.8956e-24, 4.68097e-24], [357.699, 1.34633e-23, 9.33047e-24, 3.90651e-24, 4.68081e-24], [358.524, 1.34918e-23, 9.33239e-24, 3.91757e-24, 4.68065e-24], [359.351, 1.35203e-23, 9.33432e-24, 3.92877e-24, 4.68051e-24], [360.179, 1.3549e-23, 9.33626e-24, 3.94012e-24, 4.68037e-24], [361.01, 1.35777e-23, 9.33822e-24, 3.95161e-24, 4.68024e-24], [361.842, 1.36065e-23, 9.34019e-24, 3.96325e-24, 4.68011e-24], [362.677, 1.36354e-23, 9.34217e-24, 3.97504e-24, 4.67999e-24], [363.513, 1.36643e-23, 9.34416e-24, 3.98697e-24, 4.67988e-24], [364.351, 1.36933e-23, 9.34617e-24, 3.99905e-24, 4.67978e-24], [365.192, 1.37224e-23, 9.34818e-24, 4.01128e-24, 4.67968e-24], [366.034, 1.37515e-23, 9.35021e-24, 4.02366e-24, 4.6796e-24], [366.878, 1.37808e-23, 9.35226e-24, 4.03619e-24, 4.67951e-24], [367.724, 1.38101e-23, 9.35431e-24, 4.04886e-24, 4.67944e-24], [368.572, 1.38394e-23, 9.35638e-24, 4.06169e-24, 4.67937e-24], [369.422, 1.38689e-23, 9.35847e-24, 4.07467e-24, 4.67931e-24], [370.274, 1.38984e-23, 9.36056e-24, 4.08779e-24, 4.67926e-24], [371.128, 1.3928e-23, 9.36267e-24, 4.10107e-24, 4.67922e-24], [371.983, 1.39577e-23, 9.36479e-24, 4.1145e-24, 4.67918e-24], [372.841, 1.39875e-23, 9.36692e-24, 4.12808e-24, 4.67915e-24], [373.701, 1.40173e-23, 9.36907e-24, 4.14181e-24, 4.67913e-24], [374.563, 1.40472e-23, 9.37123e-24, 4.15569e-24, 4.67911e-24], [375.426, 1.40772e-23, 9.3734e-24, 4.16973e-24, 4.6791e-24], [376.292, 1.41072e-23, 9.37559e-24, 4.18392e-24, 4.6791e-24], [377.16, 1.41374e-23, 9.37779e-24, 4.19826e-24, 4.67911e-24], [378.03, 1.41676e-23, 9.38e-24, 4.21276e-24, 4.67912e-24], [378.901, 1.41978e-23, 9.38223e-24, 4.22741e-24, 4.67915e-24], [379.775, 1.42282e-23, 9.38447e-24, 4.24222e-24, 4.67918e-24], [380.651, 1.42586e-23, 9.38673e-24, 4.25717e-24, 4.67921e-24], [381.529, 1.42891e-23, 9.38899e-24, 4.27229e-24, 4.67926e-24], [382.408, 1.43197e-23, 9.39127e-24, 4.28756e-24, 4.67931e-24], [383.29, 1.43504e-23, 9.39357e-24, 4.30298e-24, 4.67937e-24], [384.174, 1.43811e-23, 9.39588e-24, 4.31856e-24, 4.67944e-24], [385.06, 1.4412e-23, 9.3982e-24, 4.3343e-24, 4.67951e-24], [385.948, 1.44428e-23, 9.40053e-24, 4.35019e-24, 4.6796e-24], [386.838, 1.44738e-23, 9.40288e-24, 4.36624e-24, 4.67969e-24], [387.73, 1.45049e-23, 9.40525e-24, 4.38245e-24, 4.67978e-24], [388.624, 1.4536e-23, 9.40762e-24, 4.39881e-24, 4.67989e-24], [389.52, 1.45672e-23, 9.41001e-24, 4.41533e-24, 4.68001e-24], [390.419, 1.45985e-23, 9.41242e-24, 4.43201e-24, 4.68013e-24], [391.319, 1.46299e-23, 9.41484e-24, 4.44884e-24, 4.68026e-24], [392.221, 1.46613e-23, 9.41727e-24, 4.46583e-24, 4.68039e-24], [393.126, 1.46928e-23, 9.41972e-24, 4.48299e-24, 4.68054e-24], [394.032, 1.47244e-23, 9.42218e-24, 4.5003e-24, 4.68069e-24], [394.941, 1.47561e-23, 9.42466e-24, 4.51776e-24, 4.68086e-24], [395.852, 1.47878e-23, 9.42715e-24, 4.53539e-24, 4.68102e-24], [396.765, 1.48197e-23, 9.42965e-24, 4.55318e-24, 4.6812e-24], [397.679, 1.48516e-23, 9.43217e-24, 4.57112e-24, 4.68139e-24], [398.597, 1.48836e-23, 9.43471e-24, 4.58923e-24, 4.68158e-24], [399.516, 1.49157e-23, 9.43726e-24, 4.6075e-24, 4.68178e-24], [400.437, 1.49478e-23, 9.43982e-24, 4.62592e-24, 4.68199e-24], [401.36, 1.49801e-23, 9.4424e-24, 4.64451e-24, 4.68221e-24], [402.286, 1.50124e-23, 9.44499e-24, 4.66325e-24, 4.68244e-24], [403.214, 1.50448e-23, 9.4476e-24, 4.68216e-24, 4.68268e-24], [404.143, 1.50772e-23, 9.45022e-24, 4.70123e-24, 4.68292e-24], [405.075, 1.51098e-23, 9.45286e-24, 4.72045e-24, 4.68317e-24], [406.009, 1.51424e-23, 9.45551e-24, 4.73984e-24, 4.68343e-24], [406.946, 1.51752e-23, 9.45818e-24, 4.7594e-24, 4.6837e-24], [407.884, 1.5208e-23, 9.46087e-24, 4.77911e-24, 4.68398e-24], [408.825, 1.52408e-23, 9.46357e-24, 4.79898e-24, 4.68427e-24], [409.767, 1.52738e-23, 9.46628e-24, 4.81902e-24, 4.68457e-24], [410.712, 1.53068e-23, 9.46901e-24, 4.83922e-24, 4.68487e-24], [411.66, 1.534e-23, 9.47176e-24, 4.85959e-24, 4.68519e-24], [412.609, 1.53732e-23, 9.47452e-24, 4.88011e-24, 4.68552e-24], [413.56, 1.54065e-23, 9.4773e-24, 4.9008e-24, 4.68585e-24], [414.514, 1.54399e-23, 9.48009e-24, 4.92166e-24, 4.6862e-24], [415.47, 1.54733e-23, 9.48291e-24, 4.94268e-24, 4.68656e-24], [416.428, 1.55069e-23, 9.48573e-24, 4.96386e-24, 4.68693e-24], [417.388, 1.55405e-23, 9.48858e-24, 4.98521e-24, 4.68731e-24], [418.351, 1.55742e-23, 9.49144e-24, 5.00673e-24, 4.6877e-24], [419.315, 1.5608e-23, 9.49432e-24, 5.02841e-24, 4.68811e-24], [420.282, 1.56419e-23, 9.49723e-24, 5.05026e-24, 4.68853e-24], [421.252, 1.56758e-23, 9.50015e-24, 5.07229e-24, 4.68897e-24], [422.223, 1.57099e-23, 9.50309e-24, 5.09448e-24, 4.68942e-24], [423.197, 1.5744e-23, 9.50605e-24, 5.11685e-24, 4.6899e-24], [424.173, 1.57783e-23, 9.50904e-24, 5.1394e-24, 4.6904e-24], [425.151, 1.58126e-23, 9.51205e-24, 5.16213e-24, 4.69094e-24], [426.131, 1.5847e-23, 9.51509e-24, 5.18505e-24, 4.69151e-24], [427.114, 1.58815e-23, 9.51817e-24, 5.20817e-24, 4.69213e-24], [428.099, 1.59161e-23, 9.5213e-24, 5.2315e-24, 4.69281e-24], [429.086, 1.59508e-23, 9.52448e-24, 5.25507e-24, 4.69359e-24], [430.075, 1.59857e-23, 9.52774e-24, 5.27891e-24, 4.6945e-24], [431.067, 1.60207e-23, 9.53111e-24, 5.30309e-24, 4.69563e-24], [432.061, 1.60558e-23, 9.53468e-24, 5.32774e-24, 4.69711e-24], [433.057, 1.60913e-23, 9.53858e-24, 5.35312e-24, 4.69925e-24], [434.056, 1.61272e-23, 9.54317e-24, 5.37986e-24, 4.70278e-24], [435.057, 1.61643e-23, 9.54951e-24, 5.40978e-24, 4.70981e-24], [436.06, 1.62049e-23, 9.5618e-24, 5.45024e-24, 4.72887e-24], [437.066, 1.62694e-23, 9.61449e-24, 5.56082e-24, 4.8288e-24], [438.074, 1.06875e-22, 1.06061e-22, 1.05772e-22, 1.05735e-22], [439.084, 1.63376e-23, 9.6166e-24, 5.60229e-24, 4.82147e-24], [440.096, 1.63459e-23, 9.57339e-24, 5.54721e-24, 4.72876e-24], [441.111, 1.63765e-23, 9.56794e-24, 5.55746e-24, 4.71176e-24], [442.129, 1.64105e-23, 9.56815e-24, 5.57765e-24, 4.70618e-24], [443.148, 1.64455e-23, 9.57001e-24, 5.60083e-24, 4.70392e-24], [444.17, 1.6481e-23, 9.57253e-24, 5.62529e-24, 4.70297e-24], [445.194, 1.65167e-23, 9.57536e-24, 5.65044e-24, 4.70264e-24], [446.221, 1.65526e-23, 9.57837e-24, 5.67603e-24, 4.70265e-24], [447.25, 1.65887e-23, 9.5815e-24, 5.70194e-24, 4.70285e-24], [448.281, 1.66249e-23, 9.58469e-24, 5.72812e-24, 4.70318e-24], [449.315, 1.66612e-23, 9.58794e-24, 5.75454e-24, 4.7036e-24], [450.351, 1.66976e-23, 9.59124e-24, 5.78116e-24, 4.70407e-24], [451.39, 1.67341e-23, 9.59457e-24, 5.80798e-24, 4.7046e-24], [452.43, 1.67707e-23, 9.59793e-24, 5.83499e-24, 4.70515e-24], [453.474, 1.68074e-23, 9.60131e-24, 5.86218e-24, 4.70574e-24], [454.519, 1.68442e-23, 9.60473e-24, 5.88955e-24, 4.70635e-24], [455.568, 1.68811e-23, 9.60816e-24, 5.91709e-24, 4.70698e-24], [456.618, 1.69181e-23, 9.61162e-24, 5.94481e-24, 4.70763e-24], [457.671, 1.69551e-23, 9.61509e-24, 5.97271e-24, 4.70829e-24], [458.727, 1.69923e-23, 9.61859e-24, 6.00077e-24, 4.70897e-24], [459.784, 1.70295e-23, 9.62211e-24, 6.02901e-24, 4.70966e-24], [460.845, 1.70669e-23, 9.62565e-24, 6.05742e-24, 4.71037e-24], [461.907, 1.71043e-23, 9.62921e-24, 6.086e-24, 4.71108e-24], [462.972, 1.71418e-23, 9.63278e-24, 6.11475e-24, 4.71181e-24], [464.04, 1.71795e-23, 9.63638e-24, 6.14366e-24, 4.71255e-24], [465.11, 1.72172e-23, 9.64e-24, 6.17275e-24, 4.71331e-24], [466.183, 1.7255e-23, 9.64363e-24, 6.20201e-24, 4.71407e-24], [467.258, 1.72929e-23, 9.64728e-24, 6.23144e-24, 4.71484e-24], [468.335, 1.73309e-23, 9.65096e-24, 6.26103e-24, 4.71563e-24], [469.415, 1.73689e-23, 9.65465e-24, 6.2908e-24, 4.71642e-24], [470.498, 1.74071e-23, 9.65836e-24, 6.32073e-24, 4.71723e-24], [471.583, 1.74454e-23, 9.66209e-24, 6.35084e-24, 4.71804e-24], [472.67, 1.74837e-23, 9.66583e-24, 6.38111e-24, 4.71887e-24], [473.76, 1.75222e-23, 9.6696e-24, 6.41155e-24, 4.71971e-24], [474.853, 1.75607e-23, 9.67339e-24, 6.44217e-24, 4.72055e-24], [475.948, 1.75994e-23, 9.67719e-24, 6.47295e-24, 4.72141e-24], [477.045, 1.76381e-23, 9.68102e-24, 6.5039e-24, 4.72227e-24], [478.145, 1.76769e-23, 9.68486e-24, 6.53502e-24, 4.72315e-24], [479.248, 1.77158e-23, 9.68872e-24, 6.56632e-24, 4.72404e-24], [480.353, 1.77548e-23, 9.69261e-24, 6.59778e-24, 4.72493e-24], [481.461, 1.7794e-23, 9.69651e-24, 6.62941e-24, 4.72584e-24], [482.571, 1.78332e-23, 9.70043e-24, 6.66122e-24, 4.72676e-24], [483.684, 1.78725e-23, 9.70437e-24, 6.69319e-24, 4.72768e-24], [484.799, 1.79118e-23, 9.70833e-24, 6.72534e-24, 4.72862e-24], [485.917, 1.79513e-23, 9.71231e-24, 6.75765e-24, 4.72957e-24], [487.038, 1.79909e-23, 9.7163e-24, 6.79014e-24, 4.73052e-24], [488.161, 1.80306e-23, 9.72032e-24, 6.8228e-24, 4.73149e-24], [489.286, 1.80704e-23, 9.72436e-24, 6.85563e-24, 4.73247e-24], [490.415, 1.81102e-23, 9.72842e-24, 6.88863e-24, 4.73345e-24], [491.546, 1.81502e-23, 9.7325e-24, 6.9218e-24, 4.73445e-24], [492.679, 1.81903e-23, 9.73659e-24, 6.95514e-24, 4.73546e-24], [493.815, 1.82304e-23, 9.74071e-24, 6.98866e-24, 4.73648e-24], [494.954, 1.82707e-23, 9.74485e-24, 7.02234e-24, 4.7375e-24], [496.095, 1.8311e-23, 9.749e-24, 7.0562e-24, 4.73854e-24], [497.239, 1.83515e-23, 9.75318e-24, 7.09024e-24, 4.73959e-24], [498.386, 1.8392e-23, 9.75738e-24, 7.12444e-24, 4.74065e-24], [499.535, 1.84327e-23, 9.7616e-24, 7.15882e-24, 4.74172e-24], [500.687, 1.84734e-23, 9.76584e-24, 7.19337e-24, 4.7428e-24], [501.842, 1.85143e-23, 9.7701e-24, 7.22809e-24, 4.74389e-24], [502.999, 1.85552e-23, 9.77438e-24, 7.26298e-24, 4.74499e-24], [504.159, 1.85962e-23, 9.77868e-24, 7.29805e-24, 4.7461e-24], [505.322, 1.86374e-23, 9.783e-24, 7.3333e-24, 4.74722e-24], [506.487, 1.86786e-23, 9.78734e-24, 7.36871e-24, 4.74835e-24], [507.655, 1.872e-23, 9.7917e-24, 7.4043e-24, 4.74949e-24], [508.825, 1.87614e-23, 9.79608e-24, 7.44006e-24, 4.75064e-24], [509.999, 1.88029e-23, 9.80049e-24, 7.476e-24, 4.75181e-24], [511.175, 1.88446e-23, 9.80491e-24, 7.51211e-24, 4.75298e-24], [512.354, 1.88863e-23, 9.80936e-24, 7.5484e-24, 4.75416e-24], [513.535, 1.89281e-23, 9.81382e-24, 7.58486e-24, 4.75536e-24], [514.719, 1.89701e-23, 9.81831e-24, 7.6215e-24, 4.75656e-24], [515.906, 1.90121e-23, 9.82282e-24, 7.65831e-24, 4.75778e-24], [517.096, 1.90542e-23, 9.82735e-24, 7.69529e-24, 4.759e-24], [518.288, 1.90965e-23, 9.8319e-24, 7.73246e-24, 4.76024e-24], [519.484, 1.91388e-23, 9.83648e-24, 7.76979e-24, 4.76149e-24], [520.682, 1.91813e-23, 9.84107e-24, 7.80731e-24, 4.76275e-24], [521.882, 1.92238e-23, 9.84569e-24, 7.84499e-24, 4.76402e-24], [523.086, 1.92664e-23, 9.85033e-24, 7.88286e-24, 4.7653e-24], [524.292, 1.93092e-23, 9.85499e-24, 7.9209e-24, 4.76659e-24], [525.501, 1.9352e-23, 9.85967e-24, 7.95912e-24, 4.76789e-24], [526.713, 1.9395e-23, 9.86437e-24, 7.99751e-24, 4.76921e-24], [527.927, 1.9438e-23, 9.8691e-24, 8.03608e-24, 4.77053e-24], [529.145, 1.94812e-23, 9.87385e-24, 8.07483e-24, 4.77186e-24], [530.365, 1.95245e-23, 9.87862e-24, 8.11376e-24, 4.77321e-24], [531.588, 1.95678e-23, 9.88341e-24, 8.15286e-24, 4.77457e-24], [532.814, 1.96113e-23, 9.88823e-24, 8.19214e-24, 4.77594e-24], [534.043, 1.96549e-23, 9.89306e-24, 8.2316e-24, 4.77732e-24], [535.274, 1.96985e-23, 9.89792e-24, 8.27124e-24, 4.77871e-24], [536.508, 1.97423e-23, 9.9028e-24, 8.31106e-24, 4.78011e-24], [537.746, 1.97862e-23, 9.90771e-24, 8.35105e-24, 4.78152e-24], [538.986, 1.98302e-23, 9.91264e-24, 8.39122e-24, 4.78295e-24], [540.229, 1.98743e-23, 9.91759e-24, 8.43158e-24, 4.78438e-24], [541.474, 1.99185e-23, 9.92256e-24, 8.47211e-24, 4.78583e-24], [542.723, 1.99628e-23, 9.92755e-24, 8.51282e-24, 4.78729e-24], [543.974, 2.00072e-23, 9.93257e-24, 8.55371e-24, 4.78876e-24], [545.229, 2.00517e-23, 9.93762e-24, 8.59478e-24, 4.79024e-24], [546.486, 2.00964e-23, 9.94268e-24, 8.63603e-24, 4.79173e-24], [547.746, 2.01411e-23, 9.94777e-24, 8.67746e-24, 4.79323e-24], [549.01, 2.01859e-23, 9.95288e-24, 8.71907e-24, 4.79475e-24], [550.276, 2.02309e-23, 9.95801e-24, 8.76087e-24, 4.79628e-24], [551.544, 2.02759e-23, 9.96317e-24, 8.80284e-24, 4.79781e-24], [552.816, 2.03211e-23, 9.96835e-24, 8.84499e-24, 4.79936e-24], [554.091, 2.03664e-23, 9.97356e-24, 8.88733e-24, 4.80093e-24], [555.369, 2.04117e-23, 9.97879e-24, 8.92985e-24, 4.8025e-24], [556.65, 2.04572e-23, 9.98404e-24, 8.97255e-24, 4.80409e-24], [557.933, 2.05028e-23, 9.98932e-24, 9.01543e-24, 4.80568e-24], [559.22, 2.05485e-23, 9.99462e-24, 9.0585e-24, 4.80729e-24], [560.509, 2.05943e-23, 9.99994e-24, 9.10174e-24, 4.80891e-24], [561.802, 2.06402e-23, 1.00053e-23, 9.14517e-24, 4.81054e-24], [563.097, 2.06863e-23, 1.00107e-23, 9.18879e-24, 4.81219e-24], [564.396, 2.07324e-23, 1.00161e-23, 9.23258e-24, 4.81385e-24], [565.697, 2.07787e-23, 1.00215e-23, 9.27656e-24, 4.81551e-24], [567.002, 2.0825e-23, 1.00269e-23, 9.32073e-24, 4.81719e-24], [568.309, 2.08715e-23, 1.00324e-23, 9.36507e-24, 4.81889e-24], [569.62, 2.09181e-23, 1.00379e-23, 9.40961e-24, 4.82059e-24], [570.933, 2.09648e-23, 1.00434e-23, 9.45432e-24, 4.82231e-24], [572.25, 2.10116e-23, 1.0049e-23, 9.49923e-24, 4.82404e-24], [573.57, 2.10585e-23, 1.00545e-23, 9.54431e-24, 4.82578e-24], [574.892, 2.11055e-23, 1.00601e-23, 9.58959e-24, 4.82753e-24], [576.218, 2.11526e-23, 1.00657e-23, 9.63504e-24, 4.82929e-24], [577.547, 2.11999e-23, 1.00714e-23, 9.68069e-24, 4.83107e-24], [578.879, 2.12472e-23, 1.00771e-23, 9.72652e-24, 4.83286e-24], [580.214, 2.12947e-23, 1.00827e-23, 9.77253e-24, 4.83466e-24], [581.552, 2.13423e-23, 1.00885e-23, 9.81874e-24, 4.83648e-24], [582.893, 2.139e-23, 1.00942e-23, 9.86513e-24, 4.8383e-24], [584.237, 2.14378e-23, 1.01e-23, 9.9117e-24, 4.84014e-24], [585.584, 2.14857e-23, 1.01058e-23, 9.95847e-24, 4.842e-24], [586.934, 2.15338e-23, 1.01116e-23, 1.00054e-23, 4.84386e-24], [588.288, 2.15819e-23, 1.01174e-23, 1.00526e-23, 4.84574e-24], [589.644, 2.16302e-23, 1.01233e-23, 1.00999e-23, 4.84763e-24], [591.004, 2.16786e-23, 1.01292e-23, 1.01474e-23, 4.84953e-24], [592.367, 2.17271e-23, 1.01351e-23, 1.01951e-23, 4.85144e-24], [593.733, 2.17757e-23, 1.01411e-23, 1.0243e-23, 4.85337e-24], [595.102, 2.18244e-23, 1.01471e-23, 1.02911e-23, 4.85531e-24], [596.474, 2.18732e-23, 1.01531e-23, 1.03394e-23, 4.85726e-24], [597.85, 2.19222e-23, 1.01591e-23, 1.03878e-23, 4.85923e-24], [599.229, 2.19713e-23, 1.01652e-23, 1.04365e-23, 4.86121e-24], [600.61, 2.20205e-23, 1.01712e-23, 1.04853e-23, 4.8632e-24], [601.995, 2.20698e-23, 1.01773e-23, 1.05343e-23, 4.8652e-24], [603.384, 2.21192e-23, 1.01835e-23, 1.05836e-23, 4.86722e-24], [604.775, 2.21687e-23, 1.01896e-23, 1.0633e-23, 4.86925e-24], [606.17, 2.22184e-23, 1.01958e-23, 1.06826e-23, 4.8713e-24], [607.567, 2.22682e-23, 1.02021e-23, 1.07324e-23, 4.87335e-24], [608.969, 2.23181e-23, 1.02083e-23, 1.07824e-23, 4.87542e-24], [610.373, 2.23681e-23, 1.02146e-23, 1.08326e-23, 4.87751e-24], [611.78, 2.24182e-23, 1.02209e-23, 1.08829e-23, 4.8796e-24], [613.191, 2.24684e-23, 1.02272e-23, 1.09335e-23, 4.88171e-24], [614.605, 2.25188e-23, 1.02335e-23, 1.09843e-23, 4.88383e-24], [616.022, 2.25693e-23, 1.02399e-23, 1.10352e-23, 4.88597e-24], [617.443, 2.26199e-23, 1.02463e-23, 1.10864e-23, 4.88812e-24], [618.867, 2.26706e-23, 1.02528e-23, 1.11377e-23, 4.89028e-24], [620.294, 2.27215e-23, 1.02592e-23, 1.11893e-23, 4.89246e-24], [621.724, 2.27724e-23, 1.02657e-23, 1.1241e-23, 4.89465e-24], [623.158, 2.28235e-23, 1.02722e-23, 1.12929e-23, 4.89685e-24], [624.595, 2.28747e-23, 1.02788e-23, 1.1345e-23, 4.89907e-24], [626.035, 2.29261e-23, 1.02854e-23, 1.13973e-23, 4.9013e-24], [627.479, 2.29775e-23, 1.0292e-23, 1.14499e-23, 4.90354e-24], [628.926, 2.30291e-23, 1.02986e-23, 1.15026e-23, 4.9058e-24], [630.376, 2.30808e-23, 1.03052e-23, 1.15555e-23, 4.90807e-24], [631.83, 2.31326e-23, 1.03119e-23, 1.16086e-23, 4.91035e-24], [633.287, 2.31845e-23, 1.03187e-23, 1.16619e-23, 4.91265e-24], [634.747, 2.32366e-23, 1.03254e-23, 1.17154e-23, 4.91496e-24], [636.211, 2.32888e-23, 1.03322e-23, 1.1769e-23, 4.91729e-24], [637.678, 2.33411e-23, 1.0339e-23, 1.18229e-23, 4.91963e-24], [639.149, 2.33935e-23, 1.03458e-23, 1.1877e-23, 4.92198e-24], [640.623, 2.34461e-23, 1.03527e-23, 1.19313e-23, 4.92435e-24], [642.1, 2.34987e-23, 1.03596e-23, 1.19858e-23, 4.92673e-24], [643.58, 2.35515e-23, 1.03665e-23, 1.20405e-23, 4.92913e-24], [645.065, 2.36045e-23, 1.03734e-23, 1.20954e-23, 4.93154e-24], [646.552, 2.36575e-23, 1.03804e-23, 1.21504e-23, 4.93396e-24], [648.043, 2.37107e-23, 1.03874e-23, 1.22057e-23, 4.9364e-24], [649.537, 2.3764e-23, 1.03944e-23, 1.22612e-23, 4.93885e-24], [651.035, 2.38174e-23, 1.04015e-23, 1.23169e-23, 4.94132e-24], [652.537, 2.3871e-23, 1.04086e-23, 1.23728e-23, 4.9438e-24], [654.041, 2.39247e-23, 1.04157e-23, 1.24288e-23, 4.9463e-24], [655.55, 2.39785e-23, 1.04229e-23, 1.24851e-23, 4.94881e-24], [657.061, 2.40324e-23, 1.04301e-23, 1.25416e-23, 4.95133e-24], [658.576, 2.40865e-23, 1.04373e-23, 1.25983e-23, 4.95387e-24], [660.095, 2.41406e-23, 1.04445e-23, 1.26552e-23, 4.95643e-24], [661.617, 2.4195e-23, 1.04518e-23, 1.27123e-23, 4.95899e-24], [663.143, 2.42494e-23, 1.04591e-23, 1.27696e-23, 4.96158e-24], [664.672, 2.4304e-23, 1.04664e-23, 1.28271e-23, 4.96417e-24], [666.205, 2.43587e-23, 1.04738e-23, 1.28848e-23, 4.96678e-24], [667.741, 2.44135e-23, 1.04812e-23, 1.29427e-23, 4.96941e-24], [669.281, 2.44685e-23, 1.04886e-23, 1.30008e-23, 4.97205e-24], [670.824, 2.45235e-23, 1.04961e-23, 1.30591e-23, 4.97471e-24], [672.371, 2.45788e-23, 1.05036e-23, 1.31176e-23, 4.97738e-24], [673.922, 2.46341e-23, 1.05111e-23, 1.31763e-23, 4.98007e-24], [675.476, 2.46896e-23, 1.05187e-23, 1.32352e-23, 4.98277e-24], [677.034, 2.47452e-23, 1.05262e-23, 1.32944e-23, 4.98548e-24], [678.595, 2.48009e-23, 1.05339e-23, 1.33537e-23, 4.98821e-24], [680.16, 2.48568e-23, 1.05415e-23, 1.34133e-23, 4.99096e-24], [681.728, 2.49128e-23, 1.05492e-23, 1.3473e-23, 4.99372e-24], [683.3, 2.49689e-23, 1.05569e-23, 1.3533e-23, 4.9965e-24], [684.876, 2.50252e-23, 1.05646e-23, 1.35931e-23, 4.99929e-24], [686.455, 2.50816e-23, 1.05724e-23, 1.36535e-23, 5.00209e-24], [688.038, 2.51381e-23, 1.05802e-23, 1.37141e-23, 5.00492e-24], [689.625, 2.51948e-23, 1.05881e-23, 1.37749e-23, 5.00775e-24], [691.215, 2.52516e-23, 1.05959e-23, 1.38358e-23, 5.0106e-24], [692.809, 2.53085e-23, 1.06038e-23, 1.3897e-23, 5.01347e-24], [694.407, 2.53656e-23, 1.06118e-23, 1.39585e-23, 5.01636e-24], [696.008, 2.54228e-23, 1.06197e-23, 1.40201e-23, 5.01925e-24], [697.613, 2.54801e-23, 1.06277e-23, 1.40819e-23, 5.02217e-24], [699.222, 2.55376e-23, 1.06358e-23, 1.41439e-23, 5.0251e-24], [700.834, 2.55952e-23, 1.06438e-23, 1.42062e-23, 5.02804e-24], [702.45, 2.56529e-23, 1.06519e-23, 1.42687e-23, 5.03101e-24], [704.07, 2.57108e-23, 1.06601e-23, 1.43313e-23, 5.03398e-24], [705.694, 2.57688e-23, 1.06682e-23, 1.43942e-23, 5.03698e-24], [707.321, 2.58269e-23, 1.06764e-23, 1.44573e-23, 5.03998e-24], [708.952, 2.58852e-23, 1.06846e-23, 1.45206e-23, 5.04301e-24], [710.587, 2.59436e-23, 1.06929e-23, 1.45841e-23, 5.04605e-24], [712.226, 2.60022e-23, 1.07012e-23, 1.46478e-23, 5.0491e-24], [713.868, 2.60609e-23, 1.07095e-23, 1.47118e-23, 5.05218e-24], [715.514, 2.61197e-23, 1.07179e-23, 1.47759e-23, 5.05526e-24], [717.164, 2.61787e-23, 1.07263e-23, 1.48403e-23, 5.05837e-24], [718.818, 2.62378e-23, 1.07347e-23, 1.49049e-23, 5.06149e-24], [720.475, 2.6297e-23, 1.07432e-23, 1.49697e-23, 5.06463e-24], [722.137, 2.63564e-23, 1.07517e-23, 1.50347e-23, 5.06778e-24], [723.802, 2.64159e-23, 1.07602e-23, 1.50999e-23, 5.07095e-24], [725.471, 2.64756e-23, 1.07688e-23, 1.51653e-23, 5.07413e-24], [727.144, 2.65354e-23, 1.07774e-23, 1.5231e-23, 5.07733e-24], [728.821, 2.65953e-23, 1.0786e-23, 1.52969e-23, 5.08055e-24], [730.502, 2.66554e-23, 1.07947e-23, 1.53629e-23, 5.08379e-24], [732.186, 2.67156e-23, 1.08034e-23, 1.54292e-23, 5.08704e-24], [733.875, 2.6776e-23, 1.08122e-23, 1.54958e-23, 5.09031e-24], [735.567, 2.68365e-23, 1.08209e-23, 1.55625e-23, 5.09359e-24], [737.263, 2.68972e-23, 1.08298e-23, 1.56294e-23, 5.09689e-24], [738.963, 2.6958e-23, 1.08386e-23, 1.56966e-23, 5.10021e-24], [740.667, 2.70189e-23, 1.08475e-23, 1.5764e-23, 5.10354e-24], [742.375, 2.708e-23, 1.08564e-23, 1.58316e-23, 5.10689e-24], [744.087, 2.71412e-23, 1.08654e-23, 1.58994e-23, 5.11026e-24], [745.803, 2.72025e-23, 1.08743e-23, 1.59674e-23, 5.11364e-24], [747.523, 2.72641e-23, 1.08834e-23, 1.60357e-23, 5.11705e-24], [749.247, 2.73257e-23, 1.08924e-23, 1.61042e-23, 5.12046e-24], [750.975, 2.73875e-23, 1.09015e-23, 1.61729e-23, 5.1239e-24], [752.706, 2.74494e-23, 1.09107e-23, 1.62418e-23, 5.12735e-24], [754.442, 2.75115e-23, 1.09198e-23, 1.63109e-23, 5.13082e-24], [756.182, 2.75738e-23, 1.0929e-23, 1.63803e-23, 5.13431e-24], [757.926, 2.76361e-23, 1.09383e-23, 1.64498e-23, 5.13781e-24], [759.673, 2.76987e-23, 1.09475e-23, 1.65196e-23, 5.14133e-24], [761.425, 2.77613e-23, 1.09569e-23, 1.65897e-23, 5.14487e-24], [763.181, 2.78241e-23, 1.09662e-23, 1.66599e-23, 5.14843e-24], [764.941, 2.78871e-23, 1.09756e-23, 1.67304e-23, 5.152e-24], [766.705, 2.79502e-23, 1.0985e-23, 1.6801e-23, 5.15559e-24], [768.473, 2.80135e-23, 1.09945e-23, 1.68719e-23, 5.1592e-24], [770.245, 2.80769e-23, 1.1004e-23, 1.69431e-23, 5.16282e-24], [772.021, 2.81404e-23, 1.10135e-23, 1.70144e-23, 5.16647e-24], [773.801, 2.82041e-23, 1.10231e-23, 1.7086e-23, 5.17013e-24], [775.586, 2.8268e-23, 1.10327e-23, 1.71578e-23, 5.17381e-24], [777.374, 2.8332e-23, 1.10423e-23, 1.72298e-23, 5.1775e-24], [779.167, 2.83961e-23, 1.1052e-23, 1.73021e-23, 5.18122e-24], [780.964, 2.84604e-23, 1.10617e-23, 1.73745e-23, 5.18495e-24], [782.765, 2.85249e-23, 1.10715e-23, 1.74472e-23, 5.1887e-24], [784.57, 2.85895e-23, 1.10813e-23, 1.75202e-23, 5.19247e-24], [786.379, 2.86542e-23, 1.10911e-23, 1.75933e-23, 5.19625e-24], [788.192, 2.87191e-23, 1.1101e-23, 1.76667e-23, 5.20006e-24], [790.01, 2.87842e-23, 1.11109e-23, 1.77403e-23, 5.20388e-24], [791.832, 2.88494e-23, 1.11209e-23, 1.78141e-23, 5.20772e-24], [793.658, 2.89148e-23, 1.11309e-23, 1.78882e-23, 5.21158e-24], [795.488, 2.89803e-23, 1.11409e-23, 1.79625e-23, 5.21545e-24], [797.322, 2.90459e-23, 1.1151e-23, 1.8037e-23, 5.21935e-24], [799.161, 2.91118e-23, 1.11611e-23, 1.81117e-23, 5.22326e-24], [801.004, 2.91777e-23, 1.11712e-23, 1.81867e-23, 5.22719e-24], [802.851, 2.92439e-23, 1.11814e-23, 1.82619e-23, 5.23114e-24], [804.702, 2.93101e-23, 1.11916e-23, 1.83373e-23, 5.23511e-24], [806.558, 2.93766e-23, 1.12019e-23, 1.84129e-23, 5.2391e-24], [808.418, 2.94432e-23, 1.12122e-23, 1.84888e-23, 5.2431e-24], [810.282, 2.95099e-23, 1.12225e-23, 1.85649e-23, 5.24713e-24], [812.151, 2.95768e-23, 1.12329e-23, 1.86413e-23, 5.25117e-24], [814.023, 2.96439e-23, 1.12433e-23, 1.87178e-23, 5.25523e-24], [815.901, 2.97111e-23, 1.12538e-23, 1.87947e-23, 5.25931e-24], [817.782, 2.97785e-23, 1.12643e-23, 1.88717e-23, 5.26341e-24], [819.668, 2.9846e-23, 1.12748e-23, 1.8949e-23, 5.26753e-24], [821.558, 2.99137e-23, 1.12854e-23, 1.90265e-23, 5.27167e-24], [823.452, 2.99815e-23, 1.1296e-23, 1.91042e-23, 5.27583e-24], [825.351, 3.00495e-23, 1.13067e-23, 1.91821e-23, 5.28e-24], [827.255, 3.01177e-23, 1.13174e-23, 1.92603e-23, 5.2842e-24], [829.162, 3.0186e-23, 1.13281e-23, 1.93388e-23, 5.28842e-24], [831.074, 3.02545e-23, 1.13389e-23, 1.94174e-23, 5.29265e-24], [832.991, 3.03232e-23, 1.13497e-23, 1.94963e-23, 5.29691e-24], [834.912, 3.0392e-23, 1.13606e-23, 1.95755e-23, 5.30119e-24], [836.837, 3.04609e-23, 1.13715e-23, 1.96548e-23, 5.30549e-24], [838.767, 3.053e-23, 1.13825e-23, 1.97344e-23, 5.30981e-24], [840.701, 3.05993e-23, 1.13935e-23, 1.98143e-23, 5.31415e-24], [842.64, 3.06688e-23, 1.14045e-23, 1.98943e-23, 5.31852e-24], [844.583, 3.07384e-23, 1.14156e-23, 1.99746e-23, 5.32293e-24], [846.53, 3.08082e-23, 1.14267e-23, 2.00552e-23, 5.32736e-24], [848.482, 3.08781e-23, 1.14379e-23, 2.0136e-23, 5.33185e-24], [850.439, 3.09483e-23, 1.14492e-23, 2.0217e-23, 5.33642e-24], [852.4, 3.10186e-23, 1.14606e-23, 2.02984e-23, 5.34117e-24], [854.366, 3.10891e-23, 1.14722e-23, 2.03801e-23, 5.34642e-24], [856.336, 3.11603e-23, 1.14851e-23, 2.04627e-23, 5.35428e-24], [858.311, 3.12528e-23, 1.15556e-23, 2.05777e-23, 5.48441e-24], [860.29, 3.13032e-23, 1.15114e-23, 2.06288e-23, 5.37083e-24], [862.274, 3.13726e-23, 1.15181e-23, 2.07086e-23, 5.36525e-24], [864.262, 3.14436e-23, 1.1529e-23, 2.07909e-23, 5.36844e-24], [866.255, 3.15149e-23, 1.15403e-23, 2.08737e-23, 5.37259e-24], [868.253, 3.15865e-23, 1.15518e-23, 2.09568e-23, 5.37701e-24], [870.255, 3.16582e-23, 1.15634e-23, 2.10402e-23, 5.38154e-24], [872.262, 3.17301e-23, 1.1575e-23, 2.11238e-23, 5.38613e-24], [874.273, 3.18022e-23, 1.15867e-23, 2.12077e-23, 5.39076e-24], [876.289, 3.18745e-23, 1.15984e-23, 2.12918e-23, 5.39542e-24], [878.31, 3.19469e-23, 1.16102e-23, 2.13761e-23, 5.40011e-24], [880.336, 3.20195e-23, 1.16221e-23, 2.14607e-23, 5.40482e-24], [882.366, 3.20922e-23, 1.16339e-23, 2.15455e-23, 5.40956e-24], [884.4, 3.21651e-23, 1.16459e-23, 2.16306e-23, 5.41432e-24], [886.44, 3.22382e-23, 1.16578e-23, 2.17159e-23, 5.41911e-24], [888.484, 3.23115e-23, 1.16698e-23, 2.18015e-23, 5.42391e-24], [890.533, 3.23849e-23, 1.16819e-23, 2.18873e-23, 5.42874e-24], [892.586, 3.24586e-23, 1.1694e-23, 2.19734e-23, 5.43359e-24], [894.645, 3.25323e-23, 1.17061e-23, 2.20596e-23, 5.43846e-24], [896.708, 3.26063e-23, 1.17183e-23, 2.21462e-23, 5.44335e-24], [898.775, 3.26804e-23, 1.17306e-23, 2.2233e-23, 5.44826e-24], [900.848, 3.27547e-23, 1.17429e-23, 2.232e-23, 5.4532e-24], [902.925, 3.28292e-23, 1.17552e-23, 2.24073e-23, 5.45815e-24], [905.008, 3.29038e-23, 1.17676e-23, 2.24948e-23, 5.46313e-24], [907.095, 3.29786e-23, 1.178e-23, 2.25825e-23, 5.46813e-24], [909.186, 3.30536e-23, 1.17924e-23, 2.26705e-23, 5.47314e-24], [911.283, 3.31288e-23, 1.18049e-23, 2.27588e-23, 5.47818e-24], [913.384, 3.32041e-23, 1.18175e-23, 2.28473e-23, 5.48325e-24], [915.491, 3.32797e-23, 1.18301e-23, 2.2936e-23, 5.48833e-24], [917.602, 3.33553e-23, 1.18428e-23, 2.3025e-23, 5.49344e-24], [919.718, 3.34312e-23, 1.18554e-23, 2.31143e-23, 5.49856e-24], [921.839, 3.35073e-23, 1.18682e-23, 2.32038e-23, 5.50371e-24], [923.964, 3.35835e-23, 1.1881e-23, 2.32935e-23, 5.50888e-24], [926.095, 3.36599e-23, 1.18938e-23, 2.33835e-23, 5.51407e-24], [928.231, 3.37365e-23, 1.19067e-23, 2.34737e-23, 5.51928e-24], [930.371, 3.38132e-23, 1.19196e-23, 2.35642e-23, 5.52452e-24], [932.517, 3.38901e-23, 1.19326e-23, 2.36549e-23, 5.52978e-24], [934.667, 3.39673e-23, 1.19456e-23, 2.37459e-23, 5.53505e-24], [936.822, 3.40446e-23, 1.19587e-23, 2.38371e-23, 5.54036e-24], [938.983, 3.4122e-23, 1.19718e-23, 2.39286e-23, 5.54568e-24], [941.148, 3.41997e-23, 1.19849e-23, 2.40203e-23, 5.55102e-24], [943.318, 3.42775e-23, 1.19981e-23, 2.41123e-23, 5.55639e-24], [945.494, 3.43555e-23, 1.20114e-23, 2.42046e-23, 5.56178e-24], [947.674, 3.44337e-23, 1.20247e-23, 2.4297e-23, 5.56719e-24], [949.859, 3.45121e-23, 1.2038e-23, 2.43898e-23, 5.57263e-24], [952.05, 3.45907e-23, 1.20514e-23, 2.44828e-23, 5.57808e-24], [954.245, 3.46694e-23, 1.20649e-23, 2.4576e-23, 5.58356e-24], [956.445, 3.47483e-23, 1.20784e-23, 2.46695e-23, 5.58906e-24], [958.651, 3.48274e-23, 1.20919e-23, 2.47632e-23, 5.59458e-24], [960.862, 3.49067e-23, 1.21055e-23, 2.48572e-23, 5.60013e-24], [963.077, 3.49862e-23, 1.21192e-23, 2.49515e-23, 5.6057e-24], [965.298, 3.50659e-23, 1.21328e-23, 2.5046e-23, 5.61129e-24], [967.524, 3.51457e-23, 1.21466e-23, 2.51408e-23, 5.61691e-24], [969.755, 3.52258e-23, 1.21604e-23, 2.52358e-23, 5.62254e-24], [971.992, 3.5306e-23, 1.21742e-23, 2.5331e-23, 5.6282e-24], [974.233, 3.53864e-23, 1.21881e-23, 2.54266e-23, 5.63389e-24], [976.48, 3.5467e-23, 1.2202e-23, 2.55224e-23, 5.63959e-24], [978.731, 3.55478e-23, 1.2216e-23, 2.56184e-23, 5.64532e-24], [980.988, 3.56287e-23, 1.223e-23, 2.57147e-23, 5.65107e-24], [983.251, 3.57099e-23, 1.22441e-23, 2.58112e-23, 5.65685e-24], [985.518, 3.57912e-23, 1.22583e-23, 2.5908e-23, 5.66265e-24], [987.791, 3.58728e-23, 1.22724e-23, 2.60051e-23, 5.66847e-24], [990.068, 3.59545e-23, 1.22867e-23, 2.61024e-23, 5.67431e-24], [992.352, 3.60364e-23, 1.2301e-23, 2.62e-23, 5.68018e-24], [994.64, 3.61185e-23, 1.23153e-23, 2.62978e-23, 5.68607e-24], [996.934, 3.62008e-23, 1.23297e-23, 2.63959e-23, 5.69199e-24], [999.233, 3.62833e-23, 1.23441e-23, 2.64943e-23, 5.69793e-24], [1001.54, 3.6366e-23, 1.23586e-23, 2.65929e-23, 5.70389e-24], [1003.85, 3.64488e-23, 1.23731e-23, 2.66918e-23, 5.70988e-24], [1006.16, 3.65319e-23, 1.23877e-23, 2.67909e-23, 5.71589e-24], [1008.48, 3.66152e-23, 1.24023e-23, 2.68903e-23, 5.72192e-24], [1010.81, 3.66986e-23, 1.2417e-23, 2.699e-23, 5.72798e-24], [1013.14, 3.67822e-23, 1.24318e-23, 2.70899e-23, 5.73406e-24], [1015.47, 3.68661e-23, 1.24466e-23, 2.71901e-23, 5.74017e-24], [1017.82, 3.69501e-23, 1.24614e-23, 2.72905e-23, 5.7463e-24], [1020.16, 3.70343e-23, 1.24763e-23, 2.73912e-23, 5.75245e-24], [1022.52, 3.71188e-23, 1.24913e-23, 2.74921e-23, 5.75863e-24], [1024.87, 3.72034e-23, 1.25063e-23, 2.75934e-23, 5.76483e-24], [1027.24, 3.72882e-23, 1.25213e-23, 2.76949e-23, 5.77106e-24], [1029.61, 3.73732e-23, 1.25364e-23, 2.77966e-23, 5.77731e-24], [1031.98, 3.74584e-23, 1.25516e-23, 2.78986e-23, 5.78358e-24], [1034.36, 3.75438e-23, 1.25668e-23, 2.80009e-23, 5.78988e-24], [1036.74, 3.76294e-23, 1.2582e-23, 2.81034e-23, 5.79621e-24], [1039.14, 3.77152e-23, 1.25974e-23, 2.82062e-23, 5.80255e-24], [1041.53, 3.78012e-23, 1.26127e-23, 2.83093e-23, 5.80893e-24], [1043.93, 3.78875e-23, 1.26281e-23, 2.84126e-23, 5.81532e-24], [1046.34, 3.79739e-23, 1.26436e-23, 2.85162e-23, 5.82175e-24], [1048.75, 3.80605e-23, 1.26591e-23, 2.86201e-23, 5.82819e-24], [1051.17, 3.81473e-23, 1.26747e-23, 2.87242e-23, 5.83467e-24], [1053.6, 3.82343e-23, 1.26904e-23, 2.88286e-23, 5.84116e-24], [1056.03, 3.83215e-23, 1.27061e-23, 2.89333e-23, 5.84769e-24], [1058.46, 3.84089e-23, 1.27218e-23, 2.90382e-23, 5.85423e-24], [1060.9, 3.84965e-23, 1.27376e-23, 2.91434e-23, 5.86081e-24], [1063.35, 3.85843e-23, 1.27534e-23, 2.92489e-23, 5.8674e-24], [1065.8, 3.86723e-23, 1.27693e-23, 2.93546e-23, 5.87403e-24], [1068.26, 3.87606e-23, 1.27853e-23, 2.94606e-23, 5.88067e-24], [1070.72, 3.8849e-23, 1.28013e-23, 2.95669e-23, 5.88735e-24], [1073.19, 3.89376e-23, 1.28174e-23, 2.96735e-23, 5.89405e-24], [1075.67, 3.90265e-23, 1.28335e-23, 2.97803e-23, 5.90077e-24], [1078.15, 3.91155e-23, 1.28497e-23, 2.98873e-23, 5.90752e-24], [1080.63, 3.92048e-23, 1.28659e-23, 2.99947e-23, 5.9143e-24], [1083.12, 3.92942e-23, 1.28822e-23, 3.01023e-23, 5.9211e-24], [1085.62, 3.93839e-23, 1.28986e-23, 3.02102e-23, 5.92793e-24], [1088.13, 3.94738e-23, 1.2915e-23, 3.03184e-23, 5.93478e-24], [1090.63, 3.95639e-23, 1.29314e-23, 3.04268e-23, 5.94166e-24], [1093.15, 3.96542e-23, 1.29479e-23, 3.05355e-23, 5.94856e-24], [1095.67, 3.97447e-23, 1.29645e-23, 3.06445e-23, 5.95549e-24], [1098.2, 3.98354e-23, 1.29811e-23, 3.07537e-23, 5.96245e-24], [1100.73, 3.99263e-23, 1.29978e-23, 3.08633e-23, 5.96943e-24], [1103.27, 4.00175e-23, 1.30145e-23, 3.09731e-23, 5.97644e-24], [1105.81, 4.01088e-23, 1.30313e-23, 3.10831e-23, 5.98347e-24], [1108.36, 4.02004e-23, 1.30482e-23, 3.11935e-23, 5.99053e-24], [1110.92, 4.02921e-23, 1.30651e-23, 3.13041e-23, 5.99762e-24], [1113.48, 4.03841e-23, 1.30821e-23, 3.1415e-23, 6.00474e-24], [1116.05, 4.04763e-23, 1.30991e-23, 3.15262e-23, 6.01188e-24], [1118.62, 4.05687e-23, 1.31162e-23, 3.16376e-23, 6.01904e-24], [1121.2, 4.06614e-23, 1.31333e-23, 3.17493e-23, 6.02624e-24], [1123.79, 4.07542e-23, 1.31505e-23, 3.18613e-23, 6.03346e-24], [1126.38, 4.08472e-23, 1.31677e-23, 3.19736e-23, 6.0407e-24], [1128.97, 4.09405e-23, 1.31851e-23, 3.20861e-23, 6.04798e-24], [1131.58, 4.1034e-23, 1.32024e-23, 3.2199e-23, 6.05528e-24], [1134.19, 4.11277e-23, 1.32198e-23, 3.23121e-23, 6.06261e-24], [1136.8, 4.12216e-23, 1.32373e-23, 3.24254e-23, 6.06996e-24], [1139.42, 4.13158e-23, 1.32549e-23, 3.25391e-23, 6.07734e-24], [1142.05, 4.14101e-23, 1.32725e-23, 3.2653e-23, 6.08475e-24], [1144.69, 4.15047e-23, 1.32901e-23, 3.27673e-23, 6.09219e-24], [1147.33, 4.15995e-23, 1.33078e-23, 3.28817e-23, 6.09965e-24], [1149.97, 4.16945e-23, 1.33256e-23, 3.29965e-23, 6.10714e-24], [1152.62, 4.17898e-23, 1.33435e-23, 3.31116e-23, 6.11466e-24], [1155.28, 4.18852e-23, 1.33614e-23, 3.32269e-23, 6.1222e-24], [1157.94, 4.19809e-23, 1.33793e-23, 3.33425e-23, 6.12977e-24], [1160.62, 4.20768e-23, 1.33973e-23, 3.34584e-23, 6.13737e-24], [1163.29, 4.21729e-23, 1.34154e-23, 3.35746e-23, 6.145e-24], [1165.97, 4.22693e-23, 1.34335e-23, 3.36911e-23, 6.15266e-24], [1168.66, 4.23658e-23, 1.34517e-23, 3.38078e-23, 6.16034e-24], [1171.36, 4.24626e-23, 1.347e-23, 3.39248e-23, 6.16805e-24], [1174.06, 4.25596e-23, 1.34883e-23, 3.40421e-23, 6.17579e-24], [1176.77, 4.26569e-23, 1.35067e-23, 3.41597e-23, 6.18356e-24], [1179.48, 4.27544e-23, 1.35251e-23, 3.42776e-23, 6.19135e-24], [1182.2, 4.28521e-23, 1.35436e-23, 3.43957e-23, 6.19917e-24], [1184.93, 4.295e-23, 1.35622e-23, 3.45142e-23, 6.20703e-24], [1187.66, 4.30481e-23, 1.35808e-23, 3.46329e-23, 6.21491e-24], [1190.4, 4.31465e-23, 1.35995e-23, 3.47519e-23, 6.22281e-24], [1193.14, 4.32451e-23, 1.36182e-23, 3.48712e-23, 6.23075e-24], [1195.89, 4.33439e-23, 1.3637e-23, 3.49908e-23, 6.23871e-24], [1198.65, 4.3443e-23, 1.36559e-23, 3.51107e-23, 6.24671e-24], [1201.42, 4.35423e-23, 1.36748e-23, 3.52308e-23, 6.25473e-24], [1204.19, 4.36418e-23, 1.36938e-23, 3.53513e-23, 6.26278e-24], [1206.96, 4.37415e-23, 1.37128e-23, 3.5472e-23, 6.27086e-24], [1209.75, 4.38415e-23, 1.37319e-23, 3.5593e-23, 6.27897e-24], [1212.54, 4.39417e-23, 1.37511e-23, 3.57143e-23, 6.28711e-24], [1215.33, 4.40422e-23, 1.37703e-23, 3.58359e-23, 6.29528e-24], [1218.13, 4.41429e-23, 1.37897e-23, 3.59578e-23, 6.30348e-24], [1220.94, 4.42438e-23, 1.3809e-23, 3.608e-23, 6.31171e-24], [1223.76, 4.43449e-23, 1.38284e-23, 3.62024e-23, 6.31998e-24], [1226.58, 4.44463e-23, 1.38479e-23, 3.63252e-23, 6.32829e-24], [1229.41, 4.45479e-23, 1.38675e-23, 3.64482e-23, 6.33665e-24], [1232.24, 4.46498e-23, 1.38872e-23, 3.65716e-23, 6.34509e-24], [1235.09, 4.47519e-23, 1.3907e-23, 3.66953e-23, 6.35374e-24], [1237.93, 4.48544e-23, 1.39274e-23, 3.68194e-23, 6.36356e-24], [1240.79, 4.50232e-23, 1.41591e-23, 3.70241e-23, 6.82338e-24], [1243.65, 4.50599e-23, 1.39673e-23, 3.70682e-23, 6.38097e-24], [1246.52, 4.51626e-23, 1.39863e-23, 3.71927e-23, 6.38755e-24], [1249.39, 4.52659e-23, 1.40062e-23, 3.73177e-23, 6.39574e-24], [1252.27, 4.53694e-23, 1.40262e-23, 3.74431e-23, 6.40417e-24], [1255.16, 4.54731e-23, 1.40463e-23, 3.75687e-23, 6.4127e-24], [1258.06, 4.55771e-23, 1.40665e-23, 3.76947e-23, 6.42127e-24], [1260.96, 4.56813e-23, 1.40867e-23, 3.78209e-23, 6.42989e-24], [1263.86, 4.57858e-23, 1.41071e-23, 3.79475e-23, 6.43855e-24], [1266.78, 4.58905e-23, 1.41274e-23, 3.80743e-23, 6.44724e-24], [1269.7, 4.59955e-23, 1.41479e-23, 3.82015e-23, 6.45595e-24], [1272.63, 4.61007e-23, 1.41684e-23, 3.83289e-23, 6.46471e-24], [1275.56, 4.62061e-23, 1.4189e-23, 3.84567e-23, 6.47349e-24], [1278.5, 4.63118e-23, 1.42096e-23, 3.85847e-23, 6.4823e-24], [1281.45, 4.64177e-23, 1.42303e-23, 3.8713e-23, 6.49115e-24], [1284.41, 4.65239e-23, 1.42511e-23, 3.88417e-23, 6.50002e-24], [1287.37, 4.66304e-23, 1.4272e-23, 3.89706e-23, 6.50893e-24], [1290.34, 4.6737e-23, 1.42929e-23, 3.90998e-23, 6.51786e-24], [1293.31, 4.68439e-23, 1.43139e-23, 3.92294e-23, 6.52683e-24], [1296.3, 4.69511e-23, 1.43349e-23, 3.93592e-23, 6.53583e-24], [1299.29, 4.70585e-23, 1.4356e-23, 3.94893e-23, 6.54486e-24], [1302.28, 4.71662e-23, 1.43772e-23, 3.96197e-23, 6.55392e-24], [1305.28, 4.72741e-23, 1.43984e-23, 3.97505e-23, 6.56301e-24], [1308.29, 4.73823e-23, 1.44197e-23, 3.98815e-23, 6.57213e-24], [1311.31, 4.74907e-23, 1.44411e-23, 4.00128e-23, 6.58128e-24], [1314.34, 4.75993e-23, 1.44626e-23, 4.01445e-23, 6.59047e-24], [1317.37, 4.77083e-23, 1.44841e-23, 4.02764e-23, 6.59968e-24], [1320.4, 4.78174e-23, 1.45057e-23, 4.04086e-23, 6.60893e-24], [1323.45, 4.79268e-23, 1.45273e-23, 4.05412e-23, 6.61821e-24], [1326.5, 4.80365e-23, 1.4549e-23, 4.0674e-23, 6.62752e-24], [1329.56, 4.81465e-23, 1.45708e-23, 4.08072e-23, 6.63686e-24], [1332.63, 4.82566e-23, 1.45927e-23, 4.09406e-23, 6.64623e-24], [1335.7, 4.83671e-23, 1.46146e-23, 4.10744e-23, 6.65563e-24], [1338.78, 4.84778e-23, 1.46366e-23, 4.12084e-23, 6.66507e-24], [1341.87, 4.85887e-23, 1.46587e-23, 4.13428e-23, 6.67454e-24], [1344.96, 4.86999e-23, 1.46808e-23, 4.14775e-23, 6.68404e-24], [1348.06, 4.88114e-23, 1.4703e-23, 4.16125e-23, 6.69357e-24], [1351.17, 4.89231e-23, 1.47253e-23, 4.17478e-23, 6.70313e-24], [1354.29, 4.90351e-23, 1.47476e-23, 4.18834e-23, 6.71273e-24], [1357.41, 4.91473e-23, 1.477e-23, 4.20193e-23, 6.72236e-24], [1360.54, 4.92598e-23, 1.47925e-23, 4.21555e-23, 6.73201e-24], [1363.68, 4.93726e-23, 1.48151e-23, 4.2292e-23, 6.74171e-24], [1366.82, 4.94856e-23, 1.48377e-23, 4.24289e-23, 6.75143e-24], [1369.97, 4.95989e-23, 1.48604e-23, 4.2566e-23, 6.76119e-24], [1373.13, 4.97124e-23, 1.48832e-23, 4.27035e-23, 6.77098e-24], [1376.3, 4.98262e-23, 1.4906e-23, 4.28412e-23, 6.7808e-24], [1379.47, 4.99403e-23, 1.49289e-23, 4.29793e-23, 6.79065e-24], [1382.65, 5.00546e-23, 1.49519e-23, 4.31177e-23, 6.80054e-24], [1385.84, 5.01692e-23, 1.49749e-23, 4.32564e-23, 6.81046e-24], [1389.04, 5.02841e-23, 1.4998e-23, 4.33954e-23, 6.82041e-24], [1392.24, 5.03992e-23, 1.50212e-23, 4.35347e-23, 6.8304e-24], [1395.45, 5.05146e-23, 1.50445e-23, 4.36744e-23, 6.84041e-24], [1398.67, 5.06303e-23, 1.50678e-23, 4.38143e-23, 6.85047e-24], [1401.9, 5.07462e-23, 1.50912e-23, 4.39546e-23, 6.86055e-24], [1405.13, 5.08624e-23, 1.51147e-23, 4.40952e-23, 6.87067e-24], [1408.37, 5.09789e-23, 1.51383e-23, 4.42361e-23, 6.88082e-24], [1411.62, 5.10956e-23, 1.51619e-23, 4.43773e-23, 6.891e-24], [1414.87, 5.12126e-23, 1.51856e-23, 4.45188e-23, 6.90122e-24], [1418.13, 5.13299e-23, 1.52093e-23, 4.46607e-23, 6.91147e-24], [1421.4, 5.14475e-23, 1.52332e-23, 4.48028e-23, 6.92176e-24], [1424.68, 5.15653e-23, 1.52571e-23, 4.49453e-23, 6.93207e-24], [1427.97, 5.16834e-23, 1.52811e-23, 4.50881e-23, 6.94243e-24], [1431.26, 5.18017e-23, 1.53051e-23, 4.52312e-23, 6.95281e-24], [1434.56, 5.19204e-23, 1.53293e-23, 4.53747e-23, 6.96323e-24], [1437.87, 5.20393e-23, 1.53535e-23, 4.55184e-23, 6.97368e-24], [1441.18, 5.21585e-23, 1.53778e-23, 4.56625e-23, 6.98417e-24], [1444.51, 5.2278e-23, 1.54021e-23, 4.58069e-23, 6.99469e-24], [1447.84, 5.23977e-23, 1.54266e-23, 4.59517e-23, 7.00525e-24], [1451.18, 5.25177e-23, 1.54511e-23, 4.60967e-23, 7.01584e-24], [1454.52, 5.2638e-23, 1.54756e-23, 4.62421e-23, 7.02646e-24], [1457.88, 5.27586e-23, 1.55003e-23, 4.63878e-23, 7.03712e-24], [1461.24, 5.28794e-23, 1.5525e-23, 4.65338e-23, 7.04781e-24], [1464.61, 5.30006e-23, 1.55498e-23, 4.66801e-23, 7.05854e-24], [1467.99, 5.3122e-23, 1.55747e-23, 4.68268e-23, 7.0693e-24], [1471.37, 5.32437e-23, 1.55997e-23, 4.69737e-23, 7.0801e-24], [1474.77, 5.33657e-23, 1.56247e-23, 4.71211e-23, 7.09093e-24], [1478.17, 5.34879e-23, 1.56498e-23, 4.72687e-23, 7.10179e-24], [1481.57, 5.36105e-23, 1.5675e-23, 4.74167e-23, 7.11269e-24], [1484.99, 5.37333e-23, 1.57003e-23, 4.75649e-23, 7.12363e-24], [1488.42, 5.38564e-23, 1.57256e-23, 4.77136e-23, 7.1346e-24], [1491.85, 5.39798e-23, 1.5751e-23, 4.78625e-23, 7.14561e-24], [1495.29, 5.41035e-23, 1.57765e-23, 4.80118e-23, 7.15665e-24], [1498.74, 5.42274e-23, 1.58021e-23, 4.81614e-23, 7.16772e-24], [1502.19, 5.43517e-23, 1.58277e-23, 4.83113e-23, 7.17884e-24], [1505.66, 5.44762e-23, 1.58534e-23, 4.84616e-23, 7.18998e-24], [1509.13, 5.46011e-23, 1.58792e-23, 4.86122e-23, 7.20117e-24], [1512.61, 5.47262e-23, 1.59051e-23, 4.87631e-23, 7.21238e-24], [1516.1, 5.48516e-23, 1.59311e-23, 4.89143e-23, 7.22364e-24], [1519.59, 5.49773e-23, 1.59571e-23, 4.90659e-23, 7.23493e-24], [1523.1, 5.51033e-23, 1.59832e-23, 4.92178e-23, 7.24625e-24], [1526.61, 5.52295e-23, 1.60094e-23, 4.93701e-23, 7.25761e-24], [1530.13, 5.53561e-23, 1.60357e-23, 4.95227e-23, 7.26901e-24], [1533.66, 5.5483e-23, 1.6062e-23, 4.96756e-23, 7.28044e-24], [1537.19, 5.56101e-23, 1.60884e-23, 4.98288e-23, 7.29191e-24], [1540.74, 5.57376e-23, 1.61149e-23, 4.99824e-23, 7.30342e-24], [1544.29, 5.58653e-23, 1.61415e-23, 5.01364e-23, 7.31496e-24], [1547.85, 5.59934e-23, 1.61682e-23, 5.02906e-23, 7.32654e-24], [1551.42, 5.61217e-23, 1.61949e-23, 5.04452e-23, 7.33816e-24], [1555.0, 5.62503e-23, 1.62218e-23, 5.06002e-23, 7.34982e-24], [1558.59, 5.63793e-23, 1.62487e-23, 5.07554e-23, 7.36152e-24], [1562.18, 5.65085e-23, 1.62757e-23, 5.0911e-23, 7.37327e-24], [1565.78, 5.6638e-23, 1.63027e-23, 5.1067e-23, 7.38508e-24], [1569.39, 5.67679e-23, 1.633e-23, 5.12233e-23, 7.39704e-24], [1573.01, 5.68981e-23, 1.63577e-23, 5.13801e-23, 7.4099e-24], [1576.64, 5.70343e-23, 1.6405e-23, 5.15434e-23, 7.46585e-24], [1580.28, 5.71592e-23, 1.64121e-23, 5.16943e-23, 7.43309e-24], [1583.92, 5.72902e-23, 1.64393e-23, 5.18519e-23, 7.44452e-24], [1587.57, 5.74215e-23, 1.64668e-23, 5.20099e-23, 7.45639e-24], [1591.23, 5.75531e-23, 1.64944e-23, 5.21682e-23, 7.46837e-24], [1594.9, 5.76851e-23, 1.65221e-23, 5.23269e-23, 7.4804e-24], [1598.58, 5.78173e-23, 1.65499e-23, 5.24859e-23, 7.49249e-24], [1602.27, 5.79499e-23, 1.65778e-23, 5.26453e-23, 7.50461e-24], [1605.96, 5.80827e-23, 1.66057e-23, 5.2805e-23, 7.51678e-24], [1609.67, 5.82159e-23, 1.66337e-23, 5.29651e-23, 7.52898e-24], [1613.38, 5.83493e-23, 1.66619e-23, 5.31255e-23, 7.54122e-24], [1617.1, 5.84831e-23, 1.66901e-23, 5.32862e-23, 7.5535e-24], [1620.83, 5.86172e-23, 1.67183e-23, 5.34473e-23, 7.56582e-24], [1624.56, 5.87516e-23, 1.67467e-23, 5.36088e-23, 7.57818e-24], [1628.31, 5.88863e-23, 1.67752e-23, 5.37706e-23, 7.59057e-24], [1632.07, 5.90214e-23, 1.68037e-23, 5.39327e-23, 7.60301e-24], [1635.83, 5.91567e-23, 1.68323e-23, 5.40952e-23, 7.61548e-24], [1639.6, 5.92923e-23, 1.6861e-23, 5.42581e-23, 7.62799e-24], [1643.38, 5.94283e-23, 1.68898e-23, 5.44213e-23, 7.64053e-24], [1647.17, 5.95646e-23, 1.69187e-23, 5.45849e-23, 7.65312e-24], [1650.97, 5.97012e-23, 1.69476e-23, 5.47488e-23, 7.66575e-24], [1654.78, 5.98381e-23, 1.69767e-23, 5.4913e-23, 7.67841e-24], [1658.59, 5.99753e-23, 1.70058e-23, 5.50777e-23, 7.69111e-24], [1662.42, 6.01128e-23, 1.7035e-23, 5.52426e-23, 7.70385e-24], [1666.25, 6.02507e-23, 1.70643e-23, 5.54079e-23, 7.71663e-24], [1670.09, 6.03889e-23, 1.70937e-23, 5.55736e-23, 7.72945e-24], [1673.95, 6.05274e-23, 1.71231e-23, 5.57397e-23, 7.74231e-24], [1677.81, 6.06662e-23, 1.71527e-23, 5.59061e-23, 7.75521e-24], [1681.67, 6.08053e-23, 1.71823e-23, 5.60728e-23, 7.76814e-24], [1685.55, 6.09448e-23, 1.7212e-23, 5.62399e-23, 7.78112e-24], [1689.44, 6.10846e-23, 1.72418e-23, 5.64074e-23, 7.79414e-24], [1693.34, 6.12247e-23, 1.72717e-23, 5.65752e-23, 7.80719e-24], [1697.24, 6.13651e-23, 1.73017e-23, 5.67434e-23, 7.82028e-24], [1701.15, 6.15058e-23, 1.73318e-23, 5.6912e-23, 7.83342e-24], [1705.08, 6.16469e-23, 1.73619e-23, 5.70809e-23, 7.84659e-24], [1709.01, 6.17883e-23, 1.73922e-23, 5.72501e-23, 7.8598e-24], [1712.95, 6.193e-23, 1.74225e-23, 5.74198e-23, 7.87305e-24], [1716.9, 6.20721e-23, 1.74529e-23, 5.75898e-23, 7.88635e-24], [1720.86, 6.22145e-23, 1.74834e-23, 5.77601e-23, 7.89968e-24], [1724.83, 6.23572e-23, 1.7514e-23, 5.79309e-23, 7.91305e-24], [1728.8, 6.25002e-23, 1.75447e-23, 5.8102e-23, 7.92647e-24], [1732.79, 6.26436e-23, 1.75755e-23, 5.82734e-23, 7.93992e-24], [1736.79, 6.27873e-23, 1.76063e-23, 5.84452e-23, 7.95341e-24], [1740.79, 6.29314e-23, 1.76373e-23, 5.86174e-23, 7.96694e-24], [1744.81, 6.30757e-23, 1.76683e-23, 5.879e-23, 7.98052e-24], [1748.83, 6.32204e-23, 1.76994e-23, 5.89629e-23, 7.99413e-24], [1752.86, 6.33655e-23, 1.77306e-23, 5.91362e-23, 8.00779e-24], [1756.9, 6.35109e-23, 1.77619e-23, 5.93099e-23, 8.02149e-24], [1760.96, 6.36566e-23, 1.77933e-23, 5.94839e-23, 8.03522e-24], [1765.02, 6.38026e-23, 1.78248e-23, 5.96583e-23, 8.049e-24], [1769.09, 6.3949e-23, 1.78564e-23, 5.98331e-23, 8.06282e-24], [1773.17, 6.40957e-23, 1.7888e-23, 6.00083e-23, 8.07668e-24], [1777.26, 6.42428e-23, 1.79198e-23, 6.01838e-23, 8.09058e-24], [1781.35, 6.43902e-23, 1.79516e-23, 6.03597e-23, 8.10452e-24], [1785.46, 6.45379e-23, 1.79835e-23, 6.0536e-23, 8.11851e-24], [1789.58, 6.4686e-23, 1.80156e-23, 6.07126e-23, 8.13253e-24], [1793.71, 6.48344e-23, 1.80477e-23, 6.08897e-23, 8.1466e-24], [1797.84, 6.49832e-23, 1.80799e-23, 6.10671e-23, 8.16071e-24], [1801.99, 6.51323e-23, 1.81122e-23, 6.12448e-23, 8.17486e-24], [1806.14, 6.52818e-23, 1.81446e-23, 6.1423e-23, 8.18905e-24], [1810.31, 6.54316e-23, 1.8177e-23, 6.16015e-23, 8.20328e-24], [1814.48, 6.55817e-23, 1.82096e-23, 6.17805e-23, 8.21756e-24], [1818.67, 6.57322e-23, 1.82423e-23, 6.19598e-23, 8.23187e-24], [1822.86, 6.58831e-23, 1.8275e-23, 6.21394e-23, 8.24623e-24], [1827.06, 6.60342e-23, 1.83079e-23, 6.23195e-23, 8.26064e-24], [1831.28, 6.61858e-23, 1.83408e-23, 6.25e-23, 8.27508e-24], [1835.5, 6.63377e-23, 1.83739e-23, 6.26808e-23, 8.28957e-24], [1839.73, 6.64899e-23, 1.8407e-23, 6.2862e-23, 8.3041e-24], [1843.98, 6.66425e-23, 1.84402e-23, 6.30436e-23, 8.31867e-24], [1848.23, 6.67955e-23, 1.84735e-23, 6.32256e-23, 8.33328e-24], [1852.49, 6.69488e-23, 1.85069e-23, 6.34079e-23, 8.34794e-24], [1856.76, 6.71024e-23, 1.85404e-23, 6.35907e-23, 8.36264e-24], [1861.04, 6.72564e-23, 1.8574e-23, 6.37738e-23, 8.37738e-24], [1865.34, 6.74108e-23, 1.86077e-23, 6.39574e-23, 8.39217e-24], [1869.64, 6.75655e-23, 1.86415e-23, 6.41413e-23, 8.407e-24], [1873.95, 6.77206e-23, 1.86754e-23, 6.43256e-23, 8.42187e-24], [1878.27, 6.7876e-23, 1.87094e-23, 6.45103e-23, 8.43679e-24], [1882.6, 6.80318e-23, 1.87434e-23, 6.46954e-23, 8.45176e-24], [1886.94, 6.8188e-23, 1.87776e-23, 6.48809e-23, 8.46678e-24], [1891.29, 6.83445e-23, 1.88119e-23, 6.50668e-23, 8.4819e-24], [1895.65, 6.85014e-23, 1.88466e-23, 6.52531e-23, 8.49764e-24], [1900.03, 6.86589e-23, 1.88817e-23, 6.544e-23, 8.51429e-24], [1904.41, 6.88162e-23, 1.89153e-23, 6.56268e-23, 8.52735e-24], [1908.8, 6.89742e-23, 1.89499e-23, 6.58142e-23, 8.54246e-24], [1913.2, 6.91325e-23, 1.89846e-23, 6.60021e-23, 8.5577e-24], [1917.61, 6.92912e-23, 1.90195e-23, 6.61903e-23, 8.57299e-24], [1922.03, 6.94503e-23, 1.90544e-23, 6.6379e-23, 8.58834e-24], [1926.47, 6.96097e-23, 1.90894e-23, 6.6568e-23, 8.60374e-24], [1930.91, 6.97695e-23, 1.91246e-23, 6.67575e-23, 8.61917e-24], [1935.36, 6.99296e-23, 1.91598e-23, 6.69473e-23, 8.63466e-24], [1939.83, 7.00902e-23, 1.91951e-23, 6.71376e-23, 8.65019e-24], [1944.3, 7.02511e-23, 1.92306e-23, 6.73282e-23, 8.66576e-24], [1948.78, 7.04124e-23, 1.92661e-23, 6.75193e-23, 8.68138e-24], [1953.28, 7.0574e-23, 1.93017e-23, 6.77107e-23, 8.69705e-24], [1957.78, 7.07361e-23, 1.93374e-23, 6.79026e-23, 8.71276e-24], [1962.29, 7.08985e-23, 1.93733e-23, 6.80949e-23, 8.72851e-24], [1966.82, 7.10612e-23, 1.94092e-23, 6.82875e-23, 8.74431e-24], [1971.36, 7.12244e-23, 1.94452e-23, 6.84806e-23, 8.76015e-24], [1975.9, 7.13879e-23, 1.94813e-23, 6.86741e-23, 8.77604e-24], [1980.46, 7.15518e-23, 1.95175e-23, 6.8868e-23, 8.79198e-24], [1985.02, 7.17161e-23, 1.95539e-23, 6.90623e-23, 8.80796e-24], [1989.6, 7.18808e-23, 1.95903e-23, 6.9257e-23, 8.82399e-24], [1994.19, 7.20458e-23, 1.96268e-23, 6.94522e-23, 8.84006e-24], [1998.79, 7.22112e-23, 1.96634e-23, 6.96477e-23, 8.85618e-24], [2003.4, 7.2377e-23, 1.97001e-23, 6.98437e-23, 8.87234e-24], [2008.02, 7.25432e-23, 1.9737e-23, 7.004e-23, 8.88855e-24], [2012.65, 7.27098e-23, 1.97739e-23, 7.02368e-23, 8.90481e-24], [2017.29, 7.28768e-23, 1.98109e-23, 7.0434e-23, 8.92111e-24], [2021.94, 7.30441e-23, 1.98481e-23, 7.06316e-23, 8.93745e-24], [2026.6, 7.32118e-23, 1.98853e-23, 7.08296e-23, 8.95385e-24], [2031.28, 7.338e-23, 1.99226e-23, 7.10281e-23, 8.97029e-24], [2035.96, 7.35485e-23, 1.996e-23, 7.1227e-23, 8.98678e-24], [2040.66, 7.37174e-23, 1.99976e-23, 7.14263e-23, 9.00331e-24], [2045.36, 7.38866e-23, 2.00352e-23, 7.1626e-23, 9.01989e-24], [2050.08, 7.40563e-23, 2.0073e-23, 7.18261e-23, 9.03652e-24], [2054.81, 7.42264e-23, 2.01108e-23, 7.20266e-23, 9.05319e-24], [2059.54, 7.43968e-23, 2.01488e-23, 7.22276e-23, 9.06991e-24], [2064.29, 7.45677e-23, 2.01868e-23, 7.2429e-23, 9.08668e-24], [2069.05, 7.4739e-23, 2.0225e-23, 7.26308e-23, 9.1035e-24], [2073.83, 7.49106e-23, 2.02632e-23, 7.28331e-23, 9.12036e-24], [2078.61, 7.50826e-23, 2.03016e-23, 7.30358e-23, 9.13727e-24], [2083.4, 7.52551e-23, 2.03401e-23, 7.32389e-23, 9.15423e-24], [2088.21, 7.54279e-23, 2.03787e-23, 7.34424e-23, 9.17123e-24], [2093.02, 7.56011e-23, 2.04173e-23, 7.36463e-23, 9.18829e-24], [2097.85, 7.57748e-23, 2.04561e-23, 7.38507e-23, 9.20539e-24], [2102.69, 7.59488e-23, 2.0495e-23, 7.40555e-23, 9.22254e-24], [2107.53, 7.61233e-23, 2.0534e-23, 7.42608e-23, 9.23973e-24], [2112.39, 7.62981e-23, 2.05731e-23, 7.44665e-23, 9.25698e-24], [2117.27, 7.64733e-23, 2.06123e-23, 7.46726e-23, 9.27427e-24], [2122.15, 7.6649e-23, 2.06517e-23, 7.48792e-23, 9.29161e-24], [2127.04, 7.6825e-23, 2.06911e-23, 7.50861e-23, 9.309e-24], [2131.95, 7.70015e-23, 2.07306e-23, 7.52936e-23, 9.32644e-24], [2136.86, 7.71784e-23, 2.07702e-23, 7.55014e-23, 9.34393e-24], [2141.79, 7.73556e-23, 2.081e-23, 7.57097e-23, 9.36146e-24], [2146.73, 7.75333e-23, 2.08498e-23, 7.59185e-23, 9.37905e-24], [2151.68, 7.77114e-23, 2.08898e-23, 7.61276e-23, 9.39668e-24], [2156.64, 7.78899e-23, 2.09299e-23, 7.63373e-23, 9.41436e-24], [2161.61, 7.80689e-23, 2.097e-23, 7.65473e-23, 9.43209e-24], [2166.6, 7.82482e-23, 2.10103e-23, 7.67578e-23, 9.44987e-24], [2171.6, 7.84279e-23, 2.10507e-23, 7.69688e-23, 9.4677e-24], [2176.6, 7.86081e-23, 2.10912e-23, 7.71802e-23, 9.48558e-24], [2181.62, 7.87887e-23, 2.11318e-23, 7.7392e-23, 9.50351e-24], [2186.65, 7.89697e-23, 2.11726e-23, 7.76043e-23, 9.52149e-24], [2191.7, 7.91511e-23, 2.12134e-23, 7.7817e-23, 9.53952e-24], [2196.75, 7.93329e-23, 2.12543e-23, 7.80302e-23, 9.55759e-24], [2201.82, 7.95152e-23, 2.12954e-23, 7.82438e-23, 9.57572e-24], [2206.89, 7.96978e-23, 2.13365e-23, 7.84579e-23, 9.5939e-24], [2211.98, 7.98809e-23, 2.13778e-23, 7.86724e-23, 9.61212e-24], [2217.08, 8.00645e-23, 2.14192e-23, 7.88874e-23, 9.6304e-24], [2222.2, 8.02484e-23, 2.14607e-23, 7.91028e-23, 9.64873e-24], [2227.32, 8.04328e-23, 2.15023e-23, 7.93187e-23, 9.66711e-24], [2232.46, 8.06175e-23, 2.1544e-23, 7.95351e-23, 9.68554e-24], [2237.6, 8.08028e-23, 2.15858e-23, 7.97519e-23, 9.70402e-24], [2242.76, 8.09884e-23, 2.16278e-23, 7.99691e-23, 9.72256e-24], [2247.94, 8.11745e-23, 2.16698e-23, 8.01868e-23, 9.74115e-24], [2253.12, 8.1361e-23, 2.1712e-23, 8.0405e-23, 9.75986e-24], [2258.32, 8.15482e-23, 2.17552e-23, 8.06239e-23, 9.7804e-24], [2263.52, 8.17353e-23, 2.17967e-23, 8.08427e-23, 9.79722e-24], [2268.74, 8.19231e-23, 2.18392e-23, 8.10623e-23, 9.81598e-24], [2273.97, 8.21113e-23, 2.18818e-23, 8.12823e-23, 9.83481e-24], [2279.22, 8.23e-23, 2.19245e-23, 8.15028e-23, 9.85369e-24], [2284.47, 8.24891e-23, 2.19674e-23, 8.17237e-23, 9.87263e-24], [2289.74, 8.26786e-23, 2.20103e-23, 8.19451e-23, 9.89162e-24], [2295.02, 8.28686e-23, 2.20534e-23, 8.2167e-23, 9.91067e-24], [2300.31, 8.3059e-23, 2.20966e-23, 8.23894e-23, 9.92976e-24], [2305.62, 8.32499e-23, 2.21399e-23, 8.26122e-23, 9.94891e-24], [2310.94, 8.34412e-23, 2.21833e-23, 8.28355e-23, 9.96811e-24], [2316.27, 8.36329e-23, 2.22269e-23, 8.30592e-23, 9.98736e-24], [2321.61, 8.38251e-23, 2.22705e-23, 8.32835e-23, 1.00067e-23], [2326.96, 8.40177e-23, 2.23143e-23, 8.35082e-23, 1.0026e-23], [2332.33, 8.42108e-23, 2.23581e-23, 8.37333e-23, 1.00454e-23], [2337.7, 8.44043e-23, 2.24021e-23, 8.3959e-23, 1.00649e-23], [2343.1, 8.45982e-23, 2.24463e-23, 8.41851e-23, 1.00844e-23], [2348.5, 8.47926e-23, 2.24905e-23, 8.44117e-23, 1.0104e-23], [2353.91, 8.49875e-23, 2.25348e-23, 8.46388e-23, 1.01236e-23], [2359.34, 8.51828e-23, 2.25793e-23, 8.48664e-23, 1.01433e-23], [2364.78, 8.53785e-23, 2.26239e-23, 8.50944e-23, 1.0163e-23], [2370.24, 8.55748e-23, 2.26686e-23, 8.53229e-23, 1.01828e-23], [2375.7, 8.57714e-23, 2.27134e-23, 8.55519e-23, 1.02026e-23], [2381.18, 8.59685e-23, 2.27583e-23, 8.57814e-23, 1.02225e-23], [2386.67, 8.61661e-23, 2.28034e-23, 8.60114e-23, 1.02424e-23], [2392.17, 8.63641e-23, 2.28486e-23, 8.62419e-23, 1.02624e-23], [2397.69, 8.65626e-23, 2.28939e-23, 8.64728e-23, 1.02825e-23], [2403.22, 8.67615e-23, 2.29393e-23, 8.67043e-23, 1.03026e-23], [2408.76, 8.69609e-23, 2.29848e-23, 8.69362e-23, 1.03227e-23], [2414.32, 8.71608e-23, 2.30305e-23, 8.71686e-23, 1.03429e-23], [2419.88, 8.73611e-23, 2.30762e-23, 8.74015e-23, 1.03632e-23], [2425.46, 8.75619e-23, 2.31221e-23, 8.76349e-23, 1.03835e-23], [2431.06, 8.77632e-23, 2.31682e-23, 8.78688e-23, 1.04039e-23], [2436.66, 8.79649e-23, 2.32143e-23, 8.81032e-23, 1.04243e-23], [2442.28, 8.81671e-23, 2.32606e-23, 8.8338e-23, 1.04448e-23], [2447.91, 8.83697e-23, 2.33069e-23, 8.85734e-23, 1.04654e-23], [2453.56, 8.85728e-23, 2.33534e-23, 8.88093e-23, 1.04859e-23], [2459.22, 8.87764e-23, 2.34001e-23, 8.90457e-23, 1.05066e-23], [2464.89, 8.89804e-23, 2.34468e-23, 8.92825e-23, 1.05273e-23], [2470.57, 8.9185e-23, 2.34937e-23, 8.95199e-23, 1.05481e-23], [2476.27, 8.939e-23, 2.35407e-23, 8.97578e-23, 1.05689e-23], [2481.98, 8.95954e-23, 2.35878e-23, 8.99961e-23, 1.05897e-23], [2487.7, 8.98014e-23, 2.3635e-23, 9.0235e-23, 1.06107e-23], [2493.44, 9.00078e-23, 2.36824e-23, 9.04744e-23, 1.06317e-23], [2499.19, 9.02147e-23, 2.37298e-23, 9.07143e-23, 1.06527e-23], [2504.95, 9.0422e-23, 2.37774e-23, 9.09547e-23, 1.06738e-23], [2510.73, 9.06299e-23, 2.38252e-23, 9.11956e-23, 1.06949e-23], [2516.52, 9.08382e-23, 2.3873e-23, 9.1437e-23, 1.07161e-23], [2522.32, 9.1047e-23, 2.3921e-23, 9.1679e-23, 1.07374e-23], [2528.14, 9.12563e-23, 2.39691e-23, 9.19214e-23, 1.07587e-23], [2533.97, 9.14661e-23, 2.40174e-23, 9.21643e-23, 1.07801e-23], [2539.81, 9.16763e-23, 2.40657e-23, 9.24078e-23, 1.08015e-23], [2545.67, 9.18871e-23, 2.41142e-23, 9.26518e-23, 1.0823e-23], [2551.54, 9.20983e-23, 2.41628e-23, 9.28963e-23, 1.08446e-23], [2557.42, 9.231e-23, 2.42115e-23, 9.31413e-23, 1.08662e-23], [2563.32, 9.25222e-23, 2.42604e-23, 9.33869e-23, 1.08878e-23], [2569.23, 9.27349e-23, 2.43094e-23, 9.36329e-23, 1.09095e-23], [2575.16, 9.29481e-23, 2.43585e-23, 9.38795e-23, 1.09313e-23], [2581.09, 9.31618e-23, 2.44077e-23, 9.41266e-23, 1.09532e-23], [2587.05, 9.3376e-23, 2.44571e-23, 9.43742e-23, 1.0975e-23], [2593.01, 9.35906e-23, 2.45066e-23, 9.46224e-23, 1.0997e-23], [2598.99, 9.38058e-23, 2.45562e-23, 9.4871e-23, 1.1019e-23], [2604.99, 9.40215e-23, 2.4606e-23, 9.51202e-23, 1.10411e-23], [2610.99, 9.42376e-23, 2.46559e-23, 9.537e-23, 1.10632e-23], [2617.01, 9.44543e-23, 2.47059e-23, 9.56202e-23, 1.10854e-23], [2623.05, 9.46714e-23, 2.4756e-23, 9.5871e-23, 1.11076e-23], [2629.1, 9.48891e-23, 2.48063e-23, 9.61223e-23, 1.11299e-23], [2635.16, 9.51072e-23, 2.48567e-23, 9.63742e-23, 1.11523e-23], [2641.24, 9.53259e-23, 2.49073e-23, 9.66266e-23, 1.11747e-23], [2647.33, 9.55451e-23, 2.49579e-23, 9.68795e-23, 1.11972e-23], [2653.43, 9.57647e-23, 2.50087e-23, 9.7133e-23, 1.12197e-23], [2659.55, 9.59849e-23, 2.50597e-23, 9.7387e-23, 1.12423e-23], [2665.68, 9.62056e-23, 2.51107e-23, 9.76415e-23, 1.1265e-23], [2671.83, 9.64268e-23, 2.51619e-23, 9.78966e-23, 1.12876e-23], [2677.99, 9.66485e-23, 2.52132e-23, 9.81522e-23, 1.13104e-23], [2684.17, 9.68707e-23, 2.52647e-23, 9.84084e-23, 1.13332e-23], [2690.36, 9.70935e-23, 2.53163e-23, 9.86651e-23, 1.13561e-23], [2696.56, 9.73167e-23, 2.5368e-23, 9.89224e-23, 1.13791e-23], [2702.78, 9.75405e-23, 2.54198e-23, 9.91802e-23, 1.14021e-23], [2709.01, 9.77647e-23, 2.54718e-23, 9.94385e-23, 1.14252e-23], [2715.26, 9.79895e-23, 2.55239e-23, 9.96974e-23, 1.14483e-23], [2721.52, 9.82149e-23, 2.55762e-23, 9.99569e-23, 1.14715e-23], [2727.8, 9.84407e-23, 2.56286e-23, 1.00217e-22, 1.14947e-23], [2734.09, 9.8667e-23, 2.56811e-23, 1.00477e-22, 1.15181e-23], [2740.39, 9.88939e-23, 2.57338e-23, 1.00739e-22, 1.15414e-23], [2746.71, 9.91213e-23, 2.57865e-23, 1.01e-22, 1.15649e-23], [2753.04, 9.93492e-23, 2.58395e-23, 1.01262e-22, 1.15884e-23], [2759.39, 9.95777e-23, 2.58925e-23, 1.01525e-22, 1.16119e-23], [2765.76, 9.98067e-23, 2.59457e-23, 1.01789e-22, 1.16355e-23], [2772.13, 1.00036e-22, 2.59991e-23, 1.02052e-22, 1.16592e-23], [2778.53, 1.00266e-22, 2.60525e-23, 1.02317e-22, 1.1683e-23], [2784.93, 1.00497e-22, 2.61062e-23, 1.02582e-22, 1.17068e-23], [2791.36, 1.00728e-22, 2.61599e-23, 1.02847e-22, 1.17306e-23], [2797.79, 1.0096e-22, 2.62138e-23, 1.03114e-22, 1.17545e-23], [2804.24, 1.01192e-22, 2.62678e-23, 1.0338e-22, 1.17785e-23], [2810.71, 1.01424e-22, 2.6322e-23, 1.03648e-22, 1.18026e-23], [2817.19, 1.01658e-22, 2.63763e-23, 1.03915e-22, 1.18267e-23], [2823.69, 1.01891e-22, 2.64307e-23, 1.04184e-22, 1.18509e-23], [2830.2, 1.02126e-22, 2.64853e-23, 1.04453e-22, 1.18751e-23], [2836.73, 1.02361e-22, 2.654e-23, 1.04723e-22, 1.18994e-23], [2843.27, 1.02596e-22, 2.65949e-23, 1.04993e-22, 1.19238e-23], [2849.83, 1.02832e-22, 2.66499e-23, 1.05263e-22, 1.19482e-23], [2856.4, 1.03068e-22, 2.6705e-23, 1.05535e-22, 1.19727e-23], [2862.98, 1.03305e-22, 2.67603e-23, 1.05807e-22, 1.19973e-23], [2869.59, 1.03543e-22, 2.68157e-23, 1.06079e-22, 1.20219e-23], [2876.2, 1.03781e-22, 2.68712e-23, 1.06352e-22, 1.20466e-23], [2882.84, 1.0402e-22, 2.69269e-23, 1.06626e-22, 1.20714e-23], [2889.48, 1.04259e-22, 2.69828e-23, 1.069e-22, 1.20962e-23], [2896.15, 1.04499e-22, 2.70388e-23, 1.07175e-22, 1.21211e-23], [2902.83, 1.04739e-22, 2.70949e-23, 1.0745e-22, 1.2146e-23], [2909.52, 1.0498e-22, 2.71512e-23, 1.07726e-22, 1.2171e-23], [2916.23, 1.05222e-22, 2.72076e-23, 1.08003e-22, 1.21961e-23], [2922.95, 1.05464e-22, 2.72641e-23, 1.0828e-22, 1.22212e-23], [2929.69, 1.05706e-22, 2.73208e-23, 1.08558e-22, 1.22464e-23], [2936.45, 1.05949e-22, 2.73777e-23, 1.08836e-22, 1.22717e-23], [2943.22, 1.06193e-22, 2.74347e-23, 1.09115e-22, 1.2297e-23], [2950.01, 1.06437e-22, 2.74918e-23, 1.09394e-22, 1.23224e-23], [2956.81, 1.06682e-22, 2.75491e-23, 1.09674e-22, 1.23479e-23], [2963.63, 1.06927e-22, 2.76065e-23, 1.09955e-22, 1.23734e-23], [2970.46, 1.07173e-22, 2.76641e-23, 1.10237e-22, 1.2399e-23], [2977.31, 1.0742e-22, 2.77218e-23, 1.10518e-22, 1.24246e-23], [2984.18, 1.07667e-22, 2.77796e-23, 1.10801e-22, 1.24504e-23], [2991.06, 1.07914e-22, 2.78376e-23, 1.11084e-22, 1.24762e-23], [2997.96, 1.08163e-22, 2.78958e-23, 1.11368e-22, 1.2502e-23], [3004.87, 1.08411e-22, 2.79541e-23, 1.11652e-22, 1.2528e-23], [3011.8, 1.08661e-22, 2.80126e-23, 1.11937e-22, 1.2554e-23], [3018.75, 1.08911e-22, 2.80712e-23, 1.12223e-22, 1.25801e-23], [3025.71, 1.09161e-22, 2.81299e-23, 1.12509e-22, 1.26062e-23], [3032.68, 1.09412e-22, 2.81888e-23, 1.12796e-22, 1.26325e-23], [3039.68, 1.09664e-22, 2.82479e-23, 1.13083e-22, 1.26587e-23], [3046.69, 1.09916e-22, 2.83071e-23, 1.13371e-22, 1.26851e-23], [3053.71, 1.10169e-22, 2.83664e-23, 1.1366e-22, 1.27115e-23], [3060.76, 1.10423e-22, 2.84259e-23, 1.13949e-22, 1.2738e-23], [3067.81, 1.10677e-22, 2.84856e-23, 1.14239e-22, 1.27645e-23], [3074.89, 1.10931e-22, 2.85454e-23, 1.14529e-22, 1.27911e-23], [3081.98, 1.11186e-22, 2.86053e-23, 1.14821e-22, 1.28178e-23], [3089.09, 1.11442e-22, 2.86654e-23, 1.15112e-22, 1.28446e-23], [3096.21, 1.11699e-22, 2.87258e-23, 1.15405e-22, 1.28717e-23], [3103.35, 1.11955e-22, 2.8786e-23, 1.15698e-22, 1.28983e-23], [3110.51, 1.12213e-22, 2.88466e-23, 1.15991e-22, 1.29253e-23], [3117.68, 1.12471e-22, 2.89073e-23, 1.16286e-22, 1.29523e-23], [3124.87, 1.1273e-22, 2.89682e-23, 1.16581e-22, 1.29794e-23], [3132.07, 1.12989e-22, 2.90292e-23, 1.16876e-22, 1.30065e-23], [3139.3, 1.13249e-22, 2.90903e-23, 1.17172e-22, 1.30338e-23], [3146.54, 1.1351e-22, 2.91516e-23, 1.17469e-22, 1.30611e-23], [3153.79, 1.13771e-22, 2.92131e-23, 1.17767e-22, 1.30884e-23], [3161.06, 1.14032e-22, 2.92747e-23, 1.18065e-22, 1.31159e-23], [3168.35, 1.14295e-22, 2.93365e-23, 1.18363e-22, 1.31434e-23], [3175.66, 1.14558e-22, 2.93984e-23, 1.18663e-22, 1.3171e-23], [3182.98, 1.14821e-22, 2.94605e-23, 1.18963e-22, 1.31986e-23], [3190.32, 1.15085e-22, 2.95228e-23, 1.19263e-22, 1.32263e-23], [3197.68, 1.1535e-22, 2.95852e-23, 1.19565e-22, 1.32541e-23], [3205.05, 1.15616e-22, 2.96477e-23, 1.19867e-22, 1.3282e-23], [3212.44, 1.15882e-22, 2.97104e-23, 1.20169e-22, 1.33099e-23], [3219.85, 1.16148e-22, 2.97733e-23, 1.20473e-22, 1.33379e-23], [3227.28, 1.16415e-22, 2.98363e-23, 1.20776e-22, 1.3366e-23], [3234.72, 1.16683e-22, 2.98995e-23, 1.21081e-22, 1.33941e-23], [3242.18, 1.16952e-22, 2.99629e-23, 1.21386e-22, 1.34223e-23], [3249.66, 1.17221e-22, 3.00263e-23, 1.21692e-22, 1.34506e-23], [3257.15, 1.1749e-22, 3.009e-23, 1.21999e-22, 1.3479e-23], [3264.66, 1.17761e-22, 3.01538e-23, 1.22306e-22, 1.35074e-23], [3272.19, 1.18032e-22, 3.02178e-23, 1.22614e-22, 1.35359e-23], [3279.73, 1.18303e-22, 3.02819e-23, 1.22922e-22, 1.35645e-23], [3287.3, 1.18575e-22, 3.03462e-23, 1.23231e-22, 1.35931e-23], [3294.88, 1.18848e-22, 3.04107e-23, 1.23541e-22, 1.36218e-23], [3302.48, 1.19122e-22, 3.04753e-23, 1.23852e-22, 1.36506e-23], [3310.09, 1.19396e-22, 3.05401e-23, 1.24163e-22, 1.36795e-23], [3317.72, 1.1967e-22, 3.0605e-23, 1.24475e-22, 1.37084e-23], [3325.37, 1.19946e-22, 3.06701e-23, 1.24787e-22, 1.37374e-23], [3333.04, 1.20222e-22, 3.07354e-23, 1.25101e-22, 1.37665e-23], [3340.73, 1.20498e-22, 3.08008e-23, 1.25414e-22, 1.37956e-23], [3348.43, 1.20776e-22, 3.08664e-23, 1.25729e-22, 1.38248e-23], [3356.15, 1.21054e-22, 3.09321e-23, 1.26044e-22, 1.38541e-23], [3363.89, 1.21332e-22, 3.0998e-23, 1.2636e-22, 1.38835e-23], [3371.65, 1.21611e-22, 3.10641e-23, 1.26677e-22, 1.39129e-23], [3379.43, 1.21891e-22, 3.11303e-23, 1.26994e-22, 1.39425e-23], [3387.22, 1.22172e-22, 3.11967e-23, 1.27312e-22, 1.3972e-23], [3395.03, 1.22453e-22, 3.12633e-23, 1.27631e-22, 1.40017e-23], [3402.86, 1.22734e-22, 3.133e-23, 1.2795e-22, 1.40314e-23], [3410.71, 1.23017e-22, 3.13969e-23, 1.2827e-22, 1.40613e-23], [3418.57, 1.233e-22, 3.1464e-23, 1.28591e-22, 1.40911e-23], [3426.45, 1.23584e-22, 3.15312e-23, 1.28912e-22, 1.41211e-23], [3434.36, 1.23868e-22, 3.15986e-23, 1.29234e-22, 1.41511e-23], [3442.28, 1.24153e-22, 3.16662e-23, 1.29557e-22, 1.41812e-23], [3450.21, 1.24439e-22, 3.17339e-23, 1.29881e-22, 1.42114e-23], [3458.17, 1.24725e-22, 3.18018e-23, 1.30205e-22, 1.42417e-23], [3466.14, 1.25012e-22, 3.18699e-23, 1.3053e-22, 1.4272e-23], [3474.14, 1.253e-22, 3.19381e-23, 1.30856e-22, 1.43024e-23], [3482.15, 1.25588e-22, 3.20065e-23, 1.31182e-22, 1.43329e-23], [3490.18, 1.25877e-22, 3.20751e-23, 1.31509e-22, 1.43635e-23], [3498.23, 1.26167e-22, 3.21438e-23, 1.31837e-22, 1.43941e-23], [3506.29, 1.26457e-22, 3.22128e-23, 1.32165e-22, 1.44248e-23], [3514.38, 1.26748e-22, 3.22818e-23, 1.32494e-22, 1.44556e-23], [3522.48, 1.2704e-22, 3.23511e-23, 1.32824e-22, 1.44865e-23], [3530.61, 1.27332e-22, 3.24205e-23, 1.33155e-22, 1.45175e-23], [3538.75, 1.27625e-22, 3.24901e-23, 1.33486e-22, 1.45485e-23], [3546.91, 1.27919e-22, 3.25599e-23, 1.33818e-22, 1.45796e-23], [3555.09, 1.28213e-22, 3.26298e-23, 1.34151e-22, 1.46108e-23], [3563.29, 1.28508e-22, 3.26999e-23, 1.34485e-22, 1.4642e-23], [3571.5, 1.28804e-22, 3.27702e-23, 1.34819e-22, 1.46733e-23], [3579.74, 1.291e-22, 3.28407e-23, 1.35154e-22, 1.47047e-23], [3587.99, 1.29397e-22, 3.29113e-23, 1.35489e-22, 1.47362e-23], [3596.27, 1.29695e-22, 3.29821e-23, 1.35826e-22, 1.47678e-23], [3604.56, 1.29994e-22, 3.30531e-23, 1.36163e-22, 1.47994e-23], [3612.87, 1.30293e-22, 3.31242e-23, 1.36501e-22, 1.48312e-23], [3621.2, 1.30593e-22, 3.31956e-23, 1.36839e-22, 1.4863e-23], [3629.55, 1.30893e-22, 3.32671e-23, 1.37179e-22, 1.48949e-23], [3637.92, 1.31194e-22, 3.33387e-23, 1.37519e-22, 1.49268e-23], [3646.31, 1.31496e-22, 3.34106e-23, 1.3786e-22, 1.49588e-23], [3654.72, 1.31799e-22, 3.34826e-23, 1.38201e-22, 1.4991e-23], [3663.15, 1.32102e-22, 3.35548e-23, 1.38543e-22, 1.50232e-23], [3671.6, 1.32406e-22, 3.36272e-23, 1.38886e-22, 1.50554e-23], [3680.06, 1.32711e-22, 3.36998e-23, 1.3923e-22, 1.50878e-23], [3688.55, 1.33016e-22, 3.37725e-23, 1.39575e-22, 1.51202e-23], [3697.06, 1.33323e-22, 3.38455e-23, 1.3992e-22, 1.51528e-23], [3705.58, 1.33629e-22, 3.39186e-23, 1.40266e-22, 1.51854e-23], [3714.13, 1.33937e-22, 3.39919e-23, 1.40613e-22, 1.5218e-23], [3722.69, 1.34245e-22, 3.40653e-23, 1.4096e-22, 1.52508e-23], [3731.28, 1.34554e-22, 3.4139e-23, 1.41309e-22, 1.52836e-23], [3739.88, 1.34864e-22, 3.42128e-23, 1.41658e-22, 1.53166e-23], [3748.5, 1.35174e-22, 3.42868e-23, 1.42008e-22, 1.53496e-23], [3757.15, 1.35485e-22, 3.4361e-23, 1.42358e-22, 1.53827e-23], [3765.81, 1.35797e-22, 3.44354e-23, 1.4271e-22, 1.54159e-23], [3774.5, 1.3611e-22, 3.45099e-23, 1.43062e-22, 1.54492e-23], [3783.2, 1.36423e-22, 3.45847e-23, 1.43415e-22, 1.54826e-23], [3791.92, 1.36737e-22, 3.46596e-23, 1.43768e-22, 1.5516e-23], [3800.67, 1.37052e-22, 3.47347e-23, 1.44123e-22, 1.55495e-23], [3809.43, 1.37367e-22, 3.481e-23, 1.44478e-22, 1.55832e-23], [3818.22, 1.37683e-22, 3.48855e-23, 1.44834e-22, 1.56169e-23], [3827.02, 1.38e-22, 3.49612e-23, 1.45191e-22, 1.56506e-23], [3835.85, 1.38318e-22, 3.5037e-23, 1.45548e-22, 1.56845e-23], [3844.69, 1.38636e-22, 3.51131e-23, 1.45906e-22, 1.57184e-23], [3853.56, 1.38955e-22, 3.51893e-23, 1.46266e-22, 1.57525e-23], [3862.44, 1.39275e-22, 3.52657e-23, 1.46625e-22, 1.57866e-23], [3871.35, 1.39596e-22, 3.53423e-23, 1.46986e-22, 1.58208e-23], [3880.28, 1.39917e-22, 3.54191e-23, 1.47348e-22, 1.5855e-23], [3889.23, 1.40239e-22, 3.5496e-23, 1.4771e-22, 1.58894e-23], [3898.2, 1.40562e-22, 3.55732e-23, 1.48073e-22, 1.59238e-23], [3907.18, 1.40885e-22, 3.56505e-23, 1.48437e-22, 1.59584e-23], [3916.19, 1.4121e-22, 3.57281e-23, 1.48801e-22, 1.5993e-23], [3925.23, 1.41535e-22, 3.58058e-23, 1.49167e-22, 1.60277e-23], [3934.28, 1.4186e-22, 3.58837e-23, 1.49533e-22, 1.60625e-23], [3943.35, 1.42187e-22, 3.59618e-23, 1.499e-22, 1.60973e-23], [3952.44, 1.42514e-22, 3.60401e-23, 1.50268e-22, 1.61323e-23], [3961.56, 1.42842e-22, 3.61186e-23, 1.50636e-22, 1.61673e-23], [3970.69, 1.43171e-22, 3.61972e-23, 1.51006e-22, 1.62025e-23], [3979.85, 1.43501e-22, 3.62765e-23, 1.51376e-22, 1.62385e-23], [3989.03, 1.43831e-22, 3.63552e-23, 1.51747e-22, 1.6273e-23], [3998.23, 1.44162e-22, 3.64344e-23, 1.52119e-22, 1.63083e-23], [4007.45, 1.44494e-22, 3.65139e-23, 1.52492e-22, 1.63438e-23], [4016.69, 1.44827e-22, 3.65935e-23, 1.52865e-22, 1.63794e-23], [4025.95, 1.4516e-22, 3.66733e-23, 1.5324e-22, 1.6415e-23], [4035.23, 1.45494e-22, 3.67533e-23, 1.53615e-22, 1.64507e-23], [4044.54, 1.45829e-22, 3.68336e-23, 1.53991e-22, 1.64866e-23], [4053.87, 1.46165e-22, 3.6914e-23, 1.54368e-22, 1.65225e-23], [4063.21, 1.46501e-22, 3.69946e-23, 1.54745e-22, 1.65585e-23], [4072.58, 1.46838e-22, 3.70754e-23, 1.55124e-22, 1.65945e-23], [4081.97, 1.47176e-22, 3.71564e-23, 1.55503e-22, 1.66307e-23], [4091.39, 1.47515e-22, 3.72376e-23, 1.55883e-22, 1.6667e-23], [4100.82, 1.47855e-22, 3.7319e-23, 1.56264e-22, 1.67033e-23], [4110.28, 1.48195e-22, 3.74006e-23, 1.56646e-22, 1.67397e-23], [4119.76, 1.48536e-22, 3.74823e-23, 1.57028e-22, 1.67763e-23], [4129.26, 1.48878e-22, 3.75643e-23, 1.57412e-22, 1.68129e-23], [4138.78, 1.49221e-22, 3.76465e-23, 1.57796e-22, 1.68496e-23], [4148.32, 1.49564e-22, 3.77289e-23, 1.58181e-22, 1.68864e-23], [4157.89, 1.49909e-22, 3.78115e-23, 1.58567e-22, 1.69232e-23], [4167.48, 1.50254e-22, 3.78943e-23, 1.58954e-22, 1.69602e-23], [4177.09, 1.506e-22, 3.79773e-23, 1.59342e-22, 1.69973e-23], [4186.72, 1.50946e-22, 3.80604e-23, 1.5973e-22, 1.70344e-23], [4196.38, 1.51294e-22, 3.81438e-23, 1.6012e-22, 1.70717e-23], [4206.05, 1.51642e-22, 3.82274e-23, 1.6051e-22, 1.7109e-23], [4215.75, 1.51991e-22, 3.83112e-23, 1.60901e-22, 1.71464e-23], [4225.47, 1.52341e-22, 3.83952e-23, 1.61293e-22, 1.71839e-23], [4235.22, 1.52692e-22, 3.84794e-23, 1.61686e-22, 1.72215e-23], [4244.98, 1.53043e-22, 3.85638e-23, 1.6208e-22, 1.72592e-23], [4254.77, 1.53396e-22, 3.86484e-23, 1.62474e-22, 1.7297e-23], [4264.58, 1.53749e-22, 3.87332e-23, 1.6287e-22, 1.73349e-23], [4274.42, 1.54103e-22, 3.88183e-23, 1.63266e-22, 1.73729e-23], [4284.27, 1.54458e-22, 3.89035e-23, 1.63663e-22, 1.74109e-23], [4294.15, 1.54813e-22, 3.89889e-23, 1.64061e-22, 1.74491e-23], [4304.06, 1.5517e-22, 3.90745e-23, 1.6446e-22, 1.74873e-23], [4313.98, 1.55527e-22, 3.91604e-23, 1.6486e-22, 1.75257e-23], [4323.93, 1.55885e-22, 3.92464e-23, 1.65261e-22, 1.75641e-23], [4333.9, 1.56244e-22, 3.93327e-23, 1.65662e-22, 1.76026e-23], [4343.89, 1.56604e-22, 3.94191e-23, 1.66065e-22, 1.76412e-23], [4353.91, 1.56964e-22, 3.95058e-23, 1.66468e-22, 1.76799e-23], [4363.95, 1.57325e-22, 3.95927e-23, 1.66872e-22, 1.77187e-23], [4374.02, 1.57688e-22, 3.96798e-23, 1.67277e-22, 1.77576e-23], [4384.1, 1.58051e-22, 3.97671e-23, 1.67683e-22, 1.77966e-23], [4394.21, 1.58415e-22, 3.98546e-23, 1.6809e-22, 1.78357e-23], [4404.34, 1.58779e-22, 3.99423e-23, 1.68498e-22, 1.78749e-23], [4414.5, 1.59145e-22, 4.00303e-23, 1.68907e-22, 1.79142e-23], [4424.68, 1.59511e-22, 4.01184e-23, 1.69316e-22, 1.79536e-23], [4434.88, 1.59878e-22, 4.02068e-23, 1.69727e-22, 1.7993e-23], [4445.11, 1.60247e-22, 4.02954e-23, 1.70138e-22, 1.80326e-23], [4455.36, 1.60616e-22, 4.03841e-23, 1.70551e-22, 1.80723e-23], [4465.64, 1.60985e-22, 4.04731e-23, 1.70964e-22, 1.8112e-23], [4475.93, 1.61356e-22, 4.05624e-23, 1.71378e-22, 1.81519e-23], [4486.26, 1.61727e-22, 4.06518e-23, 1.71793e-22, 1.81918e-23], [4496.6, 1.621e-22, 4.07414e-23, 1.72209e-22, 1.82318e-23], [4506.97, 1.62473e-22, 4.08313e-23, 1.72626e-22, 1.8272e-23], [4517.36, 1.62847e-22, 4.09214e-23, 1.73044e-22, 1.83122e-23], [4527.78, 1.63222e-22, 4.10117e-23, 1.73463e-22, 1.83526e-23], [4538.22, 1.63598e-22, 4.11022e-23, 1.73882e-22, 1.8393e-23], [4548.69, 1.63975e-22, 4.11929e-23, 1.74303e-22, 1.84335e-23], [4559.18, 1.64352e-22, 4.12839e-23, 1.74725e-22, 1.84741e-23], [4569.69, 1.6473e-22, 4.1375e-23, 1.75147e-22, 1.85149e-23], [4580.23, 1.6511e-22, 4.14664e-23, 1.7557e-22, 1.85557e-23], [4590.79, 1.6549e-22, 4.1558e-23, 1.75995e-22, 1.85966e-23], [4601.38, 1.65871e-22, 4.16498e-23, 1.7642e-22, 1.86376e-23], [4611.99, 1.66253e-22, 4.17419e-23, 1.76846e-22, 1.86788e-23], [4622.62, 1.66636e-22, 4.18342e-23, 1.77273e-22, 1.872e-23], [4633.28, 1.67019e-22, 4.19266e-23, 1.77702e-22, 1.87613e-23], [4643.97, 1.67404e-22, 4.20194e-23, 1.78131e-22, 1.88027e-23], [4654.67, 1.67789e-22, 4.21123e-23, 1.78561e-22, 1.88442e-23], [4665.41, 1.68176e-22, 4.22054e-23, 1.78992e-22, 1.88858e-23], [4676.17, 1.68563e-22, 4.22988e-23, 1.79424e-22, 1.89276e-23], [4686.95, 1.68951e-22, 4.23924e-23, 1.79856e-22, 1.89694e-23], [4697.76, 1.6934e-22, 4.24863e-23, 1.8029e-22, 1.90113e-23], [4708.59, 1.6973e-22, 4.25803e-23, 1.80725e-22, 1.90533e-23], [4719.45, 1.70121e-22, 4.26746e-23, 1.81161e-22, 1.90954e-23], [4730.33, 1.70513e-22, 4.27691e-23, 1.81598e-22, 1.91377e-23], [4741.24, 1.70905e-22, 4.28638e-23, 1.82035e-22, 1.918e-23], [4752.17, 1.71299e-22, 4.29588e-23, 1.82474e-22, 1.92224e-23], [4763.13, 1.71693e-22, 4.3054e-23, 1.82914e-22, 1.92649e-23], [4774.12, 1.72088e-22, 4.31494e-23, 1.83354e-22, 1.93076e-23], [4785.13, 1.72485e-22, 4.32451e-23, 1.83796e-22, 1.93503e-23], [4796.16, 1.72882e-22, 4.33409e-23, 1.84238e-22, 1.93931e-23], [4807.22, 1.7328e-22, 4.3437e-23, 1.84682e-22, 1.94361e-23], [4818.31, 1.73679e-22, 4.35333e-23, 1.85126e-22, 1.94791e-23], [4829.42, 1.74079e-22, 4.36299e-23, 1.85572e-22, 1.95222e-23], [4840.55, 1.7448e-22, 4.37267e-23, 1.86018e-22, 1.95655e-23], [4851.72, 1.74881e-22, 4.38237e-23, 1.86466e-22, 1.96088e-23], [4862.9, 1.75284e-22, 4.3921e-23, 1.86914e-22, 1.96523e-23], [4874.12, 1.75688e-22, 4.40185e-23, 1.87364e-22, 1.96959e-23], [4885.36, 1.76092e-22, 4.41162e-23, 1.87814e-22, 1.97395e-23], [4896.62, 1.76498e-22, 4.42142e-23, 1.88266e-22, 1.97833e-23], [4907.91, 1.76904e-22, 4.43124e-23, 1.88718e-22, 1.98272e-23], [4919.23, 1.77312e-22, 4.44108e-23, 1.89172e-22, 1.98711e-23], [4930.58, 1.7772e-22, 4.45095e-23, 1.89626e-22, 1.99152e-23], [4941.95, 1.78129e-22, 4.46084e-23, 1.90082e-22, 1.99594e-23], [4953.34, 1.78539e-22, 4.47075e-23, 1.90539e-22, 2.00037e-23], [4964.76, 1.78951e-22, 4.48069e-23, 1.90996e-22, 2.00481e-23], [4976.21, 1.79363e-22, 4.49065e-23, 1.91455e-22, 2.00926e-23], [4987.69, 1.79776e-22, 4.50063e-23, 1.91914e-22, 2.01372e-23], [4999.19, 1.8019e-22, 4.51064e-23, 1.92375e-22, 2.01819e-23], [5010.72, 1.80605e-22, 4.52067e-23, 1.92836e-22, 2.02268e-23], [5022.27, 1.8102e-22, 4.53073e-23, 1.93299e-22, 2.02717e-23], [5033.85, 1.81437e-22, 4.54081e-23, 1.93763e-22, 2.03168e-23], [5045.46, 1.81855e-22, 4.55091e-23, 1.94228e-22, 2.03619e-23], [5057.1, 1.82274e-22, 4.56104e-23, 1.94693e-22, 2.04072e-23], [5068.76, 1.82694e-22, 4.5712e-23, 1.9516e-22, 2.04525e-23], [5080.45, 1.83114e-22, 4.58137e-23, 1.95628e-22, 2.0498e-23], [5092.16, 1.83536e-22, 4.59157e-23, 1.96097e-22, 2.05436e-23], [5103.91, 1.83959e-22, 4.6018e-23, 1.96567e-22, 2.05893e-23], [5115.68, 1.84382e-22, 4.61205e-23, 1.97038e-22, 2.06351e-23], [5127.47, 1.84807e-22, 4.62233e-23, 1.9751e-22, 2.0681e-23], [5139.3, 1.85233e-22, 4.63262e-23, 1.97983e-22, 2.0727e-23], [5151.15, 1.85659e-22, 4.64295e-23, 1.98457e-22, 2.07731e-23], [5163.03, 1.86087e-22, 4.6533e-23, 1.98932e-22, 2.08194e-23], [5174.93, 1.86515e-22, 4.66367e-23, 1.99408e-22, 2.08657e-23], [5186.87, 1.86945e-22, 4.67407e-23, 1.99886e-22, 2.09122e-23], [5198.83, 1.87375e-22, 4.68449e-23, 2.00364e-22, 2.09588e-23], [5210.81, 1.87807e-22, 4.69494e-23, 2.00843e-22, 2.10055e-23], [5222.83, 1.88239e-22, 4.70541e-23, 2.01324e-22, 2.10523e-23], [5234.87, 1.88673e-22, 4.71591e-23, 2.01806e-22, 2.10992e-23], [5246.95, 1.89107e-22, 4.72643e-23, 2.02288e-22, 2.11462e-23], [5259.05, 1.89543e-22, 4.73698e-23, 2.02772e-22, 2.11933e-23], [5271.17, 1.89979e-22, 4.74755e-23, 2.03257e-22, 2.12406e-23], [5283.33, 1.90417e-22, 4.75814e-23, 2.03743e-22, 2.12879e-23], [5295.51, 1.90855e-22, 4.76877e-23, 2.0423e-22, 2.13354e-23], [5307.72, 1.91295e-22, 4.77942e-23, 2.04718e-22, 2.1383e-23], [5319.96, 1.91735e-22, 4.79009e-23, 2.05207e-22, 2.14307e-23], [5332.23, 1.92177e-22, 4.80079e-23, 2.05697e-22, 2.14785e-23], [5344.53, 1.9262e-22, 4.81151e-23, 2.06188e-22, 2.15264e-23], [5356.85, 1.93063e-22, 4.82226e-23, 2.06681e-22, 2.15744e-23], [5369.2, 1.93508e-22, 4.83304e-23, 2.07174e-22, 2.16226e-23], [5381.59, 1.93954e-22, 4.84384e-23, 2.07669e-22, 2.16709e-23], [5394.0, 1.944e-22, 4.85466e-23, 2.08165e-22, 2.17193e-23], [5406.43, 1.94848e-22, 4.86552e-23, 2.08662e-22, 2.17677e-23], [5418.9, 1.95297e-22, 4.87639e-23, 2.09159e-22, 2.18164e-23], [5431.4, 1.95746e-22, 4.8873e-23, 2.09659e-22, 2.18651e-23], [5443.92, 1.96197e-22, 4.89823e-23, 2.10159e-22, 2.19139e-23], [5456.48, 1.96649e-22, 4.90918e-23, 2.1066e-22, 2.19629e-23], [5469.06, 1.97102e-22, 4.92017e-23, 2.11162e-22, 2.2012e-23], [5481.67, 1.97556e-22, 4.93117e-23, 2.11666e-22, 2.20612e-23], [5494.31, 1.98011e-22, 4.94221e-23, 2.12171e-22, 2.21105e-23], [5506.98, 1.98467e-22, 4.95327e-23, 2.12677e-22, 2.21599e-23], [5519.68, 1.98924e-22, 4.96435e-23, 2.13183e-22, 2.22095e-23], [5532.41, 1.99382e-22, 4.97547e-23, 2.13692e-22, 2.22591e-23], [5545.17, 1.99841e-22, 4.98661e-23, 2.14201e-22, 2.23089e-23], [5557.95, 2.00302e-22, 4.99777e-23, 2.14711e-22, 2.23588e-23], [5570.77, 2.00763e-22, 5.00896e-23, 2.15223e-22, 2.24088e-23], [5583.62, 2.01225e-22, 5.02018e-23, 2.15735e-22, 2.2459e-23], [5596.49, 2.01689e-22, 5.03143e-23, 2.16249e-22, 2.25092e-23], [5609.4, 2.02153e-22, 5.0427e-23, 2.16764e-22, 2.25596e-23], [5622.33, 2.02619e-22, 5.054e-23, 2.1728e-22, 2.26101e-23], [5635.3, 2.03086e-22, 5.06532e-23, 2.17797e-22, 2.26607e-23], [5648.29, 2.03553e-22, 5.07668e-23, 2.18316e-22, 2.27115e-23], [5661.32, 2.04022e-22, 5.08805e-23, 2.18835e-22, 2.27623e-23], [5674.37, 2.04492e-22, 5.09946e-23, 2.19356e-22, 2.28133e-23], [5687.46, 2.04963e-22, 5.11089e-23, 2.19878e-22, 2.28644e-23], [5700.57, 2.05435e-22, 5.12235e-23, 2.20401e-22, 2.29156e-23], [5713.72, 2.05908e-22, 5.13384e-23, 2.20925e-22, 2.2967e-23], [5726.9, 2.06383e-22, 5.14536e-23, 2.2145e-22, 2.30184e-23], [5740.1, 2.06858e-22, 5.1569e-23, 2.21977e-22, 2.307e-23], [5753.34, 2.07334e-22, 5.16847e-23, 2.22505e-22, 2.31217e-23], [5766.61, 2.07812e-22, 5.18007e-23, 2.23034e-22, 2.31736e-23], [5779.9, 2.08291e-22, 5.19169e-23, 2.23564e-22, 2.32255e-23], [5793.23, 2.0877e-22, 5.20334e-23, 2.24095e-22, 2.32776e-23], [5806.59, 2.09251e-22, 5.21502e-23, 2.24628e-22, 2.33298e-23], [5819.98, 2.09733e-22, 5.22673e-23, 2.25161e-22, 2.33821e-23], [5833.4, 2.10216e-22, 5.23846e-23, 2.25696e-22, 2.34346e-23], [5846.85, 2.107e-22, 5.25023e-23, 2.26232e-22, 2.34871e-23], [5860.34, 2.11186e-22, 5.26202e-23, 2.2677e-22, 2.35398e-23], [5873.85, 2.11672e-22, 5.27383e-23, 2.27308e-22, 2.35927e-23], [5887.4, 2.1216e-22, 5.28568e-23, 2.27848e-22, 2.36456e-23], [5900.97, 2.12648e-22, 5.29755e-23, 2.28389e-22, 2.36987e-23], [5914.58, 2.13138e-22, 5.30946e-23, 2.28931e-22, 2.37519e-23], [5928.22, 2.13629e-22, 5.32139e-23, 2.29475e-22, 2.38052e-23], [5941.89, 2.14121e-22, 5.33335e-23, 2.30019e-22, 2.38587e-23], [5955.59, 2.14614e-22, 5.34533e-23, 2.30565e-22, 2.39122e-23], [5969.33, 2.15109e-22, 5.35735e-23, 2.31112e-22, 2.39659e-23], [5983.09, 2.15604e-22, 5.36939e-23, 2.3166e-22, 2.40198e-23], [5996.89, 2.16101e-22, 5.38146e-23, 2.3221e-22, 2.40737e-23], [6010.72, 2.16599e-22, 5.39356e-23, 2.3276e-22, 2.41278e-23], [6024.58, 2.17097e-22, 5.40569e-23, 2.33312e-22, 2.4182e-23], [6038.47, 2.17598e-22, 5.41785e-23, 2.33866e-22, 2.42364e-23], [6052.4, 2.18099e-22, 5.43004e-23, 2.3442e-22, 2.42909e-23], [6066.35, 2.18601e-22, 5.44225e-23, 2.34976e-22, 2.43455e-23], [6080.34, 2.19105e-22, 5.4545e-23, 2.35533e-22, 2.44002e-23], [6094.36, 2.19609e-22, 5.46677e-23, 2.36091e-22, 2.4455e-23], [6108.42, 2.20115e-22, 5.47907e-23, 2.3665e-22, 2.451e-23], [6122.5, 2.20622e-22, 5.49141e-23, 2.37211e-22, 2.45652e-23], [6136.62, 2.2113e-22, 5.50377e-23, 2.37773e-22, 2.46204e-23], [6150.77, 2.2164e-22, 5.51616e-23, 2.38336e-22, 2.46758e-23], [6164.96, 2.2215e-22, 5.52858e-23, 2.38901e-22, 2.47313e-23], [6179.17, 2.22662e-22, 5.54102e-23, 2.39467e-22, 2.47869e-23], [6193.42, 2.23175e-22, 5.5535e-23, 2.40034e-22, 2.48427e-23], [6207.7, 2.23689e-22, 5.56601e-23, 2.40602e-22, 2.48986e-23], [6222.02, 2.24204e-22, 5.57854e-23, 2.41172e-22, 2.49547e-23], [6236.37, 2.24721e-22, 5.59111e-23, 2.41742e-22, 2.50108e-23], [6250.75, 2.25238e-22, 5.60371e-23, 2.42315e-22, 2.50671e-23], [6265.16, 2.25757e-22, 5.61638e-23, 2.42888e-22, 2.51246e-23], [6279.61, 2.26277e-22, 5.62899e-23, 2.43463e-22, 2.51801e-23], [6294.09, 2.26798e-22, 5.64167e-23, 2.44039e-22, 2.52368e-23], [6308.6, 2.27321e-22, 5.65439e-23, 2.44616e-22, 2.52937e-23], [6323.15, 2.27845e-22, 5.66713e-23, 2.45195e-22, 2.53507e-23], [6337.73, 2.28369e-22, 5.67991e-23, 2.45775e-22, 2.54078e-23], [6352.35, 2.28895e-22, 5.69271e-23, 2.46356e-22, 2.5465e-23], [6367.0, 2.29423e-22, 5.70555e-23, 2.46939e-22, 2.55224e-23], [6381.68, 2.29951e-22, 5.71842e-23, 2.47522e-22, 2.55799e-23], [6396.4, 2.30481e-22, 5.73131e-23, 2.48108e-22, 2.56375e-23], [6411.15, 2.31012e-22, 5.74424e-23, 2.48694e-22, 2.56953e-23], [6425.93, 2.31544e-22, 5.75719e-23, 2.49282e-22, 2.57532e-23], [6440.75, 2.32077e-22, 5.77018e-23, 2.49871e-22, 2.58113e-23], [6455.6, 2.32612e-22, 5.7832e-23, 2.50462e-22, 2.58695e-23], [6470.49, 2.33148e-22, 5.79625e-23, 2.51053e-22, 2.59278e-23], [6485.41, 2.33685e-22, 5.80933e-23, 2.51646e-22, 2.59863e-23], [6500.36, 2.34223e-22, 5.82244e-23, 2.52241e-22, 2.60449e-23], [6515.35, 2.34763e-22, 5.83558e-23, 2.52837e-22, 2.61036e-23], [6530.38, 2.35304e-22, 5.84875e-23, 2.53434e-22, 2.61625e-23], [6545.44, 2.35846e-22, 5.86196e-23, 2.54032e-22, 2.62215e-23], [6560.53, 2.36389e-22, 5.87519e-23, 2.54632e-22, 2.62807e-23], [6575.66, 2.36933e-22, 5.88846e-23, 2.55233e-22, 2.634e-23], [6590.82, 2.37479e-22, 5.90175e-23, 2.55836e-22, 2.63994e-23], [6606.02, 2.38026e-22, 5.91508e-23, 2.5644e-22, 2.6459e-23], [6621.26, 2.38575e-22, 5.92844e-23, 2.57045e-22, 2.65187e-23], [6636.52, 2.39124e-22, 5.94183e-23, 2.57652e-22, 2.65786e-23], [6651.83, 2.39675e-22, 5.95525e-23, 2.5826e-22, 2.66386e-23], [6667.17, 2.40227e-22, 5.96871e-23, 2.58869e-22, 2.66988e-23], [6682.54, 2.40781e-22, 5.98219e-23, 2.5948e-22, 2.6759e-23], [6697.95, 2.41335e-22, 5.99571e-23, 2.60092e-22, 2.68195e-23], [6713.4, 2.41891e-22, 6.00926e-23, 2.60706e-22, 2.688e-23], [6728.88, 2.42449e-22, 6.02284e-23, 2.61321e-22, 2.69408e-23], [6744.39, 2.43007e-22, 6.03645e-23, 2.61937e-22, 2.70016e-23], [6759.95, 2.43567e-22, 6.0501e-23, 2.62555e-22, 2.70626e-23], [6775.54, 2.44128e-22, 6.06377e-23, 2.63174e-22, 2.71237e-23], [6791.16, 2.4469e-22, 6.07748e-23, 2.63794e-22, 2.7185e-23], [6806.82, 2.45254e-22, 6.09122e-23, 2.64416e-22, 2.72465e-23], [6822.52, 2.45819e-22, 6.105e-23, 2.65039e-22, 2.7308e-23], [6838.25, 2.46385e-22, 6.1188e-23, 2.65664e-22, 2.73698e-23], [6854.02, 2.46953e-22, 6.13264e-23, 2.6629e-22, 2.74316e-23], [6869.82, 2.47522e-22, 6.14651e-23, 2.66918e-22, 2.74936e-23], [6885.67, 2.48092e-22, 6.16042e-23, 2.67547e-22, 2.75558e-23], [6901.54, 2.48664e-22, 6.17435e-23, 2.68177e-22, 2.76181e-23], [6917.46, 2.49236e-22, 6.18832e-23, 2.68809e-22, 2.76805e-23], [6933.41, 2.49811e-22, 6.20232e-23, 2.69442e-22, 2.77431e-23], [6949.4, 2.50386e-22, 6.21636e-23, 2.70077e-22, 2.78059e-23], [6965.43, 2.50963e-22, 6.23043e-23, 2.70713e-22, 2.78688e-23], [6981.49, 2.51541e-22, 6.24453e-23, 2.7135e-22, 2.79318e-23], [6997.59, 2.52121e-22, 6.25866e-23, 2.71989e-22, 2.7995e-23], [7013.72, 2.52701e-22, 6.27283e-23, 2.7263e-22, 2.80583e-23], [7029.9, 2.53284e-22, 6.28703e-23, 2.73272e-22, 2.81218e-23], [7046.11, 2.53867e-22, 6.30126e-23, 2.73915e-22, 2.81855e-23], [7062.36, 2.54452e-22, 6.31553e-23, 2.7456e-22, 2.82492e-23], [7078.64, 2.55038e-22, 6.32983e-23, 2.75206e-22, 2.83132e-23], [7094.97, 2.55626e-22, 6.34417e-23, 2.75854e-22, 2.83773e-23], [7111.33, 2.56215e-22, 6.35853e-23, 2.76503e-22, 2.84415e-23], [7127.73, 2.56805e-22, 6.37293e-23, 2.77153e-22, 2.85059e-23], [7144.16, 2.57397e-22, 6.38737e-23, 2.77805e-22, 2.85704e-23], [7160.64, 2.5799e-22, 6.40184e-23, 2.78459e-22, 2.86351e-23], [7177.15, 2.58584e-22, 6.41634e-23, 2.79114e-22, 2.86999e-23], [7193.7, 2.5918e-22, 6.43088e-23, 2.7977e-22, 2.87649e-23], [7210.29, 2.59777e-22, 6.44545e-23, 2.80428e-22, 2.88301e-23], [7226.92, 2.60375e-22, 6.46006e-23, 2.81088e-22, 2.88954e-23], [7243.58, 2.60975e-22, 6.4747e-23, 2.81749e-22, 2.89608e-23], [7260.28, 2.61576e-22, 6.48937e-23, 2.82411e-22, 2.90264e-23], [7277.03, 2.62179e-22, 6.50408e-23, 2.83075e-22, 2.90922e-23], [7293.81, 2.62783e-22, 6.51882e-23, 2.83741e-22, 2.91581e-23], [7310.63, 2.63389e-22, 6.5336e-23, 2.84408e-22, 2.92242e-23], [7327.49, 2.63995e-22, 6.54841e-23, 2.85076e-22, 2.92904e-23], [7344.38, 2.64604e-22, 6.56326e-23, 2.85746e-22, 2.93568e-23], [7361.32, 2.65213e-22, 6.57814e-23, 2.86418e-22, 2.94233e-23], [7378.29, 2.65824e-22, 6.59306e-23, 2.87091e-22, 2.949e-23], [7395.31, 2.66437e-22, 6.60801e-23, 2.87766e-22, 2.95568e-23], [7412.36, 2.67051e-22, 6.623e-23, 2.88442e-22, 2.96238e-23], [7429.46, 2.67666e-22, 6.63802e-23, 2.89119e-22, 2.9691e-23], [7446.59, 2.68282e-22, 6.65308e-23, 2.89799e-22, 2.97583e-23], [7463.76, 2.68901e-22, 6.66817e-23, 2.90479e-22, 2.98258e-23], [7480.97, 2.6952e-22, 6.6833e-23, 2.91162e-22, 2.98934e-23], [7498.22, 2.70141e-22, 6.69846e-23, 2.91845e-22, 2.99612e-23], [7515.51, 2.70763e-22, 6.71366e-23, 2.92531e-22, 3.00292e-23], [7532.84, 2.71387e-22, 6.72889e-23, 2.93218e-22, 3.00973e-23], [7550.22, 2.72013e-22, 6.74416e-23, 2.93906e-22, 3.01656e-23], [7567.63, 2.72639e-22, 6.75947e-23, 2.94596e-22, 3.0234e-23], [7585.08, 2.73267e-22, 6.77481e-23, 2.95288e-22, 3.03026e-23], [7602.57, 2.73897e-22, 6.79019e-23, 2.95981e-22, 3.03713e-23], [7620.1, 2.74528e-22, 6.8056e-23, 2.96676e-22, 3.04402e-23], [7637.67, 2.75161e-22, 6.82105e-23, 2.97372e-22, 3.05093e-23], [7655.28, 2.75795e-22, 6.83654e-23, 2.9807e-22, 3.05785e-23], [7672.94, 2.7643e-22, 6.85206e-23, 2.98769e-22, 3.06479e-23], [7690.63, 2.77067e-22, 6.86762e-23, 2.9947e-22, 3.07175e-23], [7708.37, 2.77705e-22, 6.88321e-23, 3.00173e-22, 3.07872e-23], [7726.14, 2.78345e-22, 6.89885e-23, 3.00877e-22, 3.08571e-23], [7743.96, 2.78986e-22, 6.91451e-23, 3.01583e-22, 3.09272e-23], [7761.82, 2.79629e-22, 6.93022e-23, 3.0229e-22, 3.09974e-23], [7779.71, 2.80273e-22, 6.94596e-23, 3.02999e-22, 3.10678e-23], [7797.65, 2.80919e-22, 6.96174e-23, 3.0371e-22, 3.11383e-23], [7815.64, 2.81566e-22, 6.97755e-23, 3.04422e-22, 3.1209e-23], [7833.66, 2.82215e-22, 6.99341e-23, 3.05136e-22, 3.12799e-23], [7851.72, 2.82865e-22, 7.0093e-23, 3.05852e-22, 3.13509e-23], [7869.83, 2.83517e-22, 7.02522e-23, 3.06569e-22, 3.14222e-23], [7887.98, 2.8417e-22, 7.04119e-23, 3.07288e-22, 3.14935e-23], [7906.17, 2.84825e-22, 7.05719e-23, 3.08008e-22, 3.15651e-23], [7924.4, 2.85481e-22, 7.07323e-23, 3.0873e-22, 3.16368e-23], [7942.67, 2.86139e-22, 7.0893e-23, 3.09454e-22, 3.17087e-23], [7960.99, 2.86798e-22, 7.10542e-23, 3.10179e-22, 3.17807e-23], [7979.35, 2.87459e-22, 7.12157e-23, 3.10906e-22, 3.18529e-23], [7997.75, 2.88122e-22, 7.13776e-23, 3.11635e-22, 3.19253e-23], [8016.19, 2.88785e-22, 7.15399e-23, 3.12365e-22, 3.19979e-23], [8034.68, 2.89451e-22, 7.17025e-23, 3.13097e-22, 3.20706e-23], [8053.2, 2.90118e-22, 7.18657e-23, 3.1383e-22, 3.21438e-23], [8071.77, 2.90786e-22, 7.20289e-23, 3.14565e-22, 3.22165e-23], [8090.39, 2.91456e-22, 7.21927e-23, 3.15302e-22, 3.22898e-23], [8109.04, 2.92128e-22, 7.23569e-23, 3.16041e-22, 3.23632e-23], [8127.74, 2.92801e-22, 7.25215e-23, 3.16781e-22, 3.24367e-23], [8146.49, 2.93475e-22, 7.26864e-23, 3.17523e-22, 3.25105e-23], [8165.27, 2.94152e-22, 7.28518e-23, 3.18267e-22, 3.25844e-23], [8184.1, 2.94829e-22, 7.30175e-23, 3.19012e-22, 3.26585e-23], [8202.97, 2.95509e-22, 7.31836e-23, 3.19759e-22, 3.27328e-23], [8221.89, 2.9619e-22, 7.33501e-23, 3.20508e-22, 3.28072e-23], [8240.85, 2.96872e-22, 7.3517e-23, 3.21258e-22, 3.28818e-23], [8259.85, 2.97556e-22, 7.36842e-23, 3.2201e-22, 3.29566e-23], [8278.9, 2.98242e-22, 7.38519e-23, 3.22764e-22, 3.30316e-23], [8297.99, 2.98929e-22, 7.402e-23, 3.2352e-22, 3.31067e-23], [8317.13, 2.99618e-22, 7.41884e-23, 3.24277e-22, 3.31821e-23], [8336.31, 3.00308e-22, 7.43572e-23, 3.25036e-22, 3.32575e-23], [8355.53, 3.01e-22, 7.45265e-23, 3.25797e-22, 3.33332e-23], [8374.8, 3.01694e-22, 7.46961e-23, 3.26559e-22, 3.34091e-23], [8394.11, 3.02389e-22, 7.48661e-23, 3.27323e-22, 3.34851e-23], [8413.47, 3.03085e-22, 7.50366e-23, 3.28089e-22, 3.35613e-23], [8432.87, 3.03784e-22, 7.52074e-23, 3.28857e-22, 3.36377e-23], [8452.31, 3.04484e-22, 7.53786e-23, 3.29626e-22, 3.37142e-23], [8471.81, 3.05185e-22, 7.55502e-23, 3.30397e-22, 3.3791e-23], [8491.34, 3.05889e-22, 7.57223e-23, 3.3117e-22, 3.38679e-23], [8510.92, 3.06593e-22, 7.58947e-23, 3.31945e-22, 3.3945e-23], [8530.55, 3.073e-22, 7.60675e-23, 3.32721e-22, 3.40222e-23], [8550.22, 3.08008e-22, 7.62407e-23, 3.33499e-22, 3.40997e-23], [8569.94, 3.08718e-22, 7.64144e-23, 3.34279e-22, 3.41773e-23], [8589.7, 3.09429e-22, 7.65884e-23, 3.35061e-22, 3.42552e-23], [8609.51, 3.10142e-22, 7.67629e-23, 3.35845e-22, 3.43331e-23], [8629.36, 3.10857e-22, 7.69377e-23, 3.3663e-22, 3.44113e-23], [8649.26, 3.11573e-22, 7.7113e-23, 3.37417e-22, 3.44897e-23], [8669.21, 3.12291e-22, 7.72887e-23, 3.38206e-22, 3.45682e-23], [8689.2, 3.13011e-22, 7.74647e-23, 3.38996e-22, 3.4647e-23], [8709.23, 3.13732e-22, 7.76412e-23, 3.39789e-22, 3.47259e-23], [8729.32, 3.14455e-22, 7.78181e-23, 3.40583e-22, 3.4805e-23], [8749.45, 3.15179e-22, 7.79955e-23, 3.41379e-22, 3.48843e-23], [8769.62, 3.15906e-22, 7.81732e-23, 3.42177e-22, 3.49637e-23], [8789.85, 3.16633e-22, 7.83513e-23, 3.42977e-22, 3.50434e-23], [8810.12, 3.17363e-22, 7.85299e-23, 3.43778e-22, 3.51232e-23], [8830.43, 3.18094e-22, 7.87089e-23, 3.44582e-22, 3.52033e-23], [8850.8, 3.18827e-22, 7.88883e-23, 3.45387e-22, 3.52835e-23], [8871.21, 3.19562e-22, 7.90681e-23, 3.46194e-22, 3.53639e-23], [8891.66, 3.20298e-22, 7.92483e-23, 3.47003e-22, 3.54445e-23], [8912.17, 3.21036e-22, 7.9429e-23, 3.47813e-22, 3.55252e-23], [8932.72, 3.21776e-22, 7.961e-23, 3.48626e-22, 3.56062e-23], [8953.32, 3.22518e-22, 7.97915e-23, 3.4944e-22, 3.56874e-23], [8973.96, 3.23261e-22, 7.99735e-23, 3.50256e-22, 3.57687e-23], [8994.66, 3.24006e-22, 8.01558e-23, 3.51075e-22, 3.58502e-23], [9015.4, 3.24752e-22, 8.03386e-23, 3.51894e-22, 3.5932e-23], [9036.19, 3.25501e-22, 8.05218e-23, 3.52716e-22, 3.60139e-23], [9057.03, 3.26251e-22, 8.07054e-23, 3.5354e-22, 3.6096e-23], [9077.91, 3.27002e-22, 8.08894e-23, 3.54366e-22, 3.61783e-23], [9098.85, 3.27756e-22, 8.10739e-23, 3.55193e-22, 3.62608e-23], [9119.83, 3.28511e-22, 8.12588e-23, 3.56022e-22, 3.63434e-23], [9140.86, 3.29268e-22, 8.14442e-23, 3.56854e-22, 3.64263e-23], [9161.94, 3.30027e-22, 8.163e-23, 3.57687e-22, 3.65094e-23], [9183.07, 3.30788e-22, 8.18162e-23, 3.58522e-22, 3.65926e-23], [9204.24, 3.3155e-22, 8.20028e-23, 3.59359e-22, 3.66761e-23], [9225.47, 3.32314e-22, 8.21899e-23, 3.60197e-22, 3.67597e-23], [9246.74, 3.3308e-22, 8.23774e-23, 3.61038e-22, 3.68436e-23], [9268.06, 3.33847e-22, 8.25653e-23, 3.61881e-22, 3.69276e-23], [9289.44, 3.34616e-22, 8.27537e-23, 3.62725e-22, 3.70119e-23], [9310.86, 3.35387e-22, 8.29426e-23, 3.63572e-22, 3.70963e-23], [9332.33, 3.3616e-22, 8.31318e-23, 3.6442e-22, 3.71809e-23], [9353.85, 3.36935e-22, 8.33215e-23, 3.65271e-22, 3.72658e-23], [9375.42, 3.37711e-22, 8.35117e-23, 3.66123e-22, 3.73508e-23], [9397.04, 3.3849e-22, 8.37023e-23, 3.66977e-22, 3.7436e-23], [9418.71, 3.3927e-22, 8.38933e-23, 3.67833e-22, 3.75214e-23], [9440.43, 3.40051e-22, 8.40848e-23, 3.68692e-22, 3.7607e-23], [9462.2, 3.40835e-22, 8.42767e-23, 3.69552e-22, 3.76929e-23], [9484.02, 3.4162e-22, 8.44691e-23, 3.70414e-22, 3.77789e-23], [9505.89, 3.42408e-22, 8.46619e-23, 3.71278e-22, 3.78651e-23], [9527.81, 3.43197e-22, 8.48552e-23, 3.72144e-22, 3.79515e-23], [9549.78, 3.43988e-22, 8.50489e-23, 3.73012e-22, 3.80381e-23], [9571.8, 3.4478e-22, 8.52431e-23, 3.73882e-22, 3.8125e-23], [9593.87, 3.45575e-22, 8.54377e-23, 3.74754e-22, 3.8212e-23], [9616.0, 3.46371e-22, 8.56328e-23, 3.75628e-22, 3.82992e-23], [9638.17, 3.47169e-22, 8.58283e-23, 3.76504e-22, 3.83867e-23], [9660.4, 3.47969e-22, 8.60243e-23, 3.77381e-22, 3.84743e-23], [9682.67, 3.48771e-22, 8.62207e-23, 3.78261e-22, 3.85621e-23], [9705.0, 3.49575e-22, 8.64177e-23, 3.79143e-22, 3.86502e-23], [9727.38, 3.50381e-22, 8.6615e-23, 3.80027e-22, 3.87384e-23], [9749.81, 3.51188e-22, 8.68128e-23, 3.80913e-22, 3.88269e-23], [9772.3, 3.51997e-22, 8.70111e-23, 3.81801e-22, 3.89155e-23], [9794.83, 3.52808e-22, 8.72099e-23, 3.82691e-22, 3.90044e-23], [9817.42, 3.53621e-22, 8.74091e-23, 3.83583e-22, 3.90935e-23], [9840.06, 3.54436e-22, 8.76087e-23, 3.84477e-22, 3.91828e-23], [9862.75, 3.55253e-22, 8.78089e-23, 3.85373e-22, 3.92723e-23], [9885.49, 3.56072e-22, 8.80095e-23, 3.86272e-22, 3.9362e-23], [9908.29, 3.56892e-22, 8.82105e-23, 3.87172e-22, 3.94519e-23], [9931.14, 3.57715e-22, 8.84121e-23, 3.88074e-22, 3.9542e-23], [9954.04, 3.58539e-22, 8.86141e-23, 3.88978e-22, 3.96323e-23], [9976.99, 3.59365e-22, 8.88165e-23, 3.89885e-22, 3.97228e-23], [10000.0, 3.60194e-22, 8.90195e-23, 3.90793e-22, 3.98136e-23]])

    f1 = data[:,0]
    S1 = data[:,col]**2
    fstop = [f1[0],f1[-1]]

# % ---- Interpolate to desired frequencies.  Set noise to infinity outside
# %      band of frequencies covered by f1.
# S = inf(size(f)); 
# k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop


# % -------------------------------------------------------------------------
# %                              LIGO3 spectrum
# % -------------------------------------------------------------------------

def ligo3spectrum(IFO,f):
# % LIGO3SPECTRUM - Strawman design noise curves for third-generation LIGO.  Data 
# % extracted from Figure 3 of T1400316-v5 and should be accurate to a few percent
# % in amplitude. These spectra are available from 10 Hz to 2000 Hz only.
# %
# % Arguments are the same as for the parent function. Recognised values of IFO
# % are 'LIGO-A', 'LIGO-Voyager', 'LIGO-CosmicExplorer' (case-insensitive).

    if IFO.upper()=='LIGO-A':
        data = np.array([[10.0, 3.9678080739705763e-23], [11.309284110571209, 3.0398221659058543e-23], [12.45718308341748, 2.5130640029321575e-23], [13.601530704926319, 2.0775855751257282e-23], [15.247762082845437, 1.6851937135883917e-23], [17.396345134910742, 1.3158664493151312e-23], [19.674020961625892, 1.0472208012050727e-23], [22.055222852758373, 8.494330785017142e-24], [24.942878136380717, 7.022383468430227e-24], [27.474594902238422, 6.030715466658035e-24], [30.799922117601024, 4.985677820619556e-24], [33.9261322841218, 4.447719945193319e-24], [39.393071773318546, 3.607682774264497e-24], [46.144732996187464, 3.0398221659058542e-24], [52.6470505637754, 2.660706299234269e-24], [60.06561861120157, 2.373614361213412e-24], [69.1344729755857, 2.077585575125728e-24], [82.4196060504594, 1.8890202226795184e-24], [95.70090187362572, 1.7505670753219003e-24], [114.0911370494335, 1.6222616615793647e-24], [132.4760604258842, 1.5616797507666476e-24], [149.8209405205523, 1.5033602172292137e-24], [173.96345134910743, 1.447218574511171e-24], [207.3929041512335, 1.419938926158929e-24], [249.42878136380742, 1.419938926158929e-24], [289.62234198427996, 1.419938926158929e-24], [342.2560588399126, 1.447218574511171e-24], [418.92608639188995, 1.5322425426005607e-24], [495.058462311237, 1.6534283032200178e-24], [595.4004524093444, 1.784198700846088e-24], [703.6039101022948, 2.0000000000000002e-24], [824.1960605045934, 2.2419027645873307e-24], [957.0090187362568, 2.5613446022056183e-24], [1063.4511794422535, 2.7639225134676383e-24], [1213.3035425912797, 3.098222762024156e-24], [1384.271807603447, 3.4729570877446746e-24], [1565.5123158441343, 3.893016048153975e-24], [1739.6345134910744, 4.3638817204696554e-24], [1984.7687969463493, 4.8916992467265186e-24], [2073.929041512335, 5.278585968283095e-24]])
    elif IFO.upper()=='LIGO-VOYAGER':
        data = np.array([[9.912500131952504, 2.3288724222705647e-22], [10.177323697119471, 1.818476451989422e-22], [10.728385021799559, 1.0274809943043322e-22], [11.611424327245047, 5.179086161171733e-23], [12.567145440192537, 2.926302725914811e-23], [13.84271807603448, 1.6534283032200178e-23], [15.518141037378271, 1.0472208012050727e-23], [17.244127344529417, 7.577786975592319e-24], [19.331232401691526, 5.379997350391837e-24], [21.862239943820917, 3.893016048153975e-24], [25.60925511753295, 2.7639225134676383e-24], [29.998479083578562, 2.077585575125728e-24], [34.52772300468027, 1.6534283032200178e-24], [40.80252343563839, 1.393173491227759e-24], [49.072670729843466, 1.196436076021249e-24], [60.595831335812754, 1.0472208012050726e-24], [74.82484080584663, 9.70466188329871e-25], [89.9908747327901, 9.16615127590918e-25], [115.09824517597313, 8.82384955464901e-25], [140.88182270776983, 8.657522562166054e-25], [180.18772035503366, 8.494330785017143e-25], [228.4434056085519, 8.494330785017143e-25], [297.35992787434714, 8.657522562166054e-25], [380.32307148494033, 8.993371995740937e-25], [495.058462311237, 9.70466188329871e-25], [627.6390027182711, 1.0673398462412604e-24], [761.5166254665346, 1.196436076021249e-24], [899.9087473279005, 1.2910627502415792e-24], [1101.5005867412315, 1.5033602172292137e-24], [1372.159447552731, 1.784198700846088e-24], [1679.541888074497, 2.1581809109852513e-24], [2002.288797503804, 2.5130640029321577e-24]])
    elif IFO.upper()=='LIGO-COSMICEXPLORER':
        data = np.array([[9.825765886595839, 1.1964360760212489e-23], [10.088272249062012, 1.0472208012050727e-23], [11.309284110571209, 5.917038237702881e-24], [12.789990709361842, 3.4729570877446746e-24], [14.3379991273717, 2.373614361213412e-24], [15.932725590172907, 1.7175693961372587e-24], [18.663468857943204, 1.2194218219772992e-24], [21.29336383289432, 9.342249965045526e-25], [24.724628281814923, 7.294801298506009e-25], [28.45761327865132, 5.805503791383494e-25], [33.62927907429953, 4.799492141562294e-25], [39.393071773318546, 4.0440369926888656e-25], [47.79575353611991, 3.5396790053924888e-25], [57.990757086193945, 3.2184114594537887e-25], [70.36039101022943, 3.0398221659058544e-25], [84.62153832015815, 2.982522404004116e-25], [105.41459956546393, 3.098222762024156e-25], [129.0289083390161, 3.1577453414167256e-25], [157.9331445149521, 3.343262611490267e-25], [193.31232401691528, 3.7476348457207434e-25], [236.6169225059843, 4.200916460642608e-25], [279.6178457673089, 4.709023113457543e-25], [333.3502332614619, 5.379997350391837e-25], [397.40803277607773, 6.264663730608179e-25], [486.4328550848352, 7.43494789957218e-25], [590.1907063072211, 8.82384955464901e-25], [728.7782784890753, 1.0878454152776334e-24], [868.822976815246, 1.2910627502415792e-24], [1000.0, 1.4750223152736534e-24], [1161.1424327245047, 1.7175693961372587e-24], [1360.153070492633, 2.0000000000000002e-24], [1621.5250570856872, 2.419215875408719e-24], [1984.7687969463493, 2.926302725914811e-24], [2037.7941227312338, 3.0398221659058542e-24]])

    f1 = data[:,0]
    S1 = data[:,1]**2
    fstop = [10,2000]

#     % ---- Interpolate to desired frequencies.  Set noise to infinity outside
#     %      band of frequencies covered by f1.
#     S = inf(size(f));
#     k = find(f>=f1(1) & f<=f1(end));
##### S = exp(interp1d(np.log(f1),np.log(S1),log(f(k))));
    S = interp1d(f1,S1)

    return S, fstop


# % -------------------------------------------------------------------------
# %                              LIGO3 spectrum (OLD)
# % -------------------------------------------------------------------------

def ligo3spectrumold(IFO,f):
# % LIGO3SPECTRUM - Design noise curve for LIGO3.  Data produced by Rana
# % for straw man challenge, November 2011.
# %
# % Arguments are the same as for the parent function.

# %ascii file of 3G LIGO Noise curves
# %Format:   frequency   strain   strain   strain
# %* Column 2:   Standard 3G: 100kg masses, 10dB squeezing, 2.5x thermal noise reduction
# %* Column 3:  mid-freq  3G: 100kg masses, 10dB squeezing, 5x thermal noise reduction
# %* Column 4: high-freq  3G: 100kg masses, 15dB squeezing, 2.5x thermal noise reduction
    data = np.array([[1.0, 2.4969431523912284e-15, 2.496943152391229e-15, 2.4969431523911306e-15], [1.0027727599724743, 2.4799308119902096e-15, 2.4799308119902104e-15, 2.4799308119901126e-15], [1.0055532081428138, 2.4640336853632634e-15, 2.4640336853632638e-15, 2.4640336853631664e-15], [1.0083413658285454, 2.4492474670157327e-15, 2.4492474670157334e-15, 2.4492474670156364e-15], [1.0111372544063049, 2.435570304876475e-15, 2.4355703048764756e-15, 2.4355703048763794e-15], [1.0139408953120004, 2.423002864140451e-15, 2.4230028641404516e-15, 2.4230028641403558e-15], [1.0167523100409763, 2.4115484077513084e-15, 2.4115484077513088e-15, 2.4115484077512137e-15], [1.0195715201481788, 2.401212894902249e-15, 2.4012128949022496e-15, 2.401212894902155e-15], [1.0223985472483206, 2.3920050991996837e-15, 2.392005099199684e-15, 2.3920050991995903e-15], [1.0252334130160465, 2.3839367484398114e-15, 2.3839367484398122e-15, 2.383936748439719e-15], [1.0280761391861009, 2.377022688300828e-15, 2.3770226883008283e-15, 2.377022688300736e-15], [1.030926747553492, 2.3712810726637166e-15, 2.371281072663717e-15, 2.3712810726636255e-15], [1.0337852599736614, 2.366733583753467e-15, 2.3667335837534675e-15, 2.3667335837533768e-15], [1.0366516983626506, 2.363405685855553e-15, 2.3634056858555534e-15, 2.363405685855464e-15], [1.039526084697268, 2.361326917026668e-15, 2.3613269170266683e-15, 2.3613269170265795e-15], [1.0424084410152596, 2.360531224006219e-15, 2.3605312240062194e-15, 2.3605312240061314e-15], [1.0452987894154762, 2.361057346473431e-15, 2.3610573464734315e-15, 2.3610573464733443e-15], [1.0481971520580433, 2.3629492579169994e-15, 2.362949257917e-15, 2.362949257916914e-15], [1.0511035511645315, 2.366256671732868e-15, 2.3662566717328684e-15, 2.366256671732783e-15], [1.0540180090181261, 2.371035622793245e-15, 2.3710356227932452e-15, 2.3710356227931616e-15], [1.0569405479637988, 2.3773491367030194e-15, 2.3773491367030198e-15, 2.3773491367029373e-15], [1.059871190408478, 2.385268001362562e-15, 2.3852680013625625e-15, 2.3852680013624813e-15], [1.0628099588212214, 2.3948716583960887e-15, 2.394871658396089e-15, 2.394871658396009e-15], [1.065756875733388, 2.406249235619524e-15, 2.4062492356195246e-15, 2.4062492356194457e-15], [1.068711963738811, 2.419500746188872e-15, 2.4195007461888724e-15, 2.4195007461887947e-15], [1.0716752454939704, 2.434738485618952e-15, 2.4347384856189526e-15, 2.434738485618876e-15], [1.0746467437181677, 2.4520886647931656e-15, 2.452088664793166e-15, 2.452088664793091e-15], [1.0776264811936995, 2.4716933257911615e-15, 2.471693325791162e-15, 2.4716933257910886e-15], [1.0806144807660316, 2.493712598364362e-15, 2.493712598364362e-15, 2.49371259836429e-15], [1.083610765343976, 2.5183273688822146e-15, 2.5183273688822154e-15, 2.5183273688821448e-15], [1.086615357899864, 2.5457424514878995e-15, 2.5457424514879e-15, 2.545742451487831e-15], [1.0896282814697247, 2.5761903743006607e-15, 2.576190374300661e-15, 2.5761903743005936e-15], [1.09264955915346, 2.6099359235066503e-15, 2.6099359235066507e-15, 2.609935923506585e-15], [1.0956792141150224, 2.647281627463616e-15, 2.647281627463616e-15, 2.6472816274635523e-15], [1.0987172695825929, 2.6885744148087997e-15, 2.6885744148088e-15, 2.6885744148087377e-15], [1.1017637488487577, 2.734213749648235e-15, 2.7342137496482355e-15, 2.7342137496481748e-15], [1.104818675270689, 2.7846616398161957e-15, 2.784661639816196e-15, 2.784661639816137e-15], [1.1078820722703218, 2.840455040420256e-15, 2.840455040420256e-15, 2.8404550404201987e-15], [1.1109539633345349, 2.902221348251617e-15, 2.9022213482516174e-15, 2.9022213482515618e-15], [1.1140343720153305, 2.9706979235505335e-15, 2.970697923550534e-15, 2.9706979235504802e-15], [1.1171233219300154, 3.0467569146370202e-15, 3.0467569146370206e-15, 3.046756914636969e-15], [1.1202208367613806, 3.1314371445458516e-15, 3.131437144545852e-15, 3.1314371445458023e-15], [1.1233269402578843, 3.2259855189898646e-15, 3.225985518989865e-15, 3.2259855189898173e-15], [1.1264416562338333, 3.3319114450995928e-15, 3.331911445099593e-15, 3.3319114450995474e-15], [1.1295650085695663, 3.45105929274363e-15, 3.4510592927436305e-15, 3.4510592927435867e-15], [1.1326970212116356, 3.585706284383107e-15, 3.585706284383107e-15, 3.585706284383065e-15], [1.1358377181729922, 3.738696869593792e-15, 3.738696869593792e-15, 3.738696869593753e-15], [1.138987123533169, 3.913630498175915e-15, 3.913630498175915e-15, 3.913630498175878e-15], [1.1421452614384655, 4.1151293027244286e-15, 4.1151293027244286e-15, 4.115129302724394e-15], [1.1453121561021333, 4.34922839209985e-15, 4.34922839209985e-15, 4.349228392099818e-15], [1.1484878318045617, 4.6239596909511725e-15, 4.6239596909511725e-15, 4.6239596909511426e-15], [1.1516723128934634, 4.9502513875414985e-15, 4.9502513875414985e-15, 4.950251387541471e-15], [1.1548656237840613, 5.343361740625097e-15, 5.343361740625097e-15, 5.343361740625071e-15], [1.1580677889592763, 5.8252583166690915e-15, 5.8252583166690915e-15, 5.8252583166690686e-15], [1.1612788329699146, 6.428759753956306e-15, 6.428759753956306e-15, 6.428759753956286e-15], [1.1644987804348552, 7.205177897026037e-15, 7.205177897026037e-15, 7.20517789702602e-15], [1.1677276560412404, 8.239477499272201e-15, 8.239477499272201e-15, 8.239477499272185e-15], [1.1709654845446629, 9.683272890605722e-15, 9.683272890605722e-15, 9.68327289060571e-15], [1.1742122907693573, 1.1836119246915019e-14, 1.1836119246915019e-14, 1.1836119246915008e-14], [1.17746809960839, 1.5384171068271614e-14, 1.5384171068271614e-14, 1.5384171068271604e-14], [1.1807329360238497, 2.232090924248319e-14, 2.232090924248319e-14, 2.2320909242483183e-14], [1.1840068250470388, 4.189264851061066e-14, 4.189264851061066e-14, 4.189264851061066e-14], [1.1872897917786656, 4.553406014761366e-13, 4.553406014761366e-13, 4.553406014761366e-13], [1.190581861389037, 4.941582493404276e-14, 4.941582493404276e-14, 4.941582493404275e-14], [1.1938830591182505, 2.3019413016073054e-14, 2.3019413016073054e-14, 2.3019413016073048e-14], [1.1971934102763888, 1.4829354505815323e-14, 1.4829354505815323e-14, 1.4829354505815313e-14], [1.2005129402437134, 1.0842734542197864e-14, 1.0842734542197864e-14, 1.0842734542197853e-14], [1.2038416744708587, 8.48624643466405e-15, 8.48624643466405e-15, 8.486246434664035e-15], [1.207179638479028, 6.9310295213142125e-15, 6.931029521314213e-15, 6.931029521314196e-15], [1.2105268578601889, 5.828683412556677e-15, 5.828683412556677e-15, 5.828683412556658e-15], [1.2138833582772688, 5.007201360424949e-15, 5.0072013604249495e-15, 5.0072013604249266e-15], [1.2172491654643527, 4.371879090695165e-15, 4.371879090695165e-15, 4.37187909069514e-15], [1.2206243052268801, 3.8662753923384135e-15, 3.8662753923384135e-15, 3.866275392338385e-15], [1.2240088034418426, 3.4546611552617052e-15, 3.4546611552617056e-15, 3.4546611552616736e-15], [1.2274026860579823, 3.1133172049619063e-15, 3.1133172049619063e-15, 3.1133172049618716e-15], [1.2308059790959915, 2.8258794758986705e-15, 2.825879475898671e-15, 2.825879475898633e-15], [1.2342187086487109, 2.580693379956886e-15, 2.580693379956886e-15, 2.580693379956845e-15], [1.2376409008813312, 2.3692334893353068e-15, 2.369233489335307e-15, 2.369233489335263e-15], [1.241072582031592, 2.1851195444902665e-15, 2.185119544490267e-15, 2.1851195444902195e-15], [1.2445137784099847, 2.0234817969936006e-15, 2.023481796993601e-15, 2.0234817969935505e-15], [1.2479645163999527, 1.8805390802660167e-15, 1.880539080266017e-15, 1.8805390802659634e-15], [1.2514248224580948, 1.753310822468676e-15, 1.7533108224686762e-15, 1.7533108224686192e-15], [1.2548947231143677, 1.6394158875698742e-15, 1.6394158875698746e-15, 1.6394158875698145e-15], [1.2583742449722883, 1.536929164167179e-15, 1.5369291641671793e-15, 1.5369291641671158e-15], [1.2618634147091403, 1.4442774443509534e-15, 1.4442774443509536e-15, 1.444277444350887e-15], [1.2653622590761755, 1.360162582843446e-15, 1.3601625828434463e-15, 1.3601625828433763e-15], [1.2688708048988218, 1.2835039465561415e-15, 1.2835039465561419e-15, 1.2835039465560685e-15], [1.2723890790768866, 1.2133947316720904e-15, 1.213394731672091e-15, 1.213394731672014e-15], [1.2759171085847645, 1.1490684003477406e-15, 1.149068400347741e-15, 1.1490684003476607e-15], [1.2794549204716437, 1.089872603724341e-15, 1.0898726037243413e-15, 1.0898726037242575e-15], [1.2830025418617126, 1.035248712994704e-15, 1.0352487129947044e-15, 1.0352487129946172e-15], [1.2865599999543698, 9.847156001930845e-16, 9.847156001930851e-16, 9.847156001929944e-16], [1.2901273220244298, 9.37856673802768e-16, 9.378566738027683e-16, 9.378566738026742e-16], [1.2937045354223349, 8.94309431852108e-16, 8.943094318521084e-16, 8.943094318520108e-16], [1.2972916675743624, 8.537569800770362e-16, 8.5375698007703665e-16, 8.537569800769356e-16], [1.300888745982837, 8.159210970533625e-16, 8.15921097053363e-16, 8.159210970532584e-16], [1.3044957982263408, 7.805565268669945e-16, 7.80556526866995e-16, 7.805565268668869e-16], [1.3081128519599237, 7.474462531111723e-16, 7.474462531111729e-16, 7.474462531110613e-16], [1.3117399349153176, 7.163975628653961e-16, 7.163975628653967e-16, 7.163975628652815e-16], [1.3153770749011469, 6.872387507942723e-16, 6.872387507942729e-16, 6.872387507941541e-16], [1.3190242998031432, 6.598163451374515e-16, 6.598163451374521e-16, 6.598163451373298e-16], [1.3226816375843584, 6.339927616761077e-16, 6.339927616761083e-16, 6.339927616759825e-16], [1.326349116285379, 6.096443105913427e-16, 6.096443105913432e-16, 6.096443105912138e-16], [1.330026764024542, 5.866594958164552e-16, 5.866594958164559e-16, 5.866594958163229e-16], [1.3337146089981489, 5.649375580168748e-16, 5.649375580168755e-16, 5.649375580167388e-16], [1.3374126794806833, 5.443872214443256e-16, 5.443872214443263e-16, 5.443872214441861e-16], [1.3411210038250272, 5.249256121562101e-16, 5.249256121562108e-16, 5.249256121560669e-16], [1.3448396104626776, 5.064773208836338e-16, 5.064773208836344e-16, 5.06477320883487e-16], [1.3485685279039668, 4.889735884881657e-16, 4.889735884881664e-16, 4.889735884880153e-16], [1.3523077847382776, 4.723515957105006e-16, 4.723515957105013e-16, 4.723515957103466e-16], [1.3560574096342655, 4.565538419701741e-16, 4.565538419701749e-16, 4.565538419700167e-16], [1.3598174313400766, 4.415276004691572e-16, 4.41527600469158e-16, 4.4152760046899623e-16], [1.3635878786835693, 4.2722443889598163e-16, 4.272244388959824e-16, 4.272244388958171e-16], [1.3673687805725343, 4.1359979670951246e-16, 4.135997967095133e-16, 4.1359979670934443e-16], [1.371160165994917, 4.0061261137230853e-16, 4.006126113723093e-16, 4.0061261137213696e-16], [1.374962064019039, 3.8822498705758823e-16, 3.8822498705758907e-16, 3.882249870574131e-16], [1.3787745037938217, 3.7640190031530885e-16, 3.7640190031530974e-16, 3.7640190031513027e-16], [1.3825975145490095, 3.6511093798675194e-16, 3.6511093798675283e-16, 3.6511093798656986e-16], [1.3864311255953936, 3.543220633314336e-16, 3.543220633314345e-16, 3.54322063331248e-16], [1.390275366325037, 3.4400740689799087e-16, 3.440074068979918e-16, 3.440074068978019e-16], [1.3941302662115003, 3.3414107915025214e-16, 3.3414107915025303e-16, 3.3414107915005966e-16], [1.3979958548100666, 3.2469900226606795e-16, 3.246990022660689e-16, 3.246990022658721e-16], [1.401872161757969, 3.156587588717018e-16, 3.156587588717028e-16, 3.1565875887150264e-16], [1.4057592167746176, 3.0699945576901564e-16, 3.069994557690166e-16, 3.0699945576881305e-16], [1.4096570496618273, 2.987016009641229e-16, 2.9870160096412387e-16, 2.9870160096391694e-16], [1.4135656903040459, 2.907469925219e-16, 2.90746992521901e-16, 2.9074699252169076e-16], [1.417485168668584, 2.8311861795597467e-16, 2.831186179559757e-16, 2.831186179557622e-16], [1.4214155148058443, 2.758005630235048e-16, 2.7580056302350586e-16, 2.758005630232891e-16], [1.4253567588495522, 2.687779289319422e-16, 2.687779289319433e-16, 2.6877792893172336e-16], [1.429308931016986, 2.6203675708444356e-16, 2.6203675708444465e-16, 2.620367570842216e-16], [1.43327206160921, 2.555639605942129e-16, 2.55563960594214e-16, 2.555639605939878e-16], [1.437246181011306, 2.493472618883339e-16, 2.4934726188833505e-16, 2.493472618881058e-16], [1.4412313196926059, 2.4337513580029077e-16, 2.433751358002919e-16, 2.433751358000596e-16], [1.445227508206926, 2.376367576192239e-16, 2.3763675761922503e-16, 2.3763675761898975e-16], [1.449234777192801, 2.321219556242744e-16, 2.321219556242756e-16, 2.3212195562403735e-16], [1.4532531573737189, 2.2682116768535213e-16, 2.2682116768535327e-16, 2.268211676851122e-16], [1.4572826795583569, 2.2172540155839023e-16, 2.2172540155839139e-16, 2.2172540155814753e-16], [1.4613233746408163, 2.1682619854437816e-16, 2.1682619854437935e-16, 2.1682619854413273e-16], [1.4653752736008616, 2.1211560021801092e-16, 2.121156002180121e-16, 2.1211560021776282e-16], [1.4694384075041558, 2.0758611796430555e-16, 2.0758611796430674e-16, 2.0758611796405484e-16], [1.4735128075025, 2.0323070509046213e-16, 2.0323070509046336e-16, 2.0323070509020888e-16], [1.4775985048340712, 1.9904273130623124e-16, 1.9904273130623247e-16, 1.9904273130597553e-16], [1.481695530823663, 1.9501595938934585e-16, 1.9501595938934709e-16, 1.9501595938908775e-16], [1.485803916882925, 1.9114452387367378e-16, 1.9114452387367503e-16, 1.9114452387341338e-16], [1.4899236945106038, 1.8742291161691512e-16, 1.8742291161691638e-16, 1.8742291161665248e-16], [1.4940548952927837, 1.8384594412222388e-16, 1.8384594412222514e-16, 1.8384594412195907e-16], [1.4981975509031311, 1.8040876150429879e-16, 1.8040876150430007e-16, 1.8040876150403195e-16], [1.5023516931031344, 1.771068080056169e-16, 1.771068080056182e-16, 1.771068080053481e-16], [1.50651735374235, 1.7393581898262932e-16, 1.7393581898263063e-16, 1.7393581898235865e-16], [1.5106945647586447, 1.7089180929529813e-16, 1.708918092952994e-16, 1.7089180929502567e-16], [1.5148833581784422, 1.679710630464255e-16, 1.6797106304642682e-16, 1.679710630461514e-16], [1.519083766116967, 1.6517012463003378e-16, 1.651701246300351e-16, 1.651701246297581e-16], [1.5232958207784915, 1.6248579106088253e-16, 1.6248579106088386e-16, 1.6248579106060541e-16], [1.5275195544565838, 1.5991510557016195e-16, 1.5991510557016328e-16, 1.5991510556988348e-16], [1.531754999534353, 1.574553524658744e-16, 1.5745535246587572e-16, 1.574553524655947e-16], [1.5360021884846993, 1.5510405327050387e-16, 1.551040532705052e-16, 1.5510405327022309e-16], [1.5402611538705626, 1.5285896416376467e-16, 1.52858964163766e-16, 1.5285896416348292e-16], [1.5445319283451722, 1.5071807477465184e-16, 1.507180747746532e-16, 1.5071807477436928e-16], [1.5488145446522963, 1.4867960838532373e-16, 1.4867960838532506e-16, 1.4867960838504046e-16], [1.5531090356264943, 1.4674202362976798e-16, 1.4674202362976931e-16, 1.4674202362948414e-16], [1.5574154341933677, 1.4490401779350205e-16, 1.4490401779350338e-16, 1.449040177932178e-16], [1.5617337733698131, 1.4316453184729759e-16, 1.4316453184729892e-16, 1.4316453184701308e-16], [1.5660640862642743, 1.415227573790002e-16, 1.4152275737900153e-16, 1.4152275737871557e-16], [1.5704064060769976, 1.3997814562392858e-16, 1.399781456239299e-16, 1.39978145623644e-16], [1.5747607661002851, 1.3853041883736647e-16, 1.385304188373678e-16, 1.385304188370821e-16], [1.579127199718751, 1.3717958430391157e-16, 1.371795843039129e-16, 1.3717958430362758e-16], [1.5835057404095767, 1.3592595133983022e-16, 1.3592595133983157e-16, 1.359259513395468e-16], [1.5878964217427678, 1.347701517186668e-16, 1.3477015171866812e-16, 1.347701517183841e-16], [1.5922992773814115, 1.337131640402556e-16, 1.3371316404025691e-16, 1.3371316403997383e-16], [1.5967143410819344, 1.3275634267304968e-16, 1.32756342673051e-16, 1.3275634267276902e-16], [1.6011416466943624, 1.3190145203447505e-16, 1.3190145203447636e-16, 1.319014520341957e-16], [1.605581228162578, 1.3115070714049648e-16, 1.3115070714049779e-16, 1.3115070714021865e-16], [1.6100331195245836, 1.3050682156238136e-16, 1.3050682156238266e-16, 1.3050682156210525e-16], [1.6144973549127595, 1.2997306418713116e-16, 1.2997306418713244e-16, 1.29973064186857e-16], [1.6189739685541273, 1.2955332650299252e-16, 1.2955332650299378e-16, 1.2955332650272051e-16], [1.6234629947706123, 1.2925220254270859e-16, 1.2925220254270984e-16, 1.2925220254243897e-16], [1.6279644679793055, 1.2907508414093833e-16, 1.2907508414093959e-16, 1.2907508414067135e-16], [1.6324784226927291, 1.2902827483405405e-16, 1.2902827483405528e-16, 1.2902827483378993e-16], [1.6370048935190997, 1.2911912659841632e-16, 1.2911912659841753e-16, 1.291191265981553e-16], [1.6415439151625941, 1.293562047530411e-16, 1.2935620475304228e-16, 1.2935620475278343e-16], [1.6460955224236158, 1.2974948783534848e-16, 1.2974948783534963e-16, 1.2974948783509446e-16], [1.6506597501990614, 1.3031061122140264e-16, 1.303106112214038e-16, 1.3031061122115252e-16], [1.6552366334825879, 1.31053165883949e-16, 1.3105316588395013e-16, 1.3105316588370304e-16], [1.6598262073648815, 1.319930672181629e-16, 1.31993067218164e-16, 1.3199306721792143e-16], [1.6644285070339269, 1.331490136853602e-16, 1.331490136853613e-16, 1.331490136851235e-16], [1.669043567775276, 1.345430616686256e-16, 1.3454306166862667e-16, 1.3454306166839393e-16], [1.673671424972319, 1.3620135220142162e-16, 1.3620135220142268e-16, 1.3620135220119532e-16], [1.6783121141065565, 1.3815503832464718e-16, 1.381550383246482e-16, 1.3815503832442652e-16], [1.68296567075787, 1.4044148059029498e-16, 1.4044148059029596e-16, 1.4044148059008033e-16], [1.6876321306047959, 1.431058055248193e-16, 1.4310580552482025e-16, 1.43105805524611e-16], [1.6923115294247986, 1.4620296223341646e-16, 1.462029622334174e-16, 1.4620296223321483e-16], [1.6970039030945445, 1.4980047311576855e-16, 1.4980047311576944e-16, 1.4980047311557395e-16], [1.701709287590178, 1.5398216803688711e-16, 1.5398216803688795e-16, 1.5398216803669988e-16], [1.7064277189875958, 1.5885333788812006e-16, 1.5885333788812087e-16, 1.588533378879406e-16], [1.7111592334627252, 1.645479792641861e-16, 1.6454797926418688e-16, 1.6454797926401477e-16], [1.7159038672918008, 1.7123919165324119e-16, 1.7123919165324193e-16, 1.7123919165307836e-16], [1.7206616568516415, 1.791544523947867e-16, 1.7915445239478743e-16, 1.7915445239463283e-16], [1.7254326386199312, 1.8859866524991514e-16, 1.8859866524991579e-16, 1.8859866524977059e-16], [1.7302168491754975, 1.9999002535068547e-16, 1.9999002535068609e-16, 1.9999002535055065e-16], [1.735014325198592, 2.1391786375543372e-16, 2.139178637554343e-16, 2.1391786375530908e-16], [1.7398251034711723, 2.3123997386544165e-16, 2.312399738654422e-16, 2.3123997386532766e-16], [1.744649220877183, 2.532548944387014e-16, 2.532548944387019e-16, 2.5325489443859844e-16], [1.74948671440284, 2.820264161272871e-16, 2.8202641612728753e-16, 2.820264161271957e-16], [1.754337621136912, 3.210443248447557e-16, 3.2104432484475607e-16, 3.210443248446763e-16], [1.7592019782710064, 3.767126799807975e-16, 3.767126799807978e-16, 3.7671267998073057e-16], [1.764079823099854, 4.621937767815877e-16, 4.62193776781588e-16, 4.621937767815337e-16], [1.7689711930215948, 6.095166454745557e-16, 6.0951664547455575e-16, 6.095166454745151e-16], [1.7738761255380655, 9.222570101641354e-16, 9.222570101641356e-16, 9.22257010164109e-16], [1.7787946582550853, 2.027967873591055e-15, 2.027967873591055e-15, 2.0279678735910432e-15], [1.7837268288827464, 7.438712740798656e-15, 7.438712740798656e-15, 7.438712740798653e-15], [1.7886726752357012, 1.2502089416223779e-15, 1.2502089416223779e-15, 1.250208941622359e-15], [1.7936322352334533, 6.645354433633157e-16, 6.645354433633159e-16, 6.645354433632806e-16], [1.7986055469006483, 4.442600121492095e-16, 4.442600121492098e-16, 4.442600121491576e-16], [1.803592648367365, 3.2895104078611573e-16, 3.2895104078611603e-16, 3.289510407860464e-16], [1.8085935778694069, 2.5815126317137154e-16, 2.5815126317137194e-16, 2.581512631712842e-16], [1.8136083737485975, 2.1034986644028965e-16, 2.1034986644029014e-16, 2.1034986644018365e-16], [1.818637074453072, 1.7597125505071937e-16, 1.7597125505071993e-16, 1.7597125505059404e-16], [1.8236797185375733, 1.5010572463646688e-16, 1.5010572463646752e-16, 1.5010572463632158e-16], [1.828736344663748, 1.2997620934296834e-16, 1.2997620934296908e-16, 1.299762093428024e-16], [1.8338069916004403, 1.1389381943349114e-16, 1.1389381943349196e-16, 1.1389381943330386e-16], [1.8388916982239938, 1.0077244790392913e-16, 1.0077244790393005e-16, 1.0077244790371984e-16], [1.8439905035185447, 8.988191335159938e-17, 8.988191335160039e-17, 8.988191335136733e-17], [1.8491034465763239, 8.071323173233307e-17, 8.071323173233419e-17, 8.071323173207754e-17], [1.854230566597955, 7.290077840785175e-17, 7.290077840785299e-17, 7.290077840757199e-17], [1.8593719028927564, 6.617515704010905e-17, 6.617515704011038e-17, 6.617515703980427e-17], [1.864527494879041, 6.033351456065114e-17, 6.033351456065258e-17, 6.033351456032055e-17], [1.8696973820844196, 5.5220205908788687e-17, 5.522020590879024e-17, 5.522020590843151e-17], [1.8748816041461034, 5.071382989426388e-17, 5.071382989426555e-17, 5.0713829893879285e-17], [1.8800802008112083, 4.671831450616121e-17, 4.6718314506163e-17, 4.6718314505748363e-17], [1.8852932119370591, 4.3156649043836076e-17, 4.315664904383799e-17, 4.315664904339413e-17], [1.890520677491496, 3.996638950652172e-17, 3.996638950652377e-17, 3.99663895060498e-17], [1.8957626375531795, 3.709637834508815e-17, 3.709637834509032e-17, 3.709637834458536e-17], [1.9010191323118995, 3.450431232629687e-17, 3.450431232629917e-17, 3.450431232576232e-17], [1.906290202068882, 3.2154913288993834e-17, 3.2154913288996274e-17, 3.21549132884266e-17], [1.9115758872370983, 3.0018534392250297e-17, 3.001853439225288e-17, 3.001853439164945e-17], [1.9168762283415768, 2.807008555352612e-17, 2.807008555352885e-17, 2.8070085552890714e-17], [1.9221912660197098, 2.6288195971604493e-17, 2.6288195971607362e-17, 2.6288195970933555e-17], [1.927521041021569, 2.4654554915810456e-17, 2.4654554915813476e-17, 2.465455491510301e-17], [1.932865594210216, 2.31533880750433e-17, 2.3153388075046475e-17, 2.3153388074298367e-17], [1.9382249665620148, 2.177103807247494e-17, 2.1771038072478268e-17, 2.1771038071691512e-17], [1.9435991991669486, 2.049562580241677e-17, 2.0495625802420262e-17, 2.0495625801593844e-17], [1.948988333228932, 1.931677504757361e-17, 1.9316775047577265e-17, 1.931677504671017e-17], [1.9543924100661287, 1.8225387064998376e-17, 1.82253870650022e-17, 1.8225387064093408e-17], [1.9598114711112677, 1.7213454947108997e-17, 1.7213454947112994e-17, 1.7213454946161486e-17], [1.9652455579119614, 1.6273909886118344e-17, 1.6273909886122513e-17, 1.6273909885127273e-17], [1.9706947121310228, 1.5400493216218487e-17, 1.5400493216222835e-17, 1.5400493215182858e-17], [1.9761589755467863, 1.4587649432879465e-17, 1.4587649432883995e-17, 1.4587649431798288e-17], [1.9816383900534287, 1.3830436403133292e-17, 1.3830436403138005e-17, 1.3830436402005604e-17], [1.9871329976612875, 1.3124449764235282e-17, 1.3124449764240185e-17, 1.3124449763060152e-17], [1.9926428404971857, 1.2465759118471515e-17, 1.2465759118476607e-17, 1.2465759117248055e-17], [1.998167960804754, 1.1850854111608152e-17, 1.1850854111613437e-17, 1.1850854110335527e-17], [2.003708400944754, 1.1276598863072545e-17, 1.1276598863078025e-17, 1.127659886174999e-17], [2.0092642033954045, 1.0740193521040772e-17, 1.074019352104645e-17, 1.0740193519667614e-17], [2.0148354107527044, 1.0239141963096673e-17, 1.0239141963102548e-17, 1.0239141961672345e-17], [2.020422065730764, 9.771224866755154e-18, 9.771224866761228e-18, 9.771224865279226e-18], [2.026024211162126, 9.334477544655644e-18, 9.334477544661916e-18, 9.334477543127852e-18], [2.0316418899981, 8.927172085171937e-18, 8.927172085178407e-18, 8.927172083592217e-18], [2.0372751453090894, 8.54780346764154e-18, 8.547803467648205e-18, 8.547803466010069e-18], [2.0429240202849193, 8.195079438500284e-18, 8.195079438507142e-18, 8.195079436817531e-18], [2.048588558235172, 7.867914046042431e-18, 7.867914046049477e-18, 7.867914044309214e-18], [2.0542688025895153, 7.565424843194902e-18, 7.565424843202131e-18, 7.565424841412451e-18], [2.059964796898038, 7.286933886375934e-18, 7.286933886383337e-18, 7.286933884545961e-18], [2.0656765848315835, 7.031972792917566e-18, 7.031972792925133e-18, 7.031972791042353e-18], [2.071404210182082, 6.800292281853434e-18, 6.800292281861151e-18, 6.800292279935922e-18], [2.07714771686289, 6.5918768311770294e-18, 6.591876831184881e-18, 6.591876829220915e-18], [2.0829071489091238, 6.406965366624081e-18, 6.406965366632048e-18, 6.40696536463392e-18], [2.088682550478, 6.2460792957283815e-18, 6.246079295736441e-18, 6.246079293709688e-18], [2.094473965849171, 6.11005978395941e-18, 6.110059783967535e-18, 6.110059781918757e-18], [2.1002814394250673, 6.000117043537639e-18, 6.000117043545798e-18, 6.000117041482736e-18], [2.106105015731236, 5.917895737755478e-18, 5.917895737763636e-18, 5.9178957356952286e-18], [2.111944739416683, 5.865562660172222e-18, 5.865562660180338e-18, 5.865562658116742e-18], [2.1178006552542152, 5.845926055352496e-18, 5.845926055360525e-18, 5.845926053313083e-18], [2.1236728081407845, 5.862601003937307e-18, 5.8626010039452014e-18, 5.862601001926348e-18], [2.1295612430978292, 5.920243376899868e-18, 5.920243376907576e-18, 5.920243374930672e-18], [2.1354660052716237, 6.024888032941462e-18, 6.024888032948931e-18, 6.0248880310280255e-18], [2.1413871399336206, 6.184448954583619e-18, 6.184448954590792e-18, 6.184448952740317e-18], [2.1473246924808, 6.4094770844569185e-18, 6.409477084463742e-18, 6.409477082698149e-18], [2.1532787084360168, 6.714340110084243e-18, 6.714340110090664e-18, 6.7143401084240375e-18], [2.1592492334483495, 7.119117515883071e-18, 7.11911751588904e-18, 7.11911751433471e-18], [2.165236313293451, 7.652760695199659e-18, 7.652760695205133e-18, 7.65276069377532e-18], [2.171239993873899, 8.35860960715257e-18, 8.358609607157511e-18, 8.358609605863045e-18], [2.1772603212195483, 9.304587046308338e-18, 9.304587046312713e-18, 9.304587045162827e-18], [2.1832973414878825, 1.0603438780115864e-17, 1.0603438780119649e-17, 1.0603438779121875e-17], [2.1893511009643696, 1.2456822685770796e-17, 1.2456822685773972e-17, 1.2456822684934129e-17], [2.1954216460628166, 1.5264040195912645e-17, 1.52640401959152e-17, 1.5264040195237463e-17], [2.2015090233257233, 1.9941766293286458e-17, 1.9941766293288384e-17, 1.9941766292775412e-17], [2.207613279424642, 2.9160912751673584e-17, 2.9160912751674884e-17, 2.9160912751328e-17], [2.2137344611605334, 5.538093720069875e-17, 5.538093720069943e-17, 5.5380937200518815e-17], [2.219872615464127, 6.445779388244643e-16, 6.445779388244644e-16, 6.44577938824449e-16], [2.226027789396278, 6.563504201442857e-17, 6.563504201442913e-17, 6.563504201428011e-17], [2.232200030148331, 3.109929949666063e-17, 3.1099299496661776e-17, 3.10992994963508e-17], [2.2383893850424825, 2.032620529302587e-17, 2.0326205293027598e-17, 2.0326205292557114e-17], [2.24459590153214, 1.5075978549581477e-17, 1.5075978549583776e-17, 1.5075978548956527e-17], [2.2508196272022887, 1.197138358786783e-17, 1.197138358787068e-17, 1.1971383587089588e-17], [2.2570606097698547, 9.92165141404759e-18, 9.921651414050979e-18, 9.921651413119046e-18], [2.263318897084073, 8.467731193749688e-18, 8.467731193753599e-18, 8.467731192673848e-18], [2.269594537126853, 7.382894352476305e-18, 7.382894352480724e-18, 7.382894351256152e-18], [2.2758875780131445, 6.542327866748694e-18, 6.542327866753605e-18, 6.5423278653871354e-18], [2.282198067991311, 5.8716930281901275e-18, 5.871693028195516e-18, 5.8716930266899875e-18], [2.288526055443496, 5.323979038827299e-18, 5.323979038833152e-18, 5.3239790371912924e-18], [2.2948715888859943, 4.868019699766396e-18, 4.8680196997727006e-18, 4.8680196979971226e-18], [2.301234716969626, 4.482348624791263e-18, 4.4823486247980046e-18, 4.482348622891204e-18], [2.3076154884801077, 4.1517061281978075e-18, 4.151706128204976e-18, 4.1517061261693256e-18], [2.3140139523384273, 3.864952683221282e-18, 3.864952683228863e-18, 3.864952681066624e-18], [2.3204301576012183, 3.613769968290026e-18, 3.613769968298008e-18, 3.613769966011329e-18], [2.3268641534611376, 3.391823378719823e-18, 3.391823378728197e-18, 3.3918233763191244e-18], [2.3333159892472404, 3.1942055676597955e-18, 3.1942055676685496e-18, 3.1942055651390348e-18], [2.3397857144253593, 3.017056925017265e-18, 3.0170569250263885e-18, 3.0170569223782965e-18], [2.3462733785984855, 2.857300735680672e-18, 2.857300735690155e-18, 2.8573007329252723e-18], [2.3527790315071453, 2.7124545844338957e-18, 2.7124545844437276e-18, 2.712454581563772e-18], [2.3593027230297854, 2.5804936131795296e-18, 2.580493613189701e-18, 2.5804936101963296e-18], [2.3658445031831525, 2.4597497588913784e-18, 2.4597497589018805e-18, 2.459749755796697e-18], [2.372404422122677, 2.348836414912167e-18, 2.34883641492299e-18, 2.3488364117075556e-18], [2.37898253014286, 2.2465913519327024e-18, 2.246591351943837e-18, 2.246591348619675e-18], [2.385578877677656, 2.1520329494879165e-18, 2.152032949499354e-18, 2.152032946067959e-18], [2.392193515300861, 2.0643262623319168e-18, 2.0643262623436476e-18, 2.064326258806492e-18], [2.3988264937265003, 1.9827564441318164e-18, 1.9827564441438323e-18, 1.982756440502372e-18], [2.405477863809216, 1.9067077380397887e-18, 1.906707738052081e-18, 1.9067077343077625e-18], [2.4121476765446594, 1.8356467238587724e-18, 1.8356467238713317e-18, 1.8356467200255967e-18], [2.4188359830698793, 1.7691088516926976e-18, 1.7691088517055154e-18, 1.7691088477598077e-18], [2.425542834663716, 1.7066875360566457e-18, 1.7066875360697134e-18, 1.7066875320254818e-18], [2.4322682827471938, 1.6480252616340744e-18, 1.6480252616473832e-18, 1.6480252575060874e-18], [2.4390123788839144, 1.5928062819443513e-18, 1.592806281957892e-18, 1.5928062777210066e-18], [2.4457751747804535, 1.5407505886402094e-18, 1.5407505886539735e-18, 1.5407505843229906e-18], [2.452556722286756, 1.49160890136611e-18, 1.4916089013800887e-18, 1.491608896956523e-18], [2.4593570733965358, 1.4451584826532479e-18, 1.445158482667432e-18, 1.4451584781528233e-18], [2.466176280247671, 1.401199623873359e-18, 1.4011996238877397e-18, 1.4011996192836563e-18], [2.4730143951226076, 1.359552680169803e-18, 1.3595526801843714e-18, 1.3595526754924113e-18], [2.479871470448757, 1.3200555569539512e-18, 1.320055556968698e-18, 1.3200555521904928e-18], [2.486747558798898, 1.2825615697692843e-18, 1.2825615697842002e-18, 1.2825615649214168e-18], [2.4936427128915843, 1.2469376143901868e-18, 1.246937614405263e-18, 1.2469376094596052e-18], [2.5005569855915426, 1.2130625959085309e-18, 1.2130625959237577e-18, 1.2130625908969682e-18], [2.5074904299100815, 1.1808260749940795e-18, 1.1808260750094477e-18, 1.1808260699033092e-18], [2.5144430990054993, 1.1501270970444948e-18, 1.150127097059995e-18, 1.1501270918763317e-18], [2.521415046183486, 1.1208731759829906e-18, 1.120873175998613e-18, 1.1208731707392914e-18], [2.5284063248975386, 1.0929794093350247e-18, 1.09297940935076e-18, 1.09297940401769e-18], [2.5354169887493656, 1.0663677051661953e-18, 1.066367705182034e-18, 1.0663676997771694e-18], [2.5424470914893007, 1.0409661046801014e-18, 1.0409661046960335e-18, 1.0409660992213727e-18], [2.5494966870167164, 1.0167081869066267e-18, 1.0167081869226423e-18, 1.0167081813802285e-18], [2.5565658293804328, 9.93532544072856e-19, 9.935325440889453e-19, 9.935325384808667e-19], [2.5636545727791344, 9.713823180319874e-19, 9.71382318048141e-19, 9.713823123765303e-19], [2.5707629715617872, 9.502047896026386e-19, 9.502047896188462e-19, 9.50204783885882e-19], [2.577891080228053, 9.299510138985377e-19, 9.299510139147894e-19, 9.299510081226944e-19], [2.585038953428708, 9.105754957531427e-19, 9.105754957694285e-19, 9.105754899204698e-19], [2.592206645966062, 8.920359002015454e-19, 8.920359002178557e-19, 8.920358943143444e-19], [2.599394212794379, 8.74292793702794e-19, 8.742927937191185e-19, 8.742927877634095e-19], [2.606601709020297, 8.573094123935547e-19, 8.573094124098835e-19, 8.573094064043739e-19], [2.6138291899032517, 8.410514541780395e-19, 8.410514541943624e-19, 8.410514481414915e-19], [2.6210767108559008, 8.254868918953989e-19, 8.254868919117062e-19, 8.254868858139542e-19], [2.6283443274445464, 8.105858051770607e-19, 8.105858051933423e-19, 8.105857990532298e-19], [2.635632095389565, 7.963202289236999e-19, 7.963202289399459e-19, 7.963202227600326e-19], [2.64294007056583, 7.8266401660317e-19, 7.826640166193706e-19, 7.826640104022543e-19], [2.6502683090031436, 7.695927168042197e-19, 7.695927168203648e-19, 7.695927105686802e-19], [2.6576168668866647, 7.570834616821355e-19, 7.570834616982155e-19, 7.570834554146332e-19], [2.6649858005573406, 7.451148661066874e-19, 7.451148661226925e-19, 7.451148598099174e-19], [2.6723751665123388, 7.336669364738929e-19, 7.336669364898136e-19, 7.336669301505841e-19], [2.679785021405479, 7.22720988275036e-19, 7.227209882908628e-19, 7.22720981927949e-19], [2.6872154220476685, 7.122595716316451e-19, 7.122595716473684e-19, 7.12259565263571e-19], [2.694666425407338, 7.022664041066538e-19, 7.022664041222643e-19, 7.022663977204131e-19], [2.702138088610878, 6.92726310191489e-19, 6.927263102069775e-19, 6.927263037899301e-19], [2.7096304689430766, 6.836251669486571e-19, 6.836251669640144e-19, 6.8362516053465435e-19], [2.717143623847559, 6.749498553605188e-19, 6.74949855375736e-19, 6.749498489369719e-19], [2.7246776109272277, 6.666882169993872e-19, 6.666882170144555e-19, 6.666882105692189e-19], [2.732232487944704, 6.588290156924487e-19, 6.588290157073594e-19, 6.588290092586038e-19], [2.7398083128227713, 6.513619039088244e-19, 6.513619039235689e-19, 6.51361897474268e-19], [2.7474051436448184, 6.442773936460235e-19, 6.442773936605937e-19, 6.442773872137396e-19], [2.755023038655287, 6.3756683164017895e-19, 6.375668316545666e-19, 6.37566825213169e-19], [2.7626620562601154, 6.312223787694128e-19, 6.3122237878361e-19, 6.312223723506938e-19], [2.770322255027187, 6.252369935634661e-19, 6.252369935774651e-19, 6.252369871560696e-19], [2.778003693686782, 6.196044197759744e-19, 6.196044197897675e-19, 6.196044133829444e-19], [2.7857064311320223, 6.143191780192213e-19, 6.143191780328014e-19, 6.143191716436133e-19], [2.7934305264193293, 6.093765615058143e-19, 6.093765615191742e-19, 6.093765551506934e-19], [2.801176038768873, 6.047726359882154e-19, 6.047726360013482e-19, 6.047726296566547e-19], [2.808943027565026, 6.005042440362737e-19, 6.00504244049173e-19, 6.005042377313536e-19], [2.8167315523568193, 5.965690138461768e-19, 5.965690138588362e-19, 5.965690075709822e-19], [2.8245416728583996, 5.929653728322544e-19, 5.929653728446677e-19, 5.929653665898742e-19], [2.832373448949487, 5.896925663175512e-19, 5.896925663297127e-19, 5.896925601110765e-19], [2.8402269406758336, 5.867506817113417e-19, 5.867506817232459e-19, 5.867506755438644e-19], [2.8481022082496827, 5.841406786436039e-19, 5.841406786552455e-19, 5.84140672518215e-19], [2.855999312050234, 5.818644256201685e-19, 5.818644256315427e-19, 5.818644195399572e-19], [2.863918312624101, 5.799247438701412e-19, 5.799247438812432e-19, 5.799247378381924e-19], [2.871859270685781, 5.783254591826676e-19, 5.783254591934931e-19, 5.783254532020617e-19], [2.8798222471181183, 5.770714626765331e-19, 5.770714626870783e-19, 5.7707145675034375e-19], [2.887807302972768, 5.761687816184346e-19, 5.761687816286955e-19, 5.761687757497275e-19], [2.895814499470671, 5.756246616092963e-19, 5.756246616192699e-19, 5.756246558011277e-19], [2.9038438980025143, 5.75447661699983e-19, 5.754476617096662e-19, 5.754476559553985e-19], [2.911895560129209, 5.756477642865234e-19, 5.756477642959136e-19, 5.756477586085562e-19], [2.9199695475823617, 5.762365019814308e-19, 5.762365019905256e-19, 5.762364963731004e-19], [2.9280659222647416, 5.772271040752601e-19, 5.7722710408405765e-19, 5.772270985395712e-19], [2.9361847462507638, 5.78634665708097e-19, 5.786346657165957e-19, 5.786346602480377e-19], [2.944326081786958, 5.804763434857641e-19, 5.804763434939629e-19, 5.804763381043044e-19], [2.9524899912924494, 5.82771582027402e-19, 5.8277158203529985e-19, 5.827715767274925e-19], [2.9606765373594364, 5.85542376854706e-19, 5.855423768623025e-19, 5.855423716392767e-19], [2.96888578275367, 5.888135801734852e-19, 5.888135801807805e-19, 5.888135750454438e-19], [2.9771177904149377, 5.926132575137446e-19, 5.92613257520739e-19, 5.926132524759752e-19], [2.985372623457542, 5.969731049609918e-19, 5.969731049676861e-19, 5.96973100016353e-19], [2.993650345170786, 6.019289389289179e-19, 6.019289389353132e-19, 6.019289340802419e-19], [3.0019510190194594, 6.075212732235498e-19, 6.075212732296477e-19, 6.075212684736407e-19], [3.0102747086443253, 6.137960017070375e-19, 6.137960017128399e-19, 6.137959970586696e-19], [3.0186214778626064, 6.208052094205719e-19, 6.208052094260814e-19, 6.208052048764883e-19], [3.026991390668475, 6.2860814088877755e-19, 6.286081408939968e-19, 6.286081364516887e-19], [3.035384511233545, 6.372723619364097e-19, 6.37272361941342e-19, 6.372723576089916e-19], [3.0438009039073624, 6.468751612993487e-19, 6.468751613039977e-19, 6.468751570842416e-19], [3.052240633217898, 6.575052514363767e-19, 6.575052514407466e-19, 6.575052473361834e-19], [3.060703763872045, 6.692648454113158e-19, 6.692648454154112e-19, 6.692648414285996e-19], [3.06919036075611, 6.82272210171712e-19, 6.822722101755378e-19, 6.822722063089956e-19], [3.077700488936319, 6.966648283765377e-19, 6.966648283800992e-19, 6.96664824636301e-19], [3.0862342136593073, 7.126033445768714e-19, 7.126033445801745e-19, 7.126033409615501e-19], [3.0947916003526226, 7.3027653212244205e-19, 7.302765321254933e-19, 7.302765286344258e-19], [3.1033727146252303, 7.499076022662979e-19, 7.499076022691039e-19, 7.499075989079282e-19], [3.111977622268012, 7.717622981284216e-19, 7.717622981309895e-19, 7.717622949019903e-19], [3.120606389254273, 7.961593913039779e-19, 7.961593913063154e-19, 7.961593882117253e-19], [3.129259081740245, 8.23484456021952e-19, 8.234844560240673e-19, 8.234844530660645e-19], [3.1379357660655964, 8.542081798739072e-19, 8.542081798758085e-19, 8.542081770565156e-19], [3.1466365087539385, 8.889110549794115e-19, 8.88911054981108e-19, 8.889110523025894e-19], [3.155361376513338, 9.28317202775566e-19, 9.283172027770669e-19, 9.283172002413267e-19], [3.1641104362368258, 9.733415325745986e-19, 9.73341532575914e-19, 9.73341530184894e-19], [3.1728837550029114, 1.025156796656292e-18, 1.0251567966574321e-18, 1.0251567944130096e-18], [3.1816814000760982, 1.0852910754835444e-18, 1.0852910754845197e-18, 1.0852910733885057e-18], [3.190503438907396, 1.1557731214726434e-18, 1.1557731214734652e-18, 1.1557731195276015e-18], [3.19934993913484, 1.23935541422824e-18, 1.23935541422892e-18, 1.239355412434877e-18], [3.208220968584011, 1.3398681438074852e-18, 1.339868143808035e-18, 1.3398681421674095e-18], [3.217116595268554, 1.4628035034374224e-18, 1.4628035034378547e-18, 1.4628035019521673e-18], [3.2260368873906975, 1.6163263710777328e-18, 1.6163263710780604e-18, 1.6163263697487533e-18], [3.2349819133417803, 1.813124661206653e-18, 1.813124661206889e-18, 1.8131246600353232e-18], [3.243951741702773, 2.0740440076438272e-18, 2.0740440076439855e-18, 2.074044006631438e-18], [3.2529464412448053, 2.435898633995364e-18, 2.435898633995458e-18, 2.4358986331431187e-18], [3.2619660809296915, 2.9703824154906375e-18, 2.9703824154906818e-18, 2.9703824147996527e-18], [3.2710107299104623, 3.838195747337749e-18, 3.838195747337758e-18, 3.838195746809048e-18], [3.2800804575318923, 5.489326605483972e-18, 5.489326605483961e-18, 5.489326605118483e-18], [3.2891753333310327, 9.846646939264654e-18, 9.846646939264639e-18, 9.846646939063207e-18], [3.2982954270377425, 5.3439562051538323e-17, 5.343956205153832e-17, 5.343956205150162e-17], [3.3074408085752287, 1.5059653827783358e-17, 1.5059653827783336e-17, 1.505965382765461e-17], [3.316611548060574, 6.505463350787216e-18, 6.50546335078715e-18, 6.50546335049255e-18], [3.3258077158052823, 4.111130646636862e-18, 4.111130646636736e-18, 4.111130646175864e-18], [3.335029382315814, 2.98544942064035e-18, 2.985449420640148e-18, 2.9854494200127236e-18], [3.344276618294126, 2.3318618517594904e-18, 2.331861851759195e-18, 2.331861850965056e-18], [3.353549494638213, 1.905243589990605e-18, 1.9052435899901987e-18, 1.905243589029303e-18], [3.3628480824426576, 1.6051281105718267e-18, 1.6051281105712916e-18, 1.6051281094437222e-18], [3.372172452999167, 1.3827160601455506e-18, 1.3827160601448696e-18, 1.382716058850837e-18], [3.3815226777971237, 1.21144558258705e-18, 1.2114455825862053e-18, 1.2114455811260521e-18], [3.3908988285241346, 1.0756212111698738e-18, 1.0756212111688478e-18, 1.0756212095430511e-18], [3.4003009770665757, 9.653728982899744e-19, 9.65372898288749e-19, 9.653728964979265e-19], [3.409729195510152, 8.741830257965086e-19, 8.74183025795066e-19, 8.741830238399778e-19], [3.41918355614044, 7.975746212057485e-19, 7.975746212040703e-19, 7.975746190856248e-19], [3.428664131443449, 7.323708246404922e-19, 7.323708246385608e-19, 7.323708223578171e-19], [3.4381709941061733, 6.762550193395091e-19, 6.762550193373064e-19, 6.762550168954797e-19], [3.447704217017154, 6.274982352439166e-19, 6.27498235241425e-19, 6.274982326398908e-19], [3.45726387326703, 5.847840375332007e-19, 5.847840375304023e-19, 5.847840347707006e-19], [3.466850036149107, 5.470925161687566e-19, 5.470925161656341e-19, 5.470925132494737e-19], [3.4764627791599123, 5.136213452070047e-19, 5.136213452035407e-19, 5.136213421328045e-19], [3.4861021759997635, 4.837307920574291e-19, 4.837307920536066e-19, 4.837307888303557e-19], [3.4957683005733315, 4.569046083056872e-19, 4.569046083014895e-19, 4.569046049279681e-19], [3.5054612269902066, 4.3272169761067997e-19, 4.3272169760609087e-19, 4.3272169408473153e-19], [3.5151810295654657, 4.108352489944986e-19, 4.108352489895021e-19, 4.108352453229304e-19], [3.524927782820246, 3.909571380808566e-19, 3.909571380754374e-19, 3.909571342664771e-19], [3.534701561482313, 3.7284610836232293e-19, 3.7284610835646617e-19, 3.7284610440814435e-19], [3.5445024404866334, 3.5629870638940784e-19, 3.5629870638309926e-19, 3.5629870229865194e-19], [3.554330494975953, 3.4114225138236004e-19, 3.4114225137558616e-19, 3.4114224715846335e-19], [3.564185800301367, 3.272293270445178e-19, 3.272293270372658e-19, 3.272293226911373e-19], [3.5740684320229046, 3.1443342582168162e-19, 3.144334258139396e-19, 3.144334213427003e-19], [3.5839784659101017, 3.0264547527327883e-19, 3.0264547526503565e-19, 3.0264547067281156e-19], [3.5939159779425878, 2.9177104658604697e-19, 2.9177104657729255e-19, 2.9177104186844656e-19], [3.6038810443106626, 2.817280957168408e-19, 2.817280957075662e-19, 2.817280908867039e-19], [3.613873741415887, 2.724451242759369e-19, 2.724451242661342e-19, 2.7244511933811e-19], [3.623894145871661, 2.6385967415611e-19, 2.6385967414577263e-19, 2.638596691156961e-19], [3.6339423345038186, 2.5591708987978393e-19, 2.559170898689066e-19, 2.5591708474214847e-19], [3.6440183843512104, 2.485694976211817e-19, 2.485694976097607e-19, 2.485694923919594e-19], [3.6541223726662997, 2.417749612267982e-19, 2.417749612148312e-19, 2.417749559118994e-19], [3.6642543769157525, 2.354967842760666e-19, 2.35496784263553e-19, 2.3549677888168424e-19], [3.6744144747810292, 2.297029339930705e-19, 2.2970293398001152e-19, 2.2970292852568687e-19], [3.684602744158982, 2.2436556814937267e-19, 2.2436556813577127e-19, 2.243655626157663e-19], [3.694819263162455, 2.1946065036590824e-19, 2.1946065035176965e-19, 2.194606447731612e-19], [3.705064110120879, 2.1496764271475832e-19, 2.1496764270008967e-19, 2.1496763707026336e-19], [3.7153373635808737, 2.1086926746285592e-19, 2.1086926744766673e-19, 2.1086926177432377e-19], [3.7256391023068485, 2.071513323703094e-19, 2.0715133235461164e-19, 2.0715132664577638e-19], [3.735969405281611, 2.0380261631134366e-19, 2.0380261629515181e-19, 2.0380261055917921e-19], [3.746328351706965, 2.0081481426965567e-19, 2.0081481425298694e-19, 2.008148084985702e-19], [3.7567160210043236, 1.9818254311706772e-19, 1.981825430999422e-19, 1.981825373361204e-19], [3.7671324928153185, 1.9590341217570802e-19, 1.9590341215814888e-19, 1.9590340639431487e-19], [3.7775778470024037, 1.9397816558268795e-19, 1.9397816556472145e-19, 1.939781598106299e-19], [3.7880521636494784, 1.9241090717118917e-19, 1.924109071528451e-19, 1.924109014186205e-19], [3.7985555230624906, 1.9120942328928394e-19, 1.9120942327059556e-19, 1.9120941756674061e-19], [3.80908800577006, 1.9038562516891518e-19, 1.9038562514991956e-19, 1.9038561948732343e-19], [3.819649692524091, 1.8995614081129384e-19, 1.89956140792032e-19, 1.8995613518197888e-19], [3.8302406643003954, 1.8994309787193434e-19, 1.8994309785245142e-19, 1.8994309230662928e-19], [3.8408610022993117, 1.9037515521185672e-19, 1.903751551922023e-19, 1.9037514972271147e-19], [3.8515107879463257, 1.9128886393329992e-19, 1.912888639135281e-19, 1.9128885853289035e-19], [3.862190102892696, 1.9273047243231177e-19, 1.9273047241248158e-19, 1.9273046713364916e-19], [3.872899029016083, 1.947583399998914e-19, 1.947583399800669e-19, 1.9475833481643164e-19], [3.8836376484211743, 1.974461991003588e-19, 1.9744619908060933e-19, 1.974461940460121e-19], [3.894406043440311, 2.008876231593578e-19, 2.0088762313975828e-19, 2.0088761824849848e-19], [3.905204296634124, 2.0520224100220462e-19, 2.0520224098283585e-19, 2.0520223624968092e-19], [3.9160324907921673, 2.1054453756404568e-19, 2.1054453754499452e-19, 2.1054453298518977e-19], [3.9268907089335445, 2.1711657759741047e-19, 2.171165775787702e-19, 2.171165732080487e-19], [3.9377790343075567, 2.251868434711344e-19, 2.25186843453005e-19, 2.2518683928759777e-19], [3.9486975503943342, 2.3511889989246714e-19, 2.3511889987495557e-19, 2.351188959316016e-19], [3.9596463409054747, 2.4741642181733947e-19, 2.4741642180056004e-19, 2.4741641809651663e-19], [3.9706254897846924, 2.627966141236448e-19, 2.627966141077195e-19, 2.62796610660773e-19], [3.9816350812084536, 2.8231534490661776e-19, 2.8231534489167635e-19, 2.823153417201526e-19], [3.9926751995866288, 3.0759212147392093e-19, 3.0759212146010176e-19, 3.075921185828768e-19], [4.0037459295631335, 3.412420826945425e-19, 3.4124208268199244e-19, 3.412420801185037e-19], [4.014847356016583, 3.8777747503492647e-19, 3.877774750238016e-19, 3.8777747279405873e-19], [4.025979564060941, 4.557055096703955e-19, 4.557055096608612e-19, 4.557055077854572e-19], [4.037142639046169, 5.632010038536288e-19, 5.632010038458604e-19, 5.632010023459829e-19], [4.048336666558885, 7.57279164380858e-19, 7.572791643750413e-19, 7.572791632724827e-19], [4.0595617324230195, 1.208589533287596e-18, 1.2085895332839273e-18, 1.2085895326010923e-18], [4.0708179227004715, 3.396920149819104e-18, 3.39692014981779e-18, 3.3969201495776614e-18], [4.082105323691766, 3.558128441803245e-18, 3.558128441801983e-18, 3.5581284415753924e-18], [4.093424021936723, 1.1109256378731482e-18, 1.1109256378690807e-18, 1.1109256371517655e-18], [4.104774104215116, 6.387343182925126e-19, 6.387343182853955e-19, 6.387343170522769e-19], [4.116155657547332, 4.386804406407861e-19, 4.386804406303627e-19, 4.386804388557433e-19], [4.1275687691950536, 3.284732296917604e-19, 3.284732296777601e-19, 3.28473227335257e-19], [4.1390135266619135, 2.5889949290406575e-19, 2.5889949288620386e-19, 2.5889948994873087e-19], [4.150490017694171, 2.1112915373450823e-19, 2.111291537124854e-19, 2.111291501522391e-19], [4.1619983302813885, 1.764035382062883e-19, 1.7640353817978983e-19, 1.764035339682359e-19], [4.173538552657098, 1.5009644108935901e-19, 1.5009644105805442e-19, 1.5009643616591423e-19], [4.185110773299484, 1.295343289214556e-19, 1.2953432888499782e-19, 1.2953432328223484e-19], [4.1967150809320595, 1.1306461828264383e-19, 1.1306461824066867e-19, 1.1306461189647483e-19], [4.208351564524347, 9.961095401402088e-20, 9.961095396614639e-20, 9.961094684892811e-20], [4.220020313292561, 8.844231209831274e-20, 8.844231204413847e-20, 8.844230412150276e-20], [4.231721416700288, 7.904488360488217e-20, 7.904488354398853e-20, 7.904487478272876e-20], [4.243454964459176, 7.104704969659588e-20, 7.104704962854336e-20, 7.1047039994625e-20], [4.255221046529626, 6.417343869415677e-20, 6.417343861848528e-20, 6.41734280770313e-20], [4.267019753121475, 5.82157116613273e-20, 5.821571157755533e-20, 5.821570009283164e-20], [4.278851174694687, 5.301337705164718e-20, 5.3013376959271044e-20, 5.3013364494671625e-20], [4.290715401960056, 4.8440835297411e-20, 4.8440835195904025e-20, 4.844082171393594e-20], [4.302612525879891, 4.4398414509120347e-20, 4.4398414397932024e-20, 4.439839986020026e-20], [4.314542637668716, 4.0806033038378486e-20, 4.0806032916933605e-20, 4.0806017284125766e-20], [4.326505828793978, 3.7598632780645764e-20, 3.759863264834354e-20, 3.759861588021434e-20], [4.338502190976736, 3.4722831779246703e-20, 3.472283163545982e-20, 3.472281369081546e-20], [4.350531816192368, 3.213443258513779e-20, 3.2134432429211494e-20, 3.213441326589384e-20], [4.362594796671283, 2.979654163900641e-20, 2.9796541470257516e-20, 2.9796521045128086e-20], [4.3746912248996175, 2.7678131786109867e-20, 2.7678131603825755e-20, 2.7678109872749633e-20], [4.386821193619955, 2.5752930759448653e-20, 2.575293056288623e-20, 2.5752907480715732e-20], [4.398984795832026, 2.399855257834854e-20, 2.3998552366733184e-20, 2.3998527887291407e-20], [4.411182124793432, 2.2395812140914143e-20, 2.2395811913438588e-20, 2.239578598950278e-20], [4.423413274020354, 2.0928179497042696e-20, 2.0928179252865925e-20, 2.092815183615071e-20], [4.43567833728827, 1.9581341710762556e-20, 1.958134144900861e-20, 1.958131249014907e-20], [4.447977408632674, 1.8342848377059623e-20, 1.8342848096816426e-20, 1.8342817545350982e-20], [4.460310582349801, 1.720182275453591e-20, 1.7201822454854018e-20, 1.720179025920722e-20], [4.472677952997345, 1.6148724786326236e-20, 1.6148724466217576e-20, 1.6148690573682692e-20], [4.4850796153951835, 1.517515546741591e-20, 1.517515512585247e-20, 1.5175119482573958e-20], [4.497515664626113, 1.4273694394055237e-20, 1.427369402996772e-20, 1.4273656580923546e-20], [4.509986196036565, 1.3437764121896788e-20, 1.3437763734173232e-20, 1.3437724423157035e-20], [4.522491305237347, 1.2661516320304057e-20, 1.2661515907788401e-20, 1.2661474677391554e-20], [4.535031088104372, 1.1939735752797182e-20, 1.1939735314287813e-20, 1.1939692105881309e-20], [4.547605640779396, 1.1267758918416247e-20, 1.126775845266449e-20, 1.1267713206380718e-20], [4.560215059670747, 1.0641404814640071e-20, 1.0641404320348628e-20, 1.0641356975063033e-20], [4.572859441454077, 1.005691577252997e-20, 1.0056915248351339e-20, 1.0056865741663934e-20], [4.5855388830730925, 9.510906700965133e-21, 9.510906145499968e-21, 9.510854413716751e-21], [4.5982534817403025, 9.000321383064839e-21, 9.00032079486026e-21, 9.000266772974511e-21], [4.611003334937763, 8.522394712147871e-21, 8.522394089695735e-21, 8.522337711369215e-21], [4.623788540417824, 8.074619950483422e-21, 8.07461929221854e-21, 8.074560489762587e-21], [4.636609196203881, 7.654720252040142e-21, 7.654719556338454e-21, 7.654658260694983e-21], [4.649465400591122, 7.26062381844642e-21, 7.26062308362311e-21, 7.260559224345545e-21], [4.6623572521472845, 6.890442161599564e-21, 6.890441385907133e-21, 6.890374891141247e-21], [4.675284849713414, 6.5424510316238226e-21, 6.542450213250101e-21, 6.5423810097148e-21], [4.688248292404616, 6.215073638928127e-21, 6.215072775994237e-21, 6.215000788962288e-21], [4.701247679610816, 5.906865856907778e-21, 5.906864947466042e-21, 5.906790100744744e-21], [4.714283110997529, 5.61650313971344e-21, 5.616502181745229e-21, 5.6164243976570606e-21], [4.727354686506616, 5.34276892932324e-21, 5.342767920736766e-21, 5.342687120100009e-21], [4.7404625063570505, 5.084544359383733e-21, 5.084543298011788e-21, 5.084459400121126e-21], [4.753606671045692, 4.840799091120134e-21, 4.840797974717758e-21, 4.840710897324901e-21], [4.766787281348056, 4.6105831400107444e-21, 4.610581966252845e-21, 4.610491625547154e-21], [4.780004438319077, 4.393019571646233e-21, 4.393018338125132e-21, 4.392924648714283e-21], [4.793258243293898, 4.187297961881904e-21, 4.187296666104829e-21, 4.187199540995535e-21], [4.806548797888636, 3.992668530549504e-21, 3.992667169936011e-21, 3.992566520514804e-21], [4.819876204001168, 3.8084368700423696e-21, 3.808435441921702e-21, 3.808331177935807e-21], [4.833240563811904, 3.633959200370276e-21, 3.6339577019786526e-21, 3.633849731516976e-21], [4.84664197978458, 3.4686380910765786e-21, 3.468636519554419e-21, 3.468524749028672e-21], [4.860080554667041, 3.311918597959268e-21, 3.311916950348349e-21, 3.3118012844743118e-21], [4.873556391492023, 3.163284769028923e-21, 3.1632830422694565e-21, 3.1631633840484452e-21], [4.887069593577949, 3.0222564797339274e-21, 3.022254670661578e-21, 3.0221309213621106e-21], [4.900620264529717, 2.888386562322164e-21, 2.8883846676649904e-21, 2.8882567268046976e-21], [4.914208508239503, 2.7612581983978103e-21, 2.761256214773129e-21, 2.7611239801009367e-21], [4.927834428887543, 2.640482547372102e-21, 2.6404804712832823e-21, 2.640343838761899e-21], [4.941498130942943, 2.5256965866709406e-21, 2.5256944145041234e-21, 2.5255532782928666e-21], [4.955199719164479, 2.4165611423234517e-21, 2.41655887034419e-21, 2.416413122782205e-21], [4.968939298601393, 2.312759090966622e-21, 2.31275671531645e-21, 2.3126062469063346e-21], [4.982716974594211, 2.2139937164114296e-21, 2.21399123310435e-21, 2.2138359324962596e-21], [4.996532852775535, 2.1199872057681195e-21, 2.1199846106870153e-21, 2.119824364663271e-21], [5.010387039070864, 2.0304792717535223e-21, 2.0304765606464865e-21, 2.030311254106762e-21], [5.0242796396994045, 1.9452258892358807e-21, 1.9452230577124785e-21, 1.9450525736595823e-21], [5.038210761174881, 1.863998135334968e-21, 1.863995178862406e-21, 1.86381939838876e-21], [5.052180510306356, 1.7865811235096197e-21, 1.786578037408844e-21, 1.786396839683687e-21], [5.06618899419905, 1.7127730230520707e-21, 1.7127698024937865e-21, 1.7125830647511798e-21], [5.080236320255153, 1.642384156281948e-21, 1.6423807962825604e-21, 1.6421883938102586e-21], [5.094322596174668, 1.5752361665087062e-21, 1.5752326619261821e-21, 1.5750344680554567e-21], [5.108447929956212, 1.5111612505209284e-21, 1.5111575960505787e-21, 1.5109534821470655e-21], [5.122612429897865, 1.4500014499743735e-21, 1.4499976401445646e-21, 1.4497874756001921e-21], [5.136816204597985, 1.3916079965979208e-21, 1.3916040257657024e-21, 1.3913876779918508e-21], [5.151059362956053, 1.335840706624997e-21, 1.3358365689716636e-21, 1.3356139033935852e-21], [5.165342014173498, 1.282567420294481e-21, 1.2825631098210583e-21, 1.2823339898736832e-21], [5.1796642677545375, 1.2316634826559177e-21, 1.231658993178577e-21, 1.2314232803038348e-21], [5.1940262335070235, 1.1830112622640038e-21, 1.1830065874094093e-21, 1.1827641410551456e-21], [5.208428021543272, 1.1364997046616738e-21, 1.1364948378622596e-21, 1.1362455154828772e-21], [5.222869742280922, 1.092023917833282e-21, 1.0920188523224827e-21, 1.0917625093814239e-21], [5.237351506443766, 1.0494847870635647e-21, 1.0494795158709662e-21, 1.0492160058452135e-21], [5.251873425062612, 1.0087886168665367e-21, 1.0087831328129794e-21, 1.008512307199657e-21], [5.266435609476128, 9.698467978550052e-22, 9.698410935476472e-22, 9.6956280187291e-22], [5.281038171331696, 9.32575496607276e-22, 9.325695644345932e-22, 9.322836542649625e-22], [5.295681222586275, 8.968953667559553e-22, 8.968891988827151e-22, 8.965955158389955e-22], [5.310364875507244, 8.627312796759709e-22, 8.62724868038165e-22, 8.624232558121835e-22], [5.325089242673285, 8.300120732864459e-22, 8.300054095861936e-22, 8.296957099604715e-22], [5.339854436975225, 7.98670317606434e-22, 7.986633933068797e-22, 7.983454461774818e-22], [5.3546605716169084, 7.686420958172426e-22, 7.686349021374224e-22, 7.683085455401862e-22], [5.3695077601160754, 7.398667996875274e-22, 7.398593275972496e-22, 7.395243977376073e-22], [5.384396116305215, 7.122869383105589e-22, 7.122791785252841e-22, 7.119355098118864e-22], [5.3993257543324535, 6.85847959188391e-22, 6.858399021641308e-22, 6.854873272465619e-22], [5.414296788662417, 6.604980807750724e-22, 6.604897167033253e-22, 6.601280665140965e-22], [5.429309334077116, 6.361881356618383e-22, 6.361794544645605e-22, 6.358085582657092e-22], [5.444363505676828, 6.128714236516147e-22, 6.128624149762587e-22, 6.124821004107876e-22], [5.459459418880967, 5.9050357402915705e-22, 5.904942272437685e-22, 5.901043203922294e-22], [5.474597189428989, 5.690424163868631e-22, 5.690327205752794e-22, 5.686330460177946e-22], [5.489776933381259, 5.484478594156075e-22, 5.484378033727618e-22, 5.480281842568155e-22], [5.504998767119951, 5.286817771149742e-22, 5.286713493423169e-22, 5.282516074566015e-22], [5.520262807349943, 5.097079019185632e-22, 5.096970906196462e-22, 5.092670464743577e-22], [5.535569171099701, 4.914917242678971e-22, 4.914805173440968e-22, 4.910399902579894e-22], [5.55091797572219, 4.74000398203137e-22, 4.73988783249604e-22, 4.735375914441678e-22], [5.56630933889576, 4.572026525708285e-22, 4.571906168726269e-22, 4.567285775737938e-22], [5.5817433786250605, 4.410687074780241e-22, 4.410562380065414e-22, 4.4058316755422395e-22], [5.597220213241938, 4.2557019564927475e-22, 4.255572790588704e-22, 4.250729930248701e-22], [5.612739961406338, 4.106800883675702e-22, 4.1066671099255384e-22, 4.10171024307066e-22], [5.628302742107234, 3.9637262570331456e-22, 3.9635877355524683e-22, 3.958515006425232e-22], [5.643908674663516, 3.8262325075637424e-22, 3.826089095217368e-22, 3.820898644452563e-22], [5.659557878724924, 3.694085476555953e-22, 3.6939370269386815e-22, 3.6886269931140284e-22], [5.675250474272955, 3.5670618307822657e-22, 3.5669081942041424e-22, 3.561476715494725e-22], [5.690986581621786, 3.4449485106806514e-22, 3.444789534156837e-22, 3.4392347500977654e-22], [5.706766321419195, 3.3275422094643665e-22, 3.327377736710467e-22, 3.321697790071114e-22], [5.72258981464749, 3.2146488812434674e-22, 3.2144787526763347e-22, 3.208671791451387e-22], [5.738457182624435, 3.1060832763702266e-22, 3.105907329115156e-22, 3.0999715086356033e-22], [5.754368547004175, 3.0016685023429833e-22, 3.001486570247189e-22, 2.9954200554162115e-22], [5.770324029778173, 2.9012356087140696e-22, 2.9010475223674118e-22, 2.89484849002457e-22], [5.786323753276149, 2.8046231945509693e-22, 2.804428781314139e-22, 2.7980954227314653e-22], [5.80236784016701, 2.711677037096929e-22, 2.7114761211380247e-22, 2.705006644651802e-22], [5.818456413459799, 2.6222497403658563e-22, 2.622042142705251e-22, 2.615434776486088e-22], [5.834589596504625, 2.5362004024891957e-22, 2.5359859410535314e-22, 2.529238936018578e-22], [5.850767512993629, 2.4533943007087965e-22, 2.4531727903951925e-22, 2.446284423263468e-22], [5.866990286961913, 2.373702592984002e-22, 2.3734738457340276e-22, 2.366442422228579e-22], [5.883258042788496, 2.2970020352433235e-22, 2.2967658601278675e-22, 2.2895897183242403e-22], [5.899570905197279, 2.223174713376896e-22, 2.2229309166925324e-22, 2.2156084305160733e-22], [5.915928999257983, 2.1521077891216686e-22, 2.1518561744990954e-22, 2.1443857573697374e-22], [5.932332450387127, 2.0836932590455472e-22, 2.0834336275700551e-22, 2.0758137361955783e-22], [5.9487813843489725, 2.0178277258856564e-22, 2.0175598762309362e-22, 2.0097890145449822e-22], [5.965275927256495, 1.954412181545066e-22, 1.954135910120358e-22, 1.9462126333649934e-22], [5.981816205572358, 1.8933518010929116e-22, 1.8930669022038746e-22, 1.8849898211516942e-22], [5.998402346109865, 1.8345557471552489e-22, 1.834262013179702e-22, 1.826029798491215e-22], [6.015034476033955, 1.777936984122896e-22, 1.7776342057007678e-22, 1.769245592411332e-22], [6.0317127228621565, 1.7234121016349143e-22, 1.7231000678736936e-22, 1.714553860002483e-22], [6.048437214465572, 1.67090114683343e-22, 1.6705796455287388e-22, 1.6618747208021403e-22], [6.065208079069868, 1.620327464913263e-22, 1.6199962827852412e-22, 1.6111315974622851e-22], [6.082025445256241, 1.5716175475215594e-22, 1.5712764704668758e-22, 1.5622510642581196e-22], [6.0988894419624184, 1.5247008885883707e-22, 1.5243497019478452e-22, 1.5151627030122897e-22], [6.115800198483637, 1.4795098471944634e-22, 1.4791483360367827e-22, 1.4697989660416808e-22], [6.132757844473645, 1.4359795171086585e-22, 1.4356074665294032e-22, 1.4260950457581195e-22], [6.14976250994568, 1.3940476026459464e-22, 1.393664798082271e-22, 1.3839887505690706e-22], [6.16681432527348, 1.3536543005231092e-22, 1.3532605280828278e-22, 1.3434203867563444e-22], [6.18391342119228, 1.314742187403067e-22, 1.3143372342092183e-22, 1.3043326460203773e-22], [6.2010599287998085, 1.27725611284371e-22, 1.2768397673915909e-22, 1.266670498405674e-22], [6.2182539795573, 1.2411430973772697e-22, 1.2407151499056663e-22, 1.230381090326649e-22], [6.235495705290497, 1.2063522354697505e-22, 1.2059124783435983e-22, 1.1954136474497173e-22], [6.252785238190661, 1.1728346031190085e-22, 1.1723828312234597e-22, 1.1617193821805737e-22], [6.270122710815596, 1.140543169868383e-22, 1.140079181012136e-22, 1.1292514055368852e-22], [6.287508256090646, 1.109432715023352e-22, 1.1089563103507946e-22, 1.0979646431902455e-22], [6.304942007309737, 1.0794597478729749e-22, 1.0789707322829078e-22, 1.0678157554782255e-22], [6.322424098136379, 1.0505824317284274e-22, 1.0500806142975888e-22, 1.0387630611994575e-22], [6.339954662604698, 1.0227605116004898e-22, 1.0222457060106286e-22, 1.0107664650090806e-22], [6.3575338351204715, 9.959552453509625e-23, 9.954272703169984e-23, 9.837873882536414e-23], [6.375161750462144, 9.701293381590592e-23, 9.695880178560037e-23, 9.577887030835346e-23], [6.392838543781877, 9.45246880153117e-23, 9.446920446401531e-23, 9.327346696933248e-23], [6.410564350606563, 9.212732870670998e-23, 9.207047727054428e-23, 9.085908765521373e-23], [6.428339306838899, 8.981752437865776e-23, 8.975928936499095e-23, 8.853241834879968e-23], [6.446163548758387, 8.759206506569097e-23, 8.753243149305289e-23, 8.629026675022845e-23], [6.464037213022405, 8.54478572430971e-23, 8.538681087983512e-23, 8.412955711895636e-23], [6.48196043666726, 8.338191897411643e-23, 8.331944637543963e-23, 8.204732536576235e-23], [6.499933357109212, 8.139137529825776e-23, 8.132746384145072e-23, 8.004071438264094e-23], [6.517956112145557, 7.94734538501578e-23, 7.940809176772988e-23, 7.81069696015919e-23], [6.53602883995566, 7.762548069856712e-23, 7.755865710904552e-23, 7.624343477117475e-23], [6.554151679102025, 7.584487639555866e-23, 7.577658133171812e-23, 7.444754794211416e-23], [6.572324768531367, 7.412915222626073e-23, 7.405937666056208e-23, 7.271683765212902e-23], [6.59054824757565, 7.247590664992466e-23, 7.240464251694032e-23, 7.104891930195905e-23], [6.608822255953189, 7.088282192305953e-23, 7.081006213873899e-23, 6.944149171294043e-23], [6.627146933769696, 6.934766089589725e-23, 6.927339937345966e-23, 6.789233385869919e-23], [6.645522421519358, 6.786826397372392e-23, 6.779249563600858e-23, 6.639930176287369e-23], [6.663948860085928, 6.644254623429226e-23, 6.636526702258077e-23, 6.496032555407858e-23], [6.682426390743791, 6.506849469365474e-23, 6.498970157273037e-23, 6.35734066719904e-23], [6.700955155159052, 6.374416571190488e-23, 6.366385667134332e-23, 6.223661521532827e-23], [6.719535295390622, 6.246768253140214e-23, 6.238585658307854e-23, 6.094808742632514e-23], [6.7381669538913105, 6.123723293961304e-23, 6.115389011143256e-23, 5.97060233026546e-23], [6.7568502735089115, 6.005106704929204e-23, 5.996620837522741e-23, 5.850868433141703e-23], [6.775585397487297, 5.890749518880737e-23, 5.882112269508777e-23, 5.735439133768155e-23], [6.7943724694675325, 5.780488589547214e-23, 5.771700258332496e-23, 5.624152244065145e-23], [6.813211633488953, 5.674166400520379e-23, 5.665227383013503e-23, 5.516851111182282e-23], [6.832103033990289, 5.571630883209488e-23, 5.562541667975058e-23, 5.413384432825856e-23], [6.851046815810759, 5.472735243128606e-23, 5.463496409029251e-23, 5.313606081467801e-23], [6.870043124191186, 5.37733779394666e-23, 5.367950007131929e-23, 5.2173749369380996e-23], [6.889092104775117, 5.285301798706343e-23, 5.2757658093315283e-23, 5.1245547267292724e-23], [6.9081939036099245, 5.196495317668276e-23, 5.18681195636407e-23, 5.035013873522945e-23], [6.927348667147947, 5.11079106226739e-23, 5.100961236396985e-23, 4.948625349423102e-23], [6.946556542247589, 5.028066254692975e-23, 5.01809094440678e-23, 4.8652665363396025e-23], [6.965817676174463, 4.948202492639315e-23, 4.9380827467761085e-23, 4.7848190921047524e-23], [6.985132216602515, 4.8710856188036476e-23, 4.860822550643068e-23, 4.7071688218625275e-23], [7.004500311615151, 4.796605594742829e-23, 4.7862003776548844e-23, 4.632205554277731e-23], [7.023922109706382, 4.724656378726747e-23, 4.7141102417360556e-23, 4.5598230222197315e-23], [7.0433977597819535, 4.655135807258553e-23, 4.6444500305643024e-23, 4.489918747526703e-23], [7.062927411160493, 4.587945479978808e-23, 4.5771213904340134e-23, 4.422393929501808e-23], [7.0825112135746515, 4.5229906476698265e-23, 4.5120296142710145e-23, 4.357153336880697e-23], [7.10214931717225, 4.4601801031352403e-23, 4.449083532531869e-23, 4.294105202923333e-23], [7.121841872517442, 4.3994260747493325e-23, 4.388195406793962e-23, 4.2331611234695086e-23], [7.14158903059185, 4.340644122476788e-23, 4.329280825859926e-23, 4.174235957583122e-23], [7.161390942795738, 4.2837530362340277e-23, 4.272258604187746e-23, 4.1172477307656806e-23], [7.181247760949164, 4.228674736430651e-23, 4.2170506825545495e-23, 4.062117540426304e-23], [7.201159637293144, 4.175334176602687e-23, 4.163582030795493e-23, 4.008769463499965e-23], [7.22112672449083, 4.1236592480224826e-23, 4.1117805525700246e-23, 3.9571304660676514e-23], [7.24114917562866, 4.07358068624518e-23, 4.061576992046192e-23, 3.9071303149342926e-23], [7.26122714421756, 4.0250319794954765e-23, 4.012904842442854e-23, 3.8587014909400747e-23], [7.281360784194092, 3.977949278885256e-23, 3.965700256423966e-23, 3.811779104087151e-23], [7.301550249921649, 3.9322713103970786e-23, 3.9199019582562645e-23, 3.766300810251253e-23], [7.321795696191643, 3.887939288660619e-23, 3.875451157758288e-23, 3.722206729623646e-23], [7.342097278224678, 3.8448968324445476e-23, 3.832291465984131e-23, 3.679439366634733e-23], [7.362455151671753, 3.8030898819332955e-23, 3.790368812680262e-23, 3.6379435315585724e-23], [7.382869472615446, 3.7624666177229224e-23, 3.749631365468234e-23, 3.5976662635666403e-23], [7.403340397571117, 3.722977381588287e-23, 3.7100294507977095e-23, 3.5585567554135726e-23], [7.423868083488107, 3.684574599010188e-23, 3.6715154766506113e-23, 3.520566279613541e-23], [7.44445268775093, 3.6472127034645884e-23, 3.6340438570193486e-23, 3.4836481161775525e-23], [7.465094368180506, 3.610848062516332e-23, 3.597570938171779e-23, 3.4477574819246535e-23], [7.48579328303534, 3.575438905685292e-23, 3.5620549266835926e-23, 3.412851461316591e-23], [7.506549591012758, 3.5409452541436876e-23, 3.5274558193051486e-23, 3.378888938919379e-23], [7.527363451250114, 3.507328852201577e-23, 3.4937353346089975e-23, 3.3458305333875884e-23], [7.548235023326006, 3.4745531006415937e-23, 3.4608568464686836e-23, 3.313638533111221e-23], [7.569164467261515, 3.4425829918606893e-23, 3.4287853193559713e-23, 3.2822768333957306e-23], [7.59015194352141, 3.411385046856404e-23, 3.3974872454550787e-23, 3.251710875331833e-23], [7.611197613015405, 3.380927254029988e-23, 3.366930583607686e-23, 3.221907586226111e-23], [7.632301637099369, 3.3511790098268306e-23, 3.337084700069038e-23, 3.1928353216884966e-23], [7.653464177576567, 3.3221110611888826e-23, 3.3079203110749534e-23, 3.1644638093523535e-23], [7.674685396698919, 3.293695449817917e-23, 3.279409427219235e-23, 3.136764094205426e-23], [7.695965457168219, 3.2659054582291057e-23, 3.251525299606284e-23, 3.1097084855476626e-23], [7.717304522137401, 3.23871555759319e-23, 3.2242423677865047e-23, 3.083270505586874e-23], [7.738702755211778, 3.2121013573339105e-23, 3.197536209436283e-23, 3.057424839634671e-23], [7.760160320450308, 3.1860395564636054e-23, 3.171383491777327e-23, 3.0321472879056605e-23], [7.781677382366838, 3.1605078966291956e-23, 3.145761924691172e-23, 3.007414718888945e-23], [7.803254105931371, 3.135485116843893e-23, 3.1206502155270514e-23, 2.983205024284069e-23], [7.824890656571345, 3.110950909888433e-23, 3.096028025547368e-23, 2.959497075531536e-23], [7.846587200172873, 3.086885880315994e-23, 3.071875927999689e-23, 2.9362706818144293e-23], [7.868343903082044, 3.06327150407015e-23, 3.048175367774182e-23, 2.913506549589951e-23], [7.890160932106173, 3.040090089642233e-23, 3.024908622612337e-23, 2.891186243598885e-23], [7.9120384545150975, 3.017324740760733e-23, 3.0020587658232093e-23, 2.869292149318567e-23], [7.933976638042457, 2.99495932055589e-23, 2.979609630496627e-23, 2.847807436823985e-23], [7.955975650886964, 2.9729784171787657e-23, 2.957545775133967e-23, 2.826716026066572e-23], [7.978035661713726, 2.951367310812893e-23, 2.935852450698533e-23, 2.806002553476545e-23], [8.000156839655501, 2.93011194206614e-23, 2.9145155690183396e-23, 2.7856523399228375e-23], [8.022339354314013, 2.9091988816717086e-23, 2.8935216725101547e-23, 2.765651359926948e-23], [8.044583375761261, 2.888615301488798e-23, 2.87285790519592e-23, 2.7459862121600985e-23], [8.066889074540805, 2.86834894673672e-23, 2.852511984942002e-23, 2.726644091130113e-23], [8.089256621669083, 2.848388109445542e-23, 2.832472176918782e-23, 2.7076127600970897e-23], [8.11168618863672, 2.8287216030538717e-23, 2.812727268211603e-23, 2.6888805250886803e-23], [8.134177947409846, 2.809338738145986e-23, 2.7932665435529416e-23, 2.670436210056794e-23], [8.156732070431408, 2.79022929926434e-23, 2.7740797621395363e-23, 2.652269133101926e-23], [8.179348730622497, 2.771383522772137e-23, 2.7551571354976955e-23, 2.634369083739486e-23], [8.202028101383677, 2.7527920757181414e-23, 2.7364893063418865e-23, 2.6167263011499047e-23], [8.224770356596302, 2.7344460356788323e-23, 2.7180673284130634e-23, 2.5993314534615026e-23], [8.247575670623867, 2.7163368715245253e-23, 2.6998826472453816e-23, 2.5821756178858894e-23], [8.270444218313328, 2.69845642508534e-23, 2.6819270818168844e-23, 2.565250261816102e-23], [8.293376174996448, 2.6807968936820947e-23, 2.664192807083748e-23, 2.5485472247579715e-23], [8.316371716491153, 2.6633508134763686e-23, 2.646672337317697e-23, 2.5320587011285215e-23], [8.339431019102854, 2.6461110436210425e-23, 2.629358510246344e-23, 2.515777223812651e-23], [8.362554259625835, 2.6290707511653402e-23, 2.6122444719529386e-23, 2.499695648550993e-23], [8.385741615344573, 2.6122233966930787e-23, 2.5953236625031155e-23, 2.4838071389952807e-23], [8.40899326403511, 2.5955627206590322e-23, 2.5785898022668285e-23, 2.4681051525543887e-23], [8.432309383966436, 2.5790827303901625e-23, 2.562036878921274e-23, 2.452583426861165e-23], [8.455690153901816, 2.562777687728832e-23, 2.545659135068757e-23, 2.4372359668981782e-23], [8.479135753100202, 2.546642097292177e-23, 2.5294510564984688e-23, 2.422057032802145e-23], [8.502646361317574, 2.5306706953103e-23, 2.513407361006297e-23, 2.407041128216014e-23], [8.526222158808341, 2.5148584390272155e-23, 2.497522987784814e-23, 2.3921829892209885e-23], [8.54986332632671, 2.4992004966341818e-23, 2.48179308735183e-23, 2.377477573839579e-23], [8.573570045128072, 2.4836922377191702e-23, 2.4662130119851144e-23, 2.3629200520536743e-23], [8.59734249697041, 2.468329224194672e-23, 2.450778306637209e-23, 2.3485057963220588e-23], [8.621180864115662, 2.453107201694997e-23, 2.4354847003332716e-23, 2.3342303725581168e-23], [8.645085329331145, 2.4380220914201096e-23, 2.420328097994069e-23, 2.320089531627149e-23], [8.669056075890941, 2.42306998239804e-23, 2.40530457269207e-23, 2.306079201207945e-23], [8.693093287577305, 2.4082471241468957e-23, 2.390410358307478e-23, 2.292195478111326e-23], [8.717197148682088, 2.3935499197330872e-23, 2.3756418425728536e-23, 2.2784346209992766e-23], [8.741367844008117, 2.3789749191781134e-23, 2.3609955604784207e-23, 2.2647930434598658e-23], [8.76560555887066, 2.364518813228025e-23, 2.3464681880288304e-23, 2.2512673074529062e-23], [8.789910479098793, 2.3501784274467734e-23, 2.3320565363264094e-23, 2.237854117122819e-23], [8.814282791036874, 2.335950716626226e-23, 2.3177575459805688e-23, 2.2245503128998335e-23], [8.838722681545931, 2.3218327594903537e-23, 2.303568281801314e-23, 2.2113528659354367e-23], [8.863230338005122, 2.307821753695499e-23, 2.2894859277926237e-23, 2.1982588728447338e-23], [8.887805948313165, 2.293915011086118e-23, 2.275507782406279e-23, 2.185265550684311e-23], [8.912449700889766, 2.2801099532134864e-23, 2.2616312540576854e-23, 2.1723702322483114e-23], [8.937161784677086, 2.2664041071014393e-23, 2.247853856877671e-23, 2.1595703615994427e-23], [8.961942389141168, 2.2527951012344108e-23, 2.234173206704052e-23, 2.1468634898333898e-23], [8.986791704273395, 2.2392806617612116e-23, 2.2205870172812625e-23, 2.13424727104352e-23], [9.01170992059197, 2.2258586089175884e-23, 2.20709309667563e-23, 2.1217194585913692e-23], [9.036697229143337, 2.2125268536292348e-23, 2.1936893438787134e-23, 2.1092779014733183e-23], [9.061753821503677, 2.1992833942988933e-23, 2.1803737455990792e-23, 2.096920540926243e-23], [9.086879889780361, 2.1861263137798592e-23, 2.1671443732219776e-23, 2.084645407232449e-23], [9.112075626613425, 2.1730537764986346e-23, 2.1539993799376113e-23, 2.0724506166763167e-23], [9.13734122517706, 2.1600640257311707e-23, 2.1409369980168584e-23, 2.0603343686375944e-23], [9.162676879181067, 2.1471553810343273e-23, 2.1279555362455914e-23, 2.0482949428882548e-23], [9.188082782872382, 2.1343262357959412e-23, 2.1150533774669073e-23, 2.0363306969644126e-23], [9.213559131036506, 2.1215750549218655e-23, 2.1022289762788927e-23, 2.0244400637331676e-23], [9.239106118999072, 2.1089003726394292e-23, 2.0894808568333535e-23, 2.0126215490579594e-23], [9.264723942627278, 2.0963007904024394e-23, 2.0768076107437786e-23, 2.0008737295383688e-23], [9.290412798331417, 2.0837749749056026e-23, 2.0642078951046754e-23, 1.9891952504411739e-23], [9.316172883066397, 2.0713216561926192e-23, 2.0516804305977447e-23, 1.977584823653102e-23], [9.342004394333213, 2.0589396258412198e-23, 2.039223999687043e-23, 1.9660412257552336e-23], [9.367907530180501, 2.0466277352448313e-23, 2.0268374448948502e-23, 1.9545632961955154e-23], [9.393882489206028, 2.0343848939496865e-23, 2.014519667153477e-23, 1.943149935508262e-23], [9.419929470558223, 2.0222100680707832e-23, 2.0022696242141592e-23, 1.9318001036257247e-23], [9.446048673937721, 2.010102278761801e-23, 1.9900863291338525e-23, 1.9205128182347147e-23], [9.472240299598859, 1.9980606007462686e-23, 1.9779688487993745e-23, 1.9092871532230783e-23], [9.498504548351246, 1.9860841608920287e-23, 1.9659163025104376e-23, 1.8981222371478545e-23], [9.524841621561281, 1.9741721368251872e-23, 1.953927860606105e-23, 1.887017251742393e-23], [9.551251721153703, 1.962323755597677e-23, 1.942002743129369e-23, 1.875971430509186e-23], [9.577735049613146, 1.9505382923724598e-23, 1.9301402185173558e-23, 1.8649840573013253e-23], [9.604291809985678, 1.9388150691437217e-23, 1.9183396023374663e-23, 1.8540544649446725e-23], [9.63092220588037, 1.927153453487557e-23, 1.906600256029649e-23, 1.8431820339295254e-23], [9.657626441470846, 1.9155528573221687e-23, 1.8949215856844352e-23, 1.8323661910582366e-23], [9.684404721496868, 1.9040127356884845e-23, 1.8833030408260327e-23, 1.8216064081362623e-23], [9.711257251265879, 1.8925325855548452e-23, 1.8717441132143332e-23, 1.810902200727884e-23], [9.73818423665459, 1.881111944606679e-23, 1.8602443356564634e-23, 1.8002531268184416e-23], [9.765185884110567, 1.869750390068715e-23, 1.848803280820796e-23, 1.7896587855869548e-23], [9.7922624006538, 1.858447537512162e-23, 1.8374205600595164e-23, 1.779118816113997e-23], [9.8194139938783, 1.8472030396774462e-23, 1.8260958222269908e-23, 1.7686328961430483e-23], [9.846640871953682, 1.8360165852784534e-23, 1.814828752499567e-23, 1.7582007407782121e-23], [9.873943243626762, 1.8248878978186952e-23, 1.8036190711894827e-23, 1.747822101242743e-23], [9.901321318223175, 1.8138167343941805e-23, 1.7924665325565308e-23, 1.7374967635974675e-23], [9.92877530564895, 1.802802884491903e-23, 1.7813709236116202e-23, 1.7272245474489436e-23], [9.956305416392148, 1.7918461687812413e-23, 1.7703320629068322e-23, 1.7170053046902206e-23], [9.98391186152445, 1.7809464378896408e-23, 1.7593497993241072e-23, 1.7068389181586103e-23], [10.011594852702796, 1.770103581718841e-23, 1.748424021517288e-23, 1.6967253113699342e-23], [10.039354602171002, 1.7593175913274852e-23, 1.7375547205957982e-23, 1.6866645130186722e-23], [10.067191322761374, 1.748588406016788e-23, 1.72674183534946e-23, 1.6766564974008666e-23], [10.09510522789637, 1.7379159700791714e-23, 1.7159853096811788e-23, 1.6667012436601294e-23], [10.1230965315902, 1.7273002527051168e-23, 1.705285112779594e-23, 1.6567987565250322e-23], [10.151165448450488, 1.71674124661451e-23, 1.6946412377573154e-23, 1.6469490648437917e-23], [10.17931219367992, 1.7062389666971817e-23, 1.68405370027983e-23, 1.6371522201804435e-23], [10.20753698307787, 1.695793448634591e-23, 1.6735225371954306e-23, 1.6274082953231605e-23], [10.235840033042102, 1.6854047475296796e-23, 1.6630478051617533e-23, 1.6177173828662845e-23], [10.26422156057037, 1.6750729365342438e-23, 1.6526295792675544e-23, 1.6080795937451483e-23], [10.292681783262129, 1.6647981054674367e-23, 1.642267951655641e-23, 1.5984950557769615e-23], [10.321220919320176, 1.6545803594487156e-23, 1.6319630301449675e-23, 1.5889639122337753e-23], [10.349839187552334, 1.644419817522528e-23, 1.621714936859028e-23, 1.5794863203726378e-23], [10.378536807373127, 1.6343166112988943e-23, 1.6115238068563824e-23, 1.5700624500351593e-23], [10.407313998805458, 1.624270883589154e-23, 1.601389786764118e-23, 1.560692482190241e-23], [10.43617098248232, 1.6142827870562714e-23, 1.5913130334207992e-23, 1.551376607504011e-23], [10.465107979648447, 1.6043524828827915e-23, 1.5812937125382207e-23, 1.542115024988613e-23], [10.494125212162036, 1.594480139433343e-23, 1.5713319973599785e-23, 1.532907940533446e-23], [10.523222902496453, 1.5846659309594867e-23, 1.5614280673513385e-23, 1.523755565613342e-23], [10.552401273741923, 1.574910036295404e-23, 1.5515821068945282e-23, 1.5146581158605894e-23], [10.581660549607243, 1.5652126375950136e-23, 1.5417943040107125e-23, 1.5056158097970907e-23], [10.611000954421502, 1.555573919077522e-23, 1.5320648491004243e-23, 1.4966288674773353e-23], [10.640422713135811, 1.5459940658118005e-23, 1.522393933713579e-23, 1.4876975092417664e-23], [10.669926051325003, 1.5364732625158736e-23, 1.512781749337726e-23, 1.4788219544685618e-23], [10.699511195189373, 1.5270116923934915e-23, 1.503228486228642e-23, 1.4700024203258803e-23], [10.729178371556443, 1.5176095360042968e-23, 1.4937343322638926e-23, 1.46123912063632e-23], [10.758927807882628, 1.5082669701671465e-23, 1.4842994718292435e-23, 1.4525322647046723e-23], [10.78875973225507, 1.4989841668956784e-23, 1.4749240847543043e-23, 1.4438820562219715e-23], [10.81867437339331, 1.4897612923820623e-23, 1.4656083452690925e-23, 1.435288692203903e-23], [10.848671960651085, 1.480598506016856e-23, 1.4563524210172185e-23, 1.4267523619767375e-23], [10.878752724018085, 1.4714959594500091e-23, 1.4471564721032278e-23, 1.418273246208832e-23], [10.908916894121692, 1.462453795703202e-23, 1.438020650182917e-23, 1.409851515969944e-23], [10.939164702228762, 1.4534721483217438e-23, 1.4289450976088808e-23, 1.401487331888952e-23], [10.96949638024741, 1.4445511405748525e-23, 1.419929946615164e-23, 1.3931808432889528e-23], [10.999912160728758, 1.4356908847115267e-23, 1.4109753185497977e-23, 1.38493218745235e-23], [11.030412276868763, 1.426891481257977e-23, 1.4020813231654263e-23, 1.3767414888816596e-23], [11.06099696250995, 1.4181530183702397e-23, 1.393248057953218e-23, 1.368608858636342e-23], [11.09166645214326, 1.40947557124151e-23, 1.3844756075307533e-23, 1.3605343937391582e-23], [11.122420980909801, 1.400859201547447e-23, 1.3757640430821783e-23, 1.3525181765864273e-23], [11.153260784602677, 1.392303956966108e-23, 1.3671134218477165e-23, 1.3445602744855198e-23], [11.184186099668795, 1.3838098707273622e-23, 1.3585237866749998e-23, 1.336660739178877e-23], [11.215197163210656, 1.3753769612340416e-23, 1.349995165605982e-23, 1.328819606483163e-23], [11.246294212988216, 1.367005231717371e-23, 1.3415275715368982e-23, 1.3210368959253615e-23], [11.277477487420656, 1.3586946699623483e-23, 1.3331210019090728e-23, 1.313312610486121e-23], [11.30874722558826, 1.3504452480646339e-23, 1.3247754384699225e-23, 1.3056467363520495e-23], [11.340103667234201, 1.34225692225522e-23, 1.3164908470704374e-23, 1.2980392427586156e-23], [11.37154705276642, 1.3341296327642873e-23, 1.3082671775170302e-23, 1.2904900818639725e-23], [11.403077623259442, 1.3260633037331426e-23, 1.3001043634753938e-23, 1.282999188662163e-23], [11.434695620456228, 1.3180578431787161e-23, 1.2920023224153259e-23, 1.2755664809798502e-23], [11.466401286770061, 1.3101131430000126e-23, 1.283960955608486e-23, 1.2681918595026715e-23], [11.498194865286347, 1.302229079022141e-23, 1.2759801481611577e-23, 1.2608752078063345e-23], [11.530076599764518, 1.2944055110997309e-23, 1.2680597690997442e-23, 1.253616392537164e-23], [11.56204673463991, 1.2866422832391643e-23, 1.2601996714892595e-23, 1.2464152634987666e-23], [11.5941055150256, 1.2789392237801642e-23, 1.2523996925986854e-23, 1.2392716539149086e-23], [11.626253186714308, 1.2712961455956936e-23, 1.2446596540928877e-23, 1.232185380611392e-23], [11.658489996180279, 1.2637128463421367e-23, 1.2369793622722256e-23, 1.225156244322381e-23], [11.690816190581181, 1.25618910873152e-23, 1.2293586083392108e-23, 1.2181840299847842e-23], [11.723232017759983, 1.2487247008377464e-23, 1.2217971686921115e-23, 1.211268507066414e-23], [11.75573772624686, 1.2413193764334707e-23, 1.2142948052566584e-23, 1.2044094299525901e-23], [11.788333565261105, 1.2339728753536178e-23, 1.2068512658391805e-23, 1.197606538329094e-23], [11.821019784713034, 1.2266849238751738e-23, 1.1994662845037684e-23, 1.1908595576034383e-23], [11.853796635205914, 1.219455235129021e-23, 1.1921395819771094e-23, 1.1841681993432697e-23], [11.886664368037867, 1.212283509526155e-23, 1.1848708660701438e-23, 1.1775321617437488e-23], [11.919623235203794, 1.2051694352033247e-23, 1.1776598321150584e-23, 1.1709511301093684e-23], [11.952673489397345, 1.1981126884826374e-23, 1.1705061634245019e-23, 1.1644247773389951e-23], [11.985815384012803, 1.1911129343472018e-23, 1.1634095317638467e-23, 1.1579527644517507e-23], [12.019049173147064, 1.1841698269262189e-23, 1.1563695978322029e-23, 1.151534741100416e-23], [12.052375111601568, 1.1772830099892437e-23, 1.1493860117544934e-23, 1.1451703460957649e-23], [12.08579345488426, 1.1704521174515437e-23, 1.1424584135851392e-23, 1.1388592079591506e-23], [12.119304459211557, 1.1636767738790317e-23, 1.1355864338111053e-23, 1.132600945448755e-23], [12.152908381510285, 1.1569565950041912e-23, 1.1287696938714248e-23, 1.1263951681193792e-23], [12.186605479419686, 1.1502911882408483e-23, 1.1220078066633986e-23, 1.1202414768812137e-23], [12.22039601129336, 1.1436801531966019e-23, 1.1153003770638959e-23, 1.1141394645152465e-23], [12.25428023620126, 1.1371230821892271e-23, 1.1086470024421654e-23, 1.1080887162514771e-23], [12.288258413931686, 1.1306195607621902e-23, 1.102047273175794e-23, 1.1020888103158483e-23], [12.322330804993253, 1.1241691681841992e-23, 1.09550077315741e-23, 1.0961393184297907e-23], [12.356497670616928, 1.1177714779636452e-23, 1.0890070803022267e-23, 1.090239806400327e-23], [12.390759272757983, 1.1114260583365585e-23, 1.0825657670469924e-23, 1.0843898346019044e-23], [12.425115874098054, 1.105132472760906e-23, 1.0761764008437781e-23, 1.0785889585166639e-23], [12.45956773804711, 1.0988902803983345e-23, 1.0698385446405982e-23, 1.0728367292467358e-23], [12.494115128745502, 1.0926990365805939e-23, 1.0635517573615016e-23, 1.0671326939854189e-23], [12.528758311065976, 1.0865582932798411e-23, 1.0573155943711428e-23, 1.0614763965462566e-23], [12.5634975506157, 1.0804675995527727e-23, 1.051129607924027e-23, 1.0558673778080681e-23], [12.59833311373833, 1.0744265019800981e-23, 1.0449933476157891e-23, 1.0503051761827416e-23], [12.633265267516004, 1.0684345450993765e-23, 1.0389063608086464e-23, 1.0447893280821871e-23], [12.668294279771418, 1.0624912718098875e-23, 1.0328681930536224e-23, 1.0393193683247153e-23], [12.703420419069896, 1.0565962237819185e-23, 1.0268783884922738e-23, 1.0338948305858933e-23], [12.73864395472141, 1.0507489418375282e-23, 1.0209364902537501e-23, 1.0285152477830755e-23], [12.773965156782666, 1.044948966327099e-23, 1.0150420408278109e-23, 1.0231801524755833e-23], [12.80938429605917, 1.0391958374901275e-23, 1.0091945824338482e-23, 1.017889077254931e-23], [12.844901644107326, 1.0334890957945897e-23, 1.0033936573667937e-23, 1.0126415550752925e-23], [12.880517473236479, 1.0278282822708876e-23, 9.976388083348414e-24, 1.007437119623481e-23], [12.916232056511028, 1.02221293882716e-23, 9.919295787806933e-24, 1.0022753056422886e-23], [12.952045667752515, 1.0166426085488774e-23, 9.86265513185462e-24, 9.971556492372229e-24], [12.987958581541713, 1.0111168359867949e-23, 9.806461573657616e-24, 9.920776881843536e-24], [13.02397107322077, 1.0056351674285045e-23, 9.750710587439057e-24, 9.870409622121753e-24], [13.06008341889526, 1.0001971511573034e-23, 9.69539766618591e-24, 9.820450132723304e-24], [13.096295895436345, 9.948023376928981e-24, 9.64051832407535e-24, 9.77089385776017e-24], [13.132608780482894, 9.894502800239001e-24, 9.586068098889304e-24, 9.721736268502546e-24], [13.169022352443584, 9.841405338231668e-24, 9.532042554150784e-24, 9.67297286559883e-24], [13.205536890499062, 9.78872657650414e-24, 9.478437281279548e-24, 9.624599181117034e-24], [13.242152674604073, 9.736462131369661e-24, 9.4252479014752e-24, 9.576610780327539e-24], [13.278869985489605, 9.684607651723164e-24, 9.37247006756622e-24, 9.529003263978422e-24], [13.315689104665065, 9.633158820562375e-24, 9.320099465662039e-24, 9.481772269444056e-24], [13.352610314420389, 9.582111356575566e-24, 9.268131816792635e-24, 9.43491347262516e-24], [13.389633897828263, 9.531461015529594e-24, 9.216562878253016e-24, 9.388422589299116e-24], [13.42676013874625, 9.481203591521463e-24, 9.165388445013386e-24, 9.342295376331038e-24], [13.463989321818982, 9.43133491819328e-24, 9.114604350895724e-24, 9.296527633011675e-24], [13.501321732480346, 9.38185086973624e-24, 9.064206469679865e-24, 9.251115201928397e-24], [13.53875765695566, 9.332747361919048e-24, 9.01419071613563e-24, 9.206053970177816e-24], [13.576297382263899, 9.284020352888957e-24, 8.964553046908717e-24, 9.161339870040723e-24], [13.613941196219844, 9.235665843985254e-24, 8.915289461316109e-24, 9.116968879886658e-24], [13.651689387436345, 9.187679880355138e-24, 8.866396002076863e-24, 9.072937024588808e-24], [13.689542245326484, 9.140058551600232e-24, 8.817868755933559e-24, 9.02924037642694e-24], [13.727500060105823, 9.092797992245792e-24, 8.769703854169726e-24, 8.985875055385574e-24], [13.765563122794626, 9.04589438215762e-24, 8.7218974731257e-24, 8.942837229561034e-24], [13.803731725220075, 8.99934394689395e-24, 8.674445834497457e-24, 8.900123115448231e-24], [13.842006160018542, 8.953142958006804e-24, 8.627345205712855e-24, 8.857728978411137e-24], [13.88038672063778, 8.907287733206403e-24, 8.58059190014102e-24, 8.815651132550578e-24], [13.918873701339232, 8.861774636516164e-24, 8.53418227727418e-24, 8.773885941000769e-24], [13.957467397200233, 8.816600078380801e-24, 8.48811274282359e-24, 8.732429815997568e-24], [13.996168104116308, 8.771760515651816e-24, 8.44237974877553e-24, 8.691279218785295e-24], [14.034976118803428, 8.727252451556084e-24, 8.396979793421344e-24, 8.650430659465396e-24], [14.073891738800274, 8.68307243564116e-24, 8.351909421233417e-24, 8.609880697061634e-24], [14.112915262470558, 8.639217063645072e-24, 8.307165222832531e-24, 8.569625939314746e-24], [14.152046989005262, 8.595682977285013e-24, 8.262743834774989e-24, 8.529663042305287e-24], [14.191287218424954, 8.552466864069613e-24, 8.21864193941633e-24, 8.489988710380494e-24], [14.230636251582093, 8.509565457051481e-24, 8.174856264625974e-24, 8.4505996957285e-24], [14.270094390163317, 8.466975534507248e-24, 8.131383583555388e-24, 8.41149279811033e-24], [14.309661936691795, 8.424693919652665e-24, 8.088220714309652e-24, 8.372664864510333e-24], [14.349339194529495, 8.382717480235569e-24, 8.045364519626208e-24, 8.334112788598026e-24], [14.389126467879539, 8.341043128202121e-24, 8.002811906504916e-24, 8.29583351042443e-24], [14.42902406178855, 8.299667819290206e-24, 7.960559825816749e-24, 8.257824016080155e-24], [14.469032282148948, 8.25858855253769e-24, 7.918605271877642e-24, 8.220081336839553e-24], [14.509151435701334, 8.217802369911649e-24, 7.876945282043214e-24, 8.182602549092381e-24], [14.54938183003681, 8.177306355764533e-24, 7.835576936208694e-24, 8.145384773496656e-24], [14.589723773599383, 8.137097636397194e-24, 7.794497356349302e-24, 8.108425174624279e-24], [14.630177575688279, 8.0971733795118e-24, 7.753703706038607e-24, 8.071720960314331e-24], [14.670743546460335, 8.057530793740117e-24, 7.713193189911805e-24, 8.035269381227739e-24], [14.7114219969324, 8.018167128060359e-24, 7.672963053191608e-24, 7.999067730072739e-24], [14.752213238983675, 7.979079671296616e-24, 7.63301058110684e-24, 7.963113341197593e-24], [14.793117585358138, 7.940265751568219e-24, 7.59333309838748e-24, 7.92740358999281e-24], [14.834135349666926, 7.901722735696827e-24, 7.5539279686922e-24, 7.891935892104628e-24], [14.875266846390744, 7.863448028665854e-24, 7.5147925940701e-24, 7.856707702977737e-24], [14.916512390882295, 7.82543907307582e-24, 7.475924414376998e-24, 7.821716517288574e-24], [14.957872299368645, 7.787693348500121e-24, 7.437320906739107e-24, 7.786959868068497e-24], [14.99934688895372, 7.75020837101197e-24, 7.398979584939234e-24, 7.752435326499488e-24], [15.040936477620672, 7.712981692517129e-24, 7.360897998908002e-24, 7.718140500911759e-24], [15.08264138423435, 7.676010900214529e-24, 7.323073734090527e-24, 7.684073036323703e-24], [15.124461928543742, 7.63929361601958e-24, 7.285504410886986e-24, 7.650230613854977e-24], [15.166398431184415, 7.602827495980991e-24, 7.248187684115608e-24, 7.616610950034347e-24], [15.208451213681004, 7.566610229712387e-24, 7.211121242390981e-24, 7.58321179627568e-24], [15.250620598449622, 7.53063953979054e-24, 7.174302807579037e-24, 7.55003093806108e-24], [15.292906908800399, 7.494913181245884e-24, 7.137730134241716e-24, 7.517066194705971e-24], [15.335310468939907, 7.459428940921612e-24, 7.101401009036391e-24, 7.484315418289164e-24], [15.377831603973645, 7.424184636986406e-24, 7.06531325019371e-24, 7.451776493520528e-24], [15.420470639908597, 7.38917811830997e-24, 7.029464706931852e-24, 7.419447336774106e-24], [15.463227903655646, 7.35440726394797e-24, 6.993853258927245e-24, 7.387325895655661e-24], [15.506103723032155, 7.319869982603636e-24, 6.958476815736267e-24, 7.355410148527761e-24], [15.549098426764417, 7.285564212045064e-24, 6.923333316308523e-24, 7.323698103672978e-24], [15.592212344490214, 7.251487918636866e-24, 6.888420728400232e-24, 7.292187799080777e-24], [15.63544580676134, 7.217639096729826e-24, 6.853737048083593e-24, 7.260877301465983e-24], [15.678799145046115, 7.184015768218798e-24, 6.819280299174637e-24, 7.229764706130217e-24], [15.722272691731966, 7.150615981967422e-24, 6.78504853278024e-24, 7.198848136093668e-24], [15.76586678012793, 7.117437813350144e-24, 6.751039826773719e-24, 7.168125741823631e-24], [15.809581744467225, 7.084479363706671e-24, 6.717252285235808e-24, 7.137595700423283e-24], [15.853417919909848, 7.051738759907493e-24, 6.683684038066343e-24, 7.107256215479128e-24], [15.897375642545086, 7.019214153793693e-24, 6.650333240395723e-24, 7.077105516108569e-24], [15.941455249394126, 6.986903721786699e-24, 6.617198072161569e-24, 7.047141856897294e-24], [15.98565707841263, 6.95480566435005e-24, 6.584276737666238e-24, 7.017363517076245e-24], [16.029981468493357, 6.922918205576669e-24, 6.5515674650409405e-24, 6.987768800227553e-24], [16.074428759468706, 6.89123959270551e-24, 6.519068505839509e-24, 6.958356033727646e-24], [16.118999292113344, 6.859768095698261e-24, 6.486778134584324e-24, 6.929123568320652e-24], [16.163693408146862, 6.828502006793588e-24, 6.4546946483345604e-24, 6.900069777596202e-24], [16.208511450236323, 6.797439640075849e-24, 6.4228163662398204e-24, 6.871193057609287e-24], [16.25345376199893, 6.766579331091266e-24, 6.391141629133212e-24, 6.8424918265067e-24], [16.29852068800467, 6.735919436371739e-24, 6.359668799107338e-24, 6.813964523875602e-24], [16.343712573778905, 6.705458333091182e-24, 6.3283962591213314e-24, 6.785609610541084e-24], [16.38902976580511, 6.675194418639345e-24, 6.297322412578468e-24, 6.757425568020684e-24], [16.43447261152742, 6.645126110253442e-24, 6.266445682973541e-24, 6.729410898232444e-24], [16.48004145935339, 6.615251844616273e-24, 6.2357645134725806e-24, 6.701564122963268e-24], [16.525736658656605, 6.585570077507848e-24, 6.2052773665696606e-24, 6.673883783617464e-24], [16.571558559779383, 6.556079283436356e-24, 6.174982723692813e-24, 6.64636844076006e-24], [16.617507514035456, 6.526777955254368e-24, 6.1448790848761775e-24, 6.619016673686241e-24], [16.66358387371266, 6.497664603863549e-24, 6.114964968370638e-24, 6.591827080246477e-24], [16.709787992075665, 6.468737757825234e-24, 6.085238910333317e-24, 6.564798276290684e-24], [16.75612022336862, 6.4399959630565686e-24, 6.0556994644772944e-24, 6.537928895472535e-24], [16.802580922817953, 6.411437782493354e-24, 6.026345201745756e-24, 6.511217588832486e-24], [16.849170446635007, 6.383061795763711e-24, 5.997174709980685e-24, 6.484663024435537e-24], [16.895889152018828, 6.354866598888187e-24, 5.968186593617265e-24, 6.458263887103289e-24], [16.94273739715891, 6.326850803991122e-24, 5.9393794733831726e-24, 6.4320188782124054e-24], [16.989715541237892, 6.299013038959355e-24, 5.910751985958599e-24, 6.405926715116137e-24], [17.036823944434367, 6.271351947178421e-24, 5.8823027837595994e-24, 6.37998613103537e-24], [17.08406296792559, 6.2438661872522546e-24, 5.854030534547746e-24, 6.354195874727254e-24], [17.131432973890288, 6.216554432709084e-24, 5.825933921238588e-24, 6.3285547101802266e-24], [17.17893432551142, 6.18941537174653e-24, 5.798011641569251e-24, 6.3030614163653945e-24], [17.226567386978957, 6.162447706950985e-24, 5.770262407869228e-24, 6.2777147869232364e-24], [17.27433252349271, 6.135650155039841e-24, 5.742684946767856e-24, 6.252513629889116e-24], [17.322230101265063, 6.109021446621202e-24, 5.71527799893985e-24, 6.227456767478015e-24], [17.370260487523836, 6.0825603259368236e-24, 5.688040318890668e-24, 6.202543035828198e-24], [17.418424050515096, 6.0562655506109185e-24, 5.6609706746461264e-24, 6.17777128469291e-24], [17.466721159505955, 6.030135891439618e-24, 5.634067847593341e-24, 6.153140377265367e-24], [17.515152184787404, 6.0041701321243246e-24, 5.607330632157926e-24, 6.128649189859041e-24], [17.563717497677175, 5.978367069084718e-24, 5.5807578356762974e-24, 6.104296611810593e-24], [17.612417470522583, 5.95272551120935e-24, 5.554348278061249e-24, 6.080081545098279e-24], [17.66125247670336, 5.927244279658598e-24, 5.52810079169067e-24, 6.0560029042474566e-24], [17.71022289063452, 5.901922207642432e-24, 5.502014221113676e-24, 6.0320596160272115e-24], [17.759329087769274, 5.876758140222363e-24, 5.476087422887396e-24, 6.008250619292322e-24], [17.808571444601842, 5.851750934109957e-24, 5.4503192653473045e-24, 5.984574864763911e-24], [17.85795033867039, 5.826899457474034e-24, 5.4247086284411486e-24, 5.9610313148368405e-24], [17.90746614855989, 5.8022025897512405e-24, 5.399254403510057e-24, 5.937618943399372e-24], [17.957119253905052, 5.777659221440396e-24, 5.373955493097826e-24, 5.9143367355495776e-24], [18.00691003539323, 5.7532682539452984e-24, 5.348810810770581e-24, 5.891183687539898e-24], [18.05683887476731, 5.7290285993954215e-24, 5.323819280952246e-24, 5.868158806559072e-24], [18.106906154828685, 5.7049391804437504e-24, 5.298979838727422e-24, 5.845261110449583e-24], [18.157112259440147, 5.680998930138048e-24, 5.274291429678213e-24, 5.822489627705695e-24], [18.20745757352885, 5.657206791720187e-24, 5.249753009716874e-24, 5.7998433971712874e-24], [18.25794248308926, 5.6335617184964804e-24, 5.225363544913288e-24, 5.777321467980241e-24], [18.3085673751861, 5.61006267364526e-24, 5.201122011339626e-24, 5.754922899285671e-24], [18.35933263795737, 5.5867086300770546e-24, 5.177027394919413e-24, 5.7326467601311666e-24], [18.410238660617242, 5.563498570305388e-24, 5.15307869125705e-24, 5.710492129416904e-24], [18.461285833459105, 5.5404314862702496e-24, 5.129274905516349e-24, 5.68845809561167e-24], [18.51247454785853, 5.51750637920741e-24, 5.105615052248765e-24, 5.666543756663014e-24], [18.563805196276277, 5.494722259498022e-24, 5.0820981552600895e-24, 5.644748219829474e-24], [18.615278172261327, 5.4720781465448815e-24, 5.058723247459631e-24, 5.6230706015675956e-24], [18.66689387045384, 5.4495730686422414e-24, 5.035489370744805e-24, 5.6015100274676714e-24], [18.718652686588264, 5.427206062814496e-24, 5.012395575858871e-24, 5.5800656319215736e-24], [18.770555017496292, 5.404976174728199e-24, 4.9894409222370436e-24, 5.558736558220131e-24], [18.822601261109934, 5.382882458528295e-24, 4.966624477931885e-24, 5.5375219582456766e-24], [18.87479181646459, 5.360923976747408e-24, 4.943945319428792e-24, 5.5164209924825725e-24], [18.92712708370206, 5.339099800178595e-24, 4.9214025315710765e-24, 5.4954328298566624e-24], [18.979607464073688, 5.317409007733746e-24, 4.8989952074068044e-24, 5.474556647534635e-24], [19.03223335994335, 5.29585068636675e-24, 4.876722448092536e-24, 5.4537916309464785e-24], [19.085005174790588, 5.274423930948104e-24, 4.8545833627807434e-24, 5.4331369736281895e-24], [19.137923313213715, 5.253127844131008e-24, 4.8325770684897925e-24, 5.4125918770095155e-24], [19.19098818093288, 5.231961536277381e-24, 4.810702690024585e-24, 5.392155550422571e-24], [19.244200184793204, 5.210924125343958e-24, 4.788959359825651e-24, 5.3718272110017815e-24], [19.297559732767876, 5.190014736790936e-24, 4.767346217917309e-24, 5.351606083590588e-24], [19.35106723396133, 5.169232503428577e-24, 4.745862411770772e-24, 5.331491400427007e-24], [19.404723098612322, 5.1485765654066726e-24, 4.724507096204521e-24, 5.311482401416441e-24], [19.458527738097096, 5.1280460700419174e-24, 4.703279433305604e-24, 5.291578333712656e-24], [19.512481564932575, 5.1076401717648005e-24, 4.6821785923298774e-24, 5.27177845179685e-24], [19.566584992779468, 5.087358032014447e-24, 4.661203749588861e-24, 5.2520820173232446e-24], [19.620838436445467, 5.067198819172971e-24, 4.640354088383639e-24, 5.232488299129928e-24], [19.675242311888432, 5.04716170841908e-24, 4.619628798887558e-24, 5.212996572925258e-24], [19.729797036219566, 5.027245881714224e-24, 4.599027078095758e-24, 5.193606121464814e-24], [19.78450302770664, 5.007450527675344e-24, 4.5785481297038176e-24, 5.174316234346088e-24], [19.839360705777157, 4.987774841494882e-24, 4.558191164047621e-24, 5.155126207869834e-24], [19.89437049102162, 4.968218024884584e-24, 4.537955397996433e-24, 5.136035345114986e-24], [19.9495328051967, 4.948779285964701e-24, 4.517840054909656e-24, 5.117042955705609e-24], [20.00484807122852, 4.929457839204593e-24, 4.4978443645200355e-24, 5.098148355808631e-24], [20.060316713215855, 4.910252905357895e-24, 4.477967562870386e-24, 5.079350868114541e-24], [20.115939156433413, 4.891163711352859e-24, 4.458208892248183e-24, 5.060649821601034e-24], [20.171715827335102, 4.872189490264796e-24, 4.438567601092898e-24, 5.042044551667946e-24], [20.227647153557268, 4.853329481213064e-24, 4.419042943941386e-24, 5.02353439992256e-24], [20.28373356392199, 4.8345829293007616e-24, 4.39963418133881e-24, 5.005118714153864e-24], [20.33997548844037, 4.81594908553928e-24, 4.3803405797907505e-24, 4.9867968482299814e-24], [20.39637335831582, 4.7974272068045814e-24, 4.361161411653518e-24, 4.968568162153916e-24], [20.452927605947405, 4.779016555737922e-24, 4.342095955130688e-24, 4.950432021817135e-24], [20.509638664933085, 4.760716400718591e-24, 4.323143494135733e-24, 4.9323877991366444e-24], [20.566506970073124, 4.742526015761463e-24, 4.3043033182933266e-24, 4.914434871797019e-24], [20.623532957373364, 4.724444680487661e-24, 4.285574722829971e-24, 4.89657262334331e-24], [20.680717064048576, 4.706471680051032e-24, 4.266957008513693e-24, 4.87880044306313e-24], [20.73805972852584, 4.6886063050662435e-24, 4.24844948163463e-24, 4.861117725871604e-24], [20.795561390447872, 4.6708478515790855e-24, 4.230051453892588e-24, 4.843523872380015e-24], [20.853222490676444, 4.6531956209902445e-24, 4.211762242372039e-24, 4.826018288781954e-24], [20.911043471295695, 4.635648919988301e-24, 4.193581169473716e-24, 4.808600386705138e-24], [20.96902477561557, 4.618207060519857e-24, 4.175507562852664e-24, 4.791269583306661e-24], [21.02716684817522, 4.600869359719537e-24, 4.157540755377476e-24, 4.774025301146107e-24], [21.085470134746387, 4.583635139860716e-24, 4.139680085060617e-24, 4.75686696813649e-24], [21.143935082336817, 4.566503728303044e-24, 4.121924895010344e-24, 4.73979401750125e-24], [21.20256213919371, 4.549474457444413e-24, 4.10427453337176e-24, 4.722805887741326e-24], [21.26135175480717, 4.532546664661949e-24, 4.086728353297025e-24, 4.70590202254789e-24], [21.3203043799136, 4.515719692272678e-24, 4.069285712864523e-24, 4.6890818708050145e-24], [21.379420466499187, 4.498992887473295e-24, 4.051945975046558e-24, 4.672344886472867e-24], [21.438700467803397, 4.4823656023022645e-24, 4.034708507658534e-24, 4.655690528599211e-24], [21.498144838322393, 4.4658371935960546e-24, 4.017572683307806e-24, 4.6391182612900276e-24], [21.557754033812554, 4.4494070229261004e-24, 4.000537879345436e-24, 4.6226275535895165e-24], [21.617528511293962, 4.433074456568623e-24, 3.9836034778237894e-24, 4.60621787949959e-24], [21.677468729053892, 4.4168388654515995e-24, 3.9667688654420524e-24, 4.5898887178992194e-24], [21.73757514665038, 4.4006996251174046e-24, 3.9500334335007644e-24, 4.573639552534546e-24], [21.797848224915658, 4.3846561156763675e-24, 3.9333965778904586e-24, 4.5574698719533095e-24], [21.858288425959778, 4.368707721765545e-24, 3.916857698983645e-24, 4.5413791694971126e-24], [21.918896213174083, 4.352853832499311e-24, 3.900416201651974e-24, 4.5253669431808326e-24], [21.979672051234797, 4.337093841442295e-24, 3.8840714951964354e-24, 4.509432695764695e-24], [22.040616406106576, 4.3214271465514114e-24, 3.8678229933060505e-24, 4.4935759346017796e-24], [22.101729745046082, 4.305853150158508e-24, 3.851670114014268e-24, 4.477796171710346e-24], [22.163012536605596, 4.290371258905122e-24, 3.8356122796788796e-24, 4.462092923637823e-24], [22.22446525063655, 4.274980883728287e-24, 3.819648916918528e-24, 4.4464657114945135e-24], [22.286088358293163, 4.259681439801354e-24, 3.803779456577012e-24, 4.4309140608540036e-24], [22.347882332036068, 4.244472346511009e-24, 3.7880033337059634e-24, 4.415437501776111e-24], [22.4098476456359, 4.229353027413992e-24, 3.772319987500225e-24, 4.4000355687251775e-24], [22.471984774176967, 4.214322910208433e-24, 3.756728861263757e-24, 4.384707800583361e-24], [22.534294194060852, 4.1993814266824236e-24, 3.741229402397216e-24, 4.369453740552955e-24], [22.59677638301011, 4.184528012694228e-24, 3.725821062326324e-24, 4.35427293616964e-24], [22.659431820071877, 4.169762108128203e-24, 3.710503296500915e-24, 4.3391649392436276e-24], [22.72226098562159, 4.1550831568665186e-24, 3.69527556432969e-24, 4.3241293058487036e-24], [22.78526436136664, 4.1404906067460735e-24, 3.6801373291557966e-24, 4.309165596236545e-24], [22.848442430350076, 4.1259839095435545e-24, 3.66508805823588e-24, 4.29427337491418e-24], [22.911795676954338, 4.111562520913264e-24, 3.650127222682072e-24, 4.279452210437885e-24], [22.9753245869049, 4.0972259003868116e-24, 3.635254297450463e-24, 4.264701675568269e-24], [23.039029647274077, 4.082973511316606e-24, 3.620468761287552e-24, 4.250021347105281e-24], [23.102911346484696, 4.0688048208607066e-24, 3.6057700967127654e-24, 4.235410805926908e-24], [23.166970174313853, 4.054719299933758e-24, 3.591157789987502e-24, 4.220869636890732e-24], [23.231206621896703, 4.040716423195405e-24, 3.576631331061729e-24, 4.2063974288799866e-24], [23.295621181730183, 4.026795669007988e-24, 3.5621902135703765e-24, 4.191993774727964e-24], [23.360214347676816, 4.012956519409284e-24, 3.54783393478255e-24, 4.177658271188039e-24], [23.424986614968454, 3.999198460082968e-24, 3.533561995580431e-24, 4.163390518916842e-24], [23.48993848021019, 3.985520980326076e-24, 3.519373900423795e-24, 4.1491901224161e-24], [23.555070441384004, 3.9719235730361526e-24, 3.505269157321758e-24, 4.13505669009034e-24], [23.620382997852694, 3.958405734649254e-24, 3.49124727780443e-24, 4.120989834058295e-24], [23.685876650363646, 3.944966965147534e-24, 3.477307776883935e-24, 4.106989170282243e-24], [23.751551901052743, 3.931606768005391e-24, 3.463450173049234e-24, 4.0930543184258646e-24], [23.817409253448133, 3.918324650181914e-24, 3.449673988210102e-24, 4.079184901922329e-24], [23.883449212474137, 3.905120122071978e-24, 3.43597874768e-24, 4.065380547838044e-24], [23.94967228445511, 3.891992697508603e-24, 3.422363980146152e-24, 4.0516408869777844e-24], [24.01607897711932, 3.878941893691708e-24, 3.4088292176555714e-24, 4.037965553664064e-24], [24.08266979960286, 3.865967231205756e-24, 3.395373995559598e-24, 4.024354185893521e-24], [24.149445262453522, 3.8530682339788716e-24, 3.3819978525128844e-24, 4.010806425269176e-24], [24.21640587763472, 3.840244429243177e-24, 3.368700330430683e-24, 3.997321916868623e-24], [24.28355215852942, 3.827495347520087e-24, 3.3554809744620384e-24, 3.983900309306345e-24], [24.350884619944075, 3.81482052260076e-24, 3.342339332980445e-24, 3.970541254708633e-24], [24.4184037781126, 3.8022194915115035e-24, 3.329274957536874e-24, 3.957244408648747e-24], [24.486110150700274, 3.789691794489052e-24, 3.316287402838855e-24, 3.944009430125756e-24], [24.554004256807737, 3.777236974962258e-24, 3.3033762267337915e-24, 3.9308359815595e-24], [24.622086616974983, 3.764854579519472e-24, 3.2905409901779758e-24, 3.917723728723639e-24], [24.690357753185307, 3.75254415789366e-24, 3.2777812571999963e-24, 3.904672340766467e-24], [24.758818188869412, 3.740305262926678e-24, 3.265096594897377e-24, 3.89168149013475e-24], [24.827468448909286, 3.728137450556797e-24, 3.25248657339278e-24, 3.878750852587829e-24], [24.896309059642295, 3.716040279787071e-24, 3.239950765824567e-24, 3.865880107131605e-24], [24.96534054886523, 3.704013312669128e-24, 3.227488748297966e-24, 3.853068936044798e-24], [25.0345634458383, 3.692056114268432e-24, 3.215100099892524e-24, 3.840317024769308e-24], [25.103978281289294, 3.680168252655005e-24, 3.2027844026162747e-24, 3.827624061970804e-24], [25.173585587417524, 3.668349298867175e-24, 3.190541241384934e-24, 3.814989739436074e-24], [25.243385897897976, 3.656598826904424e-24, 3.17837020400114e-24, 3.802413752126109e-24], [25.313379747885396, 3.644916413685624e-24, 3.1662708811320416e-24, 3.7898957980529285e-24], [25.383567674018366, 3.633301639041685e-24, 3.154242866282569e-24, 3.777435578319888e-24], [25.453950214423482, 3.621754085697243e-24, 3.1422857557721752e-24, 3.765032797120838e-24], [25.524527908719396, 3.610273339219004e-24, 3.130399148713015e-24, 3.75268716157388e-24], [25.595301298021003, 3.598858988035201e-24, 3.1185826469858536e-24, 3.7403983818744174e-24], [25.66627092494358, 3.587510623381388e-24, 3.1068358552185914e-24, 3.728166171132604e-24], [25.737437333606923, 3.576227839295697e-24, 3.095158380760592e-24, 3.715990245417425e-24], [25.80880106963962, 3.565010232591924e-24, 3.083549833655243e-24, 3.703870323693018e-24], [25.880362680183076, 3.55385740283338e-24, 3.072009826641111e-24, 3.691806127789155e-24], [25.95212271389581, 3.542768952330983e-24, 3.0605379750980807e-24, 3.6797973824454264e-24], [26.024081720957632, 3.531744486088668e-24, 3.0491338970398974e-24, 3.667843815141978e-24], [26.09624025307391, 3.5207836118228235e-24, 3.0377972130915497e-24, 3.655945156241515e-24], [26.168598863479712, 3.5098859399047945e-24, 3.0265275464753976e-24, 3.6441011388067406e-24], [26.241158106944113, 3.499051083364489e-24, 3.0153245229750112e-24, 3.632311498678352e-24], [26.31391853977442, 3.488278657857325e-24, 3.004187770907912e-24, 3.6205759743931106e-24], [26.386880719820446, 3.477568281655678e-24, 2.993116921132323e-24, 3.6088943071974105e-24], [26.460045206478824, 3.466919575612716e-24, 2.9821116070013546e-24, 3.597266240967067e-24], [26.533412560697215, 3.456332163157085e-24, 2.9711714643466194e-24, 3.585691522228549e-24], [26.606983344978673, 3.445805670259964e-24, 2.9602961314583112e-24, 3.574169900081598e-24], [26.680758123385925, 3.435339725426261e-24, 2.9494852490589414e-24, 3.562701126220593e-24], [26.75473746154571, 3.424933959675392e-24, 2.9387384603008916e-24, 3.551284954902531e-24], [26.82892192665315, 3.414588006505757e-24, 2.928055410723787e-24, 3.53992114285722e-24], [26.903312087476017, 3.404301501891597e-24, 2.917435748234566e-24, 3.528609449331972e-24], [26.977908514359157, 3.394074084256951e-24, 2.9068791231024688e-24, 3.517349636016243e-24], [27.052711779228854, 3.383905394457823e-24, 2.896385187924135e-24, 3.506141467046754e-24], [27.127722455597162, 3.373795075760725e-24, 2.8859535976130826e-24, 3.494984708948517e-24], [27.20294111856644, 3.3637427738282525e-24, 2.8755840093654874e-24, 3.48387913063831e-24], [27.27836834483358, 3.353748136693352e-24, 2.865276082649277e-24, 3.472824503358975e-24], [27.354004712694554, 3.343810814742119e-24, 2.8550294791886216e-24, 3.461820600669901e-24], [27.42985080204879, 3.333930460704164e-24, 2.8448438629284954e-24, 3.4508671984423575e-24], [27.505907194403648, 3.3241067296152455e-24, 2.834718900032027e-24, 3.439964074762817e-24], [27.582174472878886, 3.314339278821521e-24, 2.824654258837759e-24, 3.429111009997411e-24], [27.658653222211093, 3.3046277679390674e-24, 2.8146496098616466e-24, 3.418307786672317e-24], [27.735344028758195, 3.294971858852098e-24, 2.8047046257648406e-24, 3.40755418951215e-24], [27.812247480503945, 3.2853712156821165e-24, 2.7948189813407682e-24, 3.39685000536406e-24], [27.88936416706243, 3.2758255047796403e-24, 2.784992353486199e-24, 3.386195023199743e-24], [27.966694679682625, 3.2663343947009423e-24, 2.7752244211878378e-24, 3.3755890340773755e-24], [28.044239611252863, 3.2568975561854314e-24, 2.7655148655042085e-24, 3.365031831082484e-24], [28.12199955630543, 3.2475146621549387e-24, 2.7558633695330532e-24, 3.354523209389227e-24], [28.19997511102107, 3.2381853876674747e-24, 2.7462696184098633e-24, 3.344062966084679e-24], [28.278166873233687, 3.2289094099312286e-24, 2.7367332992816737e-24, 3.3336509002994286e-24], [28.356575442434746, 3.2196864082587568e-24, 2.727254101276817e-24, 3.3232868130462426e-24], [28.435201419777982, 3.210516064067818e-24, 2.7178317155041814e-24, 3.3129705072661796e-24], [28.514045408083994, 3.201398060857293e-24, 2.7084658350165977e-24, 3.302701787786125e-24], [28.593108011844834, 3.192332084187865e-24, 2.6991561547991842e-24, 3.2924804612670268e-24], [28.672389837228717, 3.1833178216721912e-24, 2.689902371756514e-24, 3.282306336211673e-24], [28.75189149208457, 3.174354962944398e-24, 2.6807041846849056e-24, 3.2721792228830528e-24], [28.831613585946755, 3.165443199654782e-24, 2.6715612942506513e-24, 3.262098933314964e-24], [28.91155673003972, 3.156582225449113e-24, 2.6624734029821952e-24, 3.2520652812813693e-24], [28.991721537282686, 3.1477717359534546e-24, 2.6534402152417783e-24, 3.242078082263125e-24], [29.07210862229439, 3.139011428750434e-24, 2.644461437213707e-24, 3.2321371533922337e-24], [29.15271860139772, 3.130301003365401e-24, 2.6355367768803444e-24, 3.222242313438635e-24], [29.23355209262449, 3.121640161259896e-24, 2.6266659440045398e-24, 3.2123933828240166e-24], [29.314609715720167, 3.1130286057973423e-24, 2.617848650109134e-24, 3.2025901835085836e-24], [29.395892092148614, 3.1044660422392816e-24, 2.6090846084746344e-24, 3.1928325390283006e-24], [29.477399845096905, 3.0959521777290788e-24, 2.6003735340885266e-24, 3.1831202744574126e-24], [29.559133599480013, 3.087486721261653e-24, 2.5917151436647358e-24, 3.173453216322221e-24], [29.64109398194568, 3.0790693836878465e-24, 2.5831091555959836e-24, 3.1638311926584227e-24], [29.723281620879174, 3.0706998776797194e-24, 2.574555289949116e-24, 3.1542540329086264e-24], [29.805697146408104, 3.062377917727787e-24, 2.5660532684482307e-24, 3.144721567940229e-24], [29.888341190407363, 3.0541032201158133e-24, 2.557602814452347e-24, 3.135233629983832e-24], [29.971214386503785, 3.0458755029112616e-24, 2.549203652937634e-24, 3.1257900526286376e-24], [30.054317370081137, 3.0376944859480672e-24, 2.5408555104828028e-24, 3.1163906707893375e-24], [30.137650778284943, 3.029559890806555e-24, 2.532558115248909e-24, 3.1070353206500415e-24], [30.22121525002737, 3.0214714408049585e-24, 2.5243111969663276e-24, 3.097723839677145e-24], [30.305011425992184, 3.013428860979537e-24, 2.51611448691074e-24, 3.088456066563329e-24], [30.389039948639557, 3.005431878069512e-24, 2.5079677178917462e-24, 3.0792318412034502e-24], [30.473301462211072, 2.997480220502587e-24, 2.4998706242315635e-24, 3.070051004667869e-24], [30.55779661273464, 2.98957361837986e-24, 2.4918229417547798e-24, 3.060913399174055e-24], [30.642526048029435, 2.9817118034637113e-24, 2.4838244077628576e-24, 3.0518188680695395e-24], [30.727490417710918, 2.9738945091542663e-24, 2.4758747610201143e-24, 3.0427672557697538e-24], [30.812690373195743, 2.9661214704870768e-24, 2.4679737417453893e-24, 3.03375840778123e-24], [30.89812656770679, 2.9583924241072156e-24, 2.4601210915856612e-24, 3.024792170621979e-24], [30.983799656278148, 2.950707108262621e-24, 2.452316553598739e-24, 3.015868391836384e-24], [31.069710295760245, 2.9430652627819634e-24, 2.444559872244885e-24, 3.0069869199254894e-24], [31.155859144824706, 2.935466629068333e-24, 2.436850793369629e-24, 2.998147604355189e-24], [31.24224686396953, 2.92791095008425e-24, 2.429189064180014e-24, 2.9893502955372593e-24], [31.32887411552413, 2.9203979703287253e-24, 2.4215744332384336e-24, 2.98059484474341e-24], [31.415741563654315, 2.91292743583337e-24, 2.4140066504430277e-24, 2.9718811041368968e-24], [31.50284987436762, 2.905499094146415e-24, 2.4064854670062534e-24, 2.9632089267363058e-24], [31.590199715518143, 2.898112694312196e-24, 2.399010635450467e-24, 2.954578166346831e-24], [31.677791756811807, 2.890767986868734e-24, 2.3915819095889316e-24, 2.9459886775980697e-24], [31.76562666981148, 2.8834647238250033e-24, 2.3841990445014737e-24, 2.9374403158691717e-24], [31.853705127942085, 2.8762026586503658e-24, 2.3768617965341225e-24, 2.9289329372738657e-24], [31.942027806495847, 2.8689815462661017e-24, 2.369569923276247e-24, 2.920466398657453e-24], [32.030595382637365, 2.861801143024661e-24, 2.3623231835442035e-24, 2.9120405575409588e-24], [32.11940853540887, 2.8546612067050117e-24, 2.3551213373728604e-24, 2.9036552721278828e-24], [32.20846794573541, 2.8475614964875537e-24, 2.3479641459956748e-24, 2.8953104012237315e-24], [32.297774296430056, 2.840501772961122e-24, 2.3408513718364345e-24, 2.887005804306255e-24], [32.38732827219921, 2.8334817980862554e-24, 2.3337827784893875e-24, 2.8787413413813423e-24], [32.47713055964776, 2.8265013352035217e-24, 2.326758130708813e-24, 2.8705168730692813e-24], [32.56718184728438, 2.8195601490118187e-24, 2.319777194394285e-24, 2.862332260516953e-24], [32.65748282552683, 2.812658005555284e-24, 2.312839736576984e-24, 2.8541873653849e-24], [32.748034186707194, 2.80579467221257e-24, 2.305945525407802e-24, 2.846082049827464e-24], [32.83883662507732, 2.798969917693811e-24, 2.2990943301432397e-24, 2.8380161765058022e-24], [32.92989083681397, 2.7921835120110105e-24, 2.292285921130982e-24, 2.829989608486428e-24], [33.02119752002424, 2.785435226484562e-24, 2.285520069799229e-24, 2.8220022093014467e-24], [33.112757374750935, 2.7787248337197957e-24, 2.2787965486429485e-24, 2.8140538428697844e-24], [33.20457110297789, 2.7720521076018797e-24, 2.2721151312139806e-24, 2.806144373507498e-24], [33.29663940863541, 2.7654168232814774e-24, 2.2654755921009925e-24, 2.798273665885792e-24], [33.38896299760559, 2.7588187571671057e-24, 2.2588777069304555e-24, 2.7904415850342147e-24], [33.48154257772779, 2.752257686913196e-24, 2.252321252342322e-24, 2.782647996305029e-24], [33.57437885880401, 2.7457333914041784e-24, 2.2458060059845765e-24, 2.7748927653363803e-24], [33.66747255260438, 2.739245650748943e-24, 2.2393317464994767e-24, 2.7671757580518642e-24], [33.76082437287263, 2.732794246277653e-24, 2.232898253511137e-24, 2.7594968406777685e-24], [33.85443503533147, 2.7263789605126397e-24, 2.226505307622653e-24, 2.751855879632884e-24], [33.948305257688176, 2.719999577178144e-24, 2.2201526903927994e-24, 2.7442527416035712e-24], [34.04243575964001, 2.7136558811762624e-24, 2.2138401843325873e-24, 2.736687293461818e-24], [34.13682726287987, 2.707347658587148e-24, 2.2075675728947555e-24, 2.729159402288377e-24], [34.23148049110166, 2.701074696649096e-24, 2.2013346404613533e-24, 2.7216689353166717e-24], [34.32639617000594, 2.6948367837591576e-24, 2.19514117233203e-24, 2.7142157599511057e-24], [34.421575027305444, 2.6886337094588992e-24, 2.188986954715849e-24, 2.7067997437402193e-24], [34.51701779273065, 2.6824652644227917e-24, 2.1828717747240954e-24, 2.699420754337236e-24], [34.61272519803552, 2.6763312404525775e-24, 2.1767954203585787e-24, 2.692078659510228e-24], [34.70869797700289, 2.6702314304726517e-24, 2.170757680498665e-24, 2.6847733271432775e-24], [34.80493686545024, 2.664165628510648e-24, 2.1647583448962748e-24, 2.677504625168539e-24], [34.90144260123526, 2.6581336296944877e-24, 2.158797204168416e-24, 2.6702724215805663e-24], [34.998215924261565, 2.6521352302477177e-24, 2.1528740497779452e-24, 2.663076584440245e-24], [35.09525757648437, 2.6461702274739453e-24, 2.1469886740429195e-24, 2.6559169818310028e-24], [35.19256830191613, 2.6402384197532385e-24, 2.1411408701077847e-24, 2.648793481857675e-24], [35.29014884663226, 2.634339606534266e-24, 2.135330431951893e-24, 2.641705952644648e-24], [35.387999958776874, 2.628473588321028e-24, 2.129557154373302e-24, 2.6346542622906907e-24], [35.48612238856848, 2.6226401666730698e-24, 2.123820832977547e-24, 2.6276382788963904e-24], [35.584516888305835, 2.616839144190519e-24, 2.1181212641812984e-24, 2.620657870522562e-24], [35.683184212373575, 2.611070324511462e-24, 2.1124582451912343e-24, 2.613712905194507e-24], [35.78212511724808, 2.6053335123024492e-24, 2.106831574009763e-24, 2.6068032508880305e-24], [35.881340361503256, 2.599628513252483e-24, 2.1012410494172406e-24, 2.5999287755196047e-24], [35.98083070581633, 2.5939551340628894e-24, 2.0956864709738885e-24, 2.5930893469233406e-24], [36.0805969129738, 2.5883131824452033e-24, 2.0901676390032524e-24, 2.58628483286678e-24], [36.18063974787708, 2.5827024671104007e-24, 2.0846843545949215e-24, 2.579515101021123e-24], [36.28095997754852, 2.577122797763469e-24, 2.079236419595391e-24, 2.5727800189559512e-24], [36.38155837113722, 2.5715739850973463e-24, 2.073823636596697e-24, 2.5660794541371128e-24], [36.48243569992493, 2.566055840785232e-24, 2.0684458089377825e-24, 2.5594132739130023e-24], [36.583592737332054, 2.56056817747673e-24, 2.063102740696064e-24, 2.5527813455159016e-24], [36.68503025892344, 2.5551108087887944e-24, 2.0577942366782992e-24, 2.5461835360400637e-24], [36.7867490424144, 2.5496835493014152e-24, 2.0525201024217684e-24, 2.5396197124516093e-24], [36.88874986767667, 2.5442862145492105e-24, 2.047280144184449e-24, 2.533089741556686e-24], [36.99103351674437, 2.5389186210236246e-24, 2.042074168937653e-24, 2.526593490047408e-24], [37.093600773820064, 2.533580586154583e-24, 2.03690198437083e-24, 2.5201308244253534e-24], [37.19645242528066, 2.528271928315987e-24, 2.0317633988752636e-24, 2.5137016110607984e-24], [37.29958925968353, 2.5229924668125168e-24, 2.0266582215482204e-24, 2.507305716135088e-24], [37.40301206777249, 2.5177420218819656e-24, 2.0215862621829067e-24, 2.5009430056888964e-24], [37.50672164248399, 2.512520414682099e-24, 2.0165473312689813e-24, 2.4946133455699314e-24], [37.61071877895303, 2.507327467293064e-24, 2.0115412399848647e-24, 2.48831660146973e-24], [37.7150042745193, 2.5021630027043132e-24, 2.0065678001934874e-24, 2.4820526388812674e-24], [37.8195789287334, 2.497026844817123e-24, 2.0016268244457957e-24, 2.475821323129245e-24], [37.92444354336279, 2.4919188184386085e-24, 1.9967181259665876e-24, 2.4696225193688067e-24], [38.0295989223982, 2.486838749270447e-24, 1.9918415186590837e-24, 2.4634560925398692e-24], [38.135045872059486, 2.4817864639138306e-24, 1.9869968170987564e-24, 2.45732190742569e-24], [38.24078520080201, 2.476761789857423e-24, 1.98218383653069e-24, 2.4512198286011293e-24], [38.346817719322786, 2.471764555479488e-24, 1.9774023928675996e-24, 2.445149720474865e-24], [38.45314424056669, 2.4667945900366197e-24, 1.972652302684523e-24, 2.439111447244713e-24], [38.55976557973272, 2.461851723665344e-24, 1.9679333832185605e-24, 2.4331048729319646e-24], [38.6666825542802, 2.4569357873766704e-24, 1.9632454523678116e-24, 2.4271298613756847e-24], [38.77389598393509, 2.4520466130496187e-24, 1.9585883286848897e-24, 2.421186276216271e-24], [38.88140669069623, 2.4471840334313367e-24, 1.9539618313790246e-24, 2.415273980915262e-24], [38.98921549884168, 2.442347882129971e-24, 1.9493657803094194e-24, 2.4093928387478025e-24], [39.09732323493505, 2.4375379936131024e-24, 1.9447999959894623e-24, 2.4035427128056985e-24], [39.20573072783178, 2.4327542032037585e-24, 1.940264299581067e-24, 2.3977234660024426e-24], [39.31443880868552, 2.427996347075708e-24, 1.935758512891545e-24, 2.3919349610683395e-24], [39.42344831095454, 2.4232642622525163e-24, 1.9312824583762796e-24, 2.386177060565585e-24], [39.53276007040804, 2.4185577866008857e-24, 1.9268359591347503e-24, 2.3804496268745263e-24], [39.6423749251327, 2.4138767588307784e-24, 1.9224188389117938e-24, 2.374752522211743e-24], [39.75229371553894, 2.4092210184897043e-24, 1.9180309220922307e-24, 2.3690856086268846e-24], [39.862517284367435, 2.40459040596108e-24, 1.913672033703971e-24, 2.363448748004483e-24], [39.9730464766956, 2.3999847624612613e-24, 1.9093419994155424e-24, 2.3578418020760178e-24], [40.08388213994402, 2.3954039300355404e-24, 1.9050406455351434e-24, 2.3522646324157623e-24], [40.19502512388305, 2.3908477515559985e-24, 1.9007677990112756e-24, 2.346717100453091e-24], [40.30647628063916, 2.3863160707184232e-24, 1.8965232874293404e-24, 2.3411990674672725e-24], [40.41823646470161, 2.3818087320395288e-24, 1.892306939014937e-24, 2.3357103946020643e-24], [40.53030653292895, 2.3773255808554973e-24, 1.888118582631243e-24, 2.3302509428723858e-24], [40.642687344555554, 2.3728664633158246e-24, 1.883958047779056e-24, 2.3248205731534956e-24], [40.755379761198334, 2.368431226386588e-24, 1.879825164596808e-24, 2.319419146216774e-24], [40.868384646863184, 2.3640197178419813e-24, 1.8757197638605122e-24, 2.31404652270211e-24], [40.9817028679517, 2.359631786264258e-24, 1.871641676984787e-24, 2.3087025631433654e-24], [41.09533529326775, 2.3552672810441256e-24, 1.867590736020951e-24, 2.3033871279820736e-24], [41.20928279402437, 2.3509260523729127e-24, 1.863566773658908e-24, 2.2981000775546752e-24], [41.32354624385002, 2.3466079512443836e-24, 1.85956962322775e-24, 2.29284127210848e-24], [41.43812651879567, 2.342312829450393e-24, 1.855599118692146e-24, 2.287610571810573e-24], [41.55302449734132, 2.338040539580858e-24, 1.8516550946590503e-24, 2.2824078367606903e-24], [41.668241060402764, 2.33379093501785e-24, 1.847737386374663e-24, 2.277232926977079e-24], [41.783777091338465, 2.329563869937967e-24, 1.8438458297219928e-24, 2.2720857024331196e-24], [41.89963347595612, 2.3253591993070687e-24, 1.8399802612291006e-24, 2.2669660230449443e-24], [42.015811102519606, 2.3211767788790083e-24, 1.8361405180625036e-24, 2.2618737486866597e-24], [42.13231086175573, 2.317016465194221e-24, 1.832326438032423e-24, 2.2568087391984115e-24], [42.24913364686103, 2.312878115575941e-24, 1.8285378595900163e-24, 2.25177085438519e-24], [42.36628035350878, 2.308761588130791e-24, 1.824774621832197e-24, 2.2467599540450778e-24], [42.48375187985562, 2.304666741745057e-24, 1.8210365644999446e-24, 2.2417758979599992e-24], [42.60154912654862, 2.3005934360835282e-24, 1.8173235279792102e-24, 2.236818545914549e-24], [42.71967299673212, 2.2965415315867645e-24, 1.8136353533030384e-24, 2.2318877576961583e-24], [42.83812439605464, 2.292510889470309e-24, 1.8099718821501664e-24, 2.2269833931131968e-24], [42.9569042326759, 2.2885013717221507e-24, 1.806332956850387e-24, 2.2221053119973884e-24], [43.076013417273685, 2.2845128411014134e-24, 1.802718420382213e-24, 2.2172533742173895e-24], [43.19545286305088, 2.280545161135177e-24, 1.7991281163738554e-24, 2.212427439682852e-24], [43.31522348574246, 2.276598196118176e-24, 1.795561889105929e-24, 2.2076273683544335e-24], [43.435326203622466, 2.2726718111109943e-24, 1.7920195835122554e-24, 2.202853020262265e-24], [43.55576193751125, 2.2687658719377887e-24, 1.7885010451804292e-24, 2.1981042555047618e-24], [43.676531610782206, 2.264880245183409e-24, 1.785006120352567e-24, 2.193380934255274e-24], [43.797636149369104, 2.2610147981947565e-24, 1.7815346559290126e-24, 2.1886829167900136e-24], [43.919076481773075, 2.257169399074683e-24, 1.778086499466492e-24, 2.1840100634714227e-24], [44.04085353906976, 2.253343916684958e-24, 1.7746614991813147e-24, 2.179362234783134e-24], [44.16296825491651, 2.2495382206407497e-24, 1.7712595039496816e-24, 2.1747392913224214e-24], [44.2854215655594, 2.2457521813117872e-24, 1.767880363308324e-24, 2.1701410938186244e-24], [44.40821440984055, 2.241985669817586e-24, 1.7645239274579965e-24, 2.1655675031335916e-24], [44.531347729205194, 2.2382385580289573e-24, 1.7611900472632035e-24, 2.1610183802838003e-24], [44.65482246770908, 2.234510718562947e-24, 1.7578785742521547e-24, 2.1564935864343805e-24], [44.7786395720255, 2.2308020247856364e-24, 1.7545893606209557e-24, 2.1519929829329035e-24], [44.902799991452675, 2.227112350805034e-24, 1.75132225923284e-24, 2.147516431288451e-24], [45.027304677921, 2.223441571473899e-24, 1.7480771236202412e-24, 2.1430637932067765e-24], [45.152154586000336, 2.219789562385149e-24, 1.7448538079845464e-24, 2.138634930584284e-24], [45.277350672907374, 2.2161561998716024e-24, 1.741652167199533e-24, 2.1342297055241368e-24], [45.40289389851291, 2.212541361003838e-24, 1.7384720568110506e-24, 2.129847980350154e-24], [45.52878522534921, 2.2089449235878977e-24, 1.73531333303912e-24, 2.125489617602761e-24], [45.655025618617444, 2.2053667661653496e-24, 1.7321758527770956e-24, 2.121154480065174e-24], [45.781616046195, 2.2018067680084118e-24, 1.729059473595183e-24, 2.1168424307556445e-24], [45.90855747864309, 2.1982648091213387e-24, 1.7259640537421503e-24, 2.11255333295014e-24], [46.03585088921392, 2.1947407702373113e-24, 1.7228894521423884e-24, 2.1082870501921498e-24], [46.16349725385834, 2.191234532815977e-24, 1.7198355284008687e-24, 2.1040434462873304e-24], [46.29149755123327, 2.1877459790425496e-24, 1.7168021428032286e-24, 2.09982238532798e-24], [46.41985276270921, 2.184274991826778e-24, 1.7137891563153548e-24, 2.095623731698931e-24], [46.54856387237781, 2.1808214547987443e-24, 1.7107964305862687e-24, 2.091447350080727e-24], [46.67763186705931, 2.177385252309112e-24, 1.707823827947976e-24, 2.0872931054601618e-24], [46.8070577363102, 2.1739662694263713e-24, 1.704871211416993e-24, 2.0831608631469586e-24], [46.93684247243074, 2.170564391935653e-24, 1.7019384446944175e-24, 2.0790504887743733e-24], [47.066987070472614, 2.1671795063356414e-24, 1.6990253921676782e-24, 2.0749618483099722e-24], [47.1974925282466, 2.1638114998379967e-24, 1.696131918910878e-24, 2.0708948080675996e-24], [47.32835984633009, 2.1604602603647916e-24, 1.69325789068657e-24, 2.0668492347107763e-24], [47.45959002807486, 2.1571256765463164e-24, 1.6904031739438728e-24, 2.0628249952665733e-24], [47.591184079614756, 2.153807637719378e-24, 1.6875676358219457e-24, 2.0588219571275273e-24], [47.723143009873326, 2.1505060339260007e-24, 1.684751144150453e-24, 2.054839988069876e-24], [47.85546783057178, 2.1472207559098274e-24, 1.6819535674484353e-24, 2.050878956247305e-24], [47.98815955623643, 2.1439516951162613e-24, 1.6791747749256234e-24, 2.0469387302171243e-24], [48.12121920420668, 2.1406987436886744e-24, 1.676414636484225e-24, 2.0430191789335094e-24], [48.25464779464278, 2.137461794466243e-24, 1.6736730227174875e-24, 2.0391201717563776e-24], [48.3884463505336, 2.134240740984025e-24, 1.6709498049120287e-24, 2.035241578472332e-24], [48.52261589770459, 2.131035477468173e-24, 1.6682448550466402e-24, 2.031383269285558e-24], [48.6571574648255, 2.127845898835776e-24, 1.665558045793557e-24, 2.027545114840191e-24], [48.79207208341836, 2.1246719006920187e-24, 1.6628892505185394e-24, 2.023726986213866e-24], [48.927360787865325, 2.1215133793273444e-24, 1.660238343280538e-24, 2.0199287549343757e-24], [49.06302461541674, 2.118370231716964e-24, 1.6576051988339802e-24, 2.0161502929875968e-24], [49.1990646061989, 2.115242355516833e-24, 1.6549896926262132e-24, 2.0123914728158123e-24], [49.33548180322215, 2.112129649062096e-24, 1.6523917007989307e-24, 2.0086521673305985e-24], [49.47227725238888, 2.109032011366268e-24, 1.6498111001891806e-24, 2.0049322499261504e-24], [49.609452002501435, 2.1059493421164487e-24, 1.6472477683274684e-24, 2.001231594470357e-24], [49.74700710527037, 2.1028815416726043e-24, 1.6447015834387876e-24, 1.997550075322536e-24], [49.884943615322264, 2.0998285110649387e-24, 1.6421724244428517e-24, 1.9938875673379332e-24], [50.02326259020798, 2.09679015199205e-24, 1.6396601709531734e-24, 1.9902439458741257e-24], [50.16196509041069, 2.0937663668176234e-24, 1.6371647032771974e-24, 1.986619086794828e-24], [50.301052179353995, 2.090757058568573e-24, 1.6346859024162727e-24, 1.9830128664761526e-24], [50.44052492341026, 2.0877621309325413e-24, 1.6322236500654232e-24, 1.9794251618161613e-24], [50.580384391908495, 2.0847814882558303e-24, 1.6297778286121491e-24, 1.9758558502344928e-24], [50.720631657142754, 2.0818150355407356e-24, 1.6273483211369568e-24, 1.9723048096852546e-24], [50.8612677943803, 2.078862678442495e-24, 1.624935011413087e-24, 1.9687719186551562e-24], [51.00229388186984, 2.0759243232677676e-24, 1.6225377839046732e-24, 1.9652570561749608e-24], [51.14371100084987, 2.0729998769719382e-24, 1.6201565237675736e-24, 1.9617601018203542e-24], [51.28552023555683, 2.0700892471557707e-24, 1.6177911168483689e-24, 1.9582809357178437e-24], [51.42772267323352, 2.0671923420642985e-24, 1.6154414496832816e-24, 1.9548194385539445e-24], [51.57031940413738, 2.064309070582536e-24, 1.6131074094980906e-24, 1.9513754915711523e-24], [51.713311521548874, 2.0614393422343364e-24, 1.610788884207097e-24, 1.9479489765806076e-24], [51.85670012177992, 2.0585830671793184e-24, 1.6084857624116475e-24, 1.9445397759653593e-24], [52.00048630418221, 2.055740156209844e-24, 1.6061979334006733e-24, 1.941147772677407e-24], [52.14467117115566, 2.0529105207491375e-24, 1.6039252871490328e-24, 1.9377728502544684e-24], [52.28925582815688, 2.0500940728478346e-24, 1.6016677143156115e-24, 1.934414892809614e-24], [52.43424138370762, 2.0472907251817053e-24, 1.599425106244577e-24, 1.931073785047658e-24], [52.579628949403435, 2.0445003910493633e-24, 1.5971973549623833e-24, 1.9277494122619016e-24], [52.7254196399219, 2.0417229843686262e-24, 1.5949843531766656e-24, 1.9244416603399466e-24], [52.8716145730314, 2.0389584196747548e-24, 1.5927859942766348e-24, 1.921150415766845e-24], [53.0182148695996, 2.036206612116709e-24, 1.5906021723299993e-24, 1.9178755656269803e-24], [53.16522165360205, 2.0334674774552752e-24, 1.5884327820837004e-24, 1.914616997610199e-24], [53.3126360521309, 2.0307409320596613e-24, 1.5862777189606685e-24, 1.9113746000125635e-24], [53.460459195403345, 2.028026892905307e-24, 1.5841368790592055e-24, 1.908148261741524e-24], [53.60869221677047, 2.0253252775704488e-24, 1.582010159152562e-24, 1.9049378723148063e-24], [53.75733625272581, 2.0226360042335916e-24, 1.5798974566859304e-24, 1.9017433218670606e-24], [53.906392442914225, 2.0199589916707705e-24, 1.577798669775325e-24, 1.898564501150985e-24], [54.05586193014044, 2.0172941592523714e-24, 1.5757136972075383e-24, 1.895401301538194e-24], [54.20574586037794, 2.0146414269403343e-24, 1.5736424384358778e-24, 1.892253615024051e-24], [54.35604538277772, 2.0120007152858812e-24, 1.5715847935812636e-24, 1.8891213342296174e-24], [54.50676164967707, 2.0093719454254943e-24, 1.569540663428501e-24, 1.8860043523995442e-24], [54.6578958166085, 2.006755039078934e-24, 1.5675099494258618e-24, 1.8829025634099788e-24], [54.80944904230848, 2.0041499185460635e-24, 1.5654925536825236e-24, 1.8798158617666395e-24], [54.96142248872637, 2.0015565067035534e-24, 1.5634883789670377e-24, 1.8767441426056666e-24], [55.11381732103337, 1.998974727002389e-24, 1.5614973287053328e-24, 1.873687301699185e-24], [55.266634707631354, 1.996404503464835e-24, 1.5595193069796783e-24, 1.870645235452022e-24], [55.41987582016204, 1.993845760680926e-24, 1.5575542185252175e-24, 1.867617840906221e-24], [55.5735418335157, 1.991298423806551e-24, 1.5556019687295084e-24, 1.864605015740928e-24], [55.72763392584031, 1.9887624185591113e-24, 1.553662463629529e-24, 1.8616066582732697e-24], [55.88215327855059, 1.9862376712162197e-24, 1.5517356099098662e-24, 1.858622667462373e-24], [56.03710107633701, 1.9837241086103487e-24, 1.549821314900997e-24, 1.8556529429006324e-24], [56.19247850717499, 1.9812216581283205e-24, 1.5479194865764414e-24, 1.85269738482926e-24], [56.348286762333814, 1.9787302477065843e-24, 1.5460300335515214e-24, 1.8497558941237867e-24], [56.50452703638593, 1.9762498058290373e-24, 1.5441528650802586e-24, 1.846828372307613e-24], [56.661200527216025, 1.9737802615231525e-24, 1.5422878910540442e-24, 1.8439147215414812e-24], [56.818308436030215, 1.97132154435775e-24, 1.5404350219986545e-24, 1.8410148446304227e-24], [56.97585196736535, 1.9688735844396392e-24, 1.5385941690720456e-24, 1.8381286450218342e-24], [57.1338323290981, 1.966436312410448e-24, 1.5367652440621433e-24, 1.8352560268052886e-24], [57.29225073245429, 1.9640096594439104e-24, 1.5349481593848442e-24, 1.8323968947139145e-24], [57.45110839201821, 1.961593557241975e-24, 1.5331428280810722e-24, 1.8295511541211093e-24], [57.61040652574184, 1.9591879380327835e-24, 1.5313491638146516e-24, 1.826718711044574e-24], [57.7701463549544, 1.9567927345668573e-24, 1.5295670808694895e-24, 1.8238994721423783e-24], [57.93032910437141, 1.9544078801142578e-24, 1.5277964941478872e-24, 1.821093344713717e-24], [58.090956002104285, 1.952033308461497e-24, 1.5260373191669058e-24, 1.81830023669837e-24], [58.2520282796697, 1.9496689539082962e-24, 1.524289472056935e-24, 1.815520056677578e-24], [58.41354717199899, 1.947314751264864e-24, 1.5225528695582109e-24, 1.8127527138697992e-24], [58.575513917447765, 1.9449706358483247e-24, 1.520827429018804e-24, 1.809998118133177e-24], [58.73792975780518, 1.942636543480061e-24, 1.519113068391825e-24, 1.8072561799637467e-24], [58.900795938303645, 1.9403124104820213e-24, 1.5174097062328023e-24, 1.8045268104931324e-24], [59.06411370762824, 1.9379981736744233e-24, 1.5157172616969147e-24, 1.8018099214898435e-24], [59.22788431792643, 1.9356937703720104e-24, 1.5140356545361381e-24, 1.799105425354352e-24], [59.39210902481753, 1.933399138381049e-24, 1.512364805097413e-24, 1.796413235122774e-24], [59.556789087402386, 1.931114215996612e-24, 1.5107046343183796e-24, 1.793733264461121e-24], [59.72192576827304, 1.9288389419989678e-24, 1.5090550637263044e-24, 1.7910654276668702e-24], [59.887520333522374, 1.9265732556508316e-24, 1.5074160154338865e-24, 1.788409639665943e-24], [60.053574052753916, 1.924317096694185e-24, 1.5057874121372314e-24, 1.785765816012119e-24], [60.22008819909143, 1.9220704053470734e-24, 1.5041691771132329e-24, 1.783133872884082e-24], [60.38706404918875, 1.919833122300633e-24, 1.502561234215684e-24, 1.7805137270867356e-24], [60.5545028832396, 1.9176051887158627e-24, 1.5009635078734984e-24, 1.7779052960436844e-24], [60.722405984987276, 1.9153865462207708e-24, 1.4993759230875917e-24, 1.7753084978019335e-24], [60.890774641734794, 1.9131771369071648e-24, 1.497798405427524e-24, 1.7727232510259214e-24], [61.059610144354366, 1.910976903327222e-24, 1.4962308810290734e-24, 1.7701494749975033e-24], [61.22891378729753, 1.908785788491136e-24, 1.4946732765911912e-24, 1.767587089612812e-24], [61.39868686860504, 1.9066037358636854e-24, 1.4931255193730854e-24, 1.7650360153799887e-24], [61.568930689916776, 1.9044306893607118e-24, 1.4915875371907998e-24, 1.762496173418175e-24], [61.73964655648184, 1.9022665933470463e-24, 1.4900592584152849e-24, 1.7599674854553093e-24], [61.91083577716837, 1.9001113926322925e-24, 1.4885406119686654e-24, 1.7574498738249837e-24], [62.08249966447375, 1.8979650324688578e-24, 1.487031527321086e-24, 1.754943261463901e-24], [62.25463953453456, 1.8958274585481835e-24, 1.4855319344883387e-24, 1.7524475719105347e-24], [62.42725670713672, 1.8936986169979573e-24, 1.4840417640287202e-24, 1.7499627293036947e-24], [62.60035250572566, 1.8915784543792467e-24, 1.4825609470396738e-24, 1.747488658375558e-24], [62.77392825741633, 1.889466917683169e-24, 1.4810894151547831e-24, 1.7450252844563152e-24], [62.947985293003484, 1.8873639543280126e-24, 1.4796271005414313e-24, 1.742572533464644e-24], [63.122524946971836, 1.885269512156205e-24, 1.4781739358966974e-24, 1.7401303319103543e-24], [63.29754855750626, 1.8831835394315297e-24, 1.4767298544454069e-24, 1.737698606886551e-24], [63.473057466502276, 1.881105984835689e-24, 1.4752947899359795e-24, 1.7352772860735425e-24], [63.649053019575966, 1.879036797465606e-24, 1.4738686766385533e-24, 1.7328662977297067e-24], [63.825536566074554, 1.8769759268306267e-24, 1.4724514493408674e-24, 1.7304655706937447e-24], [64.00250945908671, 1.8749233228492227e-24, 1.471043043345692e-24, 1.728075034378903e-24], [64.17997305545272, 1.8728789358458704e-24, 1.4696433944679799e-24, 1.725694618770289e-24], [64.35792871577537, 1.8708427165489905e-24, 1.4682524390311577e-24, 1.7233242544233758e-24], [64.53637780442985, 1.868814616086915e-24, 1.466870113864432e-24, 1.720963872460648e-24], [64.71532168957445, 1.8667945859856603e-24, 1.4654963562999988e-24, 1.7186134045683212e-24], [64.89476174316108, 1.864782578165625e-24, 1.4641311041693297e-24, 1.7162727829926383e-24], [65.0746993409458, 1.8627785449391263e-24, 1.4627742958003514e-24, 1.713941940539896e-24], [65.25513586249919, 1.860782439007098e-24, 1.4614258700145726e-24, 1.7116208105696353e-24], [65.4360726912171, 1.8587942134563907e-24, 1.4600857661237876e-24, 1.7093093269932618e-24], [65.61751121433124, 1.856813821756698e-24, 1.4587539239268557e-24, 1.7070074242714125e-24], [65.79945282291972, 1.854841217758239e-24, 1.457430283706962e-24, 1.7047150374114985e-24], [65.98189891191782, 1.8528763556880575e-24, 1.4561147862283522e-24, 1.702432101962096e-24], [66.16485088012865, 1.8509191901480305e-24, 1.454807372732995e-24, 1.7001585540111153e-24], [66.34831013023383, 1.8489696761115064e-24, 1.453507984937845e-24, 1.6978943301843657e-24], [66.53227806880426, 1.847027768920552e-24, 1.4522165650319964e-24, 1.6956393676400917e-24], [66.71675610631091, 1.8450934242834346e-24, 1.4509330556726607e-24, 1.6933936040655805e-24], [66.90174565713583, 1.8431665982716e-24, 1.4496573999831233e-24, 1.6911569776758418e-24], [67.08724813958261, 1.8412472473168253e-24, 1.448389541549186e-24, 1.6889294272101814e-24], [67.27326497588751, 1.8393353282087176e-24, 1.4471294244158309e-24, 1.6867108919257024e-24], [67.45979759223033, 1.8374307980916698e-24, 1.4458769930848026e-24, 1.684501311598758e-24], [67.64684741874527, 1.8355336144623966e-24, 1.4446321925111872e-24, 1.6823006265191693e-24], [67.83441588953205, 1.833643735167131e-24, 1.4433949681001968e-24, 1.68010877748609e-24], [68.02250444266673, 1.8317611183986536e-24, 1.4421652657046342e-24, 1.6779257058067628e-24], [68.21111452021283, 1.8298857226940364e-24, 1.4409430316212346e-24, 1.6757513532918736e-24], [68.40024756823236, 1.8280175069315746e-24, 1.4397282125878842e-24, 1.673585662252931e-24], [68.58990503679686, 1.826156430328341e-24, 1.4385207557807821e-24, 1.671428575498998e-24], [68.78008837999872, 1.82430245243752e-24, 1.4373206088112924e-24, 1.6692800363312573e-24], [68.97079905596205, 1.8224555331455325e-24, 1.4361277197227287e-24, 1.667139988544082e-24], [69.16203852685399, 1.8206156326697727e-24, 1.4349420369875483e-24, 1.66500837641684e-24], [69.353808258896, 1.818782711555782e-24, 1.4337635095043888e-24, 1.662885144713258e-24], [69.54610972237487, 1.816956730674515e-24, 1.4325920865949906e-24, 1.66077023867845e-24], [69.73894439165439, 1.8151376512200835e-24, 1.4314277180009701e-24, 1.6586636040334845e-24], [69.93231374518619, 1.8133254347069545e-24, 1.4302703538813688e-24, 1.656565186973931e-24], [70.12621926552141, 1.811520042967595e-24, 1.4291199448093649e-24, 1.6544749341648945e-24], [70.32066243932182, 1.8097214381494442e-24, 1.42797644176915e-24, 1.6523927927396475e-24], [70.51564475737139, 1.8079295827131323e-24, 1.4268397961533e-24, 1.6503187102941027e-24], [70.71116771458786, 1.806144439429356e-24, 1.4257099597598432e-24, 1.6482526348845708e-24], [70.9072328100338, 1.804365971376725e-24, 1.4245868847890534e-24, 1.6461945150252907e-24], [71.1038415469284, 1.802594141939027e-24, 1.4234705238405487e-24, 1.6441442996824429e-24], [71.30099543265885, 1.8008289148029177e-24, 1.422360829910813e-24, 1.6421019382735323e-24], [71.49869597879211, 1.7990702539556476e-24, 1.4212577563898958e-24, 1.6400673806622204e-24], [71.69694470108624, 1.79731812368226e-24, 1.4201612570584544e-24, 1.6380405771564028e-24], [71.89574311950213, 1.7955724885632477e-24, 1.4190712860852362e-24, 1.6360214785029059e-24], [72.0950927582152, 1.7938333134724017e-24, 1.4179877980240814e-24, 1.63401003588638e-24], [72.29499514562698, 1.792100563574182e-24, 1.4169107478108822e-24, 1.632006200924615e-24], [72.495451814377, 1.790374204321464e-24, 1.4158400907609095e-24, 1.6300099256648822e-24], [72.69646430135437, 1.7886542014531284e-24, 1.4147757825660596e-24, 1.6280211625819179e-24], [72.8980341477096, 1.7869405209917364e-24, 1.4137177792918166e-24, 1.6260398645735062e-24], [73.10016289886644, 1.7852331292410437e-24, 1.4126660373747499e-24, 1.62406598495771e-24], [73.30285210453371, 1.7835319927841365e-24, 1.4116205136193525e-24, 1.6220994774687171e-24], [73.50610331871738, 1.781837078480454e-24, 1.4105811651954909e-24, 1.6201402962556098e-24], [73.70991809973209, 1.7801483534641912e-24, 1.4095479496358225e-24, 1.6181883958762017e-24], [73.9142980102134, 1.778465785141725e-24, 1.408520824832672e-24, 1.6162437312961926e-24], [74.11924461712968, 1.7767893411891295e-24, 1.4074997490352954e-24, 1.6143062578844046e-24], [74.32475949179407, 1.775118989550274e-24, 1.4064846808476627e-24, 1.6123759314101094e-24], [74.5308442098767, 1.7734546984347103e-24, 1.4054755792251173e-24, 1.6104527080399647e-24], [74.73750035141659, 1.771796436314958e-24, 1.4044724034722302e-24, 1.6085365443343957e-24], [74.9447295008338, 1.770144171924997e-24, 1.4034751132395086e-24, 1.6066273972442838e-24], [75.15253324694164, 1.7684978742571703e-24, 1.402483668521473e-24, 1.6047252241086536e-24], [75.36091318295878, 1.766857512561254e-24, 1.4014980296532347e-24, 1.6028299826498531e-24], [75.56987090652161, 1.765223056341127e-24, 1.4005181573084538e-24, 1.600941630971796e-24], [75.77940801969629, 1.763594475353401e-24, 1.3995440124964014e-24, 1.5990601275563416e-24], [75.98952612899112, 1.7619717396049137e-24, 1.3985755565594096e-24, 1.5971854312599444e-24], [76.20022684536889, 1.7603548193510677e-24, 1.397612751170336e-24, 1.5953175013105318e-24], [76.41151178425912, 1.7587436850932386e-24, 1.3966555583300102e-24, 1.5934562973046844e-24], [76.62338256557078, 1.757138307576846e-24, 1.3957039403644375e-24, 1.5916017792042582e-24], [76.83584081370424, 1.7555386577898223e-24, 1.3947578599224867e-24, 1.5897539073332325e-24], [77.04888815756391, 1.7539447069597804e-24, 1.3938172799733448e-24, 1.5879126423746619e-24], [77.26252623057078, 1.7523564265524843e-24, 1.3928821638037888e-24, 1.5860779453679382e-24], [77.47675667067517, 1.75077378826994e-24, 1.3919524750158705e-24, 1.5842497777047987e-24], [77.69158112036877, 1.7491967640478317e-24, 1.3910281775243287e-24, 1.5824281011279524e-24], [77.90700122669759, 1.747625326054217e-24, 1.3901092355543929e-24, 1.5806128777260849e-24], [78.1230186412745, 1.7460594466873382e-24, 1.38919561363875e-24, 1.5788040699322237e-24], [78.33963502029187, 1.744499098573347e-24, 1.38828727661553e-24, 1.5770016405201718e-24], [78.5568520245344, 1.7429442545649253e-24, 1.3873841896256684e-24, 1.5752055526015858e-24], [78.77467131939162, 1.7413948877388884e-24, 1.3864863181110046e-24, 1.5734157696234716e-24], [78.99309457487087, 1.7398509713947293e-24, 1.3855936278110576e-24, 1.5716322553646964e-24], [79.21212346560996, 1.738312479052366e-24, 1.384706084761187e-24, 1.5698549739326888e-24], [79.43175967089007, 1.736779384450485e-24, 1.3838236552901563e-24, 1.5680838897616746e-24], [79.65200487464874, 1.7352516615446295e-24, 1.3829463060174913e-24, 1.566318967609065e-24], [79.8728607654925, 1.733729284505324e-24, 1.3820740038516304e-24, 1.564560172552718e-24], [80.09432903671009, 1.7322122277163703e-24, 1.3812067159872806e-24, 1.5628074699875002e-24], [80.3164113862853, 1.730700465772901e-24, 1.3803444099028882e-24, 1.5610608256239558e-24], [80.53910951690989, 1.7291939734797266e-24, 1.3794870533591763e-24, 1.559320205483592e-24], [80.76242513599712, 1.7276927258494933e-24, 1.3786346143956509e-24, 1.5575855758976955e-24], [80.98635995569418, 1.7261966981008324e-24, 1.3777870613296094e-24, 1.5558569035038196e-24], [81.21091569289575, 1.724705865656731e-24, 1.3769443627529197e-24, 1.5541341552422902e-24], [81.43609406925701, 1.7232202041428132e-24, 1.3761064875303689e-24, 1.5524172983549686e-24], [81.66189681120687, 1.7217396893855376e-24, 1.3752734047970378e-24, 1.5507063003819875e-24], [81.88832564996133, 1.7202642974105186e-24, 1.3744450839564227e-24, 1.5490011291580226e-24], [82.11538232153652, 1.7187940044408538e-24, 1.373621494678243e-24, 1.5473017528110586e-24], [82.3430685667621, 1.7173287868955307e-24, 1.3728026068961269e-24, 1.5456081397587704e-24], [82.57138613129474, 1.715868621387506e-24, 1.3719883908053485e-24, 1.5439202587065238e-24], [82.8003367656313, 1.7144134847224024e-24, 1.3711788168610648e-24, 1.5422380786438844e-24], [83.02992222512245, 1.7129633538965766e-24, 1.3703738557758803e-24, 1.5405615688429206e-24], [83.26014426998594, 1.7115182060957577e-24, 1.3695734785177956e-24, 1.538890698854976e-24], [83.49100466532022, 1.7100780186931043e-24, 1.3687776563083304e-24, 1.53722543850858e-24], [83.7225051811179, 1.708642769248156e-24, 1.36798636062008e-24, 1.535565757906392e-24], [83.9546475922793, 1.707212435504553e-24, 1.3671995631750848e-24, 1.533911627422757e-24], [84.18743367862642, 1.7057869953890452e-24, 1.36641723594246e-24, 1.532263017701712e-24], [84.42086522491587, 1.704366427009744e-24, 1.365639351136387e-24, 1.530619899654023e-24], [84.65494402085318, 1.7029507086545022e-24, 1.364865881214407e-24, 1.5289822444545878e-24], [84.8896718611062, 1.701539818789513e-24, 1.364096798875156e-24, 1.5273500235401628e-24], [85.12505054531917, 1.70013373605784e-24, 1.3633320770561486e-24, 1.525723208607125e-24], [85.3610818781261, 1.6987324392776357e-24, 1.36257168893266e-24, 1.524101771608929e-24], [85.59776766916488, 1.697335907441036e-24, 1.3618156079147764e-24, 1.5224856847532067e-24], [85.83510973309112, 1.6959441197124716e-24, 1.3610638076459077e-24, 1.520874920500115e-24], [86.07310988959195, 1.694557055427181e-24, 1.3603162620012698e-24, 1.5192694515597616e-24], [86.31176996340021, 1.6931746940899695e-24, 1.3595729450850815e-24, 1.5176692508894366e-24], [86.55109178430814, 1.691797015373572e-24, 1.3588338312294729e-24, 1.5160742916919876e-24], [86.79107718718166, 1.6904239991173247e-24, 1.3580988949922106e-24, 1.5144845474127908e-24], [87.03172801197421, 1.6890556253257518e-24, 1.3573681111548934e-24, 1.5128999917380147e-24], [87.27304610374105, 1.6876918741671807e-24, 1.3566414547213521e-24, 1.5113205985920772e-24], [87.51503331265343, 1.6863327259723545e-24, 1.3559189009153175e-24, 1.5097463421355275e-24], [87.75769149401253, 1.6849781612330847e-24, 1.3552004251790717e-24, 1.5081771967625634e-24], [88.0010225082639, 1.6836281606008275e-24, 1.3544860031715642e-24, 1.5066131370992727e-24], [88.24502822101164, 1.682282704885454e-24, 1.3537756107662734e-24, 1.505054138000915e-24], [88.48971050303265, 1.680941775053818e-24, 1.3530692240499964e-24, 1.5035001745501497e-24], [88.73507123029133, 1.679605352228487e-24, 1.3523668193206569e-24, 1.5019512220547976e-24], [88.98111228395335, 1.678273417686473e-24, 1.3516683730856545e-24, 1.5004072560456378e-24], [89.22783555040056, 1.6769459528578054e-24, 1.3509738620602306e-24, 1.498868252274096e-24], [89.47524292124524, 1.6756229393244426e-24, 1.3502832631657822e-24, 1.4973341867106727e-24], [89.72333629334466, 1.674304358818857e-24, 1.3495965535280167e-24, 1.4958050355421583e-24], [89.97211756881572, 1.672990193222851e-24, 1.3489137104753475e-24, 1.4942807751702915e-24], [90.2215886550493, 1.6716804245661876e-24, 1.3482347115372893e-24, 1.4927613822092894e-24], [90.47175146472509, 1.6703750350256954e-24, 1.347559534442687e-24, 1.4912468334837036e-24], [90.72260791582613, 1.6690740069234613e-24, 1.346888157118015e-24, 1.4897371060270248e-24], [90.97415993165359, 1.6677773227261763e-24, 1.3462205576859784e-24, 1.488232177078927e-24], [91.22640944084158, 1.6664849650437955e-24, 1.3455567144637788e-24, 1.4867320240839228e-24], [91.47935837737171, 1.6651969166281213e-24, 1.3448966059613689e-24, 1.4852366246888752e-24], [91.73300868058814, 1.6639131603718495e-24, 1.3442402108801834e-24, 1.4837459567416945e-24], [91.98736229521234, 1.6626336793073644e-24, 1.3435875081112805e-24, 1.4822599982888127e-24], [92.24242117135796, 1.6613584566055414e-24, 1.3429384767338472e-24, 1.4807787275735575e-24], [92.49818726454605, 1.6600874755745475e-24, 1.3422930960138827e-24, 1.4793021230344196e-24], [92.75466253571963, 1.6588207196590216e-24, 1.341651345402108e-24, 1.4778301633029448e-24], [93.01184895125905, 1.6575581724384178e-24, 1.341013204533153e-24, 1.4763628272018351e-24], [93.26974848299686, 1.6562998176264423e-24, 1.3403786532236473e-24, 1.4749000937435146e-24], [93.52836310823328, 1.6550456390695832e-24, 1.3397476714707368e-24, 1.473441942127939e-24], [93.78769480975085, 1.6537956207461363e-24, 1.3391202394504838e-24, 1.4719883517408479e-24], [94.04774557582998, 1.652549746765271e-24, 1.3384963375169358e-24, 1.4705393021522815e-24], [94.30851740026411, 1.651308001365723e-24, 1.3378759462000492e-24, 1.4690947731144063e-24], [94.57001228237493, 1.650070368914876e-24, 1.3372590462045162e-24, 1.4676547445601543e-24], [94.83223222702793, 1.6488368339076067e-24, 1.3366456184085574e-24, 1.466219196601159e-24], [95.09517924464743, 1.6476073809654399e-24, 1.3360356438618856e-24, 1.4647881095262935e-24], [95.35885535123228, 1.6463819948354545e-24, 1.335429103785071e-24, 1.4633614637998973e-24], [95.62326256837116, 1.645160660389022e-24, 1.3348259795675995e-24, 1.4619392400601635e-24], [95.88840292325813, 1.6439433626211987e-24, 1.3342262527666688e-24, 1.4605214191173255e-24], [96.15427844870824, 1.642730086649428e-24, 1.333629905105868e-24, 1.4591079819521996e-24], [96.42089118317298, 1.6415208177127206e-24, 1.333036918473764e-24, 1.4576989097143918e-24], [96.68824317075601, 1.6403155411705323e-24, 1.3324472749226775e-24, 1.456294183720699e-24], [96.95633646122876, 1.639114242502002e-24, 1.331860956666976e-24, 1.4548937854538504e-24], [97.22517311004613, 1.6379169073047166e-24, 1.331277946082394e-24, 1.4534976965604074e-24], [97.49475517836257, 1.6367235212940848e-24, 1.3306982257041905e-24, 1.4521058988496035e-24], [97.76508473304733, 1.6355340703021594e-24, 1.3301217782259177e-24, 1.4507183742915884e-24], [98.03616384670072, 1.6343485402767372e-24, 1.3295485864985526e-24, 1.449335105015933e-24], [98.30799459766979, 1.6331669172805964e-24, 1.3289786335288622e-24, 1.4479560733102191e-24], [98.5805790700644, 1.63198918749041e-24, 1.3284119024779523e-24, 1.4465812616185766e-24], [98.85391935377325, 1.6308153371959632e-24, 1.3278483766606877e-24, 1.4452106525398433e-24], [99.1280175444796, 1.6296453527991487e-24, 1.3272880395437947e-24, 1.4438442288264577e-24], [99.4028757436777, 1.6284792208132245e-24, 1.3267308747450307e-24, 1.442481973382927e-24], [99.67849605868862, 1.627316927861718e-24, 1.3261768660318115e-24, 1.441123869264356e-24], [99.95488060267655, 1.6261584606779249e-24, 1.3256259973201237e-24, 1.439769899674988e-24], [100.23203149466514, 1.6250038061037232e-24, 1.3250782526730899e-24, 1.438420047966797e-24], [100.5099508595533, 1.6238529510887848e-24, 1.3245336163002254e-24, 1.4370742976382308e-24], [100.78864082813209, 1.622705882689882e-24, 1.3239920725558806e-24, 1.4357326323326253e-24], [101.0681035371004, 1.6215625880699557e-24, 1.3234536059381863e-24, 1.4343950358369456e-24], [101.34834112908193, 1.620423054497274e-24, 1.3229182010880056e-24, 1.4330614920805553e-24], [101.62935575264139, 1.6192872693446915e-24, 1.3223858427877484e-24, 1.4317319851335975e-24], [101.91114956230064, 1.618155220088822e-24, 1.3218565159602714e-24, 1.430406499205865e-24], [102.1937247185559, 1.617026894309137e-24, 1.3213302056676094e-24, 1.429085018645512e-24], [102.47708338789354, 1.61590227968735e-24, 1.3208068971102335e-24, 1.4277675279377342e-24], [102.76122774280736, 1.614781364006503e-24, 1.3202865756255158e-24, 1.4264540117032753e-24], [103.04615996181502, 1.6136641351503693e-24, 1.3197692266868642e-24, 1.4251444546974928e-24], [103.33188222947429, 1.6125505811024675e-24, 1.3192548359029196e-24, 1.423838841808918e-24], [103.61839673640068, 1.6114406899452991e-24, 1.3187433890159487e-24, 1.4225371580580677e-24], [103.9057056792832, 1.6103344498599639e-24, 1.318234871901216e-24, 1.4212393885960994e-24], [104.19381126090249, 1.6092318491249993e-24, 1.3177292705657586e-24, 1.4199455187038453e-24], [104.48271569014622, 1.608132876115841e-24, 1.3172265711474625e-24, 1.4186555337903026e-24], [104.77242118202736, 1.6070375193041089e-24, 1.316726759913993e-24, 1.4173694193916545e-24], [105.06292995770018, 1.6059457672567775e-24, 1.3162298232616457e-24, 1.4160871611700905e-24], [105.35424424447763, 1.604857608635752e-24, 1.3157357477146354e-24, 1.4148087449125819e-24], [105.64636627584906, 1.6037730321966939e-24, 1.315244519923991e-24, 1.4135341565296449e-24], [105.93929829149607, 1.6026920267888063e-24, 1.3147561266663371e-24, 1.4122633820544236e-24], [106.23304253731082, 1.6016145813537246e-24, 1.3142705548433395e-24, 1.4109964076413973e-24], [106.5276012654124, 1.6005406849252946e-24, 1.3137877914804454e-24, 1.4097332195652346e-24], [106.82297673416481, 1.5994703266284669e-24, 1.313307823725944e-24, 1.4084738042199257e-24], [107.11917120819393, 1.598403495679004e-24, 1.312830638850283e-24, 1.4072181481174549e-24], [107.41618695840461, 1.597340181382391e-24, 1.3123562242448356e-24, 1.4059662378867165e-24], [107.71402626199877, 1.5962803731336579e-24, 1.3118845674211786e-24, 1.4047180602726762e-24], [108.01269140249205, 1.5952240604164269e-24, 1.311415656009901e-24, 1.4034736021351507e-24], [108.31218466973206, 1.5941712328023487e-24, 1.3109494777601369e-24, 1.4022328504476811e-24], [108.61250835991562, 1.5931218799505237e-24, 1.3104860205383118e-24, 1.400995792296825e-24], [108.913664775606, 1.5920759916067868e-24, 1.3100252723274893e-24, 1.3997624148807696e-24], [109.21565622575136, 1.5910335576032715e-24, 1.3095672212261377e-24, 1.3985327055085175e-24], [109.51848502570161, 1.5899945678575423e-24, 1.3091118554477957e-24, 1.3973066515988296e-24], [109.82215349722686, 1.5889590123721696e-24, 1.3086591633197358e-24, 1.3960842406794197e-24], [110.12666396853487, 1.5879268812341322e-24, 1.3082091332823197e-24, 1.3948654603856208e-24], [110.43201877428903, 1.5868981646140918e-24, 1.3077617538881973e-24, 1.3936502984597897e-24], [110.73822025562588, 1.5858728527661278e-24, 1.3073170138013328e-24, 1.3924387427500805e-24], [111.04527076017379, 1.5848509360266525e-24, 1.3068749017963025e-24, 1.3912307812097232e-24], [111.35317264207015, 1.5838324048142846e-24, 1.3064354067574036e-24, 1.3900264018959067e-24], [111.66192826198007, 1.582817249629173e-24, 1.3059985176778755e-24, 1.3888255929689875e-24], [111.97153998711427, 1.5818054610523112e-24, 1.3055642236591157e-24, 1.3876283426914997e-24], [112.28201019124681, 1.5807970297450192e-24, 1.305132513909854e-24, 1.386434639427283e-24], [112.59334125473401, 1.579791946448629e-24, 1.3047033777453723e-24, 1.3852444716406415e-24], [112.90553556453236, 1.5787902019836204e-24, 1.3042768045867052e-24, 1.3840578278954157e-24], [113.21859551421645, 1.5777917872492684e-24, 1.3038527839599818e-24, 1.382874696854015e-24], [113.5325235039981, 1.5767966932231427e-24, 1.3034313054955239e-24, 1.3816950672768174e-24], [113.84732194074395, 1.5758049109604006e-24, 1.3030123589271046e-24, 1.3805189280210364e-24], [114.16299323799461, 1.5748164315935621e-24, 1.3025959340913463e-24, 1.3793462680400868e-24], [114.47953981598275, 1.5738312463317347e-24, 1.302182020926886e-24, 1.378177076382663e-24], [114.79696410165185, 1.5728493464602263e-24, 1.3017706094735352e-24, 1.3770113421918627e-24], [115.11526852867458, 1.5718707233400306e-24, 1.3013616898718195e-24, 1.3758490547045683e-24], [115.43445553747148, 1.5708953684073717e-24, 1.3009552523620592e-24, 1.3746902032503904e-24], [115.75452757523011, 1.5699232731730988e-24, 1.300551287283587e-24, 1.3735347772511276e-24], [116.07548709592335, 1.568954429222478e-24, 1.3001497850744084e-24, 1.3723827662197625e-24], [116.39733656032847, 1.5679888282143221e-24, 1.299750736270123e-24, 1.3712341597598668e-24], [116.72007843604553, 1.567026461880808e-24, 1.299354131503468e-24, 1.3700889475646755e-24], [117.04371519751714, 1.5660673220270157e-24, 1.2989599615035554e-24, 1.3689471194164905e-24], [117.36824932604647, 1.5651114005302679e-24, 1.2985682170951352e-24, 1.3678086651858207e-24], [117.69368330981708, 1.5641586893398279e-24, 1.2981788891981171e-24, 1.3666735748306663e-24], [118.0200196439117, 1.563209180476467e-24, 1.2977919688268087e-24, 1.3655418383958113e-24], [118.34726083033092, 1.5622628660319206e-24, 1.2974074470891579e-24, 1.3644134460121131e-24], [118.67540937801333, 1.5613197381685383e-24, 1.2970253151862814e-24, 1.3632883878957284e-24], [119.00446780285364, 1.5603797891188168e-24, 1.2966455644118476e-24, 1.3621666543474825e-24], [119.33443862772297, 1.559443011184964e-24, 1.2962681861512068e-24, 1.3610482357521337e-24], [119.6653243824877, 1.5585093967384566e-24, 1.2958931718810339e-24, 1.3599331225776607e-24], [119.99712760402858, 1.557578938219689e-24, 1.295520513168509e-24, 1.3588213053746508e-24], [120.329850836261, 1.5566516281374542e-24, 1.295150201670986e-24, 1.3577127747755597e-24], [120.66349663015343, 1.555727459068729e-24, 1.29478222913497e-24, 1.3566075214941237e-24], [120.9980675437484, 1.5548064236580277e-24, 1.2944165873959352e-24, 1.355505536324617e-24], [121.33356614218042, 1.5538885146171878e-24, 1.294053268377511e-24, 1.3544068101412387e-24], [121.6699949976971, 1.552973724724894e-24, 1.2936922640908708e-24, 1.3533113338975275e-24], [122.00735668967783, 1.552062046826276e-24, 1.2933335666343587e-24, 1.352219098625651e-24], [122.34565380465433, 1.551153473832614e-24, 1.292977168192648e-24, 1.3511300954357983e-24], [122.68488893633017, 1.5502479987208432e-24, 1.2926230610365072e-24, 1.350044315515617e-24], [123.02506468560024, 1.5493456145333267e-24, 1.292271237521836e-24, 1.3489617501295457e-24], [123.36618366057162, 1.5484463143772964e-24, 1.2919216900894875e-24, 1.3478823906182293e-24], [123.70824847658253, 1.5475500914245701e-24, 1.2915744112645094e-24, 1.3468062283978751e-24], [124.05126175622327, 1.5466569389113762e-24, 1.2912293936556106e-24, 1.3457332549597984e-24], [124.39522612935595, 1.5457668501376357e-24, 1.2908866299548232e-24, 1.3446634618697012e-24], [124.74014423313427, 1.544879818466929e-24, 1.2905461129366518e-24, 1.3435968407671735e-24], [125.08601871202467, 1.543995837325928e-24, 1.2902078354578271e-24, 1.3425333833651064e-24], [125.43285221782538, 1.5431149002042662e-24, 1.289871790456637e-24, 1.3414730814491188e-24], [125.78064740968834, 1.5422370006539762e-24, 1.2895379709525855e-24, 1.340415926877038e-24], [126.12940695413792, 1.5413621322893432e-24, 1.2892063700455492e-24, 1.3393619115783456e-24], [126.47913352509224, 1.5404902887864436e-24, 1.2888769809157264e-24, 1.3383110275535844e-24], [126.82982980388394, 1.5396214638827953e-24, 1.2885497968227927e-24, 1.3372632668738835e-24], [127.18149847927971, 1.5387556513771964e-24, 1.2882248111054965e-24, 1.3362186216804627e-24], [127.53414224750246, 1.5378928451294171e-24, 1.2879020171813441e-24, 1.335177084184017e-24], [127.88776381225013, 1.537033039059554e-24, 1.2875814085458566e-24, 1.3341386466642637e-24], [128.24236588471808, 1.5361762271480756e-24, 1.2872629787723075e-24, 1.3331033014694463e-24], [128.5979511836186, 1.5353224034354673e-24, 1.286946721511187e-24, 1.3320710410158235e-24], [128.9545224352027, 1.5344715620217128e-24, 1.2866326304896695e-24, 1.331041857787091e-24], [129.31208237328067, 1.5336236970662557e-24, 1.2863206995113182e-24, 1.3300157443340894e-24], [129.67063373924256, 1.5327788027874913e-24, 1.2860109224554352e-24, 1.3289926932740906e-24], [130.03017928208018, 1.5319368734625549e-24, 1.285703293276815e-24, 1.3279726972904849e-24], [130.39072175840715, 1.5310979034272253e-24, 1.2853978060051054e-24, 1.3269557491322792e-24], [130.75226393248087, 1.5302618870753035e-24, 1.2850944547445821e-24, 1.3259418416135887e-24], [131.11480857622334, 1.5294288188585865e-24, 1.284793233673478e-24, 1.3249309676132e-24], [131.4783584692421, 1.528598693286451e-24, 1.284494137043867e-24, 1.3239231200741198e-24], [131.84291639885234, 1.527771504925658e-24, 1.284197159180926e-24, 1.3229182920031668e-24], [132.2084851600973, 1.526947248400073e-24, 1.2839022944826905e-24, 1.3219164764704532e-24], [132.57506755577066, 1.5261259183903567e-24, 1.283609537419617e-24, 1.320917666609004e-24], [132.94266639643732, 1.5253075096338423e-24, 1.2833188825341989e-24, 1.3199218556143709e-24], [133.31128450045546, 1.5244920169240412e-24, 1.2830303244405222e-24, 1.3189290367440849e-24], [133.6809246939974, 1.5236794351106219e-24, 1.2827438578238676e-24, 1.3179392033173364e-24], [134.05158981107226, 1.5228697590989917e-24, 1.2824594774403935e-24, 1.3169523487145542e-24], [134.42328269354704, 1.5220629838502136e-24, 1.2821771781166157e-24, 1.3159684663769233e-24], [134.79600619116826, 1.5212591043805616e-24, 1.2818969547491381e-24, 1.314987549806107e-24], [135.16976316158463, 1.520458115761493e-24, 1.2816188023042614e-24, 1.3140095925637334e-24], [135.54455647036787, 1.5196600131192346e-24, 1.2813427158175318e-24, 1.3130345882710662e-24], [135.92038899103565, 1.5188647916346255e-24, 1.281068690393502e-24, 1.3120625306085769e-24], [136.29726360507323, 1.518072446542967e-24, 1.2807967212052227e-24, 1.311093413315644e-24], [136.6751832019551, 1.5172829731336416e-24, 1.2805268034939564e-24, 1.3101272301900788e-24], [137.05415067916817, 1.5164963667499726e-24, 1.28025893256888e-24, 1.3091639750878137e-24], [137.4341689422328, 1.5157126227890267e-24, 1.2799931038066332e-24, 1.3082036419224985e-24], [137.81524090472604, 1.5149317367013268e-24, 1.2797293126509467e-24, 1.3072462246651783e-24], [138.19736948830368, 1.5141537039907158e-24, 1.2794675546125383e-24, 1.3062917173439196e-24], [138.58055762272204, 1.5133785202141119e-24, 1.2792078252683427e-24, 1.3053401140434521e-24], [138.9648082458616, 1.5126061809812128e-24, 1.2789501202616341e-24, 1.3043914089048061e-24], [139.35012430374826, 1.5118366819544724e-24, 1.2786944353014532e-24, 1.3034455961250193e-24], [139.73650875057695, 1.5110700188487902e-24, 1.2784407661622028e-24, 1.3025026699567333e-24], [140.12396454873382, 1.5103061874313454e-24, 1.278189108683649e-24, 1.3015626247079537e-24], [140.51249466881907, 1.5095451835212694e-24, 1.277939458770235e-24, 1.3006254547415855e-24], [140.90210208966923, 1.5087870029897054e-24, 1.2776918123910068e-24, 1.2996911544752485e-24], [141.29278979838105, 1.5080316417594397e-24, 1.2774461655791451e-24, 1.2987597183808884e-24], [141.6845607903332, 1.5072790958046712e-24, 1.2772025144318816e-24, 1.2978311409844116e-24], [142.07741806921018, 1.5065293611510926e-24, 1.2769608551099472e-24, 1.2969054168655069e-24], [142.4713646470251, 1.5057824338753455e-24, 1.2767211838374046e-24, 1.2959825406571945e-24], [142.86640354414212, 1.5050383101051794e-24, 1.2764834969013486e-24, 1.2950625070456623e-24], [143.26253778930078, 1.504296986019022e-24, 1.2762477906516083e-24, 1.294145310769825e-24], [143.659770419638, 1.503558457845976e-24, 1.2760140615003982e-24, 1.2932309466211405e-24], [144.05810448071236, 1.502822721865575e-24, 1.275782305922133e-24, 1.2923194094432774e-24], [144.4575430265271, 1.5020897744076338e-24, 1.2755525204531308e-24, 1.291410694131842e-24], [144.858089119553, 1.5013596118520606e-24, 1.2753247016913008e-24, 1.2905047956340487e-24], [145.25974583075293, 1.5006322306287167e-24, 1.2750988462957817e-24, 1.2896017089485016e-24], [145.66251623960403, 1.4999076272172953e-24, 1.2748749509868405e-24, 1.2887014291249171e-24], [146.0664034341232, 1.4991857981471006e-24, 1.2746530125456307e-24, 1.2878039512638198e-24], [146.47141051088857, 1.498466739996959e-24, 1.2744330278136727e-24, 1.2869092705162824e-24], [146.87754057506513, 1.497750449394957e-24, 1.2742149936929015e-24, 1.286017382083693e-24], [147.2847967404271, 1.4970369230184749e-24, 1.2739989071451644e-24, 1.2851282812174603e-24], [147.69318212938293, 1.4963261575938681e-24, 1.2737847651921542e-24, 1.2842419632187859e-24], [148.10269987299876, 1.4956181498963728e-24, 1.2735725649150744e-24, 1.2833584234383918e-24], [148.51335311102193, 1.494912896750104e-24, 1.2733623034543562e-24, 1.2824776572763063e-24], [148.92514499190622, 1.494210395027671e-24, 1.2731539780095374e-24, 1.2815996601815748e-24], [149.33807867283465, 1.4935106416502688e-24, 1.2729475858389038e-24, 1.2807244276520512e-24], [149.75215731974487, 1.492813633587418e-24, 1.2727431242593624e-24, 1.2798519552341727e-24], [150.16738410735286, 1.4921193678568046e-24, 1.2725405906460535e-24, 1.278982238522667e-24], [150.58376221917683, 1.4914278415243674e-24, 1.272339982432366e-24, 1.2781152731604124e-24], [151.00129484756286, 1.490739051703918e-24, 1.272141297109442e-24, 1.2772510548381326e-24], [151.41998519370793, 1.4900529955570747e-24, 1.271944532226177e-24, 1.2763895792941984e-24], [151.83983646768567, 1.489369670293331e-24, 1.27174968538886e-24, 1.2755308423144461e-24], [152.26085188847043, 1.4886890731696613e-24, 1.2715567542610658e-24, 1.274674839731912e-24], [152.68303468396155, 1.4880112014906692e-24, 1.2713657365632685e-24, 1.2738215674266488e-24], [153.10638809100928, 1.4873360526082254e-24, 1.2711766300728901e-24, 1.2729710213254853e-24], [153.53091535543794, 1.4866636239215366e-24, 1.2709894326238533e-24, 1.2721231974018636e-24], [153.95661973207294, 1.4859939128770173e-24, 1.270804142106581e-24, 1.2712780916756214e-24], [154.38350448476342, 1.4853269169680545e-24, 1.2706207564675807e-24, 1.2704357002127524e-24], [154.8115728864092, 1.4846626337350854e-24, 1.2704392737094456e-24, 1.269596019125298e-24], [155.24082821898438, 1.4840010607653416e-24, 1.2702596918905546e-24, 1.2687590445710602e-24], [155.6712737735637, 1.483342195692818e-24, 1.2700820091249292e-24, 1.2679247727534504e-24], [156.10291285034725, 1.4826860361981987e-24, 1.2699062235819774e-24, 1.267093199921335e-24], [156.5357487586853, 1.4820325800087335e-24, 1.2697323334863866e-24, 1.2662643223687862e-24], [156.96978481710477, 1.4813818248980706e-24, 1.269560337117893e-24, 1.2654381364349682e-24], [157.4050243533335, 1.4807337686863821e-24, 1.2693902328111526e-24, 1.2646146385039155e-24], [157.84147070432672, 1.4800884092399613e-24, 1.2692220189554807e-24, 1.263793825004349e-24], [158.27912721629227, 1.4794457444714297e-24, 1.2690556939947782e-24, 1.2629756924095586e-24], [158.71799724471572, 1.4788057723395065e-24, 1.268891256427255e-24, 1.262160237237198e-24], [159.15808415438727, 1.4781684908489055e-24, 1.2687287048053891e-24, 1.2613474560490849e-24], [159.59939131942622, 1.4775338980503197e-24, 1.2685680377356513e-24, 1.2605373454511177e-24], [160.04192212330793, 1.4769019920403478e-24, 1.268409253878357e-24, 1.2597299020930463e-24], [160.48567995888925, 1.4762727709613586e-24, 1.2682523519475503e-24, 1.2589251226683454e-24], [160.93066822843468, 1.475646233001457e-24, 1.268097330710872e-24, 1.2581230039140614e-24], [161.37689034364197, 1.4750223763943761e-24, 1.2679441889892924e-24, 1.2573235426106363e-24], [161.82434972566915, 1.4744011994195263e-24, 1.2677929256570363e-24, 1.2565267355818006e-24], [162.2730498051603, 1.4737827004017219e-24, 1.2676435396414714e-24, 1.2557325796943716e-24], [162.72299402227134, 1.4731668777112828e-24, 1.2674960299228978e-24, 1.2549410718581504e-24], [163.17418582669757, 1.4725537297639257e-24, 1.2673503955343241e-24, 1.2541522090257845e-24], [163.62662867769887, 1.4719432550206182e-24, 1.2672066355615958e-24, 1.253365988192583e-24], [164.08032604412728, 1.471335451987701e-24, 1.2670647491429198e-24, 1.2525824063964264e-24], [164.5352814044531, 1.470730319216591e-24, 1.266924735468963e-24, 1.2518014607176076e-24], [164.99149824679108, 1.470127855303968e-24, 1.2667865937826348e-24, 1.2510231482787295e-24], [165.44898006892848, 1.4695280588915129e-24, 1.266650323378972e-24, 1.250247466244523e-24], [165.90773037835024, 1.4689309286659823e-24, 1.2665159236049304e-24, 1.2494744118217792e-24], [166.36775269226735, 1.4683364633591294e-24, 1.2663833938594612e-24, 1.2487039822591996e-24], [166.82905053764307, 1.4677446617475784e-24, 1.2662527335931276e-24, 1.2479361748472501e-24], [167.2916274512197, 1.467155522652909e-24, 1.2661239423081316e-24, 1.247170986918087e-24], [167.75548697954665, 1.4665690449415118e-24, 1.2659970195581901e-24, 1.2464084158454173e-24], [168.22063267900643, 1.4659852275245522e-24, 1.2658719649484265e-24, 1.245648459044373e-24], [168.68706811584303, 1.4654040693580094e-24, 1.2657487781351431e-24, 1.2448911139714452e-24], [169.15479686618863, 1.4648255694425004e-24, 1.2656274588258476e-24, 1.2441363781243087e-24], [169.62382251609134, 1.4642497268233462e-24, 1.2655080067790789e-24, 1.2433842490417537e-24], [170.094148661542, 1.4636765405905095e-24, 1.2653904218042676e-24, 1.2426347243035943e-24], [170.56577890850295, 1.4631060098785282e-24, 1.2652747037616722e-24, 1.2418878015305163e-24], [171.0387168729343, 1.4625381338665493e-24, 1.2651608525622968e-24, 1.2411434783840347e-24], [171.51296618082287, 1.4619729117781712e-24, 1.2650488681677218e-24, 1.2404017525663512e-24], [171.98853046820952, 1.4614103428815676e-24, 1.2649387505900587e-24, 1.2396626218202707e-24], [172.4654133812164, 1.4608504264893629e-24, 1.2648304998918375e-24, 1.238926083929132e-24], [172.9436185760762, 1.4602931619585592e-24, 1.2647241161858896e-24, 1.2381921367166516e-24], [173.42314971915874, 1.459738548690673e-24, 1.2646195996352162e-24, 1.2374607780469164e-24], [173.9040104870004, 1.45918658613159e-24, 1.2645169504530903e-24, 1.2367320058242527e-24], [174.38620456633163, 1.4586372737714873e-24, 1.2644161689026826e-24, 1.2360058179931074e-24], [174.86973565410483, 1.4580906111449664e-24, 1.2643172552972224e-24, 1.2352822125380337e-24], [175.35460745752383, 1.457546597830922e-24, 1.264220209999758e-24, 1.2345611874835587e-24], [175.84082369407076, 1.4570052334525995e-24, 1.2641250334230982e-24, 1.2338427408941396e-24], [176.3283880915367, 1.4564665176774651e-24, 1.2640317260298077e-24, 1.2331268708740399e-24], [176.81730438804777, 1.4559304502173236e-24, 1.2639402883320681e-24, 1.2324135755673007e-24], [177.3075763320959, 1.4553970308282369e-24, 1.2638507208915628e-24, 1.2317028531576506e-24], [177.7992076825659, 1.45486625931049e-24, 1.2637630243195004e-24, 1.2309947018684156e-24], [178.29220220876573, 1.4543381355086073e-24, 1.2636771992764593e-24, 1.230289119962472e-24], [178.7865636904546, 1.4538126593113445e-24, 1.2635932464723336e-24, 1.2295861057421583e-24], [179.2822959178717, 1.4532898306517344e-24, 1.2635111666662738e-24, 1.2288856575492517e-24], [179.77940269176617, 1.4527696495069432e-24, 1.263430960666668e-24, 1.2281877737648395e-24], [180.2778878234252, 1.4522521158983935e-24, 1.2633526293309533e-24, 1.2274924528093102e-24], [180.77775513470414, 1.451737229891757e-24, 1.2632761735656877e-24, 1.2267996931422803e-24], [181.27900845805556, 1.4512249915968205e-24, 1.2632015943263403e-24, 1.226109493262511e-24], [181.78165163655783, 1.4507154011676314e-24, 1.2631288926174175e-24, 1.2254218517078725e-24], [182.28568852394608, 1.4502084588024426e-24, 1.2630580694922278e-24, 1.2247367670553049e-24], [182.79112298464014, 1.4497041647437122e-24, 1.2629891260529105e-24, 1.2240542379207273e-24], [183.2979588937755, 1.4492025192780842e-24, 1.2629220634504028e-24, 1.2233742629590016e-24], [183.80620013723257, 1.448703522736457e-24, 1.2628568828842798e-24, 1.2226968408639017e-24], [184.3158506116656, 1.4482071754939264e-24, 1.2627935856028485e-24, 1.2220219703680422e-24], [184.82691422453433, 1.4477134779698095e-24, 1.2627321729030154e-24, 1.2213496502428306e-24], [185.33939489413183, 1.447222430627715e-24, 1.2626726461301678e-24, 1.2206798792984558e-24], [185.853296549617, 1.4467340339754578e-24, 1.2626150066783154e-24, 1.2200126563837977e-24], [186.3686231310421, 1.4462482885650939e-24, 1.2625592559898345e-24, 1.219347980386405e-24], [186.88537858938514, 1.445765194992995e-24, 1.262505395555595e-24, 1.2186858502324824e-24], [187.40356688657846, 1.4452847538998256e-24, 1.262453426914846e-24, 1.2180262648868134e-24], [187.92319199554044, 1.4448069659705096e-24, 1.2624033516551305e-24, 1.2173692233527436e-24], [188.4442579002054, 1.4443318319343183e-24, 1.2623551714123615e-24, 1.2167147246721315e-24], [188.96676859555365, 1.4438593525648794e-24, 1.2623088878706477e-24, 1.2160627679253467e-24], [189.49072808764336, 1.4433895286801331e-24, 1.2622645027624094e-24, 1.215413352231196e-24], [190.01614039363972, 1.4429223611424232e-24, 1.262222017868217e-24, 1.2147664767469203e-24], [190.5430095418472, 1.4424578508584978e-24, 1.2621814350168077e-24, 1.2141221406681654e-24], [191.07133957173977, 1.4419959987795235e-24, 1.2621427560850797e-24, 1.213480343228944e-24], [191.60113453399129, 1.4415368059010978e-24, 1.2621059829980438e-24, 1.2128410837016138e-24], [192.13239849050794, 1.4410802732632972e-24, 1.2620711177287585e-24, 1.2122043613968491e-24], [192.66513551445763, 1.4406264019506951e-24, 1.2620381622983617e-24, 1.2115701756636301e-24], [193.19934969030362, 1.4401751930924104e-24, 1.262007118776043e-24, 1.2109385258892208e-24], [193.73504511383288, 1.4397266478621026e-24, 1.2619779892789665e-24, 1.210309411499131e-24], [194.27222589219016, 1.4392807674779792e-24, 1.2619507759723066e-24, 1.2096828319571082e-24], [194.81089614390748, 1.4388375532029225e-24, 1.2619254810692155e-24, 1.2090587867651341e-24], [195.3510599989371, 1.438397006344437e-24, 1.2619021068307293e-24, 1.2084372754633965e-24], [195.8927215986827, 1.4379591282547066e-24, 1.2618806555659279e-24, 1.2078182976302633e-24], [196.43588509603052, 1.4375239203306132e-24, 1.26186112963174e-24, 1.2072018528822917e-24], [196.9805546553825, 1.437091384013839e-24, 1.2618435314330025e-24, 1.2065879408742116e-24], [197.52673445268667, 1.436661520790794e-24, 1.2618278634224656e-24, 1.2059765612989017e-24], [198.07442867547059, 1.4362343321927388e-24, 1.261814128100738e-24, 1.2053677138873856e-24], [198.6236415228728, 1.435809819795831e-24, 1.2618023280163452e-24, 1.2047613984088528e-24], [199.17437720567443, 1.4353879852210954e-24, 1.261792465765598e-24, 1.2041576146706112e-24], [199.72663994633297, 1.434968830134512e-24, 1.2617845439927703e-24, 1.203556362518104e-24], [200.2804339790129, 1.4345523562470237e-24, 1.2617785653898815e-24, 1.2029576418348947e-24], [200.83576354961963, 1.4341385653146584e-24, 1.2617745326968767e-24, 1.2023614525426902e-24], [201.39263291583148, 1.4337274591384969e-24, 1.2617724487015167e-24, 1.2017677946013165e-24], [201.95104634713164, 1.4333190395646885e-24, 1.261772316239387e-24, 1.2011766680087076e-24], [202.51100812484242, 1.4329133084846467e-24, 1.2617741381939614e-24, 1.2005880728009544e-24], [203.07252254215635, 1.432510267834916e-24, 1.2617779174965048e-24, 1.200002009052257e-24], [203.63559390417055, 1.432109919597339e-24, 1.2617836571261744e-24, 1.199418476874956e-24], [204.20022652791903, 1.4317122657990797e-24, 1.261791360109939e-24, 1.1988374764195361e-24], [204.76642474240595, 1.4313173085126257e-24, 1.261801029522652e-24, 1.198259007874614e-24], [205.3341928886383, 1.4309250498559284e-24, 1.26181266848702e-24, 1.197683071466981e-24], [205.9035353196604, 1.4305354919923645e-24, 1.2618262801736037e-24, 1.1971096674615763e-24], [206.47445640058564, 1.4301486371308491e-24, 1.2618418678008725e-24, 1.196538796161515e-24], [207.0469605086315, 1.4297644875258703e-24, 1.2618594346351566e-24, 1.1959704579080972e-24], [207.62105203315247, 1.4293830454775202e-24, 1.261878983990677e-24, 1.195404653080819e-24], [208.19673537567294, 1.4290043133316118e-24, 1.2619005192296e-24, 1.1948413820973856e-24], [208.77401494992256, 1.4286282934796772e-24, 1.2619240437619732e-24, 1.1942806454137312e-24], [209.3528951818684, 1.4282549883590597e-24, 1.2619495610458198e-24, 1.1937224435240288e-24], [209.93338050975024, 1.4278844004529448e-24, 1.2619770745871126e-24, 1.1931667769607016e-24], [210.51547538411404, 1.4275165322904436e-24, 1.2620065879397537e-24, 1.1926136462944534e-24], [211.09918426784546, 1.4271513864466461e-24, 1.262038104705691e-24, 1.1920630521342837e-24], [211.68451163620526, 1.4267889655427003e-24, 1.2620716285348328e-24, 1.1915149951275096e-24], [212.27146197686284, 1.4264292722458144e-24, 1.262107163125159e-24, 1.190969475959769e-24], [212.86003978993102, 1.426072309269385e-24, 1.2621447122226772e-24, 1.1904264953550659e-24], [213.4502495879998, 1.4257180793730717e-24, 1.2621842796214476e-24, 1.1898860540757983e-24], [214.04209589617219, 1.4253665853627771e-24, 1.2622258691636877e-24, 1.1893481529227507e-24], [214.63558325209755, 1.4250178300907698e-24, 1.2622694847396544e-24, 1.1888127927351368e-24], [215.23071620600757, 1.424671816455775e-24, 1.2623151302878176e-24, 1.1882799743906428e-24], [215.82749932075072, 1.4243285474029906e-24, 1.2623628097948045e-24, 1.1877496988054268e-24], [216.42593717182646, 1.4239880259242262e-24, 1.2624125272954203e-24, 1.1872219669341786e-24], [217.0260343474219, 1.4236502550578171e-24, 1.2624642868727091e-24, 1.1866967797700911e-24], [217.6277954484452, 1.4233152378888925e-24, 1.262518092657991e-24, 1.186174138344962e-24], [218.23122508856238, 1.4229829775493316e-24, 1.2625739488308796e-24, 1.1856540437291833e-24], [218.83632789423217, 1.4226534772178799e-24, 1.2626318596193014e-24, 1.1851364970317942e-24], [219.44310850474048, 1.4223267401201078e-24, 1.2626918292995009e-24, 1.1846214994004624e-24], [220.05157157223792, 1.422002769528705e-24, 1.262753862196186e-24, 1.1841090520216117e-24], [220.66172176177344, 1.4216815687633298e-24, 1.2628179626824449e-24, 1.183599156120365e-24], [221.27356375133166, 1.4213631411908197e-24, 1.2628841351798336e-24, 1.1830918129606343e-24], [221.88710223186825, 1.4210474902252047e-24, 1.262952384158403e-24, 1.1825870238451338e-24], [222.50234190734503, 1.4207346193278341e-24, 1.2630227141367435e-24, 1.18208479011544e-24], [223.11928749476766, 1.4204245320073913e-24, 1.2630951296820144e-24, 1.1815851131519966e-24], [223.73794372421986, 1.4201172318200241e-24, 1.2631696354099927e-24, 1.1810879943741893e-24], [224.35831533890223, 1.419812722369409e-24, 1.2632462359851264e-24, 1.1805934352403653e-24], [224.98040709516565, 1.4195110073067986e-24, 1.2633249361205092e-24, 1.1801014372478749e-24], [225.60422376255025, 1.419212090331129e-24, 1.263405740578043e-24, 1.1796120019331142e-24], [226.22977012382012, 1.4189159751891339e-24, 1.2634886541683439e-24, 1.1791251308715906e-24], [226.85705097500144, 1.418622665675354e-24, 1.2635736817509108e-24, 1.1786408256779307e-24], [227.48607112541868, 1.418332165632262e-24, 1.2636608282340496e-24, 1.1781590880059435e-24], [228.1168353977306, 1.4180444789503309e-24, 1.2637500985750307e-24, 1.1776799195486667e-24], [228.7493486279691, 1.41775960956815e-24, 1.2638414977800907e-24, 1.177203322038414e-24], [229.38361566557424, 1.4174775614724597e-24, 1.263935030904439e-24, 1.1767292972468131e-24], [230.0196413734331, 1.4171983386982678e-24, 1.2640307030523683e-24, 1.1762578469848641e-24], [230.65743062791643, 1.4169219453289188e-24, 1.264128519377288e-24, 1.1757889731029738e-24], [231.2969883189152, 1.4166483854961699e-24, 1.264228485081725e-24, 1.1753226774910131e-24], [231.93831934987992, 1.4163776633803403e-24, 1.2643306054174719e-24, 1.1748589620783861e-24], [232.58142863785594, 1.4161097832102973e-24, 1.2644348856855332e-24, 1.1743978288340351e-24], [233.22632111352402, 1.4158447492636358e-24, 1.264541331236241e-24, 1.173939279766542e-24], [233.87300172123497, 1.4155825658667137e-24, 1.264649947469305e-24, 1.1734833169241462e-24], [234.5214754190502, 1.4153232373947498e-24, 1.2647607398338195e-24, 1.1730299423948119e-24], [235.17174717877793, 1.4150667682719163e-24, 1.264873713828423e-24, 1.1725791583062776e-24], [235.82382198601178, 1.4148131629714599e-24, 1.2649888750011976e-24, 1.1721309668261202e-24], [236.47770484017067, 1.4145624260157094e-24, 1.2651062289498573e-24, 1.171685370161785e-24], [237.13340075453402, 1.4143145619762743e-24, 1.2652257813217815e-24, 1.1712423705606805e-24], [237.79091475628306, 1.4140695754740559e-24, 1.2653475378139734e-24, 1.1708019703101997e-24], [238.45025188653727, 1.4138274711793649e-24, 1.265471504173279e-24, 1.1703641717377912e-24], [239.1114172003946, 1.4135882538120015e-24, 1.2655976861963014e-24, 1.1699289772110134e-24], [239.77441576696964, 1.4133519281414032e-24, 1.2657260897295532e-24, 1.169496389137608e-24], [240.43925266943162, 1.4131184989866625e-24, 1.2658567206694623e-24, 1.1690664099655405e-24], [241.10593300504524, 1.4128879712166568e-24, 1.2659895849624786e-24, 1.1686390421830628e-24], [241.77446188520764, 1.4126603497501491e-24, 1.2661246886050935e-24, 1.1682142883187794e-24], [242.4448444354894, 1.4124356395558814e-24, 1.2662620376439082e-24, 1.167792150941712e-24], [243.11708579567303, 1.4122138456526502e-24, 1.2664016381757499e-24, 1.1673726326613486e-24], [243.79119111979182, 1.4119949731094318e-24, 1.2665434963476448e-24, 1.1669557361277173e-24], [244.46716557617077, 1.4117790270454494e-24, 1.2666876183569814e-24, 1.1665414640314442e-24], [245.14501434746455, 1.4115660126302978e-24, 1.2668340104514897e-24, 1.1661298191038158e-24], [245.82474263069878, 1.411355935084011e-24, 1.2669826789293362e-24, 1.1657208041168437e-24], [246.5063556373089, 1.4111487996772017e-24, 1.2671336301392179e-24, 1.1653144218833367e-24], [247.18985859318073, 1.4109446117311076e-24, 1.2672868704804313e-24, 1.1649106752569518e-24], [247.87525673868942, 1.4107433766177342e-24, 1.2674424064028517e-24, 1.1645095671322705e-24], [248.5625553287414, 1.4105450997599289e-24, 1.267600244407122e-24, 1.1641111004448606e-24], [249.2517596328128, 1.4103497866314914e-24, 1.2677603910446152e-24, 1.1637152781713484e-24], [249.94287493499138, 1.410157442757267e-24, 1.2679228529175881e-24, 1.1633221033294756e-24], [250.63590653401644, 1.4099680737132704e-24, 1.268087636679223e-24, 1.1629315789781883e-24], [251.3308597433187, 1.4097816851267337e-24, 1.2682547490336317e-24, 1.1625437082176708e-24], [252.02773989106274, 1.4095982826762748e-24, 1.268424196736051e-24, 1.1621584941894522e-24], [252.72655232018576, 1.4094178720919498e-24, 1.2685959865927979e-24, 1.1617759400764521e-24], [253.42730238844052, 1.4092404591553616e-24, 1.2687701254613936e-24, 1.1613960491030522e-24], [254.1299954684355, 1.4090660496997847e-24, 1.2689466202506696e-24, 1.1610188245351779e-24], [254.8346369476754, 1.4088946496102474e-24, 1.269125477920768e-24, 1.1606442696803593e-24], [255.54123222860386, 1.4087262648236404e-24, 1.2693067054832365e-24, 1.1602723878878065e-24], [256.249786728644, 1.4085609013288354e-24, 1.2694903100011407e-24, 1.159903182548484e-24], [256.96030588024047, 1.4083985651667407e-24, 1.2696762985890975e-24, 1.1595366570951715e-24], [257.6727951309001, 1.4082392624304792e-24, 1.2698646784133635e-24, 1.159172815002555e-24], [258.3872599432346, 1.4080829992654303e-24, 1.2700554566919064e-24, 1.1588116597872854e-24], [259.10370579500244, 1.4079297818693656e-24, 1.2702486406944833e-24, 1.1584531950080597e-24], [259.8221381791505, 1.4077796164925089e-24, 1.2704442377427135e-24, 1.1580974242656765e-24], [260.54256260385654, 1.407632509437734e-24, 1.2706422552101607e-24, 1.1577443512031483e-24], [261.2649845925703, 1.4074884670605804e-24, 1.2708427005224093e-24, 1.1573939795057426e-24], [261.9894096840579, 1.4073474957694033e-24, 1.2710455811571097e-24, 1.157046312901072e-24], [262.71584343244194, 1.4072096020254321e-24, 1.2712509046441168e-24, 1.1567013551591535e-24], [263.44429140724617, 1.407074792342974e-24, 1.271458678565513e-24, 1.1563591100925198e-24], [264.1747591934372, 1.4069430732894017e-24, 1.2716689105556951e-24, 1.1560195815562511e-24], [264.9072523914667, 1.4068144514853518e-24, 1.2718816083015028e-24, 1.1556827734480909e-24], [265.6417766173161, 1.4066889336047845e-24, 1.2720967795422259e-24, 1.1553486897084951e-24], [266.3783375025375, 1.4065665263750863e-24, 1.2723144320697104e-24, 1.1550173343207197e-24], [267.1169406942987, 1.4064472365772297e-24, 1.272534573728481e-24, 1.1546887113109099e-24], [267.8575918554258, 1.4063310710458197e-24, 1.2727572124157337e-24, 1.1543628247481593e-24], [268.60029666444586, 1.4062180366692175e-24, 1.2729823560815006e-24, 1.1540396787445943e-24], [269.34506081563194, 1.4061081403896995e-24, 1.2732100127286812e-24, 1.1537192774554694e-24], [270.09189001904485, 1.4060013892034553e-24, 1.273440190413147e-24, 1.1534016250792067e-24], [270.8407900005798, 1.4058977901608448e-24, 1.2736728972437847e-24, 1.153086725857534e-24], [271.59176650200664, 1.4057973503663527e-24, 1.273908141382631e-24, 1.1527745840754985e-24], [272.34482528101717, 1.4057000769788159e-24, 1.2741459310449232e-24, 1.1524652040616047e-24], [273.09997211126677, 1.4056059772114444e-24, 1.2743862744991768e-24, 1.1521585901878457e-24], [273.8572127824207, 1.4055150583320298e-24, 1.2746291800672639e-24, 1.1518547468698359e-24], [274.61655310019734, 1.4054273276629352e-24, 1.274874656124535e-24, 1.1515536785668356e-24], [275.37799888641234, 1.405342792581311e-24, 1.2751227110998593e-24, 1.1512553897818817e-24], [276.1415559790249, 1.4052614605191106e-24, 1.2753733534757113e-24, 1.1509598850618256e-24], [276.9072302321802, 1.4051833389632993e-24, 1.2756265917882787e-24, 1.1506671689974604e-24], [277.67502751625665, 1.4051084354558665e-24, 1.2758824346275298e-24, 1.1503772462235551e-24], [278.44495371790964, 1.4050367575940132e-24, 1.2761408906372918e-24, 1.1500901214189807e-24], [279.21701474011604, 1.4049683130301857e-24, 1.276401968515325e-24, 1.1498057993067485e-24], [279.9912165022214, 1.404903109472269e-24, 1.2766656770134616e-24, 1.1495242846541383e-24], [280.7675649399828, 1.4048411546836273e-24, 1.2769320249375992e-24, 1.1492455822727416e-24], [281.5460660056176, 1.4047824564832668e-24, 1.2772010211478848e-24, 1.1489696970185774e-24], [282.32672566784584, 1.4047270227458732e-24, 1.2774726745587131e-24, 1.1486966337921354e-24], [283.1095499119373, 1.4046748614020213e-24, 1.2777469941388471e-24, 1.1484263975385106e-24], [283.8945447397585, 1.4046259804381732e-24, 1.278023988911532e-24, 1.1481589932474299e-24], [284.68171616981635, 1.4045803878968987e-24, 1.2783036679545157e-24, 1.14789442595339e-24], [285.4710702373075, 1.4045380918768985e-24, 1.2785860404002009e-24, 1.1476327007356999e-24], [286.26261299416086, 1.4044991005331415e-24, 1.2788711154356542e-24, 1.147373822718581e-24], [287.05635050908717, 1.4044634220770088e-24, 1.2791589023027867e-24, 1.1471177970712598e-24], [287.8522888676232, 1.4044310647763458e-24, 1.2794494102983324e-24, 1.146864629008033e-24], [288.6504341721804, 1.4044020369556083e-24, 1.2797426487740226e-24, 1.1466143237883616e-24], [289.45079254209054, 1.404376346995985e-24, 1.2800386271366236e-24, 1.146366886716969e-24], [290.2533701136521, 1.4043540033354335e-24, 1.2803373548480318e-24, 1.1461223231438837e-24], [291.05817304017927, 1.4043350144688907e-24, 1.2806388414253677e-24, 1.1458806384645767e-24], [291.86520749204647, 1.4043193889483183e-24, 1.2809430964410522e-24, 1.1456418381200133e-24], [292.67447965673824, 1.404307135382807e-24, 1.2812501295228766e-24, 1.1454059275967352e-24], [293.48599573889544, 1.4042982624387289e-24, 1.281559950354138e-24, 1.1451729124269684e-24], [294.2997619603619, 1.404292778839822e-24, 1.2818725686736741e-24, 1.1449427981886915e-24], [295.1157845602345, 1.4042906933672878e-24, 1.2821879942759756e-24, 1.1447155905057223e-24], [295.9340697949084, 1.4042920148599078e-24, 1.2825062370112539e-24, 1.1444912950478e-24], [296.75462393812705, 1.4042967522141798e-24, 1.282827306785521e-24, 1.1442699175306856e-24], [297.57745328102925, 1.4043049143843837e-24, 1.2831512135607365e-24, 1.1440514637162318e-24], [298.40256413219794, 1.4043165103827132e-24, 1.2834779673548139e-24, 1.1438359394124661e-24], [299.2299628177073, 1.4043315492793943e-24, 1.2838075782417374e-24, 1.1436233504736946e-24], [300.05965568117347, 1.4043500402027763e-24, 1.2841400563516738e-24, 1.1434137028005722e-24], [300.89164908380053, 1.4043719923394163e-24, 1.2844754118710024e-24, 1.1432070023401786e-24], [301.7259494044318, 1.4043974149342574e-24, 1.2848136550424681e-24, 1.1430032550861362e-24], [302.5625630395974, 1.4044263172906522e-24, 1.2851547961651896e-24, 1.1428024670786538e-24], [303.40149640356276, 1.4044587087705518e-24, 1.2854988455948267e-24, 1.142604644404655e-24], [304.2427559283793, 1.4044945987945465e-24, 1.2858458137436009e-24, 1.1424097931978258e-24], [305.086348063933, 1.4045339968420032e-24, 1.2861957110804017e-24, 1.1422179196387186e-24], [305.93227927799296, 1.4045769124511687e-24, 1.2865485481308963e-24, 1.1420290299548379e-24], [306.780556056263, 1.4046233552192935e-24, 1.286904335477582e-24, 1.1418431304207235e-24], [307.63118490242914, 1.4046733348027023e-24, 1.2872630837598672e-24, 1.1416602273580327e-24], [308.48417233821135, 1.4047268609169205e-24, 1.287624803674201e-24, 1.1414803271356217e-24], [309.33952490341284, 1.404783943336793e-24, 1.2879895059741086e-24, 1.1413034361696463e-24], [310.19724915596913, 1.4048445918965759e-24, 1.288357201470315e-24, 1.1411295609236351e-24], [311.05735167200066, 1.4049088164900216e-24, 1.2887279010307888e-24, 1.1409587079085674e-24], [311.91983904586056, 1.4049766270705284e-24, 1.289101615580858e-24, 1.1407908836829774e-24], [312.78471789018744, 1.4050480336512158e-24, 1.2894783561033177e-24, 1.1406260948530293e-24], [313.65199483595495, 1.4051230463050417e-24, 1.289858133638432e-24, 1.1404643480725995e-24], [314.52167653252303, 1.4052016751648807e-24, 1.290240959284114e-24, 1.1403056500433563e-24], [315.3937696476878, 1.405283930423683e-24, 1.2906268441959197e-24, 1.1401500075148693e-24], [316.2682808677349, 1.4053698223345177e-24, 1.291015799587233e-24, 1.1399974272846648e-24], [317.1452168974882, 1.40545936121072e-24, 1.2914078367292425e-24, 1.1398479161983284e-24], [318.0245844603631, 1.4055525574259787e-24, 1.29180296695111e-24, 1.1397014811495858e-24], [318.9063902984178, 1.405649421414443e-24, 1.2922012016400303e-24, 1.1395581290803887e-24], [319.7906411724034, 1.4057499636708066e-24, 1.2926025522412726e-24, 1.139417866980987e-24], [320.6773438618184, 1.4058541947504414e-24, 1.293007030258317e-24, 1.1392807018900333e-24], [321.56650516495773, 1.4059621252694796e-24, 1.2934146472529248e-24, 1.139146640894651e-24], [322.4581318989675, 1.4060737659049313e-24, 1.2938254148452082e-24, 1.1390156911305314e-24], [323.352230899896, 1.4061891273947762e-24, 1.2942393447137175e-24, 1.138887859782012e-24], [324.24880902274543, 1.4063082205380486e-24, 1.29465644859553e-24, 1.1387631540821495e-24], [325.14787314152636, 1.4064310561949726e-24, 1.2950767382863307e-24, 1.1386415813128285e-24], [326.0494301493079, 1.4065576452870525e-24, 1.2955002256405167e-24, 1.1385231488048292e-24], [326.9534869582742, 1.406687998797156e-24, 1.2959269225712103e-24, 1.1384078639379104e-24], [327.8600504997729, 1.4068221277696345e-24, 1.2963568410504335e-24, 1.1382957341409031e-24], [328.7691277243723, 1.4069600433104075e-24, 1.2967899931091133e-24, 1.1381867668917816e-24], [329.68072560191166, 1.4071017565870835e-24, 1.297226390837234e-24, 1.1380809697177706e-24], [330.59485112155687, 1.4072472788290467e-24, 1.2976660463838484e-24, 1.1379783501954007e-24], [331.511511291853, 1.4073966213275487e-24, 1.2981089719572095e-24, 1.1378789159506139e-24], [332.43071314077747, 1.4075497954358218e-24, 1.2985551798248242e-24, 1.1377826746588344e-24], [333.3524637157955, 1.4077068125691807e-24, 1.2990046823135723e-24, 1.1376896340450694e-24], [334.2767700839123, 1.4078676842051064e-24, 1.2994574918097306e-24, 1.1375998018839728e-24], [335.2036393317288, 1.408032421883344e-24, 1.2999136207591229e-24, 1.1375131859999398e-24], [336.1330785654958, 1.4082010372060284e-24, 1.3003730816671317e-24, 1.1374297942671964e-24], [337.0650949111667, 1.4083735418377519e-24, 1.3008358870988576e-24, 1.137349634609878e-24], [337.99969551445486, 1.4085499475056709e-24, 1.3013020496791158e-24, 1.1372727150021073e-24], [338.9368875408858, 1.4087302659996373e-24, 1.3017715820925991e-24, 1.1371990434681059e-24], [339.87667817585407, 1.4089145091722105e-24, 1.3022444970838803e-24, 1.1371286280822309e-24], [340.8190746246778, 1.4091026889388604e-24, 1.3027208074575768e-24, 1.1370614769691158e-24], [341.76408411265277, 1.4092948172779976e-24, 1.3032005260783748e-24, 1.1369975983037185e-24], [342.71171388510993, 1.409490906231102e-24, 1.3036836658711326e-24, 1.1369370003114268e-24], [343.6619712074682, 1.4096909679027764e-24, 1.304170239820946e-24, 1.1368796912681236e-24], [344.6148633652941, 1.40989501446091e-24, 1.3046602609732653e-24, 1.1368256795002974e-24], [345.570397664353, 1.4101030581367315e-24, 1.3051537424339501e-24, 1.136774973385121e-24], [346.52858143066896, 1.4103151112249007e-24, 1.3056506973693541e-24, 1.136727581350525e-24], [347.48942201057815, 1.4105311860836454e-24, 1.3061511390064235e-24, 1.1366835118753101e-24], [348.4529267707872, 1.4107512951348262e-24, 1.306655080632773e-24, 1.1366427734892212e-24], [349.419103098429, 1.410975450864056e-24, 1.3071625355967672e-24, 1.1366053747730439e-24], [350.3879584011181, 1.411203665820783e-24, 1.3076735173076118e-24, 1.1365713243586868e-24], [351.35950010701, 1.4114359526183911e-24, 1.3081880392354408e-24, 1.1365406309292779e-24], [352.33373566485517, 1.411672323934314e-24, 1.3087061149113862e-24, 1.1365133032192608e-24], [353.3106725440589, 1.411912792510135e-24, 1.3092277579277152e-24, 1.1364893500144893e-24], [354.2903182347373, 1.412157371151647e-24, 1.3097529819378231e-24, 1.136468780152301e-24], [355.2726802477737, 1.4124060727290238e-24, 1.3102818006564325e-24, 1.1364516025216485e-24], [356.2577661148786, 1.4126589101768792e-24, 1.3108142278596322e-24, 1.1364378260631792e-24], [357.24558338864495, 1.4129158964943568e-24, 1.3113502773849414e-24, 1.1364274597693208e-24], [358.2361396426081, 1.4131770447452904e-24, 1.3118899631314687e-24, 1.136420512684415e-24], [359.2294424713027, 1.413442368058241e-24, 1.3124332990599444e-24, 1.1364169939047898e-24], [360.2254994903217, 1.4137118796266719e-24, 1.3129802991928524e-24, 1.1364169125788864e-24], [361.2243183363729, 1.4139855927089946e-24, 1.3135309776145219e-24, 1.136420277907352e-24], [362.2259066673406, 1.4142635206287593e-24, 1.314085348471255e-24, 1.1364270991431806e-24], [363.2302721623409, 1.4145456767746495e-24, 1.3146434259713516e-24, 1.1364373855917716e-24], [364.2374225217835, 1.4148320746007458e-24, 1.3152052243853089e-24, 1.1364511466111143e-24], [365.24736546742935, 1.4151227276265142e-24, 1.3157707580458828e-24, 1.1364683916118586e-24], [366.260108742449, 1.4154176494370078e-24, 1.3163400413482004e-24, 1.1364891300574655e-24], [367.27566011148406, 1.415716853682957e-24, 1.3169130887498844e-24, 1.1365133714643282e-24], [368.2940273607055, 1.4160203540808988e-24, 1.3174899147711664e-24, 1.1365411254019078e-24], [369.31521829787255, 1.4163281644133274e-24, 1.3180705339950363e-24, 1.136572401492877e-24], [370.33924075239474, 1.416640298528806e-24, 1.3186549610673469e-24, 1.1366072094132624e-24], [371.3661025753894, 1.4169567703421328e-24, 1.3192432106969394e-24, 1.1366455588926044e-24], [372.3958116397441, 1.4172775938344693e-24, 1.3198352976558212e-24, 1.1366874597141095e-24], [373.42837584017616, 1.4176027830535116e-24, 1.3204312367792885e-24, 1.1367329217148274e-24], [374.4638030932918, 1.4179323521136089e-24, 1.3210310429660816e-24, 1.136781954785813e-24], [375.5021013376496, 1.4182663151959974e-24, 1.3216347311785615e-24, 1.1368345688723402e-24], [376.5432785338186, 1.4186046865489156e-24, 1.3222423164428687e-24, 1.1368907739740744e-24], [377.5873426644413, 1.4189474804878362e-24, 1.3228538138491229e-24, 1.1369505801453079e-24], [378.634301734294, 1.4192947113955948e-24, 1.3234692385515814e-24, 1.1370139974951455e-24], [379.68416377034896, 1.419646393722701e-24, 1.3240886057689083e-24, 1.137081036187806e-24], [380.73693682183364, 1.4200025419874552e-24, 1.3247119307843362e-24, 1.137151706442815e-24], [381.79262896029593, 1.420363170776246e-24, 1.3253392289459335e-24, 1.137226018535317e-24], [382.8512482796627, 1.4207282947437615e-24, 1.3259705156668154e-24, 1.1373039827963575e-24], [383.91280289630424, 1.4210979286132708e-24, 1.3266058064254786e-24, 1.1373856096131963e-24], [384.9773009490958, 1.4214720871769669e-24, 1.3272451167660734e-24, 1.1374709094296848e-24], [386.04475059947856, 1.421850785296157e-24, 1.3278884622986642e-24, 1.1375598927465805e-24], [387.11516003152485, 1.4222340379016956e-24, 1.3285358586996504e-24, 1.1376525701219955e-24], [388.1885374519982, 1.4226218599943317e-24, 1.3291873217121058e-24, 1.1377489521718333e-24], [389.2648910904183, 1.423014266645078e-24, 1.329842867146197e-24, 1.1378490495702556e-24], [390.3442291991237, 1.4234112729956382e-24, 1.3305025108795959e-24, 1.1379528730502007e-24], [391.4265600533332, 1.4238128942588907e-24, 1.3311662688579796e-24, 1.1380604334039728e-24], [392.51189195121265, 1.4242191457194075e-24, 1.331834157095605e-24, 1.1381717414838793e-24], [393.60023321393464, 1.4246300427340041e-24, 1.3325061916758457e-24, 1.1382868082029096e-24], [394.6915921857471, 1.4250456007323682e-24, 1.3331823887518526e-24, 1.138405644535519e-24], [395.78597723403175, 1.4254658352177222e-24, 1.3338627645472638e-24, 1.138528261518459e-24], [396.8833967493732, 1.4258907617676212e-24, 1.3345473353570257e-24, 1.1386546702517507e-24], [397.98385914561936, 1.4263203960347717e-24, 1.335236117548262e-24, 1.1387848818997107e-24], [399.0873728599491, 1.4267547537479417e-24, 1.335929127561218e-24, 1.1389189076921204e-24], [400.19394635293537, 1.4271938507130554e-24, 1.3366263819104078e-24, 1.1390567589255454e-24], [401.3035881086092, 1.4276377028143183e-24, 1.3373278971858168e-24, 1.1391984469647823e-24], [402.41630663452736, 1.4280863260155466e-24, 1.3380336900542735e-24, 1.1393439832445095e-24], [403.5321104618344, 1.4285397363616283e-24, 1.3387437772610133e-24, 1.139493379271133e-24], [404.65100814533093, 1.428997949980169e-24, 1.3394581756313939e-24, 1.139646646624855e-24], [405.773008263538, 1.4294609830833795e-24, 1.3401769020729006e-24, 1.1398037969620428e-24], [406.89811941876144, 1.4299288519702007e-24, 1.3408999735773693e-24, 1.1399648420179025e-24], [408.0263502371612, 1.4304015730286825e-24, 1.341627407223509e-24, 1.1401297936094971e-24], [409.1577093688134, 1.4308791627387688e-24, 1.3423592201798422e-24, 1.1402986636392226e-24], [410.29220548778045, 1.4313616376753821e-24, 1.3430954297079341e-24, 1.1404714640987485e-24], [411.42984729217545, 1.43184901451207e-24, 1.343836053166289e-24, 1.1406482070735718e-24], [412.5706435042283, 1.4323413100250906e-24, 1.344581108014625e-24, 1.1408289047482061e-24], [413.71460287035507, 1.4328385410981898e-24, 1.345330611818951e-24, 1.1410135694122175e-24], [414.86173416122153, 1.4333407247281292e-24, 1.346084582257357e-24, 1.141202213467173e-24], [416.01204617181537, 1.4338478780310626e-24, 1.3468430371268014e-24, 1.1413948494347312e-24], [417.16554772150755, 1.4343600182499703e-24, 1.3476059943509354e-24, 1.1415914899660303e-24], [418.3222476541254, 1.4348771627633703e-24, 1.3483734719893353e-24, 1.1417921478527095e-24], [419.48215483801613, 1.4353993290955756e-24, 1.3491454882482801e-24, 1.1419968360398313e-24], [420.64527816611815, 1.4359265349286608e-24, 1.3499220614934676e-24, 1.1422055676410685e-24], [421.81162655602776, 1.436458798116728e-24, 1.3507032102650318e-24, 1.1424183559567143e-24], [422.98120895006645, 1.4369961367028258e-24, 1.3514889532954461e-24, 1.1426352144950643e-24], [424.1540343153523, 1.4375385689390686e-24, 1.3522793095307702e-24, 1.142856156997905e-24], [425.3301116438653, 1.4380861133108937e-24, 1.3530742981562837e-24, 1.143081197471133e-24], [426.5094499525193, 1.438638788566116e-24, 1.3538739386271885e-24, 1.143310350221566e-24], [427.69205828323, 1.4391966137503004e-24, 1.3546782507059362e-24, 1.1435436299016234e-24], [428.87794570298274, 1.4397596082496914e-24, 1.355487254507654e-24, 1.1437810515636675e-24], [430.0671213039053, 1.4403277918438653e-24, 1.356300970555722e-24, 1.1440226307266068e-24], [431.2595942033339, 1.440901184770686e-24, 1.3571194198503676e-24, 1.1442683834580271e-24], [432.45537354388625, 1.4414798078068039e-24, 1.3579426239537802e-24, 1.1445183264760618e-24], [433.65446849353003, 1.442063682368286e-24, 1.358770605096339e-24, 1.1447724772767003e-24], [434.85688824565386, 1.4426528306373839e-24, 1.359603386310554e-24, 1.145030854294109e-24], [436.062642019136, 1.4432472757232497e-24, 1.3604409916008072e-24, 1.145293477104024e-24], [437.27173905841835, 1.443847041867826e-24, 1.3612834461608504e-24, 1.145560366684182e-24], [438.48418863357364, 1.4444521547116337e-24, 1.3621307766545057e-24, 1.1458315457506532e-24], [439.70000004037956, 1.4450626416407577e-24, 1.3629830115821461e-24, 1.1461070391967862e-24], [440.9191826003888, 1.4456785322441947e-24, 1.3638401817637435e-24, 1.1463868746719437e-24], [442.1417456609991, 1.4462998589242326e-24, 1.3647023209834583e-24, 1.1466710833537457e-24], [443.3676985955277, 1.4469266577208995e-24, 1.3655694668603641e-24, 1.1469597009913397e-24], [444.5970508032818, 1.4475589694415184e-24, 1.3664416620412621e-24, 1.1472527693346257e-24], [445.8298117096291, 1.4481968412313724e-24, 1.3673189558593429e-24, 1.1475503381218005e-24], [447.06599076607364, 1.4488403287946387e-24, 1.3682014066794346e-24, 1.1478524678897696e-24], [448.3055974503242, 1.449489499593055e-24, 1.3690890852756282e-24, 1.148159234021939e-24], [449.5486412663705, 1.4501444375480989e-24, 1.3699820797963123e-24, 1.1484707326987255e-24], [450.79513174455445, 1.4508052501135666e-24, 1.3708805032319857e-24, 1.148787089848073e-24], [452.0450784416419, 1.451472079192269e-24, 1.3717845049415485e-24, 1.149108474961085e-24], [453.2984909408992, 1.452145118488663e-24, 1.3726942889733604e-24, 1.1494351230531706e-24], [454.555378852163, 1.4528246420372367e-24, 1.3736101441848122e-24, 1.149767370769937e-24], [455.81575181191704, 1.4535110529737403e-24, 1.3745324957320561e-24, 1.150105718113941e-24], [457.0796194833643, 1.4542049708269573e-24, 1.375461997224378e-24, 1.1504509389269447e-24], [458.3469915565019, 1.4549073965353877e-24, 1.3763997049272688e-24, 1.150804289749777e-24], [459.6178777481936, 1.4556200457967792e-24, 1.3773474296569825e-24, 1.1511679317456858e-24], [460.8922878022477, 1.4563460803040509e-24, 1.3783085086688376e-24, 1.15154585624075e-24], [462.1702314894878, 1.4570918900874889e-24, 1.3792896870142678e-24, 1.1519461406719576e-24], [463.4517186078309, 1.4578720925087057e-24, 1.3803063940202862e-24, 1.1523872758148976e-24], [464.73675898236144, 1.4587266253777344e-24, 1.3814007843075119e-24, 1.1529197994346097e-24], [466.025362465405, 1.4598002132889938e-24, 1.3827256051327754e-24, 1.153726857536758e-24], [467.31753893660726, 1.4619974147382977e-24, 1.3852354236245662e-24, 1.1559527262357047e-24], [468.61329830300576, 1.5072013988255413e-24, 1.4330466583388866e-24, 1.2120893161372698e-24], [469.91265049910936, 1.4676244695431448e-24, 1.3915505072465786e-24, 1.161952284098947e-24], [471.21560548697295, 1.4631935994646487e-24, 1.387065224898485e-24, 1.1557884556541435e-24], [472.5221732562723, 1.4632778864919948e-24, 1.387342221327183e-24, 1.1553295375291023e-24], [473.83236382438423, 1.4638304140544447e-24, 1.3881124693207782e-24, 1.1554605728496622e-24], [475.14618723645873, 1.4645064102695553e-24, 1.3890122026593463e-24, 1.155745167182989e-24], [476.4636535655021, 1.4652314479337026e-24, 1.3899628940167767e-24, 1.1560890783300439e-24], [477.78477291244724, 1.4659818516763957e-24, 1.390939564240261e-24, 1.1564623162168983e-24], [479.10955540623695, 1.4667482096434547e-24, 1.3919322832282408e-24, 1.1568529510583745e-24], [480.4380112038972, 1.4675261895124388e-24, 1.3929364799782893e-24, 1.1572554852848486e-24], [481.7701504906184, 1.4683135790009373e-24, 1.3939498201262053e-24, 1.1576671061052069e-24], [483.10598347983205, 1.4691091583042072e-24, 1.3949710163264911e-24, 1.1580862576403818e-24], [484.4455204132876, 1.4699122146299025e-24, 1.3959993160843516e-24, 1.1585120261297963e-24], [485.7887715611344, 1.4707223126565716e-24, 1.397034259596115e-24, 1.1589438492258174e-24], [487.13574722199644, 1.4715391776261636e-24, 1.398075556419297e-24, 1.1593813679284535e-24], [488.4864577230548, 1.472362632086738e-24, 1.3991230187423462e-24, 1.159824346465682e-24], [489.8409134201254, 1.4731925598982697e-24, 1.4001765234156384e-24, 1.1602726267026009e-24], [491.1991246977368, 1.4740288848616441e-24, 1.4012359894086807e-24, 1.1607261010715131e-24], [492.5611019692135, 1.4748715575587143e-24, 1.4023013639291006e-24, 1.1611846959022084e-24], [493.9268556767514, 1.4757205469887237e-24, 1.4033726136011054e-24, 1.1616483608272821e-24], [495.2963962915019, 1.4765758351035606e-24, 1.4044497187017117e-24, 1.162117061859482e-24], [496.66973431365005, 1.4774374131489462e-24, 1.4055326693018713e-24, 1.1625907767565577e-24], [498.04688027249426, 1.4783052791608343e-24, 1.4066214626260645e-24, 1.163069491849505e-24], [499.4278447265299, 1.4791794362185463e-24, 1.4077161012100708e-24, 1.1635531998294572e-24], [500.8126382635261, 1.4800598912044201e-24, 1.4088165915929195e-24, 1.1640418981761954e-24], [502.2012715006128, 1.4809466539089237e-24, 1.4099229433732442e-24, 1.1645355880243276e-24], [503.5937550843552, 1.481839736375572e-24, 1.4110351685184955e-24, 1.16503427333332e-24], [504.99009969084153, 1.482739152415157e-24, 1.4121532808527644e-24, 1.1655379602719342e-24], [506.39031602575994, 1.4836449172408555e-24, 1.4132772956720227e-24, 1.1660466567559632e-24], [507.79441482448465, 1.4845570471913592e-24, 1.4144072294522268e-24, 1.1665603720973393e-24], [509.20240685215634, 1.485475559518388e-24, 1.4155430996253128e-24, 1.1670791167349128e-24], [510.6143029037634, 1.4864004722221023e-24, 1.4166849244056752e-24, 1.1676029020259188e-24], [512.0301138042282, 1.4873318039225439e-24, 1.4178327226546772e-24, 1.1681317400830487e-24], [513.4498504084859, 1.4882695737582951e-24, 1.4189865137738058e-24, 1.1686656436460428e-24], [514.8735236015713, 1.4892138013060084e-24, 1.4201463176198907e-24, 1.1692046259797257e-24], [516.3011442987009, 1.4901645065159939e-24, 1.4213121544372032e-24, 1.1697487007924178e-24], [517.7327234453548, 1.491121709660414e-24, 1.4224840448028985e-24, 1.1702978821702618e-24], [519.1682720173646, 1.492085431291218e-24, 1.4236620095826714e-24, 1.1708521845239537e-24], [520.607801020993, 1.4930556922058658e-24, 1.4248460698946673e-24, 1.1714116225453043e-24], [522.0513214930216, 1.4940325134191961e-24, 1.4260362470798534e-24, 1.1719762111716174e-24], [523.4988445008346, 1.495015916140238e-24, 1.4272325626776014e-24, 1.1725459655563053e-24], [524.9503811425035, 1.4960059217529341e-24, 1.4284350384053924e-24, 1.1731209010445077e-24], [526.405942546871, 1.4970025518001624e-24, 1.4296436961420224e-24, 1.1737010331528249e-24], [527.8655398736367, 1.4980058279702892e-24, 1.430858557913433e-24, 1.1742863775523171e-24], [529.3291843134474, 1.499015772085892e-24, 1.4320796458808608e-24, 1.1748769500542187e-24], [530.796887087974, 1.5000324060941813e-24, 1.433306982330781e-24, 1.1754727665978674e-24], [532.2686594500059, 1.5010557520589925e-24, 1.4345405896664769e-24, 1.176073843240506e-24], [533.7445126835313, 1.5020858321538595e-24, 1.4357804904007958e-24, 1.1766801961485613e-24], [535.2244581038278, 1.5031226686560955e-24, 1.437026707149904e-24, 1.1772918415901955e-24], [536.7085070575478, 1.5041662839417953e-24, 1.438279262628078e-24, 1.1779087959289455e-24], [538.1966709228032, 1.5052167004814288e-24, 1.4395381796430886e-24, 1.1785310756182138e-24], [539.6889611092572, 1.5062739408361526e-24, 1.44080348109231e-24, 1.1791586971965507e-24], [541.1853890582071, 1.5073380276544809e-24, 1.4420751899592796e-24, 1.1797916772835257e-24], [542.6859662426759, 1.5084089836694508e-24, 1.443353329310728e-24, 1.1804300325761481e-24], [544.1907041674971, 1.5094868316961869e-24, 1.4446379222940291e-24, 1.18107377984577e-24], [545.6996143694057, 1.5105715946297167e-24, 1.4459289921349394e-24, 1.1817229359353607e-24], [547.2127084171232, 1.5116632954429976e-24, 1.4472265621355553e-24, 1.1823775177570745e-24], [548.7299979114512, 1.512761957185322e-24, 1.448530655672653e-24, 1.1830375422901948e-24], [550.2514944853564, 1.5138676029807761e-24, 1.4498412961960923e-24, 1.183703026579241e-24], [551.7772098040593, 1.5149802560269176e-24, 1.4511585072274673e-24, 1.1843739877323185e-24], [553.307155565128, 1.5160999395936366e-24, 1.4524823123589235e-24, 1.1850504429196816e-24], [554.8413434985625, 1.517226677022037e-24, 1.4538127352520074e-24, 1.1857324093723855e-24], [556.3797853668896, 1.5183604917235955e-24, 1.4551497996368346e-24, 1.1864199043811751e-24], [557.9224929652486, 1.5195014071791782e-24, 1.4564935293110688e-24, 1.1871129452953768e-24], [559.4694781214861, 1.5206494469384012e-24, 1.457843948139268e-24, 1.1878115495220043e-24], [561.0207526962416, 1.521804634618898e-24, 1.4592010800521839e-24, 1.1885157345249104e-24], [562.5763285830457, 1.5229669939056357e-24, 1.4605649490460104e-24, 1.1892255178239756e-24], [564.1362177084021, 1.5241365485504639e-24, 1.4619355791820126e-24, 1.1899409169944815e-24], [565.7004320318874, 1.525313322371443e-24, 1.4633129945858006e-24, 1.1906619496664046e-24], [567.2689835462367, 1.5264973392525208e-24, 1.4646972194470457e-24, 1.191388633523902e-24], [568.8418842774403, 1.5276886231429686e-24, 1.466088278018904e-24, 1.192120986304742e-24], [570.4191462848314, 1.5288871980570554e-24, 1.4674861946177144e-24, 1.1928590257998386e-24], [572.0007816611834, 1.5300930880736141e-24, 1.4688909936225856e-24, 1.193602769852787e-24], [573.5868025327974, 1.5313063173357639e-24, 1.470302699475117e-24, 1.194352236359485e-24], [575.1772210596002, 1.5325269100504977e-24, 1.4717213366790077e-24, 1.1951074432676967e-24], [576.7720494352326, 1.5337548904884334e-24, 1.4731469297998297e-24, 1.1958684085767338e-24], [578.3712998871483, 1.5349902829835721e-24, 1.474579503464808e-24, 1.1966351503371417e-24], [579.9749846767038, 1.536233111932912e-24, 1.4760190823624365e-24, 1.1974076866503254e-24], [581.5831160992516, 1.5374834017962928e-24, 1.477465691242368e-24, 1.1981860356683142e-24], [583.1957064842389, 1.5387411770961278e-24, 1.4789193549151505e-24, 1.1989702155934653e-24], [584.8127681952972, 1.5400064624171965e-24, 1.4803800982520542e-24, 1.1997602446782153e-24], [586.4343136303413, 1.541279282406374e-24, 1.4818479461848095e-24, 1.2005561412248126e-24], [588.0603552216609, 1.5425596617724711e-24, 1.4833229237054905e-24, 1.2013579235851057e-24], [589.6909054360189, 1.5438476252860553e-24, 1.4848050558663424e-24, 1.2021656101603222e-24], [591.3259767747438, 1.5451431977791903e-24, 1.4862943677795565e-24, 1.2029792194008352e-24], [592.965581773829, 1.5464464041453432e-24, 1.487790884617208e-24, 1.2037987698059934e-24], [594.6097330040262, 1.5477572693391554e-24, 1.489294631611041e-24, 1.2046242799239071e-24], [596.2584430709438, 1.5490758183762399e-24, 1.4908056340523007e-24, 1.205455768351241e-24], [597.9117246151407, 1.5504020763331332e-24, 1.4923239172917081e-24, 1.2062932537330933e-24], [599.569590312227, 1.5517360683470309e-24, 1.4938495067392293e-24, 1.2071367547627652e-24], [601.2320528729574, 1.5530778196156489e-24, 1.4953824278639658e-24, 1.2079862901816056e-24], [602.8991250433324, 1.554427355397132e-24, 1.496922706194078e-24, 1.208841878778873e-24], [604.5708196046922, 1.5557847010098052e-24, 1.4984703673165773e-24, 1.2097035393915286e-24], [606.2471493738185, 1.557149881832206e-24, 1.5000254368773967e-24, 1.2105712909041636e-24], [607.9281272030281, 1.5585229233027338e-24, 1.5015879405810406e-24, 1.2114451522487474e-24], [609.6137659802778, 1.559903850919641e-24, 1.5031579041906375e-24, 1.2123251424045427e-24], [611.3040786292577, 1.5612926902408991e-24, 1.5047353535278115e-24, 1.2132112803979619e-24], [612.9990781094909, 1.562689466884012e-24, 1.5063203144725473e-24, 1.214103585302397e-24], [614.698777416437, 1.5640942065259302e-24, 1.5079128129631138e-24, 1.2150020762381026e-24], [616.403189581586, 1.5655069349028735e-24, 1.5095128749959252e-24, 1.2159067723720355e-24], [618.1123276725638, 1.5669276778102852e-24, 1.511120526625544e-24, 1.2168176929177597e-24], [619.826204793227, 1.5683564611026584e-24, 1.5127357939645067e-24, 1.2177348571352874e-24], [621.5448340837688, 1.569793310693395e-24, 1.5143587031832345e-24, 1.2186582843309434e-24], [623.2682287208142, 1.5712382525547511e-24, 1.5159892805100136e-24, 1.2195879938572656e-24], [624.9964019175254, 1.5726913127176235e-24, 1.5176275522307946e-24, 1.2205240051128336e-24], [626.7293669237033, 1.5741525172715397e-24, 1.5192735446892322e-24, 1.2214663375421948e-24], [628.4671370258833, 1.5756218923644893e-24, 1.520927284286533e-24, 1.2224150106357153e-24], [630.2097255474447, 1.5770994642027652e-24, 1.5225887974813424e-24, 1.2233700439294365e-24], [631.9571458487071, 1.578585259050977e-24, 1.5242581107897893e-24, 1.2243314570050137e-24], [633.7094113270354, 1.580079303231787e-24, 1.525935250785238e-24, 1.2252992694895225e-24], [635.4665354169437, 1.5815816231258922e-24, 1.5276202440983248e-24, 1.2262735010553917e-24], [637.2285315901945, 1.5830922451719672e-24, 1.5293131174169191e-24, 1.2272541714202975e-24], [638.9954133559067, 1.5846111958663542e-24, 1.5310138974858419e-24, 1.2282413003469623e-24], [640.7671942606546, 1.5861385017632183e-24, 1.532722611107073e-24, 1.2292349076431602e-24], [642.5438878885743, 1.5876741894742375e-24, 1.5344392851394587e-24, 1.2302350131615015e-24], [644.3255078614702, 1.5892182856685863e-24, 1.5361639464987295e-24, 1.2312416367993696e-24], [646.1120678389125, 1.5907708170728114e-24, 1.537896622157407e-24, 1.2322547984987879e-24], [647.9035815183493, 1.5923318104707658e-24, 1.5396373391447829e-24, 1.233274518246335e-24], [649.700062635206, 1.5939012927034358e-24, 1.5413861245467672e-24, 1.234300816072995e-24], [651.5015249629953, 1.5954792906689036e-24, 1.543143005505882e-24, 1.2353337120540754e-24], [653.3079823134185, 1.5970658313222083e-24, 1.544908009221171e-24, 1.2363732263090852e-24], [655.1194485364756, 1.5986609416752897e-24, 1.5466811629481548e-24, 1.237419379001639e-24], [656.9359375205668, 1.6002646487968278e-24, 1.5484624939987274e-24, 1.238472190339325e-24], [658.7574631926042, 1.6018769798121906e-24, 1.5502520297411118e-24, 1.2395316805736176e-24], [660.5840395181125, 1.6034979619033357e-24, 1.5520497975998215e-24, 1.2405978699997685e-24], [662.4156805013434, 1.6051276223087198e-24, 1.5538558250555741e-24, 1.2416707789567009e-24], [664.2524001853774, 1.606765988323156e-24, 1.5556701396452117e-24, 1.2427504278268877e-24], [666.0942126522311, 1.608413087297784e-24, 1.55749276896168e-24, 1.2438368370362709e-24], [667.9411320229706, 1.6100689466399453e-24, 1.5593237406539545e-24, 1.2449300270541447e-24], [669.7931724578128, 1.6117335938130727e-24, 1.5611630824269468e-24, 1.2460300183930464e-24], [671.650348156241, 1.6134070563366754e-24, 1.5630108220415403e-24, 1.2471368316086844e-24], [673.5126733571068, 1.6150893617861398e-24, 1.5648669873144045e-24, 1.2482504872997869e-24], [675.380162338746, 1.6167805377927398e-24, 1.5667316061180617e-24, 1.2493710061080501e-24], [677.2528294190812, 1.6184806120434945e-24, 1.5686047063807606e-24, 1.2504984087180078e-24], [679.1306889557399, 1.6201896122810798e-24, 1.5704863160864197e-24, 1.2516327158569364e-24], [681.013755346155, 1.6219075663038143e-24, 1.5723764632746613e-24, 1.2527739482947837e-24], [682.9020430276838, 1.6236345019654553e-24, 1.5742751760406327e-24, 1.2539221268440193e-24], [684.7955664777116, 1.6253704471752226e-24, 1.576182482535052e-24, 1.2550772723595869e-24], [686.6943402137695, 1.6271154298976653e-24, 1.5780984109641336e-24, 1.2562394057387828e-24], [688.5983787936386, 1.6288694781525826e-24, 1.580022989589516e-24, 1.2574085479211619e-24], [690.5076968154688, 1.6306326200149818e-24, 1.581956246728266e-24, 1.2585847198884605e-24], [692.422308917884, 1.6324048836149366e-24, 1.5838982107527656e-24, 1.2597679426644758e-24], [694.3422297801003, 1.6341862971375612e-24, 1.5858489100907243e-24, 1.2609582373149919e-24], [696.2674741220321, 1.6359768888229102e-24, 1.5878083732251108e-24, 1.2621556249476827e-24], [698.1980567044133, 1.6377766869659186e-24, 1.5897766286941137e-24, 1.263360126712022e-24], [700.1339923289031, 1.6395857199163233e-24, 1.5917537050911163e-24, 1.264571763799195e-24], [702.0752958382011, 1.6414040160785785e-24, 1.5937396310646198e-24, 1.2657905574420027e-24], [704.0219821161648, 1.6432316039117962e-24, 1.5957344353182438e-24, 1.267016528914779e-24], [705.9740660879183, 1.6450685119296534e-24, 1.59773814661064e-24, 1.2682496995332909e-24], [707.9315627199724, 1.6469147687004413e-24, 1.5997507937555965e-24, 1.2694900906547088e-24], [709.8944870203334, 1.648770402846741e-24, 1.6017724056217253e-24, 1.270737723677404e-24], [711.8628540386239, 1.6506354430456645e-24, 1.6038030111327501e-24, 1.2719926200409998e-24], [713.8366788661933, 1.6525099180285368e-24, 1.60584263926721e-24, 1.2732548012261811e-24], [715.8159766362373, 1.6543938565810205e-24, 1.6078913190586253e-24, 1.274524288754688e-24], [717.8007625239117, 1.6562872875429573e-24, 1.6099490795953678e-24, 1.2758011041891964e-24], [719.7910517464501, 1.6581902398082695e-24, 1.6120159500205844e-24, 1.2770852691332224e-24], [721.7868595632776, 1.6601027423250226e-24, 1.6140919595323087e-24, 1.2783768052310919e-24], [723.7882012761331, 1.6620248240952465e-24, 1.6161771373833018e-24, 1.2796757341678162e-24], [725.7950922291805, 1.66395651417499e-24, 1.6182715128811326e-24, 1.2809820776690553e-24], [727.8075478091324, 1.66589784167416e-24, 1.6203751153880691e-24, 1.2822958575010047e-24], [729.825583445362, 1.6678488357565647e-24, 1.6224879743211293e-24, 1.283617095470358e-24], [731.8492146100275, 1.669809525639813e-24, 1.6246101191520274e-24, 1.2849458134242142e-24], [733.8784568181841, 1.6717799405952303e-24, 1.6267415794071121e-24, 1.2862820332499876e-24], [735.9133256279105, 1.673760109947891e-24, 1.628882384667433e-24, 1.2876257768753807e-24], [737.9538366404225, 1.6757500630765045e-24, 1.6310325645686537e-24, 1.288977066268279e-24], [740.0000055001927, 1.677749829413471e-24, 1.6331921488011482e-24, 1.2903359234367236e-24], [742.051847895075, 1.679759438444659e-24, 1.6353611671097875e-24, 1.291702370428781e-24], [744.1093795564188, 1.68177891970955e-24, 1.637539649294141e-24, 1.2930764293325421e-24], [746.172616259196, 1.683808302801139e-24, 1.6397276252083874e-24, 1.294458122276048e-24], [748.2415738221157, 1.6858476173657794e-24, 1.6419251247611924e-24, 1.2958474714271637e-24], [750.3162681077515, 1.687896893103372e-24, 1.644132177915935e-24, 1.297244498993628e-24], [752.3967150226567, 1.689956159767109e-24, 1.64634881469047e-24, 1.2986492272228892e-24], [754.4829305174917, 1.6920254471636393e-24, 1.6485750651573367e-24, 1.300061678402137e-24], [756.5749305871462, 1.6941047851528504e-24, 1.6508109594435524e-24, 1.3014818748581677e-24], [758.6727312708564, 1.69619420364801e-24, 1.6530565277307924e-24, 1.302909838957402e-24], [760.7763486523318, 1.6982937326156393e-24, 1.655311800255292e-24, 1.3043455931057862e-24], [762.8857988598807, 1.7004034020754953e-24, 1.6575768073078602e-24, 1.305789159748756e-24], [765.0010980665281, 1.702523242100574e-24, 1.6598515792339026e-24, 1.307240561371184e-24], [767.1222624901465, 1.704653282817119e-24, 1.6621361464334575e-24, 1.3086998204973594e-24], [769.2493083935728, 1.7067935544045718e-24, 1.664430539361189e-24, 1.310166959690924e-24], [771.3822520847407, 1.708944087095557e-24, 1.6667347885263806e-24, 1.3116420015548291e-24], [773.5211099167981, 1.7111049111758407e-24, 1.669048924492932e-24, 1.3131249687312875e-24], [775.6658982882382, 1.7132760569844076e-24, 1.6713729778794626e-24, 1.314615883901772e-24], [777.8166336430257, 1.7154575549133617e-24, 1.6737069793592367e-24, 1.316114769786941e-24], [779.9733324707156, 1.71764943540803e-24, 1.6760509596602916e-24, 1.3176216491466556e-24], [782.1360113065883, 1.71985172896683e-24, 1.678404949565342e-24, 1.3191365447798898e-24], [784.3046867317697, 1.7220644661413687e-24, 1.6807689799118843e-24, 1.3206594795247497e-24], [786.4793753733642, 1.724287677536457e-24, 1.6831430815922674e-24, 1.322190476258457e-24], [788.6600939045758, 1.7265213938099867e-24, 1.6855272855535672e-24, 1.3237295578972671e-24], [790.8468590448429, 1.7287656456731767e-24, 1.6879216227978728e-24, 1.32527674739656e-24], [793.0396875599591, 1.7310204638903303e-24, 1.6903261243820695e-24, 1.3268320677507105e-24], [795.2385962622079, 1.7332858792790735e-24, 1.692740821418099e-24, 1.3283955419931873e-24], [797.443602010491, 1.7355619227101836e-24, 1.695165745072805e-24, 1.329967193196453e-24], [799.6547217104511, 1.7378486251078376e-24, 1.697600926568229e-24, 1.3315470444720639e-24], [801.8719723146105, 1.7401460174494488e-24, 1.7000463971814466e-24, 1.3331351189706008e-24], [804.0953708224932, 1.7424541307658162e-24, 1.7025021882447524e-24, 1.33473143988172e-24], [806.3249342807624, 1.7447729961411384e-24, 1.7049683311457038e-24, 1.3363360304341685e-24], [808.5606797833437, 1.747102644713086e-24, 1.7074448573272075e-24, 1.3379489138958144e-24], [810.8026244715643, 1.7494431076728056e-24, 1.7099317982875504e-24, 1.3395701135736662e-24], [813.0507855342759, 1.7517944162651133e-24, 1.7124291855806208e-24, 1.3411996528139616e-24], [815.3051802079947, 1.754156601788371e-24, 1.7149370508158e-24, 1.3428375550021594e-24], [817.5658257770253, 1.756529695594775e-24, 1.717455425658284e-24, 1.3444838435630723e-24], [819.8327395736032, 1.7589137290903404e-24, 1.719984341829085e-24, 1.3461385419609137e-24], [822.1059389780166, 1.7613087337350286e-24, 1.7225238311051833e-24, 1.3478016736994003e-24], [824.385441418749, 1.7637147410429314e-24, 1.7250739253197316e-24, 1.349473262321895e-24], [826.6712643726052, 1.766131782582292e-24, 1.727634656362113e-24, 1.3511533314114872e-24], [828.9634253648528, 1.7685598899758023e-24, 1.730206056178237e-24, 1.352841904591225e-24], [831.2619419693494, 1.7709990949006808e-24, 1.7327881567706637e-24, 1.354539005524245e-24], [833.5668318086839, 1.7734494290889925e-24, 1.7353809901989353e-24, 1.3562446579140566e-24], [835.8781125543051, 1.775910924327711e-24, 1.737984588579663e-24, 1.35795888550472e-24], [838.1958019266635, 1.778383612459258e-24, 1.740598984087093e-24, 1.359681712081276e-24], [840.5199176953407, 1.7808675253815774e-24, 1.7432242089532022e-24, 1.3614131614700027e-24], [842.8504776791935, 1.7833626950486833e-24, 1.7458602954682777e-24, 1.363153257538928e-24], [845.1874997464838, 1.785869153470941e-24, 1.7485072759812097e-24, 1.3649020241982852e-24], [847.5310018150162, 1.788386932715791e-24, 1.751165182900261e-24, 1.366659485401206e-24], [849.8810018522806, 1.790916064908127e-24, 1.7538340486934545e-24, 1.3684256651443765e-24], [852.2375178755827, 1.7934565822312204e-24, 1.7565139058895415e-24, 1.370200587469005e-24], [854.6005679521896, 1.7960085169274616e-24, 1.759204787078772e-24, 1.3719842764618353e-24], [856.970170199461, 1.798571901299428e-24, 1.7619067249139894e-24, 1.3737767562564791e-24], [859.3463427849952, 1.80114676771117e-24, 1.7646197521119616e-24, 1.3755780510350003e-24], [861.7291039267614, 1.803733148589736e-24, 1.76734390145495e-24, 1.377388185029878e-24], [864.1184718932453, 1.8063310764270205e-24, 1.7700792057926056e-24, 1.3792071825263923e-24], [866.5144650035863, 1.808940583782231e-24, 1.7728256980444926e-24, 1.3810350678656951e-24], [868.9171016277189, 1.811561703284643e-24, 1.7755834112029025e-24, 1.3828718654485066e-24], [871.3264001865103, 1.8141944676374327e-24, 1.7783523783367815e-24, 1.3847175997399375e-24], [873.7423791519083, 1.816838909622264e-24, 1.7811326325964003e-24, 1.3865722952755508e-24], [876.1650570470749, 1.819495062105349e-24, 1.783924207219546e-24, 1.3884359766691733e-24], [878.5944524465364, 1.8221629580452005e-24, 1.786727135539416e-24, 1.3903086686230584e-24], [881.0305839763178, 1.8248426305028703e-24, 1.789541450995061e-24, 1.3921903959412225e-24], [883.4734703140937, 1.8275341126554767e-24, 1.7923671871451597e-24, 1.3940811835471509e-24], [885.9231301893224, 1.8302374378143814e-24, 1.7952043776865288e-24, 1.3959810565076008e-24], [888.3795823834002, 1.8329526394499206e-24, 1.798053056479311e-24, 1.3978900400650273e-24], [890.842845729797, 1.835679751225713e-24, 1.8009132575819006e-24, 1.39980815968248e-24], [893.3129391142014, 1.838418807046546e-24, 1.80378501529969e-24, 1.401735441106583e-24], [895.7898814746712, 1.841169841127209e-24, 1.8066683642551308e-24, 1.4036719104577652e-24], [898.2736918017715, 1.843932888092763e-24, 1.809563339489791e-24, 1.4056175943618814e-24], [900.7643891387268, 1.8467079831279796e-24, 1.8124699766164534e-24, 1.4075725201465472e-24], [903.2619925815607, 1.8494951622061027e-24, 1.8153883120519246e-24, 1.4095367161415065e-24], [905.7665212792489, 1.8522944624484888e-24, 1.8183183833830403e-24, 1.4115102121510907e-24], [908.277994433859, 1.8551059227090535e-24, 1.821260229961434e-24, 1.4134930402219063e-24], [910.7964313007042, 1.8579295845595362e-24, 1.8242138939061928e-24, 1.4154852359372323e-24], [913.3218511884869, 1.8607654940247495e-24, 1.8271794218696784e-24, 1.4174868406967516e-24], [915.8542734594491, 1.8636137047998965e-24, 1.8301568683115015e-24, 1.4194979059435957e-24], [918.3937175295167, 1.8664742845985653e-24, 1.833146301958173e-24, 1.42151850150473e-24], [920.9402028684551, 1.8693473286733315e-24, 1.8361478195612996e-24, 1.4235487333556131e-24], [923.4937490000109, 1.8722329915410592e-24, 1.839161578179866e-24, 1.4255887853048985e-24], [926.054375502069, 1.875131571509785e-24, 1.8421878811897086e-24, 1.4276390300585487e-24], [928.6221020067956, 1.878043779079339e-24, 1.845227451387209e-24, 1.429700381891062e-24], [931.1969482007956, 1.8809718447432666e-24, 1.8482825581830752e-24, 1.4317757522927059e-24], [933.7789338252568, 1.883926705984784e-24, 1.851364330265335e-24, 1.4338794937819031e-24], [936.3680786761059, 1.8870724118643454e-24, 1.8546396813843716e-24, 1.4362271450887176e-24], [938.9644026041624, 1.890307660130647e-24, 1.858005494298657e-24, 1.4386856096705216e-24], [941.5679255152812, 1.8928139819824264e-24, 1.8606291254160124e-24, 1.440179331930082e-24], [944.1786673705163, 1.8957715397279324e-24, 1.8637112458816276e-24, 1.4422590645424627e-24], [946.7966481862651, 1.898762011506678e-24, 1.866826253499584e-24, 1.4443751147654962e-24], [949.4218880344295, 1.9017693946942806e-24, 1.869957867462912e-24, 1.4465064327140152e-24], [952.054407042562, 1.9047909041359725e-24, 1.8731032541546743e-24, 1.4486493384450753e-24], [954.6942253940282, 1.9078257644010144e-24, 1.8762616245665207e-24, 1.4508027931889164e-24], [957.3413633281527, 1.9108737073592462e-24, 1.8794327057866294e-24, 1.4529664248601616e-24], [959.9958411403818, 1.9139346338335245e-24, 1.882616396783191e-24, 1.4551400833898302e-24], [962.6576791824383, 1.9170085119838336e-24, 1.8858126650277283e-24, 1.4573237071752533e-24], [965.3268978624701, 1.920095340456147e-24, 1.8890215090060263e-24, 1.4595172746467836e-24], [968.0035176452167, 1.923195133089499e-24, 1.8922429426607624e-24, 1.4617207841680012e-24], [970.6875590521573, 1.9263079118892766e-24, 1.895476988243385e-24, 1.4639342448000911e-24], [973.3790426616767, 1.9294337035299084e-24, 1.8987236727564637e-24, 1.4661576717049266e-24], [976.0779891092141, 1.9325725374979224e-24, 1.9019830260647133e-24, 1.4683910837042504e-24], [978.78441908743, 1.9357244450520958e-24, 1.905255079837242e-24, 1.4706345019126955e-24], [981.4983533463578, 1.9388894586143772e-24, 1.908539866927987e-24, 1.4728879489370885e-24], [984.2198126935658, 1.9420676113991534e-24, 1.911837420998608e-24, 1.4751514483890743e-24], [986.9488179943173, 1.9452589371802974e-24, 1.915147776281532e-24, 1.4774250245786341e-24], [989.6853901717335, 1.9484634701401466e-24, 1.9184709674263327e-24, 1.4797087023155954e-24], [992.4295502069441, 1.9516812447694768e-24, 1.921807029397997e-24, 1.4820025067778525e-24], [995.1813191392592, 1.954912295799351e-24, 1.925155997407592e-24, 1.4843064634218454e-24], [997.9407180663225, 1.958156658154199e-24, 1.928517906864553e-24, 1.486620597920606e-24], [1000.7077681442797, 1.9614143669186426e-24, 1.9318927933429145e-24, 1.4889449361200037e-24], [1003.4824905879341, 1.9646854573136952e-24, 1.9352806925571202e-24, 1.4912795040074337e-24], [1006.2649066709158, 1.967969964679563e-24, 1.938681640344515e-24, 1.4936243276889866e-24], [1009.0550377258382, 1.971267924462978e-24, 1.9420956726524594e-24, 1.4959794333727014e-24], [1011.8529051444687, 1.974579372207788e-24, 1.9455228255287495e-24, 1.4983448473560476e-24], [1014.6585303778838, 1.9779043435478524e-24, 1.9489631351143972e-24, 1.5007205960164703e-24], [1017.4719349366449, 1.981242874201806e-24, 1.9524166376382785e-24, 1.5031067058042835e-24], [1020.2931403909539, 1.9845949999688e-24, 1.9558833694128177e-24, 1.5055032032371074e-24], [1023.1221683708197, 1.987960756725625e-24, 1.959363366831036e-24, 1.5079101148957167e-24], [1025.9590405662302, 1.991340180424071e-24, 1.9628566663638704e-24, 1.5103274674206479e-24], [1028.8037787273101, 1.994733307089273e-24, 1.966363304558475e-24, 1.512755287509697e-24], [1031.6564046644962, 1.9981401728180942e-24, 1.969883318036576e-24, 1.515193601915851e-24], [1034.5169402486965, 2.0015608137782077e-24, 1.9734167434935263e-24, 1.5176424374457468e-24], [1037.3854074114654, 2.004995266207238e-24, 1.9769636176974214e-24, 1.5201018209584401e-24], [1040.2618281451646, 2.0084435664119415e-24, 1.9805239774882576e-24, 1.5225717793643201e-24], [1043.146224503138, 2.011905750767958e-24, 1.9840978597776698e-24, 1.5250523396244696e-24], [1046.0386185998777, 2.0153818557193663e-24, 1.987685301548472e-24, 1.527543528749995e-24], [1048.9390326111945, 2.018871917778451e-24, 1.991286339854404e-24, 1.5300453738015774e-24], [1051.8474887743844, 2.0223759735255395e-24, 1.9949010118199625e-24, 1.5325579018891306e-24], [1054.7640093884065, 2.0258940596089304e-24, 1.9985293546403105e-24, 1.53508114017151e-24], [1057.6886168140447, 2.029426212744918e-24, 2.0021714055812933e-24, 1.5376151158563816e-24], [1060.6213334740894, 2.0329724697177673e-24, 2.005827201979403e-24, 1.5401598562000796e-24], [1063.5621818534983, 2.03653286737972e-24, 2.009496781241771e-24, 1.5427153885074605e-24], [1066.51118449958, 2.0401074426511734e-24, 2.013180180846334e-24, 1.5452817401319724e-24], [1069.4683640221551, 2.0436962325208424e-24, 2.016877438341995e-24, 1.547858938475633e-24], [1072.433743093743, 2.047299274045708e-24, 2.020588591348549e-24, 1.5504470109889827e-24], [1075.4073444497249, 2.050916604351427e-24, 2.0243136775570896e-24, 1.553045985171279e-24], [1078.3891908885196, 2.054548260632401e-24, 2.0280527347300704e-24, 1.5556558885704763e-24], [1081.379305271765, 2.058194280151905e-24, 2.03180580070141e-24, 1.5582767487833641e-24], [1084.3777105244842, 2.0618547002424244e-24, 2.0355729133768343e-24, 1.5609085934556492e-24], [1087.3844296352706, 2.065529558305853e-24, 2.0393541107340535e-24, 1.5635514502821822e-24], [1090.3994856564548, 2.0692188918136008e-24, 2.0431494308228625e-24, 1.5662053470069422e-24], [1093.4229017042906, 2.0729227383070718e-24, 2.0469589117656028e-24, 1.5688703114233645e-24], [1096.4547009591224, 2.0766411353976804e-24, 2.0507825917571774e-24, 1.571546371374378e-24], [1099.494906665573, 2.0803741207672695e-24, 2.054620509065453e-24, 1.5742335547526376e-24], [1102.5435421327143, 2.084121732168266e-24, 2.058472702031398e-24, 1.5769318895006631e-24], [1105.6006307342507, 2.0878840074240475e-24, 2.0623392090694423e-24, 1.5796414036110873e-24], [1108.6661959086925, 2.0916609844291992e-24, 2.0662200686677277e-24, 1.5823621251267985e-24], [1111.7402611595444, 2.0954527011496766e-24, 2.070115319388249e-24, 1.5850940821411172e-24], [1114.8228500554753, 2.0992591956233003e-24, 2.0740249998673436e-24, 1.5878373027981045e-24], [1117.9139862305096, 2.103080505959832e-24, 2.0779491488157506e-24, 1.5905918152926473e-24], [1121.0136933841986, 2.1069166703414824e-24, 2.081887805019109e-24, 1.5933576478707523e-24], [1124.1219952818105, 2.110767727022956e-24, 2.0858410073379847e-24, 1.596134828829714e-24], [1127.2389157545056, 2.1146337143321048e-24, 2.0898087947085203e-24, 1.5989233865184141e-24], [1130.3644786995237, 2.11851467066982e-24, 2.0937912061423035e-24, 1.6017233493374152e-24], [1133.4987080803692, 2.122410634510673e-24, 2.097788280727002e-24, 1.6045347457393259e-24], [1136.6416279269854, 2.126321644403072e-24, 2.1018000576264988e-24, 1.6073576042289545e-24], [1139.7932623359502, 2.1302477389696214e-24, 2.105826576081248e-24, 1.6101919533635565e-24], [1142.953635470651, 2.1341889569074395e-24, 2.1098678754085714e-24, 1.613037821753096e-24], [1146.1227715614789, 2.138145336988505e-24, 2.1139239950029877e-24, 1.6158952380604696e-24], [1149.3006949060054, 2.142116918059847e-24, 2.1179949743363978e-24, 1.6187642310017124e-24], [1152.4874298691786, 2.1461037390441102e-24, 2.1220808529586286e-24, 1.6216448293463375e-24], [1155.6830008834993, 2.1501058389397288e-24, 2.1261816704975913e-24, 1.624537061917515e-24], [1158.8874324492165, 2.1541232568211918e-24, 2.130297466659534e-24, 1.6274409575922995e-24], [1162.100749134516, 2.1581560318396145e-24, 2.1344282812296e-24, 1.630356545301974e-24], [1165.3229755756984, 2.162204203222817e-24, 2.1385741540718857e-24, 1.6332838540322326e-24], [1168.55413647738, 2.166267810275914e-24, 2.1427351251300194e-24, 1.6362229128234815e-24], [1171.7942566126735, 2.170346892381385e-24, 2.1469112344272164e-24, 1.6391737507710555e-24], [1175.0433608233852, 2.1744414889997262e-24, 2.1511025220669037e-24, 1.642136397025529e-24], [1178.3014740201988, 2.1785516396695252e-24, 2.1553090282327975e-24, 1.6451108807929667e-24], [1181.5686211828693, 2.1826773840080276e-24, 2.1595307931894376e-24, 1.648097231335185e-24], [1184.8448273604176, 2.1868187617112643e-24, 2.1637678572823054e-24, 1.6510954779699956e-24], [1188.1301176713155, 2.1909758125547453e-24, 2.1680202609385003e-24, 1.6541056500715783e-24], [1191.424517303684, 2.1951485763933356e-24, 2.1722880446665954e-24, 1.657127777070574e-24], [1194.7280515154891, 2.1993370931621748e-24, 2.1765712490575454e-24, 1.660161888454611e-24], [1198.0407456347232, 2.203541402876607e-24, 2.180869914784595e-24, 1.6632080137683615e-24], [1201.3626250596133, 2.2077615456327392e-24, 2.1851840826038233e-24, 1.666266182613928e-24], [1204.6937152588048, 2.2119975616077307e-24, 2.1895137933544124e-24, 1.6693364246511047e-24], [1208.0340417715665, 2.216249491060318e-24, 2.1938590879591588e-24, 1.6724187695977136e-24], [1211.3836302079767, 2.220517374330941e-24, 2.198220007424575e-24, 1.6755132472297926e-24], [1214.742506249129, 2.224801251842363e-24, 2.2025965928415053e-24, 1.6786198873819816e-24], [1218.1106956473193, 2.229101164099895e-24, 2.206988885385312e-24, 1.681738719947776e-24], [1221.488224226254, 2.2334171516918606e-24, 2.211396926316339e-24, 1.684869774879837e-24], [1224.8751178812356, 2.2377492552899373e-24, 2.2158207569802293e-24, 1.6880130821902573e-24], [1228.2714025793762, 2.2420975156495875e-24, 2.2202604188083358e-24, 1.691168671950949e-24], [1231.677104359784, 2.246461973610313e-24, 2.2247159533179575e-24, 1.6943365742937948e-24], [1235.0922493337655, 2.250842670096264e-24, 2.229187402112932e-24, 1.697516819411142e-24], [1238.5168636850321, 2.2552396461164827e-24, 2.233674806883863e-24, 1.700709437555986e-24], [1241.9509736698922, 2.2596529427653245e-24, 2.2381782094085132e-24, 1.7039144590423098e-24], [1245.3946056174605, 2.2640826012230097e-24, 2.2426976515523456e-24, 1.7071319142455003e-24], [1248.8477859298516, 2.2685286627557746e-24, 2.247233175268651e-24, 1.7103618336025165e-24], [1252.3105410823919, 2.2729911687164977e-24, 2.2517848225991605e-24, 1.7136042476123364e-24], [1255.782897623811, 2.277470160544949e-24, 2.256352635674268e-24, 1.716859186836213e-24], [1259.2648821764612, 2.281965679768462e-24, 2.260936656713687e-24, 1.720126681898137e-24], [1262.7565214365022, 2.2864777680020794e-24, 2.2655369280265783e-24, 1.723406763485063e-24], [1266.2578421741232, 2.2910064669492296e-24, 2.270153492012206e-24, 1.7266994623473335e-24], [1269.768871233735, 2.295551818401974e-24, 2.274786391160158e-24, 1.7300048092990184e-24], [1273.2896355341866, 2.3001138642416407e-24, 2.2794356680509783e-24, 1.7333228352183597e-24], [1276.8201620689617, 2.3046926464391565e-24, 2.2841013653564532e-24, 1.7366535710480812e-24], [1280.3604779063958, 2.309288207055627e-24, 2.2887835258402013e-24, 1.7399970477958812e-24], [1283.9106101898724, 2.313900588242757e-24, 2.2934821923580453e-24, 1.7433532965348185e-24], [1287.4705861380428, 2.3185298322435075e-24, 2.2981974078586763e-24, 1.7467223484038487e-24], [1291.0404330450226, 2.323175981392497e-24, 2.3029292153840133e-24, 1.7501042346082582e-24], [1294.6201782806154, 2.3278390781166617e-24, 2.3076776580698598e-24, 1.7534989864202376e-24], [1298.2098492905104, 2.332519164935844e-24, 2.3124427791464677e-24, 1.7569066351794706e-24], [1301.8094735964946, 2.3372162844635395e-24, 2.317224621939259e-24, 1.760327212293857e-24], [1305.4190787966718, 2.341930479407601e-24, 2.322023229869525e-24, 1.7637607492402332e-24], [1309.038692565663, 2.346661792571058e-24, 2.3268386464552193e-24, 1.767207277565231e-24], [1312.6683426548302, 2.351410266853129e-24, 2.331670915311952e-24, 1.770666828886377e-24], [1316.3080568924772, 2.3561759452502696e-24, 2.3365200801540284e-24, 1.7741394348933192e-24], [1319.957863184075, 2.3609588708576263e-24, 2.3413861847958848e-24, 1.777625127349361e-24], [1323.6177895124642, 2.365759086870544e-24, 2.3462692731535837e-24, 1.7811239380933843e-24], [1327.287863938079, 2.370576636586589e-24, 2.351169389246827e-24, 1.7846358990423174e-24], [1330.968114599157, 2.3754115634083214e-24, 2.3560865772017253e-24, 1.788161042194373e-24], [1334.658569711958, 2.3802639108464183e-24, 2.3610208812539236e-24, 1.7916993996332688e-24], [1338.359257570975, 2.3851337225246756e-24, 2.365972345753606e-24, 1.7952510035342496e-24], [1342.070206549159, 2.3900210421857713e-24, 2.3709410151712856e-24, 1.798815886172016e-24], [1345.7914450981284, 2.3949259137008766e-24, 2.3759269341074605e-24, 1.8023940799327164e-24], [1349.523001748396, 2.399848381082156e-24, 2.3809301473051765e-24, 1.8059856173307457e-24], [1353.2649051095768, 2.4047884885025938e-24, 2.3859506996699797e-24, 1.8095905310346536e-24], [1357.0171838706199, 2.4097462803259153e-24, 2.3909886363000283e-24, 1.8132088539067525e-24], [1360.7798668000157, 2.4147218011543517e-24, 2.3960440025341845e-24, 1.816840619066442e-24], [1364.552982746026, 2.4197150959077976e-24, 2.4011168440317302e-24, 1.820485859995116e-24], [1368.3365606369055, 2.424726209959896e-24, 2.406207206909415e-24, 1.82414461071693e-24], [1372.1306294811122, 2.429755189384027e-24, 2.4113151379892196e-24, 1.827816906125791e-24], [1375.9352183675444, 2.4348020814250252e-24, 2.4164406852734912e-24, 1.8315027826123255e-24], [1379.7503564657513, 2.4398669354680205e-24, 2.4215838989208127e-24, 1.835202279352167e-24], [1383.5760730261677, 2.4449498052114624e-24, 2.4267448334348307e-24, 1.838915441195601e-24], [1387.4123973803273, 2.4500507541191916e-24, 2.4319235531559313e-24, 1.8426423259184416e-24], [1391.259358941099, 2.455169871339493e-24, 2.437120148295894e-24, 1.8463830253932167e-24], [1395.1169872029009, 2.460307329478323e-24, 2.4423347931302164e-24, 1.8501377424254604e-24], [1398.985311741934, 2.465463681621647e-24, 2.4475680451726073e-24, 1.8539071857883643e-24], [1402.8643622164127, 2.4706429370769896e-24, 2.4528239431719824e-24, 1.857696660950852e-24], [1406.7541683667766, 2.4837852437666613e-24, 2.4660997616886567e-24, 1.872053769343315e-24], [1410.6547600159365, 2.4810455292007524e-24, 2.46337887076027e-24, 1.8653028443649564e-24], [1414.56616706949, 2.486267285664233e-24, 2.46867631073028e-24, 1.8691173949085133e-24], [1418.4884195159561, 2.4915128937742477e-24, 2.473997353252844e-24, 1.8729531308309536e-24], [1422.4215474270095, 2.496777721676137e-24, 2.4793373335684864e-24, 1.8768038642584145e-24], [1426.3655809576996, 2.5020612620681035e-24, 2.4846957411531185e-24, 1.8806688931664172e-24], [1430.3205503466952, 2.507363426881324e-24, 2.4900724877276655e-24, 1.8845480730809173e-24], [1434.2864859165036, 2.512684216811589e-24, 2.495467574414449e-24, 1.8884413775291167e-24], [1438.2634180737118, 2.5180236592691933e-24, 2.500881029241638e-24, 1.892348815506214e-24], [1442.2513773092217, 2.523381791658063e-24, 2.5063128903024622e-24, 1.8962704092350906e-24], [1446.25039419847, 2.5287586557425007e-24, 2.511763200081599e-24, 1.9002061866725964e-24], [1450.26049940168, 2.534154295431328e-24, 2.5172320032234376e-24, 1.9041561785622997e-24], [1454.281723664081, 2.539568755801457e-24, 2.522719345548653e-24, 1.9081204171350475e-24], [1458.3140978161587, 2.545002082627255e-24, 2.528225273580184e-24, 1.912098935482993e-24], [1462.357652773878, 2.550454322137684e-24, 2.5337498342986213e-24, 1.916091767236192e-24], [1466.412419538932, 2.555925520883513e-24, 2.5392930750084445e-24, 1.9200989463855913e-24], [1470.4784291989683, 2.5614157256610962e-24, 2.5448550432612378e-24, 1.9241205071815018e-24], [1474.555712927839, 2.5669249834667933e-24, 2.550435786809761e-24, 1.9281564840726588e-24], [1478.6443019858268, 2.572453341469209e-24, 2.5560353535799662e-24, 1.9322069116689824e-24], [1482.7442277198998, 2.5780008469912245e-24, 2.5616537916528758e-24, 1.9362718247174644e-24], [1486.8555215639399, 2.5835675474986552e-24, 2.5672911492531388e-24, 1.9403512580868046e-24], [1490.9782150389844, 2.5891534905926626e-24, 2.572947474741367e-24, 1.9444452467569605e-24], [1495.1123397534768, 2.5947587240046275e-24, 2.578622816608936e-24, 1.948553825812187e-24], [1499.2579274034972, 2.6003832955927923e-24, 2.584317223474593e-24, 1.9526770304362264e-24], [1503.4150097730176, 2.606027253340031e-24, 2.590030744082179e-24, 1.956814895909127e-24], [1507.5836187341329, 2.611690645352414e-24, 2.5957634272991614e-24, 1.9609674576049087e-24], [1511.763786247318, 2.617373519858102e-24, 2.6015153221154934e-24, 1.9651347509899292e-24], [1515.9555443616587, 2.6230759252067777e-24, 2.607286477643021e-24, 1.9693168116219543e-24], [1520.1589252151161, 2.6287979098693305e-24, 2.613076943115138e-24, 1.9735136751493155e-24], [1524.3739610347518, 2.634539522437877e-24, 2.618886767886786e-24, 1.977725377310598e-24], [1528.6006841369922, 2.640300811625565e-24, 2.62471600143423e-24, 1.9819519539343014e-24], [1532.8391269278636, 2.646081826266866e-24, 2.6305646933553307e-24, 1.986193440938836e-24], [1537.0893219032528, 2.6518826153176563e-24, 2.6364328933695984e-24, 1.9904498743324023e-24], [1541.3513016491434, 2.6577032278557053e-24, 2.64232065131867e-24, 1.9947212902132143e-24], [1545.6250988418783, 2.6635437130807464e-24, 2.6482280171663416e-24, 1.999007724769554e-24], [1549.9107462483983, 2.6694041203149808e-24, 2.6541550409990692e-24, 2.003309214280045e-24], [1554.2082767265047, 2.6752844990035444e-24, 2.6601017730264006e-24, 2.0076257951138285e-24], [1558.517723225098, 2.681184898714686e-24, 2.666068263581143e-24, 2.0119575037309164e-24], [1562.8391187844481, 2.6871053691404584e-24, 2.672054563120022e-24, 2.0163043766823697e-24], [1567.1724965364317, 2.6930459600970507e-24, 2.678060722224007e-24, 2.0206664506107728e-24], [1571.5178897047902, 2.6990067215252205e-24, 2.6840867915987134e-24, 2.0250437622504653e-24], [1575.875331605392, 2.7049877034908533e-24, 2.690132822074946e-24, 2.0294363484278862e-24], [1580.2448556464765, 2.7109889561853247e-24, 2.6961988646090395e-24, 2.0338442460619414e-24], [1584.6264953289228, 2.717010529926191e-24, 2.7022849702835287e-24, 2.0382674921644303e-24], [1589.0202842464926, 2.7230524751574024e-24, 2.708391190307347e-24, 2.0427061238402796e-24], [1593.426256086102, 2.729114842450118e-24, 2.7145175760166135e-24, 2.047160178288153e-24], [1597.8444446280669, 2.7351976825029283e-24, 2.7206641788748453e-24, 2.0516296928005184e-24], [1602.2748837463714, 2.741301046142554e-24, 2.726831050473629e-24, 2.056114704764354e-24], [1606.7176074089239, 2.7474249843242807e-24, 2.7330182425330395e-24, 2.0606152516613178e-24], [1611.172649677818, 2.753569548132442e-24, 2.739225806902105e-24, 2.0651313710682413e-24], [1615.6400447095896, 2.759734788781018e-24, 2.7454537955593837e-24, 2.0696631006575247e-24], [1620.1198267554882, 2.7659207576141194e-24, 2.751702260613433e-24, 2.074210478197506e-24], [1624.6120301617273, 2.7721275061065114e-24, 2.7579712543033087e-24, 2.0787735415529073e-24], [1629.1166893697612, 2.778355085864087e-24, 2.764260828999023e-24, 2.083352328685152e-24], [1633.633838916535, 2.7846035486245583e-24, 2.7705710372022185e-24, 2.0879468776529678e-24], [1638.1635134347634, 2.7908729462576857e-24, 2.7769019315463755e-24, 2.092557226612535e-24], [1642.7057476531827, 2.7971633307661234e-24, 2.7832535647976433e-24, 2.097183413818047e-24], [1647.2605763968268, 2.8034747542857867e-24, 2.789625989855183e-24, 2.1018254776221814e-24], [1651.828034587296, 2.8098072690864534e-24, 2.7960192597517548e-24, 2.106483456476419e-24], [1656.4081572430102, 2.8161609275722e-24, 2.8024334276541365e-24, 2.111157388931399e-24], [1661.0009794794946, 2.822535782282003e-24, 2.8088685468637065e-24, 2.1158473136375434e-24], [1665.6065365096356, 2.8289318858903287e-24, 2.8153246708170105e-24, 2.1205532693452665e-24], [1670.2248636439622, 2.835349291207665e-24, 2.8218018530862837e-24, 2.1252752949055476e-24], [1674.855996290905, 2.841788051180906e-24, 2.8283001473798134e-24, 2.130013429270254e-24], [1679.4999699570803, 2.848248218893981e-24, 2.8348196075425455e-24, 2.1347677114925632e-24], [1684.1568202475487, 2.8547298475686795e-24, 2.841360287556902e-24, 2.139538180727566e-24], [1688.8265828660983, 2.8612329905646778e-24, 2.847922241542786e-24, 2.144324876232412e-24], [1693.5092936155213, 2.8677577013805206e-24, 2.854505523758538e-24, 2.1491278373669803e-24], [1698.204988397873, 2.8743040336538967e-24, 2.8611101886012114e-24, 2.1539471035942177e-24], [1702.9137032147582, 2.8808720411625444e-24, 2.8677362906074432e-24, 2.1587827144805936e-24], [1707.6354741676114, 2.8874617778243176e-24, 2.8743838844535124e-24, 2.1636347096965e-24], [1712.37033745796, 2.8940732976981643e-24, 2.881053024956304e-24, 2.168503129016754e-24], [1717.1183293877168, 2.9007066549845524e-24, 2.88774376707371e-24, 2.1733880123210668e-24], [1721.8794863594446, 2.90736190402597e-24, 2.8944561659051188e-24, 2.1782893995944375e-24], [1726.6538448766478, 2.9140390993077362e-24, 2.9011902766922086e-24, 2.1832073309276895e-24], [1731.4414415440401, 2.920738295458223e-24, 2.9079461548191507e-24, 2.1881418465179015e-24], [1736.2423130678344, 2.9274595472498475e-24, 2.914723855813592e-24, 2.1930929866689182e-24], [1741.0564962560265, 2.9342029095993658e-24, 2.921523435346922e-24, 2.1980607917918564e-24], [1745.884028018661, 2.940968437568811e-24, 2.9283449492352066e-24, 2.203045302405699e-24], [1750.7249453681345, 2.9477561863658403e-24, 2.9351884534395096e-24, 2.2080465591377585e-24], [1755.579285419463, 2.9545662113447353e-24, 2.942054004066887e-24, 2.2130646027244064e-24], [1760.4470853905805, 2.961398568006747e-24, 2.948941657370707e-24, 2.2180994740115824e-24], [1765.32838260261, 2.968253312001372e-24, 2.9558514697519212e-24, 2.223151213955769e-24], [1770.2232144801646, 2.975130499126851e-24, 2.9627834977595433e-24, 2.2282198636245802e-24], [1775.1316185516196, 2.9820301853312047e-24, 2.9697377980916685e-24, 2.2333054641979445e-24], [1780.0536324494142, 2.988952426713561e-24, 2.9767144275967953e-24, 2.238408056969242e-24], [1784.9892939103252, 2.9958972795251785e-24, 2.9837134432748277e-24, 2.2435276833467417e-24], [1789.9386407757743, 3.002864800171655e-24, 2.9907349022792822e-24, 2.2486643848557656e-24], [1794.9017109921037, 3.0098550452147672e-24, 2.997778861919114e-24, 2.2538182031412418e-24], [1799.8785426108677, 3.0168680713754567e-24, 3.004845379661694e-24, 2.2589891799713395e-24], [1804.8691737891359, 3.0239039355385266e-24, 3.011934513137511e-24, 2.2641773572431053e-24], [1809.8736427897707, 3.0309626947588304e-24, 3.0190463201463623e-24, 2.269382776990657e-24], [1814.8919879817356, 3.038044406270931e-24, 3.0261808586670354e-24, 2.2746054813979394e-24], [1819.924247840375, 3.045149127505973e-24, 3.0333381868742258e-24, 2.27984551282036e-24], [1824.9704609477237, 3.052276916118458e-24, 3.0405183631653894e-24, 2.285102913821118e-24], [1830.030665992785, 3.0594278300368575e-24, 3.0477214462115276e-24, 2.2903777272375986e-24], [1835.1049017718517, 3.0666019275589113e-24, 3.05494749505282e-24, 2.295669996308963e-24], [1840.1932071887754, 3.0737992675493976e-24, 3.062196569297095e-24, 2.3009797649399278e-24], [1845.295621255289, 3.0810199098876134e-24, 3.0694687295689345e-24, 2.3063070782982102e-24], [1850.4121830912873, 3.0882639166044255e-24, 3.0767640386508393e-24, 2.3116519843335866e-24], [1855.542931925143, 3.0955313552639174e-24, 3.0840825648771743e-24, 2.3170145382959172e-24], [1860.687907093992, 3.1028223115870537e-24, 3.0914243948036794e-24, 2.3223948196015274e-24], [1865.8471480440505, 3.1101369576150458e-24, 3.0987897016181824e-24, 2.3277930229094137e-24], [1871.020694330902, 3.1174763487940766e-24, 3.1061795451134997e-24, 2.3332105231495835e-24], [1876.208585619815, 3.1251405839652005e-24, 3.1138951084419446e-24, 2.3390482543190998e-24], [1881.410861686032, 3.1322228610582402e-24, 3.1210263150026607e-24, 2.3440945804785422e-24], [1886.6275624150926, 3.139631003594938e-24, 3.12848426768992e-24, 2.349562431109461e-24], [1891.8587278031255, 3.1470637611495087e-24, 3.1359666284978204e-24, 2.3550492626582847e-24], [1897.1043979571534, 3.1545205308737482e-24, 3.14347279292113e-24, 2.3605542338182103e-24], [1902.3646130954155, 3.162001284831285e-24, 3.151002733435914e-24, 2.3660772713788894e-24], [1907.6394135476573, 3.169506060762358e-24, 3.158556488429394e-24, 2.3716183897910608e-24], [1912.928839755458, 3.177034911090329e-24, 3.1661341110229756e-24, 2.3771776230285874e-24], [1918.2329322725232, 3.184587892877379e-24, 3.1737356589918364e-24, 2.3827550111668617e-24], [1923.551731765012, 3.19216506504249e-24, 3.1813611919729264e-24, 2.388350596664846e-24], [1928.8852790118328, 3.1997664874066273e-24, 3.189010770506695e-24, 2.393964423089255e-24], [1934.2336149049713, 3.207392220312543e-24, 3.1966844556555315e-24, 2.3995965346060123e-24], [1939.5967804497934, 3.2150423244560647e-24, 3.204382308834439e-24, 2.4052469757543443e-24], [1944.974816765366, 3.2227168608039263e-24, 3.2121043917285484e-24, 2.4109157913368968e-24], [1950.367765084763, 3.2304158905515975e-24, 3.219850766250795e-24, 2.416603026362991e-24], [1955.7756667553956, 3.2381394751005417e-24, 3.2276214945190734e-24, 2.4223087260177206e-24], [1961.1985632393137, 3.245887676044791e-24, 3.2354166388427552e-24, 2.4280329356438307e-24], [1966.6364961135391, 3.2536605551635064e-24, 3.243236261715214e-24, 2.43377570073152e-24], [1972.0895070703693, 3.2614581744164535e-24, 3.2510804258092634e-24, 2.4395370669118363e-24], [1977.557637917712, 3.269280595941515e-24, 3.2589491939746547e-24, 2.445317079953086e-24], [1983.0409305793885, 3.2771278820526995e-24, 3.2668426292360628e-24, 2.4511157857580165e-24], [1988.5394270954766, 3.2850000952396445e-24, 3.2747607947925777e-24, 2.4569332303625486e-24], [1994.0531696226155, 3.2928972981670713e-24, 3.2827037540171417e-24, 2.462769459934908e-24], [1999.58220043433, 3.300819553674844e-24, 3.290671570456598e-24, 2.4686245207751023e-24], [2005.1265619213677, 3.308766924777662e-24, 3.2986643078313704e-24, 2.474498459314627e-24], [2010.686296592008, 3.316739474665787e-24, 3.3066820300361794e-24, 2.480391322116611e-24], [2016.2614470724022, 3.324737266705086e-24, 3.314724801140078e-24, 2.4863031558758092e-24], [2021.8520561068872, 3.332760364437465e-24, 3.322792685386865e-24, 2.492234007418742e-24], [2027.4581665583266, 3.340808831581355e-24, 3.3308857471955745e-24, 2.498183923704038e-24], [2033.079821408425, 3.348882732032326e-24, 3.339004051161068e-24, 2.5041529518228946e-24], [2038.717063758071, 3.3569821298633766e-24, 3.3471476620543216e-24, 2.510141138999058e-24], [2044.369936827659, 3.3651070893255704e-24, 3.355316644823051e-24, 2.516148532589469e-24], [2050.0384839574263, 3.373257674848733e-24, 3.363511064592398e-24, 2.5221751800847197e-24], [2055.7227486077745, 3.381433951041554e-24, 3.3717309866650245e-24, 2.5282211291090486e-24], [2061.4227743596207, 3.389635982693011e-24, 3.379976476522529e-24, 2.534286427421387e-24], [2067.1386049147113, 3.39786383477207e-24, 3.388247599825135e-24, 2.5403711229152204e-24], [2072.870284095977, 3.406117572428646e-24, 3.396544422412646e-24, 2.5464752636192717e-24], [2078.6178558478487, 3.414397260994233e-24, 3.404867010305063e-24, 2.5525988976979305e-24], [2084.3813642366154, 3.422702965982645e-24, 3.413215429703321e-24, 2.5587420734518282e-24], [2090.160853450742, 3.4310347530902905e-24, 3.4215897469895535e-24, 2.5649048393180782e-24], [2095.9563678012196, 3.439392688196775e-24, 3.4299900287276905e-24, 2.5710872438707314e-24], [2101.767951721913, 3.447776837365948e-24, 3.438416341664487e-24, 2.5772893358214605e-24], [2107.595649769876, 3.4561872668459896e-24, 3.446868752729614e-24, 2.5835111640197676e-24], [2113.4395066257207, 3.4646240430704355e-24, 3.4553473290366676e-24, 2.58975277745369e-24], [2119.2995670939376, 3.47308723265847e-24, 3.463852137883462e-24, 2.5960142252500924e-24], [2125.1758761032597, 3.4815769024158746e-24, 3.472383246752955e-24, 2.602295556675173e-24], [2131.068478706986, 3.490093119335187e-24, 3.4809407233134255e-24, 2.6085968211349563e-24], [2136.977420083348, 3.4986359505968745e-24, 3.48952463541961e-24, 2.6149180681757926e-24], [2142.902745535836, 3.5072054635696016e-24, 3.498135051112981e-24, 2.621259347484829e-24], [2148.8445004935643, 3.515801725810864e-24, 3.506772038622376e-24, 2.6276207088904212e-24], [2154.8027305116016, 3.524424805067668e-24, 3.515435666364659e-24, 2.6340022023626646e-24], [2160.777481271342, 3.5330747692774544e-24, 3.5241260029456406e-24, 2.6404038780139393e-24], [2166.7687985808366, 3.5417516865681365e-24, 3.5328431171601165e-24, 2.6468257860992985e-24], [2172.7767283751473, 3.550455625259285e-24, 3.541587077993033e-24, 2.653267977017003e-24], [2178.801316716711, 3.55918665386254e-24, 3.550357954619907e-24, 2.659730501309087e-24], [2184.8426097956767, 3.567944841082197e-24, 3.559155816407396e-24, 2.6662134096616816e-24], [2190.900653930276, 3.576730255816045e-24, 3.56798073291413e-24, 2.6727167529058286e-24], [2196.9754955671633, 3.5855429671559724e-24, 3.576832773891315e-24, 2.679240582017725e-24], [2203.067181281778, 3.5943830443885054e-24, 3.585712009283265e-24, 2.6857849481194185e-24], [2209.1757577787052, 3.6032505569957884e-24, 3.594618509228372e-24, 2.6923499024793377e-24], [2215.3012718920363, 3.612145574656034e-24, 3.603552344059558e-24, 2.6989354965130176e-24], [2221.443770585709, 3.621068167244443e-24, 3.612513584305184e-24, 2.705541781783515e-24], [2227.6033009538933, 3.630018404834082e-24, 3.62150230068992e-24, 2.71216881000237e-24], [2233.7799102213294, 3.6389963576965e-24, 3.6305185641353674e-24, 2.7188166330302797e-24], [2239.9736457437098, 3.6480020963031546e-24, 3.639562445761462e-24, 2.7254853028781467e-24], [2246.1845550080247, 3.657035691325985e-24, 3.6486340168870625e-24, 2.732174871708119e-24], [2252.4126856329426, 3.666097213639569e-24, 3.6577333490320945e-24, 2.7388853918354738e-24], [2258.658085369158, 3.675186734322214e-24, 3.6668605139186345e-24, 2.745616915730236e-24], [2264.9208020997767, 3.684304324659146e-24, 3.676015583474106e-24, 2.752369496020577e-24], [2271.2008838406605, 3.693450056145959e-24, 3.685198629834728e-24, 2.7591431854974195e-24], [2277.4983787408214, 3.702624000494266e-24, 3.694409725351168e-24, 2.7659380371217275e-24], [2283.8133350827707, 3.711826229642268e-24, 3.703648942599134e-24, 2.7727541040376475e-24], [2290.1458012828907, 3.7210568157718586e-24, 3.7129163543965015e-24, 2.779591439595912e-24], [2296.4958258918195, 3.73031583134426e-24, 3.72221203383902e-24, 2.7864500974004264e-24], [2302.8634575948063, 3.739603349173703e-24, 3.731536054374142e-24, 2.793330131406945e-24], [2309.2487452121004, 3.748919442601996e-24, 3.7408884899759554e-24, 2.8002315961563533e-24], [2315.65173769931, 3.758264185968362e-24, 3.75026941561598e-24, 2.8071545474032616e-24], [2322.0724841477954, 3.7676376561249065e-24, 3.759678908781789e-24, 2.8140990441447543e-24], [2328.5110337850238, 3.777039938862457e-24, 3.769117055916179e-24, 2.8210651572237814e-24], [2334.9674359749665, 3.7864711728568644e-24, 3.7785839964545596e-24, 2.8280530281693555e-24], [2341.441740218468, 3.795932620355006e-24, 3.788080995726338e-24, 2.8350643027693664e-24], [2347.9339961536284, 3.805421895672124e-24, 3.7976056635500695e-24, 2.8420957424171554e-24], [2354.444253556174, 3.814938464113304e-24, 3.8071574646577095e-24, 2.8491465865538183e-24], [2360.9725623398585, 3.82448566740114e-24, 3.816739748097856e-24, 2.856221264963132e-24], [2367.518972556824, 3.834062196349262e-24, 3.8263512024484366e-24, 2.8633179799489127e-24], [2374.0835343980048, 3.8436680571301526e-24, 3.835991834355369e-24, 2.870436694881523e-24], [2380.6662981934933, 3.853303311979481e-24, 3.845661706640986e-24, 2.8775774480717945e-24], [2387.2673144129444, 3.862968032339198e-24, 3.855360891350697e-24, 2.8847402900315365e-24], [2393.8866336659416, 3.872662292199824e-24, 3.8650894630816035e-24, 2.8919252745586838e-24], [2400.524306702411, 3.882386166557199e-24, 3.874847497436074e-24, 2.899132456670634e-24], [2407.180384412989, 3.8921397309412746e-24, 3.884635070549561e-24, 2.9063618919728408e-24], [2413.854917829414, 3.901923061243346e-24, 3.8944522589174886e-24, 2.913613636427072e-24], [2420.5479581249333, 3.9117362336439626e-24, 3.904299139322998e-24, 2.9208877462547435e-24], [2427.2595566146756, 3.921579324580393e-24, 3.914175788804346e-24, 2.9281842778926023e-24], [2433.9897647560642, 3.931452410729956e-24, 3.924082284638205e-24, 2.9355032879704274e-24], [2440.7386341491915, 3.941355569002484e-24, 3.934018704332096e-24, 2.9428448332999693e-24], [2447.5062165372337, 3.951288876534994e-24, 3.9439851256190525e-24, 2.9502089708680646e-24], [2454.2925638068295, 3.96125241069034e-24, 3.953981626456265e-24, 2.957595757833871e-24], [2461.0977279884937, 3.9712462490554e-24, 3.9640082850232586e-24, 2.9650052515264852e-24], [2467.921761257007, 3.981270469440842e-24, 3.974065179721657e-24, 2.9724375094441015e-24], [2474.7647159318203, 3.9913251498812736e-24, 3.9841523891753314e-24, 2.979892589253475e-24], [2481.626644477447, 4.00141036863505e-24, 3.994269992230193e-24, 2.9873705487898848e-24], [2488.5075995038815, 4.0115262041848944e-24, 4.00441806795483e-24, 2.99487144605706e-24], [2495.4076337669835, 4.021672735238349e-24, 4.014596695640938e-24, 3.002395339227446e-24], [2502.326800168901, 4.0318500407280015e-24, 4.024805954803558e-24, 3.009942286642507e-24], [2509.265151758458, 4.042058199812418e-24, 4.035045925182001e-24, 3.0175123468131054e-24], [2516.2227417315808, 4.052297291876626e-24, 4.045316686740323e-24, 3.0251055784198517e-24], [2523.199623431683, 4.062567396532487e-24, 4.055618319667714e-24, 3.0327220403134477e-24], [2530.195850350093, 4.0728685936197496e-24, 4.065950904379528e-24, 3.0403617915154308e-24], [2537.211476126466, 4.083200963206507e-24, 4.076314521517754e-24, 3.0480248912183526e-24], [2544.2465545491714, 4.0935645855897185e-24, 4.0867092519515345e-24, 3.0557113987864253e-24], [2551.301139555733, 4.1039595412960506e-24, 4.097135176777997e-24, 3.063421373755896e-24], [2558.3752852332204, 4.114385911082793e-24, 4.107592377323174e-24, 3.0711548758358102e-24], [2565.469045818684, 4.124843775937977e-24, 4.118080935142126e-24, 3.0789119649080142e-24], [2572.5824756995517, 4.135333217081343e-24, 4.128600932019897e-24, 3.0866927010282046e-24], [2579.715629414062, 4.1458543159653e-24, 4.139152449972481e-24, 3.0944971444261997e-24], [2586.868561651667, 4.156407154275272e-24, 4.149735571247164e-24, 3.1023253555064337e-24], [2594.041327253469, 4.1669918139304946e-24, 4.160350378323321e-24, 3.1101773948485308e-24], [2601.233981212618, 4.177608377084676e-24, 4.170996953913073e-24, 3.1180533232077794e-24], [2608.4465786747633, 4.188256926126749e-24, 4.1816753809620465e-24, 3.1259532015158037e-24], [2615.679174938452, 4.19893754368167e-24, 4.1923857426501634e-24, 3.133877090880994e-24], [2622.9318254555556, 4.209650312611084e-24, 4.20312812239231e-24, 3.141825052589006e-24], [2630.204585831709, 4.2203953160138244e-24, 4.213902603838834e-24, 3.149797148103402e-24], [2637.497511826724, 4.231172637227065e-24, 4.2247092708766974e-24, 3.15779343906604e-24], [2644.8106593550165, 4.241982359826348e-24, 4.235548207629506e-24, 3.165813987297814e-24], [2652.1440844860517, 4.2528245676273474e-24, 4.246419498459266e-24, 3.1738588547992785e-24], [2659.4978434447485, 4.2636993446856515e-24, 4.257323227966177e-24, 3.181928103750855e-24], [2666.87199261193, 4.274606775297895e-24, 4.268259480989754e-24, 3.1900217965137513e-24], [2674.266588524759, 4.285546944002745e-24, 4.279228342609816e-24, 3.1981399956306177e-24], [2681.681687877145, 4.296519935581423e-24, 4.2902298981470164e-24, 3.206282763825949e-24], [2689.1173475202104, 4.30752583505858e-24, 4.301264233163705e-24, 3.2144501640069767e-24], [2696.5736244626996, 4.318564727703372e-24, 4.312331433465014e-24, 3.222642259264419e-24], [2704.0505758714417, 4.3296366990303225e-24, 4.3234315850997126e-24, 3.2308591128734436e-24], [2711.5482590717634, 4.340741834800451e-24, 4.334564774361349e-24, 3.239100788294664e-24], [2719.0667315479523, 4.351880221022881e-24, 4.345731087789846e-24, 3.2473673491757897e-24], [2726.606050943674, 4.363051943956492e-24, 4.356930612173161e-24, 3.2556588593535594e-24], [2734.1662750624387, 4.3742570901127386e-24, 4.3681634345501106e-24, 3.263975382857026e-24], [2741.747461868017, 4.385495746259869e-24, 4.379429642214589e-24, 3.272316983912642e-24], [2749.3496694849173, 4.396767999429867e-24, 4.390729322722527e-24, 3.28068372695335e-24], [2756.9729561988024, 4.4080739369326066e-24, 4.4020625639060474e-24, 3.2890756766370174e-24], [2764.617380456944, 4.419413646384547e-24, 4.413429453902222e-24, 3.297492897884483e-24], [2772.2830008686838, 4.430787215778029e-24, 4.424830081222434e-24, 3.3059354559719247e-24], [2779.9698762058624, 4.44219473367026e-24, 4.436264534941632e-24, 3.31440341678388e-24], [2787.6780654032923, 4.453636289814664e-24, 4.447732905330481e-24, 3.322896847658861e-24], [2795.4076275591865, 4.465111977987906e-24, 4.4592352866860884e-24, 3.331415821178232e-24], [2803.158621935634, 4.476621917601651e-24, 4.470771798971851e-24, 3.339960444132574e-24], [2810.9311079590325, 4.48816711774421e-24, 4.482343452980542e-24, 3.348532015632322e-24], [2818.7251452205633, 4.499744956387547e-24, 4.4939476236683905e-24, 3.357126967711227e-24], [2826.5407934766367, 4.5113577238405616e-24, 4.505586604718236e-24, 3.365748316430352e-24], [2834.378112649357, 4.523005323878323e-24, 4.517260300161211e-24, 3.3743958789626915e-24], [2842.237162826967, 4.534687607012911e-24, 4.5289685607030424e-24, 3.3830694009223424e-24], [2850.1180042643346, 4.5464046463989405e-24, 4.540711459977791e-24, 3.391768926189181e-24], [2858.0206973833865, 4.558156529081149e-24, 4.552489085525519e-24, 3.400494517108198e-24], [2865.9453027735963, 4.5699433447876135e-24, 4.564301527571041e-24, 3.4092462394665964e-24], [2873.891881192427, 4.581765184167934e-24, 4.576148877259836e-24, 3.4180241601330594e-24], [2881.860493565818, 4.593622138364352e-24, 4.588031226228615e-24, 3.426828346481914e-24], [2889.8512009886285, 4.605514298876607e-24, 4.5999486664700055e-24, 3.43565886621158e-24], [2897.864064725136, 4.617441757511582e-24, 4.611901290282128e-24, 3.4445157872766846e-24], [2905.8991462094796, 4.62940460636252e-24, 4.623889190247797e-24, 3.4533991778594784e-24], [2913.956507046136, 4.641402937798725e-24, 4.635912459224194e-24, 3.462309106356278e-24], [2922.0362090104063, 4.6534368444624075e-24, 4.6479711903397275e-24, 3.471245641371856e-24], [2930.13831404887, 4.665506419265742e-24, 4.6600654769910754e-24, 3.480208851715884e-24], [2938.2628842798804, 4.677611755390351e-24, 4.672195412842667e-24, 3.489198806401178e-24], [2946.4099819940175, 4.68975294628726e-24, 4.684361091826652e-24, 3.498215574643616e-24], [2954.5796696545913, 4.7019300856768324e-24, 4.696562608142824e-24, 3.5072592258612795e-24], [2962.7720098980953, 4.714143267549489e-24, 4.708800056259349e-24, 3.516329829675074e-24], [2970.9870655347067, 4.72639258616628e-24, 4.721073530913331e-24, 3.525427455908877e-24], [2979.22489954876, 4.738678136058915e-24, 4.7333831271108585e-24, 3.5345521745896596e-24], [2987.485575099229, 4.751000012031213e-24, 4.745728940128434e-24, 3.543704055948118e-24], [2995.769155520208, 4.7633583091591275e-24, 4.758111065513025e-24, 3.5528831704192416e-24], [3004.07570432141, 4.775753122792064e-24, 4.770529599083369e-24, 3.562089588642566e-24], [3012.4052851886336, 4.7881845485530256e-24, 4.782984636930122e-24, 3.571323381462866e-24], [3020.7579619842772, 4.8006526823402026e-24, 4.795476275417458e-24, 3.580584619930909e-24], [3029.133798747799, 4.8131576203268384e-24, 4.808004611182928e-24, 3.589873375303642e-24], [3037.532859696238, 4.825699458962428e-24, 4.820569741138668e-24, 3.599189719044858e-24], [3045.9552092246786, 4.83827829497369e-24, 4.8331717624723704e-24, 3.608533722825869e-24], [3054.4009119067623, 4.8508942253649765e-24, 4.845810772647694e-24, 3.6179054585262485e-24], [3062.8700324951888, 4.8635473474191245e-24, 4.8584868694051224e-24, 3.627304998233914e-24], [3071.3626359221817, 4.876237758698671e-24, 4.871200150763177e-24, 3.6367324142463005e-24], [3079.878787300022, 4.888965557046024e-24, 4.8839507150185985e-24, 3.646187779070518e-24], [3088.418551921519, 4.901730840584708e-24, 4.896738660747581e-24, 3.655671165424271e-24], [3096.981995260536, 4.914533707720065e-24, 4.909564086806488e-24, 3.6651826462361925e-24], [3105.5691829724674, 4.927374257139926e-24, 4.922427092332526e-24, 3.674722294646586e-24], [3114.1801808947657, 4.9402525878158035e-24, 4.935327776744927e-24, 3.684290184008196e-24], [3122.8150550474224, 4.9531687990031126e-24, 4.94826623974519e-24, 3.69388638788656e-24], [3131.473871633494, 4.9661229902426426e-24, 4.9612425813185435e-24, 3.703510980060803e-24], [3140.1566970396107, 4.979115261361164e-24, 4.974256901734555e-24, 3.713164034524445e-24], [3148.863597836458, 4.992145712472165e-24, 4.9873093015478784e-24, 3.7228456254858664e-24], [3157.5946407793226, 5.0052144439771285e-24, 5.000399881599528e-24, 3.732555827369315e-24], [3166.3498928085774, 5.018321556566334e-24, 5.013528743017684e-24, 3.7422947148156426e-24], [3175.1294210502047, 5.031467151219997e-24, 5.0266959872188354e-24, 3.7520623626833596e-24], [3183.9332928163203, 5.044651329209636e-24, 5.039901715909152e-24, 3.761858846050104e-24], [3192.761575605669, 5.057874192099971e-24, 5.0531460310863874e-24, 3.7716842402143516e-24], [3201.6143371041644, 5.071135841751458e-24, 5.0664290350424054e-24, 3.781538620698341e-24], [3210.491645185386, 5.084436380323793e-24, 5.0797508303667154e-24, 3.791422063253058e-24], [3219.3935679111146, 5.097775910284412e-24, 5.0931115199549554e-24, 3.801334643867854e-24], [3228.3201735318626, 5.111154534423333e-24, 5.1065112070237704e-24, 3.811276438790909e-24], [3237.271530487362, 5.124572355891749e-24, 5.119949995149419e-24, 3.821247524580186e-24], [3246.2477074071303, 5.138029478312624e-24, 5.133427988378489e-24, 3.831247978250978e-24], [3255.248773110964, 5.1515260061779195e-24, 5.146945291625474e-24, 3.841277877808338e-24], [3264.274796609495, 5.165062046888416e-24, 5.16050201271436e-24, 3.851337304982655e-24], [3273.3258471046893, 5.1786377318400686e-24, 5.1740982834834266e-24, 3.8614263735075775e-24], [3282.401993990409, 5.192304067650987e-24, 5.187785155306772e-24, 3.8716134270636126e-24], [3291.5033068529146, 5.2059082828109474e-24, 5.201409723923084e-24, 3.881693516848053e-24], [3300.6298554714253, 5.219603346843195e-24, 5.2151250924265775e-24, 3.891871731364864e-24], [3309.781709818626, 5.2333384804196254e-24, 5.2288804375761735e-24, 3.902079905903355e-24], [3318.9589400612376, 5.2471137569832185e-24, 5.2426758331974876e-24, 3.912318075583823e-24], [3328.1616165605274, 5.260929278740623e-24, 5.2565113819029326e-24, 3.92258631392334e-24], [3337.389809872851, 5.274785151090444e-24, 5.2703871894978095e-24, 3.932884698539463e-24], [3346.6435907502123, 5.288681480346681e-24, 5.2843033627017776e-24, 3.9432133080967335e-24], [3355.9230301407815, 5.302618373312237e-24, 5.298260008721991e-24, 3.953572221734101e-24], [3365.228199189463, 5.316595937164591e-24, 5.312257235138676e-24, 3.963961518911277e-24], [3374.559169238416, 5.3306142794181444e-24, 5.3262951498674535e-24, 3.974381279357455e-24], [3383.916011827629, 5.344673507909608e-24, 5.340373861144705e-24, 3.984831583051284e-24], [3393.2987986954386, 5.358773730791917e-24, 5.3544934775214976e-24, 3.995312510212465e-24], [3402.707601779105, 5.372915056532516e-24, 5.368654107861859e-24, 4.005824141298617e-24], [3412.1424932153514, 5.38709759391215e-24, 5.3828558613415836e-24, 4.0163665570028504e-24], [3421.60354534092, 5.4013214520243574e-24, 5.397098847447707e-24, 4.0269398382537834e-24], [3431.0908306931165, 5.415586740277091e-24, 5.411383175980159e-24, 4.0375440662153284e-24], [3440.6044220103886, 5.429893568391813e-24, 5.425708957050833e-24, 4.04817932228677e-24], [3450.144392232856, 5.44424204640541e-24, 5.4400763010855255e-24, 4.0588456881035194e-24], [3459.7108145028983, 5.4586322846699844e-24, 5.4544853188237226e-24, 4.0695432455370626e-24], [3469.3037621656877, 5.4730643938541896e-24, 5.468936121319934e-24, 4.0802720766960034e-24], [3478.9233087697776, 5.4875384849439994e-24, 5.483428819944474e-24, 4.0910322639263646e-24], [3488.5695280676377, 5.5020546692434864e-24, 5.4979635263842475e-24, 4.101823889812328e-24], [3498.2424940162564, 5.516613058375918e-24, 5.5125403526438366e-24, 4.11264703717685e-24], [3507.9422807776755, 5.53121376428385e-24, 5.5271594110455985e-24, 4.1235017890821445e-24], [3517.668962719565, 5.545856899230958e-24, 5.541820814231522e-24, 4.134388228830574e-24], [3527.422614415812, 5.560542575802802e-24, 5.556524675163963e-24, 4.145306439965199e-24], [3537.2033106470635, 5.5752709069068145e-24, 5.571271107125669e-24, 4.1562565062702295e-24], [3547.0111264013317, 5.590042005774158e-24, 5.5860602237216166e-24, 4.1672385117721315e-24], [3556.8461368745375, 5.604855985960569e-24, 5.600892138879868e-24, 4.17825254074004e-24], [3566.7084174711154, 5.619712961346613e-24, 5.6157669668518305e-24, 4.1892986776862644e-24], [3576.5980438045654, 5.634613046139367e-24, 5.630684822213941e-24, 4.200377007367484e-24], [3586.5150916980556, 5.64955635487254e-24, 5.6456458198677995e-24, 4.211487614784891e-24], [3596.4596371849902, 5.6645430024080794e-24, 5.6606500750417504e-24, 4.2226305851852365e-24], [3606.4317565095985, 5.679573103937436e-24, 5.675697703292198e-24, 4.233806004061774e-24], [3616.4315261275074, 5.694646774981081e-24, 5.6907888205030824e-24, 4.245013957154411e-24], [3626.459022706351, 5.709764131391369e-24, 5.7059235428887734e-24, 4.256254530451076e-24], [3636.5143231263282, 5.7249252893523245e-24, 5.721101986993851e-24, 4.267527810188843e-24], [3646.5975044808247, 5.7401303653817934e-24, 5.7363242696952634e-24, 4.278833882854435e-24], [3656.708644076973, 5.7553794763323026e-24, 5.751590508203192e-24, 4.290172835186475e-24], [3666.847819436273, 5.770672739394068e-24, 5.7669008200640574e-24, 4.3015447541774615e-24], [3677.0151082951597, 5.786010272097516e-24, 5.782255323163062e-24, 4.312949727078035e-24], [3687.210588605619, 5.801392192320665e-24, 5.7976541357315576e-24, 4.3243878414050646e-24], [3697.4343385357906, 5.816818618303302e-24, 5.813097376361253e-24, 4.3358591849608334e-24], [3707.6864364705334, 5.832289668685562e-24, 5.8285851640428085e-24, 4.347363845884296e-24], [3717.9669610120673, 5.8478054626409925e-24, 5.844117618298995e-24, 4.3589019128283636e-24], [3728.2759909805427, 5.863366120487097e-24, 5.859694859795613e-24, 4.370473475779196e-24], [3738.613605414673, 5.878971768771253e-24, 5.875317015430563e-24, 4.38207863287848e-24], [3748.9798835723136, 5.8946228776487056e-24, 5.890984555920076e-24, 4.39371794305045e-24], [3759.3749049310973, 5.9103184976545284e-24, 5.906696531481908e-24, 4.405390060449698e-24], [3769.7987491890135, 5.9260597629757756e-24, 5.922454077271551e-24, 4.417096434458406e-24], [3780.2514962650434, 5.9418465245731636e-24, 5.93825704442311e-24, 4.428836792676345e-24], [3790.7332262997757, 5.957678874285683e-24, 5.954105525096645e-24, 4.4406111856598466e-24], [3801.244019655991, 5.973556931183284e-24, 5.96999963869831e-24, 4.452419700292888e-24], [3811.7839569192993, 5.9894808166760006e-24, 5.985939506974173e-24, 4.464262426392061e-24], [3822.3531188987677, 6.005450652897732e-24, 6.0019252523930124e-24, 4.476139454539196e-24], [3832.9515866275115, 6.0214665624385546e-24, 6.0179569978784705e-24, 4.488050875719465e-24], [3843.5794413633466, 6.037528668276108e-24, 6.034034866740391e-24, 4.499996781230718e-24], [3854.236764589383, 6.053637093755624e-24, 6.0501589826548465e-24, 4.511977262654921e-24], [3864.923638014678, 6.0697919625819315e-24, 6.0663294696561476e-24, 4.523992411847388e-24], [3875.640143574834, 6.0859933988171785e-24, 6.082546452134553e-24, 4.536042320932792e-24], [3886.3863634326476, 6.1022415268796196e-24, 6.098810054835072e-24, 4.548127082303375e-24], [3897.1623799787462, 6.118536471544263e-24, 6.115120402858106e-24, 4.56024678861899e-24], [3907.9682758321833, 6.1348783579431436e-24, 6.131477621659732e-24, 4.57240153280664e-24], [3918.8041338411126, 6.151267311566027e-24, 6.147881837052414e-24, 4.5845914080612825e-24], [3929.6700370833933, 6.1677034582612434e-24, 6.164333175205823e-24, 4.596816507846002e-24], [3940.5660688672524, 6.184186924236673e-24, 6.180831762647853e-24, 4.60907692589296e-24], [3951.4923127318966, 6.200717836060347e-24, 6.1973777262652004e-24, 4.621372756203642e-24], [3962.4488524481826, 6.217296320661934e-24, 6.213971193304873e-24, 4.633704093050104e-24], [3973.4357720192265, 6.233922505333334e-24, 6.230612291374778e-24, 4.646071030975067e-24], [3984.4531556810816, 6.250596517729838e-24, 6.2473011484448906e-24, 4.658473664793023e-24], [3995.501087903348, 6.267318485870848e-24, 6.264037892847975e-24, 4.670912089590954e-24], [4006.579653389863, 6.284088538141163e-24, 6.280822653280873e-24, 4.683386400728753e-24], [4017.688937079315, 6.300906803292124e-24, 6.2976555588056634e-24, 4.695896693840547e-24], [4028.8290241459, 6.317773410442199e-24, 6.3145367388502354e-24, 4.708443064834889e-24], [4039.9999999999977, 6.334688489078047e-24, 6.3314663232093744e-24, 4.721025609895621e-24]])
   
    f1 = data[:,0]
    if IFO.upper() in ['LIGO3', 'LIGO3-DEFAULT']:
        S1 = data[:,1]**2
    elif IFO.upper()=='LIGO3-MID':
        S1 = data[:,2]**2
    elif IFO.upper()=='LIGO3-HIGH':
        S1 = data[:,4]**2

    fstop = [f1[0],f1[-1]]

# % ---- Interpolate to desired frequencies.  Set noise to infinity outside
# %      band of frequencies covered by f1.
# S = inf(size(f));
# k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop


# % -------------------------------------------------------------------------
# %                              KAGRA spectrum
# % -------------------------------------------------------------------------

def kagraspectrum(IFO,f):
# % KAGRASPECTRUM - Design noise curve for KAGRA. Data taken from the DCC 
# % entry https://dcc.ligo.org/T1600593-v1, which was used for DCC v36 of the 
# % Observing Scenarios paper (https://dcc.ligo.org/LIGO-P1200087). 
# %
# % Arguments are the same as for the parent function.

# % The following analytic fit is copied from Manzotti & Dietz,
# % https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=p1200014
# % I have not checked its accuracy, and it is not used in this function.
# %   S.^0.5 = 6.499e-25 * ( ...
# %       + 9.72e-9*exp(-1.43-9.88*x-0.23*x.^2) ...
# %          + 1.17*exp( 0.14-3.10*x-0.26*x.^2) ...
# %          + 1.70*exp( 0.14+1.09*x-0.013*x.^2) ...
# %          + 1.25*exp(0.071+2.83*x-4.91*x.^2) ...
# %   );
# % where 
# %   x := log(f/100Hz)

# % ---- Extract pre-sampled noise spectrum from data.  First column is
# %      frequency, second column is KAGRA_Early curve, third is KAGRA_Mid curve,
# %      fourth is KAGRA_Late curve, fifth is KAGRA_Design ("KAGRA") curve.
    data = np.array([[1.0023052, 7.6534e-15, 3.8267e-15, 2.29602e-15, 7.6533992e-16], [1.0046158, 7.51513e-15, 3.75757e-15, 2.25454e-15, 7.5151318e-16], [1.0069317, 7.37938e-15, 3.68969e-15, 2.21381e-15, 7.3793811e-16], [1.0092529, 7.2461e-15, 3.62305e-15, 2.17383e-15, 7.2461015e-16], [1.0115795, 7.11525e-15, 3.55762e-15, 2.13457e-15, 7.1152481e-16], [1.0139114, 6.98678e-15, 3.49339e-15, 2.09603e-15, 6.9867768e-16], [1.0162487, 6.86064e-15, 3.43032e-15, 2.05819e-15, 6.8606444e-16], [1.0185914, 6.73681e-15, 3.3684e-15, 2.02104e-15, 6.7368084e-16], [1.0209395, 6.61523e-15, 3.30761e-15, 1.98457e-15, 6.6152273e-16], [1.023293, 6.49586e-15, 3.24793e-15, 1.94876e-15, 6.49586e-16], [1.0256519, 6.37867e-15, 3.18933e-15, 1.9136e-15, 6.3786665e-16], [1.0280163, 6.26361e-15, 3.1318e-15, 1.87908e-15, 6.2636072e-16], [1.0303861, 6.15064e-15, 3.07532e-15, 1.84519e-15, 6.1506436e-16], [1.0327614, 6.03974e-15, 3.01987e-15, 1.81192e-15, 6.0397374e-16], [1.0351422, 5.93085e-15, 2.96543e-15, 1.77926e-15, 5.9308516e-16], [1.0375284, 5.82395e-15, 2.91197e-15, 1.74718e-15, 5.8239494e-16], [1.0399202, 5.71899e-15, 2.8595e-15, 1.7157e-15, 5.7189949e-16], [1.0423174, 5.61595e-15, 2.80798e-15, 1.68479e-15, 5.6159527e-16], [1.0447202, 5.51479e-15, 2.75739e-15, 1.65444e-15, 5.5147883e-16], [1.0471285, 5.41547e-15, 2.70773e-15, 1.62464e-15, 5.4154676e-16], [1.0495424, 5.31796e-15, 2.65898e-15, 1.59539e-15, 5.3179572e-16], [1.0519619, 5.22222e-15, 2.61111e-15, 1.56667e-15, 5.2222243e-16], [1.0543869, 5.12824e-15, 2.56412e-15, 1.53847e-15, 5.1282368e-16], [1.0568175, 5.03596e-15, 2.51798e-15, 1.51079e-15, 5.0359629e-16], [1.0592537, 4.94537e-15, 2.47269e-15, 1.48361e-15, 4.9453718e-16], [1.0616956, 4.85643e-15, 2.42822e-15, 1.45693e-15, 4.856433e-16], [1.064143, 4.76912e-15, 2.38456e-15, 1.43073e-15, 4.7691165e-16], [1.0665961, 4.68339e-15, 2.3417e-15, 1.40502e-15, 4.683393e-16], [1.0690549, 4.59923e-15, 2.29962e-15, 1.37977e-15, 4.5992337e-16], [1.0715193, 4.51661e-15, 2.25831e-15, 1.35498e-15, 4.5166104e-16], [1.0739894, 4.4355e-15, 2.21775e-15, 1.33065e-15, 4.4354952e-16], [1.0764652, 4.35586e-15, 2.17793e-15, 1.30676e-15, 4.355861e-16], [1.0789467, 4.27768e-15, 2.13884e-15, 1.2833e-15, 4.2776809e-16], [1.081434, 4.20093e-15, 2.10046e-15, 1.26028e-15, 4.2009287e-16], [1.0839269, 4.12558e-15, 2.06279e-15, 1.23767e-15, 4.1255786e-16], [1.0864256, 4.05161e-15, 2.0258e-15, 1.21548e-15, 4.0516053e-16], [1.0889301, 3.97898e-15, 1.98949e-15, 1.1937e-15, 3.978984e-16], [1.0914403, 3.90769e-15, 1.95385e-15, 1.17231e-15, 3.9076902e-16], [1.0939564, 3.8377e-15, 1.91885e-15, 1.15131e-15, 3.8377002e-16], [1.0964782, 3.76899e-15, 1.8845e-15, 1.1307e-15, 3.7689903e-16], [1.0990058, 3.70154e-15, 1.85077e-15, 1.11046e-15, 3.7015374e-16], [1.1015393, 3.63532e-15, 1.81766e-15, 1.0906e-15, 3.6353191e-16], [1.1040786, 3.57031e-15, 1.78516e-15, 1.07109e-15, 3.5703129e-16], [1.1066238, 3.5065e-15, 1.75325e-15, 1.05195e-15, 3.5064972e-16], [1.1091748, 3.44385e-15, 1.72193e-15, 1.03316e-15, 3.4438506e-16], [1.1117317, 3.38235e-15, 1.69118e-15, 1.01471e-15, 3.3823519e-16], [1.1142945, 3.32198e-15, 1.66099e-15, 9.96594e-16, 3.3219806e-16], [1.1168632, 3.26272e-15, 1.63136e-15, 9.78815e-16, 3.2627165e-16], [1.1194379, 3.20454e-15, 1.60227e-15, 9.61362e-16, 3.2045396e-16], [1.1220185, 3.14743e-15, 1.57372e-15, 9.44229e-16, 3.1474305e-16], [1.124605, 3.09137e-15, 1.54569e-15, 9.27411e-16, 3.0913701e-16], [1.1271975, 3.03634e-15, 1.51817e-15, 9.10902e-16, 3.0363396e-16], [1.1297959, 2.98232e-15, 1.49116e-15, 8.94696e-16, 2.9823205e-16], [1.1324004, 2.92929e-15, 1.46465e-15, 8.78788e-16, 2.9292947e-16], [1.1350108, 2.87724e-15, 1.43862e-15, 8.63173e-16, 2.8772445e-16], [1.1376273, 2.82615e-15, 1.41308e-15, 8.47846e-16, 2.8261525e-16], [1.1402498, 2.776e-15, 1.388e-15, 8.328e-16, 2.7760016e-16], [1.1428783, 2.72678e-15, 1.36339e-15, 8.18033e-16, 2.726775e-16], [1.1455129, 2.67846e-15, 1.33923e-15, 8.03537e-16, 2.6784563e-16], [1.1481536, 2.63103e-15, 1.31551e-15, 7.89309e-16, 2.6310292e-16], [1.1508004, 2.58448e-15, 1.29224e-15, 7.75343e-16, 2.584478e-16], [1.1534533, 2.53879e-15, 1.26939e-15, 7.61636e-16, 2.5387871e-16], [1.1561122, 2.49394e-15, 1.24697e-15, 7.48182e-16, 2.4939413e-16], [1.1587774, 2.44993e-15, 1.22496e-15, 7.34978e-16, 2.4499255e-16], [1.1614486, 2.40673e-15, 1.20336e-15, 7.22018e-16, 2.4067251e-16], [1.164126, 2.36433e-15, 1.18216e-15, 7.09298e-16, 2.3643257e-16], [1.1668096, 2.32271e-15, 1.16136e-15, 6.96814e-16, 2.3227132e-16], [1.1694994, 2.28187e-15, 1.14094e-15, 6.84562e-16, 2.2818737e-16], [1.1721954, 2.24179e-15, 1.1209e-15, 6.72538e-16, 2.2417936e-16], [1.1748976, 2.20246e-15, 1.10123e-15, 6.60738e-16, 2.2024595e-16], [1.177606, 2.16386e-15, 1.08193e-15, 6.49158e-16, 2.1638584e-16], [1.1803206, 2.12598e-15, 1.06299e-15, 6.37793e-16, 2.1259775e-16], [1.1830416, 2.0888e-15, 1.0444e-15, 6.26641e-16, 2.0888042e-16], [1.1857687, 2.05233e-15, 1.02616e-15, 6.15698e-16, 2.0523261e-16], [1.1885022, 2.01653e-15, 1.00827e-15, 6.04959e-16, 2.0165312e-16], [1.191242, 1.98141e-15, 9.90704e-16, 5.94422e-16, 1.9814075e-16], [1.1939881, 1.94694e-15, 9.73472e-16, 5.84083e-16, 1.9469435e-16], [1.1967405, 1.91313e-15, 9.56564e-16, 5.73938e-16, 1.9131278e-16], [1.1994993, 1.87995e-15, 9.39975e-16, 5.63985e-16, 1.8799492e-16], [1.2022644, 1.8474e-15, 9.23698e-16, 5.54219e-16, 1.8473967e-16], [1.2050359, 1.81546e-15, 9.0773e-16, 5.44638e-16, 1.8154597e-16], [1.2078138, 1.78413e-15, 8.92064e-16, 5.35238e-16, 1.7841275e-16], [1.2105981, 1.75339e-15, 8.76695e-16, 5.26017e-16, 1.75339e-16], [1.2133889, 1.72324e-15, 8.61618e-16, 5.16971e-16, 1.7232369e-16], [1.216186, 1.69366e-15, 8.46829e-16, 5.08098e-16, 1.6936585e-16], [1.2189896, 1.66464e-15, 8.32322e-16, 4.99393e-16, 1.664645e-16], [1.2217997, 1.63619e-15, 8.18093e-16, 4.90856e-16, 1.636187e-16], [1.2246162, 1.60828e-15, 8.04138e-16, 4.82483e-16, 1.6082751e-16], [1.2274392, 1.5809e-15, 7.9045e-16, 4.7427e-16, 1.5809003e-16], [1.2302688, 1.55405e-15, 7.77027e-16, 4.66216e-16, 1.5540537e-16], [1.2331048, 1.52773e-15, 7.63863e-16, 4.58318e-16, 1.5277266e-16], [1.2359474, 1.50191e-15, 7.50955e-16, 4.50573e-16, 1.5019104e-16], [1.2387966, 1.4766e-15, 7.38298e-16, 4.42979e-16, 1.4765969e-16], [1.2416523, 1.45178e-15, 7.25889e-16, 4.35533e-16, 1.4517778e-16], [1.2445146, 1.42745e-15, 7.13723e-16, 4.28234e-16, 1.4274453e-16], [1.2473835, 1.40359e-15, 7.01796e-16, 4.21077e-16, 1.4035914e-16], [1.250259, 1.38021e-15, 6.90104e-16, 4.14063e-16, 1.3802087e-16], [1.2531412, 1.35729e-15, 6.78645e-16, 4.07187e-16, 1.3572897e-16], [1.25603, 1.33483e-15, 6.67414e-16, 4.00448e-16, 1.3348271e-16], [1.2589254, 1.31281e-15, 6.56407e-16, 3.93844e-16, 1.3128138e-16], [1.2618275, 1.29124e-15, 6.45621e-16, 3.87373e-16, 1.291243e-16], [1.2647363, 1.27011e-15, 6.35054e-16, 3.81032e-16, 1.2701078e-16], [1.2676519, 1.2494e-15, 6.24701e-16, 3.74821e-16, 1.2494017e-16], [1.2705741, 1.22912e-15, 6.14559e-16, 3.68735e-16, 1.2291183e-16], [1.2735031, 1.20925e-15, 6.04626e-16, 3.62775e-16, 1.2092513e-16], [1.2764388, 1.18979e-15, 5.94897e-16, 3.56938e-16, 1.1897947e-16], [1.2793813, 1.17074e-15, 5.85371e-16, 3.51223e-16, 1.1707426e-16], [1.2823306, 1.15209e-15, 5.76045e-16, 3.45627e-16, 1.1520892e-16], [1.2852867, 1.13383e-15, 5.66914e-16, 3.40149e-16, 1.1338288e-16], [1.2882496, 1.11596e-15, 5.57978e-16, 3.34787e-16, 1.1159562e-16], [1.2912193, 1.09847e-15, 5.49233e-16, 3.2954e-16, 1.0984659e-16], [1.2941958, 1.08135e-15, 5.40677e-16, 3.24406e-16, 1.081353e-16], [1.2971793, 1.06461e-15, 5.32306e-16, 3.19384e-16, 1.0646125e-16], [1.3001696, 1.04824e-15, 5.2412e-16, 3.14472e-16, 1.0482396e-16], [1.3031668, 1.03223e-15, 5.16115e-16, 3.09669e-16, 1.0322297e-16], [1.3061709, 1.01658e-15, 5.08289e-16, 3.04974e-16, 1.0165784e-16], [1.3091819, 1.00128e-15, 5.00641e-16, 3.00384e-16, 1.0012813e-16], [1.3121999, 9.86334e-16, 4.93167e-16, 2.959e-16, 9.8633437e-17], [1.3152248, 9.71734e-16, 4.85867e-16, 2.9152e-16, 9.717336e-17], [1.3182567, 9.57475e-16, 4.78738e-16, 2.87243e-16, 9.5747523e-17], [1.3212956, 9.43556e-16, 4.71778e-16, 2.83067e-16, 9.4355562e-17], [1.3243415, 9.29971e-16, 4.64986e-16, 2.78991e-16, 9.2997132e-17], [1.3273945, 9.16719e-16, 4.5836e-16, 2.75016e-16, 9.1671904e-17], [1.3304544, 9.03796e-16, 4.51898e-16, 2.71139e-16, 9.0379567e-17], [1.3335214, 8.91198e-16, 4.45599e-16, 2.67359e-16, 8.9119827e-17], [1.3365955, 8.78924e-16, 4.39462e-16, 2.63677e-16, 8.7892408e-17], [1.3396767, 8.66971e-16, 4.33485e-16, 2.60091e-16, 8.6697051e-17], [1.342765, 8.55335e-16, 4.27668e-16, 2.56601e-16, 8.5533517e-17], [1.3458604, 8.44016e-16, 4.22008e-16, 2.53205e-16, 8.4401585e-17], [1.3489629, 8.33011e-16, 4.16505e-16, 2.49903e-16, 8.3301051e-17], [1.3520726, 8.22317e-16, 4.11159e-16, 2.46695e-16, 8.2231733e-17], [1.3551894, 8.11935e-16, 4.05967e-16, 2.4358e-16, 8.1193467e-17], [1.3583134, 8.01861e-16, 4.00931e-16, 2.40558e-16, 8.0186111e-17], [1.3614447, 7.92095e-16, 3.96048e-16, 2.37629e-16, 7.9209543e-17], [1.3645831, 7.82637e-16, 3.91318e-16, 2.34791e-16, 7.8263661e-17], [1.3677288, 7.73484e-16, 3.86742e-16, 2.32045e-16, 7.7348388e-17], [1.3708818, 7.64637e-16, 3.82318e-16, 2.29391e-16, 7.6463667e-17], [1.374042, 7.56095e-16, 3.78047e-16, 2.26828e-16, 7.5609467e-17], [1.3772095, 7.47858e-16, 3.73929e-16, 2.24357e-16, 7.4785779e-17], [1.3803843, 7.39926e-16, 3.69963e-16, 2.21978e-16, 7.399262e-17], [1.3835664, 7.323e-16, 3.6615e-16, 2.1969e-16, 7.3230035e-17], [1.3867558, 7.24981e-16, 3.6249e-16, 2.17494e-16, 7.2498095e-17], [1.3899526, 7.17969e-16, 3.58984e-16, 2.15391e-16, 7.17969e-17], [1.3931568, 7.11266e-16, 3.55633e-16, 2.1338e-16, 7.1126579e-17], [1.3963684, 7.04873e-16, 3.52436e-16, 2.11462e-16, 7.0487294e-17], [1.3995873, 6.98792e-16, 3.49396e-16, 2.09638e-16, 6.9879239e-17], [1.4028137, 6.93026e-16, 3.46513e-16, 2.07908e-16, 6.9302644e-17], [1.4060475, 6.87578e-16, 3.43789e-16, 2.06273e-16, 6.8757776e-17], [1.4092888, 6.82449e-16, 3.41225e-16, 2.04735e-16, 6.8244939e-17], [1.4125375, 6.77645e-16, 3.38822e-16, 2.03293e-16, 6.776448e-17], [1.4157938, 6.73168e-16, 3.36584e-16, 2.0195e-16, 6.7316792e-17], [1.4190575, 6.69023e-16, 3.34512e-16, 2.00707e-16, 6.6902311e-17], [1.4223288, 6.65215e-16, 3.32608e-16, 1.99565e-16, 6.6521528e-17], [1.4256076, 6.6175e-16, 3.30875e-16, 1.98525e-16, 6.6174984e-17], [1.428894, 6.58633e-16, 3.29316e-16, 1.9759e-16, 6.5863281e-17], [1.4321879, 6.55871e-16, 3.27935e-16, 1.96761e-16, 6.5587083e-17], [1.4354894, 6.53471e-16, 3.26736e-16, 1.96041e-16, 6.5347123e-17], [1.4387986, 6.51442e-16, 3.25721e-16, 1.95433e-16, 6.5144205e-17], [1.4421154, 6.49792e-16, 3.24896e-16, 1.94938e-16, 6.4979216e-17], [1.4454398, 6.48531e-16, 3.24266e-16, 1.94559e-16, 6.4853128e-17], [1.4487719, 6.4767e-16, 3.23835e-16, 1.94301e-16, 6.476701e-17], [1.4521116, 6.4722e-16, 3.2361e-16, 1.94166e-16, 6.4722034e-17], [1.4554591, 6.47195e-16, 3.23597e-16, 1.94158e-16, 6.4719487e-17], [1.4588143, 6.47608e-16, 3.23804e-16, 1.94282e-16, 6.4760782e-17], [1.4621772, 6.48475e-16, 3.24237e-16, 1.94542e-16, 6.4847473e-17], [1.4655478, 6.49813e-16, 3.24906e-16, 1.94944e-16, 6.4981266e-17], [1.4689263, 6.5164e-16, 3.2582e-16, 1.95492e-16, 6.5164039e-17], [1.4723125, 6.53979e-16, 3.26989e-16, 1.96194e-16, 6.5397862e-17], [1.4757065, 6.5685e-16, 3.28425e-16, 1.97055e-16, 6.5685014e-17], [1.4791084, 6.6028e-16, 3.3014e-16, 1.98084e-16, 6.6028015e-17], [1.4825181, 6.64297e-16, 3.32148e-16, 1.99289e-16, 6.6429651e-17], [1.4859356, 6.6893e-16, 3.34465e-16, 2.00679e-16, 6.6893006e-17], [1.4893611, 6.74215e-16, 3.37108e-16, 2.02265e-16, 6.7421507e-17], [1.4927944, 6.8019e-16, 3.40095e-16, 2.04057e-16, 6.8018962e-17], [1.4962357, 6.86896e-16, 3.43448e-16, 2.06069e-16, 6.8689616e-17], [1.4996848, 6.94382e-16, 3.47191e-16, 2.08315e-16, 6.9438208e-17], [1.503142, 7.02701e-16, 3.5135e-16, 2.1081e-16, 7.0270051e-17], [1.5066071, 7.11911e-16, 3.55956e-16, 2.13573e-16, 7.1191104e-17], [1.5100802, 7.22081e-16, 3.6104e-16, 2.16624e-16, 7.2208083e-17], [1.5135612, 7.33286e-16, 3.66643e-16, 2.19986e-16, 7.332857e-17], [1.5170504, 7.45612e-16, 3.72806e-16, 2.23683e-16, 7.4561161e-17], [1.5205475, 7.59156e-16, 3.79578e-16, 2.27747e-16, 7.5915632e-17], [1.5240528, 7.74031e-16, 3.87016e-16, 2.32209e-16, 7.7403142e-17], [1.5275661, 7.90365e-16, 3.95182e-16, 2.37109e-16, 7.9036488e-17], [1.5310875, 8.08304e-16, 4.04152e-16, 2.42491e-16, 8.0830401e-17], [1.534617, 8.28019e-16, 4.1401e-16, 2.48406e-16, 8.2801926e-17], [1.5381546, 8.49709e-16, 4.24854e-16, 2.54913e-16, 8.4970887e-17], [1.5417005, 8.73605e-16, 4.36802e-16, 2.62081e-16, 8.7360464e-17], [1.5452544, 8.99979e-16, 4.4999e-16, 2.69994e-16, 8.9997929e-17], [1.5488166, 9.29156e-16, 4.64578e-16, 2.78747e-16, 9.2915583e-17], [1.552387, 9.6152e-16, 4.8076e-16, 2.88456e-16, 9.6151953e-17], [1.5559656, 9.97534e-16, 4.98767e-16, 2.9926e-16, 9.9753362e-17], [1.5595525, 1.03776e-15, 5.1888e-16, 3.11328e-16, 1.0377599e-16], [1.5631476, 1.08289e-15, 5.41443e-16, 3.24866e-16, 1.0828862e-16], [1.5667511, 1.13376e-15, 5.66882e-16, 3.40129e-16, 1.1337638e-16], [1.5703628, 1.19146e-15, 5.95729e-16, 3.57438e-16, 1.1914588e-16], [1.5739829, 1.25732e-15, 6.28662e-16, 3.77197e-16, 1.2573238e-16], [1.5776113, 1.3331e-15, 6.66551e-16, 3.99931e-16, 1.3331018e-16], [1.581248, 1.42108e-15, 7.10538e-16, 4.26323e-16, 1.4210766e-16], [1.5848932, 1.5243e-15, 7.6215e-16, 4.5729e-16, 1.5243004e-16], [1.5885467, 1.64695e-15, 8.23473e-16, 4.94084e-16, 1.6469462e-16], [1.5922087, 1.79487e-15, 8.97437e-16, 5.38462e-16, 1.7948737e-16], [1.5958791, 1.97658e-15, 9.88288e-16, 5.92973e-16, 1.9765766e-16], [1.599558, 2.20485e-15, 1.10242e-15, 6.61455e-16, 2.2048496e-16], [1.6032454, 2.4999e-15, 1.24995e-15, 7.49971e-16, 2.4999035e-16], [1.6069413, 2.89565e-15, 1.44782e-15, 8.68695e-16, 2.8956486e-16], [1.6106456, 3.45367e-15, 1.72683e-15, 1.0361e-15, 3.4536685e-16], [1.6143586, 4.29868e-15, 2.14934e-15, 1.2896e-15, 4.2986807e-16], [1.61808, 5.72732e-15, 2.86366e-15, 1.7182e-15, 5.7273189e-16], [1.6218101, 8.6601e-15, 4.33005e-15, 2.59803e-15, 8.6600959e-16], [1.6255488, 1.81002e-14, 9.05009e-15, 5.43005e-15, 1.810018e-15], [1.629296, 1.64906e-13, 8.24532e-14, 4.94719e-14, 1.6490642e-14], [1.6330519, 1.46074e-14, 7.30369e-15, 4.38222e-15, 1.4607385e-15], [1.6368165, 7.57974e-15, 3.78987e-15, 2.27392e-15, 7.579735e-16], [1.6405898, 5.08966e-15, 2.54483e-15, 1.5269e-15, 5.0896572e-16], [1.6443717, 3.81542e-15, 1.90771e-15, 1.14463e-15, 3.8154228e-16], [1.6481624, 3.04154e-15, 1.52077e-15, 9.12463e-16, 3.0415419e-16], [1.6519618, 2.52184e-15, 1.26092e-15, 7.56553e-16, 2.5218438e-16], [1.65577, 2.14888e-15, 1.07444e-15, 6.44665e-16, 2.1488828e-16], [1.6595869, 1.86829e-15, 9.34143e-16, 5.60486e-16, 1.8682862e-16], [1.6634127, 1.64959e-15, 8.24793e-16, 4.94876e-16, 1.6495866e-16], [1.6672472, 1.47439e-15, 7.37193e-16, 4.42316e-16, 1.4743858e-16], [1.6710906, 1.33092e-15, 6.65459e-16, 3.99275e-16, 1.3309172e-16], [1.6749429, 1.2113e-15, 6.05652e-16, 3.63391e-16, 1.211305e-16], [1.678804, 1.11008e-15, 5.55039e-16, 3.33024e-16, 1.1100788e-16], [1.6826741, 1.02332e-15, 5.1166e-16, 3.06996e-16, 1.0233207e-16], [1.686553, 9.48151e-16, 4.74076e-16, 2.84445e-16, 9.4815109e-17], [1.6904409, 8.82406e-16, 4.41203e-16, 2.64722e-16, 8.8240609e-17], [1.6943378, 8.24428e-16, 4.12214e-16, 2.47328e-16, 8.2442829e-17], [1.6982437, 7.72927e-16, 3.86463e-16, 2.31878e-16, 7.729268e-17], [1.7021585, 7.26881e-16, 3.63441e-16, 2.18064e-16, 7.2688127e-17], [1.7060824, 6.85475e-16, 3.42737e-16, 2.05642e-16, 6.8547462e-17], [1.7100153, 6.48045e-16, 3.24022e-16, 1.94413e-16, 6.4804481e-17], [1.7139573, 6.1405e-16, 3.07025e-16, 1.84215e-16, 6.1404979e-17], [1.7179084, 5.83042e-16, 2.91521e-16, 1.74912e-16, 5.8304151e-17], [1.7218686, 5.54646e-16, 2.77323e-16, 1.66394e-16, 5.5464639e-17], [1.7258379, 5.2855e-16, 2.64275e-16, 1.58565e-16, 5.285505e-17], [1.7298164, 5.04488e-16, 2.52244e-16, 1.51346e-16, 5.0448813e-17], [1.733804, 4.82233e-16, 2.41116e-16, 1.4467e-16, 4.8223287e-17], [1.7378008, 4.61591e-16, 2.30795e-16, 1.38477e-16, 4.6159063e-17], [1.7418069, 4.42394e-16, 2.21197e-16, 1.32718e-16, 4.4239407e-17], [1.7458222, 4.24498e-16, 2.12249e-16, 1.27349e-16, 4.2449814e-17], [1.7498467, 4.07777e-16, 2.03888e-16, 1.22333e-16, 4.0777654e-17], [1.7538805, 3.92119e-16, 1.96059e-16, 1.17636e-16, 3.9211874e-17], [1.7579236, 3.77428e-16, 1.88714e-16, 1.13228e-16, 3.7742762e-17], [1.761976, 3.63618e-16, 1.81809e-16, 1.09085e-16, 3.6361753e-17], [1.7660378, 3.50613e-16, 1.75306e-16, 1.05184e-16, 3.5061259e-17], [1.770109, 3.38345e-16, 1.69173e-16, 1.01504e-16, 3.3834537e-17], [1.7741895, 3.26756e-16, 1.63378e-16, 9.80267e-17, 3.2675577e-17], [1.7782794, 3.1579e-16, 1.57895e-16, 9.4737e-17, 3.1579002e-17], [1.7823788, 3.054e-16, 1.527e-16, 9.162e-17, 3.0539988e-17], [1.7864876, 2.95542e-16, 1.47771e-16, 8.86626e-17, 2.9554195e-17], [1.7906059, 2.86177e-16, 1.43089e-16, 8.58531e-17, 2.8617707e-17], [1.7947336, 2.7727e-16, 1.38635e-16, 8.3181e-17, 2.7726986e-17], [1.7988709, 2.68788e-16, 1.34394e-16, 8.06365e-17, 2.687882e-17], [1.8030177, 2.60703e-16, 1.30351e-16, 7.82109e-17, 2.6070293e-17], [1.8071741, 2.52987e-16, 1.26494e-16, 7.58962e-17, 2.5298748e-17], [1.8113401, 2.45618e-16, 1.22809e-16, 7.36853e-17, 2.4561763e-17], [1.8155157, 2.38571e-16, 1.19286e-16, 7.15714e-17, 2.3857119e-17], [1.8197009, 2.31828e-16, 1.15914e-16, 6.95484e-17, 2.3182786e-17], [1.8238957, 2.25369e-16, 1.12684e-16, 6.76107e-17, 2.25369e-17], [1.8281002, 2.19177e-16, 1.09589e-16, 6.57532e-17, 2.1917746e-17], [1.8323144, 2.13237e-16, 1.06619e-16, 6.39712e-17, 2.1323746e-17], [1.8365383, 2.07534e-16, 1.03767e-16, 6.22603e-17, 2.0753442e-17], [1.840772, 2.02055e-16, 1.01027e-16, 6.06165e-17, 2.0205488e-17], [1.8450154, 1.96786e-16, 9.83932e-17, 5.90359e-17, 1.9678638e-17], [1.8492686, 1.91717e-16, 9.58587e-17, 5.75152e-17, 1.9171738e-17], [1.8535316, 1.86837e-16, 9.34186e-17, 5.60511e-17, 1.8683715e-17], [1.8578045, 1.82136e-16, 9.10679e-17, 5.46407e-17, 1.8213574e-17], [1.8620871, 1.77604e-16, 8.88019e-17, 5.32812e-17, 1.7760387e-17], [1.8663797, 1.73233e-16, 8.66165e-17, 5.19699e-17, 1.7323292e-17], [1.8706821, 1.69015e-16, 8.45074e-17, 5.07044e-17, 1.690148e-17], [1.8749945, 1.64942e-16, 8.2471e-17, 4.94826e-17, 1.6494202e-17], [1.8793168, 1.61008e-16, 8.05038e-17, 4.83023e-17, 1.6100751e-17], [1.8836491, 1.57205e-16, 7.86023e-17, 4.71614e-17, 1.5720469e-17], [1.8879913, 1.53527e-16, 7.67637e-17, 4.60582e-17, 1.5352739e-17], [1.8923436, 1.4997e-16, 7.49849e-17, 4.49909e-17, 1.4996982e-17], [1.8967059, 1.46527e-16, 7.32633e-17, 4.3958e-17, 1.4652652e-17], [1.9010783, 1.43192e-16, 7.15962e-17, 4.29577e-17, 1.431924e-17], [1.9054607, 1.39963e-16, 6.99813e-17, 4.19888e-17, 1.3996262e-17], [1.9098533, 1.36833e-16, 6.84163e-17, 4.10498e-17, 1.3683267e-17], [1.9142559, 1.33798e-16, 6.68991e-17, 4.01395e-17, 1.3379827e-17], [1.9186687, 1.30855e-16, 6.54277e-17, 3.92566e-17, 1.3085538e-17], [1.9230917, 1.28e-16, 6.40001e-17, 3.84001e-17, 1.280002e-17], [1.9275249, 1.25229e-16, 6.26146e-17, 3.75687e-17, 1.2522913e-17], [1.9319683, 1.22539e-16, 6.12694e-17, 3.67616e-17, 1.2253877e-17], [1.936422, 1.19926e-16, 5.9963e-17, 3.59778e-17, 1.199259e-17], [1.9408859, 1.17387e-16, 5.86937e-17, 3.52162e-17, 1.1738747e-17], [1.9453601, 1.14921e-16, 5.74603e-17, 3.44762e-17, 1.1492058e-17], [1.9498446, 1.12522e-16, 5.62612e-17, 3.37567e-17, 1.1252248e-17], [1.9543395, 1.10191e-16, 5.50953e-17, 3.30572e-17, 1.1019058e-17], [1.9588447, 1.07922e-16, 5.39612e-17, 3.23767e-17, 1.0792239e-17], [1.9633603, 1.05716e-16, 5.28578e-17, 3.17147e-17, 1.0571556e-17], [1.9678863, 1.03568e-16, 5.17839e-17, 3.10704e-17, 1.0356786e-17], [1.9724227, 1.01477e-16, 5.07386e-17, 3.04431e-17, 1.0147714e-17], [1.9769696, 9.94414e-17, 4.97207e-17, 2.98324e-17, 9.944139e-18], [1.981527, 9.74587e-17, 4.87293e-17, 2.92376e-17, 9.7458661e-18], [1.9860949, 9.55271e-17, 4.77636e-17, 2.86581e-17, 9.5527112e-18], [1.9906733, 9.3645e-17, 4.68225e-17, 2.80935e-17, 9.3644979e-18], [1.9952623, 9.18106e-17, 4.59053e-17, 2.75432e-17, 9.1810582e-18], [1.9998619, 9.00223e-17, 4.50112e-17, 2.70067e-17, 9.0022314e-18], [2.004472, 8.82786e-17, 4.41393e-17, 2.64836e-17, 8.8278638e-18], [2.0090928, 8.65781e-17, 4.3289e-17, 2.59734e-17, 8.6578087e-18], [2.0137242, 8.49192e-17, 4.24596e-17, 2.54758e-17, 8.4919253e-18], [2.0183664, 8.33008e-17, 4.16504e-17, 2.49902e-17, 8.3300793e-18], [2.0230192, 8.17214e-17, 4.08607e-17, 2.45164e-17, 8.1721416e-18], [2.0276827, 8.01799e-17, 4.00899e-17, 2.4054e-17, 8.0179887e-18], [2.032357, 7.8675e-17, 3.93375e-17, 2.36025e-17, 7.8675022e-18], [2.0370421, 7.72057e-17, 3.86028e-17, 2.31617e-17, 7.7205685e-18], [2.0417379, 7.57708e-17, 3.78854e-17, 2.27312e-17, 7.5770786e-18], [2.0464446, 7.43693e-17, 3.71846e-17, 2.23108e-17, 7.4369277e-18], [2.0511622, 7.30001e-17, 3.65001e-17, 2.19e-17, 7.3000154e-18], [2.0558906, 7.16624e-17, 3.58312e-17, 2.14987e-17, 7.1662449e-18], [2.0606299, 7.03552e-17, 3.51776e-17, 2.11066e-17, 7.0355235e-18], [2.0653802, 6.90776e-17, 3.45388e-17, 2.07233e-17, 6.9077618e-18], [2.0701413, 6.78287e-17, 3.39144e-17, 2.03486e-17, 6.7828738e-18], [2.0749135, 6.66078e-17, 3.33039e-17, 1.99823e-17, 6.6607768e-18], [2.0796967, 6.54139e-17, 3.2707e-17, 1.96242e-17, 6.5413911e-18], [2.0844909, 6.42464e-17, 3.21232e-17, 1.92739e-17, 6.42464e-18], [2.0892961, 6.31045e-17, 3.15522e-17, 1.89313e-17, 6.3104496e-18], [2.0941125, 6.19875e-17, 3.09937e-17, 1.85962e-17, 6.1987486e-18], [2.0989399, 6.08947e-17, 3.04473e-17, 1.82684e-17, 6.0894682e-18], [2.1037784, 5.98254e-17, 2.99127e-17, 1.79476e-17, 5.9825422e-18], [2.1086281, 5.87791e-17, 2.93895e-17, 1.76337e-17, 5.8779066e-18], [2.113489, 5.7755e-17, 2.88775e-17, 1.73265e-17, 5.7754996e-18], [2.1183611, 5.67526e-17, 2.83763e-17, 1.70258e-17, 5.6752616e-18], [2.1232445, 5.57713e-17, 2.78857e-17, 1.67314e-17, 5.5771349e-18], [2.128139, 5.48106e-17, 2.74053e-17, 1.64432e-17, 5.481064e-18], [2.1330449, 5.38699e-17, 2.6935e-17, 1.6161e-17, 5.386995e-18], [2.1379621, 5.29487e-17, 2.64744e-17, 1.58846e-17, 5.294876e-18], [2.1428906, 5.20466e-17, 2.60233e-17, 1.5614e-17, 5.2046565e-18], [2.1478305, 5.11629e-17, 2.55814e-17, 1.53489e-17, 5.116288e-18], [2.1527817, 5.02972e-17, 2.51486e-17, 1.50892e-17, 5.0297234e-18], [2.1577444, 4.94492e-17, 2.47246e-17, 1.48347e-17, 4.9449172e-18], [2.1627185, 4.86182e-17, 2.43091e-17, 1.45855e-17, 4.8618251e-18], [2.1677041, 4.7804e-17, 2.3902e-17, 1.43412e-17, 4.7804044e-18], [2.1727012, 4.70061e-17, 2.35031e-17, 1.41018e-17, 4.7006138e-18], [2.1777098, 4.62241e-17, 2.31121e-17, 1.38672e-17, 4.6224132e-18], [2.1827299, 4.54576e-17, 2.27288e-17, 1.36373e-17, 4.5457636e-18], [2.1877616, 4.47063e-17, 2.23531e-17, 1.34119e-17, 4.4706274e-18], [2.1928049, 4.39697e-17, 2.19848e-17, 1.31909e-17, 4.396968e-18], [2.1978599, 4.32475e-17, 2.16237e-17, 1.29742e-17, 4.32475e-18], [2.2029265, 4.25394e-17, 2.12697e-17, 1.27618e-17, 4.2539391e-18], [2.2080047, 4.1845e-17, 2.09225e-17, 1.25535e-17, 4.1845017e-18], [2.2130947, 4.1164e-17, 2.0582e-17, 1.23492e-17, 4.1164056e-18], [2.2181964, 4.04962e-17, 2.02481e-17, 1.21489e-17, 4.0496193e-18], [2.2233099, 3.98411e-17, 1.99206e-17, 1.19523e-17, 3.9841122e-18], [2.2284351, 3.91985e-17, 1.95993e-17, 1.17596e-17, 3.9198548e-18], [2.2335722, 3.85682e-17, 1.92841e-17, 1.15705e-17, 3.8568182e-18], [2.2387211, 3.79497e-17, 1.89749e-17, 1.13849e-17, 3.7949744e-18], [2.2438819, 3.73429e-17, 1.86715e-17, 1.12029e-17, 3.7342963e-18], [2.2490546, 3.67476e-17, 1.83738e-17, 1.10243e-17, 3.6747573e-18], [2.2542392, 3.61633e-17, 1.80817e-17, 1.0849e-17, 3.6163319e-18], [2.2594358, 3.55899e-17, 1.7795e-17, 1.0677e-17, 3.5589949e-18], [2.2646443, 3.50272e-17, 1.75136e-17, 1.05082e-17, 3.5027221e-18], [2.2698649, 3.44749e-17, 1.72374e-17, 1.03425e-17, 3.4474899e-18], [2.2750974, 3.39327e-17, 1.69664e-17, 1.01798e-17, 3.3932752e-18], [2.2803421, 3.34005e-17, 1.67003e-17, 1.00202e-17, 3.3400557e-18], [2.2855988, 3.28781e-17, 1.6439e-17, 9.86343e-18, 3.2878095e-18], [2.2908677, 3.23651e-17, 1.61826e-17, 9.70954e-18, 3.2365155e-18], [2.2961486, 3.18615e-17, 1.59308e-17, 9.55846e-18, 3.1861531e-18], [2.3014418, 3.1367e-17, 1.56835e-17, 9.4101e-18, 3.136702e-18], [2.3067472, 3.08814e-17, 1.54407e-17, 9.26443e-18, 3.0881428e-18], [2.3120648, 3.04045e-17, 1.52023e-17, 9.12137e-18, 3.0404563e-18], [2.3173946, 2.99362e-17, 1.49681e-17, 8.98087e-18, 2.9936241e-18], [2.3227368, 2.94763e-17, 1.47381e-17, 8.84288e-18, 2.9476279e-18], [2.3280913, 2.90245e-17, 1.45122e-17, 8.70735e-18, 2.9024501e-18], [2.3334581, 2.85807e-17, 1.42904e-17, 8.57422e-18, 2.8580736e-18], [2.3388372, 2.81448e-17, 1.40724e-17, 8.44344e-18, 2.8144816e-18], [2.3442288, 2.77166e-17, 1.38583e-17, 8.31497e-18, 2.7716577e-18], [2.3496328, 2.72958e-17, 1.36479e-17, 8.18876e-18, 2.729586e-18], [2.3550493, 2.68825e-17, 1.34413e-17, 8.06475e-18, 2.688251e-18], [2.3604782, 2.64764e-17, 1.32382e-17, 7.94291e-18, 2.6476375e-18], [2.3659197, 2.60773e-17, 1.30386e-17, 7.82319e-18, 2.6077307e-18], [2.3713737, 2.56851e-17, 1.28426e-17, 7.70555e-18, 2.5685162e-18], [2.3768403, 2.52998e-17, 1.26499e-17, 7.58994e-18, 2.52998e-18], [2.3823195, 2.49211e-17, 1.24605e-17, 7.47632e-18, 2.4921082e-18], [2.3878113, 2.45489e-17, 1.22744e-17, 7.36466e-18, 2.4548876e-18], [2.3933158, 2.4183e-17, 1.20915e-17, 7.25491e-18, 2.4183049e-18], [2.3988329, 2.38235e-17, 1.19117e-17, 7.14704e-18, 2.3823476e-18], [2.4043628, 2.347e-17, 1.1735e-17, 7.04101e-18, 2.347003e-18], [2.4099054, 2.31226e-17, 1.15613e-17, 6.93677e-18, 2.312259e-18], [2.4154608, 2.2781e-17, 1.13905e-17, 6.83431e-18, 2.2781037e-18], [2.421029, 2.24452e-17, 1.12226e-17, 6.73357e-18, 2.2445255e-18], [2.4266101, 2.21151e-17, 1.10576e-17, 6.63454e-18, 2.2115131e-18], [2.432204, 2.17905e-17, 1.08953e-17, 6.53716e-18, 2.1790554e-18], [2.4378108, 2.14714e-17, 1.07357e-17, 6.44142e-18, 2.1471415e-18], [2.4434306, 2.11576e-17, 1.05788e-17, 6.34728e-18, 2.115761e-18], [2.4490632, 2.0849e-17, 1.04245e-17, 6.25471e-18, 2.0849034e-18], [2.4547089, 2.05456e-17, 1.02728e-17, 6.16367e-18, 2.0545586e-18], [2.4603676, 2.02471e-17, 1.01236e-17, 6.07415e-18, 2.0247169e-18], [2.4660393, 1.99537e-17, 9.97684e-18, 5.9861e-18, 1.9953685e-18], [2.4717241, 1.9665e-17, 9.83252e-18, 5.89951e-18, 1.966504e-18], [2.4774221, 1.93811e-17, 9.69057e-18, 5.81434e-18, 1.9381141e-18], [2.4831331, 1.91019e-17, 9.55095e-18, 5.73057e-18, 1.91019e-18], [2.4888573, 1.88272e-17, 9.41361e-18, 5.64817e-18, 1.8827227e-18], [2.4945947, 1.8557e-17, 9.27851e-18, 5.56711e-18, 1.8557037e-18], [2.5003454, 1.82912e-17, 9.14562e-18, 5.48737e-18, 1.8291244e-18], [2.5061093, 1.80297e-17, 9.01488e-18, 5.40893e-18, 1.8029768e-18], [2.5118864, 1.77725e-17, 8.88626e-18, 5.33176e-18, 1.7772526e-18], [2.5176769, 1.75194e-17, 8.75972e-18, 5.25583e-18, 1.751944e-18], [2.5234808, 1.72704e-17, 8.63521e-18, 5.18113e-18, 1.7270433e-18], [2.529298, 1.70254e-17, 8.51271e-18, 5.10763e-18, 1.702543e-18], [2.5351286, 1.67843e-17, 8.39217e-18, 5.0353e-18, 1.6784355e-18], [2.5409727, 1.65471e-17, 8.27356e-18, 4.96414e-18, 1.6547138e-18], [2.5468303, 1.63137e-17, 8.15685e-18, 4.89411e-18, 1.6313707e-18], [2.5527013, 1.6084e-17, 8.04199e-18, 4.8252e-18, 1.6083993e-18], [2.5585859, 1.58579e-17, 7.92896e-18, 4.75738e-18, 1.5857928e-18], [2.564484, 1.56354e-17, 7.81772e-18, 4.69063e-18, 1.5635445e-18], [2.5703958, 1.54165e-17, 7.70824e-18, 4.62494e-18, 1.5416481e-18], [2.5763212, 1.52009e-17, 7.60048e-18, 4.56029e-18, 1.520097e-18], [2.5822602, 1.49888e-17, 7.49442e-18, 4.49665e-18, 1.4988851e-18], [2.5882129, 1.478e-17, 7.39003e-18, 4.43402e-18, 1.4780063e-18], [2.5941794, 1.45745e-17, 7.28727e-18, 4.37236e-18, 1.4574546e-18], [2.6001596, 1.43722e-17, 7.18612e-18, 4.31167e-18, 1.4372241e-18], [2.6061535, 1.41731e-17, 7.08654e-18, 4.25192e-18, 1.4173092e-18], [2.6121614, 1.3977e-17, 6.98852e-18, 4.19311e-18, 1.3977041e-18], [2.618183, 1.3784e-17, 6.89201e-18, 4.13521e-18, 1.3784035e-18], [2.6242185, 1.3594e-17, 6.797e-18, 4.0782e-18, 1.3594018e-18], [2.630268, 1.34069e-17, 6.70347e-18, 4.02208e-18, 1.3406939e-18], [2.6363314, 1.32227e-17, 6.61137e-18, 3.96682e-18, 1.3222746e-18], [2.6424088, 1.30414e-17, 6.52069e-18, 3.91241e-18, 1.3041388e-18], [2.6485001, 1.28628e-17, 6.4314e-18, 3.85884e-18, 1.2862815e-18], [2.6546056, 1.2687e-17, 6.34348e-18, 3.80609e-18, 1.2686978e-18], [2.6607251, 1.25138e-17, 6.25691e-18, 3.75415e-18, 1.2513831e-18], [2.6668587, 1.23433e-17, 6.17166e-18, 3.703e-18, 1.2343326e-18], [2.6730064, 1.21754e-17, 6.0877e-18, 3.65262e-18, 1.2175418e-18], [2.6791683, 1.201e-17, 6.00503e-18, 3.60302e-18, 1.2010061e-18], [2.6853444, 1.18472e-17, 5.9236e-18, 3.55416e-18, 1.1847212e-18], [2.6915348, 1.16868e-17, 5.84341e-18, 3.50605e-18, 1.1686828e-18], [2.6977394, 1.15288e-17, 5.76443e-18, 3.45866e-18, 1.1528865e-18], [2.7039584, 1.13733e-17, 5.68664e-18, 3.41198e-18, 1.1373284e-18], [2.7101916, 1.122e-17, 5.61002e-18, 3.36601e-18, 1.1220042e-18], [2.7164393, 1.10691e-17, 5.53455e-18, 3.32073e-18, 1.1069101e-18], [2.7227013, 1.09204e-17, 5.46021e-18, 3.27612e-18, 1.0920421e-18], [2.7289778, 1.07739e-17, 5.38698e-18, 3.23219e-18, 1.0773964e-18], [2.7352687, 1.06297e-17, 5.31484e-18, 3.1889e-18, 1.0629692e-18], [2.7415742, 1.04875e-17, 5.24378e-18, 3.14627e-18, 1.0487568e-18], [2.7478942, 1.03475e-17, 5.17377e-18, 3.10426e-18, 1.0347557e-18], [2.7542287, 1.02096e-17, 5.10481e-18, 3.06288e-18, 1.0209622e-18], [2.7605779, 1.00737e-17, 5.03686e-18, 3.02212e-18, 1.007373e-18], [2.7669416, 9.93982e-18, 4.96992e-18, 2.98195e-18, 9.9398448e-19], [2.7733201, 9.80791e-18, 4.90396e-18, 2.94238e-18, 9.8079344e-19], [2.7797133, 9.67794e-18, 4.83898e-18, 2.90339e-18, 9.6779655e-19], [2.7861212, 9.54988e-18, 4.77495e-18, 2.86497e-18, 9.5499059e-19], [2.7925438, 9.4237e-18, 4.71186e-18, 2.82711e-18, 9.423724e-19], [2.7989813, 9.29936e-18, 4.64969e-18, 2.78981e-18, 9.2993888e-19], [2.8054336, 9.17685e-18, 4.58843e-18, 2.75306e-18, 9.1768697e-19], [2.8119008, 9.05611e-18, 4.52806e-18, 2.71684e-18, 9.0561371e-19], [2.8183829, 8.93714e-18, 4.46858e-18, 2.68115e-18, 8.9371615e-19], [2.82488, 8.81989e-18, 4.40995e-18, 2.64597e-18, 8.8199141e-19], [2.831392, 8.70434e-18, 4.35218e-18, 2.61131e-18, 8.7043668e-19], [2.837919, 8.59047e-18, 4.29524e-18, 2.57714e-18, 8.5904919e-19], [2.8444611, 8.47824e-18, 4.23913e-18, 2.54348e-18, 8.478262e-19], [2.8510183, 8.36763e-18, 4.18382e-18, 2.51029e-18, 8.3676506e-19], [2.8575905, 8.25861e-18, 4.12931e-18, 2.47759e-18, 8.2586313e-19], [2.864178, 8.15115e-18, 4.07558e-18, 2.44535e-18, 8.1511786e-19], [2.8707806, 8.04524e-18, 4.02263e-18, 2.41358e-18, 8.0452671e-19], [2.8773984, 7.94085e-18, 3.97043e-18, 2.38226e-18, 7.940872e-19], [2.8840315, 7.83794e-18, 3.91898e-18, 2.35139e-18, 7.8379689e-19], [2.8906799, 7.73651e-18, 3.86826e-18, 2.32096e-18, 7.736534e-19], [2.8973436, 7.63652e-18, 3.81827e-18, 2.29096e-18, 7.6365438e-19], [2.9040227, 7.53795e-18, 3.76898e-18, 2.26139e-18, 7.5379753e-19], [2.9107171, 7.44078e-18, 3.7204e-18, 2.23224e-18, 7.4408057e-19], [2.917427, 7.34499e-18, 3.6725e-18, 2.2035e-18, 7.3450129e-19], [2.9241524, 7.25055e-18, 3.62528e-18, 2.17517e-18, 7.250575e-19], [2.9308932, 7.15744e-18, 3.57873e-18, 2.14724e-18, 7.1574707e-19], [2.9376497, 7.06565e-18, 3.53283e-18, 2.1197e-18, 7.0656788e-19], [2.9444216, 6.97515e-18, 3.48758e-18, 2.09255e-18, 6.9751786e-19], [2.9512092, 6.88592e-18, 3.44297e-18, 2.06578e-18, 6.8859499e-19], [2.9580125, 6.79795e-18, 3.39898e-18, 2.03939e-18, 6.7979727e-19], [2.9648314, 6.7112e-18, 3.35561e-18, 2.01336e-18, 6.7112273e-19], [2.971666, 6.62567e-18, 3.31284e-18, 1.98771e-18, 6.6256946e-19], [2.9785164, 6.54133e-18, 3.27067e-18, 1.9624e-18, 6.5413556e-19], [2.9853826, 6.45816e-18, 3.22909e-18, 1.93745e-18, 6.4581917e-19], [2.9922646, 6.37616e-18, 3.18809e-18, 1.91285e-18, 6.3761847e-19], [2.9991625, 6.29529e-18, 3.14765e-18, 1.88859e-18, 6.2953166e-19], [3.0060763, 6.21554e-18, 3.10778e-18, 1.86467e-18, 6.2155698e-19], [3.013006, 6.1369e-18, 3.06846e-18, 1.84107e-18, 6.136927e-19], [3.0199517, 6.05934e-18, 3.02968e-18, 1.81781e-18, 6.0593712e-19], [3.0269134, 5.98286e-18, 2.99144e-18, 1.79486e-18, 5.9828855e-19], [3.0338912, 5.90743e-18, 2.95372e-18, 1.77223e-18, 5.9074537e-19], [3.040885, 5.83303e-18, 2.91652e-18, 1.74991e-18, 5.8330595e-19], [3.047895, 5.75966e-18, 2.87984e-18, 1.7279e-18, 5.7596871e-19], [3.0549211, 5.68729e-18, 2.84365e-18, 1.70619e-18, 5.6873208e-19], [3.0619634, 5.61592e-18, 2.80797e-18, 1.68478e-18, 5.6159454e-19], [3.069022, 5.54552e-18, 2.77277e-18, 1.66366e-18, 5.5455457e-19], [3.0760968, 5.47608e-18, 2.73805e-18, 1.64283e-18, 5.476107e-19], [3.083188, 5.40759e-18, 2.7038e-18, 1.62228e-18, 5.4076147e-19], [3.0902954, 5.34003e-18, 2.67002e-18, 1.60201e-18, 5.3400544e-19], [3.0974193, 5.27338e-18, 2.6367e-18, 1.58202e-18, 5.2734121e-19], [3.1045596, 5.20765e-18, 2.60383e-18, 1.5623e-18, 5.207674e-19], [3.1117163, 5.1428e-18, 2.57141e-18, 1.54284e-18, 5.1428264e-19], [3.1188896, 5.07883e-18, 2.53942e-18, 1.52365e-18, 5.0788561e-19], [3.1260794, 5.01572e-18, 2.50787e-18, 1.50472e-18, 5.0157497e-19], [3.1332857, 4.95347e-18, 2.47674e-18, 1.48604e-18, 4.9534945e-19], [3.1405087, 4.89205e-18, 2.44603e-18, 1.46762e-18, 4.8920777e-19], [3.1477483, 4.83146e-18, 2.41574e-18, 1.44944e-18, 4.8314869e-19], [3.1550046, 4.77168e-18, 2.38585e-18, 1.43151e-18, 4.7717096e-19], [3.1622777, 4.7127e-18, 2.35636e-18, 1.41382e-18, 4.7127339e-19], [3.1695675, 4.65452e-18, 2.32727e-18, 1.39636e-18, 4.6545478e-19], [3.1768741, 4.59711e-18, 2.29856e-18, 1.37914e-18, 4.5971398e-19], [3.1841975, 4.54047e-18, 2.27024e-18, 1.36215e-18, 4.5404982e-19], [3.1915379, 4.48458e-18, 2.2423e-18, 1.34538e-18, 4.4846118e-19], [3.1988951, 4.42944e-18, 2.21473e-18, 1.32884e-18, 4.4294694e-19], [3.2062693, 4.37503e-18, 2.18752e-18, 1.31251e-18, 4.3750602e-19], [3.2136605, 4.32134e-18, 2.16068e-18, 1.29641e-18, 4.3213734e-19], [3.2210688, 4.26837e-18, 2.13419e-18, 1.28052e-18, 4.2683983e-19], [3.2284941, 4.21609e-18, 2.10806e-18, 1.26483e-18, 4.2161247e-19], [3.2359366, 4.16451e-18, 2.08226e-18, 1.24936e-18, 4.1645422e-19], [3.2433962, 4.11361e-18, 2.05681e-18, 1.23409e-18, 4.1136408e-19], [3.250873, 4.06338e-18, 2.0317e-18, 1.21902e-18, 4.0634106e-19], [3.258367, 4.01381e-18, 2.00691e-18, 1.20415e-18, 4.0138418e-19], [3.2658783, 3.96489e-18, 1.98246e-18, 1.18947e-18, 3.9649249e-19], [3.2734069, 3.91662e-18, 1.95832e-18, 1.17499e-18, 3.9166504e-19], [3.2809529, 3.86898e-18, 1.9345e-18, 1.1607e-18, 3.8690091e-19], [3.2885163, 3.82196e-18, 1.91099e-18, 1.14659e-18, 3.8219918e-19], [3.2960971, 3.77556e-18, 1.88779e-18, 1.13267e-18, 3.7755895e-19], [3.3036954, 3.72976e-18, 1.86489e-18, 1.11893e-18, 3.7297934e-19], [3.3113112, 3.68456e-18, 1.84229e-18, 1.10537e-18, 3.6845949e-19], [3.3189446, 3.63995e-18, 1.81999e-18, 1.09199e-18, 3.6399852e-19], [3.3265955, 3.59592e-18, 1.79797e-18, 1.07878e-18, 3.5959561e-19], [3.3342641, 3.55247e-18, 1.77624e-18, 1.06575e-18, 3.5524992e-19], [3.3419504, 3.50957e-18, 1.7548e-18, 1.05288e-18, 3.5096064e-19], [3.3496544, 3.46724e-18, 1.73363e-18, 1.04018e-18, 3.4672697e-19], [3.3573761, 3.42545e-18, 1.71273e-18, 1.02764e-18, 3.4254811e-19], [3.3651157, 3.3842e-18, 1.69211e-18, 1.01527e-18, 3.3842328e-19], [3.3728731, 3.34348e-18, 1.67175e-18, 1.00305e-18, 3.3435173e-19], [3.3806484, 3.30329e-18, 1.65166e-18, 9.90994e-19, 3.303327e-19], [3.3884416, 3.26362e-18, 1.63182e-18, 9.79092e-19, 3.2636544e-19], [3.3962527, 3.22446e-18, 1.61224e-18, 9.67344e-19, 3.2244923e-19], [3.4040819, 3.1858e-18, 1.59291e-18, 9.55746e-19, 3.1858336e-19], [3.4119291, 3.14764e-18, 1.57383e-18, 9.44297e-19, 3.147671e-19], [3.4197944, 3.10996e-18, 1.55499e-18, 9.32995e-19, 3.1099977e-19], [3.4276779, 3.07277e-18, 1.5364e-18, 9.21838e-19, 3.0728068e-19], [3.4355795, 3.03606e-18, 1.51804e-18, 9.10823e-19, 3.0360916e-19], [3.4434993, 2.99981e-18, 1.49992e-18, 8.9995e-19, 2.9998454e-19], [3.4514374, 2.96403e-18, 1.48202e-18, 8.89214e-19, 2.9640617e-19], [3.4593938, 2.9287e-18, 1.46436e-18, 8.78616e-19, 2.9287341e-19], [3.4673685, 2.89382e-18, 1.44692e-18, 8.68153e-19, 2.8938561e-19], [3.4753616, 2.85939e-18, 1.4297e-18, 8.57822e-19, 2.8594216e-19], [3.4833732, 2.82539e-18, 1.41271e-18, 8.47623e-19, 2.8254245e-19], [3.4914032, 2.79182e-18, 1.39592e-18, 8.37553e-19, 2.7918586e-19], [3.4994517, 2.75868e-18, 1.37935e-18, 8.27611e-19, 2.758718e-19], [3.5075187, 2.72596e-18, 1.36299e-18, 8.17795e-19, 2.7259969e-19], [3.5156044, 2.69366e-18, 1.34684e-18, 8.08103e-19, 2.6936895e-19], [3.5237087, 2.66176e-18, 1.33089e-18, 7.98533e-19, 2.66179e-19], [3.5318317, 2.63026e-18, 1.31514e-18, 7.89084e-19, 2.6302929e-19], [3.5399734, 2.59916e-18, 1.29959e-18, 7.79754e-19, 2.5991927e-19], [3.5481339, 2.56845e-18, 1.28423e-18, 7.70541e-19, 2.5684839e-19], [3.5563132, 2.53813e-18, 1.26907e-18, 7.61444e-19, 2.5381612e-19], [3.5645113, 2.50818e-18, 1.2541e-18, 7.52461e-19, 2.5082193e-19], [3.5727284, 2.47862e-18, 1.23932e-18, 7.43592e-19, 2.478653e-19], [3.5809644, 2.44942e-18, 1.22472e-18, 7.34833e-19, 2.4494572e-19], [3.5892193, 2.42059e-18, 1.21031e-18, 7.26184e-19, 2.4206269e-19], [3.5974934, 2.39212e-18, 1.19607e-18, 7.17643e-19, 2.3921571e-19], [3.6057864, 2.36401e-18, 1.18201e-18, 7.09209e-19, 2.3640429e-19], [3.6140986, 2.33624e-18, 1.16813e-18, 7.00879e-19, 2.3362795e-19], [3.62243, 2.30883e-18, 1.15442e-18, 6.92654e-19, 2.3088621e-19], [3.6307805, 2.28175e-18, 1.14089e-18, 6.84531e-19, 2.2817862e-19], [3.6391504, 2.25501e-18, 1.12752e-18, 6.7651e-19, 2.255047e-19], [3.6475395, 2.2286e-18, 1.11431e-18, 6.68588e-19, 2.2286401e-19], [3.6559479, 2.20253e-18, 1.10127e-18, 6.60764e-19, 2.202561e-19], [3.6643757, 2.17677e-18, 1.0884e-18, 6.53037e-19, 2.1768053e-19], [3.672823, 2.15133e-18, 1.07568e-18, 6.45406e-19, 2.1513687e-19], [3.6812897, 2.12621e-18, 1.06312e-18, 6.3787e-19, 2.1262468e-19], [3.689776, 2.1014e-18, 1.05071e-18, 6.30426e-19, 2.1014355e-19], [3.6982818, 2.07689e-18, 1.03846e-18, 6.23075e-19, 2.0769306e-19], [3.7068072, 2.05269e-18, 1.02636e-18, 6.15814e-19, 2.0527281e-19], [3.7153523, 2.02879e-18, 1.0144e-18, 6.08643e-19, 2.0288239e-19], [3.7239171, 2.00518e-18, 1.0026e-18, 6.0156e-19, 2.005214e-19], [3.7325016, 1.98186e-18, 9.9094e-19, 5.94564e-19, 1.9818946e-19], [3.7411059, 1.95883e-18, 9.79423e-19, 5.87654e-19, 1.9588618e-19], [3.74973, 1.93608e-18, 9.68048e-19, 5.80829e-19, 1.9361118e-19], [3.758374, 1.9136e-18, 9.56813e-19, 5.74088e-19, 1.9136408e-19], [3.767038, 1.89141e-18, 9.45715e-19, 5.67429e-19, 1.8914452e-19], [3.7757219, 1.86948e-18, 9.34753e-19, 5.60852e-19, 1.8695214e-19], [3.7844258, 1.84783e-18, 9.23925e-19, 5.54355e-19, 1.8478657e-19], [3.7931498, 1.82644e-18, 9.1323e-19, 5.47938e-19, 1.8264746e-19], [3.801894, 1.80531e-18, 9.02665e-19, 5.41599e-19, 1.8053446e-19], [3.8106582, 1.78444e-18, 8.92229e-19, 5.35337e-19, 1.7844724e-19], [3.8194427, 1.76382e-18, 8.8192e-19, 5.29152e-19, 1.7638546e-19], [3.8282474, 1.74345e-18, 8.71736e-19, 5.23042e-19, 1.7434877e-19], [3.8370725, 1.72333e-18, 8.61677e-19, 5.17006e-19, 1.7233686e-19], [3.8459178, 1.70346e-18, 8.51739e-19, 5.11044e-19, 1.7034939e-19], [3.8547836, 1.68382e-18, 8.41923e-19, 5.05154e-19, 1.6838606e-19], [3.8636698, 1.66443e-18, 8.32225e-19, 4.99335e-19, 1.6644655e-19], [3.8725764, 1.64527e-18, 8.22645e-19, 4.93587e-19, 1.6453054e-19], [3.8815037, 1.62634e-18, 8.13181e-19, 4.87908e-19, 1.6263773e-19], [3.8904514, 1.60764e-18, 8.03831e-19, 4.82299e-19, 1.6076782e-19], [3.8994199, 1.58917e-18, 7.94595e-19, 4.76757e-19, 1.5892051e-19], [3.908409, 1.57092e-18, 7.8547e-19, 4.71282e-19, 1.5709551e-19], [3.9174188, 1.55289e-18, 7.76455e-19, 4.65873e-19, 1.5529254e-19], [3.9264494, 1.53507e-18, 7.67549e-19, 4.60529e-19, 1.535113e-19], [3.9355008, 1.51748e-18, 7.5875e-19, 4.5525e-19, 1.5175152e-19], [3.944573, 1.50009e-18, 7.50057e-19, 4.50034e-19, 1.5001292e-19], [3.9536662, 1.48291e-18, 7.41468e-19, 4.44881e-19, 1.4829523e-19], [3.9627803, 1.46594e-18, 7.32983e-19, 4.3979e-19, 1.4659817e-19], [3.9719155, 1.44918e-18, 7.24599e-19, 4.3476e-19, 1.4492149e-19], [3.9810717, 1.43261e-18, 7.16317e-19, 4.2979e-19, 1.4326492e-19], [3.990249, 1.41624e-18, 7.08133e-19, 4.2488e-19, 1.4162821e-19], [3.9994475, 1.40007e-18, 7.00047e-19, 4.20028e-19, 1.400111e-19], [4.0086672, 1.38409e-18, 6.92059e-19, 4.15235e-19, 1.3841333e-19], [4.0179081, 1.36831e-18, 6.84165e-19, 4.10499e-19, 1.3683467e-19], [4.0271703, 1.35271e-18, 6.76366e-19, 4.0582e-19, 1.3527487e-19], [4.0364539, 1.3373e-18, 6.6866e-19, 4.01196e-19, 1.3373369e-19], [4.0457589, 1.32207e-18, 6.61046e-19, 3.96628e-19, 1.3221089e-19], [4.0550854, 1.30702e-18, 6.53523e-19, 3.92114e-19, 1.3070624e-19], [4.0644333, 1.29216e-18, 6.46089e-19, 3.87654e-19, 1.2921951e-19], [4.0738028, 1.27746e-18, 6.38744e-19, 3.83246e-19, 1.2775047e-19], [4.0831939, 1.26295e-18, 6.31486e-19, 3.78892e-19, 1.2629889e-19], [4.0926066, 1.24861e-18, 6.24315e-19, 3.74589e-19, 1.2486456e-19], [4.102041, 1.23443e-18, 6.17228e-19, 3.70337e-19, 1.2344726e-19], [4.1114972, 1.22043e-18, 6.10226e-19, 3.66135e-19, 1.2204677e-19], [4.1209752, 1.20659e-18, 6.03306e-19, 3.61984e-19, 1.2066288e-19], [4.130475, 1.19291e-18, 5.96469e-19, 3.57881e-19, 1.1929538e-19], [4.1399967, 1.1794e-18, 5.89712e-19, 3.53827e-19, 1.1794406e-19], [4.1495404, 1.16605e-18, 5.83035e-19, 3.49821e-19, 1.1660873e-19], [4.1591061, 1.15285e-18, 5.76438e-19, 3.45863e-19, 1.1528918e-19], [4.1686938, 1.13981e-18, 5.69918e-19, 3.41951e-19, 1.1398521e-19], [4.1783037, 1.12693e-18, 5.63475e-19, 3.38085e-19, 1.1269663e-19], [4.1879357, 1.11419e-18, 5.57108e-19, 3.34265e-19, 1.1142324e-19], [4.1975898, 1.10161e-18, 5.50816e-19, 3.3049e-19, 1.1016486e-19], [4.2072663, 1.08917e-18, 5.44598e-19, 3.26759e-19, 1.0892129e-19], [4.216965, 1.07688e-18, 5.38453e-19, 3.23072e-19, 1.0769236e-19], [4.2266861, 1.06474e-18, 5.32381e-19, 3.19429e-19, 1.0647787e-19], [4.2364297, 1.05274e-18, 5.2638e-19, 3.15828e-19, 1.0527766e-19], [4.2461956, 1.04087e-18, 5.20449e-19, 3.12269e-19, 1.0409154e-19], [4.2559841, 1.02915e-18, 5.14588e-19, 3.08753e-19, 1.0291933e-19], [4.2657952, 1.01757e-18, 5.08796e-19, 3.05277e-19, 1.0176087e-19], [4.2756289, 1.00612e-18, 5.03071e-19, 3.01843e-19, 1.0061598e-19], [4.2854852, 9.94803e-19, 4.97414e-19, 2.98448e-19, 9.9484493e-20], [4.2953643, 9.83621e-19, 4.91823e-19, 2.95094e-19, 9.8366248e-20], [4.3052661, 9.72569e-19, 4.86297e-19, 2.91778e-19, 9.7261078e-20], [4.3151908, 9.61646e-19, 4.80835e-19, 2.88501e-19, 9.6168821e-20], [4.3251383, 9.50851e-19, 4.75438e-19, 2.85263e-19, 9.5089318e-20], [4.3351088, 9.40182e-19, 4.70103e-19, 2.82062e-19, 9.4022412e-20], [4.3451022, 9.29637e-19, 4.64831e-19, 2.78899e-19, 9.2967947e-20], [4.3551187, 9.19215e-19, 4.5962e-19, 2.75772e-19, 9.1925771e-20], [4.3651583, 9.08915e-19, 4.5447e-19, 2.72682e-19, 9.0895731e-20], [4.3752211, 8.98734e-19, 4.4938e-19, 2.69628e-19, 8.987768e-20], [4.385307, 8.88672e-19, 4.44349e-19, 2.66609e-19, 8.8871469e-20], [4.3954162, 8.78727e-19, 4.39376e-19, 2.63626e-19, 8.7876953e-20], [4.4055486, 8.68897e-19, 4.34461e-19, 2.60677e-19, 8.689399e-20], [4.4157045, 8.59182e-19, 4.29603e-19, 2.57762e-19, 8.5922437e-20], [4.4258837, 8.49579e-19, 4.24802e-19, 2.54881e-19, 8.4962155e-20], [4.4360864, 8.40087e-19, 4.20056e-19, 2.52034e-19, 8.4013006e-20], [4.4463127, 8.30706e-19, 4.15365e-19, 2.49219e-19, 8.3074854e-20], [4.4565625, 8.21433e-19, 4.10729e-19, 2.46437e-19, 8.2147566e-20], [4.4668359, 8.12267e-19, 4.06146e-19, 2.43688e-19, 8.1231008e-20], [4.477133, 8.03207e-19, 4.01616e-19, 2.4097e-19, 8.032505e-20], [4.4874539, 7.94252e-19, 3.97139e-19, 2.38283e-19, 7.9429563e-20], [4.4977985, 7.85401e-19, 3.92713e-19, 2.35628e-19, 7.8544419e-20], [4.508167, 7.76651e-19, 3.88338e-19, 2.33003e-19, 7.7669493e-20], [4.5185594, 7.68003e-19, 3.84014e-19, 2.30409e-19, 7.6804662e-20], [4.5289758, 7.59454e-19, 3.7974e-19, 2.27844e-19, 7.5949801e-20], [4.5394162, 7.51004e-19, 3.75515e-19, 2.25309e-19, 7.5104792e-20], [4.5498806, 7.42651e-19, 3.71339e-19, 2.22803e-19, 7.4269514e-20], [4.5603692, 7.34395e-19, 3.6721e-19, 2.20326e-19, 7.3443849e-20], [4.5708819, 7.26233e-19, 3.63129e-19, 2.17878e-19, 7.2627683e-20], [4.5814189, 7.18165e-19, 3.59095e-19, 2.15457e-19, 7.1820898e-20], [4.5919801, 7.1019e-19, 3.55108e-19, 2.13065e-19, 7.1023384e-20], [4.6025657, 7.02306e-19, 3.51166e-19, 2.107e-19, 7.0235027e-20], [4.6131757, 6.94513e-19, 3.47269e-19, 2.08362e-19, 6.9455718e-20], [4.6238102, 6.86809e-19, 3.43418e-19, 2.06051e-19, 6.8685348e-20], [4.6344692, 6.79194e-19, 3.3961e-19, 2.03766e-19, 6.7923809e-20], [4.6451528, 6.71665e-19, 3.35846e-19, 2.01507e-19, 6.7170995e-20], [4.6558609, 6.64223e-19, 3.32125e-19, 1.99275e-19, 6.6426801e-20], [4.6665938, 6.56866e-19, 3.28446e-19, 1.97068e-19, 6.5691124e-20], [4.6773514, 6.49594e-19, 3.2481e-19, 1.94886e-19, 6.4963861e-20], [4.6881338, 6.42404e-19, 3.21215e-19, 1.92729e-19, 6.4244913e-20], [4.6989411, 6.35297e-19, 3.17662e-19, 1.90597e-19, 6.3534179e-20], [4.7097733, 6.2827e-19, 3.14148e-19, 1.88489e-19, 6.2831562e-20], [4.7206304, 6.21324e-19, 3.10675e-19, 1.86405e-19, 6.2136963e-20], [4.7315126, 6.14458e-19, 3.07242e-19, 1.84345e-19, 6.1450288e-20], [4.7424199, 6.07669e-19, 3.03848e-19, 1.82309e-19, 6.0771442e-20], [4.7533523, 6.00958e-19, 3.00492e-19, 1.80295e-19, 6.0100332e-20], [4.7643099, 5.94323e-19, 2.97175e-19, 1.78305e-19, 5.9436865e-20], [4.7752927, 5.87764e-19, 2.93895e-19, 1.76337e-19, 5.8780951e-20], [4.7863009, 5.81279e-19, 2.90653e-19, 1.74392e-19, 5.8132499e-20], [4.7973345, 5.74868e-19, 2.87448e-19, 1.72469e-19, 5.7491421e-20], [4.8083935, 5.6853e-19, 2.84279e-19, 1.70567e-19, 5.6857629e-20], [4.819478, 5.62264e-19, 2.81146e-19, 1.68687e-19, 5.6231038e-20], [4.830588, 5.5607e-19, 2.78048e-19, 1.66829e-19, 5.561156e-20], [4.8417237, 5.49945e-19, 2.74986e-19, 1.64992e-19, 5.4999113e-20], [4.852885, 5.4389e-19, 2.71959e-19, 1.63175e-19, 5.4393613e-20], [4.8640721, 5.37903e-19, 2.68965e-19, 1.61379e-19, 5.3794978e-20], [4.8752849, 5.31985e-19, 2.66006e-19, 1.59604e-19, 5.3203126e-20], [4.8865236, 5.26133e-19, 2.6308e-19, 1.57848e-19, 5.2617977e-20], [4.8977882, 5.20348e-19, 2.60188e-19, 1.56113e-19, 5.2039453e-20], [4.9090788, 5.14628e-19, 2.57328e-19, 1.54397e-19, 5.1467474e-20], [4.9203954, 5.08973e-19, 2.545e-19, 1.527e-19, 5.0901965e-20], [4.931738, 5.03382e-19, 2.51705e-19, 1.51023e-19, 5.0342848e-20], [4.9431069, 4.97854e-19, 2.48941e-19, 1.49364e-19, 4.9790048e-20], [4.9545019, 4.92388e-19, 2.46208e-19, 1.47725e-19, 4.9243492e-20], [4.9659232, 4.86984e-19, 2.43506e-19, 1.46103e-19, 4.8703105e-20], [4.9773708, 4.81641e-19, 2.40834e-19, 1.44501e-19, 4.8168815e-20], [4.9888449, 4.76358e-19, 2.38193e-19, 1.42916e-19, 4.7640551e-20], [5.0003453, 4.71135e-19, 2.35581e-19, 1.41349e-19, 4.7118241e-20], [5.0118723, 4.65971e-19, 2.32999e-19, 1.398e-19, 4.6601817e-20], [5.0234259, 4.60865e-19, 2.30446e-19, 1.38268e-19, 4.6091208e-20], [5.0350061, 4.55816e-19, 2.27922e-19, 1.36753e-19, 4.5586347e-20], [5.046613, 4.50824e-19, 2.25426e-19, 1.35256e-19, 4.5087167e-20], [5.0582466, 4.45888e-19, 2.22958e-19, 1.33775e-19, 4.4593601e-20], [5.0699071, 4.41008e-19, 2.20518e-19, 1.32311e-19, 4.4105584e-20], [5.0815944, 4.36183e-19, 2.18105e-19, 1.30863e-19, 4.3623051e-20], [5.0933087, 4.31411e-19, 2.1572e-19, 1.29432e-19, 4.3145937e-20], [5.10505, 4.26694e-19, 2.13361e-19, 1.28017e-19, 4.2674179e-20], [5.1168184, 4.22029e-19, 2.11029e-19, 1.26617e-19, 4.2207716e-20], [5.1286138, 4.17417e-19, 2.08722e-19, 1.25233e-19, 4.1746485e-20], [5.1404365, 4.12856e-19, 2.06442e-19, 1.23865e-19, 4.1290426e-20], [5.1522864, 4.08346e-19, 2.04187e-19, 1.22512e-19, 4.0839478e-20], [5.1641637, 4.03887e-19, 2.01958e-19, 1.21175e-19, 4.0393582e-20], [5.1760683, 3.99478e-19, 1.99753e-19, 1.19852e-19, 3.9952678e-20], [5.1880004, 3.95118e-19, 1.97574e-19, 1.18544e-19, 3.951671e-20], [5.19996, 3.90807e-19, 1.95418e-19, 1.17251e-19, 3.908562e-20], [5.2119471, 3.86545e-19, 1.93287e-19, 1.15972e-19, 3.8659352e-20], [5.2239619, 3.8233e-19, 1.91179e-19, 1.14707e-19, 3.8237848e-20], [5.2360044, 3.78162e-19, 1.89095e-19, 1.13457e-19, 3.7821055e-20], [5.2480746, 3.7404e-19, 1.87034e-19, 1.12221e-19, 3.7408917e-20], [5.2601727, 3.69965e-19, 1.84997e-19, 1.10998e-19, 3.7001382e-20], [5.2722986, 3.65935e-19, 1.82982e-19, 1.09789e-19, 3.6598394e-20], [5.2844525, 3.6195e-19, 1.80989e-19, 1.08594e-19, 3.6199903e-20], [5.2966344, 3.58009e-19, 1.79019e-19, 1.07411e-19, 3.5805856e-20], [5.3088444, 3.54113e-19, 1.77071e-19, 1.06242e-19, 3.5416202e-20], [5.3210826, 3.50259e-19, 1.75144e-19, 1.05087e-19, 3.5030889e-20], [5.333349, 3.46449e-19, 1.73239e-19, 1.03943e-19, 3.4649869e-20], [5.3456436, 3.42681e-19, 1.71355e-19, 1.02813e-19, 3.4273091e-20], [5.3579666, 3.38955e-19, 1.69492e-19, 1.01695e-19, 3.3900507e-20], [5.370318, 3.35271e-19, 1.6765e-19, 1.0059e-19, 3.3532068e-20], [5.3826978, 3.31627e-19, 1.65828e-19, 9.9497e-20, 3.3167727e-20], [5.3951062, 3.28024e-19, 1.64027e-19, 9.84161e-20, 3.2807437e-20], [5.4075432, 3.24461e-19, 1.62245e-19, 9.73473e-20, 3.245115e-20], [5.4200089, 3.20938e-19, 1.60484e-19, 9.62903e-20, 3.2098822e-20], [5.4325033, 3.17454e-19, 1.58742e-19, 9.5245e-20, 3.1750406e-20], [5.4450265, 3.14008e-19, 1.57019e-19, 9.42113e-20, 3.1405858e-20], [5.4575786, 3.10601e-19, 1.55315e-19, 9.31892e-20, 3.1065134e-20], [5.4701596, 3.07231e-19, 1.53631e-19, 9.21783e-20, 3.0728189e-20], [5.4827696, 3.03899e-19, 1.51964e-19, 9.11787e-20, 3.039498e-20], [5.4954087, 3.00604e-19, 1.50317e-19, 9.01901e-20, 3.0065465e-20], [5.508077, 2.97345e-19, 1.48688e-19, 8.92125e-20, 2.9739601e-20], [5.5207744, 2.94123e-19, 1.47076e-19, 8.82458e-20, 2.9417346e-20], [5.5335011, 2.90936e-19, 1.45483e-19, 8.72897e-20, 2.909866e-20], [5.5462571, 2.87784e-19, 1.43907e-19, 8.63442e-20, 2.8783501e-20], [5.5590426, 2.84667e-19, 1.42349e-19, 8.54092e-20, 2.8471829e-20], [5.5718575, 2.81585e-19, 1.40807e-19, 8.44845e-20, 2.8163605e-20], [5.5847019, 2.78537e-19, 1.39283e-19, 8.357e-20, 2.7858789e-20], [5.597576, 2.75522e-19, 1.37776e-19, 8.26657e-20, 2.7557342e-20], [5.6104798, 2.72541e-19, 1.36286e-19, 8.17713e-20, 2.7259226e-20], [5.6234133, 2.69592e-19, 1.34811e-19, 8.08868e-20, 2.6964402e-20], [5.6363766, 2.66677e-19, 1.33354e-19, 8.00121e-20, 2.6672835e-20], [5.6493697, 2.63793e-19, 1.31912e-19, 7.91471e-20, 2.6384485e-20], [5.6623929, 2.60941e-19, 1.30486e-19, 7.82916e-20, 2.6099318e-20], [5.6754461, 2.58121e-19, 1.29076e-19, 7.74455e-20, 2.5817296e-20], [5.6885293, 2.55332e-19, 1.27681e-19, 7.66087e-20, 2.5538384e-20], [5.7016427, 2.52573e-19, 1.26302e-19, 7.57812e-20, 2.5262546e-20], [5.7147864, 2.49845e-19, 1.24938e-19, 7.49628e-20, 2.4989748e-20], [5.7279603, 2.47147e-19, 1.23589e-19, 7.41534e-20, 2.4719956e-20], [5.7411646, 2.44479e-19, 1.22255e-19, 7.33529e-20, 2.4453134e-20], [5.7543994, 2.4184e-19, 1.20935e-19, 7.25613e-20, 2.4189251e-20], [5.7676646, 2.3923e-19, 1.19631e-19, 7.17783e-20, 2.3928271e-20], [5.7809605, 2.36649e-19, 1.1834e-19, 7.1004e-20, 2.3670162e-20], [5.794287, 2.34096e-19, 1.17064e-19, 7.02382e-20, 2.3414893e-20], [5.8076442, 2.31572e-19, 1.15801e-19, 6.94808e-20, 2.316243e-20], [5.8210322, 2.29075e-19, 1.14553e-19, 6.87317e-20, 2.2912742e-20], [5.834451, 2.26605e-19, 1.13318e-19, 6.79909e-20, 2.2665798e-20], [5.8479008, 2.24163e-19, 1.12097e-19, 6.72582e-20, 2.2421566e-20], [5.8613816, 2.21747e-19, 1.10889e-19, 6.65335e-20, 2.2180017e-20], [5.8748935, 2.19358e-19, 1.09695e-19, 6.58168e-20, 2.1941119e-20], [5.8884366, 2.16995e-19, 1.08513e-19, 6.5108e-20, 2.1704843e-20], [5.9020108, 2.14658e-19, 1.07345e-19, 6.44069e-20, 2.1471159e-20], [5.9156163, 2.12347e-19, 1.06189e-19, 6.37135e-20, 2.1240039e-20], [5.9292532, 2.10061e-19, 1.05046e-19, 6.30278e-20, 2.1011452e-20], [5.9429216, 2.078e-19, 1.03916e-19, 6.23495e-20, 2.0785372e-20], [5.9566214, 2.05564e-19, 1.02798e-19, 6.16787e-20, 2.0561768e-20], [5.9703529, 2.03352e-19, 1.01692e-19, 6.10152e-20, 2.0340615e-20], [5.984116, 2.01165e-19, 1.00598e-19, 6.0359e-20, 2.0121883e-20], [5.9979108, 1.99002e-19, 9.95167e-20, 5.971e-20, 1.9905546e-20], [6.0117374, 1.96862e-19, 9.84468e-20, 5.90681e-20, 1.9691578e-20], [6.0255959, 1.94745e-19, 9.73886e-20, 5.84332e-20, 1.947995e-20], [6.0394863, 1.92652e-19, 9.63421e-20, 5.78052e-20, 1.9270638e-20], [6.0534087, 1.90582e-19, 9.53069e-20, 5.71842e-20, 1.9063614e-20], [6.0673633, 1.88534e-19, 9.42831e-20, 5.65699e-20, 1.8858854e-20], [6.08135, 1.86509e-19, 9.32705e-20, 5.59623e-20, 1.8656332e-20], [6.095369, 1.84506e-19, 9.22689e-20, 5.53614e-20, 1.8456023e-20], [6.1094202, 1.82524e-19, 9.12783e-20, 5.4767e-20, 1.8257903e-20], [6.1235039, 1.80565e-19, 9.02985e-20, 5.41791e-20, 1.8061946e-20], [6.1376201, 1.78627e-19, 8.93294e-20, 5.35976e-20, 1.7868128e-20], [6.1517687, 1.76709e-19, 8.83709e-20, 5.30225e-20, 1.7676426e-20], [6.16595, 1.74813e-19, 8.74228e-20, 5.24537e-20, 1.7486816e-20], [6.180164, 1.72938e-19, 8.64851e-20, 5.18911e-20, 1.7299274e-20], [6.1944108, 1.71083e-19, 8.55576e-20, 5.13346e-20, 1.7113777e-20], [6.2086903, 1.69248e-19, 8.46402e-20, 5.07841e-20, 1.6930303e-20], [6.2230029, 1.67433e-19, 8.37328e-20, 5.02397e-20, 1.6748829e-20], [6.2373484, 1.65638e-19, 8.28353e-20, 4.97012e-20, 1.6569333e-20], [6.2517269, 1.63862e-19, 8.19476e-20, 4.91686e-20, 1.6391792e-20], [6.2661386, 1.62106e-19, 8.10695e-20, 4.86417e-20, 1.6216185e-20], [6.2805836, 1.60369e-19, 8.02011e-20, 4.81206e-20, 1.604249e-20], [6.2950618, 1.58651e-19, 7.9342e-20, 4.76052e-20, 1.5870686e-20], [6.3095734, 1.56952e-19, 7.84923e-20, 4.70954e-20, 1.5700752e-20], [6.3241185, 1.55271e-19, 7.76519e-20, 4.65911e-20, 1.5532667e-20], [6.3386971, 1.53608e-19, 7.68206e-20, 4.60924e-20, 1.536641e-20], [6.3533093, 1.51964e-19, 7.59983e-20, 4.5599e-20, 1.5201962e-20], [6.3679552, 1.50337e-19, 7.5185e-20, 4.5111e-20, 1.5039301e-20], [6.3826349, 1.48728e-19, 7.43805e-20, 4.46283e-20, 1.4878408e-20], [6.3973484, 1.47136e-19, 7.35848e-20, 4.41509e-20, 1.4719264e-20], [6.4120958, 1.45562e-19, 7.27977e-20, 4.36786e-20, 1.4561849e-20], [6.4268772, 1.44005e-19, 7.20192e-20, 4.32115e-20, 1.4406144e-20], [6.4416927, 1.42465e-19, 7.12491e-20, 4.27494e-20, 1.4252129e-20], [6.4565423, 1.40941e-19, 7.04873e-20, 4.22924e-20, 1.4099787e-20], [6.4714262, 1.39434e-19, 6.97339e-20, 4.18403e-20, 1.3949098e-20], [6.4863443, 1.37944e-19, 6.89886e-20, 4.13932e-20, 1.3800045e-20], [6.5012969, 1.36469e-19, 6.82514e-20, 4.09508e-20, 1.3652608e-20], [6.5162839, 1.35011e-19, 6.75222e-20, 4.05133e-20, 1.350677e-20], [6.5313055, 1.33568e-19, 6.68009e-20, 4.00805e-20, 1.3362514e-20], [6.5463617, 1.32141e-19, 6.60874e-20, 3.96525e-20, 1.3219822e-20], [6.5614527, 1.3073e-19, 6.53817e-20, 3.9229e-20, 1.3078677e-20], [6.5765784, 1.29333e-19, 6.46836e-20, 3.88102e-20, 1.2939061e-20], [6.591739, 1.27952e-19, 6.39931e-20, 3.83958e-20, 1.2800957e-20], [6.6069345, 1.26586e-19, 6.331e-20, 3.7986e-20, 1.266435e-20], [6.622165, 1.25235e-19, 6.26343e-20, 3.75806e-20, 1.2529222e-20], [6.6374307, 1.23898e-19, 6.1966e-20, 3.71796e-20, 1.2395556e-20], [6.6527316, 1.22576e-19, 6.13049e-20, 3.67829e-20, 1.2263338e-20], [6.6680677, 1.21268e-19, 6.06509e-20, 3.63906e-20, 1.213255e-20], [6.6834392, 1.19974e-19, 6.00041e-20, 3.60024e-20, 1.2003178e-20], [6.6988461, 1.18694e-19, 5.93642e-20, 3.56185e-20, 1.1875205e-20], [6.7142885, 1.17428e-19, 5.87312e-20, 3.52387e-20, 1.1748615e-20], [6.7297666, 1.16176e-19, 5.81051e-20, 3.48631e-20, 1.1623395e-20], [6.7452803, 1.14937e-19, 5.74857e-20, 3.44914e-20, 1.1499528e-20], [6.7608298, 1.13712e-19, 5.68731e-20, 3.41239e-20, 1.1377e-20], [6.7764151, 1.125e-19, 5.62671e-20, 3.37602e-20, 1.1255796e-20], [6.7920363, 1.11301e-19, 5.56676e-20, 3.34005e-20, 1.1135901e-20], [6.8076936, 1.10114e-19, 5.50745e-20, 3.30447e-20, 1.1017302e-20], [6.8233869, 1.08941e-19, 5.44879e-20, 3.26928e-20, 1.0899983e-20], [6.8391165, 1.07781e-19, 5.39077e-20, 3.23446e-20, 1.078393e-20], [6.8548823, 1.06633e-19, 5.33336e-20, 3.20002e-20, 1.0669131e-20], [6.8706844, 1.05497e-19, 5.27658e-20, 3.16595e-20, 1.055557e-20], [6.886523, 1.04373e-19, 5.22041e-20, 3.13225e-20, 1.0443234e-20], [6.902398, 1.03262e-19, 5.16485e-20, 3.09891e-20, 1.0332111e-20], [6.9183097, 1.02163e-19, 5.10989e-20, 3.06593e-20, 1.0222186e-20], [6.9342581, 1.01075e-19, 5.05551e-20, 3.03331e-20, 1.0113446e-20], [6.9502432, 9.99995e-20, 5.00173e-20, 3.00104e-20, 1.0005879e-20], [6.9662651, 9.89353e-20, 4.94852e-20, 2.96911e-20, 9.8994709e-21], [6.982324, 9.78826e-20, 4.89589e-20, 2.93753e-20, 9.7942098e-21], [6.99842, 9.68413e-20, 4.84383e-20, 2.9063e-20, 9.6900829e-21], [7.014553, 9.58112e-20, 4.79232e-20, 2.87539e-20, 9.5870778e-21], [7.0307232, 9.47921e-20, 4.74137e-20, 2.84482e-20, 9.4851822e-21], [7.0469307, 9.3784e-20, 4.69097e-20, 2.81458e-20, 9.3843838e-21], [7.0631755, 9.27868e-20, 4.64111e-20, 2.78467e-20, 9.2846708e-21], [7.0794578, 9.18004e-20, 4.59179e-20, 2.75508e-20, 9.1860311e-21], [7.0957777, 9.08245e-20, 4.543e-20, 2.7258e-20, 9.0884531e-21], [7.1121351, 8.98591e-20, 4.49474e-20, 2.69684e-20, 8.9919251e-21], [7.1285303, 8.89041e-20, 4.44699e-20, 2.66819e-20, 8.8964357e-21], [7.1449633, 8.79594e-20, 4.39976e-20, 2.63985e-20, 8.8019734e-21], [7.1614341, 8.70249e-20, 4.35303e-20, 2.61182e-20, 8.7085271e-21], [7.1779429, 8.61004e-20, 4.30681e-20, 2.58409e-20, 8.6160856e-21], [7.1944898, 8.51858e-20, 4.26108e-20, 2.55665e-20, 8.524638e-21], [7.2110748, 8.42811e-20, 4.21585e-20, 2.52951e-20, 8.4341734e-21], [7.227698, 8.33861e-20, 4.1711e-20, 2.50266e-20, 8.3446811e-21], [7.2443596, 8.25007e-20, 4.12684e-20, 2.4761e-20, 8.2561504e-21], [7.2610596, 8.16248e-20, 4.08304e-20, 2.44983e-20, 8.1685709e-21], [7.277798, 8.07584e-20, 4.03972e-20, 2.42383e-20, 8.0819321e-21], [7.2945751, 7.99012e-20, 3.99687e-20, 2.39812e-20, 7.9962239e-21], [7.3113908, 7.90532e-20, 3.95447e-20, 2.37268e-20, 7.911436e-21], [7.3282453, 7.82144e-20, 3.91253e-20, 2.34752e-20, 7.8275584e-21], [7.3451387, 7.73845e-20, 3.87104e-20, 2.32263e-20, 7.7445813e-21], [7.362071, 7.65636e-20, 3.83e-20, 2.298e-20, 7.6624947e-21], [7.3790423, 7.57514e-20, 3.78939e-20, 2.27364e-20, 7.5812891e-21], [7.3960528, 7.4948e-20, 3.74922e-20, 2.24953e-20, 7.5009547e-21], [7.4131024, 7.41532e-20, 3.70949e-20, 2.22569e-20, 7.4214822e-21], [7.4301914, 7.33669e-20, 3.67017e-20, 2.2021e-20, 7.3428621e-21], [7.4473197, 7.2589e-20, 3.63128e-20, 2.17877e-20, 7.2650851e-21], [7.4644876, 7.18195e-20, 3.59281e-20, 2.15569e-20, 7.1881421e-21], [7.481695, 7.10583e-20, 3.55475e-20, 2.13285e-20, 7.1120241e-21], [7.4989421, 7.03052e-20, 3.5171e-20, 2.11026e-20, 7.036722e-21], [7.5162289, 6.95601e-20, 3.47985e-20, 2.08791e-20, 6.9622269e-21], [7.5335556, 6.88231e-20, 3.443e-20, 2.0658e-20, 6.8885302e-21], [7.5509223, 6.80939e-20, 3.40654e-20, 2.04393e-20, 6.8156231e-21], [7.568329, 6.73726e-20, 3.37048e-20, 2.02229e-20, 6.7434971e-21], [7.5857758, 6.6659e-20, 3.3348e-20, 2.00088e-20, 6.6721436e-21], [7.6032628, 6.5953e-20, 3.2995e-20, 1.9797e-20, 6.6015544e-21], [7.6207901, 6.52546e-20, 3.26459e-20, 1.95875e-20, 6.531721e-21], [7.6383578, 6.45636e-20, 3.23004e-20, 1.93803e-20, 6.4626353e-21], [7.6559661, 6.38801e-20, 3.19587e-20, 1.91752e-20, 6.3942893e-21], [7.6736149, 6.32038e-20, 3.16206e-20, 1.89724e-20, 6.3266748e-21], [7.6913044, 6.25349e-20, 3.12861e-20, 1.87717e-20, 6.2597839e-21], [7.7090347, 6.1873e-20, 3.09552e-20, 1.85731e-20, 6.1936089e-21], [7.7268059, 6.12183e-20, 3.06279e-20, 1.83767e-20, 6.1281419e-21], [7.744618, 6.05705e-20, 3.0304e-20, 1.81824e-20, 6.0633753e-21], [7.7624712, 5.99297e-20, 2.99836e-20, 1.79902e-20, 5.9993014e-21], [7.7803655, 5.92957e-20, 2.96667e-20, 1.78e-20, 5.9359129e-21], [7.7983011, 5.86685e-20, 2.93531e-20, 1.76119e-20, 5.8732023e-21], [7.816278, 5.80481e-20, 2.90429e-20, 1.74257e-20, 5.8111622e-21], [7.8342964, 5.74342e-20, 2.8736e-20, 1.72416e-20, 5.7497854e-21], [7.8523563, 5.68269e-20, 2.84324e-20, 1.70594e-20, 5.6890647e-21], [7.8704579, 5.62261e-20, 2.8132e-20, 1.68792e-20, 5.6289931e-21], [7.8886012, 5.56317e-20, 2.78348e-20, 1.67009e-20, 5.5695635e-21], [7.9067863, 5.50437e-20, 2.75409e-20, 1.65245e-20, 5.510769e-21], [7.9250133, 5.4462e-20, 2.725e-20, 1.635e-20, 5.4526027e-21], [7.9432823, 5.38864e-20, 2.69623e-20, 1.61774e-20, 5.3950578e-21], [7.9615935, 5.3317e-20, 2.66776e-20, 1.60066e-20, 5.3381277e-21], [7.9799469, 5.27537e-20, 2.6396e-20, 1.58376e-20, 5.2818057e-21], [7.9983426, 5.21964e-20, 2.61174e-20, 1.56704e-20, 5.2260853e-21], [8.0167806, 5.16451e-20, 2.58417e-20, 1.5505e-20, 5.1709599e-21], [8.0352612, 5.10997e-20, 2.5569e-20, 1.53414e-20, 5.1164231e-21], [8.0537844, 5.056e-20, 2.52992e-20, 1.51795e-20, 5.0624687e-21], [8.0723503, 5.00262e-20, 2.50323e-20, 1.50194e-20, 5.0090903e-21], [8.090959, 4.9498e-20, 2.47683e-20, 1.4861e-20, 4.9562817e-21], [8.1096106, 4.89755e-20, 2.4507e-20, 1.47042e-20, 4.9040368e-21], [8.1283052, 4.84585e-20, 2.42486e-20, 1.45491e-20, 4.8523495e-21], [8.1470428, 4.79471e-20, 2.39929e-20, 1.43957e-20, 4.8012138e-21], [8.1658237, 4.74411e-20, 2.37399e-20, 1.4244e-20, 4.7506239e-21], [8.1846479, 4.69405e-20, 2.34897e-20, 1.40938e-20, 4.7005737e-21], [8.2035154, 4.64452e-20, 2.32421e-20, 1.39452e-20, 4.6510576e-21], [8.2224265, 4.59553e-20, 2.29971e-20, 1.37983e-20, 4.6020698e-21], [8.2413812, 4.54705e-20, 2.27548e-20, 1.36529e-20, 4.5536046e-21], [8.2603795, 4.4991e-20, 2.2515e-20, 1.3509e-20, 4.5056563e-21], [8.2794216, 4.45165e-20, 2.22778e-20, 1.33667e-20, 4.4582195e-21], [8.2985077, 4.40471e-20, 2.20431e-20, 1.32259e-20, 4.4112887e-21], [8.3176377, 4.35827e-20, 2.1811e-20, 1.30866e-20, 4.3648584e-21], [8.3368118, 4.31233e-20, 2.15813e-20, 1.29488e-20, 4.3189233e-21], [8.3560302, 4.26688e-20, 2.1354e-20, 1.28124e-20, 4.273478e-21], [8.3752928, 4.22191e-20, 2.11292e-20, 1.26775e-20, 4.2285173e-21], [8.3945999, 4.17742e-20, 2.09068e-20, 1.25441e-20, 4.1840361e-21], [8.4139514, 4.1334e-20, 2.06867e-20, 1.2412e-20, 4.1400292e-21], [8.4333476, 4.08986e-20, 2.0469e-20, 1.22814e-20, 4.0964914e-21], [8.4527885, 4.04678e-20, 2.02537e-20, 1.21522e-20, 4.0534179e-21], [8.4722741, 4.00415e-20, 2.00406e-20, 1.20243e-20, 4.0108035e-21], [8.4918048, 3.96198e-20, 1.98298e-20, 1.18979e-20, 3.9686435e-21], [8.5113804, 3.92027e-20, 1.96212e-20, 1.17727e-20, 3.926933e-21], [8.5310011, 3.87899e-20, 1.94148e-20, 1.16489e-20, 3.8856672e-21], [8.5506671, 3.83816e-20, 1.92107e-20, 1.15264e-20, 3.8448412e-21], [8.5703785, 3.79776e-20, 1.90087e-20, 1.14052e-20, 3.8044505e-21], [8.5901352, 3.75779e-20, 1.88089e-20, 1.12854e-20, 3.7644904e-21], [8.6099375, 3.71825e-20, 1.86112e-20, 1.11667e-20, 3.7249563e-21], [8.6297855, 3.67913e-20, 1.84157e-20, 1.10494e-20, 3.6858437e-21], [8.6496792, 3.64042e-20, 1.82222e-20, 1.09333e-20, 3.6471481e-21], [8.6696188, 3.60213e-20, 1.80307e-20, 1.08184e-20, 3.608865e-21], [8.6896043, 3.56425e-20, 1.78413e-20, 1.07048e-20, 3.57099e-21], [8.7096359, 3.52677e-20, 1.7654e-20, 1.05924e-20, 3.5335189e-21], [8.7297137, 3.48969e-20, 1.74686e-20, 1.04812e-20, 3.4964474e-21], [8.7498378, 3.453e-20, 1.72852e-20, 1.03711e-20, 3.4597711e-21], [8.7700082, 3.41671e-20, 1.71038e-20, 1.02623e-20, 3.4234859e-21], [8.7902252, 3.3808e-20, 1.69243e-20, 1.01546e-20, 3.3875877e-21], [8.8104887, 3.34528e-20, 1.67467e-20, 1.0048e-20, 3.3520724e-21], [8.830799, 3.31014e-20, 1.6571e-20, 9.94258e-21, 3.3169359e-21], [8.8511561, 3.27537e-20, 1.63971e-20, 9.83829e-21, 3.2821742e-21], [8.8715601, 3.24097e-20, 1.62252e-20, 9.73511e-21, 3.2477834e-21], [8.8920112, 3.20693e-20, 1.6055e-20, 9.63303e-21, 3.2137594e-21], [8.9125094, 3.17327e-20, 1.58867e-20, 9.53203e-21, 3.1800986e-21], [8.9330548, 3.13996e-20, 1.57202e-20, 9.43212e-21, 3.146797e-21], [8.9536477, 3.107e-20, 1.55555e-20, 9.33327e-21, 3.1138509e-21], [8.9742879, 3.0744e-20, 1.53925e-20, 9.23548e-21, 3.0812565e-21], [8.9949758, 3.04214e-20, 1.52312e-20, 9.13873e-21, 3.0490102e-21], [9.0157114, 3.01023e-20, 1.50717e-20, 9.04302e-21, 3.0171082e-21], [9.0364947, 2.97866e-20, 1.49139e-20, 8.94833e-21, 2.9855471e-21], [9.057326, 2.94743e-20, 1.47577e-20, 8.85465e-21, 2.9543231e-21], [9.0782053, 2.91653e-20, 1.46033e-20, 8.76197e-21, 2.9234329e-21], [9.0991327, 2.88597e-20, 1.44505e-20, 8.67028e-21, 2.8928728e-21], [9.1201084, 2.85572e-20, 1.42993e-20, 8.57957e-21, 2.8626396e-21], [9.1411324, 2.82581e-20, 1.41497e-20, 8.48983e-21, 2.8327297e-21], [9.1622049, 2.79621e-20, 1.40018e-20, 8.40106e-21, 2.8031398e-21], [9.183326, 2.76693e-20, 1.38554e-20, 8.31323e-21, 2.7738667e-21], [9.2044957, 2.73796e-20, 1.37106e-20, 8.22634e-21, 2.7449069e-21], [9.2257143, 2.7093e-20, 1.35673e-20, 8.14038e-21, 2.7162573e-21], [9.2469817, 2.68095e-20, 1.34256e-20, 8.05535e-21, 2.6879148e-21], [9.2682982, 2.6529e-20, 1.32854e-20, 7.97122e-21, 2.6598761e-21], [9.2896639, 2.62516e-20, 1.31467e-20, 7.888e-21, 2.632138e-21], [9.3110788, 2.59771e-20, 1.30095e-20, 7.80567e-21, 2.6046977e-21], [9.332543, 2.57056e-20, 1.28737e-20, 7.72423e-21, 2.5775519e-21], [9.3540567, 2.54369e-20, 1.27394e-20, 7.64365e-21, 2.5506977e-21], [9.3756201, 2.51712e-20, 1.26066e-20, 7.56395e-21, 2.5241322e-21], [9.3972331, 2.49083e-20, 1.24752e-20, 7.4851e-21, 2.4978524e-21], [9.418896, 2.46483e-20, 1.23452e-20, 7.4071e-21, 2.4718554e-21], [9.4406088, 2.4391e-20, 1.22166e-20, 7.32994e-21, 2.4461384e-21], [9.4623716, 2.41366e-20, 1.20894e-20, 7.25362e-21, 2.4206986e-21], [9.4841846, 2.38848e-20, 1.19635e-20, 7.17811e-21, 2.3955332e-21], [9.5060479, 2.36358e-20, 1.1839e-20, 7.10342e-21, 2.3706395e-21], [9.5279616, 2.33895e-20, 1.17159e-20, 7.02954e-21, 2.3460149e-21], [9.5499259, 2.31458e-20, 1.15941e-20, 6.95646e-21, 2.3216566e-21], [9.5719407, 2.29048e-20, 1.14736e-20, 6.88417e-21, 2.297562e-21], [9.5940063, 2.26664e-20, 1.13544e-20, 6.81266e-21, 2.2737286e-21], [9.6161228, 2.24306e-20, 1.12365e-20, 6.74193e-21, 2.2501539e-21], [9.6382902, 2.21973e-20, 1.11199e-20, 6.67197e-21, 2.2268354e-21], [9.6605088, 2.19666e-20, 1.10046e-20, 6.60276e-21, 2.2037706e-21], [9.6827786, 2.17384e-20, 1.08905e-20, 6.53432e-21, 2.1809571e-21], [9.7050997, 2.15126e-20, 1.07777e-20, 6.46661e-21, 2.1583925e-21], [9.7274722, 2.12894e-20, 1.06661e-20, 6.39965e-21, 2.1360746e-21], [9.7498964, 2.10686e-20, 1.05557e-20, 6.33343e-21, 2.114001e-21], [9.7723722, 2.08502e-20, 1.04465e-20, 6.26792e-21, 2.0921696e-21], [9.7948999, 2.06342e-20, 1.03386e-20, 6.20314e-21, 2.070578e-21], [9.8174794, 2.04206e-20, 1.02318e-20, 6.13907e-21, 2.0492242e-21], [9.8401111, 2.02093e-20, 1.01262e-20, 6.07571e-21, 2.0281061e-21], [9.8627949, 2.00004e-20, 1.00218e-20, 6.01305e-21, 2.0072217e-21], [9.8855309, 1.97938e-20, 9.91848e-21, 5.95109e-21, 1.9865688e-21], [9.9083194, 1.95895e-20, 9.81635e-21, 5.88981e-21, 1.9661456e-21], [9.9311605, 1.93875e-20, 9.71536e-21, 5.82922e-21, 1.9459501e-21], [9.9540542, 1.91877e-20, 9.6155e-21, 5.7693e-21, 1.9259805e-21], [9.9770006, 1.89902e-20, 9.51676e-21, 5.71006e-21, 1.9062351e-21], [10.0, 1.87949e-20, 9.41914e-21, 5.65148e-21, 1.886712e-21], [10.023052, 1.86018e-20, 9.32261e-21, 5.59357e-21, 1.8674095e-21], [10.046158, 1.84109e-20, 9.22719e-21, 5.53631e-21, 1.8483261e-21], [10.069317, 1.82222e-20, 9.13285e-21, 5.47971e-21, 1.8294602e-21], [10.092529, 1.80356e-20, 9.03959e-21, 5.42375e-21, 1.8108103e-21], [10.115795, 1.78512e-20, 8.9474e-21, 5.36844e-21, 1.792375e-21], [10.139114, 1.76689e-20, 8.85628e-21, 5.31377e-21, 1.7741529e-21], [10.162487, 1.74888e-20, 8.76622e-21, 5.25973e-21, 1.7561427e-21], [10.185914, 1.73107e-20, 8.67721e-21, 5.20633e-21, 1.7383433e-21], [10.209395, 1.71348e-20, 8.58926e-21, 5.15355e-21, 1.7207535e-21], [10.23293, 1.69609e-20, 8.50234e-21, 5.10141e-21, 1.7033723e-21], [10.256519, 1.67891e-20, 8.41647e-21, 5.04988e-21, 1.6861988e-21], [10.280163, 1.66194e-20, 8.33163e-21, 4.99898e-21, 1.6692322e-21], [10.303861, 1.64517e-20, 8.24782e-21, 4.94869e-21, 1.6524717e-21], [10.327614, 1.62861e-20, 8.16503e-21, 4.89902e-21, 1.6359169e-21], [10.351422, 1.61226e-20, 8.08328e-21, 4.84997e-21, 1.6195671e-21], [10.375284, 1.59611e-20, 8.00255e-21, 4.80153e-21, 1.6034221e-21], [10.399202, 1.58016e-20, 7.92284e-21, 4.7537e-21, 1.5874816e-21], [10.423174, 1.56442e-20, 7.84415e-21, 4.70649e-21, 1.5717457e-21], [10.447202, 1.54889e-20, 7.76649e-21, 4.6599e-21, 1.5562145e-21], [10.471285, 1.53356e-20, 7.68986e-21, 4.61391e-21, 1.5408882e-21], [10.495424, 1.51843e-20, 7.61425e-21, 4.56855e-21, 1.5257675e-21], [10.519619, 1.50351e-20, 7.53967e-21, 4.5238e-21, 1.5108531e-21], [10.543869, 1.4888e-20, 7.46614e-21, 4.47968e-21, 1.4961459e-21], [10.568175, 1.4743e-20, 7.39364e-21, 4.43618e-21, 1.4816472e-21], [10.592537, 1.46001e-20, 7.3222e-21, 4.39332e-21, 1.4673586e-21], [10.616956, 1.44593e-20, 7.25181e-21, 4.35109e-21, 1.4532818e-21], [10.64143, 1.43207e-20, 7.1825e-21, 4.3095e-21, 1.4394191e-21], [10.665961, 1.41842e-20, 7.11427e-21, 4.26856e-21, 1.4257731e-21], [10.690549, 1.40499e-20, 7.04714e-21, 4.22828e-21, 1.4123468e-21], [10.715193, 1.39179e-20, 6.98113e-21, 4.18868e-21, 1.3991437e-21], [10.739894, 1.37882e-20, 6.91625e-21, 4.14975e-21, 1.3861679e-21], [10.764652, 1.36607e-20, 6.85254e-21, 4.11152e-21, 1.3734238e-21], [10.789467, 1.35357e-20, 6.79001e-21, 4.07401e-21, 1.360917e-21], [10.81434, 1.34131e-20, 6.7287e-21, 4.03722e-21, 1.3486533e-21], [10.839269, 1.3293e-20, 6.66864e-21, 4.00118e-21, 1.3366397e-21], [10.864256, 1.31754e-20, 6.60987e-21, 3.96592e-21, 1.324884e-21], [10.889301, 1.30606e-20, 6.55244e-21, 3.93147e-21, 1.3133951e-21], [10.914403, 1.29485e-20, 6.4964e-21, 3.89784e-21, 1.3021832e-21], [10.939564, 1.28394e-20, 6.4418e-21, 3.86508e-21, 1.2912597e-21], [10.964782, 1.27332e-20, 6.38871e-21, 3.83323e-21, 1.2806377e-21], [10.990058, 1.26303e-20, 6.3372e-21, 3.80232e-21, 1.2703322e-21], [11.015393, 1.25307e-20, 6.28737e-21, 3.77242e-21, 1.2603601e-21], [11.040786, 1.24346e-20, 6.2393e-21, 3.74358e-21, 1.2507409e-21], [11.066238, 1.23423e-20, 6.19311e-21, 3.71587e-21, 1.2414967e-21], [11.091748, 1.2254e-20, 6.14893e-21, 3.68936e-21, 1.2326528e-21], [11.117317, 1.21701e-20, 6.1069e-21, 3.66414e-21, 1.2242384e-21], [11.142945, 1.20908e-20, 6.06718e-21, 3.64031e-21, 1.2162869e-21], [11.168632, 1.20165e-20, 6.02998e-21, 3.61799e-21, 1.208837e-21], [11.194379, 1.19477e-20, 5.99552e-21, 3.59731e-21, 1.2019333e-21], [11.220185, 1.1885e-20, 5.96405e-21, 3.57843e-21, 1.1956276e-21], [11.24605, 1.18288e-20, 5.93589e-21, 3.56153e-21, 1.1899802e-21], [11.271975, 1.178e-20, 5.91137e-21, 3.54682e-21, 1.1850616e-21], [11.297959, 1.17394e-20, 5.89092e-21, 3.53455e-21, 1.1809543e-21], [11.324004, 1.17079e-20, 5.87502e-21, 3.52501e-21, 1.1777558e-21], [11.350108, 1.16866e-20, 5.86426e-21, 3.51855e-21, 1.1755812e-21], [11.376273, 1.16771e-20, 5.85931e-21, 3.51559e-21, 1.1745675e-21], [11.402498, 1.16809e-20, 5.861e-21, 3.5166e-21, 1.1748785e-21], [11.428783, 1.16999e-20, 5.87031e-21, 3.52219e-21, 1.1767112e-21], [11.455129, 1.17367e-20, 5.88844e-21, 3.53307e-21, 1.1803039e-21], [11.481536, 1.17941e-20, 5.91685e-21, 3.55011e-21, 1.1859472e-21], [11.508004, 1.18756e-20, 5.95731e-21, 3.57439e-21, 1.1939976e-21], [11.534533, 1.19858e-20, 6.01204e-21, 3.60722e-21, 1.2048964e-21], [11.561122, 1.21301e-20, 6.0838e-21, 3.65028e-21, 1.2191946e-21], [11.587774, 1.23155e-20, 6.17606e-21, 3.70563e-21, 1.2375874e-21], [11.614486, 1.2551e-20, 6.29326e-21, 3.77596e-21, 1.2609614e-21], [11.64126, 1.28479e-20, 6.44114e-21, 3.86468e-21, 1.290462e-21], [11.668096, 1.32213e-20, 6.6272e-21, 3.97632e-21, 1.3275903e-21], [11.694994, 1.36912e-20, 6.86145e-21, 4.11687e-21, 1.3743473e-21], [11.721954, 1.42849e-20, 7.15749e-21, 4.2945e-21, 1.4334522e-21], [11.748976, 1.50402e-20, 7.53425e-21, 4.52055e-21, 1.5086878e-21], [11.77606, 1.60112e-20, 8.01877e-21, 4.81126e-21, 1.6054649e-21], [11.803206, 1.7278e-20, 8.6511e-21, 5.19066e-21, 1.7317909e-21], [11.830416, 1.89643e-20, 9.49306e-21, 5.69583e-21, 1.9000286e-21], [11.857687, 2.12713e-20, 1.06453e-20, 6.38719e-21, 2.130315e-21], [11.885022, 2.45509e-20, 1.22837e-20, 7.37024e-21, 2.4578219e-21], [11.91242, 2.9478e-20, 1.47459e-20, 8.84751e-21, 2.9500574e-21], [11.939881, 3.75352e-20, 1.87729e-20, 1.12637e-20, 3.7552705e-21], [11.967405, 5.27387e-20, 2.63731e-20, 1.58239e-20, 5.2751085e-21], [11.994993, 9.11108e-20, 4.55576e-20, 2.73345e-20, 9.1117913e-21], [12.022644, 3.58502e-19, 1.79252e-19, 1.07551e-19, 3.5850394e-20], [12.050359, 1.82236e-19, 9.11193e-20, 5.46716e-20, 1.8223995e-20], [12.078138, 7.27431e-20, 3.63742e-20, 2.18245e-20, 7.2751778e-21], [12.105981, 4.56809e-20, 2.28446e-20, 1.37068e-20, 4.5694524e-21], [12.133889, 3.35176e-20, 1.67644e-20, 1.00586e-20, 3.3535985e-21], [12.16186, 2.66579e-20, 1.33359e-20, 8.00155e-21, 2.6680812e-21], [12.189896, 2.22846e-20, 1.11506e-20, 6.69033e-21, 2.2311737e-21], [12.217997, 1.92719e-20, 9.6454e-21, 5.78724e-21, 1.9302975e-21], [12.246162, 1.70819e-20, 8.5515e-21, 5.1309e-21, 1.7116606e-21], [12.274392, 1.54253e-20, 7.72422e-21, 4.63453e-21, 1.546335e-21], [12.302688, 1.41328e-20, 7.07892e-21, 4.24735e-21, 1.4173954e-21], [12.331048, 1.3099e-20, 6.56287e-21, 3.93772e-21, 1.3142936e-21], [12.359474, 1.22547e-20, 6.14153e-21, 3.68492e-21, 1.2301253e-21], [12.387966, 1.1553e-20, 5.7914e-21, 3.47484e-21, 1.1601912e-21], [12.416523, 1.09609e-20, 5.49597e-21, 3.29758e-21, 1.101189e-21], [12.445146, 1.04544e-20, 5.24333e-21, 3.146e-21, 1.0507357e-21], [12.473835, 1.0016e-20, 5.02467e-21, 3.0148e-21, 1.0070731e-21], [12.50259, 9.63238e-21, 4.83339e-21, 2.90003e-21, 9.6887858e-22], [12.531412, 9.29351e-21, 4.66441e-21, 2.79865e-21, 9.3514068e-22], [12.5603, 8.99148e-21, 4.51382e-21, 2.70829e-21, 9.0507458e-22], [12.589254, 8.72011e-21, 4.37852e-21, 2.62711e-21, 8.7806335e-22], [12.618275, 8.47448e-21, 4.25606e-21, 2.55364e-21, 8.5361648e-22], [12.647363, 8.25065e-21, 4.14448e-21, 2.48669e-21, 8.3133986e-22], [12.676519, 8.04541e-21, 4.04216e-21, 2.4253e-21, 8.1091389e-22], [12.705741, 7.85614e-21, 3.94781e-21, 2.36868e-21, 7.9207713e-22], [12.735031, 7.68066e-21, 3.86033e-21, 2.3162e-21, 7.7461403e-22], [12.764388, 7.5172e-21, 3.77884e-21, 2.26731e-21, 7.5834561e-22], [12.793813, 7.36424e-21, 3.70259e-21, 2.22155e-21, 7.4312219e-22], [12.823306, 7.22052e-21, 3.63094e-21, 2.17857e-21, 7.2881789e-22], [12.852867, 7.08497e-21, 3.56337e-21, 2.13802e-21, 7.1532614e-22], [12.882496, 6.95668e-21, 3.49941e-21, 2.09964e-21, 7.0255629e-22], [12.912193, 6.83487e-21, 3.43868e-21, 2.06321e-21, 6.9043079e-22], [12.941958, 6.71887e-21, 3.38084e-21, 2.02851e-21, 6.78883e-22], [12.971793, 6.6081e-21, 3.32561e-21, 1.99537e-21, 6.6785537e-22], [13.001696, 6.50207e-21, 3.27274e-21, 1.96365e-21, 6.5729802e-22], [13.031668, 6.40033e-21, 3.22201e-21, 1.93321e-21, 6.4716747e-22], [13.061709, 6.3025e-21, 3.17323e-21, 1.90394e-21, 6.3742577e-22], [13.091819, 6.20824e-21, 3.12623e-21, 1.87574e-21, 6.2803956e-22], [13.121999, 6.11727e-21, 3.08086e-21, 1.84852e-21, 6.1897949e-22], [13.152248, 6.02932e-21, 3.037e-21, 1.8222e-21, 6.1021964e-22], [13.182567, 5.94416e-21, 2.99453e-21, 1.79672e-21, 6.01737e-22], [13.212956, 5.86158e-21, 2.95334e-21, 1.772e-21, 5.9351113e-22], [13.243415, 5.78141e-21, 2.91335e-21, 1.74801e-21, 5.8552377e-22], [13.273945, 5.70346e-21, 2.87447e-21, 1.72468e-21, 5.7775859e-22], [13.304544, 5.62761e-21, 2.83664e-21, 1.70198e-21, 5.7020097e-22], [13.335214, 5.55371e-21, 2.79978e-21, 1.67987e-21, 5.6283771e-22], [13.365955, 5.48165e-21, 2.76383e-21, 1.6583e-21, 5.5565692e-22], [13.396767, 5.41132e-21, 2.72874e-21, 1.63725e-21, 5.4864786e-22], [13.42765, 5.34261e-21, 2.69447e-21, 1.61668e-21, 5.4180078e-22], [13.458604, 5.27544e-21, 2.66096e-21, 1.59658e-21, 5.3510682e-22], [13.489629, 5.20974e-21, 2.62818e-21, 1.57691e-21, 5.2855791e-22], [13.520726, 5.14542e-21, 2.59609e-21, 1.55765e-21, 5.2214667e-22], [13.551894, 5.08241e-21, 2.56465e-21, 1.53879e-21, 5.1586637e-22], [13.583134, 5.02066e-21, 2.53384e-21, 1.52031e-21, 5.0971082e-22], [13.614447, 4.96011e-21, 2.50363e-21, 1.50218e-21, 5.0367434e-22], [13.645831, 4.9007e-21, 2.47399e-21, 1.48439e-21, 4.9775169e-22], [13.677288, 4.84238e-21, 2.44489e-21, 1.46694e-21, 4.9193805e-22], [13.708818, 4.78512e-21, 2.41632e-21, 1.44979e-21, 4.8622896e-22], [13.74042, 4.72887e-21, 2.38825e-21, 1.43295e-21, 4.8062027e-22], [13.772095, 4.67359e-21, 2.36067e-21, 1.4164e-21, 4.7510814e-22], [13.803843, 4.61924e-21, 2.33355e-21, 1.40013e-21, 4.6968902e-22], [13.835664, 4.5658e-21, 2.30688e-21, 1.38413e-21, 4.6435957e-22], [13.867558, 4.51322e-21, 2.28064e-21, 1.36838e-21, 4.591167e-22], [13.899526, 4.46148e-21, 2.25482e-21, 1.35289e-21, 4.5395751e-22], [13.931568, 4.41056e-21, 2.22941e-21, 1.33765e-21, 4.4887928e-22], [13.963684, 4.36043e-21, 2.20439e-21, 1.32264e-21, 4.4387948e-22], [13.995873, 4.31106e-21, 2.17976e-21, 1.30785e-21, 4.3895571e-22], [14.028137, 4.26243e-21, 2.15549e-21, 1.29329e-21, 4.3410572e-22], [14.060475, 4.21452e-21, 2.13158e-21, 1.27895e-21, 4.293274e-22], [14.092888, 4.16731e-21, 2.10802e-21, 1.26481e-21, 4.2461875e-22], [14.125375, 4.12079e-21, 2.0848e-21, 1.25088e-21, 4.1997788e-22], [14.157938, 4.07492e-21, 2.06191e-21, 1.23715e-21, 4.1540301e-22], [14.190575, 4.0297e-21, 2.03934e-21, 1.2236e-21, 4.1089244e-22], [14.223288, 3.98511e-21, 2.01709e-21, 1.21025e-21, 4.0644458e-22], [14.256076, 3.94114e-21, 1.99514e-21, 1.19708e-21, 4.0205788e-22], [14.28894, 3.89776e-21, 1.97349e-21, 1.1841e-21, 3.9773092e-22], [14.321879, 3.85497e-21, 1.95214e-21, 1.17128e-21, 3.9346229e-22], [14.354894, 3.81275e-21, 1.93107e-21, 1.15864e-21, 3.8925068e-22], [14.387986, 3.77109e-21, 1.91027e-21, 1.14616e-21, 3.8509484e-22], [14.421154, 3.72998e-21, 1.88976e-21, 1.13385e-21, 3.8099356e-22], [14.454398, 3.68941e-21, 1.86951e-21, 1.1217e-21, 3.769457e-22], [14.487719, 3.64936e-21, 1.84952e-21, 1.10971e-21, 3.7295014e-22], [14.521116, 3.60983e-21, 1.82979e-21, 1.09787e-21, 3.6900583e-22], [14.554591, 3.5708e-21, 1.81031e-21, 1.08618e-21, 3.6511176e-22], [14.588143, 3.53226e-21, 1.79107e-21, 1.07464e-21, 3.6126695e-22], [14.621772, 3.49421e-21, 1.77208e-21, 1.06325e-21, 3.5747047e-22], [14.655478, 3.45664e-21, 1.75333e-21, 1.052e-21, 3.537214e-22], [14.689263, 3.41953e-21, 1.73481e-21, 1.04088e-21, 3.5001888e-22], [14.723125, 3.38288e-21, 1.71651e-21, 1.02991e-21, 3.4636206e-22], [14.757065, 3.34668e-21, 1.69845e-21, 1.01907e-21, 3.4275015e-22], [14.791084, 3.31093e-21, 1.6806e-21, 1.00836e-21, 3.3918235e-22], [14.825181, 3.27561e-21, 1.66297e-21, 9.97783e-22, 3.356579e-22], [14.859356, 3.24072e-21, 1.64556e-21, 9.87334e-22, 3.3217608e-22], [14.893611, 3.20625e-21, 1.62835e-21, 9.7701e-22, 3.2873617e-22], [14.927944, 3.17219e-21, 1.61135e-21, 9.66811e-22, 3.2533749e-22], [14.962357, 3.13854e-21, 1.59455e-21, 9.56733e-22, 3.2197936e-22], [14.996848, 3.10529e-21, 1.57796e-21, 9.46775e-22, 3.1866114e-22], [15.03142, 3.07243e-21, 1.56156e-21, 9.36935e-22, 3.153822e-22], [15.066071, 3.03996e-21, 1.54535e-21, 9.27212e-22, 3.1214193e-22], [15.100802, 3.00787e-21, 1.52934e-21, 9.17603e-22, 3.0893973e-22], [15.135612, 2.97616e-21, 1.51351e-21, 9.08106e-22, 3.0577502e-22], [15.170504, 2.94483e-21, 1.49787e-21, 8.98721e-22, 3.0264725e-22], [15.205475, 2.91385e-21, 1.48241e-21, 8.89444e-22, 2.9955586e-22], [15.240528, 2.88324e-21, 1.46713e-21, 8.80276e-22, 2.9650031e-22], [15.275661, 2.85298e-21, 1.45202e-21, 8.71214e-22, 2.934801e-22], [15.310875, 2.82307e-21, 1.43709e-21, 8.62256e-22, 2.9049469e-22], [15.34617, 2.7935e-21, 1.42234e-21, 8.53401e-22, 2.8754361e-22], [15.381546, 2.76427e-21, 1.40775e-21, 8.44648e-22, 2.8462636e-22], [15.417005, 2.73538e-21, 1.39333e-21, 8.35996e-22, 2.8174247e-22], [15.452544, 2.70682e-21, 1.37907e-21, 8.27442e-22, 2.7889148e-22], [15.488166, 2.67858e-21, 1.36498e-21, 8.18985e-22, 2.7607293e-22], [15.52387, 2.65067e-21, 1.35104e-21, 8.10625e-22, 2.7328638e-22], [15.559656, 2.62307e-21, 1.33727e-21, 8.0236e-22, 2.7053139e-22], [15.595525, 2.59579e-21, 1.32365e-21, 7.94188e-22, 2.6780755e-22], [15.631476, 2.56881e-21, 1.31018e-21, 7.86109e-22, 2.6511443e-22], [15.667511, 2.54214e-21, 1.29687e-21, 7.78121e-22, 2.6245163e-22], [15.703628, 2.51576e-21, 1.2837e-21, 7.70222e-22, 2.5981875e-22], [15.739829, 2.48969e-21, 1.27069e-21, 7.62413e-22, 2.572154e-22], [15.776113, 2.4639e-21, 1.25782e-21, 7.5469e-22, 2.5464118e-22], [15.81248, 2.43841e-21, 1.24509e-21, 7.47055e-22, 2.5209574e-22], [15.848932, 2.4132e-21, 1.23251e-21, 7.39505e-22, 2.4957869e-22], [15.885467, 2.38827e-21, 1.22006e-21, 7.32039e-22, 2.4708967e-22], [15.922087, 2.36362e-21, 1.20776e-21, 7.24656e-22, 2.4462833e-22], [15.958791, 2.33925e-21, 1.19559e-21, 7.17355e-22, 2.4219432e-22], [15.99558, 2.31514e-21, 1.18356e-21, 7.10136e-22, 2.3978729e-22], [16.032454, 2.2913e-21, 1.17166e-21, 7.02996e-22, 2.3740691e-22], [16.069413, 2.26773e-21, 1.15989e-21, 6.95936e-22, 2.3505284e-22], [16.106456, 2.24442e-21, 1.14826e-21, 6.88953e-22, 2.3272476e-22], [16.143586, 2.22136e-21, 1.13675e-21, 6.82048e-22, 2.3042235e-22], [16.1808, 2.19856e-21, 1.12537e-21, 6.75219e-22, 2.2814529e-22], [16.218101, 2.17602e-21, 1.11411e-21, 6.68466e-22, 2.2589327e-22], [16.255488, 2.15372e-21, 1.10298e-21, 6.61786e-22, 2.23666e-22], [16.29296, 2.13166e-21, 1.09197e-21, 6.55181e-22, 2.2146315e-22], [16.330519, 2.10985e-21, 1.08108e-21, 6.48647e-22, 2.1928445e-22], [16.368165, 2.08828e-21, 1.07031e-21, 6.42186e-22, 2.1712961e-22], [16.405898, 2.06694e-21, 1.05966e-21, 6.35795e-22, 2.1499832e-22], [16.443717, 2.04584e-21, 1.04912e-21, 6.29475e-22, 2.1289032e-22], [16.481624, 2.02497e-21, 1.03871e-21, 6.23223e-22, 2.1080533e-22], [16.519618, 2.00433e-21, 1.0284e-21, 6.1704e-22, 2.0874307e-22], [16.5577, 1.98391e-21, 1.01821e-21, 6.10924e-22, 2.0670327e-22], [16.595869, 1.96372e-21, 1.00813e-21, 6.04876e-22, 2.0468567e-22], [16.634127, 1.94375e-21, 9.98154e-22, 5.98893e-22, 2.0269001e-22], [16.672472, 1.92399e-21, 9.88292e-22, 5.92975e-22, 2.0071603e-22], [16.710906, 1.90445e-21, 9.78536e-22, 5.87122e-22, 1.9876347e-22], [16.749429, 1.88513e-21, 9.68887e-22, 5.81332e-22, 1.9683209e-22], [16.78804, 1.86601e-21, 9.59342e-22, 5.75605e-22, 1.9492164e-22], [16.826741, 1.8471e-21, 9.49902e-22, 5.69941e-22, 1.9303187e-22], [16.86553, 1.8284e-21, 9.40563e-22, 5.64338e-22, 1.9116255e-22], [16.904409, 1.8099e-21, 9.31326e-22, 5.58796e-22, 1.8931344e-22], [16.943378, 1.7916e-21, 9.2219e-22, 5.53314e-22, 1.8748431e-22], [16.982437, 1.7735e-21, 9.13152e-22, 5.47891e-22, 1.8567492e-22], [17.021585, 1.7556e-21, 9.04212e-22, 5.42527e-22, 1.8388506e-22], [17.060824, 1.73789e-21, 8.95369e-22, 5.37221e-22, 1.8211449e-22], [17.100153, 1.72037e-21, 8.86621e-22, 5.31973e-22, 1.80363e-22], [17.139573, 1.70305e-21, 8.77968e-22, 5.26781e-22, 1.7863036e-22], [17.179084, 1.68591e-21, 8.69409e-22, 5.21645e-22, 1.7691637e-22], [17.218686, 1.66895e-21, 8.60942e-22, 5.16565e-22, 1.7522082e-22], [17.258379, 1.65218e-21, 8.52566e-22, 5.1154e-22, 1.7354348e-22], [17.298164, 1.63559e-21, 8.44281e-22, 5.06568e-22, 1.7188416e-22], [17.33804, 1.61918e-21, 8.36084e-22, 5.01651e-22, 1.7024265e-22], [17.378008, 1.60295e-21, 8.27977e-22, 4.96786e-22, 1.6861876e-22], [17.418069, 1.58689e-21, 8.19956e-22, 4.91974e-22, 1.6701228e-22], [17.458222, 1.57101e-21, 8.12022e-22, 4.87213e-22, 1.6542301e-22], [17.498467, 1.55529e-21, 8.04174e-22, 4.82504e-22, 1.6385077e-22], [17.538805, 1.53975e-21, 7.96409e-22, 4.77846e-22, 1.6229535e-22], [17.579236, 1.52437e-21, 7.88729e-22, 4.73237e-22, 1.6075659e-22], [17.61976, 1.50916e-21, 7.8113e-22, 4.68678e-22, 1.5923428e-22], [17.660378, 1.49412e-21, 7.73614e-22, 4.64168e-22, 1.5772824e-22], [17.70109, 1.47924e-21, 7.66178e-22, 4.59707e-22, 1.562383e-22], [17.741895, 1.46451e-21, 7.58822e-22, 4.55293e-22, 1.5476426e-22], [17.782794, 1.44995e-21, 7.51545e-22, 4.50927e-22, 1.5330597e-22], [17.823788, 1.43554e-21, 7.44346e-22, 4.46608e-22, 1.5186323e-22], [17.864876, 1.42129e-21, 7.37224e-22, 4.42335e-22, 1.5043588e-22], [17.906059, 1.40719e-21, 7.30179e-22, 4.38107e-22, 1.4902375e-22], [17.947336, 1.39324e-21, 7.23209e-22, 4.33925e-22, 1.4762667e-22], [17.988709, 1.37944e-21, 7.16314e-22, 4.29788e-22, 1.4624446e-22], [18.030177, 1.36579e-21, 7.09492e-22, 4.25695e-22, 1.4487698e-22], [18.071741, 1.35229e-21, 7.02744e-22, 4.21647e-22, 1.4352405e-22], [18.113401, 1.33894e-21, 6.96068e-22, 4.17641e-22, 1.4218551e-22], [18.155157, 1.32572e-21, 6.89464e-22, 4.13678e-22, 1.4086121e-22], [18.197009, 1.31265e-21, 6.8293e-22, 4.09758e-22, 1.3955099e-22], [18.238957, 1.29972e-21, 6.76466e-22, 4.05879e-22, 1.382547e-22], [18.281002, 1.28693e-21, 6.70071e-22, 4.02043e-22, 1.3697217e-22], [18.323144, 1.27428e-21, 6.63744e-22, 3.98247e-22, 1.3570326e-22], [18.365383, 1.26176e-21, 6.57486e-22, 3.94491e-22, 1.3444782e-22], [18.40772, 1.24938e-21, 6.51294e-22, 3.90776e-22, 1.3320571e-22], [18.450154, 1.23713e-21, 6.45168e-22, 3.87101e-22, 1.3197677e-22], [18.492686, 1.22501e-21, 6.39107e-22, 3.83464e-22, 1.3076087e-22], [18.535316, 1.21302e-21, 6.33112e-22, 3.79867e-22, 1.2955785e-22], [18.578045, 1.20116e-21, 6.2718e-22, 3.76308e-22, 1.2836759e-22], [18.620871, 1.18943e-21, 6.21312e-22, 3.72787e-22, 1.2718993e-22], [18.663797, 1.17783e-21, 6.15507e-22, 3.69304e-22, 1.2602475e-22], [18.706821, 1.16635e-21, 6.09763e-22, 3.65858e-22, 1.2487191e-22], [18.749945, 1.15499e-21, 6.04081e-22, 3.62448e-22, 1.2373126e-22], [18.793168, 1.14375e-21, 5.98459e-22, 3.59075e-22, 1.2260269e-22], [18.836491, 1.13264e-21, 5.92897e-22, 3.55738e-22, 1.2148605e-22], [18.879913, 1.12165e-21, 5.87395e-22, 3.52437e-22, 1.2038123e-22], [18.923436, 1.11077e-21, 5.81952e-22, 3.49171e-22, 1.1928808e-22], [18.967059, 1.10001e-21, 5.76566e-22, 3.4594e-22, 1.1820649e-22], [19.010783, 1.08937e-21, 5.71238e-22, 3.42743e-22, 1.1713633e-22], [19.054607, 1.07884e-21, 5.65967e-22, 3.3958e-22, 1.1607747e-22], [19.098533, 1.06842e-21, 5.60751e-22, 3.36451e-22, 1.1502979e-22], [19.142559, 1.05812e-21, 5.55592e-22, 3.33355e-22, 1.1399318e-22], [19.186687, 1.04793e-21, 5.50488e-22, 3.30293e-22, 1.1296751e-22], [19.230917, 1.03785e-21, 5.45438e-22, 3.27263e-22, 1.1195266e-22], [19.275249, 1.02787e-21, 5.40441e-22, 3.24265e-22, 1.1094852e-22], [19.319683, 1.01801e-21, 5.35498e-22, 3.21299e-22, 1.0995497e-22], [19.36422, 1.00825e-21, 5.30608e-22, 3.18365e-22, 1.0897189e-22], [19.408859, 9.98591e-22, 5.2577e-22, 3.15462e-22, 1.0799919e-22], [19.453601, 9.89041e-22, 5.20983e-22, 3.1259e-22, 1.0703673e-22], [19.498446, 9.79593e-22, 5.16248e-22, 3.09749e-22, 1.0608442e-22], [19.543395, 9.70247e-22, 5.11563e-22, 3.06938e-22, 1.0514214e-22], [19.588447, 9.61003e-22, 5.06928e-22, 3.04157e-22, 1.0420979e-22], [19.633603, 9.51858e-22, 5.02342e-22, 3.01405e-22, 1.0328726e-22], [19.678863, 9.42812e-22, 4.97805e-22, 2.98683e-22, 1.0237444e-22], [19.724227, 9.33863e-22, 4.93317e-22, 2.9599e-22, 1.0147123e-22], [19.769696, 9.25012e-22, 4.88876e-22, 2.93326e-22, 1.0057753e-22], [19.81527, 9.16256e-22, 4.84483e-22, 2.9069e-22, 9.969324e-23], [19.860949, 9.07595e-22, 4.80136e-22, 2.88082e-22, 9.8818251e-23], [19.906733, 8.99027e-22, 4.75836e-22, 2.85502e-22, 9.7952468e-23], [19.952623, 8.90553e-22, 4.71582e-22, 2.82949e-22, 9.7095793e-23], [19.998619, 8.8217e-22, 4.67373e-22, 2.80424e-22, 9.6248128e-23], [20.04472, 8.73878e-22, 4.63209e-22, 2.77925e-22, 9.5409378e-23], [20.090928, 8.65676e-22, 4.59089e-22, 2.75453e-22, 9.4579447e-23], [20.137242, 8.57563e-22, 4.55013e-22, 2.73008e-22, 9.3758243e-23], [20.183664, 8.49538e-22, 4.50981e-22, 2.70589e-22, 9.2945672e-23], [20.230192, 8.416e-22, 4.46992e-22, 2.68195e-22, 9.2141642e-23], [20.276827, 8.33749e-22, 4.43045e-22, 2.65827e-22, 9.1346064e-23], [20.32357, 8.25983e-22, 4.39141e-22, 2.63485e-22, 9.0558847e-23], [20.370421, 8.18301e-22, 4.35278e-22, 2.61167e-22, 8.9779903e-23], [20.417379, 8.10704e-22, 4.31457e-22, 2.58874e-22, 8.9009145e-23], [20.464446, 8.03188e-22, 4.27676e-22, 2.56605e-22, 8.8246487e-23], [20.511622, 7.95755e-22, 4.23935e-22, 2.54361e-22, 8.7491843e-23], [20.558906, 7.88403e-22, 4.20235e-22, 2.52141e-22, 8.6745128e-23], [20.606299, 7.81132e-22, 4.16574e-22, 2.49945e-22, 8.6006259e-23], [20.653802, 7.7394e-22, 4.12952e-22, 2.47771e-22, 8.5275154e-23], [20.701413, 7.66826e-22, 4.09369e-22, 2.45622e-22, 8.4551731e-23], [20.749135, 7.5979e-22, 4.05825e-22, 2.43495e-22, 8.383591e-23], [20.796967, 7.52832e-22, 4.02318e-22, 2.41391e-22, 8.3127611e-23], [20.844909, 7.4595e-22, 3.98849e-22, 2.39309e-22, 8.2426756e-23], [20.892961, 7.39143e-22, 3.95417e-22, 2.3725e-22, 8.1733266e-23], [20.941125, 7.32411e-22, 3.92022e-22, 2.35213e-22, 8.1047066e-23], [20.989399, 7.25754e-22, 3.88663e-22, 2.33198e-22, 8.036808e-23], [21.037784, 7.19169e-22, 3.8534e-22, 2.31204e-22, 7.9696232e-23], [21.086281, 7.12657e-22, 3.82052e-22, 2.29231e-22, 7.9031448e-23], [21.13489, 7.06217e-22, 3.788e-22, 2.2728e-22, 7.8373656e-23], [21.183611, 6.99848e-22, 3.75583e-22, 2.2535e-22, 7.7722784e-23], [21.232445, 6.9355e-22, 3.72401e-22, 2.23441e-22, 7.7078759e-23], [21.28139, 6.87321e-22, 3.69253e-22, 2.21552e-22, 7.6441512e-23], [21.330449, 6.81162e-22, 3.66138e-22, 2.19683e-22, 7.5810973e-23], [21.379621, 6.75071e-22, 3.63057e-22, 2.17834e-22, 7.5187073e-23], [21.428906, 6.69047e-22, 3.6001e-22, 2.16006e-22, 7.4569744e-23], [21.478305, 6.63091e-22, 3.56995e-22, 2.14197e-22, 7.395892e-23], [21.527817, 6.57201e-22, 3.54012e-22, 2.12407e-22, 7.3354534e-23], [21.577444, 6.51377e-22, 3.51062e-22, 2.10637e-22, 7.2756522e-23], [21.627185, 6.45618e-22, 3.48144e-22, 2.08886e-22, 7.2164818e-23], [21.677041, 6.39924e-22, 3.45257e-22, 2.07154e-22, 7.1579359e-23], [21.727012, 6.34293e-22, 3.42401e-22, 2.05441e-22, 7.1000082e-23], [21.777098, 6.28726e-22, 3.39577e-22, 2.03746e-22, 7.0426925e-23], [21.827299, 6.23221e-22, 3.36783e-22, 2.0207e-22, 6.9859828e-23], [21.877616, 6.17779e-22, 3.34019e-22, 2.00411e-22, 6.9298729e-23], [21.928049, 6.12398e-22, 3.31285e-22, 1.98771e-22, 6.874357e-23], [21.978599, 6.07078e-22, 3.28581e-22, 1.97149e-22, 6.8194291e-23], [22.029265, 6.01818e-22, 3.25907e-22, 1.95544e-22, 6.7650835e-23], [22.080047, 5.96619e-22, 3.23261e-22, 1.93957e-22, 6.7113145e-23], [22.130947, 5.91478e-22, 3.20645e-22, 1.92387e-22, 6.6581164e-23], [22.181964, 5.86396e-22, 3.18057e-22, 1.90834e-22, 6.6054836e-23], [22.233099, 5.81373e-22, 3.15497e-22, 1.89298e-22, 6.5534108e-23], [22.284351, 5.76407e-22, 3.12966e-22, 1.87779e-22, 6.5018926e-23], [22.335722, 5.71498e-22, 3.10462e-22, 1.86277e-22, 6.4509235e-23], [22.387211, 5.66646e-22, 3.07986e-22, 1.84791e-22, 6.4004985e-23], [22.438819, 5.6185e-22, 3.05537e-22, 1.83322e-22, 6.3506122e-23], [22.490546, 5.57109e-22, 3.03115e-22, 1.81869e-22, 6.3012598e-23], [22.542392, 5.52424e-22, 3.00719e-22, 1.80432e-22, 6.2524361e-23], [22.594358, 5.47793e-22, 2.98351e-22, 1.7901e-22, 6.2041364e-23], [22.646443, 5.43217e-22, 2.96008e-22, 1.77605e-22, 6.1563556e-23], [22.698649, 5.38694e-22, 2.93692e-22, 1.76215e-22, 6.1090892e-23], [22.750974, 5.34225e-22, 2.91401e-22, 1.74841e-22, 6.0623323e-23], [22.803421, 5.29808e-22, 2.89136e-22, 1.73482e-22, 6.0160805e-23], [22.855988, 5.25444e-22, 2.86896e-22, 1.72138e-22, 5.9703293e-23], [22.908677, 5.21132e-22, 2.84682e-22, 1.70809e-22, 5.9250741e-23], [22.961486, 5.16871e-22, 2.82492e-22, 1.69495e-22, 5.8803107e-23], [23.014418, 5.12662e-22, 2.80327e-22, 1.68196e-22, 5.8360347e-23], [23.067472, 5.08503e-22, 2.78187e-22, 1.66912e-22, 5.7922421e-23], [23.120648, 5.04395e-22, 2.76071e-22, 1.65642e-22, 5.7489286e-23], [23.173946, 5.00336e-22, 2.73978e-22, 1.64387e-22, 5.7060904e-23], [23.227368, 4.96327e-22, 2.7191e-22, 1.63146e-22, 5.6637234e-23], [23.280913, 4.92367e-22, 2.69866e-22, 1.61919e-22, 5.6218239e-23], [23.334581, 4.88457e-22, 2.67845e-22, 1.60707e-22, 5.580388e-23], [23.388372, 4.84594e-22, 2.65847e-22, 1.59508e-22, 5.5394121e-23], [23.442288, 4.8078e-22, 2.63872e-22, 1.58323e-22, 5.4988926e-23], [23.496328, 4.77014e-22, 2.61921e-22, 1.57152e-22, 5.458826e-23], [23.550493, 4.73295e-22, 2.59992e-22, 1.55995e-22, 5.4192089e-23], [23.604782, 4.69623e-22, 2.58086e-22, 1.54851e-22, 5.3800381e-23], [23.659197, 4.65998e-22, 2.56202e-22, 1.53721e-22, 5.3413102e-23], [23.713737, 4.6242e-22, 2.54341e-22, 1.52604e-22, 5.3030223e-23], [23.768403, 4.58888e-22, 2.52501e-22, 1.51501e-22, 5.2651711e-23], [23.823195, 4.55402e-22, 2.50684e-22, 1.5041e-22, 5.2277539e-23], [23.878113, 4.51962e-22, 2.48889e-22, 1.49333e-22, 5.1907679e-23], [23.933158, 4.48567e-22, 2.47115e-22, 1.48269e-22, 5.1542102e-23], [23.988329, 4.45218e-22, 2.45363e-22, 1.47218e-22, 5.1180783e-23], [24.043628, 4.41913e-22, 2.43633e-22, 1.4618e-22, 5.0823696e-23], [24.099054, 4.38654e-22, 2.41924e-22, 1.45154e-22, 5.0470819e-23], [24.154608, 4.35439e-22, 2.40236e-22, 1.44141e-22, 5.0122127e-23], [24.21029, 4.32268e-22, 2.38569e-22, 1.43141e-22, 4.97776e-23], [24.266101, 4.29142e-22, 2.36923e-22, 1.42154e-22, 4.9437216e-23], [24.32204, 4.2606e-22, 2.35299e-22, 1.41179e-22, 4.9100957e-23], [24.378108, 4.23022e-22, 2.33695e-22, 1.40217e-22, 4.8768804e-23], [24.434306, 4.20027e-22, 2.32112e-22, 1.39267e-22, 4.8440742e-23], [24.490632, 4.17076e-22, 2.3055e-22, 1.3833e-22, 4.8116754e-23], [24.547089, 4.14169e-22, 2.29008e-22, 1.37405e-22, 4.7796826e-23], [24.603676, 4.11305e-22, 2.27487e-22, 1.36492e-22, 4.7480947e-23], [24.660393, 4.08485e-22, 2.25986e-22, 1.35592e-22, 4.7169105e-23], [24.717241, 4.05708e-22, 2.24506e-22, 1.34704e-22, 4.6861291e-23], [24.774221, 4.02974e-22, 2.23047e-22, 1.33828e-22, 4.6557496e-23], [24.831331, 4.00283e-22, 2.21607e-22, 1.32964e-22, 4.6257715e-23], [24.888573, 3.97636e-22, 2.20188e-22, 1.32113e-22, 4.5961944e-23], [24.945947, 3.95031e-22, 2.1879e-22, 1.31274e-22, 4.5670179e-23], [25.003454, 3.9247e-22, 2.17412e-22, 1.30447e-22, 4.5382419e-23], [25.061093, 3.89953e-22, 2.16054e-22, 1.29633e-22, 4.5098667e-23], [25.118864, 3.87478e-22, 2.14717e-22, 1.2883e-22, 4.4818926e-23], [25.176769, 3.85047e-22, 2.134e-22, 1.2804e-22, 4.45432e-23], [25.234808, 3.8266e-22, 2.12104e-22, 1.27262e-22, 4.4271497e-23], [25.29298, 3.80317e-22, 2.10828e-22, 1.26497e-22, 4.4003828e-23], [25.351286, 3.78017e-22, 2.09573e-22, 1.25744e-22, 4.3740205e-23], [25.409727, 3.75762e-22, 2.08339e-22, 1.25003e-22, 4.3480643e-23], [25.468303, 3.7355e-22, 2.07125e-22, 1.24275e-22, 4.322516e-23], [25.527013, 3.71384e-22, 2.05932e-22, 1.23559e-22, 4.2973778e-23], [25.585859, 3.69262e-22, 2.0476e-22, 1.22856e-22, 4.2726519e-23], [25.64484, 3.67186e-22, 2.0361e-22, 1.22166e-22, 4.2483411e-23], [25.703958, 3.65155e-22, 2.0248e-22, 1.21488e-22, 4.2244484e-23], [25.763212, 3.6317e-22, 2.01373e-22, 1.20824e-22, 4.2009772e-23], [25.822602, 3.61231e-22, 2.00287e-22, 1.20172e-22, 4.1779314e-23], [25.882129, 3.5934e-22, 1.99222e-22, 1.19533e-22, 4.1553152e-23], [25.941794, 3.57495e-22, 1.9818e-22, 1.18908e-22, 4.133133e-23], [26.001596, 3.55699e-22, 1.97161e-22, 1.18296e-22, 4.1113902e-23], [26.061535, 3.53952e-22, 1.96164e-22, 1.17698e-22, 4.0900921e-23], [26.121614, 3.52254e-22, 1.9519e-22, 1.17114e-22, 4.0692449e-23], [26.18183, 3.50606e-22, 1.94239e-22, 1.16543e-22, 4.0488552e-23], [26.242185, 3.49008e-22, 1.93312e-22, 1.15987e-22, 4.0289303e-23], [26.30268, 3.47463e-22, 1.92409e-22, 1.15446e-22, 4.009478e-23], [26.363314, 3.4597e-22, 1.91531e-22, 1.14919e-22, 3.9905068e-23], [26.424088, 3.44531e-22, 1.90678e-22, 1.14407e-22, 3.972026e-23], [26.485001, 3.43147e-22, 1.8985e-22, 1.1391e-22, 3.9540457e-23], [26.546056, 3.41819e-22, 1.89049e-22, 1.13429e-22, 3.9365765e-23], [26.607251, 3.40548e-22, 1.88274e-22, 1.12964e-22, 3.9196303e-23], [26.668587, 3.39336e-22, 1.87526e-22, 1.12516e-22, 3.9032198e-23], [26.730064, 3.38184e-22, 1.86807e-22, 1.12084e-22, 3.8873587e-23], [26.791683, 3.37093e-22, 1.86116e-22, 1.1167e-22, 3.8720619e-23], [26.853444, 3.36067e-22, 1.85455e-22, 1.11273e-22, 3.8573453e-23], [26.915348, 3.35105e-22, 1.84825e-22, 1.10895e-22, 3.8432263e-23], [26.977394, 3.34211e-22, 1.84226e-22, 1.10535e-22, 3.8297236e-23], [27.039584, 3.33387e-22, 1.83659e-22, 1.10196e-22, 3.8168575e-23], [27.101916, 3.32635e-22, 1.83126e-22, 1.09876e-22, 3.8046499e-23], [27.164393, 3.31957e-22, 1.82628e-22, 1.09577e-22, 3.7931244e-23], [27.227013, 3.31356e-22, 1.82166e-22, 1.093e-22, 3.7823067e-23], [27.289778, 3.30835e-22, 1.81742e-22, 1.09045e-22, 3.7722245e-23], [27.352687, 3.30398e-22, 1.81356e-22, 1.08814e-22, 3.7629079e-23], [27.415742, 3.30047e-22, 1.81011e-22, 1.08607e-22, 3.7543895e-23], [27.478942, 3.29787e-22, 1.80709e-22, 1.08425e-22, 3.7467047e-23], [27.542287, 3.29621e-22, 1.80451e-22, 1.08271e-22, 3.7398916e-23], [27.605779, 3.29555e-22, 1.8024e-22, 1.08144e-22, 3.7339922e-23], [27.669416, 3.29592e-22, 1.80078e-22, 1.08047e-22, 3.7290516e-23], [27.733201, 3.29737e-22, 1.79966e-22, 1.0798e-22, 3.7251192e-23], [27.797133, 3.29998e-22, 1.79909e-22, 1.07946e-22, 3.7222487e-23], [27.861212, 3.30378e-22, 1.7991e-22, 1.07946e-22, 3.7204988e-23], [27.925438, 3.30886e-22, 1.7997e-22, 1.07982e-22, 3.7199335e-23], [27.989813, 3.31529e-22, 1.80094e-22, 1.08056e-22, 3.7206228e-23], [28.054336, 3.32314e-22, 1.80286e-22, 1.08171e-22, 3.7226433e-23], [28.119008, 3.33249e-22, 1.80549e-22, 1.0833e-22, 3.726079e-23], [28.183829, 3.34346e-22, 1.8089e-22, 1.08534e-22, 3.7310224e-23], [28.2488, 3.35614e-22, 1.81312e-22, 1.08787e-22, 3.7375748e-23], [28.31392, 3.37064e-22, 1.81821e-22, 1.09093e-22, 3.7458481e-23], [28.37919, 3.38709e-22, 1.82424e-22, 1.09454e-22, 3.7559658e-23], [28.444611, 3.40563e-22, 1.83127e-22, 1.09876e-22, 3.7680642e-23], [28.510183, 3.42642e-22, 1.83938e-22, 1.10363e-22, 3.7822946e-23], [28.575905, 3.44963e-22, 1.84866e-22, 1.1092e-22, 3.7988249e-23], [28.64178, 3.47545e-22, 1.8592e-22, 1.11552e-22, 3.8178419e-23], [28.707806, 3.50408e-22, 1.87109e-22, 1.12266e-22, 3.8395539e-23], [28.773984, 3.53577e-22, 1.88447e-22, 1.13068e-22, 3.864194e-23], [28.840315, 3.57078e-22, 1.89946e-22, 1.13968e-22, 3.8920233e-23], [28.906799, 3.6094e-22, 1.91621e-22, 1.14973e-22, 3.9233356e-23], [28.973436, 3.65198e-22, 1.93489e-22, 1.16093e-22, 3.9584621e-23], [29.040227, 3.69888e-22, 1.95567e-22, 1.1734e-22, 3.9977775e-23], [29.107171, 3.75054e-22, 1.97878e-22, 1.18727e-22, 4.0417073e-23], [29.17427, 3.80744e-22, 2.00447e-22, 1.20268e-22, 4.0907362e-23], [29.241524, 3.87015e-22, 2.03299e-22, 1.2198e-22, 4.1454188e-23], [29.308932, 3.93929e-22, 2.06468e-22, 1.23881e-22, 4.2063921e-23], [29.376497, 4.01561e-22, 2.09991e-22, 1.25994e-22, 4.2743913e-23], [29.444216, 4.09996e-22, 2.13909e-22, 1.28346e-22, 4.3502693e-23], [29.512092, 4.19334e-22, 2.18273e-22, 1.30964e-22, 4.4350201e-23], [29.580125, 4.29691e-22, 2.23141e-22, 1.33885e-22, 4.52981e-23], [29.648314, 4.41204e-22, 2.28582e-22, 1.37149e-22, 4.6360148e-23], [29.71666, 4.54038e-22, 2.34678e-22, 1.40807e-22, 4.7552701e-23], [29.785164, 4.68387e-22, 2.41525e-22, 1.44915e-22, 4.8895336e-23], [29.853826, 4.84489e-22, 2.49244e-22, 1.49546e-22, 5.0411694e-23], [29.922646, 5.02629e-22, 2.57977e-22, 1.54786e-22, 5.2130576e-23], [29.991625, 5.23163e-22, 2.67901e-22, 1.60741e-22, 5.4087431e-23], [30.060763, 5.46532e-22, 2.79238e-22, 1.67543e-22, 5.6326389e-23], [30.13006, 5.7329e-22, 2.92265e-22, 1.75359e-22, 5.8903082e-23], [30.199517, 6.04148e-22, 3.07338e-22, 1.84403e-22, 6.1888652e-23], [30.269134, 6.4003e-22, 3.24919e-22, 1.94952e-22, 6.5375563e-23], [30.338912, 6.82158e-22, 3.4562e-22, 2.07372e-22, 6.9486228e-23], [30.40885, 7.32188e-22, 3.70269e-22, 2.22161e-22, 7.438622e-23], [30.47895, 7.92414e-22, 4.00013e-22, 2.40008e-22, 8.0305141e-23], [30.549211, 8.66113e-22, 4.36491e-22, 2.61895e-22, 8.7570907e-23], [30.619634, 9.58136e-22, 4.82131e-22, 2.89278e-22, 9.6668723e-23], [30.69022, 1.07598e-21, 5.40678e-22, 3.24407e-22, 1.083483e-22], [30.760968, 1.23187e-21, 6.1825e-22, 3.7095e-22, 1.2383267e-22], [30.83188, 1.4472e-21, 7.25544e-22, 4.35326e-22, 1.4526183e-22], [30.902954, 1.76313e-21, 8.83136e-22, 5.29882e-22, 1.7675074e-22], [30.974193, 2.27016e-21, 1.13629e-21, 6.81771e-22, 2.2735144e-22], [31.045596, 3.21398e-21, 1.60783e-21, 9.64697e-22, 3.2163124e-22], [31.117163, 5.5771e-21, 2.78903e-21, 1.67342e-21, 5.5784231e-22], [31.188896, 2.21221e-20, 1.10612e-20, 6.6367e-21, 2.212241e-21], [31.260794, 1.09969e-20, 5.49868e-21, 3.29921e-21, 1.0997551e-21], [31.332857, 4.36927e-21, 2.18522e-21, 1.31113e-21, 4.3708819e-22], [31.405087, 2.71476e-21, 1.3583e-21, 8.1498e-22, 2.7173066e-22], [31.477483, 1.96412e-21, 9.83308e-22, 5.89985e-22, 1.967574e-22], [31.550046, 1.53623e-21, 7.69686e-22, 4.61812e-22, 1.5405735e-22], [31.622777, 1.2602e-21, 6.31987e-22, 3.79192e-22, 1.2654083e-22], [31.695675, 1.06768e-21, 5.36032e-22, 3.21619e-22, 1.0737226e-22], [31.768741, 9.25985e-22, 4.65476e-22, 2.79285e-22, 9.3282408e-23], [31.841975, 8.175e-22, 4.11516e-22, 2.4691e-22, 8.2510866e-23], [31.915379, 7.31911e-22, 3.68993e-22, 2.21396e-22, 7.4025605e-23], [31.988951, 6.62768e-22, 3.34681e-22, 2.00809e-22, 6.7181553e-23], [32.062693, 6.05832e-22, 3.06461e-22, 1.83877e-22, 6.1554765e-23], [32.136605, 5.58203e-22, 2.82883e-22, 1.6973e-22, 5.6855146e-23], [32.210688, 5.1783e-22, 2.62919e-22, 1.57752e-22, 5.2877491e-23], [32.284941, 4.83219e-22, 2.45825e-22, 1.47495e-22, 4.9472557e-23], [32.359366, 4.53259e-22, 2.31044e-22, 1.38626e-22, 4.6529205e-23], [32.433962, 4.27107e-22, 2.18154e-22, 1.30892e-22, 4.3962982e-23], [32.50873, 4.04107e-22, 2.06828e-22, 1.24097e-22, 4.1708572e-23], [32.58367, 3.83747e-22, 1.9681e-22, 1.18086e-22, 3.9714673e-23], [32.658783, 3.65618e-22, 1.87895e-22, 1.12737e-22, 3.7940443e-23], [32.734069, 3.4939e-22, 1.79918e-22, 1.07951e-22, 3.6352969e-23], [32.809529, 3.34793e-22, 1.72746e-22, 1.03648e-22, 3.4925442e-23], [32.885163, 3.21608e-22, 1.66268e-22, 9.97609e-23, 3.3635813e-23], [32.960971, 3.09649e-22, 1.60392e-22, 9.62354e-23, 3.2465787e-23], [33.036954, 2.98762e-22, 1.55042e-22, 9.30253e-23, 3.1400065e-23], [33.113112, 2.88819e-22, 1.50153e-22, 9.00918e-23, 3.0425762e-23], [33.189446, 2.79708e-22, 1.4567e-22, 8.7402e-23, 2.9531953e-23], [33.265955, 2.71336e-22, 1.41547e-22, 8.4928e-23, 2.8709318e-23], [33.342641, 2.6362e-22, 1.37742e-22, 8.26455e-23, 2.7949867e-23], [33.419504, 2.56493e-22, 1.34223e-22, 8.05338e-23, 2.7246708e-23], [33.496544, 2.49892e-22, 1.30958e-22, 7.85749e-23, 2.6593876e-23], [33.573761, 2.43764e-22, 1.27922e-22, 7.67532e-23, 2.598618e-23], [33.651157, 2.38064e-22, 1.25092e-22, 7.5055e-23, 2.5419085e-23], [33.728731, 2.32751e-22, 1.22447e-22, 7.34682e-23, 2.4888617e-23], [33.806484, 2.27789e-22, 1.1997e-22, 7.19822e-23, 2.4391279e-23], [33.884416, 2.23145e-22, 1.17646e-22, 7.05878e-23, 2.3923984e-23], [33.962527, 2.18792e-22, 1.15461e-22, 6.92765e-23, 2.3483997e-23], [34.040819, 2.14704e-22, 1.13402e-22, 6.80412e-23, 2.3068891e-23], [34.119291, 2.10859e-22, 1.11459e-22, 6.68752e-23, 2.2676503e-23], [34.197944, 2.07237e-22, 1.09621e-22, 6.57726e-23, 2.2304902e-23], [34.276779, 2.0382e-22, 1.0788e-22, 6.47282e-23, 2.195236e-23], [34.355795, 2.00592e-22, 1.06229e-22, 6.37372e-23, 2.1617325e-23], [34.434993, 1.97537e-22, 1.04659e-22, 6.27955e-23, 2.1298404e-23], [34.514374, 1.94643e-22, 1.03165e-22, 6.18992e-23, 2.0994338e-23], [34.593938, 1.91897e-22, 1.01742e-22, 6.10449e-23, 2.0703992e-23], [34.673685, 1.8929e-22, 1.00382e-22, 6.02294e-23, 2.0426341e-23], [34.753616, 1.8681e-22, 9.90831e-23, 5.94498e-23, 2.0160451e-23], [34.833732, 1.84448e-22, 9.78395e-23, 5.87037e-23, 1.9905477e-23], [34.914032, 1.82198e-22, 9.66476e-23, 5.79886e-23, 1.966065e-23], [34.994517, 1.8005e-22, 9.55039e-23, 5.73023e-23, 1.9425269e-23], [35.075187, 1.77998e-22, 9.44051e-23, 5.66431e-23, 1.9198692e-23], [35.156044, 1.76037e-22, 9.33482e-23, 5.60089e-23, 1.8980336e-23], [35.237087, 1.74159e-22, 9.23305e-23, 5.53983e-23, 1.8769665e-23], [35.318317, 1.7236e-22, 9.13495e-23, 5.48097e-23, 1.8566186e-23], [35.399734, 1.70635e-22, 9.04028e-23, 5.42417e-23, 1.8369449e-23], [35.481339, 1.68979e-22, 8.94884e-23, 5.36931e-23, 1.817904e-23], [35.563132, 1.67388e-22, 8.86043e-23, 5.31626e-23, 1.7994576e-23], [35.645113, 1.65859e-22, 8.77487e-23, 5.26492e-23, 1.7815705e-23], [35.727284, 1.64387e-22, 8.69199e-23, 5.21519e-23, 1.7642102e-23], [35.809644, 1.6297e-22, 8.61164e-23, 5.16698e-23, 1.7473468e-23], [35.892193, 1.61604e-22, 8.53367e-23, 5.1202e-23, 1.7309524e-23], [35.974934, 1.60286e-22, 8.45797e-23, 5.07478e-23, 1.7150013e-23], [36.057864, 1.59015e-22, 8.38439e-23, 5.03063e-23, 1.6994696e-23], [36.140986, 1.57787e-22, 8.31283e-23, 4.9877e-23, 1.6843352e-23], [36.2243, 1.566e-22, 8.24319e-23, 4.94591e-23, 1.6695775e-23], [36.307805, 1.55452e-22, 8.17536e-23, 4.90522e-23, 1.6551774e-23], [36.391504, 1.54341e-22, 8.10925e-23, 4.86555e-23, 1.6411169e-23], [36.475395, 1.53265e-22, 8.04479e-23, 4.82687e-23, 1.6273795e-23], [36.559479, 1.52223e-22, 7.98188e-23, 4.78913e-23, 1.6139497e-23], [36.643757, 1.51212e-22, 7.92045e-23, 4.75227e-23, 1.600813e-23], [36.72823, 1.50232e-22, 7.86044e-23, 4.71627e-23, 1.5879558e-23], [36.812897, 1.49281e-22, 7.80178e-23, 4.68107e-23, 1.5753654e-23], [36.89776, 1.48357e-22, 7.74441e-23, 4.64664e-23, 1.5630301e-23], [36.982818, 1.47459e-22, 7.68826e-23, 4.61296e-23, 1.5509386e-23], [37.068072, 1.46587e-22, 7.6333e-23, 4.57998e-23, 1.5390805e-23], [37.153523, 1.45738e-22, 7.57946e-23, 4.54767e-23, 1.5274461e-23], [37.239171, 1.44913e-22, 7.5267e-23, 4.51602e-23, 1.516026e-23], [37.325016, 1.44109e-22, 7.47497e-23, 4.48498e-23, 1.5048117e-23], [37.411059, 1.43326e-22, 7.42423e-23, 4.45454e-23, 1.493795e-23], [37.4973, 1.42564e-22, 7.37445e-23, 4.42467e-23, 1.4829683e-23], [37.58374, 1.41821e-22, 7.32559e-23, 4.39535e-23, 1.4723242e-23], [37.67038, 1.41096e-22, 7.2776e-23, 4.36656e-23, 1.4618559e-23], [37.757219, 1.4039e-22, 7.23046e-23, 4.33828e-23, 1.4515571e-23], [37.844258, 1.397e-22, 7.18414e-23, 4.31049e-23, 1.4414215e-23], [37.931498, 1.39027e-22, 7.13861e-23, 4.28316e-23, 1.4314435e-23], [38.01894, 1.3837e-22, 7.09383e-23, 4.2563e-23, 1.4216176e-23], [38.106582, 1.37728e-22, 7.04978e-23, 4.22987e-23, 1.4119386e-23], [38.194427, 1.371e-22, 7.00644e-23, 4.20386e-23, 1.4024017e-23], [38.282474, 1.36487e-22, 6.96378e-23, 4.17827e-23, 1.3930021e-23], [38.370725, 1.35888e-22, 6.92178e-23, 4.15307e-23, 1.3837356e-23], [38.459178, 1.35302e-22, 6.88042e-23, 4.12825e-23, 1.3745979e-23], [38.547836, 1.34728e-22, 6.83968e-23, 4.10381e-23, 1.365585e-23], [38.636698, 1.34167e-22, 6.79953e-23, 4.07972e-23, 1.3566933e-23], [38.725764, 1.33617e-22, 6.75996e-23, 4.05598e-23, 1.347919e-23], [38.815037, 1.33079e-22, 6.72096e-23, 4.03257e-23, 1.3392588e-23], [38.904514, 1.32552e-22, 6.6825e-23, 4.0095e-23, 1.3307094e-23], [38.994199, 1.32036e-22, 6.64457e-23, 3.98674e-23, 1.3222678e-23], [39.08409, 1.3153e-22, 6.60715e-23, 3.96429e-23, 1.313931e-23], [39.174188, 1.31035e-22, 6.57023e-23, 3.94214e-23, 1.3056961e-23], [39.264494, 1.30549e-22, 6.5338e-23, 3.92028e-23, 1.2975604e-23], [39.355008, 1.30072e-22, 6.49784e-23, 3.8987e-23, 1.2895215e-23], [39.44573, 1.29605e-22, 6.46234e-23, 3.8774e-23, 1.2815767e-23], [39.536662, 1.29147e-22, 6.42728e-23, 3.85637e-23, 1.2737239e-23], [39.627803, 1.28697e-22, 6.39266e-23, 3.8356e-23, 1.2659607e-23], [39.719155, 1.28256e-22, 6.35847e-23, 3.81508e-23, 1.2582849e-23], [39.810717, 1.27822e-22, 6.32469e-23, 3.79481e-23, 1.2506946e-23], [39.90249, 1.27397e-22, 6.29131e-23, 3.77479e-23, 1.2431877e-23], [39.994475, 1.26979e-22, 6.25833e-23, 3.755e-23, 1.2357624e-23], [40.086672, 1.26569e-22, 6.22573e-23, 3.73544e-23, 1.2284167e-23], [40.179081, 1.26167e-22, 6.19351e-23, 3.71611e-23, 1.2211491e-23], [40.271703, 1.25771e-22, 6.16165e-23, 3.69699e-23, 1.2139577e-23], [40.364539, 1.25382e-22, 6.13016e-23, 3.67809e-23, 1.206841e-23], [40.457589, 1.25e-22, 6.09901e-23, 3.65941e-23, 1.1997975e-23], [40.550854, 1.24624e-22, 6.06821e-23, 3.64092e-23, 1.1928256e-23], [40.644333, 1.24255e-22, 6.03774e-23, 3.62264e-23, 1.1859239e-23], [40.738028, 1.23892e-22, 6.0076e-23, 3.60456e-23, 1.1790911e-23], [40.831939, 1.23535e-22, 5.97778e-23, 3.58667e-23, 1.1723258e-23], [40.926066, 1.23184e-22, 5.94827e-23, 3.56896e-23, 1.1656266e-23], [41.02041, 1.22839e-22, 5.91908e-23, 3.55145e-23, 1.1589925e-23], [41.114972, 1.225e-22, 5.89019e-23, 3.53411e-23, 1.1524222e-23], [41.209752, 1.22166e-22, 5.86159e-23, 3.51695e-23, 1.1459145e-23], [41.30475, 1.21837e-22, 5.83328e-23, 3.49997e-23, 1.1394684e-23], [41.399967, 1.21513e-22, 5.80526e-23, 3.48316e-23, 1.1330827e-23], [41.495404, 1.21195e-22, 5.77752e-23, 3.46651e-23, 1.1267565e-23], [41.591061, 1.20882e-22, 5.75005e-23, 3.45003e-23, 1.1204887e-23], [41.686938, 1.20573e-22, 5.72286e-23, 3.43371e-23, 1.1142783e-23], [41.783037, 1.2027e-22, 5.69593e-23, 3.41756e-23, 1.1081245e-23], [41.879357, 1.19971e-22, 5.66925e-23, 3.40155e-23, 1.1020263e-23], [41.975898, 1.19676e-22, 5.64284e-23, 3.3857e-23, 1.0959829e-23], [42.072663, 1.19386e-22, 5.61667e-23, 3.37e-23, 1.0899933e-23], [42.16965, 1.19101e-22, 5.59076e-23, 3.35445e-23, 1.0840569e-23], [42.266861, 1.1882e-22, 5.56508e-23, 3.33905e-23, 1.0781726e-23], [42.364297, 1.18543e-22, 5.53965e-23, 3.32379e-23, 1.0723399e-23], [42.461956, 1.1827e-22, 5.51444e-23, 3.30867e-23, 1.066558e-23], [42.559841, 1.18001e-22, 5.48948e-23, 3.29369e-23, 1.060826e-23], [42.657952, 1.17736e-22, 5.46473e-23, 3.27884e-23, 1.0551434e-23], [42.756289, 1.17475e-22, 5.44022e-23, 3.26413e-23, 1.0495094e-23], [42.854852, 1.17217e-22, 5.41592e-23, 3.24955e-23, 1.0439233e-23], [42.953643, 1.16964e-22, 5.39184e-23, 3.2351e-23, 1.0383846e-23], [43.052661, 1.16714e-22, 5.36798e-23, 3.22079e-23, 1.0328925e-23], [43.151908, 1.16467e-22, 5.34432e-23, 3.20659e-23, 1.0274465e-23], [43.251383, 1.16224e-22, 5.32088e-23, 3.19253e-23, 1.022046e-23], [43.351088, 1.15985e-22, 5.29763e-23, 3.17858e-23, 1.0166903e-23], [43.451022, 1.15749e-22, 5.27459e-23, 3.16476e-23, 1.011379e-23], [43.551187, 1.15516e-22, 5.25175e-23, 3.15105e-23, 1.0061114e-23], [43.651583, 1.15286e-22, 5.22911e-23, 3.13747e-23, 1.0008871e-23], [43.752211, 1.1506e-22, 5.20666e-23, 3.124e-23, 9.9570548e-24], [43.85307, 1.14837e-22, 5.1844e-23, 3.11064e-23, 9.9056607e-24], [43.954162, 1.14617e-22, 5.16233e-23, 3.0974e-23, 9.8546835e-24], [44.055486, 1.14399e-22, 5.14045e-23, 3.08427e-23, 9.8041186e-24], [44.157045, 1.14185e-22, 5.11875e-23, 3.07125e-23, 9.7539611e-24], [44.258837, 1.13974e-22, 5.09723e-23, 3.05834e-23, 9.7042064e-24], [44.360864, 1.13765e-22, 5.07589e-23, 3.04554e-23, 9.6548501e-24], [44.463127, 1.1356e-22, 5.05473e-23, 3.03284e-23, 9.6058877e-24], [44.565625, 1.13357e-22, 5.03374e-23, 3.02024e-23, 9.5573149e-24], [44.668359, 1.13156e-22, 5.01293e-23, 3.00776e-23, 9.5091276e-24], [44.77133, 1.12959e-22, 4.99228e-23, 2.99537e-23, 9.4613216e-24], [44.874539, 1.12764e-22, 4.9718e-23, 2.98308e-23, 9.4138928e-24], [44.977985, 1.12571e-22, 4.95149e-23, 2.9709e-23, 9.3668375e-24], [45.08167, 1.12382e-22, 4.93135e-23, 2.95881e-23, 9.3201517e-24], [45.185594, 1.12194e-22, 4.91136e-23, 2.94682e-23, 9.2738317e-24], [45.289758, 1.12009e-22, 4.89154e-23, 2.93492e-23, 9.2278738e-24], [45.394162, 1.11827e-22, 4.87187e-23, 2.92312e-23, 9.1822745e-24], [45.498806, 1.11647e-22, 4.85236e-23, 2.91142e-23, 9.13703e-24], [45.603692, 1.11469e-22, 4.83301e-23, 2.89981e-23, 9.0921371e-24], [45.708819, 1.11293e-22, 4.81381e-23, 2.88829e-23, 9.0475924e-24], [45.814189, 1.1112e-22, 4.79476e-23, 2.87686e-23, 9.0033924e-24], [45.919801, 1.10949e-22, 4.77587e-23, 2.86552e-23, 8.959534e-24], [46.025657, 1.1078e-22, 4.75712e-23, 2.85427e-23, 8.916014e-24], [46.131757, 1.10613e-22, 4.73851e-23, 2.84311e-23, 8.8728293e-24], [46.238102, 1.10448e-22, 4.72006e-23, 2.83203e-23, 8.8299767e-24], [46.344692, 1.10286e-22, 4.70174e-23, 2.82105e-23, 8.7874533e-24], [46.451528, 1.10125e-22, 4.68357e-23, 2.81014e-23, 8.7452562e-24], [46.558609, 1.09966e-22, 4.66554e-23, 2.79932e-23, 8.7033824e-24], [46.665938, 1.0981e-22, 4.64765e-23, 2.78859e-23, 8.661829e-24], [46.773514, 1.09655e-22, 4.6299e-23, 2.77794e-23, 8.6205933e-24], [46.881338, 1.09502e-22, 4.61228e-23, 2.76737e-23, 8.5796726e-24], [46.989411, 1.09352e-22, 4.5948e-23, 2.75688e-23, 8.5390641e-24], [47.097733, 1.09203e-22, 4.57745e-23, 2.74647e-23, 8.4987652e-24], [47.206304, 1.09056e-22, 4.56023e-23, 2.73614e-23, 8.4587733e-24], [47.315126, 1.0891e-22, 4.54315e-23, 2.72589e-23, 8.4190859e-24], [47.424199, 1.08767e-22, 4.52619e-23, 2.71572e-23, 8.3797003e-24], [47.533523, 1.08625e-22, 4.50937e-23, 2.70562e-23, 8.3406141e-24], [47.643099, 1.08485e-22, 4.49267e-23, 2.6956e-23, 8.301825e-24], [47.752927, 1.08346e-22, 4.4761e-23, 2.68566e-23, 8.2633305e-24], [47.863009, 1.0821e-22, 4.45965e-23, 2.67579e-23, 8.2251282e-24], [47.973345, 1.08075e-22, 4.44333e-23, 2.666e-23, 8.1872158e-24], [48.083935, 1.07941e-22, 4.42712e-23, 2.65627e-23, 8.149591e-24], [48.19478, 1.0781e-22, 4.41104e-23, 2.64663e-23, 8.1122516e-24], [48.30588, 1.07679e-22, 4.39509e-23, 2.63705e-23, 8.0751954e-24], [48.417237, 1.07551e-22, 4.37925e-23, 2.62755e-23, 8.0384202e-24], [48.52885, 1.07424e-22, 4.36352e-23, 2.61811e-23, 8.0019238e-24], [48.640721, 1.07298e-22, 4.34792e-23, 2.60875e-23, 7.9657041e-24], [48.752849, 1.07174e-22, 4.33243e-23, 2.59946e-23, 7.929759e-24], [48.865236, 1.07051e-22, 4.31706e-23, 2.59024e-23, 7.8940865e-24], [48.977882, 1.0693e-22, 4.3018e-23, 2.58108e-23, 7.8586845e-24], [49.090788, 1.0681e-22, 4.28665e-23, 2.57199e-23, 7.823551e-24], [49.203954, 1.06692e-22, 4.27162e-23, 2.56297e-23, 7.7886841e-24], [49.31738, 1.06575e-22, 4.2567e-23, 2.55402e-23, 7.7540818e-24], [49.431069, 1.0646e-22, 4.24188e-23, 2.54513e-23, 7.7197422e-24], [49.545019, 1.06345e-22, 4.22718e-23, 2.53631e-23, 7.6856633e-24], [49.659232, 1.06232e-22, 4.21259e-23, 2.52755e-23, 7.6518434e-24], [49.773708, 1.06121e-22, 4.1981e-23, 2.51886e-23, 7.6182806e-24], [49.888449, 1.06011e-22, 4.18372e-23, 2.51023e-23, 7.584973e-24], [50.003453, 1.05902e-22, 4.16944e-23, 2.50166e-23, 7.5519188e-24], [50.118723, 1.05794e-22, 4.15527e-23, 2.49316e-23, 7.5191163e-24], [50.234259, 1.05688e-22, 4.1412e-23, 2.48472e-23, 7.4865637e-24], [50.350061, 1.05582e-22, 4.12724e-23, 2.47634e-23, 7.4542593e-24], [50.46613, 1.05478e-22, 4.11338e-23, 2.46803e-23, 7.4222013e-24], [50.582466, 1.05376e-22, 4.09961e-23, 2.45977e-23, 7.3903881e-24], [50.699071, 1.05274e-22, 4.08595e-23, 2.45157e-23, 7.358818e-24], [50.815944, 1.05174e-22, 4.07239e-23, 2.44344e-23, 7.3274894e-24], [50.933087, 1.05074e-22, 4.05893e-23, 2.43536e-23, 7.2964005e-24], [51.0505, 1.04976e-22, 4.04557e-23, 2.42734e-23, 7.2655498e-24], [51.168184, 1.04879e-22, 4.0323e-23, 2.41938e-23, 7.2349357e-24], [51.286138, 1.04783e-22, 4.01913e-23, 2.41148e-23, 7.2045566e-24], [51.404365, 1.04688e-22, 4.00605e-23, 2.40363e-23, 7.1744109e-24], [51.522864, 1.04595e-22, 3.99307e-23, 2.39584e-23, 7.1444971e-24], [51.641637, 1.04502e-22, 3.98019e-23, 2.38811e-23, 7.1148137e-24], [51.760683, 1.0441e-22, 3.9674e-23, 2.38044e-23, 7.085359e-24], [51.880004, 1.0432e-22, 3.9547e-23, 2.37282e-23, 7.0561318e-24], [51.9996, 1.0423e-22, 3.94209e-23, 2.36525e-23, 7.0271303e-24], [52.119471, 1.04142e-22, 3.92957e-23, 2.35774e-23, 6.9983533e-24], [52.239619, 1.04054e-22, 3.91715e-23, 2.35029e-23, 6.9697991e-24], [52.360044, 1.03968e-22, 3.90481e-23, 2.34289e-23, 6.9414665e-24], [52.480746, 1.03882e-22, 3.89256e-23, 2.33554e-23, 6.9133538e-24], [52.601727, 1.03798e-22, 3.88041e-23, 2.32824e-23, 6.8854598e-24], [52.722986, 1.03714e-22, 3.86834e-23, 2.321e-23, 6.8577831e-24], [52.844525, 1.03631e-22, 3.85635e-23, 2.31381e-23, 6.8303221e-24], [52.966344, 1.03549e-22, 3.84446e-23, 2.30667e-23, 6.8030757e-24], [53.088444, 1.03469e-22, 3.83264e-23, 2.29959e-23, 6.7760423e-24], [53.210826, 1.03389e-22, 3.82092e-23, 2.29255e-23, 6.7492207e-24], [53.33349, 1.0331e-22, 3.80928e-23, 2.28557e-23, 6.7226095e-24], [53.456436, 1.03231e-22, 3.79772e-23, 2.27863e-23, 6.6962073e-24], [53.579666, 1.03154e-22, 3.78625e-23, 2.27175e-23, 6.6700129e-24], [53.70318, 1.03078e-22, 3.77485e-23, 2.26491e-23, 6.644025e-24], [53.826978, 1.03002e-22, 3.76355e-23, 2.25813e-23, 6.6182422e-24], [53.951062, 1.02927e-22, 3.75232e-23, 2.25139e-23, 6.5926633e-24], [54.075432, 1.02853e-22, 3.74117e-23, 2.2447e-23, 6.567287e-24], [54.200089, 1.0278e-22, 3.7301e-23, 2.23806e-23, 6.542112e-24], [54.325033, 1.02708e-22, 3.71912e-23, 2.23147e-23, 6.517137e-24], [54.450265, 1.02637e-22, 3.70821e-23, 2.22493e-23, 6.4923609e-24], [54.575786, 1.02566e-22, 3.69738e-23, 2.21843e-23, 6.4677824e-24], [54.701596, 1.02496e-22, 3.68663e-23, 2.21198e-23, 6.4434003e-24], [54.827696, 1.02427e-22, 3.67595e-23, 2.20557e-23, 6.4192132e-24], [54.954087, 1.02359e-22, 3.66536e-23, 2.19921e-23, 6.3952201e-24], [55.08077, 1.02291e-22, 3.65484e-23, 2.1929e-23, 6.3714198e-24], [55.207744, 1.02224e-22, 3.64439e-23, 2.18663e-23, 6.3478109e-24], [55.335011, 1.02158e-22, 3.63402e-23, 2.18041e-23, 6.3243925e-24], [55.462571, 1.02093e-22, 3.62373e-23, 2.17424e-23, 6.3011632e-24], [55.590426, 1.02028e-22, 3.6135e-23, 2.1681e-23, 6.2781219e-24], [55.718575, 1.01964e-22, 3.60336e-23, 2.16201e-23, 6.2552674e-24], [55.847019, 1.01901e-22, 3.59328e-23, 2.15597e-23, 6.2325987e-24], [55.97576, 1.01838e-22, 3.58328e-23, 2.14997e-23, 6.2101144e-24], [56.104798, 1.01776e-22, 3.57335e-23, 2.14401e-23, 6.1878136e-24], [56.234133, 1.01715e-22, 3.56349e-23, 2.13809e-23, 6.1656951e-24], [56.363766, 1.01655e-22, 3.5537e-23, 2.13222e-23, 6.1437577e-24], [56.493697, 1.01595e-22, 3.54399e-23, 2.12639e-23, 6.1220003e-24], [56.623929, 1.01536e-22, 3.53434e-23, 2.1206e-23, 6.1004219e-24], [56.754461, 1.01477e-22, 3.52476e-23, 2.11486e-23, 6.0790213e-24], [56.885293, 1.01419e-22, 3.51525e-23, 2.10915e-23, 6.0577974e-24], [57.016427, 1.01362e-22, 3.50581e-23, 2.10349e-23, 6.0367491e-24], [57.147864, 1.01305e-22, 3.49644e-23, 2.09786e-23, 6.0158754e-24], [57.279603, 1.01249e-22, 3.48713e-23, 2.09228e-23, 5.9951751e-24], [57.411646, 1.01194e-22, 3.4779e-23, 2.08674e-23, 5.9746472e-24], [57.543994, 1.01139e-22, 3.46872e-23, 2.08123e-23, 5.9542906e-24], [57.676646, 1.01085e-22, 3.45962e-23, 2.07577e-23, 5.9341043e-24], [57.809605, 1.01031e-22, 3.45058e-23, 2.07035e-23, 5.9140871e-24], [57.94287, 1.00978e-22, 3.44161e-23, 2.06496e-23, 5.8942381e-24], [58.076442, 1.00926e-22, 3.4327e-23, 2.05962e-23, 5.8745562e-24], [58.210322, 1.00874e-22, 3.42385e-23, 2.05431e-23, 5.8550404e-24], [58.34451, 1.00823e-22, 3.41507e-23, 2.04904e-23, 5.8356896e-24], [58.479008, 1.00772e-22, 3.40635e-23, 2.04381e-23, 5.8165028e-24], [58.613816, 1.00722e-22, 3.3977e-23, 2.03862e-23, 5.7974789e-24], [58.748935, 1.00672e-22, 3.38911e-23, 2.03346e-23, 5.778617e-24], [58.884366, 1.00623e-22, 3.38058e-23, 2.02835e-23, 5.759916e-24], [59.020108, 1.00575e-22, 3.37211e-23, 2.02326e-23, 5.7413749e-24], [59.156163, 1.00527e-22, 3.3637e-23, 2.01822e-23, 5.7229928e-24], [59.292532, 1.00479e-22, 3.35535e-23, 2.01321e-23, 5.7047685e-24], [59.429216, 1.00432e-22, 3.34707e-23, 2.00824e-23, 5.6867012e-24], [59.566214, 1.00386e-22, 3.33884e-23, 2.00331e-23, 5.6687898e-24], [59.703529, 1.0034e-22, 3.33068e-23, 1.99841e-23, 5.6510333e-24], [59.84116, 1.00295e-22, 3.32257e-23, 1.99354e-23, 5.6334307e-24], [59.979108, 1.0025e-22, 3.31452e-23, 1.98871e-23, 5.6159811e-24], [60.117374, 1.00205e-22, 3.30653e-23, 1.98392e-23, 5.5986835e-24], [60.255959, 1.00161e-22, 3.2986e-23, 1.97916e-23, 5.5815369e-24], [60.394863, 1.00118e-22, 3.29072e-23, 1.97443e-23, 5.5645404e-24], [60.534087, 1.00075e-22, 3.28291e-23, 1.96974e-23, 5.5476929e-24], [60.673633, 1.00032e-22, 3.27515e-23, 1.96509e-23, 5.5309936e-24], [60.8135, 9.99904e-23, 3.26744e-23, 1.96046e-23, 5.5144414e-24], [60.95369, 9.99489e-23, 3.25979e-23, 1.95588e-23, 5.4980354e-24], [61.094202, 9.99077e-23, 3.2522e-23, 1.95132e-23, 5.4817747e-24], [61.235039, 9.98671e-23, 3.24466e-23, 1.9468e-23, 5.4656583e-24], [61.376201, 9.98269e-23, 3.23718e-23, 1.94231e-23, 5.4496853e-24], [61.517687, 9.97871e-23, 3.22975e-23, 1.93785e-23, 5.4338547e-24], [61.6595, 9.97478e-23, 3.22238e-23, 1.93343e-23, 5.4181657e-24], [61.80164, 9.97089e-23, 3.21506e-23, 1.92904e-23, 5.4026172e-24], [61.944108, 9.96704e-23, 3.20779e-23, 1.92467e-23, 5.3872083e-24], [62.086903, 9.96324e-23, 3.20058e-23, 1.92035e-23, 5.3719382e-24], [62.230029, 9.95948e-23, 3.19342e-23, 1.91605e-23, 5.3568059e-24], [62.373484, 9.95576e-23, 3.18631e-23, 1.91178e-23, 5.3418104e-24], [62.517269, 9.95208e-23, 3.17925e-23, 1.90755e-23, 5.3269509e-24], [62.661386, 9.94844e-23, 3.17225e-23, 1.90335e-23, 5.3122265e-24], [62.805836, 9.94485e-23, 3.16529e-23, 1.89917e-23, 5.2976362e-24], [62.950618, 9.94129e-23, 3.15839e-23, 1.89503e-23, 5.2831792e-24], [63.095734, 9.93777e-23, 3.15154e-23, 1.89092e-23, 5.2688545e-24], [63.241185, 9.93429e-23, 3.14473e-23, 1.88684e-23, 5.2546613e-24], [63.386971, 9.93086e-23, 3.13798e-23, 1.88279e-23, 5.2405986e-24], [63.533093, 9.92745e-23, 3.13128e-23, 1.87877e-23, 5.2266656e-24], [63.679552, 9.92409e-23, 3.12462e-23, 1.87477e-23, 5.2128614e-24], [63.826349, 9.92077e-23, 3.11802e-23, 1.87081e-23, 5.199185e-24], [63.973484, 9.91748e-23, 3.11146e-23, 1.86688e-23, 5.1856357e-24], [64.120958, 9.91423e-23, 3.10495e-23, 1.86297e-23, 5.1722125e-24], [64.268772, 9.91102e-23, 3.09849e-23, 1.85909e-23, 5.1589145e-24], [64.416927, 9.90784e-23, 3.09208e-23, 1.85525e-23, 5.1457409e-24], [64.565423, 9.9047e-23, 3.08571e-23, 1.85143e-23, 5.1326908e-24], [64.714262, 9.9016e-23, 3.07939e-23, 1.84763e-23, 5.1197634e-24], [64.863443, 9.89852e-23, 3.07312e-23, 1.84387e-23, 5.1069577e-24], [65.012969, 9.89549e-23, 3.06689e-23, 1.84014e-23, 5.094273e-24], [65.162839, 9.89249e-23, 3.06071e-23, 1.83643e-23, 5.0817083e-24], [65.313055, 9.88952e-23, 3.05458e-23, 1.83275e-23, 5.0692628e-24], [65.463617, 9.88659e-23, 3.04849e-23, 1.82909e-23, 5.0569357e-24], [65.614527, 9.88369e-23, 3.04244e-23, 1.82546e-23, 5.0447261e-24], [65.765784, 9.88083e-23, 3.03644e-23, 1.82186e-23, 5.0326331e-24], [65.91739, 9.87799e-23, 3.03048e-23, 1.81829e-23, 5.020656e-24], [66.069345, 9.87519e-23, 3.02457e-23, 1.81474e-23, 5.0087938e-24], [66.22165, 9.87243e-23, 3.0187e-23, 1.81122e-23, 4.9970458e-24], [66.374307, 9.86969e-23, 3.01288e-23, 1.80773e-23, 4.9854111e-24], [66.527316, 9.86699e-23, 3.0071e-23, 1.80426e-23, 4.9738888e-24], [66.680677, 9.86432e-23, 3.00136e-23, 1.80082e-23, 4.9624782e-24], [66.834392, 9.86167e-23, 2.99566e-23, 1.7974e-23, 4.9511784e-24], [66.988461, 9.85906e-23, 2.99001e-23, 1.79401e-23, 4.9399886e-24], [67.142885, 9.85648e-23, 2.9844e-23, 1.79064e-23, 4.928908e-24], [67.297666, 9.85394e-23, 2.97883e-23, 1.7873e-23, 4.9179358e-24], [67.452803, 9.85142e-23, 2.9733e-23, 1.78398e-23, 4.9070711e-24], [67.608298, 9.84893e-23, 2.96781e-23, 1.78069e-23, 4.8963131e-24], [67.764151, 9.84647e-23, 2.96236e-23, 1.77742e-23, 4.8856611e-24], [67.920363, 9.84404e-23, 2.95696e-23, 1.77417e-23, 4.8751142e-24], [68.076936, 9.84164e-23, 2.95159e-23, 1.77095e-23, 4.8646717e-24], [68.233869, 9.83926e-23, 2.94626e-23, 1.76776e-23, 4.8543326e-24], [68.391165, 9.83692e-23, 2.94098e-23, 1.76459e-23, 4.8440963e-24], [68.548823, 9.8346e-23, 2.93573e-23, 1.76144e-23, 4.833962e-24], [68.706844, 9.83231e-23, 2.93052e-23, 1.75831e-23, 4.8239287e-24], [68.86523, 9.83005e-23, 2.92535e-23, 1.75521e-23, 4.8139959e-24], [69.02398, 9.82782e-23, 2.92022e-23, 1.75213e-23, 4.8041626e-24], [69.183097, 9.82561e-23, 2.91513e-23, 1.74908e-23, 4.7944282e-24], [69.342581, 9.82343e-23, 2.91008e-23, 1.74605e-23, 4.7847917e-24], [69.502432, 9.82128e-23, 2.90506e-23, 1.74304e-23, 4.7752526e-24], [69.662651, 9.81916e-23, 2.90008e-23, 1.74005e-23, 4.7658099e-24], [69.82324, 9.81706e-23, 2.89514e-23, 1.73709e-23, 4.7564629e-24], [69.9842, 9.81498e-23, 2.89024e-23, 1.73414e-23, 4.7472108e-24], [70.14553, 9.81294e-23, 2.88537e-23, 1.73122e-23, 4.738053e-24], [70.307232, 9.81092e-23, 2.88054e-23, 1.72832e-23, 4.7289885e-24], [70.469307, 9.80892e-23, 2.87575e-23, 1.72545e-23, 4.7200168e-24], [70.631755, 9.80695e-23, 2.87099e-23, 1.72259e-23, 4.711137e-24], [70.794578, 9.805e-23, 2.86627e-23, 1.71976e-23, 4.7023483e-24], [70.957777, 9.80308e-23, 2.86158e-23, 1.71695e-23, 4.6936501e-24], [71.121351, 9.80118e-23, 2.85693e-23, 1.71416e-23, 4.6850415e-24], [71.285303, 9.79931e-23, 2.85231e-23, 1.71139e-23, 4.6765219e-24], [71.449633, 9.79746e-23, 2.84773e-23, 1.70864e-23, 4.6680906e-24], [71.614341, 9.79564e-23, 2.84318e-23, 1.70591e-23, 4.6597466e-24], [71.779429, 9.79384e-23, 2.83867e-23, 1.7032e-23, 4.6514895e-24], [71.944898, 9.79206e-23, 2.83419e-23, 1.70051e-23, 4.6433183e-24], [72.110748, 9.79031e-23, 2.82975e-23, 1.69785e-23, 4.6352325e-24], [72.27698, 9.78858e-23, 2.82534e-23, 1.6952e-23, 4.6272312e-24], [72.443596, 9.78687e-23, 2.82096e-23, 1.69257e-23, 4.6193138e-24], [72.610596, 9.78519e-23, 2.81661e-23, 1.68997e-23, 4.6114796e-24], [72.77798, 9.78352e-23, 2.8123e-23, 1.68738e-23, 4.6037278e-24], [72.945751, 9.78188e-23, 2.80802e-23, 1.68481e-23, 4.5960577e-24], [73.113908, 9.78027e-23, 2.80378e-23, 1.68227e-23, 4.5884687e-24], [73.282453, 9.77867e-23, 2.79956e-23, 1.67974e-23, 4.58096e-24], [73.451387, 9.7771e-23, 2.79538e-23, 1.67723e-23, 4.5735309e-24], [73.62071, 9.77555e-23, 2.79123e-23, 1.67474e-23, 4.5661808e-24], [73.790423, 9.77402e-23, 2.78711e-23, 1.67227e-23, 4.5589089e-24], [73.960528, 9.77251e-23, 2.78302e-23, 1.66981e-23, 4.5517146e-24], [74.131024, 9.77102e-23, 2.77897e-23, 1.66738e-23, 4.5445972e-24], [74.301914, 9.76956e-23, 2.77494e-23, 1.66497e-23, 4.537556e-24], [74.473197, 9.76811e-23, 2.77095e-23, 1.66257e-23, 4.5305904e-24], [74.644876, 9.76669e-23, 2.76698e-23, 1.66019e-23, 4.5236996e-24], [74.81695, 9.76528e-23, 2.76305e-23, 1.65783e-23, 4.516883e-24], [74.989421, 9.7639e-23, 2.75915e-23, 1.65549e-23, 4.5101399e-24], [75.162289, 9.76254e-23, 2.75527e-23, 1.65316e-23, 4.5034697e-24], [75.335556, 9.7612e-23, 2.75143e-23, 1.65086e-23, 4.4968717e-24], [75.509223, 9.75987e-23, 2.74761e-23, 1.64857e-23, 4.4903453e-24], [75.68329, 9.75857e-23, 2.74383e-23, 1.6463e-23, 4.4838898e-24], [75.857758, 9.75729e-23, 2.74007e-23, 1.64404e-23, 4.4775045e-24], [76.032628, 9.75603e-23, 2.73635e-23, 1.64181e-23, 4.4711889e-24], [76.207901, 9.75478e-23, 2.73265e-23, 1.63959e-23, 4.4649422e-24], [76.383578, 9.75356e-23, 2.72898e-23, 1.63739e-23, 4.4587639e-24], [76.559661, 9.75235e-23, 2.72533e-23, 1.6352e-23, 4.4526533e-24], [76.736149, 9.75117e-23, 2.72172e-23, 1.63303e-23, 4.4466098e-24], [76.913044, 9.75e-23, 2.71813e-23, 1.63088e-23, 4.4406327e-24], [77.090347, 9.74885e-23, 2.71458e-23, 1.62875e-23, 4.4347215e-24], [77.268059, 9.74772e-23, 2.71105e-23, 1.62663e-23, 4.4288755e-24], [77.44618, 9.74661e-23, 2.70754e-23, 1.62452e-23, 4.4230941e-24], [77.624712, 9.74552e-23, 2.70406e-23, 1.62244e-23, 4.4173767e-24], [77.803655, 9.74444e-23, 2.70062e-23, 1.62037e-23, 4.4117226e-24], [77.983011, 9.74339e-23, 2.69719e-23, 1.61832e-23, 4.4061314e-24], [78.16278, 9.74235e-23, 2.6938e-23, 1.61628e-23, 4.4006024e-24], [78.342964, 9.74133e-23, 2.69043e-23, 1.61426e-23, 4.3951349e-24], [78.523563, 9.74033e-23, 2.68708e-23, 1.61225e-23, 4.3897285e-24], [78.704579, 9.73934e-23, 2.68376e-23, 1.61026e-23, 4.3843825e-24], [78.886012, 9.73837e-23, 2.68047e-23, 1.60828e-23, 4.3790963e-24], [79.067863, 9.73742e-23, 2.67721e-23, 1.60632e-23, 4.3738693e-24], [79.250133, 9.73649e-23, 2.67396e-23, 1.60438e-23, 4.368701e-24], [79.432823, 9.73558e-23, 2.67075e-23, 1.60245e-23, 4.3635909e-24], [79.615935, 9.73468e-23, 2.66756e-23, 1.60053e-23, 4.3585382e-24], [79.799469, 9.7338e-23, 2.66439e-23, 1.59863e-23, 4.3535426e-24], [79.983426, 9.73293e-23, 2.66125e-23, 1.59675e-23, 4.3486033e-24], [80.167806, 9.73209e-23, 2.65813e-23, 1.59488e-23, 4.3437199e-24], [80.352612, 9.73126e-23, 2.65504e-23, 1.59303e-23, 4.3388918e-24], [80.537844, 9.73044e-23, 2.65197e-23, 1.59118e-23, 4.3341185e-24], [80.723503, 9.72965e-23, 2.64893e-23, 1.58936e-23, 4.3293993e-24], [80.90959, 9.72887e-23, 2.64591e-23, 1.58755e-23, 4.3247338e-24], [81.096106, 9.7281e-23, 2.64291e-23, 1.58575e-23, 4.3201215e-24], [81.283052, 9.72735e-23, 2.63994e-23, 1.58396e-23, 4.3155617e-24], [81.470428, 9.72662e-23, 2.63699e-23, 1.5822e-23, 4.311054e-24], [81.658237, 9.7259e-23, 2.63407e-23, 1.58044e-23, 4.3065978e-24], [81.846479, 9.7252e-23, 2.63116e-23, 1.5787e-23, 4.3021926e-24], [82.035154, 9.72452e-23, 2.62828e-23, 1.57697e-23, 4.297838e-24], [82.224265, 9.72385e-23, 2.62543e-23, 1.57526e-23, 4.2935333e-24], [82.413812, 9.7232e-23, 2.62259e-23, 1.57356e-23, 4.289278e-24], [82.603795, 9.72256e-23, 2.61978e-23, 1.57187e-23, 4.2850718e-24], [82.794216, 9.72194e-23, 2.61699e-23, 1.57019e-23, 4.2809139e-24], [82.985077, 9.72134e-23, 2.61422e-23, 1.56853e-23, 4.2768041e-24], [83.176377, 9.72074e-23, 2.61148e-23, 1.56689e-23, 4.2727417e-24], [83.368118, 9.72017e-23, 2.60875e-23, 1.56525e-23, 4.2687263e-24], [83.560302, 9.71961e-23, 2.60605e-23, 1.56363e-23, 4.2647573e-24], [83.752928, 9.71906e-23, 2.60337e-23, 1.56202e-23, 4.2608344e-24], [83.945999, 9.71853e-23, 2.60071e-23, 1.56043e-23, 4.256957e-24], [84.139514, 9.71802e-23, 2.59807e-23, 1.55884e-23, 4.2531246e-24], [84.333476, 9.71752e-23, 2.59546e-23, 1.55727e-23, 4.2493368e-24], [84.527885, 9.71703e-23, 2.59286e-23, 1.55572e-23, 4.245593e-24], [84.722741, 9.71656e-23, 2.59028e-23, 1.55417e-23, 4.241893e-24], [84.918048, 9.71611e-23, 2.58773e-23, 1.55264e-23, 4.2382361e-24], [85.113804, 9.71567e-23, 2.5852e-23, 1.55112e-23, 4.2346219e-24], [85.310011, 9.71524e-23, 2.58268e-23, 1.54961e-23, 4.23105e-24], [85.506671, 9.71483e-23, 2.58019e-23, 1.54811e-23, 4.2275199e-24], [85.703785, 9.71443e-23, 2.57771e-23, 1.54663e-23, 4.2240312e-24], [85.901352, 9.71405e-23, 2.57526e-23, 1.54516e-23, 4.2205834e-24], [86.099375, 9.71368e-23, 2.57283e-23, 1.5437e-23, 4.2171761e-24], [86.297855, 9.71332e-23, 2.57041e-23, 1.54225e-23, 4.2138089e-24], [86.496792, 9.71298e-23, 2.56802e-23, 1.54081e-23, 4.2104813e-24], [86.696188, 9.71266e-23, 2.56565e-23, 1.53939e-23, 4.2071929e-24], [86.896043, 9.71234e-23, 2.56329e-23, 1.53797e-23, 4.2039432e-24], [87.096359, 9.71204e-23, 2.56095e-23, 1.53657e-23, 4.200732e-24], [87.297137, 9.71176e-23, 2.55864e-23, 1.53518e-23, 4.1975587e-24], [87.498378, 9.71149e-23, 2.55634e-23, 1.5338e-23, 4.1944229e-24], [87.700082, 9.71123e-23, 2.55406e-23, 1.53243e-23, 4.1913243e-24], [87.902252, 9.71099e-23, 2.5518e-23, 1.53108e-23, 4.1882624e-24], [88.104887, 9.71076e-23, 2.54955e-23, 1.52973e-23, 4.1852368e-24], [88.30799, 9.71054e-23, 2.54733e-23, 1.5284e-23, 4.1822472e-24], [88.511561, 9.71034e-23, 2.54512e-23, 1.52707e-23, 4.1792931e-24], [88.715601, 9.71015e-23, 2.54294e-23, 1.52576e-23, 4.1763741e-24], [88.920112, 9.70998e-23, 2.54077e-23, 1.52446e-23, 4.17349e-24], [89.125094, 9.70982e-23, 2.53862e-23, 1.52317e-23, 4.1706402e-24], [89.330548, 9.70967e-23, 2.53648e-23, 1.52189e-23, 4.1678244e-24], [89.536477, 9.70953e-23, 2.53437e-23, 1.52062e-23, 4.1650423e-24], [89.742879, 9.70941e-23, 2.53227e-23, 1.51936e-23, 4.1622934e-24], [89.949758, 9.7093e-23, 2.53019e-23, 1.51811e-23, 4.1595774e-24], [90.157114, 9.70921e-23, 2.52812e-23, 1.51687e-23, 4.156894e-24], [90.364947, 9.70912e-23, 2.52608e-23, 1.51565e-23, 4.1542428e-24], [90.57326, 9.70906e-23, 2.52405e-23, 1.51443e-23, 4.1516234e-24], [90.782053, 9.709e-23, 2.52203e-23, 1.51322e-23, 4.1490354e-24], [90.991327, 9.70896e-23, 2.52004e-23, 1.51202e-23, 4.1464787e-24], [91.201084, 9.70893e-23, 2.51806e-23, 1.51084e-23, 4.1439527e-24], [91.411324, 9.70891e-23, 2.5161e-23, 1.50966e-23, 4.1414571e-24], [91.622049, 9.70891e-23, 2.51415e-23, 1.50849e-23, 4.1389917e-24], [91.83326, 9.70891e-23, 2.51222e-23, 1.50733e-23, 4.136556e-24], [92.044957, 9.70894e-23, 2.51031e-23, 1.50619e-23, 4.1341498e-24], [92.257143, 9.70897e-23, 2.50841e-23, 1.50505e-23, 4.1317728e-24], [92.469817, 9.70902e-23, 2.50653e-23, 1.50392e-23, 4.1294245e-24], [92.682982, 9.70908e-23, 2.50467e-23, 1.5028e-23, 4.1271047e-24], [92.896639, 9.70915e-23, 2.50282e-23, 1.50169e-23, 4.1248131e-24], [93.110788, 9.70924e-23, 2.50099e-23, 1.50059e-23, 4.1225494e-24], [93.32543, 9.70934e-23, 2.49917e-23, 1.4995e-23, 4.1203132e-24], [93.540567, 9.70945e-23, 2.49737e-23, 1.49842e-23, 4.1181043e-24], [93.756201, 9.70957e-23, 2.49558e-23, 1.49735e-23, 4.1159224e-24], [93.972331, 9.70971e-23, 2.49381e-23, 1.49629e-23, 4.1137671e-24], [94.18896, 9.70985e-23, 2.49206e-23, 1.49524e-23, 4.1116382e-24], [94.406088, 9.71002e-23, 2.49032e-23, 1.49419e-23, 4.1095353e-24], [94.623716, 9.71019e-23, 2.4886e-23, 1.49316e-23, 4.1074583e-24], [94.841846, 9.71037e-23, 2.48689e-23, 1.49213e-23, 4.1054068e-24], [95.060479, 9.71057e-23, 2.48519e-23, 1.49112e-23, 4.1033805e-24], [95.279616, 9.71078e-23, 2.48351e-23, 1.49011e-23, 4.1013791e-24], [95.499259, 9.71101e-23, 2.48185e-23, 1.48911e-23, 4.0994025e-24], [95.719407, 9.71124e-23, 2.4802e-23, 1.48812e-23, 4.0974503e-24], [95.940063, 9.71149e-23, 2.47856e-23, 1.48714e-23, 4.0955222e-24], [96.161228, 9.71175e-23, 2.47694e-23, 1.48616e-23, 4.0936181e-24], [96.382902, 9.71202e-23, 2.47533e-23, 1.4852e-23, 4.0917376e-24], [96.605088, 9.71231e-23, 2.47374e-23, 1.48425e-23, 4.0898804e-24], [96.827786, 9.71261e-23, 2.47216e-23, 1.4833e-23, 4.0880464e-24], [97.050997, 9.71292e-23, 2.4706e-23, 1.48236e-23, 4.0862354e-24], [97.274722, 9.71324e-23, 2.46905e-23, 1.48143e-23, 4.0844469e-24], [97.498964, 9.71357e-23, 2.46752e-23, 1.48051e-23, 4.0826809e-24], [97.723722, 9.71392e-23, 2.46599e-23, 1.4796e-23, 4.0809371e-24], [97.948999, 9.71427e-23, 2.46449e-23, 1.47869e-23, 4.0792152e-24], [98.174794, 9.71465e-23, 2.46299e-23, 1.4778e-23, 4.077515e-24], [98.401111, 9.71503e-23, 2.46151e-23, 1.47691e-23, 4.0758363e-24], [98.627949, 9.71542e-23, 2.46004e-23, 1.47603e-23, 4.0741788e-24], [98.855309, 9.71583e-23, 2.45859e-23, 1.47515e-23, 4.0725425e-24], [99.083194, 9.71625e-23, 2.45715e-23, 1.47429e-23, 4.0709269e-24], [99.311605, 9.71668e-23, 2.45573e-23, 1.47344e-23, 4.069332e-24], [99.540542, 9.71712e-23, 2.45431e-23, 1.47259e-23, 4.0677575e-24], [99.770006, 9.71758e-23, 2.45291e-23, 1.47175e-23, 4.0662032e-24], [100.0, 9.71804e-23, 2.45153e-23, 1.47092e-23, 4.0646689e-24], [100.23052, 9.71852e-23, 2.45015e-23, 1.47009e-23, 4.0631545e-24], [100.46158, 9.71902e-23, 2.44879e-23, 1.46927e-23, 4.0616596e-24], [100.69317, 9.71952e-23, 2.44744e-23, 1.46847e-23, 4.0601842e-24], [100.92529, 9.72003e-23, 2.44611e-23, 1.46767e-23, 4.058728e-24], [101.15795, 9.72056e-23, 2.44479e-23, 1.46687e-23, 4.0572909e-24], [101.39114, 9.7211e-23, 2.44348e-23, 1.46609e-23, 4.0558726e-24], [101.62487, 9.72165e-23, 2.44218e-23, 1.46531e-23, 4.0544731e-24], [101.85914, 9.72222e-23, 2.4409e-23, 1.46454e-23, 4.053092e-24], [102.09395, 9.72279e-23, 2.43963e-23, 1.46378e-23, 4.0517294e-24], [102.3293, 9.72338e-23, 2.43837e-23, 1.46302e-23, 4.0503849e-24], [102.56519, 9.72398e-23, 2.43712e-23, 1.46227e-23, 4.0490584e-24], [102.80163, 9.72459e-23, 2.43589e-23, 1.46153e-23, 4.0477498e-24], [103.03861, 9.72522e-23, 2.43466e-23, 1.4608e-23, 4.0464589e-24], [103.27614, 9.72585e-23, 2.43346e-23, 1.46007e-23, 4.0451856e-24], [103.51422, 9.7265e-23, 2.43226e-23, 1.45935e-23, 4.0439297e-24], [103.75284, 9.72716e-23, 2.43107e-23, 1.45864e-23, 4.042691e-24], [103.99202, 9.72784e-23, 2.4299e-23, 1.45794e-23, 4.0414695e-24], [104.23174, 9.72852e-23, 2.42874e-23, 1.45724e-23, 4.0402649e-24], [104.47202, 9.72922e-23, 2.42759e-23, 1.45655e-23, 4.0390772e-24], [104.71285, 9.72992e-23, 2.42645e-23, 1.45587e-23, 4.0379061e-24], [104.95424, 9.73065e-23, 2.42533e-23, 1.4552e-23, 4.0367517e-24], [105.19619, 9.73138e-23, 2.42421e-23, 1.45453e-23, 4.0356137e-24], [105.43869, 9.73212e-23, 2.42311e-23, 1.45387e-23, 4.034492e-24], [105.68175, 9.73288e-23, 2.42202e-23, 1.45321e-23, 4.0333866e-24], [105.92537, 9.73365e-23, 2.42094e-23, 1.45257e-23, 4.0322973e-24], [106.16956, 9.73443e-23, 2.41988e-23, 1.45193e-23, 4.0312239e-24], [106.4143, 9.73522e-23, 2.41882e-23, 1.45129e-23, 4.0301664e-24], [106.65961, 9.73603e-23, 2.41778e-23, 1.45067e-23, 4.0291248e-24], [106.90549, 9.73685e-23, 2.41675e-23, 1.45005e-23, 4.0280987e-24], [107.15193, 9.73768e-23, 2.41572e-23, 1.44943e-23, 4.0270883e-24], [107.39894, 9.73852e-23, 2.41472e-23, 1.44883e-23, 4.0260933e-24], [107.64652, 9.73938e-23, 2.41372e-23, 1.44823e-23, 4.0251138e-24], [107.89467, 9.74024e-23, 2.41273e-23, 1.44764e-23, 4.0241495e-24], [108.1434, 9.74112e-23, 2.41176e-23, 1.44705e-23, 4.0232005e-24], [108.39269, 9.74201e-23, 2.41079e-23, 1.44647e-23, 4.0222666e-24], [108.64256, 9.74292e-23, 2.40984e-23, 1.4459e-23, 4.0213478e-24], [108.89301, 9.74383e-23, 2.4089e-23, 1.44534e-23, 4.020444e-24], [109.14403, 9.74476e-23, 2.40797e-23, 1.44478e-23, 4.0195552e-24], [109.39564, 9.7457e-23, 2.40705e-23, 1.44423e-23, 4.0186811e-24], [109.64782, 9.74666e-23, 2.40614e-23, 1.44368e-23, 4.017822e-24], [109.90058, 9.74762e-23, 2.40524e-23, 1.44314e-23, 4.0169775e-24], [110.15393, 9.7486e-23, 2.40435e-23, 1.44261e-23, 4.0161478e-24], [110.40786, 9.74959e-23, 2.40348e-23, 1.44209e-23, 4.0153326e-24], [110.66238, 9.7506e-23, 2.40261e-23, 1.44157e-23, 4.0145322e-24], [110.91748, 9.75161e-23, 2.40176e-23, 1.44105e-23, 4.0137462e-24], [111.17317, 9.75264e-23, 2.40092e-23, 1.44055e-23, 4.0129748e-24], [111.42945, 9.75368e-23, 2.40008e-23, 1.44005e-23, 4.0122179e-24], [111.68632, 9.75473e-23, 2.39926e-23, 1.43956e-23, 4.0114754e-24], [111.94379, 9.7558e-23, 2.39845e-23, 1.43907e-23, 4.0107474e-24], [112.20185, 9.75688e-23, 2.39765e-23, 1.43859e-23, 4.0100338e-24], [112.4605, 9.75797e-23, 2.39686e-23, 1.43812e-23, 4.0093346e-24], [112.71975, 9.75908e-23, 2.39608e-23, 1.43765e-23, 4.0086498e-24], [112.97959, 9.76019e-23, 2.39532e-23, 1.43719e-23, 4.0079793e-24], [113.24004, 9.76133e-23, 2.39456e-23, 1.43674e-23, 4.0073232e-24], [113.50108, 9.76247e-23, 2.39381e-23, 1.43629e-23, 4.0066815e-24], [113.76273, 9.76363e-23, 2.39308e-23, 1.43585e-23, 4.0060542e-24], [114.02498, 9.76479e-23, 2.39235e-23, 1.43541e-23, 4.0054413e-24], [114.28783, 9.76598e-23, 2.39164e-23, 1.43498e-23, 4.0048427e-24], [114.55129, 9.76717e-23, 2.39094e-23, 1.43456e-23, 4.0042586e-24], [114.81536, 9.76838e-23, 2.39024e-23, 1.43415e-23, 4.0036889e-24], [115.08004, 9.7696e-23, 2.38956e-23, 1.43374e-23, 4.0031337e-24], [115.34533, 9.77084e-23, 2.38889e-23, 1.43333e-23, 4.0025929e-24], [115.61122, 9.77209e-23, 2.38823e-23, 1.43294e-23, 4.0020667e-24], [115.87774, 9.77335e-23, 2.38758e-23, 1.43255e-23, 4.0015551e-24], [116.14486, 9.77462e-23, 2.38694e-23, 1.43216e-23, 4.0010581e-24], [116.4126, 9.77591e-23, 2.38631e-23, 1.43178e-23, 4.0005758e-24], [116.68096, 9.77721e-23, 2.38569e-23, 1.43141e-23, 4.0001081e-24], [116.94994, 9.77853e-23, 2.38508e-23, 1.43105e-23, 3.9996553e-24], [117.21954, 9.77986e-23, 2.38448e-23, 1.43069e-23, 3.9992173e-24], [117.48976, 9.7812e-23, 2.3839e-23, 1.43034e-23, 3.9987943e-24], [117.7606, 9.78256e-23, 2.38332e-23, 1.42999e-23, 3.9983863e-24], [118.03206, 9.78393e-23, 2.38276e-23, 1.42965e-23, 3.9979933e-24], [118.30416, 9.78531e-23, 2.3822e-23, 1.42932e-23, 3.9976155e-24], [118.57687, 9.78671e-23, 2.38166e-23, 1.42899e-23, 3.997253e-24], [118.85022, 9.78812e-23, 2.38112e-23, 1.42867e-23, 3.9969059e-24], [119.1242, 9.78955e-23, 2.3806e-23, 1.42836e-23, 3.9965743e-24], [119.39881, 9.79099e-23, 2.38009e-23, 1.42805e-23, 3.9962582e-24], [119.67405, 9.79245e-23, 2.37959e-23, 1.42775e-23, 3.9959578e-24], [119.94993, 9.79391e-23, 2.3791e-23, 1.42746e-23, 3.9956733e-24], [120.22644, 9.7954e-23, 2.37862e-23, 1.42717e-23, 3.9954048e-24], [120.50359, 9.7969e-23, 2.37815e-23, 1.42689e-23, 3.9951523e-24], [120.78138, 9.79841e-23, 2.37769e-23, 1.42662e-23, 3.9949161e-24], [121.05981, 9.79994e-23, 2.37725e-23, 1.42635e-23, 3.9946963e-24], [121.33889, 9.80148e-23, 2.37681e-23, 1.42609e-23, 3.9944931e-24], [121.6186, 9.80303e-23, 2.37639e-23, 1.42583e-23, 3.9943066e-24], [121.89896, 9.80461e-23, 2.37597e-23, 1.42558e-23, 3.994137e-24], [122.17997, 9.80619e-23, 2.37557e-23, 1.42534e-23, 3.9939845e-24], [122.46162, 9.8078e-23, 2.37518e-23, 1.42511e-23, 3.9938492e-24], [122.74392, 9.80941e-23, 2.3748e-23, 1.42488e-23, 3.9937315e-24], [123.02688, 9.81104e-23, 2.37443e-23, 1.42466e-23, 3.9936314e-24], [123.31048, 9.81269e-23, 2.37407e-23, 1.42444e-23, 3.9935492e-24], [123.59474, 9.81436e-23, 2.37373e-23, 1.42424e-23, 3.9934852e-24], [123.87966, 9.81604e-23, 2.37339e-23, 1.42404e-23, 3.9934395e-24], [124.16523, 9.81773e-23, 2.37307e-23, 1.42384e-23, 3.9934124e-24], [124.45146, 9.81944e-23, 2.37276e-23, 1.42366e-23, 3.9934041e-24], [124.73835, 9.82117e-23, 2.37246e-23, 1.42348e-23, 3.9934149e-24], [125.0259, 9.82291e-23, 2.37217e-23, 1.4233e-23, 3.9934452e-24], [125.31412, 9.82467e-23, 2.37189e-23, 1.42314e-23, 3.9934951e-24], [125.603, 9.82645e-23, 2.37163e-23, 1.42298e-23, 3.993565e-24], [125.89254, 9.82824e-23, 2.37138e-23, 1.42283e-23, 3.9936551e-24], [126.18275, 9.83005e-23, 2.37114e-23, 1.42268e-23, 3.9937658e-24], [126.47363, 9.83187e-23, 2.37091e-23, 1.42255e-23, 3.9938975e-24], [126.76519, 9.83372e-23, 2.37069e-23, 1.42242e-23, 3.9940504e-24], [127.05741, 9.83558e-23, 2.37049e-23, 1.42229e-23, 3.9942248e-24], [127.35031, 9.83745e-23, 2.3703e-23, 1.42218e-23, 3.9944213e-24], [127.64388, 9.83935e-23, 2.37012e-23, 1.42207e-23, 3.9946402e-24], [127.93813, 9.84126e-23, 2.36996e-23, 1.42197e-23, 3.9948817e-24], [128.23306, 9.84319e-23, 2.36981e-23, 1.42188e-23, 3.9951465e-24], [128.52867, 9.84514e-23, 2.36967e-23, 1.4218e-23, 3.9954348e-24], [128.82496, 9.84711e-23, 2.36954e-23, 1.42172e-23, 3.9957472e-24], [129.12193, 9.84909e-23, 2.36943e-23, 1.42166e-23, 3.996084e-24], [129.41958, 9.8511e-23, 2.36933e-23, 1.4216e-23, 3.9964458e-24], [129.71793, 9.85312e-23, 2.36924e-23, 1.42155e-23, 3.9968331e-24], [130.01696, 9.85516e-23, 2.36917e-23, 1.4215e-23, 3.9972463e-24], [130.31668, 9.85722e-23, 2.36912e-23, 1.42147e-23, 3.9976861e-24], [130.61709, 9.8593e-23, 2.36907e-23, 1.42144e-23, 3.9981528e-24], [130.91819, 9.8614e-23, 2.36905e-23, 1.42143e-23, 3.9986472e-24], [131.21999, 9.86352e-23, 2.36903e-23, 1.42142e-23, 3.9991698e-24], [131.52248, 9.86566e-23, 2.36903e-23, 1.42142e-23, 3.9997212e-24], [131.82567, 9.86782e-23, 2.36905e-23, 1.42143e-23, 4.000302e-24], [132.12956, 9.87001e-23, 2.36908e-23, 1.42145e-23, 4.000913e-24], [132.43415, 9.87221e-23, 2.36913e-23, 1.42148e-23, 4.0015547e-24], [132.73945, 9.87443e-23, 2.36919e-23, 1.42151e-23, 4.002228e-24], [133.04544, 9.87668e-23, 2.36927e-23, 1.42156e-23, 4.0029336e-24], [133.35214, 9.87895e-23, 2.36937e-23, 1.42162e-23, 4.0036721e-24], [133.65955, 9.88124e-23, 2.36948e-23, 1.42169e-23, 4.0044446e-24], [133.96767, 9.88355e-23, 2.36961e-23, 1.42176e-23, 4.0052516e-24], [134.2765, 9.88589e-23, 2.36975e-23, 1.42185e-23, 4.0060942e-24], [134.58604, 9.88824e-23, 2.36992e-23, 1.42195e-23, 4.0069732e-24], [134.89629, 9.89063e-23, 2.3701e-23, 1.42206e-23, 4.0078896e-24], [135.20726, 9.89303e-23, 2.3703e-23, 1.42218e-23, 4.0088443e-24], [135.51894, 9.89547e-23, 2.37052e-23, 1.42231e-23, 4.0098383e-24], [135.83134, 9.89792e-23, 2.37076e-23, 1.42245e-23, 4.0108728e-24], [136.14447, 9.90041e-23, 2.37102e-23, 1.42261e-23, 4.0119487e-24], [136.45831, 9.90291e-23, 2.37129e-23, 1.42278e-23, 4.0130673e-24], [136.77288, 9.90545e-23, 2.37159e-23, 1.42296e-23, 4.0142297e-24], [137.08818, 9.90801e-23, 2.37191e-23, 1.42315e-23, 4.0154372e-24], [137.4042, 9.9106e-23, 2.37225e-23, 1.42335e-23, 4.0166911e-24], [137.72095, 9.91322e-23, 2.37262e-23, 1.42357e-23, 4.0179927e-24], [138.03843, 9.91586e-23, 2.373e-23, 1.4238e-23, 4.0193435e-24], [138.35664, 9.91854e-23, 2.37341e-23, 1.42405e-23, 4.0207449e-24], [138.67558, 9.92124e-23, 2.37384e-23, 1.4243e-23, 4.0221984e-24], [138.99526, 9.92398e-23, 2.3743e-23, 1.42458e-23, 4.0237057e-24], [139.31568, 9.92674e-23, 2.37478e-23, 1.42487e-23, 4.0252684e-24], [139.63684, 9.92954e-23, 2.37529e-23, 1.42517e-23, 4.0268884e-24], [139.95873, 9.93237e-23, 2.37582e-23, 1.42549e-23, 4.0285674e-24], [140.28137, 9.93524e-23, 2.37638e-23, 1.42583e-23, 4.0303074e-24], [140.60475, 9.93813e-23, 2.37697e-23, 1.42618e-23, 4.0321103e-24], [140.92888, 9.94107e-23, 2.37759e-23, 1.42655e-23, 4.0339784e-24], [141.25375, 9.94404e-23, 2.37823e-23, 1.42694e-23, 4.0359139e-24], [141.57938, 9.94704e-23, 2.37891e-23, 1.42735e-23, 4.037919e-24], [141.90575, 9.95009e-23, 2.37962e-23, 1.42777e-23, 4.0399961e-24], [142.23288, 9.95317e-23, 2.38036e-23, 1.42822e-23, 4.042148e-24], [142.56076, 9.95629e-23, 2.38114e-23, 1.42868e-23, 4.0443772e-24], [142.8894, 9.95946e-23, 2.38195e-23, 1.42917e-23, 4.0466866e-24], [143.21879, 9.96267e-23, 2.3828e-23, 1.42968e-23, 4.0490792e-24], [143.54894, 9.96592e-23, 2.38368e-23, 1.43021e-23, 4.0515581e-24], [143.87986, 9.96921e-23, 2.3846e-23, 1.43076e-23, 4.0541266e-24], [144.21154, 9.97256e-23, 2.38557e-23, 1.43134e-23, 4.0567882e-24], [144.54398, 9.97595e-23, 2.38657e-23, 1.43194e-23, 4.0595466e-24], [144.87719, 9.97939e-23, 2.38762e-23, 1.43257e-23, 4.0624056e-24], [145.21116, 9.98288e-23, 2.38872e-23, 1.43323e-23, 4.0653693e-24], [145.54591, 9.98643e-23, 2.38986e-23, 1.43391e-23, 4.0684421e-24], [145.88143, 9.99003e-23, 2.39105e-23, 1.43463e-23, 4.0716284e-24], [146.21772, 9.99368e-23, 2.39229e-23, 1.43537e-23, 4.0749332e-24], [146.55478, 9.9974e-23, 2.39358e-23, 1.43615e-23, 4.0783616e-24], [146.89263, 1.00012e-22, 2.39493e-23, 1.43696e-23, 4.081919e-24], [147.23125, 1.0005e-22, 2.39634e-23, 1.4378e-23, 4.085611e-24], [147.57065, 1.00089e-22, 2.39781e-23, 1.43868e-23, 4.0894439e-24], [147.91084, 1.00129e-22, 2.39934e-23, 1.4396e-23, 4.0934241e-24], [148.25181, 1.0017e-22, 2.40093e-23, 1.44056e-23, 4.0975585e-24], [148.59356, 1.00211e-22, 2.4026e-23, 1.44156e-23, 4.1018544e-24], [148.93611, 1.00253e-22, 2.40433e-23, 1.4426e-23, 4.1063197e-24], [149.27944, 1.00296e-22, 2.40615e-23, 1.44369e-23, 4.1109626e-24], [149.62357, 1.00339e-22, 2.40804e-23, 1.44482e-23, 4.1157919e-24], [149.96848, 1.00384e-22, 2.41002e-23, 1.44601e-23, 4.1208173e-24], [150.3142, 1.0043e-22, 2.41208e-23, 1.44725e-23, 4.1260488e-24], [150.66071, 1.00476e-22, 2.41424e-23, 1.44855e-23, 4.1314971e-24], [151.00802, 1.00524e-22, 2.4165e-23, 1.4499e-23, 4.137174e-24], [151.35612, 1.00572e-22, 2.41886e-23, 1.45131e-23, 4.1430918e-24], [151.70504, 1.00622e-22, 2.42133e-23, 1.4528e-23, 4.1492639e-24], [152.05475, 1.00673e-22, 2.42391e-23, 1.45435e-23, 4.1557046e-24], [152.40528, 1.00726e-22, 2.42661e-23, 1.45597e-23, 4.1624292e-24], [152.75661, 1.00779e-22, 2.42945e-23, 1.45767e-23, 4.1694545e-24], [153.10875, 1.00834e-22, 2.43242e-23, 1.45945e-23, 4.1767981e-24], [153.4617, 1.00891e-22, 2.43554e-23, 1.46132e-23, 4.1844795e-24], [153.81546, 1.00949e-22, 2.43881e-23, 1.46329e-23, 4.1925194e-24], [154.17005, 1.01009e-22, 2.44224e-23, 1.46535e-23, 4.2009405e-24], [154.52544, 1.01071e-22, 2.44585e-23, 1.46751e-23, 4.209767e-24], [154.88166, 1.01134e-22, 2.44965e-23, 1.46979e-23, 4.2190256e-24], [155.2387, 1.012e-22, 2.45365e-23, 1.47219e-23, 4.228745e-24], [155.59656, 1.01268e-22, 2.45786e-23, 1.47471e-23, 4.2389566e-24], [155.95525, 1.01338e-22, 2.4623e-23, 1.47738e-23, 4.2496943e-24], [156.31476, 1.01411e-22, 2.46698e-23, 1.48019e-23, 4.2609955e-24], [156.67511, 1.01487e-22, 2.47192e-23, 1.48315e-23, 4.2729009e-24], [157.03628, 1.01565e-22, 2.47715e-23, 1.48629e-23, 4.285455e-24], [157.39829, 1.01646e-22, 2.48268e-23, 1.48961e-23, 4.2987066e-24], [157.76113, 1.01731e-22, 2.48854e-23, 1.49313e-23, 4.3127095e-24], [158.1248, 1.0182e-22, 2.49476e-23, 1.49685e-23, 4.3275226e-24], [158.48932, 1.01912e-22, 2.50136e-23, 1.50081e-23, 4.3432112e-24], [158.85467, 1.02009e-22, 2.50837e-23, 1.50502e-23, 4.3598472e-24], [159.22087, 1.02111e-22, 2.51584e-23, 1.5095e-23, 4.3775102e-24], [159.58791, 1.02217e-22, 2.52379e-23, 1.51428e-23, 4.3962887e-24], [159.9558, 1.02329e-22, 2.53229e-23, 1.51937e-23, 4.4162809e-24], [160.32454, 1.02447e-22, 2.54137e-23, 1.52482e-23, 4.4375965e-24], [160.69413, 1.02572e-22, 2.55109e-23, 1.53065e-23, 4.4603578e-24], [161.06456, 1.02704e-22, 2.56151e-23, 1.53691e-23, 4.4847019e-24], [161.43586, 1.02845e-22, 2.5727e-23, 1.54362e-23, 4.5107829e-24], [161.808, 1.02994e-22, 2.58475e-23, 1.55085e-23, 4.5387741e-24], [162.18101, 1.03153e-22, 2.59774e-23, 1.55865e-23, 4.5688714e-24], [162.55488, 1.03323e-22, 2.61178e-23, 1.56707e-23, 4.6012965e-24], [162.9296, 1.03506e-22, 2.62697e-23, 1.57618e-23, 4.6363016e-24], [163.30519, 1.03703e-22, 2.64346e-23, 1.58607e-23, 4.674174e-24], [163.68165, 1.03915e-22, 2.66139e-23, 1.59683e-23, 4.7152428e-24], [164.05898, 1.04144e-22, 2.68094e-23, 1.60856e-23, 4.7598856e-24], [164.43717, 1.04394e-22, 2.70231e-23, 1.62138e-23, 4.808538e-24], [164.81624, 1.04666e-22, 2.72574e-23, 1.63544e-23, 4.8617043e-24], [165.19618, 1.04965e-22, 2.7515e-23, 1.6509e-23, 4.9199711e-24], [165.577, 1.05293e-22, 2.77992e-23, 1.66795e-23, 4.9840238e-24], [165.95869, 1.05656e-22, 2.81137e-23, 1.68682e-23, 5.0546674e-24], [166.34127, 1.06058e-22, 2.84631e-23, 1.70779e-23, 5.1328528e-24], [166.72472, 1.06508e-22, 2.88528e-23, 1.73117e-23, 5.2197092e-24], [167.10906, 1.07012e-22, 2.92891e-23, 1.75735e-23, 5.3165866e-24], [167.49429, 1.07581e-22, 2.97799e-23, 1.78679e-23, 5.42511e-24], [167.8804, 1.08229e-22, 3.03347e-23, 1.82008e-23, 5.5472499e-24], [168.26741, 1.08969e-22, 3.0965e-23, 1.8579e-23, 5.6854158e-24], [168.6553, 1.09823e-22, 3.16854e-23, 1.90113e-23, 5.8425803e-24], [169.04409, 1.10816e-22, 3.2514e-23, 1.95084e-23, 6.0224483e-24], [169.43378, 1.11982e-22, 3.34734e-23, 2.00841e-23, 6.2296888e-24], [169.82437, 1.13363e-22, 3.45931e-23, 2.07559e-23, 6.4702609e-24], [170.21585, 1.15019e-22, 3.5911e-23, 2.15466e-23, 6.7518796e-24], [170.60824, 1.17028e-22, 3.74773e-23, 2.24864e-23, 7.0846975e-24], [171.00153, 1.19501e-22, 3.93596e-23, 2.36158e-23, 7.4823303e-24], [171.39573, 1.22595e-22, 4.16507e-23, 2.49904e-23, 7.9634445e-24], [171.79084, 1.26535e-22, 4.44813e-23, 2.66888e-23, 8.5543065e-24], [172.18686, 1.31664e-22, 4.80418e-23, 2.88251e-23, 9.2930485e-24], [172.58379, 1.38511e-22, 5.26194e-23, 3.15716e-23, 1.0237173e-23], [172.98164, 1.47934e-22, 5.86684e-23, 3.52011e-23, 1.1477585e-23], [173.3804, 1.61412e-22, 6.69522e-23, 4.01713e-23, 1.3166903e-23], [173.78008, 1.81671e-22, 7.88578e-23, 4.73147e-23, 1.5582489e-23], [174.18069, 2.14304e-22, 9.71974e-23, 5.83184e-23, 1.9286601e-23], [174.58222, 2.72703e-22, 1.28669e-22, 7.72017e-23, 2.5618763e-23], [174.98467, 3.98609e-22, 1.94126e-22, 1.16475e-22, 3.8749049e-23], [175.38805, 8.20702e-22, 4.07859e-22, 2.44715e-22, 8.153558e-23], [175.79236, 6.79937e-21, 3.39938e-21, 2.03963e-21, 6.7987215e-22], [176.1976, 6.59817e-22, 3.26798e-22, 1.96079e-22, 6.5314623e-23], [176.60378, 3.56144e-22, 1.72236e-22, 1.03342e-22, 3.4361914e-23], [177.0109, 2.52081e-22, 1.17644e-22, 7.05864e-23, 2.3403975e-23], [177.41895, 2.01458e-22, 8.99915e-23, 5.39949e-23, 1.783506e-23], [177.82794, 1.72554e-22, 7.34446e-23, 4.40668e-23, 1.4488784e-23], [178.23788, 1.54433e-22, 6.25383e-23, 3.7523e-23, 1.2272395e-23], [178.64876, 1.4234e-22, 5.48817e-23, 3.2929e-23, 1.0707957e-23], [179.06059, 1.33892e-22, 4.92629e-23, 2.95577e-23, 9.5532032e-24], [179.47336, 1.27778e-22, 4.50017e-23, 2.7001e-23, 8.6721767e-24], [179.88709, 1.23228e-22, 4.16869e-23, 2.50121e-23, 7.9826708e-24], [180.30177, 1.1976e-22, 3.90556e-23, 2.34334e-23, 7.432055e-24], [180.71741, 1.17064e-22, 3.69319e-23, 2.21592e-23, 6.9850799e-24], [181.13401, 1.14934e-22, 3.51939e-23, 2.11164e-23, 6.6172619e-24], [181.55157, 1.13226e-22, 3.37546e-23, 2.02527e-23, 6.3110782e-24], [181.97009, 1.11839e-22, 3.25502e-23, 1.95301e-23, 6.053671e-24], [182.38957, 1.10701e-22, 3.15333e-23, 1.892e-23, 5.8354042e-24], [182.81002, 1.09758e-22, 3.06678e-23, 1.84007e-23, 5.6489249e-24], [183.23144, 1.08969e-22, 2.99259e-23, 1.79555e-23, 5.4885349e-24], [183.65383, 1.08305e-22, 2.92857e-23, 1.75714e-23, 5.3497575e-24], [184.0772, 1.07743e-22, 2.873e-23, 1.7238e-23, 5.2290341e-24], [184.50154, 1.07264e-22, 2.82451e-23, 1.6947e-23, 5.1235056e-24], [184.92686, 1.06854e-22, 2.78198e-23, 1.66919e-23, 5.0308533e-24], [185.35316, 1.06501e-22, 2.74452e-23, 1.64671e-23, 4.9491804e-24], [185.78045, 1.06196e-22, 2.71138e-23, 1.62683e-23, 4.8769238e-24], [186.20871, 1.05933e-22, 2.68194e-23, 1.60916e-23, 4.8127858e-24], [186.63797, 1.05704e-22, 2.65571e-23, 1.59342e-23, 4.7556823e-24], [187.06821, 1.05506e-22, 2.63225e-23, 1.57935e-23, 4.7047023e-24], [187.49945, 1.05333e-22, 2.6112e-23, 1.56672e-23, 4.6590759e-24], [187.93168, 1.05183e-22, 2.59226e-23, 1.55536e-23, 4.6181492e-24], [188.36491, 1.05052e-22, 2.57518e-23, 1.54511e-23, 4.581364e-24], [188.79913, 1.04938e-22, 2.55973e-23, 1.53584e-23, 4.548242e-24], [189.23436, 1.04839e-22, 2.54573e-23, 1.52744e-23, 4.5183713e-24], [189.67059, 1.04753e-22, 2.53301e-23, 1.51981e-23, 4.4913957e-24], [190.10783, 1.04679e-22, 2.52143e-23, 1.51286e-23, 4.4670065e-24], [190.54607, 1.04616e-22, 2.51087e-23, 1.50652e-23, 4.4449347e-24], [190.98533, 1.04562e-22, 2.50122e-23, 1.50073e-23, 4.4249453e-24], [191.42559, 1.04516e-22, 2.49239e-23, 1.49543e-23, 4.4068325e-24], [191.86687, 1.04477e-22, 2.48429e-23, 1.49058e-23, 4.3904152e-24], [192.30917, 1.04445e-22, 2.47686e-23, 1.48612e-23, 4.3755338e-24], [192.75249, 1.0442e-22, 2.47004e-23, 1.48202e-23, 4.3620473e-24], [193.19683, 1.04399e-22, 2.46376e-23, 1.47826e-23, 4.3498305e-24], [193.6422, 1.04384e-22, 2.45798e-23, 1.47479e-23, 4.3387723e-24], [194.08859, 1.04374e-22, 2.45265e-23, 1.47159e-23, 4.3287737e-24], [194.53601, 1.04367e-22, 2.44773e-23, 1.46864e-23, 4.3197463e-24], [194.98446, 1.04364e-22, 2.44319e-23, 1.46591e-23, 4.311611e-24], [195.43395, 1.04365e-22, 2.43899e-23, 1.46339e-23, 4.3042968e-24], [195.88447, 1.04369e-22, 2.43511e-23, 1.46107e-23, 4.2977399e-24], [196.33603, 1.04376e-22, 2.43153e-23, 1.45892e-23, 4.2918831e-24], [196.78863, 1.04386e-22, 2.42821e-23, 1.45693e-23, 4.2866744e-24], [197.24227, 1.04398e-22, 2.42514e-23, 1.45508e-23, 4.282067e-24], [197.69696, 1.04413e-22, 2.4223e-23, 1.45338e-23, 4.2780187e-24], [198.1527, 1.04429e-22, 2.41967e-23, 1.4518e-23, 4.274491e-24], [198.60949, 1.04448e-22, 2.41724e-23, 1.45034e-23, 4.2714489e-24], [199.06733, 1.04469e-22, 2.41499e-23, 1.44899e-23, 4.2688608e-24], [199.52623, 1.04491e-22, 2.41291e-23, 1.44775e-23, 4.2666977e-24], [199.98619, 1.04515e-22, 2.41099e-23, 1.44659e-23, 4.2649333e-24], [200.4472, 1.04541e-22, 2.40922e-23, 1.44553e-23, 4.2635434e-24], [200.90928, 1.04568e-22, 2.40758e-23, 1.44455e-23, 4.262506e-24], [201.37242, 1.04596e-22, 2.40607e-23, 1.44364e-23, 4.2618009e-24], [201.83664, 1.04626e-22, 2.40469e-23, 1.44281e-23, 4.2614095e-24], [202.30192, 1.04657e-22, 2.40341e-23, 1.44205e-23, 4.2613149e-24], [202.76827, 1.04689e-22, 2.40224e-23, 1.44135e-23, 4.2615014e-24], [203.2357, 1.04723e-22, 2.40117e-23, 1.4407e-23, 4.2619546e-24], [203.70421, 1.04757e-22, 2.4002e-23, 1.44012e-23, 4.2626613e-24], [204.17379, 1.04792e-22, 2.39931e-23, 1.43959e-23, 4.2636093e-24], [204.64446, 1.04828e-22, 2.3985e-23, 1.4391e-23, 4.2647872e-24], [205.11622, 1.04865e-22, 2.39778e-23, 1.43867e-23, 4.2661846e-24], [205.58906, 1.04903e-22, 2.39713e-23, 1.43828e-23, 4.2677919e-24], [206.06299, 1.04942e-22, 2.39654e-23, 1.43793e-23, 4.2696002e-24], [206.53802, 1.04982e-22, 2.39603e-23, 1.43762e-23, 4.2716012e-24], [207.01413, 1.05022e-22, 2.39557e-23, 1.43734e-23, 4.2737872e-24], [207.49135, 1.05063e-22, 2.39518e-23, 1.43711e-23, 4.2761511e-24], [207.96967, 1.05105e-22, 2.39484e-23, 1.4369e-23, 4.2786864e-24], [208.44909, 1.05148e-22, 2.39456e-23, 1.43673e-23, 4.2813869e-24], [208.92961, 1.05191e-22, 2.39432e-23, 1.43659e-23, 4.2842468e-24], [209.41125, 1.05234e-22, 2.39414e-23, 1.43648e-23, 4.287261e-24], [209.89399, 1.05279e-22, 2.394e-23, 1.4364e-23, 4.2904244e-24], [210.37784, 1.05324e-22, 2.39391e-23, 1.43634e-23, 4.2937325e-24], [210.86281, 1.05369e-22, 2.39385e-23, 1.43631e-23, 4.2971809e-24], [211.3489, 1.05415e-22, 2.39384e-23, 1.43631e-23, 4.3007657e-24], [211.83611, 1.05462e-22, 2.39387e-23, 1.43632e-23, 4.3044831e-24], [212.32445, 1.05509e-22, 2.39393e-23, 1.43636e-23, 4.3083298e-24], [212.8139, 1.05557e-22, 2.39403e-23, 1.43642e-23, 4.3123024e-24], [213.30449, 1.05605e-22, 2.39417e-23, 1.4365e-23, 4.3163978e-24], [213.79621, 1.05654e-22, 2.39433e-23, 1.4366e-23, 4.3206134e-24], [214.28906, 1.05703e-22, 2.39453e-23, 1.43672e-23, 4.3249464e-24], [214.78305, 1.05752e-22, 2.39476e-23, 1.43686e-23, 4.3293943e-24], [215.27817, 1.05802e-22, 2.39502e-23, 1.43701e-23, 4.3339549e-24], [215.77444, 1.05853e-22, 2.3953e-23, 1.43718e-23, 4.338626e-24], [216.27185, 1.05904e-22, 2.39561e-23, 1.43737e-23, 4.3434055e-24], [216.77041, 1.05955e-22, 2.39595e-23, 1.43757e-23, 4.3482917e-24], [217.27012, 1.06007e-22, 2.39632e-23, 1.43779e-23, 4.3532826e-24], [217.77098, 1.06059e-22, 2.3967e-23, 1.43802e-23, 4.3583767e-24], [218.27299, 1.06111e-22, 2.39711e-23, 1.43827e-23, 4.3635724e-24], [218.77616, 1.06164e-22, 2.39755e-23, 1.43853e-23, 4.3688683e-24], [219.28049, 1.06218e-22, 2.398e-23, 1.4388e-23, 4.3742631e-24], [219.78599, 1.06272e-22, 2.39848e-23, 1.43909e-23, 4.3797553e-24], [220.29265, 1.06326e-22, 2.39898e-23, 1.43939e-23, 4.385344e-24], [220.80047, 1.0638e-22, 2.3995e-23, 1.4397e-23, 4.3910279e-24], [221.30947, 1.06435e-22, 2.40003e-23, 1.44002e-23, 4.396806e-24], [221.81964, 1.0649e-22, 2.40059e-23, 1.44035e-23, 4.4026774e-24], [222.33099, 1.06546e-22, 2.40116e-23, 1.4407e-23, 4.4086411e-24], [222.84351, 1.06602e-22, 2.40176e-23, 1.44105e-23, 4.4146964e-24], [223.35722, 1.06659e-22, 2.40237e-23, 1.44142e-23, 4.4208425e-24], [223.87211, 1.06715e-22, 2.40299e-23, 1.4418e-23, 4.4270785e-24], [224.38819, 1.06772e-22, 2.40364e-23, 1.44218e-23, 4.4334038e-24], [224.90546, 1.0683e-22, 2.4043e-23, 1.44258e-23, 4.4398179e-24], [225.42392, 1.06888e-22, 2.40497e-23, 1.44298e-23, 4.44632e-24], [225.94358, 1.06946e-22, 2.40567e-23, 1.4434e-23, 4.4529097e-24], [226.46443, 1.07004e-22, 2.40637e-23, 1.44382e-23, 4.4595864e-24], [226.98649, 1.07063e-22, 2.40709e-23, 1.44426e-23, 4.4663497e-24], [227.50974, 1.07122e-22, 2.40783e-23, 1.4447e-23, 4.4731992e-24], [228.03421, 1.07182e-22, 2.40858e-23, 1.44515e-23, 4.4801343e-24], [228.55988, 1.07241e-22, 2.40935e-23, 1.44561e-23, 4.4871548e-24], [229.08677, 1.07302e-22, 2.41012e-23, 1.44607e-23, 4.4942603e-24], [229.61486, 1.07362e-22, 2.41092e-23, 1.44655e-23, 4.5014505e-24], [230.14418, 1.07423e-22, 2.41172e-23, 1.44703e-23, 4.5087251e-24], [230.67472, 1.07484e-22, 2.41254e-23, 1.44752e-23, 4.5160839e-24], [231.20648, 1.07545e-22, 2.41337e-23, 1.44802e-23, 4.5235265e-24], [231.73946, 1.07607e-22, 2.41421e-23, 1.44853e-23, 4.5310528e-24], [232.27368, 1.07669e-22, 2.41507e-23, 1.44904e-23, 4.5386626e-24], [232.80913, 1.07732e-22, 2.41594e-23, 1.44956e-23, 4.5463557e-24], [233.34581, 1.07795e-22, 2.41682e-23, 1.45009e-23, 4.5541319e-24], [233.88372, 1.07858e-22, 2.41771e-23, 1.45063e-23, 4.5619911e-24], [234.42288, 1.07921e-22, 2.41861e-23, 1.45117e-23, 4.5699331e-24], [234.96328, 1.07985e-22, 2.41953e-23, 1.45172e-23, 4.577958e-24], [235.50493, 1.08049e-22, 2.42045e-23, 1.45227e-23, 4.5860655e-24], [236.04782, 1.08113e-22, 2.42139e-23, 1.45284e-23, 4.5942555e-24], [236.59197, 1.08178e-22, 2.42234e-23, 1.4534e-23, 4.6025281e-24], [237.13737, 1.08243e-22, 2.4233e-23, 1.45398e-23, 4.6108831e-24], [237.68403, 1.08308e-22, 2.42427e-23, 1.45456e-23, 4.6193206e-24], [238.23195, 1.08374e-22, 2.42525e-23, 1.45515e-23, 4.6278404e-24], [238.78113, 1.0844e-22, 2.42624e-23, 1.45575e-23, 4.6364426e-24], [239.33158, 1.08506e-22, 2.42725e-23, 1.45635e-23, 4.6451271e-24], [239.88329, 1.08573e-22, 2.42826e-23, 1.45696e-23, 4.653894e-24], [240.43628, 1.0864e-22, 2.42928e-23, 1.45757e-23, 4.6627432e-24], [240.99054, 1.08707e-22, 2.43031e-23, 1.45819e-23, 4.6716748e-24], [241.54608, 1.08774e-22, 2.43136e-23, 1.45881e-23, 4.6806888e-24], [242.1029, 1.08842e-22, 2.43241e-23, 1.45945e-23, 4.6897853e-24], [242.66101, 1.08911e-22, 2.43347e-23, 1.46008e-23, 4.6989642e-24], [243.2204, 1.08979e-22, 2.43455e-23, 1.46073e-23, 4.7082256e-24], [243.78108, 1.09048e-22, 2.43563e-23, 1.46138e-23, 4.7175695e-24], [244.34306, 1.09117e-22, 2.43672e-23, 1.46203e-23, 4.7269961e-24], [244.90632, 1.09186e-22, 2.43783e-23, 1.4627e-23, 4.7365054e-24], [245.47089, 1.09256e-22, 2.43894e-23, 1.46336e-23, 4.7460974e-24], [246.03676, 1.09326e-22, 2.44006e-23, 1.46404e-23, 4.7557723e-24], [246.60393, 1.09397e-22, 2.44119e-23, 1.46471e-23, 4.7655301e-24], [247.17241, 1.09468e-22, 2.44233e-23, 1.4654e-23, 4.7753708e-24], [247.74221, 1.09539e-22, 2.44348e-23, 1.46609e-23, 4.7852947e-24], [248.31331, 1.0961e-22, 2.44464e-23, 1.46678e-23, 4.7953017e-24], [248.88573, 1.09682e-22, 2.44581e-23, 1.46749e-23, 4.8053919e-24], [249.45947, 1.09754e-22, 2.44699e-23, 1.46819e-23, 4.8155656e-24], [250.03454, 1.09826e-22, 2.44817e-23, 1.4689e-23, 4.8258227e-24], [250.61093, 1.09899e-22, 2.44937e-23, 1.46962e-23, 4.8361634e-24], [251.18864, 1.09972e-22, 2.45057e-23, 1.47034e-23, 4.8465877e-24], [251.76769, 1.10045e-22, 2.45179e-23, 1.47107e-23, 4.8570959e-24], [252.34808, 1.10119e-22, 2.45301e-23, 1.47181e-23, 4.8676879e-24], [252.9298, 1.10193e-22, 2.45424e-23, 1.47255e-23, 4.878364e-24], [253.51286, 1.10267e-22, 2.45549e-23, 1.47329e-23, 4.8891242e-24], [254.09727, 1.10342e-22, 2.45674e-23, 1.47404e-23, 4.8999687e-24], [254.68303, 1.10416e-22, 2.45799e-23, 1.4748e-23, 4.9108975e-24], [255.27013, 1.10492e-22, 2.45926e-23, 1.47556e-23, 4.9219109e-24], [255.85859, 1.10567e-22, 2.46054e-23, 1.47632e-23, 4.9330088e-24], [256.4484, 1.10643e-22, 2.46183e-23, 1.4771e-23, 4.9441915e-24], [257.03958, 1.10719e-22, 2.46312e-23, 1.47787e-23, 4.9554591e-24], [257.63212, 1.10796e-22, 2.46442e-23, 1.47865e-23, 4.9668117e-24], [258.22602, 1.10873e-22, 2.46574e-23, 1.47944e-23, 4.9782495e-24], [258.82129, 1.1095e-22, 2.46706e-23, 1.48023e-23, 4.9897725e-24], [259.41794, 1.11028e-22, 2.46839e-23, 1.48103e-23, 5.0013808e-24], [260.01596, 1.11106e-22, 2.46972e-23, 1.48183e-23, 5.0130748e-24], [260.61535, 1.11184e-22, 2.47107e-23, 1.48264e-23, 5.0248543e-24], [261.21614, 1.11262e-22, 2.47243e-23, 1.48346e-23, 5.0367197e-24], [261.8183, 1.11341e-22, 2.47379e-23, 1.48428e-23, 5.048671e-24], [262.42185, 1.1142e-22, 2.47517e-23, 1.4851e-23, 5.0607083e-24], [263.0268, 1.115e-22, 2.47655e-23, 1.48593e-23, 5.0728319e-24], [263.63314, 1.1158e-22, 2.47794e-23, 1.48676e-23, 5.0850417e-24], [264.24088, 1.1166e-22, 2.47934e-23, 1.4876e-23, 5.097338e-24], [264.85001, 1.11741e-22, 2.48075e-23, 1.48845e-23, 5.109721e-24], [265.46056, 1.11822e-22, 2.48216e-23, 1.4893e-23, 5.1221906e-24], [266.07251, 1.11903e-22, 2.48359e-23, 1.49015e-23, 5.1347471e-24], [266.68587, 1.11984e-22, 2.48502e-23, 1.49101e-23, 5.1473906e-24], [267.30064, 1.12066e-22, 2.48647e-23, 1.49188e-23, 5.1601212e-24], [267.91683, 1.12149e-22, 2.48792e-23, 1.49275e-23, 5.1729391e-24], [268.53444, 1.12231e-22, 2.48938e-23, 1.49363e-23, 5.1858444e-24], [269.15348, 1.12314e-22, 2.49085e-23, 1.49451e-23, 5.1988372e-24], [269.77394, 1.12398e-22, 2.49232e-23, 1.49539e-23, 5.2119177e-24], [270.39584, 1.12481e-22, 2.49381e-23, 1.49629e-23, 5.225086e-24], [271.01916, 1.12565e-22, 2.4953e-23, 1.49718e-23, 5.2383421e-24], [271.64393, 1.12649e-22, 2.49681e-23, 1.49808e-23, 5.2516864e-24], [272.27013, 1.12734e-22, 2.49832e-23, 1.49899e-23, 5.2651188e-24], [272.89778, 1.12819e-22, 2.49984e-23, 1.4999e-23, 5.2786396e-24], [273.52687, 1.12905e-22, 2.50137e-23, 1.50082e-23, 5.2922488e-24], [274.15742, 1.1299e-22, 2.50291e-23, 1.50175e-23, 5.3059466e-24], [274.78942, 1.13076e-22, 2.50446e-23, 1.50267e-23, 5.319733e-24], [275.42287, 1.13163e-22, 2.50601e-23, 1.50361e-23, 5.3336083e-24], [276.05779, 1.1325e-22, 2.50758e-23, 1.50455e-23, 5.3475726e-24], [276.69416, 1.13337e-22, 2.50915e-23, 1.50549e-23, 5.361626e-24], [277.33201, 1.13424e-22, 2.51073e-23, 1.50644e-23, 5.3757685e-24], [277.97133, 1.13512e-22, 2.51232e-23, 1.50739e-23, 5.3900004e-24], [278.61212, 1.136e-22, 2.51392e-23, 1.50835e-23, 5.4043218e-24], [279.25438, 1.13689e-22, 2.51553e-23, 1.50932e-23, 5.4187327e-24], [279.89813, 1.13778e-22, 2.51714e-23, 1.51029e-23, 5.4332333e-24], [280.54336, 1.13867e-22, 2.51877e-23, 1.51126e-23, 5.4478238e-24], [281.19008, 1.13957e-22, 2.5204e-23, 1.51224e-23, 5.4625042e-24], [281.83829, 1.14047e-22, 2.52205e-23, 1.51323e-23, 5.4772747e-24], [282.488, 1.14137e-22, 2.5237e-23, 1.51422e-23, 5.4921353e-24], [283.1392, 1.14228e-22, 2.52536e-23, 1.51522e-23, 5.5070863e-24], [283.7919, 1.14319e-22, 2.52703e-23, 1.51622e-23, 5.5221276e-24], [284.44611, 1.1441e-22, 2.52871e-23, 1.51722e-23, 5.5372596e-24], [285.10183, 1.14502e-22, 2.53039e-23, 1.51824e-23, 5.5524821e-24], [285.75905, 1.14594e-22, 2.53209e-23, 1.51925e-23, 5.5677954e-24], [286.4178, 1.14687e-22, 2.53379e-23, 1.52028e-23, 5.5831996e-24], [287.07806, 1.1478e-22, 2.53551e-23, 1.52131e-23, 5.5986948e-24], [287.73984, 1.14873e-22, 2.53723e-23, 1.52234e-23, 5.6142811e-24], [288.40315, 1.14967e-22, 2.53896e-23, 1.52338e-23, 5.6299586e-24], [289.06799, 1.15061e-22, 2.5407e-23, 1.52442e-23, 5.6457274e-24], [289.73436, 1.15155e-22, 2.54245e-23, 1.52547e-23, 5.6615877e-24], [290.40227, 1.1525e-22, 2.54421e-23, 1.52653e-23, 5.6775395e-24], [291.07171, 1.15345e-22, 2.54598e-23, 1.52759e-23, 5.6935829e-24], [291.7427, 1.15441e-22, 2.54776e-23, 1.52865e-23, 5.7097181e-24], [292.41524, 1.15537e-22, 2.54954e-23, 1.52972e-23, 5.7259452e-24], [293.08932, 1.15633e-22, 2.55134e-23, 1.5308e-23, 5.7422642e-24], [293.76497, 1.1573e-22, 2.55314e-23, 1.53188e-23, 5.7586753e-24], [294.44216, 1.15827e-22, 2.55495e-23, 1.53297e-23, 5.7751785e-24], [295.12092, 1.15925e-22, 2.55677e-23, 1.53406e-23, 5.791774e-24], [295.80125, 1.16022e-22, 2.55861e-23, 1.53516e-23, 5.8084619e-24], [296.48314, 1.16121e-22, 2.56045e-23, 1.53627e-23, 5.8252423e-24], [297.1666, 1.16219e-22, 2.56229e-23, 1.53738e-23, 5.8421152e-24], [297.85164, 1.16318e-22, 2.56415e-23, 1.53849e-23, 5.8590808e-24], [298.53826, 1.16418e-22, 2.56602e-23, 1.53961e-23, 5.8761391e-24], [299.22646, 1.16517e-22, 2.5679e-23, 1.54074e-23, 5.8932903e-24], [299.91625, 1.16618e-22, 2.56978e-23, 1.54187e-23, 5.9105345e-24], [300.60763, 1.16718e-22, 2.57168e-23, 1.54301e-23, 5.9278717e-24], [301.3006, 1.16819e-22, 2.57358e-23, 1.54415e-23, 5.9453021e-24], [301.99517, 1.16921e-22, 2.5755e-23, 1.5453e-23, 5.9628257e-24], [302.69134, 1.17022e-22, 2.57742e-23, 1.54645e-23, 5.9804426e-24], [303.38912, 1.17124e-22, 2.57935e-23, 1.54761e-23, 5.998153e-24], [304.0885, 1.17227e-22, 2.5813e-23, 1.54878e-23, 6.0159569e-24], [304.7895, 1.1733e-22, 2.58325e-23, 1.54995e-23, 6.0338543e-24], [305.49211, 1.17433e-22, 2.58521e-23, 1.55113e-23, 6.0518455e-24], [306.19634, 1.17537e-22, 2.58718e-23, 1.55231e-23, 6.0699305e-24], [306.9022, 1.17641e-22, 2.58916e-23, 1.5535e-23, 6.0881093e-24], [307.60968, 1.17746e-22, 2.59115e-23, 1.55469e-23, 6.1063821e-24], [308.3188, 1.17851e-22, 2.59315e-23, 1.55589e-23, 6.1247489e-24], [309.02954, 1.17956e-22, 2.59516e-23, 1.55709e-23, 6.1432099e-24], [309.74193, 1.18062e-22, 2.59717e-23, 1.5583e-23, 6.1617651e-24], [310.45596, 1.18168e-22, 2.5992e-23, 1.55952e-23, 6.1804146e-24], [311.17163, 1.18275e-22, 2.60124e-23, 1.56074e-23, 6.1991585e-24], [311.88896, 1.18382e-22, 2.60329e-23, 1.56197e-23, 6.2179968e-24], [312.60794, 1.18489e-22, 2.60534e-23, 1.56321e-23, 6.2369298e-24], [313.32857, 1.18597e-22, 2.60741e-23, 1.56445e-23, 6.2559573e-24], [314.05087, 1.18705e-22, 2.60949e-23, 1.56569e-23, 6.2750796e-24], [314.77483, 1.18814e-22, 2.61157e-23, 1.56694e-23, 6.2942967e-24], [315.50046, 1.18923e-22, 2.61367e-23, 1.5682e-23, 6.3136087e-24], [316.22777, 1.19032e-22, 2.61578e-23, 1.56947e-23, 6.3330157e-24], [316.95675, 1.19142e-22, 2.61789e-23, 1.57073e-23, 6.3525178e-24], [317.68741, 1.19253e-22, 2.62002e-23, 1.57201e-23, 6.3721149e-24], [318.41975, 1.19363e-22, 2.62215e-23, 1.57329e-23, 6.3918073e-24], [319.15379, 1.19475e-22, 2.6243e-23, 1.57458e-23, 6.411595e-24], [319.88951, 1.19586e-22, 2.62645e-23, 1.57587e-23, 6.4314781e-24], [320.62693, 1.19698e-22, 2.62862e-23, 1.57717e-23, 6.4514567e-24], [321.36605, 1.19811e-22, 2.6308e-23, 1.57848e-23, 6.4715308e-24], [322.10688, 1.19924e-22, 2.63298e-23, 1.57979e-23, 6.4917006e-24], [322.84941, 1.20037e-22, 2.63518e-23, 1.58111e-23, 6.5119661e-24], [323.59366, 1.20151e-22, 2.63738e-23, 1.58243e-23, 6.5323273e-24], [324.33962, 1.20265e-22, 2.6396e-23, 1.58376e-23, 6.5527845e-24], [325.0873, 1.2038e-22, 2.64183e-23, 1.5851e-23, 6.5733376e-24], [325.8367, 1.20495e-22, 2.64407e-23, 1.58644e-23, 6.5939868e-24], [326.58783, 1.2061e-22, 2.64631e-23, 1.58779e-23, 6.6147321e-24], [327.34069, 1.20726e-22, 2.64857e-23, 1.58914e-23, 6.6355737e-24], [328.09529, 1.20842e-22, 2.65084e-23, 1.5905e-23, 6.6565115e-24], [328.85163, 1.20959e-22, 2.65312e-23, 1.59187e-23, 6.6775458e-24], [329.60971, 1.21076e-22, 2.65541e-23, 1.59325e-23, 6.6986765e-24], [330.36954, 1.21194e-22, 2.65771e-23, 1.59463e-23, 6.7199038e-24], [331.13112, 1.21312e-22, 2.66002e-23, 1.59601e-23, 6.7412278e-24], [331.89446, 1.21431e-22, 2.66234e-23, 1.59741e-23, 6.7626485e-24], [332.65955, 1.2155e-22, 2.66468e-23, 1.59881e-23, 6.7841661e-24], [333.42641, 1.21669e-22, 2.66702e-23, 1.60021e-23, 6.8057806e-24], [334.19504, 1.21789e-22, 2.66937e-23, 1.60162e-23, 6.8274921e-24], [334.96544, 1.2191e-22, 2.67174e-23, 1.60304e-23, 6.8493008e-24], [335.73761, 1.22031e-22, 2.67412e-23, 1.60447e-23, 6.8712068e-24], [336.51157, 1.22152e-22, 2.6765e-23, 1.6059e-23, 6.89321e-24], [337.28731, 1.22274e-22, 2.6789e-23, 1.60734e-23, 6.9153107e-24], [338.06484, 1.22396e-22, 2.68131e-23, 1.60879e-23, 6.937509e-24], [338.84416, 1.22519e-22, 2.68373e-23, 1.61024e-23, 6.959805e-24], [339.62527, 1.22642e-22, 2.68616e-23, 1.6117e-23, 6.9821987e-24], [340.40819, 1.22765e-22, 2.68861e-23, 1.61317e-23, 7.0046903e-24], [341.19291, 1.2289e-22, 2.69106e-23, 1.61464e-23, 7.0272799e-24], [341.97944, 1.23014e-22, 2.69353e-23, 1.61612e-23, 7.0499677e-24], [342.76779, 1.23139e-22, 2.69601e-23, 1.61761e-23, 7.0727537e-24], [343.55795, 1.23265e-22, 2.6985e-23, 1.6191e-23, 7.0956382e-24], [344.34993, 1.23391e-22, 2.701e-23, 1.6206e-23, 7.1186212e-24], [345.14374, 1.23517e-22, 2.70352e-23, 1.62211e-23, 7.1417029e-24], [345.93938, 1.23644e-22, 2.70604e-23, 1.62363e-23, 7.1648834e-24], [346.73685, 1.23771e-22, 2.70858e-23, 1.62515e-23, 7.188163e-24], [347.53616, 1.23899e-22, 2.71113e-23, 1.62668e-23, 7.2115417e-24], [348.33732, 1.24028e-22, 2.7137e-23, 1.62822e-23, 7.2350198e-24], [349.14032, 1.24157e-22, 2.71627e-23, 1.62976e-23, 7.2585974e-24], [349.94517, 1.24286e-22, 2.71886e-23, 1.63132e-23, 7.2822747e-24], [350.75187, 1.24416e-22, 2.72147e-23, 1.63288e-23, 7.306052e-24], [351.56044, 1.24546e-22, 2.72408e-23, 1.63445e-23, 7.3299293e-24], [352.37087, 1.24677e-22, 2.72671e-23, 1.63603e-23, 7.3539071e-24], [353.18317, 1.24808e-22, 2.72935e-23, 1.63761e-23, 7.3779854e-24], [353.99734, 1.2494e-22, 2.73201e-23, 1.6392e-23, 7.4021647e-24], [354.81339, 1.25072e-22, 2.73467e-23, 1.6408e-23, 7.426445e-24], [355.63132, 1.25205e-22, 2.73736e-23, 1.64241e-23, 7.4508268e-24], [356.45113, 1.25339e-22, 2.74005e-23, 1.64403e-23, 7.4753103e-24], [357.27284, 1.25472e-22, 2.74277e-23, 1.64566e-23, 7.4998958e-24], [358.09644, 1.25607e-22, 2.74549e-23, 1.6473e-23, 7.5245838e-24], [358.92193, 1.25742e-22, 2.74823e-23, 1.64894e-23, 7.5493746e-24], [359.74934, 1.25877e-22, 2.75099e-23, 1.65059e-23, 7.5742686e-24], [360.57864, 1.26013e-22, 2.75376e-23, 1.65226e-23, 7.5992662e-24], [361.40986, 1.2615e-22, 2.75655e-23, 1.65393e-23, 7.6243679e-24], [362.243, 1.26287e-22, 2.75936e-23, 1.65561e-23, 7.6495743e-24], [363.07805, 1.26424e-22, 2.76218e-23, 1.65731e-23, 7.6748859e-24], [363.91504, 1.26562e-22, 2.76502e-23, 1.65901e-23, 7.7003033e-24], [364.75395, 1.26701e-22, 2.76787e-23, 1.66072e-23, 7.7258271e-24], [365.59479, 1.2684e-22, 2.77075e-23, 1.66245e-23, 7.7514581e-24], [366.43757, 1.2698e-22, 2.77364e-23, 1.66418e-23, 7.7771971e-24], [367.2823, 1.2712e-22, 2.77656e-23, 1.66593e-23, 7.803045e-24], [368.12897, 1.27261e-22, 2.77949e-23, 1.66769e-23, 7.8290026e-24], [368.9776, 1.27403e-22, 2.78244e-23, 1.66947e-23, 7.8550712e-24], [369.82818, 1.27545e-22, 2.78542e-23, 1.67125e-23, 7.8812517e-24], [370.68072, 1.27688e-22, 2.78842e-23, 1.67305e-23, 7.9075455e-24], [371.53523, 1.27831e-22, 2.79145e-23, 1.67487e-23, 7.933954e-24], [372.39171, 1.27975e-22, 2.7945e-23, 1.6767e-23, 7.9604789e-24], [373.25016, 1.28119e-22, 2.79757e-23, 1.67854e-23, 7.9871218e-24], [374.11059, 1.28265e-22, 2.80068e-23, 1.68041e-23, 8.0138848e-24], [374.973, 1.28411e-22, 2.80381e-23, 1.68229e-23, 8.04077e-24], [375.8374, 1.28557e-22, 2.80698e-23, 1.68419e-23, 8.0677801e-24], [376.7038, 1.28705e-22, 2.81018e-23, 1.68611e-23, 8.0949178e-24], [377.57219, 1.28853e-22, 2.81342e-23, 1.68805e-23, 8.1221863e-24], [378.44258, 1.29001e-22, 2.81669e-23, 1.69001e-23, 8.1495894e-24], [379.31498, 1.29151e-22, 2.82001e-23, 1.69201e-23, 8.1771311e-24], [380.1894, 1.29301e-22, 2.82337e-23, 1.69402e-23, 8.2048163e-24], [381.06582, 1.29453e-22, 2.82679e-23, 1.69607e-23, 8.2326505e-24], [381.94427, 1.29605e-22, 2.83026e-23, 1.69816e-23, 8.2606401e-24], [382.82474, 1.29758e-22, 2.83379e-23, 1.70028e-23, 8.2887924e-24], [383.70725, 1.29912e-22, 2.83739e-23, 1.70243e-23, 8.3171162e-24], [384.59178, 1.30067e-22, 2.84107e-23, 1.70464e-23, 8.3456217e-24], [385.47836, 1.30224e-22, 2.84483e-23, 1.7069e-23, 8.3743207e-24], [386.36698, 1.30381e-22, 2.84868e-23, 1.70921e-23, 8.4032277e-24], [387.25764, 1.30541e-22, 2.85264e-23, 1.71159e-23, 8.4323595e-24], [388.15037, 1.30701e-22, 2.85673e-23, 1.71404e-23, 8.4617368e-24], [389.04514, 1.30863e-22, 2.86097e-23, 1.71658e-23, 8.4913844e-24], [389.94199, 1.31027e-22, 2.86536e-23, 1.71922e-23, 8.5213326e-24], [390.8409, 1.31194e-22, 2.86996e-23, 1.72198e-23, 8.5516189e-24], [391.74188, 1.31362e-22, 2.87479e-23, 1.72487e-23, 8.5822898e-24], [392.64494, 1.31534e-22, 2.87989e-23, 1.72793e-23, 8.613404e-24], [393.55008, 1.31709e-22, 2.88533e-23, 1.7312e-23, 8.6450358e-24], [394.4573, 1.31888e-22, 2.89117e-23, 1.7347e-23, 8.677281e-24], [395.36662, 1.32072e-22, 2.89752e-23, 1.73851e-23, 8.7102646e-24], [396.27803, 1.32262e-22, 2.9045e-23, 1.7427e-23, 8.7441522e-24], [397.19155, 1.32459e-22, 2.91228e-23, 1.74737e-23, 8.779167e-24], [398.10717, 1.32666e-22, 2.92109e-23, 1.75265e-23, 8.8156156e-24], [399.0249, 1.32885e-22, 2.93127e-23, 1.75876e-23, 8.8539287e-24], [399.94475, 1.33122e-22, 2.94328e-23, 1.76597e-23, 8.8947267e-24], [400.86672, 1.33381e-22, 2.95783e-23, 1.7747e-23, 8.9389296e-24], [401.79081, 1.33673e-22, 2.97598e-23, 1.78559e-23, 8.9879493e-24], [402.71703, 1.34014e-22, 2.99942e-23, 1.79965e-23, 9.0440432e-24], [403.64539, 1.34428e-22, 3.03099e-23, 1.81859e-23, 9.1110025e-24], [404.57589, 1.34965e-22, 3.07566e-23, 1.84539e-23, 9.1955954e-24], [405.50854, 1.35714e-22, 3.1429e-23, 1.88574e-23, 9.3108761e-24], [406.44333, 1.36874e-22, 3.25249e-23, 1.9515e-23, 9.4847096e-24], [407.38028, 1.38936e-22, 3.45119e-23, 2.07072e-23, 9.785441e-24], [408.31939, 1.43427e-22, 3.87176e-23, 2.32306e-23, 1.0418466e-23], [409.26066, 1.57151e-22, 5.02125e-23, 3.01275e-23, 1.2243674e-23], [410.2041, 2.4715e-22, 1.07745e-22, 6.46468e-23, 2.2669117e-23], [411.14972, 4.0278e-22, 1.9206e-22, 1.15236e-22, 3.9057526e-23], [412.09752, 1.66959e-22, 5.73461e-23, 3.44076e-23, 1.3492185e-23], [413.0475, 1.46234e-22, 4.07015e-23, 2.44209e-23, 1.0828084e-23], [413.99967, 1.40701e-22, 3.53615e-23, 2.12169e-23, 1.007425e-23], [414.95404, 1.38564e-22, 3.3044e-23, 1.98264e-23, 9.7797171e-24], [415.91061, 1.37583e-22, 3.18566e-23, 1.9114e-23, 9.6466141e-24], [416.86938, 1.37099e-22, 3.11807e-23, 1.87084e-23, 9.5839926e-24], [417.83037, 1.36864e-22, 3.07678e-23, 1.84607e-23, 9.5568481e-24], [418.79357, 1.36766e-22, 3.05033e-23, 1.8302e-23, 9.549458e-24], [419.75898, 1.3675e-22, 3.03286e-23, 1.81971e-23, 9.5539324e-24], [420.72663, 1.36786e-22, 3.02114e-23, 1.81268e-23, 9.5659664e-24], [421.6965, 1.36857e-22, 3.01325e-23, 1.80795e-23, 9.583053e-24], [422.66861, 1.36952e-22, 3.00804e-23, 1.80482e-23, 9.6036531e-24], [423.64297, 1.37065e-22, 3.00472e-23, 1.80283e-23, 9.6267803e-24], [424.61956, 1.3719e-22, 3.00281e-23, 1.80168e-23, 9.6517793e-24], [425.59841, 1.37326e-22, 3.00194e-23, 1.80116e-23, 9.678201e-24], [426.57952, 1.37468e-22, 3.00188e-23, 1.80113e-23, 9.7057297e-24], [427.56289, 1.37617e-22, 3.00244e-23, 1.80147e-23, 9.7341381e-24], [428.54852, 1.37771e-22, 3.0035e-23, 1.8021e-23, 9.7632595e-24], [429.53643, 1.37928e-22, 3.00496e-23, 1.80298e-23, 9.792969e-24], [430.52661, 1.38089e-22, 3.00675e-23, 1.80405e-23, 9.8231721e-24], [431.51908, 1.38253e-22, 3.0088e-23, 1.80528e-23, 9.853796e-24], [432.51383, 1.38419e-22, 3.01107e-23, 1.80664e-23, 9.8847837e-24], [433.51088, 1.38587e-22, 3.01353e-23, 1.80812e-23, 9.9160905e-24], [434.51022, 1.38757e-22, 3.01615e-23, 1.80969e-23, 9.9476807e-24], [435.51187, 1.38929e-22, 3.01891e-23, 1.81135e-23, 9.9795255e-24], [436.51583, 1.39102e-22, 3.02178e-23, 1.81307e-23, 1.0011601e-23], [437.52211, 1.39277e-22, 3.02476e-23, 1.81485e-23, 1.004389e-23], [438.5307, 1.39453e-22, 3.02782e-23, 1.81669e-23, 1.0076374e-23], [439.54162, 1.3963e-22, 3.03097e-23, 1.81858e-23, 1.0109042e-23], [440.55486, 1.39808e-22, 3.03418e-23, 1.82051e-23, 1.0141882e-23], [441.57045, 1.39987e-22, 3.03746e-23, 1.82248e-23, 1.0174886e-23], [442.58837, 1.40168e-22, 3.0408e-23, 1.82448e-23, 1.0208045e-23], [443.60864, 1.40349e-22, 3.04419e-23, 1.82651e-23, 1.0241353e-23], [444.63127, 1.40531e-22, 3.04763e-23, 1.82858e-23, 1.0274804e-23], [445.65625, 1.40714e-22, 3.05111e-23, 1.83066e-23, 1.0308394e-23], [446.68359, 1.40898e-22, 3.05463e-23, 1.83278e-23, 1.0342118e-23], [447.7133, 1.41082e-22, 3.05819e-23, 1.83491e-23, 1.0375973e-23], [448.74539, 1.41268e-22, 3.06178e-23, 1.83707e-23, 1.0409955e-23], [449.77985, 1.41454e-22, 3.06541e-23, 1.83924e-23, 1.0444062e-23], [450.8167, 1.41641e-22, 3.06907e-23, 1.84144e-23, 1.0478292e-23], [451.85594, 1.41829e-22, 3.07275e-23, 1.84365e-23, 1.0512643e-23], [452.89758, 1.42018e-22, 3.07647e-23, 1.84588e-23, 1.0547112e-23], [453.94162, 1.42207e-22, 3.08021e-23, 1.84813e-23, 1.0581698e-23], [454.98806, 1.42397e-22, 3.08398e-23, 1.85039e-23, 1.0616399e-23], [456.03692, 1.42588e-22, 3.08777e-23, 1.85266e-23, 1.0651216e-23], [457.08819, 1.42779e-22, 3.09159e-23, 1.85495e-23, 1.0686145e-23], [458.14189, 1.42971e-22, 3.09542e-23, 1.85725e-23, 1.0721187e-23], [459.19801, 1.43164e-22, 3.09928e-23, 1.85957e-23, 1.075634e-23], [460.25657, 1.43358e-22, 3.10317e-23, 1.8619e-23, 1.0791604e-23], [461.31757, 1.43552e-22, 3.10707e-23, 1.86424e-23, 1.0826978e-23], [462.38102, 1.43747e-22, 3.11099e-23, 1.86659e-23, 1.0862462e-23], [463.44692, 1.43943e-22, 3.11493e-23, 1.86896e-23, 1.0898054e-23], [464.51528, 1.44139e-22, 3.1189e-23, 1.87134e-23, 1.0933755e-23], [465.58609, 1.44336e-22, 3.12288e-23, 1.87373e-23, 1.0969563e-23], [466.65938, 1.44534e-22, 3.12688e-23, 1.87613e-23, 1.100548e-23], [467.73514, 1.44732e-22, 3.1309e-23, 1.87854e-23, 1.1041503e-23], [468.81338, 1.44931e-22, 3.13493e-23, 1.88096e-23, 1.1077633e-23], [469.89411, 1.45131e-22, 3.13899e-23, 1.88339e-23, 1.111387e-23], [470.97733, 1.45332e-22, 3.14306e-23, 1.88584e-23, 1.1150213e-23], [472.06304, 1.45533e-22, 3.14715e-23, 1.88829e-23, 1.1186662e-23], [473.15126, 1.45735e-22, 3.15126e-23, 1.89076e-23, 1.1223217e-23], [474.24199, 1.45937e-22, 3.15539e-23, 1.89323e-23, 1.1259877e-23], [475.33523, 1.4614e-22, 3.15953e-23, 1.89572e-23, 1.1296643e-23], [476.43099, 1.46344e-22, 3.16369e-23, 1.89821e-23, 1.1333515e-23], [477.52927, 1.46549e-22, 3.16786e-23, 1.90072e-23, 1.1370492e-23], [478.63009, 1.46754e-22, 3.17205e-23, 1.90323e-23, 1.1407574e-23], [479.73345, 1.4696e-22, 3.17626e-23, 1.90576e-23, 1.1444761e-23], [480.83935, 1.47167e-22, 3.18049e-23, 1.90829e-23, 1.1482053e-23], [481.9478, 1.47374e-22, 3.18473e-23, 1.91084e-23, 1.151945e-23], [483.0588, 1.47582e-22, 3.18899e-23, 1.91339e-23, 1.1556952e-23], [484.17237, 1.47791e-22, 3.19326e-23, 1.91596e-23, 1.1594559e-23], [485.2885, 1.48e-22, 3.19755e-23, 1.91853e-23, 1.163227e-23], [486.40721, 1.4821e-22, 3.20186e-23, 1.92112e-23, 1.1670087e-23], [487.52849, 1.48421e-22, 3.20618e-23, 1.92371e-23, 1.1708008e-23], [488.65236, 1.48632e-22, 3.21052e-23, 1.92631e-23, 1.1746033e-23], [489.77882, 1.48844e-22, 3.21488e-23, 1.92893e-23, 1.1784163e-23], [490.90788, 1.49057e-22, 3.21925e-23, 1.93155e-23, 1.1822398e-23], [492.03954, 1.4927e-22, 3.22364e-23, 1.93418e-23, 1.1860738e-23], [493.1738, 1.49485e-22, 3.22804e-23, 1.93682e-23, 1.1899182e-23], [494.31069, 1.497e-22, 3.23246e-23, 1.93948e-23, 1.1937731e-23], [495.45019, 1.49915e-22, 3.2369e-23, 1.94214e-23, 1.1976385e-23], [496.59232, 1.50131e-22, 3.24135e-23, 1.94481e-23, 1.2015143e-23], [497.73708, 1.50348e-22, 3.24581e-23, 1.94749e-23, 1.2054007e-23], [498.88449, 1.50566e-22, 3.2503e-23, 1.95018e-23, 1.2092974e-23], [500.03453, 1.50784e-22, 3.2548e-23, 1.95288e-23, 1.2132047e-23], [501.18723, 1.51003e-22, 3.25931e-23, 1.95559e-23, 1.2171225e-23], [502.34259, 1.51223e-22, 3.26384e-23, 1.95831e-23, 1.2210507e-23], [503.50061, 1.51444e-22, 3.26839e-23, 1.96103e-23, 1.2249894e-23], [504.6613, 1.51665e-22, 3.27295e-23, 1.96377e-23, 1.2289386e-23], [505.82466, 1.51887e-22, 3.27753e-23, 1.96652e-23, 1.2328983e-23], [506.99071, 1.52109e-22, 3.28213e-23, 1.96928e-23, 1.2368686e-23], [508.15944, 1.52333e-22, 3.28674e-23, 1.97204e-23, 1.2408493e-23], [509.33087, 1.52557e-22, 3.29136e-23, 1.97482e-23, 1.2448405e-23], [510.505, 1.52782e-22, 3.29601e-23, 1.9776e-23, 1.2488423e-23], [511.68184, 1.53007e-22, 3.30067e-23, 1.9804e-23, 1.2528546e-23], [512.86138, 1.53233e-22, 3.30534e-23, 1.9832e-23, 1.2568774e-23], [514.04365, 1.5346e-22, 3.31003e-23, 1.98602e-23, 1.2609107e-23], [515.22864, 1.53688e-22, 3.31474e-23, 1.98884e-23, 1.2649546e-23], [516.41637, 1.53916e-22, 3.31946e-23, 1.99168e-23, 1.269009e-23], [517.60683, 1.54145e-22, 3.3242e-23, 1.99452e-23, 1.273074e-23], [518.80004, 1.54375e-22, 3.32895e-23, 1.99737e-23, 1.2771496e-23], [519.996, 1.54605e-22, 3.33373e-23, 2.00024e-23, 1.2812357e-23], [521.19471, 1.54837e-22, 3.33851e-23, 2.00311e-23, 1.2853324e-23], [522.39619, 1.55068e-22, 3.34332e-23, 2.00599e-23, 1.2894396e-23], [523.60044, 1.55301e-22, 3.34814e-23, 2.00888e-23, 1.2935575e-23], [524.80746, 1.55535e-22, 3.35297e-23, 2.01178e-23, 1.297686e-23], [526.01727, 1.55769e-22, 3.35782e-23, 2.01469e-23, 1.301825e-23], [527.22986, 1.56004e-22, 3.36269e-23, 2.01762e-23, 1.3059747e-23], [528.44525, 1.56239e-22, 3.36758e-23, 2.02055e-23, 1.310135e-23], [529.66344, 1.56476e-22, 3.37248e-23, 2.02349e-23, 1.3143059e-23], [530.88444, 1.56713e-22, 3.3774e-23, 2.02644e-23, 1.3184875e-23], [532.10826, 1.56951e-22, 3.38233e-23, 2.0294e-23, 1.3226797e-23], [533.3349, 1.57189e-22, 3.38728e-23, 2.03237e-23, 1.3268825e-23], [534.56436, 1.57428e-22, 3.39225e-23, 2.03535e-23, 1.331096e-23], [535.79666, 1.57669e-22, 3.39723e-23, 2.03834e-23, 1.3353202e-23], [537.0318, 1.57909e-22, 3.40223e-23, 2.04134e-23, 1.3395551e-23], [538.26978, 1.58151e-22, 3.40725e-23, 2.04435e-23, 1.3438006e-23], [539.51062, 1.58393e-22, 3.41228e-23, 2.04737e-23, 1.3480568e-23], [540.75432, 1.58636e-22, 3.41733e-23, 2.0504e-23, 1.3523238e-23], [542.00089, 1.5888e-22, 3.42239e-23, 2.05344e-23, 1.3566014e-23], [543.25033, 1.59125e-22, 3.42748e-23, 2.05649e-23, 1.3608898e-23], [544.50265, 1.5937e-22, 3.43258e-23, 2.05955e-23, 1.3651889e-23], [545.75786, 1.59616e-22, 3.43769e-23, 2.06262e-23, 1.3694988e-23], [547.01596, 1.59863e-22, 3.44282e-23, 2.06569e-23, 1.3738194e-23], [548.27696, 1.60111e-22, 3.44797e-23, 2.06878e-23, 1.3781507e-23], [549.54087, 1.60359e-22, 3.45314e-23, 2.07188e-23, 1.3824929e-23], [550.8077, 1.60608e-22, 3.45832e-23, 2.07499e-23, 1.3868458e-23], [552.07744, 1.60858e-22, 3.46352e-23, 2.07811e-23, 1.3912095e-23], [553.35011, 1.61109e-22, 3.46874e-23, 2.08124e-23, 1.395584e-23], [554.62571, 1.6136e-22, 3.47397e-23, 2.08438e-23, 1.3999693e-23], [555.90426, 1.61613e-22, 3.47922e-23, 2.08753e-23, 1.4043654e-23], [557.18575, 1.61866e-22, 3.48449e-23, 2.0907e-23, 1.4087724e-23], [558.47019, 1.62119e-22, 3.48978e-23, 2.09387e-23, 1.4131901e-23], [559.7576, 1.62374e-22, 3.49508e-23, 2.09705e-23, 1.4176188e-23], [561.04798, 1.62629e-22, 3.5004e-23, 2.10024e-23, 1.4220583e-23], [562.34133, 1.62885e-22, 3.50573e-23, 2.10344e-23, 1.4265087e-23], [563.63766, 1.63142e-22, 3.51109e-23, 2.10665e-23, 1.4309699e-23], [564.93697, 1.634e-22, 3.51646e-23, 2.10987e-23, 1.4354421e-23], [566.23929, 1.63658e-22, 3.52184e-23, 2.11311e-23, 1.4399251e-23], [567.54461, 1.63918e-22, 3.52725e-23, 2.11635e-23, 1.4444191e-23], [568.85293, 1.64178e-22, 3.53267e-23, 2.1196e-23, 1.448924e-23], [570.16427, 1.64439e-22, 3.53811e-23, 2.12287e-23, 1.4534399e-23], [571.47864, 1.647e-22, 3.54357e-23, 2.12614e-23, 1.4579666e-23], [572.79603, 1.64963e-22, 3.54904e-23, 2.12942e-23, 1.4625044e-23], [574.11646, 1.65226e-22, 3.55453e-23, 2.13272e-23, 1.4670531e-23], [575.43994, 1.6549e-22, 3.56004e-23, 2.13603e-23, 1.4716128e-23], [576.76646, 1.65755e-22, 3.56557e-23, 2.13934e-23, 1.4761835e-23], [578.09605, 1.6602e-22, 3.57111e-23, 2.14267e-23, 1.4807652e-23], [579.4287, 1.66287e-22, 3.57668e-23, 2.14601e-23, 1.4853579e-23], [580.76442, 1.66554e-22, 3.58225e-23, 2.14935e-23, 1.4899616e-23], [582.10322, 1.66822e-22, 3.58785e-23, 2.15271e-23, 1.4945764e-23], [583.4451, 1.67091e-22, 3.59347e-23, 2.15608e-23, 1.4992023e-23], [584.79008, 1.67361e-22, 3.5991e-23, 2.15946e-23, 1.5038392e-23], [586.13816, 1.67631e-22, 3.60475e-23, 2.16285e-23, 1.5084871e-23], [587.48935, 1.67903e-22, 3.61042e-23, 2.16625e-23, 1.5131462e-23], [588.84366, 1.68175e-22, 3.6161e-23, 2.16966e-23, 1.5178164e-23], [590.20108, 1.68448e-22, 3.62181e-23, 2.17308e-23, 1.5224977e-23], [591.56163, 1.68721e-22, 3.62753e-23, 2.17652e-23, 1.5271901e-23], [592.92532, 1.68996e-22, 3.63327e-23, 2.17996e-23, 1.5318937e-23], [594.29216, 1.69271e-22, 3.63903e-23, 2.18342e-23, 1.5366084e-23], [595.66214, 1.69548e-22, 3.6448e-23, 2.18688e-23, 1.5413343e-23], [597.03529, 1.69825e-22, 3.6506e-23, 2.19036e-23, 1.5460713e-23], [598.4116, 1.70103e-22, 3.65641e-23, 2.19385e-23, 1.5508196e-23], [599.79108, 1.70381e-22, 3.66224e-23, 2.19734e-23, 1.5555791e-23], [601.17374, 1.70661e-22, 3.66809e-23, 2.20085e-23, 1.5603497e-23], [602.55959, 1.70941e-22, 3.67396e-23, 2.20437e-23, 1.5651317e-23], [603.94863, 1.71222e-22, 3.67984e-23, 2.2079e-23, 1.5699248e-23], [605.34087, 1.71505e-22, 3.68574e-23, 2.21145e-23, 1.5747292e-23], [606.73633, 1.71787e-22, 3.69167e-23, 2.215e-23, 1.5795449e-23], [608.135, 1.72071e-22, 3.69761e-23, 2.21856e-23, 1.5843719e-23], [609.5369, 1.72356e-22, 3.70356e-23, 2.22214e-23, 1.5892102e-23], [610.94202, 1.72641e-22, 3.70954e-23, 2.22573e-23, 1.5940598e-23], [612.35039, 1.72928e-22, 3.71554e-23, 2.22932e-23, 1.5989207e-23], [613.76201, 1.73215e-22, 3.72155e-23, 2.23293e-23, 1.603793e-23], [615.17687, 1.73503e-22, 3.72758e-23, 2.23655e-23, 1.6086766e-23], [616.595, 1.73792e-22, 3.73364e-23, 2.24018e-23, 1.6135717e-23], [618.0164, 1.74081e-22, 3.73971e-23, 2.24382e-23, 1.618478e-23], [619.44108, 1.74372e-22, 3.7458e-23, 2.24748e-23, 1.6233958e-23], [620.86903, 1.74663e-22, 3.7519e-23, 2.25114e-23, 1.6283251e-23], [622.30029, 1.74956e-22, 3.75803e-23, 2.25482e-23, 1.6332657e-23], [623.73484, 1.75249e-22, 3.76418e-23, 2.25851e-23, 1.6382178e-23], [625.17269, 1.75543e-22, 3.77034e-23, 2.2622e-23, 1.6431814e-23], [626.61386, 1.75838e-22, 3.77653e-23, 2.26592e-23, 1.6481564e-23], [628.05836, 1.76133e-22, 3.78273e-23, 2.26964e-23, 1.6531429e-23], [629.50618, 1.7643e-22, 3.78895e-23, 2.27337e-23, 1.658141e-23], [630.95734, 1.76728e-22, 3.79519e-23, 2.27711e-23, 1.6631506e-23], [632.41185, 1.77026e-22, 3.80145e-23, 2.28087e-23, 1.6681717e-23], [633.86971, 1.77325e-22, 3.80773e-23, 2.28464e-23, 1.6732044e-23], [635.33093, 1.77625e-22, 3.81403e-23, 2.28842e-23, 1.6782486e-23], [636.79552, 1.77926e-22, 3.82035e-23, 2.29221e-23, 1.6833044e-23], [638.26349, 1.78228e-22, 3.82669e-23, 2.29601e-23, 1.6883719e-23], [639.73484, 1.78531e-22, 3.83304e-23, 2.29983e-23, 1.693451e-23], [641.20958, 1.78835e-22, 3.83942e-23, 2.30365e-23, 1.6985417e-23], [642.68772, 1.79139e-22, 3.84582e-23, 2.30749e-23, 1.7036441e-23], [644.16927, 1.79445e-22, 3.85223e-23, 2.31134e-23, 1.7087581e-23], [645.65423, 1.79751e-22, 3.85867e-23, 2.3152e-23, 1.7138839e-23], [647.14262, 1.80058e-22, 3.86512e-23, 2.31907e-23, 1.7190213e-23], [648.63443, 1.80366e-22, 3.8716e-23, 2.32296e-23, 1.7241705e-23], [650.12969, 1.80675e-22, 3.87809e-23, 2.32686e-23, 1.7293315e-23], [651.62839, 1.80985e-22, 3.88461e-23, 2.33077e-23, 1.7345042e-23], [653.13055, 1.81296e-22, 3.89115e-23, 2.33469e-23, 1.7396887e-23], [654.63617, 1.81608e-22, 3.8977e-23, 2.33862e-23, 1.744885e-23], [656.14527, 1.81921e-22, 3.90428e-23, 2.34257e-23, 1.7500932e-23], [657.65784, 1.82234e-22, 3.91087e-23, 2.34652e-23, 1.7553132e-23], [659.1739, 1.82549e-22, 3.91749e-23, 2.3505e-23, 1.7605451e-23], [660.69345, 1.82864e-22, 3.92413e-23, 2.35448e-23, 1.7657889e-23], [662.2165, 1.83181e-22, 3.93079e-23, 2.35847e-23, 1.7710446e-23], [663.74307, 1.83498e-22, 3.93747e-23, 2.36248e-23, 1.7763123e-23], [665.27316, 1.83816e-22, 3.94417e-23, 2.3665e-23, 1.7815919e-23], [666.80677, 1.84135e-22, 3.95089e-23, 2.37053e-23, 1.7868835e-23], [668.34392, 1.84455e-22, 3.95763e-23, 2.37458e-23, 1.7921871e-23], [669.88461, 1.84776e-22, 3.9644e-23, 2.37864e-23, 1.7975028e-23], [671.42885, 1.85098e-22, 3.97118e-23, 2.38271e-23, 1.8028306e-23], [672.97666, 1.85421e-22, 3.97799e-23, 2.38679e-23, 1.8081705e-23], [674.52803, 1.85745e-22, 3.98482e-23, 2.39089e-23, 1.8135226e-23], [676.08298, 1.8607e-22, 3.99167e-23, 2.395e-23, 1.8188868e-23], [677.64151, 1.86395e-22, 3.99855e-23, 2.39913e-23, 1.8242633e-23], [679.20363, 1.86722e-22, 4.00544e-23, 2.40327e-23, 1.829652e-23], [680.76936, 1.8705e-22, 4.01237e-23, 2.40742e-23, 1.835053e-23], [682.33869, 1.87378e-22, 4.01931e-23, 2.41159e-23, 1.8404665e-23], [683.91165, 1.87708e-22, 4.02628e-23, 2.41577e-23, 1.8458923e-23], [685.48823, 1.88038e-22, 4.03327e-23, 2.41996e-23, 1.8513306e-23], [687.06844, 1.8837e-22, 4.04029e-23, 2.42418e-23, 1.8567815e-23], [688.6523, 1.88703e-22, 4.04734e-23, 2.4284e-23, 1.862245e-23], [690.2398, 1.89036e-22, 4.05441e-23, 2.43265e-23, 1.8677213e-23], [691.83097, 1.89371e-22, 4.06152e-23, 2.43691e-23, 1.8732104e-23], [693.42581, 1.89706e-22, 4.06865e-23, 2.44119e-23, 1.8787125e-23], [695.02432, 1.90043e-22, 4.07581e-23, 2.44549e-23, 1.8842277e-23], [696.62651, 1.90381e-22, 4.08301e-23, 2.44981e-23, 1.8897563e-23], [698.2324, 1.9072e-22, 4.09024e-23, 2.45415e-23, 1.8952983e-23], [699.842, 1.91059e-22, 4.09751e-23, 2.45851e-23, 1.9008542e-23], [701.4553, 1.91401e-22, 4.10483e-23, 2.4629e-23, 1.9064242e-23], [703.07232, 1.91743e-22, 4.11219e-23, 2.46731e-23, 1.9120088e-23], [704.69307, 1.92086e-22, 4.11961e-23, 2.47176e-23, 1.9176085e-23], [706.31755, 1.92431e-22, 4.12708e-23, 2.47625e-23, 1.9232241e-23], [707.94578, 1.92778e-22, 4.13464e-23, 2.48078e-23, 1.9288565e-23], [709.57777, 1.93125e-22, 4.14227e-23, 2.48536e-23, 1.9345069e-23], [711.21351, 1.93475e-22, 4.15002e-23, 2.49001e-23, 1.9401772e-23], [712.85303, 1.93826e-22, 4.15791e-23, 2.49475e-23, 1.9458699e-23], [714.49633, 1.9418e-22, 4.16598e-23, 2.49959e-23, 1.9515885e-23], [716.14341, 1.94537e-22, 4.17429e-23, 2.50458e-23, 1.9573387e-23], [717.79429, 1.94898e-22, 4.18295e-23, 2.50977e-23, 1.963129e-23], [719.44898, 1.95263e-22, 4.19213e-23, 2.51528e-23, 1.9689733e-23], [721.10748, 1.95637e-22, 4.2021e-23, 2.52126e-23, 1.9748958e-23], [722.7698, 1.96022e-22, 4.21339e-23, 2.52803e-23, 1.9809412e-23], [724.43596, 1.96429e-22, 4.22707e-23, 2.53624e-23, 1.9871998e-23], [726.10596, 1.96879e-22, 4.24556e-23, 2.54733e-23, 1.9938784e-23], [727.7798, 1.97428e-22, 4.27544e-23, 2.56527e-23, 2.0015402e-23], [729.45751, 1.98282e-22, 4.34016e-23, 2.6041e-23, 2.0122123e-23], [731.13908, 2.00707e-22, 4.57893e-23, 2.74736e-23, 2.0383369e-23], [732.82453, 2.52855e-22, 8.93346e-23, 5.36008e-23, 2.5552234e-23], [734.51387, 2.07817e-22, 5.25794e-23, 3.15476e-23, 2.1127193e-23], [736.2071, 2.00407e-22, 4.44867e-23, 2.6692e-23, 2.0421377e-23], [737.90423, 1.99766e-22, 4.34219e-23, 2.60531e-23, 2.0381237e-23], [739.60528, 1.99819e-22, 4.31404e-23, 2.58843e-23, 2.040925e-23], [741.31024, 2.00048e-22, 4.30612e-23, 2.58367e-23, 2.0454636e-23], [743.01914, 2.00342e-22, 4.30566e-23, 2.5834e-23, 2.050647e-23], [744.73197, 2.00666e-22, 4.3086e-23, 2.58516e-23, 2.0561282e-23], [746.44876, 2.01007e-22, 4.3133e-23, 2.58798e-23, 2.0617691e-23], [748.1695, 2.01357e-22, 4.31902e-23, 2.59141e-23, 2.0675063e-23], [749.89421, 2.01713e-22, 4.32535e-23, 2.59521e-23, 2.073307e-23], [751.62289, 2.02074e-22, 4.3321e-23, 2.59926e-23, 2.0791533e-23], [753.35556, 2.02437e-22, 4.33913e-23, 2.60348e-23, 2.0850344e-23], [755.09223, 2.02804e-22, 4.34636e-23, 2.60782e-23, 2.0909436e-23], [756.8329, 2.03172e-22, 4.35375e-23, 2.61225e-23, 2.0968768e-23], [758.57758, 2.03543e-22, 4.36125e-23, 2.61675e-23, 2.102831e-23], [760.32628, 2.03915e-22, 4.36885e-23, 2.62131e-23, 2.1088043e-23], [762.07901, 2.04288e-22, 4.37653e-23, 2.62592e-23, 2.1147952e-23], [763.83578, 2.04663e-22, 4.38427e-23, 2.63056e-23, 2.1208029e-23], [765.59661, 2.0504e-22, 4.39207e-23, 2.63524e-23, 2.1268264e-23], [767.36149, 2.05417e-22, 4.39992e-23, 2.63995e-23, 2.1328653e-23], [769.13044, 2.05796e-22, 4.40781e-23, 2.64469e-23, 2.1389192e-23], [770.90347, 2.06176e-22, 4.41574e-23, 2.64945e-23, 2.1449878e-23], [772.68059, 2.06558e-22, 4.42371e-23, 2.65423e-23, 2.1510708e-23], [774.4618, 2.0694e-22, 4.43172e-23, 2.65903e-23, 2.157168e-23], [776.24712, 2.07324e-22, 4.43976e-23, 2.66386e-23, 2.1632793e-23], [778.03655, 2.07708e-22, 4.44783e-23, 2.6687e-23, 2.1694046e-23], [779.83011, 2.08094e-22, 4.45593e-23, 2.67356e-23, 2.1755439e-23], [781.6278, 2.08481e-22, 4.46406e-23, 2.67844e-23, 2.1816969e-23], [783.42964, 2.08869e-22, 4.47222e-23, 2.68333e-23, 2.1878637e-23], [785.23563, 2.09258e-22, 4.48041e-23, 2.68824e-23, 2.1940443e-23], [787.04579, 2.09648e-22, 4.48862e-23, 2.69317e-23, 2.2002386e-23], [788.86012, 2.1004e-22, 4.49686e-23, 2.69812e-23, 2.2064466e-23], [790.67863, 2.10432e-22, 4.50513e-23, 2.70308e-23, 2.2126683e-23], [792.50133, 2.10826e-22, 4.51342e-23, 2.70805e-23, 2.2189036e-23], [794.32823, 2.1122e-22, 4.52174e-23, 2.71305e-23, 2.2251527e-23], [796.15935, 2.11616e-22, 4.53009e-23, 2.71805e-23, 2.2314154e-23], [797.99469, 2.12013e-22, 4.53846e-23, 2.72307e-23, 2.2376918e-23], [799.83426, 2.12411e-22, 4.54685e-23, 2.72811e-23, 2.2439819e-23], [801.67806, 2.1281e-22, 4.55527e-23, 2.73316e-23, 2.2502857e-23], [803.52612, 2.1321e-22, 4.56372e-23, 2.73823e-23, 2.2566032e-23], [805.37844, 2.13612e-22, 4.57219e-23, 2.74331e-23, 2.2629344e-23], [807.23503, 2.14014e-22, 4.58068e-23, 2.74841e-23, 2.2692793e-23], [809.0959, 2.14418e-22, 4.5892e-23, 2.75352e-23, 2.2756381e-23], [810.96106, 2.14822e-22, 4.59775e-23, 2.75865e-23, 2.2820106e-23], [812.83052, 2.15228e-22, 4.60632e-23, 2.76379e-23, 2.2883968e-23], [814.70428, 2.15635e-22, 4.61491e-23, 2.76895e-23, 2.2947969e-23], [816.58237, 2.16043e-22, 4.62353e-23, 2.77412e-23, 2.3012109e-23], [818.46479, 2.16452e-22, 4.63217e-23, 2.7793e-23, 2.3076387e-23], [820.35154, 2.16862e-22, 4.64084e-23, 2.7845e-23, 2.3140803e-23], [822.24265, 2.17273e-22, 4.64953e-23, 2.78972e-23, 2.3205359e-23], [824.13812, 2.17686e-22, 4.65825e-23, 2.79495e-23, 2.3270054e-23], [826.03795, 2.18099e-22, 4.66699e-23, 2.80019e-23, 2.3334888e-23], [827.94216, 2.18514e-22, 4.67576e-23, 2.80545e-23, 2.3399862e-23], [829.85077, 2.1893e-22, 4.68455e-23, 2.81073e-23, 2.3464975e-23], [831.76377, 2.19347e-22, 4.69336e-23, 2.81602e-23, 2.3530229e-23], [833.68118, 2.19765e-22, 4.7022e-23, 2.82132e-23, 2.3595624e-23], [835.60302, 2.20184e-22, 4.71107e-23, 2.82664e-23, 2.3661159e-23], [837.52928, 2.20605e-22, 4.71996e-23, 2.83197e-23, 2.3726834e-23], [839.45999, 2.21026e-22, 4.72887e-23, 2.83732e-23, 2.3792651e-23], [841.39514, 2.21449e-22, 4.73781e-23, 2.84269e-23, 2.385861e-23], [843.33476, 2.21872e-22, 4.74677e-23, 2.84806e-23, 2.3924709e-23], [845.27885, 2.22297e-22, 4.75576e-23, 2.85346e-23, 2.3990951e-23], [847.22741, 2.22724e-22, 4.76478e-23, 2.85887e-23, 2.4057335e-23], [849.18048, 2.23151e-22, 4.77381e-23, 2.86429e-23, 2.4123862e-23], [851.13804, 2.23579e-22, 4.78288e-23, 2.86973e-23, 2.4190531e-23], [853.10011, 2.24009e-22, 4.79197e-23, 2.87518e-23, 2.4257343e-23], [855.06671, 2.24439e-22, 4.80108e-23, 2.88065e-23, 2.4324299e-23], [857.03785, 2.24871e-22, 4.81022e-23, 2.88613e-23, 2.4391397e-23], [859.01352, 2.25304e-22, 4.81938e-23, 2.89163e-23, 2.445864e-23], [860.99375, 2.25738e-22, 4.82857e-23, 2.89714e-23, 2.4526027e-23], [862.97855, 2.26174e-22, 4.83779e-23, 2.90267e-23, 2.4593558e-23], [864.96792, 2.2661e-22, 4.84703e-23, 2.90822e-23, 2.4661234e-23], [866.96188, 2.27048e-22, 4.85629e-23, 2.91378e-23, 2.4729054e-23], [868.96043, 2.27487e-22, 4.86558e-23, 2.91935e-23, 2.479702e-23], [870.96359, 2.27927e-22, 4.8749e-23, 2.92494e-23, 2.4865131e-23], [872.97137, 2.28368e-22, 4.88424e-23, 2.93054e-23, 2.4933388e-23], [874.98378, 2.28811e-22, 4.89361e-23, 2.93616e-23, 2.5001791e-23], [877.00082, 2.29254e-22, 4.903e-23, 2.9418e-23, 2.5070341e-23], [879.02252, 2.29699e-22, 4.91242e-23, 2.94745e-23, 2.5139037e-23], [881.04887, 2.30145e-22, 4.92186e-23, 2.95312e-23, 2.520788e-23], [883.0799, 2.30592e-22, 4.93133e-23, 2.9588e-23, 2.527687e-23], [885.11561, 2.31041e-22, 4.94083e-23, 2.9645e-23, 2.5346007e-23], [887.15601, 2.3149e-22, 4.95035e-23, 2.97021e-23, 2.5415292e-23], [889.20112, 2.31941e-22, 4.95989e-23, 2.97594e-23, 2.5484726e-23], [891.25094, 2.32393e-22, 4.96947e-23, 2.98168e-23, 2.5554307e-23], [893.30548, 2.32846e-22, 4.97907e-23, 2.98744e-23, 2.5624038e-23], [895.36477, 2.33301e-22, 4.98869e-23, 2.99322e-23, 2.5693917e-23], [897.42879, 2.33756e-22, 4.99834e-23, 2.99901e-23, 2.5763945e-23], [899.49758, 2.34213e-22, 5.00802e-23, 3.00481e-23, 2.5834124e-23], [901.57114, 2.34671e-22, 5.01772e-23, 3.01063e-23, 2.5904451e-23], [903.64947, 2.3513e-22, 5.02745e-23, 3.01647e-23, 2.597493e-23], [905.7326, 2.35591e-22, 5.03721e-23, 3.02233e-23, 2.6045558e-23], [907.82053, 2.36052e-22, 5.04699e-23, 3.02819e-23, 2.6116337e-23], [909.91327, 2.36515e-22, 5.0568e-23, 3.03408e-23, 2.6187268e-23], [912.01084, 2.3698e-22, 5.06664e-23, 3.03998e-23, 2.625835e-23], [914.11324, 2.37445e-22, 5.0765e-23, 3.0459e-23, 2.6329583e-23], [916.22049, 2.37912e-22, 5.08639e-23, 3.05183e-23, 2.6400969e-23], [918.3326, 2.38379e-22, 5.0963e-23, 3.05778e-23, 2.6472506e-23], [920.44957, 2.38848e-22, 5.10624e-23, 3.06375e-23, 2.6544197e-23], [922.57143, 2.39319e-22, 5.11621e-23, 3.06973e-23, 2.661604e-23], [924.69817, 2.3979e-22, 5.12621e-23, 3.07572e-23, 2.6688037e-23], [926.82982, 2.40263e-22, 5.13623e-23, 3.08174e-23, 2.6760187e-23], [928.96639, 2.40737e-22, 5.14628e-23, 3.08777e-23, 2.6832491e-23], [931.10788, 2.41213e-22, 5.15635e-23, 3.09381e-23, 2.6904949e-23], [933.2543, 2.41689e-22, 5.16645e-23, 3.09987e-23, 2.6977562e-23], [935.40567, 2.42167e-22, 5.17658e-23, 3.10595e-23, 2.705033e-23], [937.56201, 2.42646e-22, 5.18674e-23, 3.11204e-23, 2.7123252e-23], [939.72331, 2.43127e-22, 5.19693e-23, 3.11816e-23, 2.719633e-23], [941.8896, 2.43608e-22, 5.20714e-23, 3.12428e-23, 2.7269565e-23], [944.06088, 2.44091e-22, 5.21738e-23, 3.13043e-23, 2.7342955e-23], [946.23716, 2.44575e-22, 5.22764e-23, 3.13658e-23, 2.7416501e-23], [948.41846, 2.45061e-22, 5.23793e-23, 3.14276e-23, 2.7490205e-23], [950.60479, 2.45547e-22, 5.24826e-23, 3.14895e-23, 2.7564065e-23], [952.79616, 2.46035e-22, 5.2586e-23, 3.15516e-23, 2.7638083e-23], [954.99259, 2.46525e-22, 5.26898e-23, 3.16139e-23, 2.7712259e-23], [957.19407, 2.47015e-22, 5.27938e-23, 3.16763e-23, 2.7786593e-23], [959.40063, 2.47507e-22, 5.28981e-23, 3.17389e-23, 2.7861085e-23], [961.61228, 2.48e-22, 5.30027e-23, 3.18016e-23, 2.7935736e-23], [963.82902, 2.48495e-22, 5.31076e-23, 3.18646e-23, 2.8010546e-23], [966.05088, 2.4899e-22, 5.32127e-23, 3.19276e-23, 2.8085516e-23], [968.27786, 2.49487e-22, 5.33182e-23, 3.19909e-23, 2.8160645e-23], [970.50997, 2.49986e-22, 5.34239e-23, 3.20543e-23, 2.8235935e-23], [972.74722, 2.50485e-22, 5.35299e-23, 3.21179e-23, 2.8311385e-23], [974.98964, 2.50986e-22, 5.36361e-23, 3.21817e-23, 2.8386996e-23], [977.23722, 2.51489e-22, 5.37427e-23, 3.22456e-23, 2.8462768e-23], [979.48999, 2.51992e-22, 5.38495e-23, 3.23097e-23, 2.8538701e-23], [981.74794, 2.52497e-22, 5.39566e-23, 3.2374e-23, 2.8614796e-23], [984.01111, 2.53003e-22, 5.4064e-23, 3.24384e-23, 2.8691054e-23], [986.27949, 2.53511e-22, 5.41717e-23, 3.2503e-23, 2.8767474e-23], [988.55309, 2.5402e-22, 5.42797e-23, 3.25678e-23, 2.8844057e-23], [990.83194, 2.5453e-22, 5.43879e-23, 3.26327e-23, 2.8920803e-23], [993.11605, 2.55041e-22, 5.44964e-23, 3.26979e-23, 2.8997712e-23], [995.40542, 2.55554e-22, 5.46053e-23, 3.27632e-23, 2.9074786e-23], [997.70006, 2.56068e-22, 5.47144e-23, 3.28286e-23, 2.9152024e-23], [1000.0, 2.56584e-22, 5.48238e-23, 3.28943e-23, 2.9229426e-23], [1002.3052, 2.57101e-22, 5.49334e-23, 3.29601e-23, 2.9306994e-23], [1004.6158, 2.57619e-22, 5.50434e-23, 3.30261e-23, 2.9384726e-23], [1006.9317, 2.58138e-22, 5.51537e-23, 3.30922e-23, 2.9462625e-23], [1009.2529, 2.58659e-22, 5.52642e-23, 3.31585e-23, 2.9540689e-23], [1011.5795, 2.59181e-22, 5.53751e-23, 3.3225e-23, 2.961892e-23], [1013.9114, 2.59705e-22, 5.54862e-23, 3.32917e-23, 2.9697318e-23], [1016.2487, 2.6023e-22, 5.55976e-23, 3.33586e-23, 2.9775883e-23], [1018.5914, 2.60756e-22, 5.57094e-23, 3.34256e-23, 2.9854616e-23], [1020.9395, 2.61284e-22, 5.58214e-23, 3.34928e-23, 2.9933516e-23], [1023.293, 2.61813e-22, 5.59337e-23, 3.35602e-23, 3.0012585e-23], [1025.6519, 2.62343e-22, 5.60463e-23, 3.36278e-23, 3.0091822e-23], [1028.0163, 2.62875e-22, 5.61592e-23, 3.36955e-23, 3.0171228e-23], [1030.3861, 2.63408e-22, 5.62724e-23, 3.37634e-23, 3.0250804e-23], [1032.7614, 2.63943e-22, 5.63859e-23, 3.38315e-23, 3.0330549e-23], [1035.1422, 2.64479e-22, 5.64996e-23, 3.38998e-23, 3.0410464e-23], [1037.5284, 2.65016e-22, 5.66137e-23, 3.39682e-23, 3.049055e-23], [1039.9202, 2.65555e-22, 5.67281e-23, 3.40369e-23, 3.0570807e-23], [1042.3174, 2.66095e-22, 5.68428e-23, 3.41057e-23, 3.0651235e-23], [1044.7202, 2.66636e-22, 5.69578e-23, 3.41747e-23, 3.0731835e-23], [1047.1285, 2.67179e-22, 5.70731e-23, 3.42438e-23, 3.0812606e-23], [1049.5424, 2.67724e-22, 5.71887e-23, 3.43132e-23, 3.089355e-23], [1051.9619, 2.68269e-22, 5.73045e-23, 3.43827e-23, 3.0974667e-23], [1054.3869, 2.68816e-22, 5.74207e-23, 3.44524e-23, 3.1055957e-23], [1056.8175, 2.69365e-22, 5.75372e-23, 3.45223e-23, 3.1137421e-23], [1059.2537, 2.69915e-22, 5.7654e-23, 3.45924e-23, 3.1219059e-23], [1061.6956, 2.70466e-22, 5.77711e-23, 3.46627e-23, 3.1300871e-23], [1064.143, 2.71019e-22, 5.78886e-23, 3.47331e-23, 3.1382858e-23], [1066.5961, 2.71573e-22, 5.80063e-23, 3.48038e-23, 3.146502e-23], [1069.0549, 2.72129e-22, 5.81243e-23, 3.48746e-23, 3.1547357e-23], [1071.5193, 2.72686e-22, 5.82427e-23, 3.49456e-23, 3.1629871e-23], [1073.9894, 2.73244e-22, 5.83613e-23, 3.50168e-23, 3.1712561e-23], [1076.4652, 2.73804e-22, 5.84803e-23, 3.50882e-23, 3.1795428e-23], [1078.9467, 2.74365e-22, 5.85996e-23, 3.51597e-23, 3.1878472e-23], [1081.434, 2.74928e-22, 5.87191e-23, 3.52315e-23, 3.1961695e-23], [1083.9269, 2.75492e-22, 5.88391e-23, 3.53034e-23, 3.2045095e-23], [1086.4256, 2.76058e-22, 5.89593e-23, 3.53756e-23, 3.2128674e-23], [1088.9301, 2.76625e-22, 5.90798e-23, 3.54479e-23, 3.2212433e-23], [1091.4403, 2.77194e-22, 5.92007e-23, 3.55204e-23, 3.2296371e-23], [1093.9564, 2.77764e-22, 5.93219e-23, 3.55932e-23, 3.238049e-23], [1096.4782, 2.78335e-22, 5.94435e-23, 3.56661e-23, 3.246479e-23], [1099.0058, 2.78908e-22, 5.95653e-23, 3.57392e-23, 3.2549271e-23], [1101.5393, 2.79483e-22, 5.96875e-23, 3.58125e-23, 3.2633935e-23], [1104.0786, 2.80059e-22, 5.98101e-23, 3.58861e-23, 3.2718781e-23], [1106.6238, 2.80636e-22, 5.9933e-23, 3.59598e-23, 3.2803812e-23], [1109.1748, 2.81216e-22, 6.00563e-23, 3.60338e-23, 3.2889028e-23], [1111.7317, 2.81796e-22, 6.01799e-23, 3.61079e-23, 3.2974431e-23], [1114.2945, 2.82378e-22, 6.0304e-23, 3.61824e-23, 3.3060022e-23], [1116.8632, 2.82962e-22, 6.04284e-23, 3.6257e-23, 3.3145803e-23], [1119.4379, 2.83547e-22, 6.05533e-23, 3.6332e-23, 3.3231777e-23], [1122.0185, 2.84134e-22, 6.06787e-23, 3.64072e-23, 3.3317948e-23], [1124.605, 2.84723e-22, 6.08047e-23, 3.64828e-23, 3.3404322e-23], [1127.1975, 2.85313e-22, 6.09314e-23, 3.65589e-23, 3.3490908e-23], [1129.7959, 2.85905e-22, 6.1059e-23, 3.66354e-23, 3.3577718e-23], [1132.4004, 2.865e-22, 6.11877e-23, 3.67126e-23, 3.3664777e-23], [1135.0108, 2.87097e-22, 6.13182e-23, 3.67909e-23, 3.3752126e-23], [1137.6273, 2.87699e-22, 6.14516e-23, 3.6871e-23, 3.3839847e-23], [1140.2498, 2.88305e-22, 6.15905e-23, 3.69543e-23, 3.3928126e-23], [1142.8783, 2.88924e-22, 6.17414e-23, 3.70448e-23, 3.4017446e-23], [1145.5129, 2.89573e-22, 6.19271e-23, 3.71562e-23, 3.4109448e-23], [1148.1536, 2.90359e-22, 6.22696e-23, 3.73617e-23, 3.4213023e-23], [1150.8004, 2.93501e-22, 6.53032e-23, 3.91819e-23, 3.4516831e-23], [1153.4533, 2.93343e-22, 6.45712e-23, 3.87427e-23, 3.4540335e-23], [1156.1122, 2.92137e-22, 6.26149e-23, 3.75689e-23, 3.4475119e-23], [1158.7774, 2.92585e-22, 6.25575e-23, 3.75345e-23, 3.4550305e-23], [1161.4486, 2.93149e-22, 6.26345e-23, 3.75807e-23, 3.4635423e-23], [1164.126, 2.93741e-22, 6.27425e-23, 3.76455e-23, 3.4722956e-23], [1166.8096, 2.94343e-22, 6.28616e-23, 3.77169e-23, 3.4811455e-23], [1169.4994, 2.94951e-22, 6.29857e-23, 3.77914e-23, 3.4900485e-23], [1172.1954, 2.95563e-22, 6.31124e-23, 3.78675e-23, 3.4989877e-23], [1174.8976, 2.96178e-22, 6.32409e-23, 3.79445e-23, 3.5079557e-23], [1177.606, 2.96794e-22, 6.33704e-23, 3.80223e-23, 3.5169487e-23], [1180.3206, 2.97413e-22, 6.35008e-23, 3.81005e-23, 3.5259647e-23], [1183.0416, 2.98033e-22, 6.36319e-23, 3.81791e-23, 3.5350025e-23], [1185.7687, 2.98655e-22, 6.37635e-23, 3.82581e-23, 3.5440614e-23], [1188.5022, 2.99279e-22, 6.38957e-23, 3.83374e-23, 3.553141e-23], [1191.242, 2.99905e-22, 6.40283e-23, 3.8417e-23, 3.5622409e-23], [1193.9881, 3.00532e-22, 6.41613e-23, 3.84968e-23, 3.571361e-23], [1196.7405, 3.01161e-22, 6.42948e-23, 3.85769e-23, 3.5805012e-23], [1199.4993, 3.01792e-22, 6.44286e-23, 3.86572e-23, 3.5896614e-23], [1202.2644, 3.02424e-22, 6.45628e-23, 3.87377e-23, 3.5988415e-23], [1205.0359, 3.03058e-22, 6.46974e-23, 3.88185e-23, 3.6080416e-23], [1207.8138, 3.03693e-22, 6.48324e-23, 3.88994e-23, 3.6172617e-23], [1210.5981, 3.0433e-22, 6.49677e-23, 3.89806e-23, 3.6265018e-23], [1213.3889, 3.04969e-22, 6.51034e-23, 3.90621e-23, 3.6357618e-23], [1216.186, 3.05609e-22, 6.52395e-23, 3.91437e-23, 3.6450418e-23], [1218.9896, 3.06251e-22, 6.53759e-23, 3.92255e-23, 3.6543418e-23], [1221.7997, 3.06894e-22, 6.55127e-23, 3.93076e-23, 3.6636619e-23], [1224.6162, 3.07539e-22, 6.56498e-23, 3.93899e-23, 3.6730021e-23], [1227.4392, 3.08186e-22, 6.57872e-23, 3.94723e-23, 3.6823625e-23], [1230.2688, 3.08835e-22, 6.59251e-23, 3.9555e-23, 3.691743e-23], [1233.1048, 3.09485e-22, 6.60633e-23, 3.9638e-23, 3.7011437e-23], [1235.9474, 3.10136e-22, 6.62018e-23, 3.97211e-23, 3.7105647e-23], [1238.7966, 3.10789e-22, 6.63407e-23, 3.98044e-23, 3.7200059e-23], [1241.6523, 3.11444e-22, 6.64799e-23, 3.9888e-23, 3.7294675e-23], [1244.5146, 3.12101e-22, 6.66195e-23, 3.99717e-23, 3.7389495e-23], [1247.3835, 3.12759e-22, 6.67595e-23, 4.00557e-23, 3.7484518e-23], [1250.259, 3.13419e-22, 6.68998e-23, 4.01399e-23, 3.7579747e-23], [1253.1412, 3.1408e-22, 6.70405e-23, 4.02243e-23, 3.767518e-23], [1256.03, 3.14744e-22, 6.71815e-23, 4.03089e-23, 3.7770819e-23], [1258.9254, 3.15408e-22, 6.73229e-23, 4.03937e-23, 3.7866665e-23], [1261.8275, 3.16075e-22, 6.74646e-23, 4.04788e-23, 3.7962716e-23], [1264.7363, 3.16743e-22, 6.76067e-23, 4.0564e-23, 3.8058975e-23], [1267.6519, 3.17413e-22, 6.77492e-23, 4.06495e-23, 3.8155441e-23], [1270.5741, 3.18084e-22, 6.7892e-23, 4.07352e-23, 3.8252114e-23], [1273.5031, 3.18758e-22, 6.80351e-23, 4.08211e-23, 3.8348997e-23], [1276.4388, 3.19432e-22, 6.81787e-23, 4.09072e-23, 3.8446088e-23], [1279.3813, 3.20109e-22, 6.83226e-23, 4.09935e-23, 3.8543388e-23], [1282.3306, 3.20787e-22, 6.84668e-23, 4.10801e-23, 3.8640898e-23], [1285.2867, 3.21467e-22, 6.86115e-23, 4.11669e-23, 3.8738618e-23], [1288.2496, 3.22149e-22, 6.87564e-23, 4.12539e-23, 3.8836549e-23], [1291.2193, 3.22832e-22, 6.89018e-23, 4.13411e-23, 3.8934691e-23], [1294.1958, 3.23517e-22, 6.90475e-23, 4.14285e-23, 3.9033045e-23], [1297.1793, 3.24204e-22, 6.91936e-23, 4.15161e-23, 3.9131611e-23], [1300.1696, 3.24892e-22, 6.934e-23, 4.1604e-23, 3.923039e-23], [1303.1668, 3.25582e-22, 6.94868e-23, 4.16921e-23, 3.9329382e-23], [1306.1709, 3.26274e-22, 6.9634e-23, 4.17804e-23, 3.9428588e-23], [1309.1819, 3.26968e-22, 6.97816e-23, 4.18689e-23, 3.9528007e-23], [1312.1999, 3.27663e-22, 6.99295e-23, 4.19577e-23, 3.9627642e-23], [1315.2248, 3.2836e-22, 7.00778e-23, 4.20467e-23, 3.9727491e-23], [1318.2567, 3.29059e-22, 7.02264e-23, 4.21359e-23, 3.9827556e-23], [1321.2956, 3.2976e-22, 7.03754e-23, 4.22253e-23, 3.9927837e-23], [1324.3415, 3.30462e-22, 7.05249e-23, 4.23149e-23, 4.0028335e-23], [1327.3945, 3.31166e-22, 7.06746e-23, 4.24048e-23, 4.012905e-23], [1330.4544, 3.31872e-22, 7.08248e-23, 4.24949e-23, 4.0229983e-23], [1333.5214, 3.32579e-22, 7.09753e-23, 4.25852e-23, 4.0331134e-23], [1336.5955, 3.33288e-22, 7.11262e-23, 4.26757e-23, 4.0432503e-23], [1339.6767, 3.33999e-22, 7.12775e-23, 4.27665e-23, 4.0534092e-23], [1342.765, 3.34712e-22, 7.14291e-23, 4.28575e-23, 4.06359e-23], [1345.8604, 3.35427e-22, 7.15812e-23, 4.29487e-23, 4.0737928e-23], [1348.9629, 3.36143e-22, 7.17336e-23, 4.30401e-23, 4.0840177e-23], [1352.0726, 3.36861e-22, 7.18864e-23, 4.31318e-23, 4.0942648e-23], [1355.1894, 3.37581e-22, 7.20395e-23, 4.32237e-23, 4.104534e-23], [1358.3134, 3.38302e-22, 7.21931e-23, 4.33158e-23, 4.1148254e-23], [1361.4447, 3.39026e-22, 7.2347e-23, 4.34082e-23, 4.1251391e-23], [1364.5831, 3.39751e-22, 7.25013e-23, 4.35008e-23, 4.1354751e-23], [1367.7288, 3.40478e-22, 7.2656e-23, 4.35936e-23, 4.1458335e-23], [1370.8818, 3.41207e-22, 7.28111e-23, 4.36867e-23, 4.1562143e-23], [1374.042, 3.41937e-22, 7.29666e-23, 4.37799e-23, 4.1666177e-23], [1377.2095, 3.4267e-22, 7.31224e-23, 4.38735e-23, 4.1770435e-23], [1380.3843, 3.43404e-22, 7.32787e-23, 4.39672e-23, 4.1874919e-23], [1383.5664, 3.4414e-22, 7.34353e-23, 4.40612e-23, 4.197963e-23], [1386.7558, 3.44878e-22, 7.35923e-23, 4.41554e-23, 4.2084568e-23], [1389.9526, 3.45618e-22, 7.37497e-23, 4.42498e-23, 4.2189733e-23], [1393.1568, 3.46359e-22, 7.39075e-23, 4.43445e-23, 4.2295126e-23], [1396.3684, 3.47102e-22, 7.40657e-23, 4.44394e-23, 4.2400748e-23], [1399.5873, 3.47848e-22, 7.42243e-23, 4.45346e-23, 4.2506599e-23], [1402.8137, 3.48595e-22, 7.43833e-23, 4.463e-23, 4.2612679e-23], [1406.0475, 3.49343e-22, 7.45427e-23, 4.47256e-23, 4.2718989e-23], [1409.2888, 3.50094e-22, 7.47024e-23, 4.48214e-23, 4.282553e-23], [1412.5375, 3.50847e-22, 7.48626e-23, 4.49175e-23, 4.2932303e-23], [1415.7938, 3.51601e-22, 7.50231e-23, 4.50139e-23, 4.3039307e-23], [1419.0575, 3.52357e-22, 7.51841e-23, 4.51104e-23, 4.3146543e-23], [1422.3288, 3.53115e-22, 7.53454e-23, 4.52073e-23, 4.3254012e-23], [1425.6076, 3.53875e-22, 7.55072e-23, 4.53043e-23, 4.3361715e-23], [1428.894, 3.54637e-22, 7.56693e-23, 4.54016e-23, 4.3469651e-23], [1432.1879, 3.55401e-22, 7.58319e-23, 4.54991e-23, 4.3577822e-23], [1435.4894, 3.56166e-22, 7.59948e-23, 4.55969e-23, 4.3686228e-23], [1438.7986, 3.56934e-22, 7.61582e-23, 4.56949e-23, 4.379487e-23], [1442.1154, 3.57703e-22, 7.63219e-23, 4.57932e-23, 4.3903747e-23], [1445.4398, 3.58474e-22, 7.64861e-23, 4.58917e-23, 4.4012862e-23], [1448.7719, 3.59247e-22, 7.66507e-23, 4.59904e-23, 4.4122213e-23], [1452.1116, 3.60022e-22, 7.68156e-23, 4.60894e-23, 4.4231802e-23], [1455.4591, 3.60799e-22, 7.6981e-23, 4.61886e-23, 4.434163e-23], [1458.8143, 3.61578e-22, 7.71468e-23, 4.62881e-23, 4.4451697e-23], [1462.1772, 3.62359e-22, 7.7313e-23, 4.63878e-23, 4.4562003e-23], [1465.5478, 3.63142e-22, 7.74796e-23, 4.64878e-23, 4.4672549e-23], [1468.9263, 3.63926e-22, 7.76466e-23, 4.6588e-23, 4.4783335e-23], [1472.3125, 3.64713e-22, 7.7814e-23, 4.66884e-23, 4.4894363e-23], [1475.7065, 3.65501e-22, 7.79819e-23, 4.67891e-23, 4.5005633e-23], [1479.1084, 3.66291e-22, 7.81501e-23, 4.68901e-23, 4.5117145e-23], [1482.5181, 3.67084e-22, 7.83188e-23, 4.69913e-23, 4.5228899e-23], [1485.9356, 3.67878e-22, 7.84879e-23, 4.70927e-23, 4.5340897e-23], [1489.3611, 3.68674e-22, 7.86574e-23, 4.71944e-23, 4.5453139e-23], [1492.7944, 3.69472e-22, 7.88273e-23, 4.72964e-23, 4.5565626e-23], [1496.2357, 3.70272e-22, 7.89976e-23, 4.73986e-23, 4.5678358e-23], [1499.6848, 3.71074e-22, 7.91684e-23, 4.7501e-23, 4.5791336e-23], [1503.142, 3.71878e-22, 7.93395e-23, 4.76037e-23, 4.590456e-23], [1506.6071, 3.72684e-22, 7.95111e-23, 4.77067e-23, 4.601803e-23], [1510.0802, 3.73492e-22, 7.96831e-23, 4.78099e-23, 4.6131749e-23], [1513.5612, 3.74302e-22, 7.98556e-23, 4.79133e-23, 4.6245715e-23], [1517.0504, 3.75114e-22, 8.00284e-23, 4.80171e-23, 4.6359931e-23], [1520.5475, 3.75928e-22, 8.02017e-23, 4.8121e-23, 4.6474396e-23], [1524.0528, 3.76744e-22, 8.03754e-23, 4.82253e-23, 4.6589111e-23], [1527.5661, 3.77562e-22, 8.05496e-23, 4.83298e-23, 4.6704078e-23], [1531.0875, 3.78382e-22, 8.07242e-23, 4.84345e-23, 4.6819298e-23], [1534.617, 3.79204e-22, 8.08994e-23, 4.85396e-23, 4.6934776e-23], [1538.1546, 3.80028e-22, 8.10753e-23, 4.86452e-23, 4.7050527e-23], [1541.7005, 3.80856e-22, 8.12539e-23, 4.87524e-23, 4.7166693e-23], [1545.2544, 3.81924e-22, 8.17115e-23, 4.90269e-23, 4.7302326e-23], [1548.8166, 3.82513e-22, 8.16056e-23, 4.89634e-23, 4.739931e-23], [1552.387, 3.83343e-22, 8.1781e-23, 4.90686e-23, 4.7515923e-23], [1555.9656, 3.84177e-22, 8.19583e-23, 4.9175e-23, 4.7632888e-23], [1559.5525, 3.85013e-22, 8.21361e-23, 4.92817e-23, 4.7750123e-23], [1563.1476, 3.85851e-22, 8.23145e-23, 4.93887e-23, 4.786762e-23], [1566.7511, 3.86691e-22, 8.24934e-23, 4.9496e-23, 4.7985375e-23], [1570.3628, 3.87533e-22, 8.26727e-23, 4.96036e-23, 4.8103389e-23], [1573.9829, 3.88377e-22, 8.28524e-23, 4.97114e-23, 4.8221662e-23], [1577.6113, 3.89223e-22, 8.30326e-23, 4.98196e-23, 4.8340194e-23], [1581.248, 3.90071e-22, 8.32132e-23, 4.99279e-23, 4.8458985e-23], [1584.8932, 3.90922e-22, 8.33943e-23, 5.00366e-23, 4.8578037e-23], [1588.5467, 3.91774e-22, 8.35758e-23, 5.01455e-23, 4.869735e-23], [1592.2087, 3.92629e-22, 8.37578e-23, 5.02547e-23, 4.8816924e-23], [1595.8791, 3.93485e-22, 8.39402e-23, 5.03641e-23, 4.893676e-23], [1599.558, 3.94344e-22, 8.41231e-23, 5.04738e-23, 4.9056859e-23], [1603.2454, 3.95205e-22, 8.43064e-23, 5.05838e-23, 4.9177221e-23], [1606.9413, 3.96067e-22, 8.44902e-23, 5.06941e-23, 4.9297848e-23], [1610.6456, 3.96932e-22, 8.46744e-23, 5.08046e-23, 4.941874e-23], [1614.3586, 3.97799e-22, 8.48591e-23, 5.09155e-23, 4.9539897e-23], [1618.08, 3.98668e-22, 8.50443e-23, 5.10266e-23, 4.9661321e-23], [1621.8101, 3.9954e-22, 8.52299e-23, 5.11379e-23, 4.9783012e-23], [1625.5488, 4.00413e-22, 8.5416e-23, 5.12496e-23, 4.9904972e-23], [1629.296, 4.01289e-22, 8.56026e-23, 5.13616e-23, 5.0027203e-23], [1633.0519, 4.02166e-22, 8.57897e-23, 5.14738e-23, 5.0149705e-23], [1636.8165, 4.03046e-22, 8.59773e-23, 5.15864e-23, 5.0272481e-23], [1640.5898, 4.03929e-22, 8.61656e-23, 5.16993e-23, 5.0395535e-23], [1644.3717, 4.04813e-22, 8.63545e-23, 5.18127e-23, 5.0518873e-23], [1648.1624, 4.057e-22, 8.65442e-23, 5.19265e-23, 5.0642507e-23], [1651.9618, 4.0659e-22, 8.6735e-23, 5.2041e-23, 5.0766458e-23], [1655.77, 4.07483e-22, 8.69278e-23, 5.21567e-23, 5.0890783e-23], [1659.5869, 4.08381e-22, 8.71251e-23, 5.22751e-23, 5.1015655e-23], [1663.4127, 4.09295e-22, 8.73381e-23, 5.24029e-23, 5.1141845e-23], [1667.2472, 4.10329e-22, 8.76897e-23, 5.26138e-23, 5.1277743e-23], [1671.0906, 4.13067e-22, 9.00111e-23, 5.40067e-23, 5.1550156e-23], [1674.9429, 4.12027e-22, 8.79527e-23, 5.27716e-23, 5.152019e-23], [1678.804, 4.12883e-22, 8.80902e-23, 5.28541e-23, 5.1642194e-23], [1682.6741, 4.13778e-22, 8.82706e-23, 5.29624e-23, 5.1767373e-23], [1686.553, 4.14682e-22, 8.84593e-23, 5.30756e-23, 5.1893358e-23], [1690.4409, 4.1559e-22, 8.86508e-23, 5.31905e-23, 5.2019788e-23], [1694.3378, 4.16501e-22, 8.88438e-23, 5.33063e-23, 5.2146564e-23], [1698.2437, 4.17415e-22, 8.90378e-23, 5.34227e-23, 5.2273652e-23], [1702.1585, 4.18331e-22, 8.92325e-23, 5.35395e-23, 5.2401038e-23], [1706.0824, 4.19249e-22, 8.94279e-23, 5.36567e-23, 5.2528715e-23], [1710.0153, 4.2017e-22, 8.96238e-23, 5.37743e-23, 5.265668e-23], [1713.9573, 4.21093e-22, 8.98202e-23, 5.38921e-23, 5.2784931e-23], [1717.9084, 4.22019e-22, 9.00172e-23, 5.40103e-23, 5.2913467e-23], [1721.8686, 4.22946e-22, 9.02147e-23, 5.41288e-23, 5.3042288e-23], [1725.8379, 4.23876e-22, 9.04127e-23, 5.42476e-23, 5.3171394e-23], [1729.8164, 4.24808e-22, 9.06111e-23, 5.43667e-23, 5.3300786e-23], [1733.804, 4.25742e-22, 9.08101e-23, 5.44861e-23, 5.3430463e-23], [1737.8008, 4.26679e-22, 9.10096e-23, 5.46057e-23, 5.3560427e-23], [1741.8069, 4.27618e-22, 9.12095e-23, 5.47257e-23, 5.3690678e-23], [1745.8222, 4.28559e-22, 9.141e-23, 5.4846e-23, 5.3821216e-23], [1749.8467, 4.29503e-22, 9.16109e-23, 5.49665e-23, 5.3952042e-23], [1753.8805, 4.30448e-22, 9.18123e-23, 5.50874e-23, 5.4083157e-23], [1757.9236, 4.31396e-22, 9.20142e-23, 5.52085e-23, 5.4214561e-23], [1761.976, 4.32347e-22, 9.22166e-23, 5.533e-23, 5.4346255e-23], [1766.0378, 4.33299e-22, 9.24195e-23, 5.54517e-23, 5.4478239e-23], [1770.109, 4.34254e-22, 9.26229e-23, 5.55737e-23, 5.4610515e-23], [1774.1895, 4.35211e-22, 9.28268e-23, 5.56961e-23, 5.4743082e-23], [1778.2794, 4.36171e-22, 9.30311e-23, 5.58187e-23, 5.4875942e-23], [1782.3788, 4.37132e-22, 9.3236e-23, 5.59416e-23, 5.5009095e-23], [1786.4876, 4.38096e-22, 9.34414e-23, 5.60648e-23, 5.5142542e-23], [1790.6059, 4.39063e-22, 9.36472e-23, 5.61883e-23, 5.5276284e-23], [1794.7336, 4.40032e-22, 9.38536e-23, 5.63121e-23, 5.5410321e-23], [1798.8709, 4.41003e-22, 9.40604e-23, 5.64362e-23, 5.5544654e-23], [1803.0177, 4.41976e-22, 9.42678e-23, 5.65607e-23, 5.5679284e-23], [1807.1741, 4.42952e-22, 9.44756e-23, 5.66854e-23, 5.5814211e-23], [1811.3401, 4.4393e-22, 9.4684e-23, 5.68104e-23, 5.5949436e-23], [1815.5157, 4.4491e-22, 9.48928e-23, 5.69357e-23, 5.6084959e-23], [1819.7009, 4.45893e-22, 9.51022e-23, 5.70613e-23, 5.6220783e-23], [1823.8957, 4.46878e-22, 9.5312e-23, 5.71872e-23, 5.6356906e-23], [1828.1002, 4.47866e-22, 9.55224e-23, 5.73134e-23, 5.649333e-23], [1832.3144, 4.48856e-22, 9.57333e-23, 5.744e-23, 5.6630056e-23], [1836.5383, 4.49848e-22, 9.59446e-23, 5.75668e-23, 5.6767084e-23], [1840.772, 4.50843e-22, 9.61565e-23, 5.76939e-23, 5.6904415e-23], [1845.0154, 4.5184e-22, 9.63689e-23, 5.78214e-23, 5.704205e-23], [1849.2686, 4.52839e-22, 9.65818e-23, 5.79491e-23, 5.7179989e-23], [1853.5316, 4.53841e-22, 9.67953e-23, 5.80772e-23, 5.7318234e-23], [1857.8045, 4.54845e-22, 9.70092e-23, 5.82055e-23, 5.7456784e-23], [1862.0871, 4.55852e-22, 9.72236e-23, 5.83342e-23, 5.7595641e-23], [1866.3797, 4.56861e-22, 9.74386e-23, 5.84632e-23, 5.7734805e-23], [1870.6821, 4.57872e-22, 9.76541e-23, 5.85925e-23, 5.7874277e-23], [1874.9945, 4.58886e-22, 9.78701e-23, 5.8722e-23, 5.8014058e-23], [1879.3168, 4.59903e-22, 9.80866e-23, 5.8852e-23, 5.8154148e-23], [1883.6491, 4.60921e-22, 9.83036e-23, 5.89822e-23, 5.8294548e-23], [1887.9913, 4.61942e-22, 9.85212e-23, 5.91127e-23, 5.843526e-23], [1892.3436, 4.62966e-22, 9.87393e-23, 5.92436e-23, 5.8576283e-23], [1896.7059, 4.63992e-22, 9.89579e-23, 5.93747e-23, 5.8717619e-23], [1901.0783, 4.65021e-22, 9.9177e-23, 5.95062e-23, 5.8859267e-23], [1905.4607, 4.66052e-22, 9.93967e-23, 5.9638e-23, 5.900123e-23], [1909.8533, 4.67085e-22, 9.96168e-23, 5.97701e-23, 5.9143507e-23], [1914.2559, 4.68121e-22, 9.98376e-23, 5.99025e-23, 5.92861e-23], [1918.6687, 4.6916e-22, 1.00059e-22, 6.00353e-23, 5.9429009e-23], [1923.0917, 4.70201e-22, 1.00281e-22, 6.01683e-23, 5.9572234e-23], [1927.5249, 4.71244e-22, 1.00503e-22, 6.03017e-23, 5.9715778e-23], [1931.9683, 4.7229e-22, 1.00726e-22, 6.04354e-23, 5.9859639e-23], [1936.422, 4.73338e-22, 1.00949e-22, 6.05694e-23, 6.000382e-23], [1940.8859, 4.74389e-22, 1.01173e-22, 6.07038e-23, 6.014832e-23], [1945.3601, 4.75443e-22, 1.01397e-22, 6.08384e-23, 6.0293142e-23], [1949.8446, 4.76499e-22, 1.01622e-22, 6.09734e-23, 6.0438284e-23], [1954.3395, 4.77557e-22, 1.01848e-22, 6.11087e-23, 6.0583749e-23], [1958.8447, 4.78618e-22, 1.02074e-22, 6.12444e-23, 6.0729536e-23], [1963.3603, 4.79682e-22, 1.02301e-22, 6.13803e-23, 6.0875648e-23], [1967.8863, 4.80748e-22, 1.02528e-22, 6.15166e-23, 6.1022083e-23], [1972.4227, 4.81816e-22, 1.02755e-22, 6.16532e-23, 6.1168844e-23], [1976.9696, 4.82887e-22, 1.02984e-22, 6.17902e-23, 6.1315931e-23], [1981.527, 4.83961e-22, 1.03212e-22, 6.19274e-23, 6.1463344e-23], [1986.0949, 4.85038e-22, 1.03442e-22, 6.2065e-23, 6.1611085e-23], [1990.6733, 4.86116e-22, 1.03672e-22, 6.22029e-23, 6.1759155e-23], [1995.2623, 4.87198e-22, 1.03902e-22, 6.23412e-23, 6.1907553e-23], [1999.8619, 4.88282e-22, 1.04133e-22, 6.24798e-23, 6.2056281e-23], [2004.472, 4.89368e-22, 1.04364e-22, 6.26187e-23, 6.220534e-23], [2009.0928, 4.90458e-22, 1.04597e-22, 6.27579e-23, 6.235473e-23], [2013.7242, 4.91549e-22, 1.04829e-22, 6.28975e-23, 6.2504452e-23], [2018.3664, 4.92644e-22, 1.05062e-22, 6.30374e-23, 6.2654507e-23], [2023.0192, 4.93741e-22, 1.05296e-22, 6.31777e-23, 6.2804896e-23], [2027.6827, 4.9484e-22, 1.0553e-22, 6.33182e-23, 6.295562e-23], [2032.357, 4.95943e-22, 1.05765e-22, 6.34592e-23, 6.3106679e-23], [2037.0421, 4.97048e-22, 1.06001e-22, 6.36004e-23, 6.3258073e-23], [2041.7379, 4.98155e-22, 1.06237e-22, 6.3742e-23, 6.3409805e-23], [2046.4446, 4.99265e-22, 1.06473e-22, 6.38839e-23, 6.3561875e-23], [2051.1622, 5.00378e-22, 1.0671e-22, 6.40262e-23, 6.3714282e-23], [2055.8906, 5.01493e-22, 1.06948e-22, 6.41688e-23, 6.386703e-23], [2060.6299, 5.02612e-22, 1.07186e-22, 6.43117e-23, 6.4020117e-23], [2065.3802, 5.03732e-22, 1.07425e-22, 6.4455e-23, 6.4173545e-23], [2070.1413, 5.04856e-22, 1.07664e-22, 6.45987e-23, 6.4327315e-23], [2074.9135, 5.05982e-22, 1.07904e-22, 6.47426e-23, 6.4481427e-23], [2079.6967, 5.07111e-22, 1.08145e-22, 6.48869e-23, 6.4635883e-23], [2084.4909, 5.08242e-22, 1.08386e-22, 6.50316e-23, 6.4790682e-23], [2089.2961, 5.09376e-22, 1.08628e-22, 6.51766e-23, 6.4945827e-23], [2094.1125, 5.10513e-22, 1.0887e-22, 6.53219e-23, 6.5101317e-23], [2098.9399, 5.11653e-22, 1.09113e-22, 6.54676e-23, 6.5257154e-23], [2103.7784, 5.12795e-22, 1.09356e-22, 6.56137e-23, 6.5413339e-23], [2108.6281, 5.1394e-22, 1.096e-22, 6.57601e-23, 6.5569871e-23], [2113.489, 5.15088e-22, 1.09845e-22, 6.59068e-23, 6.5726753e-23], [2118.3611, 5.16238e-22, 1.1009e-22, 6.60539e-23, 6.5883984e-23], [2123.2445, 5.17391e-22, 1.10336e-22, 6.62013e-23, 6.6041566e-23], [2128.139, 5.18547e-22, 1.10582e-22, 6.63491e-23, 6.6199499e-23], [2133.0449, 5.19706e-22, 1.10829e-22, 6.64973e-23, 6.6357785e-23], [2137.9621, 5.20867e-22, 1.11076e-22, 6.66458e-23, 6.6516424e-23], [2142.8906, 5.22031e-22, 1.11324e-22, 6.67946e-23, 6.6675417e-23], [2147.8305, 5.23198e-22, 1.11573e-22, 6.69438e-23, 6.6834765e-23], [2152.7817, 5.24368e-22, 1.11822e-22, 6.70934e-23, 6.6994468e-23], [2157.7444, 5.25541e-22, 1.12072e-22, 6.72433e-23, 6.7154528e-23], [2162.7185, 5.26716e-22, 1.12323e-22, 6.73936e-23, 6.7314945e-23], [2167.7041, 5.27894e-22, 1.12574e-22, 6.75442e-23, 6.7475721e-23], [2172.7012, 5.29075e-22, 1.12825e-22, 6.76952e-23, 6.7636855e-23], [2177.7098, 5.30258e-22, 1.13078e-22, 6.78465e-23, 6.779835e-23], [2182.7299, 5.31445e-22, 1.1333e-22, 6.79982e-23, 6.7960205e-23], [2187.7616, 5.32634e-22, 1.13584e-22, 6.81503e-23, 6.8122422e-23], [2192.8049, 5.33826e-22, 1.13838e-22, 6.83027e-23, 6.8285001e-23], [2197.8599, 5.35021e-22, 1.14093e-22, 6.84555e-23, 6.8447943e-23], [2202.9265, 5.36219e-22, 1.14348e-22, 6.86087e-23, 6.861125e-23], [2208.0047, 5.3742e-22, 1.14604e-22, 6.87622e-23, 6.8774922e-23], [2213.0947, 5.38623e-22, 1.1486e-22, 6.89161e-23, 6.893896e-23], [2218.1964, 5.3983e-22, 1.15117e-22, 6.90703e-23, 6.9103365e-23], [2223.3099, 5.41039e-22, 1.15375e-22, 6.92249e-23, 6.9268137e-23], [2228.4351, 5.42251e-22, 1.15633e-22, 6.93799e-23, 6.9433279e-23], [2233.5722, 5.43466e-22, 1.15892e-22, 6.95353e-23, 6.959879e-23], [2238.7211, 5.44684e-22, 1.16152e-22, 6.9691e-23, 6.9764673e-23], [2243.8819, 5.45904e-22, 1.16412e-22, 6.98472e-23, 6.9930927e-23], [2249.0546, 5.47128e-22, 1.16673e-22, 7.00037e-23, 7.0097556e-23], [2254.2392, 5.48355e-22, 1.16934e-22, 7.01606e-23, 7.026456e-23], [2259.4358, 5.49584e-22, 1.17197e-22, 7.03179e-23, 7.0431943e-23], [2264.6443, 5.50817e-22, 1.1746e-22, 7.04757e-23, 7.0599711e-23], [2269.8649, 5.52053e-22, 1.17723e-22, 7.06341e-23, 7.0767878e-23], [2275.0974, 5.53292e-22, 1.17989e-22, 7.07934e-23, 7.0936483e-23], [2280.3421, 5.54538e-22, 1.18259e-22, 7.09553e-23, 7.1105713e-23], [2285.5988, 5.55819e-22, 1.18567e-22, 7.11403e-23, 7.1277833e-23], [2290.8677, 5.57272e-22, 1.19074e-22, 7.14444e-23, 7.1463521e-23], [2296.1486, 5.58284e-22, 1.19062e-22, 7.14373e-23, 7.161501e-23], [2301.4418, 5.59528e-22, 1.19319e-22, 7.15913e-23, 7.178471e-23], [2306.7472, 5.60782e-22, 1.19584e-22, 7.17504e-23, 7.1955316e-23], [2312.0648, 5.6204e-22, 1.19851e-22, 7.19108e-23, 7.2126406e-23], [2317.3946, 5.63301e-22, 1.2012e-22, 7.20719e-23, 7.2297912e-23], [2322.7368, 5.64566e-22, 1.20389e-22, 7.22334e-23, 7.2469816e-23], [2328.0913, 5.65833e-22, 1.20659e-22, 7.23954e-23, 7.2642113e-23], [2333.4581, 5.67104e-22, 1.2093e-22, 7.25579e-23, 7.2814798e-23], [2338.8372, 5.68378e-22, 1.21201e-22, 7.27207e-23, 7.2987874e-23], [2344.2288, 5.69654e-22, 1.21473e-22, 7.28839e-23, 7.3161338e-23], [2349.6328, 5.70934e-22, 1.21746e-22, 7.30476e-23, 7.3335192e-23], [2355.0493, 5.72217e-22, 1.22019e-22, 7.32116e-23, 7.3509437e-23], [2360.4782, 5.73503e-22, 1.22293e-22, 7.3376e-23, 7.3684073e-23], [2365.9197, 5.74791e-22, 1.22568e-22, 7.35408e-23, 7.3859101e-23], [2371.3737, 5.76083e-22, 1.22843e-22, 7.3706e-23, 7.4034521e-23], [2376.8403, 5.77378e-22, 1.23119e-22, 7.38716e-23, 7.4210336e-23], [2382.3195, 5.78676e-22, 1.23396e-22, 7.40375e-23, 7.4386544e-23], [2387.8113, 5.79978e-22, 1.23673e-22, 7.42039e-23, 7.4563149e-23], [2393.3158, 5.81282e-22, 1.23951e-22, 7.43707e-23, 7.474015e-23], [2398.8329, 5.82589e-22, 1.2423e-22, 7.45378e-23, 7.4917548e-23], [2404.3628, 5.83899e-22, 1.24509e-22, 7.47054e-23, 7.5095344e-23], [2409.9054, 5.85213e-22, 1.24789e-22, 7.48734e-23, 7.5273539e-23], [2415.4608, 5.86529e-22, 1.2507e-22, 7.50417e-23, 7.5452134e-23], [2421.029, 5.87849e-22, 1.25351e-22, 7.52105e-23, 7.5631131e-23], [2426.6101, 5.89172e-22, 1.25633e-22, 7.53796e-23, 7.5810529e-23], [2432.204, 5.90498e-22, 1.25915e-22, 7.55492e-23, 7.599033e-23], [2437.8108, 5.91827e-22, 1.26199e-22, 7.57191e-23, 7.6170535e-23], [2443.4306, 5.93159e-22, 1.26482e-22, 7.58895e-23, 7.6351145e-23], [2449.0632, 5.94494e-22, 1.26767e-22, 7.60602e-23, 7.6532161e-23], [2454.7089, 5.95833e-22, 1.27052e-22, 7.62314e-23, 7.6713583e-23], [2460.3676, 5.97174e-22, 1.27338e-22, 7.64029e-23, 7.6895412e-23], [2466.0393, 5.98519e-22, 1.27625e-22, 7.65749e-23, 7.7077651e-23], [2471.7241, 5.99867e-22, 1.27912e-22, 7.67473e-23, 7.7260299e-23], [2477.4221, 6.01218e-22, 1.282e-22, 7.692e-23, 7.7443357e-23], [2483.1331, 6.02572e-22, 1.28489e-22, 7.70932e-23, 7.7626827e-23], [2488.8573, 6.0393e-22, 1.28778e-22, 7.72668e-23, 7.7810709e-23], [2494.5947, 6.0529e-22, 1.29068e-22, 7.74408e-23, 7.7995005e-23], [2500.3454, 6.06654e-22, 1.29359e-22, 7.76152e-23, 7.8179715e-23], [2506.1093, 6.08022e-22, 1.2965e-22, 7.77901e-23, 7.836484e-23], [2511.8864, 6.09392e-22, 1.29942e-22, 7.79653e-23, 7.8550382e-23], [2517.6769, 6.10765e-22, 1.30235e-22, 7.8141e-23, 7.8736341e-23], [2523.4808, 6.12142e-22, 1.30528e-22, 7.8317e-23, 7.8922718e-23], [2529.298, 6.13522e-22, 1.30823e-22, 7.84935e-23, 7.9109515e-23], [2535.1286, 6.14906e-22, 1.31117e-22, 7.86704e-23, 7.9296731e-23], [2540.9727, 6.16292e-22, 1.31413e-22, 7.88477e-23, 7.9484369e-23], [2546.8303, 6.17682e-22, 1.31709e-22, 7.90255e-23, 7.9672429e-23], [2552.7013, 6.19075e-22, 1.32006e-22, 7.92036e-23, 7.9860913e-23], [2558.5859, 6.20472e-22, 1.32304e-22, 7.93822e-23, 8.004982e-23], [2564.484, 6.21871e-22, 1.32602e-22, 7.95612e-23, 8.0239153e-23], [2570.3958, 6.23274e-22, 1.32901e-22, 7.97406e-23, 8.0428912e-23], [2576.3212, 6.24681e-22, 1.33201e-22, 7.99204e-23, 8.0619098e-23], [2582.2602, 6.2609e-22, 1.33501e-22, 8.01007e-23, 8.0809713e-23], [2588.2129, 6.27503e-22, 1.33802e-22, 8.02814e-23, 8.1000757e-23], [2594.1794, 6.28919e-22, 1.34104e-22, 8.04625e-23, 8.119223e-23], [2600.1596, 6.30339e-22, 1.34407e-22, 8.06441e-23, 8.1384135e-23], [2606.1535, 6.31762e-22, 1.3471e-22, 8.0826e-23, 8.1576473e-23], [2612.1614, 6.33188e-22, 1.35014e-22, 8.10084e-23, 8.1769243e-23], [2618.183, 6.34618e-22, 1.35319e-22, 8.11912e-23, 8.1962448e-23], [2624.2185, 6.36051e-22, 1.35624e-22, 8.13745e-23, 8.2156088e-23], [2630.268, 6.37487e-22, 1.3593e-22, 8.15582e-23, 8.2350165e-23], [2636.3314, 6.38927e-22, 1.36237e-22, 8.17423e-23, 8.2544678e-23], [2642.4088, 6.4037e-22, 1.36545e-22, 8.19269e-23, 8.273963e-23], [2648.5001, 6.41816e-22, 1.36853e-22, 8.21118e-23, 8.2935022e-23], [2654.6056, 6.43266e-22, 1.37162e-22, 8.22973e-23, 8.3130854e-23], [2660.7251, 6.4472e-22, 1.37472e-22, 8.24831e-23, 8.3327127e-23], [2666.8587, 6.46176e-22, 1.37782e-22, 8.26694e-23, 8.3523843e-23], [2673.0064, 6.47636e-22, 1.38094e-22, 8.28562e-23, 8.3721003e-23], [2679.1683, 6.491e-22, 1.38406e-22, 8.30433e-23, 8.3918607e-23], [2685.3444, 6.50567e-22, 1.38718e-22, 8.3231e-23, 8.4116656e-23], [2691.5348, 6.52038e-22, 1.39032e-22, 8.3419e-23, 8.4315152e-23], [2697.7394, 6.53511e-22, 1.39346e-22, 8.36075e-23, 8.4514096e-23], [2703.9584, 6.54989e-22, 1.39661e-22, 8.37965e-23, 8.4713489e-23], [2710.1916, 6.5647e-22, 1.39976e-22, 8.39858e-23, 8.4913332e-23], [2716.4393, 6.57954e-22, 1.40293e-22, 8.41757e-23, 8.5113625e-23], [2722.7013, 6.59442e-22, 1.4061e-22, 8.4366e-23, 8.531437e-23], [2728.9778, 6.60933e-22, 1.40928e-22, 8.45567e-23, 8.5515569e-23], [2735.2687, 6.62428e-22, 1.41246e-22, 8.47479e-23, 8.5717221e-23], [2741.5742, 6.63927e-22, 1.41566e-22, 8.49395e-23, 8.5919328e-23], [2747.8942, 6.65428e-22, 1.41886e-22, 8.51315e-23, 8.6121892e-23], [2754.2287, 6.66934e-22, 1.42207e-22, 8.53241e-23, 8.6324913e-23], [2760.5779, 6.68443e-22, 1.42528e-22, 8.55171e-23, 8.6528392e-23], [2766.9416, 6.69955e-22, 1.42851e-22, 8.57105e-23, 8.6732331e-23], [2773.3201, 6.71471e-22, 1.43174e-22, 8.59044e-23, 8.693673e-23], [2779.7133, 6.72991e-22, 1.43498e-22, 8.60987e-23, 8.714159e-23], [2786.1212, 6.74514e-22, 1.43823e-22, 8.62935e-23, 8.7346914e-23], [2792.5438, 6.76041e-22, 1.44148e-22, 8.64888e-23, 8.7552701e-23], [2798.9813, 6.77571e-22, 1.44474e-22, 8.66845e-23, 8.7758952e-23], [2805.4336, 6.79105e-22, 1.44801e-22, 8.68807e-23, 8.796567e-23], [2811.9008, 6.80642e-22, 1.45129e-22, 8.70773e-23, 8.8172855e-23], [2818.3829, 6.82184e-22, 1.45457e-22, 8.72744e-23, 8.8380508e-23], [2824.88, 6.83728e-22, 1.45787e-22, 8.74719e-23, 8.8588629e-23], [2831.392, 6.85277e-22, 1.46117e-22, 8.767e-23, 8.8797222e-23], [2837.919, 6.86829e-22, 1.46447e-22, 8.78685e-23, 8.9006285e-23], [2844.4611, 6.88384e-22, 1.46779e-22, 8.80674e-23, 8.9215821e-23], [2851.0183, 6.89943e-22, 1.47111e-22, 8.82668e-23, 8.9425831e-23], [2857.5905, 6.91506e-22, 1.47445e-22, 8.84667e-23, 8.9636315e-23], [2864.178, 6.93073e-22, 1.47778e-22, 8.86671e-23, 8.9847276e-23], [2870.7806, 6.94643e-22, 1.48113e-22, 8.88679e-23, 9.0058713e-23], [2877.3984, 6.96217e-22, 1.48449e-22, 8.90692e-23, 9.0270628e-23], [2884.0315, 6.97795e-22, 1.48785e-22, 8.9271e-23, 9.0483022e-23], [2890.6799, 6.99376e-22, 1.49122e-22, 8.94732e-23, 9.0695897e-23], [2897.3436, 7.00961e-22, 1.4946e-22, 8.96759e-23, 9.0909253e-23], [2904.0227, 7.0255e-22, 1.49799e-22, 8.98791e-23, 9.1123092e-23], [2910.7171, 7.04143e-22, 1.50138e-22, 9.00828e-23, 9.1337414e-23], [2917.427, 7.05739e-22, 1.50478e-22, 9.0287e-23, 9.1552222e-23], [2924.1524, 7.07339e-22, 1.50819e-22, 9.04916e-23, 9.1767515e-23], [2930.8932, 7.08942e-22, 1.51161e-22, 9.06967e-23, 9.1983296e-23], [2937.6497, 7.1055e-22, 1.51504e-22, 9.09023e-23, 9.2199564e-23], [2944.4216, 7.12161e-22, 1.51847e-22, 9.11084e-23, 9.2416323e-23], [2951.2092, 7.13776e-22, 1.52192e-22, 9.13149e-23, 9.2633572e-23], [2958.0125, 7.15395e-22, 1.52537e-22, 9.1522e-23, 9.2851314e-23], [2964.8314, 7.17018e-22, 1.52883e-22, 9.17295e-23, 9.3069549e-23], [2971.666, 7.18644e-22, 1.53229e-22, 9.19376e-23, 9.3288279e-23], [2978.5164, 7.20274e-22, 1.53577e-22, 9.21461e-23, 9.3507508e-23], [2985.3826, 7.21909e-22, 1.53925e-22, 9.23553e-23, 9.372724e-23], [2992.2646, 7.23547e-22, 1.54275e-22, 9.2565e-23, 9.3947488e-23], [2999.1625, 7.2519e-22, 1.54627e-22, 9.27761e-23, 9.4168317e-23], [3006.0763, 7.26852e-22, 1.54996e-22, 9.29978e-23, 9.4390767e-23], [3013.006, 7.28512e-22, 1.55361e-22, 9.32165e-23, 9.461333e-23], [3019.9517, 7.30139e-22, 1.55682e-22, 9.34091e-23, 9.4833488e-23], [3026.9134, 7.31794e-22, 1.56033e-22, 9.36198e-23, 9.505607e-23], [3033.8912, 7.33454e-22, 1.56387e-22, 9.38319e-23, 9.5279258e-23], [3040.885, 7.35119e-22, 1.56741e-22, 9.40447e-23, 9.5502974e-23], [3047.895, 7.36788e-22, 1.57097e-22, 9.42581e-23, 9.5727203e-23], [3054.9211, 7.3846e-22, 1.57453e-22, 9.4472e-23, 9.5951944e-23], [3061.9634, 7.40137e-22, 1.57811e-22, 9.46864e-23, 9.6177196e-23], [3069.022, 7.41817e-22, 1.58169e-22, 9.49013e-23, 9.6402958e-23], [3076.0968, 7.43502e-22, 1.58528e-22, 9.51167e-23, 9.6629233e-23], [3083.188, 7.4519e-22, 1.58888e-22, 9.53327e-23, 9.6856021e-23], [3090.2954, 7.46883e-22, 1.59249e-22, 9.55491e-23, 9.7083323e-23], [3097.4193, 7.48579e-22, 1.5961e-22, 9.57661e-23, 9.7311141e-23], [3104.5596, 7.50279e-22, 1.59973e-22, 9.59835e-23, 9.7539475e-23], [3111.7163, 7.51983e-22, 1.60336e-22, 9.62015e-23, 9.7768327e-23], [3118.8896, 7.53692e-22, 1.607e-22, 9.642e-23, 9.7997697e-23], [3126.0794, 7.55404e-22, 1.61065e-22, 9.6639e-23, 9.8227588e-23], [3133.2857, 7.5712e-22, 1.61431e-22, 9.68585e-23, 9.8458e-23], [3140.5087, 7.5884e-22, 1.61797e-22, 9.70785e-23, 9.8688934e-23], [3147.7483, 7.60565e-22, 1.62165e-22, 9.7299e-23, 9.8920392e-23], [3155.0046, 7.62293e-22, 1.62533e-22, 9.75201e-23, 9.9152375e-23], [3162.2777, 7.64025e-22, 1.62903e-22, 9.77416e-23, 9.9384884e-23], [3169.5675, 7.65762e-22, 1.63273e-22, 9.79637e-23, 9.961792e-23], [3176.8741, 7.67502e-22, 1.63644e-22, 9.81863e-23, 9.9851485e-23], [3184.1975, 7.69246e-22, 1.64016e-22, 9.84095e-23, 1.0008558e-22], [3191.5379, 7.70995e-22, 1.64389e-22, 9.86331e-23, 1.0032021e-22], [3198.8951, 7.72748e-22, 1.64762e-22, 9.88573e-23, 1.0055536e-22], [3206.2693, 7.74505e-22, 1.65137e-22, 9.9082e-23, 1.0079106e-22], [3213.6605, 7.76265e-22, 1.65512e-22, 9.93072e-23, 1.0102728e-22], [3221.0688, 7.7803e-22, 1.65888e-22, 9.95329e-23, 1.0126405e-22], [3228.4941, 7.798e-22, 1.66265e-22, 9.97592e-23, 1.0150135e-22], [3235.9366, 7.81573e-22, 1.66643e-22, 9.9986e-23, 1.0173918e-22], [3243.3962, 7.8335e-22, 1.67022e-22, 1.00213e-22, 1.0197756e-22], [3250.873, 7.85132e-22, 1.67402e-22, 1.00441e-22, 1.0221648e-22], [3258.367, 7.86917e-22, 1.67783e-22, 1.0067e-22, 1.0245594e-22], [3265.8783, 7.88707e-22, 1.68164e-22, 1.00899e-22, 1.0269595e-22], [3273.4069, 7.90501e-22, 1.68547e-22, 1.01128e-22, 1.029365e-22], [3280.9529, 7.923e-22, 1.6893e-22, 1.01358e-22, 1.0317759e-22], [3288.5163, 7.94102e-22, 1.69314e-22, 1.01589e-22, 1.0341924e-22], [3296.0971, 7.95909e-22, 1.69699e-22, 1.0182e-22, 1.0366143e-22], [3303.6954, 7.9772e-22, 1.70085e-22, 1.02051e-22, 1.0390417e-22], [3311.3112, 7.99535e-22, 1.70472e-22, 1.02283e-22, 1.0414747e-22], [3318.9446, 8.01354e-22, 1.7086e-22, 1.02516e-22, 1.0439131e-22], [3326.5955, 8.03178e-22, 1.71249e-22, 1.02749e-22, 1.0463571e-22], [3334.2641, 8.05006e-22, 1.71639e-22, 1.02983e-22, 1.0488067e-22], [3341.9504, 8.06838e-22, 1.72029e-22, 1.03217e-22, 1.0512618e-22], [3349.6544, 8.08674e-22, 1.72421e-22, 1.03452e-22, 1.0537225e-22], [3357.3761, 8.10515e-22, 1.72813e-22, 1.03688e-22, 1.0561888e-22], [3365.1157, 8.1236e-22, 1.73206e-22, 1.03924e-22, 1.0586607e-22], [3372.8731, 8.14209e-22, 1.736e-22, 1.0416e-22, 1.0611382e-22], [3380.6484, 8.16063e-22, 1.73996e-22, 1.04397e-22, 1.0636213e-22], [3388.4416, 8.17921e-22, 1.74392e-22, 1.04635e-22, 1.0661101e-22], [3396.2527, 8.19783e-22, 1.74789e-22, 1.04873e-22, 1.0686045e-22], [3404.0819, 8.2165e-22, 1.75187e-22, 1.05112e-22, 1.0711046e-22], [3411.9291, 8.23521e-22, 1.75585e-22, 1.05351e-22, 1.0736104e-22], [3419.7944, 8.25396e-22, 1.75985e-22, 1.05591e-22, 1.0761219e-22], [3427.6779, 8.27276e-22, 1.76386e-22, 1.05832e-22, 1.0786391e-22], [3435.5795, 8.2916e-22, 1.76788e-22, 1.06073e-22, 1.081162e-22], [3443.4993, 8.31048e-22, 1.7719e-22, 1.06314e-22, 1.0836906e-22], [3451.4374, 8.32941e-22, 1.77594e-22, 1.06556e-22, 1.086225e-22], [3459.3938, 8.34839e-22, 1.77998e-22, 1.06799e-22, 1.0887652e-22], [3467.3685, 8.3674e-22, 1.78404e-22, 1.07042e-22, 1.0913111e-22], [3475.3616, 8.38646e-22, 1.7881e-22, 1.07286e-22, 1.0938628e-22], [3483.3732, 8.40557e-22, 1.79217e-22, 1.0753e-22, 1.0964204e-22], [3491.4032, 8.42472e-22, 1.79625e-22, 1.07775e-22, 1.0989837e-22], [3499.4517, 8.44392e-22, 1.80035e-22, 1.08021e-22, 1.1015529e-22], [3507.5187, 8.46316e-22, 1.80445e-22, 1.08267e-22, 1.1041279e-22], [3515.6044, 8.48244e-22, 1.80856e-22, 1.08514e-22, 1.1067088e-22], [3523.7087, 8.50177e-22, 1.81268e-22, 1.08761e-22, 1.1092955e-22], [3531.8317, 8.52115e-22, 1.81681e-22, 1.09009e-22, 1.1118881e-22], [3539.9734, 8.54057e-22, 1.82095e-22, 1.09257e-22, 1.1144866e-22], [3548.1339, 8.56003e-22, 1.8251e-22, 1.09506e-22, 1.1170911e-22], [3556.3132, 8.57954e-22, 1.82926e-22, 1.09756e-22, 1.1197014e-22], [3564.5113, 8.5991e-22, 1.83343e-22, 1.10006e-22, 1.1223177e-22], [3572.7284, 8.6187e-22, 1.83761e-22, 1.10256e-22, 1.12494e-22], [3580.9644, 8.63835e-22, 1.8418e-22, 1.10508e-22, 1.1275682e-22], [3589.2193, 8.65804e-22, 1.84599e-22, 1.1076e-22, 1.1302024e-22], [3597.4934, 8.67778e-22, 1.8502e-22, 1.11012e-22, 1.1328426e-22], [3605.7864, 8.69757e-22, 1.85442e-22, 1.11265e-22, 1.1354888e-22], [3614.0986, 8.7174e-22, 1.85865e-22, 1.11519e-22, 1.1381411e-22], [3622.43, 8.73728e-22, 1.86288e-22, 1.11773e-22, 1.1407993e-22], [3630.7805, 8.7572e-22, 1.86713e-22, 1.12028e-22, 1.1434637e-22], [3639.1504, 8.77717e-22, 1.87139e-22, 1.12283e-22, 1.146134e-22], [3647.5395, 8.79719e-22, 1.87566e-22, 1.12539e-22, 1.1488105e-22], [3655.9479, 8.81725e-22, 1.87993e-22, 1.12796e-22, 1.1514931e-22], [3664.3757, 8.83737e-22, 1.88422e-22, 1.13053e-22, 1.1541817e-22], [3672.823, 8.85752e-22, 1.88852e-22, 1.13311e-22, 1.1568765e-22], [3681.2897, 8.87773e-22, 1.89283e-22, 1.1357e-22, 1.1595775e-22], [3689.776, 8.89798e-22, 1.89714e-22, 1.13829e-22, 1.1622845e-22], [3698.2818, 8.91828e-22, 1.90147e-22, 1.14088e-22, 1.1649978e-22], [3706.8072, 8.93862e-22, 1.90581e-22, 1.14348e-22, 1.1677172e-22], [3715.3523, 8.95902e-22, 1.91015e-22, 1.14609e-22, 1.1704428e-22], [3723.9171, 8.97946e-22, 1.91451e-22, 1.14871e-22, 1.1731746e-22], [3732.5016, 8.99995e-22, 1.91888e-22, 1.15133e-22, 1.1759127e-22], [3741.1059, 9.02048e-22, 1.92326e-22, 1.15395e-22, 1.178657e-22], [3749.73, 9.04106e-22, 1.92765e-22, 1.15659e-22, 1.1814075e-22], [3758.374, 9.0617e-22, 1.93204e-22, 1.15923e-22, 1.1841643e-22], [3767.038, 9.08238e-22, 1.93645e-22, 1.16187e-22, 1.1869274e-22], [3775.7219, 9.1031e-22, 1.94087e-22, 1.16452e-22, 1.1896968e-22], [3784.4258, 9.12388e-22, 1.9453e-22, 1.16718e-22, 1.1924725e-22], [3793.1498, 9.14471e-22, 1.94974e-22, 1.16984e-22, 1.1952545e-22], [3801.894, 9.16558e-22, 1.95419e-22, 1.17251e-22, 1.1980429e-22], [3810.6582, 9.1865e-22, 1.95865e-22, 1.17519e-22, 1.2008377e-22], [3819.4427, 9.20748e-22, 1.96313e-22, 1.17788e-22, 1.2036392e-22], [3828.2474, 9.22856e-22, 1.96769e-22, 1.18061e-22, 1.2064518e-22], [3837.0725, 9.24962e-22, 1.97217e-22, 1.1833e-22, 1.2092649e-22], [3845.9178, 9.27068e-22, 1.9766e-22, 1.18596e-22, 1.2120807e-22], [3854.7836, 9.29184e-22, 1.98111e-22, 1.18867e-22, 1.2149071e-22], [3863.6698, 9.31306e-22, 1.98563e-22, 1.19138e-22, 1.2177401e-22], [3872.5764, 9.33432e-22, 1.99016e-22, 1.1941e-22, 1.2205797e-22], [3881.5037, 9.35564e-22, 1.99471e-22, 1.19682e-22, 1.2234257e-22], [3890.4514, 9.377e-22, 1.99926e-22, 1.19956e-22, 1.2262783e-22], [3899.4199, 9.39841e-22, 2.00383e-22, 1.2023e-22, 1.2291373e-22], [3908.409, 9.41988e-22, 2.0084e-22, 1.20504e-22, 1.2320029e-22], [3917.4188, 9.44139e-22, 2.01299e-22, 1.20779e-22, 1.234875e-22], [3926.4494, 9.46295e-22, 2.01758e-22, 1.21055e-22, 1.2377537e-22], [3935.5008, 9.48457e-22, 2.02219e-22, 1.21332e-22, 1.2406389e-22], [3944.573, 9.50623e-22, 2.02681e-22, 1.21609e-22, 1.2435308e-22], [3953.6662, 9.52795e-22, 2.03144e-22, 1.21886e-22, 1.2464292e-22], [3962.7803, 9.54971e-22, 2.03608e-22, 1.22165e-22, 1.2493342e-22], [3971.9155, 9.57153e-22, 2.04073e-22, 1.22444e-22, 1.2522458e-22], [3981.0717, 9.59339e-22, 2.04539e-22, 1.22724e-22, 1.2551641e-22], [3990.249, 9.61531e-22, 2.05006e-22, 1.23004e-22, 1.2580891e-22], [3999.4475, 9.63728e-22, 2.05475e-22, 1.23285e-22, 1.2610207e-22], [4008.6672, 9.6593e-22, 2.05944e-22, 1.23567e-22, 1.263959e-22], [4017.9081, 9.68137e-22, 2.06415e-22, 1.23849e-22, 1.266904e-22], [4027.1703, 9.70349e-22, 2.06886e-22, 1.24132e-22, 1.2698557e-22], [4036.4539, 9.72566e-22, 2.07359e-22, 1.24415e-22, 1.2728142e-22], [4045.7589, 9.74789e-22, 2.07833e-22, 1.247e-22, 1.2757794e-22], [4055.0854, 9.77016e-22, 2.08308e-22, 1.24985e-22, 1.2787514e-22], [4064.4333, 9.79249e-22, 2.08784e-22, 1.2527e-22, 1.2817302e-22], [4073.8028, 9.81487e-22, 2.09261e-22, 1.25557e-22, 1.2847157e-22], [4083.1939, 9.8373e-22, 2.09739e-22, 1.25843e-22, 1.2877081e-22], [4092.6066, 9.85979e-22, 2.10218e-22, 1.26131e-22, 1.2907073e-22], [4102.041, 9.88233e-22, 2.10699e-22, 1.26419e-22, 1.2937133e-22], [4111.4972, 9.90491e-22, 2.1118e-22, 1.26708e-22, 1.2967263e-22], [4120.9752, 9.92756e-22, 2.11663e-22, 1.26998e-22, 1.299746e-22], [4130.475, 9.95025e-22, 2.12147e-22, 1.27288e-22, 1.3027727e-22], [4139.9967, 9.973e-22, 2.12632e-22, 1.27579e-22, 1.3058063e-22], [4149.5404, 9.9958e-22, 2.13118e-22, 1.27871e-22, 1.3088468e-22], [4159.1061, 1.00186e-21, 2.13605e-22, 1.28163e-22, 1.3118943e-22], [4168.6938, 1.00416e-21, 2.14093e-22, 1.28456e-22, 1.3149487e-22], [4178.3037, 1.00645e-21, 2.14583e-22, 1.2875e-22, 1.3180101e-22], [4187.9357, 1.00875e-21, 2.15073e-22, 1.29044e-22, 1.3210785e-22], [4197.5898, 1.01106e-21, 2.15565e-22, 1.29339e-22, 1.3241538e-22], [4207.2663, 1.01337e-21, 2.16058e-22, 1.29635e-22, 1.3272363e-22], [4216.965, 1.01569e-21, 2.16552e-22, 1.29931e-22, 1.3303257e-22], [4226.6861, 1.01801e-21, 2.17047e-22, 1.30228e-22, 1.3334222e-22], [4236.4297, 1.02034e-21, 2.17544e-22, 1.30526e-22, 1.3365258e-22], [4246.1956, 1.02267e-21, 2.18041e-22, 1.30825e-22, 1.3396364e-22], [4255.9841, 1.02501e-21, 2.1854e-22, 1.31124e-22, 1.3427542e-22], [4265.7952, 1.02736e-21, 2.19039e-22, 1.31424e-22, 1.3458791e-22], [4275.6289, 1.02971e-21, 2.1954e-22, 1.31724e-22, 1.3490111e-22], [4285.4852, 1.03206e-21, 2.20043e-22, 1.32026e-22, 1.3521503e-22], [4295.3643, 1.03442e-21, 2.20546e-22, 1.32328e-22, 1.3552966e-22], [4305.2661, 1.03679e-21, 2.2105e-22, 1.3263e-22, 1.3584501e-22], [4315.1908, 1.03916e-21, 2.21556e-22, 1.32934e-22, 1.3616109e-22], [4325.1383, 1.04154e-21, 2.22063e-22, 1.33238e-22, 1.3647788e-22], [4335.1088, 1.04392e-21, 2.22571e-22, 1.33542e-22, 1.367954e-22], [4345.1022, 1.04631e-21, 2.2308e-22, 1.33848e-22, 1.3711365e-22], [4355.1187, 1.0487e-21, 2.2359e-22, 1.34154e-22, 1.3743262e-22], [4365.1583, 1.0511e-21, 2.24102e-22, 1.34461e-22, 1.3775232e-22], [4375.2211, 1.05351e-21, 2.24614e-22, 1.34769e-22, 1.3807275e-22], [4385.307, 1.05592e-21, 2.25128e-22, 1.35077e-22, 1.3839392e-22], [4395.4162, 1.05833e-21, 2.25643e-22, 1.35386e-22, 1.3871581e-22], [4405.5486, 1.06076e-21, 2.2616e-22, 1.35696e-22, 1.3903845e-22], [4415.7045, 1.06318e-21, 2.26677e-22, 1.36006e-22, 1.3936182e-22], [4425.8837, 1.06562e-21, 2.27196e-22, 1.36318e-22, 1.3968593e-22], [4436.0864, 1.06805e-21, 2.27716e-22, 1.36629e-22, 1.4001078e-22], [4446.3127, 1.0705e-21, 2.28237e-22, 1.36942e-22, 1.4033637e-22], [4456.5625, 1.07295e-21, 2.28759e-22, 1.37256e-22, 1.4066271e-22], [4466.8359, 1.0754e-21, 2.29283e-22, 1.3757e-22, 1.4098979e-22], [4477.133, 1.07787e-21, 2.29808e-22, 1.37885e-22, 1.4131762e-22], [4487.4539, 1.08033e-21, 2.30333e-22, 1.382e-22, 1.416462e-22], [4497.7985, 1.08281e-21, 2.30861e-22, 1.38516e-22, 1.4197553e-22], [4508.167, 1.08528e-21, 2.31389e-22, 1.38833e-22, 1.4230562e-22], [4518.5594, 1.08777e-21, 2.31919e-22, 1.39151e-22, 1.4263646e-22], [4528.9758, 1.09026e-21, 2.3245e-22, 1.3947e-22, 1.4296805e-22], [4539.4162, 1.09275e-21, 2.32982e-22, 1.39789e-22, 1.4330041e-22], [4549.8806, 1.09526e-21, 2.33515e-22, 1.40109e-22, 1.4363352e-22], [4560.3692, 1.09776e-21, 2.3405e-22, 1.4043e-22, 1.4396739e-22], [4570.8819, 1.10028e-21, 2.34585e-22, 1.40751e-22, 1.4430203e-22], [4581.4189, 1.1028e-21, 2.35122e-22, 1.41073e-22, 1.4463744e-22], [4591.9801, 1.10532e-21, 2.35661e-22, 1.41396e-22, 1.4497361e-22], [4602.5657, 1.10785e-21, 2.362e-22, 1.4172e-22, 1.4531055e-22], [4613.1757, 1.11039e-21, 2.36741e-22, 1.42045e-22, 1.4564826e-22], [4623.8102, 1.11293e-21, 2.37283e-22, 1.4237e-22, 1.4598674e-22], [4634.4692, 1.11548e-21, 2.37827e-22, 1.42696e-22, 1.4632599e-22], [4645.1528, 1.11803e-21, 2.38371e-22, 1.43023e-22, 1.4666603e-22], [4655.8609, 1.12059e-21, 2.38917e-22, 1.4335e-22, 1.4700684e-22], [4666.5938, 1.12316e-21, 2.39464e-22, 1.43678e-22, 1.4734843e-22], [4677.3514, 1.12573e-21, 2.40012e-22, 1.44007e-22, 1.476908e-22], [4688.1338, 1.12831e-21, 2.40562e-22, 1.44337e-22, 1.4803395e-22], [4698.9411, 1.1309e-21, 2.41113e-22, 1.44668e-22, 1.4837789e-22], [4709.7733, 1.13349e-21, 2.41665e-22, 1.44999e-22, 1.4872262e-22], [4720.6304, 1.13608e-21, 2.42219e-22, 1.45331e-22, 1.4906814e-22], [4731.5126, 1.13868e-21, 2.42774e-22, 1.45664e-22, 1.4941444e-22], [4742.4199, 1.14129e-21, 2.4333e-22, 1.45998e-22, 1.4976156e-22], [4753.3523, 1.14391e-21, 2.43893e-22, 1.46336e-22, 1.5010979e-22], [4764.3099, 1.14653e-21, 2.44447e-22, 1.46668e-22, 1.5045822e-22], [4775.2927, 1.14915e-21, 2.45006e-22, 1.47003e-22, 1.5080761e-22], [4786.3009, 1.15179e-21, 2.45567e-22, 1.4734e-22, 1.5115789e-22], [4797.3345, 1.15443e-21, 2.46129e-22, 1.47678e-22, 1.5150897e-22], [4808.3935, 1.15707e-21, 2.46693e-22, 1.48016e-22, 1.5186086e-22], [4819.478, 1.15972e-21, 2.47258e-22, 1.48355e-22, 1.5221356e-22], [4830.588, 1.16238e-21, 2.47825e-22, 1.48695e-22, 1.5256706e-22], [4841.7237, 1.16504e-21, 2.48393e-22, 1.49036e-22, 1.5292137e-22], [4852.885, 1.16771e-21, 2.48962e-22, 1.49377e-22, 1.5327649e-22], [4864.0721, 1.17039e-21, 2.49532e-22, 1.49719e-22, 1.5363243e-22], [4875.2849, 1.17307e-21, 2.50104e-22, 1.50062e-22, 1.5398918e-22], [4886.5236, 1.17576e-21, 2.50677e-22, 1.50406e-22, 1.5434674e-22], [4897.7882, 1.17845e-21, 2.51251e-22, 1.50751e-22, 1.5470513e-22], [4909.0788, 1.18115e-21, 2.51827e-22, 1.51096e-22, 1.5506433e-22], [4920.3954, 1.18386e-21, 2.52404e-22, 1.51442e-22, 1.5542436e-22], [4931.738, 1.18657e-21, 2.52982e-22, 1.51789e-22, 1.5578521e-22], [4943.1069, 1.18929e-21, 2.53562e-22, 1.52137e-22, 1.5614689e-22], [4954.5019, 1.19202e-21, 2.54143e-22, 1.52486e-22, 1.565094e-22], [4965.9232, 1.19475e-21, 2.54726e-22, 1.52835e-22, 1.5687273e-22], [4977.3708, 1.19749e-21, 2.55309e-22, 1.53186e-22, 1.572369e-22], [4988.8449, 1.20023e-21, 2.55895e-22, 1.53537e-22, 1.576019e-22], [5000.3453, 1.20298e-21, 2.56481e-22, 1.53889e-22, 1.5796774e-22], [5011.8723, 1.20574e-21, 2.57069e-22, 1.54241e-22, 1.5833441e-22], [5023.4259, 1.2085e-21, 2.57658e-22, 1.54595e-22, 1.5870193e-22], [5035.0061, 1.21127e-21, 2.58249e-22, 1.54949e-22, 1.5907028e-22], [5046.613, 1.21405e-21, 2.58841e-22, 1.55304e-22, 1.5943948e-22], [5058.2466, 1.21683e-21, 2.59434e-22, 1.5566e-22, 1.5980953e-22], [5069.9071, 1.21962e-21, 2.60029e-22, 1.56017e-22, 1.6018042e-22], [5081.5944, 1.22242e-21, 2.60625e-22, 1.56375e-22, 1.6055216e-22], [5093.3087, 1.22522e-21, 2.61222e-22, 1.56733e-22, 1.6092475e-22], [5105.05, 1.22803e-21, 2.61821e-22, 1.57093e-22, 1.612982e-22], [5116.8184, 1.23084e-21, 2.62421e-22, 1.57453e-22, 1.616725e-22], [5128.6138, 1.23367e-21, 2.63023e-22, 1.57814e-22, 1.6204766e-22], [5140.4365, 1.2365e-21, 2.63626e-22, 1.58176e-22, 1.6242368e-22], [5152.2864, 1.23933e-21, 2.6423e-22, 1.58538e-22, 1.6280056e-22], [5164.1637, 1.24217e-21, 2.64836e-22, 1.58902e-22, 1.631783e-22], [5176.0683, 1.24502e-21, 2.65443e-22, 1.59266e-22, 1.6355691e-22], [5188.0004, 1.24788e-21, 2.66052e-22, 1.59631e-22, 1.6393638e-22], [5199.96, 1.25074e-21, 2.66662e-22, 1.59997e-22, 1.6431672e-22], [5211.9471, 1.2536e-21, 2.67274e-22, 1.60364e-22, 1.6469794e-22], [5223.9619, 1.25648e-21, 2.67886e-22, 1.60732e-22, 1.6508003e-22], [5236.0044, 1.25936e-21, 2.68501e-22, 1.611e-22, 1.6546299e-22], [5248.0746, 1.26225e-21, 2.69116e-22, 1.6147e-22, 1.6584683e-22], [5260.1727, 1.26514e-21, 2.69734e-22, 1.6184e-22, 1.6623156e-22], [5272.2986, 1.26805e-21, 2.70352e-22, 1.62211e-22, 1.6661716e-22], [5284.4525, 1.27095e-21, 2.70972e-22, 1.62583e-22, 1.6700364e-22], [5296.6344, 1.27387e-21, 2.71594e-22, 1.62956e-22, 1.6739101e-22], [5308.8444, 1.27679e-21, 2.72217e-22, 1.6333e-22, 1.6777927e-22], [5321.0826, 1.27972e-21, 2.72841e-22, 1.63705e-22, 1.6816842e-22], [5333.349, 1.28265e-21, 2.73467e-22, 1.6408e-22, 1.6855846e-22], [5345.6436, 1.2856e-21, 2.74094e-22, 1.64456e-22, 1.689494e-22], [5357.9666, 1.28854e-21, 2.74723e-22, 1.64834e-22, 1.6934123e-22], [5370.318, 1.2915e-21, 2.75353e-22, 1.65212e-22, 1.6973395e-22], [5382.6978, 1.29446e-21, 2.75984e-22, 1.65591e-22, 1.7012758e-22], [5395.1062, 1.29743e-21, 2.76617e-22, 1.6597e-22, 1.7052211e-22], [5407.5432, 1.30041e-21, 2.77252e-22, 1.66351e-22, 1.7091755e-22], [5420.0089, 1.30339e-21, 2.77888e-22, 1.66733e-22, 1.7131389e-22], [5432.5033, 1.30638e-21, 2.78525e-22, 1.67115e-22, 1.7171114e-22], [5445.0265, 1.30938e-21, 2.79164e-22, 1.67499e-22, 1.7210929e-22], [5457.5786, 1.31238e-21, 2.79805e-22, 1.67883e-22, 1.7250837e-22], [5470.1596, 1.31539e-21, 2.80447e-22, 1.68268e-22, 1.7290835e-22], [5482.7696, 1.31841e-21, 2.8109e-22, 1.68654e-22, 1.7330926e-22], [5495.4087, 1.32144e-21, 2.81735e-22, 1.69041e-22, 1.7371108e-22], [5508.077, 1.32447e-21, 2.82381e-22, 1.69429e-22, 1.7411382e-22], [5520.7744, 1.32751e-21, 2.83029e-22, 1.69817e-22, 1.7451749e-22], [5533.5011, 1.33055e-21, 2.83678e-22, 1.70207e-22, 1.7492208e-22], [5546.2571, 1.33361e-21, 2.84329e-22, 1.70598e-22, 1.753276e-22], [5559.0426, 1.33667e-21, 2.84982e-22, 1.70989e-22, 1.7573405e-22], [5571.8575, 1.33973e-21, 2.85636e-22, 1.71381e-22, 1.7614143e-22], [5584.7019, 1.34281e-21, 2.86291e-22, 1.71775e-22, 1.7654974e-22], [5597.576, 1.34589e-21, 2.86948e-22, 1.72169e-22, 1.76959e-22], [5610.4798, 1.34898e-21, 2.87606e-22, 1.72564e-22, 1.7736919e-22], [5623.4133, 1.35207e-21, 2.88266e-22, 1.7296e-22, 1.7778032e-22], [5636.3766, 1.35518e-21, 2.88928e-22, 1.73357e-22, 1.7819239e-22], [5649.3697, 1.35829e-21, 2.89591e-22, 1.73755e-22, 1.786054e-22], [5662.3929, 1.3614e-21, 2.90256e-22, 1.74153e-22, 1.7901937e-22], [5675.4461, 1.36453e-21, 2.90922e-22, 1.74553e-22, 1.7943428e-22], [5688.5293, 1.36766e-21, 2.91589e-22, 1.74954e-22, 1.7985015e-22], [5701.6427, 1.3708e-21, 2.92258e-22, 1.75355e-22, 1.8026697e-22], [5714.7864, 1.37394e-21, 2.92929e-22, 1.75758e-22, 1.8068474e-22], [5727.9603, 1.3771e-21, 2.93602e-22, 1.76161e-22, 1.8110347e-22], [5741.1646, 1.38026e-21, 2.94275e-22, 1.76565e-22, 1.8152317e-22], [5754.3994, 1.38343e-21, 2.94951e-22, 1.76971e-22, 1.8194382e-22], [5767.6646, 1.3866e-21, 2.95628e-22, 1.77377e-22, 1.8236545e-22], [5780.9605, 1.38979e-21, 2.96311e-22, 1.77787e-22, 1.8278833e-22], [5794.287, 1.39297e-21, 2.96987e-22, 1.78192e-22, 1.8321161e-22], [5807.6442, 1.39617e-21, 2.97668e-22, 1.78601e-22, 1.8363612e-22], [5821.0322, 1.39938e-21, 2.98351e-22, 1.79011e-22, 1.8406161e-22], [5834.451, 1.40259e-21, 2.99036e-22, 1.79422e-22, 1.8448809e-22], [5847.9008, 1.40581e-21, 2.99723e-22, 1.79834e-22, 1.8491555e-22], [5861.3816, 1.40904e-21, 3.00411e-22, 1.80246e-22, 1.8534398e-22], [5874.8935, 1.41227e-21, 3.011e-22, 1.8066e-22, 1.857734e-22], [5888.4366, 1.41551e-21, 3.01792e-22, 1.81075e-22, 1.862038e-22], [5902.0108, 1.41876e-21, 3.02484e-22, 1.81491e-22, 1.8663519e-22], [5915.6163, 1.42202e-21, 3.03179e-22, 1.81907e-22, 1.8706758e-22], [5929.2532, 1.42528e-21, 3.03875e-22, 1.82325e-22, 1.8750095e-22], [5942.9216, 1.42856e-21, 3.04573e-22, 1.82744e-22, 1.8793531e-22], [5956.6214, 1.43184e-21, 3.05272e-22, 1.83163e-22, 1.8837068e-22], [5970.3529, 1.43512e-21, 3.05973e-22, 1.83584e-22, 1.8880704e-22], [5984.116, 1.43842e-21, 3.06675e-22, 1.84005e-22, 1.892444e-22], [5997.9108, 1.44172e-21, 3.07379e-22, 1.84428e-22, 1.8968277e-22], [6011.7374, 1.44503e-21, 3.08085e-22, 1.84851e-22, 1.9012214e-22], [6025.5959, 1.44835e-21, 3.08792e-22, 1.85275e-22, 1.9056252e-22], [6039.4863, 1.45168e-21, 3.09501e-22, 1.85701e-22, 1.9100391e-22], [6053.4087, 1.45501e-21, 3.10212e-22, 1.86127e-22, 1.9144631e-22], [6067.3633, 1.45835e-21, 3.10924e-22, 1.86555e-22, 1.9188973e-22], [6081.35, 1.4617e-21, 3.11638e-22, 1.86983e-22, 1.9233417e-22], [6095.369, 1.46506e-21, 3.12354e-22, 1.87412e-22, 1.9277962e-22], [6109.4202, 1.46842e-21, 3.13071e-22, 1.87843e-22, 1.932261e-22], [6123.5039, 1.47179e-21, 3.1379e-22, 1.88274e-22, 1.936736e-22], [6137.6201, 1.47517e-21, 3.14511e-22, 1.88706e-22, 1.9412213e-22], [6151.7687, 1.47856e-21, 3.15233e-22, 1.8914e-22, 1.9457169e-22], [6165.95, 1.48196e-21, 3.15957e-22, 1.89574e-22, 1.9502228e-22], [6180.164, 1.48536e-21, 3.16683e-22, 1.9001e-22, 1.9547391e-22], [6194.4108, 1.48877e-21, 3.1741e-22, 1.90446e-22, 1.9592657e-22], [6208.6903, 1.49219e-21, 3.18139e-22, 1.90883e-22, 1.9638027e-22], [6223.0029, 1.49562e-21, 3.1887e-22, 1.91322e-22, 1.9683501e-22], [6237.3484, 1.49905e-21, 3.19602e-22, 1.91761e-22, 1.9729079e-22], [6251.7269, 1.5025e-21, 3.20336e-22, 1.92202e-22, 1.9774762e-22], [6266.1386, 1.50595e-21, 3.21072e-22, 1.92643e-22, 1.982055e-22], [6280.5836, 1.50941e-21, 3.21809e-22, 1.93085e-22, 1.9866443e-22], [6295.0618, 1.51287e-21, 3.22548e-22, 1.93529e-22, 1.9912442e-22], [6309.5734, 1.51635e-21, 3.23289e-22, 1.93973e-22, 1.9958545e-22], [6324.1185, 1.51983e-21, 3.24032e-22, 1.94419e-22, 2.0004755e-22], [6338.6971, 1.52332e-21, 3.24776e-22, 1.94866e-22, 2.0051071e-22], [6353.3093, 1.52682e-21, 3.25522e-22, 1.95313e-22, 2.0097493e-22], [6367.9552, 1.53033e-21, 3.2627e-22, 1.95762e-22, 2.0144022e-22], [6382.6349, 1.53384e-21, 3.27019e-22, 1.96212e-22, 2.0190657e-22], [6397.3484, 1.53737e-21, 3.2777e-22, 1.96662e-22, 2.02374e-22], [6412.0958, 1.5409e-21, 3.28523e-22, 1.97114e-22, 2.028425e-22], [6426.8772, 1.54444e-21, 3.29278e-22, 1.97567e-22, 2.0331207e-22], [6441.6927, 1.54799e-21, 3.30034e-22, 1.98021e-22, 2.0378272e-22], [6456.5423, 1.55154e-21, 3.30793e-22, 1.98476e-22, 2.0425446e-22], [6471.4262, 1.55511e-21, 3.31553e-22, 1.98932e-22, 2.0472727e-22], [6486.3443, 1.55868e-21, 3.32314e-22, 1.99389e-22, 2.0520117e-22], [6501.2969, 1.56226e-21, 3.33078e-22, 1.99847e-22, 2.0567616e-22], [6516.2839, 1.56585e-21, 3.33843e-22, 2.00306e-22, 2.0615224e-22], [6531.3055, 1.56945e-21, 3.3461e-22, 2.00766e-22, 2.0662941e-22], [6546.3617, 1.57305e-21, 3.35379e-22, 2.01227e-22, 2.0710768e-22], [6561.4527, 1.57667e-21, 3.36149e-22, 2.0169e-22, 2.0758705e-22], [6576.5784, 1.58029e-21, 3.36922e-22, 2.02153e-22, 2.0806751e-22], [6591.739, 1.58392e-21, 3.37696e-22, 2.02617e-22, 2.0854908e-22], [6606.9345, 1.58756e-21, 3.38472e-22, 2.03083e-22, 2.0903176e-22], [6622.165, 1.59121e-21, 3.39249e-22, 2.0355e-22, 2.0951554e-22], [6637.4307, 1.59486e-21, 3.40029e-22, 2.04017e-22, 2.1000044e-22], [6652.7316, 1.59853e-21, 3.4081e-22, 2.04486e-22, 2.1048645e-22], [6668.0677, 1.6022e-21, 3.41593e-22, 2.04956e-22, 2.1097357e-22], [6683.4392, 1.60588e-21, 3.42378e-22, 2.05427e-22, 2.1146181e-22], [6698.8461, 1.60957e-21, 3.43165e-22, 2.05899e-22, 2.1195118e-22], [6714.2885, 1.61327e-21, 3.43953e-22, 2.06372e-22, 2.1244167e-22], [6729.7666, 1.61698e-21, 3.44744e-22, 2.06846e-22, 2.1293328e-22], [6745.2803, 1.6207e-21, 3.45536e-22, 2.07322e-22, 2.1342602e-22], [6760.8298, 1.62442e-21, 3.4633e-22, 2.07798e-22, 2.139199e-22], [6776.4151, 1.62815e-21, 3.47126e-22, 2.08275e-22, 2.1441491e-22], [6792.0363, 1.63189e-21, 3.47923e-22, 2.08754e-22, 2.1491106e-22], [6807.6936, 1.63564e-21, 3.48723e-22, 2.09234e-22, 2.1540834e-22], [6823.3869, 1.6394e-21, 3.49524e-22, 2.09715e-22, 2.1590677e-22], [6839.1165, 1.64317e-21, 3.50328e-22, 2.10197e-22, 2.1640634e-22], [6854.8823, 1.64695e-21, 3.51133e-22, 2.1068e-22, 2.1690706e-22], [6870.6844, 1.65073e-21, 3.5194e-22, 2.11164e-22, 2.1740893e-22], [6886.523, 1.65453e-21, 3.52749e-22, 2.11649e-22, 2.1791196e-22], [6902.398, 1.65833e-21, 3.5356e-22, 2.12136e-22, 2.1841615e-22], [6918.3097, 1.66214e-21, 3.54373e-22, 2.12624e-22, 2.1892155e-22], [6934.2581, 1.66596e-21, 3.55186e-22, 2.13112e-22, 2.1942797e-22], [6950.2432, 1.66979e-21, 3.56003e-22, 2.13602e-22, 2.1993563e-22], [6966.2651, 1.67363e-21, 3.56821e-22, 2.14093e-22, 2.2044445e-22], [6982.324, 1.67747e-21, 3.57641e-22, 2.14585e-22, 2.2095445e-22], [6998.42, 1.68133e-21, 3.58463e-22, 2.15078e-22, 2.2146561e-22], [7014.553, 1.68519e-21, 3.59287e-22, 2.15572e-22, 2.2197795e-22], [7030.7232, 1.68907e-21, 3.60113e-22, 2.16068e-22, 2.2249147e-22], [7046.9307, 1.69295e-21, 3.60941e-22, 2.16564e-22, 2.2300617e-22], [7063.1755, 1.69684e-21, 3.6177e-22, 2.17062e-22, 2.2352205e-22], [7079.4578, 1.70074e-21, 3.62602e-22, 2.17561e-22, 2.2403912e-22], [7095.7777, 1.70465e-21, 3.63435e-22, 2.18061e-22, 2.2455737e-22], [7112.1351, 1.70857e-21, 3.64271e-22, 2.18562e-22, 2.2507681e-22], [7128.5303, 1.7125e-21, 3.65108e-22, 2.19065e-22, 2.2559745e-22], [7144.9633, 1.71644e-21, 3.65947e-22, 2.19568e-22, 2.2611928e-22], [7161.4341, 1.72038e-21, 3.66789e-22, 2.20073e-22, 2.2664231e-22], [7177.9429, 1.72434e-21, 3.67632e-22, 2.20579e-22, 2.2716655e-22], [7194.4898, 1.7283e-21, 3.68477e-22, 2.21086e-22, 2.2769199e-22], [7211.0748, 1.73227e-21, 3.69324e-22, 2.21594e-22, 2.2821863e-22], [7227.698, 1.73626e-21, 3.70173e-22, 2.22104e-22, 2.2874649e-22], [7244.3596, 1.74025e-21, 3.71024e-22, 2.22614e-22, 2.2927556e-22], [7261.0596, 1.74425e-21, 3.71877e-22, 2.23126e-22, 2.2980584e-22], [7277.798, 1.74826e-21, 3.72732e-22, 2.23639e-22, 2.3033734e-22], [7294.5751, 1.75228e-21, 3.73589e-22, 2.24153e-22, 2.3087007e-22], [7311.3908, 1.75631e-21, 3.74448e-22, 2.24669e-22, 2.3140401e-22], [7328.2453, 1.76034e-21, 3.75308e-22, 2.25185e-22, 2.3193919e-22], [7345.1387, 1.76439e-21, 3.76171e-22, 2.25703e-22, 2.3247559e-22], [7362.071, 1.76845e-21, 3.77036e-22, 2.26222e-22, 2.3301323e-22], [7379.0423, 1.77251e-21, 3.77903e-22, 2.26742e-22, 2.335521e-22], [7396.0528, 1.77659e-21, 3.78772e-22, 2.27263e-22, 2.3409221e-22], [7413.1024, 1.78067e-21, 3.79643e-22, 2.27786e-22, 2.3463356e-22], [7430.1914, 1.78477e-21, 3.80516e-22, 2.28309e-22, 2.3517616e-22], [7447.3197, 1.78887e-21, 3.81391e-22, 2.28834e-22, 2.3572e-22], [7464.4876, 1.79298e-21, 3.82268e-22, 2.29361e-22, 2.362651e-22], [7481.695, 1.79711e-21, 3.83146e-22, 2.29888e-22, 2.3681144e-22], [7498.9421, 1.80124e-21, 3.84027e-22, 2.30416e-22, 2.3735904e-22], [7516.2289, 1.80538e-21, 3.8491e-22, 2.30946e-22, 2.379079e-22], [7533.5556, 1.80953e-21, 3.85796e-22, 2.31477e-22, 2.3845802e-22], [7550.9223, 1.81369e-21, 3.86683e-22, 2.3201e-22, 2.3900941e-22], [7568.329, 1.81786e-21, 3.87572e-22, 2.32543e-22, 2.3956206e-22], [7585.7758, 1.82204e-21, 3.88463e-22, 2.33078e-22, 2.4011598e-22], [7603.2628, 1.82623e-21, 3.89356e-22, 2.33614e-22, 2.4067118e-22], [7620.7901, 1.83043e-21, 3.90252e-22, 2.34151e-22, 2.4122765e-22], [7638.3578, 1.83464e-21, 3.91149e-22, 2.34689e-22, 2.417854e-22], [7655.9661, 1.83886e-21, 3.92048e-22, 2.35229e-22, 2.4234443e-22], [7673.6149, 1.84309e-21, 3.9295e-22, 2.3577e-22, 2.4290475e-22], [7691.3044, 1.84733e-21, 3.93854e-22, 2.36312e-22, 2.4346636e-22], [7709.0347, 1.85158e-21, 3.94759e-22, 2.36856e-22, 2.4402925e-22], [7726.8059, 1.85584e-21, 3.95667e-22, 2.374e-22, 2.4459344e-22], [7744.618, 1.8601e-21, 3.96577e-22, 2.37946e-22, 2.4515893e-22], [7762.4712, 1.86438e-21, 3.97489e-22, 2.38493e-22, 2.4572572e-22], [7780.3655, 1.86867e-21, 3.98403e-22, 2.39042e-22, 2.4629381e-22], [7798.3011, 1.87297e-21, 3.99319e-22, 2.39592e-22, 2.468632e-22], [7816.278, 1.87727e-21, 4.00238e-22, 2.40143e-22, 2.4743391e-22], [7834.2964, 1.88159e-21, 4.01158e-22, 2.40695e-22, 2.4800592e-22], [7852.3563, 1.88592e-21, 4.02081e-22, 2.41249e-22, 2.4857925e-22], [7870.4579, 1.89026e-21, 4.03006e-22, 2.41803e-22, 2.491539e-22], [7888.6012, 1.8946e-21, 4.03933e-22, 2.4236e-22, 2.4972987e-22], [7906.7863, 1.89896e-21, 4.04862e-22, 2.42917e-22, 2.5030717e-22], [7925.0133, 1.90333e-21, 4.05793e-22, 2.43476e-22, 2.5088579e-22], [7943.2823, 1.90771e-21, 4.06726e-22, 2.44036e-22, 2.5146574e-22], [7961.5935, 1.91209e-21, 4.07661e-22, 2.44597e-22, 2.5204703e-22], [7979.9469, 1.91649e-21, 4.08599e-22, 2.45159e-22, 2.5262965e-22], [7998.3426, 1.9209e-21, 4.09539e-22, 2.45723e-22, 2.5321361e-22], [8016.7806, 1.92532e-21, 4.10481e-22, 2.46289e-22, 2.5379891e-22], [8035.2612, 1.92975e-21, 4.11425e-22, 2.46855e-22, 2.5438556e-22], [8053.7844, 1.93419e-21, 4.12371e-22, 2.47423e-22, 2.5497356e-22], [8072.3503, 1.93863e-21, 4.1332e-22, 2.47992e-22, 2.5556291e-22], [8090.959, 1.94309e-21, 4.14271e-22, 2.48562e-22, 2.5615361e-22], [8109.6106, 1.94756e-21, 4.15223e-22, 2.49134e-22, 2.5674568e-22], [8128.3052, 1.95204e-21, 4.16179e-22, 2.49707e-22, 2.573391e-22], [8147.0428, 1.95654e-21, 4.17141e-22, 2.50284e-22, 2.579342e-22], [8165.8237, 1.96103e-21, 4.18095e-22, 2.50857e-22, 2.5853005e-22], [8184.6479, 1.96554e-21, 4.19057e-22, 2.51434e-22, 2.5912757e-22], [8203.5154, 1.97007e-21, 4.20021e-22, 2.52013e-22, 2.5972647e-22], [8222.4265, 1.9746e-21, 4.20987e-22, 2.52592e-22, 2.6032675e-22], [8241.3812, 1.97914e-21, 4.21956e-22, 2.53173e-22, 2.6092841e-22], [8260.3795, 1.98369e-21, 4.22926e-22, 2.53756e-22, 2.6153145e-22], [8279.4216, 1.98826e-21, 4.23899e-22, 2.5434e-22, 2.6213588e-22], [8298.5077, 1.99283e-21, 4.24874e-22, 2.54925e-22, 2.6274169e-22], [8317.6377, 1.99741e-21, 4.25852e-22, 2.55511e-22, 2.633489e-22], [8336.8118, 2.00201e-21, 4.26831e-22, 2.56099e-22, 2.6395751e-22], [8356.0302, 2.00662e-21, 4.27813e-22, 2.56688e-22, 2.6456752e-22], [8375.2928, 2.01123e-21, 4.28798e-22, 2.57279e-22, 2.6517893e-22], [8394.5999, 2.01586e-21, 4.29784e-22, 2.5787e-22, 2.6579174e-22], [8413.9514, 2.0205e-21, 4.30773e-22, 2.58464e-22, 2.6640597e-22], [8433.3476, 2.02514e-21, 4.31764e-22, 2.59058e-22, 2.670216e-22], [8452.7885, 2.0298e-21, 4.32757e-22, 2.59654e-22, 2.6763866e-22], [8472.2741, 2.03447e-21, 4.33753e-22, 2.60252e-22, 2.6825713e-22], [8491.8048, 2.03915e-21, 4.34751e-22, 2.6085e-22, 2.6887702e-22], [8511.3804, 2.04385e-21, 4.35751e-22, 2.6145e-22, 2.6949834e-22], [8531.0011, 2.04855e-21, 4.36753e-22, 2.62052e-22, 2.7012109e-22], [8550.6671, 2.05326e-21, 4.37758e-22, 2.62655e-22, 2.7074527e-22], [8570.3785, 2.05799e-21, 4.38765e-22, 2.63259e-22, 2.7137089e-22], [8590.1352, 2.06272e-21, 4.39775e-22, 2.63865e-22, 2.7199794e-22], [8609.9375, 2.06747e-21, 4.40787e-22, 2.64472e-22, 2.7262644e-22], [8629.7855, 2.07222e-21, 4.41801e-22, 2.6508e-22, 2.7325638e-22], [8649.6792, 2.07699e-21, 4.42817e-22, 2.6569e-22, 2.7388777e-22], [8669.6188, 2.08177e-21, 4.43836e-22, 2.66302e-22, 2.7452062e-22], [8689.6043, 2.08656e-21, 4.44857e-22, 2.66914e-22, 2.7515492e-22], [8709.6359, 2.09136e-21, 4.45881e-22, 2.67529e-22, 2.7579068e-22], [8729.7137, 2.09617e-21, 4.46907e-22, 2.68144e-22, 2.764279e-22], [8749.8378, 2.101e-21, 4.47935e-22, 2.68761e-22, 2.7706658e-22], [8770.0082, 2.10583e-21, 4.48966e-22, 2.69379e-22, 2.7770674e-22], [8790.2252, 2.11068e-21, 4.49999e-22, 2.69999e-22, 2.7834836e-22], [8810.4887, 2.11553e-21, 4.51034e-22, 2.70621e-22, 2.7899147e-22], [8830.799, 2.1204e-21, 4.52072e-22, 2.71243e-22, 2.7963605e-22], [8851.1561, 2.12528e-21, 4.53112e-22, 2.71867e-22, 2.8028211e-22], [8871.5601, 2.13017e-21, 4.54155e-22, 2.72493e-22, 2.8092966e-22], [8892.0112, 2.13507e-21, 4.552e-22, 2.7312e-22, 2.815787e-22], [8912.5094, 2.13998e-21, 4.56247e-22, 2.73748e-22, 2.8222924e-22], [8933.0548, 2.14491e-21, 4.57297e-22, 2.74378e-22, 2.8288127e-22], [8953.6477, 2.14984e-21, 4.58349e-22, 2.7501e-22, 2.835348e-22], [8974.2879, 2.15479e-21, 4.59404e-22, 2.75643e-22, 2.8418983e-22], [8994.9758, 2.15975e-21, 4.60461e-22, 2.76277e-22, 2.8484637e-22], [9015.7114, 2.16472e-21, 4.61521e-22, 2.76913e-22, 2.8550442e-22], [9036.4947, 2.1697e-21, 4.62583e-22, 2.7755e-22, 2.8616398e-22], [9057.326, 2.17469e-21, 4.63647e-22, 2.78188e-22, 2.8682506e-22], [9078.2053, 2.1797e-21, 4.64714e-22, 2.78829e-22, 2.8748767e-22], [9099.1327, 2.18471e-21, 4.65784e-22, 2.7947e-22, 2.8815179e-22], [9120.1084, 2.18974e-21, 4.66856e-22, 2.80113e-22, 2.8881744e-22], [9141.1324, 2.19478e-21, 4.6793e-22, 2.80758e-22, 2.8948463e-22], [9162.2049, 2.19983e-21, 4.69007e-22, 2.81404e-22, 2.9015335e-22], [9183.326, 2.20489e-21, 4.70086e-22, 2.82052e-22, 2.9082361e-22], [9204.4957, 2.20997e-21, 4.71168e-22, 2.82701e-22, 2.9149541e-22], [9225.7143, 2.21505e-21, 4.72252e-22, 2.83351e-22, 2.9216875e-22], [9246.9817, 2.22015e-21, 4.73339e-22, 2.84004e-22, 2.9284365e-22], [9268.2982, 2.22526e-21, 4.74429e-22, 2.84657e-22, 2.9352009e-22], [9289.6639, 2.23038e-21, 4.7552e-22, 2.85312e-22, 2.941981e-22], [9311.0788, 2.23552e-21, 4.76615e-22, 2.85969e-22, 2.9487766e-22], [9332.543, 2.24066e-21, 4.77712e-22, 2.86627e-22, 2.9555879e-22], [9354.0567, 2.24582e-21, 4.78811e-22, 2.87287e-22, 2.9624148e-22], [9375.6201, 2.25099e-21, 4.79913e-22, 2.87948e-22, 2.9692575e-22], [9397.2331, 2.25617e-21, 4.81018e-22, 2.88611e-22, 2.9761158e-22], [9418.896, 2.26136e-21, 4.82125e-22, 2.89275e-22, 2.98299e-22], [9440.6088, 2.26656e-21, 4.83234e-22, 2.89941e-22, 2.98988e-22], [9462.3716, 2.27178e-21, 4.84346e-22, 2.90608e-22, 2.9967859e-22], [9484.1846, 2.27701e-21, 4.85462e-22, 2.91277e-22, 3.0037078e-22], [9506.0479, 2.28225e-21, 4.86578e-22, 2.91947e-22, 3.0106452e-22], [9527.9616, 2.2875e-21, 4.87698e-22, 2.92619e-22, 3.0175988e-22], [9549.9259, 2.29277e-21, 4.88821e-22, 2.93293e-22, 3.0245684e-22], [9571.9407, 2.29804e-21, 4.89946e-22, 2.93968e-22, 3.0315541e-22], [9594.0063, 2.30333e-21, 4.91074e-22, 2.94644e-22, 3.0385558e-22], [9616.1228, 2.30864e-21, 4.92204e-22, 2.95322e-22, 3.0455736e-22], [9638.2902, 2.31395e-21, 4.93337e-22, 2.96002e-22, 3.0526076e-22], [9660.5088, 2.31928e-21, 4.94472e-22, 2.96683e-22, 3.0596578e-22], [9682.7786, 2.32461e-21, 4.9561e-22, 2.97366e-22, 3.0667241e-22], [9705.0997, 2.32996e-21, 4.96751e-22, 2.98051e-22, 3.0738068e-22], [9727.4722, 2.33533e-21, 4.97895e-22, 2.98737e-22, 3.0809057e-22], [9749.8964, 2.3407e-21, 4.99041e-22, 2.99424e-22, 3.088021e-22], [9772.3722, 2.34609e-21, 5.00189e-22, 3.00114e-22, 3.0951526e-22], [9794.8999, 2.35149e-21, 5.01341e-22, 3.00804e-22, 3.1023007e-22], [9817.4794, 2.3569e-21, 5.02495e-22, 3.01497e-22, 3.1094652e-22], [9840.1111, 2.36233e-21, 5.03651e-22, 3.02191e-22, 3.1166462e-22], [9862.7949, 2.36777e-21, 5.04811e-22, 3.02886e-22, 3.1238437e-22], [9885.5309, 2.37322e-21, 5.05973e-22, 3.03584e-22, 3.1310578e-22], [9908.3194, 2.37868e-21, 5.07137e-22, 3.04282e-22, 3.1382885e-22], [9931.1605, 2.38415e-21, 5.08305e-22, 3.04983e-22, 3.1455358e-22], [9954.0542, 2.38964e-21, 5.09475e-22, 3.05685e-22, 3.1527998e-22], [9977.0006, 2.39514e-21, 5.10647e-22, 3.06388e-22, 3.1600805e-22], [10000.0, 2.40066e-21, 5.11823e-22, 3.07094e-22, 3.167378e-22]])
            
    f1 = data[:,0]
    if IFO.upper()=='KAGRA_EARLY':
        S1 = data[:,1]**2
    elif IFO.upper()=='KAGRA_MID':
        S1 = data[:,2]**2
    elif IFO.upper()=='KAGRA_LATE':
        S1 = data[:,3]**2
    elif IFO.upper()=='KAGRA':
        S1 = data[:,4]**2
    else:
        raise ValueError('IFO '+IFO+' not recognized.')
    fstop = [f1[0],f1[-1]]

    # % ---- Interpolate to desired frequencies.  Set noise to infinity outside
    # %      band of frequencies covered by f1.
    # S = inf(size(f)); 
    # k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop


    # % -------------------------------------------------------------------------
    # %                              ETB spectrum
    # % -------------------------------------------------------------------------

def etb(f):
    # % ETB - Design noise curve "B" for ET.
    # %
    # % Arguments are the same as for the parent function.

    # % ---- Extract pre-sampled noise spectrum from data.  First column is
    # %      frequency, second column is effective noise amplitude for a 5-year
    # %      integration period with SNR=10.  (We'll need to remove these
    # %      scalings.)  Data supplied by B. S. Sathyaprakash.
    data = np.array([[1.0, 4.052959e-24], [1.003076, 3.910123e-24], [1.006161, 3.772411e-24], [1.009256, 3.639651e-24], [1.01236, 3.511674e-24], [1.015474, 3.388315e-24], [1.018598, 3.269419e-24], [1.021731, 3.154832e-24], [1.024873, 3.044407e-24], [1.028026, 2.938001e-24], [1.031188, 2.835476e-24], [1.03436, 2.736698e-24], [1.037541, 2.64154e-24], [1.040733, 2.549875e-24], [1.043934, 2.461583e-24], [1.047145, 2.376549e-24], [1.050366, 2.294658e-24], [1.053596, 2.215802e-24], [1.056837, 2.139876e-24], [1.060088, 2.066778e-24], [1.063348, 1.996409e-24], [1.066619, 1.928674e-24], [1.0699, 1.863482e-24], [1.073191, 1.800742e-24], [1.076492, 1.740371e-24], [1.079803, 1.682284e-24], [1.083124, 1.6264e-24], [1.086456, 1.572644e-24], [1.089797, 1.520939e-24], [1.093149, 1.471213e-24], [1.096512, 1.423396e-24], [1.099885, 1.37742e-24], [1.103268, 1.333216e-24], [1.106661, 1.290716e-24], [1.110065, 1.249858e-24], [1.113479, 1.210583e-24], [1.116904, 1.172834e-24], [1.12034, 1.136553e-24], [1.123786, 1.101689e-24], [1.127242, 1.068189e-24], [1.13071, 1.036003e-24], [1.134188, 1.005084e-24], [1.137676, 9.753842e-25], [1.141175, 9.468589e-25], [1.144686, 9.194648e-25], [1.148206, 8.931599e-25], [1.151738, 8.679035e-25], [1.155281, 8.436564e-25], [1.158834, 8.203807e-25], [1.162399, 7.980397e-25], [1.165974, 7.765978e-25], [1.169561, 7.560208e-25], [1.173158, 7.362755e-25], [1.176766, 7.173297e-25], [1.180386, 6.991523e-25], [1.184017, 6.817133e-25], [1.187658, 6.649835e-25], [1.191312, 6.489347e-25], [1.194976, 6.335398e-25], [1.198651, 6.187722e-25], [1.202338, 6.046072e-25], [1.206036, 5.910222e-25], [1.209746, 5.779931e-25], [1.213467, 5.654966e-25], [1.2172, 5.5351e-25], [1.220943, 5.420115e-25], [1.224699, 5.3098e-25], [1.228466, 5.203951e-25], [1.232244, 5.102373e-25], [1.236035, 5.004875e-25], [1.239837, 4.911275e-25], [1.24365, 4.821397e-25], [1.247475, 4.735071e-25], [1.251313, 4.652136e-25], [1.255161, 4.572434e-25], [1.259022, 4.495814e-25], [1.262895, 4.422133e-25], [1.266779, 4.351251e-25], [1.270676, 4.283035e-25], [1.274584, 4.217358e-25], [1.278504, 4.154098e-25], [1.282437, 4.093138e-25], [1.286382, 4.034366e-25], [1.290338, 3.977675e-25], [1.294307, 3.922962e-25], [1.298288, 3.870132e-25], [1.302282, 3.819092e-25], [1.306287, 3.769757e-25], [1.310305, 3.722043e-25], [1.314336, 3.67587e-25], [1.318378, 3.631159e-25], [1.322433, 3.587838e-25], [1.326501, 3.545837e-25], [1.330581, 3.505089e-25], [1.334674, 3.465531e-25], [1.338779, 3.427105e-25], [1.342897, 3.389754e-25], [1.347028, 3.353424e-25], [1.351171, 3.318065e-25], [1.355327, 3.283628e-25], [1.359496, 3.250068e-25], [1.363677, 3.217342e-25], [1.367872, 3.18541e-25], [1.372079, 3.154233e-25], [1.376299, 3.123775e-25], [1.380533, 3.094001e-25], [1.384779, 3.06488e-25], [1.389038, 3.03638e-25], [1.393311, 3.008473e-25], [1.397596, 2.981131e-25], [1.401895, 2.954329e-25], [1.406207, 2.928043e-25], [1.410533, 2.902251e-25], [1.414871, 2.876931e-25], [1.419223, 2.852062e-25], [1.423588, 2.827624e-25], [1.427967, 2.8036e-25], [1.432359, 2.779973e-25], [1.436765, 2.756725e-25], [1.441185, 2.733842e-25], [1.445617, 2.711309e-25], [1.450064, 2.689113e-25], [1.454524, 2.66724e-25], [1.458998, 2.64568e-25], [1.463486, 2.624419e-25], [1.467987, 2.603447e-25], [1.472502, 2.582755e-25], [1.477032, 2.562332e-25], [1.481575, 2.542169e-25], [1.486132, 2.522258e-25], [1.490703, 2.502591e-25], [1.495288, 2.48316e-25], [1.499888, 2.463957e-25], [1.504501, 2.444977e-25], [1.509129, 2.426212e-25], [1.51377, 2.407657e-25], [1.518427, 2.389305e-25], [1.523097, 2.371152e-25], [1.527782, 2.353191e-25], [1.532481, 2.335418e-25], [1.537195, 2.317829e-25], [1.541923, 2.300418e-25], [1.546666, 2.283182e-25], [1.551423, 2.266116e-25], [1.556195, 2.249216e-25], [1.560982, 2.23248e-25], [1.565783, 2.215903e-25], [1.570599, 2.199482e-25], [1.57543, 2.183215e-25], [1.580276, 2.167097e-25], [1.585137, 2.151127e-25], [1.590012, 2.135301e-25], [1.594903, 2.119617e-25], [1.599809, 2.104072e-25], [1.604729, 2.088664e-25], [1.609665, 2.073391e-25], [1.614616, 2.058251e-25], [1.619583, 2.043241e-25], [1.624564, 2.028359e-25], [1.629561, 2.013604e-25], [1.634574, 1.998973e-25], [1.639601, 1.984465e-25], [1.644645, 1.970078e-25], [1.649703, 1.955811e-25], [1.654777, 1.941661e-25], [1.659867, 1.927628e-25], [1.664973, 1.913709e-25], [1.670094, 1.899904e-25], [1.675231, 1.88621e-25], [1.680384, 1.872627e-25], [1.685552, 1.859153e-25], [1.690737, 1.845787e-25], [1.695937, 1.832527e-25], [1.701154, 1.819373e-25], [1.706386, 1.806324e-25], [1.711635, 1.793377e-25], [1.7169, 1.780532e-25], [1.722181, 1.767789e-25], [1.727478, 1.755145e-25], [1.732791, 1.7426e-25], [1.738121, 1.730153e-25], [1.743467, 1.717803e-25], [1.74883, 1.705549e-25], [1.754209, 1.693389e-25], [1.759605, 1.681324e-25], [1.765017, 1.669351e-25], [1.770446, 1.657471e-25], [1.775892, 1.645683e-25], [1.781354, 1.633985e-25], [1.786833, 1.622376e-25], [1.792329, 1.610856e-25], [1.797842, 1.599425e-25], [1.803372, 1.588081e-25], [1.808919, 1.576823e-25], [1.814483, 1.56565e-25], [1.820064, 1.554563e-25], [1.825662, 1.54356e-25], [1.831278, 1.532641e-25], [1.836911, 1.521804e-25], [1.842561, 1.51105e-25], [1.848228, 1.500376e-25], [1.853913, 1.489784e-25], [1.859615, 1.479271e-25], [1.865335, 1.468838e-25], [1.871073, 1.458483e-25], [1.876828, 1.448206e-25], [1.882601, 1.438007e-25], [1.888392, 1.427884e-25], [1.8942, 1.417837e-25], [1.900026, 1.407866e-25], [1.90587, 1.39797e-25], [1.911733, 1.388148e-25], [1.917613, 1.378399e-25], [1.923511, 1.368724e-25], [1.929428, 1.359121e-25], [1.935362, 1.34959e-25], [1.941315, 1.34013e-25], [1.947286, 1.33074e-25], [1.953276, 1.321421e-25], [1.959284, 1.312172e-25], [1.96531, 1.302991e-25], [1.971355, 1.293879e-25], [1.977419, 1.284835e-25], [1.983501, 1.275859e-25], [1.989602, 1.266949e-25], [1.995722, 1.258105e-25], [2.00186, 1.249328e-25], [2.008018, 1.240615e-25], [2.014194, 1.231968e-25], [2.02039, 1.223384e-25], [2.026604, 1.214865e-25], [2.032838, 1.206409e-25], [2.03909, 1.198015e-25], [2.045362, 1.189684e-25], [2.051654, 1.181415e-25], [2.057964, 1.173207e-25], [2.064294, 1.165059e-25], [2.070644, 1.156973e-25], [2.077013, 1.148946e-25], [2.083401, 1.140978e-25], [2.089809, 1.13307e-25], [2.096237, 1.12522e-25], [2.102685, 1.117428e-25], [2.109153, 1.109693e-25], [2.11564, 1.102016e-25], [2.122148, 1.094396e-25], [2.128675, 1.086832e-25], [2.135223, 1.079324e-25], [2.14179, 1.071871e-25], [2.148378, 1.064473e-25], [2.154986, 1.05713e-25], [2.161615, 1.049841e-25], [2.168263, 1.042605e-25], [2.174933, 1.035423e-25], [2.181622, 1.028294e-25], [2.188333, 1.021218e-25], [2.195064, 1.014193e-25], [2.201815, 1.007221e-25], [2.208588, 1.000299e-25], [2.215381, 9.934287e-26], [2.222196, 9.866088e-26], [2.229031, 9.798389e-26], [2.235887, 9.731188e-26], [2.242764, 9.664481e-26], [2.249662, 9.598265e-26], [2.256582, 9.532534e-26], [2.263523, 9.467285e-26], [2.270485, 9.402516e-26], [2.277469, 9.338221e-26], [2.284474, 9.274398e-26], [2.291501, 9.211042e-26], [2.298549, 9.148152e-26], [2.305619, 9.085721e-26], [2.312711, 9.023748e-26], [2.319825, 8.962229e-26], [2.32696, 8.90116e-26], [2.334118, 8.840537e-26], [2.341297, 8.780358e-26], [2.348498, 8.72062e-26], [2.355722, 8.661317e-26], [2.362968, 8.602448e-26], [2.370236, 8.544008e-26], [2.377527, 8.485996e-26], [2.38484, 8.428406e-26], [2.392175, 8.371237e-26], [2.399533, 8.314485e-26], [2.406913, 8.258147e-26], [2.414317, 8.202219e-26], [2.421743, 8.146699e-26], [2.429192, 8.091582e-26], [2.436664, 8.036868e-26], [2.444159, 7.982551e-26], [2.451676, 7.928629e-26], [2.459217, 7.8751e-26], [2.466782, 7.821959e-26], [2.474369, 7.769205e-26], [2.48198, 7.716834e-26], [2.489614, 7.664843e-26], [2.497272, 7.613229e-26], [2.504953, 7.56199e-26], [2.512658, 7.511123e-26], [2.520387, 7.460624e-26], [2.528139, 7.410491e-26], [2.535915, 7.360721e-26], [2.543715, 7.311312e-26], [2.551539, 7.26226e-26], [2.559388, 7.213563e-26], [2.56726, 7.165218e-26], [2.575156, 7.117222e-26], [2.583077, 7.069573e-26], [2.591022, 7.022268e-26], [2.598992, 6.975305e-26], [2.606986, 6.928681e-26], [2.615005, 6.882393e-26], [2.623048, 6.836438e-26], [2.631116, 6.790815e-26], [2.639209, 6.74552e-26], [2.647327, 6.700551e-26], [2.65547, 6.655906e-26], [2.663638, 6.611582e-26], [2.671831, 6.567576e-26], [2.680049, 6.523887e-26], [2.688292, 6.480512e-26], [2.696561, 6.437448e-26], [2.704855, 6.394693e-26], [2.713175, 6.352244e-26], [2.721521, 6.3101e-26], [2.729891, 6.268257e-26], [2.738288, 6.226715e-26], [2.746711, 6.185469e-26], [2.755159, 6.144519e-26], [2.763634, 6.103861e-26], [2.772134, 6.063494e-26], [2.780661, 6.023416e-26], [2.789214, 5.983623e-26], [2.797793, 5.944114e-26], [2.806399, 5.904887e-26], [2.815031, 5.86594e-26], [2.823689, 5.82727e-26], [2.832375, 5.788875e-26], [2.841087, 5.750753e-26], [2.849826, 5.712903e-26], [2.858591, 5.675322e-26], [2.867384, 5.638007e-26], [2.876203, 5.600957e-26], [2.88505, 5.564171e-26], [2.893924, 5.527645e-26], [2.902826, 5.491378e-26], [2.911754, 5.455368e-26], [2.92071, 5.419613e-26], [2.929694, 5.38411e-26], [2.938705, 5.348859e-26], [2.947744, 5.313857e-26], [2.956811, 5.279102e-26], [2.965906, 5.244592e-26], [2.975029, 5.210326e-26], [2.98418, 5.176301e-26], [2.993358, 5.142516e-26], [3.002566, 5.108969e-26], [3.011801, 5.075657e-26], [3.021065, 5.04258e-26], [3.030357, 5.009736e-26], [3.039678, 4.977122e-26], [3.049028, 4.944737e-26], [3.058406, 4.912579e-26], [3.067813, 4.880646e-26], [3.07725, 4.848937e-26], [3.086715, 4.81745e-26], [3.096209, 4.786183e-26], [3.105733, 4.755135e-26], [3.115285, 4.724304e-26], [3.124868, 4.693688e-26], [3.134479, 4.663285e-26], [3.14412, 4.633095e-26], [3.153791, 4.603114e-26], [3.163492, 4.573343e-26], [3.173222, 4.543779e-26], [3.182983, 4.51442e-26], [3.192773, 4.485265e-26], [3.202594, 4.456312e-26], [3.212444, 4.427561e-26], [3.222325, 4.399008e-26], [3.232237, 4.370654e-26], [3.242179, 4.342496e-26], [3.252151, 4.314533e-26], [3.262154, 4.286762e-26], [3.272188, 4.259184e-26], [3.282253, 4.231797e-26], [3.292349, 4.204598e-26], [3.302476, 4.177586e-26], [3.312634, 4.150761e-26], [3.322823, 4.12412e-26], [3.333043, 4.097663e-26], [3.343295, 4.071388e-26], [3.353579, 4.045292e-26], [3.363894, 4.019376e-26], [3.374241, 3.993638e-26], [3.384619, 3.968076e-26], [3.39503, 3.94269e-26], [3.405473, 3.917477e-26], [3.415947, 3.892436e-26], [3.426454, 3.867567e-26], [3.436994, 3.842867e-26], [3.447565, 3.818336e-26], [3.458169, 3.793972e-26], [3.468806, 3.769774e-26], [3.479476, 3.745741e-26], [3.490178, 3.721871e-26], [3.500914, 3.698163e-26], [3.511682, 3.674617e-26], [3.522483, 3.65123e-26], [3.533318, 3.628002e-26], [3.544186, 3.604932e-26], [3.555087, 3.582017e-26], [3.566022, 3.559258e-26], [3.576991, 3.536653e-26], [3.587993, 3.5142e-26], [3.599029, 3.491899e-26], [3.6101, 3.469748e-26], [3.621204, 3.447747e-26], [3.632342, 3.425894e-26], [3.643515, 3.404188e-26], [3.654721, 3.382628e-26], [3.665963, 3.361212e-26], [3.677239, 3.339941e-26], [3.68855, 3.318812e-26], [3.699895, 3.297826e-26], [3.711275, 3.276979e-26], [3.722691, 3.256273e-26], [3.734141, 3.235704e-26], [3.745627, 3.215274e-26], [3.757148, 3.194979e-26], [3.768704, 3.174821e-26], [3.780296, 3.154796e-26], [3.791924, 3.134906e-26], [3.803587, 3.115147e-26], [3.815287, 3.09552e-26], [3.827022, 3.076024e-26], [3.838793, 3.056657e-26], [3.850601, 3.037419e-26], [3.862445, 3.018309e-26], [3.874325, 2.999325e-26], [3.886242, 2.980467e-26], [3.898195, 2.961734e-26], [3.910186, 2.943125e-26], [3.922213, 2.924639e-26], [3.934277, 2.906275e-26], [3.946378, 2.888032e-26], [3.958517, 2.869909e-26], [3.970693, 2.851906e-26], [3.982906, 2.834022e-26], [3.995157, 2.816255e-26], [4.007446, 2.798605e-26], [4.019772, 2.781072e-26], [4.032136, 2.763653e-26], [4.044538, 2.746349e-26], [4.056979, 2.729158e-26], [4.069458, 2.71208e-26], [4.081975, 2.695113e-26], [4.09453, 2.678258e-26], [4.107124, 2.661513e-26], [4.119757, 2.644878e-26], [4.132429, 2.628351e-26], [4.14514, 2.611932e-26], [4.15789, 2.59562e-26], [4.170679, 2.579415e-26], [4.183507, 2.563315e-26], [4.196375, 2.547319e-26], [4.209282, 2.531428e-26], [4.22223, 2.515641e-26], [4.235217, 2.499955e-26], [4.248244, 2.484372e-26], [4.26131, 2.46889e-26], [4.274418, 2.453508e-26], [4.287565, 2.438226e-26], [4.300753, 2.423042e-26], [4.313982, 2.407957e-26], [4.327251, 2.39297e-26], [4.340561, 2.378079e-26], [4.353912, 2.363285e-26], [4.367304, 2.348586e-26], [4.380737, 2.333982e-26], [4.394211, 2.319472e-26], [4.407728, 2.305056e-26], [4.421285, 2.290732e-26], [4.434884, 2.2765e-26], [4.448525, 2.262361e-26], [4.462208, 2.248312e-26], [4.475934, 2.234353e-26], [4.489701, 2.220484e-26], [4.50351, 2.206704e-26], [4.517363, 2.193012e-26], [4.531257, 2.179408e-26], [4.545195, 2.165891e-26], [4.559175, 2.152461e-26], [4.573199, 2.139116e-26], [4.587265, 2.125857e-26], [4.601375, 2.112683e-26], [4.615528, 2.099593e-26], [4.629725, 2.086586e-26], [4.643965, 2.073663e-26], [4.658249, 2.060821e-26], [4.672578, 2.048062e-26], [4.68695, 2.035384e-26], [4.701366, 2.022787e-26], [4.715827, 2.01027e-26], [4.730332, 1.997832e-26], [4.744882, 1.985474e-26], [4.759476, 1.973194e-26], [4.774116, 1.960992e-26], [4.7888, 1.948868e-26], [4.80353, 1.93682e-26], [4.818305, 1.924849e-26], [4.833126, 1.912954e-26], [4.847992, 1.901134e-26], [4.862903, 1.889389e-26], [4.877861, 1.877718e-26], [4.892864, 1.866121e-26], [4.907914, 1.854598e-26], [4.92301, 1.843147e-26], [4.938153, 1.831769e-26], [4.953342, 1.820463e-26], [4.968578, 1.809228e-26], [4.98386, 1.798064e-26], [4.99919, 1.78697e-26], [5.014567, 1.775946e-26], [5.029991, 1.764992e-26], [5.045462, 1.754107e-26], [5.060981, 1.74329e-26], [5.076548, 1.732541e-26], [5.092163, 1.721861e-26], [5.107826, 1.711247e-26], [5.123537, 1.7007e-26], [5.139296, 1.690219e-26], [5.155104, 1.679804e-26], [5.17096, 1.669455e-26], [5.186865, 1.65917e-26], [5.202819, 1.64895e-26], [5.218822, 1.638794e-26], [5.234875, 1.628702e-26], [5.250977, 1.618673e-26], [5.267128, 1.608707e-26], [5.283329, 1.598803e-26], [5.29958, 1.588962e-26], [5.31588, 1.579182e-26], [5.332231, 1.569463e-26], [5.348632, 1.559805e-26], [5.365084, 1.550207e-26], [5.381586, 1.54067e-26], [5.398139, 1.531192e-26], [5.414743, 1.521773e-26], [5.431398, 1.512413e-26], [5.448104, 1.503112e-26], [5.464862, 1.493869e-26], [5.481671, 1.484683e-26], [5.498532, 1.475554e-26], [5.515445, 1.466483e-26], [5.532409, 1.457468e-26], [5.549426, 1.448509e-26], [5.566496, 1.439607e-26], [5.583617, 1.430759e-26], [5.600792, 1.421967e-26], [5.618019, 1.413229e-26], [5.635299, 1.404546e-26], [5.652633, 1.395917e-26], [5.670019, 1.387342e-26], [5.687459, 1.37882e-26], [5.704953, 1.370351e-26], [5.722501, 1.361934e-26], [5.740102, 1.35357e-26], [5.757758, 1.345258e-26], [5.775468, 1.336998e-26], [5.793233, 1.328788e-26], [5.811052, 1.32063e-26], [5.828926, 1.312523e-26], [5.846855, 1.304466e-26], [5.864839, 1.296458e-26], [5.882878, 1.288501e-26], [5.900973, 1.280593e-26], [5.919124, 1.272734e-26], [5.93733, 1.264923e-26], [5.955592, 1.257162e-26], [5.973911, 1.249448e-26], [5.992286, 1.241782e-26], [6.010717, 1.234163e-26], [6.029205, 1.226592e-26], [6.047751, 1.219068e-26], [6.066352, 1.21159e-26], [6.085012, 1.204158e-26], [6.103728, 1.196773e-26], [6.122503, 1.189433e-26], [6.141335, 1.182139e-26], [6.160224, 1.17489e-26], [6.179172, 1.167685e-26], [6.198179, 1.160526e-26], [6.217243, 1.15341e-26], [6.236367, 1.146339e-26], [6.255549, 1.139311e-26], [6.27479, 1.132327e-26], [6.29409, 1.125385e-26], [6.31345, 1.118487e-26], [6.332869, 1.111632e-26], [6.352348, 1.104818e-26], [6.371887, 1.098047e-26], [6.391486, 1.091318e-26], [6.411146, 1.08463e-26], [6.430866, 1.077984e-26], [6.450646, 1.071378e-26], [6.470487, 1.064814e-26], [6.490389, 1.05829e-26], [6.510353, 1.051806e-26], [6.530378, 1.045362e-26], [6.550464, 1.038958e-26], [6.570613, 1.032593e-26], [6.590823, 1.026268e-26], [6.611095, 1.019982e-26], [6.63143, 1.013734e-26], [6.651827, 1.007525e-26], [6.672288, 1.001355e-26], [6.692811, 9.95222e-27], [6.713397, 9.891273e-27], [6.734046, 9.8307e-27], [6.754759, 9.770502e-27], [6.775536, 9.710675e-27], [6.796376, 9.651216e-27], [6.817281, 9.592123e-27], [6.83825, 9.533395e-27], [6.859284, 9.475029e-27], [6.880382, 9.417022e-27], [6.901545, 9.359373e-27], [6.922773, 9.302079e-27], [6.944066, 9.245137e-27], [6.965426, 9.188548e-27], [6.98685, 9.132306e-27], [7.008341, 9.07641e-27], [7.029897, 9.020859e-27], [7.05152, 8.96565e-27], [7.07321, 8.910781e-27], [7.094966, 8.85625e-27], [7.116789, 8.802054e-27], [7.138679, 8.748192e-27], [7.160637, 8.694661e-27], [7.182662, 8.641461e-27], [7.204755, 8.588588e-27], [7.226916, 8.53604e-27], [7.249145, 8.483815e-27], [7.271442, 8.431912e-27], [7.293808, 8.380328e-27], [7.316243, 8.329062e-27], [7.338746, 8.278111e-27], [7.361319, 8.227474e-27], [7.383962, 8.177148e-27], [7.406674, 8.127132e-27], [7.429456, 8.077424e-27], [7.452307, 8.028022e-27], [7.47523, 7.978923e-27], [7.498222, 7.930126e-27], [7.521286, 7.88163e-27], [7.54442, 7.833432e-27], [7.567626, 7.78553e-27], [7.590903, 7.737923e-27], [7.614251, 7.690609e-27], [7.637672, 7.643586e-27], [7.661164, 7.596852e-27], [7.684729, 7.550406e-27], [7.708366, 7.504245e-27], [7.732076, 7.458368e-27], [7.755859, 7.412773e-27], [7.779714, 7.367458e-27], [7.803644, 7.322422e-27], [7.827647, 7.277663e-27], [7.851723, 7.233179e-27], [7.875874, 7.188969e-27], [7.900099, 7.14503e-27], [7.924399, 7.101362e-27], [7.948773, 7.057962e-27], [7.973222, 7.014829e-27], [7.997747, 6.971962e-27], [8.022347, 6.929357e-27], [8.047022, 6.887015e-27], [8.071774, 6.844933e-27], [8.096602, 6.80311e-27], [8.121506, 6.761543e-27], [8.146486, 6.720233e-27], [8.171544, 6.679176e-27], [8.196678, 6.638372e-27], [8.22189, 6.597818e-27], [8.247179, 6.557514e-27], [8.272546, 6.517458e-27], [8.297992, 6.477647e-27], [8.323515, 6.438082e-27], [8.349117, 6.39876e-27], [8.374798, 6.359679e-27], [8.400557, 6.320839e-27], [8.426396, 6.282237e-27], [8.452315, 6.243873e-27], [8.478313, 6.205744e-27], [8.504391, 6.16785e-27], [8.530549, 6.130189e-27], [8.556788, 6.092759e-27], [8.583107, 6.05556e-27], [8.609508, 6.018589e-27], [8.63599, 5.981845e-27], [8.662553, 5.945327e-27], [8.689197, 5.909034e-27], [8.715924, 5.872964e-27], [8.742733, 5.837115e-27], [8.769624, 5.801487e-27], [8.796598, 5.766078e-27], [8.823656, 5.730886e-27], [8.850796, 5.695911e-27], [8.87802, 5.661151e-27], [8.905327, 5.626605e-27], [8.932719, 5.592271e-27], [8.960195, 5.558148e-27], [8.987755, 5.524234e-27], [9.0154, 5.49053e-27], [9.04313, 5.457032e-27], [9.070945, 5.42374e-27], [9.098846, 5.390653e-27], [9.126833, 5.35777e-27], [9.154906, 5.325089e-27], [9.183065, 5.292608e-27], [9.211311, 5.260327e-27], [9.239644, 5.228245e-27], [9.268063, 5.19636e-27], [9.296571, 5.164671e-27], [9.325166, 5.133177e-27], [9.353848, 5.101877e-27], [9.38262, 5.070768e-27], [9.411479, 5.039852e-27], [9.440428, 5.009125e-27], [9.469465, 4.978587e-27], [9.498592, 4.948238e-27], [9.527808, 4.918074e-27], [9.557114, 4.888097e-27], [9.58651, 4.858303e-27], [9.615997, 4.828693e-27], [9.645575, 4.799265e-27], [9.675243, 4.770018e-27], [9.705003, 4.740951e-27], [9.734854, 4.712062e-27], [9.764797, 4.683351e-27], [9.794832, 4.654817e-27], [9.82496, 4.626459e-27], [9.85518, 4.598274e-27], [9.885493, 4.570264e-27], [9.915899, 4.542425e-27], [9.946399, 4.514758e-27], [9.976993, 4.487261e-27], [10.007681, 4.459933e-27], [10.038463, 4.432773e-27], [10.06934, 4.405781e-27], [10.100312, 4.378954e-27], [10.131379, 4.352293e-27], [10.162542, 4.325795e-27], [10.1938, 4.299461e-27], [10.225155, 4.273288e-27], [10.256606, 4.247277e-27], [10.288154, 4.221426e-27], [10.319799, 4.195734e-27], [10.351541, 4.1702e-27], [10.383381, 4.144823e-27], [10.415319, 4.119602e-27], [10.447355, 4.094537e-27], [10.479489, 4.069626e-27], [10.511723, 4.044868e-27], [10.544055, 4.020263e-27], [10.576487, 3.995809e-27], [10.609019, 3.971506e-27], [10.641651, 3.947352e-27], [10.674383, 3.923347e-27], [10.707216, 3.89949e-27], [10.74015, 3.875779e-27], [10.773185, 3.852215e-27], [10.806322, 3.828796e-27], [10.83956, 3.805521e-27], [10.872901, 3.78239e-27], [10.906345, 3.759401e-27], [10.939891, 3.736553e-27], [10.973541, 3.713846e-27], [11.007294, 3.691279e-27], [11.041151, 3.668851e-27], [11.075112, 3.646562e-27], [11.109177, 3.624409e-27], [11.143347, 3.602394e-27], [11.177623, 3.580513e-27], [11.212004, 3.558768e-27], [11.24649, 3.537156e-27], [11.281083, 3.515678e-27], [11.315782, 3.494332e-27], [11.350587, 3.473118e-27], [11.3855, 3.452034e-27], [11.42052, 3.431081e-27], [11.455648, 3.410257e-27], [11.490884, 3.38956e-27], [11.526229, 3.368992e-27], [11.561682, 3.34855e-27], [11.597244, 3.328235e-27], [11.632915, 3.308045e-27], [11.668696, 3.287979e-27], [11.704588, 3.268037e-27], [11.740589, 3.248218e-27], [11.776702, 3.228521e-27], [11.812925, 3.208945e-27], [11.84926, 3.189491e-27], [11.885707, 3.170156e-27], [11.922265, 3.150941e-27], [11.958937, 3.131844e-27], [11.995721, 3.112866e-27], [12.032618, 3.094004e-27], [12.069628, 3.075258e-27], [12.106753, 3.056629e-27], [12.143991, 3.038114e-27], [12.181345, 3.019714e-27], [12.218813, 3.001427e-27], [12.256396, 2.983253e-27], [12.294095, 2.965191e-27], [12.33191, 2.947241e-27], [12.369841, 2.929402e-27], [12.407889, 2.911673e-27], [12.446054, 2.894053e-27], [12.484336, 2.876542e-27], [12.522736, 2.85914e-27], [12.561254, 2.841844e-27], [12.599891, 2.824656e-27], [12.638646, 2.807574e-27], [12.677521, 2.790598e-27], [12.716515, 2.773726e-27], [12.75563, 2.756959e-27], [12.794864, 2.740295e-27], [12.834219, 2.723734e-27], [12.873696, 2.707276e-27], [12.913293, 2.690919e-27], [12.953013, 2.674664e-27], [12.992854, 2.658509e-27], [13.032818, 2.642454e-27], [13.072906, 2.626498e-27], [13.113116, 2.610641e-27], [13.15345, 2.594882e-27], [13.193908, 2.579221e-27], [13.234491, 2.563656e-27], [13.275198, 2.548188e-27], [13.316031, 2.532815e-27], [13.356989, 2.517538e-27], [13.398073, 2.502355e-27], [13.439284, 2.487267e-27], [13.480621, 2.472271e-27], [13.522086, 2.457369e-27], [13.563678, 2.442558e-27], [13.605398, 2.42784e-27], [13.647246, 2.413213e-27], [13.689223, 2.398676e-27], [13.731329, 2.384229e-27], [13.773565, 2.369872e-27], [13.81593, 2.355604e-27], [13.858426, 2.341424e-27], [13.901052, 2.327332e-27], [13.94381, 2.313328e-27], [13.986699, 2.29941e-27], [14.02972, 2.285579e-27], [14.072874, 2.271833e-27], [14.11616, 2.258172e-27], [14.159579, 2.244597e-27], [14.203132, 2.231105e-27], [14.246819, 2.217698e-27], [14.29064, 2.204373e-27], [14.334596, 2.191131e-27], [14.378687, 2.177972e-27], [14.422914, 2.164894e-27], [14.467277, 2.151897e-27], [14.511776, 2.138981e-27], [14.556412, 2.126145e-27], [14.601186, 2.113389e-27], [14.646097, 2.100712e-27], [14.691146, 2.088114e-27], [14.736334, 2.075594e-27], [14.781661, 2.063152e-27], [14.827127, 2.050788e-27], [14.872734, 2.0385e-27], [14.91848, 2.026288e-27], [14.964367, 2.014153e-27], [15.010395, 2.002093e-27], [15.056565, 1.990108e-27], [15.102877, 1.978198e-27], [15.149331, 1.966361e-27], [15.195929, 1.954599e-27], [15.242669, 1.942909e-27], [15.289553, 1.931293e-27], [15.336582, 1.919748e-27], [15.383755, 1.908276e-27], [15.431073, 1.896875e-27], [15.478537, 1.885545e-27], [15.526147, 1.874286e-27], [15.573903, 1.863096e-27], [15.621806, 1.851977e-27], [15.669857, 1.840927e-27], [15.718055, 1.829946e-27], [15.766401, 1.819033e-27], [15.814897, 1.808188e-27], [15.863541, 1.797411e-27], [15.912335, 1.786701e-27], [15.961279, 1.776058e-27], [16.010374, 1.765482e-27], [16.059619, 1.754971e-27], [16.109016, 1.744527e-27], [16.158565, 1.734147e-27], [16.208267, 1.723832e-27], [16.258121, 1.713582e-27], [16.308129, 1.703396e-27], [16.35829, 1.693273e-27], [16.408606, 1.683214e-27], [16.459077, 1.673217e-27], [16.509702, 1.663284e-27], [16.560484, 1.653412e-27], [16.611422, 1.643602e-27], [16.662516, 1.633853e-27], [16.713768, 1.624166e-27], [16.765177, 1.614539e-27], [16.816744, 1.604972e-27], [16.86847, 1.595466e-27], [16.920355, 1.586019e-27], [16.9724, 1.576631e-27], [17.024604, 1.567302e-27], [17.07697, 1.558031e-27], [17.129496, 1.548819e-27], [17.182184, 1.539664e-27], [17.235034, 1.530567e-27], [17.288046, 1.521527e-27], [17.341222, 1.512544e-27], [17.394561, 1.503617e-27], [17.448064, 1.494746e-27], [17.501732, 1.485931e-27], [17.555565, 1.477171e-27], [17.609563, 1.468467e-27], [17.663728, 1.459817e-27], [17.718059, 1.451221e-27], [17.772557, 1.44268e-27], [17.827223, 1.434192e-27], [17.882057, 1.425758e-27], [17.93706, 1.417377e-27], [17.992231, 1.409048e-27], [18.047573, 1.400772e-27], [18.103085, 1.392549e-27], [18.158767, 1.384377e-27], [18.214621, 1.376256e-27], [18.270647, 1.368187e-27], [18.326845, 1.360169e-27], [18.383215, 1.352201e-27], [18.439759, 1.344284e-27], [18.496478, 1.336416e-27], [18.55337, 1.328598e-27], [18.610438, 1.32083e-27], [18.667681, 1.313111e-27], [18.7251, 1.30544e-27], [18.782696, 1.297818e-27], [18.840468, 1.290245e-27], [18.898419, 1.282719e-27], [18.956548, 1.275241e-27], [19.014856, 1.26781e-27], [19.073343, 1.260426e-27], [19.132009, 1.253089e-27], [19.190857, 1.245799e-27], [19.249885, 1.238555e-27], [19.309095, 1.231356e-27], [19.368487, 1.224204e-27], [19.428062, 1.217097e-27], [19.48782, 1.210034e-27], [19.547762, 1.203017e-27], [19.607888, 1.196045e-27], [19.668199, 1.189116e-27], [19.728695, 1.182232e-27], [19.789378, 1.175392e-27], [19.850247, 1.168595e-27], [19.911304, 1.161841e-27], [19.972548, 1.15513e-27], [20.033981, 1.148462e-27], [20.095603, 1.141837e-27], [20.157414, 1.135254e-27], [20.219415, 1.128713e-27], [20.281607, 1.122213e-27], [20.343991, 1.115755e-27], [20.406566, 1.109339e-27], [20.469333, 1.102963e-27], [20.532294, 1.096628e-27], [20.595449, 1.090333e-27], [20.658797, 1.084079e-27], [20.722341, 1.077865e-27], [20.78608, 1.071691e-27], [20.850015, 1.065556e-27], [20.914147, 1.05946e-27], [20.978475, 1.053404e-27], [21.043002, 1.047386e-27], [21.107728, 1.041407e-27], [21.172652, 1.035466e-27], [21.237776, 1.029564e-27], [21.3031, 1.023699e-27], [21.368626, 1.017872e-27], [21.434353, 1.012083e-27], [21.500282, 1.006331e-27], [21.566413, 1.000616e-27], [21.632749, 9.949374e-28], [21.699288, 9.892957e-28], [21.766032, 9.836903e-28], [21.832981, 9.781211e-28], [21.900136, 9.725878e-28], [21.967498, 9.670902e-28], [22.035067, 9.616281e-28], [22.102844, 9.562014e-28], [22.170829, 9.508097e-28], [22.239023, 9.454529e-28], [22.307427, 9.401308e-28], [22.376042, 9.348431e-28], [22.444867, 9.295897e-28], [22.513904, 9.243703e-28], [22.583154, 9.191848e-28], [22.652617, 9.14033e-28], [22.722293, 9.089146e-28], [22.792183, 9.038295e-28], [22.862289, 8.987774e-28], [22.93261, 8.937581e-28], [23.003147, 8.887716e-28], [23.073902, 8.838175e-28], [23.144874, 8.788957e-28], [23.216064, 8.74006e-28], [23.287474, 8.691481e-28], [23.359102, 8.64322e-28], [23.430952, 8.595273e-28], [23.503022, 8.547641e-28], [23.575314, 8.500319e-28], [23.647828, 8.453307e-28], [23.720566, 8.406603e-28], [23.793527, 8.360205e-28], [23.866712, 8.314111e-28], [23.940123, 8.268319e-28], [24.013759, 8.222828e-28], [24.087622, 8.177635e-28], [24.161712, 8.13274e-28], [24.23603, 8.088139e-28], [24.310577, 8.043832e-28], [24.385353, 7.999817e-28], [24.460359, 7.956092e-28], [24.535595, 7.912654e-28], [24.611063, 7.869504e-28], [24.686763, 7.826638e-28], [24.762696, 7.784056e-28], [24.838863, 7.741755e-28], [24.915264, 7.699734e-28], [24.9919, 7.65799e-28], [25.068771, 7.616524e-28], [25.145879, 7.575332e-28], [25.223224, 7.534414e-28], [25.300807, 7.493767e-28], [25.378629, 7.45339e-28], [25.45669, 7.413282e-28], [25.534991, 7.37344e-28], [25.613533, 7.333864e-28], [25.692317, 7.294551e-28], [25.771343, 7.255501e-28], [25.850612, 7.216711e-28], [25.930124, 7.17818e-28], [26.009882, 7.139906e-28], [26.089885, 7.101889e-28], [26.170133, 7.064125e-28], [26.250629, 7.026615e-28], [26.331372, 6.989356e-28], [26.412364, 6.952348e-28], [26.493604, 6.915587e-28], [26.575095, 6.879074e-28], [26.656836, 6.842806e-28], [26.738829, 6.806782e-28], [26.821074, 6.771001e-28], [26.903571, 6.735461e-28], [26.986323, 6.700161e-28], [27.069329, 6.665099e-28], [27.152591, 6.630274e-28], [27.236108, 6.595685e-28], [27.319882, 6.56133e-28], [27.403915, 6.527207e-28], [27.488205, 6.493316e-28], [27.572755, 6.459655e-28], [27.657565, 6.426222e-28], [27.742636, 6.393017e-28], [27.827968, 6.360037e-28], [27.913563, 6.327282e-28], [27.999421, 6.294751e-28], [28.085543, 6.262441e-28], [28.17193, 6.230352e-28], [28.258583, 6.198482e-28], [28.345503, 6.16683e-28], [28.432689, 6.135395e-28], [28.520144, 6.104176e-28], [28.607868, 6.07317e-28], [28.695862, 6.042377e-28], [28.784126, 6.011796e-28], [28.872662, 5.981425e-28], [28.96147, 5.951263e-28], [29.050552, 5.92131e-28], [29.139907, 5.891562e-28], [29.229537, 5.86202e-28], [29.319443, 5.832682e-28], [29.409626, 5.803547e-28], [29.500085, 5.774614e-28], [29.590824, 5.745881e-28], [29.681841, 5.717348e-28], [29.773138, 5.689013e-28], [29.864716, 5.660875e-28], [29.956575, 5.632932e-28], [30.048718, 5.605184e-28], [30.141143, 5.577629e-28], [30.233853, 5.550267e-28], [30.326848, 5.523096e-28], [30.420129, 5.496115e-28], [30.513697, 5.469323e-28], [30.607553, 5.442718e-28], [30.701697, 5.4163e-28], [30.796131, 5.390067e-28], [30.890856, 5.364019e-28], [30.985872, 5.338154e-28], [31.08118, 5.312471e-28], [31.176781, 5.28697e-28], [31.272677, 5.261648e-28], [31.368867, 5.236505e-28], [31.465353, 5.21154e-28], [31.562136, 5.186751e-28], [31.659217, 5.162139e-28], [31.756596, 5.1377e-28], [31.854275, 5.113436e-28], [31.952254, 5.089344e-28], [32.050535, 5.065423e-28], [32.149118, 5.041673e-28], [32.248004, 5.018092e-28], [32.347194, 4.99468e-28], [32.446689, 4.971435e-28], [32.546491, 4.948356e-28], [32.646599, 4.925442e-28], [32.747015, 4.902693e-28], [32.847741, 4.880108e-28], [32.948776, 4.857684e-28], [33.050121, 4.835422e-28], [33.151779, 4.81332e-28], [33.253749, 4.791378e-28], [33.356033, 4.769594e-28], [33.458631, 4.747967e-28], [33.561545, 4.726497e-28], [33.664776, 4.705182e-28], [33.768324, 4.684022e-28], [33.87219, 4.663015e-28], [33.976376, 4.642162e-28], [34.080883, 4.62146e-28], [34.185711, 4.600908e-28], [34.290861, 4.580507e-28], [34.396335, 4.560254e-28], [34.502133, 4.54015e-28], [34.608257, 4.520192e-28], [34.714707, 4.500381e-28], [34.821485, 4.480715e-28], [34.92859, 4.461194e-28], [35.036026, 4.441816e-28], [35.143792, 4.42258e-28], [35.251889, 4.403486e-28], [35.360319, 4.384534e-28], [35.469082, 4.365721e-28], [35.57818, 4.347047e-28], [35.687613, 4.328511e-28], [35.797383, 4.310113e-28], [35.907491, 4.291851e-28], [36.017937, 4.273724e-28], [36.128724, 4.255733e-28], [36.23985, 4.237875e-28], [36.351319, 4.220151e-28], [36.46313, 4.202559e-28], [36.575286, 4.185098e-28], [36.687786, 4.167768e-28], [36.800633, 4.150567e-28], [36.913826, 4.133496e-28], [37.027368, 4.116552e-28], [37.141259, 4.099736e-28], [37.2555, 4.083046e-28], [37.370093, 4.066482e-28], [37.485038, 4.050043e-28], [37.600336, 4.033728e-28], [37.71599, 4.017536e-28], [37.831999, 4.001467e-28], [37.948365, 3.985519e-28], [38.065088, 3.969692e-28], [38.182171, 3.953986e-28], [38.299614, 3.938398e-28], [38.417418, 3.92293e-28], [38.535585, 3.907579e-28], [38.654115, 3.892345e-28], [38.773009, 3.877228e-28], [38.89227, 3.862226e-28], [39.011897, 3.847338e-28], [39.131892, 3.832565e-28], [39.252256, 3.817905e-28], [39.37299, 3.803358e-28], [39.494096, 3.788923e-28], [39.615574, 3.774598e-28], [39.737426, 3.760384e-28], [39.859653, 3.74628e-28], [39.982255, 3.732284e-28], [40.105235, 3.718397e-28], [40.228593, 3.704617e-28], [40.352331, 3.690943e-28], [40.476449, 3.677376e-28], [40.600948, 3.663914e-28], [40.725831, 3.650557e-28], [40.851098, 3.637303e-28], [40.97675, 3.624153e-28], [41.102789, 3.611106e-28], [41.229215, 3.59816e-28], [41.35603, 3.585315e-28], [41.483236, 3.572571e-28], [41.610832, 3.559927e-28], [41.738821, 3.547381e-28], [41.867204, 3.534934e-28], [41.995982, 3.522585e-28], [42.125155, 3.510333e-28], [42.254726, 3.498178e-28], [42.384696, 3.486118e-28], [42.515065, 3.474153e-28], [42.645835, 3.462283e-28], [42.777008, 3.450507e-28], [42.908584, 3.438823e-28], [43.040565, 3.427233e-28], [43.172951, 3.415734e-28], [43.305745, 3.404326e-28], [43.438947, 3.393009e-28], [43.572559, 3.381782e-28], [43.706582, 3.370645e-28], [43.841018, 3.359596e-28], [43.975866, 3.348635e-28], [44.11113, 3.337762e-28], [44.24681, 3.326976e-28], [44.382906, 3.316275e-28], [44.519422, 3.305661e-28], [44.656357, 3.295132e-28], [44.793714, 3.284687e-28], [44.931493, 3.274325e-28], [45.069696, 3.264048e-28], [45.208324, 3.253852e-28], [45.347378, 3.243739e-28], [45.48686, 3.233708e-28], [45.626772, 3.223757e-28], [45.767113, 3.213886e-28], [45.907886, 3.204096e-28], [46.049092, 3.194384e-28], [46.190733, 3.184751e-28], [46.332809, 3.175196e-28], [46.475322, 3.165719e-28], [46.618273, 3.156318e-28], [46.761665, 3.146994e-28], [46.905497, 3.137745e-28], [47.049772, 3.128572e-28], [47.19449, 3.119473e-28], [47.339654, 3.110449e-28], [47.485264, 3.101498e-28], [47.631321, 3.09262e-28], [47.777829, 3.083815e-28], [47.924786, 3.075081e-28], [48.072196, 3.066419e-28], [48.22006, 3.057828e-28], [48.368378, 3.049308e-28], [48.517152, 3.040857e-28], [48.666384, 3.032475e-28], [48.816075, 3.024162e-28], [48.966226, 3.015918e-28], [49.116839, 3.007741e-28], [49.267916, 2.999632e-28], [49.419457, 2.991589e-28], [49.571464, 2.983612e-28], [49.723939, 2.975702e-28], [49.876882, 2.967856e-28], [50.030297, 2.960076e-28], [50.184183, 2.952359e-28], [50.338542, 2.944707e-28], [50.493376, 2.937117e-28], [50.648687, 2.929591e-28], [50.804475, 2.922126e-28], [50.960742, 2.914724e-28], [51.11749, 2.907383e-28], [51.27472, 2.900103e-28], [51.432434, 2.892883e-28], [51.590633, 2.885723e-28], [51.749319, 2.878623e-28], [51.908492, 2.871582e-28], [52.068155, 2.864599e-28], [52.228309, 2.857674e-28], [52.388956, 2.850807e-28], [52.550097, 2.843998e-28], [52.711734, 2.837245e-28], [52.873868, 2.830548e-28], [53.0365, 2.823907e-28], [53.199633, 2.817322e-28], [53.363268, 2.810792e-28], [53.527405, 2.804316e-28], [53.692048, 2.797894e-28], [53.857197, 2.791526e-28], [54.022854, 2.785212e-28], [54.189021, 2.77895e-28], [54.355699, 2.772741e-28], [54.522889, 2.766583e-28], [54.690594, 2.760478e-28], [54.858814, 2.754423e-28], [55.027552, 2.748419e-28], [55.196809, 2.742466e-28], [55.366587, 2.736563e-28], [55.536886, 2.730709e-28], [55.70771, 2.724904e-28], [55.879059, 2.719148e-28], [56.050935, 2.713441e-28], [56.22334, 2.707782e-28], [56.396275, 2.70217e-28], [56.569742, 2.696605e-28], [56.743742, 2.691087e-28], [56.918278, 2.685616e-28], [57.09335, 2.680191e-28], [57.268961, 2.674811e-28], [57.445113, 2.669477e-28], [57.621806, 2.664188e-28], [57.799042, 2.658943e-28], [57.976824, 2.653743e-28], [58.155152, 2.648586e-28], [58.334029, 2.643473e-28], [58.513456, 2.638403e-28], [58.693435, 2.633376e-28], [58.873968, 2.628392e-28], [59.055056, 2.623449e-28], [59.236701, 2.618548e-28], [59.418905, 2.613689e-28], [59.601669, 2.608871e-28], [59.784995, 2.604093e-28], [59.968885, 2.599356e-28], [60.153341, 2.594659e-28], [60.338364, 2.590002e-28], [60.523956, 2.585384e-28], [60.710119, 2.580805e-28], [60.896855, 2.576265e-28], [61.084165, 2.571763e-28], [61.272051, 2.567299e-28], [61.460516, 2.562874e-28], [61.649559, 2.558485e-28], [61.839185, 2.554134e-28], [62.029393, 2.54982e-28], [62.220187, 2.545542e-28], [62.411567, 2.541301e-28], [62.603536, 2.537095e-28], [62.796096, 2.532925e-28], [62.989248, 2.528791e-28], [63.182994, 2.524691e-28], [63.377336, 2.520627e-28], [63.572275, 2.516597e-28], [63.767815, 2.5126e-28], [63.963955, 2.508638e-28], [64.160699, 2.50471e-28], [64.358049, 2.500815e-28], [64.556005, 2.496952e-28], [64.75457, 2.493123e-28], [64.953746, 2.489326e-28], [65.153534, 2.485562e-28], [65.353937, 2.481829e-28], [65.554957, 2.478128e-28], [65.756594, 2.474458e-28], [65.958852, 2.47082e-28], [66.161732, 2.467212e-28], [66.365237, 2.463636e-28], [66.569367, 2.460089e-28], [66.774125, 2.456573e-28], [66.979512, 2.453086e-28], [67.185532, 2.449629e-28], [67.392185, 2.446202e-28], [67.599474, 2.442804e-28], [67.8074, 2.439434e-28], [68.015966, 2.436093e-28], [68.225173, 2.432781e-28], [68.435024, 2.429497e-28], [68.645521, 2.426241e-28], [68.856665, 2.423012e-28], [69.068458, 2.419811e-28], [69.280903, 2.416637e-28], [69.494001, 2.41349e-28], [69.707755, 2.41037e-28], [69.922166, 2.407277e-28], [70.137236, 2.40421e-28], [70.352969, 2.401169e-28], [70.569364, 2.398154e-28], [70.786426, 2.395165e-28], [71.004155, 2.392201e-28], [71.222553, 2.389262e-28], [71.441624, 2.386349e-28], [71.661368, 2.38346e-28], [71.881788, 2.380596e-28], [72.102886, 2.377756e-28], [72.324665, 2.374941e-28], [72.547125, 2.37215e-28], [72.77027, 2.369382e-28], [72.994101, 2.366638e-28], [73.21862, 2.363918e-28], [73.44383, 2.361221e-28], [73.669733, 2.358547e-28], [73.896331, 2.355896e-28], [74.123625, 2.353267e-28], [74.351619, 2.350661e-28], [74.580314, 2.348078e-28], [74.809713, 2.345516e-28], [75.039817, 2.342977e-28], [75.270628, 2.340459e-28], [75.50215, 2.337963e-28], [75.734384, 2.335488e-28], [75.967332, 2.333035e-28], [76.200997, 2.330603e-28], [76.43538, 2.328191e-28], [76.670485, 2.325801e-28], [76.906312, 2.323431e-28], [77.142865, 2.321081e-28], [77.380146, 2.318752e-28], [77.618156, 2.316442e-28], [77.856898, 2.314153e-28], [78.096375, 2.311883e-28], [78.336589, 2.309633e-28], [78.577541, 2.307403e-28], [78.819234, 2.305192e-28], [79.061671, 2.303e-28], [79.304853, 2.300827e-28], [79.548784, 2.298673e-28], [79.793465, 2.296537e-28], [80.038898, 2.29442e-28], [80.285086, 2.292322e-28], [80.532032, 2.290242e-28], [80.779737, 2.28818e-28], [81.028204, 2.286136e-28], [81.277435, 2.28411e-28], [81.527433, 2.282101e-28], [81.7782, 2.28011e-28], [82.029738, 2.278137e-28], [82.28205, 2.276181e-28], [82.535138, 2.274242e-28], [82.789004, 2.27232e-28], [83.043651, 2.270415e-28], [83.299082, 2.268527e-28], [83.555298, 2.266656e-28], [83.812302, 2.264801e-28], [84.070097, 2.262963e-28], [84.328684, 2.26114e-28], [84.588067, 2.259335e-28], [84.848248, 2.257545e-28], [85.10923, 2.255771e-28], [85.371013, 2.254013e-28], [85.633603, 2.25227e-28], [85.896999, 2.250544e-28], [86.161206, 2.248832e-28], [86.426226, 2.247137e-28], [86.692061, 2.245456e-28], [86.958713, 2.243791e-28], [87.226186, 2.24214e-28], [87.494481, 2.240505e-28], [87.763602, 2.238884e-28], [88.03355, 2.237278e-28], [88.304329, 2.235687e-28], [88.57594, 2.234111e-28], [88.848387, 2.232548e-28], [89.121672, 2.231001e-28], [89.395798, 2.229467e-28], [89.670767, 2.227947e-28], [89.946581, 2.226442e-28], [90.223244, 2.224951e-28], [90.500758, 2.223473e-28], [90.779126, 2.222009e-28], [91.058349, 2.220559e-28], [91.338432, 2.219122e-28], [91.619376, 2.217699e-28], [91.901184, 2.216289e-28], [92.183859, 2.214893e-28], [92.467404, 2.21351e-28], [92.75182, 2.21214e-28], [93.037112, 2.210783e-28], [93.32328, 2.209439e-28], [93.61033, 2.208108e-28], [93.898262, 2.20679e-28], [94.187079, 2.205484e-28], [94.476786, 2.204192e-28], [94.767383, 2.202911e-28], [95.058874, 2.201644e-28], [95.351261, 2.200388e-28], [95.644548, 2.199145e-28], [95.938737, 2.197915e-28], [96.233831, 2.196696e-28], [96.529833, 2.19549e-28], [96.826745, 2.194296e-28], [97.12457, 2.193114e-28], [97.423312, 2.191943e-28], [97.722972, 2.190785e-28], [98.023554, 2.189638e-28], [98.32506, 2.188503e-28], [98.627494, 2.18738e-28], [98.930858, 2.186268e-28], [99.235156, 2.185168e-28], [99.540389, 2.18408e-28], [99.846561, 2.183002e-28], [100.15367, 2.181936e-28], [100.46173, 2.180882e-28], [100.77074, 2.179838e-28], [101.0807, 2.178806e-28], [101.39161, 2.177785e-28], [101.70347, 2.176775e-28], [102.0163, 2.175776e-28], [102.33008, 2.174788e-28], [102.64484, 2.173811e-28], [102.96056, 2.172844e-28], [103.27725, 2.171889e-28], [103.59492, 2.170944e-28], [103.91356, 2.17001e-28], [104.23318, 2.169086e-28], [104.55379, 2.168173e-28], [104.87538, 2.167271e-28], [105.19796, 2.166379e-28], [105.52154, 2.165497e-28], [105.84611, 2.164626e-28], [106.17168, 2.163765e-28], [106.49824, 2.162915e-28], [106.82582, 2.162074e-28], [107.1544, 2.161244e-28], [107.48399, 2.160424e-28], [107.8146, 2.159615e-28], [108.14622, 2.158815e-28], [108.47886, 2.158025e-28], [108.81253, 2.157246e-28], [109.14722, 2.156476e-28], [109.48294, 2.155716e-28], [109.81969, 2.154966e-28], [110.15748, 2.154226e-28], [110.49631, 2.153496e-28], [110.83618, 2.152775e-28], [111.1771, 2.152065e-28], [111.51906, 2.151363e-28], [111.86208, 2.150672e-28], [112.20615, 2.14999e-28], [112.55128, 2.149318e-28], [112.89747, 2.148656e-28], [113.24473, 2.148002e-28], [113.59306, 2.147359e-28], [113.94245, 2.146725e-28], [114.29292, 2.1461e-28], [114.64447, 2.145485e-28], [114.9971, 2.144879e-28], [115.35082, 2.144282e-28], [115.70562, 2.143695e-28], [116.06151, 2.143117e-28], [116.4185, 2.142549e-28], [116.77659, 2.141989e-28], [117.13578, 2.141439e-28], [117.49607, 2.140898e-28], [117.85747, 2.140366e-28], [118.21998, 2.139844e-28], [118.58361, 2.13933e-28], [118.94836, 2.138826e-28], [119.31423, 2.13833e-28], [119.68122, 2.137844e-28], [120.04934, 2.137367e-28], [120.4186, 2.136898e-28], [120.78899, 2.136439e-28], [121.16052, 2.135989e-28], [121.53319, 2.135547e-28], [121.90701, 2.135115e-28], [122.28198, 2.134692e-28], [122.6581, 2.134277e-28], [123.03538, 2.133871e-28], [123.41382, 2.133474e-28], [123.79342, 2.133086e-28], [124.17419, 2.132707e-28], [124.55613, 2.132337e-28], [124.93925, 2.131975e-28], [125.32355, 2.131623e-28], [125.70902, 2.131279e-28], [126.09569, 2.130944e-28], [126.48354, 2.130617e-28], [126.87259, 2.130299e-28], [127.26283, 2.129991e-28], [127.65427, 2.12969e-28], [128.04692, 2.129399e-28], [128.44077, 2.129116e-28], [128.83584, 2.128842e-28], [129.23212, 2.128576e-28], [129.62962, 2.12832e-28], [130.02834, 2.128071e-28], [130.42829, 2.127832e-28], [130.82947, 2.127601e-28], [131.23188, 2.127379e-28], [131.63553, 2.127165e-28], [132.04042, 2.12696e-28], [132.44656, 2.126764e-28], [132.85395, 2.126576e-28], [133.26259, 2.126397e-28], [133.67248, 2.126227e-28], [134.08364, 2.126065e-28], [134.49606, 2.125911e-28], [134.90975, 2.125767e-28], [135.32472, 2.125631e-28], [135.74096, 2.125503e-28], [136.15848, 2.125384e-28], [136.57728, 2.125274e-28], [136.99737, 2.125172e-28], [137.41876, 2.125079e-28], [137.84144, 2.124994e-28], [138.26542, 2.124918e-28], [138.6907, 2.12485e-28], [139.1173, 2.124791e-28], [139.5452, 2.124741e-28], [139.97442, 2.124699e-28], [140.40496, 2.124666e-28], [140.83683, 2.124641e-28], [141.27002, 2.124625e-28], [141.70455, 2.124618e-28], [142.14041, 2.124619e-28], [142.57762, 2.124629e-28], [143.01617, 2.124647e-28], [143.45606, 2.124674e-28], [143.89731, 2.124709e-28], [144.33992, 2.124754e-28], [144.78389, 2.124806e-28], [145.22922, 2.124868e-28], [145.67593, 2.124938e-28], [146.12401, 2.125016e-28], [146.57346, 2.125103e-28], [147.0243, 2.125199e-28], [147.47653, 2.125304e-28], [147.93015, 2.125417e-28], [148.38516, 2.125539e-28], [148.84157, 2.12567e-28], [149.29939, 2.125809e-28], [149.75861, 2.125957e-28], [150.21925, 2.126113e-28], [150.6813, 2.126279e-28], [151.14477, 2.126453e-28], [151.60967, 2.126635e-28], [152.076, 2.126827e-28], [152.54377, 2.127027e-28], [153.01297, 2.127236e-28], [153.48362, 2.127454e-28], [153.95571, 2.127681e-28], [154.42926, 2.127916e-28], [154.90426, 2.12816e-28], [155.38072, 2.128413e-28], [155.85865, 2.128675e-28], [156.33805, 2.128946e-28], [156.81892, 2.129225e-28], [157.30128, 2.129513e-28], [157.78511, 2.129811e-28], [158.27044, 2.130117e-28], [158.75725, 2.130432e-28], [159.24557, 2.130756e-28], [159.73539, 2.131089e-28], [160.22671, 2.131431e-28], [160.71954, 2.131782e-28], [161.21389, 2.132142e-28], [161.70976, 2.132511e-28], [162.20716, 2.132889e-28], [162.70609, 2.133276e-28], [163.20655, 2.133672e-28], [163.70855, 2.134077e-28], [164.21209, 2.134491e-28], [164.71719, 2.134915e-28], [165.22383, 2.135347e-28], [165.73204, 2.135789e-28], [166.24181, 2.136239e-28], [166.75314, 2.136699e-28], [167.26605, 2.137169e-28], [167.78054, 2.137647e-28], [168.29661, 2.138135e-28], [168.81426, 2.138632e-28], [169.33351, 2.139138e-28], [169.85436, 2.139653e-28], [170.37681, 2.140178e-28], [170.90086, 2.140712e-28], [171.42653, 2.141256e-28], [171.95381, 2.141809e-28], [172.48272, 2.142371e-28], [173.01325, 2.142943e-28], [173.54541, 2.143524e-28], [174.07921, 2.144115e-28], [174.61466, 2.144715e-28], [175.15175, 2.145325e-28], [175.69049, 2.145944e-28], [176.23089, 2.146573e-28], [176.77295, 2.147212e-28], [177.31668, 2.14786e-28], [177.86208, 2.148518e-28], [178.40916, 2.149185e-28], [178.95792, 2.149862e-28], [179.50837, 2.150549e-28], [180.06051, 2.151246e-28], [180.61435, 2.151952e-28], [181.16989, 2.152668e-28], [181.72715, 2.153394e-28], [182.28611, 2.15413e-28], [182.8468, 2.154876e-28], [183.40921, 2.155632e-28], [183.97335, 2.156397e-28], [184.53923, 2.157173e-28], [185.10684, 2.157958e-28], [185.67621, 2.158754e-28], [186.24732, 2.15956e-28], [186.82019, 2.160375e-28], [187.39482, 2.161201e-28], [187.97122, 2.162037e-28], [188.54939, 2.162883e-28], [189.12935, 2.163739e-28], [189.71108, 2.164605e-28], [190.2946, 2.165482e-28], [190.87992, 2.166369e-28], [191.46704, 2.167266e-28], [192.05597, 2.168174e-28], [192.64671, 2.169092e-28], [193.23926, 2.17002e-28], [193.83364, 2.170959e-28], [194.42984, 2.171908e-28], [195.02788, 2.172868e-28], [195.62776, 2.173838e-28], [196.22948, 2.174819e-28], [196.83305, 2.175811e-28], [197.43849, 2.176813e-28], [198.04578, 2.177825e-28], [198.65494, 2.178849e-28], [199.26597, 2.179883e-28], [199.87889, 2.180928e-28], [200.49369, 2.181983e-28], [201.11038, 2.18305e-28], [201.72896, 2.184127e-28], [202.34945, 2.185215e-28], [202.97185, 2.186314e-28], [203.59616, 2.187424e-28], [204.2224, 2.188545e-28], [204.85056, 2.189677e-28], [205.48065, 2.19082e-28], [206.11268, 2.191974e-28], [206.74665, 2.19314e-28], [207.38257, 2.194316e-28], [208.02045, 2.195504e-28], [208.66029, 2.196703e-28], [209.3021, 2.197913e-28], [209.94589, 2.199134e-28], [210.59165, 2.200367e-28], [211.2394, 2.201611e-28], [211.88914, 2.202867e-28], [212.54088, 2.204134e-28], [213.19463, 2.205413e-28], [213.85039, 2.206703e-28], [214.50816, 2.208004e-28], [215.16796, 2.209318e-28], [215.82978, 2.210642e-28], [216.49364, 2.211979e-28], [217.15955, 2.213327e-28], [217.8275, 2.214687e-28], [218.4975, 2.216059e-28], [219.16957, 2.217443e-28], [219.84371, 2.218838e-28], [220.51991, 2.220246e-28], [221.1982, 2.221665e-28], [221.87858, 2.223097e-28], [222.56104, 2.22454e-28], [223.24561, 2.225996e-28], [223.93228, 2.227463e-28], [224.62107, 2.228943e-28], [225.31197, 2.230435e-28], [226.005, 2.231939e-28], [226.70016, 2.233456e-28], [227.39745, 2.234985e-28], [228.09689, 2.236526e-28], [228.79849, 2.238079e-28], [229.50224, 2.239645e-28], [230.20816, 2.241224e-28], [230.91624, 2.242815e-28], [231.62651, 2.244419e-28], [232.33896, 2.246035e-28], [233.0536, 2.247664e-28], [233.77044, 2.249305e-28], [234.48949, 2.25096e-28], [235.21074, 2.252627e-28], [235.93422, 2.254307e-28], [236.65992, 2.255999e-28], [237.38785, 2.257705e-28], [238.11802, 2.259424e-28], [238.85044, 2.261155e-28], [239.58511, 2.2629e-28], [240.32204, 2.264658e-28], [241.06123, 2.266429e-28], [241.8027, 2.268213e-28], [242.54646, 2.27001e-28], [243.29249, 2.271821e-28], [244.04083, 2.273645e-28], [244.79146, 2.275482e-28], [245.54441, 2.277333e-28], [246.29967, 2.279197e-28], [247.05725, 2.281075e-28], [247.81716, 2.282966e-28], [248.57941, 2.284871e-28], [249.34401, 2.28679e-28], [250.11095, 2.288722e-28], [250.88026, 2.290668e-28], [251.65193, 2.292627e-28], [252.42598, 2.294601e-28], [253.2024, 2.296588e-28], [253.98122, 2.298589e-28], [254.76243, 2.300605e-28], [255.54604, 2.302634e-28], [256.33206, 2.304677e-28], [257.1205, 2.306735e-28], [257.91137, 2.308806e-28], [258.70467, 2.310892e-28], [259.50041, 2.312992e-28], [260.29859, 2.315107e-28], [261.09924, 2.317235e-28], [261.90234, 2.319378e-28], [262.70792, 2.321536e-28], [263.51597, 2.323708e-28], [264.3265, 2.325895e-28], [265.13954, 2.328096e-28], [265.95507, 2.330312e-28], [266.77311, 2.332543e-28], [267.59366, 2.334788e-28], [268.41674, 2.337048e-28], [269.24236, 2.339323e-28], [270.07051, 2.341613e-28], [270.90121, 2.343918e-28], [271.73446, 2.346238e-28], [272.57028, 2.348573e-28], [273.40866, 2.350923e-28], [274.24963, 2.353288e-28], [275.09318, 2.355668e-28], [275.93933, 2.358064e-28], [276.78808, 2.360475e-28], [277.63944, 2.362902e-28], [278.49342, 2.365343e-28], [279.35003, 2.367801e-28], [280.20927, 2.370273e-28], [281.07115, 2.372762e-28], [281.93569, 2.375266e-28], [282.80288, 2.377785e-28], [283.67274, 2.380321e-28], [284.54528, 2.382872e-28], [285.4205, 2.385439e-28], [286.29841, 2.388022e-28], [287.17903, 2.390621e-28], [288.06235, 2.393236e-28], [288.94839, 2.395866e-28], [289.83715, 2.398514e-28], [290.72865, 2.401177e-28], [291.62289, 2.403856e-28], [292.51988, 2.406552e-28], [293.41963, 2.409264e-28], [294.32215, 2.411992e-28], [295.22744, 2.414737e-28], [296.13552, 2.417498e-28], [297.04639, 2.420276e-28], [297.96006, 2.42307e-28], [298.87654, 2.425882e-28], [299.79584, 2.428709e-28], [300.71797, 2.431554e-28], [301.64294, 2.434415e-28], [302.57075, 2.437294e-28], [303.50142, 2.440189e-28], [304.43494, 2.443101e-28], [305.37134, 2.44603e-28], [306.31062, 2.448977e-28], [307.25279, 2.45194e-28], [308.19785, 2.454921e-28], [309.14583, 2.457919e-28], [310.09672, 2.460934e-28], [311.05053, 2.463967e-28], [312.00728, 2.467017e-28], [312.96697, 2.470085e-28], [313.92961, 2.47317e-28], [314.89521, 2.476273e-28], [315.86378, 2.479393e-28], [316.83534, 2.482531e-28], [317.80988, 2.485688e-28], [318.78741, 2.488861e-28], [319.76796, 2.492053e-28], [320.75152, 2.495263e-28], [321.73811, 2.498491e-28], [322.72773, 2.501737e-28], [323.72039, 2.505001e-28], [324.71611, 2.508283e-28], [325.71489, 2.511583e-28], [326.71674, 2.514902e-28], [327.72168, 2.518239e-28], [328.7297, 2.521595e-28], [329.74083, 2.524969e-28], [330.75506, 2.528361e-28], [331.77242, 2.531773e-28], [332.7929, 2.535203e-28], [333.81653, 2.538651e-28], [334.8433, 2.542119e-28], [335.87323, 2.545605e-28], [336.90633, 2.54911e-28], [337.94261, 2.552634e-28], [338.98207, 2.556178e-28], [340.02473, 2.55974e-28], [341.0706, 2.563322e-28], [342.11968, 2.566922e-28], [343.17199, 2.570542e-28], [344.22754, 2.574182e-28], [345.28634, 2.577841e-28], [346.34839, 2.581519e-28], [347.41371, 2.585216e-28], [348.4823, 2.588934e-28], [349.55418, 2.592671e-28], [350.62936, 2.596428e-28], [351.70785, 2.600204e-28], [352.78965, 2.604001e-28], [353.87478, 2.607817e-28], [354.96325, 2.611653e-28], [356.05507, 2.615509e-28], [357.15024, 2.619386e-28], [358.24879, 2.623282e-28], [359.35071, 2.627199e-28], [360.45602, 2.631136e-28], [361.56473, 2.635093e-28], [362.67685, 2.639071e-28], [363.7924, 2.643069e-28], [364.91137, 2.647088e-28], [366.03379, 2.651127e-28], [367.15965, 2.655187e-28], [368.28898, 2.659268e-28], [369.42179, 2.663369e-28], [370.55808, 2.667492e-28], [371.69786, 2.671635e-28], [372.84115, 2.675799e-28], [373.98796, 2.679985e-28], [375.13829, 2.684191e-28], [376.29216, 2.688419e-28], [377.44959, 2.692667e-28], [378.61057, 2.696938e-28], [379.77512, 2.701229e-28], [380.94325, 2.705542e-28], [382.11498, 2.709877e-28], [383.29031, 2.714233e-28], [384.46926, 2.71861e-28], [385.65183, 2.72301e-28], [386.83804, 2.727431e-28], [388.0279, 2.731874e-28], [389.22142, 2.736339e-28], [390.41861, 2.740826e-28], [391.61948, 2.745335e-28], [392.82405, 2.749866e-28], [394.03232, 2.754419e-28], [395.24431, 2.758995e-28], [396.46002, 2.763592e-28], [397.67948, 2.768212e-28], [398.90268, 2.772855e-28], [400.12965, 2.77752e-28], [401.36039, 2.782208e-28], [402.59492, 2.786918e-28], [403.83324, 2.791651e-28], [405.07538, 2.796407e-28], [406.32133, 2.801185e-28], [407.57112, 2.805987e-28], [408.82475, 2.810811e-28], [410.08224, 2.815659e-28], [411.34359, 2.82053e-28], [412.60883, 2.825424e-28], [413.87795, 2.830341e-28], [415.15098, 2.835281e-28], [416.42793, 2.840245e-28], [417.7088, 2.845232e-28], [418.99361, 2.850243e-28], [420.28238, 2.855278e-28], [421.57511, 2.860336e-28], [422.87181, 2.865418e-28], [424.17251, 2.870524e-28], [425.4772, 2.875653e-28], [426.78591, 2.880807e-28], [428.09864, 2.885984e-28], [429.41541, 2.891186e-28], [430.73623, 2.896412e-28], [432.06111, 2.901662e-28], [433.39007, 2.906937e-28], [434.72312, 2.912235e-28], [436.06027, 2.917559e-28], [437.40153, 2.922907e-28], [438.74691, 2.928279e-28], [440.09643, 2.933676e-28], [441.45011, 2.939098e-28], [442.80795, 2.944545e-28], [444.16996, 2.950016e-28], [445.53616, 2.955513e-28], [446.90657, 2.961034e-28], [448.28119, 2.966581e-28], [449.66004, 2.972152e-28], [451.04313, 2.97775e-28], [452.43047, 2.983372e-28], [453.82209, 2.98902e-28], [455.21798, 2.994693e-28], [456.61816, 3.000392e-28], [458.02266, 3.006116e-28], [459.43147, 3.011866e-28], [460.84462, 3.017642e-28], [462.26211, 3.023444e-28], [463.68396, 3.029271e-28], [465.11019, 3.035125e-28], [466.5408, 3.041005e-28], [467.97581, 3.04691e-28], [469.41524, 3.052842e-28], [470.8591, 3.058801e-28], [472.30739, 3.064785e-28], [473.76014, 3.070797e-28], [475.21736, 3.076834e-28], [476.67906, 3.082898e-28], [478.14526, 3.088989e-28], [479.61597, 3.095107e-28], [481.0912, 3.101251e-28], [482.57096, 3.107423e-28], [484.05528, 3.113621e-28], [485.54417, 3.119846e-28], [487.03763, 3.126099e-28], [488.53569, 3.132379e-28], [490.03836, 3.138686e-28], [491.54565, 3.14502e-28], [493.05757, 3.151382e-28], [494.57415, 3.157771e-28], [496.09539, 3.164188e-28], [497.62131, 3.170632e-28], [499.15192, 3.177105e-28], [500.68724, 3.183605e-28], [502.22728, 3.190133e-28], [503.77206, 3.196689e-28], [505.32159, 3.203273e-28], [506.87589, 3.209885e-28], [508.43497, 3.216526e-28], [509.99884, 3.223194e-28], [511.56753, 3.229892e-28], [513.14103, 3.236617e-28], [514.71938, 3.243371e-28], [516.30259, 3.250154e-28], [517.89066, 3.256965e-28], [519.48362, 3.263806e-28], [521.08148, 3.270675e-28], [522.68425, 3.277573e-28], [524.29195, 3.2845e-28], [525.9046, 3.291456e-28], [527.52221, 3.298441e-28], [529.14479, 3.305456e-28], [530.77236, 3.3125e-28], [532.40495, 3.319573e-28], [534.04255, 3.326676e-28], [535.68519, 3.333808e-28], [537.33288, 3.34097e-28], [538.98564, 3.348162e-28], [540.64348, 3.355384e-28], [542.30642, 3.362636e-28], [543.97448, 3.369917e-28], [545.64767, 3.377229e-28], [547.326, 3.384571e-28], [549.0095, 3.391943e-28], [550.69818, 3.399346e-28], [552.39205, 3.406779e-28], [554.09112, 3.414243e-28], [555.79543, 3.421737e-28], [557.50498, 3.429261e-28], [559.21978, 3.436817e-28], [560.93986, 3.444403e-28], [562.66524, 3.452021e-28], [564.39591, 3.459669e-28], [566.13192, 3.467349e-28], [567.87326, 3.475059e-28], [569.61995, 3.482801e-28], [571.37202, 3.490575e-28], [573.12948, 3.498379e-28], [574.89235, 3.506216e-28], [576.66064, 3.514084e-28], [578.43436, 3.521983e-28], [580.21354, 3.529915e-28], [581.9982, 3.537878e-28], [583.78834, 3.545874e-28], [585.58399, 3.553901e-28], [587.38517, 3.561961e-28], [589.19188, 3.570052e-28], [591.00415, 3.578177e-28], [592.82199, 3.586333e-28], [594.64543, 3.594522e-28], [596.47448, 3.602744e-28], [598.30915, 3.610998e-28], [600.14946, 3.619286e-28], [601.99543, 3.627606e-28], [603.84709, 3.635959e-28], [605.70443, 3.644345e-28], [607.5675, 3.652764e-28], [609.43629, 3.661216e-28], [611.31083, 3.669702e-28], [613.19113, 3.678221e-28], [615.07722, 3.686774e-28], [616.96911, 3.69536e-28], [618.86682, 3.70398e-28], [620.77037, 3.712634e-28], [622.67977, 3.721322e-28], [624.59504, 3.730044e-28], [626.51621, 3.738799e-28], [628.44328, 3.747589e-28], [630.37629, 3.756414e-28], [632.31523, 3.765272e-28], [634.26014, 3.774165e-28], [636.21104, 3.783092e-28], [638.16793, 3.792055e-28], [640.13085, 3.801051e-28], [642.0998, 3.810083e-28], [644.07481, 3.81915e-28], [646.05589, 3.828251e-28], [648.04307, 3.837388e-28], [650.03635, 3.84656e-28], [652.03577, 3.855767e-28], [654.04134, 3.86501e-28], [656.05308, 3.874288e-28], [658.07101, 3.883602e-28], [660.09514, 3.892951e-28], [662.1255, 3.902336e-28], [664.1621, 3.911758e-28], [666.20497, 3.921215e-28], [668.25412, 3.930708e-28], [670.30958, 3.940237e-28], [672.37136, 3.949803e-28], [674.43947, 3.959405e-28], [676.51395, 3.969043e-28], [678.59481, 3.978719e-28], [680.68208, 3.98843e-28], [682.77576, 3.998179e-28], [684.87588, 4.007964e-28], [686.98246, 4.017787e-28], [689.09552, 4.027646e-28], [691.21508, 4.037543e-28], [693.34116, 4.047477e-28], [695.47378, 4.057449e-28], [697.61296, 4.067457e-28], [699.75872, 4.077504e-28], [701.91107, 4.087588e-28], [704.07005, 4.09771e-28], [706.23567, 4.10787e-28], [708.40795, 4.118068e-28], [710.58691, 4.128304e-28], [712.77258, 4.138579e-28], [714.96497, 4.148891e-28], [717.1641, 4.159242e-28], [719.36999, 4.169632e-28], [721.58267, 4.18006e-28], [723.80216, 4.190527e-28], [726.02847, 4.201033e-28], [728.26163, 4.211578e-28], [730.50166, 4.222162e-28], [732.74858, 4.232785e-28], [735.00241, 4.243447e-28], [737.26317, 4.254149e-28], [739.53089, 4.26489e-28], [741.80558, 4.275671e-28], [744.08727, 4.286492e-28], [746.37598, 4.297352e-28], [748.67172, 4.308253e-28], [750.97453, 4.319193e-28], [753.28442, 4.330174e-28], [755.60142, 4.341195e-28], [757.92554, 4.352256e-28], [760.25681, 4.363358e-28], [762.59525, 4.3745e-28], [764.94089, 4.385683e-28], [767.29374, 4.396907e-28], [769.65382, 4.408172e-28], [772.02117, 4.419478e-28], [774.3958, 4.430825e-28], [776.77773, 4.442213e-28], [779.16699, 4.453643e-28], [781.56359, 4.465114e-28], [783.96757, 4.476627e-28], [786.37895, 4.488181e-28], [788.79774, 4.499778e-28], [791.22397, 4.511416e-28], [793.65766, 4.523097e-28], [796.09884, 4.53482e-28], [798.54752, 4.546585e-28], [801.00374, 4.558392e-28], [803.46752, 4.570242e-28], [805.93887, 4.582135e-28], [808.41782, 4.59407e-28], [810.9044, 4.606049e-28], [813.39863, 4.61807e-28], [815.90052, 4.630134e-28], [818.41012, 4.642242e-28], [820.92743, 4.654394e-28], [823.45249, 4.666588e-28], [825.98531, 4.678827e-28], [828.52592, 4.691109e-28], [831.07435, 4.703434e-28], [833.63062, 4.715804e-28], [836.19475, 4.728218e-28], [838.76676, 4.740677e-28], [841.34669, 4.753179e-28], [843.93456, 4.765726e-28], [846.53038, 4.778318e-28], [849.13419, 4.790954e-28], [851.746, 4.803636e-28], [854.36585, 4.816362e-28], [856.99376, 4.829133e-28], [859.62975, 4.84195e-28], [862.27385, 4.854812e-28], [864.92608, 4.867719e-28], [867.58647, 4.880672e-28], [870.25504, 4.893671e-28], [872.93182, 4.906715e-28], [875.61684, 4.919806e-28], [878.31011, 4.932942e-28], [881.01167, 4.946125e-28], [883.72154, 4.959355e-28], [886.43974, 4.97263e-28], [889.1663, 4.985953e-28], [891.90125, 4.999322e-28], [894.64461, 5.012738e-28], [897.39641, 5.026201e-28], [900.15667, 5.039711e-28], [902.92543, 5.053269e-28], [905.7027, 5.066874e-28], [908.48851, 5.080526e-28], [911.28289, 5.094226e-28], [914.08587, 5.107974e-28], [916.89747, 5.12177e-28], [919.71771, 5.135614e-28], [922.54664, 5.149506e-28], [925.38426, 5.163446e-28], [928.23061, 5.177435e-28], [931.08571, 5.191473e-28], [933.9496, 5.205559e-28], [936.8223, 5.219695e-28], [939.70383, 5.233879e-28], [942.59423, 5.248112e-28], [945.49351, 5.262395e-28], [948.40172, 5.276727e-28], [951.31886, 5.291109e-28], [954.24499, 5.305541e-28], [957.18011, 5.320022e-28], [960.12426, 5.334553e-28], [963.07746, 5.349135e-28], [966.03975, 5.363767e-28], [969.01115, 5.378449e-28], [971.99169, 5.393182e-28], [974.9814, 5.407965e-28], [977.98031, 5.4228e-28], [980.98844, 5.437685e-28], [984.00582, 5.452621e-28], [987.03248, 5.467609e-28], [990.06845, 5.482648e-28], [993.11376, 5.497739e-28], [996.16844, 5.512882e-28], [999.23251, 5.528076e-28], [1002.306, 5.543322e-28], [1005.389, 5.558621e-28], [1008.4814, 5.573972e-28], [1011.5833, 5.589375e-28], [1014.6948, 5.604831e-28], [1017.8159, 5.620339e-28], [1020.9465, 5.635901e-28], [1024.0868, 5.651515e-28], [1027.2368, 5.667183e-28], [1030.3964, 5.682904e-28], [1033.5658, 5.698679e-28], [1036.7449, 5.714507e-28], [1039.9337, 5.730389e-28], [1043.1324, 5.746325e-28], [1046.341, 5.762315e-28], [1049.5594, 5.77836e-28], [1052.7876, 5.794458e-28], [1056.0259, 5.810612e-28], [1059.2741, 5.82682e-28], [1062.5322, 5.843083e-28], [1065.8004, 5.859401e-28], [1069.0787, 5.875774e-28], [1072.367, 5.892202e-28], [1075.6655, 5.908686e-28], [1078.9741, 5.925226e-28], [1082.2928, 5.941822e-28], [1085.6218, 5.958473e-28], [1088.961, 5.975181e-28], [1092.3105, 5.991945e-28], [1095.6703, 6.008765e-28], [1099.0404, 6.025642e-28], [1102.4209, 6.042576e-28], [1105.8118, 6.059567e-28], [1109.2131, 6.076615e-28], [1112.6249, 6.09372e-28], [1116.0472, 6.110882e-28], [1119.48, 6.128102e-28], [1122.9234, 6.14538e-28], [1126.3773, 6.162716e-28], [1129.8419, 6.18011e-28], [1133.3171, 6.197562e-28], [1136.803, 6.215073e-28], [1140.2997, 6.232642e-28], [1143.8071, 6.25027e-28], [1147.3253, 6.267957e-28], [1150.8543, 6.285703e-28], [1154.3942, 6.303508e-28], [1157.9449, 6.321373e-28], [1161.5066, 6.339297e-28], [1165.0792, 6.357281e-28], [1168.6628, 6.375325e-28], [1172.2575, 6.393429e-28], [1175.8632, 6.411594e-28], [1179.48, 6.429819e-28], [1183.1079, 6.448104e-28], [1186.7469, 6.46645e-28], [1190.3972, 6.484858e-28], [1194.0587, 6.503326e-28], [1197.7314, 6.521856e-28], [1201.4155, 6.540447e-28], [1205.1109, 6.5591e-28], [1208.8176, 6.577815e-28], [1212.5358, 6.596592e-28], [1216.2654, 6.615431e-28], [1220.0064, 6.634333e-28], [1223.759, 6.653297e-28], [1227.5231, 6.672324e-28], [1231.2988, 6.691414e-28], [1235.0861, 6.710567e-28], [1238.885, 6.729783e-28], [1242.6957, 6.749063e-28], [1246.518, 6.768407e-28], [1250.3521, 6.787814e-28], [1254.1981, 6.807286e-28], [1258.0558, 6.826821e-28], [1261.9254, 6.846421e-28], [1265.8069, 6.866086e-28], [1269.7003, 6.885815e-28], [1273.6058, 6.90561e-28], [1277.5232, 6.925469e-28], [1281.4527, 6.945394e-28], [1285.3942, 6.965385e-28], [1289.3479, 6.985441e-28], [1293.3138, 7.005563e-28], [1297.2918, 7.025751e-28], [1301.2821, 7.046006e-28], [1305.2847, 7.066327e-28], [1309.2995, 7.086715e-28], [1313.3268, 7.107169e-28], [1317.3664, 7.127691e-28], [1321.4184, 7.14828e-28], [1325.4829, 7.168937e-28], [1329.5599, 7.189661e-28], [1333.6494, 7.210453e-28], [1337.7515, 7.231313e-28], [1341.8663, 7.252241e-28], [1345.9937, 7.273238e-28], [1350.1337, 7.294303e-28], [1354.2866, 7.315437e-28], [1358.4522, 7.336641e-28], [1362.6306, 7.357913e-28], [1366.8218, 7.379255e-28], [1371.026, 7.400667e-28], [1375.2431, 7.422148e-28], [1379.4731, 7.4437e-28], [1383.7162, 7.465322e-28], [1387.9723, 7.487014e-28], [1392.2415, 7.508777e-28], [1396.5238, 7.530611e-28], [1400.8193, 7.552516e-28], [1405.1281, 7.574492e-28], [1409.45, 7.59654e-28], [1413.7853, 7.61866e-28], [1418.1339, 7.640851e-28], [1422.4959, 7.663114e-28], [1426.8713, 7.68545e-28], [1431.2601, 7.707859e-28], [1435.6625, 7.73034e-28], [1440.0784, 7.752894e-28], [1444.5079, 7.775522e-28], [1448.951, 7.798223e-28], [1453.4077, 7.820997e-28], [1457.8782, 7.843845e-28], [1462.3624, 7.866768e-28], [1466.8604, 7.889764e-28], [1471.3723, 7.912835e-28], [1475.898, 7.935981e-28], [1480.4377, 7.959202e-28], [1484.9913, 7.982498e-28], [1489.5589, 8.005869e-28], [1494.1406, 8.029316e-28], [1498.7364, 8.052839e-28], [1503.3463, 8.076438e-28], [1507.9703, 8.100113e-28], [1512.6086, 8.123864e-28], [1517.2612, 8.147692e-28], [1521.9281, 8.171597e-28], [1526.6093, 8.19558e-28], [1531.305, 8.219639e-28], [1536.015, 8.243776e-28], [1540.7396, 8.267992e-28], [1545.4787, 8.292285e-28], [1550.2324, 8.316656e-28], [1555.0007, 8.341106e-28], [1559.7836, 8.365635e-28], [1564.5813, 8.390243e-28], [1569.3937, 8.41493e-28], [1574.221, 8.439696e-28], [1579.063, 8.464542e-28], [1583.92, 8.489468e-28], [1588.7919, 8.514475e-28], [1593.6788, 8.539561e-28], [1598.5807, 8.564729e-28], [1603.4978, 8.589977e-28], [1608.4299, 8.615306e-28], [1613.3772, 8.640716e-28], [1618.3397, 8.666209e-28], [1623.3175, 8.691783e-28], [1628.3106, 8.717439e-28], [1633.319, 8.743178e-28], [1638.3429, 8.768999e-28], [1643.3822, 8.794902e-28], [1648.437, 8.820889e-28], [1653.5074, 8.84696e-28], [1658.5933, 8.873113e-28], [1663.6949, 8.899351e-28], [1668.8122, 8.925672e-28], [1673.9452, 8.952078e-28], [1679.0941, 8.978568e-28], [1684.2587, 9.005143e-28], [1689.4393, 9.031803e-28], [1694.6357, 9.058549e-28], [1699.8482, 9.08538e-28], [1705.0767, 9.112297e-28], [1710.3213, 9.1393e-28], [1715.582, 9.166389e-28], [1720.8588, 9.193564e-28], [1726.152, 9.220827e-28], [1731.4614, 9.248176e-28], [1736.7871, 9.275614e-28], [1742.1292, 9.303138e-28], [1747.4877, 9.330751e-28], [1752.8628, 9.358451e-28], [1758.2543, 9.38624e-28], [1763.6625, 9.414118e-28], [1769.0872, 9.442084e-28], [1774.5287, 9.47014e-28], [1779.9869, 9.498284e-28], [1785.4619, 9.526519e-28], [1790.9537, 9.554844e-28], [1796.4624, 9.583258e-28], [1801.9881, 9.611764e-28], [1807.5308, 9.64036e-28], [1813.0905, 9.669047e-28], [1818.6673, 9.697826e-28], [1824.2612, 9.726696e-28], [1829.8724, 9.755658e-28], [1835.5008, 9.784712e-28], [1841.1466, 9.813859e-28], [1846.8097, 9.843098e-28], [1852.4902, 9.87243e-28], [1858.1882, 9.901855e-28], [1863.9037, 9.931374e-28], [1869.6368, 9.960987e-28], [1875.3876, 9.990693e-28], [1881.156, 1.002049e-27], [1886.9422, 1.005039e-27], [1892.7461, 1.008038e-27], [1898.5679, 1.011047e-27], [1904.4077, 1.014065e-27], [1910.2653, 1.017092e-27], [1916.1411, 1.02013e-27], [1922.0348, 1.023177e-27], [1927.9467, 1.026233e-27], [1933.8768, 1.0293e-27], [1939.8252, 1.032376e-27], [1945.7918, 1.035461e-27], [1951.7768, 1.038557e-27], [1957.7802, 1.041662e-27], [1963.802, 1.044777e-27], [1969.8424, 1.047902e-27], [1975.9013, 1.051037e-27], [1981.9789, 1.054182e-27], [1988.0752, 1.057337e-27], [1994.1903, 1.060502e-27], [2000.3241, 1.063677e-27], [2006.4768, 1.066861e-27], [2012.6484, 1.070056e-27], [2018.8391, 1.073261e-27], [2025.0487, 1.076477e-27], [2031.2775, 1.079702e-27], [2037.5254, 1.082937e-27], [2043.7926, 1.086183e-27], [2050.079, 1.089439e-27], [2056.3847, 1.092706e-27], [2062.7099, 1.095982e-27], [2069.0545, 1.099269e-27], [2075.4186, 1.102566e-27], [2081.8023, 1.105874e-27], [2088.2056, 1.109192e-27], [2094.6286, 1.112521e-27], [2101.0714, 1.11586e-27], [2107.534, 1.11921e-27], [2114.0165, 1.12257e-27], [2120.5189, 1.125941e-27], [2127.0413, 1.129323e-27], [2133.5838, 1.132715e-27], [2140.1464, 1.136118e-27], [2146.7292, 1.139531e-27], [2153.3322, 1.142956e-27], [2159.9556, 1.146391e-27], [2166.5993, 1.149837e-27], [2173.2634, 1.153294e-27], [2179.9481, 1.156761e-27], [2186.6533, 1.16024e-27], [2193.3791, 1.16373e-27], [2200.1256, 1.16723e-27], [2206.8929, 1.170742e-27], [2213.681, 1.174264e-27], [2220.49, 1.177798e-27], [2227.3199, 1.181343e-27], [2234.1708, 1.184899e-27], [2241.0428, 1.188466e-27], [2247.9359, 1.192045e-27], [2254.8503, 1.195634e-27], [2261.7859, 1.199235e-27], [2268.7428, 1.202848e-27], [2275.7211, 1.206471e-27], [2282.7209, 1.210106e-27], [2289.7422, 1.213753e-27], [2296.7852, 1.217411e-27], [2303.8497, 1.22108e-27], [2310.9361, 1.224761e-27], [2318.0442, 1.228454e-27], [2325.1741, 1.232158e-27], [2332.3261, 1.235874e-27], [2339.5, 1.239601e-27], [2346.6959, 1.24334e-27], [2353.914, 1.247091e-27], [2361.1543, 1.250854e-27], [2368.4169, 1.254628e-27], [2375.7018, 1.258415e-27], [2383.0091, 1.262213e-27], [2390.3389, 1.266023e-27], [2397.6913, 1.269845e-27], [2405.0662, 1.273679e-27], [2412.4639, 1.277525e-27], [2419.8843, 1.281383e-27], [2427.3275, 1.285253e-27], [2434.7936, 1.289136e-27], [2442.2827, 1.29303e-27], [2449.7948, 1.296937e-27], [2457.33, 1.300856e-27], [2464.8884, 1.304787e-27], [2472.4701, 1.308731e-27], [2480.075, 1.312687e-27], [2487.7034, 1.316655e-27], [2495.3552, 1.320636e-27], [2503.0306, 1.324629e-27], [2510.7295, 1.328635e-27], [2518.4522, 1.332653e-27], [2526.1986, 1.336684e-27], [2533.9688, 1.340727e-27], [2541.763, 1.344783e-27], [2549.5811, 1.348852e-27], [2557.4232, 1.352934e-27], [2565.2895, 1.357028e-27], [2573.1799, 1.361135e-27], [2581.0947, 1.365255e-27], [2589.0338, 1.369388e-27], [2596.9973, 1.373534e-27], [2604.9852, 1.377693e-27], [2612.9978, 1.381865e-27], [2621.035, 1.38605e-27], [2629.097, 1.390247e-27], [2637.1837, 1.394459e-27], [2645.2953, 1.398683e-27], [2653.4318, 1.40292e-27], [2661.5934, 1.407171e-27], [2669.7801, 1.411435e-27], [2677.992, 1.415712e-27], [2686.2291, 1.420003e-27], [2694.4915, 1.424307e-27], [2702.7794, 1.428624e-27], [2711.0928, 1.432955e-27], [2719.4317, 1.4373e-27], [2727.7963, 1.441658e-27], [2736.1866, 1.446029e-27], [2744.6027, 1.450415e-27], [2753.0448, 1.454814e-27], [2761.5127, 1.459226e-27], [2770.0067, 1.463653e-27], [2778.5269, 1.468093e-27], [2787.0732, 1.472548e-27], [2795.6459, 1.477016e-27], [2804.2449, 1.481498e-27], [2812.8704, 1.485994e-27], [2821.5223, 1.490504e-27], [2830.201, 1.495028e-27], [2838.9062, 1.499566e-27], [2847.6383, 1.504119e-27], [2856.3972, 1.508685e-27], [2865.1831, 1.513266e-27], [2873.996, 1.517861e-27], [2882.836, 1.522471e-27], [2891.7032, 1.527094e-27], [2900.5977, 1.531733e-27], [2909.5195, 1.536385e-27], [2918.4688, 1.541053e-27], [2927.4456, 1.545734e-27], [2936.45, 1.550431e-27], [2945.4821, 1.555142e-27], [2954.542, 1.559867e-27], [2963.6297, 1.564608e-27], [2972.7454, 1.569363e-27], [2981.8892, 1.574133e-27], [2991.061, 1.578918e-27], [3000.2611, 1.583717e-27], [3009.4895, 1.588532e-27], [3018.7463, 1.593362e-27], [3028.0315, 1.598206e-27], [3037.3453, 1.603066e-27], [3046.6877, 1.607941e-27], [3056.0589, 1.612831e-27], [3065.4589, 1.617737e-27], [3074.8878, 1.622657e-27], [3084.3457, 1.627593e-27], [3093.8327, 1.632544e-27], [3103.3489, 1.637511e-27], [3112.8944, 1.642493e-27], [3122.4692, 1.647491e-27], [3132.0735, 1.652504e-27], [3141.7073, 1.657533e-27], [3151.3708, 1.662577e-27], [3161.0639, 1.667637e-27], [3170.7869, 1.672713e-27], [3180.5398, 1.677805e-27], [3190.3227, 1.682912e-27], [3200.1357, 1.688036e-27], [3209.9788, 1.693175e-27], [3219.8523, 1.69833e-27], [3229.7561, 1.703502e-27], [3239.6903, 1.708689e-27], [3249.6552, 1.713893e-27], [3259.6506, 1.719113e-27], [3269.6769, 1.724348e-27], [3279.7339, 1.729601e-27], [3289.8219, 1.734869e-27], [3299.941, 1.740154e-27], [3310.0911, 1.745455e-27], [3320.2725, 1.750773e-27], [3330.4852, 1.756108e-27], [3340.7293, 1.761458e-27], [3351.0049, 1.766826e-27], [3361.3121, 1.77221e-27], [3371.651, 1.777611e-27], [3382.0217, 1.783029e-27], [3392.4243, 1.788463e-27], [3402.859, 1.793915e-27], [3413.3257, 1.799383e-27], [3423.8246, 1.804869e-27], [3434.3558, 1.810371e-27], [3444.9194, 1.815891e-27], [3455.5154, 1.821427e-27], [3466.1441, 1.826981e-27], [3476.8055, 1.832552e-27], [3487.4996, 1.83814e-27], [3498.2267, 1.843746e-27], [3508.9867, 1.849369e-27], [3519.7799, 1.85501e-27], [3530.6062, 1.860668e-27], [3541.4659, 1.866343e-27], [3552.3589, 1.872037e-27], [3563.2855, 1.877747e-27], [3574.2456, 1.883476e-27], [3585.2395, 1.889222e-27], [3596.2672, 1.894987e-27], [3607.3288, 1.900769e-27], [3618.4244, 1.906569e-27], [3629.5542, 1.912387e-27], [3640.7182, 1.918223e-27], [3651.9165, 1.924077e-27], [3663.1493, 1.92995e-27], [3674.4166, 1.93584e-27], [3685.7186, 1.941749e-27], [3697.0553, 1.947676e-27], [3708.427, 1.953622e-27], [3719.8336, 1.959586e-27], [3731.2752, 1.965568e-27], [3742.7521, 1.971569e-27], [3754.2643, 1.977589e-27], [3765.8119, 1.983627e-27], [3777.395, 1.989684e-27], [3789.0137, 1.99576e-27], [3800.6682, 2.001855e-27], [3812.3585, 2.007968e-27], [3824.0848, 2.0141e-27], [3835.8471, 2.020252e-27], [3847.6456, 2.026423e-27], [3859.4804, 2.032612e-27], [3871.3516, 2.038821e-27], [3883.2594, 2.045049e-27], [3895.2037, 2.051297e-27], [3907.1848, 2.057563e-27], [3919.2028, 2.06385e-27], [3931.2577, 2.070155e-27], [3943.3497, 2.07648e-27], [3955.4788, 2.082825e-27], [3967.6453, 2.08919e-27], [3979.8493, 2.095574e-27], [3992.0907, 2.101978e-27], [4004.3698, 2.108401e-27], [4016.6867, 2.114845e-27], [4029.0414, 2.121309e-27], [4041.4342, 2.127792e-27], [4053.8651, 2.134296e-27], [4066.3342, 2.14082e-27], [4078.8417, 2.147364e-27], [4091.3876, 2.153928e-27], [4103.9721, 2.160513e-27], [4116.5954, 2.167118e-27], [4129.2574, 2.173744e-27], [4141.9584, 2.18039e-27], [4154.6985, 2.187056e-27], [4167.4778, 2.193744e-27], [4180.2963, 2.200452e-27], [4193.1543, 2.20718e-27], [4206.0519, 2.21393e-27], [4218.9891, 2.2207e-27], [4231.9661, 2.227492e-27], [4244.983, 2.234304e-27], [4258.04, 2.241137e-27], [4271.1371, 2.247992e-27], [4284.2745, 2.254868e-27], [4297.4524, 2.261765e-27], [4310.6707, 2.268684e-27], [4323.9297, 2.275623e-27], [4337.2295, 2.282585e-27], [4350.5702, 2.289568e-27], [4363.9519, 2.296572e-27], [4377.3748, 2.303598e-27], [4390.839, 2.310646e-27], [4404.3446, 2.317716e-27], [4417.8918, 2.324808e-27], [4431.4806, 2.331921e-27], [4445.1112, 2.339057e-27], [4458.7837, 2.346214e-27], [4472.4983, 2.353394e-27], [4486.2551, 2.360596e-27], [4500.0541, 2.36782e-27], [4513.8957, 2.375067e-27], [4527.7798, 2.382336e-27], [4541.7066, 2.389627e-27], [4555.6762, 2.396941e-27], [4569.6888, 2.404278e-27], [4583.7446, 2.411637e-27], [4597.8435, 2.419019e-27], [4611.9858, 2.426424e-27], [4626.1716, 2.433852e-27], [4640.4011, 2.441303e-27], [4654.6743, 2.448777e-27], [4668.9914, 2.456274e-27], [4683.3526, 2.463794e-27], [4697.7579, 2.471337e-27], [4712.2075, 2.478904e-27], [4726.7016, 2.486494e-27], [4741.2403, 2.494107e-27], [4755.8236, 2.501744e-27], [4770.4519, 2.509405e-27], [4785.1251, 2.517089e-27], [4799.8435, 2.524797e-27], [4814.6071, 2.532529e-27], [4829.4162, 2.540285e-27], [4844.2708, 2.548065e-27], [4859.1711, 2.555869e-27], [4874.1172, 2.563697e-27], [4889.1093, 2.571549e-27], [4904.1475, 2.579425e-27], [4919.2319, 2.587326e-27], [4934.3628, 2.595251e-27], [4949.5402, 2.603201e-27], [4964.7643, 2.611175e-27], [4980.0352, 2.619174e-27], [4995.3531, 2.627197e-27], [5010.7181, 2.635246e-27], [5026.1303, 2.643319e-27], [5041.59, 2.651417e-27], [5057.0972, 2.65954e-27], [5072.6521, 2.667689e-27], [5088.2549, 2.675862e-27], [5103.9056, 2.684061e-27], [5119.6045, 2.692285e-27], [5135.3517, 2.700534e-27], [5151.1473, 2.708809e-27], [5166.9915, 2.71711e-27], [5182.8844, 2.725436e-27], [5198.8262, 2.733788e-27], [5214.8171, 2.742165e-27], [5230.8571, 2.750569e-27], [5246.9465, 2.758998e-27], [5263.0854, 2.767454e-27], [5279.2739, 2.775935e-27], [5295.5122, 2.784443e-27], [5311.8004, 2.792977e-27], [5328.1387, 2.801538e-27], [5344.5273, 2.810125e-27], [5360.9663, 2.818738e-27], [5377.4559, 2.827378e-27], [5393.9962, 2.836045e-27], [5410.5874, 2.844738e-27], [5427.2296, 2.853458e-27], [5443.923, 2.862205e-27], [5460.6677, 2.87098e-27], [5477.4639, 2.879781e-27], [5494.3118, 2.888609e-27], [5511.2116, 2.897465e-27], [5528.1633, 2.906348e-27], [5545.1671, 2.915258e-27], [5562.2233, 2.924196e-27], [5579.3319, 2.933162e-27], [5596.4931, 2.942155e-27], [5613.7071, 2.951176e-27], [5630.9741, 2.960225e-27], [5648.2942, 2.969302e-27], [5665.6675, 2.978407e-27], [5683.0943, 2.98754e-27], [5700.5747, 2.996701e-27], [5718.1088, 3.00589e-27], [5735.6969, 3.015108e-27], [5753.3391, 3.024354e-27], [5771.0356, 3.033629e-27], [5788.7865, 3.042932e-27], [5806.5919, 3.052264e-27], [5824.4522, 3.061625e-27], [5842.3674, 3.071018e-27], [5860.3377, 3.080439e-27], [5878.3632, 3.08989e-27], [5896.4443, 3.09937e-27], [5914.5809, 3.10888e-27], [5932.7733, 3.118418e-27], [5951.0217, 3.127986e-27], [5969.3262, 3.137584e-27], [5987.687, 3.147211e-27], [6006.1042, 3.156868e-27], [6024.5781, 3.166554e-27], [6043.1089, 3.176271e-27], [6061.6966, 3.186017e-27], [6080.3416, 3.195794e-27], [6099.0438, 3.2056e-27], [6117.8036, 3.215437e-27], [6136.6211, 3.225304e-27], [6155.4965, 3.235202e-27], [6174.4299, 3.24513e-27], [6193.4216, 3.255089e-27], [6212.4717, 3.265078e-27], [6231.5804, 3.275098e-27], [6250.7478, 3.285149e-27], [6269.9742, 3.295232e-27], [6289.2598, 3.305345e-27], [6308.6046, 3.315489e-27], [6328.009, 3.325665e-27], [6347.4731, 3.335872e-27], [6366.997, 3.34611e-27], [6386.581, 3.35638e-27], [6406.2252, 3.366682e-27], [6425.9298, 3.377015e-27], [6445.6951, 3.38738e-27], [6465.5211, 3.397778e-27], [6485.4081, 3.408207e-27], [6505.3563, 3.418668e-27], [6525.3659, 3.429162e-27], [6545.437, 3.439688e-27], [6565.5698, 3.450247e-27], [6585.7646, 3.460838e-27], [6606.0215, 3.471461e-27], [6626.3406, 3.482118e-27], [6646.7223, 3.492807e-27], [6667.1667, 3.503529e-27], [6687.674, 3.514284e-27], [6708.2443, 3.525073e-27], [6728.8779, 3.535894e-27], [6749.575, 3.546749e-27], [6770.3357, 3.557638e-27], [6791.1603, 3.56856e-27], [6812.0489, 3.579515e-27], [6833.0018, 3.590505e-27], [6854.0192, 3.601528e-27], [6875.1012, 3.612585e-27], [6896.248, 3.623676e-27], [6917.4599, 3.634802e-27], [6938.737, 3.645962e-27], [6960.0796, 3.657156e-27], [6981.4878, 3.668385e-27], [7002.9619, 3.679648e-27], [7024.502, 3.690946e-27], [7046.1083, 3.702279e-27], [7067.7812, 3.713646e-27], [7089.5207, 3.725049e-27], [7111.327, 3.736487e-27], [7133.2004, 3.74796e-27], [7155.1412, 3.759468e-27], [7177.1494, 3.771012e-27], [7199.2252, 3.782592e-27], [7221.369, 3.794207e-27], [7243.5809, 3.805858e-27], [7265.8612, 3.817545e-27], [7288.2099, 3.829268e-27], [7310.6274, 3.841027e-27], [7333.1139, 3.852822e-27], [7355.6695, 3.864653e-27], [7378.2945, 3.876521e-27], [7400.9891, 3.888426e-27], [7423.7535, 3.900367e-27], [7446.5879, 3.912345e-27], [7469.4925, 3.92436e-27], [7492.4676, 3.936412e-27], [7515.5134, 3.948501e-27], [7538.63, 3.960627e-27], [7561.8178, 3.972791e-27], [7585.0769, 3.984992e-27], [7608.4075, 3.997231e-27], [7631.8099, 4.009507e-27], [7655.2842, 4.021821e-27], [7678.8308, 4.034174e-27], [7702.4498, 4.046564e-27], [7726.1414, 4.058992e-27], [7749.906, 4.071459e-27], [7773.7436, 4.083964e-27], [7797.6545, 4.096508e-27], [7821.639, 4.10909e-27], [7845.6972, 4.121711e-27], [7869.8295, 4.134371e-27], [7894.0359, 4.14707e-27], [7918.3169, 4.159808e-27], [7942.6725, 4.172585e-27], [7967.103, 4.185401e-27], [7991.6087, 4.198257e-27], [8016.1898, 4.211153e-27], [8040.8464, 4.224088e-27], [8065.5789, 4.237064e-27], [8090.3875, 4.250079e-27], [8115.2724, 4.263134e-27], [8140.2338, 4.27623e-27], [8165.272, 4.289365e-27], [8190.3872, 4.302542e-27], [8215.5797, 4.315758e-27], [8240.8496, 4.329016e-27], [8266.1973, 4.342314e-27], [8291.623, 4.355654e-27], [8317.1268, 4.369034e-27], [8342.7091, 4.382456e-27], [8368.3701, 4.395919e-27], [8394.11, 4.409423e-27], [8419.9291, 4.422969e-27], [8445.8276, 4.436557e-27], [8471.8058, 4.450186e-27], [8497.8639, 4.463858e-27], [8524.0021, 4.477572e-27], [8550.2207, 4.491328e-27], [8576.52, 4.505126e-27], [8602.9001, 4.518967e-27], [8629.3614, 4.53285e-27], [8655.9041, 4.546776e-27], [8682.5285, 4.560745e-27], [8709.2347, 4.574757e-27], [8736.0231, 4.588812e-27], [8762.8938, 4.602911e-27], [8789.8472, 4.617053e-27], [8816.8836, 4.631238e-27], [8844.003, 4.645467e-27], [8871.2059, 4.65974e-27], [8898.4925, 4.674057e-27], [8925.863, 4.688418e-27], [8953.3177, 4.702823e-27], [8980.8568, 4.717273e-27], [9008.4807, 4.731767e-27], [9036.1895, 4.746305e-27], [9063.9835, 4.760889e-27], [9091.863, 4.775517e-27], [9119.8283, 4.79019e-27], [9147.8796, 4.804909e-27], [9176.0172, 4.819673e-27], [9204.2413, 4.834482e-27], [9232.5522, 4.849337e-27], [9260.9503, 4.864238e-27], [9289.4356, 4.879184e-27], [9318.0086, 4.894177e-27], [9346.6695, 4.909215e-27], [9375.4185, 4.9243e-27], [9404.256, 4.939432e-27], [9433.1821, 4.95461e-27], [9462.1973, 4.969835e-27], [9491.3016, 4.985106e-27], [9520.4955, 5.000425e-27], [9549.7792, 5.015791e-27], [9579.153, 5.031204e-27], [9608.6171, 5.046665e-27], [9638.1718, 5.062173e-27], [9667.8175, 5.077729e-27], [9697.5543, 5.093333e-27], [9727.3826, 5.108985e-27], [9757.3027, 5.124685e-27], [9787.3147, 5.140433e-27], [9817.4191, 5.15623e-27], [9847.6161, 5.172076e-27], [9877.906, 5.18797e-27], [9908.289, 5.203913e-27], [9938.7655, 5.219906e-27], [9969.3357, 5.235947e-27], [10000.0, 5.252038e-27]])

    fiveyr = 5 * 365.25 * 86400
    snr = 10
    f1 = data[:,0]
    S1 = fiveyr*(data[:,2]/10)**2
    fstop = [f1[0],f1[-1]]

    # %---- Interpolate to desired frequencies.  Set noise to infinity outside
    # %      band of frequencies covered by f1.
    # S = inf(size(f)); 
    # k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop

    
# % -------------------------------------------------------------------------
# %                              ETD spectrum
# % -------------------------------------------------------------------------

def etd(f):
# % ETD - Design noise curve "D" for ET.
# %
# % Arguments are the same as for the parent function.

# % ---- Extract pre-sampled noise spectrum from data.  First column is
# %      frequency, second column is effective noise amplitude for a 5-year
# %      integration period with SNR=10.  (We'll need to remove these
# %      scalings.)  Data supplied by B. S. Sathyaprakash.
    data = np.array([[1.0, 2.951094e-20], [1.003076, 2.992112e-20], [1.006161, 3.021851e-20], [1.009256, 3.050756e-20], [1.01236, 3.081522e-20], [1.015474, 3.099403e-20], [1.018598, 3.112464e-20], [1.021731, 3.113518e-20], [1.024873, 3.109389e-20], [1.028026, 3.106523e-20], [1.031188, 3.088506e-20], [1.03436, 3.046101e-20], [1.037541, 3.0082e-20], [1.040733, 2.970451e-20], [1.043934, 2.928851e-20], [1.047145, 2.90148e-20], [1.050366, 2.860714e-20], [1.053596, 2.798962e-20], [1.056837, 2.740944e-20], [1.060088, 2.67701e-20], [1.063348, 2.606687e-20], [1.066619, 2.549494e-20], [1.0699, 2.490659e-20], [1.073191, 2.457898e-20], [1.076492, 3.016319e-20], [1.079803, 2.329392e-20], [1.083124, 2.230382e-20], [1.086456, 2.174331e-20], [1.089797, 2.103648e-20], [1.093149, 2.016804e-20], [1.096512, 1.968525e-20], [1.099885, 1.874334e-20], [1.103268, 1.775369e-20], [1.106661, 1.678721e-20], [1.110065, 1.586929e-20], [1.113479, 1.496301e-20], [1.116904, 1.40904e-20], [1.12034, 1.322406e-20], [1.123786, 1.213252e-20], [1.127242, 1.112174e-20], [1.13071, 1.024744e-20], [1.134188, 9.22745e-21], [1.137676, 8.383408e-21], [1.141175, 7.629707e-21], [1.144686, 6.811006e-21], [1.148206, 6.153557e-21], [1.151738, 5.566647e-21], [1.155281, 4.958615e-21], [1.158834, 4.477038e-21], [1.162399, 4.064043e-21], [1.165974, 3.646402e-21], [1.169561, 3.311999e-21], [1.173158, 3.013304e-21], [1.176766, 2.715794e-21], [1.180386, 2.474642e-21], [1.184017, 2.258262e-21], [1.187658, 2.04474e-21], [1.191312, 1.869453e-21], [1.194976, 1.711444e-21], [1.198651, 1.556481e-21], [1.202338, 1.427722e-21], [1.206036, 1.311146e-21], [1.209746, 1.197309e-21], [1.213467, 1.104935e-21], [1.2172, 1.029731e-21], [1.220943, 9.555007e-22], [1.224699, 8.93124e-22], [1.228466, 8.361043e-22], [1.232244, 7.79897e-22], [1.236035, 7.323545e-22], [1.239837, 6.887888e-22], [1.24365, 6.458856e-22], [1.247475, 6.094664e-22], [1.251313, 5.7606e-22], [1.255161, 5.431839e-22], [1.259022, 5.153034e-22], [1.262895, 4.89761e-22], [1.266779, 4.646325e-22], [1.270676, 4.44735e-22], [1.274584, 4.303032e-22], [1.278504, 4.158979e-22], [1.282437, 4.04775e-22], [1.286382, 3.950916e-22], [1.290338, 3.854272e-22], [1.294307, 3.791325e-22], [1.298288, 3.743214e-22], [1.302282, 3.695179e-22], [1.306287, 3.689215e-22], [1.310305, 3.701767e-22], [1.314336, 3.714139e-22], [1.318378, 3.795316e-22], [1.322433, 3.906622e-22], [1.326501, 4.016967e-22], [1.330581, 4.287603e-22], [1.334674, 4.621643e-22], [1.338779, 4.948977e-22], [1.342897, 6.011248e-22], [1.347028, 7.333729e-22], [1.351171, 8.550193e-22], [1.355327, 1.090507e-21], [1.359496, 1.316665e-21], [1.363677, 1.474657e-21], [1.367872, 1.3488e-21], [1.372079, 1.017119e-21], [1.376299, 5.443354e-22], [1.380533, 3.19649e-22], [1.384779, 2.537994e-22], [1.389038, 1.875733e-22], [1.393311, 1.497312e-22], [1.397596, 1.25554e-22], [1.401895, 1.014496e-22], [1.406207, 8.565099e-23], [1.410533, 7.384629e-23], [1.414871, 6.209802e-23], [1.419223, 5.37975e-23], [1.423588, 4.717226e-23], [1.427967, 4.058264e-23], [1.432359, 3.570309e-23], [1.436765, 3.166395e-23], [1.441185, 2.764783e-23], [1.445617, 2.457709e-23], [1.450064, 2.197624e-23], [1.454524, 1.939086e-23], [1.458998, 1.735936e-23], [1.463486, 1.559216e-23], [1.467987, 1.384142e-23], [1.472502, 1.24464e-23], [1.477032, 1.123353e-23], [1.481575, 1.003238e-23], [1.486132, 9.061863e-24], [1.490703, 8.210868e-24], [1.495288, 7.368398e-24], [1.499888, 6.679989e-24], [1.504501, 6.072352e-24], [1.509129, 5.471049e-24], [1.51377, 4.975106e-24], [1.518427, 4.53505e-24], [1.523097, 4.099782e-24], [1.527782, 3.737394e-24], [1.532481, 3.41328e-24], [1.537195, 3.093147e-24], [1.541923, 2.825239e-24], [1.546666, 2.585767e-24], [1.551423, 2.349401e-24], [1.556195, 2.150459e-24], [1.560982, 1.972148e-24], [1.565783, 1.796297e-24], [1.570599, 1.647561e-24], [1.57543, 1.513962e-24], [1.580276, 1.382347e-24], [1.585137, 1.27056e-24], [1.590012, 1.16999e-24], [1.594903, 1.071049e-24], [1.599809, 9.873562e-25], [1.604729, 9.133269e-25], [1.609665, 8.403461e-25], [1.614616, 7.779554e-25], [1.619583, 7.216551e-25], [1.624564, 6.662647e-25], [1.629561, 6.188009e-25], [1.634574, 5.759674e-25], [1.639601, 5.339369e-25], [1.644645, 4.978568e-25], [1.649703, 4.653117e-25], [1.654777, 4.334821e-25], [1.659867, 4.061219e-25], [1.664973, 3.814642e-25], [1.670094, 3.574438e-25], [1.675231, 3.370782e-25], [1.680384, 3.195184e-25], [1.685552, 3.023209e-25], [1.690737, 2.874592e-25], [1.695937, 2.74018e-25], [1.701154, 2.609014e-25], [1.706386, 2.495285e-25], [1.711635, 2.392325e-25], [1.7169, 2.292177e-25], [1.722181, 2.20491e-25], [1.727478, 2.125714e-25], [1.732791, 2.04887e-25], [1.738121, 1.981434e-25], [1.743467, 1.919973e-25], [1.74883, 1.860419e-25], [1.754209, 1.807443e-25], [1.759605, 1.75835e-25], [1.765017, 1.710926e-25], [1.770446, 1.668486e-25], [1.775892, 1.629327e-25], [1.781354, 1.591453e-25], [1.786833, 1.557102e-25], [1.792329, 1.525107e-25], [1.797842, 1.494095e-25], [1.803372, 1.465568e-25], [1.808919, 1.438734e-25], [1.814483, 1.412653e-25], [1.820064, 1.388332e-25], [1.825662, 1.365235e-25], [1.831278, 1.342722e-25], [1.836911, 1.321732e-25], [1.842561, 1.302171e-25], [1.848228, 1.282888e-25], [1.853913, 1.26438e-25], [1.859615, 1.246425e-25], [1.865335, 1.228721e-25], [1.871073, 1.211623e-25], [1.876828, 1.194975e-25], [1.882601, 1.178553e-25], [1.888392, 1.162616e-25], [1.8942, 1.14705e-25], [1.900026, 1.131693e-25], [1.90587, 1.11673e-25], [1.911733, 1.10208e-25], [1.917613, 1.087623e-25], [1.923511, 1.073502e-25], [1.929428, 1.059666e-25], [1.935362, 1.046006e-25], [1.941315, 1.03262e-25], [1.947286, 1.019466e-25], [1.953276, 1.006479e-25], [1.959284, 9.937305e-26], [1.96531, 9.8119e-26], [1.971355, 9.688082e-26], [1.977419, 9.566372e-26], [1.983501, 9.446546e-26], [1.989602, 9.328235e-26], [1.995722, 9.211817e-26], [2.00186, 9.097126e-26], [2.008018, 8.98388e-26], [2.014194, 8.8722e-26], [2.02039, 8.761846e-26], [2.026604, 8.652969e-26], [2.032838, 8.545746e-26], [2.03909, 8.440072e-26], [2.045362, 8.335787e-26], [2.051654, 8.233008e-26], [2.057964, 8.13166e-26], [2.064294, 8.031628e-26], [2.070644, 7.932984e-26], [2.077013, 7.835677e-26], [2.083401, 7.73962e-26], [2.089809, 7.644859e-26], [2.096237, 7.551356e-26], [2.102685, 7.459044e-26], [2.109153, 7.367945e-26], [2.11564, 7.278025e-26], [2.122148, 7.189248e-26], [2.128675, 7.101628e-26], [2.135223, 7.015142e-26], [2.14179, 6.929749e-26], [2.148378, 6.845455e-26], [2.154986, 6.762241e-26], [2.161615, 6.680075e-26], [2.168263, 6.598955e-26], [2.174933, 6.518868e-26], [2.181622, 6.439786e-26], [2.188333, 6.361705e-26], [2.195064, 6.284611e-26], [2.201815, 6.208484e-26], [2.208588, 6.133315e-26], [2.215381, 6.059096e-26], [2.222196, 5.985804e-26], [2.229031, 5.913429e-26], [2.235887, 5.841964e-26], [2.242764, 5.771389e-26], [2.249662, 5.701696e-26], [2.256582, 5.632876e-26], [2.263523, 5.564913e-26], [2.270485, 5.497797e-26], [2.277469, 5.431521e-26], [2.284474, 5.366068e-26], [2.291501, 5.301431e-26], [2.298549, 5.2376e-26], [2.305619, 5.174562e-26], [2.312711, 5.112309e-26], [2.319825, 5.050835e-26], [2.32696, 4.990123e-26], [2.334118, 4.930164e-26], [2.341297, 4.870951e-26], [2.348498, 4.812472e-26], [2.355722, 4.754718e-26], [2.362968, 4.697682e-26], [2.370236, 4.641352e-26], [2.377527, 4.58572e-26], [2.38484, 4.530779e-26], [2.392175, 4.476518e-26], [2.399533, 4.422928e-26], [2.406913, 4.370004e-26], [2.414317, 4.317734e-26], [2.421743, 4.26611e-26], [2.429192, 4.215125e-26], [2.436664, 4.164771e-26], [2.444159, 4.115039e-26], [2.451676, 4.065923e-26], [2.459217, 4.017415e-26], [2.466782, 3.969505e-26], [2.474369, 3.922188e-26], [2.48198, 3.875456e-26], [2.489614, 3.8293e-26], [2.497272, 3.783715e-26], [2.504953, 3.738692e-26], [2.512658, 3.694225e-26], [2.520387, 3.650307e-26], [2.528139, 3.60693e-26], [2.535915, 3.564087e-26], [2.543715, 3.521774e-26], [2.551539, 3.479982e-26], [2.559388, 3.438704e-26], [2.56726, 3.397936e-26], [2.575156, 3.357669e-26], [2.583077, 3.317897e-26], [2.591022, 3.278616e-26], [2.598992, 3.239818e-26], [2.606986, 3.201496e-26], [2.615005, 3.163647e-26], [2.623048, 3.126263e-26], [2.631116, 3.089337e-26], [2.639209, 3.052866e-26], [2.647327, 3.016843e-26], [2.65547, 2.981262e-26], [2.663638, 2.946118e-26], [2.671831, 2.911406e-26], [2.680049, 2.877119e-26], [2.688292, 2.843253e-26], [2.696561, 2.809803e-26], [2.704855, 2.776761e-26], [2.713175, 2.744126e-26], [2.721521, 2.71189e-26], [2.729891, 2.680049e-26], [2.738288, 2.648598e-26], [2.746711, 2.617532e-26], [2.755159, 2.586846e-26], [2.763634, 2.556536e-26], [2.772134, 2.526597e-26], [2.780661, 2.497023e-26], [2.789214, 2.467811e-26], [2.797793, 2.438956e-26], [2.806399, 2.410453e-26], [2.815031, 2.382299e-26], [2.823689, 2.354488e-26], [2.832375, 2.327016e-26], [2.841087, 2.299881e-26], [2.849826, 2.273076e-26], [2.858591, 2.246597e-26], [2.867384, 2.220442e-26], [2.876203, 2.194605e-26], [2.88505, 2.169083e-26], [2.893924, 2.143872e-26], [2.902826, 2.118967e-26], [2.911754, 2.094366e-26], [2.92071, 2.070064e-26], [2.929694, 2.046058e-26], [2.938705, 2.022343e-26], [2.947744, 1.998916e-26], [2.956811, 1.975775e-26], [2.965906, 1.952914e-26], [2.975029, 1.93033e-26], [2.98418, 1.908021e-26], [2.993358, 1.885983e-26], [3.002566, 1.864211e-26], [3.011801, 1.842704e-26], [3.021065, 1.821457e-26], [3.030357, 1.800468e-26], [3.039678, 1.779733e-26], [3.049028, 1.759248e-26], [3.058406, 1.739012e-26], [3.067813, 1.719021e-26], [3.07725, 1.699271e-26], [3.086715, 1.67976e-26], [3.096209, 1.660484e-26], [3.105733, 1.641441e-26], [3.115285, 1.622629e-26], [3.124868, 1.604043e-26], [3.134479, 1.585681e-26], [3.14412, 1.567541e-26], [3.153791, 1.549619e-26], [3.163492, 1.531913e-26], [3.173222, 1.514421e-26], [3.182983, 1.497139e-26], [3.192773, 1.480065e-26], [3.202594, 1.463196e-26], [3.212444, 1.446531e-26], [3.222325, 1.430065e-26], [3.232237, 1.413797e-26], [3.242179, 1.397724e-26], [3.252151, 1.381844e-26], [3.262154, 1.366155e-26], [3.272188, 1.350654e-26], [3.282253, 1.335339e-26], [3.292349, 1.320207e-26], [3.302476, 1.305256e-26], [3.312634, 1.290484e-26], [3.322823, 1.275889e-26], [3.333043, 1.261468e-26], [3.343295, 1.24722e-26], [3.353579, 1.233142e-26], [3.363894, 1.219232e-26], [3.374241, 1.205487e-26], [3.384619, 1.191907e-26], [3.39503, 1.178489e-26], [3.405473, 1.16523e-26], [3.415947, 1.152129e-26], [3.426454, 1.139185e-26], [3.436994, 1.126394e-26], [3.447565, 1.113755e-26], [3.458169, 1.101267e-26], [3.468806, 1.088927e-26], [3.479476, 1.076733e-26], [3.490178, 1.064684e-26], [3.500914, 1.052778e-26], [3.511682, 1.041013e-26], [3.522483, 1.029388e-26], [3.533318, 1.017899e-26], [3.544186, 1.006547e-26], [3.555087, 9.953297e-27], [3.566022, 9.842443e-27], [3.576991, 9.732899e-27], [3.587993, 9.624651e-27], [3.599029, 9.517677e-27], [3.6101, 9.411966e-27], [3.621204, 9.307501e-27], [3.632342, 9.204265e-27], [3.643515, 9.102246e-27], [3.654721, 9.001427e-27], [3.665963, 8.90179e-27], [3.677239, 8.803328e-27], [3.68855, 8.706021e-27], [3.699895, 8.609854e-27], [3.711275, 8.514818e-27], [3.722691, 8.420895e-27], [3.734141, 8.32807e-27], [3.745627, 8.236335e-27], [3.757148, 8.145672e-27], [3.768704, 8.056068e-27], [3.780296, 7.967513e-27], [3.791924, 7.879992e-27], [3.803587, 7.79349e-27], [3.815287, 7.707999e-27], [3.827022, 7.623505e-27], [3.838793, 7.539993e-27], [3.850601, 7.457455e-27], [3.862445, 7.375877e-27], [3.874325, 7.295246e-27], [3.886242, 7.215554e-27], [3.898195, 7.136787e-27], [3.910186, 7.058932e-27], [3.922213, 6.981981e-27], [3.934277, 6.905923e-27], [3.946378, 6.830743e-27], [3.958517, 6.756435e-27], [3.970693, 6.682987e-27], [3.982906, 6.610385e-27], [3.995157, 6.538624e-27], [4.007446, 6.467691e-27], [4.019772, 6.397575e-27], [4.032136, 6.328269e-27], [4.044538, 6.259761e-27], [4.056979, 6.19204e-27], [4.069458, 6.125099e-27], [4.081975, 6.058928e-27], [4.09453, 5.993516e-27], [4.107124, 5.928856e-27], [4.119757, 5.864938e-27], [4.132429, 5.801752e-27], [4.14514, 5.739291e-27], [4.15789, 5.677545e-27], [4.170679, 5.616504e-27], [4.183507, 5.556162e-27], [4.196375, 5.49651e-27], [4.209282, 5.437538e-27], [4.22223, 5.37924e-27], [4.235217, 5.321607e-27], [4.248244, 5.264629e-27], [4.26131, 5.208302e-27], [4.274418, 5.152616e-27], [4.287565, 5.097561e-27], [4.300753, 5.043134e-27], [4.313982, 4.989325e-27], [4.327251, 4.936126e-27], [4.340561, 4.883531e-27], [4.353912, 4.831533e-27], [4.367304, 4.780123e-27], [4.380737, 4.729295e-27], [4.394211, 4.679043e-27], [4.407728, 4.629358e-27], [4.421285, 4.580236e-27], [4.434884, 4.531669e-27], [4.448525, 4.483648e-27], [4.462208, 4.436171e-27], [4.475934, 4.389228e-27], [4.489701, 4.342814e-27], [4.50351, 4.296923e-27], [4.517363, 4.251549e-27], [4.531257, 4.206684e-27], [4.545195, 4.162324e-27], [4.559175, 4.118462e-27], [4.573199, 4.075093e-27], [4.587265, 4.03221e-27], [4.601375, 3.989809e-27], [4.615528, 3.947882e-27], [4.629725, 3.906426e-27], [4.643965, 3.865434e-27], [4.658249, 3.824901e-27], [4.672578, 3.784821e-27], [4.68695, 3.74519e-27], [4.701366, 3.706001e-27], [4.715827, 3.66725e-27], [4.730332, 3.628932e-27], [4.744882, 3.591042e-27], [4.759476, 3.553574e-27], [4.774116, 3.516525e-27], [4.7888, 3.479889e-27], [4.80353, 3.443661e-27], [4.818305, 3.407837e-27], [4.833126, 3.372412e-27], [4.847992, 3.337381e-27], [4.862903, 3.302741e-27], [4.877861, 3.268486e-27], [4.892864, 3.234613e-27], [4.907914, 3.201117e-27], [4.92301, 3.167993e-27], [4.938153, 3.135238e-27], [4.953342, 3.102848e-27], [4.968578, 3.070817e-27], [4.98386, 3.039144e-27], [4.99919, 3.007823e-27], [5.014567, 2.97685e-27], [5.029991, 2.946222e-27], [5.045462, 2.915935e-27], [5.060981, 2.885985e-27], [5.076548, 2.856369e-27], [5.092163, 2.827083e-27], [5.107826, 2.798123e-27], [5.123537, 2.769486e-27], [5.139296, 2.741169e-27], [5.155104, 2.713167e-27], [5.17096, 2.685479e-27], [5.186865, 2.6581e-27], [5.202819, 2.631027e-27], [5.218822, 2.604257e-27], [5.234875, 2.577787e-27], [5.250977, 2.551613e-27], [5.267128, 2.525733e-27], [5.283329, 2.500144e-27], [5.29958, 2.474842e-27], [5.31588, 2.449825e-27], [5.332231, 2.425091e-27], [5.348632, 2.400634e-27], [5.365084, 2.376454e-27], [5.381586, 2.352548e-27], [5.398139, 2.328913e-27], [5.414743, 2.305545e-27], [5.431398, 2.282444e-27], [5.448104, 2.259604e-27], [5.464862, 2.237026e-27], [5.481671, 2.214705e-27], [5.498532, 2.19264e-27], [5.515445, 2.170828e-27], [5.532409, 2.149268e-27], [5.549426, 2.127955e-27], [5.566496, 2.106889e-27], [5.583617, 2.086067e-27], [5.600792, 2.065487e-27], [5.618019, 2.045147e-27], [5.635299, 2.025045e-27], [5.652633, 2.005179e-27], [5.670019, 1.985547e-27], [5.687459, 1.966147e-27], [5.704953, 1.946978e-27], [5.722501, 1.928037e-27], [5.740102, 1.909323e-27], [5.757758, 1.890835e-27], [5.775468, 1.87257e-27], [5.793233, 1.854528e-27], [5.811052, 1.836707e-27], [5.828926, 1.819104e-27], [5.846855, 1.801721e-27], [5.864839, 1.784553e-27], [5.882878, 1.767602e-27], [5.900973, 1.750865e-27], [5.919124, 1.734341e-27], [5.93733, 1.71803e-27], [5.955592, 1.70193e-27], [5.973911, 1.686041e-27], [5.992286, 1.670361e-27], [6.010717, 1.654891e-27], [6.029205, 1.639628e-27], [6.047751, 1.624573e-27], [6.066352, 1.609725e-27], [6.085012, 1.595082e-27], [6.103728, 1.580646e-27], [6.122503, 1.566416e-27], [6.141335, 1.552389e-27], [6.160224, 1.538568e-27], [6.179172, 1.524951e-27], [6.198179, 1.511537e-27], [6.217243, 1.498327e-27], [6.236367, 1.485321e-27], [6.255549, 1.472517e-27], [6.27479, 1.459917e-27], [6.29409, 1.447519e-27], [6.31345, 1.435324e-27], [6.332869, 1.423331e-27], [6.352348, 1.41154e-27], [6.371887, 1.39995e-27], [6.391486, 1.388563e-27], [6.411146, 1.377377e-27], [6.430866, 1.366392e-27], [6.450646, 1.355608e-27], [6.470487, 1.345024e-27], [6.490389, 1.33464e-27], [6.510353, 1.324456e-27], [6.530378, 1.314471e-27], [6.550464, 1.304683e-27], [6.570613, 1.295094e-27], [6.590823, 1.285701e-27], [6.611095, 1.276504e-27], [6.63143, 1.267502e-27], [6.651827, 1.258694e-27], [6.672288, 1.250079e-27], [6.692811, 1.241655e-27], [6.713397, 1.233421e-27], [6.734046, 1.225376e-27], [6.754759, 1.217518e-27], [6.775536, 1.209845e-27], [6.796376, 1.202355e-27], [6.817281, 1.195047e-27], [6.83825, 1.187918e-27], [6.859284, 1.180967e-27], [6.880382, 1.17419e-27], [6.901545, 1.167587e-27], [6.922773, 1.161153e-27], [6.944066, 1.154888e-27], [6.965426, 1.148788e-27], [6.98685, 1.14285e-27], [7.008341, 1.137072e-27], [7.029897, 1.131451e-27], [7.05152, 1.125984e-27], [7.07321, 1.120668e-27], [7.094966, 1.115501e-27], [7.116789, 1.110479e-27], [7.138679, 1.1056e-27], [7.160637, 1.10086e-27], [7.182662, 1.096256e-27], [7.204755, 1.091786e-27], [7.226916, 1.087446e-27], [7.249145, 1.083234e-27], [7.271442, 1.079146e-27], [7.293808, 1.075179e-27], [7.316243, 1.071331e-27], [7.338746, 1.067599e-27], [7.361319, 1.063979e-27], [7.383962, 1.06047e-27], [7.406674, 1.057068e-27], [7.429456, 1.05377e-27], [7.452307, 1.050575e-27], [7.47523, 1.047478e-27], [7.498222, 1.044479e-27], [7.521286, 1.041574e-27], [7.54442, 1.038761e-27], [7.567626, 1.036038e-27], [7.590903, 1.033402e-27], [7.614251, 1.030851e-27], [7.637672, 1.028383e-27], [7.661164, 1.025996e-27], [7.684729, 1.023688e-27], [7.708366, 1.021457e-27], [7.732076, 1.019301e-27], [7.755859, 1.017218e-27], [7.779714, 1.015206e-27], [7.803644, 1.013264e-27], [7.827647, 1.01139e-27], [7.851723, 1.009581e-27], [7.875874, 1.007838e-27], [7.900099, 1.006157e-27], [7.924399, 1.004537e-27], [7.948773, 1.002977e-27], [7.973222, 1.001475e-27], [7.997747, 1.000029e-27], [8.022347, 9.986376e-28], [8.047022, 9.972994e-28], [8.071774, 9.960126e-28], [8.096602, 9.947753e-28], [8.121506, 9.93586e-28], [8.146486, 9.924429e-28], [8.171544, 9.913444e-28], [8.196678, 9.902888e-28], [8.22189, 9.892748e-28], [8.247179, 9.883008e-28], [8.272546, 9.873656e-28], [8.297992, 9.864681e-28], [8.323515, 9.856073e-28], [8.349117, 9.847824e-28], [8.374798, 9.839928e-28], [8.400557, 9.832377e-28], [8.426396, 9.825168e-28], [8.452315, 9.818297e-28], [8.478313, 9.81176e-28], [8.504391, 9.805551e-28], [8.530549, 9.799667e-28], [8.556788, 9.794105e-28], [8.583107, 9.788858e-28], [8.609508, 9.783923e-28], [8.63599, 9.779293e-28], [8.662553, 9.774961e-28], [8.689197, 9.770925e-28], [8.715924, 9.767174e-28], [8.742733, 9.763703e-28], [8.769624, 9.760505e-28], [8.796598, 9.757574e-28], [8.823656, 9.754901e-28], [8.850796, 9.752481e-28], [8.87802, 9.750305e-28], [8.905327, 9.748367e-28], [8.932719, 9.746659e-28], [8.960195, 9.745175e-28], [8.987755, 9.743908e-28], [9.0154, 9.74285e-28], [9.04313, 9.741996e-28], [9.070945, 9.741338e-28], [9.098846, 9.74087e-28], [9.126833, 9.740585e-28], [9.154906, 9.740478e-28], [9.183065, 9.740541e-28], [9.211311, 9.74077e-28], [9.239644, 9.741157e-28], [9.268063, 9.741697e-28], [9.296571, 9.742383e-28], [9.325166, 9.743211e-28], [9.353848, 9.744175e-28], [9.38262, 9.745269e-28], [9.411479, 9.746487e-28], [9.440428, 9.747824e-28], [9.469465, 9.749275e-28], [9.498592, 9.750835e-28], [9.527808, 9.752498e-28], [9.557114, 9.75426e-28], [9.58651, 9.756116e-28], [9.615997, 9.75806e-28], [9.645575, 9.760088e-28], [9.675243, 9.762195e-28], [9.705003, 9.764378e-28], [9.734854, 9.76663e-28], [9.764797, 9.768948e-28], [9.794832, 9.771327e-28], [9.82496, 9.773763e-28], [9.85518, 9.776251e-28], [9.885493, 9.778787e-28], [9.915899, 9.781369e-28], [9.946399, 9.78399e-28], [9.976993, 9.786647e-28], [10.007681, 9.789336e-28], [10.038463, 9.792054e-28], [10.06934, 9.794796e-28], [10.100312, 9.79756e-28], [10.131379, 9.800341e-28], [10.162542, 9.803135e-28], [10.1938, 9.805938e-28], [10.225155, 9.808749e-28], [10.256606, 9.811563e-28], [10.288154, 9.814376e-28], [10.319799, 9.817187e-28], [10.351541, 9.81999e-28], [10.383381, 9.822782e-28], [10.415319, 9.825562e-28], [10.447355, 9.828326e-28], [10.479489, 9.831069e-28], [10.511723, 9.833792e-28], [10.544055, 9.836488e-28], [10.576487, 9.839156e-28], [10.609019, 9.841794e-28], [10.641651, 9.844398e-28], [10.674383, 9.846965e-28], [10.707216, 9.849493e-28], [10.74015, 9.851979e-28], [10.773185, 9.85442e-28], [10.806322, 9.856815e-28], [10.83956, 9.85916e-28], [10.872901, 9.861454e-28], [10.906345, 9.863692e-28], [10.939891, 9.865874e-28], [10.973541, 9.867997e-28], [11.007294, 9.87006e-28], [11.041151, 9.872058e-28], [11.075112, 9.87399e-28], [11.109177, 9.875855e-28], [11.143347, 9.877649e-28], [11.177623, 9.879371e-28], [11.212004, 9.88102e-28], [11.24649, 9.882592e-28], [11.281083, 9.884086e-28], [11.315782, 9.8855e-28], [11.350587, 9.886832e-28], [11.3855, 9.88808e-28], [11.42052, 9.889243e-28], [11.455648, 9.890317e-28], [11.490884, 9.891303e-28], [11.526229, 9.892198e-28], [11.561682, 9.893001e-28], [11.597244, 9.893709e-28], [11.632915, 9.894322e-28], [11.668696, 9.894837e-28], [11.704588, 9.895253e-28], [11.740589, 9.895568e-28], [11.776702, 9.895782e-28], [11.812925, 9.895892e-28], [11.84926, 9.895897e-28], [11.885707, 9.895796e-28], [11.922265, 9.895587e-28], [11.958937, 9.89527e-28], [11.995721, 9.894842e-28], [12.032618, 9.894303e-28], [12.069628, 9.89365e-28], [12.106753, 9.892884e-28], [12.143991, 9.892002e-28], [12.181345, 9.891005e-28], [12.218813, 9.889889e-28], [12.256396, 9.888655e-28], [12.294095, 9.887302e-28], [12.33191, 9.885828e-28], [12.369841, 9.884232e-28], [12.407889, 9.882513e-28], [12.446054, 9.880671e-28], [12.484336, 9.878704e-28], [12.522736, 9.876612e-28], [12.561254, 9.874393e-28], [12.599891, 9.872048e-28], [12.638646, 9.869573e-28], [12.677521, 9.866972e-28], [12.716515, 9.86424e-28], [12.75563, 9.861378e-28], [12.794864, 9.858387e-28], [12.834219, 9.855263e-28], [12.873696, 9.852008e-28], [12.913293, 9.848621e-28], [12.953013, 9.8451e-28], [12.992854, 9.841446e-28], [13.032818, 9.837661e-28], [13.072906, 9.83374e-28], [13.113116, 9.829687e-28], [13.15345, 9.825501e-28], [13.193908, 9.821177e-28], [13.234491, 9.816722e-28], [13.275198, 9.812137e-28], [13.316031, 9.80741e-28], [13.356989, 9.802557e-28], [13.398073, 9.797585e-28], [13.439284, 9.792474e-28], [13.480621, 9.787219e-28], [13.522086, 9.781831e-28], [13.563678, 9.776336e-28], [13.605398, 9.770716e-28], [13.647246, 9.764963e-28], [13.689223, 9.759103e-28], [13.731329, 9.753125e-28], [13.773565, 9.74703e-28], [13.81593, 9.74084e-28], [13.858426, 9.734551e-28], [13.901052, 9.728166e-28], [13.94381, 9.721713e-28], [13.986699, 9.715193e-28], [14.02972, 9.708615e-28], [14.072874, 9.702021e-28], [14.11616, 9.695424e-28], [14.159579, 9.688848e-28], [14.203132, 9.682364e-28], [14.246819, 9.676013e-28], [14.29064, 9.669861e-28], [14.334596, 9.664058e-28], [14.378687, 9.658729e-28], [14.422914, 9.654062e-28], [14.467277, 9.650463e-28], [14.511776, 9.648395e-28], [14.556412, 9.648521e-28], [14.601186, 9.652375e-28], [14.646097, 9.662402e-28], [14.691146, 9.681712e-28], [14.736334, 9.72068e-28], [14.781661, 9.807499e-28], [14.827127, 1.064759e-27], [14.872734, 1.516649e-27], [14.91848, 1.974594e-27], [14.964367, 1.876603e-27], [15.010395, 1.376186e-27], [15.056565, 1.031027e-27], [15.102877, 9.828599e-28], [15.149331, 9.665147e-28], [15.195929, 9.579684e-28], [15.242669, 9.529715e-28], [15.289553, 9.494343e-28], [15.336582, 9.46728e-28], [15.383755, 9.444711e-28], [15.431073, 9.424871e-28], [15.478537, 9.406898e-28], [15.526147, 9.390129e-28], [15.573903, 9.374132e-28], [15.621806, 9.35869e-28], [15.669857, 9.343619e-28], [15.718055, 9.328778e-28], [15.766401, 9.314093e-28], [15.814897, 9.2995e-28], [15.863541, 9.28494e-28], [15.912335, 9.270384e-28], [15.961279, 9.255806e-28], [16.010374, 9.241178e-28], [16.059619, 9.226488e-28], [16.109016, 9.211722e-28], [16.158565, 9.196868e-28], [16.208267, 9.181919e-28], [16.258121, 9.166869e-28], [16.308129, 9.151711e-28], [16.35829, 9.136441e-28], [16.408606, 9.121057e-28], [16.459077, 9.105555e-28], [16.509702, 9.089933e-28], [16.560484, 9.07419e-28], [16.611422, 9.058324e-28], [16.662516, 9.042335e-28], [16.713768, 9.026223e-28], [16.765177, 9.009985e-28], [16.816744, 8.993623e-28], [16.86847, 8.977139e-28], [16.920355, 8.960529e-28], [16.9724, 8.943796e-28], [17.024604, 8.926941e-28], [17.07697, 8.909963e-28], [17.129496, 8.892864e-28], [17.182184, 8.875646e-28], [17.235034, 8.858306e-28], [17.288046, 8.84085e-28], [17.341222, 8.82328e-28], [17.394561, 8.805591e-28], [17.448064, 8.787788e-28], [17.501732, 8.769877e-28], [17.555565, 8.751855e-28], [17.609563, 8.73372e-28], [17.663728, 8.715476e-28], [17.718059, 8.697133e-28], [17.772557, 8.678688e-28], [17.827223, 8.660138e-28], [17.882057, 8.641493e-28], [17.93706, 8.622751e-28], [17.992231, 8.603914e-28], [18.047573, 8.584988e-28], [18.103085, 8.565974e-28], [18.158767, 8.546873e-28], [18.214621, 8.527692e-28], [18.270647, 8.508431e-28], [18.326845, 8.489093e-28], [18.383215, 8.469684e-28], [18.439759, 8.450207e-28], [18.496478, 8.430663e-28], [18.55337, 8.41106e-28], [18.610438, 8.3914e-28], [18.667681, 8.371687e-28], [18.7251, 8.351929e-28], [18.782696, 8.332128e-28], [18.840468, 8.312291e-28], [18.898419, 8.292424e-28], [18.956548, 8.272533e-28], [19.014856, 8.252624e-28], [19.073343, 8.232706e-28], [19.132009, 8.212788e-28], [19.190857, 8.192876e-28], [19.249885, 8.172983e-28], [19.309095, 8.153119e-28], [19.368487, 8.133296e-28], [19.428062, 8.113529e-28], [19.48782, 8.093833e-28], [19.547762, 8.074226e-28], [19.607888, 8.054732e-28], [19.668199, 8.035374e-28], [19.728695, 8.016179e-28], [19.789378, 7.997187e-28], [19.850247, 7.978439e-28], [19.911304, 7.959983e-28], [19.972548, 7.94189e-28], [20.033981, 7.924241e-28], [20.095603, 7.907134e-28], [20.157414, 7.890713e-28], [20.219415, 7.875163e-28], [20.281607, 7.860711e-28], [20.343991, 7.847708e-28], [20.406566, 7.836664e-28], [20.469333, 7.828219e-28], [20.532294, 7.823462e-28], [20.595449, 7.824244e-28], [20.658797, 7.832931e-28], [20.722341, 7.854401e-28], [20.78608, 7.899394e-28], [20.850015, 7.990139e-28], [20.914147, 8.192164e-28], [20.978475, 8.59472e-28], [21.043002, 9.90733e-28], [21.107728, 1.383108e-27], [21.172652, 1.683521e-27], [21.237776, 1.635834e-27], [21.3031, 1.232569e-27], [21.368626, 8.466577e-28], [21.434353, 7.858574e-28], [21.500282, 7.694076e-28], [21.566413, 7.599311e-28], [21.632749, 7.532754e-28], [21.699288, 7.482607e-28], [21.766032, 7.443135e-28], [21.832981, 7.410058e-28], [21.900136, 7.381299e-28], [21.967498, 7.355681e-28], [22.035067, 7.332307e-28], [22.102844, 7.310633e-28], [22.170829, 7.290325e-28], [22.239023, 7.271116e-28], [22.307427, 7.252823e-28], [22.376042, 7.235331e-28], [22.444867, 7.218543e-28], [22.513904, 7.202389e-28], [22.583154, 7.186826e-28], [22.652617, 7.171816e-28], [22.722293, 7.157332e-28], [22.792183, 7.143356e-28], [22.862289, 7.129877e-28], [22.93261, 7.116883e-28], [23.003147, 7.104371e-28], [23.073902, 7.092337e-28], [23.144874, 7.080781e-28], [23.216064, 7.069703e-28], [23.287474, 7.059106e-28], [23.359102, 7.048991e-28], [23.430952, 7.039363e-28], [23.503022, 7.030225e-28], [23.575314, 7.021582e-28], [23.647828, 7.013438e-28], [23.720566, 7.005799e-28], [23.793527, 6.998667e-28], [23.866712, 6.992049e-28], [23.940123, 6.985948e-28], [24.013759, 6.980368e-28], [24.087622, 6.975314e-28], [24.161712, 6.970789e-28], [24.23603, 6.966796e-28], [24.310577, 6.963338e-28], [24.385353, 6.960418e-28], [24.460359, 6.958036e-28], [24.535595, 6.956196e-28], [24.611063, 6.954896e-28], [24.686763, 6.954139e-28], [24.762696, 6.953923e-28], [24.838863, 6.954248e-28], [24.915264, 6.955112e-28], [24.9919, 6.956514e-28], [25.068771, 6.95845e-28], [25.145879, 6.960918e-28], [25.223224, 6.963914e-28], [25.300807, 6.967434e-28], [25.378629, 6.971471e-28], [25.45669, 6.976022e-28], [25.534991, 6.981079e-28], [25.613533, 6.986636e-28], [25.692317, 6.992686e-28], [25.771343, 6.99922e-28], [25.850612, 7.006231e-28], [25.930124, 7.013709e-28], [26.009882, 7.021645e-28], [26.089885, 7.030029e-28], [26.170133, 7.038851e-28], [26.250629, 7.048099e-28], [26.331372, 7.057764e-28], [26.412364, 7.067832e-28], [26.493604, 7.078292e-28], [26.575095, 7.089131e-28], [26.656836, 7.100338e-28], [26.738829, 7.111897e-28], [26.821074, 7.123796e-28], [26.903571, 7.136021e-28], [26.986323, 7.148558e-28], [27.069329, 7.161391e-28], [27.152591, 7.174507e-28], [27.236108, 7.18789e-28], [27.319882, 7.201524e-28], [27.403915, 7.215395e-28], [27.488205, 7.229485e-28], [27.572755, 7.243779e-28], [27.657565, 7.25826e-28], [27.742636, 7.272911e-28], [27.827968, 7.287715e-28], [27.913563, 7.302655e-28], [27.999421, 7.317713e-28], [28.085543, 7.332871e-28], [28.17193, 7.348112e-28], [28.258583, 7.363415e-28], [28.345503, 7.378765e-28], [28.432689, 7.39414e-28], [28.520144, 7.409523e-28], [28.607868, 7.424893e-28], [28.695862, 7.440233e-28], [28.784126, 7.455521e-28], [28.872662, 7.470739e-28], [28.96147, 7.485866e-28], [29.050552, 7.500883e-28], [29.139907, 7.515768e-28], [29.229537, 7.530503e-28], [29.319443, 7.545067e-28], [29.409626, 7.559439e-28], [29.500085, 7.5736e-28], [29.590824, 7.587529e-28], [29.681841, 7.601206e-28], [29.773138, 7.61461e-28], [29.864716, 7.627723e-28], [29.956575, 7.640524e-28], [30.048718, 7.652994e-28], [30.141143, 7.665113e-28], [30.233853, 7.676862e-28], [30.326848, 7.688223e-28], [30.420129, 7.699177e-28], [30.513697, 7.709706e-28], [30.607553, 7.719794e-28], [30.701697, 7.729421e-28], [30.796131, 7.738573e-28], [30.890856, 7.747234e-28], [30.985872, 7.755386e-28], [31.08118, 7.763017e-28], [31.176781, 7.770111e-28], [31.272677, 7.776656e-28], [31.368867, 7.782637e-28], [31.465353, 7.788044e-28], [31.562136, 7.792865e-28], [31.659217, 7.797089e-28], [31.756596, 7.800707e-28], [31.854275, 7.803709e-28], [31.952254, 7.806088e-28], [32.050535, 7.807836e-28], [32.149118, 7.808947e-28], [32.248004, 7.809416e-28], [32.347194, 7.809237e-28], [32.446689, 7.808407e-28], [32.546491, 7.806922e-28], [32.646599, 7.804782e-28], [32.747015, 7.801985e-28], [32.847741, 7.798529e-28], [32.948776, 7.794417e-28], [33.050121, 7.789649e-28], [33.151779, 7.784227e-28], [33.253749, 7.778155e-28], [33.356033, 7.771436e-28], [33.458631, 7.764074e-28], [33.561545, 7.756076e-28], [33.664776, 7.747446e-28], [33.768324, 7.738192e-28], [33.87219, 7.728321e-28], [33.976376, 7.71784e-28], [34.080883, 7.706759e-28], [34.185711, 7.695086e-28], [34.290861, 7.682832e-28], [34.396335, 7.670006e-28], [34.502133, 7.656619e-28], [34.608257, 7.642682e-28], [34.714707, 7.628206e-28], [34.821485, 7.613205e-28], [34.92859, 7.597689e-28], [35.036026, 7.581671e-28], [35.143792, 7.565165e-28], [35.251889, 7.548183e-28], [35.360319, 7.530739e-28], [35.469082, 7.512846e-28], [35.57818, 7.494517e-28], [35.687613, 7.475768e-28], [35.797383, 7.456611e-28], [35.907491, 7.43706e-28], [36.017937, 7.41713e-28], [36.128724, 7.396835e-28], [36.23985, 7.376188e-28], [36.351319, 7.355204e-28], [36.46313, 7.333897e-28], [36.575286, 7.312281e-28], [36.687786, 7.290369e-28], [36.800633, 7.268176e-28], [36.913826, 7.245714e-28], [37.027368, 7.222998e-28], [37.141259, 7.20004e-28], [37.2555, 7.176855e-28], [37.370093, 7.153454e-28], [37.485038, 7.129851e-28], [37.600336, 7.106059e-28], [37.71599, 7.082089e-28], [37.831999, 7.057954e-28], [37.948365, 7.033666e-28], [38.065088, 7.009236e-28], [38.182171, 6.984677e-28], [38.299614, 6.959998e-28], [38.417418, 6.935212e-28], [38.535585, 6.910329e-28], [38.654115, 6.885359e-28], [38.773009, 6.860313e-28], [38.89227, 6.8352e-28], [39.011897, 6.81003e-28], [39.131892, 6.784813e-28], [39.252256, 6.759557e-28], [39.37299, 6.734272e-28], [39.494096, 6.708966e-28], [39.615574, 6.683647e-28], [39.737426, 6.658324e-28], [39.859653, 6.633004e-28], [39.982255, 6.607695e-28], [40.105235, 6.582404e-28], [40.228593, 6.557139e-28], [40.352331, 6.531905e-28], [40.476449, 6.50671e-28], [40.600948, 6.48156e-28], [40.725831, 6.456461e-28], [40.851098, 6.431419e-28], [40.97675, 6.406439e-28], [41.102789, 6.381527e-28], [41.229215, 6.356688e-28], [41.35603, 6.331927e-28], [41.483236, 6.307249e-28], [41.610832, 6.282658e-28], [41.738821, 6.258159e-28], [41.867204, 6.233756e-28], [41.995982, 6.209453e-28], [42.125155, 6.185253e-28], [42.254726, 6.16116e-28], [42.384696, 6.137178e-28], [42.515065, 6.11331e-28], [42.645835, 6.089558e-28], [42.777008, 6.065927e-28], [42.908584, 6.042418e-28], [43.040565, 6.019034e-28], [43.172951, 5.995778e-28], [43.305745, 5.972652e-28], [43.438947, 5.949659e-28], [43.572559, 5.9268e-28], [43.706582, 5.904077e-28], [43.841018, 5.881492e-28], [43.975866, 5.859048e-28], [44.11113, 5.836745e-28], [44.24681, 5.814586e-28], [44.382906, 5.792572e-28], [44.519422, 5.770706e-28], [44.656357, 5.748989e-28], [44.793714, 5.727426e-28], [44.931493, 5.70602e-28], [45.069696, 5.684785e-28], [45.208324, 5.663789e-28], [45.347378, 5.643485e-28], [45.48686, 5.62363e-28], [45.626772, 5.603081e-28], [45.767113, 5.582178e-28], [45.907886, 5.561338e-28], [46.049092, 5.540696e-28], [46.190733, 5.520173e-28], [46.332809, 5.499967e-28], [46.475322, 5.480103e-28], [46.618273, 5.460444e-28], [46.761665, 5.440951e-28], [46.905497, 5.421617e-28], [47.049772, 5.402439e-28], [47.19449, 5.383414e-28], [47.339654, 5.364543e-28], [47.485264, 5.345824e-28], [47.631321, 5.327255e-28], [47.777829, 5.308837e-28], [47.924786, 5.290569e-28], [48.072196, 5.27245e-28], [48.22006, 5.25448e-28], [48.368378, 5.236658e-28], [48.517152, 5.218983e-28], [48.666384, 5.201455e-28], [48.816075, 5.184073e-28], [48.966226, 5.166836e-28], [49.116839, 5.149744e-28], [49.267916, 5.132795e-28], [49.419457, 5.115989e-28], [49.571464, 5.099326e-28], [49.723939, 5.082804e-28], [49.876882, 5.066422e-28], [50.030297, 5.05018e-28], [50.184183, 5.034076e-28], [50.338542, 5.01811e-28], [50.493376, 5.002281e-28], [50.648687, 4.986588e-28], [50.804475, 4.97103e-28], [50.960742, 4.955606e-28], [51.11749, 4.940315e-28], [51.27472, 4.925157e-28], [51.432434, 4.910129e-28], [51.590633, 4.895231e-28], [51.749319, 4.880462e-28], [51.908492, 4.865822e-28], [52.068155, 4.851308e-28], [52.228309, 4.83692e-28], [52.388956, 4.822657e-28], [52.550097, 4.808518e-28], [52.711734, 4.794502e-28], [52.873868, 4.780608e-28], [53.0365, 4.766834e-28], [53.199633, 4.75318e-28], [53.363268, 4.739645e-28], [53.527405, 4.726227e-28], [53.692048, 4.712926e-28], [53.857197, 4.69974e-28], [54.022854, 4.686668e-28], [54.189021, 4.67371e-28], [54.355699, 4.660864e-28], [54.522889, 4.648129e-28], [54.690594, 4.635504e-28], [54.858814, 4.622988e-28], [55.027552, 4.61058e-28], [55.196809, 4.598279e-28], [55.366587, 4.586084e-28], [55.536886, 4.573994e-28], [55.70771, 4.562007e-28], [55.879059, 4.550123e-28], [56.050935, 4.538341e-28], [56.22334, 4.52666e-28], [56.396275, 4.515078e-28], [56.569742, 4.503595e-28], [56.743742, 4.49221e-28], [56.918278, 4.480921e-28], [57.09335, 4.469728e-28], [57.268961, 4.458629e-28], [57.445113, 4.447624e-28], [57.621806, 4.436712e-28], [57.799042, 4.425891e-28], [57.976824, 4.415161e-28], [58.155152, 4.40452e-28], [58.334029, 4.393969e-28], [58.513456, 4.383505e-28], [58.693435, 4.373127e-28], [58.873968, 4.362836e-28], [59.055056, 4.35263e-28], [59.236701, 4.342507e-28], [59.418905, 4.332468e-28], [59.601669, 4.322511e-28], [59.784995, 4.312635e-28], [59.968885, 4.30284e-28], [60.153341, 4.293124e-28], [60.338364, 4.283486e-28], [60.523956, 4.273926e-28], [60.710119, 4.264444e-28], [60.896855, 4.255037e-28], [61.084165, 4.245705e-28], [61.272051, 4.236447e-28], [61.460516, 4.227263e-28], [61.649559, 4.218152e-28], [61.839185, 4.209112e-28], [62.029393, 4.200143e-28], [62.220187, 4.191245e-28], [62.411567, 4.182416e-28], [62.603536, 4.173655e-28], [62.796096, 4.164962e-28], [62.989248, 4.156337e-28], [63.182994, 4.147777e-28], [63.377336, 4.139283e-28], [63.572275, 4.130854e-28], [63.767815, 4.122488e-28], [63.963955, 4.114186e-28], [64.160699, 4.105947e-28], [64.358049, 4.097769e-28], [64.556005, 4.089652e-28], [64.75457, 4.081596e-28], [64.953746, 4.073599e-28], [65.153534, 4.065662e-28], [65.353937, 4.057782e-28], [65.554957, 4.049961e-28], [65.756594, 4.042196e-28], [65.958852, 4.034488e-28], [66.161732, 4.026835e-28], [66.365237, 4.019237e-28], [66.569367, 4.011694e-28], [66.774125, 4.004204e-28], [66.979512, 3.996768e-28], [67.185532, 3.989384e-28], [67.392185, 3.982052e-28], [67.599474, 3.974771e-28], [67.8074, 3.967541e-28], [68.015966, 3.960361e-28], [68.225173, 3.953231e-28], [68.435024, 3.946149e-28], [68.645521, 3.939116e-28], [68.856665, 3.932131e-28], [69.068458, 3.925193e-28], [69.280903, 3.918302e-28], [69.494001, 3.911456e-28], [69.707755, 3.904657e-28], [69.922166, 3.897903e-28], [70.137236, 3.891193e-28], [70.352969, 3.884528e-28], [70.569364, 3.877906e-28], [70.786426, 3.871327e-28], [71.004155, 3.86479e-28], [71.222553, 3.858296e-28], [71.441624, 3.851844e-28], [71.661368, 3.845432e-28], [71.881788, 3.839062e-28], [72.102886, 3.832731e-28], [72.324665, 3.82644e-28], [72.547125, 3.820189e-28], [72.77027, 3.813976e-28], [72.994101, 3.807802e-28], [73.21862, 3.801666e-28], [73.44383, 3.795567e-28], [73.669733, 3.789506e-28], [73.896331, 3.783481e-28], [74.123625, 3.777493e-28], [74.351619, 3.77154e-28], [74.580314, 3.765623e-28], [74.809713, 3.759741e-28], [75.039817, 3.753894e-28], [75.270628, 3.748081e-28], [75.50215, 3.742302e-28], [75.734384, 3.736557e-28], [75.967332, 3.730845e-28], [76.200997, 3.725166e-28], [76.43538, 3.719519e-28], [76.670485, 3.713904e-28], [76.906312, 3.708321e-28], [77.142865, 3.70277e-28], [77.380146, 3.697249e-28], [77.618156, 3.69176e-28], [77.856898, 3.686301e-28], [78.096375, 3.680871e-28], [78.336589, 3.675472e-28], [78.577541, 3.670102e-28], [78.819234, 3.664761e-28], [79.061671, 3.659449e-28], [79.304853, 3.654166e-28], [79.548784, 3.648911e-28], [79.793465, 3.643683e-28], [80.038898, 3.638484e-28], [80.285086, 3.633312e-28], [80.532032, 3.628166e-28], [80.779737, 3.623048e-28], [81.028204, 3.617956e-28], [81.277435, 3.61289e-28], [81.527433, 3.60785e-28], [81.7782, 3.602836e-28], [82.029738, 3.597848e-28], [82.28205, 3.592884e-28], [82.535138, 3.587946e-28], [82.789004, 3.583032e-28], [83.043651, 3.578143e-28], [83.299082, 3.573277e-28], [83.555298, 3.568436e-28], [83.812302, 3.563619e-28], [84.070097, 3.558825e-28], [84.328684, 3.554054e-28], [84.588067, 3.549306e-28], [84.848248, 3.544581e-28], [85.10923, 3.539878e-28], [85.371013, 3.535198e-28], [85.633603, 3.530541e-28], [85.896999, 3.525905e-28], [86.161206, 3.52129e-28], [86.426226, 3.516698e-28], [86.692061, 3.512126e-28], [86.958713, 3.507576e-28], [87.226186, 3.503047e-28], [87.494481, 3.498538e-28], [87.763602, 3.49405e-28], [88.03355, 3.489582e-28], [88.304329, 3.485135e-28], [88.57594, 3.480707e-28], [88.848387, 3.4763e-28], [89.121672, 3.471912e-28], [89.395798, 3.467543e-28], [89.670767, 3.463194e-28], [89.946581, 3.458864e-28], [90.223244, 3.454553e-28], [90.500758, 3.45026e-28], [90.779126, 3.445987e-28], [91.058349, 3.441732e-28], [91.338432, 3.437495e-28], [91.619376, 3.433276e-28], [91.901184, 3.429075e-28], [92.183859, 3.424892e-28], [92.467404, 3.420727e-28], [92.75182, 3.41658e-28], [93.037112, 3.41245e-28], [93.32328, 3.408337e-28], [93.61033, 3.404241e-28], [93.898262, 3.400163e-28], [94.187079, 3.396101e-28], [94.476786, 3.392056e-28], [94.767383, 3.388028e-28], [95.058874, 3.384016e-28], [95.351261, 3.38002e-28], [95.644548, 3.376041e-28], [95.938737, 3.372078e-28], [96.233831, 3.368131e-28], [96.529833, 3.3642e-28], [96.826745, 3.360284e-28], [97.12457, 3.356384e-28], [97.423312, 3.3525e-28], [97.722972, 3.348631e-28], [98.023554, 3.344778e-28], [98.32506, 3.34094e-28], [98.627494, 3.337116e-28], [98.930858, 3.333308e-28], [99.235156, 3.329515e-28], [99.540389, 3.325736e-28], [99.846561, 3.321973e-28], [100.15367, 3.318223e-28], [100.46173, 3.314489e-28], [100.77074, 3.310768e-28], [101.0807, 3.307063e-28], [101.39161, 3.303371e-28], [101.70347, 3.299693e-28], [102.0163, 3.29603e-28], [102.33008, 3.29238e-28], [102.64484, 3.288744e-28], [102.96056, 3.285122e-28], [103.27725, 3.281514e-28], [103.59492, 3.277919e-28], [103.91356, 3.274338e-28], [104.23318, 3.27077e-28], [104.55379, 3.267216e-28], [104.87538, 3.263675e-28], [105.19796, 3.260147e-28], [105.52154, 3.256633e-28], [105.84611, 3.253131e-28], [106.17168, 3.249643e-28], [106.49824, 3.246167e-28], [106.82582, 3.242704e-28], [107.1544, 3.239254e-28], [107.48399, 3.235817e-28], [107.8146, 3.232392e-28], [108.14622, 3.22898e-28], [108.47886, 3.225581e-28], [108.81253, 3.222194e-28], [109.14722, 3.218819e-28], [109.48294, 3.215457e-28], [109.81969, 3.212107e-28], [110.15748, 3.208769e-28], [110.49631, 3.205444e-28], [110.83618, 3.20213e-28], [111.1771, 3.198829e-28], [111.51906, 3.19554e-28], [111.86208, 3.192262e-28], [112.20615, 3.188997e-28], [112.55128, 3.185743e-28], [112.89747, 3.182501e-28], [113.24473, 3.179271e-28], [113.59306, 3.176052e-28], [113.94245, 3.172846e-28], [114.29292, 3.16965e-28], [114.64447, 3.166467e-28], [114.9971, 3.163294e-28], [115.35082, 3.160134e-28], [115.70562, 3.156984e-28], [116.06151, 3.153846e-28], [116.4185, 3.15072e-28], [116.77659, 3.147605e-28], [117.13578, 3.1445e-28], [117.49607, 3.141408e-28], [117.85747, 3.138326e-28], [118.21998, 3.135255e-28], [118.58361, 3.132196e-28], [118.94836, 3.129147e-28], [119.31423, 3.12611e-28], [119.68122, 3.123084e-28], [120.04934, 3.120068e-28], [120.4186, 3.117064e-28], [120.78899, 3.11407e-28], [121.16052, 3.111087e-28], [121.53319, 3.108115e-28], [121.90701, 3.105154e-28], [122.28198, 3.102203e-28], [122.6581, 3.099264e-28], [123.03538, 3.096334e-28], [123.41382, 3.093416e-28], [123.79342, 3.090508e-28], [124.17419, 3.087611e-28], [124.55613, 3.084725e-28], [124.93925, 3.081849e-28], [125.32355, 3.078983e-28], [125.70902, 3.076128e-28], [126.09569, 3.073284e-28], [126.48354, 3.07045e-28], [126.87259, 3.067626e-28], [127.26283, 3.064813e-28], [127.65427, 3.06201e-28], [128.04692, 3.059218e-28], [128.44077, 3.056436e-28], [128.83584, 3.053664e-28], [129.23212, 3.050903e-28], [129.62962, 3.048151e-28], [130.02834, 3.045411e-28], [130.42829, 3.04268e-28], [130.82947, 3.039959e-28], [131.23188, 3.037249e-28], [131.63553, 3.034549e-28], [132.04042, 3.03186e-28], [132.44656, 3.02918e-28], [132.85395, 3.02651e-28], [133.26259, 3.023851e-28], [133.67248, 3.021202e-28], [134.08364, 3.018563e-28], [134.49606, 3.015934e-28], [134.90975, 3.013315e-28], [135.32472, 3.010706e-28], [135.74096, 3.008107e-28], [136.15848, 3.005518e-28], [136.57728, 3.00294e-28], [136.99737, 3.000371e-28], [137.41876, 2.997812e-28], [137.84144, 2.995264e-28], [138.26542, 2.992725e-28], [138.6907, 2.990196e-28], [139.1173, 2.987678e-28], [139.5452, 2.985169e-28], [139.97442, 2.98267e-28], [140.40496, 2.980181e-28], [140.83683, 2.977703e-28], [141.27002, 2.975234e-28], [141.70455, 2.972775e-28], [142.14041, 2.970326e-28], [142.57762, 2.967886e-28], [143.01617, 2.965457e-28], [143.45606, 2.963038e-28], [143.89731, 2.960628e-28], [144.33992, 2.958229e-28], [144.78389, 2.955839e-28], [145.22922, 2.95346e-28], [145.67593, 2.95109e-28], [146.12401, 2.94873e-28], [146.57346, 2.94638e-28], [147.0243, 2.94404e-28], [147.47653, 2.94171e-28], [147.93015, 2.939389e-28], [148.38516, 2.937079e-28], [148.84157, 2.934779e-28], [149.29939, 2.932488e-28], [149.75861, 2.930207e-28], [150.21925, 2.927936e-28], [150.6813, 2.925675e-28], [151.14477, 2.923424e-28], [151.60967, 2.921183e-28], [152.076, 2.918952e-28], [152.54377, 2.916731e-28], [153.01297, 2.914519e-28], [153.48362, 2.912317e-28], [153.95571, 2.910126e-28], [154.42926, 2.907944e-28], [154.90426, 2.905772e-28], [155.38072, 2.90361e-28], [155.85865, 2.901459e-28], [156.33805, 2.899317e-28], [156.81892, 2.897185e-28], [157.30128, 2.895062e-28], [157.78511, 2.89295e-28], [158.27044, 2.890848e-28], [158.75725, 2.888756e-28], [159.24557, 2.886674e-28], [159.73539, 2.884601e-28], [160.22671, 2.882539e-28], [160.71954, 2.880487e-28], [161.21389, 2.878445e-28], [161.70976, 2.876412e-28], [162.20716, 2.87439e-28], [162.70609, 2.872378e-28], [163.20655, 2.870376e-28], [163.70855, 2.868384e-28], [164.21209, 2.866402e-28], [164.71719, 2.86443e-28], [165.22383, 2.862468e-28], [165.73204, 2.860516e-28], [166.24181, 2.858574e-28], [166.75314, 2.856643e-28], [167.26605, 2.854721e-28], [167.78054, 2.85281e-28], [168.29661, 2.850909e-28], [168.81426, 2.849018e-28], [169.33351, 2.847137e-28], [169.85436, 2.845266e-28], [170.37681, 2.843406e-28], [170.90086, 2.841556e-28], [171.42653, 2.839716e-28], [171.95381, 2.837886e-28], [172.48272, 2.836067e-28], [173.01325, 2.834258e-28], [173.54541, 2.832459e-28], [174.07921, 2.83067e-28], [174.61466, 2.828892e-28], [175.15175, 2.827124e-28], [175.69049, 2.825367e-28], [176.23089, 2.823619e-28], [176.77295, 2.821883e-28], [177.31668, 2.820156e-28], [177.86208, 2.81844e-28], [178.40916, 2.816735e-28], [178.95792, 2.81504e-28], [179.50837, 2.813355e-28], [180.06051, 2.811681e-28], [180.61435, 2.810018e-28], [181.16989, 2.808365e-28], [181.72715, 2.806722e-28], [182.28611, 2.80509e-28], [182.8468, 2.803469e-28], [183.40921, 2.801858e-28], [183.97335, 2.800258e-28], [184.53923, 2.798669e-28], [185.10684, 2.79709e-28], [185.67621, 2.795522e-28], [186.24732, 2.793965e-28], [186.82019, 2.792418e-28], [187.39482, 2.790882e-28], [187.97122, 2.789357e-28], [188.54939, 2.787843e-28], [189.12935, 2.78634e-28], [189.71108, 2.784847e-28], [190.2946, 2.783365e-28], [190.87992, 2.781894e-28], [191.46704, 2.780434e-28], [192.05597, 2.778985e-28], [192.64671, 2.777547e-28], [193.23926, 2.776121e-28], [193.83364, 2.774704e-28], [194.42984, 2.7733e-28], [195.02788, 2.771906e-28], [195.62776, 2.770523e-28], [196.22948, 2.769151e-28], [196.83305, 2.767791e-28], [197.43849, 2.766441e-28], [198.04578, 2.765103e-28], [198.65494, 2.763776e-28], [199.26597, 2.76246e-28], [199.87889, 2.761156e-28], [200.49369, 2.759863e-28], [201.11038, 2.758581e-28], [201.72896, 2.75731e-28], [202.34945, 2.756051e-28], [202.97185, 2.754803e-28], [203.59616, 2.753567e-28], [204.2224, 2.752342e-28], [204.85056, 2.751129e-28], [205.48065, 2.749927e-28], [206.11268, 2.748737e-28], [206.74665, 2.747558e-28], [207.38257, 2.746391e-28], [208.02045, 2.745236e-28], [208.66029, 2.744092e-28], [209.3021, 2.74296e-28], [209.94589, 2.74184e-28], [210.59165, 2.740731e-28], [211.2394, 2.739635e-28], [211.88914, 2.73855e-28], [212.54088, 2.737477e-28], [213.19463, 2.736416e-28], [213.85039, 2.735366e-28], [214.50816, 2.734329e-28], [215.16796, 2.733304e-28], [215.82978, 2.732291e-28], [216.49364, 2.731289e-28], [217.15955, 2.7303e-28], [217.8275, 2.729323e-28], [218.4975, 2.728358e-28], [219.16957, 2.727406e-28], [219.84371, 2.726465e-28], [220.51991, 2.725537e-28], [221.1982, 2.724621e-28], [221.87858, 2.723718e-28], [222.56104, 2.722827e-28], [223.24561, 2.721948e-28], [223.93228, 2.721082e-28], [224.62107, 2.720228e-28], [225.31197, 2.719386e-28], [226.005, 2.718558e-28], [226.70016, 2.717741e-28], [227.39745, 2.716938e-28], [228.09689, 2.716147e-28], [228.79849, 2.715369e-28], [229.50224, 2.714603e-28], [230.20816, 2.71385e-28], [230.91624, 2.71311e-28], [231.62651, 2.712383e-28], [232.33896, 2.711669e-28], [233.0536, 2.710968e-28], [233.77044, 2.710279e-28], [234.48949, 2.709604e-28], [235.21074, 2.708942e-28], [235.93422, 2.708292e-28], [236.65992, 2.707656e-28], [237.38785, 2.707033e-28], [238.11802, 2.706424e-28], [238.85044, 2.705827e-28], [239.58511, 2.705244e-28], [240.32204, 2.704674e-28], [241.06123, 2.704117e-28], [241.8027, 2.703574e-28], [242.54646, 2.703045e-28], [243.29249, 2.702528e-28], [244.04083, 2.702026e-28], [244.79146, 2.701537e-28], [245.54441, 2.701061e-28], [246.29967, 2.700599e-28], [247.05725, 2.700151e-28], [247.81716, 2.699717e-28], [248.57941, 2.699296e-28], [249.34401, 2.69889e-28], [250.11095, 2.698497e-28], [250.88026, 2.698118e-28], [251.65193, 2.697753e-28], [252.42598, 2.697402e-28], [253.2024, 2.697065e-28], [253.98122, 2.696742e-28], [254.76243, 2.696433e-28], [255.54604, 2.696139e-28], [256.33206, 2.695859e-28], [257.1205, 2.695593e-28], [257.91137, 2.695341e-28], [258.70467, 2.695104e-28], [259.50041, 2.694881e-28], [260.29859, 2.694673e-28], [261.09924, 2.694479e-28], [261.90234, 2.6943e-28], [262.70792, 2.694135e-28], [263.51597, 2.693985e-28], [264.3265, 2.69385e-28], [265.13954, 2.693729e-28], [265.95507, 2.693623e-28], [266.77311, 2.693532e-28], [267.59366, 2.693457e-28], [268.41674, 2.693395e-28], [269.24236, 2.693349e-28], [270.07051, 2.693318e-28], [270.90121, 2.693303e-28], [271.73446, 2.693302e-28], [272.57028, 2.693316e-28], [273.40866, 2.693346e-28], [274.24963, 2.693391e-28], [275.09318, 2.693451e-28], [275.93933, 2.693527e-28], [276.78808, 2.693618e-28], [277.63944, 2.693725e-28], [278.49342, 2.693847e-28], [279.35003, 2.693985e-28], [280.20927, 2.694138e-28], [281.07115, 2.694308e-28], [281.93569, 2.694493e-28], [282.80288, 2.694693e-28], [283.67274, 2.69491e-28], [284.54528, 2.695143e-28], [285.4205, 2.695391e-28], [286.29841, 2.695656e-28], [287.17903, 2.695936e-28], [288.06235, 2.696233e-28], [288.94839, 2.696546e-28], [289.83715, 2.696875e-28], [290.72865, 2.697221e-28], [291.62289, 2.697583e-28], [292.51988, 2.697961e-28], [293.41963, 2.698356e-28], [294.32215, 2.698767e-28], [295.22744, 2.699195e-28], [296.13552, 2.699639e-28], [297.04639, 2.7001e-28], [297.96006, 2.700578e-28], [298.87654, 2.701073e-28], [299.79584, 2.701584e-28], [300.71797, 2.702113e-28], [301.64294, 2.702658e-28], [302.57075, 2.703221e-28], [303.50142, 2.7038e-28], [304.43494, 2.704397e-28], [305.37134, 2.70501e-28], [306.31062, 2.705642e-28], [307.25279, 2.70629e-28], [308.19785, 2.706956e-28], [309.14583, 2.707639e-28], [310.09672, 2.708339e-28], [311.05053, 2.709058e-28], [312.00728, 2.709793e-28], [312.96697, 2.710547e-28], [313.92961, 2.711318e-28], [314.89521, 2.712107e-28], [315.86378, 2.712914e-28], [316.83534, 2.713739e-28], [317.80988, 2.714581e-28], [318.78741, 2.715442e-28], [319.76796, 2.716321e-28], [320.75152, 2.717218e-28], [321.73811, 2.718133e-28], [322.72773, 2.719067e-28], [323.72039, 2.720018e-28], [324.71611, 2.720989e-28], [325.71489, 2.721977e-28], [326.71674, 2.722984e-28], [327.72168, 2.72401e-28], [328.7297, 2.725055e-28], [329.74083, 2.726118e-28], [330.75506, 2.7272e-28], [331.77242, 2.728301e-28], [332.7929, 2.72942e-28], [333.81653, 2.730559e-28], [334.8433, 2.731717e-28], [335.87323, 2.732894e-28], [336.90633, 2.73409e-28], [337.94261, 2.735305e-28], [338.98207, 2.73654e-28], [340.02473, 2.737794e-28], [341.0706, 2.739067e-28], [342.11968, 2.74036e-28], [343.17199, 2.741673e-28], [344.22754, 2.743005e-28], [345.28634, 2.744357e-28], [346.34839, 2.745729e-28], [347.41371, 2.747121e-28], [348.4823, 2.748533e-28], [349.55418, 2.749965e-28], [350.62936, 2.751417e-28], [351.70785, 2.752889e-28], [352.78965, 2.754382e-28], [353.87478, 2.755895e-28], [354.96325, 2.757428e-28], [356.05507, 2.758982e-28], [357.15024, 2.760557e-28], [358.24879, 2.762152e-28], [359.35071, 2.763769e-28], [360.45602, 2.765406e-28], [361.56473, 2.767064e-28], [362.67685, 2.768744e-28], [363.7924, 2.770445e-28], [364.91137, 2.772167e-28], [366.03379, 2.773911e-28], [367.15965, 2.775676e-28], [368.28898, 2.777463e-28], [369.42179, 2.779272e-28], [370.55808, 2.781104e-28], [371.69786, 2.782957e-28], [372.84115, 2.784833e-28], [373.98796, 2.786732e-28], [375.13829, 2.788654e-28], [376.29216, 2.790599e-28], [377.44959, 2.792567e-28], [378.61057, 2.794559e-28], [379.77512, 2.796574e-28], [380.94325, 2.798615e-28], [382.11498, 2.80068e-28], [383.29031, 2.80277e-28], [384.46926, 2.804886e-28], [385.65183, 2.807029e-28], [386.83804, 2.809199e-28], [388.0279, 2.811397e-28], [389.22142, 2.813624e-28], [390.41861, 2.815881e-28], [391.61948, 2.81817e-28], [392.82405, 2.820493e-28], [394.03232, 2.822851e-28], [395.24431, 2.825248e-28], [396.46002, 2.827688e-28], [397.67948, 2.830173e-28], [398.90268, 2.832711e-28], [400.12965, 2.835308e-28], [401.36039, 2.837976e-28], [402.59492, 2.840726e-28], [403.83324, 2.843579e-28], [405.07538, 2.846562e-28], [406.32133, 2.849715e-28], [407.57112, 2.853098e-28], [408.82475, 2.856808e-28], [410.08224, 2.861003e-28], [411.34359, 2.865962e-28], [412.60883, 2.872214e-28], [413.87795, 2.880861e-28], [415.15098, 2.894533e-28], [416.42793, 2.920867e-28], [417.7088, 2.990483e-28], [418.99361, 3.340482e-28], [420.28238, 4.970461e-27], [421.57511, 3.279467e-28], [422.87181, 2.986934e-28], [424.17251, 2.930246e-28], [425.4772, 2.911874e-28], [426.78591, 2.904877e-28], [428.09864, 2.902411e-28], [429.41541, 2.902104e-28], [430.73623, 2.902958e-28], [432.06111, 2.904496e-28], [433.39007, 2.906467e-28], [434.72312, 2.908729e-28], [436.06027, 2.911196e-28], [437.40153, 2.913815e-28], [438.74691, 2.916549e-28], [440.09643, 2.919375e-28], [441.45011, 2.922278e-28], [442.80795, 2.925244e-28], [444.16996, 2.928266e-28], [445.53616, 2.931337e-28], [446.90657, 2.934452e-28], [448.28119, 2.937608e-28], [449.66004, 2.940803e-28], [451.04313, 2.944033e-28], [452.43047, 2.947297e-28], [453.82209, 2.950594e-28], [455.21798, 2.953923e-28], [456.61816, 2.957283e-28], [458.02266, 2.960673e-28], [459.43147, 2.964093e-28], [460.84462, 2.967542e-28], [462.26211, 2.971019e-28], [463.68396, 2.974525e-28], [465.11019, 2.978059e-28], [466.5408, 2.981622e-28], [467.97581, 2.985212e-28], [469.41524, 2.98883e-28], [470.8591, 2.992476e-28], [472.30739, 2.996149e-28], [473.76014, 2.99985e-28], [475.21736, 3.003578e-28], [476.67906, 3.007333e-28], [478.14526, 3.011117e-28], [479.61597, 3.014927e-28], [481.0912, 3.018765e-28], [482.57096, 3.02263e-28], [484.05528, 3.026523e-28], [485.54417, 3.030444e-28], [487.03763, 3.034391e-28], [488.53569, 3.038367e-28], [490.03836, 3.04237e-28], [491.54565, 3.0464e-28], [493.05757, 3.050459e-28], [494.57415, 3.054545e-28], [496.09539, 3.058658e-28], [497.62131, 3.0628e-28], [499.15192, 3.066969e-28], [500.68724, 3.071166e-28], [502.22728, 3.075392e-28], [503.77206, 3.079645e-28], [505.32159, 3.083926e-28], [506.87589, 3.088236e-28], [508.43497, 3.092574e-28], [509.99884, 3.09694e-28], [511.56753, 3.101334e-28], [513.14103, 3.105757e-28], [514.71938, 3.110208e-28], [516.30259, 3.114688e-28], [517.89066, 3.119196e-28], [519.48362, 3.123733e-28], [521.08148, 3.128299e-28], [522.68425, 3.132893e-28], [524.29195, 3.137517e-28], [525.9046, 3.142169e-28], [527.52221, 3.14685e-28], [529.14479, 3.151561e-28], [530.77236, 3.1563e-28], [532.40495, 3.161069e-28], [534.04255, 3.165867e-28], [535.68519, 3.170694e-28], [537.33288, 3.175551e-28], [538.98564, 3.180437e-28], [540.64348, 3.185353e-28], [542.30642, 3.190298e-28], [543.97448, 3.195273e-28], [545.64767, 3.200278e-28], [547.326, 3.205313e-28], [549.0095, 3.210377e-28], [550.69818, 3.215472e-28], [552.39205, 3.220596e-28], [554.09112, 3.225751e-28], [555.79543, 3.230936e-28], [557.50498, 3.236151e-28], [559.21978, 3.241397e-28], [560.93986, 3.246673e-28], [562.66524, 3.251979e-28], [564.39591, 3.257316e-28], [566.13192, 3.262683e-28], [567.87326, 3.268082e-28], [569.61995, 3.273511e-28], [571.37202, 3.278971e-28], [573.12948, 3.284461e-28], [574.89235, 3.289983e-28], [576.66064, 3.295536e-28], [578.43436, 3.30112e-28], [580.21354, 3.306735e-28], [581.9982, 3.312382e-28], [583.78834, 3.31806e-28], [585.58399, 3.323769e-28], [587.38517, 3.32951e-28], [589.19188, 3.335282e-28], [591.00415, 3.341086e-28], [592.82199, 3.346922e-28], [594.64543, 3.35279e-28], [596.47448, 3.358689e-28], [598.30915, 3.364621e-28], [600.14946, 3.370584e-28], [601.99543, 3.37658e-28], [603.84709, 3.382608e-28], [605.70443, 3.388668e-28], [607.5675, 3.39476e-28], [609.43629, 3.400885e-28], [611.31083, 3.407042e-28], [613.19113, 3.413232e-28], [615.07722, 3.419455e-28], [616.96911, 3.42571e-28], [618.86682, 3.431998e-28], [620.77037, 3.438319e-28], [622.67977, 3.444673e-28], [624.59504, 3.45106e-28], [626.51621, 3.45748e-28], [628.44328, 3.463933e-28], [630.37629, 3.470419e-28], [632.31523, 3.476939e-28], [634.26014, 3.483492e-28], [636.21104, 3.490079e-28], [638.16793, 3.496699e-28], [640.13085, 3.503353e-28], [642.0998, 3.510041e-28], [644.07481, 3.516762e-28], [646.05589, 3.523517e-28], [648.04307, 3.530306e-28], [650.03635, 3.53713e-28], [652.03577, 3.543987e-28], [654.04134, 3.550879e-28], [656.05308, 3.557805e-28], [658.07101, 3.564765e-28], [660.09514, 3.571759e-28], [662.1255, 3.578789e-28], [664.1621, 3.585852e-28], [666.20497, 3.592951e-28], [668.25412, 3.600084e-28], [670.30958, 3.607252e-28], [672.37136, 3.614455e-28], [674.43947, 3.621693e-28], [676.51395, 3.628966e-28], [678.59481, 3.636274e-28], [680.68208, 3.643617e-28], [682.77576, 3.650996e-28], [684.87588, 3.65841e-28], [686.98246, 3.66586e-28], [689.09552, 3.673345e-28], [691.21508, 3.680866e-28], [693.34116, 3.688422e-28], [695.47378, 3.696015e-28], [697.61296, 3.703643e-28], [699.75872, 3.711307e-28], [701.91107, 3.719007e-28], [704.07005, 3.726744e-28], [706.23567, 3.734516e-28], [708.40795, 3.742325e-28], [710.58691, 3.750171e-28], [712.77258, 3.758053e-28], [714.96497, 3.765971e-28], [717.1641, 3.773926e-28], [719.36999, 3.781918e-28], [721.58267, 3.789946e-28], [723.80216, 3.798012e-28], [726.02847, 3.806114e-28], [728.26163, 3.814254e-28], [730.50166, 3.82243e-28], [732.74858, 3.830644e-28], [735.00241, 3.838896e-28], [737.26317, 3.847184e-28], [739.53089, 3.855511e-28], [741.80558, 3.863874e-28], [744.08727, 3.872276e-28], [746.37598, 3.880715e-28], [748.67172, 3.889192e-28], [750.97453, 3.897707e-28], [753.28442, 3.90626e-28], [755.60142, 3.914852e-28], [757.92554, 3.923481e-28], [760.25681, 3.932149e-28], [762.59525, 3.940855e-28], [764.94089, 3.9496e-28], [767.29374, 3.958383e-28], [769.65382, 3.967205e-28], [772.02117, 3.976066e-28], [774.3958, 3.984965e-28], [776.77773, 3.993904e-28], [779.16699, 4.002882e-28], [781.56359, 4.011899e-28], [783.96757, 4.020955e-28], [786.37895, 4.030051e-28], [788.79774, 4.039186e-28], [791.22397, 4.048361e-28], [793.65766, 4.057575e-28], [796.09884, 4.06683e-28], [798.54752, 4.076125e-28], [801.00374, 4.08546e-28], [803.46752, 4.094835e-28], [805.93887, 4.104251e-28], [808.41782, 4.113709e-28], [810.9044, 4.123208e-28], [813.39863, 4.13275e-28], [815.90052, 4.142334e-28], [818.41012, 4.151964e-28], [820.92743, 4.16164e-28], [823.45249, 4.171368e-28], [825.98531, 4.181156e-28], [828.52592, 4.191021e-28], [831.07435, 4.201007e-28], [833.63062, 4.211257e-28], [836.19475, 4.222444e-28], [838.76676, 4.243432e-28], [841.34669, 4.290291e-28], [843.93456, 4.253396e-28], [846.53038, 4.261331e-28], [849.13419, 4.270914e-28], [851.746, 4.280859e-28], [854.36585, 4.290949e-28], [856.99376, 4.301125e-28], [859.62975, 4.311365e-28], [862.27385, 4.321659e-28], [864.92608, 4.332002e-28], [867.58647, 4.342391e-28], [870.25504, 4.352826e-28], [872.93182, 4.363306e-28], [875.61684, 4.373829e-28], [878.31011, 4.384396e-28], [881.01167, 4.395007e-28], [883.72154, 4.405661e-28], [886.43974, 4.416359e-28], [889.1663, 4.4271e-28], [891.90125, 4.437885e-28], [894.64461, 4.448713e-28], [897.39641, 4.459585e-28], [900.15667, 4.470501e-28], [902.92543, 4.48146e-28], [905.7027, 4.492464e-28], [908.48851, 4.503511e-28], [911.28289, 4.514602e-28], [914.08587, 4.525738e-28], [916.89747, 4.536917e-28], [919.71771, 4.548141e-28], [922.54664, 4.55941e-28], [925.38426, 4.570723e-28], [928.23061, 4.58208e-28], [931.08571, 4.593483e-28], [933.9496, 4.60493e-28], [936.8223, 4.616422e-28], [939.70383, 4.627959e-28], [942.59423, 4.639542e-28], [945.49351, 4.65117e-28], [948.40172, 4.662843e-28], [951.31886, 4.674561e-28], [954.24499, 4.686325e-28], [957.18011, 4.698135e-28], [960.12426, 4.709991e-28], [963.07746, 4.721893e-28], [966.03975, 4.733841e-28], [969.01115, 4.745835e-28], [971.99169, 4.757875e-28], [974.9814, 4.769961e-28], [977.98031, 4.782095e-28], [980.98844, 4.794274e-28], [984.00582, 4.806501e-28], [987.03248, 4.818774e-28], [990.06845, 4.831094e-28], [993.11376, 4.843462e-28], [996.16844, 4.855876e-28], [999.23251, 4.868338e-28], [1002.306, 4.880847e-28], [1005.389, 4.893404e-28], [1008.4814, 4.906008e-28], [1011.5833, 4.918661e-28], [1014.6948, 4.931361e-28], [1017.8159, 4.944109e-28], [1020.9465, 4.956905e-28], [1024.0868, 4.96975e-28], [1027.2368, 4.982643e-28], [1030.3964, 4.995584e-28], [1033.5658, 5.008574e-28], [1036.7449, 5.021613e-28], [1039.9337, 5.034701e-28], [1043.1324, 5.047837e-28], [1046.341, 5.061023e-28], [1049.5594, 5.074258e-28], [1052.7876, 5.087542e-28], [1056.0259, 5.100876e-28], [1059.2741, 5.114259e-28], [1062.5322, 5.127692e-28], [1065.8004, 5.141175e-28], [1069.0787, 5.154708e-28], [1072.367, 5.168291e-28], [1075.6655, 5.181925e-28], [1078.9741, 5.195608e-28], [1082.2928, 5.209342e-28], [1085.6218, 5.223127e-28], [1088.961, 5.236963e-28], [1092.3105, 5.250849e-28], [1095.6703, 5.264786e-28], [1099.0404, 5.278775e-28], [1102.4209, 5.292814e-28], [1105.8118, 5.306906e-28], [1109.2131, 5.321048e-28], [1112.6249, 5.335242e-28], [1116.0472, 5.349489e-28], [1119.48, 5.363786e-28], [1122.9234, 5.378137e-28], [1126.3773, 5.392539e-28], [1129.8419, 5.406993e-28], [1133.3171, 5.4215e-28], [1136.803, 5.43606e-28], [1140.2997, 5.450672e-28], [1143.8071, 5.465338e-28], [1147.3253, 5.480056e-28], [1150.8543, 5.494827e-28], [1154.3942, 5.509652e-28], [1157.9449, 5.52453e-28], [1161.5066, 5.539461e-28], [1165.0792, 5.554446e-28], [1168.6628, 5.569485e-28], [1172.2575, 5.584578e-28], [1175.8632, 5.599726e-28], [1179.48, 5.614927e-28], [1183.1079, 5.630183e-28], [1186.7469, 5.645493e-28], [1190.3972, 5.660858e-28], [1194.0587, 5.676278e-28], [1197.7314, 5.691753e-28], [1201.4155, 5.707283e-28], [1205.1109, 5.722868e-28], [1208.8176, 5.738509e-28], [1212.5358, 5.754205e-28], [1216.2654, 5.769957e-28], [1220.0064, 5.785766e-28], [1223.759, 5.80163e-28], [1227.5231, 5.81755e-28], [1231.2988, 5.833528e-28], [1235.0861, 5.849562e-28], [1238.885, 5.865655e-28], [1242.6957, 5.881807e-28], [1246.518, 5.898021e-28], [1250.3521, 5.914307e-28], [1254.1981, 5.930712e-28], [1258.0558, 5.947811e-28], [1261.9254, 5.967517e-28], [1265.8069, 5.98009e-28], [1269.7003, 5.996479e-28], [1273.6058, 6.013046e-28], [1277.5232, 6.029694e-28], [1281.4527, 6.046407e-28], [1285.3942, 6.063183e-28], [1289.3479, 6.080018e-28], [1293.3138, 6.096913e-28], [1297.2918, 6.113868e-28], [1301.2821, 6.130881e-28], [1305.2847, 6.147955e-28], [1309.2995, 6.165087e-28], [1313.3268, 6.182279e-28], [1317.3664, 6.199531e-28], [1321.4184, 6.216843e-28], [1325.4829, 6.234215e-28], [1329.5599, 6.251647e-28], [1333.6494, 6.269139e-28], [1337.7515, 6.286692e-28], [1341.8663, 6.304306e-28], [1345.9937, 6.32198e-28], [1350.1337, 6.339715e-28], [1354.2866, 6.357512e-28], [1358.4522, 6.375369e-28], [1362.6306, 6.393288e-28], [1366.8218, 6.411269e-28], [1371.026, 6.429312e-28], [1375.2431, 6.447416e-28], [1379.4731, 6.465583e-28], [1383.7162, 6.483812e-28], [1387.9723, 6.502103e-28], [1392.2415, 6.520457e-28], [1396.5238, 6.538874e-28], [1400.8193, 6.557354e-28], [1405.1281, 6.575897e-28], [1409.45, 6.594503e-28], [1413.7853, 6.613173e-28], [1418.1339, 6.631906e-28], [1422.4959, 6.650703e-28], [1426.8713, 6.669565e-28], [1431.2601, 6.68849e-28], [1435.6625, 6.70748e-28], [1440.0784, 6.726534e-28], [1444.5079, 6.745653e-28], [1448.951, 6.764837e-28], [1453.4077, 6.784087e-28], [1457.8782, 6.803401e-28], [1462.3624, 6.822781e-28], [1466.8604, 6.842227e-28], [1471.3723, 6.86174e-28], [1475.898, 6.881319e-28], [1480.4377, 6.900964e-28], [1484.9913, 6.920675e-28], [1489.5589, 6.940453e-28], [1494.1406, 6.960298e-28], [1498.7364, 6.98021e-28], [1503.3463, 7.000188e-28], [1507.9703, 7.020234e-28], [1512.6086, 7.040347e-28], [1517.2612, 7.060528e-28], [1521.9281, 7.080777e-28], [1526.6093, 7.101094e-28], [1531.305, 7.121479e-28], [1536.015, 7.141933e-28], [1540.7396, 7.162455e-28], [1545.4787, 7.183046e-28], [1550.2324, 7.203705e-28], [1555.0007, 7.224436e-28], [1559.7836, 7.245252e-28], [1564.5813, 7.266138e-28], [1569.3937, 7.287094e-28], [1574.221, 7.30812e-28], [1579.063, 7.329216e-28], [1583.92, 7.350382e-28], [1588.7919, 7.371618e-28], [1593.6788, 7.392925e-28], [1598.5807, 7.414303e-28], [1603.4978, 7.435753e-28], [1608.4299, 7.457273e-28], [1613.3772, 7.478865e-28], [1618.3397, 7.500529e-28], [1623.3175, 7.522264e-28], [1628.3106, 7.544072e-28], [1633.319, 7.565952e-28], [1638.3429, 7.587905e-28], [1643.3822, 7.60993e-28], [1648.437, 7.632029e-28], [1653.5074, 7.6542e-28], [1658.5933, 7.676446e-28], [1663.6949, 7.698766e-28], [1668.8122, 7.721164e-28], [1673.9452, 7.743653e-28], [1679.0941, 7.76664e-28], [1684.2587, 7.788923e-28], [1689.4393, 7.811484e-28], [1694.6357, 7.834233e-28], [1699.8482, 7.857068e-28], [1705.0767, 7.87998e-28], [1710.3213, 7.902969e-28], [1715.582, 7.926033e-28], [1720.8588, 7.949174e-28], [1726.152, 7.972391e-28], [1731.4614, 7.995684e-28], [1736.7871, 8.019053e-28], [1742.1292, 8.0425e-28], [1747.4877, 8.066023e-28], [1752.8628, 8.089624e-28], [1758.2543, 8.113302e-28], [1763.6625, 8.137058e-28], [1769.0872, 8.160892e-28], [1774.5287, 8.184803e-28], [1779.9869, 8.208793e-28], [1785.4619, 8.232862e-28], [1790.9537, 8.257009e-28], [1796.4624, 8.281235e-28], [1801.9881, 8.305541e-28], [1807.5308, 8.329926e-28], [1813.0905, 8.35439e-28], [1818.6673, 8.378935e-28], [1824.2612, 8.403559e-28], [1829.8724, 8.428264e-28], [1835.5008, 8.45305e-28], [1841.1466, 8.477916e-28], [1846.8097, 8.502864e-28], [1852.4902, 8.527892e-28], [1858.1882, 8.553003e-28], [1863.9037, 8.578195e-28], [1869.6368, 8.603468e-28], [1875.3876, 8.628825e-28], [1881.156, 8.654263e-28], [1886.9422, 8.679785e-28], [1892.7461, 8.70539e-28], [1898.5679, 8.731077e-28], [1904.4077, 8.756848e-28], [1910.2653, 8.782702e-28], [1916.1411, 8.808641e-28], [1922.0348, 8.834664e-28], [1927.9467, 8.860771e-28], [1933.8768, 8.886963e-28], [1939.8252, 8.91324e-28], [1945.7918, 8.939665e-28], [1951.7768, 8.966203e-28], [1957.7802, 8.992826e-28], [1963.802, 9.019536e-28], [1969.8424, 9.046331e-28], [1975.9013, 9.073212e-28], [1981.9789, 9.10018e-28], [1988.0752, 9.127234e-28], [1994.1903, 9.154376e-28], [2000.3241, 9.181604e-28], [2006.4768, 9.208919e-28], [2012.6484, 9.236323e-28], [2018.8391, 9.263815e-28], [2025.0487, 9.291394e-28], [2031.2775, 9.319063e-28], [2037.5254, 9.346819e-28], [2043.7926, 9.374665e-28], [2050.079, 9.4026e-28], [2056.3847, 9.430624e-28], [2062.7099, 9.458739e-28], [2069.0545, 9.486943e-28], [2075.4186, 9.515238e-28], [2081.8023, 9.543624e-28], [2088.2056, 9.572102e-28], [2094.6286, 9.60068e-28], [2101.0714, 9.672075e-28], [2107.534, 9.658089e-28], [2114.0165, 9.68692e-28], [2120.5189, 9.715852e-28], [2127.0413, 9.744878e-28], [2133.5838, 9.773997e-28], [2140.1464, 9.803209e-28], [2146.7292, 9.832515e-28], [2153.3322, 9.861914e-28], [2159.9556, 9.891407e-28], [2166.5993, 9.920995e-28], [2173.2634, 9.950676e-28], [2179.9481, 9.980452e-28], [2186.6533, 1.001032e-27], [2193.3791, 1.004029e-27], [2200.1256, 1.007035e-27], [2206.8929, 1.010051e-27], [2213.681, 1.013076e-27], [2220.49, 1.016111e-27], [2227.3199, 1.019156e-27], [2234.1708, 1.02221e-27], [2241.0428, 1.025275e-27], [2247.9359, 1.028349e-27], [2254.8503, 1.031432e-27], [2261.7859, 1.034526e-27], [2268.7428, 1.037629e-27], [2275.7211, 1.040742e-27], [2282.7209, 1.043865e-27], [2289.7422, 1.046998e-27], [2296.7852, 1.050141e-27], [2303.8497, 1.053294e-27], [2310.9361, 1.056457e-27], [2318.0442, 1.05963e-27], [2325.1741, 1.062813e-27], [2332.3261, 1.066006e-27], [2339.5, 1.06921e-27], [2346.6959, 1.072423e-27], [2353.914, 1.075647e-27], [2361.1543, 1.078881e-27], [2368.4169, 1.082125e-27], [2375.7018, 1.08538e-27], [2383.0091, 1.088645e-27], [2390.3389, 1.09192e-27], [2397.6913, 1.095205e-27], [2405.0662, 1.098501e-27], [2412.4639, 1.101808e-27], [2419.8843, 1.105125e-27], [2427.3275, 1.108452e-27], [2434.7936, 1.11179e-27], [2442.2827, 1.115139e-27], [2449.7948, 1.118498e-27], [2457.33, 1.121868e-27], [2464.8884, 1.125248e-27], [2472.4701, 1.128639e-27], [2480.075, 1.132041e-27], [2487.7034, 1.135454e-27], [2495.3552, 1.138877e-27], [2503.0306, 1.142312e-27], [2510.7295, 1.145757e-27], [2518.4522, 1.149216e-27], [2526.1986, 1.152681e-27], [2533.9688, 1.156158e-27], [2541.763, 1.159647e-27], [2549.5811, 1.163147e-27], [2557.4232, 1.166658e-27], [2565.2895, 1.17018e-27], [2573.1799, 1.173713e-27], [2581.0947, 1.177258e-27], [2589.0338, 1.180814e-27], [2596.9973, 1.18438e-27], [2604.9852, 1.187959e-27], [2612.9978, 1.191548e-27], [2621.035, 1.195149e-27], [2629.097, 1.198761e-27], [2637.1837, 1.202385e-27], [2645.2953, 1.20602e-27], [2653.4318, 1.209666e-27], [2661.5934, 1.213324e-27], [2669.7801, 1.216994e-27], [2677.992, 1.220675e-27], [2686.2291, 1.224367e-27], [2694.4915, 1.228072e-27], [2702.7794, 1.231788e-27], [2711.0928, 1.235515e-27], [2719.4317, 1.239255e-27], [2727.7963, 1.243006e-27], [2736.1866, 1.246769e-27], [2744.6027, 1.250544e-27], [2753.0448, 1.254331e-27], [2761.5127, 1.25813e-27], [2770.0067, 1.261941e-27], [2778.5269, 1.265764e-27], [2787.0732, 1.269599e-27], [2795.6459, 1.273446e-27], [2804.2449, 1.277305e-27], [2812.8704, 1.281176e-27], [2821.5223, 1.285059e-27], [2830.201, 1.288955e-27], [2838.9062, 1.292863e-27], [2847.6383, 1.296783e-27], [2856.3972, 1.300715e-27], [2865.1831, 1.30466e-27], [2873.996, 1.308618e-27], [2882.836, 1.312587e-27], [2891.7032, 1.31657e-27], [2900.5977, 1.320565e-27], [2909.5195, 1.324572e-27], [2918.4688, 1.328592e-27], [2927.4456, 1.332625e-27], [2936.45, 1.33667e-27], [2945.4821, 1.340729e-27], [2954.542, 1.344799e-27], [2963.6297, 1.348882e-27], [2972.7454, 1.352979e-27], [2981.8892, 1.357088e-27], [2991.061, 1.36121e-27], [3000.2611, 1.365346e-27], [3009.4895, 1.369494e-27], [3018.7463, 1.373655e-27], [3028.0315, 1.377829e-27], [3037.3453, 1.382017e-27], [3046.6877, 1.386217e-27], [3056.0589, 1.390431e-27], [3065.4589, 1.394658e-27], [3074.8878, 1.398899e-27], [3084.3457, 1.403152e-27], [3093.8327, 1.407419e-27], [3103.3489, 1.4117e-27], [3112.8944, 1.415993e-27], [3122.4692, 1.420301e-27], [3132.0735, 1.424622e-27], [3141.7073, 1.428956e-27], [3151.3708, 1.433304e-27], [3161.0639, 1.437666e-27], [3170.7869, 1.442041e-27], [3180.5398, 1.44643e-27], [3190.3227, 1.450833e-27], [3200.1357, 1.45525e-27], [3209.9788, 1.459681e-27], [3219.8523, 1.464125e-27], [3229.7561, 1.468584e-27], [3239.6903, 1.473056e-27], [3249.6552, 1.477543e-27], [3259.6506, 1.482043e-27], [3269.6769, 1.486558e-27], [3279.7339, 1.491087e-27], [3289.8219, 1.49563e-27], [3299.941, 1.500187e-27], [3310.0911, 1.504759e-27], [3320.2725, 1.509345e-27], [3330.4852, 1.513945e-27], [3340.7293, 1.51856e-27], [3351.0049, 1.52319e-27], [3361.3121, 1.527855e-27], [3371.651, 1.532492e-27], [3382.0217, 1.537165e-27], [3392.4243, 1.541852e-27], [3402.859, 1.546555e-27], [3413.3257, 1.551272e-27], [3423.8246, 1.556004e-27], [3434.3558, 1.56075e-27], [3444.9194, 1.565512e-27], [3455.5154, 1.570289e-27], [3466.1441, 1.57508e-27], [3476.8055, 1.579887e-27], [3487.4996, 1.584708e-27], [3498.2267, 1.589545e-27], [3508.9867, 1.594397e-27], [3519.7799, 1.599264e-27], [3530.6062, 1.604146e-27], [3541.4659, 1.609044e-27], [3552.3589, 1.613957e-27], [3563.2855, 1.618886e-27], [3574.2456, 1.62383e-27], [3585.2395, 1.628789e-27], [3596.2672, 1.633764e-27], [3607.3288, 1.638754e-27], [3618.4244, 1.643761e-27], [3629.5542, 1.648783e-27], [3640.7182, 1.65382e-27], [3651.9165, 1.658874e-27], [3663.1493, 1.663943e-27], [3674.4166, 1.669028e-27], [3685.7186, 1.674129e-27], [3697.0553, 1.679246e-27], [3708.427, 1.684379e-27], [3719.8336, 1.689528e-27], [3731.2752, 1.694694e-27], [3742.7521, 1.699875e-27], [3754.2643, 1.705073e-27], [3765.8119, 1.710287e-27], [3777.395, 1.715518e-27], [3789.0137, 1.720765e-27], [3800.6682, 1.726028e-27], [3812.3585, 1.731307e-27], [3824.0848, 1.736604e-27], [3835.8471, 1.741917e-27], [3847.6456, 1.747246e-27], [3859.4804, 1.752593e-27], [3871.3516, 1.757955e-27], [3883.2594, 1.763335e-27], [3895.2037, 1.768732e-27], [3907.1848, 1.774146e-27], [3919.2028, 1.779576e-27], [3931.2577, 1.785024e-27], [3943.3497, 1.790488e-27], [3955.4788, 1.79597e-27], [3967.6453, 1.801469e-27], [3979.8493, 1.806985e-27], [3992.0907, 1.812519e-27], [4004.3698, 1.81807e-27], [4016.6867, 1.823638e-27], [4029.0414, 1.829224e-27], [4041.4342, 1.834827e-27], [4053.8651, 1.840448e-27], [4066.3342, 1.846086e-27], [4078.8417, 1.851742e-27], [4091.3876, 1.857416e-27], [4103.9721, 1.863107e-27], [4116.5954, 1.868817e-27], [4129.2574, 1.874544e-27], [4141.9584, 1.880289e-27], [4154.6985, 1.886053e-27], [4167.4778, 1.891834e-27], [4180.2963, 1.897634e-27], [4193.1543, 1.903451e-27], [4206.0519, 1.909287e-27], [4218.9891, 1.915141e-27], [4231.9661, 1.921014e-27], [4244.983, 1.926905e-27], [4258.04, 1.932814e-27], [4271.1371, 1.938742e-27], [4284.2745, 1.944689e-27], [4297.4524, 1.950654e-27], [4310.6707, 1.956638e-27], [4323.9297, 1.962641e-27], [4337.2295, 1.968662e-27], [4350.5702, 1.974703e-27], [4363.9519, 1.980762e-27], [4377.3748, 1.98684e-27], [4390.839, 1.992938e-27], [4404.3446, 1.999054e-27], [4417.8918, 2.00519e-27], [4431.4806, 2.011345e-27], [4445.1112, 2.017519e-27], [4458.7837, 2.023713e-27], [4472.4983, 2.029926e-27], [4486.2551, 2.036159e-27], [4500.0541, 2.042411e-27], [4513.8957, 2.048683e-27], [4527.7798, 2.054974e-27], [4541.7066, 2.061286e-27], [4555.6762, 2.067617e-27], [4569.6888, 2.073968e-27], [4583.7446, 2.080338e-27], [4597.8435, 2.086729e-27], [4611.9858, 2.09314e-27], [4626.1716, 2.099571e-27], [4640.4011, 2.106022e-27], [4654.6743, 2.112494e-27], [4668.9914, 2.118986e-27], [4683.3526, 2.125498e-27], [4697.7579, 2.13203e-27], [4712.2075, 2.138584e-27], [4726.7016, 2.145157e-27], [4741.2403, 2.151752e-27], [4755.8236, 2.158367e-27], [4770.4519, 2.165003e-27], [4785.1251, 2.171659e-27], [4799.8435, 2.178337e-27], [4814.6071, 2.185036e-27], [4829.4162, 2.191755e-27], [4844.2708, 2.198496e-27], [4859.1711, 2.205258e-27], [4874.1172, 2.212041e-27], [4889.1093, 2.218846e-27], [4904.1475, 2.225672e-27], [4919.2319, 2.232519e-27], [4934.3628, 2.239388e-27], [4949.5402, 2.246279e-27], [4964.7643, 2.253191e-27], [4980.0352, 2.260125e-27], [4995.3531, 2.267081e-27], [5010.7181, 2.274058e-27], [5026.1303, 2.281058e-27], [5041.59, 2.28808e-27], [5057.0972, 2.295123e-27], [5072.6521, 2.302189e-27], [5088.2549, 2.309277e-27], [5103.9056, 2.316388e-27], [5119.6045, 2.323521e-27], [5135.3517, 2.330676e-27], [5151.1473, 2.337854e-27], [5166.9915, 2.345054e-27], [5182.8844, 2.352277e-27], [5198.8262, 2.359523e-27], [5214.8171, 2.366791e-27], [5230.8571, 2.374083e-27], [5246.9465, 2.381397e-27], [5263.0854, 2.388735e-27], [5279.2739, 2.396096e-27], [5295.5122, 2.40348e-27], [5311.8004, 2.410887e-27], [5328.1387, 2.418317e-27], [5344.5273, 2.425771e-27], [5360.9663, 2.433248e-27], [5377.4559, 2.440749e-27], [5393.9962, 2.448274e-27], [5410.5874, 2.455823e-27], [5427.2296, 2.463395e-27], [5443.923, 2.470991e-27], [5460.6677, 2.478611e-27], [5477.4639, 2.486255e-27], [5494.3118, 2.493923e-27], [5511.2116, 2.501616e-27], [5528.1633, 2.509332e-27], [5545.1671, 2.517073e-27], [5562.2233, 2.524839e-27], [5579.3319, 2.532629e-27], [5596.4931, 2.540444e-27], [5613.7071, 2.548283e-27], [5630.9741, 2.556147e-27], [5648.2942, 2.564036e-27], [5665.6675, 2.57195e-27], [5683.0943, 2.579889e-27], [5700.5747, 2.587853e-27], [5718.1088, 2.595842e-27], [5735.6969, 2.603857e-27], [5753.3391, 2.611896e-27], [5771.0356, 2.619962e-27], [5788.7865, 2.628053e-27], [5806.5919, 2.636169e-27], [5824.4522, 2.644311e-27], [5842.3674, 2.652479e-27], [5860.3377, 2.660673e-27], [5878.3632, 2.668893e-27], [5896.4443, 2.677139e-27], [5914.5809, 2.685411e-27], [5932.7733, 2.693709e-27], [5951.0217, 2.702033e-27], [5969.3262, 2.710384e-27], [5987.687, 2.718762e-27], [6006.1042, 2.727166e-27], [6024.5781, 2.735596e-27], [6043.1089, 2.744054e-27], [6061.6966, 2.752538e-27], [6080.3416, 2.761049e-27], [6099.0438, 2.769587e-27], [6117.8036, 2.778152e-27], [6136.6211, 2.786745e-27], [6155.4965, 2.795364e-27], [6174.4299, 2.804011e-27], [6193.4216, 2.812686e-27], [6212.4717, 2.821388e-27], [6231.5804, 2.830118e-27], [6250.7478, 2.838876e-27], [6269.9742, 2.847661e-27], [6289.2598, 2.856475e-27], [6308.6046, 2.865316e-27], [6328.009, 2.874186e-27], [6347.4731, 2.883083e-27], [6366.997, 2.89201e-27], [6386.581, 2.900964e-27], [6406.2252, 2.909947e-27], [6425.9298, 2.918959e-27], [6445.6951, 2.927999e-27], [6465.5211, 2.937069e-27], [6485.4081, 2.946167e-27], [6505.3563, 2.955294e-27], [6525.3659, 2.964451e-27], [6545.437, 2.973636e-27], [6565.5698, 2.982851e-27], [6585.7646, 2.992096e-27], [6606.0215, 3.001369e-27], [6626.3406, 3.010673e-27], [6646.7223, 3.020006e-27], [6667.1667, 3.029369e-27], [6687.674, 3.038763e-27], [6708.2443, 3.048186e-27], [6728.8779, 3.057639e-27], [6749.575, 3.067123e-27], [6770.3357, 3.076637e-27], [6791.1603, 3.086181e-27], [6812.0489, 3.095756e-27], [6833.0018, 3.105362e-27], [6854.0192, 3.114998e-27], [6875.1012, 3.124665e-27], [6896.248, 3.134364e-27], [6917.4599, 3.144093e-27], [6938.737, 3.153854e-27], [6960.0796, 3.163646e-27], [6981.4878, 3.17347e-27], [7002.9619, 3.183325e-27], [7024.502, 3.193212e-27], [7046.1083, 3.203131e-27], [7067.7812, 3.213082e-27], [7089.5207, 3.223064e-27], [7111.327, 3.233079e-27], [7133.2004, 3.243126e-27], [7155.1412, 3.253206e-27], [7177.1494, 3.263318e-27], [7199.2252, 3.273462e-27], [7221.369, 3.28364e-27], [7243.5809, 3.29385e-27], [7265.8612, 3.304093e-27], [7288.2099, 3.31437e-27], [7310.6274, 3.324679e-27], [7333.1139, 3.335022e-27], [7355.6695, 3.345398e-27], [7378.2945, 3.355808e-27], [7400.9891, 3.366252e-27], [7423.7535, 3.37673e-27], [7446.5879, 3.387241e-27], [7469.4925, 3.397787e-27], [7492.4676, 3.408367e-27], [7515.5134, 3.418981e-27], [7538.63, 3.42963e-27], [7561.8178, 3.440313e-27], [7585.0769, 3.451031e-27], [7608.4075, 3.461784e-27], [7631.8099, 3.472572e-27], [7655.2842, 3.483395e-27], [7678.8308, 3.494254e-27], [7702.4498, 3.505147e-27], [7726.1414, 3.516077e-27], [7749.906, 3.527042e-27], [7773.7436, 3.538043e-27], [7797.6545, 3.54908e-27], [7821.639, 3.560153e-27], [7845.6972, 3.571262e-27], [7869.8295, 3.582408e-27], [7894.0359, 3.59359e-27], [7918.3169, 3.604809e-27], [7942.6725, 3.616064e-27], [7967.103, 3.627357e-27], [7991.6087, 3.638687e-27], [8016.1898, 3.650054e-27], [8040.8464, 3.661458e-27], [8065.5789, 3.6729e-27], [8090.3875, 3.68438e-27], [8115.2724, 3.695898e-27], [8140.2338, 3.707454e-27], [8165.272, 3.719047e-27], [8190.3872, 3.73068e-27], [8215.5797, 3.742351e-27], [8240.8496, 3.75406e-27], [8266.1973, 3.765808e-27], [8291.623, 3.777595e-27], [8317.1268, 3.789422e-27], [8342.7091, 3.801288e-27], [8368.3701, 3.813193e-27], [8394.11, 3.825138e-27], [8419.9291, 3.837122e-27], [8445.8276, 3.849147e-27], [8471.8058, 3.861212e-27], [8497.8639, 3.873317e-27], [8524.0021, 3.885463e-27], [8550.2207, 3.897649e-27], [8576.52, 3.909876e-27], [8602.9001, 3.922145e-27], [8629.3614, 3.934454e-27], [8655.9041, 3.946805e-27], [8682.5285, 3.959198e-27], [8709.2347, 3.971633e-27], [8736.0231, 3.984109e-27], [8762.8938, 3.996628e-27], [8789.8472, 4.009189e-27], [8816.8836, 4.021792e-27], [8844.003, 4.034439e-27], [8871.2059, 4.047128e-27], [8898.4925, 4.059861e-27], [8925.863, 4.072637e-27], [8953.3177, 4.085456e-27], [8980.8568, 4.09832e-27], [9008.4807, 4.111227e-27], [9036.1895, 4.124179e-27], [9063.9835, 4.137175e-27], [9091.863, 4.150216e-27], [9119.8283, 4.163301e-27], [9147.8796, 4.176432e-27], [9176.0172, 4.189609e-27], [9204.2413, 4.20283e-27], [9232.5522, 4.216098e-27], [9260.9503, 4.229412e-27], [9289.4356, 4.242772e-27], [9318.0086, 4.256179e-27], [9346.6695, 4.269632e-27], [9375.4185, 4.283133e-27], [9404.256, 4.296681e-27], [9433.1821, 4.310277e-27], [9462.1973, 4.32392e-27], [9491.3016, 4.337612e-27], [9520.4955, 4.351352e-27], [9549.7792, 4.365141e-27], [9579.153, 4.378979e-27], [9608.6171, 4.392866e-27], [9638.1718, 4.406802e-27], [9667.8175, 4.420789e-27], [9697.5543, 4.434826e-27], [9727.3826, 4.448913e-27], [9757.3027, 4.463051e-27], [9787.3147, 4.47724e-27], [9817.4191, 4.491481e-27], [9847.6161, 4.505773e-27], [9877.906, 4.520118e-27], [9908.289, 4.534515e-27], [9938.7655, 4.548965e-27], [9969.3357, 4.563468e-27], [10000.0, 4.578024e-27]])
    
    fiveyr = 5 * 365.25 * 86400
    snr = 10
    f1 = data[:,0]
    S1 = fiveyr*(data[:,1]/10)**2
    fstop = [f1[0],f1[-1]]

    # % ---- Interpolate to desired frequencies.  Set noise to infinity outside
    # %      band of frequencies covered by f1.
    # S = inf(size(f)); 
    # k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop
